// "$Header:$"
//----------------------------  function.cc  ---------------------------
//   $Id: function.cc,v 1.28 2003/02/06 17:37:39 wolf Exp $
//   Version: $Name: Version-4-0-0 $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//----------------------------  function.cc  ---------------------------
//----------------------------  tensor_function.cc  ------------------------
//    $Id: tensor_function.cc,v 1.20 2003/01/14 19:51:07 wolf Exp $
//    Version: $Name: Version-4-0-0 $
//
//    Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal authors
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//----------------------------  tensor_function.cc  ------------------------
//
//modified from deal.II/base/source/function.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <FunctionSystem.H>
#include <DDecl.H>

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void FunctionSystem::value(NumPtr<REAL> &return_values,
const NumPtr<Point<SPACEDIM> > &points,int component) const {
  if (return_values.getNumber()!=points.getNumber()) {
    return_values.cleanup();
    return_values.allocate(points.getNumber());
  }
  for (int i=0;i<points.getNumber();i++) {
    return_values[i]=value(points[i],component);
  }
}

void FunctionSystem::value(NumPtr<Vector> &return_values,
const NumPtr<Point<SPACEDIM> > &points) const {
  if (return_values.getNumber()!=points.getNumber()) {
    return_values.cleanup();
    return_values.allocate(points.getNumber());
    for (int i=0;i<return_values.getNumber();i++) {
      return_values[i].reinit(n_components,true);
    }
  }
  for (int i=0;i<points.getNumber();i++) {
#ifdef INDEF
    return_values[i]=HUGE_VAL;
#endif
    value(points[i],return_values[i]);
  }
}

void FunctionSystem::gradient(
NumPtr<Tensor<1,SPACEDIM> > &return_gradients,
const NumPtr<Point<SPACEDIM> > &points,int component) const {
  if (return_gradients.getNumber()!=points.getNumber()) {
    return_gradients.cleanup();
    return_gradients.allocate(points.getNumber());
  }
  for (int i=0;i<points.getNumber();i++) {
    return_gradients[i]=gradient(points[i],component);
  }
}

void FunctionSystem::gradient(
NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &return_gradients,
const NumPtr<Point<SPACEDIM> > &points) const {
  if (return_gradients.getNumber()!=points.getNumber()) {
    return_gradients.cleanup();
    return_gradients.allocate(points.getNumber());
  }
  for (int i=0;i<points.getNumber();i++) {
    if (return_gradients[i].getNumber()!=n_components) {
      return_gradients[i].cleanup();
      return_gradients[i].allocate(n_components);
    }
    for (int component=0;component<n_components;component++) {
      return_gradients[i][component]=gradient(points[i],component);
    }
  };
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Tensor<1,SPACEDIM> ConstantFunctionSystem::zero_tensor;
void ConstantFunctionSystem::gradient(
NumPtr<Tensor<1,SPACEDIM> > &return_gradients,
const Point<SPACEDIM>&) const {
  if (return_gradients.getNumber()!=n_components) {
    return_gradients.cleanup();
    return_gradients.allocate(n_components);
  }
  for (int c=0;c<n_components;c++) return_gradients[c].initialize(0.);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ConstantComponentFunctionSystem::value(
NumPtr<Vector> &return_values,const NumPtr<Point<SPACEDIM> > &points) 
const{
  if (return_values.getNumber()!=points.getNumber()) {
    return_values.cleanup();
    return_values.allocate(points.getNumber());
  }
  for (int i=0;i<points.getNumber();i++) {
    if (return_values[i].size()!=n_components) {
      return_values[i].reinit(n_components,true);
    }
    return_values[i]=0.;
    return_values[i][selected]=function_value;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void TimeDependentFunctionSystem::value(NumPtr<REAL> &return_values,
const NumPtr<Point<SPACEDIM> > &points,REAL time,int component) const {
  if (return_values.getNumber()!=points.getNumber()) {
    return_values.cleanup();
    return_values.allocate(points.getNumber());
  }
  for (int i=0;i<points.getNumber();i++) {
    return_values[i]=value(points[i],time,component);
  }
}

void TimeDependentFunctionSystem::value(NumPtr<Vector> &return_values,
const NumPtr<Point<SPACEDIM> > &points,REAL time) const {
  if (return_values.getNumber()!=points.getNumber()) {
    return_values.cleanup();
    return_values.allocate(points.getNumber());
    for (int i=0;i<return_values.getNumber();i++) {
      return_values[i].reinit(n_components,true);
    }
  }
  for (int i=0;i<points.getNumber();i++) {
#ifdef INDEF
    return_values[i]=HUGE_VAL;
#endif
    value(points[i],time,return_values[i]);
  }
}

void TimeDependentFunctionSystem::gradient(
NumPtr<Tensor<1,SPACEDIM> > &return_gradients,
const NumPtr<Point<SPACEDIM> > &points,REAL time,int component) const {
  if (return_gradients.getNumber()!=points.getNumber()) {
    return_gradients.cleanup();
    return_gradients.allocate(points.getNumber());
  }
  for (int i=0;i<points.getNumber();i++) {
    return_gradients[i]=gradient(points[i],time,component);
  }
}

void TimeDependentFunctionSystem::gradient(
NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &return_gradients,
const NumPtr<Point<SPACEDIM> > &points,REAL time) const {
  if (return_gradients.getNumber()!=points.getNumber()) {
    return_gradients.cleanup();
    return_gradients.allocate(points.getNumber());
  }
  for (int i=0;i<points.getNumber();i++) {
    if (return_gradients[i].getNumber()!=n_components) {
      return_gradients[i].cleanup();
      return_gradients[i].allocate(n_components);
    }
    for (int component=0;component<n_components;component++) {
      return_gradients[i][component]=gradient(points[i],time,component);
    }
  };
}

REAL huge[SPACEDIM]={D_DECL(HUGE_VAL,HUGE_VAL,HUGE_VAL)};
Tensor<1,SPACEDIM> TensorFunctionSystem::bogus(huge);
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void TensorFunctionSystem::value(
NumPtr<Tensor<1,SPACEDIM> > &return_values,
const NumPtr<Point<SPACEDIM> > &points,int component) const {
  if (return_values.getNumber()!=points.getNumber()) {
    return_values.cleanup();
    return_values.allocate(points.getNumber());
  }
  for (int i=0;i<points.getNumber();i++) {
    return_values[i]=value(points[i],component);
  }
}

void TensorFunctionSystem::value(
NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &return_values,
const NumPtr<Point<SPACEDIM> > &points) const {
  if (return_values.getNumber()!=points.getNumber()) {
    return_values.cleanup();
    return_values.allocate(points.getNumber());
    for (int i=0;i<return_values.getNumber();i++) {
      return_values[i].allocate(n_components);
    }
  }
  for (int i=0;i<points.getNumber();i++) {
#ifdef INDEF
    return_values[i].initialize(bogus);
#endif
    value(points[i],return_values[i]);
  }
}

void TensorFunctionSystem::gradient(
NumPtr<Tensor<2,SPACEDIM> > &return_gradients,
const NumPtr<Point<SPACEDIM> > &points,int component) const {
  if (return_gradients.getNumber()!=points.getNumber()) {
    return_gradients.cleanup();
    return_gradients.allocate(points.getNumber());
  }
  for (int i=0;i<points.getNumber();i++) {
    return_gradients[i]=gradient(points[i],component);
  }
}

void TensorFunctionSystem::gradient(
NumPtr<NumPtr<Tensor<2,SPACEDIM> > > &return_gradients,
const NumPtr<Point<SPACEDIM> > &points) const {
  if (return_gradients.getNumber()!=points.getNumber()) {
    return_gradients.cleanup();
    return_gradients.allocate(points.getNumber());
  }
  for (int i=0;i<points.getNumber();i++) {
    if (return_gradients[i].getNumber()!=n_components) {
      return_gradients[i].cleanup();
      return_gradients[i].allocate(n_components);
    }
    for (int component=0;component<n_components;component++) {
      return_gradients[i][component]=gradient(points[i],component);
    }
  };
}

Tensor<1,SPACEDIM> ConstantTensorFunctionSystem::zero_value;
Tensor<2,SPACEDIM> ConstantTensorFunctionSystem::zero_gradient;
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ConstantTensorFunctionSystem::gradient(
NumPtr<Tensor<2,SPACEDIM> > &return_gradients,
const Point<SPACEDIM>&) const {
  if (return_gradients.getNumber()!=n_components) {
    return_gradients.cleanup();
    return_gradients.allocate(n_components);
  }
  for (int c=0;c<n_components;c++) return_gradients[c].initialize(0.);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ConstantComponentTensorFunctionSystem::value(
NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &return_values,
const NumPtr<Point<SPACEDIM> > &points) const{
  if (return_values.getNumber()!=points.getNumber()) {
    return_values.cleanup();
    return_values.allocate(points.getNumber());
  }
  for (int i=0;i<points.getNumber();i++) {
    if (return_values[i].getNumber()!=n_components) {
      return_values[i].allocate(n_components);
    }
    return_values[i].initialize(zero_value);
    return_values[i][selected]=function_value;
  }
}

Tensor<1,SPACEDIM> ConstantComponentTensorFunctionSystem::zero_value;
#include "NumPtr.C"
INSTANTIATE_NUMPTR(Vector);
