#ifndef FUNCTION_SYSTEM_H
#define FUNCTION_SYSTEM_H "$Header:$"
//----------------------------  function.h  ---------------------------
//    $Id: function.h,v 1.41 2003/02/06 17:37:50 wolf Exp $
//    Version: $Name: Version-4-0-0 $
//
//    Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal authors
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//----------------------------  function.h  ---------------------------
//----------------------------  tensor_function.h  -------------------------    --
//    $Id: tensor_function.h,v 1.23 2002/07/23 14:15:43 wolf Exp $
//    Version: $Name: Version-4-0-0 $
//
//    Copyright (C) 1998, 1999, 2000, 2001, 2002 by the deal authors
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//----------------------------  tensor_function.h  -------------------------
//
//modified from deal.II/base/include/base/function.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <Vector.H>
#include <NumPtr.H>
#include <Point.H>
#include <Tensor.H>

//compare to DEAL Function
class FunctionSystem {
  protected:
    int n_components;
    explicit FunctionSystem(const FunctionSystem &bc) : 
    n_components(bc.n_components) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    FunctionSystem(int nc=1) : n_components(nc) {}
    virtual ~FunctionSystem() {}
    int nComponents() const { return n_components; }
    virtual REAL value(const Point<SPACEDIM>&,int) const {
      OBSOLETE("not defined by user");
      return HUGE_VAL;
    }
    virtual void value(const Point<SPACEDIM>&,Vector &v) const {
      OBSOLETE("not defined by user");
      v=HUGE_VAL;
    }
    void value(NumPtr<REAL>&,const NumPtr<Point<SPACEDIM> >&,int) const;
    void value(NumPtr<Vector>&,const NumPtr<Point<SPACEDIM> >&) const;
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM>&,int) const
    {
      OBSOLETE("not defined by user");
      Tensor<1,SPACEDIM> return_gradient;
      return_gradient.initialize(HUGE_VAL);
      return return_gradient;
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> > &return_gradients,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not defined by user");
      return_gradients.allocate(n_components);
      for (int c=0;c<n_components;c++) {
        return_gradients[c].initialize(HUGE_VAL);
      }
    }
    void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
      const NumPtr<Point<SPACEDIM> >&,int) const;
    void gradient(NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&,
      const NumPtr<Point<SPACEDIM> >&) const;
    virtual void printOn(ostream &os) const {
      os << "\tFunctionSystem: this = " << this
         << "\n\tn_components = " << n_components << endl;
    }
};

//compare to DEAL ConstantFunction
class ConstantFunctionSystem : public FunctionSystem {
  private:
    static Tensor<1,SPACEDIM> zero_tensor;
  protected:
    const REAL function_value;
  public:
    ConstantFunctionSystem(REAL fv,int nc=1) : FunctionSystem(nc), 
      function_value(fv) {}
    virtual ~ConstantFunctionSystem() {}
    virtual REAL value(const Point<SPACEDIM>&,int component) const {
      return function_value; 
    }
    virtual void value(const Point<SPACEDIM>&,Vector &return_value) const {
      CHECK_SAME(return_value.size(),n_components);
      return_value=function_value;
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM>&,int) const {
      return zero_tensor;
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
      const Point<SPACEDIM>&) const;
    virtual void printOn(ostream &os) const {
      os << "\tConstantFunctionSystem: function_value = " 
         << function_value << endl;
      FunctionSystem::printOn(os);
    }
};

//compare to DEAL ComponentSelectFunction
class ConstantComponentFunctionSystem : 
public ConstantFunctionSystem {
  protected:
    const int selected;
  public:
    ConstantComponentFunctionSystem(int s,REAL fv,int nc) : 
    ConstantFunctionSystem(fv,nc),
      selected(s) {}
    virtual REAL value(const Point<SPACEDIM>&,int component) const {
      return (component==selected ? function_value : 0.); 
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &return_value) 
    const {
      CHECK_SAME(return_value.size(),n_components);
      return_value=0.;
      return_value[selected]=function_value;
    }
    virtual void value(NumPtr<REAL> &return_values,
    const NumPtr<Point<SPACEDIM> > &points,int component) const {
      if (component==selected) {
        value(return_values,points,component);
      } else {
        return_values.initialize(0.);
      }
    }
    virtual void value(NumPtr<Vector>&,const NumPtr<Point<SPACEDIM> >&)
      const;
    virtual void printOn(ostream &os) const {
      os << "\tConstantComponentFunctionSystem: selected = " << selected 
         << endl;
      ConstantFunctionSystem::printOn(os);
    }
};

//not in DEAL
class TimeDependentFunctionSystem : public FunctionSystem {
  protected:
    explicit TimeDependentFunctionSystem(
    const TimeDependentFunctionSystem &bc) : FunctionSystem(bc) {
      OBSOLETE("copy constructor disallowed");
    }
    virtual REAL value(const Point<SPACEDIM> &point,int component) const {
      OBSOLETE("prohibit FunctionSystem version");
      return FunctionSystem::value(point,component);
    }
    virtual void value(const Point<SPACEDIM> &point,Vector &v) const {
      OBSOLETE("prohibit FunctionSystem version");
      FunctionSystem::value(point,v);
    }
    void value(NumPtr<REAL> &return_values,
    const NumPtr<Point<SPACEDIM> > &points,int component) const {
      OBSOLETE("prohibit FunctionSystem version");
      FunctionSystem::value(return_values,points,component);
    }
    void value(NumPtr<Vector> &return_values,
    const NumPtr<Point<SPACEDIM> > &points) const {
      OBSOLETE("prohibit FunctionSystem version");
      FunctionSystem::value(return_values,points);
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &point,
    int component) const {
      OBSOLETE("prohibit FunctionSystem version");
      return FunctionSystem::gradient(point,component);
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> > &return_gradients,
    const Point<SPACEDIM> &points) const {
      OBSOLETE("prohibit FunctionSystem version");
      FunctionSystem::gradient(return_gradients,points);
    }
    void gradient(NumPtr<Tensor<1,SPACEDIM> > &return_gradients,
    const NumPtr<Point<SPACEDIM> > &points,int component) const {
      OBSOLETE("prohibit FunctionSystem version");
      FunctionSystem::gradient(return_gradients,points,component);
    }
    void gradient(NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &return_gradients,
    const NumPtr<Point<SPACEDIM> > &points) const {
      OBSOLETE("prohibit FunctionSystem version");
      FunctionSystem::gradient(return_gradients,points);
    }
  public:
    TimeDependentFunctionSystem(int nc=1) : FunctionSystem(nc) {}
    virtual ~TimeDependentFunctionSystem() {}
    int nComponents() const { return n_components; }
    virtual REAL value(const Point<SPACEDIM>&,REAL,int) const {
      OBSOLETE("not defined by user");
      return HUGE_VAL;
    }
    virtual void value(const Point<SPACEDIM>&,REAL,Vector &v) const {
      OBSOLETE("not defined by user");
      v=HUGE_VAL;
    }
    void value(NumPtr<REAL>&,const NumPtr<Point<SPACEDIM> >&,REAL,int)
      const;
    void value(NumPtr<Vector>&,const NumPtr<Point<SPACEDIM> >&,REAL) const;
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM>&,REAL,int)
    const {
      OBSOLETE("not defined by user");
      Tensor<1,SPACEDIM> return_gradient;
      return_gradient.initialize(HUGE_VAL);
      return return_gradient;
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&,REAL) const {
      OBSOLETE("not defined by user");
    }
    void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
      const NumPtr<Point<SPACEDIM> >&,REAL,int) const;
    void gradient(NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&,
      const NumPtr<Point<SPACEDIM> >&,REAL) const;
    virtual void printOn(ostream &os) const {
      os << "\tFunctionSystem: this = " << this << endl;
      FunctionSystem::printOn(os);
    }
};

//compare to DEAL TensorFunction
class TensorFunctionSystem {
  private:
    static Tensor<1,SPACEDIM> bogus;
  protected:
    int n_components;
    explicit TensorFunctionSystem(const TensorFunctionSystem &bc) : 
    n_components(bc.n_components) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    TensorFunctionSystem(int nc=1) : n_components(nc) {}
    virtual ~TensorFunctionSystem () {}
    int nComponents() const { return n_components; }
    virtual Tensor<1,SPACEDIM> value(const Point<SPACEDIM>&,int) const {
      OBSOLETE("not defined by user");
      return bogus;
    }
    virtual void value(const Point<SPACEDIM>&,
    NumPtr<Tensor<1,SPACEDIM> > &v) const {
      OBSOLETE("not defined by user");
      v.initialize(bogus);
    }
    void value(NumPtr<Tensor<1,SPACEDIM> >&,
      const NumPtr<Point<SPACEDIM> >&,int) const;
    void value(NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&,
      const NumPtr<Point<SPACEDIM> >&) const;
    virtual Tensor<2,SPACEDIM> gradient(const Point<SPACEDIM>&,int) const
    {
      OBSOLETE("not defined by user");
      Tensor<2,SPACEDIM> return_gradient;
      return_gradient.initialize(HUGE_VAL);
      return return_gradient;
    }
    virtual void gradient(NumPtr<Tensor<2,SPACEDIM> > &return_gradients,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not defined by user");
      return_gradients.allocate(n_components);
      for (int c=0;c<n_components;c++) {
        return_gradients[c].initialize(HUGE_VAL);
      }
    }
    void gradient(NumPtr<Tensor<2,SPACEDIM> >&,
      const NumPtr<Point<SPACEDIM> >&,int) const;
    void gradient(NumPtr<NumPtr<Tensor<2,SPACEDIM> > >&,
      const NumPtr<Point<SPACEDIM> >&) const;
    virtual void printOn(ostream &os) const {
      os << "TensorFunctionSystem: this = " << this
         << "\n\tn_components = " << n_components << endl;
    }
};

//compare to DEAL TensorFunction
class ConstantTensorFunctionSystem : public TensorFunctionSystem {
  private:
    static Tensor<1,SPACEDIM> zero_value;
    static Tensor<2,SPACEDIM> zero_gradient;
  protected:
    const Tensor<1,SPACEDIM> function_value;
  public:
    ConstantTensorFunctionSystem(const Tensor<1,SPACEDIM> &fv,int nc=1) :
      TensorFunctionSystem(nc),function_value(fv) {}
    virtual ~ConstantTensorFunctionSystem() {}
    virtual Tensor<1,SPACEDIM> value(const Point<SPACEDIM>&,int component)
    const {
      return function_value; 
    }
    virtual void value(const Point<SPACEDIM>&,
    NumPtr<Tensor<1,SPACEDIM> > &return_value) const {
      CHECK_SAME(return_value.getNumber(),n_components);
      return_value.initialize(function_value);
    }
    virtual Tensor<2,SPACEDIM> gradient(const Point<SPACEDIM>&,int) const {
      return zero_gradient;
    }
    virtual void gradient(NumPtr<Tensor<2,SPACEDIM> >&,
      const Point<SPACEDIM>&) const;
    virtual void printOn(ostream &os) const {
      os << "\tConstantTensorFunctionSystem: function_value = " 
         << function_value << endl;
      TensorFunctionSystem::printOn(os);
    }
};

//not in DEAL
class ConstantComponentTensorFunctionSystem : 
public ConstantTensorFunctionSystem {
  private:
    static Tensor<1,SPACEDIM> zero_value;
  protected:
    const int selected;
  public:
    ConstantComponentTensorFunctionSystem(int s,
    const Tensor<1,SPACEDIM> &fv,int nc) : 
      ConstantTensorFunctionSystem(fv,nc),selected(s) {}
    virtual Tensor<1,SPACEDIM> value(const Point<SPACEDIM>&,int component)
    const {
      return (component==selected ? function_value : zero_value); 
    }
    virtual void value(const Point<SPACEDIM> &p,
    NumPtr<Tensor<1,SPACEDIM> > &return_value) const {
      CHECK_SAME(return_value.getNumber(),n_components);
      return_value.initialize(zero_value);
      return_value[selected]=function_value;
    }
    virtual void value(NumPtr<Tensor<1,SPACEDIM> > &return_values,
    const NumPtr<Point<SPACEDIM> > &points,int component) const {
      if (component==selected) {
        value(return_values,points,component);
      } else {
        return_values.initialize(zero_value);
      }
    }
    virtual void value(NumPtr<NumPtr<Tensor<1,SPACEDIM> > >&,
      const NumPtr<Point<SPACEDIM> >&) const;
    virtual void printOn(ostream &os) const {
      os << "\tConstantComponentTensorFunctionSystem: selected = "
         << selected << endl;
      ConstantTensorFunctionSystem::printOn(os);
    }
};

#endif
