#include <MaterialModel.H>
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void LaplaceMaterialModel::elementStiffnessMatrixAndRHS(
const NumPtr<Point<SPACEDIM> > &q_points,
const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_matrix,
Vector &cell_rhs) const {
  int dofs_per_cell=cell_rhs.size();
  for (int q=0;q<q_points.getNumber();q++) {
    REAL rhs_value=rightHandSideValue(q_points[q]);
    REAL jxw=jxw_values[q];
    for (int i=0;i<dofs_per_cell;i++) {
      Tensor<1,SPACEDIM> sgi=sg(i,q);
      sgi*=jxw;
      REAL si=sv(i,q)*jxw;
      for (int j=0;j<dofs_per_cell;j++) cell_matrix(i,j)+=sgi*sg(j,q);
      cell_rhs[i]+=si*rhs_value;
    }
  }
}
