#include <iostream>
#include <stdlib.h>
#include "DDecl.H"
#include "GUIEnumInputParameter.H"
#include "GUIVirtualInput.H"
#include "InputParameter.H"
#include "Palette.H"
#include "Quadrature.H"
#include "SetTraps.H"
#include "Tracer.H"
#include "XYGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKGUI.H"
#include "GTKGUIVirtualInput.H"
#include "GTKWindow.H"
#else
#include "GUI.H"
#include "GUIVirtualInput.H"
#endif

GUI_INPUT_PARAMETER_LIST_TYPE *param_list=0;
BOOLEAN skip_gui=FALSE;
double winsize=0.5;
char* display_name=0;

//enum QUADRATURE_RULE{GAUSSIAN_QUADRATURE,LOBATTO_QUADRATURE,
//  NEWTON_COTES_QUADRATURE};
//static const char *quadrature_name[3]=
//  {"gaussian","lobatto","newton-cotes"};
//QUADRATURE_RULE quadrature_rule=GAUSSIAN_QUADRATURE;
//int iquadrature_rule=quadrature_rule;
#if (SPACEDIM==2)
enum SHAPE{TRIANGLE_SHAPE,QUADRILATERAL_SHAPE};
static const char *shape_name[2]={"triangle","quadrilateral"};
SHAPE shape=QUADRILATERAL_SHAPE;
#elif (SPACEDIM==3)
enum SHAPE{TETRAHEDRON_SHAPE,PRISM_SHAPE,HEXAHEDRON_SHAPE};
static const char *shape_name[3]={"tetrahedron","prism","hexahedron"};
SHAPE shape=HEXAHEDRON_SHAPE;
#endif
#if (SPACEDIM>1)
int ishape=shape;
#endif

int order=2;
int number_refinements=2;
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void checkMainInput() {
//TRACER_CALL(tr,"checkMainInput");
#if (SPACEDIM>1)
  shape=static_cast<SHAPE>(ishape);
#endif
  order=2*(order/2);
#ifdef DEBUG
//cout << "\torder = " << order << endl;
#endif
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
void runInterval(double *hr,double *gq_work,double *lq_work,
double *ncq_work,double *gq_error,double *lq_error,double *ncq_error,
double &ymin,double &ymax) {
  double log10=log(10.);
  GaussianQuadrature<SPACEDIM> gq(order/2);  // ==>gq.order()==order
  const NumPtr< Point<SPACEDIM> > &gqp=gq.points();
  const NumPtr<REAL> &gqw=gq.weights();
  double ngq=static_cast<double>(gq.numberPoints());

  LobattoQuadrature<SPACEDIM> lq(order/2+1); // ==>lq.order()==order
  const NumPtr< Point<SPACEDIM> > &lqp=lq.points();
  const NumPtr<REAL> &lqw=lq.weights();
  double nlq=static_cast<double>(lq.numberPoints());

//order-1 would be sufficient, but order is more efficient:
  NewtonCotesQuadrature<SPACEDIM> ncq(max(2,min(8,order)));
  const NumPtr< Point<SPACEDIM> > &ncqp=ncq.points();
  const NumPtr<REAL> &ncqw=ncq.weights();
  double nncq=static_cast<double>(ncq.numberPoints());

  double h=1.;
  double n=1.;
  for (int r=0;r<number_refinements;r++,h*=0.5,n*=2.) {
    double nelements=n;
    hr[r]=-log(h)/log10;
    double exact=exp(h)-1.; // int_0^h exp(x) dx / h
    double gq_val=0.;
    for (int q=0;q<gq.numberPoints();q++) {
      gq_val+=exp(h*gqp[q][0])*gqw[q];
    }
    double err=abs(gq_val*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    gq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,gq_error[r]);
    ymin=min(ymin,gq_error[r]);
    gq_work[r]=log(ngq*nelements)/log10;

    double lq_val=0.;
    for (int q=0;q<lq.numberPoints();q++) {
      lq_val+=exp(h*lqp[q][0])*lqw[q];
    }
    err=abs(lq_val*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    lq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,lq_error[r]);
    ymin=min(ymin,lq_error[r]);
    lq_work[r]=log(nlq*nelements)/log10;

    double ncq_val=0.;
    for (int q=0;q<ncq.numberPoints();q++) {
      ncq_val+=exp(h*ncqp[q][0])*ncqw[q];
    }
    err=abs(ncq_val*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    ncq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,ncq_error[r]);
    ymin=min(ymin,ncq_error[r]);
    ncq_work[r]=log(nncq*nelements)/log10;
  }
}
#endif
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void runQuadrilateral(double *hr,double *gq_work,double *lq_work,
double *ncq_work,double *gq_error,double *lq_error,double *ncq_error,
double &ymin,double &ymax) {
//TRACER_CALL(tr,"runQuadrilateral");
  double log10=log(10.);
  GaussianQuadrature<SPACEDIM> gq(order/2);  // ==>gq.order()==order
  const NumPtr< Point<SPACEDIM> > &gqp=gq.points();
  const NumPtr<REAL> &gqw=gq.weights();
  double ngq=static_cast<double>(gq.numberPoints());
#ifdef DEBUG
//cout << "\tngq = " << ngq << endl;
//for (int q=0;q<gqp.getNumber();q++) {
//  cout << "\tgqp[" << q << "] = " << gqp[q] << endl;
//}
//for (int q=0;q<gqw.getNumber();q++) {
//  cout << "\tgqw[" << q << "] = " << gqw[q] << endl;
//}
#endif

  LobattoQuadrature<SPACEDIM> lq(order/2+1); // ==>lq.order()==order
  const NumPtr< Point<SPACEDIM> > &lqp=lq.points();
  const NumPtr<REAL> &lqw=lq.weights();
  double nlq=static_cast<double>(lq.numberPoints());

//order-1 would be sufficient, but order is more efficient:
  NewtonCotesQuadrature<SPACEDIM> ncq(max(2,min(8,order)));
  const NumPtr< Point<SPACEDIM> > &ncqp=ncq.points();
  const NumPtr<REAL> &ncqw=ncq.weights();
  double nncq=static_cast<double>(ncq.numberPoints());

  double h=1.;
  double n=1.;
  for (int r=0;r<number_refinements;r++,h*=0.5,n*=2.) {
    double nelements=n*n;
    hr[r]=-log(h)/log10;
    double exact=pow(exp(h)-1.,SPACEDIM);
    double gq_val=0.;
    for (int q=0;q<gq.numberPoints();q++) {
      gq_val+=exp(h*(gqp[q][0]+gqp[q][1]))*gqw[q];
    }
    double err=abs(gq_val*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    gq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,gq_error[r]);
    ymin=min(ymin,gq_error[r]);
    gq_work[r]=log(ngq*nelements)/log10;

    double lq_val=0.;
    for (int q=0;q<lq.numberPoints();q++) {
      lq_val+=exp(h*(lqp[q][0]+lqp[q][1]))*lqw[q];
    }
    err=abs(lq_val*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    lq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,lq_error[r]);
    ymin=min(ymin,lq_error[r]);
    lq_work[r]=log(nlq*nelements)/log10;

    double ncq_val=0.;
    for (int q=0;q<ncq.numberPoints();q++) {
      ncq_val+=exp(h*(ncqp[q][0]+ncqp[q][1]))*ncqw[q];
    }
    err=abs(ncq_val*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    ncq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,ncq_error[r]);
    ymin=min(ymin,ncq_error[r]);
    ncq_work[r]=log(nncq*nelements)/log10;
  }
}
#endif
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void runTriangle(double *hr,double *gq_work,double *lq_work,
double *ncq_work,double *gq_error,double *lq_error,double *ncq_error,
double &ymin,double &ymax) {
//TRACER_CALL(tr,"runTriangle");
  double log10=log(10.);
  TriangleGaussianQuadrature gq(min(31,order));
  const NumPtr< Point<SPACEDIM> > &gqp=gq.points();
  const NumPtr<REAL> &gqw=gq.weights();
  double ngq=static_cast<double>(gq.numberPoints());

  TriangleLobattoQuadrature lq(min(19,order));
  const NumPtr< Point<SPACEDIM> > &lqp=lq.points();
  const NumPtr<REAL> &lqw=lq.weights();
  double nlq=static_cast<double>(lq.numberPoints());

//order-1 would be sufficient, but order is more efficient:
  TriangleNewtonCotesQuadrature ncq(min(6,order+2));
  const NumPtr< Point<SPACEDIM> > &ncqp=ncq.points();
  const NumPtr<REAL> &ncqw=ncq.weights();
  double nncq=static_cast<double>(ncq.numberPoints());
#ifdef DEBUG
//cout << "\tquadrature rule orders = " << gq.order() << " "
//  << lq.order() << " " << ncq.order() << endl;
#endif

  double h=1.;
  double n=1.;
  for (int r=0;r<number_refinements;r++,h*=0.5,n*=2.) {
    double nelements=n*(n+1.)*0.5;
    hr[r]=-log(h)/log10;
    double exact=1.-exp(h)*(1.-h);
    double gq_val=0.;
    for (int q=0;q<gq.numberPoints();q++) {
//    cout << "\tgqp[" << q << "] = " << gqp[q] << " gqw = " << gqw[q]
//         << " gq_val inc = " << exp(h*(gqp[q][0]+gqp[q][1]))*gqw[q]
//         << endl;
      gq_val+=exp(h*(gqp[q][0]+gqp[q][1]))*gqw[q];
    }
//  cout << "\tgq_val = " << gq_val << endl;
    double err=abs(gq_val*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    gq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,gq_error[r]);
    ymin=min(ymin,gq_error[r]);
    gq_work[r]=log(ngq*nelements)/log10;
//  cout << "\tgq_val = " << gq_val << " exact = " << exact/(h*h) << endl;

    double lq_val=0.;
    for (int q=0;q<lq.numberPoints();q++) {
      lq_val+=exp(h*(lqp[q][0]+lqp[q][1]))*lqw[q];
    }
//  cout << "\tlq_val = " << lq_val << endl;
    err=abs(lq_val*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    lq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,lq_error[r]);
    ymin=min(ymin,lq_error[r]);
    lq_work[r]=log(nlq*nelements)/log10;

    double ncq_val=0.;
    for (int q=0;q<ncq.numberPoints();q++) {
      ncq_val+=exp(h*(ncqp[q][0]+ncqp[q][1]))*ncqw[q];
    }
//  cout << "\tncq_val = " << ncq_val << endl;
    err=abs(ncq_val*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    ncq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,ncq_error[r]);
    ymin=min(ymin,ncq_error[r]);
    ncq_work[r]=log(nncq*nelements)/log10;
  }
}
#endif
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void runHexahedron(double *hr,double *gq_work,double *lq_work,
double *ncq_work,double *gq_error,double *lq_error,double *ncq_error,
double &ymin,double &ymax) {
//TRACER_CALL(tr,"runHexahedron");
  double log10=log(10.);
  GaussianQuadrature<SPACEDIM> gq(order/2);  // ==>gq.order()==order
  const NumPtr< Point<SPACEDIM> > &gqp=gq.points();
  const NumPtr<REAL> &gqw=gq.weights();
  double ngq=static_cast<double>(gq.numberPoints());
#ifdef DEBUG
//cout << "\tngq = " << ngq << endl;
//for (int q=0;q<gqp.getNumber();q++) {
//  cout << "\tgqp[" << q << "] = " << gqp[q] << endl;
//}
//for (int q=0;q<gqw.getNumber();q++) {
//  cout << "\tgqw[" << q << "] = " << gqw[q] << endl;
//}
#endif

  LobattoQuadrature<SPACEDIM> lq(order/2+1); // ==>lq.order()==order
  const NumPtr< Point<SPACEDIM> > &lqp=lq.points();
  const NumPtr<REAL> &lqw=lq.weights();
  double nlq=static_cast<double>(lq.numberPoints());

//order-1 would be sufficient, but order is more efficient:
  NewtonCotesQuadrature<SPACEDIM> ncq(max(2,min(8,order)));
  const NumPtr< Point<SPACEDIM> > &ncqp=ncq.points();
  const NumPtr<REAL> &ncqw=ncq.weights();
  double nncq=static_cast<double>(ncq.numberPoints());

  double h=1.;
  double n=1.;
  for (int r=0;r<number_refinements;r++,h*=0.5,n*=2.) {
    double nelements=n*n*n;
    hr[r]=-log(h)/log10;
    double exact=pow(exp(h)-1.,SPACEDIM);
    double gq_val=0.;
    for (int q=0;q<gq.numberPoints();q++) {
      gq_val+=exp(h*(gqp[q][0]+gqp[q][1]))*gqw[q];
    }
    double err=abs(gq_val*h*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    gq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,gq_error[r]);
    ymin=min(ymin,gq_error[r]);
    gq_work[r]=log(ngq*nelements)/log10;

    double lq_val=0.;
    for (int q=0;q<lq.numberPoints();q++) {
      lq_val+=exp(h*(lqp[q][0]+lqp[q][1]))*lqw[q];
    }
    err=abs(lq_val*h*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    lq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,lq_error[r]);
    ymin=min(ymin,lq_error[r]);
    lq_work[r]=log(nlq*nelements)/log10;

    double ncq_val=0.;
    for (int q=0;q<ncq.numberPoints();q++) {
      ncq_val+=exp(h*(ncqp[q][0]+ncqp[q][1]))*ncqw[q];
    }
    err=abs(ncq_val*h*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    ncq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,ncq_error[r]);
    ymin=min(ymin,ncq_error[r]);
    ncq_work[r]=log(nncq*nelements)/log10;
  }
}
#endif
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void runTetrahedron(double *hr,double *gq_work,double *lq_work,
double *ncq_work,double *gq_error,double *lq_error,double *ncq_error,
double &ymin,double &ymax) {
//TRACER_CALL(tr,"runTetrahedron");
  double log10=log(10.);
  TetrahedronProductGaussianQuadrature 
    gq(order==(order/2)*2 ? order/2+2 : (order+3)/2);
  const NumPtr< Point<SPACEDIM> > &gqp=gq.points();
  const NumPtr<REAL> &gqw=gq.weights();
  double ngq=static_cast<double>(gq.numberPoints());

  TetrahedronLobattoQuadrature lq(2);
  const NumPtr< Point<SPACEDIM> > &lqp=lq.points();
  const NumPtr<REAL> &lqw=lq.weights();
  double nlq=static_cast<double>(lq.numberPoints());

//order-1 would be sufficient, but order is more efficient:
  TetrahedronNewtonCotesQuadrature ncq(max(2,min(4,order)));
  const NumPtr< Point<SPACEDIM> > &ncqp=ncq.points();
  const NumPtr<REAL> &ncqw=ncq.weights();
  double nncq=static_cast<double>(ncq.numberPoints());
#ifdef DEBUG
//cout << "\tquadrature rule orders = " << gq.order() << " "
//  << lq.order() << " " << ncq.order() << endl;
#endif

  double h=1.;
  double n=1.;
  for (int r=0;r<number_refinements;r++,h*=0.5,n*=2.) {
    double nelements=n*(n+1)*(n+2)/6.;
    hr[r]=-log(h)/log10;
    double exact=exp(h)*(1.-h*(1.-h*0.5))-1.;
    double gq_val=0.;
    for (int q=0;q<gq.numberPoints();q++) {
      gq_val+=exp(h*(gqp[q][0]+gqp[q][1]+gqp[q][2]))*gqw[q];
    }
    double err=abs(gq_val*h*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    gq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,gq_error[r]);
    ymin=min(ymin,gq_error[r]);
    gq_work[r]=log(ngq*nelements)/log10;

    double lq_val=0.;
    for (int q=0;q<lq.numberPoints();q++) {
      lq_val+=exp(h*(lqp[q][0]+lqp[q][1]+lqp[q][2]))*lqw[q];
    }
    err=abs(lq_val*h*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    lq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,lq_error[r]);
    ymin=min(ymin,lq_error[r]);
    lq_work[r]=log(nlq*nelements)/log10;

    double ncq_val=0.;
    for (int q=0;q<ncq.numberPoints();q++) {
      ncq_val+=exp(h*(ncqp[q][0]+ncqp[q][1]))*ncqw[q];
    }
    err=abs(ncq_val*h*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    ncq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,ncq_error[r]);
    ymin=min(ymin,ncq_error[r]);
    ncq_work[r]=log(nncq*nelements)/log10;
  }
}
#endif
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void runPrism(double *hr,double *gq_work,double *lq_work,
double *ncq_work,double *gq_error,double *lq_error,double *ncq_error,
double &ymin,double &ymax) {
//TRACER_CALL(tr,"runPrism");
  double log10=log(10.);
  TriangleProductGaussianQuadrature 
    gqt(order==(order/2)*2 ? order/2+1 : (order+1)/2);
  GaussianQuadrature<1> gq1(order/2);
  ProductQuadrature gq(gqt,gq1);
  const NumPtr< Point<SPACEDIM> > &gqp=gq.points();
  const NumPtr<REAL> &gqw=gq.weights();
  double ngq=static_cast<double>(gq.numberPoints());

  TriangleLobattoQuadrature lqt(max(2,min(3,order)));
  LobattoQuadrature<1> lq1(order/2+1); // ==>lq.order()==order
  ProductQuadrature lq(lqt,lq1);
  const NumPtr< Point<SPACEDIM> > &lqp=lq.points();
  const NumPtr<REAL> &lqw=lq.weights();
  double nlq=static_cast<double>(lq.numberPoints());

//order-1 would be sufficient, but order is more efficient:
  TriangleNewtonCotesQuadrature ncqt(max(2,min(7,order)));
  NewtonCotesQuadrature<1> ncq1(max(2,min(8,order)));
  ProductQuadrature ncq(ncqt,ncq1);
  const NumPtr< Point<SPACEDIM> > &ncqp=ncq.points();
  const NumPtr<REAL> &ncqw=ncq.weights();
  double nncq=static_cast<double>(ncq.numberPoints());
#ifdef DEBUG
//cout << "\tquadrature rule orders = " << gq.order() << " "
//  << lq.order() << " " << ncq.order() << endl;
#endif

  double h=1.;
  double n=1.;
  for (int r=0;r<number_refinements;r++,h*=0.5,n*=2.) {
    double nelements=n*n*(n+1)*0.5;
    hr[r]=-log(h)/log10;
    double exact=exp(h)*(2.-h)-exp(2.*h)*(1.-h)-1.;
    double gq_val=0.;
    for (int q=0;q<gq.numberPoints();q++) {
      gq_val+=exp(h*(gqp[q][0]+gqp[q][1]+gqp[q][2]))*gqw[q];
    }
    double err=abs(gq_val*h*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    gq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,gq_error[r]);
    ymin=min(ymin,gq_error[r]);
    gq_work[r]=log(ngq*nelements)/log10;

    double lq_val=0.;
    for (int q=0;q<lq.numberPoints();q++) {
      lq_val+=exp(h*(lqp[q][0]+lqp[q][1]+lqp[q][2]))*lqw[q];
    }
    err=abs(lq_val*h*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    lq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,lq_error[r]);
    ymin=min(ymin,lq_error[r]);
    lq_work[r]=log(nlq*nelements)/log10;

    double ncq_val=0.;
    for (int q=0;q<ncq.numberPoints();q++) {
      ncq_val+=exp(h*(ncqp[q][0]+ncqp[q][1]))*ncqw[q];
    }
    err=abs(ncq_val*h*h*h-exact);
    if (err<=0.) err=exact*DBL_EPSILON;
    ncq_error[r]=log(err*nelements)/log10;
    ymax=max(ymax,ncq_error[r]);
    ymin=min(ymin,ncq_error[r]);
    ncq_work[r]=log(nncq*nelements)/log10;
  }
}
#endif
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void runMain(BOOLEAN /*called_before*/) {
//TRACER_CALL(tr,"runMain");
  double *hr=OPERATOR_NEW_BRACKET(double,number_refinements);
  double *gq_work=OPERATOR_NEW_BRACKET(double,number_refinements);
  double *lq_work=OPERATOR_NEW_BRACKET(double,number_refinements);
  double *ncq_work=OPERATOR_NEW_BRACKET(double,number_refinements);
  double *gq_error=OPERATOR_NEW_BRACKET(double,number_refinements);
  double *lq_error=OPERATOR_NEW_BRACKET(double,number_refinements);
  double *ncq_error=OPERATOR_NEW_BRACKET(double,number_refinements);
  double ymax=0.;
  double ymin=-1.;
#if (SPACEDIM==1)
  runInterval(hr,gq_work,lq_work,ncq_work,gq_error,lq_error,ncq_error,
    ymin,ymax);
#elif (SPACEDIM==2)
  switch (shape) {
    case TRIANGLE_SHAPE:
      runTriangle(hr,gq_work,lq_work,ncq_work,gq_error,lq_error,ncq_error,
        ymin,ymax);
      break;
    default:
      runQuadrilateral(hr,gq_work,lq_work,ncq_work,gq_error,lq_error,
        ncq_error,ymin,ymax);
  }
#else
  switch (shape) {
    case TETRAHEDRON_SHAPE:
      runTetrahedron(hr,gq_work,lq_work,ncq_work,gq_error,lq_error,
        ncq_error,ymin,ymax);
      break;
    case PRISM_SHAPE: {
      runPrism(hr,gq_work,lq_work,ncq_work,gq_error,lq_error,ncq_error,
        ymin,ymax);
      break;
    }
    default:
      runHexahedron(hr,gq_work,lq_work,ncq_work,gq_error,lq_error,
        ncq_error,ymin,ymax);
  }
#endif
  cout << "\tGaussianQuadrature in red" << endl;
  cout << "\tLobattoQuadrature in green" << endl;
  cout << "\tNewtonCotesQuadrature in blue" << endl;
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool egt("quadrature rules for integral of exponential",
    "- log_10(h)","log_10(error)",hr[0],hr[number_refinements-1],
    ymin,ymax,&cmap,0,winsize);
  egt.newPage();
  egt.setbgColor("white");
  egt.setfgColor("black");
  egt.drawAxes();
  egt.setfgColor("red");
  egt.movePen(hr[0],gq_error[0]);
  for (int r=1;r<number_refinements;r++) egt.drawLine(hr[r],gq_error[r]);
  egt.setfgColor("green");
  egt.movePen(hr[0],lq_error[0]);
  for (int r=1;r<number_refinements;r++) egt.drawLine(hr[r],lq_error[r]);
  egt.setfgColor("blue");
  egt.movePen(hr[0],ncq_error[0]);
  for (int r=1;r<number_refinements;r++) egt.drawLine(hr[r],ncq_error[r]);
  egt.flush();

  XYGraphTool wgt("quadrature rules for integral of exponential",
    "log_10(work)","log_10(error)",
    min(gq_work[0],min(lq_work[0],ncq_work[0])),
    max(gq_work[number_refinements-1],
        max(lq_work[number_refinements-1],ncq_work[number_refinements-1])),
    ymin,ymax,&cmap,0,winsize);
  wgt.newPage();
  wgt.setbgColor("white");
  wgt.setfgColor("black");
  wgt.drawAxes();
  wgt.setfgColor("red");
  wgt.movePen(hr[0],gq_error[0]);
  for (int r=1;r<number_refinements;r++) {
    wgt.drawLine(gq_work[r],gq_error[r]);
  }
  wgt.setfgColor("green");
  wgt.movePen(hr[0],lq_error[0]);
  for (int r=1;r<number_refinements;r++) {
    wgt.drawLine(lq_work[r],lq_error[r]);
  }
  wgt.setfgColor("blue");
  wgt.movePen(hr[0],ncq_error[0]);
  for (int r=1;r<number_refinements;r++) {
    wgt.drawLine(ncq_work[r],ncq_error[r]);
  }
  wgt.flush();
  XYGraphTool::WINDOW_TYPE::QuitButton qb;

  delete [] hr;
  delete [] gq_work;
  delete [] lq_work;
  delete [] ncq_work;
  delete [] gq_error;
  delete [] lq_error;
  delete [] ncq_error;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void cleanup() {
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void shutdown() {
  while (param_list->notEmpty()) delete param_list->delAfter(0);
  delete param_list; param_list=0;
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int main(int argc,char* argv[]) {
#ifdef DEBUG
//setTraps();
#endif
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  {
//  define input parameters and bounds
    param_list=OPERATOR_NEW GUI_INPUT_PARAMETER_LIST_TYPE("Main List");
    { const char *group="Grid Parameters";
#if (SPACEDIM==1)
#elif (SPACEDIM==2)
      param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
        ishape,"shape",shape_name[0],shape_name[1]));
#else
      param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
        ishape,"shape",shape_name[0],shape_name[1],shape_name[2]));
#endif
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(
        number_refinements,"number_refinements",2,INT_MAX,group));
    }
    { const char *group="Quadrature Parameters";
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(
        order,"order",2,INT_MAX,group));
    }
    { const char *group="Graphics Parameters";
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(winsize,
        "winsize",0.,1.,group) );
    }

//  read input parameters 
    if (argc>1) {
      ifstream in_file;
      in_file.open(argv[1],ios::in);
      if (in_file) {
        istream &infile(in_file);
        infile.clear(ios::goodbit);
        infile.seekg(0,ios::beg);
        char name[LENGTH_NAME], comment[LENGTH_COMMENT];
        BOOLEAN found_main=FALSE;
        while ( in_file >> setw(LENGTH_NAME) >> name ) {
          if ( strcmp(name,"Main") == 0 ) {
            in_file.getline( comment, LENGTH_COMMENT);
            found_main=TRUE;
            while ( in_file >> setw(LENGTH_NAME) >> name ) {
              if ( strcmp(name,"end") == 0 ) break;
              else if (strcmp(name,"skip_gui") == 0) {
                int iskip_gui; // type bool not read correctly
                in_file >> iskip_gui;
                skip_gui= iskip_gui!=0;
              } else param_list->formattedRead(in_file,name);
              in_file.getline( comment, LENGTH_COMMENT);
            }
          } else in_file.getline(comment,LENGTH_COMMENT);
        }
        if ( !found_main ) skip_gui=FALSE;
      }
      if (skip_gui) checkMainInput();
    } else skip_gui=FALSE;

    if (skip_gui) {
      runMain(FALSE);
      cleanup();
      shutdown();
    } else {
#ifdef USE_GTK
      GTKGUI gui(argv[0],display_name,param_list,&runMain,
      &checkMainInput,&cleanup,&shutdown,TRUE);
#else
      GUI gui(argv[0],display_name,param_list,&runMain,&checkMainInput,
      &cleanup,&shutdown,TRUE);
#endif
      gui.createFileMenu();
      gui.createViewMenu();
      gui.createHelpMenu();
      gui.createMainWindow(argc,argv);
      gui.eventLoop();
    }
  } // scope of MemoryDebugger
  return EXIT_SUCCESS;
}

template class InputParameter<int>;
template class InputParameter<double>;
template class InputParameter<bool>;
