#ifndef SHAPE_H
#define SHAPE_H "$Header:$"
//---------------------------  geometry_info.h  ---------------------------
//  $Id: geometry_info.h,v 1.41 2003/05/05 13:51:32 wolf Exp $
//  Version: $Name: Version-4-0-0 $
//
//  Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
//
//---------------------------  geometry_info.h  ---------------------------
//
//modified from deal.II/deal.II/include/grid/geometry_info.h
//  by John Trangenstein, August 2009

#include <map>
#include "AreaGraphTool.H"
#include "DDecl.H"
#include "NumPtr.H"
#include "Point.H"
#include "Tensor.H"
#include "VolGraphTool.H"

#define SHAPE_CLASS D_CHOOSE(Shape1,Shape2,Shape3)
#define CODIMENSION1_SHAPE_CLASS D_CHOOSE(Shape,Shape1,Shape2)
#define CODIMENSION2_SHAPE_CLASS D_CHOOSE(Shape,Shape,Shape1)

inline int childrenPerShape() { return 1<<SPACEDIM; }
inline int childrenPerCodimension1Shape() {
  return 1<<(SPACEDIM-1);
}
#if (SPACEDIM==3)
inline int childrenPerCodimension2Shape() { return 2; }
#endif

//compare to DEAL GeometryInfo<0>
class Shape {
  protected:
    int element_data_index;
    int dof_index;
    NumPtr<int> vertex_indices;
  public:
    inline Shape(int edi,int nv):
      element_data_index(edi),dof_index(-1),vertex_indices(nv) {}
    virtual inline ~Shape() { }

    inline int elementDataIndex() const { return element_data_index; }

    inline int dofIndex() const { return dof_index; }
    inline void setDofIndex(int d) { dof_index=d; }

    inline int numberVertices() const {
      return vertex_indices.getNumber();
    }
    inline int vertexIndex(int v) const {
      return vertex_indices[v];
    }
    inline void setVertexIndex(int v,int index) {
      vertex_indices[v]=index;
    }

    int vertexEqualToVertexIndex(int) const;
    int sharedVertexIndex(const Shape*) const;
    int sharedFaceElementIndex(const Shape*) const;

    virtual int numberCodimension1Shapes() const { return 2; }
    virtual int codimension1ShapeIndex(int f) const {
      return vertexIndex(f);
    }
    virtual int numberInteriorChildVertices() const {return 0;}
    virtual void interiorChildVertex(int &c,int &v) 
    const { 
      OBSOLETE("this shape has no interior child vertex");
    }
    virtual void printOn(ostream &os) const;

    virtual bool contains(const Point<SPACEDIM>&,
    const NumPtr<Point<SPACEDIM> >&) const {
      OBSOLETE("should have been defined in derived class");
      return false;
    }
    virtual Point<SPACEDIM> parentCoordinateForCanonical(
    const Point<SPACEDIM>&,const Shape*,int,const NumPtr<int>&) const {
      OBSOLETE("should have been defined in derived class");
      Point<SPACEDIM> dummy;
      return dummy;
    }
    virtual Point<SPACEDIM> childCoordinateOfCanonical(
    const Point<SPACEDIM>&,const Shape*,int,const NumPtr<int>&) const {
      OBSOLETE("should have been defined in derived class");
      Point<SPACEDIM> dummy;
      return dummy;
    }
    virtual int elementVertexFromFace(int,int) const {
      OBSOLETE("should have been defined in derived class");
      return INT_MAX;
    }
    virtual int elementVertexFromLine(int,int) const {
      OBSOLETE("should have been defined in derived class");
      return INT_MAX;
    }
    virtual int childOnFace(int,int) const {
      OBSOLETE("should have been defined in derived class");
      return INT_MAX;
    }
    virtual int numberInteriorChildCodimension1Shapes() const {
      OBSOLETE("should have been defined in derived class");
      return INT_MAX;
    }
    virtual void interiorChildCodimension1ShapeNeighbors(int,int&,int&) 
    const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual int numberEquallySpacedLatticePoints(int) const {
      OBSOLETE("should have been defined in derived class");
      return INT_MAX;
    }
    virtual void equallySpacedLatticePoints(int,
    NumPtr<Point<SPACEDIM> >&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual int numberInteriorEquallySpacedLatticePoints(int) const {
      OBSOLETE("should have been defined in derived class");
      return INT_MAX;
    }
//  renumber[shape function index = lattice index] = dof index
//  called from ShapeFunction::shapeFunctionsOrder for non-hierarchical sf
    virtual void reorderEquallySpacedLatticePoints(int,
    NumPtr<int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual void interiorLatticePointIndices(int,NumPtr<int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual void computeInteriorLatticePoints(int,int,
    NumPtr<Point<SPACEDIM> >&) const {
      OBSOLETE("should have been defined in derived class");
    }
//  called from MGTransferPrebuilt::buildMatrices ?
//  for any child c
//    for any vertex number v
//      vertex v of returned child shape must be the same as
//      parentCoordinateOfCanonical(canonicalVertex(v),c)
    virtual const Shape* reorientFor(const Shape*,int,
    const NumPtr<int>&) const {
      OBSOLETE("should have been defined in derived class");
      return 0;
    }
    virtual int numberDimensions() const {
      OBSOLETE("should have been defined in derived class");
      return INT_MAX;
    }
    Point<SPACEDIM> center(const NumPtr<Point<SPACEDIM> >&) const;
    virtual Point<SPACEDIM> canonicalCenter() const {
      OBSOLETE("should have been defined in derived class");
      return Point<SPACEDIM>();
    }
    virtual REAL diameter(const NumPtr<Point<SPACEDIM> >&) const {
      OBSOLETE("should have been defined in derived class");
      return DBL_MAX;
    }
    virtual REAL volume(const NumPtr<Point<SPACEDIM> >&) const {
      OBSOLETE("should have been defined in derived class");
      return DBL_MAX;
    }
    virtual void initializeMappingSupportPoints(int,
    const NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual void getIntermediatePoints(const NumPtr<Point<SPACEDIM> >&,
    NumPtr<Point<SPACEDIM> >&) const {
//    the 3D classes have not defined this function yet: not needed
      OBSOLETE("should have been defined in derived class");
    }
#if (SPACEDIM==2)
    virtual void draw(const NumPtr<Point<2> >&,AreaGraphTool&) const {
      OBSOLETE("should have been defined in derived class");
    }
#endif
#if (SPACEDIM==3)
    virtual void draw(const VolGraphTool::WINDOW_TYPE::COLOR_TYPE&,
    const NumPtr<Point<3> >&,VolGraphTool&) const {
      OBSOLETE("should have been defined in derived class");
    }
#endif
};

//compare to DEAL GeometryInfo<1>
class Shape1 : public Shape {
  private:
  public:
    inline Shape1(int edi,int nv): Shape(edi,nv) {}
    virtual inline ~Shape1() { }
    inline static int childrenPerShape() { return 2; }
    inline static int childrenPerCodimension1Shape() { return 1; }

    int faceWithIndexEqualTo(int) const;

    virtual int numberDimensions() const { return 1; }
    virtual bool containsCanonical(const Point<1>&) const {
      OBSOLETE("should have been defined in derived class");
      return false;
    }
    virtual Point<SPACEDIM> faceCenter(
    const NumPtr<Point<SPACEDIM> > &vertices,int f) const {
      return vertices[vertexIndex(f)];
    }
    virtual Point<SPACEDIM> interiorChildPoint(
    const NumPtr<Point<SPACEDIM> >&) const {
      OBSOLETE("this shape has no interior child vertex");
    }
    virtual int elementLineFromFace(int,int) const {
      OBSOLETE("should have been defined in derived class");
      return INT_MAX;
    }
    virtual Shape1* newChildShape(int,int,const NumPtr<int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual int equallySpacedLatticePointIndexOfVertex(int,int) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual bool isParentEquallySpacedLatticePoint(int,int,int,int&) 
    const {
      OBSOLETE("should have been defined in derived class");
    }
//  called from FE_L::constraints
//    renumber[dof number] = shape function number
    virtual void reorderChildEquallySpacedLatticePoints(int,const Shape*,
    int,const NumPtr<int>&,NumPtr<int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
//  called from FE_H::constraints
//    renumber[dof number] = shape function number
    virtual void reorderChildDofPoints(int,const Shape*,int,
    const NumPtr<int>&,NumPtr<int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual void printOn(ostream &os) const;
};

#if (SPACEDIM>1)
//compare to DEAL GeometryInfo<2>
class Shape2 : public Shape1 {
  protected:
    NumPtr<int> codimension1_element_indices;
  private:
    virtual Point<SPACEDIM> interiorChildPoint(
    const NumPtr<Point<SPACEDIM> >&) const {
      OBSOLETE("this routine applies to Shape1 only");
    }
    virtual Point<SPACEDIM> interiorChildPoint(
    const NumPtr<Point<SPACEDIM> >&,const NumPtr<int>&) const {
      OBSOLETE("this shape has no interior child vertex");
    }
    virtual Shape1* newChildShape(int,int,const NumPtr<int>&) const {
      OBSOLETE("this routine applies to Shape1 only");
    }
  public:
    inline Shape2(int edi,int nf,int nv):
      Shape1(edi,nv),codimension1_element_indices(nf) {}
    virtual inline ~Shape2() { }
    inline int numberCodimension1Shapes() const {
      return codimension1_element_indices.getNumber();
    }
    inline void setCodimension1ShapeIndex(int f,int index) {
      codimension1_element_indices[f]=index;
    }
    inline static int childrenPerShape() { return 4; }
    inline static int childrenPerCodimension1Shape() { return 2; }

    virtual int numberDimensions() const { return 2; }
    virtual int codimension1ShapeIndex(int f) const {
      return codimension1_element_indices[f];
    }
    virtual bool lineTangentPointsBackward(int l) const {
      return vertexIndex(elementVertexFromLine(l,0))
            >vertexIndex(elementVertexFromLine(l,1));
    }
    virtual bool containsCanonical(const Point<1>&) const {
      OBSOLETE("should have been defined in derived class");
      return false;
    }
    virtual bool containsCanonical(const Point<2>&) const {
      OBSOLETE("should have been defined in derived class");
      return false;
    }
    virtual int numberVerticesOnFace(int) const {
      return 2;
    }
    virtual void interiorChildVertexOnFace(int,int&,int&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual int numberInteriorChildVerticesOnFace(int) const {
      return 0;
    }
//  no re-ordering of vertices:
    virtual Shape1* newLineElement(int) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual Shape2* newChildShape(int,int,const NumPtr<int>&,
    const NumPtr<int>&,const map<pair<int,int>,int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual Shape1* newInteriorCodimension1Shape(int,int,
    const NumPtr<int>&,const NumPtr<int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
//  see FiniteElement::setBoundaryValues calling
//    ShapeFunction::nonzeroShapesOnFace
//  have to worry about orientation of codimension 1 Shapes
    virtual void equallySpacedLatticePointIndicesOfFace(int,int,
    NumPtr<int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
//  see Mapping*::InternalData::computeMappingSupportPoints;
//  have to worry about orientation of codimension 1 Shapes
    virtual void equallySpacedLatticePointIndicesOfFaceInterior(
    int,int,NumPtr<int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual void projectToFace(int,const NumPtr<Point<1> >&,
    NumPtr<Point<2> >&) const {
      OBSOLETE("should have been defined in derived class");
    }
#if (SPACEDIM==3)
    virtual void getMappedPoints(const NumPtr<Point<3> >&,
    const NumPtr<Point<2> >&,NumPtr<Point<3> >&,NumPtr<Point<3> >&,
    NumPtr<Point<3> >&) const {
//    the 3D classes have not defined this function yet: not needed
      OBSOLETE("should have been defined in derived class");
    }
#endif
    virtual void printOn(ostream &os) const;
# ifdef DEBUG
    virtual void checkFaceVertices(int,const Shape1*) const {
      OBSOLETE("should have been defined in derived class");
    }
# endif
};
#endif

#if (SPACEDIM==3)
//compare to DEAL GeometryInfo<3>
class Shape3 : public Shape2 {
  protected:
    NumPtr<int> codimension2_element_indices;
  private:
    virtual Point<SPACEDIM> interiorChildPoint(
    const NumPtr<Point<SPACEDIM> >&) const {
      OBSOLETE("this routine applies to Shape1 only");
    }
    virtual Point<SPACEDIM> interiorChildPoint(
    const NumPtr<Point<SPACEDIM> >&,const NumPtr<int>&) const {
      OBSOLETE("this shape has no interior child vertex");
    }
    virtual Point<SPACEDIM> interiorChildPoint(
    const NumPtr<Point<SPACEDIM> >&,const NumPtr<int>&,
    const NumPtr<int>&) const {
      OBSOLETE("this shape has no interior child vertex");
    }
    virtual Shape1* newChildShape(int,int,const NumPtr<int>&) 
    const {
      OBSOLETE("this routine applies to Shape1 only");
    }
    virtual Shape2* newChildShape(int,int,const NumPtr<int>&,
    const NumPtr<int>&,const map<pair<int,int>,int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual Shape1* newInteriorCodimension1Shape(int,int,
    const NumPtr<int>&,const NumPtr<int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
#ifdef DEBUG
    virtual void checkFaceVertices(int,const Shape1*) const {
      OBSOLETE("this routine applies to Shape2 only");
    }
#endif
  public:
    inline Shape3(int edi,int nl,int nf,int nv): Shape2(edi,nf,nv),
      codimension2_element_indices(nl) {}
    virtual inline ~Shape3() { }
    virtual Shape2* newFaceElement(int) const {
      OBSOLETE("should have been defined in derived class");
    }
    inline int numberCodimension2Shapes() const {
      return codimension2_element_indices.getNumber();
    }
    inline int codimension2ShapeIndex(int l) const {
      return codimension2_element_indices[l];
    }
    inline void setCodimension2ShapeIndex(int l,int index) {
      codimension2_element_indices[l]=index;
    }
    inline static int childrenPerShape() { return 8; }
    inline static int childrenPerCodimension1Shape() { return 4; }

    int sharedLineElementIndex(const Shape3*) const;
    int lineWithIndexEqualTo(int) const;
    virtual int lineFromFace(const Shape2*,int,int) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual void facesFromLine(int,int&,int&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual Tensor<1,SPACEDIM> faceTangent(int f,int n) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual bool faceTangentsWindBackward(int f) const {
      OBSOLETE("should have been defined in derived class");
    }

    virtual int numberDimensions() const { return 3; }
    virtual bool containsCanonical(const Point<1>&) const {
      OBSOLETE("should have been defined in derived class");
      return false;
    }
    virtual bool containsCanonical(const Point<2>&) const {
      OBSOLETE("should have been defined in derived class");
      return false;
    }
    virtual bool containsCanonical(const Point<3>&) const {
      OBSOLETE("should have been defined in derived class");
      return false;
    }
    virtual int numberInteriorChildCodimension2Shapes() const {
      return 0;
    }
    int numberInteriorChildVerticesOnLine(int) const {
      return 1;
    }
    virtual void interiorChildCodimension2ShapeNeighbors(int,int&,int&) 
    const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual void equallySpacedLatticePointIndicesOfLineInterior(
    int,int,NumPtr<int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual Shape1* newInteriorCodimension2Shape(int,int,
    const NumPtr<int>&,const NumPtr<int>&,const NumPtr<int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual Shape2* newInteriorCodimension1Shape(int,int,
    const NumPtr<int>&,const NumPtr<int>&,const NumPtr<int>&,
    const map<pair<int,int>,int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual Shape3* newChildShape(int,int,const NumPtr<int>&,
    const NumPtr<int>&,const NumPtr<int>&,
    const map<pair<pair<int,int>,pair<int,int> >,int>&,
    const map<pair<int,int>,int>&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual void projectToFace(int f,
    const NumPtr<Point<1> > &fp,NumPtr<Point<2> > &ep) const {
      Shape2::projectToFace(f,fp,ep);
    }
    virtual void projectToFace(int,const NumPtr<Point<2> >&,
    NumPtr<Point<3> >&) const {
      OBSOLETE("should have been defined in derived class");
    }
    virtual void printOn(ostream &os) const;
#ifdef DEBUG
    virtual void checkFaceVertices(int,const Shape2*) const=0;
#endif
};
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// The layout of an interval is
//   0 -->-- 1
// The vertices are
//   v0 = Point(0.)
//   v1 = Point(1.)
// The vertices are the faces.
// When refined, the layout is
//   *-0-*-1-*
// The children are
//   I0 = ( 0. , 0.5 )
//   I1 = ( 0.5, 1.  )
//compare to DEAL GeometryInfo<1>
class Interval : public Shape1 {
  private:
    static const Point<1> canonical_vertices[2];
    static const Point<SPACEDIM> canonical_center;
    static const int child_neighbors[2];

  public: 
//  i0,i1 are indices of tessellation vertices, taken from CellData
    Interval(int,const int v[2]);
    Interval(const int v[2],int); // use sparingly; avoids vertex reorder
    virtual ~Interval() {}

    inline static int verticesPerShape() { return 2; }
    inline static int facesPerShape() { return 2; }
    inline static int linesPerShape() { return 1; }
    inline static Point<1> parentCoordinateOfCanonical(
    const Point<1> &child,int c) {
      CHECK_BOUNDS(c,0,2);
      return (c==1 ? Point<1>((child[0]+1.)*0.5) : child*0.5);
    }
    inline static Point<1> canonicalVertex(int v) {
      CHECK_BOUNDS(v,0,2);
      return canonical_vertices[v];
    }
    inline static int staticElementVertexFromLine(int line,int vertex) {
      CHECK_BOUNDS(vertex,0,2);
      return vertex;
    }
    inline static int staticElementLineFromFace(int f,int l) {
      return 0;
    }
    inline static int staticElementVertexFromFace(int f,int v) {
      return f;
    }
    inline static int staticChildOnFace(int f,int) {
      return f;
    }
    inline static int staticNumberInteriorChildVertices() {
      return 1;
    }
    inline static void staticInteriorChildCodimension1ShapeNeighbors(
    int i,int &c0,int &c1) {
      CHECK_BOUNDS(i,0,1);
      c0=child_neighbors[0];
      c1=child_neighbors[1];
    }
    inline static int staticNumberEquallySpacedLatticePoints(int n) {
      return n+1;
    }
    static void staticInteriorLatticePointIndices(int,NumPtr<int>&);
    inline static int staticEquallySpacedLatticePointIndexOfVertex(int qv,
    int degree) {
      return (qv==0 ? 0 : degree);
    }
    inline static Point<SPACEDIM> staticCenter() { 
      return canonical_center; 
    }
    static void staticEquallySpacedLatticePoints(int,NumPtr<Point<1> >&);
    static int staticNumberInteriorEquallySpacedLatticePoints(int n) {
      return n-1;
    }

    virtual bool contains(const Point<SPACEDIM>&,
      const NumPtr<Point<SPACEDIM> >&) const;
    virtual bool containsCanonical(const Point<1> &p) const {
      return (p[0] >= 0.) && (p[0] <= 1.); 
    }
    virtual int numberInteriorChildVertices() const {return 1;}
    virtual void interiorChildVertex(int &c,int &v) const {
      c=0; v=1;
    }
    virtual Point<SPACEDIM> canonicalCenter() const { 
      return canonical_center; 
    }
    virtual REAL diameter(const NumPtr<Point<SPACEDIM> >&) const;
    virtual REAL volume(const NumPtr<Point<SPACEDIM> >&) const;
    virtual Shape1* newChildShape(int c,int child_element_index,
    const NumPtr<int> &interior_child_vertex_indices) const {
      if (c==0) {
        int v[2]={vertexIndex(0),interior_child_vertex_indices[0]};
        return OPERATOR_NEW Interval(child_element_index,v);
      }
      int v[2]={interior_child_vertex_indices[0],vertexIndex(1)};
      return OPERATOR_NEW Interval(child_element_index,v);
    }

    Point<SPACEDIM> parentCoordinateForCanonical(const Point<SPACEDIM>&,
      const Shape*,int,const NumPtr<int>&) const;
    Point<SPACEDIM> childCoordinateOfCanonical(const Point<SPACEDIM>&,
      const Shape*,int,const NumPtr<int>&) const;
    virtual void printOn(ostream &os) const;
    virtual int numberEquallySpacedLatticePoints(int n) const {
      return staticNumberEquallySpacedLatticePoints(n);
    }
    void equallySpacedLatticePoints(int n,
    NumPtr<Point<SPACEDIM> > &points) const {
#if (SPACEDIM==1)
      staticEquallySpacedLatticePoints(n,points);
#else
      OBSOLETE("wrong number of dimensions");
#endif
    }
    virtual int equallySpacedLatticePointIndexOfVertex(int qv,
    int degree) const {
      return staticEquallySpacedLatticePointIndexOfVertex(qv,degree);
    }
    void reorderDofs(int,NumPtr<int>&) const;
    virtual void reorderEquallySpacedLatticePoints(int,
      NumPtr<int>&) const;
    virtual void reorderChildEquallySpacedLatticePoints(int,
      const Shape*,int,const NumPtr<int>&,NumPtr<int>&) const;
    virtual void reorderChildDofPoints(int,const Shape*,int,
      const NumPtr<int>&,NumPtr<int>&) const;
    virtual int numberInteriorEquallySpacedLatticePoints(int n) const {
      return staticNumberInteriorEquallySpacedLatticePoints(n);
    }
    virtual void interiorLatticePointIndices(int n,NumPtr<int> &v) const {
      staticInteriorLatticePointIndices(n,v);
    }
    virtual void computeInteriorLatticePoints(int d,int f,
      NumPtr<Point<SPACEDIM> > &msp) const;
    virtual const Shape* reorientFor(const Shape*,int,
      const NumPtr<int>&) const;
    virtual void initializeMappingSupportPoints(int,
      const NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&) const;
    virtual void getIntermediatePoints(const NumPtr<Point<SPACEDIM> >&,
      NumPtr<Point<SPACEDIM> >&) const;
    virtual void getMappedPoints(const NumPtr<Point<SPACEDIM> >&,
      const NumPtr<Point<1> >&,NumPtr<Point<SPACEDIM> >&,
      NumPtr<Point<SPACEDIM> >&) const;
    virtual int numberInteriorChildCodimension1Shapes() const {
      return 1;
    }
    virtual int elementVertexFromFace(int f,int v) const {
      return staticElementVertexFromFace(f,v);
    }
    virtual int elementVertexFromLine(int f,int l) const {
      return staticElementVertexFromLine(f,l);
    }
    virtual int elementLineFromFace(int f,int fl) const {
      return staticElementLineFromFace(f,fl);
    }
    virtual void interiorChildCodimension1ShapeNeighbors(int i,int &c0,
    int &c1) const {
      staticInteriorChildCodimension1ShapeNeighbors(i,c0,c1);
    }
    virtual int childOnFace(int f,int s) const {
      return staticChildOnFace(f,s);
    }
#ifdef DEBUG
# if (SPACEDIM==2)
    virtual void draw(const NumPtr<Point<2> >&,AreaGraphTool&) const;
# endif
# if (SPACEDIM==3)
    virtual void draw(const VolGraphTool::WINDOW_TYPE::COLOR_TYPE&,
      const NumPtr<Point<3> >&,VolGraphTool&) const;
# endif
#endif
    Point<SPACEDIM> interiorChildPoint(
    const NumPtr<Point<SPACEDIM> > &vertices) const {
      return (vertices[vertexIndex(0)]+vertices[vertexIndex(1)])*0.5;
    }
};
#if (SPACEDIM>1)
bool findOtherVertexOnLine(const Interval*,int,/*int&,*/int&);
#endif

#ifdef MAKE_LINE_PAIR
#undef MAKE_LINE_PAIR
#endif
#define MAKE_LINE_PAIR(v0,v1) make_pair(min(v0,v1),max(v0,v1))

#if (SPACEDIM>1)
//The layout of a triangle is
//    2
//    .\
//  0 :  ^  2
//    |    \
//    0-->--1
//       1
//The vertices are
//  v0 = Point( 0., 0. )
//  v1 = Point( 1., 0. )
//  v2 = Point( 0., 1. )
//The lines are
//  l0 = v2 --> v0
//  l1 = v0 --> v1
//  l2 = v1 --> v2
//The lines are the faces.
//The midpoints of the faces are
//  v3 = Point( 0., .5 )
//  v4 = Point( .5, 0. )
//  v5 = Point( .5, .5 )
//The interior child lines are
//  cl0 = ( v5, v3 )
//  cl1 = ( v3, v4 )
//  cl2 = ( v4, v5 )
//When refined, the layout is
//    @
//    |\
//    | 2\
//    @---@\
//    |\ 3| \
//    | 0\| 1\
//    @---@---@
// The children are
//   T0 = ( v0, v4, v3 )
//   T1 = ( v1, v5, v4 )
//   T2 = ( v2, v3, v5 )
//   T3 = ( v3, v4, v5 )
//The child cells on face are ordered in the direction of the face

//not in DEAL
class Triangle : public Shape2 {
  private:
    static const Point<2> canonical_vertices[3];
    static const Point<SPACEDIM> canonical_center;
    static const int child_neighbors[3][2];

    static const int interior_child_vertices[3][2];
    static const int canonical_vertex_indices[3][2];
    static const int opposite_face[3];
    static const int children[3][2];
    static const Tensor<1,SPACEDIM> face_normals[3];
    static const Tensor<1,SPACEDIM> line_tangents[3];

    virtual Shape1* newChildShape(int,int,const NumPtr<int>&) const {
      OBSOLETE("this routine applies to Shape1 only");
    }
    virtual bool containsCanonical(const Point<1>&) const {
      OBSOLETE("should not be calling this");
      return false;
    }
  public:
    Triangle(int,const int l[3],const int v[3]);
//  use sparingly: avoids vertex reorder
    Triangle(const int l[3],const int v[3],int); 
    virtual ~Triangle() { }
    virtual Shape1* newLineElement(int) const;

    inline static int verticesPerShape() { return 3; }
    inline static int facesPerShape() { return 3; }
    inline static int linesPerShape() { return 3; }
    inline static int linesPerFace(int) { return 1; }
    inline static int staticNumberVerticesOnFace(int) {
      return 2;
    }
    inline static Point<2> parentCoordinateOfCanonical(
    const Point<2> &p,int c) {
      CHECK_BOUNDS(c,0,4);
      switch (c) {
        case 1: return Point<2>(1.-(p[0]+p[1])*0.5,     p[0]      *0.5);
        case 2: return Point<2>(    p[1]      *0.5, 1.-(p[0]+p[1])*0.5);
        case 3: return Point<2>(   (p[0]+p[1])*0.5,(1.- p[0]     )*0.5);
        default: return p*0.5;
      }
    }
    inline static Point<2> canonicalVertex(int v) {
      CHECK_BOUNDS(v,0,3);
      return canonical_vertices[v];
    }
    inline static int staticElementVertexFromLine(int line,int vertex) {
      CHECK_BOUNDS(line,0,3);
      CHECK_BOUNDS(vertex,0,2);
      return canonical_vertex_indices[line][vertex];
    }
    inline static int staticElementLineFromFace(int f,int) {
      return f;
    }
    inline static int staticElementVertexFromFace(int f,int v) {
      return canonical_vertex_indices[f][v];
    }
    inline static int staticChildOnFace(int f,int s) {
      CHECK_BOUNDS(f,0,3);
      CHECK_BOUNDS(s,0,2);
      return children[f][s];
    }
    inline static int staticNumberInteriorChildVerticesOnFace(int) {
      return 1;
    }
    inline static int staticNumberInteriorChildCodimension1Shapes() {
      return 3;
    }
    inline static void staticInteriorChildCodimension1ShapeNeighbors(
    int i,int &c0,int &c1) {
      CHECK_BOUNDS(i,0,3);
      c0=child_neighbors[i][0];
      c1=child_neighbors[i][1];
    }
    inline static int staticNumberEquallySpacedLatticePoints(int n) {
      return ((n+1)*(n+2))/2;
    }
    inline static Point<SPACEDIM> staticCenter() { 
      return canonical_center; 
    }
//  unit outer normal * face length / canonical interval length
    inline static Tensor<1,SPACEDIM> faceNormal(int f) {
      return face_normals[f];
    }
    inline static Tensor<1,SPACEDIM> lineTangent(int l) {
      return line_tangents[l];
    }

    static void staticEquallySpacedLatticePoints(int,NumPtr<Point<2> >&);
    static void staticInteriorChildVertexOnFace(int f,int &c,int &v) {
      CHECK_BOUNDS(f,0,3);
      c=interior_child_vertices[f][0];
      v=interior_child_vertices[f][1];
    }
    static void staticProjectToFace(int,const NumPtr<Point<1> >&,
      NumPtr<Point<2> >&);
    static void projectToSubface(int,int,
      const NumPtr<Point<1> >&,NumPtr<Point<2> >&);
    static void staticInteriorLatticePointIndices(int,NumPtr<int>&);
    static int staticEquallySpacedLatticePointIndexOfVertex(int,int);
    static void staticEquallySpacedLatticePointIndicesOfFace(int,int,
      NumPtr<int>&);
    static void staticEquallySpacedLatticePointIndicesOfFaceInterior(
      int,int,NumPtr<int>&);
    static int staticNumberInteriorEquallySpacedLatticePoints(int n) {
      return ((n-1)*(n-2))/2;
    }

    virtual bool contains(const Point<SPACEDIM>&,
      const NumPtr<Point<SPACEDIM> >&) const;
    virtual bool containsCanonical(const Point<2> &p) const {
      return (p[0] >= 0.) && (p[1] >= 0.) && (p[0]+p[1] <= 1.);
    }
    virtual Point<SPACEDIM> canonicalCenter() const { 
      return canonical_center; 
    }
    virtual REAL diameter(const NumPtr<Point<SPACEDIM> >&) const;
    virtual REAL volume(const NumPtr<Point<SPACEDIM> >&) const;
    virtual bool lineTangentPointsBackward(int) const;
    Shape2* newChildShape(int,int,const NumPtr<int>&,
      const NumPtr<int>&,const map<pair<int,int>,int>&) const;
    Point<SPACEDIM> parentCoordinateForCanonical(const Point<SPACEDIM>&,
      const Shape*,int,const NumPtr<int>&) const;
    Point<SPACEDIM> childCoordinateOfCanonical(const Point<SPACEDIM>&,
      const Shape*,int,const NumPtr<int>&) const;
    virtual void printOn(ostream &os) const;
    virtual int numberEquallySpacedLatticePoints(int n) const {
      return staticNumberEquallySpacedLatticePoints(n);
    }
    void equallySpacedLatticePoints(int n,
    NumPtr<Point<SPACEDIM> > &points) const {
#if (SPACEDIM==2)
      staticEquallySpacedLatticePoints(n,points);
#else
      OBSOLETE("wrong number of dimensions");
#endif
    }
    virtual int numberInteriorEquallySpacedLatticePoints(int n) const {
      return staticNumberInteriorEquallySpacedLatticePoints(n);
    }
    void reorderFaceDofs(int,NumPtr<int>&) const;
    void reorderDofs(int,NumPtr<int>&) const;
    virtual void reorderEquallySpacedLatticePoints(int,NumPtr<int>&)
      const;
    void reorderChildEquallySpacedLatticePoints(int,const Shape*,int,
      const NumPtr<int>&,NumPtr<int>&) const;
    virtual void reorderChildDofPoints(int,const Shape*,int,
      const NumPtr<int>&,NumPtr<int>&) const;
    virtual void interiorLatticePointIndices(int n,NumPtr<int> &v) const {
      staticInteriorLatticePointIndices(n,v);
    }
    virtual void computeInteriorLatticePoints(int d,int f,
      NumPtr<Point<SPACEDIM> > &msp) const;
#if (SPACEDIM==3)
    void lenoirCanonicalFacePoints(int,const NumPtr<Point<2> >&,
      NumPtr<Point<1> >&) const;
    void lenoirInterpolationForFace(int,int,
      const NumPtr<Point<2> >&,const NumPtr<Point<3> >&,
      const NumPtr<Point<3> >&,NumPtr<Point<3> >&) const;
#endif
    virtual const Shape* reorientFor(const Shape*,int,
      const NumPtr<int>&) const;
    virtual void initializeMappingSupportPoints(int,
      const NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&) const;
    virtual void getIntermediatePoints(const NumPtr<Point<SPACEDIM> >&,
      NumPtr<Point<SPACEDIM> >&) const;
# if (SPACEDIM==3)
    virtual void getMappedPoints(const NumPtr<Point<3> >&,
      const NumPtr<Point<2> >&,NumPtr<Point<3> >&,NumPtr<Point<3> >&,
      NumPtr<Point<3> >&) const;
# endif
    virtual Point<SPACEDIM> faceCenter(const NumPtr<Point<SPACEDIM> >&,
      int) const;
    virtual int numberInteriorChildCodimension1Shapes() const {
      return 3;
    }
    virtual int elementVertexFromFace(int f,int v) const {
      return staticElementVertexFromFace(f,v);
    }
    virtual int elementVertexFromLine(int f,int l) const {
      return staticElementVertexFromLine(f,l);
    }
    virtual int elementLineFromFace(int f,int fl) const {
      return staticElementLineFromFace(f,fl);
    }
    virtual void interiorChildCodimension1ShapeNeighbors(int i,
    int &c0,int &c1) const {
      staticInteriorChildCodimension1ShapeNeighbors(i,c0,c1);
    }
    virtual int childOnFace(int f,int s) const {
      return staticChildOnFace(f,s);
    }

    virtual int numberInteriorChildVerticesOnFace(int f) const {
      return staticNumberInteriorChildVerticesOnFace(f);
    }
    virtual int equallySpacedLatticePointIndexOfVertex(int qv,int degree) 
    const {
      return staticEquallySpacedLatticePointIndexOfVertex(qv,degree);
    }
    virtual void equallySpacedLatticePointIndicesOfFace(int,int,
      NumPtr<int>&) const;
    virtual void equallySpacedLatticePointIndicesOfFaceInterior(
      int,int,NumPtr<int>&) const;
    virtual void projectToFace(int f,const NumPtr<Point<1> > &fp,
    NumPtr<Point<2> > &ep) const {
      staticProjectToFace(f,fp,ep);
    }
#ifdef DEBUG
    virtual void checkFaceVertices(int,const Shape1*) const;
# if (SPACEDIM==2)
    virtual void draw(const NumPtr<Point<2> >&,AreaGraphTool&) const;
# endif
# if (SPACEDIM==3)
    void draw(const VolGraphTool::WINDOW_TYPE::COLOR_TYPE&,
      const NumPtr<Point<3> >&,VolGraphTool&) const;
# endif
#endif

    virtual void interiorChildVertexOnFace(int f,int &c,int &v) const {
      staticInteriorChildVertexOnFace(f,c,v);
    }
    virtual Shape1* newInteriorCodimension1Shape(int i,
    int codimension1_element_index,
    const NumPtr<int> &interior_child_vertex_indices,
    const NumPtr<int> &interior_child_line_vertex_indices) const {
      int v[2]={interior_child_line_vertex_indices[(i+2)%3],
                interior_child_line_vertex_indices[i]};
      return OPERATOR_NEW Interval(codimension1_element_index,v);
    }
};

//Layout of a quadrilateral is
//      3
//   2-->--3
//   |     |
// 0 ^     ^ 1
//   |     |
//   0-->--1
//      2
//The vertices are
//  v0 = Point( 0., 0. )
//  v1 = Point( 1., 0. )
//  v2 = Point( 0., 1. )
//  v3 = Point( 1., 1. )
//The lines are
//  l0 = v0 --> v2
//  l1 = v1 --> v3
//  l2 = v0 --> v1
//  l3 = v2 --> v3
//The lines are the faces.
//The midpoints of the faces are
//  v4 = Point( 0., .5 )
//  v5 = Point( 1., .5 )
//  v6 = Point( .5, 0. )
//  v7 = Point( .5, 1. )
//The interior child point is
//  v8 = Point( .5, .5 )
//The interior child lines are
//  cl0 = ( v4, v8 )
//  cl1 = ( v5, v8 )
//  cl2 = ( v6, v8 )
//  cl3 = ( v7, v8 )
//When refined, the layout is
//  *--*--*
//  | 2|3 |
//  *--*--*
//  | 0|1 |
//  *--*--*
// The children are
//   Q0 = ( v0, v6, v4, v8 )
//   Q1 = ( v1, v5, v6, v8 )
//   Q2 = ( v2, v4, v7, v8 )
//   Q3 = ( v3, v7, v5, v8 )
//compare to DEAL GeometryInfo<2>
class Quadrilateral : public Shape2 {
  private:
    static const Point<2> canonical_vertices[4];
    static const Point<SPACEDIM> canonical_center;
    static const int child_neighbors[4][2];

    static const int interior_child_vertices[4][2];
    static const int opposite_vertex[4];
    static const int canonical_vertex_indices[4][2];
    static const int opposite_face[4];
    static const int children[4][2];
    static const Tensor<1,SPACEDIM> face_normals[4];
    static const Tensor<1,SPACEDIM> line_tangents[4];

    virtual Shape1* newChildShape(int,int,const NumPtr<int>&) const {
      OBSOLETE("this routine applies to Shape1 only");
    }
    virtual Point<SPACEDIM> interiorChildPoint(
    const NumPtr<Point<SPACEDIM> >&) const {
      OBSOLETE("this routine applies to Shape1 only");
    }
    virtual bool containsCanonical(const Point<1>&) const {
      OBSOLETE("should not be calling this");
      return false;
    }
  public:
    Quadrilateral(int,const int l[4],const int v[4]);
//  use sparingly: avoids vertex reorder
    Quadrilateral(const int l[4],const int v[4],int);
    virtual ~Quadrilateral() { }
    virtual Shape1* newLineElement(int) const;

    inline static int verticesPerShape() {return 4;}
    inline static int facesPerShape() { return 4; }
    inline static int linesPerShape() { return 4; }
    inline static int linesPerFace(int) {
      return 1;
    }
    inline static int staticNumberVerticesOnFace(int) {
      return 2;
    }
    inline static Point<2> parentCoordinateOfCanonical(
    const Point<2> &p,int c) {
      CHECK_BOUNDS(c,0,4);
      switch (c) {
        case 1: return Point<2>(1.-p[1]*0.5,   p[0]*0.5 );
        case 2: return Point<2>(   p[1]*0.5,1.-p[0]*0.5 );
        case 3: return Point<2>(1.-p[0]*0.5,1.-p[1]*0.5 );
        default: return p*0.5;
      }
    }
    inline static int oppositeVertex(int v) {
      CHECK_BOUNDS(v,0,4);
      return opposite_vertex[v];
    }
    inline static Point<2> canonicalVertex(int v) {
      CHECK_BOUNDS(v,0,4);
      return canonical_vertices[v];
    }
    inline static int staticElementVertexFromLine(int line,int vertex) {
      CHECK_BOUNDS(line,0,4);
      CHECK_BOUNDS(vertex,0,2);
      return canonical_vertex_indices[line][vertex];
    }
    inline static int staticElementLineFromFace(int f,int) {
      return f;
    }
    inline static int staticElementVertexFromFace(int f,int v) {
      return canonical_vertex_indices[f][v];
    }
    inline static int staticChildOnFace(int f,int s) {
      CHECK_BOUNDS(f,0,4);
      CHECK_BOUNDS(s,0,2);
      return children[f][s];
    }
    inline static int staticNumberInteriorChildVerticesOnFace(int) {
      return 1;
    }
    inline static int staticNumberInteriorChildCodimension1Shapes() {
      return 4;
    }
    inline static void staticInteriorChildCodimension1ShapeNeighbors(
    int i,int &c0,int &c1) {
      CHECK_BOUNDS(i,0,4);
      c0=child_neighbors[i][0];
      c1=child_neighbors[i][1];
    }
    inline static int staticNumberEquallySpacedLatticePoints(int n0,
    int n1) {
      return (n0+1)*(n1+1);
    }
    inline static int staticNumberEquallySpacedLatticePoints(int n) {
      return staticNumberEquallySpacedLatticePoints(n,n);
    }
    inline static Point<SPACEDIM> staticCenter() { 
      return canonical_center; 
    }
//  unit outer normal * face length / canonical interval length
    inline static Tensor<1,SPACEDIM> faceNormal(int f) {
      return face_normals[f];
    }
    inline static Tensor<1,SPACEDIM> lineTangent(int l) {
      return line_tangents[l];
    }
    inline static int staticNumberInteriorChildVertices() {
      return 1;
    }
    inline static void staticInteriorChildVertex(int,int &c,int &v) {
      c=0;
      v=3;
    }

    static void staticEquallySpacedLatticePoints(int,int,
      NumPtr<Point<2> >&);
    static void staticEquallySpacedLatticePoints(int n,
    NumPtr<Point<2> > & points) {
      staticEquallySpacedLatticePoints(n,n,points);
    }
    static void staticInteriorChildVertexOnFace(int f,int &c, int &v) {
      CHECK_BOUNDS(f,0,4);
      c=interior_child_vertices[f][0]; 
      v=interior_child_vertices[f][1]; 
    }
    static void staticProjectToFace(int,const NumPtr<Point<1> >&,
      NumPtr<Point<2> >&);
    static void projectToSubface(int,int,
      const NumPtr<Point<1> >&,NumPtr<Point<2> >&);
    static void staticInteriorLatticePointIndices(int,int,NumPtr<int>&);
    static void staticInteriorLatticePointIndices(int n,NumPtr<int> &v) {
      staticInteriorLatticePointIndices(n,n,v);
    }
    static int staticEquallySpacedLatticePointIndexOfVertex(int,int,int);
    static int staticEquallySpacedLatticePointIndexOfVertex(int v,int n) {
      staticEquallySpacedLatticePointIndexOfVertex(v,n,n);
    }
    static void staticEquallySpacedLatticePointIndicesOfFace(int,int,int,
      NumPtr<int>&);
    static void staticEquallySpacedLatticePointIndicesOfFace(int f,int n,
    NumPtr<int> &indices) {
      staticEquallySpacedLatticePointIndicesOfFace(f,n,n,indices);
    }
    static void staticEquallySpacedLatticePointIndicesOfFaceInterior(
      int,int,int,NumPtr<int>&);
    static void staticEquallySpacedLatticePointIndicesOfFaceInterior(
    int f,int n,NumPtr<int> &indices) {
      staticEquallySpacedLatticePointIndicesOfFaceInterior(f,n,n,indices);
    }
    static int staticNumberInteriorEquallySpacedLatticePoints(int n0,
    int n1) {
      return (n1-1)*(n0-1);
    }
    static int staticNumberInteriorEquallySpacedLatticePoints(int n) {
      return staticNumberInteriorEquallySpacedLatticePoints(n,n);
    }

    virtual bool contains(const Point<SPACEDIM>&,
      const NumPtr<Point<SPACEDIM> >&) const;
    virtual bool containsCanonical(const Point<2> &p) const {
      return (p[0]>=0.) && (p[0]<=1.) && (p[1]>=0.) && (p[1]<=1.);
    }
    virtual Point<SPACEDIM> canonicalCenter() const { 
      return canonical_center; 
    }
    virtual REAL diameter(const NumPtr<Point<SPACEDIM> >&) const;
    virtual REAL volume(const NumPtr<Point<SPACEDIM> >&) const;
    virtual bool lineTangentPointsBackward(int) const;
    virtual Shape2* newChildShape(int,int,const NumPtr<int>&,
      const NumPtr<int>&,const map<pair<int,int>,int>&) const;
    Point<SPACEDIM> parentCoordinateForCanonical(const Point<SPACEDIM>&,
      const Shape*,int,const NumPtr<int>&) const;
    Point<SPACEDIM> childCoordinateOfCanonical(const Point<SPACEDIM>&,
      const Shape*,int,const NumPtr<int>&) const;
    virtual void printOn(ostream &os) const;
    virtual int numberEquallySpacedLatticePoints(int n) const {
      return staticNumberEquallySpacedLatticePoints(n,n);
    }
    void equallySpacedLatticePoints(int n,
    NumPtr<Point<SPACEDIM> > &points) const {
#if (SPACEDIM==2)
      staticEquallySpacedLatticePoints(n,n,points);
#else
      OBSOLETE("wrong number of dimensions");
#endif
    }
    virtual int numberInteriorEquallySpacedLatticePoints(int n) const {
      return staticNumberInteriorEquallySpacedLatticePoints(n,n);
    }
    void reorderFaceDofs(int,NumPtr<int>&) const;
    void reorderDofs(int,int,NumPtr<int>&) const;
    void reorderDofs(int n,NumPtr<int> &renumber) const {
      reorderDofs(n,n,renumber);
    }
    void reorderEquallySpacedLatticePoints(int,int,NumPtr<int>&) const;
    virtual void reorderEquallySpacedLatticePoints(int n,
    NumPtr<int> &renumber) const {
      reorderEquallySpacedLatticePoints(n,n,renumber);
    }
    void reorderChildEquallySpacedLatticePoints(int,const Shape*,int,
      const NumPtr<int>&,NumPtr<int>&) const;
    virtual void reorderChildDofPoints(int,const Shape*,int,
      const NumPtr<int>&,NumPtr<int>&) const;
    virtual void interiorLatticePointIndices(int n,NumPtr<int> &v) const {
      staticInteriorLatticePointIndices(n,n,v);
    }
    void computeInteriorLatticePoints(int d0,int d1,int bf,
      NumPtr<Point<SPACEDIM> > &msp) const;
    virtual void computeInteriorLatticePoints(int d,int bf,
    NumPtr<Point<SPACEDIM> > &msp) const {
      computeInteriorLatticePoints(d,d,bf,msp);
    }
    virtual const Shape* reorientFor(const Shape*,int,
      const NumPtr<int>&) const;
    virtual void initializeMappingSupportPoints(int,
      const NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&) const;
    virtual void getIntermediatePoints(const NumPtr<Point<SPACEDIM> >&,
      NumPtr<Point<SPACEDIM> >&) const;
# if (SPACEDIM==3)
    virtual void getMappedPoints(const NumPtr<Point<3> >&,
      const NumPtr<Point<2> >&,NumPtr<Point<3> >&,NumPtr<Point<3> >&,
      NumPtr<Point<3> >&) const;
# endif
    virtual Point<SPACEDIM> faceCenter(const NumPtr<Point<SPACEDIM> >&,
      int) const;
    virtual int numberInteriorChildCodimension1Shapes() const{
      return 4;
    }
    virtual int elementVertexFromFace(int f,int v) const {
      return staticElementVertexFromFace(f,v);
    }
    virtual int elementVertexFromLine(int f,int l) const {
      return staticElementVertexFromLine(f,l);
    }
    virtual int elementLineFromFace(int f,int fl) const {
      return staticElementLineFromFace(f,fl);
    }
    virtual void interiorChildCodimension1ShapeNeighbors(int i,int &c0,
    int &c1) const {
      staticInteriorChildCodimension1ShapeNeighbors(i,c0,c1);
    }
    virtual int childOnFace(int f,int s) const {
      return staticChildOnFace(f,s);
    }
    virtual int numberInteriorChildVerticesOnFace(int f) const {
      return staticNumberInteriorChildVerticesOnFace(f);
    }
    virtual int equallySpacedLatticePointIndexOfVertex(int qv,int degree) 
    const {
      return staticEquallySpacedLatticePointIndexOfVertex(qv,degree,
        degree);
    }
    void equallySpacedLatticePointIndicesOfFace(int,int,int,
      NumPtr<int>&) const;
    virtual void equallySpacedLatticePointIndicesOfFace(int f,int n,
    NumPtr<int> &indices) const {
      equallySpacedLatticePointIndicesOfFace(f,n,n,indices);
    }
    void equallySpacedLatticePointIndicesOfFaceInterior(int,int,int,
      NumPtr<int>&) const;
    virtual void equallySpacedLatticePointIndicesOfFaceInterior(
    int f,int n,NumPtr<int> &indices) const{
      equallySpacedLatticePointIndicesOfFaceInterior(f,n,n,indices);
    }
    virtual void projectToFace(int f,const NumPtr<Point<1> > &fp,
    NumPtr<Point<2> > &ep) const {
      staticProjectToFace(f,fp,ep);
    }
    virtual void projectToFaceForBDFM(int,const NumPtr<Point<1> >&,
      NumPtr<Point<2> >&) const;
#ifdef DEBUG
    virtual void checkFaceVertices(int,const Shape1*) const;
#if (SPACEDIM==2)
    virtual void draw(const NumPtr<Point<2> >&,AreaGraphTool&) const;
#endif
#if (SPACEDIM==3)
    void draw(const VolGraphTool::WINDOW_TYPE::COLOR_TYPE&,
      const NumPtr<Point<3> >&,VolGraphTool&) const;
#endif
#endif

    virtual void interiorChildVertexOnFace(int f,int &c,int &v) const {
      staticInteriorChildVertexOnFace(f,c,v);
    }
    virtual Shape1* newInteriorCodimension1Shape(int i,
    int codimension1_element_index,
    const NumPtr<int> &interior_child_vertex_indices,
    const NumPtr<int> &interior_child_line_vertex_indices) const {
      int v[2]={interior_child_line_vertex_indices[i],
                interior_child_vertex_indices[0]};
      return OPERATOR_NEW Interval(codimension1_element_index,v);
    }

    virtual int numberInteriorChildVertices() const {return 1;}
    virtual void interiorChildVertex(int &c,int &v) const {
      c=0;
      v=3;
    }
    virtual Point<SPACEDIM> interiorChildPoint(
    const NumPtr<Point<SPACEDIM> > &vertices,
    const NumPtr<int> &interior_points_of_faces) const {
//    solve 9-point Laplacian for interior child point
      return (vertices[vertexIndex(0)]+vertices[vertexIndex(1)]
             +vertices[vertexIndex(2)]+vertices[vertexIndex(3)]
             +(vertices[interior_points_of_faces[0]]
              +vertices[interior_points_of_faces[1]]
              +vertices[interior_points_of_faces[2]]
              +vertices[interior_points_of_faces[3]])*3.)*0.0625;
    }
};
#endif

#if (SPACEDIM==3)
//The layout of a tetrahedron is
//             3
//           / | \
//         /   ^3  \
//      4      0      5
//     /   0<     <2   \
//   /   .            .  \
// 1----------->-----------2
//             1
//The vertices are
//  v0 = Point( 0., 0., 0. )
//  v1 = Point( 1., 0., 0. )
//  v2 = Point( 0., 1., 0. )
//  v3 = Point( 0., 0., 1. )
//The lines are
//  L0 = v0 --> v1
//  L1 = v1 --> v2
//  L2 = v2 --> v0
//  L3 = v0 --> v3
//  L4 = v3 --> v1
//  L5 = v3 --> v2
//The faces are (right-hand rule for inward normal)
//  T0 = ( v0 , v2 , v3 ) ==> inward normal is first axis vector 
//  T1 = ( v0 , v3 , v1 ) ==> inward normal is second axis vector
//  T2 = ( v0 , v1 , v2 ) ==> inward normal is third axis vector
//  T3 = ( v1 , v3 , v2 ) ==> inward normal is - sum of axis vectors
//For mesh refinement, see
//  Endres, Lance and Krysl, Petr 
//    "Octasection-based Refinement of Finite Shape
//    Approximations on Tetrahedral Meshes that Guarantees Shape 
//    Quality", Int. J. Numer. Meth. Engng. 59 vol. 1 (2003), 69-82
//  hogwarts.ucsd.edu/~pkrysl/publications/tetras.pdf
//Let the line midpoints be
//  v4 = Point( 0.5, 0. , 0. )
//  v5 = Point( 0.5, 0.5, 0. )
//  v6 = Point( 0. , 0.5, 0. )
//  v7 = Point( 0. , 0. , 0.5)
//  v8 = Point( 0.5, 0. , 0.5)
//  v9 = Point( 0. , 0.5, 0.5)
//The interior child line is
//  l0 = v4 --> v9
//The interior child faces are
//  T0 = ( v4 , v7 , v6 )
//  T1 = ( v4 , v5 , v8 )
//  T2 = ( v6 , v9 , v5 )
//  T3 = ( v7 , v8 , v9 )
//  T4 = ( v7 , v4 , v9 )
//  T5 = ( v8 , v9 , v4 )
//  T6 = ( v9 , v4 , v5 )
//  T7 = ( v6 , v9 , v4 )
//In terms of vertices, the children are
//  T0 = ( v0, v4, v6, v7 ) has vertex at vertex 0
//  T1 = ( v4, v1, v5, v8 ) has vertex at vertex 1
//  T2 = ( v6, v5, v2, v9 ) has vertex at vertex 2
//  T3 = ( v7, v8, v9, v3 ) has vertex at vertex 3
//  T4 = ( v8, v7, v9, v4 ) has face on face 1
//  T5 = ( v8, v9, v5, v4 ) has face on face 3
//  T6 = ( v6, v9, v4, v5 ) has face on face 2
//  T7 = ( v9, v6, v4, v7 ) has face on face 0
//or in terms of faces the children are
//  T0 : x    >=0. ,   y  >=0. ,     z>=0. , x+y+z<=0.5
//  T1 : x    > 0.5,   y  >=0. ,     z>=0. , x+y+z<=1.
//  T2 : x    >=0. ,   y  > 0.5,     z>=0. , x+y+z<=1.
//  T3 : x    >=0. ,   y  >=0. ,     z> 0.5, x+y+z<=1.
//  T4 : x+y  <=0.5,   y  >=0. ,     z<=0.5, x  +z> 0.5
//  T5 : x+y  > 0.5, x+y+z<=1. , x    <=0.5, x  +z> 0.5
//  T6 : y    <=0.5, x+y  > 0.5,     z>=0. , x  +z<=0.5
//  T7 : x  +z<=0.5, x    >=0. , x+y  <=0.5, x+y+z>=0.5
//In terms of the transformation c = A p + b (p=parent, c=child coord)
//     |     A    |   b  |     inv(A)     | eig(A)
//     -----------------------------------------------------------------
//  T0:|  2  0  0 |   0  |  1/2   0    0  |
//     |  0  2  0 |   0  |   0   1/2   0  | 2,2,2
//     |  0  0  2 |   0  |   0    0   1/2 |
//     -----------------------------------------------------------------
//  T1:|  2  0  0 |  -1  |  1/2   0    0  |
//     |  0  2  0 |   0  |   0   1/2   0  | 2,2,2
//     |  0  0  2 |   0  |   0    0   1/2 |
//     -----------------------------------------------------------------
//  T2:|  2  0  0 |   0  |  1/2   0    0  |
//     |  0  2  0 |  -1  |   0   1/2   0  | 2,2,2
//     |  0  0  2 |   0  |   0    0   1/2 |
//     -----------------------------------------------------------------
//  T3:|  2  0  0 |   0  |  1/2   0    0  |
//     |  0  2  0 |   0  |   0   1/2   0  | 2,2,2
//     |  0  0  2 |  -1  |   0    0   1/2 |
//     -----------------------------------------------------------------
//  T4:| -2 -2  0 |   1  | -1/2 -1/2   0  |
//     |  0  2  0 |   0  |   0   1/2   0  | -2,-2,2
//     |  0  0 -2 |   1  |   0    0  -1/2 |
//     -----------------------------------------------------------------
//  T5:| -2  0  0 |   1  | -1/2   0    0  |
//     |  2  2  0 |  -1  |  1/2  1/2   0  | -2,-2,2
//     | -2 -2 -2 |   2  |   0  -1/2 -1/2 |
//     -----------------------------------------------------------------
//  T6:|  0  0  2 |   0  |   0   1/2  1/2 |
//     |  0 -2  0 |   1  |   0  -1/2   0  | -2,-2,2
//     |  2  2  0 |  -1  |  1/2   0    0  |
//     -----------------------------------------------------------------
//  T7:| -2  0 -2 |   1  |   0   1/2   0  |
//     |  2  0  0 |   0  |   0  -1/2 -1/2 | -2,-2,2
//     | -2 -2  0 |   1  | -1/2 -1/2   0  |
//not in DEAL
class Tetrahedron : public Shape3 {
  private:
    static const Point<3> canonical_vertices[4];
    static const Point<SPACEDIM> canonical_center;
    static const int child_neighbors[8][2];

    static const int canonical_vertex_indices[6][2];
    static const int opposite_face[4];
    static const int children[4][4];
    static const Tensor<1,SPACEDIM> face_normals[4];
    static const Tensor<1,SPACEDIM> line_tangents[6];
    static const int faces_from_line[6][2];

    static const int face_vertex_to_cell_vertex[4][3];
    static const int face_line_to_cell_line[4][3];

    virtual Shape1* newChildShape(int,int,const NumPtr<int>&) 
    const {
      OBSOLETE("this routine applies to Shape1 only");
    }
    virtual Shape2* newChildShape(int,int,const NumPtr<int>&,
    const NumPtr<int>&,const map<pair<int,int>,int>&) const {
      OBSOLETE("this routine applies to Shape2 only");
    }
    virtual Shape1* newInteriorCodimension1Shape(int,int,
    const NumPtr<int>&,const NumPtr<int>&) const {
      OBSOLETE("should not be calling this");
    }
#ifdef DEBUG
    virtual void checkFaceVertices(int,const Shape1*) const {
      OBSOLETE("this routine applies to Shape2 only");
    }
#endif
    virtual bool containsCanonical(const Point<1>&) const {
      OBSOLETE("should not be calling this");
      return false;
    }
    virtual bool containsCanonical(const Point<2>&) const {
      OBSOLETE("should not be calling this");
      return false;
    }
  public:
    Tetrahedron(int,const int l[6],const int f[4],const int v[4]);
//  use sparingly: avoids vertex reorder
    Tetrahedron(const int l[6],const int f[4],const int v[4],int);
    virtual ~Tetrahedron() { }
    virtual Shape2* newFaceElement(int) const;
    virtual Shape1* newLineElement(int) const;

    inline static int verticesPerShape() {return 4;}
    inline static int facesPerShape() { return 4; }
    inline static int linesPerShape() { return 6; }
    inline static int linesPerFace(int) {
      return 3;
    }
    inline static int staticNumberVerticesOnFace(int) {
      return 3;
    }
    inline static Point<3> parentCoordinateOfCanonical(
    const Point<3> &p,int c) {
      CHECK_BOUNDS(c,0,8);
      switch (c) {
//      v1,v4,v8,v5:
        case 1: return Point<3>(1.-(p[0]+p[1]+p[2])*0.5,p[2]*0.5,p[1]*0.5);
//      v2,v5,v9,v6:
        case 2: return Point<3>(p[0]*0.5,1.-(p[0]+p[1]+p[2])*0.5,p[1]*0.5);
//      v3,v7,v9,v8:
        case 3: return Point<3>(p[2]*0.5,p[1]*0.5,1.-(p[0]+p[1]+p[2])*0.5);
//      v4,v7,v8,v9:
        case 4: return Point<3>((1.-p[0]-p[2])*0.5,p[2]*0.5,
                                (p[0]+p[1]+p[2])*0.5);
//      v4,v5,v9,v8:
        case 5: return Point<3>((1.-p[1])*0.5,(p[0]+p[1])*0.5,
                                (p[1]+p[2])*0.5);
//      v4,v5,v6,v9:
        case 6: return Point<3>((1.-p[1]-p[2])*0.5,(p[0]+p[1]+p[2])*0.5,
                                p[2]*0.5);
//      v4,v6,v7,v9:
        case 7: return Point<3>((1.-p[0]-p[1]-p[2])*0.5,(p[0]+p[2])*0.5,
                                (p[1]+p[2])*0.5);
//      v0,v4,v6,v7:
        default: return p*0.5;
      }
    }
    inline static Point<3> canonicalVertex(int v) {
      CHECK_BOUNDS(v,0,4);
      return canonical_vertices[v];
    }
    inline static int staticElementVertexFromLine(int line,int vertex) {
      CHECK_BOUNDS(line,0,6);
      CHECK_BOUNDS(vertex,0,2);
      return canonical_vertex_indices[line][vertex];
    }
    inline static int staticElementLineFromFace(int f,int fl) {
      CHECK_BOUNDS(f,0,4);
      CHECK_BOUNDS(fl,0,3);
      return face_line_to_cell_line[f][fl];
    }
    inline static int staticElementVertexFromFace(int face,
    int face_vertex) {
      CHECK_BOUNDS(face,0,4);
      CHECK_BOUNDS(face_vertex,0,3);
      return face_vertex_to_cell_vertex[face][face_vertex];
    }
    inline static int staticChildOnFace(int f,int s) {
      CHECK_BOUNDS(f,0,4);
      CHECK_BOUNDS(s,0,4);
      return children[f][s];
    }
    inline static int staticNumberInteriorChildVerticesOnFace(int) {
      return 0;
    }
    inline static int staticNumberInteriorChildCodimension1Shapes() { 
      return 8; 
    }
    inline static void staticInteriorChildCodimension1ShapeNeighbors(
    int i,int &c0,int &c1) {
      CHECK_BOUNDS(i,0,8);
      c0=child_neighbors[i][0];
      c1=child_neighbors[i][1];
    }
    inline static int staticNumberEquallySpacedLatticePoints(int n) {
      return ((n+1)*(n+2)*(n+3))/6;
    }
    inline static Point<SPACEDIM> staticCenter() { 
      return canonical_center; 
    }
//  unit outer normal * face area / canonical triangle area
    inline static Tensor<1,SPACEDIM> faceNormal(int f) {
      return face_normals[f];
    }
    inline static Tensor<1,SPACEDIM> lineTangent(int l) {
      return line_tangents[l];
    }
    
    inline static int staticNumberCodimension2Shapes() { 
      return 6; 
    }
    inline static int staticNumberInteriorChildCodimension2Shapes() {
      return 1;
    }
    inline void staticFacesFromLine(int l,int &f0,int &f1) const {
      f0=faces_from_line[l][0];
      f1=faces_from_line[l][1];
    }

    int lineFromFace(const Shape2*,int,int) const;
    void facesFromLine(int l,int &f0,int &f1) const {
      staticFacesFromLine(l,f0,f1);
    }
    static void staticEquallySpacedLatticePoints(int,NumPtr<Point<3> >&);
    static void staticProjectToLine(int,const NumPtr<Point<1> >&,
      NumPtr<Point<3> >&);
    static void staticProjectToFace(int,const NumPtr<Point<2> >&,
      NumPtr<Point<3> >&);
    static void projectToSubface(int,int,
      const NumPtr<Point<2> >&,NumPtr<Point<3> >&);
    static void staticInteriorLatticePointIndices(int,NumPtr<int>&);
    static int staticEquallySpacedLatticePointIndexOfVertex(int,int);
    static void staticEquallySpacedLatticePointIndicesOfFace(int,int,
      NumPtr<int>&);
    static void staticEquallySpacedLatticePointIndicesOfFaceInterior(
      int,int,NumPtr<int>&);
    static void staticEquallySpacedLatticePointIndicesOfLine(int,int,
      NumPtr<int>&);
    static void staticEquallySpacedLatticePointIndicesOfLineInterior(int,
      int,NumPtr<int>&);
    static int staticNumberInteriorEquallySpacedLatticePoints(int n) {
      return ((n-1)*(n-2)*(n-3))/6;
    }

    Tensor<1,SPACEDIM> faceTangent(int f,int n) const;
    virtual bool faceTangentsWindBackward(int) const;
    void homogeneousCoordinate(int,const Point<2>&,Point<2>&) const;
    void homogeneousCoordinateDerivative(int,const Point<2>&,
      Tensor<2,2>&) const;
    virtual void initializeMappingSupportPoints(int,
      const NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&) const;
    void lenoirCanonicalFacePoints(int,const NumPtr<Point<3> >&,
      NumPtr<Point<2> >&) const;
    void lenoirCanonicalLinePoints(int,const NumPtr<Point<3> >&,
      NumPtr<Point<1> >&) const;
    void lenoirInterpolationForFace(int,int,
      const NumPtr<Point<SPACEDIM> >&,const NumPtr<Point<SPACEDIM> >&,
      const NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&) const;
    void lenoirInterpolationForLine(int,int,
      const NumPtr<Point<SPACEDIM> >&,const NumPtr<Point<SPACEDIM> >&,
      const NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&) const;

    virtual bool contains(const Point<3>&,const NumPtr<Point<3> >&) const;
    virtual bool containsCanonical(const Point<3> &p) const {
      return (p[0] >= 0.) && (p[1] >= 0.) && (p[2] >= 0.) &&
        (p[0]+p[1]+p[2] <= 1.);
    }
    virtual Point<SPACEDIM> canonicalCenter() const { 
      return canonical_center; 
    }
    virtual REAL diameter(const NumPtr<Point<SPACEDIM> >&) const;
    virtual REAL volume(const NumPtr<Point<SPACEDIM>  >&) const;
    virtual bool lineTangentPointsBackward(int) const;
    virtual Shape3* newChildShape(int,int,const NumPtr<int>&,
      const NumPtr<int>&,const NumPtr<int>&,
      const map<pair<pair<int,int>,pair<int,int> >,int>&,
      const map<pair<int,int>,int>&) const;
    Point<SPACEDIM> parentCoordinateForCanonical(const Point<SPACEDIM>&,
      const Shape*,int,const NumPtr<int>&) const;
    Point<SPACEDIM> childCoordinateOfCanonical(const Point<SPACEDIM>&,
      const Shape*,int,const NumPtr<int>&) const;
    virtual void printOn(ostream &os) const;
    virtual int numberEquallySpacedLatticePoints(int n) const {
      return staticNumberEquallySpacedLatticePoints(n);
    }
    void equallySpacedLatticePoints(int n,
    NumPtr<Point<SPACEDIM> > &points) const {
#if (SPACEDIM==3)
      staticEquallySpacedLatticePoints(n,points);
#else
      OBSOLETE("wrong number of dimensions");
#endif
    }
    virtual int numberInteriorEquallySpacedLatticePoints(int n) const {
      return staticNumberInteriorEquallySpacedLatticePoints(n);
    }
    void reorderLineDofs(int,NumPtr<int>&) const;
    void reorderHdivFaceDofs(int,int,NumPtr<int>&) const;
    void reorderHcurlFaceDofs(int,int,NumPtr<int>&) const;
    void reorderHcurlFaceDofs2(int,int,NumPtr<int>&) const;
    void reorderEquallySpacedLatticePoints(int,NumPtr<int>&) const;
    void reorderChildEquallySpacedLatticePoints(int,const Shape*,int,
      const NumPtr<int>&,NumPtr<int>&) const;
    virtual void reorderChildDofPoints(int,const Shape*,int,
      const NumPtr<int>&,NumPtr<int>&) const;
    virtual void interiorLatticePointIndices(int n,NumPtr<int> &v) const {
      staticInteriorLatticePointIndices(n,v);
    }
    virtual const Shape* reorientFor(const Shape*,int,
      const NumPtr<int>&) const;
    virtual int numberVerticesOnFace(int f) const {
      return 3;
    }
    virtual Point<SPACEDIM> faceCenter(const NumPtr<Point<SPACEDIM> >&,
      int) const;
    virtual int numberInteriorChildCodimension1Shapes() const {
      return 8;
    }
    virtual int elementVertexFromFace(int f,int v) const {
      return staticElementVertexFromFace(f,v);
    }
    virtual int elementVertexFromLine(int f,int l) const {
      return staticElementVertexFromLine(f,l);
    }
    virtual int elementLineFromFace(int f,int fl) const {
      return staticElementLineFromFace(f,fl);
    }
    virtual void interiorChildCodimension1ShapeNeighbors(int i,int &c0,
    int &c1) const {
      staticInteriorChildCodimension1ShapeNeighbors(i,c0,c1);
    }
    virtual int childOnFace(int f,int s) const {
      return staticChildOnFace(f,s);
    }
    virtual int numberInteriorChildVerticesOnFace(int f) const {
      return staticNumberInteriorChildVerticesOnFace(f);
    }
    virtual int equallySpacedLatticePointIndexOfVertex(int qv,
    int degree) const {
      return staticEquallySpacedLatticePointIndexOfVertex(qv,degree);
    }
    virtual void equallySpacedLatticePointIndicesOfFace(int,int,
      NumPtr<int>&) const;
    virtual void equallySpacedLatticePointIndicesOfFaceInterior(int,int,
      NumPtr<int>&) const;
    virtual void equallySpacedLatticePointIndicesOfLineInterior(int,int,
      NumPtr<int>&) const;
    virtual void projectToFace(int f,const NumPtr<Point<1> > &fp,
    NumPtr<Point<2> > &ep) const {
      Shape3::projectToFace(f,fp,ep);
    }
    virtual void projectToFace(int f,const NumPtr<Point<2> > &fp,
    NumPtr<Point<3> > &ep) const {
      staticProjectToFace(f,fp,ep);
    }
#ifdef DEBUG
    virtual void checkFaceVertices(int,const Shape2*) const;
    virtual void draw(const VolGraphTool::WINDOW_TYPE::COLOR_TYPE&,
      const NumPtr<Point<3> >&,VolGraphTool&) const;
#endif

    virtual Shape2* newInteriorCodimension1Shape(int,int,
      const NumPtr<int>&,const NumPtr<int>&,const NumPtr<int>&,
      const map<pair<int,int>,int>&) const;

    virtual int numberCodimension2Shapes() const {
      return 6;
    }
    virtual int numberInteriorChildCodimension2Shapes() const {
      return 1;
    }
    virtual Shape1* newInteriorCodimension2Shape(int i,
    int codimension1_element_index,
    const NumPtr<int> &interior_child_vertex_indices,
    const NumPtr<int> &interior_child_face_vertex_indices,
    const NumPtr<int> &interior_child_line_vertex_indices) const {
      CHECK_BOUNDS(i,0,1);
      int v[2]={interior_child_line_vertex_indices[0],
                interior_child_line_vertex_indices[5]};
      return OPERATOR_NEW Interval(codimension1_element_index,v);
    }
};

//The vertices of a pyramid are
//  v0 = Point(0. ,0. ,0. )
//  v1 = Point(1. ,0. ,0. )
//  v2 = Point(1. ,1. ,0. )
//  v3 = Point(0. ,1. ,0. )
//  v4 = Point(0. ,0. ,1. )
//The lines are
//  L0 = v0 --> v1
//  L1 = v1 --> v2
//  L2 = v3 --> v2
//  L3 = v0 --> v3
//  L4 = v0 --> v4
//  L5 = v1 --> v4
//  L6 = v2 --> v4
//  L7 = v3 --> v4
//The faces are
//  Q0 = ( v0, v1, v2, v3) : { L0, L1, L2, L3 }
//  T1 = ( v0, v1, v4 ) : { L0, L4, L5 }
//  T2 = ( v0, v3, v4 ) : { L3, L4, L7 }
//  T3 = ( v1, v2, v4 ) : { L1, L5, L6 }
//  T4 = ( v2, v3, v4 ) : { L2, L6, L7 }
//The children are problematic. We either subdivide into 4 pyramids
//  that don't cut the lengths well, or into 5 pyramids and 4 tetrahedra
//  see http://www.ems-i.com/gmshelp/modules/3d_mesh_module/creating-editing_3d_meshes/refine_elements.htm
//class PyramidInfo: public CanonicalShape3 {
//private:
//  static const Point vertices[5] =
//    { Point<3>(0.,0.,0.), Point<3>(1.,0.,0.), Point<3>(1.,1.,0.),
//      Point<3>(0.,1.,0.), Point<3>(0.,0.,1.) };
//  static const int canonical_vertex_indices[8][2] =
//    { {0,1}, {1,2}, {3,2}, {0,3}, {0,4}, {1,4}, {2,4}, {3,4} };
//  static const subcells[5][4] =
//    { {0, 1, 2, 4}, {0, 1, 3, 5}, {0, 2, 3, 6}, {1, 2, 3, 7} };
//  static const int dx2deal[5] = { 0, 3, 1, 2};
//public:
//  DeclException3 (ExcInvalidCoordinate, REAL,REAL,REAL,
//    << "The coordinates must satisfy 0 <= x_0,x_1,x_2 "
//    << "&& max(x_0,x_1)+x_2 <= 1, "
//    << "but here we have x_0=" << arg1
//    << "and x_1=" << arg2
//    << "and x_2=" << arg3);
//
//  inline static int verticesPerShape() { return 5; }
//  Point<3> unit_cell_vertex (int v) {
//    CHECK_BOUNDS(v,0,5);
//    return vertices[v];
//  }
//  bool is_inside_unit_cell(const Point<3> &p) {
//    return (p[0] >= 0.) && (p[1] >= 0.) && (p[2] >= 0.) &&
//      (std::max(p[0],p[1])+p[2] <= 1.);
//  }

//  inline static int linesPerShape() { return 8; }
//  int vertices_adjacent_to_line(int line,int vertex) {
//    CHECK_BOUNDS(line,0,8);
//    CHECK_BOUNDS(vertex,0,2);
//    return canonical_vertex_indices[line][vertex];
//  }

//  inline static int facesPerShape() { return 5; }
//  int opposite_face(int face) {
//    see CellAccessor::neighbor_of_neighbor; the value isn't important
//    Assert (false, ExcInternalError());
//    CHECK_BOUNDS(face,0,5);
//    return face;
//  }   
//  inline static int verticesPerFace(int face) { 
//    CHECK_BOUNDS(face,0,5);
//    return (face==0 ? 4 : 3);
//  }
//  inline static int linesPerFace(int face) { 
//    CHECK_BOUNDS(face,0,5);
//    return (face==0 ? 4 : 3); 
//  }

//  int quads_per_cell() { return 5; }

//  int child_cell_on_face (int face,int subface) {
//    Assert (false, ExcInternalError());
//    return subcells[face][subface];
//  }
//  int child_cell_from_point(const Point<3> &p) {
//    Assert (false, ExcInternalError());
//    Assert ((p[0] >= 0 && p[1] >= 0 && p[2] >= 0),
//      ExcInvalidCoordinate(p[0],p[1],p[2]));
//    Assert (std::max(p[0],p[1])+p[2] <= 1 , 
//      ExcInvalidCoordinate(p[0],p[1],p[2]));
//    return 0;
//  }
//  Point<3> cell_to_child_coordinates (const Point<3> &p,
//  int child_index) {
//    CHECK_BOUNDS(c,0,8);
//    Assert (false, ExcInternalError());
//    return Point<3>(0.,0.,0.);
//  }
//  Point<3> child_to_cell_coordinates (const Point<3> &p,
//  int child_index) {
//    CHECK_BOUNDS(c,0,8);
//    Assert (false, ExcInternalError());
//    return Point<3>(0.,0.,0.);
//  }
//
//  int dx_to_deal(int i) {
//    Assert (false, ExcInternalError());
//    CHECK_BOUNDS(i,0,5);
//  }
//};

//The vertices of a prism are
//  v0 = Point( 0., 0., 0. )
//  v1 = Point( 1., 0., 0. )
//  v2 = Point( 0., 1., 0. )
//  v3 = Point( 0., 0., 1. )
//  v4 = Point( 1., 0., 1. )
//  v5 = Point( 0., 1., 1. )
//The lines are
//  l0 : v0 --> v1
//  l1 : v1 --> v2
//  l2 : v2 --> v0
//  l3 : v3 --> v4
//  l4 : v4 --> v5
//  l5 : v5 --> v3
//  l6 : v0 --> v3
//  l7 : v1 --> v4
//  l8 : v2 --> v5
//The faces are
//  T0 = ( v0 , v2 , v1 )
//  T1 = ( v3 , v4 , v5 )
//  Q2 = ( v0 , v3 , v2 , v5 )
//  Q3 = ( v0 , v1 , v3 , v4 )
//  Q4 = ( v1 , v2 , v4 , v5 )
//Let the line midpoints be
//  v6  = Point( 0.5, 0. , 0.  )
//  v7  = Point( 0.5, 0.5, 0.  )
//  v8  = Point( 0. , 0.5, 0.  )
//  v9  = Point( 0.5, 0. , 1.  )
//  v10 = Point( 0.5, 0.5, 1.  )
//  v11 = Point( 0. , 0.5, 1.  )
//  v12 = Point( 0. , 0. , 0.5 )
//  v13 = Point( 1. , 0. , 0.5 )
//  v14 = Point( 0. , 1. , 0.5 )
//and let the quadrilateral face midpoints be
//  v15 = Point( 0. , 0.5, 0.5 )
//  v16 = Point( 0.5, 0.0, 0.5 )
//  v17 = Point( 0.5, 0.5, 0.5 )
//The interior child lines are
//  l0 = v17 --> v15
//  l1 = v15 --> v16
//  l2 = v16 --> v17
//The interior child faces are
//  T0 = ( v12 , v15 , v16 )
//  T1 = ( v16 , v17 , v13 )
//  T2 = ( v15 , v14 , v17 )
//  T3 = ( v17 , v16 , v15 )
//  Q4 = ( v6  , v16 , v7  , v17 )
//  Q5 = ( v8  , v7  , v15 , v17 )
//  Q6 = ( v8  , v6  , v15 , v16 )
//  Q7 = ( v16 , v9  , v17 , v10 )
//  Q8 = ( v15 , v17 , v11 , v10 )
//  Q9 = ( v15 , v16 , v11 , v9  )
//The children are
//  W0 = { v0 , v6 , v8 , v12, v16, v15 }
//  W1 = { v6 , v1 , v7 , v16, v13, v17 }
//  W2 = { v8 , v7 , v2 , v15, v17, v14 }
//  W3 = { v17, v16, v15, v7 , v6 , v8  }
//  W4 = { v12, v16, v15, v3 , v9 , v11 }
//  W5 = { v16, v13, v17, v9 , v4 , v10 }
//  W6 = { v15, v17, v14, v11, v10, v5  }
//  W7 = { v10, v9 , v11, v17, v16, v15 }
//In terms of the transformation c = A p + b (p=parent, c=child coord)
//     |     A    |   b  |     inv(A)     | eig(A)
//     -----------------------------------------------------------------
//  W0:|  2  0  0 |   0  |  1/2   0    0  |
//     |  0  2  0 |   0  |   0   1/2   0  | 2,2,2
//     |  0  0  2 |   0  |   0    0   1/2 |
//     -----------------------------------------------------------------
//  W1:|  2  0  0 |  -1  |  1/2   0    0  |
//     |  0  2  0 |   0  |   0   1/2   0  | 2,2,2
//     |  0  0  2 |   0  |   0    0   1/2 |
//     -----------------------------------------------------------------
//  W2:|  2  0  0 |   0  |  1/2   0    0  |
//     |  0  2  0 |  -1  |   0   1/2   0  | 2,2,2
//     |  0  0  2 |   0  |   0    0   1/2 |
//     -----------------------------------------------------------------
//  W3:|  0 -2  0 |   1  |   0  -1/2   0  |
//     | -2  0  0 |   1  | -1/2   0    0  | -2,-2,2
//     |  0  0 -2 |   1  |   0    0  -1/2 |
//     -----------------------------------------------------------------
//  W4:|  2  0  0 |   0  |  1/2   0    0  |
//     |  0  2  0 |   0  |   0   1/2   0  | 2,2,2
//     |  0  0  2 |  -1  |   0    0   1/2 |
//     -----------------------------------------------------------------
//  W5:|  2  0  0 |  -1  |  1/2   0    0  |
//     |  0  2  0 |   0  |   0   1/2   0  | 2,2,2
//     |  0  0  2 |  -1  |   0    0   1/2 |
//     -----------------------------------------------------------------
//  W6:|  2  0  0 |   0  |  1/2   0    0  |
//     |  0  2  0 |  -1  |   0   1/2   0  | 2,2,2
//     |  0  0  2 |  -1  |   0    0   1/2 |
//     -----------------------------------------------------------------
//  W7:|  0 -2  0 |   1  |   0  -1/2   0  |
//     | -2  0  0 |   1  | -1/2   0    0  | -2,-2,2
//     |  0  0 -2 |   2  |   0    0  -1/2 |

//not in DEAL
class Prism : public Shape3 {
  private:
    static const Point<3> canonical_vertices[6];
    static const Point<SPACEDIM> canonical_center;
    static const int child_neighbors[10][2];

    static const int interior_child_vertices[5][2];
    static const int canonical_vertex_indices[9][2];
    static const int opposite_face[5];
    static const int children[5][4];
    static const Tensor<1,SPACEDIM> face_normals[5];
    static const Tensor<1,SPACEDIM> line_tangents[9];
    static const int faces_from_line[9][2];

    static const int face_vertex_to_cell_vertex[5][4];
    static const int face_line_to_cell_line[5][4];
    static const int interior_child_line[3][2];

    virtual Shape1* newInteriorCodimension1Shape(int,int,
    const NumPtr<int>&,const NumPtr<int>&) const {
      OBSOLETE("should not be calling this");
    }
    virtual Shape1* newChildShape(int,int,const NumPtr<int>&) 
    const {
      OBSOLETE("this routine applies to Shape1 only");
    }
    virtual Shape2* newChildShape(int,int,const NumPtr<int>&,
    const NumPtr<int>&,const map<pair<int,int>,int>&) const {
      OBSOLETE("should not be calling this");
    }
#ifdef DEBUG
    virtual void checkFaceVertices(int,const Shape1*) const {
      OBSOLETE("this routine applies to Shape2 only");
    }
#endif
    virtual bool containsCanonical(const Point<1>&) const {
      OBSOLETE("should not be calling this");
      return false;
    }
    virtual bool containsCanonical(const Point<2>&) const {
      OBSOLETE("should not be calling this");
      return false;
    }
  public:
    Prism(int,const int l[9],const int f[5],const int v[6]);
//  use sparingly: avoids vertex reorder
    Prism(const int l[9],const int f[5],const int v[6],int);
    virtual ~Prism() { }
    virtual Shape2* newFaceElement(int) const;
    virtual Shape1* newLineElement(int) const;

    inline static int verticesPerShape() {return 6;}
    inline static int facesPerShape() { return 5; }
    inline static int linesPerShape() { return 9; }
    inline static int linesPerFace(int face) {
      CHECK_BOUNDS(face,0,5);
      return (face<2 ? 3 : 4);
    }
    inline static int staticNumberVerticesOnFace(int face) {
      CHECK_BOUNDS(face,0,5);
      return (face<2 ? 3 : 4);
    }
    inline static bool isTriangleFace(int f) { return f>=0 && f<2; }
    inline static bool isTriangleLine(int l) { return l>=0 && l<6; }
    inline static Point<3> parentCoordinateOfCanonical(const Point<3> &p,
    int c) {
      CHECK_BOUNDS(c,0,8);
      switch (c) {
        case 1: // v1, v7, v6, v13, v17, v16
          return Point<3>(1.-(p[0]+p[1])*0.5,p[0]*0.5,p[2]*0.5);
        case 2: // v2, v8, v7, v14, v15, v17
          return Point<3>(p[1]*0.5,1.-(p[0]+p[1])*0.5,p[2]*0.5);
        case 3: // v6, v7, v8, v16, v17, v15
          return Point<3>((1.-p[1])*0.5,(p[0]+p[1])*0.5,p[2]*0.5);
        case 4: // v3, v11, v9, v12, v15, v16
          return Point<3>(p[1]*0.5,p[0]*0.5,1.-p[2]*0.5);
        case 5: // v4, v9, v10, v13, v16, v17
          return Point<3>(1.-(p[0]+p[1])*0.5,p[1]*0.5,1.-p[2]*0.5);
        case 6: // v5, v10, v11, v14, v17, v15
          return Point<3>(p[0]*0.5,1.-(p[0]+p[1])*0.5,1.-p[2]*0.5);
        case 7: // v9, v11, v10, v16, v15, v17
          return Point<3>((1.-p[1])*0.5,(1.-p[0])*0.5,1.-p[2]*0.5);
        default: return p*0.5;
      }
    }
    inline static int oppositeFace(int face) {
      CHECK_BOUNDS(face,0,5);
      return opposite_face[face];
    }
    inline static Point<3> canonicalVertex(int v) {
      CHECK_BOUNDS(v,0,6);
      return canonical_vertices[v];
    }
    inline static int staticElementVertexFromLine(int line,int v) {
      CHECK_BOUNDS(line,0,9);
      CHECK_BOUNDS(v,0,2);
      return canonical_vertex_indices[line][v];
    }
    inline static int staticElementLineFromFace(int f,int fl) {
      CHECK_BOUNDS(f,0,5);
      CHECK_BOUNDS(fl,0,4);
      if (f<2) {
        CHECK_BOUNDS(fl,0,3);
      }
      return face_line_to_cell_line[f][fl];
    }
    inline static int staticElementVertexFromFace(int face,
    int face_vertex) {
      CHECK_BOUNDS(face,0,5);
      CHECK_BOUNDS(face_vertex,0,4);
      if (face<2) {
        CHECK_BOUNDS(face_vertex,0,3);
      }
      return face_vertex_to_cell_vertex[face][face_vertex];
    }
    inline static int staticChildOnFace(int f,int s) {
      CHECK_BOUNDS(f,0,5);
      CHECK_BOUNDS(s,0,4);
      return children[f][s];
    }
    inline static int staticNumberInteriorChildVerticesOnFace(int f) {
      return (f<2 ? 0 : 1 );
    }
    inline static int staticNumberInteriorChildCodimension1Shapes() {
      return 10;
    }
    inline static void staticInteriorChildCodimension1ShapeNeighbors(
    int i,int &c0,int &c1) {
      CHECK_BOUNDS(i,0,10);
      c0=child_neighbors[i][0];
      c1=child_neighbors[i][1];
    }
    inline static int staticNumberEquallySpacedLatticePoints(int n01,
    int n2) {
      return (n2+1)*((n01+1)*(n01+2))/2;
    }
    inline static int staticNumberEquallySpacedLatticePoints(int n) {
      staticNumberEquallySpacedLatticePoints(n,n);
    }
    inline static Point<SPACEDIM> staticCenter() { 
      return canonical_center; 
    }
//  unit outer normal * face area / canonical triangle/quadrilateral area
    inline static Tensor<1,SPACEDIM> faceNormal(int f) {
      return face_normals[f];
    }
    inline static Tensor<1,SPACEDIM> lineTangent(int l) {
      return line_tangents[l];
    }
    inline static int numberInteriorChildLinesOnFace(int f) {
      return (f<2 ? 3 : 4);
    }
    inline static int staticNumberCodimension2Shapes() { 
      return 9; 
    }
    inline static int staticNumberInteriorChildCodimension2Shapes() {
      return 3;
    }
    inline void staticFacesFromLine(int l,int &f0,int &f1) const {
      f0=faces_from_line[l][0];
      f1=faces_from_line[l][1];
    }

    int lineFromFace(const Shape2*,int,int) const;
    void facesFromLine(int l,int &f0,int &f1) const {
      staticFacesFromLine(l,f0,f1);
    }
    static void staticEquallySpacedLatticePoints(int,int,
      NumPtr<Point<3> >&);
    static void staticEquallySpacedLatticePoints(int n,
    NumPtr<Point<3> > &points) {
      staticEquallySpacedLatticePoints(n,n,points);
    }
    static void staticInteriorChildVertexOnFace(int f,int &c, int &v) {
      CHECK_BOUNDS(f,2,5);
      c=interior_child_vertices[f][0]; 
      v=interior_child_vertices[f][1]; 
    }
    static void staticProjectToLine(int,const NumPtr<Point<1> >&,
      NumPtr<Point<3> >&);
    static void staticProjectToFace(int,const NumPtr<Point<2> >&,
      NumPtr<Point<3> >&);
    static void projectToSubface(int,int,const NumPtr<Point<2> >&,
      NumPtr<Point<3> >&);
    static void staticInteriorLatticePointIndices(int n,NumPtr<int> &v);
    static int staticEquallySpacedLatticePointIndexOfVertex(int,int);
    static void staticEquallySpacedLatticePointIndicesOfFace(int,int,
      NumPtr<int>&);
    static void staticEquallySpacedLatticePointIndicesOfFaceInterior(
      int,int,NumPtr<int>&);
    static void staticEquallySpacedLatticePointIndicesOfLine(int,int,
      NumPtr<int>&);
    static void staticEquallySpacedLatticePointIndicesOfLineInterior(
      int,int,NumPtr<int>&);

    static void staticInteriorChildCodimension2ShapeNeighbors(int i,
    int &c0,int &c1) {
      c0=interior_child_line[i][0];
      c1=interior_child_line[i][1];
    }
    inline static int staticNumberInteriorEquallySpacedLatticePoints(
    int n) {
      return (n-1)*((n-1)*(n-2))/2;
    }
    inline static int triangularFaceContainingLine(int l) { 
      return (l<3 ? 0 : (l<6 ? 1 : -1)); 
    }
    inline static int triangularFace(int i) {
      CHECK_BOUNDS(i,0,2);
      return i;
    }
    static int lineInBothFaces(int,int);
    Tensor<1,SPACEDIM> faceTangent(int f,int n) const;
    virtual bool faceTangentsWindBackward(int) const;
    void homogeneousCoordinate(int,const Point<2>&,Point<2>&) const;
    void homogeneousCoordinateDerivative(int,const Point<2>&,
      Tensor<2,2>&) const;
    void projectToFaceForBDFM(int,const NumPtr<Point<2> >&,
      NumPtr<Point<3> >&) const;
    void projectToFaceForHcurl2(int,const NumPtr<Point<2> >&,
      NumPtr<Point<3> >&) const;
    virtual void initializeMappingSupportPoints(int,
      const NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&) const;

    virtual bool contains(const Point<3>&,const NumPtr<Point<3> >&) const;
    virtual bool containsCanonical(const Point<3> &p) const {
      return (p[0] >= 0.) && (p[1] >= 0.) && (p[2] >= 0.) &&
        (p[0]+p[1] <= 1.) && (p[2] <=1.);
    }
    virtual Point<SPACEDIM> canonicalCenter() const { 
      return canonical_center; 
    }
    virtual REAL diameter(const NumPtr<Point<SPACEDIM> >&) const;
    virtual REAL volume(const NumPtr<Point<SPACEDIM> >&) const;
    virtual bool lineTangentPointsBackward(int) const;
    virtual Shape3* newChildShape(int,int,const NumPtr<int>&,
      const NumPtr<int>&,const NumPtr<int>&,
      const map<pair<pair<int,int>,pair<int,int> >,int>&,
      const map<pair<int,int>,int>&) const;
    Point<SPACEDIM> parentCoordinateForCanonical(const Point<SPACEDIM>&,
      const Shape*,int,const NumPtr<int>&) const;
    Point<SPACEDIM> childCoordinateOfCanonical(const Point<SPACEDIM>&,
      const Shape*,int,const NumPtr<int>&) const;
    virtual void printOn(ostream &os) const;
    virtual int numberEquallySpacedLatticePoints(int n01,int n2) const {
      return staticNumberEquallySpacedLatticePoints(n01,n2);
    }
    virtual int numberEquallySpacedLatticePoints(int n) const {
      return staticNumberEquallySpacedLatticePoints(n,n);
    }
    void equallySpacedLatticePoints(int n,
    NumPtr<Point<SPACEDIM> > &points) const {
#if (SPACEDIM==3)
      staticEquallySpacedLatticePoints(n,points);
#else
      OBSOLETE("wrong number of dimensions");
#endif
    }
    virtual int numberInteriorEquallySpacedLatticePoints(int n) const {
      return staticNumberInteriorEquallySpacedLatticePoints(n);
    }
    void reorderLineDofs(int,NumPtr<int>&) const;
    void reorderRTFaceDofs(int,int,NumPtr<int>&) const;
    void reorderBDFMFaceDofs(int,int,NumPtr<int>&) const;
    void reorderHcurlFaceDofs(int,int,NumPtr<int>&) const;
    void reorderHcurlFaceDofs2(int,int,NumPtr<int>&) const;
    virtual void reorderEquallySpacedLatticePoints(int,NumPtr<int>&) const;
    void reorderChildEquallySpacedLatticePoints(int,const Shape*,int,
      const NumPtr<int>&,NumPtr<int>&) const;
    virtual void reorderChildDofPoints(int,const Shape*,int,
      const NumPtr<int>&,NumPtr<int>&) const;
    virtual void interiorLatticePointIndices(int n,NumPtr<int> &v) const {
      staticInteriorLatticePointIndices(n,v);
    }
    virtual void computeInteriorLatticePoints(int,int,
      NumPtr<Point<SPACEDIM> >&) const;
    virtual const Shape* reorientFor(const Shape*,int,
      const NumPtr<int>&) const;
    virtual int numberVerticesOnFace(int f) const {
      return staticNumberVerticesOnFace(f);
    }
    virtual Point<SPACEDIM> faceCenter(const NumPtr<Point<SPACEDIM> >&,
      int) const;
    virtual int numberInteriorChildCodimension1Shapes() const {
      return 10;
    }
    virtual int elementVertexFromFace(int f,int v) const {
      return staticElementVertexFromFace(f,v);
    }
    virtual int elementVertexFromLine(int f,int l) const {
      return staticElementVertexFromLine(f,l);
    }
    virtual int elementLineFromFace(int f,int fl) const {
      return staticElementLineFromFace(f,fl);
    }
    virtual void interiorChildCodimension1ShapeNeighbors(int i,int &c0,
    int &c1) const {
      staticInteriorChildCodimension1ShapeNeighbors(i,c0,c1);
    }
    virtual int childOnFace(int f,int s) const {
      return staticChildOnFace(f,s);
    }
    virtual int numberInteriorChildVerticesOnFace( int f) const {
      return staticNumberInteriorChildVerticesOnFace(f);
    }
    virtual int equallySpacedLatticePointIndexOfVertex(int qv,int degree) 
    const {
      return staticEquallySpacedLatticePointIndexOfVertex(qv,degree);
    }
    virtual void equallySpacedLatticePointIndicesOfFace(int,int,
      NumPtr<int>&) const;
    virtual void equallySpacedLatticePointIndicesOfFaceInterior(
      int,int,NumPtr<int>&) const;
    virtual void equallySpacedLatticePointIndicesOfLineInterior(
      int,int,NumPtr<int>&) const;
    virtual void projectToFace(int f,const NumPtr<Point<1> > &fp,
    NumPtr<Point<2> > &ep) const {
      Shape3::projectToFace(f,fp,ep);
    }
    virtual void projectToFace(int f,const NumPtr<Point<2> > &fp,
    NumPtr<Point<3> > &ep) const {
      staticProjectToFace(f,fp,ep);
    }
#ifdef DEBUG
    virtual void checkFaceVertices(int,const Shape2*) const;
    virtual void draw(const VolGraphTool::WINDOW_TYPE::COLOR_TYPE&,
      const NumPtr<Point<3> >&,VolGraphTool&) const;
#endif
    virtual void interiorChildVertexOnFace(int f,int &c,int &v) const {
      staticInteriorChildVertexOnFace(f,c,v);
    }
    virtual Shape2* newInteriorCodimension1Shape(int,int,
      const NumPtr<int>&,const NumPtr<int>&,const NumPtr<int>&,
      const map<pair<int,int>,int>&) const;

    virtual int numberCodimension2Shapes() const { return 9; }
    virtual int numberInteriorChildCodimension2Shapes() const {
      return 3;
    }
    Shape1* newInteriorCodimension2Shape(int i,
    int codimension1_element_index,
    const NumPtr<int> &interior_child_vertex_indices,
    const NumPtr<int> &interior_child_face_vertex_indices,
    const NumPtr<int> &interior_child_line_vertex_indices) const {
      CHECK_BOUNDS(i,0,3);
      int v[2]={interior_child_face_vertex_indices[2+(i+2)%3],
                interior_child_face_vertex_indices[2+i]};
      return OPERATOR_NEW Interval(codimension1_element_index,v);
    }
    virtual void interiorChildCodimension2ShapeNeighbors(int i,int &c0,
    int &c1) const {
      staticInteriorChildCodimension2ShapeNeighbors(i,c0,c1);
    }
};

//The vertices of a hexahedron are
//  v0 = Point( 0., 0., 0. )
//  v1 = Point( 1., 0., 0. )
//  v2 = Point( 0., 1., 0. )
//  v3 = Point( 1., 1., 0. )
//  v4 = Point( 0., 0., 1. )
//  v5 = Point( 1., 0., 1. )
//  v6 = Point( 0., 1., 1. )
//  v7 = Point( 1., 1., 1. )
//The lines are
//  l0 : v0 --> v2
//  l1 : v1 --> v3
//  l2 : v0 --> v1
//  l3 : v2 --> v3
//  l4 : v4 --> v6
//  l5 : v5 --> v7
//  l6 : v4 --> v5
//  l7 : v6 --> v7
//  l8 : v0 --> v4
//  l9 : v1 --> v5
//  l10: v2 --> v6
//  l11: v3 --> v7
//The faces are
//  Q0 = ( v0 , v4 , v2 , v6 )
//  Q1 = ( v1 , v5 , v3 , v7 )
//  Q2 = ( v0 , v1 , v4 , v5 )
//  Q3 = ( v2 , v3 , v6 , v7 )
//  Q4 = ( v0 , v2 , v1 , v3 )
//  Q5 = ( v4 , v6 , v5 , v7 )
//Let the line midpoints be
//  v8  = Point( 0. , 0.5, 0.  )
//  v9  = Point( 1. , 0.5, 0.  )
//  v10 = Point( 0.5, 0. , 0.  )
//  v11 = Point( 0.5, 0.5, 0.  )
//  v12 = Point( 0. , 0.5, 1.  )
//  v13 = Point( 1. , 0.5, 1.  )
//  v14 = Point( 0.5, 0. , 1.  )
//  v15 = Point( 0.5, 1. , 1.  )
//  v16 = Point( 0. , 0. , 0.5 )
//  v17 = Point( 1. , 0. , 0.5 )
//  v18 = Point( 0. , 1. , 0.5 )
//  v19 = Point( 1. , 1. , 0.5 )
//let the face midpoints be
//  v20 = Point( 0. , 0.5, 0.5 )
//  v21 = Point( 1. , 0.5, 0.5 )
//  v22 = Point( 0.5, 0. , 0.5 )
//  v23 = Point( 0.5, 1. , 0.5 )
//  v24 = Point( 0.5, 0.5, 0.  )
//  v25 = Point( 0.5, 0.5, 1.  )
//and let the centroid be
//  v26 = Point( 0.5, 0.5, 0.5 )
//The interior child lines are
//  l0 = v20 --> v26
//  l1 = v21 --> v26
//  l2 = v22 --> v26
//  l3 = v23 --> v26
//  l4 = v24 --> v26
//  l5 = v25 --> v26
//The interior child faces are
//  Q0 = { v10, v22, v24, v26 }
//  Q1 = { v22, v14, v26, v25 }
//  Q2 = { v24, v26, v11, v23 }
//  Q3 = { v26, v25, v23, v15 }
//  Q4 = { v8 , v24, v20, v26 }
//  Q5 = { v24, v9 , v26, v21 }
//  Q6 = { v20, v26, v12, v25 }
//  Q7 = { v26, v21, v25, v13 }
//  Q8 = { v16, v20, v22, v26 }
//  Q9 = { v20, v18, v26, v23 }
//  Q10= { v22, v26, v17, v21 }
//  Q11= { v26, v23, v21, v19 }
//The children are
//  H0 = { v0 , v10, v8 , v24, v16, v22, v20, v26 }
//  H1 = { v1 , v9 , v10, v24, v17, v21, v22, v26 }
//  H2 = { v2 , v8 , v11, v24, v18, v20, v23, v26 }
//  H3 = { v3 , v11, v9 , v24, v19, v23, v21, v26 }
//  H4 = { v4 , v12, v14, v25, v16, v20, v22, v26 }
//  H5 = { v5 , v14, v13, v25, v17, v22, v21, v26 }
//  H6 = { v6 , v15, v12, v25, v18, v23, v20, v26 }
//  H7 = { v7 , v13, v15, v25, v19, v21, v23, v26 }
//compare to DEAL GeometryInfo<3>
class Hexahedron : public Shape3 {
  private:
    static const Point<3> canonical_vertices[8];
    static const Point<SPACEDIM> canonical_center;
    static const int child_neighbors[12][2];

    static const int interior_child_vertices[6][2];
    static const int opposite_vertex[8];
    static const int canonical_vertex_indices[12][2];
    static const int opposite_face[6];
    static const int children[6][4];
    static const Tensor<1,SPACEDIM> face_normals[6];
    static const Tensor<1,SPACEDIM> line_tangents[12];
    static const int faces_from_line[12][2];

    static const int face_vertex_to_cell_vertex[6][4];
    static const int face_line_to_cell_line[6][4];
    static const int interior_child_line[6][2];

    virtual Shape1* newInteriorCodimension1Shape(int,int,
    const NumPtr<int>&,const NumPtr<int>&) const {
      OBSOLETE("should not be calling this");
    }
    virtual Point<SPACEDIM> interiorChildPoint(
    const NumPtr<Point<SPACEDIM> >&) const {
      OBSOLETE("this routine applies to Shape1 only");
    }
    virtual Point<SPACEDIM> interiorChildPoint(
    const NumPtr<Point<SPACEDIM> >&,const NumPtr<int>&) const {
      OBSOLETE("this shape has no interior child vertex");
    }
    virtual Shape1* newChildShape(int,int,const NumPtr<int>&) const {
      OBSOLETE("this routine applies to Shape1 only");
    }
    virtual Shape2* newChildShape(int,int,const NumPtr<int>&,
    const NumPtr<int>&,const map<pair<int,int>,int>&) const {
      OBSOLETE("should not be calling this");
    }
#ifdef DEBUG
    virtual void checkFaceVertices(int,const Shape1*) const {
      OBSOLETE("this routine applies to Shape2 only");
    }
#endif
    virtual bool containsCanonical(const Point<1>&) const {
      OBSOLETE("should not be calling this");
      return false;
    }
    virtual bool containsCanonical(const Point<2>&) const {
      OBSOLETE("should not be calling this");
      return false;
    }
  public:
    Hexahedron(int,const int l[12],const int f[6],const int v[8]);
//  use sparingly: avoids vertex reorder
    Hexahedron(const int l[12],const int f[6],const int v[8],int);
    virtual ~Hexahedron() { }
    virtual Shape2* newFaceElement(int) const;
    virtual Shape1* newLineElement(int) const;

    inline static int verticesPerShape() {return 8;}
    inline static int facesPerShape() { return 6; }
    inline static int linesPerShape() { return 12; }
    inline static int linesPerFace(int) { return 4; }
    inline static int staticNumberVerticesOnFace(int face) {
      return 4;
    }
    inline static Point<3> parentCoordinateOfCanonical(
    const Point<3> &p,int c) {
      CHECK_BOUNDS(c,0,8);
      switch (c) {
        case 1:
          return Point<3>(1.-p[1]*0.5,p[0]*0.5,p[2]*0.5);
        case 2:
          return Point<3>(p[1]*0.5,1.-p[0]*0.5,p[2]*0.5);
        case 3:
          return Point<3>(1.-p[0]*0.5,1.-p[1]*0.5,p[2]*0.5);
        case 4:
          return Point<3>(p[1]*0.5,p[0]*0.5,1.-p[2]*0.5);
        case 5:
          return Point<3>(1.-p[0]*0.5,p[1]*0.5,1.-p[2]*0.5);
        case 6:
          return Point<3>(p[0]*0.5,1.-p[1]*0.5,1.-p[2]*0.5);
        case 7:
          return Point<3>(1.-p[1]*0.5,1.-p[0]*0.5,1.-p[2]*0.5);
        default: return p*0.5;
      }
    }
    inline static int oppositeVertex(int v) {
      CHECK_BOUNDS(v,0,8);
      return opposite_vertex[v];
    }
    inline static Point<3> canonicalVertex(int v) {
      CHECK_BOUNDS(v,0,8);
      return canonical_vertices[v];
    }
    inline static int staticElementVertexFromLine(int line,int vertex) {
      CHECK_BOUNDS(line,0,12);
      CHECK_BOUNDS(vertex,0,2);
      return canonical_vertex_indices[line][vertex];
    }
    inline static int staticElementLineFromFace(int f,int fl) {
      CHECK_BOUNDS(f,0,6);
      CHECK_BOUNDS(fl,0,4);
      return face_line_to_cell_line[f][fl];
    }
    inline static int staticElementVertexFromFace(int face,
    int face_vertex) {
      CHECK_BOUNDS(face,0,6);
      CHECK_BOUNDS(face_vertex,0,4);
      return face_vertex_to_cell_vertex[face][face_vertex];
    }
    static inline int staticChildOnFace(int f,int s) {
      CHECK_BOUNDS(f,0,6);
      CHECK_BOUNDS(s,0,4);
      return children[f][s];
    }
    inline static int staticNumberInteriorChildVerticesOnFace(int) {
      return 1;
    }
    inline static int staticNumberInteriorChildCodimension1Shapes() {
      return 12;
    }
    inline static void staticInteriorChildCodimension1ShapeNeighbors(
    int i,int &c0,int &c1) {
      CHECK_BOUNDS(i,0,12);
      c0=child_neighbors[i][0];
      c1=child_neighbors[i][1];
    }
    inline static int staticNumberEquallySpacedLatticePoints(int n0,
    int n1,int n2) {
      return (n0+1)*(n1+1)*(n2+1);
    }
    inline static int staticNumberEquallySpacedLatticePoints(int n) {
      return staticNumberEquallySpacedLatticePoints(n,n,n);
    }
    inline static Point<SPACEDIM> staticCenter() { 
      return canonical_center; 
    }
//  unit outer normal * face area / canonical quadrilateral area
    inline static Tensor<1,SPACEDIM> faceNormal(int f) {
      return face_normals[f];
    }
    inline static Tensor<1,SPACEDIM> lineTangent(int l) {
      return line_tangents[l];
    }
    inline static int numberInteriorChildLinesOnFace(int f) {
      return 4;
    }
    inline static int staticNumberCodimension2Shapes() { 
      return 12; 
    }
    inline static int staticNumberInteriorChildCodimension2Shapes() {
      return 6;
    }

    inline static int staticNumberInteriorChildVertices() {
      return 1;
    }
    inline static void staticInteriorChildVertex(int,int &c,int &v) {
      c=0;
      v=7;
    }
    inline void staticFacesFromLine(int l,int &f0,int &f1) const {
      f0=faces_from_line[l][0];
      f1=faces_from_line[l][1];
    }

    int lineFromFace(const Shape2*,int,int) const;
    void facesFromLine(int l,int &f0,int &f1) const {
      staticFacesFromLine(l,f0,f1);
    }
    static void staticEquallySpacedLatticePoints(int,int,int,
      NumPtr<Point<3> >&);
    static void staticEquallySpacedLatticePoints(int n,
    NumPtr<Point<3> > & points) {
      staticEquallySpacedLatticePoints(n,n,n,points);
    }
    static void staticInteriorChildVertexOnFace(int f,int &c,int &v) {
      CHECK_BOUNDS(f,0,6);
      c=interior_child_vertices[f][0];
      v=interior_child_vertices[f][1];
    }
    static void staticProjectToLine(int,const NumPtr<Point<1> >&,
      NumPtr<Point<3> >&);
    static void staticProjectToFace(int,const NumPtr<Point<2> >&,
      NumPtr<Point<3> >&);
    static void projectToSubface(int,int,const NumPtr<Point<2> >&,
    NumPtr<Point<3> >&);
    static void staticInteriorLatticePointIndices(int,int,int,
      NumPtr<int>&);
    static void staticInteriorLatticePointIndices(int n,NumPtr<int> &v) {
      staticInteriorLatticePointIndices(n,n,n,v);
    }
    static int staticEquallySpacedLatticePointIndexOfVertex(int,int,int,
      int);
    static int staticEquallySpacedLatticePointIndexOfVertex(int v,int n) {
      staticEquallySpacedLatticePointIndexOfVertex(v,n,n,n);
    }
    static void staticEquallySpacedLatticePointIndicesOfFace(int,int,
      int,int,NumPtr<int>&);
    static void staticEquallySpacedLatticePointIndicesOfFace(int f,int n,
    NumPtr<int> &indices) {
      staticEquallySpacedLatticePointIndicesOfFace(f,n,n,n,indices);
    }
    static void staticEquallySpacedLatticePointIndicesOfFaceInterior(
      int,int,int,int,NumPtr<int>&);
    static void staticEquallySpacedLatticePointIndicesOfFaceInterior(
    int f,int n,NumPtr<int> &indices) {
      staticEquallySpacedLatticePointIndicesOfFaceInterior(f,n,n,n,
        indices);
    }
    static void staticEquallySpacedLatticePointIndicesOfLine(int,int,int,
      int,NumPtr<int>&);
    static void staticEquallySpacedLatticePointIndicesOfLine(int l,int n,
    NumPtr<int> &indices) {
      staticEquallySpacedLatticePointIndicesOfLine(l,n,n,n,indices);
    }
    static void staticEquallySpacedLatticePointIndicesOfLineInterior(int,
      int,int,int,NumPtr<int>&);
    static void staticEquallySpacedLatticePointIndicesOfLineInterior(
    int l,int n,NumPtr<int> &indices) {
      staticEquallySpacedLatticePointIndicesOfLineInterior(l,n,n,n,
        indices);
    }

    inline static void staticInteriorChildCodimension2ShapeNeighbors(
    int i,int &c0,int &c1) {
      c0=interior_child_line[i][0];
      c1=interior_child_line[i][1];
    }
    static int staticNumberInteriorEquallySpacedLatticePoints(int n0,
    int n1,int n2) {
      return (n0-1)*(n1-1)*(n2-1);
    }
    static int staticNumberInteriorEquallySpacedLatticePoints(int n) {
      return staticNumberInteriorEquallySpacedLatticePoints(n,n,n);
    }

    Tensor<1,SPACEDIM> faceTangent(int f,int n) const;
    virtual bool faceTangentsWindBackward(int) const;
    virtual bool contains(const Point<3>&,const NumPtr<Point<3> >&) const;
    virtual bool containsCanonical(const Point<3> &p) const {
      return (p[0] >= 0.) && (p[0] <= 1.) &&
             (p[1] >= 0.) && (p[1] <= 1.) &&
             (p[2] >= 0.) && (p[2] <= 1.);
    }
    virtual Point<SPACEDIM> canonicalCenter() const { 
      return canonical_center; 
    }
    virtual REAL diameter(const NumPtr<Point<SPACEDIM> >&) const;
    virtual REAL volume(const NumPtr<Point<SPACEDIM> >&) const;
    virtual bool lineTangentPointsBackward(int) const;
    Shape3* newChildShape(int,int,const NumPtr<int>&,
      const NumPtr<int>&,const NumPtr<int>&,
      const map<pair<pair<int,int>,pair<int,int> >,int>&,
      const map<pair<int,int>,int>&) const;
    Point<SPACEDIM> parentCoordinateForCanonical(const Point<SPACEDIM>&,
      const Shape*,int,const NumPtr<int>&) const;
    Point<SPACEDIM> childCoordinateOfCanonical(const Point<SPACEDIM>&,
      const Shape*,int,const NumPtr<int>&) const;
    virtual void printOn(ostream &os) const;
    virtual int numberEquallySpacedLatticePoints(int n) const {
      return staticNumberEquallySpacedLatticePoints(n,n,n);
    }
    void equallySpacedLatticePoints(int n,
    NumPtr<Point<SPACEDIM> > &points) const {
#if (SPACEDIM==3)
      staticEquallySpacedLatticePoints(n,n,n,points);
#else
      OBSOLETE("wrong number of dimensions");
#endif
    }
    virtual int numberInteriorEquallySpacedLatticePoints(int n) const {
      return staticNumberInteriorEquallySpacedLatticePoints(n,n,n);
    }
    void reorderLineDofs(int,NumPtr<int>&) const;
    void reorderHdivFaceDofs(int,int,NumPtr<int>&) const;
    void reorderHcurlFaceDofs(int,int,int,NumPtr<int>&) const;
    void reorderEquallySpacedLatticePoints(int,int,int,NumPtr<int>&) const;
    void reorderEquallySpacedLatticePoints(int n,NumPtr<int> &renumber)
    const {
      reorderEquallySpacedLatticePoints(n,n,n,renumber);
    }
    void reorderChildEquallySpacedLatticePoints(int,const Shape*,int,
      const NumPtr<int>&,NumPtr<int>&) const;
    virtual void reorderChildDofPoints(int,const Shape*,int,
      const NumPtr<int>&,NumPtr<int>&) const;
    virtual void interiorLatticePointIndices(int n,NumPtr<int> &v) const {
      staticInteriorLatticePointIndices(n,n,n,v);
    }
    void computeInteriorLatticePoints(int,int,int,int,
      NumPtr<Point<SPACEDIM> >&) const;
    virtual void computeInteriorLatticePoints(int d,int bf,
    NumPtr<Point<SPACEDIM> > &msp) const {
      computeInteriorLatticePoints(d,d,d,bf,msp);
    }
    virtual const Shape* reorientFor(const Shape*,int,
      const NumPtr<int>&) const;
    int numberVerticesOnFace(int f) const {
      return 4;
    }
    virtual void initializeMappingSupportPoints(int,
      const NumPtr<Point<SPACEDIM> >&,NumPtr<Point<SPACEDIM> >&) const;
    virtual Point<SPACEDIM> faceCenter(const NumPtr<Point<SPACEDIM> >&,
      int) const;
    virtual int numberInteriorChildCodimension1Shapes() const {
      return 12;
    }
    virtual int elementVertexFromFace(int f,int v) const {
      return staticElementVertexFromFace(f,v);
    }
    virtual int elementVertexFromLine(int f,int l) const {
      return staticElementVertexFromLine(f,l);
    }
    virtual int elementLineFromFace(int f,int fl) const {
      return staticElementLineFromFace(f,fl);
    }
    void interiorChildCodimension1ShapeNeighbors(int i,int &c0,int &c1) 
    const {
      staticInteriorChildCodimension1ShapeNeighbors(i,c0,c1);
    }
    inline int childOnFace(int f,int s) const {
      return staticChildOnFace(f,s);
    }
    virtual int numberInteriorChildVerticesOnFace(int f) const {
      return staticNumberInteriorChildVerticesOnFace(f);
    }
    virtual int equallySpacedLatticePointIndexOfVertex(int qv,int degree) 
    const {
      return staticEquallySpacedLatticePointIndexOfVertex(qv,
        degree,degree,degree);
    }
    void equallySpacedLatticePointIndicesOfFace(int,int,int,int,
      NumPtr<int>&) const;
    virtual void equallySpacedLatticePointIndicesOfFace(int f,int n,
    NumPtr<int> &indices) const {
      equallySpacedLatticePointIndicesOfFace(f,n,n,n,indices);
    }
    void equallySpacedLatticePointIndicesOfFaceInterior(int,int,int,int,
      NumPtr<int>&) const;
    virtual void equallySpacedLatticePointIndicesOfFaceInterior(
    int f,int n,NumPtr<int> &indices) const {
      equallySpacedLatticePointIndicesOfFaceInterior(f,n,n,n,indices);
    }
    void equallySpacedLatticePointIndicesOfLineInterior(int,int,int,int,
      NumPtr<int>&) const;
    virtual void equallySpacedLatticePointIndicesOfLineInterior(
    int l,int n,NumPtr<int> &indices) const {
      equallySpacedLatticePointIndicesOfLineInterior(l,n,n,n,indices);
    }
    virtual void projectToFace(int f,const NumPtr<Point<1> > &fp,
    NumPtr<Point<2> > &ep) const {
      Shape3::projectToFace(f,fp,ep);
    }
    virtual void projectToFace(int f,const NumPtr<Point<2> > &fp,
    NumPtr<Point<3> > &ep) const {
      staticProjectToFace(f,fp,ep);
    }
    virtual void projectToFaceForBDFM(int,const NumPtr<Point<2> >&,
      NumPtr<Point<3> >&) const;
#ifdef DEBUG
    virtual void checkFaceVertices(int,const Shape2*) const;
    virtual void draw(const VolGraphTool::WINDOW_TYPE::COLOR_TYPE&,
      const NumPtr<Point<3> >&,VolGraphTool&) const;
#endif
    void interiorChildVertexOnFace(int f,int &c,int &v) const {
      staticInteriorChildVertexOnFace(f,c,v);
    }
    virtual Shape2* newInteriorCodimension1Shape(int,int,
      const NumPtr<int>&,const NumPtr<int>&,const NumPtr<int>&,
      const map<pair<int,int>,int>&) const;

    virtual int numberCodimension2Shapes() const {
      return 12;
    }
    virtual int numberInteriorChildCodimension2Shapes() const {
      return 6;
    }
    virtual Shape1* newInteriorCodimension2Shape(int i,
    int codimension1_element_index,
    const NumPtr<int> &interior_child_vertex_indices,
    const NumPtr<int> &interior_child_face_vertex_indices,
    const NumPtr<int> &interior_child_line_vertex_indices) const {
      int v[2]={interior_child_face_vertex_indices[i],
                interior_child_vertex_indices[0]};
      return OPERATOR_NEW Interval(codimension1_element_index,v);
    }
    virtual void interiorChildCodimension2ShapeNeighbors(int i,int &c0,
    int &c1) const {
      staticInteriorChildCodimension2ShapeNeighbors(i,c0,c1);
    }

    virtual int numberInteriorChildVertices() const {return 1;}
    virtual Point<SPACEDIM> interiorChildPoint(
    const NumPtr<Point<SPACEDIM> > &vertices,
    const NumPtr<int> &interior_points_of_faces,
    const NumPtr<int> &interior_points_of_element_lines) const {
//    solve 27-point Laplacian for interior child point
      Point<SPACEDIM> pv(0.,0.,0.),pf(0.,0.,0.),pl(0.,0.,0.);
      for (int v=0;v<numberVertices();v++) {
        pv+=vertices[vertexIndex(v)];
      }
      for (int f=0;f<numberCodimension1Shapes();f++) {
        pf+=vertices[interior_points_of_faces[f]];
      }
      for (int l=0;l<staticNumberCodimension2Shapes();l++) {
        pl+=vertices[interior_points_of_element_lines[l]];
      }
      return pv*0.0078125+pl*7/192.+pf/12.;
    }
    inline void interiorChildVertex(int &c,int &v) const {
      c=0;
      v=7;
    }
};
#endif

inline Point<SPACEDIM> canonicalVertex(int nv,int v) {
#if (SPACEDIM==1)
  return Interval::canonicalVertex(v);
#elif (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    return Quadrilateral::canonicalVertex(v);
  } else {
    return Triangle::canonicalVertex(v);
  }
#else
  if (nv==Hexahedron::verticesPerShape()) {
    return Hexahedron::canonicalVertex(v);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    return Tetrahedron::canonicalVertex(v);
  } else {
    return Prism::canonicalVertex(v);
  }
#endif
}

inline int facesPerShape(int nv) {
#if (SPACEDIM==1)
  return Interval::facesPerShape();
#elif (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    return Quadrilateral::facesPerShape();
  } else {
    return Triangle::facesPerShape();
  }
#else
  if (nv==Hexahedron::verticesPerShape()) {
    return Hexahedron::facesPerShape();
  } else if (nv==Tetrahedron::verticesPerShape()) {
    return Tetrahedron::facesPerShape();
  } else {
    return Prism::facesPerShape();
  }
#endif
}
inline int linesPerShape(int nv) {
#if (SPACEDIM==1)
  return Interval::linesPerShape();
#elif (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    return Quadrilateral::linesPerShape();
  } else {
    return Triangle::linesPerShape();
  }
#else
  if (nv==Hexahedron::verticesPerShape()) {
    return Hexahedron::linesPerShape();
  } else if (nv==Tetrahedron::verticesPerShape()) {
    return Tetrahedron::linesPerShape();
  } else {
    return Prism::linesPerShape();
  }
#endif
}
#if (SPACEDIM>1)
inline static int linesPerFace(int nv,int f) {
# if (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    return Quadrilateral::linesPerFace(f);
  } else {
    return Triangle::linesPerFace(f);
  }
# else
  if (nv==Hexahedron::verticesPerShape()) {
    return Hexahedron::linesPerFace(f);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    return Tetrahedron::linesPerFace(f);
  } else {
    return Prism::linesPerFace(f);
  }
# endif
}
#endif
#if (SPACEDIM>1)
inline static int elementLineFromFace(int nv,int f,int l) {
# if (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    return Quadrilateral::staticElementLineFromFace(f,l);
  } else {
    return Triangle::staticElementLineFromFace(f,l);
  }
# else
  if (nv==Hexahedron::verticesPerShape()) {
    return Hexahedron::staticElementLineFromFace(f,l);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    return Tetrahedron::staticElementLineFromFace(f,l);
  } else {
    return Prism::staticElementLineFromFace(f,l);
  }
# endif
}
#endif
#if (SPACEDIM>1)
inline int childOnFace(int nv,int f,int s) {
# if (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    return Quadrilateral::staticChildOnFace(f,s);
  } else {
    return Triangle::staticChildOnFace(f,s);
  }
# else
  if (nv==Hexahedron::verticesPerShape()) {
    return Hexahedron::staticChildOnFace(f,s);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    return Tetrahedron::staticChildOnFace(f,s);
  } else {
    return Prism::staticChildOnFace(f,s);
  }
# endif
}
#endif
#if (SPACEDIM>1)
inline int numberVerticesOnFace(int nv,int f) {
# if (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    return Quadrilateral::staticNumberVerticesOnFace(f);
  } else {
    return Triangle::staticNumberVerticesOnFace(f);
  }
# else
  if (nv==Hexahedron::verticesPerShape()) {
    return Hexahedron::staticNumberVerticesOnFace(f);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    return Tetrahedron::staticNumberVerticesOnFace(f);
  } else {
    return Prism::staticNumberVerticesOnFace(f);
  }
# endif
}
#endif
#if (SPACEDIM>1)
inline int elementVertexFromFace(int nv,int f,int v) {
# if (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    return Quadrilateral::staticElementVertexFromFace(f,v);
  } else {
    return Triangle::staticElementVertexFromFace(f,v);
  }
# else
  if (nv==Hexahedron::verticesPerShape()) {
    return Hexahedron::staticElementVertexFromFace(f,v);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    return Tetrahedron::staticElementVertexFromFace(f,v);
  } else {
    return Prism::staticElementVertexFromFace(f,v);
  }
# endif
}
#endif
inline int numberEquallySpacedLatticePoints(int nv,int n) {
#if (SPACEDIM==1)
  return Interval::staticNumberEquallySpacedLatticePoints(n);
#elif (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    return Quadrilateral::staticNumberEquallySpacedLatticePoints(n);
  } else {
    return Triangle::staticNumberEquallySpacedLatticePoints(n);
  }
#else
  if (nv==Hexahedron::verticesPerShape()) {
    return Hexahedron::staticNumberEquallySpacedLatticePoints(n);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    return Tetrahedron::staticNumberEquallySpacedLatticePoints(n);
  } else {
    return Prism::staticNumberEquallySpacedLatticePoints(n);
  }
#endif
}
#if (SPACEDIM>1)
inline int numberEquallySpacedLatticePointsOnFace(int nv,int f,int n) {
# if (SPACEDIM==2)
  return Interval::staticNumberEquallySpacedLatticePoints(n);
# else
  if (nv==Hexahedron::verticesPerShape()) {
    return Quadrilateral::staticNumberEquallySpacedLatticePoints(n);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    return Triangle::staticNumberEquallySpacedLatticePoints(n);
  } else if (Prism::staticNumberVerticesOnFace(f)==
  Quadrilateral::verticesPerShape()) {
    return Quadrilateral::staticNumberEquallySpacedLatticePoints(n);
  } else {
    return Triangle::staticNumberEquallySpacedLatticePoints(n);
  }
# endif
}
#endif
#if (SPACEDIM>1)
inline int numberInteriorEquallySpacedLatticePointsOnFace(int nv,int f,
int n) {
# if (SPACEDIM==2)
  return Interval::staticNumberInteriorEquallySpacedLatticePoints(n);
# else
  if (nv==Hexahedron::verticesPerShape()) {
    return 
      Quadrilateral::staticNumberInteriorEquallySpacedLatticePoints(n);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    return Triangle::staticNumberInteriorEquallySpacedLatticePoints(n);
  } else if (Prism::staticNumberVerticesOnFace(f)==
  Quadrilateral::verticesPerShape()) {
    return 
      Quadrilateral::staticNumberInteriorEquallySpacedLatticePoints(n);
  } else {
    return Triangle::staticNumberInteriorEquallySpacedLatticePoints(n);
  }
# endif
}
#endif
#if (SPACEDIM>1)
#endif
#if (SPACEDIM==3)
#endif
#if (SPACEDIM>1)
inline void equallySpacedLatticePointIndicesOfFace(int nv,int f,int n,
NumPtr<int> &indices) {
# if (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    Quadrilateral::staticEquallySpacedLatticePointIndicesOfFace(f,n,
      indices);
  } else {
    Triangle::staticEquallySpacedLatticePointIndicesOfFace(f,n,indices);
  }
# else
  if (nv==Hexahedron::verticesPerShape()) {
    Hexahedron::staticEquallySpacedLatticePointIndicesOfFace(f,n,indices);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    Tetrahedron::staticEquallySpacedLatticePointIndicesOfFace(f,n,indices);
  } else {
    Prism::staticEquallySpacedLatticePointIndicesOfFace(f,n,indices);
  }
# endif
}
#endif
#if (SPACEDIM>1)
inline void equallySpacedLatticePointIndicesOfFaceInterior(int nv,int f,
int n,NumPtr<int> &indices) {
# if (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    Quadrilateral::staticEquallySpacedLatticePointIndicesOfFaceInterior(
      f,n,indices);
  } else {
    Triangle::staticEquallySpacedLatticePointIndicesOfFaceInterior(f,n,
      indices);
  }
# else
  if (nv==Hexahedron::verticesPerShape()) {
    Hexahedron::staticEquallySpacedLatticePointIndicesOfFaceInterior(
      f,n,indices);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    Tetrahedron::staticEquallySpacedLatticePointIndicesOfFaceInterior(
      f,n,indices);
  } else {
    Prism::staticEquallySpacedLatticePointIndicesOfFaceInterior(f,n,
      indices);
  }
# endif
}
#endif
#if (SPACEDIM==3)
inline void projectToLine(int nv,int l,const NumPtr<Point<1> > &q_points,
NumPtr<Point<3> > &q_points_line) {
  if (nv==Hexahedron::verticesPerShape()) {
    Hexahedron::staticProjectToLine(l,q_points,q_points_line);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    Tetrahedron::staticProjectToLine(l,q_points,q_points_line);
  } else {
    Prism::staticProjectToLine(l,q_points,q_points_line);
  }
}
#endif
#if (SPACEDIM>1)
inline void projectToFace(int nv,int f,
const NumPtr<Point<SPACEDIM-1> > &q_points,
NumPtr<Point<SPACEDIM> > &q_points_face) {
# if (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    Quadrilateral::staticProjectToFace(f,q_points,q_points_face);
  } else {
    Triangle::staticProjectToFace(f,q_points,q_points_face);
  }
# else
  if (nv==Hexahedron::verticesPerShape()) {
    Hexahedron::staticProjectToFace(f,q_points,q_points_face);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    Tetrahedron::staticProjectToFace(f,q_points,q_points_face);
  } else {
    Prism::staticProjectToFace(f,q_points,q_points_face);
  }
# endif
}
#endif
#if (SPACEDIM>1)
inline void projectToSubface(int nv,int f,int s,
const NumPtr<Point<SPACEDIM-1> > &q_points,
NumPtr<Point<SPACEDIM> > &q_points_subface) {
# if (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    Quadrilateral::projectToSubface(f,s,q_points,q_points_subface);
  } else {
    Triangle::projectToSubface(f,s,q_points,q_points_subface);
  }
# else
  if (nv==Hexahedron::verticesPerShape()) {
    Hexahedron::projectToSubface(f,s,q_points,q_points_subface);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    Tetrahedron::projectToSubface(f,s,q_points,q_points_subface);
  } else {
    Prism::projectToSubface(f,s,q_points,q_points_subface);
  }
# endif
}
#endif
inline void equallySpacedLatticePoints(int nv,int n,
NumPtr<Point<SPACEDIM> >&eslp) {
#if (SPACEDIM==1)
  Interval::staticEquallySpacedLatticePoints(n,eslp);
#elif (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    Quadrilateral::staticEquallySpacedLatticePoints(n,eslp);
  } else {
    Triangle::staticEquallySpacedLatticePoints(n,eslp);
  }
#else
  if (nv==Hexahedron::verticesPerShape()) {
    Hexahedron::staticEquallySpacedLatticePoints(n,eslp);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    Tetrahedron::staticEquallySpacedLatticePoints(n,eslp);
  } else {
    Prism::staticEquallySpacedLatticePoints(n,eslp);
  }
#endif
}
#if (SPACEDIM>1)
inline void equallySpacedLatticePointsForFace(int fnv,int n,
NumPtr<Point<SPACEDIM-1> >&feslp) {
#if (SPACEDIM==2)
  Interval::staticEquallySpacedLatticePoints(n,feslp);
#else
  if (fnv==Quadrilateral::verticesPerShape()) {
    Quadrilateral::staticEquallySpacedLatticePoints(n,feslp);
  } else {
    Triangle::staticEquallySpacedLatticePoints(n,feslp);
  }
#endif
}
#endif
inline Point<SPACEDIM> parentCoordinateOfCanonical(int nv,
const Point<SPACEDIM> &cp,int c) {
#if (SPACEDIM==1)
  return Interval::parentCoordinateOfCanonical(cp,c);
#elif (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    return Quadrilateral::parentCoordinateOfCanonical(cp,c);
  } else {
    return Triangle::parentCoordinateOfCanonical(cp,c);
  }
#else
  if (nv==Hexahedron::verticesPerShape()) {
    return Hexahedron::parentCoordinateOfCanonical(cp,c);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    return Tetrahedron::parentCoordinateOfCanonical(cp,c);
  } else {
    return Prism::parentCoordinateOfCanonical(cp,c);
  }
#endif
}
#if (SPACEDIM>1)
inline Point<SPACEDIM-1> parentCoordinateOfCanonical(int nv,
const Point<SPACEDIM-1> &cp,int c) {
#if (SPACEDIM==2)
  return Interval::parentCoordinateOfCanonical(cp,c);
#else
  if (nv==Quadrilateral::verticesPerShape()) {
    return Quadrilateral::parentCoordinateOfCanonical(cp,c);
  } else {
    return Triangle::parentCoordinateOfCanonical(cp,c);
  }
#endif
}
#endif
#if (SPACEDIM>1)
//  unit outer normal * face measure / canonical face measure
inline static Tensor<1,SPACEDIM> faceNormal(int nv,int f) {
#if (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    return Quadrilateral::faceNormal(f);
  } else {
    return Triangle::faceNormal(f);
  }
#else
  if (nv==Hexahedron::verticesPerShape()) {
    return Hexahedron::faceNormal(f);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    return Tetrahedron::faceNormal(f);
  } else {
    return Prism::faceNormal(f);
  }
#endif
}
#endif
#if (SPACEDIM>=2)
inline Tensor<1,SPACEDIM> lineTangent(int nv,int l) {
# if (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    return Quadrilateral::lineTangent(l);
  } else {
    return Triangle::lineTangent(l);
  }
# else
  if (nv==Hexahedron::verticesPerShape()) {
    return Hexahedron::lineTangent(l);
  } else if (nv==Tetrahedron::verticesPerShape()) {
    return Tetrahedron::lineTangent(l);
  } else {
    return Prism::lineTangent(l);
  }
# endif
}
#endif
#endif
