#ifndef TENSOR_SHAPE_FUNCTION_H
#define TENSOR_SHAPE_FUNCTION_H "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <Matrix.H>
#include <ShapeFunction.H>
//TensorShapeFunction                        requiresShapeToComputeValues=F
//  --> HdivDofTensorShapeFunction
//      --> RTSimplexDofShapeFunction
//          --> BDFMSimplexDofShapeFunction
//      --> RTParallelpipedDofShapeFunction
//      --> BDFMParallelpipedDofShapeFunction
//  --> HdivDualTensorShapeFunction
//      --> RTSimplexDualShapeFunction
//          -->BDFMSimplexDualShapeFunction
//      --> RTParallelpipedDualShapeFunction
//      --> RTPrismDualShapeFunction
//      --> BDFMParallelpipedDualShapeFunction
//      --> BDFMPrismDualShapeFunction
//  --> HdivPrismDofShapeFunction
//      --> RTPrismDofShapeFunction
//      --> BDFMPrismDofShapeFunction
//  --> RTSimplexShapeFunction
//      --> BDFMSimplexShapeFunction
//  --> RTParallelpipedShapeFunction
//  --> RTPrismShapeFunction
//  --> BDFMParallelpipedShapeFunction
//  --> BDFMPrismShapeFunction
//  --> NedelecXShapeFunction2
//  --> NedelecXShapeFunction3
//  --> HcurlDofTensorShapeFunction
//      --> Nedelec1SimplexDofShapeFunction
//      --> Nedelec1ParallelpipedDofShapeFunction
//      --> Nedelec2SimplexDofShapeFunction
//      --> Nedelec2ParallelpipedDofShapeFunction
//  --> HcurlPrismDofShapeFunction
//      --> Nedelec1PrismDofShapeFunction
//      --> Nedelec2PrismDofShapeFunction
//  --> HcurlDualTensorShapeFunction         requiresShapeToComputeValues=T
//      --> Nedelec1SimplexDualShapeFunction
//      --> Nedelec1ParallelpipedDualShapeFunction
//      --> Nedelec1PrismDualShapeFunction
//      --> HcurlDualTensorShapeFunction
//      --> Nedelec2ParallelpipedDualShapeFunction
//      --> Nedelec2PrismDualShapeFunction
//  --> HcurlMultiplierTetrahedronShapeFunction
//  --> HcurlMultiplierPrismShapeFunction
//  --> HcurlMultiplierHexahedronShapeFunction
//  --> Nedelec1SimplexShapeFunction
//  --> Nedelec1ParallelpipedShapeFunction
//  --> Nedelec1PrismShapeFunction
//  --> Nedelec2SimplexShapeFunction
//  --> Nedelec2ParallelpipedShapeFunction
//  --> Nedelec2PrismShapeFunction

template<int dim> class TensorShapeFunction {
  public:
    TensorShapeFunction() {}
    virtual ~TensorShapeFunction() {}
    virtual bool requiresShapeToComputeValues() const { return false; }
    virtual void computeValues(const Point<dim>&,NumPtr<Tensor<1,dim> >&)
      const=0;
    virtual void computeGrads(const Point<dim>&,NumPtr<Tensor<2,dim> >&) 
      const=0;
    virtual void computeGradGrads(const Point<dim>&,
      NumPtr<Tensor<3,dim> >&) const=0;
    virtual int getNumber() const=0;
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const {
      OBSOLETE("should be defined in derived class");
    }
    virtual void nonzeroShapesOnFace(const Shape1*,int,NumPtr<int>&)
    const {
      OBSOLETE("should be defined in derived class");
    }
    virtual void printOn(ostream&) const;
};

class HdivDofTensorShapeFunction : public TensorShapeFunction<SPACEDIM> {
  public:
    HdivDofTensorShapeFunction() {}
    virtual ~HdivDofTensorShapeFunction() {}
#if (SPACEDIM>1)
    virtual void computeFaceValues(const Point<SPACEDIM-1>&,
      NumPtr<REAL>&) const=0;
    virtual void computeFaceGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<1,SPACEDIM-1> >&) const=0;
    virtual void computeFaceGradGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<2,SPACEDIM-1> >&) const=0;
    virtual int getFaceNumber() const=0;
#endif
//  renumber[fe_s] = dof_s:
    virtual void printOn(ostream&) const;
};

class HdivDualTensorShapeFunction : public TensorShapeFunction<SPACEDIM> {
  public:
    HdivDualTensorShapeFunction() {}
    virtual ~HdivDualTensorShapeFunction() {}
    virtual int getNumberInteriorDofs() const=0;
    virtual int getNumberFaceDofs() const=0;
#if (SPACEDIM>1)
    virtual void computeFaceValues(int,const Point<SPACEDIM-1>&,
      NumPtr<REAL>&) const=0;
#endif
    virtual void computeCoefficientsFor(const Shape*) {
      OBSOLETE("should have been defined in derived class");
    }
    virtual void printOn(ostream&) const;
};

#if (SPACEDIM==3)
class HdivPrismDofShapeFunction : public TensorShapeFunction<3> {
  public:
    HdivPrismDofShapeFunction() {}
    virtual ~HdivPrismDofShapeFunction() {}
    virtual void computeTriangleFaceValues(const Point<2>&,NumPtr<REAL>&)
      const=0;
    virtual void computeTriangleFaceGrads(const Point<2>&,
      NumPtr<Tensor<1,2> >&) const=0;
    virtual void computeTriangleFaceGradGrads(const Point<2>&,
      NumPtr<Tensor<2,2> >&) const=0;
    virtual int getTriangleFaceNumber() const=0;
    virtual void computeQuadrilateralFaceValues(const Point<2>&,
      NumPtr<REAL>&) const=0;
    virtual void computeQuadrilateralFaceGrads(const Point<2>&,
      NumPtr<Tensor<1,2> >&) const=0;
    virtual void computeQuadrilateralFaceGradGrads(const Point<2>&,
      NumPtr<Tensor<2,2> >&) const=0;
    virtual int getQuadrilateralFaceNumber() const=0;
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const=0;
    virtual void printOn(ostream&) const;
};
#endif

//basis for Raviart-Thomas shape functions, but not dual to dof polys
class RTSimplexShapeFunction : public TensorShapeFunction<SPACEDIM> {
  protected:
    ShapeFunction<SPACEDIM> *simplex_polys;
#if (SPACEDIM>=2)
# if (SPACEDIM==2)
    HomogeneousLagrangePolynomials2 *homo_polys;
# elif (SPACEDIM==3)
    HomogeneousLagrangePolynomials3 *homo_polys;
# endif
#endif
    int order;
    int n_polys;
  public:
    RTSimplexShapeFunction(int);
    virtual ~RTSimplexShapeFunction() {
      delete simplex_polys; simplex_polys=0;
#if (SPACEDIM>=2)
      delete homo_polys; homo_polys=0;
#endif
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void printOn(ostream&) const;
};

//basis for Raviart-Thomas degree of freedom polynomials
class RTSimplexDofShapeFunction : public HdivDofTensorShapeFunction {
  protected:
#if (SPACEDIM>=2)
    ShapeFunction<SPACEDIM-1> *face_polys;
#endif
    ShapeFunction<SPACEDIM> *interior_polys;
    int order;
    int n_polys;
#if (SPACEDIM>=2)
    int n_face_polys;
#endif
  public:
    RTSimplexDofShapeFunction(int);
    virtual ~RTSimplexDofShapeFunction() {
#if (SPACEDIM>=2)
      delete face_polys; face_polys=0;
#endif
      delete interior_polys; interior_polys=0;
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
#if (SPACEDIM>=2)
    void computeFaceValues(const Point<SPACEDIM-1>&,
      NumPtr<REAL >&) const;
    void computeFaceGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<1,SPACEDIM-1> >&) const;
    void computeFaceGradGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<2,SPACEDIM-1> >&) const;
    int getFaceNumber() const { return n_face_polys; }
#endif
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const;
    void printOn(ostream&) const;
};

class RTSimplexDualShapeFunction : public HdivDualTensorShapeFunction {
  protected:
    RTSimplexDofShapeFunction dof_polys;
    ShapeFunction<SPACEDIM> *simplex_polys;
#if (SPACEDIM>=2)
# if (SPACEDIM==2)
    HomogeneousLagrangePolynomials2 *homo_polys;
# elif (SPACEDIM==3)
    HomogeneousLagrangePolynomials3 *homo_polys;
# endif
#endif
    LAMatrix *soln;
    int order;
    int n_polys;
  public:
    RTSimplexDualShapeFunction(int);
    virtual ~RTSimplexDualShapeFunction() {
      delete simplex_polys; simplex_polys=0;
#if (SPACEDIM>=2)
      delete homo_polys; homo_polys=0;
#endif
      delete soln; soln=0;
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    virtual void shapeFunctionsOrder(const Shape *e,NumPtr<int> &renumber)
    const {
      dof_polys.shapeFunctionsOrder(e,renumber);
    }
    virtual void nonzeroShapesOnFace(const Shape1*,int,NumPtr<int>&)
      const;
    int getNumber() const { return n_polys; }
    int getNumberInteriorDofs() const { return dof_polys.getNumber(); }
    int getNumberFaceDofs() const {
#if (SPACEDIM>=2)
      return dof_polys.getFaceNumber();
#else
      return 1;
#endif
    }
#if (SPACEDIM>1)
    virtual void computeFaceValues(int,const Point<SPACEDIM-1> &p,
    NumPtr<REAL> &v) const {
      dof_polys.computeFaceValues(p,v);
    }
#endif
    void printOn(ostream&) const;
};

#if (SPACEDIM>=2)
//basis for Raviart-Thomas shape functions, but not dual to dof polys
class RTParallelpipedShapeFunction : public TensorShapeFunction<SPACEDIM> {
  protected:
    TensorProductPolynomials<SPACEDIM> *polys[SPACEDIM];
    int order;
    int n_polys;
  public:
    RTParallelpipedShapeFunction(int);
    virtual ~RTParallelpipedShapeFunction() {
      for (int d=0;d<SPACEDIM;d++) {
        delete polys[d]; polys[d]=0;
      }
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM>=2)
//basis for Raviart-Thomas degree of freedom polynomials
class RTParallelpipedDofShapeFunction : public HdivDofTensorShapeFunction {
  protected:
    TensorProductPolynomials<SPACEDIM-1> *face_polys;
    TensorProductPolynomials<SPACEDIM> *interior_polys[SPACEDIM];
    int order;
    int n_polys;
    int n_face_polys;
  public:
    RTParallelpipedDofShapeFunction(int);
    virtual ~RTParallelpipedDofShapeFunction() {
      delete face_polys; face_polys=0;
      for (int d=0;d<SPACEDIM;d++) {
        if (interior_polys[d]) delete interior_polys[d];
        interior_polys[d]=0;
      }
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void computeFaceValues(const Point<SPACEDIM-1>&,
      NumPtr<REAL >&) const;
    void computeFaceGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<1,SPACEDIM-1> >&) const;
    void computeFaceGradGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<2,SPACEDIM-1> >&) const;
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const;
    int getFaceNumber() const { return n_face_polys; }
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM>=2)
class RTParallelpipedDualShapeFunction : 
public HdivDualTensorShapeFunction {
  protected:
    RTParallelpipedDofShapeFunction dof_polys;
    TensorProductPolynomials<SPACEDIM> *polys[SPACEDIM];
    LAMatrix *soln;
    int order;
    int n_polys;
  public:
    RTParallelpipedDualShapeFunction(int);
    virtual ~RTParallelpipedDualShapeFunction() {
      for (int d=0;d<SPACEDIM;d++) {
        delete polys[d]; polys[d]=0;
      }
      delete soln; soln=0;
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    virtual void shapeFunctionsOrder(const Shape *e,NumPtr<int> &renumber)
    const {
      dof_polys.shapeFunctionsOrder(e,renumber);
    }
    virtual void nonzeroShapesOnFace(const Shape1*,int,NumPtr<int>&)
      const;
    int getNumber() const { return n_polys; }
    int getNumberInteriorDofs() const { return dof_polys.getNumber(); }
    int getNumberFaceDofs() const { return dof_polys.getFaceNumber(); }
#if (SPACEDIM>1)
    virtual void computeFaceValues(int,const Point<SPACEDIM-1> &p,
    NumPtr<REAL> &v) const {
      dof_polys.computeFaceValues(p,v);
    }
#endif
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
//basis for Raviart-Thomas shape functions, but not dual to dof polys
class RTPrismShapeFunction : public TensorShapeFunction<SPACEDIM> {
  protected:
    C0LagrangePrismPolynomials *prism_poly_01;
    HomogeneousLagrangePolynomials2 *homo_poly;
    C0LagrangeIntervalPolynomials *mono_poly;
    C0LagrangePrismPolynomials *prism_poly_2;
    int order;
    int n_polys;
  public:
    RTPrismShapeFunction(int);
    virtual ~RTPrismShapeFunction() {
      delete prism_poly_01; prism_poly_01=0;
      delete homo_poly; homo_poly=0;
      delete mono_poly; mono_poly=0;
      delete prism_poly_2; prism_poly_2=0;
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
//basis for Raviart-Thomas degree of freedom polynomials
class RTPrismDofShapeFunction : public HdivPrismDofShapeFunction {
  protected:
    C0LagrangeTrianglePolynomials *triangle_face_polys;
    C0LagrangeQuadrilateralPolynomials *quadrilateral_face_polys;
    C0LagrangePrismPolynomials *interior_polys01;
    C0LagrangePrismPolynomials *interior_polys2;
    int order;
    int n_polys;
    int n_triangle_face_polys;
    int n_quadrilateral_face_polys;
  public:
    RTPrismDofShapeFunction(int);
    virtual ~RTPrismDofShapeFunction() {
      delete triangle_face_polys; triangle_face_polys=0;
      delete quadrilateral_face_polys; quadrilateral_face_polys=0;
      if (interior_polys01) delete interior_polys01; interior_polys01=0;
      if (interior_polys2) delete interior_polys2; interior_polys2=0;
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void computeTriangleFaceValues(const Point<SPACEDIM-1>&,
      NumPtr<REAL>&) const;
    void computeTriangleFaceGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<1,SPACEDIM-1> >&) const;
    void computeTriangleFaceGradGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<2,SPACEDIM-1> >&) const;
    int getTriangleFaceNumber() const { return n_triangle_face_polys; }
    void computeQuadrilateralFaceValues(const Point<SPACEDIM-1>&,
      NumPtr<REAL>&) const;
    void computeQuadrilateralFaceGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<1,SPACEDIM-1> >&) const;
    void computeQuadrilateralFaceGradGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<2,SPACEDIM-1> >&) const;
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const;
    int getQuadrilateralFaceNumber() const { 
      return n_quadrilateral_face_polys; 
    }
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
//basis for Raviart-Thomas shape functions, but not dual to dof polys
class RTPrismDualShapeFunction : public HdivDualTensorShapeFunction {
  protected:
    RTPrismDofShapeFunction dof_polys;
    C0LagrangePrismPolynomials *prism_poly_01;
    HomogeneousLagrangePolynomials2 *homo_poly;
    C0LagrangeIntervalPolynomials *mono_poly;
    C0LagrangePrismPolynomials *prism_poly_2;
    LAMatrix *soln;
    int order;
    int n_polys;
  public:
    RTPrismDualShapeFunction(int);
    virtual ~RTPrismDualShapeFunction() {
      delete prism_poly_01; prism_poly_01=0;
      delete homo_poly; homo_poly=0;
      delete mono_poly; mono_poly=0;
      delete prism_poly_2; prism_poly_2=0;
      delete soln; soln=0;
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    virtual void shapeFunctionsOrder(const Shape *e,NumPtr<int> &renumber)
    const {
      dof_polys.shapeFunctionsOrder(e,renumber);
    }
    virtual void nonzeroShapesOnFace(const Shape1*,int,NumPtr<int>&)
      const;
    int getNumber() const { return n_polys; }
    int getNumberInteriorDofs() const { return dof_polys.getNumber(); }
    int getNumberFaceDofs() const { 
      OBSOLETE("should not be called");
    }
    virtual void computeFaceValues(int nv,const Point<SPACEDIM-1> &p,
    NumPtr<REAL> &v) const {
      if (nv==3) dof_polys.computeTriangleFaceValues(p,v);
      else dof_polys.computeQuadrilateralFaceValues(p,v);
    }
    void printOn(ostream&) const;
};
#endif

//basis for Raviart-Thomas shape functions, but not dual to dof polys
class BDFMSimplexShapeFunction : public RTSimplexShapeFunction {
  public:
    BDFMSimplexShapeFunction(int degree) : RTSimplexShapeFunction(degree) {}
    virtual ~BDFMSimplexShapeFunction() {}
    void printOn(ostream&) const;
};

//basis for Raviart-Thomas degree of freedom polynomials
class BDFMSimplexDofShapeFunction : public RTSimplexDofShapeFunction {
  public:
    BDFMSimplexDofShapeFunction(int degree) : 
      RTSimplexDofShapeFunction(degree) {}
    virtual ~BDFMSimplexDofShapeFunction() {}
    void printOn(ostream&) const;
};

class BDFMSimplexDualShapeFunction : public RTSimplexDualShapeFunction {
  public:
    BDFMSimplexDualShapeFunction(int degree) : 
      RTSimplexDualShapeFunction(degree) {}
    virtual ~BDFMSimplexDualShapeFunction() {}
    void printOn(ostream&) const;
};

#if (SPACEDIM>=2)
//Brezzi-Douglas-Fortin-Marini shape functions, but not dual to dof polys
class BDFMParallelpipedShapeFunction : 
public TensorShapeFunction<SPACEDIM> {
  protected:
    ShapeFunction<SPACEDIM> *polys;
# if (SPACEDIM==2)
    HomogeneousLagrangePolynomials2 *homo_polys;
# elif (SPACEDIM==3)
    HomogeneousLagrangePolynomials3 *homo_polys;
# endif
    int order;
    int n_polys;
  public:
    BDFMParallelpipedShapeFunction(int);
    virtual ~BDFMParallelpipedShapeFunction() {
      delete polys; polys=0;
      delete homo_polys; homo_polys=0;
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM>=2)
//basis for Raviart-Thomas degree of freedom polynomials
class BDFMParallelpipedDofShapeFunction : 
public HdivDofTensorShapeFunction {
  protected:
    ShapeFunction<SPACEDIM-1> *face_polys;
    ShapeFunction<SPACEDIM> *interior_polys;
    int order;
    int n_polys;
    int n_face_polys;
  public:
    BDFMParallelpipedDofShapeFunction(int);
    virtual ~BDFMParallelpipedDofShapeFunction() {
      delete face_polys; face_polys=0;
      if (interior_polys) delete interior_polys; interior_polys=0;
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void computeFaceValues(const Point<SPACEDIM-1>&,
      NumPtr<REAL >&) const;
    void computeFaceGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<1,SPACEDIM-1> >&) const;
    void computeFaceGradGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<2,SPACEDIM-1> >&) const;
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const;
    int getFaceNumber() const { return n_face_polys; }
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM>=2)
class BDFMParallelpipedDualShapeFunction : 
public HdivDualTensorShapeFunction {
  protected:
    BDFMParallelpipedDofShapeFunction dof_polys;
    ShapeFunction<SPACEDIM> *polys;
# if (SPACEDIM==2)
    HomogeneousLagrangePolynomials2 *homo_polys;
# elif (SPACEDIM==3)
    HomogeneousLagrangePolynomials3 *homo_polys;
# endif
    LAMatrix *soln;
    int order;
    int n_polys;
  public:
    BDFMParallelpipedDualShapeFunction(int);
    virtual ~BDFMParallelpipedDualShapeFunction() {
      delete polys; polys=0;
      delete homo_polys; homo_polys=0;
      if (soln) delete soln; soln=0;
    }
    virtual bool requiresShapeToComputeValues() const { return true; }
    virtual void computeCoefficientsFor(const Shape*);
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    virtual void shapeFunctionsOrder(const Shape *e,NumPtr<int> &renumber)
    const {
      dof_polys.shapeFunctionsOrder(e,renumber);
    }
    virtual void nonzeroShapesOnFace(const Shape1*,int,NumPtr<int>&)
      const;
    int getNumber() const { return n_polys; }
    int getNumberInteriorDofs() const { return dof_polys.getNumber(); }
    int getNumberFaceDofs() const { return dof_polys.getFaceNumber(); }
    virtual void computeFaceValues(int,const Point<SPACEDIM-1> &p,
    NumPtr<REAL> &v) const {
      dof_polys.computeFaceValues(p,v);
    }
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
//basis for Raviart-Thomas shape functions, but not dual to dof polys
class BDFMPrismShapeFunction : public TensorShapeFunction<SPACEDIM> {
  protected:
    C0LagrangeTetrahedronPolynomials *poly;
    HomogeneousLagrangePolynomials3 *homo_poly;
    int order;
    int n_polys;
  public:
    BDFMPrismShapeFunction(int);
    virtual ~BDFMPrismShapeFunction() {
      delete poly; poly=0;
      delete homo_poly; homo_poly=0;
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
//basis for Raviart-Thomas degree of freedom polynomials
class BDFMPrismDofShapeFunction : public HdivPrismDofShapeFunction {
  protected:
    C0LagrangeTrianglePolynomials *face_polys;
    C0LagrangeTetrahedronPolynomials *interior_polys;
    int order;
    int n_polys;
    int n_face_polys;
  public:
    BDFMPrismDofShapeFunction(int);
    virtual ~BDFMPrismDofShapeFunction() {
      delete face_polys; face_polys=0;
      if (interior_polys) delete interior_polys; interior_polys=0;
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void computeTriangleFaceValues(const Point<SPACEDIM-1>&,
      NumPtr<REAL>&) const;
    void computeTriangleFaceGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<1,SPACEDIM-1> >&) const;
    void computeTriangleFaceGradGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<2,SPACEDIM-1> >&) const;
    int getTriangleFaceNumber() const { return n_face_polys; }
    void computeQuadrilateralFaceValues(const Point<SPACEDIM-1>&,
      NumPtr<REAL>&) const;
    void computeQuadrilateralFaceGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<1,SPACEDIM-1> >&) const;
    void computeQuadrilateralFaceGradGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<2,SPACEDIM-1> >&) const;
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const;
    int getQuadrilateralFaceNumber() const { return n_face_polys; }
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
//basis for Raviart-Thomas shape functions, but not dual to dof polys
class BDFMPrismDualShapeFunction : public HdivDualTensorShapeFunction {
  protected:
    BDFMPrismDofShapeFunction dof_polys;
    C0LagrangeTetrahedronPolynomials *poly;
    HomogeneousLagrangePolynomials3 *homo_poly;
    LAMatrix *soln;
    int order;
    int n_polys;
  public:
    BDFMPrismDualShapeFunction(int);
    virtual ~BDFMPrismDualShapeFunction() {
      delete poly; poly=0;
      delete homo_poly; homo_poly=0;
      if (soln) delete soln; soln=0;
    }
    virtual bool requiresShapeToComputeValues() const { return true; }
    virtual void computeCoefficientsFor(const Shape*);
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    virtual void shapeFunctionsOrder(const Shape *e,NumPtr<int> &renumber)
    const {
      dof_polys.shapeFunctionsOrder(e,renumber);
    }
    virtual void nonzeroShapesOnFace(const Shape1*,int,NumPtr<int>&)
      const;
    int getNumber() const { return n_polys; }
    int getNumberInteriorDofs() const { return dof_polys.getNumber(); }
    int getNumberFaceDofs() const {
      OBSOLETE("should not be called");
    }
    virtual void computeFaceValues(int nv,const Point<SPACEDIM-1> &p,
    NumPtr<REAL> &v) const {
      if (nv==3) dof_polys.computeTriangleFaceValues(p,v);
      else dof_polys.computeQuadrilateralFaceValues(p,v);
    }
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM>=2)
class NedelecXShapeFunction2 : public TensorShapeFunction<2> {
  protected:
    HomogeneousLagrangePolynomials2 *homo_poly;
    int order;
    int n_polys;
  public:
    NedelecXShapeFunction2(int);
    virtual ~NedelecXShapeFunction2() {
      delete homo_poly; homo_poly=0;
    }
    virtual void computeValues(const Point<2>&,
      NumPtr<Tensor<1,2> >&) const;
    virtual void computeGrads(const Point<2>&,
      NumPtr<Tensor<2,2> >&) const;
    virtual void computeGradGrads(const Point<2>&,
      NumPtr<Tensor<3,2> >&) const;
    virtual int getNumber() const { return n_polys; }
    virtual void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
class NedelecXShapeFunction3 : public TensorShapeFunction<3> {
  protected:
    HomogeneousLagrangePolynomials3 *homo_poly;
    HomogeneousLagrangePolynomials2 *homo_poly_01;
    int order;
    int n_polys;
  public:
    NedelecXShapeFunction3(int);
    virtual ~NedelecXShapeFunction3() {
      delete homo_poly; homo_poly=0;
      delete homo_poly_01; homo_poly_01=0;
    }
    virtual void computeValues(const Point<3>&,
      NumPtr<Tensor<1,3> >&) const;
    virtual void computeGrads(const Point<3>&,
      NumPtr<Tensor<2,3> >&) const;
    virtual void computeGradGrads(const Point<3>&,
      NumPtr<Tensor<3,3> >&) const;
    virtual int getNumber() const { return n_polys; }
    virtual void printOn(ostream&) const;
};
#endif

class HcurlDofTensorShapeFunction : public TensorShapeFunction<SPACEDIM> {
  public:
#if (SPACEDIM>1)
    virtual void computeEdgeValues(const Point<1>&,NumPtr<REAL>&) const=0;
    virtual void computeEdgeGrads(const Point<1>&,
      NumPtr<Tensor<1,1> >&) const=0;
    virtual void computeEdgeGradGrads(const Point<1>&,
      NumPtr<Tensor<2,1> >&) const=0;
#endif
#if (SPACEDIM==3)
    virtual void computeFaceValues(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<1,2> >&) const=0;
    virtual void computeFaceGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<2,2> >&) const=0;
    virtual void computeFaceGradGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<3,2> >&) const=0;
#endif
  public:
    HcurlDofTensorShapeFunction() {}
    virtual ~HcurlDofTensorShapeFunction() {}
#if (SPACEDIM>1)
    virtual int getEdgeNumber() const=0;
#endif
#if (SPACEDIM==3)
    virtual int getFaceNumber() const=0;
#endif
    virtual void printOn(ostream&) const;
};

#if (SPACEDIM==3)
class HcurlPrismDofShapeFunction : public TensorShapeFunction<3> {
  public:
    virtual void computeEdgeValues(const Point<1>&,NumPtr<REAL>&) const=0;
    virtual void computeEdgeGrads(const Point<1>&,
      NumPtr<Tensor<1,1> >&) const=0;
    virtual void computeEdgeGradGrads(const Point<1>&,
      NumPtr<Tensor<2,1> >&) const=0;
    virtual void computeFaceValues(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<1,2> >&) const=0;
    virtual void computeFaceGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<2,2> >&) const=0;
    virtual void computeFaceGradGrads(const Point<SPACEDIM-1>&,
      NumPtr<Tensor<3,2> >&) const=0;
  public:
    HcurlPrismDofShapeFunction() {}
    virtual ~HcurlPrismDofShapeFunction() {}
    virtual int getEdgeNumber() const=0;
    virtual void computeTriangleFaceValues(const Point<2>&,
      NumPtr<Tensor<1,2> >&) const=0;
    virtual void computeTriangleFaceGrads(const Point<2>&,
      NumPtr<Tensor<2,2> >&) const=0;
    virtual void computeTriangleFaceGradGrads(const Point<2>&,
      NumPtr<Tensor<3,2> >&) const=0;
    virtual int getTriangleFaceNumber() const=0;
    virtual void computeQuadrilateralFaceValues(const Point<2>&,
      NumPtr<Tensor<1,2> >&) const=0;
    virtual void computeQuadrilateralFaceGrads(const Point<2>&,
      NumPtr<Tensor<2,2> >&) const=0;
    virtual void computeQuadrilateralFaceGradGrads(const Point<2>&,
      NumPtr<Tensor<3,2> >&) const=0;
    virtual int getQuadrilateralFaceNumber() const=0;
    virtual void printOn(ostream&) const;
};
#endif

class HcurlDualTensorShapeFunction : public TensorShapeFunction<SPACEDIM> {
  protected:
  public:
    HcurlDualTensorShapeFunction() {}
    virtual ~HcurlDualTensorShapeFunction() {}
    virtual bool requiresShapeToComputeValues() const { return true; }
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const;
    virtual void computeCoefficientsFor(const Shape*)=0;
    virtual int getNumberInteriorDofs() const=0;
    virtual void computeValues(const Point<SPACEDIM>&,
    NumPtr<Tensor<1,SPACEDIM> >&) const {
      OBSOLETE("should be defined in derived class");
    }
    virtual void computeGrads(const Point<SPACEDIM>&,
    NumPtr<Tensor<2,SPACEDIM> >&) const {
      OBSOLETE("should be defined in derived class");
    }
    virtual void computeGradGrads(const Point<SPACEDIM>&,
    NumPtr<Tensor<3,SPACEDIM> >&) const {
      OBSOLETE("should be defined in derived class");
    }
    virtual int getNumberEdgeDofs() const=0;
#if (SPACEDIM>1)
    virtual void computeEdgeValues(const Point<1>&,NumPtr<REAL>&) const=0;
#endif
#if (SPACEDIM==3)
    virtual int getNumberFaceDofs() const=0;
    virtual void computeFaceValues(int,const Point<2>&,
      NumPtr<Tensor<1,2> >&) const=0;
#endif
    virtual void printOn(ostream&) const;
};

#if (SPACEDIM==3)
class HcurlMultiplierTetrahedronShapeFunction : 
public TensorShapeFunction<3> {
  private:
    int order;
    int n_polys;
    C0LagrangeTetrahedronPolynomials *polys;
  public:
    HcurlMultiplierTetrahedronShapeFunction(int ord) : order(ord) {
      polys=OPERATOR_NEW C0LagrangeTetrahedronPolynomials(ord);
      n_polys=3*polys->getNumber();
    }
    virtual ~HcurlMultiplierTetrahedronShapeFunction() {
      delete polys; polys=0;
    }
    virtual void computeValues(const Point<3>&,NumPtr<Tensor<1,3> >&)
      const;
    virtual void computeGrads(const Point<3>&,NumPtr<Tensor<2,3> >&) 
      const;
    virtual void computeGradGrads(const Point<3>&,
      NumPtr<Tensor<3,3> >&) const;
    virtual int getNumber() const { return n_polys; }
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const {
      OBSOLETE("not programmed");
    }
    virtual void nonzeroShapesOnFace(const Shape1*,int,NumPtr<int>&)
    const {
      OBSOLETE("not programmed");
    }
    virtual void printOn(ostream&) const;
};

class HcurlMultiplierPrismShapeFunction : public TensorShapeFunction<3> {
  private:
    int order01,order2;
    int n_polys;
    C0LagrangePrismPolynomials *polys01;
    C0LagrangePrismPolynomials *polys2;
  public:
    HcurlMultiplierPrismShapeFunction(int ord01,int ord2) :
    order01(ord01),order2(ord2) {
      polys01=OPERATOR_NEW C0LagrangePrismPolynomials(ord01);
      polys2=OPERATOR_NEW C0LagrangePrismPolynomials(ord01-1,ord2);
      n_polys=2*polys01->getNumber()+polys2->getNumber();
    }
    virtual ~HcurlMultiplierPrismShapeFunction() {
      delete polys01; polys01=0;
      delete polys2; polys2=0;
    }
    virtual void computeValues(const Point<3>&,NumPtr<Tensor<1,3> >&)
      const;
    virtual void computeGrads(const Point<3>&,NumPtr<Tensor<2,3> >&) 
      const;
    virtual void computeGradGrads(const Point<3>&,
      NumPtr<Tensor<3,3> >&) const;
    virtual int getNumber() const { return n_polys; }
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const {
      OBSOLETE("not programmed");
    }
    virtual void nonzeroShapesOnFace(const Shape1*,int,NumPtr<int>&)
    const {
      OBSOLETE("not programmed");
    }
    virtual void printOn(ostream&) const;
};

class HcurlMultiplierHexahedronShapeFunction : 
public TensorShapeFunction<3> {
  private:
    int ordera,orderb;
    int n_polys;
    C0LagrangeHexahedronPolynomials *polys[3];
  public:
    HcurlMultiplierHexahedronShapeFunction(int orda,int ordb) :
    ordera(orda),orderb(ordb) {
      polys[0]=
        OPERATOR_NEW C0LagrangeHexahedronPolynomials(ordb,orda,orda);
      polys[1]=
        OPERATOR_NEW C0LagrangeHexahedronPolynomials(orda,ordb,orda);
      polys[2]=
        OPERATOR_NEW C0LagrangeHexahedronPolynomials(orda,orda,ordb);
      n_polys=
        polys[0]->getNumber()+polys[1]->getNumber()+polys[2]->getNumber();
    }
    virtual ~HcurlMultiplierHexahedronShapeFunction() {
      for (int d=0;d<3;d++) {
        delete polys[d]; polys[d]=0;
      }
    }
    virtual void computeValues(const Point<3>&,NumPtr<Tensor<1,3> >&)
      const;
    virtual void computeGrads(const Point<3>&,NumPtr<Tensor<2,3> >&) 
      const;
    virtual void computeGradGrads(const Point<3>&,
      NumPtr<Tensor<3,3> >&) const;
    virtual int getNumber() const { return n_polys; }
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const {
      OBSOLETE("not programmed");
    }
    virtual void nonzeroShapesOnFace(const Shape1*,int,NumPtr<int>&)
    const {
      OBSOLETE("not programmed");
    }
    virtual void printOn(ostream&) const;
};
#endif

//basis for Nedelec_1 shape functions, but not dual to dof polys
class Nedelec1SimplexShapeFunction : public TensorShapeFunction<SPACEDIM> {
  protected:
    ShapeFunction<SPACEDIM> *simplex_polys;
#if (SPACEDIM>1)
    D_CHOOSE(void,NedelecXShapeFunction2,NedelecXShapeFunction3) *x_polys;
#endif
    int order;
    int n_polys;
  public:
    Nedelec1SimplexShapeFunction(int);
    virtual ~Nedelec1SimplexShapeFunction() {
      delete simplex_polys; simplex_polys=0;
#if (SPACEDIM>1)
      delete x_polys; x_polys=0;
#endif
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void printOn(ostream&) const;
};

//basis for Nedelec_1 degree of freedom polynomials
class Nedelec1SimplexDofShapeFunction : 
public HcurlDofTensorShapeFunction {
  protected:
#if (SPACEDIM>1)
    C0LagrangeIntervalPolynomials *edge_polys;
#endif
#if (SPACEDIM==3)
    ShapeFunction<2> *face_polys;
#endif
    ShapeFunction<SPACEDIM> *interior_polys;
    int order;
    int n_polys;
    int n_edge_polys;
#if (SPACEDIM==3)
    int n_face_polys;
#endif
  public:
    Nedelec1SimplexDofShapeFunction(int);
    virtual ~Nedelec1SimplexDofShapeFunction() {
#if (SPACEDIM>1)
      delete edge_polys; edge_polys=0;
#endif
#if (SPACEDIM==3)
      delete face_polys; face_polys=0;
#endif
      delete interior_polys; interior_polys=0;
    }
#if (SPACEDIM>1)
    void computeEdgeValues(const Point<1>&,NumPtr<REAL >&) const;
    void computeEdgeGrads(const Point<1>&,NumPtr<Tensor<1,1> >&) const;
    void computeEdgeGradGrads(const Point<1>&,NumPtr<Tensor<2,1> >&)
      const;
    int getEdgeNumber() const { return n_edge_polys; }
#endif
#if (SPACEDIM==3)
    void computeFaceValues(const Point<2>&,NumPtr<Tensor<1,2> >&) const;
    void computeFaceGrads(const Point<2>&,NumPtr<Tensor<2,2> >&) const;
    void computeFaceGradGrads(const Point<2>&,NumPtr<Tensor<3,2> >&)
      const;
    int getFaceNumber() const { return n_face_polys; }
#endif
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const;
    void printOn(ostream&) const;
};

class Nedelec1SimplexDualShapeFunction : 
public HcurlDualTensorShapeFunction {
  protected:
    Nedelec1SimplexDofShapeFunction dof_polys;
    ShapeFunction<SPACEDIM> *simplex_polys;
#if (SPACEDIM>1)
    D_CHOOSE(void,NedelecXShapeFunction2,NedelecXShapeFunction3) *x_polys;
#endif
    LAMatrix *soln;
    int order;
    int n_polys;
    bool soln_ok;
  public:
    Nedelec1SimplexDualShapeFunction(int);
    virtual ~Nedelec1SimplexDualShapeFunction() {
      delete simplex_polys; simplex_polys=0;
#if (SPACEDIM>1)
      delete x_polys; x_polys=0;
#endif
      if (soln) delete soln; soln=0;
    }
    virtual void computeCoefficientsFor(const Shape*);
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
#if (SPACEDIM>1)
    void computeEdgeValues(const Point<1> &p,NumPtr<REAL > &v) const {
      dof_polys.computeEdgeValues(p,v);
    }
#endif
    virtual void shapeFunctionsOrder(const Shape *e,NumPtr<int> &renumber)
    const {
      dof_polys.shapeFunctionsOrder(e,renumber);
    }
    int getNumber() const { return n_polys; }
    int getNumberInteriorDofs() const { return dof_polys.getNumber(); }
    int getNumberEdgeDofs() const { 
#if (SPACEDIM>=2)
      return dof_polys.getEdgeNumber();
#else
      return 1;
#endif
    }
#if (SPACEDIM==3)
    int getNumberFaceDofs() const { return dof_polys.getFaceNumber(); }
    virtual void computeFaceValues(int,const Point<SPACEDIM-1> &p,
    NumPtr<Tensor<1,2> > &v) const {
      dof_polys.computeFaceValues(p,v);
    }
#endif
    void printOn(ostream&) const;
};

#if (SPACEDIM>=2)
class Nedelec1ParallelpipedShapeFunction : 
public TensorShapeFunction<SPACEDIM> {
  protected:
    TensorProductPolynomials<SPACEDIM> *polys[SPACEDIM];
    int order;
    int n_polys;
  public:
    Nedelec1ParallelpipedShapeFunction(int);
    virtual ~Nedelec1ParallelpipedShapeFunction() {
      for (int d=0;d<SPACEDIM;d++) {
        delete polys[d]; polys[d]=0;
      }
    }
    void computeValues(const Point<SPACEDIM>&,
      NumPtr<Tensor<1,SPACEDIM> >&) const;
    void computeGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<2,SPACEDIM> >&) const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM>=2)
//basis for Raviart-Thomas degree of freedom polynomials
class Nedelec1ParallelpipedDofShapeFunction : 
public HcurlDofTensorShapeFunction {
  protected:
    C0LagrangeIntervalPolynomials *edge_polys;
# if (SPACEDIM==3)
    TensorProductPolynomials<2> *face_polys[2];
# endif
    TensorProductPolynomials<SPACEDIM> *interior_polys[SPACEDIM];
    int order;
    int n_polys;
    int n_edge_polys;
# if (SPACEDIM==3)
    int n_face_polys;
# endif
  public:
    Nedelec1ParallelpipedDofShapeFunction(int);
    virtual ~Nedelec1ParallelpipedDofShapeFunction() {
      delete edge_polys; edge_polys=0;
# if (SPACEDIM==3)
      if (face_polys[0]) delete face_polys[0]; face_polys[0]=0;
      if (face_polys[1]) delete face_polys[1]; face_polys[1]=0;
# endif
      for (int d=0;d<SPACEDIM;d++) {
        if (interior_polys[d]) delete interior_polys[d];
        interior_polys[d]=0;
      }
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void computeEdgeValues(const Point<1>&,NumPtr<REAL >&) const;
    void computeEdgeGrads(const Point<1>&,NumPtr<Tensor<1,1> >&) const;
    void computeEdgeGradGrads(const Point<1>&,NumPtr<Tensor<2,1> >&)
      const;
    int getEdgeNumber() const { return n_edge_polys; }
# if (SPACEDIM==3)
    void computeFaceValues(const Point<2>&,NumPtr<Tensor<1,2> >&) const;
    void computeFaceGrads(const Point<2>&,NumPtr<Tensor<2,2> >&) const;
    void computeFaceGradGrads(const Point<2>&,NumPtr<Tensor<3,2> >&)
      const;
    int getFaceNumber() const { return n_face_polys; }
# endif
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const;
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM>=2)
class Nedelec1ParallelpipedDualShapeFunction : 
public HcurlDualTensorShapeFunction {
  protected:
    Nedelec1ParallelpipedDofShapeFunction dof_polys;
    TensorProductPolynomials<SPACEDIM> *polys[SPACEDIM];
    LAMatrix *soln;
    int order;
    int n_polys;
    bool soln_ok;
  public:
    Nedelec1ParallelpipedDualShapeFunction(int);
    virtual ~Nedelec1ParallelpipedDualShapeFunction() {
      for (int d=0;d<SPACEDIM;d++) {
        delete polys[d]; polys[d]=0;
      }
      if (soln) delete soln; soln=0;
    }
    virtual void computeCoefficientsFor(const Shape*);
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    void computeEdgeValues(const Point<1> &p,NumPtr<REAL > &v) const {
      dof_polys.computeEdgeValues(p,v);
    }
    virtual void shapeFunctionsOrder(const Shape *e,NumPtr<int> &renumber)
    const {
      dof_polys.shapeFunctionsOrder(e,renumber);
    }
    int getNumber() const { return n_polys; }
    int getNumberInteriorDofs() const { return dof_polys.getNumber(); }
    int getNumberEdgeDofs() const { return dof_polys.getEdgeNumber(); }
# if (SPACEDIM==3)
    int getNumberFaceDofs() const { return dof_polys.getFaceNumber(); }
    virtual void computeFaceValues(int,const Point<2> &p,
    NumPtr<Tensor<1,2> > &v) const {
      dof_polys.computeFaceValues(p,v);
    }
# endif
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
class Nedelec1PrismShapeFunction : public TensorShapeFunction<3> {
  protected:
    C0LagrangePrismPolynomials *prism_poly_01;
    C0LagrangePrismPolynomials *prism_poly_2;
    HomogeneousLagrangePolynomials2 *homo_poly;
    C0LagrangeIntervalPolynomials *mono_poly;
    int order;
    int n_polys;
  public:
    Nedelec1PrismShapeFunction(int);
    virtual ~Nedelec1PrismShapeFunction() {
      delete prism_poly_01; prism_poly_01=0;
      delete homo_poly; homo_poly=0;
      delete mono_poly; mono_poly=0;
      delete prism_poly_2; prism_poly_2=0;
    }
    void computeValues(const Point<3>&,NumPtr<Tensor<1,3> >&)
      const;
    void computeGrads(const Point<3>&,NumPtr<Tensor<2,3> >&)
      const;
    void computeGradGrads(const Point<3>&,
      NumPtr<Tensor<3,3> >&) const;
    int getNumber() const { return n_polys; }
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
class Nedelec1PrismDofShapeFunction : 
public HcurlPrismDofShapeFunction {
  protected:
    C0LagrangeIntervalPolynomials *edge_polys;
    C0LagrangeTrianglePolynomials *triangle_face_polys;
    C0LagrangeQuadrilateralPolynomials *quadrilateral_face_polys[2];
    C0LagrangePrismPolynomials *interior_polys01;
    C0LagrangePrismPolynomials *interior_polys2;
    int order;
    int n_polys;
    int n_edge_polys;
    int n_triangle_face_polys;
    int n_quadrilateral_face_polys;

    virtual void computeFaceValues(const Point<2>&,
    NumPtr<Tensor<1,2> >&) const {
      OBSOLETE("inappropriate for this class");
    };
    virtual void computeFaceGrads(const Point<2>&,
    NumPtr<Tensor<2,2> >&) const {
      OBSOLETE("inappropriate for this class");
    };
    virtual void computeFaceGradGrads(const Point<2>&,
    NumPtr<Tensor<3,2> >&) const {
      OBSOLETE("inappropriate for this class");
    };
  public:
    Nedelec1PrismDofShapeFunction(int);
    virtual ~Nedelec1PrismDofShapeFunction() {
      delete edge_polys; edge_polys=0;
      delete triangle_face_polys; triangle_face_polys=0;
      delete quadrilateral_face_polys[0]; quadrilateral_face_polys[0]=0;
      delete quadrilateral_face_polys[1]; quadrilateral_face_polys[1]=0;
      if (interior_polys01) delete interior_polys01; interior_polys01=0;
      if (interior_polys2) delete interior_polys2; interior_polys2=0;
    }
    void computeValues(const Point<3>&,NumPtr<Tensor<1,3> >&) const;
    void computeGrads(const Point<3>&,NumPtr<Tensor<2,3> >&) const;
    void computeGradGrads(const Point<3>&,NumPtr<Tensor<3,3> >&) const;
    int getNumber() const { return n_polys; }
    void computeEdgeValues(const Point<1>&,NumPtr<REAL>&) const;
    void computeEdgeGrads(const Point<1>&,NumPtr<Tensor<1,1> >&) const;
    void computeEdgeGradGrads(const Point<1>&,NumPtr<Tensor<2,1> >&) const;
    int getEdgeNumber() const { return n_edge_polys; }
    void computeTriangleFaceValues(const Point<2>&,NumPtr<Tensor<1,2> >&)
      const;
    void computeTriangleFaceGrads(const Point<2>&,NumPtr<Tensor<2,2> >&)
      const;
    void computeTriangleFaceGradGrads(const Point<2>&,
      NumPtr<Tensor<3,2> >&) const;
    int getTriangleFaceNumber() const { return n_triangle_face_polys; }
    void computeQuadrilateralFaceValues(const Point<2>&,
      NumPtr<Tensor<1,2> >&) const;
    void computeQuadrilateralFaceGrads(const Point<2>&,
      NumPtr<Tensor<2,2> >&) const;
    void computeQuadrilateralFaceGradGrads(const Point<2>&,
      NumPtr<Tensor<3,2> >&) const;
    int getQuadrilateralFaceNumber() const { 
      return n_quadrilateral_face_polys; 
    }
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const;
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
class Nedelec1PrismDualShapeFunction : 
public HcurlDualTensorShapeFunction {
  protected:
    Nedelec1PrismDofShapeFunction dof_polys;
    C0LagrangePrismPolynomials *prism_poly_01;
    C0LagrangePrismPolynomials *prism_poly_2;
    HomogeneousLagrangePolynomials2 *homo_poly;
    C0LagrangeIntervalPolynomials *mono_poly;
    LAMatrix *soln;
    int order;
    int n_polys;
  public:
    Nedelec1PrismDualShapeFunction(int);
    virtual ~Nedelec1PrismDualShapeFunction() {
      delete prism_poly_01; prism_poly_01=0;
      delete prism_poly_2; prism_poly_2=0;
      delete homo_poly; homo_poly=0;
      delete mono_poly; mono_poly=0;
      if (soln) delete soln; soln=0;
    }
    void computeCoefficientsFor(const Shape*);
    void computeValues(const Point<3>&,NumPtr<Tensor<1,3> >&) const;
    void computeGrads(const Point<3>&,NumPtr<Tensor<2,3> >&) const;
    void computeGradGrads(const Point<3>&,NumPtr<Tensor<3,3> >&) const;
    void computeEdgeValues(const Point<1> &p,NumPtr<REAL > &v) const {
      dof_polys.computeEdgeValues(p,v);
    }
    virtual void shapeFunctionsOrder(const Shape *e,NumPtr<int> &renumber)
    const {
      dof_polys.shapeFunctionsOrder(e,renumber);
    }
    int getNumber() const { return n_polys; }
    int getNumberInteriorDofs() const { return dof_polys.getNumber(); }
    int getNumberFaceDofs() const { 
      OBSOLETE("should not be called");
    }
    int getNumberEdgeDofs() const { return dof_polys.getEdgeNumber(); }
    virtual void computeFaceValues(int,const Point<SPACEDIM-1>&,
      NumPtr<REAL>&) const;
    virtual void computeFaceValues(int nv,const Point<2> &p,
    NumPtr<Tensor<1,2> > &v) const {
      if (nv==3) dof_polys.computeTriangleFaceValues(p,v);
      else dof_polys.computeQuadrilateralFaceValues(p,v);
    }
    void printOn(ostream&) const;
};
#endif

//basis for Nedelec_2 shape functions, but not dual to dof polys
class Nedelec2SimplexShapeFunction : public TensorShapeFunction<SPACEDIM> {
  protected:
    ShapeFunction<SPACEDIM> *simplex_polys;
    int order;
    int n_polys;
  public:
    Nedelec2SimplexShapeFunction(int);
    virtual ~Nedelec2SimplexShapeFunction() {
      delete simplex_polys; simplex_polys=0;
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void printOn(ostream&) const;
};

//basis for Nedelec_2 degree of freedom polynomials
class Nedelec2SimplexDofShapeFunction : 
public HcurlDofTensorShapeFunction {
  protected:
#if (SPACEDIM>1)
    C0LagrangeIntervalPolynomials *edge_polys;
#endif
#if (SPACEDIM==3)
    C0LagrangeTrianglePolynomials *face_simplex_polys;
    HomogeneousLagrangePolynomials2 *face_homo_polys;
#endif
    ShapeFunction<SPACEDIM> *interior_simplex_polys;
#if (SPACEDIM>1)
    ShapeFunction<SPACEDIM> *interior_homo_polys;
#endif
    int order;
    int n_polys;
#if (SPACEDIM>=2)
    int n_edge_polys;
#endif
#if (SPACEDIM==3)
    int n_face_polys;
#endif
  public:
    Nedelec2SimplexDofShapeFunction(int);
    virtual ~Nedelec2SimplexDofShapeFunction() {
#if (SPACEDIM>1)
      delete edge_polys; edge_polys=0;
#endif
#if (SPACEDIM==3)
      if (face_simplex_polys) delete face_simplex_polys;
        face_simplex_polys=0;
      if (face_homo_polys) delete face_homo_polys; face_homo_polys=0;
#endif
      if (interior_simplex_polys) delete interior_simplex_polys;
        interior_simplex_polys=0;
#if (SPACEDIM>1)
      if (interior_homo_polys) delete interior_homo_polys;
        interior_homo_polys=0;
#endif
    }
#if (SPACEDIM>1)
    void computeEdgeValues(const Point<1>&,NumPtr<REAL >&) const;
    void computeEdgeGrads(const Point<1>&,NumPtr<Tensor<1,1> >&) const;
    void computeEdgeGradGrads(const Point<1>&,NumPtr<Tensor<2,1> >&)
      const;
    int getEdgeNumber() const { return n_edge_polys; }
#endif
#if (SPACEDIM==3)
    void computeFaceValues(const Point<2>&,NumPtr<Tensor<1,2> >&) const;
    void computeFaceGrads(const Point<2>&,NumPtr<Tensor<2,2> >&) const;
    void computeFaceGradGrads(const Point<2>&,NumPtr<Tensor<3,2> >&)
      const;
    int getFaceNumber() const { return n_face_polys; }
#endif
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const;
    void printOn(ostream&) const;
};

class Nedelec2SimplexDualShapeFunction : 
public HcurlDualTensorShapeFunction {
  protected:
    Nedelec2SimplexDofShapeFunction dof_polys;
    ShapeFunction<SPACEDIM> *simplex_polys;
    LAMatrix *soln;
    int order;
    int n_polys;
    bool soln_ok;
#if (SPACEDIM==3)
    virtual void computeFaceValues(int,const Point<2> &p,
    NumPtr<Tensor<1,2> > &v) const {
      OBSOLETE("requires shape");
    }
#endif
  public:
    Nedelec2SimplexDualShapeFunction(int);
    virtual ~Nedelec2SimplexDualShapeFunction() {
      delete simplex_polys; simplex_polys=0;
      if (soln) delete soln; soln=0;
    }
    void computeCoefficientsFor(const Shape*);
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
#if (SPACEDIM>1)
    void computeEdgeValues(const Point<1> &p,NumPtr<REAL > &v) const {
      dof_polys.computeEdgeValues(p,v);
    }
#endif
    virtual void shapeFunctionsOrder(const Shape *e,NumPtr<int> &renumber)
    const {
      dof_polys.shapeFunctionsOrder(e,renumber);
    }
    int getNumber() const { return n_polys; }
    int getNumberInteriorDofs() const { return dof_polys.getNumber(); }
    int getNumberEdgeDofs() const {
#if (SPACEDIM>=2)
      return dof_polys.getEdgeNumber();
#else
      return 1;
#endif
    }
#if (SPACEDIM==3)
    int getNumberFaceDofs() const { return dof_polys.getFaceNumber(); }
    void computeFaceValues(const Point<2> &p,NumPtr<Tensor<1,2> > &v)
    const {
      dof_polys.computeFaceValues(p,v);
    }
#endif
    void printOn(ostream&) const;
};

#if (SPACEDIM>=2)
class Nedelec2ParallelpipedShapeFunction : 
public TensorShapeFunction<SPACEDIM> {
  protected:
    TensorProductPolynomials<SPACEDIM> *polys;
    int order;
    int n_polys;
  public:
    Nedelec2ParallelpipedShapeFunction(int);
    virtual ~Nedelec2ParallelpipedShapeFunction() {
      delete polys; polys=0;
    }
    void computeValues(const Point<SPACEDIM>&,
      NumPtr<Tensor<1,SPACEDIM> >&) const;
    void computeGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<2,SPACEDIM> >&) const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM>=2)
//basis for Raviart-Thomas degree of freedom polynomials
class Nedelec2ParallelpipedDofShapeFunction : 
public HcurlDofTensorShapeFunction {
  protected:
    C0LagrangeIntervalPolynomials *edge_polys;
# if (SPACEDIM==3)
    TensorProductPolynomials<2> *face_polys[2];
# endif
    TensorProductPolynomials<SPACEDIM> *interior_polys[SPACEDIM];
    int order;
    int n_polys;
    int n_edge_polys;
# if (SPACEDIM==3)
    int n_face_polys;
# endif
  public:
    Nedelec2ParallelpipedDofShapeFunction(int);
    virtual ~Nedelec2ParallelpipedDofShapeFunction() {
      delete edge_polys; edge_polys=0;
# if (SPACEDIM==3)
      if (face_polys[0]) delete face_polys[0]; face_polys[0]=0;
      if (face_polys[1]) delete face_polys[1]; face_polys[1]=0;
# endif
      for (int d=0;d<SPACEDIM;d++) {
        if (interior_polys[d]) delete interior_polys[d];
        interior_polys[d]=0;
      }
    }
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    int getNumber() const { return n_polys; }
    void computeEdgeValues(const Point<1>&,NumPtr<REAL >&) const;
    void computeEdgeGrads(const Point<1>&,NumPtr<Tensor<1,1> >&) const;
    void computeEdgeGradGrads(const Point<1>&,NumPtr<Tensor<2,1> >&)
      const;
    int getEdgeNumber() const { return n_edge_polys; }
# if (SPACEDIM==3)
    void computeFaceValues(const Point<2>&,NumPtr<Tensor<1,2> >&) const;
    void computeFaceGrads(const Point<2>&,NumPtr<Tensor<2,2> >&) const;
    void computeFaceGradGrads(const Point<2>&,NumPtr<Tensor<3,2> >&)
      const;
    int getFaceNumber() const { return n_face_polys; }
# endif
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const;
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM>=2)
class Nedelec2ParallelpipedDualShapeFunction : 
public HcurlDualTensorShapeFunction {
  protected:
    Nedelec2ParallelpipedDofShapeFunction dof_polys;
    TensorProductPolynomials<SPACEDIM> *polys;
    LAMatrix *soln;
    int order;
    int n_polys;
  public:
    Nedelec2ParallelpipedDualShapeFunction(int);
    virtual ~Nedelec2ParallelpipedDualShapeFunction() {
      delete polys; polys=0;
      if (soln) delete soln; soln=0;
    }
    void computeCoefficientsFor(const Shape*);
    void computeValues(const Point<SPACEDIM>&,NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void computeGrads(const Point<SPACEDIM>&,NumPtr<Tensor<2,SPACEDIM> >&)
      const;
    void computeGradGrads(const Point<SPACEDIM>&,
      NumPtr<Tensor<3,SPACEDIM> >&) const;
    void computeEdgeValues(const Point<1> &p,NumPtr<REAL > &v) const {
      dof_polys.computeEdgeValues(p,v);
    }
    virtual void shapeFunctionsOrder(const Shape *e,NumPtr<int> &renumber)
    const {
      dof_polys.shapeFunctionsOrder(e,renumber);
    }
    int getNumber() const { return n_polys; }
    int getNumberInteriorDofs() const { return dof_polys.getNumber(); }
    int getNumberEdgeDofs() const { return dof_polys.getEdgeNumber(); }
# if (SPACEDIM==3)
    int getNumberFaceDofs() const { return dof_polys.getFaceNumber(); }
    virtual void computeFaceValues(int,const Point<2> &p,
    NumPtr<Tensor<1,2> > &v) const {
      dof_polys.computeFaceValues(p,v);
    }
# endif
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
class Nedelec2PrismShapeFunction : public TensorShapeFunction<3> {
  protected:
    C0LagrangePrismPolynomials *prism_poly;
    int order;
    int n_polys;
  public:
    Nedelec2PrismShapeFunction(int);
    virtual ~Nedelec2PrismShapeFunction() {
      delete prism_poly; prism_poly=0;
    }
    void computeValues(const Point<3>&,NumPtr<Tensor<1,3> >&)
      const;
    void computeGrads(const Point<3>&,NumPtr<Tensor<2,3> >&)
      const;
    void computeGradGrads(const Point<3>&,
      NumPtr<Tensor<3,3> >&) const;
    int getNumber() const { return n_polys; }
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
class Nedelec2PrismDofShapeFunction : 
public HcurlPrismDofShapeFunction {
  protected:
    C0LagrangeIntervalPolynomials *edge_polys;
    C0LagrangeTrianglePolynomials *triangle_face_polys;
    HomogeneousLagrangePolynomials2 *triangle_face_homo_polys;
    C0LagrangeQuadrilateralPolynomials *quadrilateral_face_polys[2];
    C0LagrangeTrianglePolynomials *interior_polys01;
    C0LagrangeIntervalPolynomials *interior_polys01_2;
    HomogeneousLagrangePolynomials2 *interior_homo_polys_01;
    C0LagrangePrismPolynomials *interior_polys2;
    int order;
    int n_polys;
    int n_edge_polys;
    int n_triangle_face_polys;
    int n_quadrilateral_face_polys;

    virtual void computeFaceValues(const Point<2>&,
    NumPtr<Tensor<1,2> >&) const {
      OBSOLETE("inappropriate for this class");
    };
    virtual void computeFaceGrads(const Point<2>&,
    NumPtr<Tensor<2,2> >&) const {
      OBSOLETE("inappropriate for this class");
    };
    virtual void computeFaceGradGrads(const Point<2>&,
    NumPtr<Tensor<3,2> >&) const {
      OBSOLETE("inappropriate for this class");
    };
  public:
    Nedelec2PrismDofShapeFunction(int);
    virtual ~Nedelec2PrismDofShapeFunction() {
      delete edge_polys; edge_polys=0;
      delete triangle_face_polys; triangle_face_polys=0;
      delete triangle_face_homo_polys; triangle_face_homo_polys=0;
      delete quadrilateral_face_polys[0]; quadrilateral_face_polys[0]=0;
      delete quadrilateral_face_polys[1]; quadrilateral_face_polys[1]=0;
      if (interior_polys01) delete interior_polys01; interior_polys01=0;
      if (interior_polys01_2) delete interior_polys01_2;
        interior_polys01_2=0;
      if (interior_homo_polys_01) delete interior_homo_polys_01;
        interior_homo_polys_01=0;
      if (interior_polys2) delete interior_polys2; interior_polys2=0;
    }
    void computeValues(const Point<3>&,NumPtr<Tensor<1,3> >&) const;
    void computeGrads(const Point<3>&,NumPtr<Tensor<2,3> >&) const;
    void computeGradGrads(const Point<3>&,NumPtr<Tensor<3,3> >&) const;
    int getNumber() const { return n_polys; }
    void computeEdgeValues(const Point<1>&,NumPtr<REAL>&) const;
    void computeEdgeGrads(const Point<1>&,NumPtr<Tensor<1,1> >&) const;
    void computeEdgeGradGrads(const Point<1>&,NumPtr<Tensor<2,1> >&) const;
    int getEdgeNumber() const { return n_edge_polys; }
    void computeTriangleFaceValues(const Point<2>&,NumPtr<Tensor<1,2> >&)
      const;
    void computeTriangleFaceGrads(const Point<2>&,NumPtr<Tensor<2,2> >&)
      const;
    void computeTriangleFaceGradGrads(const Point<2>&,
      NumPtr<Tensor<3,2> >&) const;
    int getTriangleFaceNumber() const { return n_triangle_face_polys; }
    void computeQuadrilateralFaceValues(const Point<2>&,
      NumPtr<Tensor<1,2> >&) const;
    void computeQuadrilateralFaceGrads(const Point<2>&,
      NumPtr<Tensor<2,2> >&) const;
    void computeQuadrilateralFaceGradGrads(const Point<2>&,
      NumPtr<Tensor<3,2> >&) const;
    int getQuadrilateralFaceNumber() const { 
      return n_quadrilateral_face_polys; 
    }
//  renumber[fe_s] = dof_s:
    virtual void shapeFunctionsOrder(const Shape*,NumPtr<int>&) const;
    void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
class Nedelec2PrismDualShapeFunction : 
public HcurlDualTensorShapeFunction {
  protected:
    Nedelec2PrismDofShapeFunction dof_polys;
    C0LagrangePrismPolynomials *prism_poly;
    LAMatrix *soln;
    int order;
    int n_polys;
  public:
    Nedelec2PrismDualShapeFunction(int);
    virtual ~Nedelec2PrismDualShapeFunction() {
      delete prism_poly; prism_poly=0;
      if (soln!=0) delete soln; soln=0;
    }
    void computeCoefficientsFor(const Shape*);
    void computeValues(const Point<3>&,NumPtr<Tensor<1,3> >&) const;
    void computeGrads(const Point<3>&,NumPtr<Tensor<2,3> >&) const;
    void computeGradGrads(const Point<3>&,NumPtr<Tensor<3,3> >&) const;
    void computeEdgeValues(const Point<1> &p,NumPtr<REAL > &v) const {
      dof_polys.computeEdgeValues(p,v);
    }
    virtual void shapeFunctionsOrder(const Shape *e,NumPtr<int> &renumber)
    const {
      dof_polys.shapeFunctionsOrder(e,renumber);
    }
    int getNumber() const { return n_polys; }
    int getNumberInteriorDofs() const { return dof_polys.getNumber(); }
    int getNumberFaceDofs() const {
      OBSOLETE("should not be called");
    }
    int getNumberEdgeDofs() const { return dof_polys.getEdgeNumber(); }
    virtual void computeFaceValues(int f,const Point<2> &p,
    NumPtr<Tensor<1,2> > &v) const {
      if (f<2) dof_polys.computeTriangleFaceValues(p,v);
      else dof_polys.computeQuadrilateralFaceValues(p,v);
    }
    void printOn(ostream&) const;
};
#endif

#endif
