// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <cmath>
#include <iostream>
#include <math.h> // for HUGE_VAL,M_PI
#include <stdlib.h>

#include "Debug.H"
#include "MemoryDebugger.H"
#include "Quadrature.H"
#include "SetTraps.H"
#include "Tracer.H"
#include "Types.H"

#include "XYGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKWindow.H"
#else
#include "XTWindow.H"
#endif

void fcn(const int &n,const REAL *x,REAL *fvec,int &iflag) {
  switch (n) {
    case 4: {
      REAL b6=x[0];
      REAL w0=x[1];
      REAL w3=x[2];
      REAL w6=x[3];
      fvec[0]=(w0+3.*w3+6.*w6)*0.5-0.5; // 0,0,0
      fvec[1]=(w0+w6*2.*b6*(1.-b6))*0.5-0.25; // 1,1,0
      fvec[2]=(w0+
             +(c*(1.+c*(-5.+8.*c))-0.0625)*w2-1./960.;
      fvec[3]=(0.015625-4.*b*b*b*(1.-3.*b))*w1
             +(0.015625-4.*c*c*c*(1.-3.*c))*w2-73./26880.;
      if (iflag==0) {
        cout << endl;
        cout << "\tx = " << x[0] << " " << x[1] << " " << x[2] << " "
             << x[3] << endl;
        cout << "\tf = " << fvec[0] << " " << fvec[1] << " " << fvec[2]
             << " " << fvec[3] << endl;
      }
      break;
    }
    case 3: {
      REAL b=x[0];
      REAL c=x[1];
      REAL w=x[2];
      fvec[0]=b*(1.-2.*b)*(0.25-w)+c*(1.-2.*c)*w-.025;
      fvec[1]=b*b*(3.-8.*b)*(0.25-w)+c*c*(3.-8.*c)*w-1./120.;
//    fvec[2]=b*(1.+b*(-5.+8.*b))*(0.25-w)+c*(1.+c*(-5.+8.*c))*w-1./60.;
      fvec[2]=(__cmath_power(1.-3.*b,3)+3.*__cmath_power(b,3))*(0.25-w)
              +(__cmath_power(1.-3.*c,3)+3.*__cmath_power(c,3))*w-.05;
      break;
    }
    case 2: {
      REAL b=x[0];
      REAL w=x[1];
      fvec[0]=4.*w*__cmath_power(1.-4.*b,2)-0.2;
      fvec[1]=4.*w*(1.+b*b*(-48.+128.*b))-7./15.;
      break;
    }
  }
}
extern "C" {
  void F77NAME(hybrd)(
    void (*)(const int&,const REAL*,REAL*,int&),
    const int&,REAL*,REAL*,const REAL&,const int&,const int&,const int&,
    const REAL&,REAL*,const int&,REAL&,const int&,int&,int&,REAL*,
    const int&,REAL*,const int&,REAL*,REAL*,REAL*,REAL*,REAL*);
}

int main(int argc,char* argv[]) {
  cout.precision(16);
//cout << "\tin main" << endl;
#ifdef DEBUG
//setTraps();
#endif
#if MEM_DEBUG
  MemoryDebugger md(1);
#endif
  plot9b();
  int n=2;
  REAL *x=OPERATOR_NEW_BRACKET(REAL,n);
  REAL *fvec=OPERATOR_NEW_BRACKET(REAL,n);
  REAL xtol=sqrt(DBL_EPSILON);
  int maxfev=100;
  int ml=n-1;
  int mu=n-1;
  REAL epsfcn=exp(log(DBL_EPSILON)/3.);
  REAL *diag=OPERATOR_NEW_BRACKET(REAL,n);
  int mode=1;
  REAL factor=100.;
  int nprint=0;
  int info=0;
  int nfev=0;
  REAL *fjac=OPERATOR_NEW_BRACKET(REAL,n*n);
  int ldfjac=n;
  REAL *r=OPERATOR_NEW_BRACKET(REAL,(n*(n+1))/2);
  int lr=(n*(n+1))/2;
  REAL *qtf=OPERATOR_NEW_BRACKET(REAL,n);
  REAL *wa1=OPERATOR_NEW_BRACKET(REAL,n);
  REAL *wa2=OPERATOR_NEW_BRACKET(REAL,n);
  REAL *wa3=OPERATOR_NEW_BRACKET(REAL,n);
  REAL *wa4=OPERATOR_NEW_BRACKET(REAL,n);
//for (int i=0;i<600;i++) {
    x[0]=drand48();
    if (n>1) x[1]=drand48();
    if (n>2) x[2]=drand48();
    if (n>3) x[3]=drand48();
    if (n==4) {
      x[0]=1./3.;
      x[1]=0.125;
      x[2]=0.09;
      x[3]=0.16;
      nprint=1;
    }
    if (n>1) {
      F77NAME(hybrd)(fcn,n,x,fvec,xtol,maxfev,ml,mu,epsfcn,diag,mode,
        factor,nprint,info,nfev,fjac,ldfjac,r,lr,qtf,wa1,wa2,wa3,wa4);
      cout << "\n\ninfo = " << info << "\n\n" << endl;
    }
    switch (n) {
      case 4: {
        if (x[0]>0. && x[0]<1./3. && x[1]>0. && x[1]<1./3. && x[2]>0.
        && x[2]<0.25 && x[3]>0. && x[3]<0.25) {
          epsfcn=DBL_EPSILON;
          F77NAME(hybrd)(fcn,n,x,fvec,xtol,maxfev,ml,mu,epsfcn,diag,
            mode,factor,nprint,info,nfev,fjac,ldfjac,r,lr,qtf,wa1,wa2,
            wa3,wa4);
          if (info==1 || info==2) {
            int iflag=0;
            fcn(n,x,fvec,iflag);
            cout << "\tinfo = " << info << endl;
            cout << "\tb = " << 1.-3.*x[0] << " " << x[0] << endl;
            cout << "\tc = " << 1.-3.*x[1] << " " << x[1] << endl;
            cout << "\tw = " << 0.25-x[2] << " " << x[2] << endl;
            cout << "\tfvec = " << fvec[0] << " " << fvec[1] << " " 
                 << fvec[2] << " " << fvec[3] << endl;
            REAL b2=x[0];
            REAL b1=1.-3.*b2;
            REAL c2=x[1];
            REAL c1=1.-3.*c2;
            REAL w1=x[2];
            REAL w2=x[3];
            REAL w3=1.-4.*(w1+w2);
            cout << "\tobj = " 
                 << ((b1*b1+3.*b2*b2)*w1+(c1*c1+3.*c2*c2)*w2+w3/16.)/6.
                    -1./60. << endl;
          }
          epsfcn=exp(log(DBL_EPSILON)/3.);
        }
        x[0]=0.2841716607157711313;
        x[1]=0.20043760551370549452;
        x[2]=-1.079341461647618871;
        x[3]=1.785252549402819083;
        check9(x[0],x[1],x[2],x[3]);
        break;
      }
      case 3: {
        if (x[0]>0. && x[0]<1./3. && x[1]>0. && x[1]<1./3. && x[2]>0.
        && x[2]<0.25 && (info==1 || info==2)) {
          epsfcn=DBL_EPSILON;
          F77NAME(hybrd)(fcn,n,x,fvec,xtol,maxfev,ml,mu,epsfcn,diag,
            mode,factor,nprint,info,nfev,fjac,ldfjac,r,lr,qtf,wa1,wa2,
            wa3,wa4);
          if (info==1 || info==2) {
            int iflag=0;
            fcn(n,x,fvec,iflag);
            cout << "\tinfo = " << info << endl;
            cout << "\tb = " << 1.-3.*x[0] << " " << x[0] << endl;
            cout << "\tc = " << 1.-3.*x[1] << " " << x[1] << endl;
            cout << "\tw = " << 0.25-x[2] << " " << x[2] << endl;
            cout << "\tfvec = " << fvec[0] << " " << fvec[1] << " " 
                 << fvec[2] << endl;
            cout << "\tobj = " 
                 << ((1.-3.*x[0])*__cmath_power(x[0],3))*(0.25-x[2])
                   +((1.-3.*x[1])*__cmath_power(x[1],3))*x[2]-1./3360.
                 << endl;
          }
          epsfcn=exp(log(DBL_EPSILON)/3.);
        }
        x[0]=1./3.;
        x[1]=0.125;
        x[2]=0.16;
        check8(x[0],x[1],x[2]);
        break;
      }
      case 2: {
        if (x[0]>0. && x[0]<1./3. && x[1]>0. && x[1]<0.25 && 
        (info==1 || info==2)) {
          epsfcn=DBL_EPSILON;
          F77NAME(hybrd)(fcn,n,x,fvec,xtol,maxfev,ml,mu,epsfcn,diag,
            mode,factor,nprint,info,nfev,fjac,ldfjac,r,lr,qtf,wa1,wa2,
            wa3,wa4);
          if (info==1 || info==2) {
            int iflag=0;
            int nn=3;
            REAL xx[3]={x[0],0.,x[1]};
            REAL ff[3];
            fcn(nn,xx,ff,iflag);
            cout << "\tinfo = " << info << endl;
            cout << "\tb = " << 1.-3.*x[0] << " " << x[0] << endl;
            cout << "\tc = " << 1. << " " << 0. << endl;
            cout << "\tw = " << 0.25-x[1] << " " << x[1] << endl;
            cout << "\tfvec = " << ff[0] << " " << ff[1] << " " << ff[2] 
                 << endl;
          }
          epsfcn=exp(log(DBL_EPSILON)/3.);
        }
//      x[0]=(1.-sqrt(.2))*0.25;
//      x[1]=0.25;
//      check4(x[0],x[1]);
//      x[0]=1./6.;
//      x[1]=0.45;
//      check5(x[0],x[1]);
//        x[0]=-0.0854101966249684545;
//        x[1]=0.36180339887498948482;
        x[0]=1.-0.41458980337503154554;
        x[1]=0.13819660112501051518;
        check12(x[0],x[1]);
        break;
      }
      case 1: {
        REAL w=1.;
        check1(w);
        break;
      }
    }
//}
  delete [] wa4;
  delete [] wa3;
  delete [] wa2;
  delete [] wa1;
  delete [] qtf;
  delete [] r;
  delete [] fjac;
  delete [] diag;
  delete [] fvec;
  delete [] x;
  
  return EXIT_SUCCESS;
}
