// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <cmath>
#include <iostream>
#include <math.h> // for HUGE_VAL,M_PI
#include <stdlib.h>

#include "Debug.H"
#include "MemoryDebugger.H"
#include "Quadrature.H"
#include "SetTraps.H"
#include "Tracer.H"
#include "Types.H"

#include "XYGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKWindow.H"
#else
#include "XTWindow.H"
#endif

void check1(REAL w) {
  cout << "check1: w = " << w << endl;
  cout << "\talpha=[0,0,0,0] : " << w/6.-1./6. << endl;
  cout << "\talpha=[1,0,0,0] : " << w/24.-1./24. << endl;
  cout << "\talpha=[1,1,0,0] : " << w/96.-1./120. << endl;
}

void check4(REAL b2,REAL w) {
  REAL b1=1.-3.*b2;
  cout << "check4: b1 = " << b1 
       << "\n\tb2 = " << b2 
       << "\n\tw = " << w << endl;
  cout << "\talpha=[0,0,0,0] : " << 4.*w/6.-1./6. << endl;
  cout << "\talpha=[1,0,0,0] : " << (b1+3.*b2)*w/6.-1./24. << endl;
  cout << "\talpha=[1,1,0,0] : " << 2.*b2*(b1+b2)*w/6.-1./120. << endl;
  cout << "\talpha=[2,0,0,0] : " << (b1*b1+3.*b2*b2)*w/6.-2./120. 
       << endl;
}

void check5(REAL b2,REAL w2) {
  REAL b1=1.-3.*b2;
  REAL w1=1.-4.*w2;
  cout << "check5:" << endl;
  cout << "check5: b1 = " << b1 
       << "\n\tb2 = " << b2 
       << "\n\tw1 = " << w1 
       << "\n\tw2 = " << w2 << endl;
  cout << "\talpha=[0,0,0,0] : " << (w1+4.*w2)/6.-1./6. << endl;
  cout << "\talpha=[1,0,0,0] : " << (w1/4.+(b1+3.*b2)*w2)/6.-1./24. 
       << endl;
  cout << "\talpha=[1,1,0,0] : " << (w1/16+2.*b2*(b1+b2)*w2)/6.-1./120.
       << endl;
  cout << "\talpha=[2,0,0,0] : " 
       << (w1/16.+(b1*b1+3.*b2*b2)*w2)/6.-2./120. << endl;
  cout << "\talpha=[1,1,1,0] : " 
       << (w1/64.+b2*b2*(3.*b1+b2)*w2)/6.-1./720. << endl;
  cout << "\talpha=[2,1,0,0] : " 
       << (w1/64.+b2*(b1*b1+b1*b2+2.*b2*b2)*w2)/6.-2./720. << endl;
  cout << "\talpha=[3,0,0,0] : "
       << (w1/64.+(__cmath_power(b1,3)+3.*__cmath_power(b2,3))*w2)/6.
          -6./720.  << endl;;
  cout << "\talpha=[1,1,1,1] : "
       << (w1/256.+4.*b1*__cmath_power(b2,3)*w2)/6.-1./5040. << endl;
}

void check8(REAL b2,REAL c2,REAL w2) {
  REAL b1=1.-3.*b2;
  REAL c1=1.-3.*c2;
  REAL w1=0.25-w2;
  cout << "check8: b1 = " << b1 
       << "\n\tb2 = " << b2 
       << "\n\tc1 = " << c1 
       << "\n\tc2 = " << c2 
       << "\n\tw1 = " << w1 
       << "\n\tw2 = " << w2 << endl;
  cout << "\talpha=[0,0,0,0] : " << 4.*(w1+w2)/6.-1./6. << endl;
  cout << "\talpha=[1,0,0,0] : " 
       << ((b1+3.*b2)*w1+(c1+3.*c2)*w2)/6.-1./24. << endl;
  cout << "\talpha=[1,1,0,0] : " 
       << (2.*b2*(b1+b2)*w1+2.*c2*(c1+c2)*w2)/6.-1./120. << endl;
  cout << "\talpha=[2,0,0,0] : " 
       << ((b1*b1+3.*b2*b2)*w1+(c1*c1+3.*c2*c2)*w2)/6.-2./120. << endl;
  cout << "\talpha=[1,1,1,0] : " 
       << (b2*b2*(3.*b1+b2)*w1+c2*c2*(3.*c1+c2)*w2)/6.-1./720. << endl;
  cout << "\talpha=[2,1,0,0] : " 
       << (b2*(b1*b1+b1*b2+2.*b2*b2)*w1
          +c2*(c1*c1+c1*c2+2.*c2*c2)*w2)/6.-2./720. << endl;
  cout << "\talpha=[3,0,0,0] : "
       << ((__cmath_power(b1,3)+3.*__cmath_power(b2,3))*w1
          +(__cmath_power(c1,3)+3.*__cmath_power(c2,3))*w2)/6.-6./720.
       << endl;;
  cout << "\talpha=[1,1,1,1] : "
       << (4.*b1*__cmath_power(b2,3)*w1+4.*c1*__cmath_power(c2,3)*w2)/6.
          -1./5040. << endl;
}

void check9(REAL b2,REAL c2,REAL w1,REAL w2) {
  REAL b1=1.-3.*b2;
  REAL c1=1.-3.*c2;
  REAL w3=1.-4.*(w1+w2);
  cout << "check9: b1 = " << b1 
       << "\n\tb2 = " << b2 
       << "\n\tc1 = " << c1 
       << "\n\tc2 = " << c2 
       << "\n\tw1 = " << w1 
       << "\n\tw2 = " << w2 
       << "\n\tw3 = " << w3 << endl;
  cout << "\talpha=[0,0,0,0] : " << (4.*(w1+w2)+w3)/6.-1./6. << endl;
  cout << "\talpha=[1,0,0,0] : " 
       << ((b1+3.*b2)*w1+(c1+3.*c2)*w2+w3/4.)/6.-1./24. << endl;
  cout << "\talpha=[1,1,0,0] : " 
       << (2.*b2*(b1+b2)*w1+2.*c2*(c1+c2)*w2+w3/16.)/6.-1./120. << endl;
  cout << "\talpha=[2,0,0,0] : " 
       << ((b1*b1+3.*b2*b2)*w1+(c1*c1+3.*c2*c2)*w2+w3/16.)/6.-2./120. 
       << endl;
  cout << "\talpha=[1,1,1,0] : " 
       << (b2*b2*(3.*b1+b2)*w1+c2*c2*(3.*c1+c2)*w2+w3/64.)/6.-1./720. 
       << endl;
  cout << "\talpha=[2,1,0,0] : " 
       << (b2*(b1*b1+b1*b2+2.*b2*b2)*w1
          +c2*(c1*c1+c1*c2+2.*c2*c2)*w2+w3/64.)/6.-2./720. << endl;
  cout << "\talpha=[3,0,0,0] : "
       << ((__cmath_power(b1,3)+3.*__cmath_power(b2,3))*w1
          +(__cmath_power(c1,3)+3.*__cmath_power(c2,3))*w2+w3/64.)/6.
          -6./720. << endl;
  cout << "\talpha=[1,1,1,1] : "
       << (4.*b1*__cmath_power(b2,3)*w1+4.*c1*__cmath_power(c2,3)*w2
          +w3/256.)/6.-1./5040. << endl;
  cout << "\talpha=[2,1,1,0] : "
       << (b2*b2*(b1*b1+2.*b1*b2+b2*b2)*w1
          +c2*c2*(c1*c1+2.*c1*c2+c2*c2)*w2+w3/256.)/6.-2./5040. << endl;
  cout << "\talpha=[3,1,0,0] : "
       << (b2*(__cmath_power(b1,3)+b1*b2*b2+2.*__cmath_power(b2,3))*w1
          +c2*(__cmath_power(c1,3)+c1*c2*c2+2.*__cmath_power(c2,3))*w2
          +w3/256.)/6.-6./5040. << endl;
}

void check12(REAL b2,REAL b3) {
  REAL b1=1.-b2-2.*b3;
  REAL w=1./12.;
  cout << "check12: b1 = " << b1 
       << "\n\tb2 = " << b2 
       << "\n\tb3 = " << b3 
       << "\n\tw = " << w << endl;
  cout << "\talpha=[0,0,0,0] : " << 12.*w/6.-1./6. << endl;
  cout << "\talpha=[1,0,0,0] : " << (b1+b2+2.*b3)*3.*w/6.-1./24. 
       << endl;
  cout << "\talpha=[1,1,0,0] : " 
       << (b1*b2+2.*b1*b3+2.*b2*b3+b3*b3)*2.*w/6.-1./120. << endl;
  cout << "\talpha=[2,0,0,0] : " 
       << (b1*b1+b2*b2+2.*b3*b3)*3.*w/6.-2./120.  << endl;
  cout << "\talpha=[1,1,1,0] : " 
       << (6.*b1*b2+3.*b1*b3+2.*b2*b3+b3*b3)*b3*w/6.-1./720. 
       << endl;
  cout << "\talpha=[2,1,0,0] : " 
       << (b1*b1*b2+2.*b1*b1*b3+b1*b2*b2+2.*b1*b3*b3+2.*b2*b2*b3
          +2.*b2*b3*b3+2.*b3*b3*b3)*w/6.-2./720. << endl;
}

void check12b(REAL b2,REAL c2,REAL d2,REAL w2,REAL w3) {
  REAL b1=1.-3.*b2;
  REAL c1=1.-3.*c2;
  REAL d1=1.-3.*d2;
  REAL w1=0.25-w2-w3;
  cout << "check12b: b1 = " << b1
       << "\n\tb2 = " << b2 
       << "\n\tc1 = " << c1 
       << "\n\tc2 = " << c2 
       << "\n\td1 = " << d1 
       << "\n\td2 = " << d2 
       << "\n\tw1 = " << w1
       << "\n\tw2 = " << w2
       << "\n\tw3 = " << w3 << endl;
  cout << "\talpha=[0,0,0,0] : " << (w1+w2+w3)*4./6.-1./6. << endl;
  cout << "\talpha=[1,0,0,0] : " 
       << ((b1*3.*b2)*w1+(c1+3.*c2)*w2+(d1+3.*d2)*w3)/6.-1./24. 
       << endl;
  cout << "\talpha=[1,1,0,0] : " 
       << (b2*(b1+b2)*w1+c2*(c1+c2)*w2+d2*(d1+d2)*w3)*2./6.-1./120.
       << endl;
  cout << "\talpha=[2,0,0,0] : " 
       << ((b1*b1+3.*b2*b2)*w1+(c1*c1+3.*c2*c2)*w2+(d1*d1+3.*d2*d2)*w3)
          /6.-2./120. << endl;
  cout << "\talpha=[1,1,1,0] : " 
       << (b2*b2*(3.*b1+b2)*w1+c2*c2*(3.*c1+c2)*w2+d2*d2*(3.*d1+d2)*w3)
          /6.-1./720. << endl;
  cout << "\talpha=[2,1,0,0] : " 
       << (b2*(b1*b1+b1*b2+2.*b2*b2)*w1
          +c2*(c1*c1+c1*c2+2.*c2*c2)*w2
          +d2*(d1*d1+d1*d2+2.*d2*d2)*w3)/6.-2./720.
       << endl;
  cout << "\talpha=[3,0,0,0] : " 
       << ((__cmath_power(b1,3)+3.*__cmath_power(b2,3))*w1
          +(__cmath_power(c1,3)+3.*__cmath_power(c2,3))*w2
          +(__cmath_power(d1,3)+3.*__cmath_power(d2,3))*w3)/6.-6./720.
       << endl;
  cout << "\talpha=[1,1,1,1] : " 
       << (b1*__cmath_power(b2,3)*w1
          +c1*__cmath_power(c2,3)*w2
          +d1*__cmath_power(d2,3)*w3)*4./6.-1./5040.
       << endl;
  cout << "\talpha=[2,1,1,0] : " 
       << (b2*b2*__cmath_power(b1+b2,2)*w1
          +c2*c2*__cmath_power(c1+c2,2)*w2
          +d2*d2*__cmath_power(d1+d2,2)*w3)/6.-2./5040.
       << endl;
  cout << "\talpha=[2,2,0,0] : " 
       << (b2*b2*(b1*b1+b2*b2)*w1
          +c2*c2*(c1*c1+c2*c2)*w2
          +d2*d2*(d1*d1+d2*d2)*w3)*2./6.-4./5040.
       << endl;
  cout << "\talpha=[3,1,0,0] : " 
       << (b2*(__cmath_power(b1,3)+b1*b2*b2+2.*__cmath_power(b2,3))*w1
          +c2*(__cmath_power(c1,3)+c1*c2*c2+2.*__cmath_power(c2,3))*w2
          +d2*(__cmath_power(d1,3)+d1*d2*d2+2.*__cmath_power(d2,3))*w3)
          /6.-6./5040.
       << endl;
  cout << "\talpha=[4,0,0,0] : " 
       << ((__cmath_power(b1,4)+3.*__cmath_power(b2,4))*w1
          +(__cmath_power(c1,4)+3.*__cmath_power(c2,4))*w2
          +(__cmath_power(d1,4)+3.*__cmath_power(d2,4))*w3)/6.-24./5040.
       << endl;
}

void plot9b() {
  REAL c2min=0.;
  REAL c2max=0.5;
  REAL ymin=0.;
  REAL ymax=0.5;
  REAL winsize=0.5;
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  {
    XYGraphTool gt("b2 vs c2","c2","b2",c2min,c2max,ymin,ymax,&cmap,
      0,winsize);
    gt.setbgColor("white");
    gt.newPage();
    gt.setfgColor("black");
    gt.drawAxes();
    gt.setfgColor("blue");
    REAL dc2=gt.getPixelWidth();
    bool move_pen=true;
    for (REAL c2=c2min+dc2;c2<=c2max;c2+=dc2) {
      REAL a=3.+c2*(-27.+45.*c2);
      REAL minus_b_over_a=(1.+c2*(-12.+27.*c2))/a;
      REAL c_over_a=(c2*(-1.+3.*c2))/a;
      REAL disc=minus_b_over_a*minus_b_over_a-c_over_a;
      if (disc>=0.) {
        disc=sqrt(disc);
        if (minus_b_over_a<0.) disc=-disc;
        REAL b2=1./(minus_b_over_a+disc);
        REAL b1=-2.*b2+1.;
        REAL w3=-(1./180.)*(-45.*__cmath_power(c2,3)
          +36.*__cmath_power(c2,2)
          -9.*c2+1.+270.*b2*__cmath_power(c2,3)
          -207.*b2*__cmath_power(c2,2)
          +45.*b2*c2-3.*b2)
          /(135.*__cmath_power(c2,5)-297.*__cmath_power(c2,4)
           +261.*__cmath_power(c2,3)-114.*__cmath_power(c2,2)+24.*c2-2.)
          /__cmath_power(c2,2);
        REAL c1=1.-2.*c2; 
        REAL w1=(1./180.)*(405.*__cmath_power(c2,5)*b2
          -108.*__cmath_power(c2,5)-1053.*__cmath_power(c2,4)*b2
          +252.*__cmath_power(c2,4)-216.*__cmath_power(c2,3)
          +918.*b2*__cmath_power(c2,3)+84*__cmath_power(c2,2)
          -342.*b2*__cmath_power(c2,2)-15.*c2+54.*b2*c2+1.-3.*b2)
          /(3.*__cmath_power(c2,2)-3.*c2+1.)
          /(-1.+3.*c2)/__cmath_power(c2,2); 
        REAL w2=-(1./180.)*(78732.*__cmath_power(c2,7)
          -29160.*__cmath_power(c2,8)-90072.*__cmath_power(c2,6)
          +56376.*__cmath_power(c2,5)-20628.*__cmath_power(c2,4)
          +4446.*__cmath_power(c2,3)-549.*__cmath_power(c2,2)+36.*c2-1.
          +18225.*__cmath_power(c2,8)*b2-69255.*__cmath_power(c2,7)*b2
          +105462.*__cmath_power(c2,6)*b2-84726.*__cmath_power(c2,5)*b2
          +38718.*__cmath_power(c2,4)*b2-10152.*b2*__cmath_power(c2,3)
          +1476.*b2*__cmath_power(c2,2)-108.*b2*c2+3.*b2)
          /__cmath_power(c2,2)
          /(-1.+3.*c2)
          /(135.*__cmath_power(c2,5)-297.*__cmath_power(c2,4)
           +261.*__cmath_power(c2,3)-114.*__cmath_power(c2,2)+24.*c2
           -2.);
        if (b2>=0. && b1>=0. && c1>=0. && w1>=0. && w2>=0. && w3>=0.) {
          if (move_pen) gt.movePen(c2,b2);
          else gt.drawLine(c2,b2);
          cout << "\tc2,c1,b2,b1,w1,w2,w3 = " << c2 << " " << c1 << " "
               << b2 << " " << b1 << " " << w1 << " " << w2 << " "
               << w3 << endl;
          move_pen=false;
        } else move_pen=true;
      } else move_pen=true;
    }
    gt.setfgColor("red");
    move_pen=true;
    for (REAL c2=c2min+dc2;c2<=c2max;c2+=dc2) {
      REAL a=3.+c2*(-27.+45.*c2);
      REAL minus_b_over_a=(1.+c2*(-12.+27.*c2))/a;
      REAL c_over_a=(c2*(-1.+3.*c2))/a;
      REAL disc=minus_b_over_a*minus_b_over_a-c_over_a;
      if (disc>=0.) {
        disc=sqrt(disc);
        if (minus_b_over_a<0.) disc=-disc;
        REAL rt1=1./(minus_b_over_a+disc);
        REAL b2=c_over_a/rt1;
        REAL b1=-2.*b2+1.;
        REAL w3=-(1./180.)*(-45.*__cmath_power(c2,3)
          +36.*__cmath_power(c2,2)
          -9.*c2+1.+270.*b2*__cmath_power(c2,3)
          -207.*b2*__cmath_power(c2,2)
          +45.*b2*c2-3.*b2)
          /(135.*__cmath_power(c2,5)-297.*__cmath_power(c2,4)
           +261.*__cmath_power(c2,3)-114.*__cmath_power(c2,2)+24.*c2-2.)
          /__cmath_power(c2,2);
        REAL c1=1.-2.*c2; 
        REAL w1=(1./180.)*(405.*__cmath_power(c2,5)*b2
          -108.*__cmath_power(c2,5)-1053.*__cmath_power(c2,4)*b2
          +252.*__cmath_power(c2,4)-216.*__cmath_power(c2,3)
          +918.*b2*__cmath_power(c2,3)+84*__cmath_power(c2,2)
          -342.*b2*__cmath_power(c2,2)-15.*c2+54.*b2*c2+1.-3.*b2)
          /(3.*__cmath_power(c2,2)-3.*c2+1.)
          /(-1.+3.*c2)/__cmath_power(c2,2); 
        REAL w2=-(1./180.)*(78732.*__cmath_power(c2,7)
          -29160.*__cmath_power(c2,8)-90072.*__cmath_power(c2,6)
          +56376.*__cmath_power(c2,5)-20628.*__cmath_power(c2,4)
          +4446.*__cmath_power(c2,3)-549.*__cmath_power(c2,2)+36.*c2-1.
          +18225.*__cmath_power(c2,8)*b2-69255.*__cmath_power(c2,7)*b2
          +105462.*__cmath_power(c2,6)*b2-84726.*__cmath_power(c2,5)*b2
          +38718.*__cmath_power(c2,4)*b2-10152.*b2*__cmath_power(c2,3)
          +1476.*b2*__cmath_power(c2,2)-108.*b2*c2+3.*b2)
          /__cmath_power(c2,2)
          /(-1.+3.*c2)
          /(135.*__cmath_power(c2,5)-297.*__cmath_power(c2,4)
           +261.*__cmath_power(c2,3)-114.*__cmath_power(c2,2)+24.*c2
           -2.);
        if (b2>=0. && b1>=0. && c1>=0. && w1>=0. && w2>=0. && w3>=0.) {
          if (move_pen) gt.movePen(c2,b2);
          else gt.drawLine(c2,b2);
          cout << "\tc2,c1,b2,b1,w1,w2,w3 = " << c2 << " " << c1 << " "
               << b2 << " " << b1 << " " << w1 << " " << w2 << " "
               << w3 << endl;
          if (move_pen) gt.movePen(c2,b2);
          else gt.drawLine(c2,b2);
          move_pen=false;
          cout << "\tc2,b2 = " << c2 << " " << b2 << endl;
        } else move_pen=true;
      } else move_pen=true;
    }
    gt.flush();
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
}

void plot12b() {
  REAL d2min=0.;
  REAL d2max=1./3.;
  REAL ymin=0.;
  REAL ymax=1./3.;
  REAL winsize=0.5;
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  {
    XYGraphTool gt("c2 vs d2","d2","d2",d2min,d2max,ymin,ymax,&cmap,
      0,winsize);
    gt.setbgColor("white");
    gt.newPage();
    gt.setfgColor("black");
    gt.drawAxes();
    gt.setfgColor("blue");
    REAL dd2=gt.getPixelWidth();
    bool move_pen=true;
    for (REAL d2=d2min;d2<=d2max;d2+=dd2) {
      REAL a=56.*(1+d2*(-10.+20.*d2));
      REAL minus_b_over_a=(25.+d2*(-266.+560.*d2))/a;
      REAL c_over_a=2.+(-25.*d2+56.*d2)/a;
      REAL disc=minus_b_over_a*minus_b_over_a-c_over_a;
      if (disc>=0.) {
        disc=sqrt(disc);
        if (minus_b_over_a<0.) disc=-disc;
        REAL rt1=1./(minus_b_over_a+disc);
        if (rt1>=0. && 3.*rt1<=1.) {
          if (move_pen) gt.movePen(d2,rt1);
          else gt.drawLine(d2,rt1);
          cout << "\td2,rt1 = " << d2 << " " << rt1 << endl;
          move_pen=false;
        } else move_pen=true;
      } else move_pen=true;
    }
    gt.setfgColor("red");
    move_pen=true;
    for (REAL d2=d2min;d2<=d2max;d2+=dd2) {
      REAL a=56.*(1+d2*(-10.+20.*d2));
      REAL minus_b_over_a=(25.+d2*(-266.+560.*d2))/a;
      REAL c_over_a=2.+(-25.*d2+56.*d2)/a;
      REAL disc=minus_b_over_a*minus_b_over_a-c_over_a;
      if (disc>=0.) {
        disc=sqrt(disc);
        if (minus_b_over_a<0.) disc=-disc;
        REAL rt1=1./(minus_b_over_a+disc);
        REAL rt2=c_over_a/rt1;
        if (rt2>=0. && 3.*rt2<=1.) {
          if (move_pen) gt.movePen(d2,rt2);
          else gt.drawLine(d2,rt2);
          move_pen=false;
          cout << "\td2,rt2 = " << d2 << " " << rt2 << endl;
        } else move_pen=true;
      } else move_pen=true;
    }
    gt.flush();
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
  {
    XYGraphTool gt("b2 vs d2","d2","b2",d2min,d2max,ymin,ymax,&cmap,
      0,winsize);
    gt.setbgColor("white");
    gt.newPage();
    gt.setfgColor("black");
    gt.drawAxes();
    gt.setfgColor("blue");
    REAL dd2=gt.getPixelWidth();
    bool move_pen=true;
    for (REAL d2=d2min;d2<=d2max;d2+=dd2) {
      REAL a=56.*(1+d2*(-10.+20.*d2));
      REAL minus_b_over_a=(25.+d2*(-266.+560.*d2))/a;
      REAL c_over_a=2.+(-25.*d2+56.*d2)/a;
      REAL disc=minus_b_over_a*minus_b_over_a-c_over_a;
      if (disc>=0.) {
        disc=sqrt(disc);
        if (minus_b_over_a<0.) disc=-disc;
        REAL rt1=1./(minus_b_over_a+disc);
        if (rt1>=0. && 3.*rt1<=1.) {
          REAL b2=-(-25.+d2*(266.-560.*d2)+56.*rt1*(1+d2*(-10.+20.*d2)))
                  /(56.*(1.+d2*(-10.+20.*d2)));
          if (move_pen) gt.movePen(d2,b2);
          else gt.drawLine(d2,b2);
          cout << "\td2,b2 = " << d2 << " " << b2 << endl;
          move_pen=false;
        } else move_pen=true;
      } else move_pen=true;
    }
    gt.setfgColor("red");
    move_pen=true;
    for (REAL d2=d2min;d2<=d2max;d2+=dd2) {
      REAL a=56.*(1+d2*(-10.+20.*d2));
      REAL minus_b_over_a=(25.+d2*(-266.+560.*d2))/a;
      REAL c_over_a=2.+(-25.*d2+56.*d2)/a;
      REAL disc=minus_b_over_a*minus_b_over_a-c_over_a;
      if (disc>=0.) {
        disc=sqrt(disc);
        if (minus_b_over_a<0.) disc=-disc;
        REAL rt1=1./(minus_b_over_a+disc);
        REAL rt2=c_over_a/rt1;
        if (rt2>=0. && 3.*rt2<=1.) {
          REAL b2=-(-25.+d2*(266.-560.*d2)+56.*rt2*(1+d2*(-10.+20.*d2)))
                  /(56.*(1.+d2*(-10.+20.*d2)));
          if (move_pen) gt.movePen(d2,b2);
          else gt.drawLine(d2,b2);
          cout << "\td2,b2 = " << d2 << " " << b2 << endl;
          move_pen=false;
          cout << "\td2,b2 = " << d2 << " " << b2 << endl;
        } else move_pen=true;
      } else move_pen=true;
    }
    gt.flush();
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
}

void fcn(const int &n,const REAL *x,REAL *fvec,int &iflag) {
  switch (n) {
    case 4: {
      REAL b=x[0];
      REAL c=x[1];
      REAL w1=x[2];
      REAL w2=x[3];
      fvec[0]=(0.25-2.*b*(1.-2.*b))*w1
             +(0.25-2.*c*(1.-2.*c))*w2-0.0125;
      fvec[1]=(0.0625-b*b*(3.-8.*b))*w1
             +(0.0625-c*c*(3.-8.*c))*w2-7./960.;
      fvec[2]=(b*(1.+b*(-5.+8.*b))-0.0625)*w1
             +(c*(1.+c*(-5.+8.*c))-0.0625)*w2-1./960.;
      fvec[3]=(0.015625-4.*b*b*b*(1.-3.*b))*w1
             +(0.015625-4.*c*c*c*(1.-3.*c))*w2-73./26880.;
      if (iflag==0) {
        cout << endl;
        cout << "\tx = " << x[0] << " " << x[1] << " " << x[2] << " "
             << x[3] << endl;
        cout << "\tf = " << fvec[0] << " " << fvec[1] << " " << fvec[2]
             << " " << fvec[3] << endl;
      }
      break;
    }
    case 3: {
      REAL b=x[0];
      REAL c=x[1];
      REAL w=x[2];
      fvec[0]=b*(1.-2.*b)*(0.25-w)+c*(1.-2.*c)*w-.025;
      fvec[1]=b*b*(3.-8.*b)*(0.25-w)+c*c*(3.-8.*c)*w-1./120.;
//    fvec[2]=b*(1.+b*(-5.+8.*b))*(0.25-w)+c*(1.+c*(-5.+8.*c))*w-1./60.;
      fvec[2]=(__cmath_power(1.-3.*b,3)+3.*__cmath_power(b,3))*(0.25-w)
              +(__cmath_power(1.-3.*c,3)+3.*__cmath_power(c,3))*w-.05;
      break;
    }
    case 2: {
      REAL b=x[0];
      REAL w=x[1];
      fvec[0]=4.*w*__cmath_power(1.-4.*b,2)-0.2;
      fvec[1]=4.*w*(1.+b*b*(-48.+128.*b))-7./15.;
      break;
    }
  }
}
extern "C" {
  void F77NAME(hybrd)(
    void (*)(const int&,const REAL*,REAL*,int&),
    const int&,REAL*,REAL*,const REAL&,const int&,const int&,const int&,
    const REAL&,REAL*,const int&,REAL&,const int&,int&,int&,REAL*,
    const int&,REAL*,const int&,REAL*,REAL*,REAL*,REAL*,REAL*);
}

int main(int argc,char* argv[]) {
  cout.precision(16);
//cout << "\tin main" << endl;
#ifdef DEBUG
//setTraps();
#endif
#if MEM_DEBUG
  MemoryDebugger md(1);
#endif
  plot9b();
  int n=2;
  REAL *x=OPERATOR_NEW_BRACKET(REAL,n);
  REAL *fvec=OPERATOR_NEW_BRACKET(REAL,n);
  REAL xtol=sqrt(DBL_EPSILON);
  int maxfev=100;
  int ml=n-1;
  int mu=n-1;
  REAL epsfcn=exp(log(DBL_EPSILON)/3.);
  REAL *diag=OPERATOR_NEW_BRACKET(REAL,n);
  int mode=1;
  REAL factor=100.;
  int nprint=0;
  int info=0;
  int nfev=0;
  REAL *fjac=OPERATOR_NEW_BRACKET(REAL,n*n);
  int ldfjac=n;
  REAL *r=OPERATOR_NEW_BRACKET(REAL,(n*(n+1))/2);
  int lr=(n*(n+1))/2;
  REAL *qtf=OPERATOR_NEW_BRACKET(REAL,n);
  REAL *wa1=OPERATOR_NEW_BRACKET(REAL,n);
  REAL *wa2=OPERATOR_NEW_BRACKET(REAL,n);
  REAL *wa3=OPERATOR_NEW_BRACKET(REAL,n);
  REAL *wa4=OPERATOR_NEW_BRACKET(REAL,n);
//for (int i=0;i<600;i++) {
    x[0]=drand48();
    if (n>1) x[1]=drand48();
    if (n>2) x[2]=drand48();
    if (n>3) x[3]=drand48();
    if (n==4) {
      x[0]=1./3.;
      x[1]=0.125;
      x[2]=0.09;
      x[3]=0.16;
      nprint=1;
    }
    if (n>1) {
      F77NAME(hybrd)(fcn,n,x,fvec,xtol,maxfev,ml,mu,epsfcn,diag,mode,
        factor,nprint,info,nfev,fjac,ldfjac,r,lr,qtf,wa1,wa2,wa3,wa4);
      cout << "\n\ninfo = " << info << "\n\n" << endl;
    }
    switch (n) {
      case 4: {
        if (x[0]>0. && x[0]<1./3. && x[1]>0. && x[1]<1./3. && x[2]>0.
        && x[2]<0.25 && x[3]>0. && x[3]<0.25) {
          epsfcn=DBL_EPSILON;
          F77NAME(hybrd)(fcn,n,x,fvec,xtol,maxfev,ml,mu,epsfcn,diag,
            mode,factor,nprint,info,nfev,fjac,ldfjac,r,lr,qtf,wa1,wa2,
            wa3,wa4);
          if (info==1 || info==2) {
            int iflag=0;
            fcn(n,x,fvec,iflag);
            cout << "\tinfo = " << info << endl;
            cout << "\tb = " << 1.-3.*x[0] << " " << x[0] << endl;
            cout << "\tc = " << 1.-3.*x[1] << " " << x[1] << endl;
            cout << "\tw = " << 0.25-x[2] << " " << x[2] << endl;
            cout << "\tfvec = " << fvec[0] << " " << fvec[1] << " " 
                 << fvec[2] << " " << fvec[3] << endl;
            REAL b2=x[0];
            REAL b1=1.-3.*b2;
            REAL c2=x[1];
            REAL c1=1.-3.*c2;
            REAL w1=x[2];
            REAL w2=x[3];
            REAL w3=1.-4.*(w1+w2);
            cout << "\tobj = " 
                 << ((b1*b1+3.*b2*b2)*w1+(c1*c1+3.*c2*c2)*w2+w3/16.)/6.
                    -1./60. << endl;
          }
          epsfcn=exp(log(DBL_EPSILON)/3.);
        }
        x[0]=0.2841716607157711313;
        x[1]=0.20043760551370549452;
        x[2]=-1.079341461647618871;
        x[3]=1.785252549402819083;
        check9(x[0],x[1],x[2],x[3]);
        break;
      }
      case 3: {
        if (x[0]>0. && x[0]<1./3. && x[1]>0. && x[1]<1./3. && x[2]>0.
        && x[2]<0.25 && (info==1 || info==2)) {
          epsfcn=DBL_EPSILON;
          F77NAME(hybrd)(fcn,n,x,fvec,xtol,maxfev,ml,mu,epsfcn,diag,
            mode,factor,nprint,info,nfev,fjac,ldfjac,r,lr,qtf,wa1,wa2,
            wa3,wa4);
          if (info==1 || info==2) {
            int iflag=0;
            fcn(n,x,fvec,iflag);
            cout << "\tinfo = " << info << endl;
            cout << "\tb = " << 1.-3.*x[0] << " " << x[0] << endl;
            cout << "\tc = " << 1.-3.*x[1] << " " << x[1] << endl;
            cout << "\tw = " << 0.25-x[2] << " " << x[2] << endl;
            cout << "\tfvec = " << fvec[0] << " " << fvec[1] << " " 
                 << fvec[2] << endl;
            cout << "\tobj = " 
                 << ((1.-3.*x[0])*__cmath_power(x[0],3))*(0.25-x[2])
                   +((1.-3.*x[1])*__cmath_power(x[1],3))*x[2]-1./3360.
                 << endl;
          }
          epsfcn=exp(log(DBL_EPSILON)/3.);
        }
        x[0]=1./3.;
        x[1]=0.125;
        x[2]=0.16;
        check8(x[0],x[1],x[2]);
        break;
      }
      case 2: {
        if (x[0]>0. && x[0]<1./3. && x[1]>0. && x[1]<0.25 && 
        (info==1 || info==2)) {
          epsfcn=DBL_EPSILON;
          F77NAME(hybrd)(fcn,n,x,fvec,xtol,maxfev,ml,mu,epsfcn,diag,
            mode,factor,nprint,info,nfev,fjac,ldfjac,r,lr,qtf,wa1,wa2,
            wa3,wa4);
          if (info==1 || info==2) {
            int iflag=0;
            int nn=3;
            REAL xx[3]={x[0],0.,x[1]};
            REAL ff[3];
            fcn(nn,xx,ff,iflag);
            cout << "\tinfo = " << info << endl;
            cout << "\tb = " << 1.-3.*x[0] << " " << x[0] << endl;
            cout << "\tc = " << 1. << " " << 0. << endl;
            cout << "\tw = " << 0.25-x[1] << " " << x[1] << endl;
            cout << "\tfvec = " << ff[0] << " " << ff[1] << " " << ff[2] 
                 << endl;
          }
          epsfcn=exp(log(DBL_EPSILON)/3.);
        }
//      x[0]=(1.-sqrt(.2))*0.25;
//      x[1]=0.25;
//      check4(x[0],x[1]);
//      x[0]=1./6.;
//      x[1]=0.45;
//      check5(x[0],x[1]);
//        x[0]=-0.0854101966249684545;
//        x[1]=0.36180339887498948482;
        x[0]=1.-0.41458980337503154554;
        x[1]=0.13819660112501051518;
        check12(x[0],x[1]);
        break;
      }
      case 1: {
        REAL w=1.;
        check1(w);
        break;
      }
    }
//}
  delete [] wa4;
  delete [] wa3;
  delete [] wa2;
  delete [] wa1;
  delete [] qtf;
  delete [] r;
  delete [] fjac;
  delete [] diag;
  delete [] fvec;
  delete [] x;
  
  return EXIT_SUCCESS;
}
