// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <cmath>
#include <iostream>
#include <iomanip>
#include <math.h> // for HUGE_VAL,M_PI
#include <stdlib.h>

#if (SPACEDIM==2)
//#include "dunavant.H"
//#include "fekete.H"
#include "ncc_triangle.H"
//#include "wandzura.H"
#endif
#if (SPACEDIM==3)
//#include "felippa.H"
//#include "gm_rules.H"
//#include "keast.H"
//#include "ncc_tetrahedron.H"
#include "gm_rules.H"
//#include "keast.H"
#endif
#include "Debug.H"
#include "MemoryDebugger.H"
#include "Quadrature.H"
#include "SetTraps.H"
#include "Types.H"

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int main(int argc,char* argv[]) {
//cout << "\tin main" << endl;
#ifdef DEBUG
//setTraps();
#endif
#if MEM_DEBUG
  MemoryDebugger md(1);
#endif
  cout << setprecision(20);
#if (SPACEDIM==2)
//for (int r=1;r<=20;r++) {
//  int on=dunavant_order_num(r);
//  double *xy=OPERATOR_NEW_BRACKET(double,2*on);
//  double *w=OPERATOR_NEW_BRACKET(double,on);
//  dunavant_rule(r,on,xy,w);
//for (int r=1;r<=7;r++) {
//  int on=fekete_order_num(r);
//  double *xy=OPERATOR_NEW_BRACKET(double,3*on);
//  fekete_rule(r,on,xy,w);
//  double *w=OPERATOR_NEW_BRACKET(double,on);
//for (int r=1;r<=7;r++) {
//  int on=wandzura_order_num(r);
//  double *xy=OPERATOR_NEW_BRACKET(double,2*on);
//  double *w=OPERATOR_NEW_BRACKET(double,on);
//  wandzura_rule(r,on,xy,w);
  for (int r=1;r<=9;r++) {
    int on=ncc_triangle_order_num(r);
    double *xy=OPERATOR_NEW_BRACKET(double,2*on);
    double *w=OPERATOR_NEW_BRACKET(double,on);
    ncc_triangle_rule(r,on,xy,w);
    cout << "\nr = " << r << endl;
    for (int q=0;q<on;q++) {
      cout << "\txy = " << xy[2*q] << " " << xy[2*q+1] << " w = "
           << w[q] << endl;
    }
    delete [] xy;
    delete [] w;
  }
#endif
#if (SPACEDIM==3)
/*
  for (int r=1;r<=20;r++) {
    int on=keast_order_num(r);
    double *xyz=OPERATOR_NEW_BRACKET(double,3*on);
    double *w=OPERATOR_NEW_BRACKET(double,on);
    keast_rule(r,on,xyz,w);
    cout << "\nr = " << r << endl;
    for (int q=0;q<on;q++) {
      cout << "\txyz = " << xyz[3*q] << " " << xyz[3*q+1] << " "
           << xyz[3*q+2] << " w = " << w[q] << endl;
    }
    delete [] xyz;
    delete [] w;
  }
*/
//
  for (int r=1;r<=20;r++) {
    int on=gm_rule_size(r,3);
    double *xyz=OPERATOR_NEW_BRACKET(double,3*on);
    double *w=OPERATOR_NEW_BRACKET(double,on);
    gm_rule_set(r,3,on,w,xyz);
    cout << "\nr = " << r << endl;
    for (int q=0;q<on;q++) {
      cout << "\txyz = " << xyz[3*q] << " " << xyz[3*q+1] << " "
           << xyz[3*q+2] << " w = " << w[q] << endl;
    }
    delete [] xyz;
    delete [] w;
  }
//
/*
  for (int r=1;r<=7;r++) {
    int on=ncc_tetrahedron_order_num(r);
    double *xyz=OPERATOR_NEW_BRACKET(double,3*on);
    double *w=OPERATOR_NEW_BRACKET(double,on);
    ncc_tetrahedron_rule(r,on,xyz,w);
    cout << "\nr = " << r << endl;
    for (int q=0;q<on;q++) {
      cout << "\txyz = " << xyz[3*q] << " " << xyz[3*q+1] << " "
           << xyz[3*q+2] << " w = " << w[q] << endl;
    }
    delete [] xyz;
    delete [] w;
  }
*/
/*
  {
    double xyz[3*1];
    double w[1];
    tetr_unit_o01(w,xyz);
    cout << "\nr = " << 1 << endl;
    cout << "\txyz = " << xyz[0] << " " << xyz[1] << " " << xyz[2]
         << " w = " << w[0] << endl;
  }
  {
    double xyz[3*4];
    double w[4];
    tetr_unit_o04(w,xyz);
    cout << "\nr = " << 2 << endl;
    for (int q=0;q<4;q++) {
      cout << "\txyz = " << xyz[3*q] << " " << xyz[3*q+1] << " " 
           << xyz[3*q+2] << " w = " << w[q] << endl;
    }
  }
  {
    double xyz[3*8];
    double w[8];
    tetr_unit_o08(w,xyz);
    cout << "\nr = " << 3 << endl;
    for (int q=0;q<8;q++) {
      cout << "\txyz = " << xyz[3*q] << " " << xyz[3*q+1] << " " 
           << xyz[3*q+2] << " w = " << w[q] << endl;
    }
  }
  {
    double xyz[3*8];
    double w[8];
    tetr_unit_o08b(w,xyz);
    cout << "\nr = " << 4 << endl;
    for (int q=0;q<8;q++) {
      cout << "\txyz = " << xyz[3*q] << " " << xyz[3*q+1] << " " 
           << xyz[3*q+2] << " w = " << w[q] << endl;
    }
  }
  {
    double xyz[3*14];
    double w[14];
    tetr_unit_o14(w,xyz);
    cout << "\nr = " << 5 << endl;
    for (int q=0;q<14;q++) {
      cout << "\txyz = " << xyz[3*q] << " " << xyz[3*q+1] << " " 
           << xyz[3*q+2] << " w = " << w[q] << endl;
    }
  }
  {
    double xyz[3*14];
    double w[14];
    tetr_unit_o14b(w,xyz);
    cout << "\nr = " << 6 << endl;
    for (int q=0;q<14;q++) {
      cout << "\txyz = " << xyz[3*q] << " " << xyz[3*q+1] << " " 
           << xyz[3*q+2] << " w = " << w[q] << endl;
    }
  }
  {
    double xyz[3*15];
    double w[15];
    tetr_unit_o15(w,xyz);
    cout << "\nr = " << 7 << endl;
    for (int q=0;q<15;q++) {
      cout << "\txyz = " << xyz[3*q] << " " << xyz[3*q+1] << " " 
           << xyz[3*q+2] << " w = " << w[q] << endl;
    }
  }
  {
    double xyz[3*15];
    double w[15];
    tetr_unit_o15b(w,xyz);
    cout << "\nr = " << 8 << endl;
    for (int q=0;q<15;q++) {
      cout << "\txyz = " << xyz[3*q] << " " << xyz[3*q+1] << " " 
           << xyz[3*q+2] << " w = " << w[q] << endl;
    }
  }
  {
    double xyz[3*24];
    double w[24];
    tetr_unit_o24(w,xyz);
    cout << "\nr = " << 9 << endl;
    for (int q=0;q<24;q++) {
      cout << "\txyz = " << xyz[3*q] << " " << xyz[3*q+1] << " " 
           << xyz[3*q+2] << " w = " << w[q] << endl;
    }
  }
*/
#endif
  return EXIT_SUCCESS;
}
