// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <set>

#include "Boundary.H"
#include "MemoryDebugger.H"
#include "PermutationGenerator.H"
#include "Shape.H"
#include "Tracer.H"

#include "AreaGraphTool.H"
#include "VolGraphTool.H"
#include "XYGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKWindow.H"
#endif

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
//for (int nv=2;nv<=8;nv++) {
//  PermutationGenerator pg(nv);
//  while (pg.hasMore()) {
//    const NumPtr<int> &p=pg.getNext();
//    for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//    cout << endl;
//  }
//  cout << endl;
//}

#if (SPACEDIM==2)
//
#ifdef DEBUG
  cout << "Triangle :" << endl;
#endif
  for (int degree=1;degree<=6;degree++) 
//int degree=2;
  {
#ifdef DEBUG
    cout << "\tdegree = " << degree << endl;
#endif
    int neslp=Triangle::staticNumberEquallySpacedLatticePoints(degree);
    int face_neslp=
      Interval::staticNumberEquallySpacedLatticePoints(degree);
    int face_interior_neslp=
      Interval::staticNumberInteriorEquallySpacedLatticePoints(degree);
    set<int> all_integers;
    for (int i=0;i<neslp;i++) all_integers.insert(i);

    for (int i=0;i<Triangle::verticesPerShape();i++) {
      int j=Triangle::staticEquallySpacedLatticePointIndexOfVertex(i,
        degree);
      CHECK_TEST(j>=0 && j<neslp);
    }

    NumPtr<set<int> > static_face_eslp_sets(Triangle::facesPerShape());
    NumPtr<set<int> > static_face_interior_eslp_sets(
      Triangle::facesPerShape());
    for (int f=0;f<Triangle::facesPerShape();f++) {
      set<int> &static_face_eslp_set=static_face_eslp_sets[f];
      NumPtr<int> eslpi;
      Triangle::staticEquallySpacedLatticePointIndicesOfFace(f,degree,
        eslpi);
      CHECK_SAME(eslpi.getNumber(),face_neslp);
      for (int i=0;i<eslpi.getNumber();i++) {
        static_face_eslp_set.insert(eslpi[i]);
      }
      set<int> intersection;
      set_intersection(all_integers.begin(),all_integers.end(),
        static_face_eslp_set.begin(),static_face_eslp_set.end(),
        inserter(intersection,intersection.begin()));
      CHECK_TEST(intersection.size()==eslpi.getNumber());

      set<int> &static_face_interior_eslp_set=
        static_face_interior_eslp_sets[f];
      NumPtr<int> eslpi2;
      Triangle::staticEquallySpacedLatticePointIndicesOfFaceInterior(f,
        degree,eslpi2);
      CHECK_SAME(eslpi2.getNumber(),face_interior_neslp);
      for (int i=0;i<eslpi2.getNumber();i++) {
        static_face_interior_eslp_set.insert(eslpi2[i]);
      }
      set<int> intersection2;
      set_intersection(static_face_eslp_set.begin(),
        static_face_eslp_set.end(),static_face_interior_eslp_set.begin(),
        static_face_interior_eslp_set.end(),
        inserter(intersection2,intersection2.begin()));
      CHECK_TEST(intersection2.size()==eslpi2.getNumber());
    }

    PermutationGenerator pg(2);
    while (pg.hasMore()) {
      const NumPtr<int> &p=pg.getNext();
      int faces[3]={0,1,2};
      int vertices[3]={0,p[0]+1,p[1]+1};
      Triangle e(0,faces,vertices);

      for (int i=0;i<e.numberVertices();i++) {
        int j=e.equallySpacedLatticePointIndexOfVertex(i,degree);
        CHECK_TEST(j>=0 && j<neslp);
      }
      for (int f=0;f<e.numberCodimension1Shapes();f++) {
        NumPtr<int> eslpi;
        e.equallySpacedLatticePointIndicesOfFace(f,degree,eslpi);
        CHECK_SAME(eslpi.getNumber(),face_neslp);
        set<int> eslpi_face_set;
        for (int i=0;i<eslpi.getNumber();i++) {
          eslpi_face_set.insert(eslpi[i]);
        }
        set<int> &static_face_eslp_set=static_face_eslp_sets[f];
        set<int> intersection;
        set_intersection(static_face_eslp_set.begin(),
          static_face_eslp_set.end(),eslpi_face_set.begin(),
          eslpi_face_set.end(),
          inserter(intersection,intersection.begin()));
        CHECK_TEST(intersection.size()==face_neslp);

        NumPtr<int> eslpi2;
        e.equallySpacedLatticePointIndicesOfFaceInterior(f,degree,eslpi2);
        CHECK_SAME(eslpi2.getNumber(),face_interior_neslp);
        set<int> eslpi_face_interior_set;
        for (int i=0;i<eslpi2.getNumber();i++) {
          eslpi_face_interior_set.insert(eslpi2[i]);
        }
        set<int> &static_face_interior_eslp_set=
          static_face_interior_eslp_sets[f];
        set<int> intersection2;
        set_intersection(static_face_interior_eslp_set.begin(),
          static_face_interior_eslp_set.end(),
          eslpi_face_interior_set.begin(),eslpi_face_interior_set.end(),
          inserter(intersection2,intersection2.begin()));
        CHECK_TEST(intersection2.size()==face_interior_neslp);
      }
    }
  }
//
//
#ifdef DEBUG
  cout << "Quadrilateral :" << endl;
#endif
  for (int degree=1;degree<=6;degree++) 
//int degree=2;
  {
#ifdef DEBUG
    cout << "\tdegree = " << degree << endl;
#endif
    int neslp=
      Quadrilateral::staticNumberEquallySpacedLatticePoints(degree);
    int face_neslp=
      Interval::staticNumberEquallySpacedLatticePoints(degree);
    int face_interior_neslp=
      Interval::staticNumberInteriorEquallySpacedLatticePoints(degree);
    set<int> all_integers;
    for (int i=0;i<neslp;i++) all_integers.insert(i);

    for (int i=0;i<Quadrilateral::verticesPerShape();i++) {
      int j=Quadrilateral::staticEquallySpacedLatticePointIndexOfVertex(i,
        degree);
      CHECK_TEST(j>=0 && j<neslp);
    }

    NumPtr<set<int> > static_face_eslp_sets(Quadrilateral::facesPerShape());
    NumPtr<set<int> > static_face_interior_eslp_sets(
      Quadrilateral::facesPerShape());
    for (int f=0;f<Quadrilateral::facesPerShape();f++) {
      set<int> &static_face_eslp_set=static_face_eslp_sets[f];
      NumPtr<int> eslpi;
      Quadrilateral::staticEquallySpacedLatticePointIndicesOfFace(f,degree,
        eslpi);
      CHECK_SAME(eslpi.getNumber(),face_neslp);
      for (int i=0;i<eslpi.getNumber();i++) {
        static_face_eslp_set.insert(eslpi[i]);
      }
      set<int> intersection;
      set_intersection(all_integers.begin(),all_integers.end(),
        static_face_eslp_set.begin(),static_face_eslp_set.end(),
        inserter(intersection,intersection.begin()));
      CHECK_TEST(intersection.size()==eslpi.getNumber());

      set<int> &static_face_interior_eslp_set=
        static_face_interior_eslp_sets[f];
      NumPtr<int> eslpi2;
      Quadrilateral::staticEquallySpacedLatticePointIndicesOfFaceInterior(f,
        degree,eslpi2);
      CHECK_SAME(eslpi2.getNumber(),face_interior_neslp);
      for (int i=0;i<eslpi2.getNumber();i++) {
        static_face_interior_eslp_set.insert(eslpi2[i]);
      }
      set<int> intersection2;
      set_intersection(static_face_eslp_set.begin(),
        static_face_eslp_set.end(),static_face_interior_eslp_set.begin(),
        static_face_interior_eslp_set.end(),
        inserter(intersection2,intersection2.begin()));
      CHECK_TEST(intersection2.size()==eslpi2.getNumber());
    }

    PermutationGenerator pg(3);
    while (pg.hasMore()) {
      const NumPtr<int> &p=pg.getNext();
      int faces[4]={0,1,2,3};
      int vertices[4]={0,p[0]+1,p[1]+1,p[2]+1};
      Quadrilateral e(0,faces,vertices);

      for (int i=0;i<e.numberVertices();i++) {
        int j=e.equallySpacedLatticePointIndexOfVertex(i,degree);
        CHECK_TEST(j>=0 && j<neslp);
      }
      for (int f=0;f<e.numberCodimension1Shapes();f++) {
        NumPtr<int> eslpi;
        e.equallySpacedLatticePointIndicesOfFace(f,degree,eslpi);
        CHECK_SAME(eslpi.getNumber(),face_neslp);
        set<int> eslpi_face_set;
        for (int i=0;i<eslpi.getNumber();i++) {
          eslpi_face_set.insert(eslpi[i]);
        }
        set<int> &static_face_eslp_set=static_face_eslp_sets[f];
        set<int> intersection;
        set_intersection(static_face_eslp_set.begin(),
          static_face_eslp_set.end(),eslpi_face_set.begin(),
          eslpi_face_set.end(),
          inserter(intersection,intersection.begin()));
        CHECK_TEST(intersection.size()==face_neslp);

        NumPtr<int> eslpi2;
        e.equallySpacedLatticePointIndicesOfFaceInterior(f,degree,eslpi2);
        CHECK_SAME(eslpi2.getNumber(),face_interior_neslp);
        set<int> eslpi_face_interior_set;
        for (int i=0;i<eslpi2.getNumber();i++) {
          eslpi_face_interior_set.insert(eslpi2[i]);
        }
        set<int> &static_face_interior_eslp_set=
          static_face_interior_eslp_sets[f];
        set<int> intersection2;
        set_intersection(static_face_interior_eslp_set.begin(),
          static_face_interior_eslp_set.end(),
          eslpi_face_interior_set.begin(),eslpi_face_interior_set.end(),
          inserter(intersection2,intersection2.begin()));
        CHECK_TEST(intersection2.size()==face_interior_neslp);
      }
    }
  }
//
#endif

#if (SPACEDIM==3)
//
#ifdef DEBUG
  cout << "Tetrahedron :" << endl;
#endif
  for (int degree=1;degree<=6;degree++) 
//int degree=2;
  {
#ifdef DEBUG
    cout << "\tdegree = " << degree << endl;
#endif
    int neslp=Tetrahedron::staticNumberEquallySpacedLatticePoints(degree);
    int face_neslp=
      Triangle::staticNumberEquallySpacedLatticePoints(degree);
    int face_interior_neslp=
      Triangle::staticNumberInteriorEquallySpacedLatticePoints(degree);
    int line_interior_neslp=
      Interval::staticNumberInteriorEquallySpacedLatticePoints(degree);
    set<int> all_integers;
    for (int i=0;i<neslp;i++) all_integers.insert(i);

    for (int i=0;i<Tetrahedron::verticesPerShape();i++) {
      int j=Tetrahedron::staticEquallySpacedLatticePointIndexOfVertex(i,
        degree);
      CHECK_TEST(j>=0 && j<neslp);
    }

    NumPtr<set<int> > static_face_eslp_sets(Tetrahedron::facesPerShape());
    NumPtr<set<int> > static_face_interior_eslp_sets(
      Tetrahedron::facesPerShape());
    for (int f=0;f<Tetrahedron::facesPerShape();f++) {
      set<int> &static_face_eslp_set=static_face_eslp_sets[f];
      NumPtr<int> eslpi;
      Tetrahedron::staticEquallySpacedLatticePointIndicesOfFace(f,degree,
        eslpi);
      CHECK_SAME(eslpi.getNumber(),face_neslp);
      for (int i=0;i<eslpi.getNumber();i++) {
        static_face_eslp_set.insert(eslpi[i]);
      }
      set<int> intersection;
      set_intersection(all_integers.begin(),all_integers.end(),
        static_face_eslp_set.begin(),static_face_eslp_set.end(),
        inserter(intersection,intersection.begin()));
      CHECK_TEST(intersection.size()==eslpi.getNumber());

      set<int> &static_face_interior_eslp_set=
        static_face_interior_eslp_sets[f];
      NumPtr<int> eslpi2;
      Tetrahedron::staticEquallySpacedLatticePointIndicesOfFaceInterior(f,
        degree,eslpi2);
      CHECK_SAME(eslpi2.getNumber(),face_interior_neslp);
      for (int i=0;i<eslpi2.getNumber();i++) {
        static_face_interior_eslp_set.insert(eslpi2[i]);
      }
      set<int> intersection2;
      set_intersection(static_face_eslp_set.begin(),
        static_face_eslp_set.end(),static_face_interior_eslp_set.begin(),
        static_face_interior_eslp_set.end(),
        inserter(intersection2,intersection2.begin()));
      CHECK_TEST(intersection2.size()==eslpi2.getNumber());
    }

    PermutationGenerator pg(2);
    while (pg.hasMore()) {
      const NumPtr<int> &p=pg.getNext();
      int lines[6]={0,1,2,3,4,5};
      int faces[4]={0,1,2,3};
      int vertices[4]={0,1,p[0]+2,p[1]+2};
      Tetrahedron e(0,lines,faces,vertices);

      for (int i=0;i<e.numberVertices();i++) {
        int j=e.equallySpacedLatticePointIndexOfVertex(i,degree);
        CHECK_TEST(j>=0 && j<neslp);
      }

      for (int f=0;f<e.numberCodimension1Shapes();f++) {
        NumPtr<int> eslpi;
        e.equallySpacedLatticePointIndicesOfFace(f,degree,eslpi);
        CHECK_SAME(eslpi.getNumber(),face_neslp);
        set<int> eslpi_face_set;
        for (int i=0;i<eslpi.getNumber();i++) {
          eslpi_face_set.insert(eslpi[i]);
        }
        set<int> &static_face_eslp_set=static_face_eslp_sets[f];
        set<int> intersection;
        set_intersection(static_face_eslp_set.begin(),
          static_face_eslp_set.end(),eslpi_face_set.begin(),
          eslpi_face_set.end(),
          inserter(intersection,intersection.begin()));
        CHECK_TEST(intersection.size()==face_neslp);

        NumPtr<int> eslpi2;
        e.equallySpacedLatticePointIndicesOfFaceInterior(f,degree,eslpi2);
        CHECK_SAME(eslpi2.getNumber(),face_interior_neslp);
        set<int> eslpi_face_interior_set;
        for (int i=0;i<eslpi2.getNumber();i++) {
          eslpi_face_interior_set.insert(eslpi2[i]);
        }
        set<int> &static_face_interior_eslp_set=
          static_face_interior_eslp_sets[f];
        set<int> intersection2;
        set_intersection(static_face_interior_eslp_set.begin(),
          static_face_interior_eslp_set.end(),
          eslpi_face_interior_set.begin(),eslpi_face_interior_set.end(),
          inserter(intersection2,intersection2.begin()));
        CHECK_TEST(intersection2.size()==face_interior_neslp);
      }

      for (int l=0;l<e.numberCodimension2Shapes();l++) {
        NumPtr<int> eslpi2;
        e.equallySpacedLatticePointIndicesOfLineInterior(l,degree,eslpi2);
        CHECK_SAME(eslpi2.getNumber(),line_interior_neslp);
        set<int> eslpi_line_interior_set;
        for (int i=0;i<eslpi2.getNumber();i++) {
          eslpi_line_interior_set.insert(eslpi2[i]);
        }
        for (int f=0;f<Tetrahedron::facesPerShape();f++) {
          for (int fl=0;fl<Triangle::facesPerShape();fl++) {
            int l2=Tetrahedron::staticElementLineFromFace(f,fl);
            if (l2==l) {
              set<int> &static_face_eslp_set=static_face_eslp_sets[f];
              set<int> intersection2;
              set_intersection(static_face_eslp_set.begin(),
                static_face_eslp_set.end(),
                eslpi_line_interior_set.begin(),
                eslpi_line_interior_set.end(),
                inserter(intersection2,intersection2.begin()));
              CHECK_TEST(intersection2.size()==eslpi2.getNumber());
            }
          }
        }
      }
    }
  }
//
//
#ifdef DEBUG
  cout << "Prism :" << endl;
#endif
  for (int degree=1;degree<=6;degree++) 
//int degree=2;
  {
#ifdef DEBUG
    cout << "\tdegree = " << degree << endl;
#endif
    int neslp=Prism::staticNumberEquallySpacedLatticePoints(degree);
    int triangle_face_neslp=
      Triangle::staticNumberEquallySpacedLatticePoints(degree);
    int triangle_face_interior_neslp=
      Triangle::staticNumberInteriorEquallySpacedLatticePoints(degree);
    int quadrilateral_face_neslp=
      Quadrilateral::staticNumberEquallySpacedLatticePoints(degree);
    int quadrilateral_face_interior_neslp=
      Quadrilateral::staticNumberInteriorEquallySpacedLatticePoints(degree);
    int line_interior_neslp=
      Interval::staticNumberInteriorEquallySpacedLatticePoints(degree);
    set<int> all_integers;
    for (int i=0;i<neslp;i++) all_integers.insert(i);

    for (int i=0;i<Prism::verticesPerShape();i++) {
      int j=Prism::staticEquallySpacedLatticePointIndexOfVertex(i,
        degree);
      CHECK_TEST(j>=0 && j<neslp);
    }

    NumPtr<set<int> > static_face_eslp_sets(Prism::facesPerShape());
    NumPtr<set<int> > static_face_interior_eslp_sets(
      Prism::facesPerShape());
    for (int f=0;f<Prism::facesPerShape();f++) {
      set<int> &static_face_eslp_set=static_face_eslp_sets[f];
      NumPtr<int> eslpi;
      Prism::staticEquallySpacedLatticePointIndicesOfFace(f,degree,
        eslpi);
      CHECK_SAME(eslpi.getNumber(),
                (f<2 ? triangle_face_neslp : quadrilateral_face_neslp));
      for (int i=0;i<eslpi.getNumber();i++) {
        static_face_eslp_set.insert(eslpi[i]);
      }
      set<int> intersection;
      set_intersection(all_integers.begin(),all_integers.end(),
        static_face_eslp_set.begin(),static_face_eslp_set.end(),
        inserter(intersection,intersection.begin()));
      CHECK_TEST(intersection.size()==eslpi.getNumber());
#ifdef DEBUG
//    cout << "\tstatic_face_eslp_sets[" << f << "] = ";
//    for (set<int>::const_iterator sfes_iterator=
//    static_face_eslp_set.begin();
//    sfes_iterator!=static_face_eslp_set.end();sfes_iterator++) {
//      cout << *sfes_iterator << " ";
//    }
//    cout << endl;
#endif

      set<int> &static_face_interior_eslp_set=
        static_face_interior_eslp_sets[f];
      NumPtr<int> eslpi2;
      Prism::staticEquallySpacedLatticePointIndicesOfFaceInterior(f,
        degree,eslpi2);
      CHECK_SAME(eslpi2.getNumber(),(f<2 ? triangle_face_interior_neslp : 
        quadrilateral_face_interior_neslp));
      for (int i=0;i<eslpi2.getNumber();i++) {
        static_face_interior_eslp_set.insert(eslpi2[i]);
      }
      set<int> intersection2;
      set_intersection(static_face_eslp_set.begin(),
        static_face_eslp_set.end(),static_face_interior_eslp_set.begin(),
        static_face_interior_eslp_set.end(),
        inserter(intersection2,intersection2.begin()));
      CHECK_TEST(intersection2.size()==eslpi2.getNumber());
    }

    PermutationGenerator pg(5);
    while (pg.hasMore()) {
      const NumPtr<int> &p=pg.getNext();
      int lines[9]={0,1,2,3,4,5,6,7,8};
      int faces[5]={0,1,2,3,4};
      int vertices[6]={0,p[0]+1,p[1]+1,p[2]+1,p[3]+1,p[4]+1};
      Prism e(0,lines,faces,vertices);
#ifdef DEBUG
//    e.printOn(cout);
#endif

      for (int i=0;i<e.numberVertices();i++) {
        int j=e.equallySpacedLatticePointIndexOfVertex(i,degree);
        CHECK_TEST(j>=0 && j<neslp);
      }

      for (int f=0;f<e.numberCodimension1Shapes();f++) {
        NumPtr<int> eslpi;
        e.equallySpacedLatticePointIndicesOfFace(f,degree,eslpi);
        int face_neslp=
          (f<2 ? triangle_face_neslp : quadrilateral_face_neslp);
        CHECK_SAME(eslpi.getNumber(),face_neslp);
        set<int> eslpi_face_set;
        for (int i=0;i<eslpi.getNumber();i++) {
          eslpi_face_set.insert(eslpi[i]);
        }
#ifdef DEBUG
//      cout << "\teslpi_face_set[" << f << "] = ";
//      for (set<int>::const_iterator efs_iterator=eslpi_face_set.begin();
//      efs_iterator!=eslpi_face_set.end();efs_iterator++) {
//        cout << *efs_iterator << " ";
//      }
//      cout << endl;
#endif
        set<int> &static_face_eslp_set=static_face_eslp_sets[f];
        set<int> intersection;
        set_intersection(static_face_eslp_set.begin(),
          static_face_eslp_set.end(),eslpi_face_set.begin(),
          eslpi_face_set.end(),
          inserter(intersection,intersection.begin()));
#ifdef DEBUG
//      cout << "\tintersection = ";
//      for (set<int>::const_iterator i_iterator=intersection.begin();
//      i_iterator!=intersection.end();i_iterator++) {
//        cout << *i_iterator << " ";
//      }
//      cout << endl;
#endif
        CHECK_TEST(intersection.size()==face_neslp);

        int face_interior_neslp=(f<2 ? triangle_face_interior_neslp : 
          quadrilateral_face_interior_neslp);
        NumPtr<int> eslpi2;
        e.equallySpacedLatticePointIndicesOfFaceInterior(f,degree,eslpi2);
        CHECK_SAME(eslpi2.getNumber(),face_interior_neslp);
        set<int> eslpi_face_interior_set;
        for (int i=0;i<eslpi2.getNumber();i++) {
          eslpi_face_interior_set.insert(eslpi2[i]);
        }
        set<int> &static_face_interior_eslp_set=
          static_face_interior_eslp_sets[f];
        set<int> intersection2;
        set_intersection(static_face_interior_eslp_set.begin(),
          static_face_interior_eslp_set.end(),
          eslpi_face_interior_set.begin(),eslpi_face_interior_set.end(),
          inserter(intersection2,intersection2.begin()));
        CHECK_TEST(intersection2.size()==face_interior_neslp);
      }

      for (int l=0;l<e.numberCodimension2Shapes();l++) {
        NumPtr<int> eslpi2;
        e.equallySpacedLatticePointIndicesOfLineInterior(l,degree,eslpi2);
        CHECK_SAME(eslpi2.getNumber(),line_interior_neslp);
        set<int> eslpi_line_interior_set;
        for (int i=0;i<eslpi2.getNumber();i++) {
          eslpi_line_interior_set.insert(eslpi2[i]);
        }
        for (int f=0;f<Prism::facesPerShape();f++) {
          for (int fl=0;fl<Triangle::facesPerShape();fl++) {
            int l2=Prism::staticElementLineFromFace(f,fl);
            if (l2==l) {
              set<int> &static_face_eslp_set=static_face_eslp_sets[f];
              set<int> intersection2;
              set_intersection(static_face_eslp_set.begin(),
                static_face_eslp_set.end(),
                eslpi_line_interior_set.begin(),
                eslpi_line_interior_set.end(),
                inserter(intersection2,intersection2.begin()));
              CHECK_TEST(intersection2.size()==eslpi2.getNumber());
            }
          }
        }
      }
    }
  }
//
//
#ifdef DEBUG
  cout << "Hexahedron :" << endl;
#endif
  for (int degree=1;degree<=6;degree++) 
//int degree=2;
  {
#ifdef DEBUG
    cout << "\tdegree = " << degree << endl;
#endif
    int neslp=Hexahedron::staticNumberEquallySpacedLatticePoints(degree);
    int face_neslp=
      Quadrilateral::staticNumberEquallySpacedLatticePoints(degree);
    int face_interior_neslp=
      Quadrilateral::staticNumberInteriorEquallySpacedLatticePoints(degree);
    int line_interior_neslp=
      Interval::staticNumberInteriorEquallySpacedLatticePoints(degree);
    set<int> all_integers;
    for (int i=0;i<neslp;i++) all_integers.insert(i);

    for (int i=0;i<Hexahedron::verticesPerShape();i++) {
      int j=Hexahedron::staticEquallySpacedLatticePointIndexOfVertex(i,
        degree);
      CHECK_TEST(j>=0 && j<neslp);
    }

    NumPtr<set<int> > static_face_eslp_sets(Hexahedron::facesPerShape());
    NumPtr<set<int> > static_face_interior_eslp_sets(
      Hexahedron::facesPerShape());
    for (int f=0;f<Hexahedron::facesPerShape();f++) {
      set<int> &static_face_eslp_set=static_face_eslp_sets[f];
      NumPtr<int> eslpi;
      Hexahedron::staticEquallySpacedLatticePointIndicesOfFace(f,degree,
        eslpi);
      CHECK_SAME(eslpi.getNumber(),face_neslp);
      for (int i=0;i<eslpi.getNumber();i++) {
        static_face_eslp_set.insert(eslpi[i]);
      }
      set<int> intersection;
      set_intersection(all_integers.begin(),all_integers.end(),
        static_face_eslp_set.begin(),static_face_eslp_set.end(),
        inserter(intersection,intersection.begin()));
      CHECK_TEST(intersection.size()==eslpi.getNumber());

      set<int> &static_face_interior_eslp_set=
        static_face_interior_eslp_sets[f];
      NumPtr<int> eslpi2;
      Hexahedron::staticEquallySpacedLatticePointIndicesOfFaceInterior(f,
        degree,eslpi2);
      CHECK_SAME(eslpi2.getNumber(),face_interior_neslp);
      for (int i=0;i<eslpi2.getNumber();i++) {
        static_face_interior_eslp_set.insert(eslpi2[i]);
      }
      set<int> intersection2;
      set_intersection(static_face_eslp_set.begin(),
        static_face_eslp_set.end(),static_face_interior_eslp_set.begin(),
        static_face_interior_eslp_set.end(),
        inserter(intersection2,intersection2.begin()));
      CHECK_TEST(intersection2.size()==eslpi2.getNumber());
    }

    PermutationGenerator pg(6);
    while (pg.hasMore()) {
      const NumPtr<int> &p=pg.getNext();
      int lines[12]={0,1,2,3,4,5,6,7,8,9,10,11};
      int faces[6]={0,1,2,3,4,5};
      int vertices[8]={0,1,p[0]+2,p[1]+2,p[2]+2,p[3]+2,p[4]+2,p[5]+2};
      Hexahedron e(0,lines,faces,vertices);

      for (int i=0;i<e.numberVertices();i++) {
        int j=e.equallySpacedLatticePointIndexOfVertex(i,degree);
        CHECK_TEST(j>=0 && j<neslp);
      }

      for (int f=0;f<e.numberCodimension1Shapes();f++) {
        NumPtr<int> eslpi;
        e.equallySpacedLatticePointIndicesOfFace(f,degree,eslpi);
        CHECK_SAME(eslpi.getNumber(),face_neslp);
        set<int> eslpi_face_set;
        for (int i=0;i<eslpi.getNumber();i++) {
          eslpi_face_set.insert(eslpi[i]);
        }
        set<int> &static_face_eslp_set=static_face_eslp_sets[f];
        set<int> intersection;
        set_intersection(static_face_eslp_set.begin(),
          static_face_eslp_set.end(),eslpi_face_set.begin(),
          eslpi_face_set.end(),
          inserter(intersection,intersection.begin()));
        CHECK_TEST(intersection.size()==face_neslp);

        NumPtr<int> eslpi2;
        e.equallySpacedLatticePointIndicesOfFaceInterior(f,degree,eslpi2);
        CHECK_SAME(eslpi2.getNumber(),face_interior_neslp);
        set<int> eslpi_face_interior_set;
        for (int i=0;i<eslpi2.getNumber();i++) {
          eslpi_face_interior_set.insert(eslpi2[i]);
        }
        set<int> &static_face_interior_eslp_set=
          static_face_interior_eslp_sets[f];
        set<int> intersection2;
        set_intersection(static_face_interior_eslp_set.begin(),
          static_face_interior_eslp_set.end(),
          eslpi_face_interior_set.begin(),eslpi_face_interior_set.end(),
          inserter(intersection2,intersection2.begin()));
        CHECK_TEST(intersection2.size()==face_interior_neslp);
      }

      for (int l=0;l<e.numberCodimension2Shapes();l++) {
        NumPtr<int> eslpi2;
        e.equallySpacedLatticePointIndicesOfLineInterior(l,degree,eslpi2);
        CHECK_SAME(eslpi2.getNumber(),line_interior_neslp);
        set<int> eslpi_line_interior_set;
        for (int i=0;i<eslpi2.getNumber();i++) {
          eslpi_line_interior_set.insert(eslpi2[i]);
        }
        for (int f=0;f<Hexahedron::facesPerShape();f++) {
          for (int fl=0;fl<Quadrilateral::facesPerShape();fl++) {
            int l2=Hexahedron::staticElementLineFromFace(f,fl);
            if (l2==l) {
              set<int> &static_face_eslp_set=static_face_eslp_sets[f];
              set<int> intersection2;
              set_intersection(static_face_eslp_set.begin(),
                static_face_eslp_set.end(),
                eslpi_line_interior_set.begin(),
                eslpi_line_interior_set.end(),
                inserter(intersection2,intersection2.begin()));
              CHECK_TEST(intersection2.size()==eslpi2.getNumber());
            }
          }
        }
      }
    }
  }
//
#endif

  return EXIT_SUCCESS;
}

#include "NumPtr.C"
INSTANTIATE_NUMPTR(set<int>)

template insert_iterator<set<int, less<int>, allocator<int> > > 
  set_intersection<
    _Rb_tree_const_iterator<int>, 
    _Rb_tree_const_iterator<int>, 
    insert_iterator<set<int, less<int>, allocator<int> > > >
  ( _Rb_tree_const_iterator<int>, 
    _Rb_tree_const_iterator<int>, 
    _Rb_tree_const_iterator<int>, 
    _Rb_tree_const_iterator<int>, 
    insert_iterator<set<int, less<int>, allocator<int> > >
  );
