// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "AreaGraphTool.H"
#include "Boundary.H"
#include "MemoryDebugger.H"
#include "Shape.H"
#include "ShapeFunction.H"
#include "Tracer.H"

#include "VolGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKWindow.H"
#endif

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
#if (SPACEDIM==3)
//
  {
    Palette pal;
    AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
    REAL rlo[2]; rlo[0]=0.; rlo[1]=0.;
    REAL rhi[2]; rhi[0]=1.; rhi[1]=1.;
    AreaGraphTool gt("testLenoirCanonicalFacePoints","x","y",rlo,rhi,
      &cmap,0,0.5);

    int nv=3;
    NumPtr<Point<3> > vertices(nv);
    vertices[0]=Point<3>(0.,0.,0.);
    vertices[1]=Point<3>(1.,0.,0.);
    vertices[2]=Point<3>(0.,1.,0.);
    int edi=0;
    int f[3]={0,1,2};
    int v[6]={0,1,2};
    Triangle e(edi,f,v);

    for (int degree=1;degree<=6;degree++) 
//  int degree=2;
    {
      cout << "\tdegree = " << degree << endl;
      NumPtr<Point<2> > eslp;
      Triangle::staticEquallySpacedLatticePoints(degree,eslp);
      for (int bf=0;bf<e.numberCodimension1Shapes();bf++) 
//    int bf=0;
      {
        cout << "\t\tbf = " << bf << endl;
        NumPtr<Point<1> > canonical_face_points;
        e.lenoirCanonicalFacePoints(bf,eslp,canonical_face_points);
#ifdef DEBUG
//      for (int i=0;i<canonical_face_points.getNumber();i++) {
//        cout << "\tcanonical_face_points[" << i << "] = "
//             << canonical_face_points[i] << endl;
//      }
#endif
        Shape1 *new_line_element=e.newLineElement(bf);
        int neslp=
          new_line_element->numberEquallySpacedLatticePoints(degree);
        NumPtr<Point<3> > face_mapping_support_points(neslp);
        for (int v=0;v<new_line_element->numberVertices();v++) {
          int iv=new_line_element->
            equallySpacedLatticePointIndexOfVertex(v,degree);
          face_mapping_support_points[iv]=
            vertices[new_line_element->vertexIndex(v)];
        }
        new_line_element->computeInteriorLatticePoints(degree,bf,
          face_mapping_support_points);

        C0LagrangeIntervalPolynomials c0lip(degree);
        NumPtr<Point<3> > new_points(canonical_face_points.getNumber());
        for (int ij=0;ij<new_points.getNumber();ij++) {
          NumPtr<REAL> values;
          c0lip.computeValues(canonical_face_points[ij],values);
          new_points[ij]=0.;
          for (int v=0;v<values.getNumber();v++) {
            new_points[ij]+=face_mapping_support_points[v]*values[v];
          }
        }
#ifdef DEBUG
//      for (int i=0;i<new_points.getNumber();i++) {
//        cout << "\tnew_points[" << i << "] = " << new_points[i] << endl;
//      }
#endif
        gt.newPage();
        gt.setfgColor("black");
        gt.drawAxes();
        gt.setfgColor("blue");
        for (int i=0;i<new_points.getNumber();i++) {
          const Point<2> &eslpi=eslp[i];
          gt.movePen(eslpi[0],eslpi[1]);
          const Point<3> &npi=new_points[i];
          gt.drawLine(npi[0],npi[1]);
        }
        gt.flush();
        AreaGraphTool::WINDOW_TYPE::QuitButton qb;

        delete new_line_element; new_line_element=0;
      }
    }
  }
//
#endif

#if (SPACEDIM==3)
//
  {
    BOOLEAN use_lighting=TRUE;
    BOOLEAN rotate_best=FALSE;
    BOOLEAN show_bounding_box=TRUE;
    Palette pal;
    REAL winsize=0.5;
    Vector3 low((1.-sqrt(3.))*0.5,(1.-sqrt(3.))*0.5,(1.-sqrt(3.))*0.5);
    Vector3 high((1.+sqrt(3.))*0.5,(1.+sqrt(3.))*0.5,(1.+sqrt(3.))*0.5);
    VolGraphTool vgt("lenoirCanonicalFacePoints",low,high,pal,
        use_lighting,winsize,rotate_best,show_bounding_box);

    int nv=4;
    NumPtr<Point<3> > vertices(nv);
    vertices[0]=Tetrahedron::canonicalVertex(0);
    vertices[1]=Tetrahedron::canonicalVertex(1);
    vertices[2]=Tetrahedron::canonicalVertex(2);
    vertices[3]=Tetrahedron::canonicalVertex(3);

//  Point<3> center(0.5,0.5,0.5);
//  BallBoundary bb(center,sqrt(center.square()));
//  Point<3> center3(0.,0.,0.);
//  BallBoundary bb3(center3,1.);
    Point<3> centroid(1./3.,1./3.,1./3.);
    PlaneBoundary pb0(centroid,vertices[0],vertices[2],vertices[3]);
    PlaneBoundary pb1(centroid,vertices[0],vertices[1],vertices[3]);
    PlaneBoundary pb2(centroid,vertices[0],vertices[1],vertices[2]);
    PlaneBoundary pb3(centroid,vertices[1],vertices[2],vertices[3]);

    int edi=0;
    int l[6]={0,1,2,3,4,5};
    int f[4]={0,1,2,3};
    int v[4]={0,1,2,3};
    Tetrahedron e(edi,l,f,v);
    for (int degree=1;degree<=6;degree++) 
//  int degree=3;
    {
      cout << "\tdegree = " << degree << endl;
      NumPtr<Point<3> > eslp;
      Tetrahedron::staticEquallySpacedLatticePoints(degree,eslp);
#ifdef DEBUG
//    for (int i=0;i<eslp.getNumber();i++) {
//      cout << "\teslp[" << i << "] = " << eslp[i] << endl;
//    }
#endif
      for (int bf=0;bf<e.numberCodimension1Shapes();bf++) 
//    int bf=0;
      {
        cout << "\tbf = " << bf << endl;
        NumPtr<Point<2> > canonical_face_points;
        e.lenoirCanonicalFacePoints(bf,eslp,canonical_face_points);
#ifdef DEBUG
//      for (int i=0;i<canonical_face_points.getNumber();i++) {
//        cout << "\tcanonical_face_points[" << i << "] = "
//             << canonical_face_points[i] << endl;
//      }
#endif
        Shape2 *new_face_element=e.newFaceElement(bf);
        NumPtr<Point<3> > fv(new_face_element->numberVertices());
        for (int i=0;i<new_face_element->numberVertices();i++) {
          fv[i]=vertices[new_face_element->vertexIndex(i)];
        }
        int neslp=
          new_face_element->numberEquallySpacedLatticePoints(degree);
        NumPtr<Point<3> > face_mapping_support_points(neslp);
        for (int v=0;v<new_face_element->numberVertices();v++) {
          int iv=new_face_element->
            equallySpacedLatticePointIndexOfVertex(v,degree);
          face_mapping_support_points[iv]=
            vertices[new_face_element->vertexIndex(v)];
        }
        for (int l=0;l<new_face_element->numberCodimension1Shapes();l++) {
          Shape1 *new_line_element=new_face_element->newLineElement(l);
          int nieslp=new_line_element->
            numberInteriorEquallySpacedLatticePoints(degree);
          NumPtr<Point<3> > fsp(nieslp);
          switch (bf) {
            case 0:
              pb0.getIntermediatePointsOn(*new_line_element,vertices,fsp);
              break;
            case 1:
              pb1.getIntermediatePointsOn(*new_line_element,vertices,fsp);
              break;
            case 2:
              pb2.getIntermediatePointsOn(*new_line_element,vertices,fsp);
              break;
            case 3:
              pb3.getIntermediatePointsOn(*new_line_element,vertices,fsp);
              break;
          }
          NumPtr<int> indices;
          new_face_element->equallySpacedLatticePointIndicesOfFaceInterior(
            l,degree,indices);
          for (int i=0;i<fsp.getNumber();i++) {
            int iv=indices[i];
            face_mapping_support_points[iv]=fsp[i];
          }
          delete new_line_element; new_line_element=0;
        }
        switch (bf) {
          case 0:
            pb0.computeInteriorLatticePoints(*new_face_element,degree,bf,
              face_mapping_support_points);
            break;
          case 1:
            pb1.computeInteriorLatticePoints(*new_face_element,degree,bf,
              face_mapping_support_points);
            break;
          case 2:
            pb2.computeInteriorLatticePoints(*new_face_element,degree,bf,
              face_mapping_support_points);
            break;
          case 3:
            pb3.computeInteriorLatticePoints(*new_face_element,degree,bf,
              face_mapping_support_points);
            break;
        }
#ifdef DEBUG
//      for (int i=0;i<face_mapping_support_points.getNumber();i++) {
//        cout << "\tface_mapping_support_points[" << i << "] = "
//             << face_mapping_support_points[i] << endl;
//      }
#endif
        C0LagrangeTrianglePolynomials c0ltp(degree);
        NumPtr<Point<3> > 
          new_points(canonical_face_points.getNumber());
        for (int ijk=0;ijk<new_points.getNumber();ijk++) {
          NumPtr<REAL> values;
          c0ltp.computeValues(canonical_face_points[ijk],values);
          new_points[ijk]=0.;
          for (int v=0;v<values.getNumber();v++) {
            new_points[ijk]+=face_mapping_support_points[v]*values[v];
          }
        }
#ifdef DEBUG
//      for (int i=0;i<new_points.getNumber();i++) {
//        cout << "\tnew_points[" << i << "] = " << new_points[i] << endl;
//      }
#endif
        VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
          vgt.getWindow()->getColor("black");

        vgt.newPage();
        vgt.expose();
        for (BOOLEAN do_best=FALSE;;do_best=TRUE) {
          VolGraphTool::DrawDelimiter dd(do_best,&vgt);
//        e.draw(xc,vertices,vgt);
          vgt.setfgColor("blue");
          for (int i=0;i<new_points.getNumber();i++) {
            Vector3 v0,v1;
            const Point<3> &eslpi=eslp[i];
            const Point<3> &bfp=new_points[i];
            for (int j=0;j<3;j++) {
              v0[j]=eslpi[j];
              v1[j]=bfp[j];
            }
            vgt.drawLine(v0,v1);
          }
          if (do_best==TRUE) break;
        }
        vgt.eventLoop();

        delete new_face_element; new_face_element=0;
      }
    }
  }
//
#endif

  return EXIT_SUCCESS;
}
