// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "Boundary.H"
#include "MemoryDebugger.H"
#include "Shape.H"
#include "ShapeFunction.H"
#include "Tracer.H"

#include "VolGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKWindow.H"
#endif

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif

  BOOLEAN use_lighting=TRUE;
  BOOLEAN rotate_best=FALSE;
  BOOLEAN show_bounding_box=TRUE;
  Palette pal;
  REAL winsize=0.5;
#if (SPACEDIM==3)
//
  {
    Vector3 low((1.-sqrt(3.))*0.5,(1.-sqrt(3.))*0.5,(1.-sqrt(3.))*0.5);
    Vector3 high((1.+sqrt(3.))*0.5,(1.+sqrt(3.))*0.5,(1.+sqrt(3.))*0.5);
    VolGraphTool *vgt=OPERATOR_NEW 
      VolGraphTool("Tetrahedron::lenoirCanonicalLinePoints",low,high,pal,
        use_lighting,winsize,rotate_best,show_bounding_box);

    int nv=4;
    NumPtr<Point<3> > vertices(nv);
    vertices[0]=Tetrahedron::canonicalVertex(0);
    vertices[1]=Tetrahedron::canonicalVertex(1);
    vertices[2]=Tetrahedron::canonicalVertex(2);
    vertices[3]=Tetrahedron::canonicalVertex(3);

    int edi=0;
    int l[6]={0,1,2,3,4,5};
    int f[4]={0,1,2,3};
    int v[4]={0,1,2,3};
    Tetrahedron e(edi,l,f,v);
    for (int degree=1;degree<=10;degree++) 
//  int degree=2;
    {
      cout << "\tdegree = " << degree << endl;
      NumPtr<Point<SPACEDIM> > eslp;
      Tetrahedron::staticEquallySpacedLatticePoints(degree,eslp);
#ifdef DEBUG
//    for (int i=0;i<eslp.getNumber();i++) {
//      cout << "\teslp[" << i << "] = " << eslp[i] << endl;
//    }
#endif
      for (int bl=0;bl<e.numberCodimension2Shapes();bl++) 
//    int bl=5;
      {
        cout << "\t\tbl = " << bl << endl;
//      Interval line_element(0,lv);
        Shape1 *new_line_element=e.newLineElement(bl);
        NumPtr<Point<1> > canonical_line_points;
        e.lenoirCanonicalLinePoints(bl,eslp,canonical_line_points);
#ifdef DEBUG
//      for (int i=0;i<canonical_line_points.getNumber();i++) {
//        cout << "\tcanonical_line_points[" << i << "] = "
//             << canonical_line_points[i] << endl;
//      }
#endif
        NumPtr<Point<SPACEDIM> > 
          line_mapping_support_points(degree+1);
        for (int v=0;v<new_line_element->numberVertices();v++) {
          int iv=new_line_element->
            equallySpacedLatticePointIndexOfVertex(v,degree);
          line_mapping_support_points[iv]=
            vertices[new_line_element->vertexIndex(v)];
        }
        new_line_element->computeInteriorLatticePoints(degree,bl,
          line_mapping_support_points);
        C0LagrangeIntervalPolynomials c0lip(degree);
        NumPtr<Point<3> > new_points(canonical_line_points.getNumber());
        for (int ijk=0;ijk<new_points.getNumber();ijk++) {
          NumPtr<REAL> values;
          c0lip.computeValues(canonical_line_points[ijk],values);
          new_points[ijk]=0.;
          for (int v=0;v<values.getNumber();v++) {
            new_points[ijk]+=line_mapping_support_points[v]*values[v];
          }
        }
#ifdef DEBUG
//      for (int i=0;i<new_points.getNumber();i++) {
//        cout << "\tnew_points[" << i << "] = " 
//             << new_points[i] << endl;
//      }
#endif
        VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
          vgt->getWindow()->getColor("black");

        vgt->newPage();
        vgt->expose();
        for (BOOLEAN do_best=FALSE;;do_best=TRUE) {
          VolGraphTool::DrawDelimiter dd(do_best,vgt);
//        e.draw(xc,vertices,*vgt);
          vgt->setfgColor("blue");
          for (int i=0;i<new_points.getNumber();i++) {
            Vector3 v0,v1;
            const Point<SPACEDIM> &eslpi=eslp[i];
            const Point<SPACEDIM> &npi=new_points[i];
            for (int j=0;j<3;j++) {
              v0[j]=eslpi[j];
              v1[j]=npi[j];
            }
            vgt->drawLine(v0,v1);
          }
          if (do_best==TRUE) break;
        }
        vgt->eventLoop();

        delete new_line_element; new_line_element=0;
      }
    }
    delete vgt;
  }
//
/*obsolete:
  {
    Vector3 low((1.-sqrt(3.))*0.5,(1.-sqrt(3.))*0.5,(1.-sqrt(3.))*0.5);
    Vector3 high((1.+sqrt(3.))*0.5,(1.+sqrt(3.))*0.5,(1.+sqrt(3.))*0.5);
    VolGraphTool *vgt=OPERATOR_NEW 
      VolGraphTool("Prism::lenoirCanonicalLinePoints",low,high,pal,
        use_lighting,winsize,rotate_best,show_bounding_box);

    Point<3> center(0.5,0.5,0.5);
    BallBoundary bb(center,sqrt(center.square()));

    int nv=6;
    NumPtr<Point<3> > vertices(nv);
    vertices[0]=Prism::canonicalVertex(0);
    vertices[1]=Prism::canonicalVertex(1);
    vertices[2]=Prism::canonicalVertex(2);
    vertices[3]=Prism::canonicalVertex(3);
    vertices[4]=Prism::canonicalVertex(4);
    vertices[5]=Prism::canonicalVertex(5);
    int edi=0;
    int l[9]={0,1,2,3,4,5,6,7,8};
    int f[5]={0,1,2,3,4};
    int v[6]={0,1,2,3,4,5};
    Prism e(edi,l,f,v);
    for (int degree=1;degree<=10;degree++) 
//  int degree=2;
    {
      cout << "\tdegree = " << degree << endl;
      NumPtr<Point<SPACEDIM> > eslp;
      Prism::staticEquallySpacedLatticePoints(degree,eslp);
#ifdef DEBUG
      for (int i=0;i<eslp.getNumber();i++) {
        cout << "\teslp[" << i << "] = " << eslp[i] << endl;
      }
#endif
//    for (int bl=0;bl<e.numberCodimension2Shapes();bl++) 
      int bl=8;
      {
        cout << "\tbl = " << bl << endl;
        NumPtr<Point<1> > canonical_line_points;
//      Prism::lenoirCanonicalLinePoints(bl,eslp,canonical_line_points);
        OBSOLETE("switch to Triangle");
#ifdef DEBUG
        for (int i=0;i<canonical_line_points.getNumber();i++) {
          cout << "\tcanonical_line_points[" << i << "] = "
               << canonical_line_points[i] << endl;
        }
#endif

        int fv[2];
        for (int i=0;i<2;i++) {
          fv[i]=e.vertexIndex(Prism::staticElementVertexFromLine(bl,i));
        }
        Interval line_element(0,fv);

        NumPtr<Point<SPACEDIM> > ball_line_points;
//      bb.getMappedPointsOn(line_element,vertices,
//        canonical_line_points,ball_line_points);
        OBSOLETE("getMappedPointsOn is obsolete");
#ifdef DEBUG
        for (int i=0;i<ball_line_points.getNumber();i++) {
          cout << "\tball_line_points[" << i << "] = " 
               << ball_line_points[i] << endl;
        }
#endif
        VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
          vgt->getWindow()->getColor("black");

        vgt->newPage();
        vgt->expose();
        for (BOOLEAN do_best=FALSE;;do_best=TRUE) {
          VolGraphTool::DrawDelimiter dd(do_best,vgt);
//        e.draw(xc,vertices,*vgt);
          vgt->setfgColor("blue");
          for (int i=0;i<ball_line_points.getNumber();i++) {
            Vector3 v0,v1;
            const Point<SPACEDIM> &eslpi=eslp[i];
            const Point<SPACEDIM> &blp=ball_line_points[i];
            for (int j=0;j<3;j++) {
              v0[j]=eslpi[j];
              v1[j]=blp[j];
            }
            vgt->drawLine(v0,v1);
          }
          if (do_best==TRUE) break;
        }
        vgt->eventLoop();
      }
    }
    delete vgt;
  }
*/
#endif

  return EXIT_SUCCESS;
}
