// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "Boundary.H"
#include "MemoryDebugger.H"
#include "Shape.H"
#include "ShapeFunction.H"
#include "Tracer.H"

#include "VolGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKWindow.H"
#endif

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif

  BOOLEAN use_lighting=TRUE;
  BOOLEAN rotate_best=FALSE;
  BOOLEAN show_bounding_box=TRUE;
  Palette pal;
  REAL winsize=0.5;
#if (SPACEDIM==3)
//
//Vector3 low((1.-sqrt(3.))*0.5,(1.-sqrt(3.))*0.5,(1.-sqrt(3.))*0.5);
//Vector3 high((1.+sqrt(3.))*0.5,(1.+sqrt(3.))*0.5,(1.+sqrt(3.))*0.5);
//VolGraphTool *vgt=OPERATOR_NEW 
//  VolGraphTool("lenoirFaceInterpolation",low,high,pal,
//    use_lighting,winsize,rotate_best,show_bounding_box);
  {

//  Point<3> center(0.5,0.5,0.5);
//  BallBoundary bb(center,sqrt(center.square()));
//  Point<3> center3(0.,0.,0.);
//  BallBoundary bb3(center3,1.);

    int nv=4;
    NumPtr<Point<3> > vertices(nv);
    vertices[0]=Tetrahedron::canonicalVertex(0);
    vertices[1]=Tetrahedron::canonicalVertex(1);
    vertices[2]=Tetrahedron::canonicalVertex(2);
    vertices[3]=Tetrahedron::canonicalVertex(3);
#ifdef DEBUG
//  for (int i=0;i<4;i++) {
//    cout << "\tvertices[" << i << "] = " << vertices[i] << endl;
//  }
#endif
    int edi=0;
    int l[6]={0,1,2,3,4,5};
    int f[4]={0,1,2,3};
    int v[4]={0,1,2,3};
    Tetrahedron e(edi,l,f,v);
    for (int degree=1;degree<=6;degree++) 
//  int degree=2;
    {
      cout << "\tdegree = " << degree << endl;
      C0LagrangeTetrahedronPolynomials ltp(degree);
      NumPtr<int> renumber;
      ltp.shapeFunctionsOrder(&e,renumber);
      NumPtr<int> renumber_inverse(renumber.getNumber());
      for (int i=0;i<renumber.getNumber();i++) {
        renumber_inverse[renumber[i]]=i;
      }
#ifdef DEBUG
//    for (int i=0;i<renumber.getNumber();i++) {
//      cout << "\trenumber[" << i << "] = " << renumber[i] << endl;
//    }
//    for (int i=0;i<renumber_inverse.getNumber();i++) {
//      cout << "\trenumber_inverse[" << i << "] = " 
//           << renumber_inverse[i] << endl;
//    }
#endif

      NumPtr<Point<SPACEDIM> > eslp;
      Tetrahedron::staticEquallySpacedLatticePoints(degree,eslp);
#ifdef DEBUG
//    for (int i=0;i<eslp.getNumber();i++) {
//      cout << "\teslp[" << i << "] = " << eslp[i] << endl;
//    }
#endif
      for (int bf=0;bf<e.numberCodimension1Shapes();bf++) 
//    int bf=0;
      {
        cout << "\t\tbf = " << bf << endl;

        NumPtr<Point<SPACEDIM> > 
          higher_order_mapping_support_points(eslp.getNumber());
        e.initializeMappingSupportPoints(degree,vertices,
          higher_order_mapping_support_points);

        NumPtr<Point<3> > *old_points=OPERATOR_NEW NumPtr<Point<3> >;
          old_points->allocate(eslp.getNumber());
        NumPtr<Point<3> > *new_points=OPERATOR_NEW NumPtr<Point<3> >;
          new_points->allocate(eslp.getNumber());

        NumPtr<Point<2> > canonical_face_points;
        e.lenoirCanonicalFacePoints(bf,eslp,canonical_face_points);

        Shape2 *new_face_element=e.newFaceElement(bf);

        C0LagrangeTrianglePolynomials c0ltp(1);
        NumPtr<Point<3> > fv(new_face_element->numberVertices());
        for (int i=0;i<new_face_element->numberVertices();i++) {
          fv[i]=vertices[new_face_element->vertexIndex(i)];
        }
        for (int ijk=0;ijk<old_points->getNumber();ijk++) {
          NumPtr<REAL> values;
          c0ltp.computeValues(canonical_face_points[ijk],values);
          (*old_points)[ijk]=
            fv[0]*values[0]+fv[1]*values[1]+fv[2]*values[2];
        }
        for (int m=2;m<=degree;m++) {
          int neslp=new_face_element->numberEquallySpacedLatticePoints(m);
          NumPtr<Point<3> > face_mapping_support_points(neslp);
          for (int v=0;v<new_face_element->numberVertices();v++) {
            int iv=
              new_face_element->equallySpacedLatticePointIndexOfVertex(v,m);
            face_mapping_support_points[iv]=
              vertices[new_face_element->vertexIndex(v)];
          }
          for (int l=0;l<new_face_element->numberCodimension1Shapes();l++) {
            int line_element_index=
              new_face_element->codimension1ShapeIndex(l);
            Shape1 *new_line_element=new_face_element->newLineElement(l);
            int nieslp=
              new_line_element->numberInteriorEquallySpacedLatticePoints(m);
            NumPtr<Point<SPACEDIM> > fsp(nieslp);
            new_line_element->getIntermediatePoints(vertices,fsp);
            NumPtr<int> indices;
            new_face_element->
              equallySpacedLatticePointIndicesOfFaceInterior(l,m,indices);
            for (int i=0;i<fsp.getNumber();i++) {
              int iv=indices[i];
              face_mapping_support_points[iv]=fsp[i];
            }
            delete new_line_element; new_line_element=0;
          }
          new_face_element->computeInteriorLatticePoints(m,bf,
            face_mapping_support_points);
          C0LagrangeTrianglePolynomials c0ltp(m);
          for (int ijk=0;ijk<new_points->getNumber();ijk++) {
            NumPtr<REAL> values;
            c0ltp.computeValues(canonical_face_points[ijk],values);
            (*new_points)[ijk]=0.;
            for (int v=0;v<values.getNumber();v++) {
              (*new_points)[ijk]+=face_mapping_support_points[v]*values[v];
            }
          }
          e.lenoirInterpolationForFace(bf,m,eslp,
            *old_points,*new_points,higher_order_mapping_support_points);
          swap(old_points,new_points);
        }
        delete old_points; old_points=0;
        delete new_points; new_points=0;

        for (int i=0;i<eslp.getNumber();i++) {
          REAL error=
            eslp[i].distance(higher_order_mapping_support_points[i]);
          if (error>256.*DBL_EPSILON) {
            cout << "\tdegree = " << degree << endl;
            cout << "\teslp[" << i << "] = " << eslp[i] << " "
                 << higher_order_mapping_support_points[i] << " " << error
                 << " " << error/DBL_EPSILON << endl;
            OBSOLETE("bad higher_order_mapping_support_points");
          }
        }
/*
        VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
          vgt->getWindow()->getColor("black");

        vgt->newPage();
        vgt->expose();
        for (BOOLEAN do_best=FALSE;;do_best=TRUE) {
          VolGraphTool::DrawDelimiter dd(do_best,vgt);
//        e.draw(xc,vertices,*vgt);
          vgt->setfgColor("blue");

          int ijk=0;
          for (int k=0;k<degree;k++) {
            for (int j=0;j+k<degree;j++) {
              for (int i=0;i+j+k<degree;i++) {
                Vector3 v0,v1;
                const Point<SPACEDIM> &homsp0=
                 higher_order_mapping_support_points[renumber[ijk]];
                const Point<SPACEDIM> &homsp1=
                  higher_order_mapping_support_points[renumber[ijk+1]];
#ifdef DEBUG
//              if (do_best) {
//                cout << "\ti,j,k = " << i << " " << j << " " << k
//                     << " homsp0[" << renumber[ijk] << "] = " << homsp0
//                     << " homsp1[" << renumber[ijk+1] << "] = " 
//                     << homsp1 << endl;
//              }
#endif
                for (int j=0;j<3;j++) {
                  v0[j]=homsp0[j];
                  v1[j]=homsp1[j];
                }
                vgt->drawLine(v0,v1);
                ijk++;
              }
              ijk++;
            }
            ijk++;
          }
          ijk=0;
          for (int k=0;k<degree;k++) {
            for (int j=0;j+k<degree;j++) {
              for (int i=0;i+j+k<degree;i++) {
                Vector3 v0,v1;
                const Point<SPACEDIM> &homsp0=
                 higher_order_mapping_support_points[renumber[ijk]];
                const Point<SPACEDIM> &homsp1=
                  higher_order_mapping_support_points[
                    renumber[ijk+degree+1-j-k]];
#ifdef DEBUG
//              if (do_best) {
//                cout << "\ti,j,k = " << i << " " << j << " " << k
//                     << " homsp0[" << renumber[ijk] << "] = " << homsp0
//                     << " homsp1[" << renumber[ijk+degree+1-j-k] 
//                     << "] = " << homsp1 << endl;
//              }
#endif
                for (int j=0;j<3;j++) {
                  v0[j]=homsp0[j];
                  v1[j]=homsp1[j];
                }
                vgt->drawLine(v0,v1);
                ijk++;
              }
              ijk++;
            }
            ijk++;
          }
          ijk=0;
          for (int k=0;k<degree;k++) {
            for (int j=0;j+k<degree;j++) {
              for (int i=0;i+j+k<degree;i++) {
                Vector3 v0,v1;
                const Point<SPACEDIM> &homsp0=
                 higher_order_mapping_support_points[renumber[ijk]];
                const Point<SPACEDIM> &homsp1=
                  higher_order_mapping_support_points[
                    renumber[ijk+(degree*(degree+3)+k*(k-3))/2+1
                                    -j-k*degree]];
#ifdef DEBUG
//              if (do_best) {
//                cout << "\ti,j,k = " << i << " " << j << " " << k
//                     << " homsp0[" << renumber[ijk] << "] = " << homsp0
//                     << " homsp1[" 
//              << renumber[ijk+(degree*(degree+3)+k*(k-3))/2+1 -j-k*degree]
//                     << "] = " << homsp1 << endl;
//              }
#endif
                for (int j=0;j<3;j++) {
                  v0[j]=homsp0[j];
                  v1[j]=homsp1[j];
                }
                vgt->drawLine(v0,v1);
                ijk++;
              }
              ijk++;
            }
            ijk++;
          }

          if (do_best==TRUE) break;
        }
        vgt->eventLoop();
*/
        delete new_face_element; new_face_element=0;
      }
    }
  }
//
/*
  {
    Point<3> center(0.5,0.5,0.5);
    BallBoundary bb(center,sqrt(center.square()));
    Point<3> center4(0.,0.,0.5);
    BallBoundary bb4(center4,sqrt(1.5));

    int nv=6;
    NumPtr<Point<3> > vertices(nv);
    vertices[0]=Prism::canonicalVertex(0);
    vertices[1]=Prism::canonicalVertex(1);
    vertices[2]=Prism::canonicalVertex(2);
    vertices[3]=Prism::canonicalVertex(3);
    vertices[4]=Prism::canonicalVertex(4);
    vertices[5]=Prism::canonicalVertex(5);
#ifdef DEBUG
//  for (int i=0;i<6;i++) {
//    cout << "\tvertices[" << i << "] = " << vertices[i] << endl;
//  }
#endif
    int edi=0;
    int l[9]={0,1,2,3,4,5,6,7,8};
    int f[5]={0,1,2,3,4};
    int v[6]={0,1,2,3,4,5};
    Prism e(edi,l,f,v);
    for (int degree=2;degree<=10;degree++) 
//  int degree=2;
    {
      cout << "\tdegree = " << degree << endl;
      C0LagrangePrismPolynomials lwp(degree);
      NumPtr<int> renumber;
      lwp.shapeFunctionsOrder(&e,renumber);
      NumPtr<int> renumber_inverse(renumber.getNumber());
      for (int i=0;i<renumber.getNumber();i++) {
        renumber_inverse[renumber[i]]=i;
      }
#ifdef DEBUG
//    for (int i=0;i<renumber.getNumber();i++) {
//      cout << "\trenumber[" << i << "] = " << renumber[i] << endl;
//    }
//    for (int i=0;i<renumber_inverse.getNumber();i++) {
//      cout << "\trenumber_inverse[" << i << "] = "
//           << renumber_inverse[i] << endl;
//    }
#endif
      NumPtr<Point<SPACEDIM> > eslp;
      Prism::staticEquallySpacedLatticePoints(degree,eslp);
#ifdef DEBUG
//    for (int i=0;i<eslp.getNumber();i++) {
//      cout << "\teslp[" << i << "] = " << eslp[i] << endl;
//    }
#endif
//    for (int bf=0;bf<e.numberCodimension1Shapes();bf++) 
      int bf=4;
      {
        cout << "\tbf = " << bf << endl;
        NumPtr<Point<SPACEDIM> > 
          higher_order_mapping_support_points(eslp.getNumber());
        if (Prism::staticNumberVerticesOnFace(bf)==3) {
          TRACER_CALL(t,"triangular face");

          int opposite_bf=Prism::oppositeFace(bf);
          NumPtr<int> ofeslpi;
          e.staticEquallySpacedLatticePointIndicesOfFace(opposite_bf,
            degree,ofeslpi);
          for (int i=0;i<ofeslpi.getNumber();i++) {
            int ijk=ofeslpi[i];
            int dof=renumber[ijk];
            const Point<SPACEDIM> &eslpi=eslp[ijk];
            REAL bc2=1.-eslpi[0]-eslpi[1];
            higher_order_mapping_support_points[dof]=
              (vertices[1]*eslpi[0]+vertices[2]*eslpi[1]
             +vertices[0]*bc2)*(1.-eslpi[2])
             +(vertices[4]*eslpi[0]+vertices[5]*eslpi[1]
             +vertices[3]*bc2)*eslpi[2];
          }

          int fl[3]={0,1,2};
          int fv[3];
          for (int i=0;i<3;i++) {
            fv[i]=e.vertexIndex(Prism::staticElementVertexFromFace(bf,i));
          }
#ifdef DEBUG
//        for (int i=0;i<3;i++) {
//          cout << "\tfv[" << i << "] = " << fv[i] << endl;
//        }
#endif
          Triangle face_element(0,fl,fv);
#ifdef DEBUG
//        cout << "\tface_element:" << endl;
//        face_element.printOn(cout);
#endif
          NumPtr<int> feslpi;
          e.staticEquallySpacedLatticePointIndicesOfFace(bf,degree,feslpi);
          NumPtr<Point<2> > canonical_face_points(feslpi.getNumber());
          for (int i=0;i<feslpi.getNumber();i++) {
            const Point<SPACEDIM> &eslpi=eslp[feslpi[i]];
            Point<2> &cfp=canonical_face_points[i];
            cfp[0]=eslpi[0];
            cfp[1]=eslpi[1];
          }
          NumPtr<Point<SPACEDIM> > bp;
//        bb.getMappedPointsOn(face_element,vertices,canonical_face_points,
//          bp);
          OBSOLETE("getMappedPointsOn is obsolete");

          for (int i=0;i<feslpi.getNumber();i++) {
            higher_order_mapping_support_points[renumber[feslpi[i]]]=bp[i];
          }
//        Prism::linearInterpolationFromTriangularFacesForInterior(degree,
//          eslp,higher_order_mapping_support_points);
          OBSOLETE("getMappedPointsOn is obsolete");
#ifdef DEBUG
//        for (int i=0;i<higher_order_mapping_support_points.getNumber();
//        i++) {
//          cout << "\thigher_order_mapping_support_points[" << i 
//               << "] = " << higher_order_mapping_support_points[i] 
//               << endl;
//        }
#endif
        } else {
          for (int i=0;i<eslp.getNumber();i++) {
            const Point<3> &eslpi=eslp[i];
            REAL bc2=1.-eslpi[0]-eslpi[1];
            higher_order_mapping_support_points[renumber[i]]=
              (vertices[1]*eslpi[0]+vertices[2]*eslpi[1]
              +vertices[0]*bc2)*(1.-eslpi[2])
             +(vertices[4]*eslpi[0]+vertices[5]*eslpi[1]
              +vertices[3]*bc2)*eslpi[2];
          }
#ifdef DEBUG
//        for (int i=0;i<higher_order_mapping_support_points.getNumber();
//        i++) {
//          cout << "\thigher_order_mapping_support_points[" << i << "] = "
//               << higher_order_mapping_support_points[i] << endl;
//        }
#endif
          NumPtr<Point<SPACEDIM> > *old_points=
            OPERATOR_NEW NumPtr<Point<SPACEDIM> >;
          old_points->allocate(eslp.getNumber());
          NumPtr<Point<SPACEDIM> > *new_points=
            OPERATOR_NEW NumPtr<Point<SPACEDIM> >;
          new_points->allocate(eslp.getNumber());

          NumPtr<Point<2> > canonical_face_points;
//        Prism::lenoirCanonicalFacePoints(bf,eslp,canonical_face_points);
          OBSOLETE("switch to Triangle");
          int fl[4]={0,1,2,3};
          int fv[4];
          for (int i=0;i<4;i++) {
            fv[i]=e.vertexIndex(Prism::staticElementVertexFromFace(bf,i));
          }
#ifdef DEBUG
//        for (int i=0;i<4;i++) {
//          cout << "\tfv[" << i << "] = " << fv[i] << endl;
//        }
#endif
          Quadrilateral face_element(0,fl,fv);
#ifdef DEBUG
//        cout << "\tface_element:" << endl;
//        face_element.printOn(cout);
#endif
          C0LagrangeQuadrilateralPolynomials c0lqp(1);
          const Point<SPACEDIM> &fv0=
            vertices[face_element.vertexIndex(0)];
          const Point<SPACEDIM> &fv1=
            vertices[face_element.vertexIndex(1)];
          const Point<SPACEDIM> &fv2=
            vertices[face_element.vertexIndex(2)];
          const Point<SPACEDIM> &fv3=
            vertices[face_element.vertexIndex(3)];
#ifdef DEBUG
//        cout << "\tfv0 = " << fv0 << endl;
//        cout << "\tfv1 = " << fv1 << endl;
//        cout << "\tfv2 = " << fv2 << endl;
//        cout << "\tfv3 = " << fv3 << endl;
#endif
          for (int i=0;i<old_points->getNumber();i++) {
            NumPtr<REAL> values;
            c0lqp.computeValues(canonical_face_points[i],values);
            (*old_points)[i]=fv0*values[0]+fv1*values[1]
                            +fv2*values[2]+fv3*values[3];
          }
#ifdef DEBUG
//        for (int i=0;i<old_points->getNumber();i++) {
//          cout << "\told_points[" << i << "] = "
//               << (*old_points)[i] << endl; 
//        }
#endif
          for (int m=2;m<=degree;m++) {
            C0LagrangeQuadrilateralPolynomials c0lqp(m);

            NumPtr<Point<2> > feslp;
            Quadrilateral::staticEquallySpacedLatticePoints(m,feslp);
#ifdef DEBUG
//          for (int i=0;i<feslp.getNumber();i++) {
//            cout << "\tfeslp[" << i << "] = "
//                 << feslp[i] << endl; 
//          }
#endif
            NumPtr<Point<SPACEDIM> > fsp;
            if (bf==4) {
//            bb4.getMappedPointsOn(face_element,vertices,feslp,fsp);
              OBSOLETE("getMappedPointsOn is obsolete");
            } else {
//            bb.getMappedPointsOn(face_element,vertices,feslp,fsp);
              OBSOLETE("getMappedPointsOn is obsolete");
            }
#ifdef DEBUG
//          for (int i=0;i<fsp.getNumber();i++) {
//            cout << "\tfsp[" << i << "] = "
//                 << fsp[i] << endl; 
//          }
#endif

            for (int i=0;i<new_points->getNumber();i++) {
              NumPtr<REAL> values;
              c0lqp.computeValues(canonical_face_points[i],values);
              (*new_points)[i]=0.;
              for (int v=0;v<values.getNumber();v++) {
                (*new_points)[i]+=fsp[v]*values[v];
              }
            }
#ifdef DEBUG
//          for (int i=0;i<new_points->getNumber();i++) {
//            cout << "\tnew_points[" << i << "] = "
//                 << (*new_points)[i] << endl;
//          }
#endif
//          Prism::lenoirInterpolationForFace(bf,m,eslp,
//            *old_points,*new_points,higher_order_mapping_support_points);
            OBSOLETE("switch to Triangle");
#ifdef DEBUG
//          for (int i=0;i<higher_order_mapping_support_points.getNumber();
//          i++) {
//            cout << "\thigher_order_mapping_support_points[" << i 
//                 << "] = " << higher_order_mapping_support_points[i] 
//                 << endl;
//          }
#endif
            swap(old_points,new_points);
          }
          delete old_points; old_points=0;
          delete new_points; new_points=0;
        }
#ifdef DEBUG
//      for (int i=0;i<higher_order_mapping_support_points.getNumber();
//      i++) {
//        cout << "\thigher_order_mapping_support_points[" << i 
//             << "] = " << higher_order_mapping_support_points[i] 
//             << endl;
//      }
#endif

        VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
          vgt->getWindow()->getColor("black");

        vgt->newPage();
        vgt->expose();
        for (BOOLEAN do_best=FALSE;;do_best=TRUE) {
          VolGraphTool::DrawDelimiter dd(do_best,vgt);
//        e.draw(xc,vertices,*vgt);
          vgt->setfgColor("blue");
          int ijk=0;
          for (int k=0;k<=degree;k++) {
            for (int j=0;j<=degree;j++) {
              for (int i=0;i+j<degree;i++) {
                Vector3 v0,v1;
                const Point<SPACEDIM> &homsp0=
                higher_order_mapping_support_points[renumber[ijk]];
                const Point<SPACEDIM> &homsp1=
                higher_order_mapping_support_points[renumber[ijk+1]];
#ifdef DEBUG
//              if (do_best) {
//                cout << "\ti,j,k = " << i << " " << j << " " << k
//                     << " homsp0[" << renumber[ijk] << "] = " << homsp0
//                     << " homsp1[" << renumber[ijk+1] << "] = " 
//                     << homsp1 << endl;
//              }
#endif
                for (int v=0;v<3;v++) {
                  v0[v]=homsp0[v];
                  v1[v]=homsp1[v];
                }
                vgt->drawLine(v0,v1);
                ijk++;
              }
              ijk++;
            }
          }
          ijk=0;
          for (int k=0;k<=degree;k++) {
            for (int j=0;j<degree;j++) {
              int nl=degree+1-j;
              for (int i=0;i+j<degree;i++) {
                Vector3 v0,v1;
                Point<SPACEDIM> homsp0=
                  higher_order_mapping_support_points[renumber[ijk]];
                const Point<SPACEDIM> &homsp1=
                  higher_order_mapping_support_points[
                    renumber[ijk+nl]];
#ifdef DEBUG
//              if (do_best) {
//                cout << "\ti,j,k = " << i << " " << j << " " << k
//                     << " homsp0[" << renumber[ijk] << "] = " << homsp0
//                     << " homsp1[" << renumber[ijk+degree+1-j-k] 
//                     << "] = " << homsp1 << endl;
//              }
#endif
                for (int v=0;v<3;v++) {
                  v0[v]=homsp0[v];
                  v1[v]=homsp1[v];
                }
                vgt->drawLine(v0,v1);
                if (i+j+1>=degree) {
                  homsp0=
                    higher_order_mapping_support_points[renumber[ijk+1]];
                  for (int v=0;v<3;v++) v0[v]=homsp0[v];
                  vgt->drawLine(v0,v1);
                }
                ijk++;
              }
              ijk++;
            }
            ijk++;
          }
          ijk=0;
          int nt=((degree+1)*(degree+2))/2;
          for (int k=0;k<degree;k++) {
            for (int j=0;j<degree;j++) {
              for (int i=0;i+j<degree;i++) {
                Vector3 v0,v1;
                Point<SPACEDIM> homsp0=
                  higher_order_mapping_support_points[renumber[ijk]];
                Point<SPACEDIM> homsp1=
                  higher_order_mapping_support_points[renumber[ijk+nt]];
#ifdef DEBUG
//              if (do_best) {
//                cout << "\ti,j,k = " << i << " " << j << " " << k
//                     << " homsp0[" << renumber[ijk] << "] = " << homsp0
//                     << " homsp1[" 
//              << renumber[ijk+(degree*(degree+3)+k*(k-3))/2+1 -j-k*degree]
//                     << "] = " << homsp1 << endl;
//              }
#endif
                for (int v=0;v<3;v++) {
                  v0[v]=homsp0[v];
                  v1[v]=homsp1[v];
                }
                vgt->drawLine(v0,v1);
                if (i+j+1>=degree) {
                  homsp0=
                    higher_order_mapping_support_points[renumber[ijk+1]];
                  homsp1=
                    higher_order_mapping_support_points[renumber[ijk+1+nt]];
                  for (int v=0;v<3;v++) {
                    v0[v]=homsp0[v];
                    v1[v]=homsp1[v];
                  }
                  vgt->drawLine(v0,v1);
                }
                ijk++;
              }
              ijk++;
            }
            ijk++;
          }
          if (do_best==TRUE) break;
        }
        vgt->eventLoop();
      }
    }
  }
  delete vgt;
*/
#endif

  return EXIT_SUCCESS;
}
