// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "Boundary.H"
#include "MemoryDebugger.H"
#include "Shape.H"
#include "ShapeFunction.H"
#include "Tracer.H"

#include "VolGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKWindow.H"
#endif

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif

  BOOLEAN use_lighting=TRUE;
  BOOLEAN rotate_best=FALSE;
  BOOLEAN show_bounding_box=TRUE;
  Palette pal;
  REAL winsize=0.5;
#if (SPACEDIM==3)
/*
  {
    Vector3 low((1.-sqrt(3.))*0.5,(1.-sqrt(3.))*0.5,(1.-sqrt(3.))*0.5);
    Vector3 high((1.+sqrt(3.))*0.5,(1.+sqrt(3.))*0.5,(1.+sqrt(3.))*0.5);
    VolGraphTool *vgt=OPERATOR_NEW 
      VolGraphTool("Tetrahedron::lenoirCanonicalLinePoints",low,high,pal,
        use_lighting,winsize,rotate_best,show_bounding_box);

    Point<3> center(0.5,0.5,0.5);
    BallBoundary bb(center,sqrt(center.square()));

    int nv=4;
    NumPtr<Point<3> > vertices(nv);
    vertices[0]=Tetrahedron::canonicalVertex(0);
    vertices[1]=Tetrahedron::canonicalVertex(1);
    vertices[2]=Tetrahedron::canonicalVertex(2);
    vertices[3]=Tetrahedron::canonicalVertex(3);
#ifdef DEBUG
//  for (int i=0;i<4;i++) {
//    cout << "\tvertices[" << i << "] = " << vertices[i] << endl;
//  }
#endif
    int edi=0;
    int l[6]={0,1,2,3,4,5};
    int f[4]={0,1,2,3};
    int v[4]={0,1,2,3};
    Tetrahedron e(edi,l,f,v);
    for (int degree=1;degree<=10;degree++) 
//  int degree=2;
    {
      cout << "\tdegree = " << degree << endl;
      C0LagrangeTetrahedronPolynomials ltp(degree);
      NumPtr<int> renumber;
      ltp.shapeFunctionsOrder(&e,renumber);
      NumPtr<int> renumber_inverse(renumber.getNumber());
      for (int i=0;i<renumber.getNumber();i++) {
        renumber_inverse[renumber[i]]=i;
      }
#ifdef DEBUG
//    for (int i=0;i<renumber.getNumber();i++) {
//      cout << "\trenumber[" << i << "] = " << renumber[i] << endl;
//    }
//    for (int i=0;i<renumber_inverse.getNumber();i++) {
//      cout << "\trenumber_inverse[" << i << "] = " 
//           << renumber_inverse[i] << endl;
//    }
#endif

      NumPtr<Point<SPACEDIM> > eslp;
      Tetrahedron::staticEquallySpacedLatticePoints(degree,eslp);
#ifdef DEBUG
//    for (int i=0;i<eslp.getNumber();i++) {
//      cout << "\teslp[" << i << "] = " << eslp[i] << endl;
//    }
#endif
//    for (int bl=0;bl<e.numberCodimension2Shapes();bl++) 
      int bl=5;
      {
        cout << "\tbl = " << bl << endl;
        int fv[2];
        for (int i=0;i<2;i++) {
          fv[i]=e.vertexIndex(Tetrahedron::elementVertexFromLine(bl,i));
        }
#ifdef DEBUG
//      for (int i=0;i<3;i++) {
//        cout << "\tfv[" << i << "] = " << fv[i] << endl;
//      }
#endif
        Interval line_element(0,fv);
#ifdef DEBUG
//      cout << "\tline_element:" << endl;
//      line_element.printOn(cout);
#endif

        NumPtr<Point<SPACEDIM> > 
          higher_order_mapping_support_points(eslp.getNumber());
        for (int ijk=0;ijk<eslp.getNumber();ijk++) {
          const Point<3> &eslpi=eslp[renumber_inverse[ijk]];
          higher_order_mapping_support_points[ijk]=
            vertices[1]*eslpi[0]+vertices[2]*eslpi[1]+vertices[3]*eslpi[2]
            +vertices[0]*(1.-eslpi[0]-eslpi[1]-eslpi[2]);
        }
#ifdef DEBUG
//      for (int i=0;i<higher_order_mapping_support_points.getNumber();
//      i++) {
//        cout << "\thigher_order_mapping_support_points[" << i << "] = "
//             << higher_order_mapping_support_points[i] << endl;
//      }
#endif

        NumPtr<Point<SPACEDIM> > *old_points=
          OPERATOR_NEW NumPtr<Point<SPACEDIM> >;
        old_points->
          allocate(higher_order_mapping_support_points.getNumber());
        NumPtr<Point<SPACEDIM> > *new_points=
          OPERATOR_NEW NumPtr<Point<SPACEDIM> >;
        new_points->allocate(old_points->getNumber());

        NumPtr<Point<1> > canonical_line_points;
        Tetrahedron::lenoirCanonicalLinePoints(bl,eslp,
          canonical_line_points);
#ifdef DEBUG
//      for (int i=0;i<canonical_line_points.getNumber();i++) {
//        cout << "\tcanonical_line_points[" << i << "] = "
//             << canonical_line_points[i] << endl;
//      }
#endif

        C0LagrangeIntervalPolynomials c0lip(1);
        Point<SPACEDIM> fv0=vertices[line_element.vertexIndex(0)];
        Point<SPACEDIM> fv1=vertices[line_element.vertexIndex(1)];
#ifdef DEBUG
//      cout << "\tfv0 = " << fv0 << endl;
//      cout << "\tfv1 = " << fv1 << endl;
#endif
        for (int ijk=0;ijk<old_points->getNumber();ijk++) {
          NumPtr<REAL> values;
          c0lip.computeValues(canonical_line_points[ijk],values);
          (*old_points)[ijk]=fv0*values[0]+fv1*values[1];
        }
#ifdef DEBUG
//      for (int i=0;i<old_points->getNumber();i++) {
//        cout << "\told_points[" << i << "] = "
//             << (*old_points)[i] << endl;
//      }
#endif
        for (int m=2;m<=degree;m++) {
          C0LagrangeIntervalPolynomials c0lip(m);
          NumPtr<int> frenumber;
          c0lip.shapeFunctionsOrder(&line_element,frenumber);

          NumPtr<Point<1> > leslp;
          Interval::staticEquallySpacedLatticePoints(m,leslp);
#ifdef DEBUG
//        for (int i=0;i<leslp.getNumber();i++) {
//          cout << "\tleslp[" << i << "] = " << leslp[i] << endl;
//        }
#endif
          NumPtr<Point<SPACEDIM> > lsp(leslp.getNumber());
          bb.getIntermediatePointsOn(line_element,vertices,leslp,lsp);
#ifdef DEBUG
//        for (int i=0;i<lsp.getNumber();i++) {
//          cout << "\tlsp[" << i << "] = " << lsp[i] << endl;
//        }
#endif
          for (int ijk=0;ijk<new_points->getNumber();ijk++) {
            NumPtr<REAL> values;
            c0lip.computeValues(canonical_line_points[ijk],values);
            (*new_points)[ijk]=0.;
            for (int v=0;v<values.getNumber();v++) {
              (*new_points)[ijk]+=lsp[v]*values[v];
            }
          }
#ifdef DEBUG
//        for (int i=0;i<new_points->getNumber();i++) {
//          cout << "\tnew_points[" << i << "] = "
//               << (*new_points)[i] << endl;
//        }
#endif
          Tetrahedron::lenoirInterpolationForLine(bl,m,renumber,eslp,
            *old_points,*new_points,higher_order_mapping_support_points);
#ifdef DEBUG
//        for (int i=0;i<higher_order_mapping_support_points.getNumber();
//        i++) {
//          cout << "\thigher_order_mapping_support_points[" << i << "] = "
//               << higher_order_mapping_support_points[i] << endl;
//        }
#endif
          swap(old_points,new_points);
        }

        VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
          vgt->getWindow()->getColor("black");

        vgt->newPage();
        vgt->expose();
        for (BOOLEAN do_best=FALSE;;do_best=TRUE) {
          VolGraphTool::DrawDelimiter dd(do_best,vgt);
//        e.draw(xc,vertices,*vgt);
          vgt->setfgColor("blue");

          int ijk=0;
          for (int k=0;k<degree;k++) {
            for (int j=0;j+k<degree;j++) {
              for (int i=0;i+j+k<degree;i++) {
                Vector3 v0,v1;
                const Point<SPACEDIM> &homsp0=
                 higher_order_mapping_support_points[renumber[ijk]];
                const Point<SPACEDIM> &homsp1=
                  higher_order_mapping_support_points[renumber[ijk+1]];
#ifdef DEBUG
//              if (do_best) {
//                cout << "\ti,j,k = " << i << " " << j << " " << k
//                     << " homsp0[" << renumber[ijk] << "] = " << homsp0
//                     << " homsp1[" << renumber[ijk+1] << "] = " 
//                     << homsp1 << endl;
//              }
#endif
                for (int j=0;j<3;j++) {
                  v0[j]=homsp0[j];
                  v1[j]=homsp1[j];
                }
                vgt->drawLine(v0,v1);
                ijk++;
              }
              ijk++;
            }
            ijk++;
          }
          ijk=0;
          for (int k=0;k<degree;k++) {
            for (int j=0;j+k<degree;j++) {
              for (int i=0;i+j+k<degree;i++) {
                Vector3 v0,v1;
                const Point<SPACEDIM> &homsp0=
                 higher_order_mapping_support_points[renumber[ijk]];
                const Point<SPACEDIM> &homsp1=
                  higher_order_mapping_support_points[
                    renumber[ijk+degree+1-j-k]];
#ifdef DEBUG
//              if (do_best) {
//                cout << "\ti,j,k = " << i << " " << j << " " << k
//                     << " homsp0[" << renumber[ijk] << "] = " << homsp0
//                     << " homsp1[" << renumber[ijk+degree+1-j-k] 
//                     << "] = " << homsp1 << endl;
//              }
#endif
                for (int j=0;j<3;j++) {
                  v0[j]=homsp0[j];
                  v1[j]=homsp1[j];
                }
                vgt->drawLine(v0,v1);
                ijk++;
              }
              ijk++;
            }
            ijk++;
          }
          ijk=0;
          for (int k=0;k<degree;k++) {
            for (int j=0;j+k<degree;j++) {
              for (int i=0;i+j+k<degree;i++) {
                Vector3 v0,v1;
                const Point<SPACEDIM> &homsp0=
                 higher_order_mapping_support_points[renumber[ijk]];
                const Point<SPACEDIM> &homsp1=
                  higher_order_mapping_support_points[
                    renumber[ijk+(degree*(degree+3)+k*(k-3))/2+1
                                    -j-k*degree]];
#ifdef DEBUG
//              if (do_best) {
//                cout << "\ti,j,k = " << i << " " << j << " " << k
//                     << " homsp0[" << renumber[ijk] << "] = " << homsp0
//                     << " homsp1[" 
//              << renumber[ijk+(degree*(degree+3)+k*(k-3))/2+1 -j-k*degree]
//                     << "] = " << homsp1 << endl;
//              }
#endif
                for (int j=0;j<3;j++) {
                  v0[j]=homsp0[j];
                  v1[j]=homsp1[j];
                }
                vgt->drawLine(v0,v1);
                ijk++;
              }
              ijk++;
            }
            ijk++;
          }

          if (do_best==TRUE) break;
        }
        vgt->eventLoop();
      }
    }
    delete vgt;
  }
*/
//
  {
    Vector3 low((1.-sqrt(3.))*0.5,(1.-sqrt(3.))*0.5,(1.-sqrt(3.))*0.5);
    Vector3 high((1.+sqrt(3.))*0.5,(1.+sqrt(3.))*0.5,(1.+sqrt(3.))*0.5);
    VolGraphTool *vgt=OPERATOR_NEW 
      VolGraphTool("Prism::lenoirCanonicalFacePoints",low,high,pal,
        use_lighting,winsize,rotate_best,show_bounding_box);

    Point<3> center(0.5,0.5,0.5);
    BallBoundary bb(center,sqrt(center.square()));
    Point<3> center4(0.,0.,0.5);
    BallBoundary bb4(center4,sqrt(1.5));

    int nv=6;
    NumPtr<Point<3> > vertices(nv);
    vertices[0]=Prism::canonicalVertex(0);
    vertices[1]=Prism::canonicalVertex(1);
    vertices[2]=Prism::canonicalVertex(2);
    vertices[3]=Prism::canonicalVertex(3);
    vertices[4]=Prism::canonicalVertex(4);
    vertices[5]=Prism::canonicalVertex(5);
#ifdef DEBUG
//  for (int i=0;i<6;i++) {
//    cout << "\tvertices[" << i << "] = " << vertices[i] << endl;
//  }
#endif
    int edi=0;
    int l[9]={0,1,2,3,4,5,6,7,8};
    int f[5]={0,1,2,3,4};
    int v[6]={0,1,2,3,4,5};
    Prism e(edi,l,f,v);
    for (int degree=2;degree<=10;degree++) 
//  int degree=2;
    {
      cout << "\tdegree = " << degree << endl;
      C0LagrangePrismPolynomials lwp(degree);
      NumPtr<int> renumber;
      lwp.shapeFunctionsOrder(&e,renumber);
      NumPtr<int> renumber_inverse(renumber.getNumber());
      for (int i=0;i<renumber.getNumber();i++) {
        renumber_inverse[renumber[i]]=i;
      }
#ifdef DEBUG
//    for (int i=0;i<renumber.getNumber();i++) {
//      cout << "\trenumber[" << i << "] = " << renumber[i] << endl;
//    }
//    for (int i=0;i<renumber_inverse.getNumber();i++) {
//      cout << "\trenumber_inverse[" << i << "] = "
//           << renumber_inverse[i] << endl;
//    }
#endif
      NumPtr<Point<SPACEDIM> > eslp;
      Prism::staticEquallySpacedLatticePoints(degree,eslp);
#ifdef DEBUG
//    for (int i=0;i<eslp.getNumber();i++) {
//      cout << "\teslp[" << i << "] = " << eslp[i] << endl;
//    }
#endif
//    for (int bl=0;bl<e.numberCodimension2Shapes();bl++) 
      int bl=8;
      {
        cout << "\tbl = " << bl << endl;
        int fv[2];
        for (int i=0;i<2;i++) {
          fv[i]=e.vertexIndex(Prism::staticElementVertexFromLine(bl,i));
        }
#ifdef DEBUG
//      for (int i=0;i<3;i++) {
//        cout << "\tfv[" << i << "] = " << fv[i] << endl;
//      }
#endif
        Interval line_element(0,fv);
#ifdef DEBUG
//      cout << "\tline_element:" << endl;
//      line_element.printOn(cout);
#endif
        NumPtr<Point<SPACEDIM> > 
        higher_order_mapping_support_points(eslp.getNumber());
        int tf=Prism::triangularFaceContainingLine(bl);
        if (tf>=0) {
          TRACER_CALL(t,"triangular face");

          int opposite_tf=Prism::oppositeFace(tf);
          NumPtr<int> ofeslpi;
          e.staticEquallySpacedLatticePointIndicesOfFace(opposite_tf,
            degree,ofeslpi);
          for (int i=0;i<ofeslpi.getNumber();i++) {
            int ijk=ofeslpi[i];
            int dof=renumber[ijk];
            const Point<SPACEDIM> &eslpi=eslp[ijk];
            REAL bc2=1.-eslpi[0]-eslpi[1];
            higher_order_mapping_support_points[dof]=
              (vertices[1]*eslpi[0]+vertices[2]*eslpi[1]
             +vertices[0]*bc2)*(1.-eslpi[2])
             +(vertices[4]*eslpi[0]+vertices[5]*eslpi[1]
             +vertices[3]*bc2)*eslpi[2];
          }

          NumPtr<int> feslpi;
          e.staticEquallySpacedLatticePointIndicesOfFace(tf,degree,feslpi);
          for (int i=0;i<feslpi.getNumber();i++) {
            int ijk=feslpi[i];
            int dof=renumber[ijk];
            const Point<SPACEDIM> &eslpi=eslp[ijk];
            REAL bc2=1.-eslpi[0]-eslpi[1];
            higher_order_mapping_support_points[dof]=
              (vertices[1]*eslpi[0]+vertices[2]*eslpi[1]
             +vertices[0]*bc2)*(1.-eslpi[2])
             +(vertices[4]*eslpi[0]+vertices[5]*eslpi[1]
             +vertices[3]*bc2)*eslpi[2];
          }
          NumPtr<Point<SPACEDIM> > *old_points=
            OPERATOR_NEW NumPtr<Point<SPACEDIM> >;
          old_points->allocate(feslpi.getNumber());
          NumPtr<Point<SPACEDIM> > *new_points=
            OPERATOR_NEW NumPtr<Point<SPACEDIM> >;
          new_points->allocate(feslpi.getNumber());

          NumPtr<Point<1> > canonical_line_points;
//        Prism::lenoirCanonicalLinePoints(bl,eslp,canonical_line_points);
          OBSOLETE("switch to Triangle::lenoirCanonicalLinePoints");

          C0LagrangeIntervalPolynomials c0lip(1);
          Point<SPACEDIM> fv0=vertices[line_element.vertexIndex(0)];
          Point<SPACEDIM> fv1=vertices[line_element.vertexIndex(1)];
#ifdef DEBUG
//        cout << "\tfv0 = " << fv0 << endl;
//        cout << "\tfv1 = " << fv1 << endl;
//        cout << "\tfv2 = " << fv2 << endl;
#endif
          for (int ijk=0;ijk<old_points->getNumber();ijk++) {
            NumPtr<REAL> values;
            c0lip.computeValues(canonical_line_points[ijk],values);
            (*old_points)[ijk]=fv0*values[0]+fv1*values[1];
          }
#ifdef DEBUG
//        for (int i=0;i<old_points->getNumber();i++) {
//          cout << "\told_points[" << i << "] = "
//               << (*old_points)[i] << endl;
//        }
#endif
          for (int m=2;m<=degree;m++) {
            C0LagrangeIntervalPolynomials c0lip(m);
            NumPtr<int> frenumber;
            c0lip.shapeFunctionsOrder(&line_element,frenumber);

            NumPtr<Point<1> > leslp;
            Interval::staticEquallySpacedLatticePoints(m,leslp);
#ifdef DEBUG 
//          for (int i=0;i<leslp.getNumber();i++) {
//            cout << "\tleslp[" << i << "] = " << leslp[i] << endl;
//          }
#endif
            NumPtr<Point<SPACEDIM> > lsp;
//          bb.getMappedPointsOn(line_element,vertices,leslp,lsp);
            OBSOLETE("getMappedPointsOn is obsolete");
#ifdef DEBUG
//          for (int i=0;i<lsp.getNumber();i++) {
//            cout << "\tlsp[" << i << "] = " << lsp[i] << endl;
//          }
#endif
            for (int ijk=0;ijk<new_points->getNumber();ijk++) {
              NumPtr<REAL> values;
              c0lip.computeValues(canonical_line_points[ijk],values);
              (*new_points)[ijk]=0.;
              for (int v=0;v<values.getNumber();v++) {
                (*new_points)[ijk]+=lsp[v]*values[v];
              }
            }
#ifdef DEBUG
//          for (int i=0;i<new_points->getNumber();i++) {
//            cout << "\tnew_points[" << i << "] = "
//                 << (*new_points)[i] << endl;
//          }
#endif
//          Prism::lenoirInterpolationForLine(bl,m,feslpi,eslp,
//            *old_points,*new_points,higher_order_mapping_support_points);
            OBSOLETE("switch to Triangle::lenoirInterpolationForLine");
#ifdef DEBUG
//          for (int i=0;i<higher_order_mapping_support_points.getNumber();
//          i++) {
//            cout << "\thigher_order_mapping_support_points[" << i 
//                 << "] = " << higher_order_mapping_support_points[i] 
//                 << endl;
//          }
#endif
            swap(old_points,new_points);
          }
//        Prism::linearInterpolationFromTriangularFacesForInterior(degree,
//          eslp,higher_order_mapping_support_points);
          OBSOLETE("Prism::linearInterpolationFromTriangularFacesForInterior is obsolete");
        } else {
          NumPtr<Point<1> > leslp;
          Interval::staticEquallySpacedLatticePoints(degree,leslp);
#ifdef DEBUG 
//        for (int i=0;i<leslp.getNumber();i++) {
//          cout << "\tleslp[" << i << "] = " << leslp[i] << endl;
//        }
#endif
          NumPtr<Point<SPACEDIM> > lsp;
//        bb.getMappedPointsOn(line_element,vertices,leslp,lsp);
          OBSOLETE("getMappedPointsOn is obsolete");
#ifdef DEBUG
//        for (int i=0;i<lsp.getNumber();i++) {
//          cout << "\tlsp[" << i << "] = " << lsp[i] << endl;
//        }
#endif
//        e.linearInterpolationFromLineForInterior(degree,bl,
//          line_element,lsp,vertices,higher_order_mapping_support_points);
          OBSOLETE("linearInterpolationFromLineForInterior is obsolete");
        }
#ifdef DEBUG
//      for (int i=0;i<higher_order_mapping_support_points.getNumber();
//      i++) {
//        cout << "\thigher_order_mapping_support_points[" << i 
//             << "] = " << higher_order_mapping_support_points[i] 
//             << endl;
//      }
#endif

        VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
          vgt->getWindow()->getColor("black");

        vgt->newPage();
        vgt->expose();
        for (BOOLEAN do_best=FALSE;;do_best=TRUE) {
          VolGraphTool::DrawDelimiter dd(do_best,vgt);
//        e.draw(xc,vertices,*vgt);
          vgt->setfgColor("blue");
          int ijk=0;
          for (int k=0;k<=degree;k++) {
            for (int j=0;j<=degree;j++) {
              for (int i=0;i+j<degree;i++) {
                Vector3 v0,v1;
                const Point<SPACEDIM> &homsp0=
                higher_order_mapping_support_points[renumber[ijk]];
                const Point<SPACEDIM> &homsp1=
                higher_order_mapping_support_points[renumber[ijk+1]];
#ifdef DEBUG
//              if (do_best) {
//                cout << "\ti,j,k = " << i << " " << j << " " << k
//                     << " homsp0[" << renumber[ijk] << "] = " << homsp0
//                     << " homsp1[" << renumber[ijk+1] << "] = " 
//                     << homsp1 << endl;
//              }
#endif
                for (int v=0;v<3;v++) {
                  v0[v]=homsp0[v];
                  v1[v]=homsp1[v];
                }
                vgt->drawLine(v0,v1);
                ijk++;
              }
              ijk++;
            }
          }
          ijk=0;
          for (int k=0;k<=degree;k++) {
            for (int j=0;j<degree;j++) {
              int nl=degree+1-j;
              for (int i=0;i+j<degree;i++) {
                Vector3 v0,v1;
                Point<SPACEDIM> homsp0=
                  higher_order_mapping_support_points[renumber[ijk]];
                const Point<SPACEDIM> &homsp1=
                  higher_order_mapping_support_points[
                    renumber[ijk+nl]];
#ifdef DEBUG
//              if (do_best) {
//                cout << "\ti,j,k = " << i << " " << j << " " << k
//                     << " homsp0[" << renumber[ijk] << "] = " << homsp0
//                     << " homsp1[" << renumber[ijk+degree+1-j-k] 
//                     << "] = " << homsp1 << endl;
//              }
#endif
                for (int v=0;v<3;v++) {
                  v0[v]=homsp0[v];
                  v1[v]=homsp1[v];
                }
                vgt->drawLine(v0,v1);
                if (i+j+1>=degree) {
                  homsp0=
                    higher_order_mapping_support_points[renumber[ijk+1]];
                  for (int v=0;v<3;v++) v0[v]=homsp0[v];
                  vgt->drawLine(v0,v1);
                }
                ijk++;
              }
              ijk++;
            }
            ijk++;
          }
          ijk=0;
          int nt=((degree+1)*(degree+2))/2;
          for (int k=0;k<degree;k++) {
            for (int j=0;j<degree;j++) {
              for (int i=0;i+j<degree;i++) {
                Vector3 v0,v1;
                Point<SPACEDIM> homsp0=
                  higher_order_mapping_support_points[renumber[ijk]];
                Point<SPACEDIM> homsp1=
                  higher_order_mapping_support_points[renumber[ijk+nt]];
#ifdef DEBUG
//              if (do_best) {
//                cout << "\ti,j,k = " << i << " " << j << " " << k
//                     << " homsp0[" << renumber[ijk] << "] = " << homsp0
//                     << " homsp1[" 
//              << renumber[ijk+(degree*(degree+3)+k*(k-3))/2+1 -j-k*degree]
//                     << "] = " << homsp1 << endl;
//              }
#endif
                for (int v=0;v<3;v++) {
                  v0[v]=homsp0[v];
                  v1[v]=homsp1[v];
                }
                vgt->drawLine(v0,v1);
                if (i+j+1>=degree) {
                  homsp0=
                    higher_order_mapping_support_points[renumber[ijk+1]];
                  homsp1=
                    higher_order_mapping_support_points[renumber[ijk+1+nt]];
                  for (int v=0;v<3;v++) {
                    v0[v]=homsp0[v];
                    v1[v]=homsp1[v];
                  }
                  vgt->drawLine(v0,v1);
                }
                ijk++;
              }
              ijk++;
            }
            ijk++;
          }
          if (do_best==TRUE) break;
        }
        vgt->eventLoop();
      }
    }
    delete vgt;
  }
//
#endif

  return EXIT_SUCCESS;
}
