// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "MemoryDebugger.H"
#include "Shape.H"
#include "Tracer.H"

int main(int argc,char* argv[]) {
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif

#if (SPACEDIM==1)
    cout << "Interval:" << endl;
    CHECK_SAME(Interval::verticesPerShape(),2);
    CHECK_SAME(Interval::facesPerShape(),2);
    CHECK_SAME(Interval::linesPerShape(),1);
    CHECK_SAME(Interval::staticNumberInteriorChildVertices(),1);
    Point<1> p=Interval::staticCenter();
    CHECK_SAME(p[0],0.5);
//  CHECK_TEST(Interval::containsCanonical(p));
    p[0]=drand48();
//  for (int c=0;c<2;c++) {
//    Point<1> cp=Interval::childCoordinateOfCanonical(p,c);
//    CHECK_SAME(Interval::parentCoordinateOfCanonical(cp,c),p);
//  }
    for (int f=0;f<2;f++) {
//    CHECK_SAME(f+Interval::oppositeFace(f),1);
      CHECK_SAME(Interval::staticElementLineFromFace(f,0),0);
      CHECK_SAME(Interval::staticElementVertexFromFace(f,0),f);
      CHECK_SAME(Interval::staticChildOnFace(f,0),f);
    }
    for (int v=0;v<2;v++) {
      CHECK_SAME(Interval::staticElementVertexFromLine(0,v),v);
      Point<1> p=Interval::canonicalVertex(v);
//    CHECK_TEST(Interval::containsCanonical(p));
//    CHECK_SAME(Interval::childShapeOfCanonical(p),v);
    }
    for (int n=1;n<=6;n++) {
      CHECK_SAME(Interval::staticNumberEquallySpacedLatticePoints(n),n+1);
      CHECK_SAME(
        Interval::staticEquallySpacedLatticePointIndexOfVertex(0,n),0);
      CHECK_SAME(
        Interval::staticEquallySpacedLatticePointIndexOfVertex(1,n),n);
    }

    int vertices[2]={0,1};
    Interval e(0,vertices);
    CHECK_SAME(e.numberInteriorChildVertices(),1);
    int c=INT_MAX;
    int v=INT_MAX;
    e.interiorChildVertex(c,v);
    CHECK_SAME(c,0);
    CHECK_SAME(v,1);

    NumPtr<Point<1> > canonical_vertices(2);
    for (int i=0;i<2;i++) {
      canonical_vertices[i]=Interval::canonicalVertex(i);
    }
    for (int i=0;i<2;i++) {
      CHECK_TEST(e.contains(canonical_vertices[i],canonical_vertices));
    }
    CHECK_TEST(e.contains(e.center(canonical_vertices),canonical_vertices));
#endif

#if (SPACEDIM==2)
    {
      cout << "Triangle:" << endl;
      CHECK_SAME(Triangle::verticesPerShape(),3);
      CHECK_SAME(Triangle::facesPerShape(),3);
      CHECK_SAME(Triangle::linesPerShape(),3);
      CHECK_SAME(Triangle::staticNumberInteriorChildCodimension1Shapes(),3);
      Point<2> p=Triangle::staticCenter();
      CHECK_SAME(p[0],1./3.);
      CHECK_SAME(p[1],1./3.);
      p[0]=drand48();
      p[1]=drand48();
//    for (int c=0;c<4;c++) {
//      Point<2> cp=Triangle::childCoordinateOfCanonical(p,c);
//      CHECK_SAME(Triangle::parentCoordinateOfCanonical(cp,c),p);
//    }
      for (int f=0;f<3;f++) {
        CHECK_SAME(Triangle::linesPerFace(f),1);
        CHECK_SAME(Triangle::staticNumberVerticesOnFace(f),2);
//      CHECK_SAME(Triangle::oppositeFace(f),(f+1)%3);
        CHECK_SAME(Triangle::staticElementLineFromFace(f,0),f);
        CHECK_SAME(Triangle::staticNumberInteriorChildVerticesOnFace(f),1);
//      for (int v=0;v<2;v++) {
//        CHECK_SAME(Triangle::staticElementVertexFromFace(f,v),(f+v+2)%3);
//      }
        CHECK_SAME(Triangle::staticElementVertexFromFace(f,0),
          (f==2 ? 1 : 0));
        CHECK_SAME(Triangle::staticElementVertexFromFace(f,1),
          (f==1 ? 1 : 2));
        for (int s=0;s<2;s++) {
          CHECK_SAME(Triangle::staticChildOnFace(f,s),(f+s+2)%3);
        }
//      Tensor<1,2> cp;
//      crossProduct(cp,Triangle::faceTangent(f,0));
//      CHECK_SAME(cp,Triangle::faceNormal(f));
        int c=INT_MAX,v=INT_MAX;
        Triangle::staticInteriorChildVertexOnFace(f,c,v);
        CHECK_SAME(c,3);
        CHECK_SAME(v,f);
      }
      for (int cf=0;cf<3;cf++) {
        int c0=INT_MAX;
        int c1=INT_MAX;
        Triangle::staticInteriorChildCodimension1ShapeNeighbors(cf,c0,c1);
        CHECK_SAME(c0,3);
        CHECK_SAME(c1,(cf+2)%3);
      }
      for (int v=0;v<3;v++) {
        Point<2> p=Triangle::canonicalVertex(v);
//      CHECK_TEST(Triangle::containsCanonical(p));
//      CHECK_SAME(Triangle::childShapeOfCanonical(p),v);
      }
      for (int l=0;l<3;l++) {
//      for (int v=0;v<2;v++) {
//        CHECK_SAME(Triangle::staticElementVertexFromLine(l,v),(l+v+2)%3);
//      }
        CHECK_SAME(Triangle::staticElementVertexFromFace(l,0),
          (l==2 ? 1 : 0));
        CHECK_SAME(Triangle::staticElementVertexFromFace(l,1),
          (l==1 ? 1 : 2));
      }
      for (int n=1;n<=6;n++) {
        CHECK_SAME(Triangle::staticNumberEquallySpacedLatticePoints(n),
          ((n+1)*(n+2))/2);
        CHECK_SAME(
          Triangle::staticEquallySpacedLatticePointIndexOfVertex(0,n),0);
        CHECK_SAME(
          Triangle::staticEquallySpacedLatticePointIndexOfVertex(1,n),n);
        CHECK_SAME(Triangle::staticEquallySpacedLatticePointIndexOfVertex(
          2,n),(n*(n+3))/2);
      }

      int vertices[3]={0,1,2};
      Triangle e(0,vertices,vertices);
      CHECK_SAME(e.numberInteriorChildVertices(),0);

      NumPtr<Point<2> > canonical_vertices(3);
      for (int i=0;i<3;i++) {
        canonical_vertices[i]=Triangle::canonicalVertex(i);
      }
      for (int i=0;i<3;i++) {
        CHECK_TEST(e.contains(canonical_vertices[i],canonical_vertices));
      }
      CHECK_TEST(e.contains(e.center(canonical_vertices),
                            canonical_vertices));
    }
    {
      cout << "Quadrilateral:" << endl;
      CHECK_SAME(Quadrilateral::verticesPerShape(),4);
      CHECK_SAME(Quadrilateral::facesPerShape(),4);
      CHECK_SAME(Quadrilateral::linesPerShape(),4);
      CHECK_SAME(Quadrilateral::staticNumberInteriorChildVertices(),1);
      CHECK_SAME(
        Quadrilateral::staticNumberInteriorChildCodimension1Shapes(),4);
      int c=INT_MAX,v=INT_MAX;
      Quadrilateral::staticInteriorChildVertex(0,c,v);
      CHECK_SAME(c,0);
      CHECK_SAME(v,3);
      Point<2> p=Quadrilateral::staticCenter();
      CHECK_SAME(p[0],0.5);
      CHECK_SAME(p[1],0.5);
      p[0]=drand48();
      p[1]=drand48();
//    for (int c=0;c<4;c++) {
//      Point<2> cp=Quadrilateral::childCoordinateOfCanonical(p,c);
//      CHECK_SAME(Quadrilateral::parentCoordinateOfCanonical(cp,c),p);
//    }
      for (int f=0;f<3;f++) {
        CHECK_SAME(Quadrilateral::linesPerFace(f),1);
        CHECK_SAME(Quadrilateral::staticNumberVerticesOnFace(f),2);
//      CHECK_SAME(Quadrilateral::oppositeFace(f),(f+1)%2+2*(f/2));
        CHECK_SAME(Quadrilateral::staticElementLineFromFace(f,0),f);
        CHECK_SAME(
          Quadrilateral::staticNumberInteriorChildVerticesOnFace(f),1);
        for (int v=0;v<2;v++) {
          CHECK_SAME(Quadrilateral::staticElementVertexFromFace(f,v),
            (f<2 ? f+2*v : 2*(f-2)+v));
        }
        for (int s=0;s<2;s++) {
          CHECK_SAME(Quadrilateral::staticChildOnFace(f,s),
            (f<2 ? f+2*s : 2*(f-2)+s));
        }
//      Tensor<1,2> cp;
//      crossProduct(cp,Quadrilateral::faceTangent(f,0));
//      CHECK_SAME(cp,Quadrilateral::faceNormal(f));
        int c=INT_MAX,v=INT_MAX;
        Quadrilateral::staticInteriorChildVertexOnFace(f,c,v);
        CHECK_SAME(c,(f%2==0 ? 0 : 3));
        CHECK_SAME(v,(f/2==0 ? 2 : 1));
      }
      for (int cf=0;cf<4;cf++) {
        int c0=INT_MAX;
        int c1=INT_MAX;
        Quadrilateral::staticInteriorChildCodimension1ShapeNeighbors(cf,c0,c1);
        CHECK_SAME(c0,(cf<2 ? cf : 2*(cf%2)));
        CHECK_SAME(c1,(cf<2 ? cf+2 : 1+2*(cf%2)));
      }
      for (int v=0;v<4;v++) {
        Point<2> p=Quadrilateral::canonicalVertex(v);
//      CHECK_TEST(Quadrilateral::containsCanonical(p));
//      CHECK_SAME(Quadrilateral::childShapeOfCanonical(p),v);
        CHECK_SAME(Quadrilateral::oppositeVertex(v),3-v);
      }
      for (int l=0;l<4;l++) {
        for (int v=0;v<2;v++) {
          CHECK_SAME(Quadrilateral::staticElementVertexFromLine(l,v),
            (l<2 ? l+2*v : 2*(l-2)+v));
        }
      }
      for (int n=1;n<=6;n++) {
        CHECK_SAME(Quadrilateral::staticNumberEquallySpacedLatticePoints(n),
          (n+1)*(n+1));
        CHECK_SAME(
          Quadrilateral::staticEquallySpacedLatticePointIndexOfVertex(0,n),0);
        CHECK_SAME(
          Quadrilateral::staticEquallySpacedLatticePointIndexOfVertex(1,n),n);
        CHECK_SAME(Quadrilateral::staticEquallySpacedLatticePointIndexOfVertex(
          2,n),n*(n+1));
        CHECK_SAME(Quadrilateral::staticEquallySpacedLatticePointIndexOfVertex(
          3,n),n*(n+2));
      }

      int vertices[4]={0,1,2,3};
      Quadrilateral e(0,vertices,vertices);
      CHECK_SAME(e.numberInteriorChildVertices(),1);
      c=INT_MAX;
      v=INT_MAX;
      e.interiorChildVertex(c,v);
      CHECK_SAME(c,0);
      CHECK_SAME(v,3);

      NumPtr<Point<2> > canonical_vertices(4);
      for (int i=0;i<4;i++) {
        canonical_vertices[i]=Quadrilateral::canonicalVertex(i);
      }
      for (int i=0;i<4;i++) {
        CHECK_TEST(e.contains(canonical_vertices[i],canonical_vertices));
      }
      CHECK_TEST(e.contains(e.center(canonical_vertices),
                            canonical_vertices));
    }
#endif

#if (SPACEDIM==3)
//
    {
      cout << "Tetrahedron:" << endl;
      CHECK_SAME(Tetrahedron::verticesPerShape(),4);
      CHECK_SAME(Tetrahedron::facesPerShape(),4);
      CHECK_SAME(Tetrahedron::linesPerShape(),6);
      CHECK_SAME(
        Tetrahedron::staticNumberInteriorChildCodimension1Shapes(),8);
      CHECK_SAME(Tetrahedron::staticNumberCodimension2Shapes(),6);
      CHECK_SAME(Tetrahedron::staticNumberInteriorChildCodimension2Shapes(),1);
      Point<3> p=Tetrahedron::staticCenter();
      CHECK_SAME(p[0],0.25);
      CHECK_SAME(p[1],0.25);
      CHECK_SAME(p[2],0.25);
      p[0]=drand48();
      p[1]=drand48();
//    for (int c=0;c<8;c++) {
//      Point<3> cp=Tetrahedron::childCoordinateOfCanonical(p,c);
//      CHECK_SAME(Tetrahedron::parentCoordinateOfCanonical(cp,c),p);
//    }
      for (int f=0;f<4;f++) {
        CHECK_SAME(Tetrahedron::linesPerFace(f),3);
        CHECK_SAME(Tetrahedron::staticNumberVerticesOnFace(f),3);
//      CHECK_SAME(Tetrahedron::oppositeFace(f),(f+1)%4);
        CHECK_SAME(Tetrahedron::staticElementLineFromFace(f,0),
//        (f<3 ? (f+2)%4 : 4)
          (f<2 ? 3 : 2*(f-1))
          );
        CHECK_SAME(Tetrahedron::staticElementLineFromFace(f,1),
//        (f<2 ? 5-f : 1+4*(f-2))
          (f<2 ? (f+2)%3 : f-2)
          );
        CHECK_SAME(Tetrahedron::staticElementLineFromFace(f,2),
//        (f<2 ? 3-3*f : 4-f)
          (f<2 ? 5-f : 4*f-7)
          );
        CHECK_SAME(
          Tetrahedron::staticNumberInteriorChildVerticesOnFace(f),0);
        CHECK_SAME(Tetrahedron::staticElementVertexFromFace(f,0),
          (f<3 ? 0 : 1));
        CHECK_SAME(Tetrahedron::staticElementVertexFromFace(f,1),
//        (f%2==0 ? 2-f/2 : 3)
          (f<2 ? 2-f : f-1)
          );
        CHECK_SAME(Tetrahedron::staticElementVertexFromFace(f,2),
//        (f<2 ? 3-2*f : 2)
          (f==2 ? 2 : 3)
          );
        CHECK_SAME(Tetrahedron::staticChildOnFace(f,0),(f<3 ? 0 : 1));
        CHECK_SAME(Tetrahedron::staticChildOnFace(f,1),
          (f<2 ? f+2 : 2*f-3));
        CHECK_SAME(Tetrahedron::staticChildOnFace(f,2),
          (f<2 ? 3-2*f : 2));
        CHECK_SAME(Tetrahedron::staticChildOnFace(f,3),
          (f<2 ? 7-3*f : 8-f));
//      Tensor<1,3> ft0=Tetrahedron::faceTangent(f,0);
//      REAL dot=(ft0[0]*ft0[0]+ft0[1]*ft0[1]+ft0[2]*ft0[2]-1.)/DBL_EPSILON;
//      CHECK_TEST(dot<16.);
//      Tensor<1,3> ft1=Tetrahedron::faceTangent(f,1);
//      dot=(ft1[0]*ft1[0]+ft1[1]*ft1[1]+ft1[2]*ft1[2]-1.)/DBL_EPSILON;
//      CHECK_TEST(dot<16.);
//      dot=(ft0[0]*ft1[0]+ft0[1]*ft1[1]+ft0[2]*ft1[2])/DBL_EPSILON;
//      CHECK_TEST(dot<16.);
//      Tensor<1,3> cp;
//      crossProduct(cp,ft0,ft1);
        Tensor<1,3> n=Tetrahedron::faceNormal(f);
//      REAL dif=
//        sqrt(pow(n[0]-cp[0],2)+pow(n[1]-cp[1],2)+pow(n[2]-cp[2],2))
//        /DBL_EPSILON;
//      CHECK_TEST(dif<16.);
      }
      for (int cf=0;cf<8;cf++) {
        int c0=INT_MAX;
        int c1=INT_MAX;
        Tetrahedron::staticInteriorChildCodimension1ShapeNeighbors(cf,c0,c1);
        CHECK_SAME(c0,(cf<3 ? 4 : (cf<5 ? 5 : (cf<7 ? 6 : 7 ))));
        CHECK_SAME(c1,(cf<2 ? 5-2*cf : (cf<4 ? 19-6*cf : 
          (cf<6 ? 22-4*cf : 49-7*cf ))));
        CHECK_SAME(c1,(cf<2 ? 5-2*cf : (cf<4 ? 19-6*cf : 
          (cf<6 ? 22-4*cf : 49-7*cf))));
      }
      for (int v=0;v<4;v++) {
        Point<3> p=Tetrahedron::canonicalVertex(v);
//      CHECK_TEST(Tetrahedron::containsCanonical(p));
//      CHECK_SAME(Tetrahedron::childShapeOfCanonical(p),v);
      }
      for (int l=0;l<6;l++) {
        CHECK_SAME(Tetrahedron::staticElementVertexFromLine(l,0),
//        (l<3 ? l : (l==3 ? 0 : 3))
          (l==5 ? 2 : ((l==1 || l==4) ? 1 : 0 ))
          );
        CHECK_SAME(Tetrahedron::staticElementVertexFromLine(l,1),
//        (l<3 ? (l+1)%3 : (l==3 ? 3 : l-3))
          (l==0 ? 1 : (l<3 ? 2 : 3))
          );
      }
      for (int n=1;n<=6;n++) {
        CHECK_SAME(Tetrahedron::staticNumberEquallySpacedLatticePoints(n),
          ((n+1)*(n+2)*(n+3))/6);
        CHECK_SAME(
          Tetrahedron::staticEquallySpacedLatticePointIndexOfVertex(0,n),0);
        CHECK_SAME(
          Tetrahedron::staticEquallySpacedLatticePointIndexOfVertex(1,n),n);
        CHECK_SAME(Tetrahedron::staticEquallySpacedLatticePointIndexOfVertex(
          2,n),(n*(n+3))/2);
        CHECK_SAME(Tetrahedron::staticEquallySpacedLatticePointIndexOfVertex(
          3,n),(n*(11+n*(6+n)))/6);
      }

      int vertices[4]={0,1,2,3};
      int lines[6]={0,1,2,3,4,5};
      Tetrahedron e(0,lines,vertices,vertices);
      CHECK_SAME(e.numberInteriorChildVertices(),0);

      NumPtr<Point<3> > canonical_vertices(4);
      for (int i=0;i<4;i++) {
        canonical_vertices[i]=Tetrahedron::canonicalVertex(i);
      }
      for (int i=0;i<4;i++) {
        CHECK_TEST(e.contains(canonical_vertices[i],canonical_vertices));
      }
      CHECK_TEST(e.contains(e.center(canonical_vertices),
                            canonical_vertices));
    }
//
//
    {
      cout << "Prism:" << endl;
      CHECK_SAME(Prism::verticesPerShape(),6);
      CHECK_SAME(Prism::facesPerShape(),5);
      CHECK_SAME(Prism::linesPerShape(),9);
      CHECK_SAME(
        Prism::staticNumberInteriorChildCodimension1Shapes(),10);
      CHECK_SAME(Prism::staticNumberCodimension2Shapes(),9);
      CHECK_SAME(Prism::staticNumberInteriorChildCodimension2Shapes(),3);
      Point<3> p=Prism::staticCenter();
      CHECK_SAME(p[0],1./3.);
      CHECK_SAME(p[1],1./3.);
      CHECK_SAME(p[2],0.5);
      p[0]=drand48();
      p[1]=drand48();
//    for (int c=0;c<8;c++) {
//      Point<3> cp=Prism::childCoordinateOfCanonical(p,c);
//      CHECK_SAME(Prism::parentCoordinateOfCanonical(cp,c),p);
//    }
      for (int f=0;f<5;f++) {
        CHECK_SAME(Prism::linesPerFace(f),(f<2 ? 3 : 4));
        CHECK_SAME(Prism::staticNumberVerticesOnFace(f),(f<2 ? 3 : 4));
//      CHECK_SAME(Prism::oppositeFace(f),(f<2 ? 1-f : 2+(f-1)%3));
        CHECK_SAME(Prism::staticElementLineFromFace(f,0),
//        (f<2 ? 2+f : (f==2 ? 2 : f+3))
          (f<2 ? 3*f+2 : (f==4 ? 7 : 6))
          );
        CHECK_SAME(Prism::staticElementLineFromFace(f,1),
//        (f<2 ? 1+3*f : (f==2 ? 5 : f+4))
          (f<2 ? 3*f : (f==3 ? 7 : 8))
          );
        CHECK_SAME(Prism::staticElementLineFromFace(f,2),
//        (f<2 ? 5*f : (f==2 ? 6 : f-3))
          (f<2 ? 3*f+1 : f%3)
          );
        if (f>=2) {
          CHECK_SAME(Prism::staticElementLineFromFace(f,3),
//          (f<2 ? 9 : (f==2 ? 8 : f))
            (f<2 ? 9 : 3+f%3)
            );
        }
        CHECK_SAME(
          Prism::staticNumberInteriorChildVerticesOnFace(f),(f<2 ? 0 : 1));
        CHECK_SAME(Prism::staticElementVertexFromFace(f,0),
//        (f<2 ? 3*f : (f==2 ? 0 : f-3)));
          (f<2 ? 3*f : (f==4 ? 1 : 0)));
        CHECK_SAME(Prism::staticElementVertexFromFace(f,1),
//        (f<2 ? 2+2*f : (f==2 ? 3 : f-2)));
          (f<2 ? 3*f+1 : (f==3 ? 1 : 2)));
        CHECK_SAME(Prism::staticElementVertexFromFace(f,2),
//        (f<2 ? 1+4*f : (f==2 ? 2 : f)));
          (f<2 ? 3*f+2 : (f==4 ? 4 : 3)));
        if (f>=2) {
          CHECK_SAME(Prism::staticElementVertexFromFace(f,3),
//          (f<2 ? 6 : (f==2 ? 5 : f+1))
            (f<2 ? 6 : (f==3 ? 4 : 5))
            );
        }
        CHECK_SAME(Prism::staticChildOnFace(f,0),
          (f<2 ? 4*f : (f==2 ? 0 : f-3)));
        CHECK_SAME(Prism::staticChildOnFace(f,1),
          (f<2 ? 1+4*f : (f==2 ? 4 : f-2)));
        CHECK_SAME(Prism::staticChildOnFace(f,2),
          (f<2 ? 2+4*f : (f==2 ? 2 : f+1)));
        CHECK_SAME(Prism::staticChildOnFace(f,3),
          (f<2 ? 3+4*f : (f==2 ? 6 : f+2)));
//      Tensor<1,3> ft0=Prism::faceTangent(f,0);
//      REAL dot=(ft0[0]*ft0[0]+ft0[1]*ft0[1]+ft0[2]*ft0[2]-1.)/DBL_EPSILON;
//      CHECK_TEST(dot<16.);
//      Tensor<1,3> ft1=Prism::faceTangent(f,1);
//      dot=(ft1[0]*ft1[0]+ft1[1]*ft1[1]+ft1[2]*ft1[2]-1.)/DBL_EPSILON;
//      CHECK_TEST(dot<16.);
//      dot=(ft0[0]*ft1[0]+ft0[1]*ft1[1]+ft0[2]*ft1[2])/DBL_EPSILON;
//      CHECK_TEST(dot<16.);
//      Tensor<1,3> cp;
//      crossProduct(cp,ft0,ft1);
        Tensor<1,3> n=Prism::faceNormal(f);
//      REAL dif=
//        sqrt(pow(n[0]-cp[0],2)+pow(n[1]-cp[1],2)+pow(n[2]-cp[2],2))
//        /DBL_EPSILON;
//      CHECK_TEST(dif<16.);
        if (f>=2) {
          int c=INT_MAX,v=INT_MAX;
          Prism::staticInteriorChildVertexOnFace(f,c,v);
          CHECK_SAME(c,(f==4 ? 1 : 0));
          CHECK_SAME(v,(f==3 ? 4 : 5));
        }
        CHECK_SAME(Prism::numberInteriorChildLinesOnFace(f),(f<2 ? 3 : 4));
      }
      for (int cf=0;cf<10;cf++) {
        int c0=INT_MAX;
        int c1=INT_MAX;
        Prism::staticInteriorChildCodimension1ShapeNeighbors(cf,c0,c1);
        CHECK_SAME(c0,(cf<4 ? cf : (cf<7 ? cf%3 : 4+cf%3)));
        CHECK_SAME(c1,(cf<4 ? cf+4 : (cf<7 ? 3 : 7)));
      }
//    for (int v=0;v<6;v++) {
//      Point<3> p=Prism::canonicalVertex(v);
//      CHECK_TEST(Prism::containsCanonical(p));
//      CHECK_SAME(Prism::childShapeOfCanonical(p),(v<3 ? v : v+1));
//    }
      for (int l=0;l<9;l++) {
//      for (int v=0;v<2;v++) {
//        CHECK_SAME(Prism::staticElementVertexFromLine(l,v),
//          (l<3 ? (l+v)%3 : (l<6 ? 3+(l+v)%3 : l-6+3*v))
//          );
//      }
        CHECK_SAME(Prism::staticElementVertexFromLine(l,0),
            (l<3 ? l%2 : (l<6 ? (l+1)%2+3 : l-6)));
        CHECK_SAME(Prism::staticElementVertexFromLine(l,1),
            (l<3 ? (l==0 ? 1 : 2) : (l<6 ? (l==3 ? 4 : 5) : l-3)));
      }
      for (int n=1;n<=6;n++) {
        CHECK_SAME(Prism::staticNumberEquallySpacedLatticePoints(n),
          (n+1)*((n+1)*(n+2))/2);
        CHECK_SAME(
          Prism::staticEquallySpacedLatticePointIndexOfVertex(0,n),0);
        CHECK_SAME(
          Prism::staticEquallySpacedLatticePointIndexOfVertex(1,n),n);
        CHECK_SAME(Prism::staticEquallySpacedLatticePointIndexOfVertex(
          2,n),(n*(n+3))/2);
        CHECK_SAME(Prism::staticEquallySpacedLatticePointIndexOfVertex(
          3,n),n*((n+1)*(n+2))/2);
        CHECK_SAME(Prism::staticEquallySpacedLatticePointIndexOfVertex(
          4,n),n+n*((n+1)*(n+2))/2);
        CHECK_SAME(Prism::staticEquallySpacedLatticePointIndexOfVertex(
          5,n),(n+1)*((n+1)*(n+2))/2-1);
      }

      int vertices[6]={0,1,2,3,4,5};
      int faces[5]={0,1,2,3,4};
      int lines[9]={0,1,2,3,4,5,6,7,8};
      Prism e(0,lines,vertices,vertices);
      CHECK_SAME(e.numberInteriorChildVertices(),0);

      NumPtr<Point<3> > canonical_vertices(6);
      for (int i=0;i<6;i++) {
        canonical_vertices[i]=Prism::canonicalVertex(i);
      }
      for (int i=0;i<6;i++) {
        CHECK_TEST(e.contains(canonical_vertices[i],canonical_vertices));
      }
      CHECK_TEST(e.contains(e.center(canonical_vertices),
                            canonical_vertices));
    }
//
//
    {
      cout << "Hexahedron:" << endl;
      CHECK_SAME(Hexahedron::verticesPerShape(),8);
      CHECK_SAME(Hexahedron::facesPerShape(),6);
      CHECK_SAME(Hexahedron::linesPerShape(),12);
      CHECK_SAME(Hexahedron::staticNumberInteriorChildVertices(),1);
      CHECK_SAME(
        Hexahedron::staticNumberInteriorChildCodimension1Shapes(),12);
      CHECK_SAME(Hexahedron::staticNumberCodimension2Shapes(),12);
      CHECK_SAME(Hexahedron::staticNumberInteriorChildCodimension2Shapes(),6);
      int c=INT_MAX,v=INT_MAX;
      Hexahedron::staticInteriorChildVertex(0,c,v);
      CHECK_SAME(c,0);
      CHECK_SAME(v,7);
      Point<3> p=Hexahedron::staticCenter();
      CHECK_SAME(p[0],0.5);
      CHECK_SAME(p[1],0.5);
      CHECK_SAME(p[2],0.5);
      p[0]=drand48();
      p[1]=drand48();
//    for (int c=0;c<8;c++) {
//      Point<3> cp=Hexahedron::childCoordinateOfCanonical(p,c);
//      CHECK_SAME(Hexahedron::parentCoordinateOfCanonical(cp,c),p);
//    }
      for (int f=0;f<6;f++) {
        CHECK_SAME(Hexahedron::linesPerFace(f),4);
        CHECK_SAME(Hexahedron::staticNumberVerticesOnFace(f),4);
//      CHECK_SAME(Hexahedron::oppositeFace(f),
//        (f<2 ? 1-f : (f<4 ? 5-f : 9-f)));
        CHECK_SAME(Hexahedron::staticElementLineFromFace(f,0),
//        (f<2 ? f : (f<4 ? 4+2*f : 4*f-14))
          (f<2 ? f+8 : (f<4 ? 2*f+4 : 4*f-16))
          );
        CHECK_SAME(Hexahedron::staticElementLineFromFace(f,1),
//        (f<2 ? 4+f : (f<4 ? 5+2*f : 4*f-13))
          (f<2 ? f+10 : (f<4 ? 2*f+5 : 4*f-15))
          );
        CHECK_SAME(Hexahedron::staticElementLineFromFace(f,2),
//        (f<2 ? 8+f : (f<4 ? f : 4*f-16))
          (f<4 ? f : 4*f-14)
          );
        CHECK_SAME(Hexahedron::staticElementLineFromFace(f,3),
//        (f<2 ? 10+f : (f<4 ? 4+f : 4*f-15))
          (f<4 ? f+4 : 4*f-13)
          );
        CHECK_SAME(
          Hexahedron::staticNumberInteriorChildVerticesOnFace(f),1);
        CHECK_SAME(Hexahedron::staticElementVertexFromFace(f,0),
          (f<2 ? f : (f<4 ? 2*f-4 : 4*f-16))
          );
        CHECK_SAME(Hexahedron::staticElementVertexFromFace(f,1),
//        (f<2 ? 4+f : (f<4 ? 2*f-3 : 4*f-14))
          (f<2 ? f+2 : (f<4 ? 2*f-3 : 4*f-15))
          );
        CHECK_SAME(Hexahedron::staticElementVertexFromFace(f,2),
//        (f<2 ? 2+f : (f<4 ? 2*f : 4*f-15))
          (f<2 ? f+4 : (f<4 ? 2*f : 4*f-14))
          );
        CHECK_SAME(Hexahedron::staticElementVertexFromFace(f,3),
//        (f<2 ? 6+f : (f<4 ? 2*f+1 : 4*f-13))
          (f<2 ? f+6 : (f<4 ? 2*f+1 : 4*f-13))
          );
        CHECK_SAME(Hexahedron::staticChildOnFace(f,0),
          (f<2 ? f : (f<4 ? 2*f-4 : 4*f-16)));
        CHECK_SAME(Hexahedron::staticChildOnFace(f,1),
          (f<2 ? 2+f : (f<4 ? 2*f-3 : 4*f-15)));
        CHECK_SAME(Hexahedron::staticChildOnFace(f,2),
          (f<2 ? 4+f : (f<4 ? 2*f : 4*f-14)));
        CHECK_SAME(Hexahedron::staticChildOnFace(f,3),
          (f<2 ? 6+f : (f<4 ? 2*f+1 : 4*f-13)));
//      Tensor<1,3> ft0=Hexahedron::faceTangent(f,0);
//      REAL dot=(ft0[0]*ft0[0]+ft0[1]*ft0[1]+ft0[2]*ft0[2]-1.)/DBL_EPSILON;
//      CHECK_TEST(dot<16.);
//      Tensor<1,3> ft1=Hexahedron::faceTangent(f,1);
//      dot=(ft1[0]*ft1[0]+ft1[1]*ft1[1]+ft1[2]*ft1[2]-1.)/DBL_EPSILON;
//      CHECK_TEST(dot<16.);
//      dot=(ft0[0]*ft1[0]+ft0[1]*ft1[1]+ft0[2]*ft1[2])/DBL_EPSILON;
//      CHECK_TEST(dot<16.);
//      Tensor<1,3> cp;
//      crossProduct(cp,ft0,ft1);
        Tensor<1,3> n=Hexahedron::faceNormal(f);
//      REAL dif=
//        sqrt(pow(n[0]-cp[0],2)+pow(n[1]-cp[1],2)+pow(n[2]-cp[2],2))
//        /DBL_EPSILON;
//      CHECK_TEST(dif<16.);
        int c=INT_MAX,v=INT_MAX;
        Hexahedron::staticInteriorChildVertexOnFace(f,c,v);
        CHECK_SAME(c,(f%2==0 ? 0 : 1<<(f/2)));
        CHECK_SAME(v,(f%2==1 ? 7 : 7-(1<<(f/2))));
        CHECK_SAME(Hexahedron::numberInteriorChildLinesOnFace(f),4);
      }
      for (int cf=0;cf<12;cf++) {
        int c0=INT_MAX;
        int c1=INT_MAX;
        Hexahedron::staticInteriorChildCodimension1ShapeNeighbors(cf,c0,c1);
        CHECK_SAME(c0,(cf<2 ? 4*cf : (cf<4 ? 4*cf-6 : (cf<6 ? cf-4 : 
          (cf<8 ? cf-2 : (cf<10 ? 2*cf-16 : 2*cf-19))))));
        CHECK_SAME(c1,(cf<2 ? 4*cf+1 : (cf<4 ? 4*cf-5 : (cf<6 ? cf-2 : 
          (cf<8 ? cf : (cf<10 ? 2*cf-12 : 2*cf-15))))));
      }
      for (int v=0;v<8;v++) {
        Point<3> p=Hexahedron::canonicalVertex(v);
//      CHECK_TEST(Hexahedron::containsCanonical(p));
//      CHECK_SAME(Hexahedron::childShapeOfCanonical(p),v);
        CHECK_SAME(Hexahedron::oppositeVertex(v),7-v);
      }
      for (int l=0;l<12;l++) {
        for (int v=0;v<2;v++) {
          CHECK_SAME(Hexahedron::staticElementVertexFromLine(l,v),
            (l<2 ? l+2*v : (l<4 ? 2*(l-2)+v : (l<6 ? l+2*v : (l<8 ? 2*l-8+v
            : l-8+4*v ))))); 
        }
      }
      for (int n=1;n<=6;n++) {
        CHECK_SAME(Hexahedron::staticNumberEquallySpacedLatticePoints(n),
          (n+1)*(n+1)*(n+1));
        CHECK_SAME(
          Hexahedron::staticEquallySpacedLatticePointIndexOfVertex(0,n),0);
        CHECK_SAME(
          Hexahedron::staticEquallySpacedLatticePointIndexOfVertex(1,n),n);
        CHECK_SAME(Hexahedron::staticEquallySpacedLatticePointIndexOfVertex(
          2,n),n*(n+1));
        CHECK_SAME(Hexahedron::staticEquallySpacedLatticePointIndexOfVertex(
          3,n),n*(n+2));
      }

      int vertices[8]={0,1,2,3,4,5,6,7};
      int faces[6]={0,1,2,3,4,5};
      int lines[12]={0,1,2,3,4,5,6,7,8,9,10,11};
      Hexahedron e(0,lines,vertices,vertices);
      CHECK_SAME(e.numberInteriorChildVertices(),1);
      c=INT_MAX;
      v=INT_MAX;
      e.interiorChildVertex(c,v);
      CHECK_SAME(c,0);
      CHECK_SAME(v,7);

      NumPtr<Point<3> > canonical_vertices(8);
      for (int i=0;i<8;i++) {
        canonical_vertices[i]=Hexahedron::canonicalVertex(i);
      }
      for (int i=0;i<8;i++) {
        CHECK_TEST(e.contains(canonical_vertices[i],canonical_vertices));
      }
      CHECK_TEST(e.contains(e.center(canonical_vertices),
                            canonical_vertices));
    }
#endif
  }
  return EXIT_SUCCESS;
}
