// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "Boundary.H"
#include "MemoryDebugger.H"
#include "PermutationGenerator.H"
#include "Shape.H"
#include "Tracer.H"

#include "AreaGraphTool.H"
#include "VolGraphTool.H"
#include "XYGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKWindow.H"
#endif

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
//for (int nv=2;nv<=8;nv++) {
//  PermutationGenerator pg(nv);
//  while (pg.hasMore()) {
//    const NumPtr<int> &p=pg.getNext();
//    for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//    cout << endl;
//  }
//  cout << endl;
//}

/* Interval::computeInteriorLatticePoints is obsolete
#if (SPACEDIM==1)
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool gt("testShapeInteriorComputeLattice","x","y",
    0.,1.,-1.,1.,&cmap,0,0.5);
  int nv=2;
  Point<1> vertices[2]={Point<1>(0.),Point<1>(1.)};
  int edi=0;
  int v[2];
  for (int i=0;i<nv;i++) v[i]=i;
  Interval e(edi,v);
  for (int degree=1;degree<=6;degree++) {
    cout << "\ndegree = " << degree << endl;
    NumPtr<Point<1> > mapping_support_points(
      e.numberEquallySpacedLatticePoints(degree));
    for (int v=0;v<nv;v++) {
      int iv=e.equallySpacedLatticePointIndexOfVertex(v,degree);
      mapping_support_points[iv]=vertices[e.vertexIndex(v)];
    }
#ifdef DEBUG
//  cout << "\tmapping_support_points =";
//  for (int i=0;i<mapping_support_points.getNumber();i++) {
//    cout << " " << mapping_support_points[i];
//  }
//  cout << endl;
#endif
    e.computeInteriorLatticePoints(degree,mapping_support_points);
#ifdef DEBUG
//    cout << "\tmapping_support_points =";
//    for (int i=0;i<mapping_support_points.getNumber();i++) {
//      cout << " " << mapping_support_points[i];
//    }
//    cout << endl;
#endif
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    gt.setfgColor("red");
    for (int iv=0;iv<mapping_support_points.getNumber();iv++) {
      char s[4];
      snprintf(s,4,"%d",iv);
      gt.putString(mapping_support_points[iv][0],0.05,s);
    }
    gt.flush();
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#endif
*/

#if (SPACEDIM==2)
//Palette pal;
//AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
//REAL rlo[2]; rlo[0]=0.; rlo[1]=0.;
//REAL rhi[2]; rhi[0]=1.; rhi[1]=1.;
//AreaGraphTool gt("testShapeInteriorComputeLattice","x","y",rlo,rhi,
//  &cmap,0,0.5);

//
  {
    PermutationGenerator pg(2);
    int nv=3;
    while (pg.hasMore()) {
      const NumPtr<int> &p=pg.getNext();
//    NumPtr<int> p(2); p[0]=1; p[1]=0;
#ifdef DEBUG
      cout << "\tpermutation = " << p[0] << " " << p[1] << endl;
#endif
      NumPtr<Point<2> > vertices(nv);
      vertices[0]=Triangle::canonicalVertex(0);
      vertices[p[0]+1]=Triangle::canonicalVertex(1);
      vertices[p[1]+1]=Triangle::canonicalVertex(2);
      Point<2> center=Triangle::staticCenter();
      int vertex_indices[3][3]={ 
        {0,p[0]+1,p[1]+1}, {p[0]+1,p[1]+1,0}, {p[1]+1,0,p[0]+1}
      };
      int edi=0;
      for (int r=0;r<3;r++) {
        cout << "\t\tr = " << r << endl;
        int l[3]={0,1,2};
        Triangle e(edi,l,vertex_indices[r]);
        for (int bf=-1;bf<3;bf++) {
          cout << "\t\t\tbf = " << bf << endl;
          for (int degree=1;degree<=6;degree++) 
//        int degree=3;
          {
            cout << "\t\t\t\tdegree = " << degree << endl;
            NumPtr<Point<2> > mapping_support_points(
              e.numberEquallySpacedLatticePoints(degree));
            NumPtr<Point<2> > equally_spaced_lattice_points;
            e.equallySpacedLatticePoints(degree,
              equally_spaced_lattice_points);
            for (int v=0;v<nv;v++) {
              int iv=
                e.equallySpacedLatticePointIndexOfVertex(v,degree);
              mapping_support_points[iv]=vertices[e.vertexIndex(v)];
#ifdef DEBUG
//            cout << "\tmapping_support_points[" << iv << "] = "
//                 << mapping_support_points[iv] << endl;
#endif
            }
            if (degree>1) {
              for (int f=0;f<e.numberCodimension1Shapes();f++) {
                int fv[2];
                fv[0]=e.vertexIndex(e.elementVertexFromLine(f,0));
                fv[1]=e.vertexIndex(e.elementVertexFromLine(f,1));
                Interval face_element(0,fv);
                int nieslp=face_element.
                  numberInteriorEquallySpacedLatticePoints(degree);
                NumPtr<Point<2> > fsp(nieslp);
                PlaneBoundary sb(center,vertices[fv[0]],vertices[fv[1]]);
                sb.getIntermediatePointsOn(face_element,vertices,fsp);
#ifdef DEBUG
//              cout << "\tface_element:" << endl;
//              face_element.printOn(cout);
#endif
                NumPtr<int> indices;
                e.equallySpacedLatticePointIndicesOfFaceInterior(f,degree,
                  indices);
                for (int i=0;i<fsp.getNumber();i++) {
                  mapping_support_points[indices[i]]=fsp[i];
#ifdef DEBUG
//                cout << "\tfsp[" << i << "] = " << fsp[i] << endl;
//                cout << "\tmapping_support_points[" << iv << "] = "
//                     << mapping_support_points[iv] << endl;
#endif
                }
              }
            }
            e.computeInteriorLatticePoints(degree,bf,
              mapping_support_points);
#ifdef DEBUG
//          for (int iv=0;iv<mapping_support_points.getNumber();iv++) {
//            cout << "\tmapping_support_points[" << iv << "] = "
//                 << mapping_support_points[iv] << endl;
//          }
#endif
/*
            gt.newPage();
            gt.setbgColor("white");
            gt.setfgColor("black");
            gt.drawAxes();
            gt.setfgColor("blue");
            for (int i=0;
            i<equally_spaced_lattice_points.getNumber(); i++) {
              char s[6];
              snprintf(s,6,"%d",i);
              gt.putString(equally_spaced_lattice_points[i][0],
                           equally_spaced_lattice_points[i][1],s);
//            cout << "\tequally_spaced_lattice_points[" << i << "] = " 
//                 << equally_spaced_lattice_points[i] << endl;
            }
            gt.setfgColor("red");
            for (int i=0;i<mapping_support_points.getNumber();i++)
            {
              char s[6];
              snprintf(s,6,"%d",i);
              gt.putString(mapping_support_points[i][0],
                         mapping_support_points[i][1],s);
//            cout << "\tmapping_support_points[" << i << "] = " 
//                 << mapping_support_points[i] << endl;
            }
            gt.flush();
            AreaGraphTool::WINDOW_TYPE::QuitButton qb;
*/
            for (int i=0;i<mapping_support_points.getNumber();i++)
            {
              REAL error=equally_spaced_lattice_points[i].
                distance(mapping_support_points[i]);
              if (error>256.*DBL_EPSILON) {
                cout << "\tdegree = " << degree << endl;
                cout << "\tequally_spaced_lattice_points[" << i << "] = "
                     << equally_spaced_lattice_points[i] << " "
                     << mapping_support_points[i] << " " << error
                     << " " << error/DBL_EPSILON << endl;
                OBSOLETE("bad mapping_support_point");
              }
            }
          }
        }
      }
    }
  }
//
//
  {
    PermutationGenerator pg(3);
    int nv=4;
    while (pg.hasMore()) 
    {
      const NumPtr<int> &p=pg.getNext();
//    NumPtr<int> p(3); p[0]=0; p[1]=2; p[2]=1;
#ifdef DEBUG
      cout << "\tpermutation = " << p[0] << " " << p[1] << " " << p[2]
           << endl;
#endif
      NumPtr<Point<2> > vertices(nv);
      vertices[0]=Quadrilateral::canonicalVertex(0);
      vertices[p[0]+1]=Quadrilateral::canonicalVertex(1);
      vertices[p[1]+1]=Quadrilateral::canonicalVertex(2);
      vertices[p[2]+1]=Quadrilateral::canonicalVertex(3);
      Point<2> center=Quadrilateral::staticCenter();
      int vertex_indices[4][4]={ 
        {0,p[0]+1,p[1]+1,p[2]+1}, {p[0]+1,p[2]+1,0,p[1]+1}, 
        {p[2]+1,p[1]+1,p[0]+1,0}, {p[1]+1,0,p[2]+1,p[0]+1} 
      };
      int edi=0;
      int l[4]={0,1,2,3};
      for (int r=0;r<4;r++) 
//    int r=0;
      {
        cout << "\t\tr = " << r << endl;
        Quadrilateral e(edi,l,vertex_indices[r]);
        for (int bf=-1;bf<1;bf++) {
          cout << "\t\t\tbf = " << bf << endl;
          for (int degree=1;degree<=6;degree++) 
//        int degree=3;
          {
            cout << "\t\t\t\tdegree = " << degree << endl;
            NumPtr<Point<2> > mapping_support_points(
              e.numberEquallySpacedLatticePoints(degree));
            NumPtr<Point<2> > equally_spaced_lattice_points;
            e.equallySpacedLatticePoints(degree,
              equally_spaced_lattice_points);
            for (int v=0;v<nv;v++) {
              int iv=
                e.equallySpacedLatticePointIndexOfVertex(v,degree);
              mapping_support_points[iv]=vertices[e.vertexIndex(v)];
#ifdef DEBUG
//            cout << "\tmapping_support_points[" << iv << "] = "
//                 << mapping_support_points[iv] << endl;
#endif
            }
            if (degree>1) {
              for (int f=0;f<e.numberCodimension1Shapes();f++) {
                int fv[2];
                fv[0]=e.vertexIndex(e.elementVertexFromLine(f,0));
                fv[1]=e.vertexIndex(e.elementVertexFromLine(f,1));
                Interval face_element(0,fv);
                int nieslp=face_element.
                  numberInteriorEquallySpacedLatticePoints(degree);
                NumPtr<Point<2> > fsp(nieslp);
                PlaneBoundary sb(center,vertices[fv[0]],vertices[fv[1]]);
                sb.getIntermediatePointsOn(face_element,vertices,fsp);
                NumPtr<int> indices;
                e.equallySpacedLatticePointIndicesOfFaceInterior(f,degree,
                  indices);
                for (int i=0;i<fsp.getNumber();i++) {
                  mapping_support_points[indices[i]]=fsp[i];
#ifdef DEBUG
//                cout << "\tfsp[" << i << "] = " << fsp[i] << endl;
//                cout << "\tmapping_support_points[" << indices[i] 
//                     << "] = " << mapping_support_points[indices[i]] 
//                     << endl;
#endif
                }
              }
            }
            e.computeInteriorLatticePoints(degree,bf,
              mapping_support_points);
#ifdef DEBUG
//          for (int i=0;i<mapping_support_points.getNumber();i++) {
//            cout << "\tmapping_support_points[" << i << "] = "
//                 << mapping_support_points[i] << endl;
//          }
#endif
/*
            gt.newPage();
            gt.setbgColor("white");
            gt.setfgColor("black");
            gt.drawAxes();
            for (int i=0;
            i<equally_spaced_lattice_points.getNumber();i++) {
              char s[6];
              snprintf(s,6,"%d",i);
              gt.putString(equally_spaced_lattice_points[i][0],
                           equally_spaced_lattice_points[i][1],s);
//            cout << "\tequally_spaced_lattice_points[" << i << "] = " 
//                 << equally_spaced_lattice_points[i] << endl;
            }
            gt.setfgColor("red");
            for (int i=0;i<mapping_support_points.getNumber();
            i++) {
              char s[6];
              snprintf(s,6,"%d",i);
              gt.putString(mapping_support_points[i][0],
                         mapping_support_points[i][1],s);
//            cout << "\tmapping_support_points[" << i << "] = " 
//                 << mapping_support_points[i] << endl;
            }
            gt.flush();
            AreaGraphTool::WINDOW_TYPE::QuitButton qb;
*/
            for (int i=0;i<mapping_support_points.getNumber();i++)
            {
              REAL error=equally_spaced_lattice_points[i].
                distance(mapping_support_points[i]);
              if (error>256.*DBL_EPSILON) {
                cout << "\tdegree = " << degree << endl;
                cout << "\tequally_spaced_lattice_points[" << i << "] = "
                     << equally_spaced_lattice_points[i] << " "
                     << mapping_support_points[i] << " " << error
                     << " " << error/DBL_EPSILON << endl;
                OBSOLETE("bad mapping_support_point");
              }
            }
          }
        }
      }
    }
  }
//
#endif

#if (SPACEDIM==3)
/*
  {
    PermutationGenerator pg(2);
    int nv=4;
    while (pg.hasMore()) 
    {
      const NumPtr<int> &p=pg.getNext();
//    NumPtr<int> p(2); p[0]=1; p[1]=0;
#ifdef DEBUG
      cout << "\tpermutation = " << p[0] << " " << p[1] << endl;
#endif
      NumPtr<Point<3> > vertices(nv);
      vertices[0]=Tetrahedron::canonicalVertex(0);
      vertices[1]=Tetrahedron::canonicalVertex(1);
      vertices[p[0]+2]=Tetrahedron::canonicalVertex(2);
      vertices[p[1]+2]=Tetrahedron::canonicalVertex(3);
      Point<3> center=Tetrahedron::staticCenter();
      int vertex_indices[12][4]={
        {0,1,p[0]+2,p[1]+2},{0,p[0]+2,p[1]+2,1},{0,p[1]+2,1,p[0]+2},
        {1,0,p[1]+2,p[0]+2},{1,p[0]+2,0,p[1]+2},{1,p[1]+2,p[0]+2,0},
        {p[0]+2,0,1,p[1]+2},{p[0]+2,1,p[1]+2,0},{p[0]+2,p[1]+2,0,1},
        {p[1]+2,0,p[0]+2,1},{p[1]+2,1,0,p[0]+2},{p[1]+2,p[0]+2,1,0}
      };
      int edi=0;
      int l[6]={0,1,2,3,4,5};
      int f[4]={0,1,2,3};
      for (int r=0;r<12;r++) 
//    int r=0;
      {
        cout << "\t\tr = " << r << endl;
        Tetrahedron e(edi,l,f,vertex_indices[r]);
//      since there is no Tetrahedron::computeInteriorLatticePoints
//        we must keep degree<=3
        for (int degree=1;degree<=3;degree++) 
//      int degree=3;
        {
          cout << "\t\t\tdegree = " << degree << endl;
          for (int bf=0;bf<e.numberCodimension1Shapes();bf++) 
//        int bf=0;
          {
            cout << "\t\t\t\tbf = " << bf << endl;
            NumPtr<Point<3> > mapping_support_points(
              e.numberEquallySpacedLatticePoints(degree));
            NumPtr<Point<3> > equally_spaced_lattice_points;
            e.equallySpacedLatticePoints(degree,
              equally_spaced_lattice_points);
            for (int v=0;v<nv;v++) {
              int iv=
                e.equallySpacedLatticePointIndexOfVertex(v,degree);
              mapping_support_points[iv]=vertices[e.vertexIndex(v)];
#ifdef DEBUG
//            cout << "\tmapping_support_points[" << iv << "] = "
//                 << mapping_support_points[iv] << endl;
#endif
            }
            if (degree>1) {
//            TRACER_CALL(t,"testShapeInteriorComputeLattice");
              for (int l=0;l<e.numberCodimension2Shapes();l++) {
                int lv[2];
                lv[0]=e.vertexIndex(e.elementVertexFromLine(l,0));
                lv[1]=e.vertexIndex(e.elementVertexFromLine(l,1));
                Interval line_element(0,lv);
                int nieslp=line_element.
                  numberInteriorEquallySpacedLatticePoints(degree);
                NumPtr<Point<3> > fsp(nieslp);

                int fv[3];
                fv[0]=e.vertexIndex(e.elementVertexFromFace(bf,0));
                fv[1]=e.vertexIndex(e.elementVertexFromFace(bf,1));
                fv[2]=e.vertexIndex(e.elementVertexFromFace(bf,2));
                PlaneBoundary sb(center,vertices[fv[0]],
                  vertices[fv[1]],vertices[fv[2]]);
                sb.getIntermediatePointsOn(line_element,vertices,fsp);
                NumPtr<int> indices;
                e.equallySpacedLatticePointIndicesOfLineInterior(l,degree,
                  indices);
#ifdef DEBUG
//              for (int i=0;i<indices.getNumber();i++) {
//                cout << "\tindices[" << i << "] = " << indices[i] 
//                     << endl;
//              }
#endif
                for (int i=0;i<fsp.getNumber();i++) {
                  int iv=indices[i];
                  mapping_support_points[iv]=fsp[i];
#ifdef DEBUG
//                cout << "\tmapping_support_points[" << iv << "] = "
//                     << mapping_support_points[iv] << endl;
#endif
                }
              }
            }
            if (degree>2) {
//            TRACER_CALL(t,"testShapeInteriorComputeLattice");
              for (int f=0;f<e.numberCodimension1Shapes();f++) {
                int fl[3]={0,1,2};
                int fv[3];
                fv[0]=e.vertexIndex(e.elementVertexFromFace(f,0));
                fv[1]=e.vertexIndex(e.elementVertexFromFace(f,1));
                fv[2]=e.vertexIndex(e.elementVertexFromFace(f,2));
//              cout << "\tfv = " << fv[0] << " " << fv[1] << " " << fv[2]
//                   << endl;
                Triangle face_element(0,fl,fv);
                int nieslp=face_element.
                  numberInteriorEquallySpacedLatticePoints(degree);
                NumPtr<Point<3> > fsp(nieslp);
                PlaneBoundary sb(center,vertices[fv[0]],
                  vertices[fv[1]],vertices[fv[2]]);
                sb.getIntermediatePointsOn(face_element,vertices,fsp);
                NumPtr<int> indices;
                e.equallySpacedLatticePointIndicesOfFaceInterior(f,degree,
                  indices);
#ifdef DEBUG
//              for (int i=0;i<indices.getNumber();i++) {
//                cout << "\tindices[" << i << "] = " << indices[i] 
//                     << endl;
//              }
#endif
                for (int i=0;i<fsp.getNumber();i++) {
                  int iv=indices[i];
                  mapping_support_points[iv]=fsp[i];
#ifdef DEBUG
//                cout << "\tmapping_support_points[" << iv << "] = "
//                     << mapping_support_points[iv] << endl;
#endif
                }
              }
            }
//          for (int i=0;
//          i<nv+9*(degree-1)+((degree-1)*(degree-2))
//          +3*(degree-1)*(degree-1);i++) {
//            cout << "\tmapping_support_points[" << i << "] = "
//                 << mapping_support_points[i] << endl;
//          }
//
            if (degree>3) {
              e.computeInteriorLatticePoints(degree,bf,
                mapping_support_points);
#ifdef DEBUG
//            for (int iv=0;iv<mapping_support_points.getNumber();iv++) {
//              cout << "\tmapping_support_points[" << iv << "] = "
//                   << mapping_support_points[iv] << endl;
//            }
#endif
            }
//
            for (int i=0;i<mapping_support_points.getNumber();
            i++) {
//            cout << "\t\tmapping_support_points[" << i << "] = " 
//                 << mapping_support_points[i] << endl;
              Point<SPACEDIM> &msp=mapping_support_points[i];
              CHECK_TEST(abs(msp[0])+abs(msp[1])+abs(msp[2])<SPACEDIM+1);
              REAL error=equally_spaced_lattice_points[i].
                distance(mapping_support_points[i]);
              if (error>DBL_EPSILON) {
                cout << "\t\tmapping_support_points[" << i << "] = " 
                     << mapping_support_points[i] << endl;
                cout << "\tequally_spaced_lattice_points[" << i << "] = "
                     << equally_spaced_lattice_points[i] << endl;
                cout << "\terror = " << error << " " << error/DBL_EPSILON
                     << endl;
                OBSOLETE("bad mapping_support_point");
              }
            }
          }
        }
      }
    }
  }
*/
/*
  {
    PermutationGenerator pg(5);
    int nv=6;
    while (pg.hasMore()) 
    {
      const NumPtr<int> &p=pg.getNext();
//    NumPtr<int> p(5); p[0]=0; p[1]=1; p[2]=3; p[3]=2; p[4]=4;
#ifdef DEBUG
      cout << "\tpermutation = " << p[0] << " " << p[1] << " " << p[2]
           << " " << p[3] << " " << p[4] << endl;
#endif
      NumPtr<Point<3> > vertices(nv);
      vertices[0]=Prism::canonicalVertex(0);
      vertices[p[0]+1]=Prism::canonicalVertex(1);
      vertices[p[1]+1]=Prism::canonicalVertex(2);
      vertices[p[2]+1]=Prism::canonicalVertex(3);
      vertices[p[3]+1]=Prism::canonicalVertex(4);
      vertices[p[4]+1]=Prism::canonicalVertex(5);
      Point<3> center=Prism::staticCenter();
      int vertex_indices[6][6]={
        {0,p[0]+1,p[1]+1,p[2]+1,p[3]+1,p[4]+1},
        {p[0]+1,p[1]+1,0,p[3]+1,p[4]+1,p[2]+1},
        {p[1]+1,0,p[0]+1,p[4]+1,p[2]+1,p[3]+1},
        {p[2]+1,p[4]+1,p[3]+1,0,p[1]+1,p[0]+1},
        {p[3]+1,p[2]+1,p[4]+1,p[0]+1,0,p[1]+1},
        {p[4]+1,p[3]+1,p[2]+1,p[1]+1,p[0]+1,0}
      };
      int edi=0;
      int l[9]={0,1,2,3,4,5,6,7,8};
      int f[5]={0,1,2,3,4};
      for (int r=0;r<6;r++)
//    int r=0;
      {
        cout << "\t\tr = " << r << endl;
        Prism e(edi,l,f,vertex_indices[r]);
#ifdef DEBUG
//      e.printOn(cout);
#endif
        for (int degree=1;degree<=6;degree++) 
//      int degree=3;
        {
          cout << "\t\t\tdegree = " << degree << endl;
          for (int bf=0;bf<e.numberCodimension1Shapes();bf++) 
//        int bf=4;
          {
            cout << "\t\t\t\tbf = " << bf << endl;
            NumPtr<Point<3> > mapping_support_points(
              e.numberEquallySpacedLatticePoints(degree));
            NumPtr<Point<3> > equally_spaced_lattice_points;
            e.equallySpacedLatticePoints(degree,
              equally_spaced_lattice_points);
            for (int v=0;v<nv;v++) {
              int iv=
                e.equallySpacedLatticePointIndexOfVertex(v,degree);
              mapping_support_points[iv]=vertices[e.vertexIndex(v)];
#ifdef DEBUG
//            cout << "\tmapping_support_points[" << iv << "] = "
//                 << mapping_support_points[iv] << endl;
#endif
            }
            if (degree>1) {
//            TRACER_CALL(t,"testShapeInteriorComputeLattice");
              for (int l=0;l<e.numberCodimension2Shapes();l++) {
                int lv[2];
                lv[0]=e.vertexIndex(e.elementVertexFromLine(l,0));
                lv[1]=e.vertexIndex(e.elementVertexFromLine(l,1));
                Interval face_element(0,lv);
                int nieslp=face_element.
                numberInteriorEquallySpacedLatticePoints(degree);
                NumPtr<Point<3> > fsp(nieslp);

                int fv[3];
                fv[0]=e.vertexIndex(e.elementVertexFromFace(bf,0));
                fv[1]=e.vertexIndex(e.elementVertexFromFace(bf,1));
                fv[2]=e.vertexIndex(e.elementVertexFromFace(bf,2));
                PlaneBoundary sb(center,vertices[fv[0]],
                  vertices[fv[1]],vertices[fv[2]]);
                sb.getIntermediatePointsOn(face_element,vertices,fsp);
                NumPtr<int> indices;
                e.equallySpacedLatticePointIndicesOfLineInterior(l,degree,
                  indices);
#ifdef DEBUG
//              cout << "\tl = " << l << endl;
//              for (int i=0;i<indices.getNumber();i++) {
//                cout << "\tindices[" << i << "] = " << indices[i] 
//                     << endl;
//              }
#endif
                for (int i=0;i<fsp.getNumber();i++) {
                  int iv=indices[i];
                  mapping_support_points[iv]=fsp[i];
#ifdef DEBUG
//                cout << "\tmapping_support_points[" << iv << "] = " 
//                     << mapping_support_points[iv] << endl;
#endif
                }
              }
            }
            if (degree>2) {
//            TRACER_CALL(t,"testShapeInteriorComputeLattice");
              for (int f=0;f<2;f++) {
                int fl[3]={0,1,2};
                int fv[3];
                fv[0]=e.vertexIndex(e.elementVertexFromFace(f,0));
                fv[1]=e.vertexIndex(e.elementVertexFromFace(f,1));
                fv[2]=e.vertexIndex(e.elementVertexFromFace(f,2));
//              cout << "\tfv = " << fv[0] << " " << fv[1] << " " << fv[2]
//                   << endl;
                Triangle face_element(0,fl,fv);
                int nieslp=face_element.
                  numberInteriorEquallySpacedLatticePoints(degree);
                NumPtr<Point<3> > fsp(nieslp);
                PlaneBoundary sb(center,vertices[fv[0]],
                  vertices[fv[1]],vertices[fv[2]]);
                sb.getIntermediatePointsOn(face_element,vertices,fsp);
                NumPtr<int> indices;
                e.equallySpacedLatticePointIndicesOfFaceInterior(f,degree,
                  indices);
#ifdef DEBUG
//              cout << "\tf = " << f << endl;
//              for (int i=0;i<indices.getNumber();i++) {
//                cout << "\tindices[" << i << "] = " << indices[i] 
//                     << endl;
//              }
#endif
                for (int i=0;i<fsp.getNumber();i++) {
                  int iv=indices[i];
                  mapping_support_points[iv]=fsp[i];
#ifdef DEBUG
//                cout << "\tmapping_support_points[" << iv << "] = " 
//                     << mapping_support_points[iv] << endl;
#endif
                }
              }
            }
            if (degree>1) {
//            TRACER_CALL(t,"testShapeInteriorComputeLattice");
              for (int f=2;f<e.numberCodimension1Shapes();f++) {
                int fl[4]={0,1,2,3};
                int fv[4];
                fv[0]=e.vertexIndex(e.elementVertexFromFace(f,0));
                fv[1]=e.vertexIndex(e.elementVertexFromFace(f,1));
                fv[2]=e.vertexIndex(e.elementVertexFromFace(f,2));
                fv[3]=e.vertexIndex(e.elementVertexFromFace(f,3));
//              cout << "\tfv = " << fv[0] << " " << fv[1] << " " << fv[2]
//                   << " " << fv[3] << endl;
                Quadrilateral face_element(0,fl,fv);
                int nieslp=face_element.
                numberInteriorEquallySpacedLatticePoints(degree);
                NumPtr<Point<3> > fsp(nieslp);
                PlaneBoundary sb(center,vertices[fv[0]],
                  vertices[fv[1]],vertices[fv[2]]);
                sb.getIntermediatePointsOn(face_element,vertices,fsp);
                NumPtr<int> indices;
                e.equallySpacedLatticePointIndicesOfFaceInterior(f,degree,
                  indices);
#ifdef DEBUG
//              cout << "\tf = " << f << endl;
//              for (int i=0;i<indices.getNumber();i++) {
//                cout << "\tindices[" << i << "] = " << indices[i] 
//                     << endl;
//              }
#endif
                for (int i=0;i<fsp.getNumber();i++) {
                  int iv=indices[i];
                  mapping_support_points[iv]=fsp[i];
#ifdef DEBUG
//                cout << "\tmapping_support_points[" << iv << "] = " 
//                     << mapping_support_points[iv] << endl;
#endif
                }
              }
            }
//
            if (degree>2) {
              e.computeInteriorLatticePoints(degree,bf,
                mapping_support_points);
            }
//
#ifdef DEBUG
//          for (int i=0;i<mapping_support_points.getNumber();i++) {
//            cout << "\tmapping_support_points[" << i << "] = "
//                 << mapping_support_points[i] << endl;
//          }
#endif
            for (int i=0;i<mapping_support_points.getNumber();i++)
            {
              Point<SPACEDIM> &msp=mapping_support_points[i];
//            CHECK_TEST(abs(msp[0])+abs(msp[1])+abs(msp[2])<SPACEDIM+1);
              REAL error=equally_spaced_lattice_points[i].
                distance(mapping_support_points[i]);
              if (error>16.*DBL_EPSILON) {
                cout << "\tmapping_support_points[" << i << "] = " 
                     << mapping_support_points[i] << endl;
                cout << "\tequally_spaced_lattice_points[" << i << "] = "
                     << equally_spaced_lattice_points[i] << endl;
                cout << "\terror = " << error << " " << error/DBL_EPSILON
                     << endl;
                OBSOLETE("bad mapping_support_point");
              }
            }
          }
        }
      }
    }
  }
*/
//
  {
    int nv=8;
    {
      NumPtr<Point<3> > vertices(nv);
      for (int i=0;i<8;i++) {
        vertices[i]=Hexahedron::canonicalVertex(i);
      }
      Point<3> center=Hexahedron::staticCenter();
      int vertex_indices[8]={ 0,1,2,3,4,5,6,7 };
      int edi=0;
      int l[12]={0,1,2,3,4,5,6,7,8,9,10,11};
      int f[6]={0,1,2,3,4,5};
      Hexahedron e(edi,l,f,vertex_indices);
      for (int bf=-1;bf<1;bf++) 
//    int bf=-1;
      {
        cout << "\tbf = " << bf << endl;
        for (int degree=1;degree<=6;degree++) 
//      int degree=2;
        {
          cout << "\t\tdegree = " << degree << endl;
          for (int f=0;f<e.numberCodimension1Shapes();f++) {
            NumPtr<Point<3> > mapping_support_points(
              e.numberEquallySpacedLatticePoints(degree));
            NumPtr<Point<3> > equally_spaced_lattice_points;
            e.equallySpacedLatticePoints(degree,
              equally_spaced_lattice_points);
            for (int v=0;v<nv;v++) {
              int iv=
                e.equallySpacedLatticePointIndexOfVertex(v,degree);
              mapping_support_points[iv]=vertices[e.vertexIndex(v)];
            }
            if (degree>1) {
              for (int l=0;l<e.numberCodimension2Shapes();l++) {
                int lv[2];
                lv[0]=e.vertexIndex(e.elementVertexFromLine(l,0));
                lv[1]=e.vertexIndex(e.elementVertexFromLine(l,1));
                Interval face_element(0,lv);
                int nieslp=face_element.
                numberInteriorEquallySpacedLatticePoints(degree);
                  NumPtr<Point<3> > fsp(nieslp);
                int fv[3];
                fv[0]=e.vertexIndex(e.elementVertexFromFace(f,0));
                fv[1]=e.vertexIndex(e.elementVertexFromFace(f,1));
                fv[2]=e.vertexIndex(e.elementVertexFromFace(f,2));
                PlaneBoundary sb(center,vertices[fv[0]],vertices[fv[1]],
                  vertices[fv[2]]);
                sb.getIntermediatePointsOn(face_element,vertices,fsp);
                NumPtr<int> indices;
                e.equallySpacedLatticePointIndicesOfLineInterior(l,degree,
                  indices);
#ifdef DEBUG
//              cout << "\tl = " << l << endl;
//              for (int i=0;i<indices.getNumber();i++) {
//                cout << "\tindices[" << i << "] = " << indices[i] 
//                     << endl;
//              }
#endif
                for (int i=0;i<fsp.getNumber();i++) {
                  int iv=indices[i];
                  mapping_support_points[iv]=fsp[i];
#ifdef DEBUG
//                cout << "\tmapping_support_points[" << iv << "] = " 
//                     << mapping_support_points[iv] << endl;
#endif
                }
              }
            }
            if (degree>1) {
              for (int f=0;f<e.numberCodimension1Shapes();f++) {
                int fl[4]={0,1,2,3};
                int fv[4];
                fv[0]=e.vertexIndex(e.elementVertexFromFace(f,0));
                fv[1]=e.vertexIndex(e.elementVertexFromFace(f,1));
                fv[2]=e.vertexIndex(e.elementVertexFromFace(f,2));
                fv[3]=e.vertexIndex(e.elementVertexFromFace(f,3));
#ifdef DEBUG
//              cout << "\tfv = " << fv[0] << " " << fv[1] << " " << fv[2]
//                   << " " << fv[3] << endl;
#endif
                Quadrilateral face_element(0,fl,fv);
                int nieslp=face_element.
                  numberInteriorEquallySpacedLatticePoints(degree);
                NumPtr<Point<3> > fsp(nieslp);
                PlaneBoundary sb(center,vertices[fv[0]],vertices[fv[1]],
                  vertices[fv[2]]);
                sb.getIntermediatePointsOn(face_element,vertices,fsp);
                NumPtr<int> indices;
                e.equallySpacedLatticePointIndicesOfFaceInterior(f,degree,
                  indices);
#ifdef DEBUG
//              cout << "\tf = " << f << endl;
//              for (int i=0;i<indices.getNumber();i++) {
//                cout << "\tindices[" << i << "] = " << indices[i] 
//                     << endl;
//              }
#endif
                for (int i=0;i<fsp.getNumber();i++) {
                  int iv=indices[i];
                  mapping_support_points[iv]=fsp[i];
#ifdef DEBUG
//                cout << "\tmapping_support_points[" << iv << "] = " 
//                     << mapping_support_points[iv] << endl;
#endif
                }
              }
            }
#ifdef DEBUG
//          for (int i=0;
//          i<nv+12*(degree-1)+6*(degree-1)*(degree-1);i++) {
//            cout << "\tmapping_support_points[" << i << "] = "
//                 << mapping_support_points[i] << endl;
//          }
#endif
            if (degree>1) {
              e.computeInteriorLatticePoints(degree,bf,
                mapping_support_points);
            }
#ifdef DEBUG
//          for (int i=0;i<mapping_support_points.getNumber();i++)
//          {
//            cout << "\tmapping_support_points[" << i << "] = "
//                 << mapping_support_points[i] << endl;
//          }
#endif
            for (int i=0;i<mapping_support_points.getNumber();i++)
            {
              Point<SPACEDIM> &msp=mapping_support_points[i];
              CHECK_TEST(abs(msp[0])+abs(msp[1])+abs(msp[2])<SPACEDIM+1);
              REAL error=equally_spaced_lattice_points[i].
                distance(mapping_support_points[i]);
              if (error>16.*DBL_EPSILON) {
                cout << "\tmapping_support_points[" << i << "] = " 
                     << mapping_support_points[i] << endl;
                cout << "\tequally_spaced_lattice_points[" << i << "] = "
                     << equally_spaced_lattice_points[i] << endl;
                cout << "\terror = " << error << " " << error/DBL_EPSILON
                     << endl;
                OBSOLETE("bad mapping_support_point");
              }
            }
          }
        }
      }
    }
  }
//
#endif

  return EXIT_SUCCESS;
}
