// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "MemoryDebugger.H"
#include "Shape.H"
#include "Tracer.H"

int main(int argc,char* argv[]) {
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif

#if (SPACEDIM==1)
    cout << "Interval:" << endl;
    for (int n=1;n<=6;n++) {
      int neslp=Interval::staticNumberEquallySpacedLatticePoints(n);
      cout << "\nn = " << n << endl;
      cout << "\tnumberEquallySpacedLatticePoints(" << n << ") = "
           << neslp << endl;
      NumPtr<Point<1> > eslp;
      Interval::staticEquallySpacedLatticePoints(n,eslp);
      for (int i=0;i<neslp;i++) {
        cout << "\teslp[" << i << "] = " << eslp[i] << endl;
      }
//    cout << "\tisParentEquallySpacedLatticePoint:" << endl;
//    for (int c=0;c<childrenPerShape();c++) {
//      for (int i=0;i<neslp;i++) {
//        int I=INT_MAX;
//        if (Interval::staticIsParentEquallySpacedLatticePoint(c,n,i,I)) {
//          cout << "\tc,i,I = " << c << " " << i << " " << I << endl;
//        }
//      }
//    }
    }
#endif

#if (SPACEDIM==2)
    cout << "Triangle:" << endl;
    for (int n=1;n<=6;n++) {
      int neslp=Triangle::staticNumberEquallySpacedLatticePoints(n);
      cout << "\nn = " << n << endl;
      cout << "\tnumberEquallySpacedLatticePoints(" << n << ") = "
           << neslp << endl;
      NumPtr<Point<2> > eslp;
      Triangle::staticEquallySpacedLatticePoints(n,eslp);
      for (int i=0;i<neslp;i++) {
        cout << "\teslp[" << i << "] = " << eslp[i] << endl;
      }
//    cout << "\tisParentEquallySpacedLatticePoint:" << endl;
//    for (int c=0;c<childrenPerShape();c++) {
//      for (int i=0;i<neslp;i++) {
//        int K=INT_MAX;
//        if (Triangle::staticIsParentEquallySpacedLatticePoint(c,n,i,K)) {
//          cout << "\tc,i,K = " << c << " " << i << " " << K << endl;
//        }
//      }
//    }
      cout << "\tequallySpacedLatticePointIndexOfVertex:" << endl;
      for (int v=0;v<Triangle::verticesPerShape();v++) {
        cout << "\tv = " << v << " ==> " 
             << Triangle::staticEquallySpacedLatticePointIndexOfVertex(v,n)
             << endl;
      }
      cout << "\tstaticEquallySpacedLatticePointIndicesOfFace:" << endl;
      for (int f=0;f<Triangle::facesPerShape();f++) {
        NumPtr<int> indices;
        Triangle::staticEquallySpacedLatticePointIndicesOfFace(f,n,indices);
        if (n>1) {
          switch (f) {
            case 0:
              CHECK_SAME(indices[0],0);
              CHECK_SAME(indices[n],(n*(n+3))/2);
              break;
            case 1:
              CHECK_SAME(indices[0],0);
              CHECK_SAME(indices[n],n);
              break;
            case 2:
              CHECK_SAME(indices[0],n);
              CHECK_SAME(indices[n],(n*(n+3))/2);
              break;
          }
        }
//      cout << "\tstaticEquallySpacedLatticePointIndicesOfFace(" << f
//           << "):" << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tindices[" << i << "] = " << indices[i] << endl;
//      }
      }
      cout << "\tequallySpacedLatticePointIndicesOfFaceInterior:" << endl;
      for (int f=0;f<Triangle::facesPerShape();f++) {
        NumPtr<int> indices;
        Triangle::staticEquallySpacedLatticePointIndicesOfFaceInterior(f,n,
          indices);
        if (n>1) {
          switch (f) {
            case 0:
              CHECK_SAME(indices[0],n+1);
              CHECK_SAME(indices[n-2],((n-1)*(n+4))/2);
              break;
            case 1:
              CHECK_SAME(indices[0],1);
              CHECK_SAME(indices[n-2],n-1);
              break;
            case 2:
              CHECK_SAME(indices[0],2*n);
              CHECK_SAME(indices[n-2],((n-1)*(n+4))/2+1);
              break;
          }
        }
//      cout << "\tequallySpacedLatticePointIndicesOfFaceInterior(" << f
//           << "):" << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tindices[" << i << "] = " << indices[i] << endl;
//      }
      }
//    cout << "\tequallySpacedLatticePointIndicesOfFaceInterior:" << endl;
//    for (int f=0;f<Triangle::facesPerShape();f++) {
//      cout << "\tf = " << f << endl;
//      NumPtr<int> indices;
//      Triangle::equallySpacedLatticePointIndicesOfFaceInterior(f,n,
//        indices);      
//      for (int v=0;v<n-1;v++) {
//        cout << "\t\tv = " << v << " ==> " << indices[v] << endl;
//      }
//    }
    }
    cout << "Quadrilateral:" << endl;
    for (int n=1;n<=6;n++) {
      int neslp=Quadrilateral::staticNumberEquallySpacedLatticePoints(n);
      cout << "\nn = " << n << endl;
      cout << "\tnumberEquallySpacedLatticePoints(" << n << ") = "
           << neslp << endl;
      NumPtr<Point<2> > eslp;
      Quadrilateral::staticEquallySpacedLatticePoints(n,eslp);
      for (int i=0;i<neslp;i++) {
        cout << "\teslp[" << i << "] = " << eslp[i] << endl;
      }
//    cout << "\tisParentEquallySpacedLatticePoint:" << endl;
//    for (int c=0;c<childrenPerShape();c++) {
//      for (int i=0;i<neslp;i++) {
//        int K=INT_MAX;
//        if (Quadrilateral::staticIsParentEquallySpacedLatticePoint(
//        c,n,i,K)) {
//          cout << "\tc,i,K = " << c << " " << i << " " << K << endl;
//        }
//      }
//    }
      cout << "\tequallySpacedLatticePointIndexOfVertex:" << endl;
      for (int v=0;v<Quadrilateral::verticesPerShape();v++) {
        cout << "\tv = " << v << " ==> " 
          << Quadrilateral::staticEquallySpacedLatticePointIndexOfVertex(
             v,n) << endl;
      }
      for (int f=0;f<Quadrilateral::facesPerShape();f++) {
        NumPtr<int> indices;
        Quadrilateral::staticEquallySpacedLatticePointIndicesOfFace(f,n,
          indices);
        if (n>1) {
          switch (f) {
            case 0:
              CHECK_SAME(indices[0],0);
              CHECK_SAME(indices[n],n*(n+1));
              break;
            case 1:
              CHECK_SAME(indices[0],n);
              CHECK_SAME(indices[n],n*(n+2));
              break;
            case 2:
              CHECK_SAME(indices[0],0);
              CHECK_SAME(indices[n],n);
              break;
            case 3:
              CHECK_SAME(indices[0],n*(n+1));
              CHECK_SAME(indices[n],n*(n+2));
              break;
          }
        }
///     cout << "\tstaticEquallySpacedLatticePointIndicesOfFace(" << f
//           << "):" << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tindices[" << i << "] = " << indices[i] << endl;
//      }
      }
      for (int f=0;f<Quadrilateral::facesPerShape();f++) {
        NumPtr<int> indices;
        Quadrilateral::staticEquallySpacedLatticePointIndicesOfFaceInterior(
          f,n,indices);
        if (n>1) {
          switch (f) {
            case 0:
              CHECK_SAME(indices[0],n+1);
              CHECK_SAME(indices[n-2],(n-1)*(n+1));
              break;
            case 1:
              CHECK_SAME(indices[0],2*n+1);
              CHECK_SAME(indices[n-2],n*(n+1)-1);
              break;
            case 2:
              CHECK_SAME(indices[0],1);
              CHECK_SAME(indices[n-2],n-1);
              break;
            case 3:
              CHECK_SAME(indices[0],n*(n+1)+1);
              CHECK_SAME(indices[n-2],n*(n+2)-1);
              break;
          }
        }
///     cout << "\tequallySpacedLatticePointIndicesOfFaceInterior(" << f
//           << "):" << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tindices[" << i << "] = " << indices[i] << endl;
//      }
      }
//    cout << "\tequallySpacedLatticePointIndicesOfFaceInterior:" << endl;
//    for (int f=0;f<Quadrilateral::facesPerShape();f++) {
//      cout << "\tf = " << f << endl;
//      NumPtr<int> indices;
//      Quadrilateral::equallySpacedLatticePointIndicesOfFaceInterior(f,n,
//        indices);
//      for (int v=0;v<n-1;v++) {
//        cout << "\t\tv = " << v << " ==> " << indices[v] << endl;
//      }
//    }
    }
#endif

#if (SPACEDIM==3)
//
    cout << "Tetrahedron:" << endl;
    for (int n=1;n<=6;n++) {
      int neslp=Tetrahedron::staticNumberEquallySpacedLatticePoints(n);
      cout << "\nn = " << n << endl;
      cout << "\tnumberEquallySpacedLatticePoints(" << n << ") = "
           << neslp << endl;
      NumPtr<Point<3> > eslp;
      Tetrahedron::staticEquallySpacedLatticePoints(n,eslp);
      for (int i=0;i<neslp;i++) {
        cout << "\teslp[" << i << "] = " << eslp[i] << endl;
      }
//    cout << "\tisParentEquallySpacedLatticePoint:" << endl;
//    for (int c=0;c<childrenPerShape();c++) {
//      for (int i=0;i<neslp;i++) {
//        int K=INT_MAX;
//        if (Tetrahedron::staticIsParentEquallySpacedLatticePoint(c,n,i,K))
//        {
//          cout << "\tc,i,K = " << c << " " << i << " " << K << endl;
//        }
//      }
//    }
      cout << "\tequallySpacedLatticePointIndexOfVertex:" << endl;
      for (int v=0;v<Tetrahedron::verticesPerShape();v++) {
        cout << "\tv = " << v << " ==> " 
          << Tetrahedron::staticEquallySpacedLatticePointIndexOfVertex(v,n)
          << endl;
      }
      for (int f=0;f<Tetrahedron::facesPerShape();f++) {
        NumPtr<int> indices;
        Tetrahedron::staticEquallySpacedLatticePointIndicesOfFace(f,n,
          indices);
//      cout << "\tstaticEquallySpacedLatticePointIndicesOfFace(" << f
//           << "):" << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tindices[" << i << "] = " << indices[i] << endl;
//      }
        if (n>2) {
          switch (f) {
            case 0:
              CHECK_SAME(indices[0],0);
              CHECK_SAME(indices[indices.getNumber()-1],(n*(11+n*(6+n)))/6);
              break;
            case 1:
              CHECK_SAME(indices[0],0);
              CHECK_SAME(indices[indices.getNumber()-1],(n*(11+n*(6+n)))/6);
              break;
            case 2:
              CHECK_SAME(indices[0],0);
              CHECK_SAME(indices[indices.getNumber()-1],(n*(n+3))/2);
              break;
            case 3:
              CHECK_SAME(indices[0],n);
              CHECK_SAME(indices[indices.getNumber()-1],(n*(11+n*(6+n)))/6);
              break;
          }
        }
      }
      for (int f=0;f<Tetrahedron::facesPerShape();f++) {
        NumPtr<int> indices;
        Tetrahedron::staticEquallySpacedLatticePointIndicesOfFaceInterior(
          f,n,indices);
//      cout << "\tequallySpacedLatticePointIndicesOfFaceInterior(" << f
//           << "):" << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tindices[" << i << "] = " << indices[i] << endl;
//      }
        if (n>2) {
          switch (f) {
            case 0:
              CHECK_SAME(indices[0],(n*(n+5))/2+1);
              CHECK_SAME(indices[indices.getNumber()-1],(n*(11+n*(6+n)))/6-6);
              break;
            case 1:
              CHECK_SAME(indices[0],(n*(n+3))/2+2);
              CHECK_SAME(indices[indices.getNumber()-1],(n*(11+n*(6+n)))/6-8);
              break;
            case 2:
              CHECK_SAME(indices[0],n+2);
              CHECK_SAME(indices[indices.getNumber()-1],(n*(n+3))/2-4);
              break;
            case 3:
              CHECK_SAME(indices[0],(n*(n+7))/2-1);
              CHECK_SAME(indices[indices.getNumber()-1],(n*(11+n*(6+n)))/6-5);
              break;
          }
        }
      }
      for (int l=0;l<Tetrahedron::linesPerShape();l++) {
        NumPtr<int> indices;
        Tetrahedron::staticEquallySpacedLatticePointIndicesOfLine(l,n,
          indices);
        switch (l) {
          case 0:
            CHECK_SAME(indices[0],0);
            CHECK_SAME(indices[n],n);
            break;
          case 1:
            CHECK_SAME(indices[0],n);
            CHECK_SAME(indices[n],(n*(n+3))/2);
            break;
          case 2:
            CHECK_SAME(indices[0],0);
            CHECK_SAME(indices[n],(n*(n+3))/2);
            break;
          case 3:
            CHECK_SAME(indices[0],0);
            CHECK_SAME(indices[n],(n*(11+n*(6+n)))/6);
            break;
          case 4:
            CHECK_SAME(indices[0],n);
            CHECK_SAME(indices[n],(n*(11+n*(6+n)))/6);
            break;
          case 5:
            CHECK_SAME(indices[0],(n*(n+3))/2);
            CHECK_SAME(indices[n],(n*(11+n*(6+n)))/6);
            break;
        }
        cout << "\tstaticEquallySpacedLatticePointIndicesOfLine(" << l
             << "):" << endl;
        for (int i=0;i<indices.getNumber();i++) {
          cout << "\tindices[" << i << "] = " << indices[i] << endl;
        }
      }
//    if (n>2) {
//      cout << "\tequallySpacedLatticePointIndexOfFace:" << endl;
//      for (int f=0;f<Tetrahedron::facesPerShape();f++) {
//        cout << "\tf = " << f << endl;
//        for (int v=0;v<((n-2)*(n-1))/2;v++) {
//          cout << "\t\tv = " << v << " ==> " 
//               << Tetrahedron::
//                  equallySpacedLatticePointIndexOfFace(f,v,n)
//               << endl;
//        }
//      }
//    }
//    if (n>1) {
//      cout << "\tequallySpacedLatticePointIndexOfLine:" << endl;
//      for (int l=0;l<Tetrahedron::linesPerShape();l++) {
//        cout << "\tl = " << l << endl;
//        for (int v=0;v<n-1;v++) {
//          cout << "\t\tv = " << v << " ==> " 
//               << Tetrahedron::
//                  equallySpacedLatticePointIndexOfLine(l,v,n)
//               << endl;
//        }
//      }
//    }
    }
//
//
    cout << "Prism:" << endl;
    for (int n=1;n<=6;n++) {
      int neslp=Prism::staticNumberEquallySpacedLatticePoints(n);
      cout << "\nn = " << n << endl;
      cout << "\tnumberEquallySpacedLatticePoints(" << n << ") = "
           << neslp << endl;
      NumPtr<Point<3> > eslp;
      Prism::staticEquallySpacedLatticePoints(n,eslp);
      for (int i=0;i<neslp;i++) {
        cout << "\teslp[" << i << "] = " << eslp[i] << endl;
      }
//    cout << "\tisParentEquallySpacedLatticePoint:" << endl;
//    for (int c=0;c<childrenPerShape();c++) {
//      for (int i=0;i<neslp;i++) {
//        int K=INT_MAX;
//        if (Prism::staticIsParentEquallySpacedLatticePoint(c,n,i,K)){
//          cout << "\tc,i,K = " << c << " " << i << " " << K << endl;
//        }
//      }
//    }
      cout << "\tequallySpacedLatticePointIndexOfVertex:" << endl;
      for (int v=0;v<Prism::verticesPerShape();v++) {
        cout << "\tv = " << v << " ==> " 
          << Prism::staticEquallySpacedLatticePointIndexOfVertex(v,n)
          << endl;
      }
      for (int f=0;f<Prism::facesPerShape();f++) {
        NumPtr<int> indices;
        Prism::staticEquallySpacedLatticePointIndicesOfFace(f,n,indices);
//      cout << "\tstaticEquallySpacedLatticePointIndicesOfFace(" << f
//           << "):" << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tindices[" << i << "] = " << indices[i] << endl;
//      }
        if (n>1) {
          switch (f) {
            case 0:
              if (n>2) {
                CHECK_SAME(indices[0],0);
                CHECK_SAME(indices[indices.getNumber()-1],(n*(n+3))/2);
              }
              break;
            case 1:
              if (n>2) {
                CHECK_SAME(indices[0],(n*(2+n*(3+n)))/2);
               CHECK_SAME(indices[indices.getNumber()-1],(n*(5+n*(4+n)))/2);
              }
              break;
            case 2:
              CHECK_SAME(indices[0],0);
              CHECK_SAME(indices[indices.getNumber()-1],(n*(5+n*(4+n)))/2);
              break;
            case 3:
              CHECK_SAME(indices[0],0);
              CHECK_SAME(indices[indices.getNumber()-1],(n*(4+n*(3+n)))/2);
              break;
            case 4:
              CHECK_SAME(indices[0],n);
              CHECK_SAME(indices[indices.getNumber()-1],(n*(5+n*(4+n)))/2);
              break;
          }
        }
      }
      for (int f=0;f<Prism::facesPerShape();f++) {
        NumPtr<int> indices;
        Prism::staticEquallySpacedLatticePointIndicesOfFaceInterior(f,n,
          indices);
//      cout << "\tequallySpacedLatticePointIndicesOfFaceInterior(" << f
//           << "):" << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tindices[" << i << "] = " << indices[i] << endl;
//      }
        if (n>1) {
          switch (f) {
            case 0:
              if (n>2) {
                CHECK_SAME(indices[0],n+2);
                CHECK_SAME(indices[indices.getNumber()-1],(n*(n+3))/2-4);
              }
              break;
            case 1:
              if (n>2) {
                CHECK_SAME(indices[0],(n*(4+n*(3+n)))/2+2);
                CHECK_SAME(indices[indices.getNumber()-1],(n*(5+n*(4+n)))/2-4);
              }
              break;
            case 2:
              CHECK_SAME(indices[0],(n*(n+5))/2+2);
              CHECK_SAME(indices[indices.getNumber()-1],(n*(2+n*(3+n)))/2-3);
              break;
            case 3:
              CHECK_SAME(indices[0],(n*(n+3))/2+2);
              CHECK_SAME(indices[indices.getNumber()-1],(n*(1+n*(2+n)))/2-2);
              break;
            case 4:
              CHECK_SAME(indices[0],(n*(n+7))/2+1);
              CHECK_SAME(indices[indices.getNumber()-1],(n*(2+n*(3+n)))/2-2);
              break;
          }
        }
      }
      for (int l=0;l<Prism::linesPerShape();l++) {
        NumPtr<int> indices;
        Prism::staticEquallySpacedLatticePointIndicesOfLine(l,n,indices);
//      cout << "\tstaticEquallySpacedLatticePointIndicesOfLine(" << l
//           << "):" << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tindices[" << i << "] = " << indices[i] << endl;
//      }
        switch (l) {
          case 0:
            CHECK_SAME(indices[0],0);
            CHECK_SAME(indices[n],n);
            break;
          case 1:
            CHECK_SAME(indices[0],n);
            CHECK_SAME(indices[n],(n*(n+3))/2);
            break;
          case 2:
            CHECK_SAME(indices[0],0);
            CHECK_SAME(indices[n],(n*(n+3))/2);
            break;
          case 3:
            CHECK_SAME(indices[0],n*((n+1)*(n+2))/2);
            CHECK_SAME(indices[n],n*((n+1)*(n+2))/2+n);
            break;
          case 4:
            CHECK_SAME(indices[0],n*((n+1)*(n+2))/2+n);
            CHECK_SAME(indices[n],(n+1)*((n+1)*(n+2))/2-1);
            break;
          case 5:
            CHECK_SAME(indices[0],n*((n+1)*(n+2))/2);
            CHECK_SAME(indices[n],(n+1)*((n+1)*(n+2))/2-1);
            break;
          case 6:
            CHECK_SAME(indices[0],0);
            CHECK_SAME(indices[n],n*((n+1)*(n+2))/2);
            break;
          case 7:
            CHECK_SAME(indices[0],n);
            CHECK_SAME(indices[n],n*((n+1)*(n+2))/2+n);
            break;
          case 8:
            CHECK_SAME(indices[0],(n*(n+3))/2);
            CHECK_SAME(indices[n],(n+1)*((n+1)*(n+2))/2-1);
            break;
        }
      }
//    cout << "\tequallySpacedLatticePointIndexOfFace:" << endl;
//    for (int f=0;f<Prism::facesPerShape();f++) {
//      cout << "\tf = " << f << endl;
//      int nv=(f<2U ? ((n-2)*(n-1))/2 : (n-1)*(n-1)); 
//      for (int v=0;v<nv;v++) {
//        cout << "\t\tv = " << v << " ==> " 
//             << Prism::equallySpacedLatticePointIndexOfFace(f,v,n)
//             << endl;
//      }
//    }
//    cout << "\tequallySpacedLatticePointIndexOfLine:" << endl;
//    for (int l=0;l<Prism::linesPerShape();l++) {
//      cout << "\tl = " << l << endl;
//      for (int v=0;v<n-1;v++) {
//        cout << "\t\tv = " << v << " ==> " 
//             << Prism::equallySpacedLatticePointIndexOfLine(l,v,n)
//             << endl;
//      }
//    }
    }
//
//
    cout << "Hexahedron:" << endl;
    for (int n=1;n<=6;n++) {
      int neslp=Hexahedron::staticNumberEquallySpacedLatticePoints(n);
      cout << "\nn = " << n << endl;
      cout << "\tnumberEquallySpacedLatticePoints(" << n << ") = "
           << neslp << endl;
      NumPtr<Point<3> > eslp;
      Hexahedron::staticEquallySpacedLatticePoints(n,eslp);
      for (int i=0;i<neslp;i++) {
        cout << "\teslp[" << i << "] = " << eslp[i] << endl;
      }
//    cout << "\tisParentEquallySpacedLatticePoint:" << endl;
//    for (int c=0;c<childrenPerShape();c++) {
//      for (int i=0;i<neslp;i++) {
//        int K=INT_MAX;
//        if (Hexahedron::
//        staticIsParentEquallySpacedLatticePoint(c,n,i,K)) {
//          cout << "\tc,i,K = " << c << " " << i << " " << K << endl;
//        }
//      }
//    }
      cout << "\tequallySpacedLatticePointIndexOfVertex:" << endl;
      for (int v=0;v<Hexahedron::verticesPerShape();v++) {
        cout << "\tv = " << v << " ==> " << Hexahedron::
          staticEquallySpacedLatticePointIndexOfVertex(v,n)
          << endl;
      }
      for (int f=0;f<Hexahedron::facesPerShape();f++) {
        NumPtr<int> indices;
        Hexahedron::staticEquallySpacedLatticePointIndicesOfFace(f,n,
          indices);
//      cout << "\tstaticEquallySpacedLatticePointIndicesOfFace(" << f
//           << "):" << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tindices[" << i << "] = " << indices[i] << endl;
//      }
        if (n>1) {
          switch (f) {
            case 0:
              CHECK_SAME(indices[0],0);
              CHECK_SAME(indices[indices.getNumber()-1],n*(n+1)*(n+2));
              break;
            case 1:
              CHECK_SAME(indices[0],n);
              CHECK_SAME(indices[indices.getNumber()-1],n*(3+n*(3+n)));
              break;
            case 2:
              CHECK_SAME(indices[0],0);
              CHECK_SAME(indices[indices.getNumber()-1],n*(2+n*(2+n)));
              break;
            case 3:
              CHECK_SAME(indices[0],n*(n+1));
              CHECK_SAME(indices[indices.getNumber()-1],n*(3+n*(3+n)));
              break;
            case 4:
              CHECK_SAME(indices[0],0);
              CHECK_SAME(indices[indices.getNumber()-1],n*(2+n));
              break;
            case 5:
              CHECK_SAME(indices[0],n*(n+1)*(n+1));
              CHECK_SAME(indices[indices.getNumber()-1],n*(3+n*(3+n)));
              break;
          }
        }
      }
      for (int f=0;f<Hexahedron::facesPerShape();f++) {
        NumPtr<int> indices;
        Hexahedron::staticEquallySpacedLatticePointIndicesOfFaceInterior(
          f,n,indices);
//      cout << "\tequallySpacedLatticePointIndicesOfFaceInterior(" << f
//           << "):" << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tindices[" << i << "] = " << indices[i] << endl;
//      }
        if (n>1) {
          switch (f) {
            case 0:
              CHECK_SAME(indices[0],n*(3+n)+2);
              CHECK_SAME(indices[indices.getNumber()-1],n*(-1+n*(2+n))-2);
              break;
            case 1:
              CHECK_SAME(indices[0],n*(4+n)+2);
              CHECK_SAME(indices[indices.getNumber()-1],n*n*(2+n)-2);
              break;
            case 2:
              CHECK_SAME(indices[0],n*(2+n)+2);
              CHECK_SAME(indices[indices.getNumber()-1],n*n*(1+n)-2);
              break;
            case 3:
              CHECK_SAME(indices[0],n*(3+2*n)+2);
              CHECK_SAME(indices[indices.getNumber()-1],n*(1+n*(2+n))-2);
              break;
            case 4:
              CHECK_SAME(indices[0],n+2);
              CHECK_SAME(indices[indices.getNumber()-1],n*(1+n)-2);
              break;
            case 5:
              CHECK_SAME(indices[0],n*(2+n*(2+n))+2);
              CHECK_SAME(indices[indices.getNumber()-1],n*(2+n*(3+n))-2);
              break;
          }
        }
      }
      for (int l=0;l<Hexahedron::linesPerShape();l++) {
        NumPtr<int> indices;
        Hexahedron::staticEquallySpacedLatticePointIndicesOfLine(l,n,
          indices);
//      cout << "\tstaticEquallySpacedLatticePointIndicesOfLine(" << l
//           << "):" << endl;
//      for (int i=0;i<indices.getNumber();i++) {
//        cout << "\tindices[" << i << "] = " << indices[i] << endl;
//      }
        switch (l) {
          case 0:
            CHECK_SAME(indices[0],0);
            CHECK_SAME(indices[n],n*(n+1));
            break;
          case 1:
            CHECK_SAME(indices[0],n);
            CHECK_SAME(indices[n],n*(n+2));
            break;
          case 2:
            CHECK_SAME(indices[0],0);
            CHECK_SAME(indices[n],n);
            break;
          case 3:
            CHECK_SAME(indices[0],n*(n+1));
            CHECK_SAME(indices[n],n*(n+2));
            break;
          case 4:
            CHECK_SAME(indices[0],n*(n+1)*(n+1));
            CHECK_SAME(indices[n],n*(n+1)*(n+2));
            break;
          case 5:
            CHECK_SAME(indices[0],n*(n+1)*(n+1)+n);
            CHECK_SAME(indices[n],n*(3+n*(3+n)));
            break;
          case 6:
            CHECK_SAME(indices[0],n*(n+1)*(n+1));
            CHECK_SAME(indices[n],n*(n+1)*(n+1)+n);
            break;
          case 7:
            CHECK_SAME(indices[0],n*(n+1)*(n+2));
            CHECK_SAME(indices[n],n*(3+n*(3+n)));
            break;
          case 8:
            CHECK_SAME(indices[0],0);
            CHECK_SAME(indices[n],n*(n+1)*(n+1));
            break;
          case 9:
            CHECK_SAME(indices[0],n);
            CHECK_SAME(indices[n],n*(n+1)*(n+1)+n);
            break;
          case 10:
            CHECK_SAME(indices[0],n*(n+1));
            CHECK_SAME(indices[n],n*(n+1)*(n+2));
            break;
          case 11:
            CHECK_SAME(indices[0],n*(n+2));
            CHECK_SAME(indices[n],n*(3+n*(3+n)));
            break;
        }
      }
//    cout << "\tequallySpacedLatticePointIndexOfFace:" << endl;
//    for (int f=0;f<Hexahedron::facesPerShape();f++) {
//      cout << "\tf = " << f << endl;
//      for (int v=0;v<(n-1)*(n-1);v++) {
//        cout << "\t\tv = " << v << " ==> " << Hexahedron::
//          staticEquallySpacedLatticePointIndexOfFace(f,v,n)
//          << endl;
//      }
//    }
//    cout << "\tequallySpacedLatticePointIndexOfLine:" << endl;
//    for (int l=0;l<Hexahedron::linesPerShape();l++) {
//      cout << "\tl = " << l << endl;
//      for (int v=0;v<n-1;v++) {
//        cout << "\t\tv = " << v << " ==> " << Hexahedron::
//          staticEquallySpacedLatticePointIndexOfLine(l,v,n)
//          << endl;
//      }
//    }
    }
//
#endif

  }
  return EXIT_SUCCESS;
}
