// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "MemoryDebugger.H"
#include "PermutationGenerator.H"
#include "Shape.H"
#include "Tracer.H"

void checkPermutation(NumPtr<int> &renumber) {
  NumPtr<int> renumber_inverse(renumber.getNumber());
  renumber_inverse.initialize(-1);
  for (int i=0;i<renumber.getNumber();i++) {
    renumber_inverse[renumber[i]]=i;
  }
  for (int i=0;i<renumber_inverse.getNumber();i++) {
    CHECK_TEST(renumber_inverse[i]!=-1);
  }
}
double factorial(int n) {
  if (n<=1) return 1.;
  return static_cast<double>(n)*factorial(n-1);
}

int main(int argc,char* argv[]) {
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  for (int nv=2;nv<=8;nv++) {
//    PermutationGenerator pg(nv);
//    while (pg.hasMore()) {
//      const NumPtr<int> &p=pg.getNext();
//      for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//      cout << endl;
//    }
//    cout << endl;
//  }

    int edi=0;
#if (SPACEDIM==1)
    { // Interval:
      int nv=2;
      cout << "Interval:" << endl;
      for (int n=1;n<=6;n++) {
        PermutationGenerator pg(nv);
        cout << "\tn = " << n << endl;
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          int v[2];
          for (int i=0;i<nv;i++) v[i]=p[i];
          Interval e(edi,v);
          {
            NumPtr<int> renumber(n+1);
            e.reorderDofs(n,renumber);
            checkPermutation(renumber);
          }
          {
            NumPtr<int> renumber;
            e.reorderEquallySpacedLatticePoints(n,renumber);
            checkPermutation(renumber);
          }
/*
          {
            NumPtr<int> renumber;
            e.reorderDoFPoints(n,renumber);
#ifdef DEBUG
//          cout << "\trenumber = ";
//          for (int i=0;i<renumber.getNumber();i++) {
//            cout << " " << renumber[i];
//          }
//          cout << endl;
#endif
          }
*/
        }
        int pv[2]={0,1};
        Interval parent(0,pv);
        int cv0[2]={0,2};
        Interval child0(1,cv0);
        int cv1[2]={2,1};
        Interval child1(2,cv1);
        NumPtr<int> line_interior_child_vertices;
        {
          NumPtr<int> renumber;
          parent.reorderChildEquallySpacedLatticePoints(n,&child0,0,
            line_interior_child_vertices,renumber);
          checkPermutation(renumber);
        }
        {
          NumPtr<int> renumber;
          parent.reorderChildEquallySpacedLatticePoints(n,&child1,1,
            line_interior_child_vertices,renumber);
          checkPermutation(renumber);
        }
      }
    }
#endif

#if (SPACEDIM==2)
    { // Triangle:  
      int nv=3;
      cout << "Triangle:" << endl;
      for (int n=1;n<=6;n++) {
        int l[3]={0,1,2};
        PermutationGenerator pg(nv);
        cout << "\tn = " << n << endl;
        int pv[3]={0,1,2};
        Triangle parent(0,pv,pv);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          int v[3];
          for (int i=0;i<nv;i++) v[i]=p[i];
          Triangle e(edi,l,v);
          {
            NumPtr<int> renumber(3*(n+1));
            e.reorderFaceDofs(n,renumber);
            checkPermutation(renumber);
          }
          {
            NumPtr<int> renumber(((n+2)*(n+1))/2);
            e.reorderDofs(n,renumber);
            checkPermutation(renumber);
          }
          {
            NumPtr<int> renumber;
            e.reorderEquallySpacedLatticePoints(n,renumber);
            checkPermutation(renumber);
          }
//        {
//          NumPtr<int> renumber;
//          e.reorderEquallySpacedLatticePoints2(n,renumber);
//          checkPermutation(renumber);
//        }
/*
          {
            NumPtr<int> renumber;
            e.reorderDoFPoints(n,renumber);
          }
*/
          NumPtr<int> p2(p.getNumber());
          p2.copyFrom(p);
          for (int i=0;i<p2.getNumber();i++) p2[i]+=3;
          NumPtr<int> 
            line_interior_child_vertices(Triangle::linesPerShape());
          line_interior_child_vertices[0]=p2[0];
          line_interior_child_vertices[1]=p2[1];
          line_interior_child_vertices[2]=p2[2];
          {
            int cv[3]={0,p2[1],p2[0]};
            Triangle child(1,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,0,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[3]={1,p2[2],p2[1]};
            Triangle child(1,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,1,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[3]={2,p2[0],p2[2]};
            Triangle child(1,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,2,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
        }
      }
    }
    { // Quadrilateral:
      int nv=4;
      cout << "Quadrilateral:" << endl;
      for (int n=1;n<=6;n++) {
        int l[4]={0,1,2,3};
        PermutationGenerator pg(nv);
        cout << "\tn = " << n << endl;
        int pv[4]={0,1,2,3};
        Quadrilateral parent(0,pv,pv);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          int v[4];
          for (int i=0;i<nv;i++) v[i]=p[i];
          Quadrilateral e(edi,l,v);
          {
            NumPtr<int> renumber(4*(n+1));
            e.reorderFaceDofs(n,renumber);
            checkPermutation(renumber);
          }
          {
            NumPtr<int> renumber((n+1)*(n+1));
            e.reorderDofs(n,renumber);
            checkPermutation(renumber);
          }
          {
            NumPtr<int> renumber;
            e.reorderEquallySpacedLatticePoints(n,renumber);
          }
/*
          {
            NumPtr<int> renumber;
            e.reorderDoFPoints(n,renumber);
          }
*/
        }
        PermutationGenerator pg2(5);
        while (pg2.hasMore()) {
          const NumPtr<int> &p=pg2.getNext();
          NumPtr<int> p2(p.getNumber());
          p2.copyFrom(p);
          for (int i=0;i<p2.getNumber();i++) p2[i]+=4;
          NumPtr<int> 
            line_interior_child_vertices(Quadrilateral::linesPerShape());
          line_interior_child_vertices[0]=p2[0];
          line_interior_child_vertices[1]=p2[1];
          line_interior_child_vertices[2]=p2[2];
          line_interior_child_vertices[3]=p2[3];
          {
            int cv[4]={0,p2[2],p2[0],p2[4]};
            Quadrilateral child(1,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,0,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[4]={1,p2[1],p2[2],p2[4]};
            Quadrilateral child(1,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,1,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[4]={2,p2[0],p2[3],p2[4]};
            Quadrilateral child(1,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,2,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[4]={3,p2[3],p2[1],p2[4]};
            Quadrilateral child(1,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,3,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
        }
      }
      for (int n1=1;n1<=6;n1++) {
        for (int n0=1;n0<=6;n0++) {
          int l[4]={0,1,2,3};
          PermutationGenerator pg(nv);
          cout << "\tn = " << n0 << " " << n1 << endl;
          int pv[4]={0,1,2,3};
          Quadrilateral parent(0,pv,pv);
          while (pg.hasMore()) {
            const NumPtr<int> &p=pg.getNext();
            int v[4];
            for (int i=0;i<nv;i++) v[i]=p[i];
            Quadrilateral e(edi,l,v);
            {
              NumPtr<int> renumber;
              e.reorderEquallySpacedLatticePoints(n0,n1,renumber);
            }
/*
            {
              NumPtr<int> renumber;
              e.reorderDoFPoints(n0,n1,renumber);
          }
*/
          }
/*
          PermutationGenerator pg2(5);
          while (pg2.hasMore()) {
            const NumPtr<int> &p=pg2.getNext();
            NumPtr<int> p2(p.getNumber());
            p2.copyFrom(p);
            for (int i=0;i<p2.getNumber();i++) p2[i]+=4;
            NumPtr<int> 
              line_interior_child_vertices(Quadrilateral::linesPerShape());
            line_interior_child_vertices[0]=p2[0];
            line_interior_child_vertices[1]=p2[1];
            line_interior_child_vertices[2]=p2[2];
            line_interior_child_vertices[3]=p2[3];
            {
              int cv[4]={0,p2[2],p2[0],p2[4]};
              Quadrilateral child(1,cv,cv);
              NumPtr<int> renumber;
              parent.reorderChildEquallySpacedLatticePoints(n0,n1,&child,0,
                line_interior_child_vertices,renumber);
              checkPermutation(renumber);
            }
            {
              int cv[4]={1,p2[1],p2[2],p2[4]};
              Quadrilateral child(1,cv,cv);
              NumPtr<int> renumber;
              parent.reorderChildEquallySpacedLatticePoints(n0,n1,&child,1,
                line_interior_child_vertices,renumber);
              checkPermutation(renumber);
            }
            {
              int cv[4]={2,p2[0],p2[3],p2[4]};
              Quadrilateral child(1,cv,cv);
              NumPtr<int> renumber;
              parent.reorderChildEquallySpacedLatticePoints(n0,n1,&child,2,
                line_interior_child_vertices,renumber);
              checkPermutation(renumber);
            }
            {
              int cv[4]={3,p2[3],p2[1],p2[4]};
              Quadrilateral child(1,cv,cv);
              NumPtr<int> renumber;
              parent.reorderChildEquallySpacedLatticePoints(n0,n1,&child,3,
                line_interior_child_vertices,renumber);
              checkPermutation(renumber);
            }
          }
*/
        }
      }
    }
#endif

#if (SPACEDIM==3)
//
    REAL eps=sqrt(DBL_EPSILON);
    { // Tetrahedron:
      int nv=4;
      cout << "Tetrahedron:" << endl;
      for (int n=1;n<=6;n++)
//    int n=3;
      {
        int l[6]={0,1,2,3,4,5};
        int f[4]={0,1,2,3};
        PermutationGenerator pg(nv);
        cout << "\tn = " << n << endl;
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          int v[4];
          for (int i=0;i<nv;i++) v[i]=p[i];
          Tetrahedron e(edi,l,f,v);
          {
            NumPtr<int> renumber(2*(n+2)*(n+1));
            e.reorderHdivFaceDofs(0,n,renumber);
            checkPermutation(renumber);
          }
          {
            NumPtr<int> renumber(4*(n+2)*(n+1));
            e.reorderHcurlFaceDofs(0,n,renumber);
            checkPermutation(renumber);
          }
          {
            NumPtr<int> renumber(4*((n+2)*(n+1)+(n+1)));
            e.reorderHcurlFaceDofs2(0,n,renumber);
            checkPermutation(renumber);
          }
          {
            NumPtr<int> renumber;
            e.reorderEquallySpacedLatticePoints(n,renumber);
            checkPermutation(renumber);
          }
          for (int f=0;f<4;f++) {
            Point<2> p; p[0]=drand48(); p[1]=drand48();
            Point<2> hp;
            Tensor<2,2> dhpdp;
            e.homogeneousCoordinate(f,p,hp);
            e.homogeneousCoordinateDerivative(f,p,dhpdp);
            for (int d=0;d<2;d++) {
              Point<2> pp(p); pp[d]+=eps;
              Point<2> hpp;
              e.homogeneousCoordinate(f,pp,hpp);
              for (int i=0;i<2;i++) {
                REAL exact=dhpdp[i][d];
                REAL diff=(hpp[i]-hp[i])/eps;
                REAL error=abs(diff-exact);
                if (error>eps*16.) {
                  cout << "\thomogeneousCoordinateDerivative(" << f << ","
                       << p << ")[" << i << "][" << d << "] = "
                       << exact << " " << diff << ", error = "
                       << error <<  endl;
                  CHECK_TEST(error<=eps*16.);
                }
              }
            }
          }
//        {
//          NumPtr<int> renumber;
//          e.reorderEquallySpacedLatticePoints2(n,renumber);
//          checkPermutation(renumber);
//        }
//        {
//          NumPtr<int> renumber;
//          e.reorderDoFPoints(n,renumber);
//        }
        }
        PermutationGenerator pg2(Tetrahedron::linesPerShape());
        int pv[4]={0,1,2,3};
        Tetrahedron parent(0,l,pv,pv);
        while (pg2.hasMore()) {
          const NumPtr<int> &p=pg2.getNext();
          NumPtr<int> p2(p.getNumber());
          p2.copyFrom(p);
          for (int i=0;i<p2.getNumber();i++) p2[i]+=4;
          NumPtr<int>
            line_interior_child_vertices(Tetrahedron::linesPerShape());
          for (int i=0;i<line_interior_child_vertices.getNumber();i++) {
            line_interior_child_vertices[i]=p2[i];
          }
          {
            int cv[4]={0,p2[0],p2[2],p2[3]};
            Tetrahedron child(1,l,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,0,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[4]={p2[0],1,p2[1],p2[4]};
            Tetrahedron child(1,l,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,1,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[4]={p2[2],p2[1],2,p2[5]};
            Tetrahedron child(1,l,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,2,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[4]={p2[3],p2[4],p2[5],3};
            Tetrahedron child(1,l,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,3,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[4]={p2[4],p2[3],p2[5],p2[0]};
            Tetrahedron child(1,l,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,4,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[4]={p2[4],p2[5],p2[1],p2[0]};
            Tetrahedron child(1,l,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,5,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[4]={p2[2],p2[5],p2[0],p2[1]};
            Tetrahedron child(1,l,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,6,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[4]={p2[5],p2[2],p2[0],p2[3]};
            Tetrahedron child(1,l,cv,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,7,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
        }
      }
    }
//
//
    { // Prism:
      int nv=6;
      cout << "Prism:" << endl;
      for (int n=1;n<=6;n++) {
        int l[9]={0,1,2,3,4,5,6,7,8};
        int f[5]={0,1,2,3,4};
        PermutationGenerator pg(nv);
        cout << "\tn = " << n << endl;
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          int v[6];
          for (int i=0;i<nv;i++) v[i]=p[i];
          Prism e(edi,l,f,v);
//        {
//          NumPtr<int> renumber((n+2)*(n+1)+3*(n+1)*(n+1));
//          e.reorderHdivFaceDofs(0,n,renumber);
//          checkPermutation(renumber);
//        }
          {
            NumPtr<int> renumber(8*(n+1)*n);
            e.reorderHcurlFaceDofs(0,n,renumber);
            checkPermutation(renumber);
          }
          {
            NumPtr<int> renumber(8*(n+1)*(n-1));
            e.reorderHcurlFaceDofs2(0,n,renumber);
            checkPermutation(renumber);
          }
          {
            NumPtr<int> renumber;
            e.reorderEquallySpacedLatticePoints(n,renumber);
          }
          for (int f=0;f<2;f++) {
            Point<2> p; p[0]=drand48(); p[1]=drand48();
            Point<2> hp;
            Tensor<2,2> dhpdp;
            e.homogeneousCoordinate(f,p,hp);
            e.homogeneousCoordinateDerivative(f,p,dhpdp);
            for (int d=0;d<2;d++) {
              Point<2> pp(p); pp[d]+=eps;
              Point<2> hpp;
              e.homogeneousCoordinate(f,pp,hpp);
              for (int i=0;i<2;i++) {
                REAL exact=dhpdp[i][d];
                REAL diff=(hpp[i]-hp[i])/eps;
                REAL error=abs(diff-exact);
                if (error>eps*16.) {
                  cout << "\thomogeneousCoordinateDerivative(" << f << ","
                       << p << ")[" << i << "][" << d << "] = "
                       << exact << " " << diff << ", error = "
                       << error <<  endl;
                  CHECK_TEST(error<=eps*16.);
                }
              }
            }
          }
//        {
//          NumPtr<int> renumber;
//          e.reorderDoFPoints(n,renumber);
//        }
        }
        int lps=Prism::linesPerShape();
        int pv[6]={0,1,2,3,4,5};
        Prism parent(0,l,f,pv);
        for (int count=0;count<1000;count++) {
          NumPtr<int> p(lps);
          RandomPermutation(p);

//        cout << "\tp[" << count << "] = ";
//        for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//        cout << endl;

          NumPtr<int> p2(p.getNumber());
          p2.copyFrom(p);
          for (int i=0;i<p2.getNumber();i++) p2[i]+=6;
          NumPtr<int>
            line_interior_child_vertices(Prism::linesPerShape());
          for (int i=0;i<line_interior_child_vertices.getNumber();i++) {
            line_interior_child_vertices[i]=p2[i];
          }
          {
            int cv[6]={0,p2[0],p2[2],p2[6],16,15};
            Prism child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,0,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[6]={p2[0],1,p2[1],16,p2[7],17};
            Prism child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,1,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[6]={p2[2],p2[1],2,15,17,p2[8]};
            Prism child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,2,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[6]={17,16,15,p2[1],p2[0],p2[2]};
            Prism child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,3,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[6]={p2[6],16,15,3,p2[3],p2[5]};
            Prism child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,4,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[6]={16,p2[7],17,p2[3],4,p2[4]};
            Prism child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,5,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[6]={15,17,p2[8],p2[5],p2[4],5};
            Prism child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,6,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[6]={p2[4],p2[3],p2[5],17,16,15};
            Prism child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,7,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
        }
      }
    }
//
    { // Hexahedron:
      int nv=8;
      cout << "Hexahedron:" << endl;
      for (int n=1;n<=6;n++) 
//    int n=2;
      {
        cerr << "\tn = " << n << endl;
        int l[12]={0,1,2,3,4,5,6,7,8,9,10,11};
        int f[6]={0,1,2,3,4,5};
        PermutationGenerator pg(nv);
        while (pg.hasMore()) {
          const NumPtr<int> &p=pg.getNext();
          int v[8];
          for (int i=0;i<nv;i++) v[i]=p[i];
#ifdef DEBUG
//        cout << "\tv =";
//        for (int i=0;i<nv;i++) cout << " " << v[i];
//        cout << endl;
#endif
          Hexahedron e(edi,l,f,v);
          {
            NumPtr<int> renumber(6*(n+1)*(n+1));
            e.reorderHdivFaceDofs(0,n,renumber);
            checkPermutation(renumber);
          }
          {
            NumPtr<int> renumber(12*(n+1)*(n+1));
            e.reorderHcurlFaceDofs(0,n,n,renumber);
            checkPermutation(renumber);
          }
          {
            NumPtr<int> renumber;
            e.reorderEquallySpacedLatticePoints(n,renumber);
          }
//        {
//          NumPtr<int> renumber;
//          e.reorderDoFPoints(n,renumber);
//        }
        }
        int lps=Hexahedron::linesPerShape();
        int pv[8]={0,1,2,3,4,5,6,7};
        Hexahedron parent(0,l,f,pv);
        for (int count=0;count<1000;count++) {
          NumPtr<int> p(lps);
          RandomPermutation(p);

//        cout << "\tp[" << count << "] = ";
//        for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//        cout << endl;

          NumPtr<int> p2(p.getNumber());
          p2.copyFrom(p);
          for (int i=0;i<p2.getNumber();i++) p2[i]+=8;
          NumPtr<int>
            line_interior_child_vertices(lps);
          for (int i=0;i<line_interior_child_vertices.getNumber();i++) {
            line_interior_child_vertices[i]=p2[i];
          }
          {
            int cv[8]={0,p2[2],p2[0],24,p2[8],22,20,26};
            Hexahedron child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,0,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[8]={1,p2[1],p2[2],24,p2[9],21,22,26};
            Hexahedron child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,1,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[8]={2,p2[0],p2[3],24,p2[10],20,23,26};
            Hexahedron child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,2,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[8]={3,p2[3],p2[1],24,p2[11],23,21,26};
            Hexahedron child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,3,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[8]={4,p2[4],p2[6],25,p2[10],20,22,26};
            Hexahedron child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,4,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[8]={5,p2[6],p2[5],25,p2[9],22,21,26};
            Hexahedron child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,5,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[8]={6,p2[7],p2[4],25,p2[2],23,20,26};
            Hexahedron child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,6,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
          {
            int cv[8]={7,p2[5],p2[7],25,p2[11],21,23,26};
            Hexahedron child(1,l,f,cv);
            NumPtr<int> renumber;
            parent.reorderChildEquallySpacedLatticePoints(n,&child,7,
              line_interior_child_vertices,renumber);
            checkPermutation(renumber);
          }
        }
      }
    }
    { // Hexahedron:
      int nv=8;
      cout << "Hexahedron:" << endl;
      for (int n2=1;n2<=6;n2++) {
        for (int n1=1;n1<=6;n1++) {
          for (int n0=1;n0<=6;n0++) {
            cerr << "\tn = " << n0 << " " << n1 << " " << n2 << endl;
            int l[12]={0,1,2,3,4,5,6,7,8,9,10,11};
            int f[6]={0,1,2,3,4,5};
            PermutationGenerator pg(nv);
            while (pg.hasMore()) {
              const NumPtr<int> &p=pg.getNext();
              int v[8];
              for (int i=0;i<nv;i++) v[i]=p[i];
              Hexahedron e(edi,l,f,v);
              {
                NumPtr<int> renumber;
                e.reorderEquallySpacedLatticePoints(n0,n1,n2,renumber);
              }
//            {
//              NumPtr<int> renumber;
//              e.reorderDoFPoints(n0,n1,n2,renumber);
//            }
            }
/*
            int lps=Hexahedron::linesPerShape();
            int pv[8]={0,1,2,3,4,5,6,7};
            Hexahedron parent(0,l,f,pv);
            for (int count=0;count<1000;count++) {
              NumPtr<int> p(lps);
              RandomPermutation(p);

//            cout << "\tp[" << count << "] = ";
//            for (int i=0;i<p.getNumber();i++) cout << p[i] << " ";
//            cout << endl;

              NumPtr<int> p2(p.getNumber());
              p2.copyFrom(p);
              for (int i=0;i<p2.getNumber();i++) p2[i]+=8;
              NumPtr<int> line_interior_child_vertices(lps);
              for (int i=0;i<line_interior_child_vertices.getNumber();i++) {
                line_interior_child_vertices[i]=p2[i];
              }
              {
                int cv[8]={0,p2[2],p2[0],24,p2[8],22,20,26};
                Hexahedron child(1,l,f,cv);
                NumPtr<int> renumber;
                parent.reorderChildEquallySpacedLatticePoints(n0,n1,n2,
                  &child,0,line_interior_child_vertices,renumber);
                checkPermutation(renumber);
              }
              {
                int cv[8]={1,p2[1],p2[2],24,p2[9],21,22,26};
                Hexahedron child(1,l,f,cv);
                NumPtr<int> renumber;
                parent.reorderChildEquallySpacedLatticePoints(n0,n1,n2,
                  &child,1,line_interior_child_vertices,renumber);
                checkPermutation(renumber);
              }
              {
                int cv[8]={2,p2[0],p2[3],24,p2[10],20,23,26};
                Hexahedron child(1,l,f,cv);
                NumPtr<int> renumber;
                parent.reorderChildEquallySpacedLatticePoints(n0,n1,n2,
                  &child,2,line_interior_child_vertices,renumber);
                checkPermutation(renumber);
              }
              {
                int cv[8]={3,p2[3],p2[1],24,p2[11],23,21,26};
                Hexahedron child(1,l,f,cv);
                NumPtr<int> renumber;
                parent.reorderChildEquallySpacedLatticePoints(n0,n1,n2,
                  &child,3,line_interior_child_vertices,renumber);
                checkPermutation(renumber);
              }
              {
                int cv[8]={4,p2[4],p2[6],25,p2[10],20,22,26};
                Hexahedron child(1,l,f,cv);
                NumPtr<int> renumber;
                parent.reorderChildEquallySpacedLatticePoints(n0,n1,n2,
                  &child,4,line_interior_child_vertices,renumber);
                checkPermutation(renumber);
              }
              {
                int cv[8]={5,p2[6],p2[5],25,p2[9],22,21,26};
                Hexahedron child(1,l,f,cv);
                NumPtr<int> renumber;
                parent.reorderChildEquallySpacedLatticePoints(n0,n1,n2,
                  &child,5,line_interior_child_vertices,renumber);
                checkPermutation(renumber);
              }
              {
                int cv[8]={6,p2[7],p2[4],25,p2[2],23,20,26};
                Hexahedron child(1,l,f,cv);
                NumPtr<int> renumber;
                parent.reorderChildEquallySpacedLatticePoints(n0,n1,n2,
                  &child,6,line_interior_child_vertices,renumber);
                checkPermutation(renumber);
              }
              {
                int cv[8]={7,p2[5],p2[7],25,p2[11],21,23,26};
                Hexahedron child(1,l,f,cv);
                NumPtr<int> renumber;
                parent.reorderChildEquallySpacedLatticePoints(n0,n1,n2,
                  &child,7,line_interior_child_vertices,renumber);
                checkPermutation(renumber);
              }
            }
*/
          }
        }
      }
    }
#endif

  }
  return EXIT_SUCCESS;
}
