// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <float.h>
#include <stdlib.h>
#include "Debug.H"
#include "MemoryDebugger.H"
#include "Tensor.H"
#include "Tracer.H"

void rank2Case() {
//TRACER_CALL(z,"rank2Case");
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    TRACER_CALL(z,"constructors");
    Tensor<2,SPACEDIM> t0;
    cout << "Tensor<2," << SPACEDIM << "> = " << t0 
         << endl;

    REAL d[SPACEDIM][SPACEDIM];
    for (int j=0;j<SPACEDIM;j++) {
      for (int i=0;i<SPACEDIM;i++) d[j][i]=i+j;
    }
    Tensor<2,SPACEDIM> t2(d);
    cout << "Tensor<2," << SPACEDIM 
         << ">(REAL[][]) = " << t2 << endl;
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"initializations");
    Tensor<2,SPACEDIM> t1;
    t1.initialize(-1.);
    for (int j=0;j<SPACEDIM;j++) {
      for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t1[j][i],-1.);
    }

    int n=SPACEDIM*SPACEDIM;
    LAVector v(n);
    t1.unroll(v);
    for (int i=0;i<n;i++) CHECK_SAME(v[i],-1.);
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"operators");
    Tensor<1,SPACEDIM> a1,b1,s1,d1,p1,q1;
    for (int i=0;i<SPACEDIM;i++) {
      a1[i]=i+1;
      b1[i]=4-i;
    }
    s1=a1+b1;
    d1=b1-a1;
    product(p1,-a1,5.);
    quotient(q1,s1,2.);
    Tensor<2,SPACEDIM> a2,b2,s2,d2,p2,q2;
    for (int i=0;i<SPACEDIM;i++) {
      a2[i]=a1;
      b2[i]=b1;
      s2[i]=s1;
      d2[i]=d1;
      p2[i]=p1;
      q2[i]=q1;
    }

    Tensor<2,SPACEDIM> t1,t2;
    t1=a2;
    t2=b2;
    CHECK_TEST(!(t1==t2));
    CHECK_TEST(t1!=t2);

    t1+=t2;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t1[i],s1);

    t2-=t1;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t2[i],-a1);

    t2*=5.;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t2[i],p1);

    t1/=2.;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t1[i],q1);

    t1=a2;
    t2=b2;

    Tensor<2,SPACEDIM> t=t1+t2;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t[i],s1);

    Tensor<2,SPACEDIM> tt=t2-t1;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(tt[i],d1);

    Tensor<2,SPACEDIM> ttt=-t2;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(ttt[i],-b1);
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"products and quotients");
    Tensor<1,SPACEDIM> a1,b1,s1,d1,p1,q1;
    for (int i=0;i<SPACEDIM;i++) {
      a1[i]=i+1;
      b1[i]=4-i;
      p1[i]=5*i+5;
      q1[i]=2.-0.5*static_cast<REAL>(i);
    }
    s1=a1+b1;
    d1=b1-s1;
    Tensor<2,SPACEDIM> a2,b2,s2,d2,p2,q2;
    for (int i=0;i<SPACEDIM;i++) {
      a2[i]=a1;
      b2[i]=b1;
      s2[i]=s1;
      d2[i]=d1;
      p2[i]=p1;
      q2[i]=q1;
    }

    Tensor<2,SPACEDIM> t;
    product(t,a2,5.);
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t[i],p1);
    product(t,5.,a2);
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t[i],p1);

    quotient(t,b2,2.);
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t[i],q1);
  }
}
void rank3Case() {
//TRACER_CALL(z,"rank3Case");
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    TRACER_CALL(z,"constructors");
    Tensor<3,SPACEDIM> t0;
    cout << "Tensor<3," << SPACEDIM << "> = " << t0 << endl;

    REAL d[SPACEDIM][SPACEDIM][SPACEDIM];
    for (int k=0;k<SPACEDIM;k++) {
      for (int j=0;j<SPACEDIM;j++) {
        for (int i=0;i<SPACEDIM;i++) {
          d[k][j][i]=i+j+k;
        }
      }
    }
    Tensor<3,SPACEDIM> t3(d);
    cout << "Tensor<3," << SPACEDIM << ">(REAL[][][]) = " << t3 << endl;
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"initializations");
    Tensor<3,SPACEDIM> t1;
    t1.initialize(-1.);
    for (int k=0;k<SPACEDIM;k++) {
      for (int j=0;j<SPACEDIM;j++) {
        for (int i=0;i<SPACEDIM;i++) {
          CHECK_SAME(t1[k][j][i],-1.);
        }
      }
    }

    int n=SPACEDIM*SPACEDIM*SPACEDIM;
    LAVector v(n);
    t1.unroll(v);
    for (int i=0;i<n;i++) CHECK_SAME(v[i],-1.);
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"operators");
    Tensor<1,SPACEDIM> a1,b1,s1,d1,p1,q1;
    for (int i=0;i<SPACEDIM;i++) {
      a1[i]=i+1;
      b1[i]=4-i;
    }
    s1=a1+b1;
    d1=b1-a1;
    product(p1,a1,-5.);
    quotient(q1,s1,2.);
    Tensor<2,SPACEDIM> a2,b2,s2,d2,p2,q2;
    for (int i=0;i<SPACEDIM;i++) {
      a2[i]=a1;
      b2[i]=b1;
      s2[i]=s1;
      d2[i]=d1;
      p2[i]=p1;
      q2[i]=q1;
    }
    Tensor<3,SPACEDIM> a3,b3,s3,d3,p3,q3;
    for (int i=0;i<SPACEDIM;i++) {
      a3[i]=a2;
      b3[i]=b2;
      s3[i]=s2;
      d3[i]=d2;
      p3[i]=p2;
      q3[i]=q2;
    }
    Tensor<4,SPACEDIM> a4,b4,s4,d4,p4,q4;
    for (int i=0;i<SPACEDIM;i++) {
      a4[i]=a3;
      b4[i]=b3;
      s4[i]=s3;
      d4[i]=d3;
      p4[i]=p3;
      q4[i]=q3;
    }

    Tensor<3,SPACEDIM> t1,t2;
    t1=a3;
    t2=b3;
    CHECK_TEST(!(t1==t2));
    CHECK_TEST(t1!=t2);

    t1+=t2;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t1[i],s2);

    t2-=t1;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t2[i],-a2);

    t2*=5.;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t2[i],p2);

    t1/=2.;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t1[i],q2);

    t1=a3;
    t2=b3;

    Tensor<3,SPACEDIM> t=t1+t2;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t[i],s2);

    Tensor<3,SPACEDIM> tt=t2-t1;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(tt[i],d2);

    Tensor<3,SPACEDIM> ttt=-t2;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(ttt[i],-b2);
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"products and quotients");
    Tensor<1,SPACEDIM> a1,b1,s1,d1,p1,q1;
    for (int i=0;i<SPACEDIM;i++) {
      a1[i]=i+1;
      b1[i]=4-i;
      p1[i]=5*i+5;
      q1[i]=2.-0.5*static_cast<REAL>(i);
    }
    s1=a1+b1;
    d1=b1-s1;
    Tensor<2,SPACEDIM> a2,b2,s2,d2,p2,q2;
    for (int i=0;i<SPACEDIM;i++) {
      a2[i]=a1;
      b2[i]=b1;
      s2[i]=s1;
      d2[i]=d1;
      p2[i]=p1;
      q2[i]=q1;
    }
    Tensor<3,SPACEDIM> a3,b3,s3,d3,p3,q3;
    for (int i=0;i<SPACEDIM;i++) {
      a3[i]=a2;
      b3[i]=b2;
      s3[i]=s2;
      d3[i]=d2;
      p3[i]=p2;
      q3[i]=q2;
    }
    Tensor<4,SPACEDIM> a4,b4,s4,d4,p4,q4;
    for (int i=0;i<SPACEDIM;i++) {
      a4[i]=a3;
      b4[i]=b3;
      s4[i]=s3;
      d4[i]=d3;
      p4[i]=p3;
      q4[i]=q3;
    }

    Tensor<3,SPACEDIM> t;
    product(t,a3,5.);
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t[i],p2);
    product(t,5.,a3);
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t[i],p2);

    quotient(t,b3,2.);
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t[i],q2);
  }
}

void rank4Case() {
//TRACER_CALL(z,"rank4Case");
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    TRACER_CALL(z,"constructors");
    Tensor<4,SPACEDIM> t0;
    cout << "Tensor<4," << SPACEDIM << "> = " << t0 << endl;
    for (int l=0;l<SPACEDIM;l++) {
      for (int k=0;k<SPACEDIM;k++) {
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            CHECK_SAME(t0[l][k][j][i],0.);
          }
        }
      }
    }

    REAL d[SPACEDIM][SPACEDIM][SPACEDIM][SPACEDIM];
    for (int l=0;l<SPACEDIM;l++) {
      for (int k=0;k<SPACEDIM;k++) {
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            d[l][k][j][i]=i+j+k+l;
          }
        }
      }
    }
    Tensor<4,SPACEDIM> t4(d);
    for (int l=0;l<SPACEDIM;l++) {
      for (int k=0;k<SPACEDIM;k++) {
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            CHECK_SAME(t4[l][k][j][i],d[l][k][j][i]);
          }
        }
      }
    }
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"initializations");
    Tensor<4,SPACEDIM> t1;
    t1.initialize(-1.);
    for (int l=0;l<SPACEDIM;l++) {
      for (int k=0;k<SPACEDIM;k++) {
        for (int j=0;j<SPACEDIM;j++) {
          for (int i=0;i<SPACEDIM;i++) {
            CHECK_SAME(t1[l][k][j][i],-1.);
          }
        }
      }
    }

    int n=SPACEDIM*SPACEDIM*SPACEDIM*SPACEDIM;
    LAVector v(n);
    t1.unroll(v);
    for (int i=0;i<n;i++) CHECK_SAME(v[i],-1.);
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"operators");
    Tensor<1,SPACEDIM> a1,b1,s1,d1,p1,q1;
    for (int i=0;i<SPACEDIM;i++) {
      a1[i]=i+1;
      b1[i]=4-i;
      p1[i]=5*i+5;
      q1[i]=2.5;
    }
    s1=a1+b1;
    d1=b1-a1;
    Tensor<2,SPACEDIM> a2,b2,s2,d2,p2,q2;
    for (int i=0;i<SPACEDIM;i++) {
      a2[i]=a1;
      b2[i]=b1;
      s2[i]=s1;
      d2[i]=d1;
      p2[i]=p1;
      q2[i]=q1;
    }
    Tensor<3,SPACEDIM> a3,b3,s3,d3,p3,q3;
    for (int i=0;i<SPACEDIM;i++) {
      a3[i]=a2;
      b3[i]=b2;
      s3[i]=s2;
      d3[i]=d2;
      p3[i]=p2;
      q3[i]=q2;
    }
    Tensor<4,SPACEDIM> a4,b4,s4,d4,p4,q4;
    for (int i=0;i<SPACEDIM;i++) {
      a4[i]=a3;
      b4[i]=b3;
      s4[i]=s3;
      d4[i]=d3;
      p4[i]=p3;
      q4[i]=q3;
    }

    Tensor<4,SPACEDIM> t1,t2;
    t1=a4;
    t2=b4;
    CHECK_TEST(!(t1==t2));
    CHECK_TEST(t1!=t2);

    t1+=t2;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t1[i],s3);

    t2-=t1;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t2[i],-a3);

    t2*=5.;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t2[i],-p3);

    t1/=2.;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t1[i],q3);

    t1=a4;
    t2=b4;

    Tensor<4,SPACEDIM> t=t1+t2;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t[i],s3);

    Tensor<4,SPACEDIM> tt=t2-t1;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(tt[i],d3);

    Tensor<4,SPACEDIM> ttt=-t2;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(ttt[i],-b3);
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"products and quotients");
    Tensor<1,SPACEDIM> a1,b1,s1,d1,p1,q1;
    for (int i=0;i<SPACEDIM;i++) {
      a1[i]=i+1;
      b1[i]=4-i;
      p1[i]=5*i+5;
      q1[i]=2.-0.5*static_cast<REAL>(i);
    }
    s1=a1+b1;
    d1=b1-s1;
    Tensor<2,SPACEDIM> a2,b2,s2,d2,p2,q2;
    for (int i=0;i<SPACEDIM;i++) {
      a2[i]=a1;
      b2[i]=b1;
      s2[i]=s1;
      d2[i]=d1;
      p2[i]=p1;
      q2[i]=q1;
    }
    Tensor<3,SPACEDIM> a3,b3,s3,d3,p3,q3;
    for (int i=0;i<SPACEDIM;i++) {
      a3[i]=a2;
      b3[i]=b2;
      s3[i]=s2;
      d3[i]=d2;
      p3[i]=p2;
      q3[i]=q2;
    }
    Tensor<4,SPACEDIM> a4,b4,s4,d4,p4,q4;
    for (int i=0;i<SPACEDIM;i++) {
      a4[i]=a3;
      b4[i]=b3;
      s4[i]=s3;
      d4[i]=d3;
      p4[i]=p3;
      q4[i]=q3;
    }

    Tensor<4,SPACEDIM> t;
    product(t,a4,5.);
    for (int i=0;i<SPACEDIM;i++) {
      CHECK_SAME(t[i],p3);
    }
    product(t,5.,a4);
    for (int i=0;i<SPACEDIM;i++) {
      CHECK_SAME(t[i],p3);
    }

    quotient(t,b4,2.);
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t[i],q3);
  }
}

int main( int /*argc*/, char** /*argv*/) {
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    TRACER_CALL(z,"constructors");
    Tensor<1,SPACEDIM> t0;
    cout << "Tensor<1," << SPACEDIM << "> = " << t0 << endl;

    Tensor<1,SPACEDIM> t1(false);
    cout << "Tensor<1," << SPACEDIM << ">(false) = " << t1 << endl;

    REAL d[SPACEDIM]; 
    for (int i=0;i<SPACEDIM;i++) d[i]=static_cast<REAL>(i);
    Tensor<1,SPACEDIM> t2(d);
    cout << "Tensor<1," << SPACEDIM << ">(REAL[]) = " << t2 << endl;

    Point<SPACEDIM> p; 
    for (int i=0;i<SPACEDIM;i++) p[i]=-static_cast<REAL>(i);
    Tensor<1,SPACEDIM> t3(p);
    cout << "Tensor<1," << SPACEDIM << ">(Point) = " << t3 << endl;

    Tensor<1,SPACEDIM> t4(t3);
    cout << "Tensor<1," << SPACEDIM << ">(Tensor) = " << t4 << endl;
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"initializations");
    Tensor<1,SPACEDIM> t1;
    t1.initialize(-1.);
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t1[i],-1.);

    LAVector v(SPACEDIM);
    t1.unroll(v);
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t1[i],v[i]);
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"operators");
    Tensor<1,SPACEDIM> t1,t2;
    for (int i=0;i<SPACEDIM;i++) {
        t1[i]=i+1;
        t2[i]=4-i;
    }
    CHECK_TEST(!(t1==t2));
    CHECK_TEST(t1!=t2);

    t1+=t2;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t1[i],5.);

    t2-=t1;
    for (int i=0;i<SPACEDIM;i++) {
      CHECK_SAME(t2[i],-static_cast<REAL>(i+1));
    }

    t2*=5.;
    for (int i=0;i<SPACEDIM;i++) {
      CHECK_SAME(t2[i],-static_cast<REAL>(5*(i+1)));
    }

    t1/=2.;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t1[i],2.5);

    for (int i=0;i<SPACEDIM;i++) {
      t1[i]=i+1;
      t2[i]=4-i;
    }
    Tensor<1,SPACEDIM> t=t1+t2;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t[i],5.);

    Tensor<1,SPACEDIM> tt=t2-t1;
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(tt[i],3-2*i);

    Tensor<1,SPACEDIM> ttt=-t2;
    for (int i=0;i<SPACEDIM;i++) {
      CHECK_SAME(ttt[i],static_cast<REAL>(i)-4.);
    }
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"products and quotients");
    Tensor<1,SPACEDIM> t1,t2;
    for (int i=0;i<SPACEDIM;i++) {
      t1[i]=i+1;
      t2[i]=4-i;
    }

    REAL s;
    product(s,t1,t2);
    switch (SPACEDIM) {
      case 1:
        CHECK_SAME(s,4.);
        break;
      case 2:
        CHECK_SAME(s,10.);
        break;
      case 3:
        CHECK_SAME(s,16.);
    }

    Tensor<1,SPACEDIM> t;
    product(t,t1,5.);
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t[i],5*i+5);

    product(t,t2,2.);
    for (int i=0;i<SPACEDIM;i++) CHECK_SAME(t[i],8-2*i);

    quotient(t,t2,2.);
    for (int i=0;i<SPACEDIM;i++) {
      CHECK_SAME(t[i],2.-static_cast<REAL>(i)*0.5);
    }
  }

  rank2Case();
  rank3Case();
  rank4Case();

  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"products");
    Tensor<1,SPACEDIM> a1,b1;
    for (int i=0;i<SPACEDIM;i++) {
      a1[i]=i+1;
      b1[i]=4-i;
    }
    Tensor<2,SPACEDIM> a2,b2;
    for (int i=0;i<SPACEDIM;i++) {
      a2[i]=a1;
      b2[i]=b1;
    }
    Tensor<3,SPACEDIM> a3,b3;
    for (int i=0;i<SPACEDIM;i++) {
      a3[i]=a2;
      b3[i]=b2;
    }
    Tensor<4,SPACEDIM> a4,b4;
    for (int i=0;i<SPACEDIM;i++) {
      a4[i]=a3;
      b4[i]=b3;
    }

    Tensor<1,SPACEDIM> t;
    product(t,a2,a1);
    switch (SPACEDIM) {
      case 1:
        CHECK_SAME(t[0],1.);
        break;
      case 2: {
        REAL ans[2]={ 5.,5. };
        for (int i=0;i<2;i++) CHECK_SAME(t[i],ans[i]);
        break;
      }
      case 3: {
        REAL ans[3]={ 14.,14.,14. };
        for (int i=0;i<3;i++) CHECK_SAME(t[i],ans[i]);
        break;
      }
    }

    product(t,a1,a2);
    switch (SPACEDIM) {
      case 1:
        CHECK_SAME(t[0],1.);
        break;
      case 2: {
        REAL ans[2]={ 3.,6. };
        for (int i=0;i<2;i++) CHECK_SAME(t[i],ans[i]);
        break;
      }
      case 3: {
        REAL ans[3]={ 6.,12.,18. };
        for (int i=0;i<3;i++) CHECK_SAME(t[i],ans[i]);
        break;
      }
    }

    Tensor<2,SPACEDIM> tt;
    product(tt,a2,b2);
    switch (SPACEDIM) {
      case 1:
        CHECK_SAME(tt[0][0],4.);
        break;
      case 2: {
        REAL ans[4]={ 12., 9., 12., 9. };
        for (int j=0;j<2;j++) {
          for (int i=0;i<2;i++) {
            CHECK_SAME(tt[j][i],ans[i+2*j]);
          }
        }
        break;
      }
      case 3: {
        REAL ans[9]={ 24., 18., 12., 24., 18., 12., 24., 18., 12. };
        for (int j=0;j<2;j++) {
          for (int i=0;i<2;i++) {
            CHECK_SAME(tt[j][i],ans[i+3*j]);
          }
        }
        break;
      }
    }

    for (int i2=1;i2<=2;i2++) {
      for (int i1=1;i1<=2;i1++) {
        product(tt,a2,i1,b2,i2);
        switch (SPACEDIM) {
          case 1:
            CHECK_SAME(tt[0][0],4.);
            break;
          case 2: {
            REAL ans11[4]={ 8., 6., 16., 12. };
            REAL ans12[4]={ 7., 7., 14., 14. };
            REAL ans21[4]={ 12., 9., 12., 9. };
            REAL ans22[4]={ 10., 10., 10., 10.};
            for (int j=0;j<2;j++) {
              for (int i=0;i<2;i++) {
                if (i1==1) {
                  if (i2==1) { CHECK_SAME(tt[j][i],ans11[i+2*j]); }
                  else { CHECK_SAME(tt[j][i],ans12[i+2*j]); }
                } else {
                  if (i2==1) { CHECK_SAME(tt[j][i],ans21[i+2*j]); }
                  else { CHECK_SAME(tt[j][i],ans22[i+2*j]); }
                }
              }
            }
            break;
          }
          case 3: {
            REAL ans11[9]={ 12., 9., 6., 24., 18., 12., 36., 27., 18. };
            REAL ans12[9]={ 9., 9., 9., 18., 18., 18., 27., 27., 27. };
            REAL ans21[9]={24., 18., 12., 24., 18., 12., 24., 18., 12.};
            REAL ans22[9]={16., 16., 16., 16., 16., 16., 16., 16., 16.};
            for (int j=0;j<3;j++) {
              for (int i=0;i<3;i++) {
                if (i1==1) {
                  if (i2==1) { CHECK_SAME(tt[j][i],ans11[i+3*j]); }
                  else { CHECK_SAME(tt[j][i],ans12[i+3*j]); }
                } else {
                  if (i2==1) { CHECK_SAME(tt[j][i],ans21[i+3*j]); }
                  else { CHECK_SAME(tt[j][i],ans22[i+3*j]); }
                }
              }
            }
            break;
          }
        }
      }
    }

    Tensor<3,SPACEDIM> ttt;
    for (int k=1;k<=3;k++) {
      product(tt,a3,k,a1);
      switch (SPACEDIM) {
        case 1:
          CHECK_SAME(tt[0][0],1.);
          break;
        case 2: {
          REAL ans1[4]={ 3., 6., 3., 6. };
          REAL ans2[4]={ 3., 6., 3., 6. };
          REAL ans3[4]={ 5., 5., 5., 5. };
          for (int j=0;j<2;j++) {
            for (int i=0;i<2;i++) {
              if (k==1) { CHECK_SAME(tt[j][i],ans1[i+2*j]); }
              else if (k==2) { CHECK_SAME(tt[j][i],ans2[i+2*j]); }
              else { CHECK_SAME(tt[j][i],ans3[i+2*j]); }
            }
          }
          break;
        }
        case 3: {
          REAL ans1[9]={ 6., 12., 18., 6., 12., 18., 6., 12., 18. };
          REAL ans2[9]={ 6., 12., 18., 6., 12., 18., 6., 12., 18. };
          REAL ans3[9]={14.,14.,14.,14.,14.,14.,14.,14.,14.};
          for (int j=0;j<3;j++) {
            for (int i=0;i<3;i++) {
              if (k==1) { CHECK_SAME(tt[j][i],ans1[i+3*j]); }
              else if (k==2) { CHECK_SAME(tt[j][i],ans2[i+3*j]); }
              else { CHECK_SAME(tt[j][i],ans3[i+3*j]); }
            }
          }
          break;
        }
      }
    }

    product(ttt,a3,a2);
    switch (SPACEDIM) {
      case 1:
        CHECK_SAME(ttt[0][0][0],1.);
        break;
      case 2: {
        REAL ans[8]={ 3.,6., 3., 6., 3., 6., 3., 6. };
        for (int k=0;k<2;k++) {
          for (int j=0;j<2;j++) {
            for (int i=0;i<2;i++) {
              CHECK_SAME(ttt[k][j][i],ans[i+2*(j+2*k)]);
            }
          }
        }
        break;
      }
      case 3: {
        REAL ans[27]={ 6., 12., 18.,
                       6., 12., 18.,
                       6., 12., 18.,
                       6., 12., 18.,
                       6., 12., 18.,
                       6., 12., 18.,
                       6., 12., 18.,
                       6., 12., 18.,
                       6., 12., 18. };
        for (int k=0;k<3;k++) {
          for (int j=0;j<3;j++) {
            for (int i=0;i<3;i++) {
              CHECK_SAME(ttt[k][j][i],ans[i+3*(j+3*k)]);
            }
          }
        }
        break;
      }
    }

    product(ttt,a2,a3);
    switch (SPACEDIM) {
      case 1:
        CHECK_SAME(ttt[0][0][0],1.);
        break;
      case 2: {
        REAL ans[8]={ 3.,6., 3., 6., 3., 6., 3., 6. };
        for (int k=0;k<2;k++) {
          for (int j=0;j<2;j++) {
            for (int i=0;i<2;i++) {
              CHECK_SAME(ttt[k][j][i],ans[i+2*(j+2*k)]);
            }
          }
        }
        break;
      }
      case 3: {
        REAL ans[27]={ 6., 12., 18.,
                       6., 12., 18.,
                       6., 12., 18.,
                       6., 12., 18.,
                       6., 12., 18.,
                       6., 12., 18.,
                       6., 12., 18.,
                       6., 12., 18.,
                       6., 12., 18. };
        for (int k=0;k<3;k++) {
          for (int j=0;j<3;j++) {
            for (int i=0;i<3;i++) {
              CHECK_SAME(ttt[k][j][i],ans[i+3*(j+3*k)]);
            }
          }
        }
        break;
      }
    }

    Tensor<4,SPACEDIM> tttt;
    product(tttt,a3,b3);
    switch (SPACEDIM) {
      case 1:
        CHECK_SAME(tttt[0][0][0][0],4.);
        break;
      case 2: {
        REAL ans[16]=
          { 12.,9.,12.,9.,12.,9.,12.,9.,12.,9.,12.,9.,12.,9.,12.,9. };
        for (int l=0;l<2;l++) {
          for (int k=0;k<2;k++) {
            for (int j=0;j<2;j++) {
              for (int i=0;i<2;i++) {
                CHECK_SAME(tttt[l][k][j][i],ans[i+2*(j+2*(k+2*l))]);
              }
            }
          }
        }
        break;
      }
      case 3: {
        REAL ans[81]={ 24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12.,
                       24., 18., 12. };
        for (int l=0;l<2;l++) {
          for (int k=0;k<3;k++) {
            for (int j=0;j<3;j++) {
              for (int i=0;i<3;i++) {
                CHECK_SAME(tttt[l][k][j][i],ans[i+3*(j+3*(k+3*l))]);
              }
            }
          }
        }
        break;
      }
    }
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"outerProducts");
    Tensor<1,SPACEDIM> a1,b1;
    for (int i=0;i<SPACEDIM;i++) {
      a1[i]=i+1;
      b1[i]=4-i;
    }
    Tensor<2,SPACEDIM> a2,b2;
    for (int i=0;i<SPACEDIM;i++) {
      a2[i]=a1;
      b2[i]=b1;
    }

    Tensor<2,SPACEDIM> tt;
    outerProduct(tt,a1,b1);
    switch (SPACEDIM) {
      case 1:
        CHECK_SAME(tt[0][0],4.);
        break;
      case 2: {
        REAL ans[4]={ 4.,3.,8.,6. };
        for (int j=0;j<2;j++) {
          for (int i=0;i<2;i++) {
            CHECK_SAME(tt[j][i],ans[i+2*j]);
          }
        }
        break;
      }
      case 3: {
        REAL ans[9]={ 4.,3.,2.,8.,6.,4.,12.,9.,6. };
        for (int j=0;j<3;j++) {
          for (int i=0;i<3;i++) {
            CHECK_SAME(tt[j][i],ans[i+3*j]);
          }
        }
        break;
      }
    }

    Tensor<3,SPACEDIM> ttt;
    outerProduct(ttt,a1,a2);
    switch (SPACEDIM) {
      case 1:
        CHECK_SAME(ttt[0][0][0],1.);
        break;
      case 2: {
        REAL ans[8]={ 1.,2.,1.,2.,2.,4.,2.,4. };
        for (int k=0;k<2;k++) {
          for (int j=0;j<2;j++) {
            for (int i=0;i<2;i++) {
              CHECK_SAME(ttt[k][j][i],ans[i+2*(j+2*k)]);
            }
          }
        }
        break;
      }
      case 3: {
        REAL ans[27]={ 1.,2.,3.,1.,2.,3.,1.,2.,3.,
                       2.,4.,6.,2.,4.,6.,2.,4.,6.,
                       3.,6.,9.,3.,6.,9.,3.,6.,9. };
        for (int k=0;k<3;k++) {
          for (int j=0;j<3;j++) {
            for (int i=0;i<3;i++) {
              CHECK_SAME(ttt[k][j][i],ans[i+3*(j+3*k)]);
            }
          }
        }
        break;
      }
    }

    outerProduct(ttt,a2,a1);
    switch (SPACEDIM) {
      case 1:
        CHECK_SAME(ttt[0][0][0],1.);
        break;
      case 2: {
        REAL ans[8]={ 1.,2.,2.,4.,1.,2.,2.,4. };
        for (int k=0;k<2;k++) {
          for (int j=0;j<2;j++) {
            for (int i=0;i<2;i++) {
              CHECK_SAME(ttt[k][j][i],ans[i+2*(j+2*k)]);
            }
          }
        }
        break;
      }
      case 3: {
        REAL ans[27]={ 1.,2.,3.,2.,4.,6.,3.,6.,9.,
                       1.,2.,3.,2.,4.,6.,3.,6.,9.,
                       1.,2.,3.,2.,4.,6.,3.,6.,9. };
        for (int k=0;k<3;k++) {
          for (int j=0;j<3;j++) {
            for (int i=0;i<3;i++) {
              CHECK_SAME(ttt[k][j][i],ans[i+3*(j+3*k)]);
            }
          }
        }
        break;
      }
    }

    Tensor<1,SPACEDIM> t;
    outerProduct(t,5.,a1);
    switch (SPACEDIM) {
      case 1:
        CHECK_SAME(t[0],5.);
        break;
      case 2: {
        REAL ans[2]={ 5.,10. };
        for (int i=0;i<2;i++) CHECK_SAME(t[i],ans[i]);
        break;
      }
      case 3: {
        REAL ans[3]={ 5.,10.,15. };
        for (int i=0;i<3;i++) CHECK_SAME(t[i],ans[i]);
        break;
      }
    }

    outerProduct(t,a1,5.);
    switch (SPACEDIM) {
      case 1:
        CHECK_SAME(t[0],5.);
        break;
      case 2: {
        REAL ans[2]={ 5.,10. };
        for (int i=0;i<2;i++) CHECK_SAME(t[i],ans[i]);
        break;
      }
      case 3: {
        REAL ans[3]={ 5.,10.,15. };
        for (int i=0;i<3;i++) CHECK_SAME(t[i],ans[i]);
        break;
      }
    }
  }
  if (SPACEDIM==2) {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    TRACER_CALL(z,"crossProducts");
    Tensor<1,2> a1,b1;
    for (int i=0;i<2;i++) {
      a1[i]=i+1;
      b1[i]=4-i;
    }

    Tensor<1,2> t;
    crossProduct(t,a1);
    REAL ans[2]={ 2.,-1. };
    for (int i=0;i<2;i++) CHECK_SAME(t[i],ans[i]);
  }
  if (SPACEDIM==3) {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"crossProducts");
    Tensor<1,3> a1,b1;
    for (int i=0;i<3;i++) {
      a1[i]=i+1;
      b1[i]=4-i;
    }

    Tensor<1,3> t;
    crossProduct(t,a1,b1);
    REAL ans[3]={ -5.,10.,-5. };
    for (int i=0;i<3;i++) CHECK_SAME(t[i],ans[i]);
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"trace");
    Tensor<2,SPACEDIM> tt;
    for (int j=0;j<SPACEDIM;j++) {
      for (int i=0;i<SPACEDIM;i++) {
        tt[j][i]=1.;
      }
    }
    CHECK_SAME(trace(tt),SPACEDIM);
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"determinant, invert, transpose");
    switch (SPACEDIM) {
      case 1: {
        Tensor<2,1> t,u;
        t[0][0]=1.;
        CHECK_SAME(determinant(t),1.);
        transpose(u,t);
        CHECK_SAME(u[0][0],1.);
        break;
      }
      case 2: {
        Tensor<2,2> tt,uu;
        tt[0][0]=1.;
        tt[1][0]=2.;
        tt[0][1]=3.;
        tt[1][1]=4.;
        CHECK_SAME(determinant(tt),-2.);
        REAL ansinv[4]={ -2.,1.5,1.,-0.5 };
        Tensor<2,2> tti=invert(tt);
        for (int j=0;j<2;j++) { 
          for (int i=0;i<2;i++) {
            CHECK_SAME(tti[j][i],ansinv[i+2*j]);
          }
        }
        REAL anstrans[4]={ 1.,2.,3.,4. };
        transpose(uu,tt);
        for (int j=0;j<2;j++) { 
          for (int i=0;i<2;i++) {
            CHECK_SAME(uu[j][i],anstrans[i+2*j]);
          }
        }
        break;
      }
      case 3: {
        Tensor<2,3> ttt,uuu;
        ttt[0][0]=1.;
        ttt[1][0]=2.;
        ttt[2][0]=3.;
        ttt[0][1]=4.;
        ttt[1][1]=5.;
        ttt[2][1]=6.;
        ttt[0][2]=7.;
        ttt[1][2]=8.;
        ttt[2][2]=10.;
        CHECK_SAME(determinant(ttt),-3.);

        REAL ansinv[9]={-2./3.,-2./3.,1.,-4./3.,11./3.,-2.,1.,-2.,1.};
        Tensor<2,3> tti=invert(ttt);
        for (int j=0;j<3;j++) { 
          for (int i=0;i<3;i++) {
            CHECK_SAME(tti[j][i],ansinv[i+3*j]);
          }
        }
        REAL anstrans[9]={ 1.,2.,3.,4.,5.,6.,7.,8.,10. };
        transpose(uuu,ttt);
        for (int j=0;j<3;j++) { 
          for (int i=0;i<3;i++) {
            CHECK_SAME(uuu[j][i],anstrans[i+3*j]);
          }
        }
      }
    }
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(z,"adjointProduct");
    switch (SPACEDIM) {
      case 1: {
        Tensor<2,1> t;
        t[0][0]=1.;
        Tensor<1,1> d,s;
        s[0]=1.;
        adjointProduct(d,t,s);
        CHECK_SAME(s[0],1.);
        break;
      }
      case 2: {
        Tensor<2,2> t;
        t[0][0]=1.;
        t[1][0]=2.;
        t[0][1]=3.;
        t[1][1]=4.;
        REAL ans0[2]={4.,-3.};
        REAL ans1[2]={-2.,1.};
        Tensor<1,2> d,s;
        s[0]=1.;
        s[1]=0.;
        adjointProduct(d,t,s);
        for (int i=0;i<2;i++) {
          CHECK_SAME(d[i],ans0[i]);
        }
        s[0]=0.;
        s[1]=1.;
        adjointProduct(d,t,s);
        for (int i=0;i<2;i++) {
          CHECK_SAME(d[i],ans1[i]);
        }
        break;
      }
      case 3: {
        Tensor<2,3> t;
        t[0][0]=1.;
        t[1][0]=2.;
        t[2][0]=3.;
        t[0][1]=4.;
        t[1][1]=5.;
        t[2][1]=6.;
        t[0][2]=7.;
        t[1][2]=8.;
        t[2][2]=10.;
        REAL ans0[3]={2.,2.,-3.};
        REAL ans1[3]={4.,-11.,6.};
        REAL ans2[3]={-3.,6.,-3.};
        Tensor<1,3> d,s;
        s[0]=1.;
        s[1]=0.;
        s[2]=0.;
        adjointProduct(d,t,s);
        for (int i=0;i<3;i++) {
          CHECK_SAME(d[i],ans0[i]);
        }
        s[0]=0.;
        s[1]=1.;
        s[2]=0.;
        adjointProduct(d,t,s);
        for (int i=0;i<3;i++) {
          CHECK_SAME(d[i],ans1[i]);
        }
        s[0]=0.;
        s[1]=0.;
        s[2]=1.;
        adjointProduct(d,t,s);
        for (int i=0;i<3;i++) {
          CHECK_SAME(d[i],ans2[i]);
        }
        break;

      }
    }
  }
  return 1;
}

template double __cmath_power<double>(double,unsigned);
