// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <float.h>
#include <stdlib.h>
#include "Debug.H"
#include "MemoryDebugger.H"
#include "Vector.H"

int main( int /*argc*/, char** /*argv*/) {
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(t,"constructors");
    Vector v0;
    CHECK_SAME(v0.size(),0);

    Vector v1(4);
    CHECK_SAME(v1.size(),4);
    CHECK_TEST(!v1.allZero());
    v1=0.;
    CHECK_TEST(v1.allZero());

    for (int i=0;i<4;i++) v1[i]=i+4;
    Vector v2(v1);
    CHECK_SAME(v1.size(),v2.size());
    for (int i=0;i<4;i++) {
      CHECK_SAME(v1[i],v2[i]);
    }
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(t,"initializations");
    Vector v1(4);
    CHECK_SAME(v1.size(),4);
    v1=1.;
    for (int i=0;i<4;i++) CHECK_SAME(v1[i],1.);

    v1.reinit(3);
    CHECK_TEST(v1.allZero());
    CHECK_SAME(v1.size(),3);
    for (int i=0;i<3;i++) CHECK_SAME(v1[i],0.);
//  CHECK_SAME(v1[3],1.);

    Vector v2(v1);
    CHECK_SAME(v2.size(),3);
    for (int i=0;i<3;i++) CHECK_SAME(v2[i],0.);

    v2=2.;
    v2.reinit(2,true);
    CHECK_SAME(v2.size(),2);
//  for (int i=0;i<2;i++) CHECK_SAME(v2[i],2.);
//  CHECK_TEST(!v2.allZero());

//  v1.reinit(v2,true);
//  CHECK_SAME(v1.size(),v2.size());
//  for (int i=0;i<2;i++) CHECK_SAME(v1[i],0.);
//  v2.reinit(v1);
//  for (int i=0;i<2;i++) CHECK_SAME(v2[i],0.);

    v1=1.;
    v2.reinit(v1.size(),true);
    v2=v1;
    CHECK_SAME(v1.size(),v2.size());
    for (int i=0;i<v1.size();i++) CHECK_SAME(v2[i],v1[i]);

    v1.equ(2.,v2);
    for (int i=0;i<v1.size();i++) CHECK_SAME(v1[i],2.*v2[i]);
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(t,"operators");
    Vector v1(4),v2(4);
    for (int i=0;i<4;i++) {
      v1[i]=i+1;
      v2[i]=4-i;
    }
    CHECK_SAME(v1*v2,20.);
    v1+=v2;
    for (int i=0;i<4;i++) CHECK_SAME(v1[i],5.);
    v2-=v1;
    for (int i=0;i<4;i++) {
      CHECK_SAME(v2[i],-static_cast<REAL>(i+1));
    }
    v2.scale(5.);
    for (int i=0;i<4;i++) {
      CHECK_SAME(v2[i],-static_cast<REAL>(5*(i+1)));
    }
//  v1/=2.;
//  for (int i=0;i<4;i++) CHECK_SAME(v1[i],2.5);
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(t,"norms");
    Vector v1(4);
    for (int i=0;i<4;i++) v1[i]=i+1;
//  CHECK_SAME(v1.meanValue(),2.5);
//  CHECK_SAME(v1.l1Norm(),10.);
    double exact=sqrt(30.);
    double error=abs(v1.l2Norm()-exact)/exact;
    CHECK_TEST(error<=4.*DBL_EPSILON);
    CHECK_SAME(v1.linftyNorm(),4.);
//  CHECK_TEST(abs(pow(v1.lpNorm(3.),3.)-100.)<DBL_EPSILON*256.);
  }
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
//  TRACER_CALL(t,"arithmetic operations");
    Vector v1(4),v2(4);
    for (int i=0;i<4;i++) {
//    v1[i]=i+1;
      v1[i]=i;
      v2[i]=4-i;
    }
//  v1.add(-1.);
//  for (int i=0;i<4;i++) {
//    CHECK_SAME(v1[i],static_cast<REAL>(i));
//  }
    v2.add(1.,v1);
    for (int i=0;i<4;i++) CHECK_SAME(v2[i],4.);
    v1.add(2.,v2);
    for (int i=0;i<4;i++) {
      CHECK_SAME(v1[i],static_cast<REAL>(i+8));
    }
//  v2.scale(v1);
//  for (int i=0;i<4;i++) {
//    CHECK_SAME(v2[i],static_cast<REAL>(32+4*i));
//  }
//  v1.ratio(v1,v2);
//  for (int i=0;i<4;i++) {
//    CHECK_SAME(v1[i],0.25);
//  }
  }

  return 1;
}

#include "NumPtr.C"
INSTANTIATE_NUMPTR(REAL)
