// "$Header: /home/faculty/johnt/cvs/deal_new/examples/EllipticInCircle.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"

/* $Id: step-5.cc,v 1.31 2003/01/09 20:32:45 wolf Exp $ */
/* Author: Wolfgang Bangerth, University of Heidelberg, 1999 */

/*    $Id: step-5.cc,v 1.31 2003/01/09 20:32:45 wolf Exp $       */
/*    Version: $Name: Version-4-0-0 $                                              */
/*                                                                */
/*    Copyright (C) 1999, 2000, 2001, 2002, 2003 by the deal.II authors */
/*                                                                */
/*    This file is subject to QPL and may not be  distributed     */
/*    without copyright and license information. Please refer     */
/*    to the file deal.II/doc/license.html for the  text  and     */
/*    further information on this license.                        */
//
//modified from deal.II/examples/step-5/step-5.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>
#include <map>

#include <AreaGraphTool.H>
#include "Boundary.H"
#include <BoundaryCondition.H>
#include <DataOut.H>
#include <DofHandler.H>
#include <FEValues.H>
#include <FiniteElement.H>
#include <Mapping.H>
#include "MemoryDebugger.H"
#include "NumPtr.H"
#include <Point.H>
#include <Precondition.H>
#include "RefinementStudy.H"
#include "SetTraps.H"
#include <Tessellation.H>
#include <Tracer.H>

//problem is
//  - div [ k(x) grad u ] = 1 in domain
//  u(x) = 0                  on boundary

class VariableCoefficientMaterialModel : public MaterialModel {
  private:
    inline REAL coefficientValue(const Point<SPACEDIM> &p) const {
      return (p.square()<0.25 ? 20. : 1.);
    }
  public:
    void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_matrix,
    Vector &cell_rhs) const {
      int dofs_per_cell=cell_rhs.size();
      int n_q_points=q_points.getNumber();
      for (int q=0;q<n_q_points;q++) {
        REAL coefficient_value=coefficientValue(q_points[q]);
        REAL jxw=jxw_values[q];
        REAL rhsv_jxw=jxw; // i.e., rhs = 1.
        for (int i=0;i<dofs_per_cell;i++) {
          Tensor<1,SPACEDIM> sgi=sg(i,q);
          sgi*=coefficient_value*jxw;
          for (int j=0;j<dofs_per_cell;j++) {
            cell_matrix(i,j)+=sgi*sg(j,q);
          }
          cell_rhs[i]+=sv(i,q)*rhsv_jxw;
        }
      }
    }
};

int main (int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif

#if (SPACEDIM==2)
  int n_vertices=25;
  NumPtr<Point<SPACEDIM> > vertices(n_vertices);
  REAL rthalf=sqrt(0.5);
  REAL other1=0.2668;
  REAL other2=0.3139;
  REAL other3=0.6621;
  int v=0;
  vertices[v++]=Point<2>(-rthalf,-rthalf);
  vertices[v++]=Point<2>( rthalf,-rthalf);
  vertices[v++]=Point<2>(-other1,-other1);
  vertices[v++]=Point<2>( other1,-other1);
  vertices[v++]=Point<2>(-other1, other1);
  vertices[v++]=Point<2>( other1, other1);
  vertices[v++]=Point<2>(-rthalf, rthalf);
  vertices[v++]=Point<2>( rthalf, rthalf);
  vertices[v++]=Point<2>( 0.    ,-1.    );
  vertices[v++]=Point<2>( 0.5   ,-0.5   );
  vertices[v++]=Point<2>( 0.    ,-other2);
  vertices[v++]=Point<2>(-0.5   ,-0.5   );
  vertices[v++]=Point<2>( 0.    ,-other3);
  vertices[v++]=Point<2>(-other2, 0.    );
  vertices[v++]=Point<2>(-0.5   , 0.5   );
  vertices[v++]=Point<2>(-1.    , 0.    );
  vertices[v++]=Point<2>(-other3, 0.    );
  vertices[v++]=Point<2>( other2, 0.    );
  vertices[v++]=Point<2>( 0.    , other2);
  vertices[v++]=Point<2>( 0.    , 0.    );
  vertices[v++]=Point<2>( 1.    , 0.    );
  vertices[v++]=Point<2>( 0.5   , 0.5   );
  vertices[v++]=Point<2>( other3, 0.    );
  vertices[v++]=Point<2>( 0.    , 1.    );
  vertices[v++]=Point<2>( 0.    , other3);
# ifdef DEBUG
//{
//  TRACER_CALL(t,"main");
//  for (int i=0;i<vertices.getNumber();i++) {
//    cout << "\tvertices[" << i << "] = " << vertices[i] << endl;
//  }
//}
# endif

  int n_cells=20;
  NumPtr<CellData> cells(n_cells);
  for (int i=0;i<cells.getNumber();i++) cells[i].allocate(4);
  int c=0;
  cells[c  ].vertex(0)= 0;
  cells[c  ].vertex(1)= 8;
  cells[c  ].vertex(3)=12;
  cells[c++].vertex(2)=11;
  cells[c  ].vertex(0)= 8;
  cells[c  ].vertex(1)= 1;
  cells[c  ].vertex(3)= 9;
  cells[c++].vertex(2)=12;
  cells[c  ].vertex(0)=12;
  cells[c  ].vertex(1)= 9;
  cells[c  ].vertex(3)= 3;
  cells[c++].vertex(2)=10;
  cells[c  ].vertex(0)=11;
  cells[c  ].vertex(1)=12;
  cells[c  ].vertex(3)=10;
  cells[c++].vertex(2)= 2;
  cells[c  ].vertex(0)= 0;
  cells[c  ].vertex(1)=11;
  cells[c  ].vertex(3)=16;
  cells[c++].vertex(2)=15;
  cells[c  ].vertex(0)=11;
  cells[c  ].vertex(1)= 2;
  cells[c  ].vertex(3)=13;
  cells[c++].vertex(2)=16;
  cells[c  ].vertex(0)=16;
  cells[c  ].vertex(1)=13;
  cells[c  ].vertex(3)= 4;
  cells[c++].vertex(2)=14;
  cells[c  ].vertex(0)=15;
  cells[c  ].vertex(1)=16;
  cells[c  ].vertex(3)=14;
  cells[c++].vertex(2)= 6;
  cells[c  ].vertex(0)= 2;
  cells[c  ].vertex(1)=10;
  cells[c  ].vertex(3)=19;
  cells[c++].vertex(2)=13;
  cells[c  ].vertex(0)=10;
  cells[c  ].vertex(1)= 3;
  cells[c  ].vertex(3)=17;
  cells[c++].vertex(2)=19;
  cells[c  ].vertex(0)=19;
  cells[c  ].vertex(1)=17;
  cells[c  ].vertex(3)= 5;
  cells[c++].vertex(2)=18;
  cells[c  ].vertex(0)=13;
  cells[c  ].vertex(1)=19;
  cells[c  ].vertex(3)=18;
  cells[c++].vertex(2)= 4;
  cells[c  ].vertex(0)= 1;
  cells[c  ].vertex(1)=20;
  cells[c  ].vertex(3)=22;
  cells[c++].vertex(2)= 9;
  cells[c  ].vertex(0)=20;
  cells[c  ].vertex(1)= 7;
  cells[c  ].vertex(3)=21;
  cells[c++].vertex(2)=22;
  cells[c  ].vertex(0)=22;
  cells[c  ].vertex(1)=21;
  cells[c  ].vertex(3)= 5;
  cells[c++].vertex(2)=17;
  cells[c  ].vertex(0)= 9;
  cells[c  ].vertex(1)=22;
  cells[c  ].vertex(3)=17;
  cells[c++].vertex(2)= 3;
  cells[c  ].vertex(0)= 6;
  cells[c  ].vertex(1)=14;
  cells[c  ].vertex(3)=24;
  cells[c++].vertex(2)=23;
  cells[c  ].vertex(0)=14;
  cells[c  ].vertex(1)= 4;
  cells[c  ].vertex(3)=18;
  cells[c++].vertex(2)=24;
  cells[c  ].vertex(0)=24;
  cells[c  ].vertex(1)=18;
  cells[c  ].vertex(3)= 5;
  cells[c++].vertex(2)=21;
  cells[c  ].vertex(0)=23;
  cells[c  ].vertex(1)=24;
  cells[c  ].vertex(3)=21;
  cells[c++].vertex(2)= 7;
#else
  OBSOLETE("case not programmed");
#endif

//create tessellation
  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
  VariableCoefficientMaterialModel vcmm;
  tess.setMaterialModel(&vcmm);

//ball boundary
  BallBoundary boundary(Point<2>(0.,0.),1.);

//Dirichlet boundary condition
  ConstantFunctionSystem cfs(0.);
  NumPtr<BoundaryCondition*> boundary_condition(1);
  boundary_condition[0]=OPERATOR_NEW BoundaryCondition(&boundary);
  boundary_condition[0]->setFunctionSystem(&cfs);
  boundary_condition[0]->makeEssential();
//boundary_condition[0]->setMaterialModel(&vcmm);

  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
//    replace bogus_boundary_condition in Tessellation::create
      codimension1_level->setBoundaryCondition(boundary_condition[0]);
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
  tess.checkFaceBoundaryCondition(boundary_condition);
#endif

//(linear or) tensor product linear coordinate mapping
  MappingL mapping;

//Lagrange polynomials to represent solution of Laplace equation
  int degree=1;
  FE_L fe(degree);

//Gaussian quadrature
  GaussianQuadrature<SPACEDIM> quadrature_formula(
    GaussianQuadrature<SPACEDIM>::argGivenOrder(
    D_CHOOSE(2*degree-1,2*degree+1,2*degree+1)));
  map<int,Quadrature<SPACEDIM> > qmap;
  int nv=D_CHOOSE(Interval::verticesPerShape(),
                  Quadrilateral::verticesPerShape(),
                  Hexahedron::verticesPerShape());
  qmap[nv]=quadrature_formula;

//graphics setup
  Palette pal;
#if (SPACEDIM==1)
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool plot_gt("solution plot","x","y",0.,1.,0.,1.,&cmap,0,0.5);
#elif (SPACEDIM==2)
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]={0.,0.},rhi[2]={1.,1.};
  AreaGraphTool mesh_gt("grid","x0","x1",rlo,rhi,&cmap,0,0.5);
//AreaGraphTool contour_gt("solution contours","x0","x1",rlo,rhi,&cmap,
//  0,0.5);
  AreaGraphTool plot_gt("solution plot","x0","y1",rlo,rhi,&cmap,0,0.5);
//BOOLEAN use_lighting=FALSE;
//Vector3 vlow(0.,0.,0.);
//Vector3 vhigh(1.,1.,1.);
//VolGraphTool surface_gt("solution surface",vlow,vhigh,pal,use_lighting,
//  0.5);
//{ VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
//  surface_gt.drawBox(vlow,vhigh);
//}
#else
  BOOLEAN use_lighting=FALSE;
  Vector3 vlow(0.,0.,0.);
  Vector3 vhigh(1.,1.,1.);
  VolGraphTool mesh_gt("grid",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&mesh_gt);
    mesh_gt.drawBox(vlow,vhigh);
  }
  VolGraphTool plot_gt("solution plot",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&plot_gt);
    plot_gt.drawBox(vlow,vhigh);
  }
  VolGraphTool surface_gt("solution surface",vlow,vhigh,pal,use_lighting,
    0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
    surface_gt.drawBox(vlow,vhigh);
  }
#endif

//store info for RefinementStudy
  int ncycles=3;
  NumPtr<Vector> solutions(ncycles);
  NumPtr<DofHandler*> dof_handlers(ncycles);
  NumPtr<Tessellation*> tessellations(ncycles);
  NumPtr<LEVEL_CLASS*> coarsest_levels(ncycles);

//loop over mesh refinements
  for (int cycle=0;cycle<ncycles;cycle++) {
    cout << "Cycle " << cycle << ':' << endl;
    cout << "   Number of active cells: "
         << tess.numberActiveShapes() << endl
         << "   Total number of cells: "
         << tess.numberShapes() << endl;

    LEVEL_CLASS *copy_coarser_level=0;
    for (LEVEL_CLASS *level=tess.coarsestLevel();level;
    level=level->finerLevel()) {
      if (copy_coarser_level==0) {
        copy_coarser_level=coarsest_levels[cycle]=
          OPERATOR_NEW LEVEL_CLASS();
        coarsest_levels[cycle]->copyLevel(*level);
      } else {
        LEVEL_CLASS *cl=OPERATOR_NEW LEVEL_CLASS(copy_coarser_level);
        cl->copyLevel(*level);
        copy_coarser_level=cl;
      }
    }
    tessellations[cycle]=OPERATOR_NEW Tessellation();
    tessellations[cycle]->copyTessellation(tess,coarsest_levels[cycle]);

//  data structure to keep track of degrees of freedom in solver
    dof_handlers[cycle]=OPERATOR_NEW DofHandler(*tessellations[cycle],fe);
    dof_handlers[cycle]->distributeDofs();
    int n_dofs=dof_handlers[cycle]->numberDofs();
    cout << "   Number of degrees of freedom: " << n_dofs << endl;

//data structure to keep track of linear system storage
    SparsityPattern *sparsity_pattern=OPERATOR_NEW SparsityPattern(n_dofs,
      n_dofs,dof_handlers[cycle]->maxCouplingsBetweenDofs());
    dof_handlers[cycle]->makeSparsityPattern(*sparsity_pattern);
    sparsity_pattern->compress();

//storage for stiffness matrix, right-hand side
    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs,0.);

//  quadrature rule for Neumann bc; 
//    quadrature rule not used, but number of quadrature points is used
    map<int,Quadrature<SPACEDIM-1> > nqmap;
#if (SPACEDIM==2)
    nqmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
      GaussianQuadrature<1>::argGivenOrder(2*degree+1));
#elif (SPACEDIM==3)
    nqmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<2>(
      GaussianQuadrature<2>::argGivenOrder(2*degree+1));
#endif
//  assemble linear system
    dof_handlers[cycle]->assembleLinearSystem(mapping,qmap,nqmap,
      system_matrix,system_rhs);

//compute essential boundary values at mapped equally spaced points on bdry 
//calls FunctionSystem in BoundaryCondition's to set boundary values
    map<int,REAL> boundary_values;
    dof_handlers[cycle]->
      interpolateBoundaryValues(mapping,boundary_values);

//modify linear system and right-hand side for essential boundary conditions
    solutions[cycle].reinit(n_dofs);
    bool preserve_symmetry=true;
    dof_handlers[cycle]->applyBoundaryValues(boundary_values,
      system_matrix,system_rhs,preserve_symmetry);
    dof_handlers[cycle]->applyBoundaryValues(boundary_values,
      solutions[cycle]);

//  solve linear system
    SolverControl solver_control(1000,DBL_EPSILON*1.e4);
    CGSolver cg (solver_control);

    PreconditionSSOR preconditioner(system_matrix,1.2);;
    cg.solve(system_matrix,solutions[cycle],system_rhs,preconditioner);

//  compute solution values for graphical display
    DataOut data_out(*dof_handlers[cycle],mapping);
    data_out.addDataVector(solutions[cycle],"solution");
    int n_subdivisions=2;
    data_out.buildPatches(n_subdivisions);

    bool quit_button=D_CHOOSE(false,false,true);
#if (SPACEDIM>=2)
    data_out.drawGrid(&mesh_gt,quit_button);
#endif
#if (SPACEDIM==2)
//  data_out.contourPlot("solution",&contour_gt);
#endif
    data_out.plot("solution",&plot_gt,quit_button);
#if (SPACEDIM>=2)
//  data_out.plotSurface("solution",&surface_gt);
#endif
#if (SPACEDIM==1)
    if (!quit_button) {
      XYGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#elif (SPACEDIM==2)
    if (!quit_button) {
      AreaGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#endif
    delete sparsity_pattern;

    if (cycle<ncycles-1) tess.refineGlobal(1);
  }
  {
    NORM_TYPE nt=L2_NORM;
    REAL norm_exponent=2.; // needed for lp norms
    RefinementStudy rs(mapping,qmap,dof_handlers,solutions,nt,
      norm_exponent);
  }

  for (int i=0;i<tessellations.getNumber();i++) {
    delete tessellations[i]; tessellations[i]=0;
  }
  for (int i=0;i<dof_handlers.getNumber();i++) {
    delete dof_handlers[i]; dof_handlers[i]=0;
  }
  delete boundary_data; boundary_data=0;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
#include "NumPtr.C"
INSTANTIATE_NUMPTR(Tessellation*);
