// "$Header: /home/faculty/johnt/cvs/deal_new/examples/RectangularDomain.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2010 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>

#include "Boundary.H"
//#include "ConstraintMatrix.H"
#include "HcurlDataOut.H"
#include "HcurlDofHandler.H"
#include "HcurlFiniteElement.H"
#include "GenerateTess.H"
//#include <KellyErrorEstimator.H>
#include "Mapping.H"
#include "MemoryDebugger.H"
//#include "Multigrid.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "TimedObject.H"
#include "TimeIntegrator.H"
#include "Tracer.H"
#include "Vector.H"

#if (SPACEDIM==1)
#elif (SPACEDIM==2)
  enum MESH_ELEMENTS { 
    QUADRILATERAL_MESH_ELEMENTS,
    TRIANGULAR_MESH_ELEMENTS 
  };
  MESH_ELEMENTS mesh_elements=QUADRILATERAL_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=TRIANGULAR_MESH_ELEMENTS;
#else
  enum MESH_ELEMENTS { 
    HEXAHEDRAL_MESH_ELEMENTS,
    PRISM_MESH_ELEMENTS,
    TETRAHEDRAL_MESH_ELEMENTS 
  };
  MESH_ELEMENTS mesh_elements=HEXAHEDRAL_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=PRISM_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=TETRAHEDRAL_MESH_ELEMENTS;
//int dir=0;
#endif
int degree=0;
int nelements=20;
//int nelements=20;

//specify H
class ConstantMagneticFieldStrengthFunctionSystem : 
public TensorFunctionSystem {
  private:
    Tensor<1,SPACEDIM> val;
  protected:
    explicit ConstantMagneticFieldStrengthFunctionSystem(
    const FunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    ConstantMagneticFieldStrengthFunctionSystem(
      const Tensor<1,SPACEDIM> &H) : val(H) {;}
    virtual ~ConstantMagneticFieldStrengthFunctionSystem() {;}
    virtual Tensor<1,SPACEDIM> value(const Point<SPACEDIM> &p,int) const {
      return val;
    }
    virtual void value(const Point<SPACEDIM> &p,
    NumPtr<Tensor<1,SPACEDIM> > &function_values) const {
      function_values[0]=val;
    }
    virtual Tensor<2,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("why called?");
      Tensor<2,SPACEDIM> g;
      return g;
    }
    virtual void gradient(NumPtr<Tensor<2,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
    virtual void printOn(ostream &os) const {
      os << "ConstantMagneticFieldStrengthFunctionSystem: val = "
         << val << endl;
      TensorFunctionSystem::printOn(os);
    }
};
class VariableMagneticFieldStrengthFunctionSystem : 
public TensorFunctionSystem {
  protected:
    explicit VariableMagneticFieldStrengthFunctionSystem(
    const FunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    VariableMagneticFieldStrengthFunctionSystem() {;}
    virtual ~VariableMagneticFieldStrengthFunctionSystem() {;}
    virtual Tensor<1,SPACEDIM> value(const Point<SPACEDIM> &p,int) const {
      Tensor<1,SPACEDIM> function_values;
      function_values.initialize(0.);
      function_values[0]=cos(0.5*M_PI*p[0]);
      return function_values;
    }
    virtual void value(const Point<SPACEDIM> &p,
    NumPtr<Tensor<1,SPACEDIM> > &function_values) const {
      function_values[0].initialize(0.);
      function_values[0][0]=cos(0.5*M_PI*p[0]);
    }
    virtual Tensor<2,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("why called?");
      Tensor<2,SPACEDIM> g;
      g.initialize(0.);
      g[0][0]=-0.5*M_PI*sin(0.5*M_PI*p[0]);
      return g;
    }
    virtual void gradient(NumPtr<Tensor<2,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
    virtual void printOn(ostream &os) const {
      os << "VariableMagneticFieldStrengthFunctionSystem" << endl;
      TensorFunctionSystem::printOn(os);
    }
};
//specify E
class ConstantElectricFieldStrengthFunctionSystem : 
public TensorFunctionSystem {
  private:
    Tensor<1,SPACEDIM> val;
  protected:
    explicit ConstantElectricFieldStrengthFunctionSystem(
    const ConstantElectricFieldStrengthFunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    ConstantElectricFieldStrengthFunctionSystem(
      const Tensor<1,SPACEDIM> &E) : val(E) {}
    virtual ~ConstantElectricFieldStrengthFunctionSystem() {}
    virtual Tensor<1,SPACEDIM> value(const Point<SPACEDIM> &p,int) const {
      return val;
    }
    virtual void value(const Point<SPACEDIM> &p,
    NumPtr<Tensor<1,SPACEDIM> > &function_values) const {
      function_values[0]=val;
    }
    virtual Tensor<2,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("should not be called");
      Tensor<2,SPACEDIM> g;
      return g;
    }
    virtual void gradient(NumPtr<Tensor<2,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("should not be called");
    }
    virtual void printOn(ostream &os) const {
      os << "ConstantElectricFieldStrengthFunctionSystem: val = " << val
         << endl;
      TensorFunctionSystem::printOn(os);
    }
};
class VariableElectricFieldStrengthFunctionSystem : 
public TensorFunctionSystem {
  protected:
    explicit VariableElectricFieldStrengthFunctionSystem(
    const VariableElectricFieldStrengthFunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    VariableElectricFieldStrengthFunctionSystem() {;}
    virtual ~VariableElectricFieldStrengthFunctionSystem() {}
    virtual Tensor<1,SPACEDIM> value(const Point<SPACEDIM> &p,int) const {
      Tensor<1,SPACEDIM> function_value;
      function_value.initialize(0.);
      function_value[0]=cos(0.25*M_PI*(p[0]+1.));
      return function_value;
    }
    virtual void value(const Point<SPACEDIM> &p,
    NumPtr<Tensor<1,SPACEDIM> > &function_values) const {
      function_values[0].initialize(0.);
      function_values[0][0]=sin(M_PI*p[0]);
    }
    virtual Tensor<2,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("should not be called");
      Tensor<2,SPACEDIM> g;
      g.initialize(0.);
      g[0][0]=-0.25*M_PI*sin(.25*M_PI*(p[0]+1.));
      return g;
    }
    virtual void gradient(NumPtr<Tensor<2,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("should not be called");
    }
    virtual void printOn(ostream &os) const {
      os << "VariableElectricFieldStrengthFunctionSystem" << endl;
      TensorFunctionSystem::printOn(os);
    }
};

Tensor<1,SPACEDIM> exactElectricFieldStrength(const Point<SPACEDIM> &x,
const REAL t) {
  Tensor<1,SPACEDIM> return_value;
#if (SPACEDIM==1)
  return_value[0]=1.+sin(M_PI*(x[0]+1.)*0.25)*cos(M_PI*t*0.25);
#else
  OBSOLETE("not programmed");
#endif
  return return_value;
}

HCURL_MULTIPLIER_DATA_TYPE exactMagneticFieldStrength(
const Point<SPACEDIM> &x,const REAL t) {
#if (SPACEDIM==1)
  return 1.-cos(M_PI*(x[0]+1.)*0.25)*sin(M_PI*t*0.25);
#else
  OBSOLETE("not programmed");
#endif
}

// 1D:
//   partial_t { epsilon E_1 } + partial_x H_2 + sigma_e E_1 = - J_1
//   partial_t { mu H_2 } + partial_x E_1 + sigma_m H_2 = - M_2
//   solution with sigma_e = sigma_m = 0, J_1 = M_2 = 0:
//     E_1=e+C sin{[2n+1] pi [x+1]/4} cos{ [2n+1] pi t/[4 sqrt(epsilon mu)}
//     H_2=h-C sqrt{epsilon/mu}
//           cos{[2n+1] pi [x+1]/4} sin{ [2n+1] pi t/[4 sqrt(epsilon mu)}
//   initial conditions:
//     E_1(x,0)=e+C sin{[2n+1] pi [x+1]/4}
//     H_2(x,0)=h
//   boundary conditions:
//     E_1(-1,t)=e
//     H_2( 1,t)=h
//   We take n=0, C=1=e=h
// 2D:
//   partial_t { epsilon E_0 } - partial_{x_1} H_2 + sigma_e E_0 = - J_0
//   partial_t { epsilon E_1 } + partial_{x_1} H_2 + sigma_e E_1 = - J_1
//   partial_t { mu H_2 } + partial_{x_0} E_1 - partial_{x_1} E_0 
//     + sigma_m H_2 = - M_2
//   solution with sigma_e = sigma_m = 0, J_0 = J_1 = M_2 = 0:
//     E_0=e0+C sqrt{mu/epsilon} cos([2m+1] pi [x_0+1]/4)
//       sin(n pi [x1+1]/2) cos(n pi t / [ 2 sqrt{epsilon mu} ] )
//     E_1=e1+C sqrt{mu/epsilon} sin([2m+1] pi [x_0+1]/4)
//       sin(n pi [x1+1]/2) cos( n pi t / [ 2 sqrt{epsilon mu} ] )
//     H_2=h2+C cos([2m+1] pi [x_0+1]/4) cos(n pi [x1+1]/2)
//       sin( n pi t / [ 2 sqrt{epsilon mu} ] )
//   initial conditions:
//     E_0(x_0,x_1,0)=e0+C sqrt{mu/epsilon}
//       cos([2m+1] pi [x_0+1]/4) sin(n pi [x1+1]/2)
//     E_1(x_0,x_1,0)=e1+C sqrt{mu/epsilon}
//       sin([2m+1] pi [x_0+1]/4) cos(n pi [x1+1]/2)
//     H_2(x_0,x_1,0)=h2
//   boundary conditions:
//     E_0(x_0,-1,t)=e0
//     E_0(x_0, 1,t)=e0
//     E_1(-1,x_1,t)=e1
//     H_2( 1,x_1,t)=h2
class TransientMaxwellModel : public HcurlTransientMaterialModel {
  private:
    REAL permittivity; // epsilon
    REAL permeability; // mu
    REAL electric_conductivity; // sigma_e
    REAL magnetic_conductivity; // sigma_m
    Tensor<1,SPACEDIM> applied_electric_current_density; // J_a
    Tensor<1,SPACEDIM> applied_magnetic_current_density; // M_a

    void elementL2ProjectionMatrixAndRHS(const NumPtr<Point<SPACEDIM> >&,
    const NumPtr<REAL>&,const Array<2,REAL>&,Matrix&,Vector&) const {
      OBSOLETE("does not apply to this class");
    }
#if (SPACEDIM<=2)
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&,const NumPtr<REAL>&,
    const Array<2,Tensor<1,SPACEDIM> >*,const Array<2,REAL>*,
    Matrix&,Vector&) const {
      OBSOLETE("does not apply to this class");
    }
#elif (SPACEDIM==3)
    virtual void elementStiffnessMatrixAndRHS(const NumPtr<Point<3> >&,
    const NumPtr<REAL>&,const Array<2,Tensor<1,3> >*,
    const Array<2,Tensor<1,3> >*,Matrix&,Vector&) const {
      OBSOLETE("should have been defined in derived class");
    }
#endif
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &shape_values,
    const Array<2,Tensor<1,SPACEDIM> > &shape_gradients,
    Matrix &stiffness_matrix,Vector &rhs) const {
      OBSOLETE("does not apply to this class");
    }
    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,
    const Array<2,REAL> &shape_values,
    const Array<2,Tensor<1,SPACEDIM> > &shape_gradients,
    Matrix &mass_matrix,Matrix &stiffness_matrix,Vector &rhs) const {
      OBSOLETE("does not apply to this class");
    }
#if (SPACEDIM==2)
  virtual void elementStiffnessMatrixAndRHS(
  const NumPtr<Point<SPACEDIM> >&,const NumPtr<REAL>&,
  const Array<2,Tensor<1,SPACEDIM> >*,
  const Array<2,Tensor<2,SPACEDIM> >*,const Array<2,REAL>*,
  Matrix&,Vector&) const {
    OBSOLETE("does not apply to this class");
  }
#endif
  public:
    TransientMaxwellModel() : permittivity(1.),permeability(1.),
    electric_conductivity(0.),magnetic_conductivity(0.) {
      applied_electric_current_density[0]=0.;
      for (int d=1;d<SPACEDIM;d++) applied_electric_current_density[d]=0.;
      for (int d=0;d<SPACEDIM;d++) applied_magnetic_current_density[d]=0.;
    }
    REAL speedOfLight() const {
      return 1./sqrt(permittivity*permeability);
    }
//  used for L2 projection of initial data
    void elementL2ProjectionMatrixAndRHS(const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,Tensor<1,SPACEDIM> > *sv,
    const Array<2,HCURL_MULTIPLIER_DATA_TYPE> *msv,
    Matrix &cell_mass_matrix,Vector &cell_rhs) const {
      TRACER_CALL(t,"TransientMaxwellModel::elementL2ProjectionMatrixAndRHS");
#ifdef DEBUG
//    for (int q=0;q<q_points.getNumber();q++) {
//      cout << "\tq_points[" << q << "] = " << q_points[q] << endl;
//    }
//    for (int i=0;i<sv->size(0);i++) {
//      cout << "\tsv(" << i << ",*) = ";
//      for (int q=0;q<sv->size(1);q++) {
//        cout << (*sv)(i,q) << " ";
//      }
//      cout << endl;
//    }
#endif
      int dpc=sv->size(0);
      int mdpc=msv->size(0);
      CHECK_SAME(cell_mass_matrix.size(0),SPACEDIM*dpc+mdpc);
      CHECK_SAME(cell_mass_matrix.size(0),cell_rhs.size());
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
#ifdef DEBUG
//      cout << "\tjxw[" << q << "] = " << jxw << endl;
#endif
#if (SPACEDIM==1)
        REAL Eq=1.+sin(0.25*M_PI*(q_points[q][0]+1.));
        REAL Hq=1.;
        for (int i=0;i<dpc;i++) {
          Tensor<1,SPACEDIM> svi=(*sv)(i,q);
          svi*=jxw;
          for (int j=0;j<dpc;j++) {
            Tensor<1,SPACEDIM> svj=(*sv)(j,q);
            cell_mass_matrix(i,j)+=(svi*svj)*permittivity;
          }
          cell_rhs[i]+=svi[0]*Eq*permittivity;
        }
        for (int i=0;i<mdpc;i++) {
          HCURL_MULTIPLIER_DATA_TYPE msvi=(*msv)(i,q);
          msvi*=jxw;
          int dpcpi=dpc+i;
          for (int j=0;j<mdpc;j++) {
            HCURL_MULTIPLIER_DATA_TYPE msvj=(*msv)(j,q);
            cell_mass_matrix(dpcpi,dpc+j)+=(msvi*msvj)*permeability;
          }
          cell_rhs[dpcpi]+=msvi*Hq*permeability;
        }
#elif (SPACEDIM==2)
        REAL Eq0=sqrt(permeability/permittivity)
          *cos(0.25*M_PI*(q_points[q][0]+1.))
          *sin(0.5*M_PI*(q_points[q][1]+1.));
        REAL Eq1=1.+sqrt(permeability/permittivity)
          *sin(0.25*M_PI*(q_points[q][0]+1.))
          *cos(0.5*M_PI*(q_points[q][1]+1.));
        REAL Hq=1.+cos(0.25*M_PI*(q_points[q][0]+1.))
          *cos(0.5*M_PI*(q_points[q][1]+1.));
        for (int i=0;i<dpc;i++) {
          Tensor<1,SPACEDIM> svi=(*sv)(i,q);
          svi*=jxw;
          for (int j=0;j<dpc;j++) {
            Tensor<1,SPACEDIM> svj=(*sv)(j,q);
            cell_mass_matrix(i,j)+=(svi*svj)*permittivity;
            cell_mass_matrix(dpc+i,dpc+j)+=(svi*svj)*permittivity;
          }
          cell_rhs[i]+=svi[0]*Eq0*permittivity;
          cell_rhs[dpc+i]+=svi[0]*Eq1*permittivity;
        }
        for (int i=0;i<mdpc;i++) {
          HCURL_MULTIPLIER_DATA_TYPE msvi=(*msv)(i,q);
          msvi*=jxw;
          int dpcpi=2*dpc+i;
          for (int j=0;j<mdpc;j++) {
            HCURL_MULTIPLIER_DATA_TYPE msvj=(*msv)(j,q);
            cell_mass_matrix(dpcpi,2*dpc+j)+=(msvi*msvj)*permeability;
          }
          cell_rhs[dpcpi]+=msvi*Hq*permeability;
        }
#else
        OBSOLETE("not programmed");
#endif
      }
#ifdef DEBUG
/*
      for (int i=0;i<2*dpc;i++) {
        cout << "\tcell_mass_matrix(" << i << ",*) = ";
        for (int j=0;j<2*dpc;j++) {
          cout << cell_mass_matrix(i,j) << " ";
        }
        cout << endl;
      }
      for (int i=0;i<2*dpc;i++) {
        cout << "\tcell_rhs[" << i << "] = " << cell_rhs[i] << endl;
      }
*/
#endif
    }
#if (SPACEDIM==1)
    void elementStiffnessMatrixAndRHS(const NumPtr<Point<1> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,Tensor<1,1> > *sv,
    const Array<2,Tensor<2,1> > *sg,const Array<2,REAL> *msv,
    Matrix &cell_mass_matrix,Matrix &cell_stiffness_matrix,
    Vector &cell_rhs) const {
//    TRACER_CALL(t,"TransientMaxwellModel::elementStiffnessMatrixAndRHS");
      int dpc=sv->size(0);
      int mdpc=msv->size(0);
      CHECK_SAME(cell_mass_matrix.size(0),dpc+mdpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),dpc+mdpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),cell_rhs.size());
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        for (int i=0;i<dpc;i++) {
          Tensor<1,1> svi=(*sv)(i,q);
          svi*=jxw;
          for (int j=0;j<dpc;j++) {
            REAL dot=svi*(*sv)(j,q);
            cell_mass_matrix(i,j)+=dot*permittivity;
            cell_stiffness_matrix(i,j)+=dot*electric_conductivity;
          }
          Tensor<2,1> sgi=(*sg)(i,q);
          sgi*=jxw;
          for (int j=0;j<mdpc;j++) {
            REAL msvj=(*msv)(j,q);
            cell_stiffness_matrix(i,j+dpc)-=sgi[0][0]*msvj;
          }
          cell_rhs[i]-=svi*applied_electric_current_density;
        }
        for (int i=0;i<mdpc;i++) {
          int dpcpi=dpc+i;
          REAL msvi=(*msv)(i,q);
          msvi*=jxw;
          for (int j=0;j<dpc;j++) {
            cell_stiffness_matrix(dpcpi,j)+=msvi*(*sg)(j,q)[0][0];
          }
          for (int j=0;j<mdpc;j++) {
            int dpcpj=dpc+j;
            REAL dot=msvi*(*msv)(j,q);
            cell_mass_matrix(dpcpi,dpcpj)+=dot*permeability;
            cell_stiffness_matrix(dpcpi,dpcpj)+=dot*magnetic_conductivity;
          }
          cell_rhs[dpcpi]-=msvi*applied_magnetic_current_density[0];
        }
      }
    }
#endif
#if (SPACEDIM==2)
    void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<2> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,Tensor<1,2> > *sv,
    const Array<2,Tensor<2,2> > *sg,const Array<2,REAL> *msv,
    Matrix &cell_mass_matrix,Matrix &cell_stiffness_matrix,
    Vector &cell_rhs) const {
//    TRACER_CALL(t,"TransientMaxwellModel::elementStiffnessMatrixAndRHS");
      int dpc=sv->size(0);
      int mdpc=msv->size(0);
      CHECK_SAME(cell_mass_matrix.size(0),2*dpc+mdpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),2*dpc+mdpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),cell_rhs.size());
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        for (int i=0;i<dpc;i++) {
          int dpcpi=dpc+i;
          Tensor<1,2> svi=(*sv)(i,q);
          Tensor<2,2> sgi=(*sg)(i,q);
          svi*=jxw;
          for (int j=0;j<dpc;j++) {
            int dpcpj=dpc+j;
            const Tensor<1,2> &svj=(*sv)(j,q);
            REAL dot=svi*svj;
            cell_mass_matrix(i,j)+=dot*permittivity;
            cell_mass_matrix(dpcpi,dpcpj)+=dot*permittivity;

            cell_stiffness_matrix(i,j)+=dot*electric_conductivity;
            cell_stiffness_matrix(dpcpi,dpcpj)+=dot*electric_conductivity;
          }
          for (int j=0;j<mdpc;j++) {
            int dpcpj=2*dpc+j;
            const REAL &msvj=(*msv)(j,q);
            cell_stiffness_matrix(i,dpcpj)+=sgi[0][1]*msvj;
            cell_stiffness_matrix(i+dpc,dpcpj)-=sgi[1][0]*msvj;
          }
          cell_rhs[i]-=svi[0]*applied_electric_current_density[0];
          cell_rhs[i+dpc]-=svi[1]*applied_electric_current_density[1];
        }
        for (int i=0;i<mdpc;i++) {
          int dpcpi=2*dpc+i;
          REAL msvi=(*msv)(i,q)*jxw;
          for (int j=0;j<dpc;j++) {
            const Tensor<2,2> &sgj=(*sg)(j,q);
            cell_stiffness_matrix(dpcpi,j)-=msvi*sgj[0][1];
            cell_stiffness_matrix(dpcpi,j+dpc)+=msvi*sgj[1][0];
          }
          for (int j=0;j<mdpc;j++) {
            REAL msvj=(*msv)(j,q);
            int dpcpj=2*dpc+j;
            const Tensor<2,2> &sgj=(*sg)(j,q);
            cell_mass_matrix(dpcpi,dpcpj)+=msvi*permeability*msvj;
            cell_stiffness_matrix(dpcpi,dpcpj)+=
              msvi*magnetic_conductivity*msvj;
          }
          cell_rhs[dpcpi]-=msvi*applied_magnetic_current_density[0];
        }
      }
    }
#elif (SPACEDIM==3)
    void elementStiffnessMatrixAndRHS(const NumPtr<Point<3> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,Tensor<1,3> > *sv,
    const Array<2,Tensor<1,3> > *sc,const Array<2,Tensor<1,3> > *msv,
    Matrix &cell_mass_matrix,Matrix &cell_stiffness_matrix,
    Vector &cell_rhs) const {
//    TRACER_CALL(t,"TransientMaxwellModel::elementStiffnessMatrixAndRHS");
#   ifdef DEBUG
/*
      for (int q=0;q<q_points.getNumber();q++) {
        cout << "\tq_points[" << q << "] = " << q_points[q] << endl;
      }
      for (int q=0;q<jxw_values.getNumber();q++) {
        cout << "\tjxw_values[" << q << "] = " << jxw_values[q] << endl;
      }
      for (int q=0;q<sv->size(1);q++) {
        cout << "\tsv(*," << q << ") = ";
        for (int i=0;i<sv->size(0);i++) cout << (*sv)(i,q) << " ";
        cout << endl;
      }
      for (int q=0;q<sc->size(1);q++) {
        cout << "\tsc(*," << q << ") = ";
        for (int i=0;i<sc->size(0);i++) cout << (*sc)(i,q) << " ";
        cout << endl;
      }
*/
#   endif
      int dpc=sv->size(0);
      int mdpc=msv->size(0);
      CHECK_SAME(cell_mass_matrix.size(0),dpc+mdpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),dpc+mdpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),cell_rhs.size());
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        for (int i=0;i<dpc;i++) {
          Tensor<1,3> svi=(*sv)(i,q);
          Tensor<1,3> sci=(*sc)(i,q);
          svi*=jxw;
          sci*=jxw;
          for (int j=0;j<dpc;j++) {
            const Tensor<1,3> &svj=(*sv)(j,q);
            REAL dot=svi*svj;
            cell_mass_matrix(i,j)+=dot*permittivity;
            cell_stiffness_matrix(i,j)+=dot*electric_conductivity;
          }
          for (int j=0;j<mdpc;j++) {
            cell_stiffness_matrix(i,dpc+j)+=sci*(*msv)(j,q);
          }
          cell_rhs[i]-=svi*applied_electric_current_density;
        }
        for (int i=0;i<mdpc;i++) {
          int dpcpi=dpc+i;
          Tensor<1,3> msvi=(*msv)(i,q)*jxw;
          for (int j=0;j<dpc;j++) {
            cell_stiffness_matrix(dpcpi,j)+=msvi*(*sc)(j,q);
          }
          for (int j=0;j<mdpc;j++) {
            int dpcpj=dpc+j;
            const Tensor<1,3> &msvj=(*msv)(j,q);
            REAL dot=msvi*msvj;
            cell_mass_matrix(dpcpi,dpcpj)+=dot*permeability;
            cell_stiffness_matrix(dpcpi,dpcpj)+=dot*magnetic_conductivity;
          }
          cell_rhs[dpcpi]-=msvi*applied_magnetic_current_density;
        }
      }
#   ifdef DEBUG
/*
      for (int i=0;i<cell_matrix.size(0);i++) {
        cout << "\tcell_matrix(" << i << ",*) = ";
        for (int j=0;j<cell_matrix.size(1);j++) {
          cout << cell_matrix(i,j) << " ";
        }
        cout << endl;
      }
      for (int i=0;i<cell_rhs.size();i++) {
        cout << "\tcell_rhs[" << i << "] = " << cell_rhs[i] << endl;
      }
*/
#   endif
    }
#endif
};

#ifdef TIMING
//the most expensive routines are the linear system solve and
//  interpolateBoundaryValues
//the latter requires computation of the shape functions on an element
//  at iteratively computed canonical points corresponding to boundary:
//TimedObjectList tl;
#endif

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
#ifdef TIMING
//TimedObject *z3=
//  OPERATOR_NEW TimedObject("main interpolateBoundaryValues",tl);
//TimedObject *z5=OPERATOR_NEW TimedObject("main solve",tl);
//TimedObject *z10=OPERATOR_NEW TimedObject("main refine",tl);
#endif
  Tessellation tess;
  const LEVEL_CLASS *level=tess.coarsestLevel();

//rectangular domain consisting of one interval/quadrilateral/hexagon
  Point<SPACEDIM> low(D_DECL(-1.,-1.,-1.));
//Point<SPACEDIM> low(0.,0.,0.);
  Point<SPACEDIM> edge[SPACEDIM];
    edge[0]=Point<SPACEDIM>(D_DECL(2.,0.,0.));
#if (SPACEDIM>=2)
    edge[1]=Point<SPACEDIM>(D_DECL(0.,2.,0.));
#endif
#if (SPACEDIM==3)
    edge[2]=Point<SPACEDIM>(0.,0.,2.);
#endif
  int number_elements[SPACEDIM]={D_DECL(nelements,nelements,nelements)};
  bool set_face_ids=true;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  BoundaryData *boundary_data=0;
#if (SPACEDIM==1)
  boundary_data=generateParallelpiped(tess,low,edge,number_elements,
    set_face_ids,boundary,boundary_condition);
  for (int i=0;i<2;i++) {
    const Boundary *b=boundary_condition[i]->getBoundary();
    delete boundary_condition[i];
    boundary_condition[i]=
      OPERATOR_NEW HcurlBoundaryCondition(boundary[i]);
    boundary_condition[i]->setBoundary(b);
    level->CODIMENSION1_LEVEL_CLASS::setBoundaryCondition(i,
      boundary_condition[i]);
  }
#elif (SPACEDIM==2)
  switch (mesh_elements) {
    case QUADRILATERAL_MESH_ELEMENTS: {
      boundary_data=generateParallelpiped(tess,low,edge,number_elements,
        set_face_ids,boundary,boundary_condition);
      for (int i=0;i<2*SPACEDIM;i++) {
        const Boundary *b=boundary_condition[i]->getBoundary();
        delete boundary_condition[i];
        boundary_condition[i]=
          OPERATOR_NEW HcurlBoundaryCondition(boundary[i]);
        boundary_condition[i]->setBoundary(b);
      }
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
//    tess.printCodimension2Elements();

//    for (int i=0;i<boundary_condition.getNumber();i++) {
//      cout << "\tboundary_condition[" << i << "] = "
//           << boundary_condition[i] << endl;
//      boundary_condition[i]->printOn(cout);
//      cout << "\tboundary = " << boundary_condition[i]->getBoundary()
//           << endl;
//      boundary_condition[i]->getBoundary()->printOn(cout);
//    }
#endif
      const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
//        Point<SPACEDIM> center=face_element->center(vertices);
          bool found=false;
          for (int i=0;i<boundary_condition.getNumber();i++) {
            const BoundaryCondition *bc=boundary_condition[i];
            const Boundary *b=bc->getBoundary();
            bool all_on_boundary=true;
            for (int v=0;v<face_element->numberVertices();v++) {
              int vi=face_element->vertexIndex(v);
              const Point<SPACEDIM> &ver=vertices[vi];
              if (abs(b->constraint(ver))>1.e-5) {
                all_on_boundary=false;
                break;
              }
            }
            if (all_on_boundary) {
              codimension1_level->setBoundaryCondition(bc);
              found=true;
              break;
            }
          }
        }
      }
      break;
    }
    case TRIANGULAR_MESH_ELEMENTS: {
//    boundary_data=generateParallelpipedTriangles(tess,low,edge,
//      number_elements,set_face_ids);
//    the following tessellation has all possible faces on outer boundary
      NumPtr<Point<2> > vertices(5);
      for (int i=0;i<vertices.getNumber();i++) vertices[i]=low;
      vertices[1]+=edge[0];
      vertices[3]+=edge[1];
      vertices[4]+=edge[0]; vertices[4]+=edge[1];
      vertices[2]=(vertices[0]+vertices[1]+vertices[3]+vertices[4])*0.25;
      NumPtr<CellData> cells(4);
      for (int c=0;c<cells.getNumber();c++) cells[c].allocate(3);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;

      cells[1].vertex(0)=1;
      cells[1].vertex(1)=4;
      cells[1].vertex(2)=2;

      cells[2].vertex(0)=2;
      cells[2].vertex(1)=4;
      cells[2].vertex(2)=3;

      cells[3].vertex(0)=0;
      cells[3].vertex(1)=2;
      cells[3].vertex(2)=3;
      boundary_data=tess.create(vertices,cells);
      boundary.allocate(4);
      boundary_condition.allocate(6);
      Point<2> center(vertices[2]);
      boundary[0]=OPERATOR_NEW PlaneBoundary(center,vertices[0],
        vertices[3]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(center,vertices[1],
        vertices[4]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(center,vertices[0],
        vertices[1]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(center,vertices[3],
        vertices[4]);
      for (int i=0;i<4;i++) {
        boundary_condition[i]=
          OPERATOR_NEW HcurlBoundaryCondition(boundary[i]);
      }

      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<SPACEDIM> center=face_element->center(vertices);
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              break;
            }
          }
        }
      }
      tess.synchronize(boundary_data);
      break;
    }
    default:
      OBSOLETE("bad mesh_elements");
  }
#else
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS: {
      boundary_data=generateParallelpiped(tess,low,edge,number_elements,
        set_face_ids,boundary,boundary_condition);
      for (int i=0;i<2*SPACEDIM;i++) {
        const Boundary *b=boundary_condition[i]->getBoundary();
        delete boundary_condition[i];
        boundary_condition[i]=
          OPERATOR_NEW HcurlBoundaryCondition(boundary[i]);
        boundary_condition[i]->setBoundary(b);
      }
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
//    tess.printCodimension2Elements();

//    for (int i=0;i<boundary_condition.getNumber();i++) {
//      cout << "\tboundary_condition[" << i << "] = "
//           << boundary_condition[i] << endl;
//      boundary_condition[i]->printOn(cout);
//      cout << "\tboundary = " << boundary_condition[i]->getBoundary()
//           << endl;
//      boundary_condition[i]->getBoundary()->printOn(cout);
//    }
#endif
      const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
//        Point<SPACEDIM> center=face_element->center(vertices);
          bool found=false;
          for (int i=0;i<boundary_condition.getNumber();i++) {
            const BoundaryCondition *bc=boundary_condition[i];
            const Boundary *b=bc->getBoundary();
            bool all_on_boundary=true;
            for (int v=0;v<face_element->numberVertices();v++) {
              int vi=face_element->vertexIndex(v);
              const Point<SPACEDIM> &ver=vertices[vi];
              if (abs(b->constraint(ver))>1.e-5) {
                all_on_boundary=false;
                break;
              }
            }
            if (all_on_boundary) {
              codimension1_level->setBoundaryCondition(bc);
              found=true;
              break;
            }
          }
        }
      }
      for (bool doit=tess.firstCodimension2ActiveShape();doit;
      doit=tess.nextCodimension2ActiveShape()) {
//      TRACER_CALL(t,"setBoundaryCondition");
        Level1 *codimension2_level=
          tess.presentCodimension2Level();
        pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
          codimension2_level->getBoundaryConditions();
        const BoundaryCondition *bc=bcs.first;
        if (bc==0) continue;
        Shape1 *face_element=codimension2_level->getShape();
        codimension2_level->setBoundaryCondition(0);
        int count=0;
        for (int i=0;i<boundary.getNumber();i++) {
          const BoundaryCondition *bc=boundary_condition[i];
          const Boundary *b=bc->getBoundary();
          bool all_on_boundary=true;
          for (int v=0;v<face_element->numberVertices();v++) {
            int vi=face_element->vertexIndex(v);
            const Point<SPACEDIM> &ver=vertices[vi];
            if (abs(b->constraint(ver))>1.e-5) {
              all_on_boundary=false;
              break;
            }
          }
          if (all_on_boundary) {
            codimension2_level->setBoundaryCondition(bc);
            count++;
          }
        }
      }
      break;
    }
    case PRISM_MESH_ELEMENTS: {
//    boundary_data=generateParallelpipedPrisms(tess,low,edge,
//      number_elements,dir,set_face_ids);
//    the following tessellation has all possible faces on outer boundary
      NumPtr<Point<3> > vertices(10);
      for (int i=0;i<vertices.getNumber();i++) vertices[i]=low;
      vertices[1]+=edge[0];
      vertices[2]+=edge[0]*0.5; vertices[2]+=edge[1]*0.5; 
      vertices[3]+=edge[1];
      vertices[4]+=edge[0]; vertices[4]+=edge[1];
      vertices[5]+=edge[2];
      vertices[6]+=edge[0]; vertices[6]+=edge[2];
      vertices[7]+=edge[0]*0.5; vertices[7]+=edge[1]*0.5; 
                                vertices[7]+=edge[2];
      vertices[8]+=edge[1]; vertices[8]+=edge[2];
      vertices[9]+=edge[0]; vertices[9]+=edge[1]; vertices[9]+=edge[2];
      NumPtr<CellData> cells(4);
      for (int c=0;c<cells.getNumber();c++) cells[c].allocate(6);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;
      cells[0].vertex(3)=5;
      cells[0].vertex(4)=6;
      cells[0].vertex(5)=7;

      cells[1].vertex(0)=0;
      cells[1].vertex(1)=2;
      cells[1].vertex(2)=3;
      cells[1].vertex(3)=5;
      cells[1].vertex(4)=7;
      cells[1].vertex(5)=8;

      cells[2].vertex(0)=2;
      cells[2].vertex(1)=4;
      cells[2].vertex(2)=3;
      cells[2].vertex(3)=7;
      cells[2].vertex(4)=9;
      cells[2].vertex(5)=8;

      cells[3].vertex(0)=1;
      cells[3].vertex(1)=4;
      cells[3].vertex(2)=2;
      cells[3].vertex(3)=6;
      cells[3].vertex(4)=9;
      cells[3].vertex(5)=7;
      boundary_data=tess.create(vertices,cells);
      boundary.allocate(6);
      boundary_condition.allocate(6);
      Point<3> center(low+(edge[0]+edge[1]+edge[2])*0.5);
      boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1],
        low+edge[2]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
        low+edge[0]+edge[1],low+edge[0]+edge[2]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[2],
        low+edge[0]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
        low+edge[1]+edge[2],low+edge[1]+edge[0]);
      boundary[4]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0],
        low+edge[1]);
      boundary[5]=OPERATOR_NEW PlaneBoundary(center,low+edge[2],
        low+edge[2]+edge[0],low+edge[2]+edge[1]);
      for (int i=0;i<2*SPACEDIM;i++) {
        boundary_condition[i]=
          OPERATOR_NEW HcurlBoundaryCondition(boundary[i]);
      }
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<SPACEDIM> center=face_element->center(vertices);
          bool found=false;
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              found=true;
              break;
            }
          }
        }
      }
      tess.synchronize(boundary_data);
      break;
    }
    case TETRAHEDRAL_MESH_ELEMENTS: {
//    boundary_data=generateParallelpipedTetrahedra(tess,low,edge,
//      number_elements,set_face_ids);
//    the following tessellation has all possible faces on outer boundary
      NumPtr<Point<3> > vertices(8);
      for (int i=0;i<vertices.getNumber();i++) vertices[i]=low;
      vertices[1]+=edge[0];
      vertices[2]+=edge[1];
      vertices[3]+=edge[0]; vertices[3]+=edge[1];
      vertices[4]+=edge[2];
      vertices[5]+=edge[0]; vertices[5]+=edge[2];
      vertices[6]+=edge[1]; vertices[6]+=edge[2];
      vertices[7]+=edge[0]; vertices[7]+=edge[1]; vertices[7]+=edge[2];
      NumPtr<CellData> cells(6);
      for (int c=0;c<cells.getNumber();c++) cells[c].allocate(4);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;
      cells[0].vertex(3)=4;

      cells[1].vertex(0)=1;
      cells[1].vertex(1)=4;
      cells[1].vertex(2)=5;
      cells[1].vertex(3)=6;

      cells[2].vertex(0)=1;
      cells[2].vertex(1)=2;
      cells[2].vertex(2)=4;
      cells[2].vertex(3)=6;

      cells[3].vertex(0)=1;
      cells[3].vertex(1)=2;
      cells[3].vertex(2)=7;
      cells[3].vertex(3)=3;

      cells[4].vertex(0)=1;
      cells[4].vertex(1)=5;
      cells[4].vertex(2)=7;
      cells[4].vertex(3)=6;

      cells[5].vertex(0)=1;
      cells[5].vertex(1)=2;
      cells[5].vertex(2)=6;
      cells[5].vertex(3)=7;
      boundary_data=tess.create(vertices,cells);
      boundary.allocate(6);
      boundary_condition.allocate(6);
      Point<3> center(low+(edge[0]+edge[1]+edge[2])*0.5);
      boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1],
        low+edge[2]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
        low+edge[0]+edge[1],low+edge[0]+edge[2]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[2],
        low+edge[0]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
        low+edge[1]+edge[2],low+edge[1]+edge[0]);
      boundary[4]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0],
        low+edge[1]);
      boundary[5]=OPERATOR_NEW PlaneBoundary(center,low+edge[2],
        low+edge[2]+edge[0],low+edge[2]+edge[1]);
      for (int i=0;i<2*SPACEDIM;i++) {
        boundary_condition[i]=
          OPERATOR_NEW HcurlBoundaryCondition(boundary[i]);
      }

      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<SPACEDIM> center=face_element->center(vertices);
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              break;
            }
          }
        }
      }
      tess.synchronize(boundary_data);
      break;
    }
    default:
      OBSOLETE("bad mesh_elements");
  }
#endif
#if (SPACEDIM>=2)
  {
//  TRACER_CALL(t,"checkFaceBoundaryCondition");
    tess.checkFaceBoundaryCondition(boundary_condition);
  }
#endif
#if (SPACEDIM==3)
  tess.checkLineBoundaryCondition(boundary_condition);
#endif
#ifdef DEBUG
//tess.plotShapes();
//tess.printVertices();
//tess.printElements();
# if (SPACEDIM>1)
//tess.printCodimension1Elements();
# endif
# if (SPACEDIM==3)
//tess.printCodimension2Elements();
# endif
#endif

  TransientMaxwellModel tmm;
  tess.setMaterialModel(&tmm);

//specified E at left is essential boundary condition
  Tensor<1,SPACEDIM> E;
#if (SPACEDIM==1)
  E[0]=1.;
#elif (SPACEDIM==2)
  E[1]=1.;
#else
  E[1]=1.;
#endif
  ConstantElectricFieldStrengthFunctionSystem sefsfs(E);
  HcurlBoundaryCondition *ebc=
    dynamic_cast<HcurlBoundaryCondition*>(boundary_condition[0]);
  CHECK_POINTER(ebc);
  ebc->setFunctionSystem(&sefsfs);
  ebc->makeEssential();

//specified tangential H on right is natural boundary condition
  Tensor<1,SPACEDIM> H;
#if (SPACEDIM==1)
  H[0]=1.;
#elif (SPACEDIM==2)
  H[0]=1.;
#else
  H[1]=1.;
#endif
  ConstantMagneticFieldStrengthFunctionSystem fmfsfs(H);
  HcurlBoundaryCondition *nbc=
    dynamic_cast<HcurlBoundaryCondition*>(boundary_condition[1]);
  nbc->setFunctionSystem(&fmfsfs);
  nbc->makeNatural();

//other boundaries (if they exist) are homogeneous and essential
  Tensor<1,SPACEDIM> Z; D_ASSIGN(Z[0]=0.,Z[1]=0.,Z[2]=0.);
  ConstantElectricFieldStrengthFunctionSystem zefsfs(Z);
  for (int d=2;d<2*SPACEDIM;d++) {
    HcurlBoundaryCondition *zbc=
      dynamic_cast<HcurlBoundaryCondition*>(boundary_condition[d]);
    zbc->setFunctionSystem(&zefsfs);
    zbc->makeEssential();
  }
#ifdef DEBUG
//tess.coarsestLevel()->
//  CODIMENSION1_LEVEL_CLASS::printFaceBoundaryConditions();
#endif

//(linear or) tensor product linear coordinate mapping
  MappingL mapping;

//set polynomials for each shape, face quadratures and dofs per object
  Nedelec1FiniteElement fe(degree);
//Nedelec2FiniteElement fe(degree);

//interior quadrature rules
  map<int,Quadrature<SPACEDIM> > qmap;
#if (SPACEDIM==1)
//qmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
//  GaussianQuadrature<1>::argGivenOrder(2*degree+3));
  qmap[Interval::verticesPerShape()]=NewtonCotesQuadrature<1>(
    NewtonCotesQuadrature<1>::argGivenOrder(2*degree+2));
#elif (SPACEDIM==2)
  switch (mesh_elements) {
    case QUADRILATERAL_MESH_ELEMENTS:
//    need order>=2*degree+1
      qmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<2>(
        GaussianQuadrature<2>::argGivenOrder(2*degree+3));
//    qmap[Quadrilateral::verticesPerShape()]=
//      NewtonCotesQuadrature<2>(max(2,2*degree+1));
      break;
    case TRIANGULAR_MESH_ELEMENTS:
      qmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(
        TriangleGaussianQuadrature::argGivenOrder(2*degree+3));
      break;
    default:
      OBSOLETE("bad mesh_elements");
  }
#else
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS:
//    need order>=2*degree+1
      qmap[Hexahedron::verticesPerShape()]=GaussianQuadrature<SPACEDIM>(
        GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+3));
//    qmap[Hexahedron::verticesPerShape()]=
//      NewtonCotesQuadrature<SPACEDIM>(max(2,2*degree+1));
      break;
    case PRISM_MESH_ELEMENTS:
      qmap[Prism::verticesPerShape()]=
        ProductQuadrature(
          TriangleGaussianQuadrature(
            TriangleGaussianQuadrature::argGivenOrder(2*degree+3)),
          GaussianQuadrature<1>(
            GaussianQuadrature<1>::argGivenOrder(2*degree+3)));
      break;
    case TETRAHEDRAL_MESH_ELEMENTS:
      qmap[Tetrahedron::verticesPerShape()]=TetrahedronGaussianQuadrature(
        TetrahedronGaussianQuadrature::argGivenOrder(2*degree+3));
      break;
    default:
      OBSOLETE("bad mesh_elements");
  }
#endif

//boundary quadrature rules
  map<int,Quadrature<SPACEDIM-1> > nqmap;
#if (SPACEDIM>=2)
//natural bc's
//should be exact for polynomials of shape function degree
//  plus exact solution degree-1
//  plus Jacobian degree
# if (SPACEDIM==2)
  nqmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
        GaussianQuadrature<1>::argGivenOrder(2*degree+2));
# else
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS: {
      int nv=Quadrilateral::verticesPerShape();
      nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(
        GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
      break;
    }
    case PRISM_MESH_ELEMENTS: {
      int nv=Quadrilateral::verticesPerShape();
      nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(
        GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
      nv=Triangle::verticesPerShape();
      nqmap[nv]=TriangleGaussianQuadrature(
        TriangleGaussianQuadrature::argGivenOrder(2*degree));
      break;
    }
    case TETRAHEDRAL_MESH_ELEMENTS: {
//    face shape function is polynomial of order=degree
//    exact solution is polynomial of order=solution order-1
//    TriangleProductGaussianQuadrature has order 2*n-1 
//      where n=constructor argument
      int nv=Triangle::verticesPerShape();
      nqmap[nv]=TriangleGaussianQuadrature(
        TriangleGaussianQuadrature::argGivenOrder(2*degree));
      break;
    }
    default:
      OBSOLETE("bad mesh_elements");
  }
# endif
#endif
# ifdef DEBUG
//for (map<int,Quadrature<SPACEDIM-1> >::const_iterator fi=nqmap.begin();
//fi!=nqmap.end();fi++) {
//  fi->second.printOn(cout);
//}
# endif

//graphics setup
  Palette pal;
  NumPtr<VirtualGraphTool*> plot_gt(D_CHOOSE(2,2,6));
#if (SPACEDIM==1)
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  plot_gt[0]=OPERATOR_NEW XYGraphTool(
    "electric field intensity E vs position","x","E",0.,1.,0.,1.,&cmap,
    0,0.5);
  plot_gt[1]=OPERATOR_NEW XYGraphTool(
    "magnetic field intensity H vs position","x","H",0.,1.,0.,1.,&cmap,
    0,0.5);
#elif (SPACEDIM==2)
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]={0.,0.},rhi[2]={1.,1.};
  AreaGraphTool mesh_gt("grid","x0","x1",rlo,rhi,&cmap,0,0.5);
  plot_gt[0]=OPERATOR_NEW AreaGraphTool("electric field intensity",
    "x0","x1",rlo,rhi,&cmap,0,0.5);
  plot_gt[1]=OPERATOR_NEW AreaGraphTool("magnetic field intensity",
    "x0","x1",rlo,rhi,&cmap,0,0.5);
#else
  BOOLEAN use_lighting=TRUE;
  Vector3 vlow(0.,0.,0.);
  Vector3 vhigh(1.,1.,1.);
  VolGraphTool mesh_gt("grid",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&mesh_gt);
    mesh_gt.drawBox(vlow,vhigh);
  }
  VolGraphTool *pgt0=OPERATOR_NEW VolGraphTool(
    "electric field intensity 0",vlow,vhigh,pal,use_lighting,0.5);
  plot_gt[0]=pgt0;
  { VolGraphTool::DrawDelimiter dd(TRUE,pgt0);
    pgt0->drawBox(vlow,vhigh);
  }
  VolGraphTool *pgt1=OPERATOR_NEW VolGraphTool(
    "electric field intensity 1",vlow,vhigh,pal,use_lighting,0.5);
  plot_gt[1]=pgt1;
  { VolGraphTool::DrawDelimiter dd(TRUE,pgt1);
    pgt1->drawBox(vlow,vhigh);
  }
  VolGraphTool *pgt2=OPERATOR_NEW VolGraphTool(
    "electric field intensity 2",vlow,vhigh,pal,use_lighting,0.5);
  plot_gt[2]=pgt1;
  { VolGraphTool::DrawDelimiter dd(TRUE,pgt2);
    pgt2->drawBox(vlow,vhigh);
  }
  VolGraphTool *pgt3=OPERATOR_NEW VolGraphTool(
    "magnetic field intensity 0",vlow,vhigh,pal,use_lighting,0.5);
  plot_gt[3]=pgt3;
  { VolGraphTool::DrawDelimiter dd(TRUE,pgt3);
    pgt3->drawBox(vlow,vhigh);
  }
  VolGraphTool *pgt4=OPERATOR_NEW VolGraphTool(
    "magnetic field intensity 1",vlow,vhigh,pal,use_lighting,0.5);
  plot_gt[4]=pgt4;
  { VolGraphTool::DrawDelimiter dd(TRUE,pgt4);
    pgt4->drawBox(vlow,vhigh);
  }
  VolGraphTool *pgt5=OPERATOR_NEW VolGraphTool(
    "magnetic field intensity 2",vlow,vhigh,pal,use_lighting,0.5);
  plot_gt[5]=pgt5;
  { VolGraphTool::DrawDelimiter dd(TRUE,pgt5);
    pgt5->drawBox(vlow,vhigh);
  }
//VolGraphTool surface_gt("solution surface",vlow,vhigh,pal,use_lighting,
//  0.5);
//{ VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
//  surface_gt.drawBox(vlow,vhigh);
//}
#endif

#ifdef DEBUG
//  tess.printVertices();
//  tess.printElements();
//  tess.printCodimension1Elements();
//  tess.printCodimension2Elements();
//  for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
//    LEVEL_CLASS *present_level=tess.presentLevel();
//    const SHAPE_CLASS *element=present_level->getShape();
//    element->printOn(cout);
//    for (int v=0;v<element->numberVertices();v++) {
//      cout << "\tvertex[" << v << "] = " 
//           << tess.getVertices()[element->vertexIndex(v)] << endl;
//    }
//  }
#endif

//  data structure to keep track of degrees of freedom in solver
    HcurlDofHandler hcurl_dof_handler(tess,fe);
    hcurl_dof_handler.distributeDofs();
    int n_dofs=hcurl_dof_handler.numberDofs();
    int n_soln_dofs=hcurl_dof_handler.totalSolutionDofs();
    int ns=D_CHOOSE(n_dofs,n_dofs+n_soln_dofs,3*n_dofs);

#ifdef DEBUG
//  hcurl_dof_handler.printDofs();

    cout << "\tnumber of active elements = " 
         << tess.numberActiveShapes() << endl;
    cout << "\tNumber of degrees of freedom = " << n_dofs << endl;
    cout << "\tassemble initial linear system..." << endl;
#endif
//  data structure to keep track of linear system storage
    int band_width=hcurl_dof_handler.maxCouplingsBetweenDofs();
#ifdef DEBUG
//  cout << "\tband_width = " << band_width << endl;
#endif
    SparsityPattern *mass_sparsity_pattern=OPERATOR_NEW SparsityPattern(
      ns,ns,band_width,true);
    SparsityPattern *stiffness_sparsity_pattern=
      OPERATOR_NEW SparsityPattern(ns,ns,band_width,true);
    hcurl_dof_handler.makeTransientSparsityPatterns(
      *mass_sparsity_pattern,*stiffness_sparsity_pattern);

    mass_sparsity_pattern->compress();
    stiffness_sparsity_pattern->compress();
#ifdef DEBUG
/*
    for (int i=0;i<mass_sparsity_pattern->nRows();i++) {
      cout << "\tmass_sparsity_pattern[" << i << "]: ";
      for (int j=mass_sparsity_pattern->getRowstartIndex(i);
      j<mass_sparsity_pattern->getRowstartIndex(i+1);j++) {
        cout << mass_sparsity_pattern->getColumnNumber(j) << " ";
      }
      cout << endl;
    }
    for (int i=0;i<stiffness_sparsity_pattern->nRows();i++) {
      cout << "\tstiffness_sparsity_pattern[" << i << "]: ";
      for (int j=stiffness_sparsity_pattern->getRowstartIndex(i);
      j<stiffness_sparsity_pattern->getRowstartIndex(i+1);j++) {
        cout << stiffness_sparsity_pattern->getColumnNumber(j) << " ";
      }
      cout << endl;
    }
*/
#endif
//  storage for mass matrix, stiffness matrix, right-hand side and solution
    SparseMatrix system_mass_matrix(mass_sparsity_pattern);
    SparseMatrix system_stiffness_matrix(stiffness_sparsity_pattern);
    Vector system_rhs(ns,0.);
    Vector initial_solution(ns,0.);

//  assemble initial linear system
    hcurl_dof_handler.assembleLinearSystemForInitialData(mapping,qmap,
      system_mass_matrix,system_rhs);
#ifdef DEBUG
/*
    for (int i=0;i<system_mass_matrix.size(0);i++) {
      cout << "\tsystem_mass_matrix(" << i << ", "; 
      int count=0;
      for (int j=mass_sparsity_pattern->getRowstartIndex(i);
      j<mass_sparsity_pattern->getRowstartIndex(i+1);j++) {
        if (count++ >0) cout << " , ";
        cout << mass_sparsity_pattern->getColumnNumber(j) << " = "
             << system_mass_matrix(i,
                  mass_sparsity_pattern->getColumnNumber(j));
      }
      cout << " )" << endl;
    }
    for (int i=0;i<system_rhs.size();i++) {
      cout << "\tsystem_rhs(" << i << ") = " << system_rhs[i] << endl;
    }
*/
#endif
//  compute essential boundary values at mapped points on bdry
//  calls FunctionSystem in BoundaryCondition's to set boundary values
    map<int,REAL> boundary_values;
    {
//    TRACER_CALL(t,"interpolateBoundaryValues");
#ifdef TIMING
//    Timer y(z3);
#endif
      hcurl_dof_handler.interpolateBoundaryValues(mapping,boundary_values);
#ifdef DEBUG
/*
      for (map<int,REAL>::const_iterator bv_iterator=
      boundary_values.begin();bv_iterator!=boundary_values.end();
      bv_iterator++) {
        cout << "\tboundary_values[" << bv_iterator->first << "] = "
             << bv_iterator->second << endl;
      }
*/
#endif
    }

//  modify initial linear system and rhs for essential bc's
    bool preserve_symmetry=true;
    hcurl_dof_handler.applyBoundaryValues(boundary_values,
      system_mass_matrix,system_rhs,preserve_symmetry);
    hcurl_dof_handler.applyBoundaryValues(boundary_values,
      initial_solution);
#ifdef DEBUG
# if (SPACEDIM==1)
/*
    int tsd=hcurl_dof_handler.totalSolutionDofs();
    REAL dx=2./static_cast<REAL>(nelements);
    Point<SPACEDIM> x; x[0]=-1.;
    for (int i=0;i<tsd;i++) {
      Tensor<1,SPACEDIM> efs=exactElectricFieldStrength(x,0.);
      initial_solution[i]=dx*efs[0]; // dx*efs = tcurl{efs}
      x[0]+=dx;
    }
    x[0]=-1.+dx*0.5;
    int n=initial_solution.size()-tsd;
    for (int i=0;i<n;i++) {
      initial_solution[i+tsd]=exactMagneticFieldStrength(x,0.);
      x[0]+=dx;
    }
    for (int i=0;i<initial_solution.size();i++) {
      cout << "\tinitial_solution(" << i << ") = "
           << initial_solution[i] << endl;
    }
    Vector Axmb(initial_solution.size());
    system_mass_matrix.vmult(Axmb,initial_solution);
    Axmb-=system_rhs;
    for (int i=0;i<Axmb.size();i++) {
      cout << "\tAxmb(" << i << ") = " << Axmb[i] << endl;
    }
*/   
# endif
/*
    for (int i=0;i<system_mass_matrix.size(0);i++) {
      cout << "\tsystem_mass_matrix(" << i << ", "; 
      int count=0;
      for (int j=mass_sparsity_pattern->getRowstartIndex(i);
      j<mass_sparsity_pattern->getRowstartIndex(i+1);j++) {
        if (count++ >0) cout << " , ";
        cout << mass_sparsity_pattern->getColumnNumber(j) << " = "
             << system_mass_matrix(i,
                  mass_sparsity_pattern->getColumnNumber(j));
      }
      cout << " )" << endl;
    }
    for (int i=0;i<system_rhs.size();i++) {
      cout << "\tsystem_rhs(" << i << ") = " << system_rhs[i] << endl;
    }
*/
#endif
    PreconditionIdentity pi;
    RelativeErrorControl cg_solver_control(system_rhs.size(),
      DBL_EPSILON*1.e6,false,false);
    CGSolver cg(cg_solver_control);
    {
//    TRACER_CALL(t,"solve");
#ifdef DEBUG
      cout << "\tsolve initial linear system..." << endl;
#endif
#ifdef TIMING
//    Timer y(z5);
#endif
      cg.solve(system_mass_matrix,initial_solution,system_rhs,pi);
    }
#ifdef DEBUG
/*
    for (int i=0;i<initial_solution.size();i++) {
      cout << "\tsolution(" << i << ") = " << initial_solution[i]
           << endl;
    }
*/
#endif
//  compute solution values for graphical display
    HcurlDataOut data_out(hcurl_dof_handler,mapping);
    int total_solution_dofs=hcurl_dof_handler.totalSolutionDofs();
    Vector *efield=initial_solution.subVector(0,total_solution_dofs);
    Vector *hfield=initial_solution.subVector(total_solution_dofs,
      n_dofs-total_solution_dofs);
    data_out.addDataVector(*efield,"electric field intensity",
      HcurlDataOut::TYPE_DOF_DATA);
    data_out.addDataVector(*hfield,"magnetic field intensity",
      HcurlDataOut::TYPE_MULTIPLIER_DATA);
//  int n_subdivisions=degree+1;
    int n_subdivisions=5;
    data_out.buildPatches(n_subdivisions);
#ifdef DEBUG
//  data_out.printOn(cout);
#endif
    bool quit_button=false;
#if (SPACEDIM>1)
    data_out.drawGrid(&mesh_gt,quit_button);
#endif
    data_out.plot("electric field intensity",plot_gt,quit_button);
    NumPtr<VirtualGraphTool*> plot_gt2=plot_gt+D_CHOOSE(1,1,3);
    data_out.plot("magnetic field intensity",plot_gt2,quit_button);
//  data_out.plotSurface("solution",&surface_gt);
#if (SPACEDIM==1)
    {
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#elif (SPACEDIM==2)
    {
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#endif
    delete efield; efield=0;
    delete hfield; hfield=0;
    data_out.clearInputDataReferences();

#ifdef DEBUG
    cout << "\tassemble transient linear system..." << endl;
#endif
    system_mass_matrix.reinit();
    hcurl_dof_handler.assembleTransientLinearSystem(mapping,qmap,nqmap,
      system_mass_matrix,system_stiffness_matrix,system_rhs);
#ifdef DEBUG
//
    for (int i=0;i<system_mass_matrix.size(0);i++) {
      cout << "\tsystem_mass_matrix(" << i << ", "; 
      int count=0;
      for (int j=mass_sparsity_pattern->getRowstartIndex(i);
      j<mass_sparsity_pattern->getRowstartIndex(i+1);j++) {
        if (count++ >0) cout << " , ";
        cout << mass_sparsity_pattern->getColumnNumber(j) << " = "
             << system_mass_matrix(i,
                  mass_sparsity_pattern->getColumnNumber(j));
      }
      cout << " )" << endl;
    }
    for (int i=0;i<system_stiffness_matrix.size(0);i++) {
      cout << "\tsystem_stiffness_matrix(" << i << ", "; 
      int count=0;
      for (int j=stiffness_sparsity_pattern->getRowstartIndex(i);
      j<stiffness_sparsity_pattern->getRowstartIndex(i+1);j++) {
        if (count++ >0) cout << " , ";
        cout << stiffness_sparsity_pattern->getColumnNumber(j) << " = "
             << system_stiffness_matrix(i,
                  stiffness_sparsity_pattern->getColumnNumber(j));
      }
      cout << endl;
    }
    for (int i=0;i<system_rhs.size();i++) {
      cout << "\tsystem_rhs(" << i << ") = " << system_rhs[i] << endl;
    }
//
#endif
//  modify transient linear system and rhs for essential bc's
    hcurl_dof_handler.applyBoundaryValues(boundary_values,
      system_stiffness_matrix,system_rhs,preserve_symmetry);
#ifdef DEBUG
//
    for (int i=0;i<system_mass_matrix.size(0);i++) {
      cout << "\tsystem_mass_matrix(" << i << ", "; 
      int count=0;
      for (int j=mass_sparsity_pattern->getRowstartIndex(i);
      j<mass_sparsity_pattern->getRowstartIndex(i+1);j++) {
        if (count++ >0) cout << " , ";
        cout << mass_sparsity_pattern->getColumnNumber(j) << " = "
             << system_mass_matrix(i,
                  mass_sparsity_pattern->getColumnNumber(j));
      }
      cout << " )" << endl;
    }
    for (int i=0;i<system_stiffness_matrix.size(0);i++) {
      cout << "\tsystem_stiffness_matrix(" << i << ", "; 
      int count=0;
      for (int j=stiffness_sparsity_pattern->getRowstartIndex(i);
      j<stiffness_sparsity_pattern->getRowstartIndex(i+1);j++) {
        if (count++ >0) cout << " , ";
        cout << stiffness_sparsity_pattern->getColumnNumber(j) << " = "
             << system_stiffness_matrix(i,
                  stiffness_sparsity_pattern->getColumnNumber(j));
      }
      cout << endl;
    }
    for (int i=0;i<system_rhs.size();i++) {
      cout << "\tsystem_rhs(" << i << ") = " << system_rhs[i] << endl;
    }

//  int offset=system_stiffness_matrix.size(0);
//  Vector x(offset); x=0.;
//  Vector Ax(offset);
//  system_stiffness_matrix.vmult(Ax,x);
//  for (int i=0;i<Ax.size();i++) {
//    cout << "\tAx[" << i << "] = " << Ax[i] << endl;
//  }
//
#endif
//  prepare to integrate in time
    NumPtr<Vector*> solution;
    TimeControl tc(20,10.,DBL_EPSILON);
    GMRESSolver gmres(cg_solver_control);
    CrankNicolsonTimeIntegrator ti(tc,system_mass_matrix,
      system_stiffness_matrix,system_rhs,boundary_values,pi,gmres,
      solution);
    *solution[0]=initial_solution;

    REAL dt=0.1/(static_cast<REAL>(nelements)*tmm.speedOfLight());
    while (true) {
//    TRACER_CALL(t,"integrate");
#ifdef DEBUG
      cout << "\tstep = " << tc.getStep() << " time = " << tc.getTime()
           << endl;
#endif
#ifdef TIMING
//    Timer y(z5);
#endif
# if (SPACEDIM==1)
//
      REAL time=tc.getTime();
      int tsd=hcurl_dof_handler.totalSolutionDofs();
      REAL dx=2./static_cast<REAL>(nelements);
      Point<SPACEDIM> x; x[0]=-1.;
      Vector &old_solution=*solution[0];
      Vector &new_solution=*solution[1];
      for (int i=0;i<tsd;i++) {
        Tensor<1,SPACEDIM> efs=exactElectricFieldStrength(x,time);
        old_solution[i]=dx*efs[0]; // dx*efs = tcurl{efs}
        efs=exactElectricFieldStrength(x,time+dt);
        new_solution=dx*efs[0];
        x[0]+=dx;
      }
      x[0]=-1.+dx*0.5;
      int n=initial_solution.size()-tsd;
      for (int i=0;i<n;i++) {
        old_solution[i+tsd]=exactMagneticFieldStrength(x,time);
        new_solution[i+tsd]=exactMagneticFieldStrength(x,time+dt);
        x[0]+=dx;
      }
      for (int i=0;i<old_solution.size();i++) {
        cout << "\told_solution(" << i << ") = "
             << old_solution[i] << endl;
      }
      for (int i=0;i<new_solution.size();i++) {
        cout << "\tnew_solution(" << i << ") = "
             << new_solution[i] << endl;
      }
      Vector soln_ave(old_solution.size());
      soln_ave=old_solution;
      soln_ave+=new_solution;
      soln_ave.scale(0.5);
      Vector b(old_solution.size());
      system_stiffness_matrix.vmult(b,soln_ave);
      b.scale(-1.);
      b+=system_rhs;
      Vector dudt(old_solution.size());
      dudt=new_solution;
      dudt-=old_solution;
      dudt.scale(1./dt);
      Vector Axmb(old_solution.size());
      system_mass_matrix.vmult(Axmb,dudt);
      Axmb-=b;
      for (int i=0;i<Axmb.size();i++) {
        cout << "\tAxmb(" << i << ") = " << Axmb[i] << endl;
      }
//   
# endif
      TIME_INTEGRATOR_STATE state=ti.takeStep(dt);
#ifdef DEBUG
/*
      for (int i=0;i<solution[0]->size();i++) {
        cout << "\tsolution(" << i << ") = " << (*solution[0])[i]
             << endl;
      }
*/
#endif
      Vector *efield=solution[0]->subVector(0,total_solution_dofs);
      Vector *hfield=solution[0]->subVector(total_solution_dofs,
        n_dofs-total_solution_dofs);
      data_out.replaceDataVector(*efield,"electric field intensity");
      data_out.replaceDataVector(*hfield,"magnetic field intensity");
      data_out.reBuildPatches();
      data_out.plot("electric field intensity",plot_gt,quit_button);
      data_out.plot("magnetic field intensity",plot_gt2,quit_button);
/*
//    data_out.plotSurface("solution",&surface_gt);
*/
      delete efield; efield=0;
      delete hfield; hfield=0;
      data_out.clearInputDataReferences();
      if (state==TIME_INTEGRATOR_DONE) break;
    }
#if (SPACEDIM==1)
    {
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#elif (SPACEDIM==2)
    {
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#endif

  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
//  if (boundary_condition[i]->isNatural()) {
//    delete boundary_condition[i]->getFunctionSystem();
//  }
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  for (int d=0;d<plot_gt.getNumber();d++) {
    delete plot_gt[d];
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
