#ifndef MOONEY_RIVLIN_H
#define MOONEY_RIVLIN_H "$Header:$"

#include "NonlinearDofHandler.H"
#include "MaterialModel.H"
#include "Types.H"

class MooneyRivlinMaterialModel : public NonlinearMaterialModel {
  private:
    REAL lame;
    REAL shear;
    Point<SPACEDIM> (*bodyForce)(const Point<SPACEDIM>&);

    Tensor<2,3> greenDeformationTensor(const NumPtr<Tensor<1,SPACEDIM> >&)
      const;
    void freeEnergy(const Tensor<2,3>&,NumPtr<REAL>&) const;
    Tensor<2,3> stressTensor(const NumPtr<Tensor<1,SPACEDIM> >&) const;
  public:
    MooneyRivlinMaterialModel(REAL l,REAL s,
    Point<SPACEDIM> (*bf)(const Point<SPACEDIM>&)) : lame(l),shear(s),
      bodyForce(bf) {}
    void elementResidual(const NumPtr<Point<SPACEDIM> > &q_points,
      const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
      const Array<2,Tensor<1,SPACEDIM> > &sg,
      const NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &input_gradients,
      Vector &cell_output) const;
#ifdef DEBUG
    bool checkElementVolumes(DofHandler&,const Vector&) const;
#endif
    void printOn(ostream&) const;
};

//the following is used for debugging the nonlinear solver
class LinearElasticityMaterialModel : public NonlinearMaterialModel {
  private:
    REAL lame;
    REAL shear;
    Point<SPACEDIM> (*bodyForce)(const Point<SPACEDIM>&);

    Tensor<2,3> infinitesimalStrainTensor(
      const NumPtr<Tensor<1,SPACEDIM> >&) const;
    Tensor<2,3> stressTensor(
      const NumPtr<Tensor<1,SPACEDIM> >&) const;
  public:
    LinearElasticityMaterialModel(REAL l,REAL s,
    Point<SPACEDIM> (*bf)(const Point<SPACEDIM>&),
    Point<SPACEDIM> (*bofo)(const Point<SPACEDIM>&)) : lame(l),shear(s),
      bodyForce(bf) {}
    void elementResidual(const NumPtr<Point<SPACEDIM> > &q_points,
      const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
      const Array<2,Tensor<1,SPACEDIM> > &sg,
      const NumPtr<NumPtr<Tensor<1,SPACEDIM> > > &input_gradients,
      Vector &cell_output) const;
#ifdef DEBUG
    bool checkElementVolumes(DofHandler&,const Vector&) const;
    void checkLinearity(NonlinearDofHandler&) const;
    void checkJacobian(NonlinearDofHandler&) const;
#endif
    void printOn(ostream&) const;
};
#endif
