// "$Header: /home/faculty/johnt/cvs/deal_new/examples/RectangularDomain.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>

#include "Boundary.H"
#include "HdivDataOut.H"
#include "HdivDofHandler.H"
#include "HdivFiniteElement.H"
#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "TimedObject.H"
#include "Tracer.H"
#include "Vector.H"

#if (SPACEDIM==1)
# define N_CYCLES 1
#elif (SPACEDIM==2)
# define N_CYCLES 1
  enum MESH_ELEMENTS { 
    QUADRILATERAL_MESH_ELEMENTS,
    TRIANGLE_MESH_ELEMENTS 
  };
//MESH_ELEMENTS mesh_elements=QUADRILATERAL_MESH_ELEMENTS;
  MESH_ELEMENTS mesh_elements=TRIANGLE_MESH_ELEMENTS;
#else
# define N_CYCLES 1
  enum MESH_ELEMENTS { 
    HEXAHEDRAL_MESH_ELEMENTS,
    PRISM_MESH_ELEMENTS,
    TETRAHEDRAL_MESH_ELEMENTS 
  };
  MESH_ELEMENTS mesh_elements=HEXAHEDRAL_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=PRISM_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=TETRAHEDRAL_MESH_ELEMENTS;
//int dir=0;
#endif
int degree=0;
int nelements=4;

//specify pressure
class SpecifiedPressureFunctionSystem : public FunctionSystem {
  private:
    REAL val;
  protected:
    explicit SpecifiedPressureFunctionSystem(const FunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    SpecifiedPressureFunctionSystem(REAL v) : val(v) {;}
    virtual ~SpecifiedPressureFunctionSystem() {;}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      return val; // pressure in injector on left
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      function_values[0]=val;
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,
    int) const {
      OBSOLETE("why called?");
      Tensor<1,SPACEDIM> g;
      return g;
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
    virtual void printOn(ostream &os) const {
      os << "SpecifiedPressureFunctionSystem:" << endl;
      FunctionSystem::printOn(os);
    }
};
//produce at specified velocity on right
class SpecifiedNormalVelocityFunctionSystem : public FunctionSystem {
  private:
    REAL val;
  protected:
    explicit SpecifiedNormalVelocityFunctionSystem(
    const SpecifiedNormalVelocityFunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    SpecifiedNormalVelocityFunctionSystem(REAL v) : val(v) {}
    virtual ~SpecifiedNormalVelocityFunctionSystem() {}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      return val;
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      function_values[0]=val;
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("should not be called");
      Tensor<1,SPACEDIM> g;
      return g;
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("should not be called");
    }
    virtual void printOn(ostream &os) const {
      os << "SpecifiedNormalVelocityFunctionSystem: val = " << val
         << endl;
      FunctionSystem::printOn(os);
    }
};

class PorousMaterialModel : public HdivMaterialModel {
  private:
    Tensor<1,SPACEDIM> gravity;
    REAL density;
    REAL permeability;
    REAL viscosity;

    void elementStiffnessMatrixAndRHS(const NumPtr<Point<SPACEDIM> >&,
    const NumPtr<REAL>&,const Array<2,REAL>&,
    const Array<2,Tensor<1,SPACEDIM> >&,Matrix&,Vector&) const {
      OBSOLETE("does not apply to this class");
    }
    void elementStiffnessMatrixAndRHS(const NumPtr<Point<3> >&,
    const NumPtr<double>&,const Array<2,Tensor<1,3> >*,
    const Array<2,Tensor<1,3> >*,Matrix&,Vector&) const {
      OBSOLETE("does not apply to this class");
    }
  public:
    PorousMaterialModel() : density(1.),permeability(1.),viscosity(1.) {
      gravity.initialize(0.); // gravity[SPACEDIM-1]=1.;
    }
    void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,
    const Array<2,Tensor<1,SPACEDIM> > *sv,const Array<2,REAL> *sd,
    const Array<2,Tensor<2,SPACEDIM> > *sg,const Array<2,REAL> *msv,
    Matrix &cell_matrix,Vector &cell_rhs) const {
      int dpc=sv->size(0);
      int mdpc=msv->size(0);
      CHECK_SAME(cell_matrix.size(0),dpc+mdpc);
      CHECK_SAME(cell_matrix.size(0),cell_rhs.size());
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        for (int i=0;i<dpc;i++) {
          Tensor<1,SPACEDIM> svi=(*sv)(i,q);
          REAL sdi=(*sd)(i,q);
          svi*=jxw;
          sdi*=jxw;
          for (int j=0;j<dpc;j++) {
            cell_matrix(i,j)+=(svi*(*sv)(j,q))*viscosity/permeability;
          }
          for (int j=0;j<mdpc;j++) cell_matrix(i,j+dpc)-=sdi*(*msv)(j,q);
#if (SPACEDIM>1)
          cell_rhs[i]-=(svi*gravity)*density;
#endif
        }
      }
      for (int i=0;i<dpc;i++) {
        for (int j=0;j<mdpc;j++) {
          cell_matrix(j+dpc,i)=cell_matrix(i,j+dpc);
        }
      }
    }
};

#ifdef TIMING
//the most expensive routines are the linear system solve and
//  interpolateBoundaryValues
//the latter requires computation of the shape functions on an element
//  at iteratively computed canonical points corresponding to boundary:
//TimedObjectList tl;
#endif

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
#ifdef TIMING
//TimedObject *z3=
//  OPERATOR_NEW TimedObject("main interpolateBoundaryValues",tl);
//TimedObject *z5=OPERATOR_NEW TimedObject("main solve",tl);
//TimedObject *z10=OPERATOR_NEW TimedObject("main refine",tl);
#endif
  Tessellation tess;

//rectangular domain consisting of one interval/quadrilateral/hexagon
  Point<SPACEDIM> low(D_DECL(-1.,-1.,-1.));
//Point<SPACEDIM> low(D_DECL(0.,0.,0.));
  Point<SPACEDIM> edge[SPACEDIM];
    edge[0]=Point<SPACEDIM>(D_DECL(2.,0.,0.));
//  edge[0]=Point<SPACEDIM>(D_DECL(1.,0.,0.));
#if (SPACEDIM>1)
    edge[1]=Point<SPACEDIM>(D_DECL(0.,2.,0.));
//  edge[1]=Point<SPACEDIM>(D_DECL(0.,1.,0.));
#endif
#if (SPACEDIM==3)
    edge[2]=Point<SPACEDIM>(D_DECL(0.,0.,2.));
#endif
  int number_elements[SPACEDIM]={D_DECL(nelements,nelements,nelements)};
  bool set_face_ids=true;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
#if (SPACEDIM==1)
  BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,boundary,boundary_condition);
#elif (SPACEDIM==2)
  BoundaryData *boundary_data=0;
  if (mesh_elements==QUADRILATERAL_MESH_ELEMENTS) {
    boundary_data=generateParallelpiped(tess,low,edge,number_elements,
      set_face_ids,boundary,boundary_condition);
  } else {
    boundary_data=generateParallelpipedTriangles(tess,low,edge,
      number_elements,set_face_ids,boundary,boundary_condition);
  }
#else
  BoundaryData *boundary_data=0;
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS:
      boundary_data=generateParallelpiped(tess,low,edge,number_elements,
        set_face_ids,boundary,boundary_condition);
      break;
    case PRISM_MESH_ELEMENTS: {
      NumPtr<Point<3> > vertices(10);
      for (int i=0;i<vertices.getNumber();i++) vertices[i]=low;
      vertices[1]+=edge[0];
      vertices[2]+=edge[0]*0.5; vertices[2]+=edge[1]*0.5; 
      vertices[3]+=edge[1];
      vertices[4]+=edge[0]; vertices[4]+=edge[1];
      vertices[5]+=edge[2];
      vertices[6]+=edge[0]; vertices[6]+=edge[2];
      vertices[7]+=edge[0]*0.5; vertices[7]+=edge[1]*0.5; 
                                vertices[7]+=edge[2];
      vertices[8]+=edge[1]; vertices[8]+=edge[2];
      vertices[9]+=edge[0]; vertices[9]+=edge[1]; vertices[9]+=edge[2];
      NumPtr<CellData> cells(4);
      for (int c=0;c<cells.getNumber();c++) cells[c].allocate(6);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;
      cells[0].vertex(3)=5;
      cells[0].vertex(4)=6;
      cells[0].vertex(5)=7;

      cells[1].vertex(0)=0;
      cells[1].vertex(1)=2;
      cells[1].vertex(2)=3;
      cells[1].vertex(3)=5;
      cells[1].vertex(4)=7;
      cells[1].vertex(5)=8;

      cells[2].vertex(0)=2;
      cells[2].vertex(1)=4;
      cells[2].vertex(2)=3;
      cells[2].vertex(3)=7;
      cells[2].vertex(4)=9;
      cells[2].vertex(5)=8;

      cells[3].vertex(0)=1;
      cells[3].vertex(1)=4;
      cells[3].vertex(2)=2;
      cells[3].vertex(3)=6;
      cells[3].vertex(4)=9;
      cells[3].vertex(5)=7;
      boundary_data=tess.create(vertices,cells);
      boundary.allocate(6);
      boundary_condition.allocate(6);
      Point<3> center(low+(edge[0]+edge[1]+edge[2])*0.5);
      boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1],
        low+edge[2]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
        low+edge[0]+edge[1],low+edge[0]+edge[2]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[2],
        low+edge[0]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
        low+edge[1]+edge[2],low+edge[1]+edge[0]);
      boundary[4]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0],
        low+edge[1]);
      boundary[5]=OPERATOR_NEW PlaneBoundary(center,low+edge[2],
        low+edge[2]+edge[0],low+edge[2]+edge[1]);
      for (int i=0;i<2*SPACEDIM;i++) {
        boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
      }
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<SPACEDIM> center=face_element->center(vertices);
          bool found=false;
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              found=true;
              break;
            }
          }
        }
      }
      tess.synchronize(boundary_data);
      break;
    }
    case TETRAHEDRAL_MESH_ELEMENTS: {
      NumPtr<Point<3> > vertices(8);
      for (int i=0;i<vertices.getNumber();i++) vertices[i]=low;
      vertices[1]+=edge[0];
      vertices[2]+=edge[1];
      vertices[3]+=edge[0]; vertices[3]+=edge[1];
      vertices[4]+=edge[2];
      vertices[5]+=edge[0]; vertices[5]+=edge[2];
      vertices[6]+=edge[1]; vertices[6]+=edge[2];
      vertices[7]+=edge[0]; vertices[7]+=edge[1]; vertices[7]+=edge[2];
      NumPtr<CellData> cells(6);
      for (int c=0;c<cells.getNumber();c++) cells[c].allocate(4);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;
      cells[0].vertex(3)=4;

      cells[1].vertex(0)=1;
      cells[1].vertex(1)=4;
      cells[1].vertex(2)=5;
      cells[1].vertex(3)=6;

      cells[2].vertex(0)=1;
      cells[2].vertex(1)=2;
      cells[2].vertex(2)=4;
      cells[2].vertex(3)=6;

      cells[3].vertex(0)=1;
      cells[3].vertex(1)=2;
      cells[3].vertex(2)=7;
      cells[3].vertex(3)=3;

      cells[4].vertex(0)=1;
      cells[4].vertex(1)=5;
      cells[4].vertex(2)=7;
      cells[4].vertex(3)=6;

      cells[5].vertex(0)=1;
      cells[5].vertex(1)=2;
      cells[5].vertex(2)=6;
      cells[5].vertex(3)=7;
      boundary_data=tess.create(vertices,cells);
      boundary.allocate(6);
      boundary_condition.allocate(6);
      Point<3> center(low+(edge[0]+edge[1]+edge[2])*0.5);
      boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1],
        low+edge[2]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
        low+edge[0]+edge[1],low+edge[0]+edge[2]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[2],
        low+edge[0]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
        low+edge[1]+edge[2],low+edge[1]+edge[0]);
      boundary[4]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0],
        low+edge[1]);
      boundary[5]=OPERATOR_NEW PlaneBoundary(center,low+edge[2],
        low+edge[2]+edge[0],low+edge[2]+edge[1]);
      for (int i=0;i<2*SPACEDIM;i++) {
        boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
      }

      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<SPACEDIM> center=face_element->center(vertices);
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              break;
            }
          }
        }
      }
      tess.synchronize(boundary_data);
      break;
    }
    default:
      OBSOLETE("bad mesh_elements");
  }
#endif
#if (SPACEDIM>1)
  {
    tess.checkFaceBoundaryCondition(boundary_condition);
  }
#endif
#if (SPACEDIM==3)
  tess.checkLineBoundaryCondition(boundary_condition);
#endif
#ifdef DEBUG
//tess.plotShapes();
//tess.printVertices();
//tess.printElements();
# if (SPACEDIM>1)
//tess.printCodimension1Elements();
# endif
#endif

  PorousMaterialModel pmm;
  tess.setMaterialModel(&pmm);

//specified pressure at left is natural boundary condition
  SpecifiedPressureFunctionSystem spfs(-1.); // p=1, sign flip is for Darcy
  boundary_condition[0]->setFunctionSystem(&spfs);
  boundary_condition[0]->makeNatural();

//specified normal velocity on right is essential boundary condition
//HdivBoundaryCondition is essential in normal direction by design
  SpecifiedNormalVelocityFunctionSystem snvfs(1.);
  boundary_condition[1]->setFunctionSystem(&snvfs);
  boundary_condition[1]->makeEssential();
#if (SPACEDIM>1)
  SpecifiedNormalVelocityFunctionSystem znvfs(0.);
  for (int d=2;d<2*SPACEDIM;d++) {
    boundary_condition[d]->setFunctionSystem(&znvfs);
    boundary_condition[d]->makeEssential();
  }
#endif

//(linear or) tensor product linear coordinate mapping
  MappingL mapping;

//polynomials to represent solution of problem
  RTFiniteElement fe(degree);
//BDFMFiniteElement fe(degree);

//quadature rules
  map<int,Quadrature<SPACEDIM> > qmap;
#if (SPACEDIM==1)
//need order>=2*degree+1
  qmap[Interval::verticesPerShape()]=GaussianQuadrature<SPACEDIM>(
    GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+3));
//qmap[2]=NewtonCotesQuadrature<1>(3); // simpson rule
#elif (SPACEDIM==2)
  if (mesh_elements==QUADRILATERAL_MESH_ELEMENTS) {
//  need order>=2*degree+1
    qmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<SPACEDIM>(
      GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+3));
  } else {
    qmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(
      TriangleGaussianQuadrature::argGivenOrder(2*degree+3)); 
//  qmap[Triangle::verticesPerShape()]=
//    TriangleFullNewtonCotesQuadrature(5);
  }
#else
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS:
//    need order>=2*degree+1
      qmap[Hexahedron::verticesPerShape()]=GaussianQuadrature<SPACEDIM>(
        GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+3));
//    qmap[Hexahedron::verticesPerShape()]=
//      NewtonCotesQuadrature<SPACEDIM>(max(2,2*degree+1));
      break;
    case PRISM_MESH_ELEMENTS:
      qmap[Prism::verticesPerShape()]=
        ProductQuadrature(
          TriangleGaussianQuadrature(
            TriangleGaussianQuadrature::argGivenOrder(2*degree+3)),
          GaussianQuadrature<1>(
            GaussianQuadrature<1>::argGivenOrder(2*degree+3)));
      break;
    case TETRAHEDRAL_MESH_ELEMENTS:
      qmap[Tetrahedron::verticesPerShape()]=TetrahedronGaussianQuadrature(
        TetrahedronGaussianQuadrature::argGivenOrder(2*degree+3));
      break;
    default:
      OBSOLETE("bad mesh_elements");
  }
#endif

  map<int,Quadrature<SPACEDIM-1> > nqmap;
#if (SPACEDIM>1)
//natural bc's
//should be exact for polynomials of shape function degree
//  plus exact solution degree-1
//  plus Jacobian degree
# if (SPACEDIM==2)
  {
    int nv=Interval::verticesPerShape();
    if (mesh_elements==QUADRILATERAL_MESH_ELEMENTS) {
      nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(
        GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
    } else {
      int nv=Interval::verticesPerShape();
      nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(
        GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
    }
  }
# else
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS: {
      int nv=Quadrilateral::verticesPerShape();
      nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(
        GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
      break;
    }
    case PRISM_MESH_ELEMENTS: {
      int nv=Quadrilateral::verticesPerShape();
      nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(
        GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
      nv=Triangle::verticesPerShape();
      nqmap[nv]=TriangleGaussianQuadrature(
        TriangleGaussianQuadrature::argGivenOrder(2*degree));
      break;
    }
    case TETRAHEDRAL_MESH_ELEMENTS: {
//    face shape function is polynomial of order=degree
//    exact solution is polynomial of order=solution order-1
//    TriangleProductGaussianQuadrature has order 2*n-1 
//      where n=constructor argument
      int nv=Triangle::verticesPerShape();
      nqmap[nv]=TriangleGaussianQuadrature(
        TriangleGaussianQuadrature::argGivenOrder(2*degree));
      break;
    }
    default:
      OBSOLETE("bad mesh_elements");
  }
# endif
#endif

  int n_active_cells[N_CYCLES];
  int n_dofs[N_CYCLES];
  REAL log_n_dofs[N_CYCLES];
  REAL L2_error[N_CYCLES];
  REAL H1_error[N_CYCLES];
  REAL Linfty_error[N_CYCLES];

//graphics setup
  Palette pal;
  NumPtr<VirtualGraphTool*> plot_gt(2);
#if (SPACEDIM==1)
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  plot_gt[0]=OPERATOR_NEW XYGraphTool("solution","x","u",0.,1.,0.,1.,
    &cmap,0,0.5);
  plot_gt[1]=OPERATOR_NEW XYGraphTool("multiplier","x","m",0.,1.,0.,1.,
    &cmap,0,0.5);
#elif (SPACEDIM==2)
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]={0.,0.},rhi[2]={1.,1.};
  AreaGraphTool mesh_gt("grid","x0","x1",rlo,rhi,&cmap,0,0.5);
  plot_gt[0]=OPERATOR_NEW AreaGraphTool("velocity","x0","x1",rlo,rhi,
    &cmap,0,0.5);
  plot_gt[1]=OPERATOR_NEW AreaGraphTool("pressure","x0","x1",rlo,rhi,
    &cmap,0,0.5);
//AreaGraphTool contour_gt("solution contours","x0","x1",rlo,rhi,&cmap,
//  0,0.5);
//BOOLEAN use_lighting=FALSE;
//Vector3 vlow(0.,0.,0.);
//Vector3 vhigh(1.,1.,1.);
//VolGraphTool surface_gt("solution surface",vlow,vhigh,pal,use_lighting,
//  0.5);
//{ VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
//  surface_gt.drawBox(vlow,vhigh);
//}
#else
//
  BOOLEAN use_lighting=TRUE;
  Vector3 vlow(0.,0.,0.);
  Vector3 vhigh(1.,1.,1.);
  VolGraphTool mesh_gt("grid",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&mesh_gt);
    mesh_gt.drawBox(vlow,vhigh);
  }
  VolGraphTool *pgt0=OPERATOR_NEW VolGraphTool("velocity",vlow,vhigh,
    pal,use_lighting,0.5);
  plot_gt[0]=pgt0;
  { VolGraphTool::DrawDelimiter dd(TRUE,pgt0);
    pgt0->drawBox(vlow,vhigh);
  }
  VolGraphTool *pgt1=OPERATOR_NEW VolGraphTool("pressure",vlow,vhigh,
    pal,use_lighting,0.5);
  plot_gt[1]=pgt1;
  { VolGraphTool::DrawDelimiter dd(TRUE,pgt1);
    pgt1->drawBox(vlow,vhigh);
  }
//VolGraphTool surface_gt("solution surface",vlow,vhigh,pal,use_lighting,
//  0.5);
//{ VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
//  surface_gt.drawBox(vlow,vhigh);
//}
//
#endif

//In the KellyErrorEstimator the square of the jump in the solution 
//  gradient is integrated over each face.
//The square of the jump has order 2*degree, so we want the quadrature rule
//  to have order 2*degree+1
  map<int,Quadrature<SPACEDIM-1> > error_estimator_quadrature_map;
#if (SPACEDIM==2)
  error_estimator_quadrature_map[Interval::verticesPerShape()]=
    GaussianQuadrature<1>(2*degree+2);
#endif
#if (SPACEDIM==3)
  error_estimator_quadrature_map[Quadrilateral::verticesPerShape()]=
    GaussianQuadrature<2>(2*degree+2);
  error_estimator_quadrature_map[Triangle::verticesPerShape()]=
    TriangleProductGaussianQuadrature(2*degree+2);
#endif

//loop over local mesh refinements
  for (int cycle=0;cycle<N_CYCLES;cycle++) {
#ifdef DEBUG
    cout << "cycle = " << cycle << endl;
#endif

//  data structure to keep track of degrees of freedom in solver
    HdivDofHandler hdiv_dof_handler(tess,fe);
    hdiv_dof_handler.distributeDofs();
    n_dofs[cycle]=hdiv_dof_handler.numberDofs();

#ifdef DEBUG
    cout << "\tnumber of active elements = " 
         << tess.numberActiveShapes() << endl;
    cout << "\tNumber of degrees of freedom = " << n_dofs[cycle]
         << endl;
    cout << "\tassemble linear system..." << endl;
#endif
//  data structure to keep track of linear system storage
    int band_width=hdiv_dof_handler.maxCouplingsBetweenDofs();
    int tsd=hdiv_dof_handler.totalSolutionDofs();
    SparsityPattern *sparsity_pattern_A=
      OPERATOR_NEW SparsityPattern(tsd,tsd,band_width,true);
    SparsityPattern *sparsity_pattern_B=
      OPERATOR_NEW SparsityPattern(hdiv_dof_handler.numberDofs()-tsd,
        tsd,band_width,false);
    hdiv_dof_handler.makeSparsityPattern(*sparsity_pattern_A,
      *sparsity_pattern_B);
//
    sparsity_pattern_A->compress();
    sparsity_pattern_B->compress();

//  storage for stiffness matrix, right-hand side and solution
    SparseMatrix system_matrix_A(sparsity_pattern_A);
    SparseMatrix system_matrix_B(sparsity_pattern_B);
    Vector system_rhs(n_dofs[cycle],0.);
    Vector solution(n_dofs[cycle],0.);

//  assemble linear system
    hdiv_dof_handler.assembleLinearSystem(mapping,qmap,nqmap,
      system_matrix_A,system_matrix_B,system_rhs);

//  replace stiffness equations with hanging node constraint equations
//  hanging_node_constraints.condense(system_matrix_A,system_matrix_B);
//  hanging_node_constraints.condense(system_rhs);

//  compute essential boundary values at mapped points on bdry
//  calls FunctionSystem in BoundaryCondition's to set boundary values
    map<int,REAL> boundary_values;
    {
#ifdef TIMING
//    Timer y(z3);
#endif
      hdiv_dof_handler.interpolateBoundaryValues(mapping,boundary_values);
    }

//  modify linear system and rhs for essential boundary conditions
    bool preserve_symmetry=true;
    hdiv_dof_handler.applyBoundaryValues(boundary_values,system_matrix_A,
      system_matrix_B,system_rhs,preserve_symmetry);
    hdiv_dof_handler.applyBoundaryValues(boundary_values,solution);

    {
#ifdef DEBUG
      cout << "\tsolve linear system..." << endl;
#endif
#ifdef TIMING
//    Timer y(z5);
#endif
      RelativeErrorControl cg_solver_control(solution.size(),
        DBL_EPSILON*1.e6,false,false);
      HdivCGSolver cg(cg_solver_control);
      PreconditionIdentity pi;
      cg.solve(system_matrix_A,system_matrix_B,solution,system_rhs,pi);
    }

    n_active_cells[cycle]=tess.numberActiveShapes();
    {
//    compute solution values for graphical display
      HdivDataOut data_out(hdiv_dof_handler,mapping);
      data_out.addDataVector(solution,"solution");
#if (SPACEDIM==1)
      int n_subdivisions=640/n_active_cells[cycle];
//    int n_subdivisions=2;
#else
//    int n_subdivisions=degree+1;
      int n_subdivisions=5;
#endif
      data_out.buildPatches(n_subdivisions);
#if (SPACEDIM>=2)
      data_out.drawGrid(&mesh_gt);
#endif
#if (SPACEDIM==2)
//    data_out.contourPlot("solution",&contour_gt);
#endif
//    3D clip plane plots of the multiplier are VERY slow, so commented out
//    see HdivDataOut::plot(const char*,int,NumPtr<VirtualGraphTool*>&)
      bool quit_button=D_CHOOSE(false,false,true);
      data_out.plot("solution",plot_gt,quit_button);
#if (SPACEDIM==1)
      if (!quit_button) {
      XYGraphTool::WINDOW_TYPE::QuitButton qb;
      }
#elif (SPACEDIM==2)
      if (!quit_button) {
      AreaGraphTool::WINDOW_TYPE::QuitButton qb;
      }
#endif
#if (SPACEDIM>=2)
//    data_out.plotSurface("solution",&surface_gt);
#endif
    }
    if (cycle<N_CYCLES-1) {
#ifdef DEBUG
      cout << "\trefine mesh..." << endl;
#endif
#ifdef TIMING
//    Timer y(z10);
#endif
      Vector estimated_error_per_cell;
      OBSOLETE("fix the following");
//    KellyErrorEstimator error_estimator(mapping,mg_dof_handler,
//      error_estimator_quadrature_map,solution,estimated_error_per_cell);
//    refine those 60 per cent of cells with the highest error values,
//      and coarsen the 3 per cent of cells with the lowest values.
      tess.refineAndCoarsenFixedFraction(estimated_error_per_cell,
        0.6,0.03);
      tess.executeCoarseningAndRefinement();
    }
    delete sparsity_pattern_A;
    delete sparsity_pattern_B;
  }

  if (N_CYCLES>1) {
    Palette pal;
    XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);

    XYGraphTool nacgt("number active cells vs cycle","cycle",
      "number active cells",0,N_CYCLES-1,n_active_cells[0],
      n_active_cells[N_CYCLES-1],&cmap,0,0.5);
    nacgt.newPage();
    nacgt.setbgColor("white");
    nacgt.setfgColor("black");
    nacgt.drawAxes();
    nacgt.setfgColor("blue");
    nacgt.movePen(0,n_active_cells[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      nacgt.drawLine(cycle,n_active_cells[cycle]);
    }
    nacgt.flush();

    XYGraphTool ndgt("number degrees of freedom vs cycle","cycle",
      "number degress of freedom",0,N_CYCLES-1,n_dofs[0],
      n_dofs[N_CYCLES-1],&cmap,0,0.5);
    ndgt.newPage();
    ndgt.setbgColor("white");
    ndgt.setfgColor("black");
    ndgt.drawAxes();
    ndgt.setfgColor("blue");
    ndgt.movePen(0,n_dofs[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ndgt.drawLine(cycle,n_dofs[cycle]);
    }
    ndgt.flush();

    double log10=log(10.);
    for (int cycle=0;cycle<N_CYCLES;cycle++) {
      log_n_dofs[cycle]=log(static_cast<REAL>(n_dofs[cycle]))
        /(static_cast<REAL>(SPACEDIM)*log10);
      L2_error[cycle]=(L2_error[cycle]>0. ? log(L2_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      H1_error[cycle]=(H1_error[cycle]>0. ? log(H1_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      Linfty_error[cycle]=(Linfty_error[cycle]>0. 
        ? log(Linfty_error[cycle]) : log(DBL_EPSILON))/log10;
    }

    REAL ymin=L2_error[0];
    REAL ymax=L2_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,L2_error[cycle]);
      ymax=max(ymax,L2_error[cycle]);
    }
    XYGraphTool l2gt("log_10(L2 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(L2 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    l2gt.newPage();
    l2gt.setbgColor("white");
    l2gt.setfgColor("black");
    l2gt.drawAxes();
    l2gt.setfgColor("blue");
    l2gt.movePen(log_n_dofs[0],L2_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      l2gt.drawLine(log_n_dofs[cycle],L2_error[cycle]);
    }
    l2gt.flush();

    ymin=H1_error[0];
    ymax=H1_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,H1_error[cycle]);
      ymax=max(ymax,H1_error[cycle]);
    }
    XYGraphTool h1gt("log_10(H1 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(H1 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    h1gt.newPage();
    h1gt.setbgColor("white");
    h1gt.setfgColor("black");
    h1gt.drawAxes();
    h1gt.setfgColor("blue");
    h1gt.movePen(log_n_dofs[0],H1_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      h1gt.drawLine(log_n_dofs[cycle],H1_error[cycle]);
    }
    h1gt.flush();

    ymin=Linfty_error[0];
    ymax=Linfty_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,Linfty_error[cycle]);
      ymax=max(ymax,Linfty_error[cycle]);
    }
    XYGraphTool linftygt(
      "log_10(Linfinity error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(Linfinity error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    linftygt.newPage();
    linftygt.setbgColor("white");
    linftygt.setfgColor("black");
    linftygt.drawAxes();
    linftygt.setfgColor("blue");
    linftygt.movePen(log_n_dofs[0],Linfty_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      linftygt.drawLine(log_n_dofs[cycle],Linfty_error[cycle]);
    }
    linftygt.flush();

    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#ifdef TIMING
//tl.printOn(cout);
//tl.~TimedObjectList();
#endif

  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  for (int d=0;d<plot_gt.getNumber();d++) {
    delete plot_gt[d];
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
