// "$Header: /home/faculty/johnt/cvs/deal_new/examples/RectangularDomain.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>

#include "Boundary.H"
#include "ConstraintMatrix.H"
#include "DataOut.H"
#include "FiniteElement.H"
#include "GenerateTess.H"
#include <KellyErrorEstimator.H>
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Multigrid.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "TimedObject.H"
#include "Tracer.H"
#include "Vector.H"

#if (SPACEDIM==1)
# define N_CYCLES 7
#elif (SPACEDIM==2)
# define N_CYCLES 4
  enum MESH_ELEMENTS { 
    QUADRILATERAL_MESH_ELEMENTS,
    TRIANGLE_MESH_ELEMENTS 
  };
  MESH_ELEMENTS mesh_elements=QUADRILATERAL_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=TRIANGLE_MESH_ELEMENTS;
#else
# define N_CYCLES 3
  enum MESH_ELEMENTS { 
    HEXAHEDRAL_MESH_ELEMENTS,
    PRISM_MESH_ELEMENTS,
    TETRAHEDRAL_MESH_ELEMENTS 
  };
  MESH_ELEMENTS mesh_elements=HEXAHEDRAL_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=PRISM_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=TETRAHEDRAL_MESH_ELEMENTS;
  int dir=0;
#endif
enum TEST_PROBLEM { 
  LINEAR_SOLUTION,    // -u''= 0, u (-1)=-1, u (1)=1 ==> u(x)=x
  QUADRATIC_SOLUTION, // -u''= 2, u (-1)= 1, u'(1)=2 ==> u(x)=x^2
  CUBIC_SOLUTION,     // -u''=6x, u'(-1)= 3, u (1)=1 ==> u(x)=x^3
  SINH_SOLUTION//-u''+u=0, u(-1)=-sinh(1), u(1)=sinh(1) ==> u(x)=sinh(x)
};
//TEST_PROBLEM test_problem=LINEAR_SOLUTION;
//TEST_PROBLEM test_problem=QUADRATIC_SOLUTION;
//TEST_PROBLEM test_problem=CUBIC_SOLUTION;
  TEST_PROBLEM test_problem=SINH_SOLUTION;
int degree=1;
int nelements=4;

REAL rightHandSideValue(const Point<SPACEDIM> &p) {
#if (SPACEDIM==1)
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return -2.;
//    return -4.;
    case CUBIC_SOLUTION:
      return -6.*p[0];
    case LINEAR_SOLUTION:
    case SINH_SOLUTION:
    default:
      return 0.;
  }
#elif (SPACEDIM==2)
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return -12.; // - 2 * trace of A=[[1,2],[2,5]]
//    return -4.;
    case CUBIC_SOLUTION:
      return -6.*(p[0]+p[1]);
    case LINEAR_SOLUTION:
    case SINH_SOLUTION:
    default:
      return 0.;
  }
#else
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return -40.; // - 2 * trace of A=[[1,2,3],[2,5,8],[3,8,14]]
//    return -4.;
    case CUBIC_SOLUTION:
      return -6.*(p[0]+p[1]+p[2]);
    case LINEAR_SOLUTION:
    case SINH_SOLUTION:
    default:
      return 0.;
  }
#endif
}

REAL exactSolutionValue(const Point<SPACEDIM> &p) {
#if (SPACEDIM==1)
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return pow(p[0],2);
//    return (1.-2.*p[0])*(1.-p[0]);
    case CUBIC_SOLUTION:
      return pow(p[0],3);
    case SINH_SOLUTION:
      return sinh(p[0]);
    case LINEAR_SOLUTION:
    default:
      return 1.-p[0];
  }
#elif (SPACEDIM==2)
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return p[0]*(   p[0]+2.*p[1])
            +p[1]*(2.*p[0]+5.*p[1]);
//    return (1.-2.*p[0])*(1.-p[0]);
    case CUBIC_SOLUTION:
      return pow(p[0],3)+pow(p[1],3);
    case SINH_SOLUTION:
      return sin(M_PI*p[0])*sinh(M_PI*p[1]);
    case LINEAR_SOLUTION:
    default:
      return p[0]+2.*p[1];
//    return 1.-p[0];
  }
#else
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      return p[0]*(   p[0]+2.*p[1]+ 3.*p[2])
            +p[1]*(2.*p[0]+5.*p[1]+ 8.*p[2])
            +p[2]*(3.*p[0]+8.*p[1]+14.*p[2]);
//    return (1.-2.*p[0])*(1.-p[0]);
    case CUBIC_SOLUTION:
      return pow(p[0],3)+pow(p[1],3)+pow(p[2],3);
    case SINH_SOLUTION:
      return sin(M_PI*p[0])*cos(M_PI*p[1])*sinh(M_SQRT2*M_PI*p[2]);
    case LINEAR_SOLUTION:
    default:
      return p[0]+2.*p[1]+3.*p[2];
//    return 1.-p[0];
  }
#endif
}

Tensor<1,SPACEDIM> exactSolutionGradient(const Point<SPACEDIM> &p) {
#if (SPACEDIM==1)
  switch (test_problem) {
    case QUADRATIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val(p);
      return_val*=2.;
//    Tensor<1,SPACEDIM> return_val;
//    return_val[0]=-3.+4.*p[0];
      return return_val;
    }
    case CUBIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val(p);
      return_val*=3.*p[0];
      return return_val;
    }
    case SINH_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=cosh(p[0]);
      return return_val;
    }
    case LINEAR_SOLUTION:
    default: {
      Tensor<1,SPACEDIM> return_val;
      return_val.initialize(-1.);
      return return_val;
    }
  }
#elif (SPACEDIM==2)
  switch (test_problem) {
    case QUADRATIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=2.*p[0]+ 4.*p[1];
      return_val[1]=4.*p[0]+10.*p[1]; // 2 * A * x
//    return_val[0]=-3.+4.*p[0];
//    return_val[1]=0.;
      return return_val;
    }
    case CUBIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=3.*p[0]*p[0];
      return_val[1]=3.*p[1]*p[1];
      return return_val;
    }
    case SINH_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=M_PI*cos(M_PI*p[0])*sinh(M_PI*p[1]);
      return_val[1]=M_PI*sin(M_PI*p[0])*cosh(M_PI*p[1]);
      return return_val;
    }
    case LINEAR_SOLUTION:
    default: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=1.;
      return_val[1]=2.;
//    return_val[0]=-1.;
//    return_val[1]=0.;
      return return_val;
    }
  }
#else
  switch (test_problem) {
    case QUADRATIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=2.*p[0]+ 4.*p[1]+ 6.*p[2];
      return_val[1]=4.*p[0]+10.*p[1]+16.*p[2];
      return_val[2]=6.*p[0]+16.*p[1]+28.*p[2]; // 2 * A * x
//    return_val[0]=-3.+4.*p[0];
//    return_val[1]=0.;
      return return_val;
    }
    case CUBIC_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=3.*p[0]*p[0];
      return_val[1]=3.*p[1]*p[1];
      return_val[2]=3.*p[2]*p[2];
      return return_val;
    }
    case SINH_SOLUTION: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=M_PI*cos(M_PI*p[0])*cos(M_PI*p[1])
                        *sinh(M_SQRT2*M_PI*p[2]);
      return_val[1]=-M_PI*sin(M_PI*p[0])*sin(M_PI*p[1])
                         *sinh(M_SQRT2*M_PI*p[2]);
      return_val[2]=M_SQRT2*M_PI*sin(M_PI*p[0])*cos(M_PI*p[1])
                           *cosh(M_SQRT2*M_PI*p[2]);
      return return_val;
    }
    case LINEAR_SOLUTION:
    default: {
      Tensor<1,SPACEDIM> return_val;
      return_val[0]=1.;
      return_val[1]=2.;
      return_val[2]=3.;
//    return_val[0]=-1.;
//    return_val[1]=0.;
//    return_val[2]=0.;
      return return_val;
    }
  }
#endif
}

class ExactFunctionSystem : public FunctionSystem {
  protected:
    explicit ExactFunctionSystem(const FunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    ExactFunctionSystem() {;}
    virtual ~ExactFunctionSystem() {;}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      return exactSolutionValue(p);
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      function_values[0]=exactSolutionValue(p);
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,
    int) const {
      return exactSolutionGradient(p);
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
};
class NeumannFunctionSystem : public FunctionSystem {
  private:
    int dir;
    bool left_hand;
  protected:
    explicit NeumannFunctionSystem(const NeumannFunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    NeumannFunctionSystem(int d,bool lh) : dir(d),left_hand(lh) {}
    virtual ~NeumannFunctionSystem() {}
    virtual REAL value(const Point<SPACEDIM> &p,int) const {
      Tensor<1,SPACEDIM> g(exactSolutionGradient(p));
      return (left_hand ? -g[dir] : g[dir]);
    }
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      Tensor<1,SPACEDIM> g(exactSolutionGradient(p));
      function_values[0]=(left_hand ? -g[dir] : g[dir]);
    }
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("should not be called");
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("should not be called");
    }
};

class SimpleMaterialModel : public MaterialModel {
  public:
    void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,REAL> &sv,
    const Array<2,Tensor<1,SPACEDIM> > &sg,Matrix &cell_matrix,
    Vector &cell_rhs) const {
      int dofs_per_cell=cell_rhs.size();
      for (int q=0;q<q_points.getNumber();q++) {
        REAL rhs_value=rightHandSideValue(q_points[q]);
        REAL jxw=jxw_values[q];
        for (int i=0;i<dofs_per_cell;i++) {
          Tensor<1,SPACEDIM> sgi=sg(i,q);
          sgi*=jxw;
          REAL si=sv(i,q)*jxw;
          for (int j=0;j<dofs_per_cell;j++) {
            cell_matrix(i,j)+=sgi*sg(j,q);
#if (SPACEDIM==1)
            if (test_problem==SINH_SOLUTION) {
              cell_matrix(i,j)+=si*sv(j,q);
            }
#endif
          }
          cell_rhs[i]+=si*rhs_value;
        }
      }
    }
};

#ifdef TIMING
//the most expensive routines are the linear system solve and
//  interpolateBoundaryValues
//the latter requires computation of the shape functions on an element
//  at iteratively computed canonical points corresponding to boundary:
//TimedObjectList tl;
#endif

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
#ifdef TIMING
//TimedObject *z3=
//  OPERATOR_NEW TimedObject("main interpolateBoundaryValues",tl);
//TimedObject *z5=OPERATOR_NEW TimedObject("main solve",tl);
//TimedObject *z10=OPERATOR_NEW TimedObject("main refine",tl);
#endif
  switch (test_problem) {
    case QUADRATIC_SOLUTION:
      cout << "\tquadratic solution" << endl;
      break;
    case CUBIC_SOLUTION:
      cout << "\tcubic solution" << endl;
      break;
    case SINH_SOLUTION:
      cout << "\tsinh solution" << endl;
      break;
    case LINEAR_SOLUTION:
    default:
      cout << "\tlinear solution" << endl;
  }
  cout << "\tdegree = " << degree << endl;
#if (SPACEDIM==2)
  switch (mesh_elements) {
    case QUADRILATERAL_MESH_ELEMENTS:
      cout << "\tquadrilateral mesh elements" << endl;
      break;
    case TRIANGLE_MESH_ELEMENTS:
      cout << "\ttriangle mesh elements" << endl;
      break;
    default:
      OBSOLETE("unknown mesh elements");
  }
#endif
#if (SPACEDIM==3)
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS:
      cout << "\thexahedral mesh elements" << endl;
      break;
    case PRISM_MESH_ELEMENTS:
      cout << "\tprism mesh elements" << endl;
      break;
    case TETRAHEDRAL_MESH_ELEMENTS:
      cout << "\ttetrahedral mesh elements" << endl;
      break;
    default:
      OBSOLETE("unknown mesh elements");
  }
#endif

  Tessellation tess;

//rectangular domain consisting of one interval/quadrilateral/hexagon
  Point<SPACEDIM> low(D_DECL(-1.,-1.,-1.));
//Point<SPACEDIM> low(D_DECL(0.,0.,0.));
  Point<SPACEDIM> edge[SPACEDIM];
    edge[0]=Point<SPACEDIM>(D_DECL(2.,0.,0.));
//  edge[0]=Point<SPACEDIM>(D_DECL(1.,0.,0.));
#if (SPACEDIM>1)
    edge[1]=Point<SPACEDIM>(D_DECL(0.,2.,0.));
//  edge[1]=Point<SPACEDIM>(D_DECL(0.,1.,0.));
#endif
#if (SPACEDIM==3)
    edge[2]=Point<SPACEDIM>(D_DECL(0.,0.,2.));
#endif
  int number_elements[SPACEDIM]={D_DECL(nelements,nelements,nelements)};
  bool set_face_ids=true;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
#if (SPACEDIM==1)
  BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,boundary,boundary_condition);
#elif (SPACEDIM==2)
  BoundaryData *boundary_data=0;
  if (mesh_elements==QUADRILATERAL_MESH_ELEMENTS) {
    boundary_data=generateParallelpiped(tess,low,edge,number_elements,
      set_face_ids,boundary,boundary_condition);
  } else {
    boundary_data=generateParallelpipedTriangles(tess,low,edge,
      number_elements,set_face_ids,boundary,boundary_condition);
  }
#else
  BoundaryData *boundary_data=0;
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS:
      boundary_data=generateParallelpiped(tess,low,edge,number_elements,
        set_face_ids,boundary,boundary_condition);
      break;
    case PRISM_MESH_ELEMENTS: {
//    the following tessellation has all possible faces on outer boundary
      NumPtr<Point<3> > vertices(10);
      for (int i=0;i<vertices.getNumber();i++) vertices[i]=low;
      vertices[1]+=edge[0];
      vertices[2]+=edge[0]*0.5; vertices[2]+=edge[1]*0.5; 
      vertices[3]+=edge[1];
      vertices[4]+=edge[0]; vertices[4]+=edge[1];
      vertices[5]+=edge[2];
      vertices[6]+=edge[0]; vertices[6]+=edge[2];
      vertices[7]+=edge[0]*0.5; vertices[7]+=edge[1]*0.5; 
                                vertices[7]+=edge[2];
      vertices[8]+=edge[1]; vertices[8]+=edge[2];
      vertices[9]+=edge[0]; vertices[9]+=edge[1]; vertices[9]+=edge[2];
      NumPtr<CellData> cells(4);
      for (int c=0;c<cells.getNumber();c++) cells[c].allocate(6);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;
      cells[0].vertex(3)=5;
      cells[0].vertex(4)=6;
      cells[0].vertex(5)=7;

      cells[1].vertex(0)=0;
      cells[1].vertex(1)=2;
      cells[1].vertex(2)=3;
      cells[1].vertex(3)=5;
      cells[1].vertex(4)=7;
      cells[1].vertex(5)=8;

      cells[2].vertex(0)=2;
      cells[2].vertex(1)=4;
      cells[2].vertex(2)=3;
      cells[2].vertex(3)=7;
      cells[2].vertex(4)=9;
      cells[2].vertex(5)=8;

      cells[3].vertex(0)=1;
      cells[3].vertex(1)=4;
      cells[3].vertex(2)=2;
      cells[3].vertex(3)=6;
      cells[3].vertex(4)=9;
      cells[3].vertex(5)=7;
      boundary_data=tess.create(vertices,cells);
      boundary.allocate(6);
      boundary_condition.allocate(6);
      Point<3> center(low+(edge[0]+edge[1]+edge[2])*0.5);
      boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1],
        low+edge[2]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
        low+edge[0]+edge[1],low+edge[0]+edge[2]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[2],
        low+edge[0]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
        low+edge[1]+edge[2],low+edge[1]+edge[0]);
      boundary[4]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0],
        low+edge[1]);
      boundary[5]=OPERATOR_NEW PlaneBoundary(center,low+edge[2],
        low+edge[2]+edge[0],low+edge[2]+edge[1]);
      for (int i=0;i<2*SPACEDIM;i++) {
        boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
      }
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<SPACEDIM> center=face_element->center(vertices);
          bool found=false;
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              found=true;
              break;
            }
          }
        }
      }
      tess.synchronize(boundary_data);
      break;
    }
    case TETRAHEDRAL_MESH_ELEMENTS: {
//    the following tessellation has all possible faces on outer boundary
      NumPtr<Point<3> > vertices(8);
      for (int i=0;i<vertices.getNumber();i++) vertices[i]=low;
      vertices[1]+=edge[0];
      vertices[2]+=edge[1];
      vertices[3]+=edge[0]; vertices[3]+=edge[1];
      vertices[4]+=edge[2];
      vertices[5]+=edge[0]; vertices[5]+=edge[2];
      vertices[6]+=edge[1]; vertices[6]+=edge[2];
      vertices[7]+=edge[0]; vertices[7]+=edge[1]; vertices[7]+=edge[2];
      NumPtr<CellData> cells(6);
      for (int c=0;c<cells.getNumber();c++) cells[c].allocate(4);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;
      cells[0].vertex(3)=4;

      cells[1].vertex(0)=1;
      cells[1].vertex(1)=4;
      cells[1].vertex(2)=5;
      cells[1].vertex(3)=6;

      cells[2].vertex(0)=1;
      cells[2].vertex(1)=2;
      cells[2].vertex(2)=4;
      cells[2].vertex(3)=6;

      cells[3].vertex(0)=1;
      cells[3].vertex(1)=2;
      cells[3].vertex(2)=7;
      cells[3].vertex(3)=3;

      cells[4].vertex(0)=1;
      cells[4].vertex(1)=5;
      cells[4].vertex(2)=7;
      cells[4].vertex(3)=6;

      cells[5].vertex(0)=1;
      cells[5].vertex(1)=2;
      cells[5].vertex(2)=6;
      cells[5].vertex(3)=7;
      boundary_data=tess.create(vertices,cells);
      boundary.allocate(6);
      boundary_condition.allocate(6);
      Point<3> center(low+(edge[0]+edge[1]+edge[2])*0.5);
      boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1],
        low+edge[2]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
        low+edge[0]+edge[1],low+edge[0]+edge[2]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[2],
        low+edge[0]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
        low+edge[1]+edge[2],low+edge[1]+edge[0]);
      boundary[4]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0],
        low+edge[1]);
      boundary[5]=OPERATOR_NEW PlaneBoundary(center,low+edge[2],
        low+edge[2]+edge[0],low+edge[2]+edge[1]);
      for (int i=0;i<2*SPACEDIM;i++) {
        boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
      }

      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<SPACEDIM> center=face_element->center(vertices);
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              break;
            }
          }
        }
      }
      tess.synchronize(boundary_data);
      break;
    }
    default:
      OBSOLETE("bad mesh_elements");
  }
#endif
#if (SPACEDIM>1)
  tess.checkFaceBoundaryCondition(boundary_condition);
#endif
#if (SPACEDIM==3)
  tess.checkLineBoundaryCondition(boundary_condition);
#endif

  SimpleMaterialModel smm;
  tess.setMaterialModel(&smm);

  ExactFunctionSystem efs;
  switch (test_problem) {
    case QUADRATIC_SOLUTION: {
//    DIRICHLET on left, back and bottom, NEUMANN on right, front and top
      for (int d=0;d<SPACEDIM;d++) {
        boundary_condition[2*d]->setFunctionSystem(&efs);
        boundary_condition[2*d]->makeEssential();
        boundary_condition[2*d+1]->setFunctionSystem(
          OPERATOR_NEW NeumannFunctionSystem(d,false));
        boundary_condition[2*d+1]->makeNatural();
      }
      break;
    }
    case CUBIC_SOLUTION: {
//    NEUMANN on left, back and bottom, DIRICHLET on right, front and top
      for (int d=0;d<SPACEDIM;d++) {
        boundary_condition[2*d]->makeNatural();
        boundary_condition[2*d+1]->setFunctionSystem(
          OPERATOR_NEW NeumannFunctionSystem(d,true));
        boundary_condition[2*d+1]->setFunctionSystem(&efs);
        boundary_condition[2*d+1]->makeEssential();
      }
      break;
    }
    case LINEAR_SOLUTION: // DIRICHLET on both
    case SINH_SOLUTION:
    default: {
      for (int d=0;d<2*SPACEDIM;d++) {
        boundary_condition[d]->setFunctionSystem(&efs);
        boundary_condition[d]->makeEssential();
      }
      break;
    }
  }

//(linear or) tensor product linear coordinate mapping
  MappingL mapping;

//polynomials to represent solution of problem
  FE_L fe(degree);
//FE_H fe(degree);

//quadrature rules
  map<int,Quadrature<SPACEDIM> > qmap;
#if (SPACEDIM==1)
//need order>=2*degree-1
  qmap[Interval::verticesPerShape()]=GaussianQuadrature<SPACEDIM>(
    GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree-1));
#elif (SPACEDIM==2)
  if (mesh_elements==QUADRILATERAL_MESH_ELEMENTS) {
//  need order>=2*degree+1
    qmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<SPACEDIM>(
      GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+1));
  } else {
    qmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(
      TriangleGaussianQuadrature::argGivenOrder(2*degree-1)); 
  }
#else
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS:
//    need order>=2*degree+1
      qmap[Hexahedron::verticesPerShape()]=GaussianQuadrature<SPACEDIM>(
        GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+1));
//    qmap[Hexahedron::verticesPerShape()]=
//      NewtonCotesQuadrature<SPACEDIM>(max(2,2*degree+1));
      break;
    case PRISM_MESH_ELEMENTS:
      qmap[Prism::verticesPerShape()]=
        ProductQuadrature(
          TriangleGaussianQuadrature(
            TriangleGaussianQuadrature::argGivenOrder(2*degree-1)),
          GaussianQuadrature<1>(
            GaussianQuadrature<1>::argGivenOrder(2*degree-1)));
      break;
    case TETRAHEDRAL_MESH_ELEMENTS:
      qmap[Tetrahedron::verticesPerShape()]=TetrahedronGaussianQuadrature(
        TetrahedronGaussianQuadrature::argGivenOrder(2*degree-1));
      break;
    default:
      OBSOLETE("bad mesh_elements");
  }
#endif

  map<int,Quadrature<SPACEDIM-1> > nqmap;
#if (SPACEDIM>1)
//natural bc's
//should be exact for polynomials of shape function degree
//  plus exact solution degree-1
//  plus Jacobian degree
# if (SPACEDIM==2)
  {
    int nv=Interval::verticesPerShape();
    if (mesh_elements==QUADRILATERAL_MESH_ELEMENTS) {
      nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(
        GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+1));
    } else {
      int nv=Interval::verticesPerShape();
      nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(
        GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+1));
    }
  }
# else
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS: {
      int nv=Quadrilateral::verticesPerShape();
      nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(
        GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+1));
      break;
    }
    case PRISM_MESH_ELEMENTS: {
      int nv=Quadrilateral::verticesPerShape();
      nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(
        GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+1));
      nv=Triangle::verticesPerShape();
      nqmap[nv]=TriangleGaussianQuadrature(
        TriangleGaussianQuadrature::argGivenOrder(2*degree-1));
      break;
    }
    case TETRAHEDRAL_MESH_ELEMENTS: {
//    face shape function is polynomial of order=degree
//    exact solution is polynomial of order=solution order-1
//    TriangleProductGaussianQuadrature has order 2*n-1 
//      where n=constructor argument
      int nv=Triangle::verticesPerShape();
      nqmap[nv]=TriangleGaussianQuadrature(
        TriangleGaussianQuadrature::argGivenOrder(2*degree-1));
      break;
    }
    default:
      OBSOLETE("bad mesh_elements");
  }
# endif
#endif

  int n_active_cells[N_CYCLES];
  int n_dofs[N_CYCLES];
  REAL log_n_dofs[N_CYCLES];
  REAL L2_error[N_CYCLES];
  REAL H1_error[N_CYCLES];
  REAL Linfty_error[N_CYCLES];

//graphics setup
  Palette pal;
#if (SPACEDIM==1)
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
//XYGraphTool plot_gt("solution plot","x","y",0.,1.,0.,1.,&cmap,0,0.5);
#elif (SPACEDIM==2)
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]={0.,0.},rhi[2]={1.,1.};
  AreaGraphTool mesh_gt("grid","x0","x1",rlo,rhi,&cmap,0,0.5);
//AreaGraphTool contour_gt("solution contours","x0","x1",rlo,rhi,&cmap,
//  0,0.5);
//AreaGraphTool plot_gt("solution plot","x0","y1",rlo,rhi,&cmap,0,0.5);
//BOOLEAN use_lighting=FALSE;
//Vector3 vlow(0.,0.,0.);
//Vector3 vhigh(1.,1.,1.);
//VolGraphTool surface_gt("solution surface",vlow,vhigh,pal,use_lighting,
//  0.5);
//{ VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
//  surface_gt.drawBox(vlow,vhigh);
//}
#else
//
  BOOLEAN use_lighting=TRUE;
  Vector3 vlow(0.,0.,0.);
  Vector3 vhigh(1.,1.,1.);
  VolGraphTool mesh_gt("grid",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&mesh_gt);
    mesh_gt.drawBox(vlow,vhigh);
  }
//VolGraphTool plot_gt("solution plot",vlow,vhigh,pal,use_lighting,0.5);
//{ VolGraphTool::DrawDelimiter dd(TRUE,&plot_gt);
//  plot_gt.drawBox(vlow,vhigh);
//}
//VolGraphTool surface_gt("solution surface",vlow,vhigh,pal,use_lighting,
//  0.5);
//{ VolGraphTool::DrawDelimiter dd(TRUE,&surface_gt);
//  surface_gt.drawBox(vlow,vhigh);
//}
//
#endif

//In the KellyErrorEstimator the square of the jump in the solution 
//  gradient is integrated over each face.
//The square of the jump has order 2*degree, so we want the quadrature rule
//  to have order 2*degree+1
  map<int,Quadrature<SPACEDIM-1> > error_estimator_quadrature_map;
#if (SPACEDIM==2)
  error_estimator_quadrature_map[Interval::verticesPerShape()]=
    GaussianQuadrature<1>(2*degree+1);
#endif
#if (SPACEDIM==3)
  error_estimator_quadrature_map[Quadrilateral::verticesPerShape()]=
    GaussianQuadrature<2>(2*degree+1);
  error_estimator_quadrature_map[Triangle::verticesPerShape()]=
    TriangleProductGaussianQuadrature(2*degree+1);
#endif

//loop over local mesh refinements
  for (int cycle=0;cycle<N_CYCLES;cycle++) {
#ifdef DEBUG
    cout << "cycle = " << cycle << endl;
#endif

//  data structure to keep track of degrees of freedom in solver
    MGDofHandler mg_dof_handler(tess,fe);
    mg_dof_handler.distributeDofs();
    n_dofs[cycle]=mg_dof_handler.numberDofs();

#ifdef DEBUG
    cout << "\tnumber of active elements = " 
         << tess.numberActiveShapes() << endl;
    cout << "\tNumber of degrees of freedom = " << n_dofs[cycle]
         << endl;
    cout << "\tassemble linear system..." << endl;
#endif
//  data structure to keep track of conforming grid constraints for
//    hanging nodes
    ConstraintMatrix hanging_node_constraints;
    hanging_node_constraints.makeHangingNodeConstraints(mg_dof_handler);
    hanging_node_constraints.close();

//  data structure to keep track of linear system storage
    SparsityPattern *sparsity_pattern=
      OPERATOR_NEW SparsityPattern(n_dofs[cycle],n_dofs[cycle],
      mg_dof_handler.maxCouplingsBetweenDofs());
    mg_dof_handler.DofHandler::makeSparsityPattern(*sparsity_pattern);

//  insert hanging node constraint equations into sparsity pattern
    hanging_node_constraints.condense(*sparsity_pattern);
    sparsity_pattern->compress();

//  storage for stiffness matrix, right-hand side and solution
    SparseMatrix system_matrix(sparsity_pattern);
    Vector system_rhs(n_dofs[cycle],0.);
    Vector solution(n_dofs[cycle],0.);

//  assemble linear system
    mg_dof_handler.assembleLinearSystem(mapping,qmap,nqmap,
      system_matrix,system_rhs);

//  replace stiffness equations with hanging node constraint equations
    hanging_node_constraints.condense(system_matrix);
    hanging_node_constraints.condense(system_rhs);

//  compute essential boundary values at mapped points on bdry
//  calls FunctionSystem in BoundaryCondition's to set boundary values
    map<int,REAL> boundary_values;
    {
#ifdef TIMING
//    Timer y(z3);
#endif
      mg_dof_handler.interpolateBoundaryValues(mapping,boundary_values);
    }

//  modify linear system and rhs for essential boundary conditions
    bool preserve_symmetry=true;
    mg_dof_handler.applyBoundaryValues(boundary_values,system_matrix,
      system_rhs,preserve_symmetry);
    mg_dof_handler.applyBoundaryValues(boundary_values,solution);


    {
#ifdef DEBUG
      cout << "\tsolve linear system..." << endl;
#endif
#ifdef TIMING
//    Timer y(z5);
#endif
      RelativeErrorControl cg_solver_control(solution.size(),
        DBL_EPSILON*1.e2,false,false);
      CGSolver cg(cg_solver_control);
//    PreconditionIdentity pi;
      PreconditionJacobi pjac(system_matrix);
      cg.solve(system_matrix,solution,system_rhs,pjac);
    }

//  apply hanging node constraints to solution of linear system
    hanging_node_constraints.distribute(solution);
    {
#ifdef DEBUG
      cout << "\testimate error..." << endl;
#endif
      n_active_cells[cycle]=tess.numberActiveShapes();
      FunctionSystem *weight=0;
      L2_error[cycle]=mg_dof_handler.integrateDifference(mapping,solution,
        efs,qmap,L2_NORM,weight,1.);
#ifdef DEBUG
      cout << "\tL2_error[" << cycle << "] = " << L2_error[cycle] << endl;
#endif
    
      H1_error[cycle]=mg_dof_handler.integrateDifference(mapping,solution,
        efs,qmap,H1_NORM,weight,1.);
   
      Linfty_error[cycle]=mg_dof_handler.integrateDifference(mapping,
        solution,efs,qmap,LINFTY_NORM,weight,1.);
    }

//  compute solution values for graphical display
    DataOut data_out(mg_dof_handler,mapping);
    data_out.addDataVector(solution,"solution");
#if (SPACEDIM==1)
    int n_subdivisions=640/n_active_cells[cycle];
#else
    int n_subdivisions=degree;
#endif
    data_out.buildPatches(n_subdivisions);
    bool quit_button=D_CHOOSE(false,false,true);
#if (SPACEDIM>=2)
    data_out.drawGrid(&mesh_gt,quit_button);
#endif
#if (SPACEDIM==2)
//  data_out.contourPlot("solution",&contour_gt,quit_button);
#endif
//  data_out.plot("solution",&plot_gt,quit_button);
#if (SPACEDIM>=2)
//  data_out.plotSurface("solution",&surface_gt);
#endif
#if (SPACEDIM==1)
//  if (!quit_button) {
//    XYGraphTool::WINDOW_TYPE::QuitButton qb;
//  }
#elif (SPACEDIM==2)
    if (!quit_button) {
      AreaGraphTool::WINDOW_TYPE::QuitButton qb;
    }
#endif
//
    if (cycle<N_CYCLES-1) {
#ifdef DEBUG
      cout << "\trefine mesh..." << endl;
#endif
#ifdef TIMING
//    Timer y(z10);
#endif
//    tess.refineGlobal(1);
      Vector estimated_error_per_cell;
      KellyErrorEstimator error_estimator(mapping,mg_dof_handler,
        error_estimator_quadrature_map,solution,estimated_error_per_cell);
//    refine those 60 per cent of cells with the highest error values,
//      and coarsen the 3 per cent of cells with the lowest values.
      tess.refineAndCoarsenFixedFraction(estimated_error_per_cell,
        0.6,0.03);
      tess.executeCoarseningAndRefinement();
    }
    delete sparsity_pattern;
  }

  if (N_CYCLES>1) {
    Palette pal;
    XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);

    XYGraphTool nacgt("number active cells vs cycle","cycle",
      "number active cells",0,N_CYCLES-1,n_active_cells[0],
      n_active_cells[N_CYCLES-1],&cmap,0,0.5);
    nacgt.newPage();
    nacgt.setbgColor("white");
    nacgt.setfgColor("black");
    nacgt.drawAxes();
    nacgt.setfgColor("blue");
    nacgt.movePen(0,n_active_cells[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      nacgt.drawLine(cycle,n_active_cells[cycle]);
    }
    nacgt.flush();

    XYGraphTool ndgt("number degrees of freedom vs cycle","cycle",
      "number degress of freedom",0,N_CYCLES-1,n_dofs[0],
      n_dofs[N_CYCLES-1],&cmap,0,0.5);
    ndgt.newPage();
    ndgt.setbgColor("white");
    ndgt.setfgColor("black");
    ndgt.drawAxes();
    ndgt.setfgColor("blue");
    ndgt.movePen(0,n_dofs[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ndgt.drawLine(cycle,n_dofs[cycle]);
    }
    ndgt.flush();

    double log10=log(10.);
    for (int cycle=0;cycle<N_CYCLES;cycle++) {
      log_n_dofs[cycle]=log(static_cast<REAL>(n_dofs[cycle]))
        /(static_cast<REAL>(SPACEDIM)*log10);
      L2_error[cycle]=(L2_error[cycle]>0. ? log(L2_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      H1_error[cycle]=(H1_error[cycle]>0. ? log(H1_error[cycle])
                                          : log(DBL_EPSILON))/log10;
      Linfty_error[cycle]=(Linfty_error[cycle]>0. 
        ? log(Linfty_error[cycle]) : log(DBL_EPSILON))/log10;
    }

    REAL ymin=L2_error[0];
    REAL ymax=L2_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,L2_error[cycle]);
      ymax=max(ymax,L2_error[cycle]);
    }
    XYGraphTool l2gt("log_10(L2 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(L2 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    l2gt.newPage();
    l2gt.setbgColor("white");
    l2gt.setfgColor("black");
    l2gt.drawAxes();
    l2gt.setfgColor("blue");
    l2gt.movePen(log_n_dofs[0],L2_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      l2gt.drawLine(log_n_dofs[cycle],L2_error[cycle]);
    }
    l2gt.flush();

    ymin=H1_error[0];
    ymax=H1_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,H1_error[cycle]);
      ymax=max(ymax,H1_error[cycle]);
    }
    XYGraphTool h1gt("log_10(H1 error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(H1 error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    h1gt.newPage();
    h1gt.setbgColor("white");
    h1gt.setfgColor("black");
    h1gt.drawAxes();
    h1gt.setfgColor("blue");
    h1gt.movePen(log_n_dofs[0],H1_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      h1gt.drawLine(log_n_dofs[cycle],H1_error[cycle]);
    }
    h1gt.flush();

    ymin=Linfty_error[0];
    ymax=Linfty_error[0];
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      ymin=min(ymin,Linfty_error[cycle]);
      ymax=max(ymax,Linfty_error[cycle]);
    }
    XYGraphTool linftygt(
      "log_10(Linfinity error) vs log_10(degrees of freedom)",
      "log_10(degrees of freedom)","log_10(Linfinity error)",
      log_n_dofs[0],log_n_dofs[N_CYCLES-1],ymin,ymax,&cmap,0,0.5);
    linftygt.newPage();
    linftygt.setbgColor("white");
    linftygt.setfgColor("black");
    linftygt.drawAxes();
    linftygt.setfgColor("blue");
    linftygt.movePen(log_n_dofs[0],Linfty_error[0]);
    for (int cycle=1;cycle<N_CYCLES;cycle++) {
      linftygt.drawLine(log_n_dofs[cycle],Linfty_error[cycle]);
    }
    linftygt.flush();

    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#ifdef TIMING
//tl.printOn(cout);
//tl.~TimedObjectList();
#endif

  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    if (boundary_condition[i]->isNatural()) {
      delete boundary_condition[i]->getFunctionSystem();
    }
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
