// "$Header: /home/faculty/johnt/cvs/deal_new/examples/RectangularDomain.C,v 1.1 2009/08/20 17:27:34 johnt Exp $"
//**********************************************************************
// Copyright 2010 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <iostream>

#include "Boundary.H"
#include "HcurlDataOut.H"
#include "HcurlDofHandler.H"
#include "HcurlFiniteElement.H"
#include "GenerateTess.H"
#include "Mapping.H"
#include "MemoryDebugger.H"
#include "Precondition.H"
#include "SetTraps.H"
#include "SparseMatrix.H"
#include "SparsityPattern.H"
#include "TimedObject.H"
#include "Tracer.H"
#include "Vector.H"

#if (SPACEDIM==1)
#elif (SPACEDIM==2)
  enum MESH_ELEMENTS { 
    QUADRILATERAL_MESH_ELEMENTS,
    TRIANGULAR_MESH_ELEMENTS 
  };
  MESH_ELEMENTS mesh_elements=QUADRILATERAL_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=TRIANGULAR_MESH_ELEMENTS;
#else
  enum MESH_ELEMENTS { 
    HEXAHEDRAL_MESH_ELEMENTS,
    PRISM_MESH_ELEMENTS,
    TETRAHEDRAL_MESH_ELEMENTS 
  };
  MESH_ELEMENTS mesh_elements=HEXAHEDRAL_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=PRISM_MESH_ELEMENTS;
//MESH_ELEMENTS mesh_elements=TETRAHEDRAL_MESH_ELEMENTS;
//int dir=0;
#endif
int degree=0;
int nelements=4;

//specify H
class ConstantMagneticFieldStrengthFunctionSystem : 
public D_CHOOSE(FunctionSystem,FunctionSystem,TensorFunctionSystem) {
  private:
    HCURL_MULTIPLIER_DATA_TYPE val;
  protected:
    explicit ConstantMagneticFieldStrengthFunctionSystem(
    const D_CHOOSE(FunctionSystem,FunctionSystem,TensorFunctionSystem)&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    ConstantMagneticFieldStrengthFunctionSystem(
      const HCURL_MULTIPLIER_DATA_TYPE &H) : val(H) {;}
    virtual ~ConstantMagneticFieldStrengthFunctionSystem() {;}
    virtual HCURL_MULTIPLIER_DATA_TYPE value(const Point<SPACEDIM> &p,int)
    const {
      return val;
    }
#if (SPACEDIM<=2)
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
      function_values[0]=val;
    }
#else
    virtual void value(const Point<SPACEDIM> &p,
    NumPtr<Tensor<1,3> > &function_values) const {
      function_values[0]=val;
    }
#endif
#if (SPACEDIM<=2)
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("why called?");
//    Tensor<1,SPACEDIM> g;
//    return g;
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
#else
    virtual Tensor<2,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("why called?");
      Tensor<2,SPACEDIM> g;
      return g;
    }
    virtual void gradient(NumPtr<Tensor<2,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
#endif
    virtual void printOn(ostream &os) const {
      os << "ConstantMagneticFieldStrengthFunctionSystem: val = "
         << val << endl;
      D_CHOOSE(FunctionSystem,FunctionSystem,TensorFunctionSystem)::
        printOn(os);
    }
};
class VariableMagneticFieldStrengthFunctionSystem : 
public D_CHOOSE(FunctionSystem,FunctionSystem,TensorFunctionSystem) {
  protected:
    explicit VariableMagneticFieldStrengthFunctionSystem(
    const D_CHOOSE(FunctionSystem,FunctionSystem,TensorFunctionSystem)&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    VariableMagneticFieldStrengthFunctionSystem() {;}
    virtual ~VariableMagneticFieldStrengthFunctionSystem() {;}
    virtual HCURL_MULTIPLIER_DATA_TYPE value(const Point<SPACEDIM> &p,int)
    const {
      HCURL_MULTIPLIER_DATA_TYPE function_values;
#if (SPACEDIM==1)
      function_values=(-sin(p[0]-1.)+cos(p[0]+1.))/cos(2.);
#elif (SPACEDIM==2)
      function_values=cos(p[0])*exp(-p[1]);
#else
      OBSOLETE("not programmed");
#endif
      return function_values;
    }
#if (SPACEDIM<=2)
    virtual void value(const Point<SPACEDIM> &p,Vector &function_values)
    const {
# if (SPACEDIM==1)
      function_values[0]=(-sin(p[0]-1.)+cos(p[0]+1.))/cos(2.);
#elif (SPACEDIM==2)
      function_values[0]=cos(p[0])*exp(-p[1]);
# else
      OBSOLETE("not programmed");
# endif
    }
#else
    virtual void value(const Point<SPACEDIM> &p,
    NumPtr<Tensor<1,SPACEDIM> > &function_values) const {
      OBSOLETE("not programmed");
    }
#endif
#if (SPACEDIM<=2)
    virtual Tensor<1,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("why called?");
    }
    virtual void gradient(NumPtr<Tensor<1,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
#else
    virtual Tensor<2,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("why called?");
    }
    virtual void gradient(NumPtr<Tensor<2,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("not programmed");
    }
#endif
    virtual void printOn(ostream &os) const {
      os << "VariableMagneticFieldStrengthFunctionSystem" << endl;
#if (SPACEDIM<=2)
      FunctionSystem::printOn(os);
#else
      TensorFunctionSystem::printOn(os);
#endif
    }
};
//specify E
class ConstantElectricFieldStrengthFunctionSystem : 
public TensorFunctionSystem {
  private:
    Tensor<1,SPACEDIM> val;
  protected:
    explicit ConstantElectricFieldStrengthFunctionSystem(
    const ConstantElectricFieldStrengthFunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    ConstantElectricFieldStrengthFunctionSystem(
      const Tensor<1,SPACEDIM> &E) : val(E) {}
    virtual ~ConstantElectricFieldStrengthFunctionSystem() {}
    virtual Tensor<1,SPACEDIM> value(const Point<SPACEDIM> &p,int) const {
      return val;
    }
    virtual void value(const Point<SPACEDIM> &p,
    NumPtr<Tensor<1,SPACEDIM> > &function_values) const {
      function_values[0]=val;
    }
    virtual Tensor<2,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("should not be called");
      Tensor<2,SPACEDIM> g;
      return g;
    }
    virtual void gradient(NumPtr<Tensor<2,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("should not be called");
    }
    virtual void printOn(ostream &os) const {
      os << "ConstantElectricFieldStrengthFunctionSystem: val = " << val
         << endl;
      TensorFunctionSystem::printOn(os);
    }
};
class VariableElectricFieldStrengthFunctionSystem : 
public TensorFunctionSystem {
  protected:
    explicit VariableElectricFieldStrengthFunctionSystem(
    const VariableElectricFieldStrengthFunctionSystem&) {
      OBSOLETE("copy constructor disallowed");
    }
  public:
    VariableElectricFieldStrengthFunctionSystem() {;}
    virtual ~VariableElectricFieldStrengthFunctionSystem() {}
    virtual Tensor<1,SPACEDIM> value(const Point<SPACEDIM> &p,int) const {
      Tensor<1,SPACEDIM> function_value;
#if (SPACEDIM==1)
      function_value[0]=(cos(p[0]-1.)+sin(p[0]+1.))/cos(2.);
#elif (SPACEDIM==2)
      function_value[0]=-cos(p[0])*exp(-p[1]);
      function_value[1]=-sin(p[0])*exp(-p[1]);
#else
      OBSOLETE("not programmed");
#endif
      return function_value;
    }
    virtual void value(const Point<SPACEDIM> &p,
    NumPtr<Tensor<1,SPACEDIM> > &function_values) const {
#if (SPACEDIM==1)
      function_values[0][0]=(cos(p[0]-1.)+sin(p[0]+1.))/cos(2.);
#elif (SPACEDIM==2)
      function_values[0][0]=-cos(p[0])*exp(-p[1]);
      function_values[0][1]=-sin(p[0])*exp(-p[1]);
#else
      OBSOLETE("not programmed");
#endif
    }
    virtual Tensor<2,SPACEDIM> gradient(const Point<SPACEDIM> &p,int)
    const {
      OBSOLETE("should not be called");
    }
    virtual void gradient(NumPtr<Tensor<2,SPACEDIM> >&,
    const Point<SPACEDIM>&) const {
      OBSOLETE("should not be called");
    }
    virtual void printOn(ostream &os) const {
      os << "VariableElectricFieldStrengthFunctionSystem" << endl;
      TensorFunctionSystem::printOn(os);
    }
};

// 1D:
//   - epsilon E_1 - partial_x (i H_2 ) = i J_1
//   - mu (i H_2) + partial_x E_1 = - M_2
//   solution with J_1 = M_2 = 0:
//     E_1={ E_1(-1) cos(sqrt[epsilon mu] [x-1])
//         + i H_2(1) sqrt(mu/epsilon) sin(sqrt[epsilon mu] [x+1]) }
//        /cos(2 sqrt[epsilon mu])
//   i H_2={- E_1(-1) sqrt(epsilon/mu) sin{sqrt(epsilon mu) [x-1]}
//         + i H_2(1) cos(sqrt[epsilon mu] [x+1]) }
//        /cos(2 sqrt[epsilon mu])
//   We take E_1(-1)=1, i H_2(1)=1
// 2D:
//   - epsilon E_0 + partial_{x_1} (i H_2) = i J_0
//   - epsilon E_1 - partial_{x_0} (i H_2) = i J_1
//   - mu (i H_2) + partial_{x_0} E_1 - partial_{x_1} E_0 = - M_2
//   solution with J_0 = J_1 = M_2 = 0:
//     E_0=-sqrt{mu/epsilon} cos(sqrt[epsilon mu] x_0)
//                           exp(-sqrt[epsilon mu] x_1)
//     E_1=-sqrt{mu/epsilon} sin(sqrt[epsilon mu] x_0)
//                           exp(-sqrt[epsilon mu] x_1)
//     i H_2=cos(sqrt[epsilon mu] x_0) exp(-sqrt[epsilon mu] x_1)
class TimeHarmonicMaxwellModel : public HcurlMaterialModel {
  private:
    REAL permittivity; // epsilon
    REAL permeability; // mu
    REAL cinv; // sqrt(permittivity*permeability)
    REAL rat; // sqrt(permeability/permittivity)
    Tensor<1,SPACEDIM> applied_electric_current_density; // J_a
    HCURL_MULTIPLIER_DATA_TYPE applied_magnetic_current_density; // M_a

    virtual void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<SPACEDIM> >&, const NumPtr<REAL>&,
    const Array<2,REAL>&,const Array<2,Tensor<1,SPACEDIM> >&,
    Matrix&,Vector&) const {
      OBSOLETE("prohibit MaterialModel version of this function");
    }
  public:
    TimeHarmonicMaxwellModel() : permittivity(1.),permeability(1.) {
      cinv=sqrt(permittivity*permeability);
      rat=sqrt(permeability/permittivity);
      applied_electric_current_density[0]=0.;
      for (int d=1;d<SPACEDIM;d++) applied_electric_current_density[d]=0.;
#if (SPACEDIM!=3)
      applied_magnetic_current_density=0.;
#else
      for (int d=0;d<3;d++) applied_magnetic_current_density[d]=0.;
#endif
    }
    Tensor<1,SPACEDIM> exactElectricFieldStrength(
    const Point<SPACEDIM> &x) const {
      Tensor<1,SPACEDIM> ret;
#if (SPACEDIM==1)
      ret[0]=(cos(cinv*(x[0]-1.))+rat*sin(cinv*(x[0]+1.)))/cos(2.*cinv);
#elif (SPACEDIM==2)
      ret[0]=-rat*cos(cinv*x[0])*exp(-cinv*x[1]);
      ret[1]=-rat*sin(cinv*x[0])*exp(-cinv*x[1]);
#else
      OBSOLETE("not programmed");
#endif
      return ret;
    }
    HCURL_MULTIPLIER_DATA_TYPE exactMagneticFieldStrength(
    const Point<SPACEDIM> &x) const {
      HCURL_MULTIPLIER_DATA_TYPE ret;
#if (SPACEDIM==1)
      ret=(-sin(cinv*(x[0]-1.))/rat+cos(cinv*(x[0]+1.)))/cos(2.*cinv);
#elif (SPACEDIM==2)
      ret=cos(cinv*x[0])*exp(-cinv*x[1]);
#else
      OBSOLETE("not programmed");
#endif
      return ret;
    }
#if (SPACEDIM==1)
    void elementStiffnessMatrixAndRHS(const NumPtr<Point<1> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,Tensor<1,1> > *sv,
    const Array<2,HCURL_MULTIPLIER_DATA_TYPE> *sc,const Array<2,REAL> *msv,
    Matrix &cell_stiffness_matrix,Vector &cell_rhs) const {
      int dpc=sv->size(0);
      int mdpc=msv->size(0);
      CHECK_SAME(cell_stiffness_matrix.size(0),dpc+mdpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),cell_rhs.size());
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        for (int i=0;i<dpc;i++) {
          Tensor<1,1> svi=(*sv)(i,q);
          svi*=jxw;
          for (int j=0;j<dpc;j++) {
            REAL dot=svi*(*sv)(j,q);
            cell_stiffness_matrix(i,j)-=dot*permittivity;
          }
          HCURL_MULTIPLIER_DATA_TYPE sci=(*sc)(i,q);
          sci*=jxw;
          for (int j=0;j<mdpc;j++) {
            REAL msvj=(*msv)(j,q);
            cell_stiffness_matrix(i,j+dpc)+=sci*msvj;
          }
          cell_rhs[i]-=svi*applied_electric_current_density;
        }
        for (int i=0;i<mdpc;i++) {
          int dpcpi=dpc+i;
          REAL msvi=(*msv)(i,q);
          msvi*=jxw;
          for (int j=0;j<dpc;j++) {
            cell_stiffness_matrix(dpcpi,j)+=msvi*(*sc)(j,q);
          }
          for (int j=0;j<mdpc;j++) {
            int dpcpj=dpc+j;
            REAL dot=msvi*(*msv)(j,q);
            cell_stiffness_matrix(dpcpi,dpcpj)-=dot*permeability;
          }
          cell_rhs[dpcpi]-=msvi*applied_magnetic_current_density;
        }
      }
    }
#endif
#if (SPACEDIM==2)
    void elementStiffnessMatrixAndRHS(
    const NumPtr<Point<2> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,Tensor<1,2> > *sv,
    const Array<2,HCURL_MULTIPLIER_DATA_TYPE> *sc,const Array<2,REAL> *msv,
    Matrix &cell_stiffness_matrix,Vector &cell_rhs) const {
      int dpc=sv->size(0);
      int mdpc=msv->size(0);
      CHECK_SAME(cell_stiffness_matrix.size(0),dpc+mdpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),cell_rhs.size());
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        for (int i=0;i<dpc;i++) {
          Tensor<1,2> svi=(*sv)(i,q);
          HCURL_MULTIPLIER_DATA_TYPE sci=(*sc)(i,q);
          svi*=jxw;
          for (int j=0;j<dpc;j++) {
            cell_stiffness_matrix(i,j)-=(svi*(*sv)(j,q))*permittivity;
          }
          for (int j=0;j<mdpc;j++) {
            cell_stiffness_matrix(i,dpc+j)+=sci*(*msv)(j,q);
          }
          cell_rhs[i]+=svi*applied_electric_current_density;
        }
        for (int i=0;i<mdpc;i++) {
          int dpcpi=dpc+i;
          REAL msvi=(*msv)(i,q)*jxw;
          for (int j=0;j<dpc;j++) {
            HCURL_MULTIPLIER_DATA_TYPE scj=(*sc)(j,q);
            cell_stiffness_matrix(dpcpi,j)+=msvi*scj;
          }
          for (int j=0;j<mdpc;j++) {
            cell_stiffness_matrix(dpcpi,dpc+j)-=
              msvi*permeability*(*msv)(j,q);
          }
          cell_rhs[dpcpi]-=msvi*applied_magnetic_current_density;
        }
      }
    }
#elif (SPACEDIM==3)
    void elementStiffnessMatrixAndRHS(const NumPtr<Point<3> > &q_points,
    const NumPtr<REAL> &jxw_values,const Array<2,Tensor<1,3> > *sv,
    const Array<2,Tensor<1,3> > *sc,const Array<2,Tensor<1,3> > *msv,
    Matrix &cell_stiffness_matrix,Vector &cell_rhs) const {
      int dpc=sv->size(0);
      int mdpc=msv->size(0);
      CHECK_SAME(cell_stiffness_matrix.size(0),dpc+mdpc);
      CHECK_SAME(cell_stiffness_matrix.size(0),cell_rhs.size());
      for (int q=0;q<q_points.getNumber();q++) {
        REAL jxw=jxw_values[q];
        for (int i=0;i<dpc;i++) {
          Tensor<1,3> svi=(*sv)(i,q);
          Tensor<1,3> sci=(*sc)(i,q);
          svi*=jxw;
          sci*=jxw;
          for (int j=0;j<dpc;j++) {
            const Tensor<1,3> &svj=(*sv)(j,q);
            REAL dot=svi*svj;
            cell_stiffness_matrix(i,j)-=dot*permittivity;
          }
          for (int j=0;j<mdpc;j++) {
            cell_stiffness_matrix(i,dpc+j)+=sci*(*msv)(j,q);
          }
          cell_rhs[i]+=svi*applied_electric_current_density;
        }
        for (int i=0;i<mdpc;i++) {
          int dpcpi=dpc+i;
          Tensor<1,3> msvi=(*msv)(i,q)*jxw;
          for (int j=0;j<dpc;j++) {
            cell_stiffness_matrix(dpcpi,j)+=msvi*(*sc)(j,q);
          }
          for (int j=0;j<mdpc;j++) {
            int dpcpj=dpc+j;
            const Tensor<1,3> &msvj=(*msv)(j,q);
            REAL dot=msvi*msvj;
            cell_stiffness_matrix(dpcpi,dpcpj)-=dot*permeability;
          }
          cell_rhs[dpcpi]-=msvi*applied_magnetic_current_density;
        }
      }
    }
#endif
};

#ifdef TIMING
//the most expensive routines are the linear system solve and
//  interpolateBoundaryValues
//the latter requires computation of the shape functions on an element
//  at iteratively computed canonical points corresponding to boundary:
//TimedObjectList tl;
#endif

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
#ifdef TIMING
//TimedObject *z3=
//  OPERATOR_NEW TimedObject("main interpolateBoundaryValues",tl);
//TimedObject *z5=OPERATOR_NEW TimedObject("main solve",tl);
//TimedObject *z10=OPERATOR_NEW TimedObject("main refine",tl);
#endif
  Tessellation tess;
  const LEVEL_CLASS *level=tess.coarsestLevel();

//rectangular domain consisting of one interval/quadrilateral/hexagon
  Point<SPACEDIM> low(D_DECL(-1.,-1.,-1.));
//Point<SPACEDIM> low(0.,0.,0.);
  Point<SPACEDIM> edge[SPACEDIM];
    edge[0]=Point<SPACEDIM>(D_DECL(2.,0.,0.));
#if (SPACEDIM>=2)
    edge[1]=Point<SPACEDIM>(D_DECL(0.,2.,0.));
#endif
#if (SPACEDIM==3)
    edge[2]=Point<SPACEDIM>(0.,0.,2.);
#endif
  int number_elements[SPACEDIM]={D_DECL(nelements,nelements,nelements)};
  bool set_face_ids=true;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
  BoundaryData *boundary_data=0;
#if (SPACEDIM==1)
  boundary_data=generateParallelpiped(tess,low,edge,number_elements,
    set_face_ids,boundary,boundary_condition);
  for (int i=0;i<2;i++) {
    const Boundary *b=boundary_condition[i]->getBoundary();
    delete boundary_condition[i];
    boundary_condition[i]=
      OPERATOR_NEW HcurlBoundaryCondition(boundary[i]);
    boundary_condition[i]->setBoundary(b);
    level->CODIMENSION1_LEVEL_CLASS::setBoundaryCondition(i,
      boundary_condition[i]);
  }
#elif (SPACEDIM==2)
  switch (mesh_elements) {
    case QUADRILATERAL_MESH_ELEMENTS: {
      boundary_data=generateParallelpiped(tess,low,edge,number_elements,
        set_face_ids,boundary,boundary_condition);
      for (int i=0;i<2*SPACEDIM;i++) {
        const Boundary *b=boundary_condition[i]->getBoundary();
        delete boundary_condition[i];
        boundary_condition[i]=
          OPERATOR_NEW HcurlBoundaryCondition(boundary[i]);
        boundary_condition[i]->setBoundary(b);
      }
      const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          bool found=false;
          for (int i=0;i<boundary_condition.getNumber();i++) {
            const BoundaryCondition *bc=boundary_condition[i];
            const Boundary *b=bc->getBoundary();
            bool all_on_boundary=true;
            for (int v=0;v<face_element->numberVertices();v++) {
              int vi=face_element->vertexIndex(v);
              const Point<SPACEDIM> &ver=vertices[vi];
              if (abs(b->constraint(ver))>1.e-5) {
                all_on_boundary=false;
                break;
              }
            }
            if (all_on_boundary) {
              codimension1_level->setBoundaryCondition(bc);
              found=true;
              break;
            }
          }
        }
      }
      break;
    }
    case TRIANGULAR_MESH_ELEMENTS: {
//    the following tessellation has all possible faces on outer boundary
      NumPtr<Point<2> > vertices(5);
      for (int i=0;i<vertices.getNumber();i++) vertices[i]=low;
      vertices[1]+=edge[0];
      vertices[3]+=edge[1];
      vertices[4]+=edge[0]; vertices[4]+=edge[1];
      vertices[2]=(vertices[0]+vertices[1]+vertices[3]+vertices[4])*0.25;
      NumPtr<CellData> cells(4);
      for (int c=0;c<cells.getNumber();c++) cells[c].allocate(3);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;

      cells[1].vertex(0)=1;
      cells[1].vertex(1)=4;
      cells[1].vertex(2)=2;

      cells[2].vertex(0)=2;
      cells[2].vertex(1)=4;
      cells[2].vertex(2)=3;

      cells[3].vertex(0)=0;
      cells[3].vertex(1)=2;
      cells[3].vertex(2)=3;
      boundary_data=tess.create(vertices,cells);
      boundary.allocate(4);
      boundary_condition.allocate(6);
      Point<2> center(vertices[2]);
      boundary[0]=OPERATOR_NEW PlaneBoundary(center,vertices[0],
        vertices[3]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(center,vertices[1],
        vertices[4]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(center,vertices[0],
        vertices[1]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(center,vertices[3],
        vertices[4]);
      for (int i=0;i<4;i++) {
        boundary_condition[i]=
          OPERATOR_NEW HcurlBoundaryCondition(boundary[i]);
      }

      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<SPACEDIM> center=face_element->center(vertices);
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              break;
            }
          }
        }
      }
      tess.synchronize(boundary_data);
      break;
    }
    default:
      OBSOLETE("bad mesh_elements");
  }
#else
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS: {
      boundary_data=generateParallelpiped(tess,low,edge,number_elements,
        set_face_ids,boundary,boundary_condition);
      for (int i=0;i<2*SPACEDIM;i++) {
        const Boundary *b=boundary_condition[i]->getBoundary();
        delete boundary_condition[i];
        boundary_condition[i]=
          OPERATOR_NEW HcurlBoundaryCondition(boundary[i]);
        boundary_condition[i]->setBoundary(b);
      }
      const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          bool found=false;
          for (int i=0;i<boundary_condition.getNumber();i++) {
            const BoundaryCondition *bc=boundary_condition[i];
            const Boundary *b=bc->getBoundary();
            bool all_on_boundary=true;
            for (int v=0;v<face_element->numberVertices();v++) {
              int vi=face_element->vertexIndex(v);
              const Point<SPACEDIM> &ver=vertices[vi];
              if (abs(b->constraint(ver))>1.e-5) {
                all_on_boundary=false;
                break;
              }
            }
            if (all_on_boundary) {
              codimension1_level->setBoundaryCondition(bc);
              found=true;
              break;
            }
          }
        }
      }
      for (bool doit=tess.firstCodimension2ActiveShape();doit;
      doit=tess.nextCodimension2ActiveShape()) {
        Level1 *codimension2_level=
          tess.presentCodimension2Level();
        pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
          codimension2_level->getBoundaryConditions();
        const BoundaryCondition *bc=bcs.first;
        if (bc==0) continue;
        Shape1 *face_element=codimension2_level->getShape();
        codimension2_level->setBoundaryCondition(0);
        int count=0;
        for (int i=0;i<boundary.getNumber();i++) {
          const BoundaryCondition *bc=boundary_condition[i];
          const Boundary *b=bc->getBoundary();
          bool all_on_boundary=true;
          for (int v=0;v<face_element->numberVertices();v++) {
            int vi=face_element->vertexIndex(v);
            const Point<SPACEDIM> &ver=vertices[vi];
            if (abs(b->constraint(ver))>1.e-5) {
              all_on_boundary=false;
              break;
            }
          }
          if (all_on_boundary) {
            codimension2_level->setBoundaryCondition(bc);
            count++;
          }
        }
      }
      break;
    }
    case PRISM_MESH_ELEMENTS: {
//    the following tessellation has all possible faces on outer boundary
      NumPtr<Point<3> > vertices(10);
      for (int i=0;i<vertices.getNumber();i++) vertices[i]=low;
      vertices[1]+=edge[0];
      vertices[2]+=edge[0]*0.5; vertices[2]+=edge[1]*0.5; 
      vertices[3]+=edge[1];
      vertices[4]+=edge[0]; vertices[4]+=edge[1];
      vertices[5]+=edge[2];
      vertices[6]+=edge[0]; vertices[6]+=edge[2];
      vertices[7]+=edge[0]*0.5; vertices[7]+=edge[1]*0.5; 
                                vertices[7]+=edge[2];
      vertices[8]+=edge[1]; vertices[8]+=edge[2];
      vertices[9]+=edge[0]; vertices[9]+=edge[1]; vertices[9]+=edge[2];
      NumPtr<CellData> cells(4);
      for (int c=0;c<cells.getNumber();c++) cells[c].allocate(6);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;
      cells[0].vertex(3)=5;
      cells[0].vertex(4)=6;
      cells[0].vertex(5)=7;

      cells[1].vertex(0)=0;
      cells[1].vertex(1)=2;
      cells[1].vertex(2)=3;
      cells[1].vertex(3)=5;
      cells[1].vertex(4)=7;
      cells[1].vertex(5)=8;

      cells[2].vertex(0)=2;
      cells[2].vertex(1)=4;
      cells[2].vertex(2)=3;
      cells[2].vertex(3)=7;
      cells[2].vertex(4)=9;
      cells[2].vertex(5)=8;

      cells[3].vertex(0)=1;
      cells[3].vertex(1)=4;
      cells[3].vertex(2)=2;
      cells[3].vertex(3)=6;
      cells[3].vertex(4)=9;
      cells[3].vertex(5)=7;
      boundary_data=tess.create(vertices,cells);
      boundary.allocate(6);
      boundary_condition.allocate(6);
      Point<3> center(low+(edge[0]+edge[1]+edge[2])*0.5);
      boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1],
        low+edge[2]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
        low+edge[0]+edge[1],low+edge[0]+edge[2]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[2],
        low+edge[0]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
        low+edge[1]+edge[2],low+edge[1]+edge[0]);
      boundary[4]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0],
        low+edge[1]);
      boundary[5]=OPERATOR_NEW PlaneBoundary(center,low+edge[2],
        low+edge[2]+edge[0],low+edge[2]+edge[1]);
      for (int i=0;i<2*SPACEDIM;i++) {
        boundary_condition[i]=
          OPERATOR_NEW HcurlBoundaryCondition(boundary[i]);
      }
      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<SPACEDIM> center=face_element->center(vertices);
          bool found=false;
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              found=true;
              break;
            }
          }
        }
      }
      tess.synchronize(boundary_data);
      break;
    }
    case TETRAHEDRAL_MESH_ELEMENTS: {
//    the following tessellation has all possible faces on outer boundary
      NumPtr<Point<3> > vertices(8);
      for (int i=0;i<vertices.getNumber();i++) vertices[i]=low;
      vertices[1]+=edge[0];
      vertices[2]+=edge[1];
      vertices[3]+=edge[0]; vertices[3]+=edge[1];
      vertices[4]+=edge[2];
      vertices[5]+=edge[0]; vertices[5]+=edge[2];
      vertices[6]+=edge[1]; vertices[6]+=edge[2];
      vertices[7]+=edge[0]; vertices[7]+=edge[1]; vertices[7]+=edge[2];
      NumPtr<CellData> cells(6);
      for (int c=0;c<cells.getNumber();c++) cells[c].allocate(4);
      cells[0].vertex(0)=0;
      cells[0].vertex(1)=1;
      cells[0].vertex(2)=2;
      cells[0].vertex(3)=4;

      cells[1].vertex(0)=1;
      cells[1].vertex(1)=4;
      cells[1].vertex(2)=5;
      cells[1].vertex(3)=6;

      cells[2].vertex(0)=1;
      cells[2].vertex(1)=2;
      cells[2].vertex(2)=4;
      cells[2].vertex(3)=6;

      cells[3].vertex(0)=1;
      cells[3].vertex(1)=2;
      cells[3].vertex(2)=7;
      cells[3].vertex(3)=3;

      cells[4].vertex(0)=1;
      cells[4].vertex(1)=5;
      cells[4].vertex(2)=7;
      cells[4].vertex(3)=6;

      cells[5].vertex(0)=1;
      cells[5].vertex(1)=2;
      cells[5].vertex(2)=6;
      cells[5].vertex(3)=7;
      boundary_data=tess.create(vertices,cells);
      boundary.allocate(6);
      boundary_condition.allocate(6);
      Point<3> center(low+(edge[0]+edge[1]+edge[2])*0.5);
      boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1],
        low+edge[2]);
      boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
        low+edge[0]+edge[1],low+edge[0]+edge[2]);
      boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[2],
        low+edge[0]);
      boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
        low+edge[1]+edge[2],low+edge[1]+edge[0]);
      boundary[4]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0],
        low+edge[1]);
      boundary[5]=OPERATOR_NEW PlaneBoundary(center,low+edge[2],
        low+edge[2]+edge[0],low+edge[2]+edge[1]);
      for (int i=0;i<2*SPACEDIM;i++) {
        boundary_condition[i]=
          OPERATOR_NEW HcurlBoundaryCondition(boundary[i]);
      }

      for (bool doit=tess.firstCodimension1ActiveShape();doit;
      doit=tess.nextCodimension1ActiveShape()) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          tess.presentCodimension1Level();
        if (codimension1_level->getBoundaryCondition()!=0) {
//        replace bogus_boundary_condition in Tessellation::create
          CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape();
          Point<SPACEDIM> center=face_element->center(vertices);
          for (int i=0;i<boundary.getNumber();i++) {
            if (abs(boundary[i]->constraint(center))<1.e-5) {
              codimension1_level->
                setBoundaryCondition(boundary_condition[i]);
              break;
            }
          }
        }
      }
      tess.synchronize(boundary_data);
      break;
    }
    default:
      OBSOLETE("bad mesh_elements");
  }
#endif
#if (SPACEDIM>=2)
  {
    tess.checkFaceBoundaryCondition(boundary_condition);
  }
#endif
#if (SPACEDIM==3)
  tess.checkLineBoundaryCondition(boundary_condition);
#endif

  TimeHarmonicMaxwellModel thmm;
  tess.setMaterialModel(&thmm);

//specified E at left is essential boundary condition
  HcurlBoundaryCondition *ebc=
    dynamic_cast<HcurlBoundaryCondition*>(boundary_condition[0]);
  CHECK_POINTER(ebc);
#if (SPACEDIM!=2)
  Tensor<1,SPACEDIM> E;
  E[0]=1.;
  ConstantElectricFieldStrengthFunctionSystem sefsfs(E);
  ebc->setFunctionSystem(&sefsfs);
#else
  VariableElectricFieldStrengthFunctionSystem vefsfs;
  ebc->setFunctionSystem(&vefsfs);
#endif
  ebc->makeEssential();

//specified tangential H on right is natural boundary condition
  HcurlBoundaryCondition *nbc=
    dynamic_cast<HcurlBoundaryCondition*>(boundary_condition[1]);
  CHECK_POINTER(nbc);
#if (SPACEDIM!=2)
  HCURL_MULTIPLIER_DATA_TYPE H;
# if (SPACEDIM==1)
  H=1.;
# else
  H[0]=1.;
# endif
  ConstantMagneticFieldStrengthFunctionSystem fmfsfs(H);
  nbc->setFunctionSystem(&fmfsfs);
#else
  VariableMagneticFieldStrengthFunctionSystem vmfsfs;
  nbc->setFunctionSystem(&vmfsfs);
#endif
  nbc->makeNatural();

//other boundaries (if they exist) are homogeneous and essential
#if (SPACEDIM>=2)
# if (SPACEDIM==2)
  VariableElectricFieldStrengthFunctionSystem zefsfs;
# else
  Tensor<1,3> zero;
  ConstantElectricFieldStrengthFunctionSystem zefsfs(zero);
# endif
  for (int d=2;d<2*SPACEDIM;d++) {
    HcurlBoundaryCondition *zbc=
      dynamic_cast<HcurlBoundaryCondition*>(boundary_condition[d]);
    zbc->setFunctionSystem(&zefsfs);
    zbc->makeEssential();
  }
#endif

//(linear or) tensor product linear coordinate mapping
  MappingL mapping;

//set polynomials for each shape, face quadratures and dofs per object
  Nedelec1FiniteElement fe(degree);
//Nedelec2FiniteElement fe(degree);

//interior quadrature rules
  map<int,Quadrature<SPACEDIM> > qmap;
#if (SPACEDIM==1)
//qmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
//  GaussianQuadrature<1>::argGivenOrder(2*degree+2));
  qmap[Interval::verticesPerShape()]=NewtonCotesQuadrature<1>(
    NewtonCotesQuadrature<1>::argGivenOrder(2*degree+2));
#elif (SPACEDIM==2)
  switch (mesh_elements) {
    case QUADRILATERAL_MESH_ELEMENTS:
//    need order>=2*degree+1
      qmap[Quadrilateral::verticesPerShape()]=GaussianQuadrature<2>(
        GaussianQuadrature<2>::argGivenOrder(2*degree+3));
//    qmap[Quadrilateral::verticesPerShape()]=
//      NewtonCotesQuadrature<2>(max(2,2*degree+1));
      break;
    case TRIANGULAR_MESH_ELEMENTS:
      qmap[Triangle::verticesPerShape()]=TriangleGaussianQuadrature(
        TriangleGaussianQuadrature::argGivenOrder(2*degree+3));
      break;
    default:
      OBSOLETE("bad mesh_elements");
  }
#else
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS:
//    need order>=2*degree+1
      qmap[Hexahedron::verticesPerShape()]=GaussianQuadrature<SPACEDIM>(
        GaussianQuadrature<SPACEDIM>::argGivenOrder(2*degree+3));
//    qmap[Hexahedron::verticesPerShape()]=
//      NewtonCotesQuadrature<SPACEDIM>(max(2,2*degree+1));
      break;
    case PRISM_MESH_ELEMENTS:
      qmap[Prism::verticesPerShape()]=
        ProductQuadrature(
          TriangleGaussianQuadrature(
            TriangleGaussianQuadrature::argGivenOrder(2*degree+3)),
          GaussianQuadrature<1>(
            GaussianQuadrature<1>::argGivenOrder(2*degree+3)));
      break;
    case TETRAHEDRAL_MESH_ELEMENTS:
      qmap[Tetrahedron::verticesPerShape()]=TetrahedronGaussianQuadrature(
        TetrahedronGaussianQuadrature::argGivenOrder(2*degree+3));
      break;
    default:
      OBSOLETE("bad mesh_elements");
  }
#endif

//boundary quadrature rules
  map<int,Quadrature<SPACEDIM-1> > nqmap;
#if (SPACEDIM>=2)
//natural bc's
//should be exact for polynomials of shape function degree
//  plus exact solution degree-1
//  plus Jacobian degree
# if (SPACEDIM==2)
  nqmap[Interval::verticesPerShape()]=GaussianQuadrature<1>(
        GaussianQuadrature<1>::argGivenOrder(2*degree+2));
# else
  switch (mesh_elements) {
    case HEXAHEDRAL_MESH_ELEMENTS: {
      int nv=Quadrilateral::verticesPerShape();
      nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(
        GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
      break;
    }
    case PRISM_MESH_ELEMENTS: {
      int nv=Quadrilateral::verticesPerShape();
      nqmap[nv]=GaussianQuadrature<SPACEDIM-1>(
        GaussianQuadrature<SPACEDIM-1>::argGivenOrder(2*degree+2));
      nv=Triangle::verticesPerShape();
      nqmap[nv]=TriangleGaussianQuadrature(
        TriangleGaussianQuadrature::argGivenOrder(2*degree));
      break;
    }
    case TETRAHEDRAL_MESH_ELEMENTS: {
//    face shape function is polynomial of order=degree
//    exact solution is polynomial of order=solution order-1
//    TriangleProductGaussianQuadrature has order 2*n-1 
//      where n=constructor argument
      int nv=Triangle::verticesPerShape();
      nqmap[nv]=TriangleGaussianQuadrature(
        TriangleGaussianQuadrature::argGivenOrder(2*degree));
      break;
    }
    default:
      OBSOLETE("bad mesh_elements");
  }
# endif
#endif

//graphics setup
  Palette pal;
#if (SPACEDIM==1)
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool *plot_gt=OPERATOR_NEW XYGraphTool(
    "electric field intensity E vs position","x","E",0.,1.,0.,1.,&cmap,
    0,0.5);
  XYGraphTool *multiplier_plot_gt=OPERATOR_NEW XYGraphTool(
    "magnetic field intensity H vs position","x","H",0.,1.,0.,1.,&cmap,
    0,0.5);
#elif (SPACEDIM==2)
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]={0.,0.},rhi[2]={1.,1.};
  AreaGraphTool mesh_gt("grid","x0","x1",rlo,rhi,&cmap,0,0.5);
  AreaGraphTool *plot_gt=OPERATOR_NEW AreaGraphTool(
    "electric field intensity","x0","x1",rlo,rhi,&cmap,0,0.5);
  AreaGraphTool *multiplier_plot_gt=OPERATOR_NEW AreaGraphTool(
    "magnetic field intensity","x0","x1",rlo,rhi,&cmap,0,0.5);
#else
  BOOLEAN use_lighting=TRUE;
  Vector3 vlow(0.,0.,0.);
  Vector3 vhigh(1.,1.,1.);
  VolGraphTool mesh_gt("grid",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,&mesh_gt);
    mesh_gt.drawBox(vlow,vhigh);
  }
  VolGraphTool *plot_gt=OPERATOR_NEW VolGraphTool(
    "electric field intensity",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,plot_gt);
    plot_gt->drawBox(vlow,vhigh);
  }
  VolGraphTool *multiplier_plot_gt=OPERATOR_NEW VolGraphTool(
    "magnetic field intensity",vlow,vhigh,pal,use_lighting,0.5);
  { VolGraphTool::DrawDelimiter dd(TRUE,multiplier_plot_gt);
    multiplier_plot_gt->drawBox(vlow,vhigh);
  }
#endif

//data structure to keep track of degrees of freedom in solver
  HcurlDofHandler hcurl_dof_handler(tess,fe);
  hcurl_dof_handler.distributeDofs();
  int n_dofs=hcurl_dof_handler.numberDofs();
#ifdef DEBUG
//hcurl_dof_handler.printDofs();

  cout << "\tnumber of active elements = " 
       << tess.numberActiveShapes() << endl;
  cout << "\tNumber of degrees of freedom = " << n_dofs << endl;
  cout << "\tassemble initial linear system..." << endl;
#endif
//data structure to keep track of linear system storage
  int band_width=hcurl_dof_handler.maxCouplingsBetweenDofs();
  SparsityPattern *stiffness_sparsity_pattern=
    OPERATOR_NEW SparsityPattern(n_dofs,n_dofs,band_width,true);
  hcurl_dof_handler.makeSparsityPattern(*stiffness_sparsity_pattern);
  stiffness_sparsity_pattern->compress();
//storage for stiffness matrix, right-hand side and solution
  SparseMatrix system_stiffness_matrix(stiffness_sparsity_pattern);
  Vector system_rhs(n_dofs,0.);
  Vector solution(n_dofs,0.);

//assemble initial linear system
  hcurl_dof_handler.assembleLinearSystem(mapping,qmap,nqmap,
    system_stiffness_matrix,system_rhs);
//compute essential boundary values at mapped points on bdry
//calls FunctionSystem in BoundaryCondition's to set boundary values
  map<int,REAL> boundary_values;
  {
//  TRACER_CALL(t,"interpolateBoundaryValues");
#ifdef TIMING
//  Timer y(z3);
#endif
    hcurl_dof_handler.interpolateBoundaryValues(mapping,boundary_values);
  }

//modify initial linear system and rhs for essential bc's
  bool preserve_symmetry=false;
  hcurl_dof_handler.applyBoundaryValues(boundary_values,
    system_stiffness_matrix,system_rhs,preserve_symmetry);
  hcurl_dof_handler.applyBoundaryValues(boundary_values,solution);
  PreconditionIdentity pi;
  RelativeErrorControl gmres_solver_control(2*system_rhs.size(),
    1.e-5,false,false);
//GMRESSolver solver(gmres_solver_control);
  BicgstabSolver solver(gmres_solver_control);
  {
#ifdef DEBUG
    cout << "\tsolve linear system..." << endl;
#endif
#ifdef TIMING
//  Timer y(z5);
#endif
    solver.solve(system_stiffness_matrix,solution,system_rhs,pi);
  }
//compute solution values for graphical display
  HcurlDataOut data_out(hcurl_dof_handler,mapping);
  int total_solution_dofs=hcurl_dof_handler.totalSolutionDofs();
  Vector *efield=solution.subVector(0,total_solution_dofs);
  Vector *hfield=solution.subVector(total_solution_dofs,
    n_dofs-total_solution_dofs);
  data_out.addDataVector(*efield,"electric field intensity",
    HcurlDataOut::TYPE_DOF_DATA);
  data_out.addDataVector(*hfield,"magnetic field intensity",
    HcurlDataOut::TYPE_MULTIPLIER_DATA);
  int n_subdivisions=degree+1;
//int n_subdivisions=5;
  data_out.buildPatches(n_subdivisions);
  bool quit_button=D_CHOOSE(false,false,true);
#if (SPACEDIM>1)
  data_out.drawGrid(&mesh_gt,quit_button);
#endif
  data_out.plot("electric field intensity",plot_gt,quit_button);
  data_out.plot("magnetic field intensity",multiplier_plot_gt,quit_button);
#if (SPACEDIM==1)
  if (!quit_button) {
  XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#elif (SPACEDIM==2)
  if (!quit_button) {
  AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#endif
  delete efield; efield=0;
  delete hfield; hfield=0;
  data_out.clearInputDataReferences();

  delete stiffness_sparsity_pattern;
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  delete plot_gt;
  delete multiplier_plot_gt;
  return EXIT_SUCCESS;
}

template class _Rb_tree<
  int, 
  pair<int const, double>, 
  _Select1st<pair<int const, double> >, 
  less<int>, 
  allocator<pair<int const, double> > 
>;
