//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//computes f(y) and returns || f(y) ||
double dfnrmk(int fi,int la,double dt,double &fsize,
const double *saturationold,const double *x, const double *y, double *f, 
double *mobo,double *mobw,double *po,double *velo,double *velw,
bool debug) {
//TRACER_CALL(tr,"dfnrmk");
  int neq=2*(la-fi);
  int n_interior_boundary_cells=la-fi+2;
  double *pressurenew=
    OPERATOR_NEW_BRACKET(double,n_interior_boundary_cells);
  double *saturationnew=
    OPERATOR_NEW_BRACKET(double,n_interior_boundary_cells);

  const double *sold=saturationold+(1-fi);
  double *pnew=pressurenew+(1-fi);
  double *snew=saturationnew+(1-fi);
  int ij=0;
  pnew[fi-1]=F77NAME(blcom).pres_left;
  snew[fi-1]=sold[fi-1];
  for (int i=fi;i<la;i++) {
    snew[i]=y[ij++];
    pnew[i]=y[ij++];
  }
  pnew[la]=F77NAME(blcom).pres_right;
  snew[la]=sold[la];
#ifdef DEBUG
//if (debug) {
//  for (int i=fi-1;i<=la;i++) {
//    cout << "\tsnew[" << i << "] = " << snew[i] << endl;
//  }
//  for (int i=fi-1;i<=la;i++) {
//    cout << "\tpnew[" << i << "] = " << pnew[i] << endl;
//  }
//}
#endif
  double fmax=0.;
  fsize=0.;
  F77NAME(func)(fi,la,dt,fmax,fsize, 
    pressurenew,saturationold,saturationnew,x, 
    f, mobo,mobw,po,velo,velw);
#ifdef DEBUG
//if (debug) {
//  for (int i=0;i<neq;i++) {
//    cout << "\tf[" << i << "] = " << f[i] << endl;
//  }
//}
#endif
  delete [] pressurenew;
  delete [] saturationnew;
  return F77NAME(dnrm2)(neq,f,1)
        /sqrt(static_cast<double>(neq));
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//back-tracking line search to minimize || f(y+step*rl) ||
//step_norm=weighted RMS norm of p
bool dlinsk(int fi,int la,double dt,const double *step,
const double *saturationold,const double *x, double &fnrm, 
double *y, double *f, double *mobo,double *mobw,double *po,double *velo,
double *velw) {
//TRACER_CALL(tr,"dlinsk");
  int neq=2*(la-fi);
  double step_norm=F77NAME(dnrm2)(neq,step,1);
  double y_norm=F77NAME(dnrm2)(neq,y,1);
#ifdef DEBUG
//for (int i=0;i<neq;i++) {
//  cout << "\ty[" << i << "]  = " << y[i] << endl;
//}
//for (int i=0;i<neq;i++) {
//  cout << "\tstep[" << i << "]  = " << step[i] << endl;
//}
//cout << "\tstptol = " << y_norm*stptol << endl;
//cout << "\tstep_norm = " << step_norm << endl;
#endif
  if (step_norm<y_norm*stptol) return true;
  double alpha=2.e-4;
  double f1nrm=0.5*fnrm*fnrm;
  bool success=false;
  double *ynew=OPERATOR_NEW_BRACKET(double,neq);
  double fsize=HUGE_VAL;
  for (double rl=1.;rl*step_norm>=y_norm*stptol;rl*=0.5) {
    for (int i=0;i<neq;i++) ynew[i]=y[i]-rl*step[i];
    double fnrmp=dfnrmk(fi,la,dt,fsize,saturationold,x, ynew, f, 
      mobo,mobw,po,velo,velw, false);
    double f1nrmp=0.5*fnrmp*fnrmp;
#ifdef DEBUG
//  cout << "\n\trl = " << rl << endl;
//  cout << "\tf1nrmp = " << f1nrmp << " "
//       << f1nrm*(1.-alpha*rl) << endl;
//  for (int i=0;i<neq;i++) {
//    cout << "\tynew[" << i << "]  = " << ynew[i] << endl;
//  }
//  for (int i=0;i<neq;i++) {
//    cout << "\tf[" << i << "]  = " << f[i] << endl;
//  }
#endif
    if (f1nrmp<=f1nrm*(1.-alpha*rl)) {
      success=true; 
      memcpy(y,ynew,neq*sizeof(double));
      fnrm=fnrmp;
      break;
    }
  }
  delete [] ynew;
  return success;
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//approximate jacobian times vector via difference quotient:
//  jacobian * v approx [f(y+v*cj)-f(y)]/cj
void datv(int fi,int la,double dt,const double *f,
const double *saturationold,const double *v,const double *x,
const double *y,double *jv,double *fn,double *mobo,double *mobw,
double *po,double *velo,double *velw) {
//TRACER_CALL(tr,"datv");
  int n=la-fi;
  int neq=2*n;
  const double *so=saturationold+(1-fi);
  double cj=sqrt(DBL_EPSILON)*F77NAME(dnrm2)(neq,y,1);
  double *saturationnew=OPERATOR_NEW_BRACKET(double,n+2);
  double *pressurenew=OPERATOR_NEW_BRACKET(double,n+2);
  double *sn=saturationnew+(1-fi);
  double *pn=pressurenew+(1-fi);
  int ij=0;
  sn[fi-1]=so[fi-1];
  pn[fi-1]=F77NAME(blcom).pres_left;
  for (int i=fi;i<la;i++) {
    sn[i]=y[ij]+v[ij]*cj;
    ij++;
    pn[i]=y[ij]+v[ij]*cj;
    ij++;
  }
  sn[la]=so[la];
  pn[la]=F77NAME(blcom).pres_right;
  double fmax=0.;
  double fsize=0.;
  F77NAME(func)(fi,la,dt,fmax,fsize, 
    pressurenew,saturationold,saturationnew,x, fn, 
    mobo,mobw,po,velo,velw);
  for (int ij=0;ij<neq;ij++) jv[ij]=(fn[ij]-f[ij])/cj;
  delete [] saturationnew;
  delete [] pressurenew;
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//solve jacobian*soln=rhs using generalized minimum residual method
int dspigm(int fi,int la,int kmp,int maxl,const double *f,
const double *rhs,const double *saturationold,const double *x,
const double *y, double *dl,double *soln,
double *mobo,double *mobw,double *po,double *velo,double *velw) {
//TRACER_CALL(tr,"dspigm");
  ASSERT(kmp<=maxl);
  bool restart=true;
  int n=la-fi;
  int neq=2*n;
  int maxlp1=maxl+1;
  double *fn=OPERATOR_NEW_BRACKET(double,neq);
  double *q=OPERATOR_NEW_BRACKET(double,2*neq);
  double *hes=OPERATOR_NEW_BRACKET(double,maxlp1*maxl);
  double *v=OPERATOR_NEW_BRACKET(double,neq*maxlp1);
  double *work=OPERATOR_NEW_BRACKET(double,maxlp1);
  double cj=sqrt(DBL_EPSILON)*F77NAME(dnrm2)(neq,y,1);
#ifdef DEBUG
/*
  double *dmobods=OPERATOR_NEW_BRACKET(double,n+2);
  double *dmobwds=OPERATOR_NEW_BRACKET(double,n+2);
  double *dpods=OPERATOR_NEW_BRACKET(double,n+2);
  double *dvelodpl=OPERATOR_NEW_BRACKET(double,n+1);
  double *dvelodpr=OPERATOR_NEW_BRACKET(double,n+1);
  double *dvelodsl=OPERATOR_NEW_BRACKET(double,n+1);
  double *dvelodsr=OPERATOR_NEW_BRACKET(double,n+1);
  double *dvelwdpl=OPERATOR_NEW_BRACKET(double,n+1);
  double *dvelwdpr=OPERATOR_NEW_BRACKET(double,n+1);
  double *dvelwdsl=OPERATOR_NEW_BRACKET(double,n+1);
  double *unsplit_matrix=OPERATOR_NEW_BRACKET(double,10*2*n);
  double *saturationnew=OPERATOR_NEW_BRACKET(double,n+2);
  double *pressurenew=OPERATOR_NEW_BRACKET(double,n+2);
  double *sn=saturationnew+(1-fi);
  double *pn=pressurenew+(1-fi);
  const double *so=saturationold+(1-fi);
  int ij=0;
  sn[fi-1]=so[fi-1];
  pn[fi-1]=F77NAME(blcom).pres_left;
  for (int i=fi;i<la;i++) {
    sn[i]=y[ij++];
    pn[i]=y[ij++];
  }
  sn[la]=so[la];
  pn[la]=F77NAME(blcom).pres_right;
  F77NAME(jacobian)(fi,la,dt, pressurenew,saturationnew,x, unsplit_matrix,
    dmobods,dmobwds,dpods,dvelodpl,dvelodpr,dvelodsl,dvelodsr,
    dvelwdpl,dvelwdpr,dvelwdsl,mobo,mobw,po,velo,velw);
//for (int i=0;i<10;i++) {
//  cout << "matrix[" << i << ",*] = ";
//  for (int j=0;j<neq;j++) cout << unsplit_matrix[i+j*10] << " ";
//  cout << endl;
//}
*/
#endif

  int iflag=0;
  int lgmr=0;
  for (int i=0;i<neq;i++) soln[i]=0.;
  memcpy(v,rhs,neq*sizeof(double));
  double rnrm=F77NAME(dnrm2)(neq,v,1);
//if (rnrm>eplin) 
  {
    double da=1./rnrm;
    F77NAME(dscal)(neq,da,v,1);
    for (int i=0;i<maxlp1*maxl;i++) hes[i]=0.;
    double prod=1.;
    bool failed=false;
    int ll=0;
    double rho=HUGE_VAL;
    double snormw=HUGE_VAL;
    for (;ll<maxl;ll++) {
      lgmr=ll;
      int llp1=ll+1;
      double *v_ll=v+ll*neq;
      double *v_llp1=v+llp1*neq;
//    v_llp1 = jacobian * v_ll
      datv(fi,la,dt,f,saturationold,v_ll,x,y, v_llp1, fn, 
        mobo,mobw,po,velo,velw);
#ifdef DEBUG
/*
      for (int i=0;i<neq;i++) {
        double jv=0.;
        for (int j=max(i-3,0);j<=min(i+3,neq-1);j++) {
          jv+=unsplit_matrix[6+i-j+j*10]*v_ll[j];
        }
        cout << "\tv_llp1[" << i << "] = " << v_llp1[i] << " " << jv 
             << endl;
      }
*/
#endif
//    Gram-Schmidt orthogonalization of v_llp1 against cols 0,...,ll of v
//    snormw = || v_llp1 ||
      F77NAME(dorth)(v_llp1,v,hes,neq,llp1,maxlp1,kmp,snormw);
      hes[llp1+ll*maxlp1]=snormw;
//    update QR factorization
      int info=INT_MAX;
      F77NAME(dheqr)(hes,maxlp1,llp1,q,info,llp1);
      if (info==llp1) {
        failed=true;
        break;
      }
      prod*=q[2*ll+1];
      rho=abs(prod*rnrm);
      if (llp1>kmp && kmp<maxl) {
        TRACER_CALL(tr,"dspigm ll>kmp && kmp<maxl");
        if (ll==kmp) {
          memcpy(dl,v,neq*sizeof(double));
          for (int i=0;i<kmp;i++) {
            int ip1=i+1;
            int i2=i*2;
            double s=q[i2+1];
            double c=q[i2];
            double *v_ip1=v+ip1*neq;
            for (int k=0;k<neq;k++) dl[k]=s*dl[k]+c*v_ip1[k];
          }
        }
        double s=q[2*ll+1];
        double c=q[2*ll]/snormw;
        for (int k=0;k<neq;k++) dl[k]=s*dl[k]+c*v_llp1[k];
        rho*=F77NAME(dnrm2)(neq,dl,1);
      }
      if (rho<=eplin*rnrm || llp1==maxl) break;
      F77NAME(dscal)(neq,1./snormw,v_llp1,1);
    }
    if (rho>=rnrm) failed=true;
    if (failed) {
      TRACER_CALL(tr,"dspigm failed");
      iflag=2;
      for (int i=0;i<neq;i++) soln[i]=0.;
    } else if (rho>eplin*rnrm) {
      iflag=1;
      if (restart) {
        if (kmp==maxl) {
          memcpy(dl,v,neq*sizeof(double));
          for (int i=0;i<maxl-1;i++) {
            int ip1=i+1;
            int i2=i*2;
            double s=q[i2+1];
            double c=q[i2];
            double *v_ip1=v+ip1*neq;
            for (int k=0;k<neq;k++) {
              dl[k]=s*dl[k]+c*v_ip1[k];
            }
          }
          double s=q[2*maxl-1];
          double c=q[2*maxl-2]/snormw;
          double *v_maxl=v+maxl*neq;
          for (int k=0;k<neq;k++) {
            dl[k]=s*dl[k]+c*v_maxl[k];
          }
        }
        F77NAME(dscal)(neq,rnrm*prod,dl,1);
      }
    }
    if (!failed) {
      ll=lgmr;
      int llp1=ll+1;
      for (int k=0;k<=llp1;k++) work[k]=0.;
      work[0]=rnrm;
      F77NAME(dhels)(hes,maxlp1,llp1,q,work);
      for (int k=0;k<neq;k++) soln[k]=0.;
      double *vi=v;
      for (int i=0;i<=ll;i++,vi+=neq) {
        F77NAME(daxpy)(neq,work[i],vi,1,soln,1);
      }
#ifdef DEBUG
/*
      for (int i=0;i<neq;i++) {
        double jv=0.;
        for (int j=max(i-3,0);j<=min(i+3,neq-1);j++) {
          jv+=unsplit_matrix[6+i-j+j*10]*soln[j];
        }
        cout << "\trhs[" << i << "] = " << rhs[i] << " " << jv 
             << endl;
      }
      ASSERT(false);
*/
#endif
    }
  }
  delete [] fn;
  delete [] work;
  delete [] q;
  delete [] hes;
  delete [] v;
#ifdef DEBUG
/*
  delete [] dmobods;
  delete [] dmobwds;
  delete [] dpods;
  delete [] dvelodpl;
  delete [] dvelodpr;
  delete [] dvelodsl;
  delete [] dvelodsr;
  delete [] dvelwdpl;
  delete [] dvelwdpr;
  delete [] dvelwdsl;
  delete [] unsplit_matrix;
  delete [] saturationnew;
  delete [] pressurenew;
  ASSERT(false);
*/
#endif
  return iflag;
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//uses dspigm with restarts to solve linear system 
//  jacobian(y) * soln = rhs
//f = f(y) is used to approximate jacobian via differences
void dslvk(int fi,int la,int kmp,int maxl,
const double *f,const double *rhs,const double *saturationold,
const double *x,const double *y, double *soln,
double *mobo,double *mobw,double *po,double *velo,double *velw) {
//TRACER_CALL(tr,"dslvk");
  bool restart=true;
  int n=la-fi;
  int neq=2*n;
  double *dl=OPERATOR_NEW_BRACKET(double,neq);
  double *r=OPERATOR_NEW_BRACKET(double,neq);
  double *z=OPERATOR_NEW_BRACKET(double,neq);
  memcpy(r,rhs,neq*sizeof(double));
  for (int i=0;i<neq;i++) soln[i]=0.;
  int iflag=1;
  for (int nrsts=0;iflag==1 && nrsts<nrmax;nrsts++) {
    if (nrsts>0) memcpy(r,dl,neq*sizeof(double));
    iflag=dspigm(fi,la,kmp,maxl,f,r,saturationold,x,y, dl,z, 
      mobo,mobw,po,velo,velw);
    for (int i=0;i<neq;i++) soln[i]+=z[i];
  }
  delete [] dl;
  delete [] r;
  delete [] z;
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void krylovSubspaceIteration(const int &fi,const int &la,
const double &dt,const double *pressureold,const double *saturationold,
const double *x,double *pressurenew,double *saturationnew) {
//TRACER_CALL(tr,"krylovSubspaceIteration");
  const double *sold=saturationold+(1-fi);
  const double *pold=pressureold+(1-fi);
  double *snew=saturationnew+(1-fi);
  double *pnew=pressurenew+(1-fi);
#ifdef DEBUG
//cout << "\tsaturationold = " << saturationold << endl;
//for (int i=fi-1;i<=la;i++) {
//  cout << "\tsaturationold[" << i << "]  = " << sold[i] << endl;
//}
//cout << "\tpressureold = " << pressureold << endl;
//for (int i=fi-1;i<=la;i++) {
//  cout << "\tpressureold[" << i << "]  = " << pold[i] << endl;
//}
//cout << "\tsaturationnew = " << saturationnew << endl;
//cout << "\tpressurenew = " << pressurenew << endl;
#endif
  int n=la-fi;
  double *mobo=OPERATOR_NEW_BRACKET(double,n+2);
  double *mobw=OPERATOR_NEW_BRACKET(double,n+2);
  double *po=OPERATOR_NEW_BRACKET(double,n+2);
  double *velo=OPERATOR_NEW_BRACKET(double,n+1);
  double *velw=OPERATOR_NEW_BRACKET(double,n+1);
  int neq=2*n;
  double *f=OPERATOR_NEW_BRACKET(double,neq);
  double *fn=OPERATOR_NEW_BRACKET(double,neq);
  double *step=OPERATOR_NEW_BRACKET(double,neq);
  double *y=OPERATOR_NEW_BRACKET(double,neq);
  double *yn=OPERATOR_NEW_BRACKET(double,neq);

  int ij=0;
  for (int i=fi;i<la;i++) {
    y[ij++]=snew[i];
    y[ij++]=pnew[i];
  }
  bool line_search=true;
  double rate=ONE;
  double fsize=HUGE_VAL;
  double fnrm=dfnrmk(fi,la,dt,fsize,saturationold,x,y, f, 
      mobo,mobw,po,velo,velw, true);
#ifdef DEBUG
//cout << "\tfnrm[0] = " << fnrm << endl;
#endif
//bool success=(fnrm<=epcon*fsize);
  if (fnrm>epcon*fsize) {
#ifdef DEBUG
//  cout << "\tepcon = " << epcon*fsize << endl;
/*
    double *dmobods=OPERATOR_NEW_BRACKET(double,n+2);
    double *dmobwds=OPERATOR_NEW_BRACKET(double,n+2);
    double *dpods=OPERATOR_NEW_BRACKET(double,n+2);
    double *dvelodpl=OPERATOR_NEW_BRACKET(double,n+1);
    double *dvelodpr=OPERATOR_NEW_BRACKET(double,n+1);
    double *dvelodsl=OPERATOR_NEW_BRACKET(double,n+1);
    double *dvelodsr=OPERATOR_NEW_BRACKET(double,n+1);
    double *dvelwdpl=OPERATOR_NEW_BRACKET(double,n+1);
    double *dvelwdpr=OPERATOR_NEW_BRACKET(double,n+1);
    double *dvelwdsl=OPERATOR_NEW_BRACKET(double,n+1);
    double *unsplit_matrix=OPERATOR_NEW_BRACKET(double,10*2*n);
    double *saturationnew=OPERATOR_NEW_BRACKET(double,n+2);
    double *pressurenew=OPERATOR_NEW_BRACKET(double,n+2);
    double *sn=saturationnew+(1-fi);
    double *pn=pressurenew+(1-fi);
    const double *so=saturationold+(1-fi);
    int ij=0;
    sn[fi-1]=so[fi-1];
    pn[fi-1]=F77NAME(blcom).pres_left;
    for (int i=fi;i<la;i++) {
      sn[i]=y[ij++];
      pn[i]=y[ij++];
    }
    sn[la]=so[la];
    pn[la]=F77NAME(blcom).pres_right;
    F77NAME(jacobian)(fi,la,dt, pressurenew,saturationnew,x,unsplit_matrix,
      dmobods,dmobwds,dpods,dvelodpl,dvelodpr,dvelodsl,dvelodsr,
      dvelwdpl,dvelwdpr,dvelwdsl,mobo,mobw,po,velo,velw);
//  for (int i=0;i<10;i++) {
//    cout << "matrix[" << i << ",*] = ";
//    for (int j=0;j<neq;j++) cout << unsplit_matrix[i+j*10] << " ";
//    cout << endl;
//  }
*/
#endif
    double fnrm0=fnrm;
    int maxl=min(5,neq); // max no. dspigm iterations
    int kmp=maxl; // no. orthogonalization vectors in dspigm
    for (int m=0;m<itmax;m++) {
      dslvk(fi,la,kmp,maxl,f,f,saturationold,x,
        y, step, mobo,mobw,po,velo,velw);
#ifdef DEBUG
/*
      for (int i=0;i<neq;i++) {
        double jv=0.;
        for (int j=max(i-3,0);j<=min(i+3,neq-1);j++) {
          jv+=unsplit_matrix[6+i-j+j*10]*step[j];
        }
        cout << "\tf[" << i << "] = " << f[i] << " " << jv 
             << endl;
      }
      ASSERT(false);
*/
#endif
//    double step_norm=F77NAME(dnrm2)(neq,step,1)
//                    /sqrt(static_cast<double>(neq));
//    if (step_norm<=0.) {
//      success=true;
//      break;
//    }
      double oldfnm=fnrm;
      memcpy(yn,y,neq*sizeof(double));
      bool ls_success=dlinsk(fi,la,dt,step,saturationold,x, 
        fnrm,yn,fn, mobo,mobw,po,velo,velw);
//    ASSERT(ls_success);
#ifdef DEBUG
//    cout << "\tfnrm[" << m+1 << "] = " << fnrm << endl;
#endif
      double *temp=yn;
      yn=y;
      y=temp;
      temp=fn;
      fn=f;
      f=temp;
      rate=fnrm/oldfnm;
      if (fnrm<=epcon*fsize) {
//      success=true;
        break;
      }
    }
#ifdef DEBUG
/*
    delete [] dmobods;
    delete [] dmobwds;
    delete [] dpods;
    delete [] dvelodpl;
    delete [] dvelodpr;
    delete [] dvelodsl;
    delete [] dvelodsr;
    delete [] dvelwdpl;
    delete [] dvelwdpr;
    delete [] dvelwdsl;
    delete [] unsplit_matrix;
    delete [] saturationnew;
    delete [] pressurenew;
    ASSERT(false);
*/
#endif
  }
//if (success) 
  {
//  TRACER_CALL(tr,"krylovSubspaceIteration success");
    int ij=0;
    snew[fi-1]=sold[fi-1];
    pnew[fi-1]=pold[fi-1];
    for (int i=fi;i<la;i++) {
      saturationnew[i]=y[ij++];
      pressurenew[i]=y[ij++];
    }
    snew[la]=snew[la-1];
    pnew[la]=pold[la];
  }
#ifdef DEBUG
//for (int i=fi-1;i<=la;i++) {
//  cout << "\tsaturationnew[" << i << "]  = " << snew[i] << endl;
//}
//for (int i=fi-1;i<=la;i++) {
//  cout << "\tpressurenew[" << i << "]  = " << pnew[i] << endl;
//}
#endif
  delete [] mobo;
  delete [] mobw;
  delete [] po;
  delete [] velo;
  delete [] velw;
  delete [] f;
  delete [] fn;
//delete [] savr;
  delete [] step;
  delete [] y;
  delete [] yn;
}
