// "$Header:$"
//-------------------------  dof_constraints.cc  ---------------------------
//   $Id: dof_constraints.cc,v 1.70 2003/02/24 22:47:20 wolf Exp $
//   Version: $Name: Version-4-0-0 $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//-------------------------  dof_constraints.cc  ---------------------------
//
//modified from deal.II/deal.II/source/dofs/dof_constraints.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <ConstraintMatrix.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ConstraintMatrix::addLine(int line) {
  CHECK_TEST(!sorted);
  for (CONSTRAINT_LINE_LIST_ENTRY *item=lines.first();item;
  item=lines.next(item)) {
    if (item->getValue()==line) return;
  }
  lines.append(OPERATOR_NEW CONSTRAINT_LINE_DATA_LIST(),line);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ConstraintMatrix::addEntry(int line,int column,REAL value) {
  CHECK_TEST(!sorted);
  CONSTRAINT_LINE_LIST_ENTRY *line_ptr;
  for (line_ptr=lines.last();line_ptr;line_ptr=lines.prev(line_ptr)) {
    if (line_ptr->getValue()==line) break;
  }
  CHECK_SAME(line_ptr->getValue(),line);
  CONSTRAINT_LINE_DATA_LIST *entries=line_ptr->getData();
  for (CONSTRAINT_LINE_DATA_LIST_ENTRY *entry=entries->first();entry;
  entry=entries->next(entry)) {
    CONSTRAINT_LINE_DATA *p=entry->getData();
    if (p->first==column) {
      CHECK_SAME(p->second,value);
      return;
    }
  }
  entries->append(OPERATOR_NEW CONSTRAINT_LINE_DATA(column,value));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ConstraintMatrix::addEntries(int line,
const CONSTRAINT_LINE_DATA_LIST &col_val_pairs) {
  CHECK_TEST(!sorted);
  CONSTRAINT_LINE_LIST_ENTRY *line_ptr;
  for (line_ptr=lines.last();line_ptr;line_ptr=lines.prev(line_ptr)) {
    if (line_ptr->getValue()==line) break;
  }
  CHECK_SAME(line_ptr->getValue(),line);

  CONSTRAINT_LINE_DATA_LIST *entries=line_ptr->getData();
  for (CONSTRAINT_LINE_DATA_LIST_ENTRY *col_val_pair=
  col_val_pairs.first();col_val_pair;
  col_val_pair=col_val_pairs.next(col_val_pair)) {
    CONSTRAINT_LINE_DATA *col_val_pair_data=col_val_pair->getData();
    bool found=false;
    for (CONSTRAINT_LINE_DATA_LIST_ENTRY *p=entries->first();p;
    p=entries->next(p)) {
      CONSTRAINT_LINE_DATA *p_data=p->getData();
      if (p_data->first==col_val_pair_data->first) {
        CHECK_SAME(p_data->second,col_val_pair_data->second);
        found=true;
        break;
      }
    }
    if (!found) {
      entries->
        append(OPERATOR_NEW CONSTRAINT_LINE_DATA(*col_val_pair_data));
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ConstraintMatrix::close() {
  if (sorted) return;
  for (CONSTRAINT_LINE_LIST_ENTRY *line=lines.first();line;
  line=lines.next(line)) {
    CONSTRAINT_LINE_DATA_LIST *line_entries=line->getData();
    for (CONSTRAINT_LINE_DATA_LIST_ENTRY *line_entry=
    line_entries->first();line_entry;) {
      CONSTRAINT_LINE_DATA *line_entry_data=line_entry->getData();
      if (line_entry_data->second==0) {
        delete line_entries->delAndNext(line_entry);
      } else line_entry=line_entries->next(line_entry);
    }
    line_entries->sort(false); // largest last
  }
  lines.sort(false); // largest last
#ifdef DEBUG
  for (CONSTRAINT_LINE_LIST_ENTRY *line=lines.first();line;
  line=lines.next(line)) {
    CONSTRAINT_LINE_DATA_LIST *line_entries=line->getData();
    for (CONSTRAINT_LINE_DATA_LIST_ENTRY *line_entry=
    line_entries->first();line_entry;
    line_entry=line_entries->next(line_entry)) {
      int test_line=line_entry->getData()->first;
      CONSTRAINT_LINE_LIST_ENTRY *le=lines.first();
      for (;le;le=lines.next(le)) {
        if (!(test_line<=le->getValue())) break; 
      }
#ifdef DEBUG
      if (le) {
        CHECK_TEST(le->getValue()!=test_line);
      }
#endif
    }
  }
#endif
  sorted=true;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ConstraintMatrix::merge(ConstraintMatrix &other_constraints) {
  NIDLList<int> this_dofs;
  for (CONSTRAINT_LINE_LIST_ENTRY *line_entry=lines.first();
  line_entry;line_entry=lines.next(line_entry)) {
    this_dofs.
      append(OPERATOR_NEW int(line_entry->getValue()));
  }
#ifdef DEBUG
  for (CONSTRAINT_LINE_LIST_ENTRY *line_entry=
  other_constraints.lines.first();line_entry;
  line_entry=other_constraints.lines.next(line_entry)) {
    int line=line_entry->getValue();
    for (NIDLListNode<int> *this_dofs_entry=
    this_dofs.first();this_dofs_entry;
    this_dofs_entry=this_dofs.next(this_dofs_entry)) {
      CHECK_TEST(*(this_dofs_entry->getData()) != line);
    }
    CONSTRAINT_LINE_DATA_LIST *line_data_list=line_entry->getData();
    for (CONSTRAINT_LINE_DATA_LIST_ENTRY *line_data_list_entry=
    line_data_list->first();line_data_list_entry;
    line_data_list_entry=line_data_list->next(line_data_list_entry)) {
      int ef=line_data_list_entry->getData()->first;
      for (NIDLListNode<int> *this_dofs_entry=
      this_dofs.first();this_dofs_entry;
      this_dofs_entry=this_dofs.next(this_dofs_entry)) {
        CHECK_TEST(*(this_dofs_entry->getData()) != ef);
      }
    }
  }
#endif
  bool object_was_sorted=sorted;
  sorted=false;
  for (CONSTRAINT_LINE_LIST_ENTRY *line=lines.first();line;
  line=lines.next(line)) {
    CONSTRAINT_LINE_DATA_LIST *line_data_list=line->getData();
    CONSTRAINT_LINE_DATA_LIST tmp;
    NumPtr<CONSTRAINT_LINE_LIST_ENTRY*> 
      tmp_other_lines(line_data_list->length());
    tmp_other_lines.initialize(0);
    bool entries_to_resolve=false;
    int i=0;
    const CONSTRAINT_LINE_LIST &other_constraints_lines=
      other_constraints.lines;
    for (CONSTRAINT_LINE_DATA_LIST_ENTRY *line_data_list_entry=
    line_data_list->first();line_data_list_entry;
    line_data_list_entry=line_data_list->next(line_data_list_entry),i++)
    {
      int index_comparison=
        line_data_list_entry->getData()->first;
      if (other_constraints.sorted) {
        CONSTRAINT_LINE_LIST_ENTRY *other_constraints_line=
          other_constraints_lines.first();
        for (;other_constraints_line;other_constraints_line=
        other_constraints_lines.next(other_constraints_line)) {
          if (other_constraints_line->getValue()>=index_comparison) {
            break;
          }
        }
        if (other_constraints_line && 
        other_constraints_line->getValue()==index_comparison)
        {
          tmp_other_lines[i]=other_constraints_line; 
          entries_to_resolve=true;
        }
      } else {
        for (CONSTRAINT_LINE_LIST_ENTRY *other_constraints_line=
        other_constraints_lines.first();other_constraints_line;
        other_constraints_line=
        other_constraints_lines.next(other_constraints_line)) {
          if (other_constraints_line->getValue()==index_comparison) {
            tmp_other_lines[i]=other_constraints_line;
            entries_to_resolve=true;
            break;
          }
        }
      }
    }
    if (!entries_to_resolve) continue;
    i=0;
    for (CONSTRAINT_LINE_DATA_LIST_ENTRY *line_data_list_entry=
    line_data_list->first();line_data_list_entry;i++) {
      if (!tmp_other_lines[i]) {
        tmp.append(line_data_list->delAndNext(line_data_list_entry));
      } else {
        CONSTRAINT_LINE_DATA *line_data_list_entry_data=
          line_data_list_entry->getData();
        CHECK_SAME(tmp_other_lines[i]->getValue(),
          line_data_list_entry_data->first);
        REAL weight=line_data_list_entry_data->second;
        CONSTRAINT_LINE_DATA_LIST *other_constraints_lines=
          tmp_other_lines[i]->getData();
        for (CONSTRAINT_LINE_DATA_LIST_ENTRY 
        *other_constraints_line_data=other_constraints_lines->first();
        other_constraints_line_data;other_constraints_line_data=
        other_constraints_lines->next(other_constraints_line_data)) {
          CONSTRAINT_LINE_DATA *other_data=
            other_constraints_line_data->getData();
          tmp.append(OPERATOR_NEW 
            CONSTRAINT_LINE_DATA(other_data->first,other_data->second));
        }
        line_data_list_entry=line_data_list->next(line_data_list_entry);
      }
    }
    while (line_data_list->notEmpty()) {
      delete line_data_list->delAfter(0);
    }
    line_data_list->join(tmp);
  }
  lines.join(other_constraints.lines);
  if (object_was_sorted) close();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ConstraintMatrix::shift(int offset) {
  for (CONSTRAINT_LINE_LIST_ENTRY *line=lines.first();line;
  line=lines.next(line)) {
    line->setValue(line->getValue()+offset);
    CONSTRAINT_LINE_DATA_LIST *line_data_list=line->getData();
    for (CONSTRAINT_LINE_DATA_LIST_ENTRY *line_data_entry=
    line_data_list->first();line_data_entry;line_data_entry=
    line_data_list->next(line_data_entry)) {
      line_data_entry->getData()->first+=offset;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ConstraintMatrix::clear() {
  while (lines.notEmpty()) {
    CONSTRAINT_LINE_DATA_LIST *line_data_list=lines.delAfter(0);
    while (line_data_list->notEmpty()) {
      delete line_data_list->delAfter(0);
    }
    delete line_data_list;
  }
  sorted=false;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ConstraintMatrix::condense(SparsityPattern &sparsity) const {
  int n_rows=sparsity.nRows();
  CHECK_TEST(sorted);
  CHECK_TEST(!sparsity.isCompressed());
  CHECK_SAME(n_rows,sparsity.nCols());
  NumPtr<CONSTRAINT_LINE_LIST_ENTRY*> distribute(n_rows);
  distribute.initialize(0);
  for (CONSTRAINT_LINE_LIST_ENTRY *line=lines.first();line;
  line=lines.next(line)) {
    distribute[line->getValue()]=line;
  }
  for (int row=0;row<n_rows;row++) {
    int row_begin=sparsity.getRowstartIndex(row);
    int row_end=sparsity.getRowstartIndex(row+1);
    if (distribute[row]==0) {
//    entries for B N^T
      for (int j=row_begin;j<row_end;j++){
        int column=sparsity.getColumnNumber(j);
        if (column==SparsityPattern::invalid_entry) break;
        else {
          if (distribute[column]!=0) {
            CONSTRAINT_LINE_DATA_LIST *cldl=
              distribute[column]->getData(); 
            for (CONSTRAINT_LINE_DATA_LIST_ENTRY *q=cldl->first();q;
            q=cldl->next(q)) {
              sparsity.add(row,q->getData()->first);
            }
          }
        }
      }
    } else {
      for (int j=row_begin;j<row_end;j++) {
        int column=sparsity.getColumnNumber(j);
        if (column==SparsityPattern::invalid_entry)
        {
            break;
        } else {
          if (distribute[column]==0) {
//          entries for N B
            CONSTRAINT_LINE_DATA_LIST *cldl=
              distribute[row]->getData();
            for (CONSTRAINT_LINE_DATA_LIST_ENTRY *q=cldl->first();q;
            q=cldl->next(q)) {
              sparsity.add(q->getData()->first,column);
            }
          } else {
//          entries for 0 or I
            CONSTRAINT_LINE_DATA_LIST *cldlp=distribute[row]->getData();
            for (CONSTRAINT_LINE_DATA_LIST_ENTRY *p=cldlp->first();p;
            p=cldlp->next(p)) {
              int pfirst=p->getData()->first;
              CONSTRAINT_LINE_DATA_LIST *cldlq=
                distribute[column]->getData();
              for (CONSTRAINT_LINE_DATA_LIST_ENTRY *q=cldlq->first();q;
              q=cldlq->next(q)) {
                sparsity.add(pfirst,q->getData()->first);
              }
            }
          }
        }
      }
    }
  }
  sparsity.compress();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//Suppose that hanging node basis function v_i satisfies
//  v_i = sum_K c_{iK} v_K
//where v_K are non-hanging node basis functions.
//Then for all non-hanging nodes I and J and hanging nodes i and j
//  sum_J int_Omega grad v_I . D grad v_J dx u_J
//  + sum_j int_Omega grad v_I . D grad v_j dx u_j
//  = sum_J int_Omega grad v_I . D grad v_J dx u_J
//  + sum_j [ sum_K int_Omega grad v_I . D grad v_K dx c_{jK} ] u_j
//and 
//  sum_J int_Omega grad v_i . D grad v_J dx u_J
//  + sum_j int_Omega grad v_i . D grad v_j dx u_j
//  = sum_J [ sum_K c_{iK} int_Omega grad v_K . D grad v_J dx ] u_J
//  + sum_j 
//    [ sum_K sum_L c_{iK} int_Omega grad v_K . D grad v_L dx c_{jL} ] u_j
void ConstraintMatrix::condense(SparseMatrix &uncondensed) const {
  const SparsityPattern &sparsity=uncondensed.getSparsityPattern();
  const int n_rows=sparsity.nRows();
  CHECK_TEST(sorted);
  CHECK_TEST(sparsity.isCompressed());
  CHECK_SAME(n_rows,sparsity.nCols());
  NumPtr<CONSTRAINT_LINE_LIST_ENTRY*> distribute(n_rows);
  distribute.initialize(0);
  for (CONSTRAINT_LINE_LIST_ENTRY *c=lines.first();c;c=lines.next(c)) {
    distribute[c->getValue()]=c;
  }
  for (int row=0;row<n_rows;row++) {
    int row_begin=sparsity.getRowstartIndex(row);
    int row_end=sparsity.getRowstartIndex(row+1);
    if (distribute[row]==0) {
      for (int j=row_begin;j<row_end;j++) {
        int column=sparsity.getColumnNumber(j);
        CHECK_TEST(column!=SparsityPattern::invalid_entry);
        REAL &Arc=uncondensed.globalEntry(j);
        CONSTRAINT_LINE_LIST_ENTRY *c=distribute[column];
        if (c) {
          CONSTRAINT_LINE_DATA_LIST *cldl=c->getData();
          for (CONSTRAINT_LINE_DATA_LIST_ENTRY *q=cldl->first();q;
          q=cldl->next(q)) {
            CONSTRAINT_LINE_DATA *q_data=q->getData();
            uncondensed.add(row,q_data->first,Arc*q_data->second);
          }
          Arc=0.;
        }
      }
    } else {
      for (int j=row_begin;j<row_end;j++) {
        int column=sparsity.getColumnNumber(j);
        CHECK_TEST(column!=SparsityPattern::invalid_entry);
        CONSTRAINT_LINE_LIST_ENTRY *c=distribute[column];
        REAL &Arc=uncondensed.globalEntry(j);
        if (c==0) {
          CONSTRAINT_LINE_DATA_LIST *cldl=distribute[row]->getData();
          for (CONSTRAINT_LINE_DATA_LIST_ENTRY *q=cldl->first();q;
          q=cldl->next(q)) {
            CONSTRAINT_LINE_DATA *q_data=q->getData();
            uncondensed.add(q_data->first,column,(q_data->second)*Arc);
          }
          Arc=0.;
        } else {
//        0 or I
          CONSTRAINT_LINE_DATA_LIST *cldlp=distribute[row]->getData();
          CONSTRAINT_LINE_DATA_LIST *cldlq=
            distribute[column]->getData();
          for (CONSTRAINT_LINE_DATA_LIST_ENTRY *p=cldlp->first();p;
          p=cldlp->next(p)) {
            CONSTRAINT_LINE_DATA *p_data=p->getData();
            int pfirst=p_data->first;
            REAL psecond=p_data->second*Arc;
            for (CONSTRAINT_LINE_DATA_LIST_ENTRY *q=cldlq->first();q;
            q=cldlq->next(q)) {
              CONSTRAINT_LINE_DATA *q_data=q->getData();
              uncondensed.add(pfirst,q_data->first,
                psecond*q_data->second);
            }
          }
//        can't set diagonal to common diagonal value size, since
//          right hand side is condensed separately
          Arc=(row==column ? 1 : 0. );
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool ConstraintMatrix::isConstrained(int index) const {
  if (lines.empty()) return false;
    for (CONSTRAINT_LINE_LIST_ENTRY *line=lines.first();line;
    line=lines.next(line)) {
      if (line->getValue()==index) return true;
    }
    return false;
//}
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
bool ConstraintMatrix::isIdentityConstrained(int index) const {
  if (lines.empty()) return false;
    for (CONSTRAINT_LINE_LIST_ENTRY *line=lines.first();line;
    line=lines.next(line)) {
      if (line->getValue()==index) {
        CONSTRAINT_LINE_DATA_LIST *cldl=line->getData();
        return cldl->length()==1 && 
               cldl->first()->getData()->second==1.0;
      }
    }
    return false;
//}
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int ConstraintMatrix::maxConstraintIndirections() const {
  int return_value=0;
  for (CONSTRAINT_LINE_LIST_ENTRY *line=lines.first();line;
  line=lines.next(line)) {
    return_value=max(return_value,
      static_cast<int>(line->getData()->length()));
  }
  return return_value;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
void ConstraintMatrix::makeHangingNodeConstraints(
const DofHandler &dof_handler) { }
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
void ConstraintMatrix::makeHangingNodeConstraints(
const DofHandler &dof_handler) { 
  FiniteElement &fe=dof_handler.getFe();
  Tessellation &tess=dof_handler.getTessellation();
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape()) {
    LEVEL_CLASS *level=tess.presentLevel();
    CODIMENSION1_LEVEL_CLASS *face_level=
      static_cast<CODIMENSION1_LEVEL_CLASS*>(level);
    const SHAPE_CLASS *element=level->getShape();
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int face_index=element->codimension1ShapeIndex(f);
      if (face_level->hasChildren(face_index)) {
        const CODIMENSION1_SHAPE_CLASS *face_element=
          face_level->getShape(face_index);
        NumPtr<int> dofs_on_mother;
        dof_handler.
          getDofIndices(face_element,face_level,dofs_on_mother);
        NumPtr<int> dofs_on_children;
        dof_handler.dofsOnChildren(face_element,face_level,
          dofs_on_children);
        Matrix constraint_matrix;
        fe.constraints(level,element,face_element,f,constraint_matrix);
        for (int row=0;row<dofs_on_children.getNumber();row++) {
          int r=dofs_on_children[row];
          addLine(r);
          for (int i=0;i<dofs_on_mother.getNumber();i++) {
            addEntry(r,dofs_on_mother[i],constraint_matrix(row,i));
          }
        }
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ConstraintMatrix::printOn(ostream &out) const {
  out << "ConstraintMatrix: sorted = " << sorted << endl;
  for (CONSTRAINT_LINE_LIST_ENTRY *line=lines.first();line;
  line=lines.next(line)) {
    int l=line->getValue();
    CONSTRAINT_LINE_DATA_LIST *cldl=line->getData();
    for (CONSTRAINT_LINE_DATA_LIST_ENTRY *j=cldl->first();j;
    j=cldl->next(j)) {
      CONSTRAINT_LINE_DATA *jd=j->getData();
      out << "\t" << l << " " << jd->first << ":  " << jd->second 
          << endl;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//Suppose that hanging node basis function v_i satisfies
//  v_i = sum_K c_{iK} v_K
template <class VectorType> void ConstraintMatrix::setResidual(
const VectorType &x,VectorType &r) const {
  CHECK_TEST(sorted);
  if (lines.empty()) return;
  for (CONSTRAINT_LINE_LIST_ENTRY *line=lines.first();line;
  line=lines.next(line)) {
    int l=line->getValue();
    r[l]=x[l];
    CONSTRAINT_LINE_DATA_LIST *cldl=line->getData();
    for (CONSTRAINT_LINE_DATA_LIST_ENTRY *j=cldl->first();j;
    j=cldl->next(j)) {
      CONSTRAINT_LINE_DATA *jd=j->getData();
      r[l]-=jd->second*x[jd->first];
    } // r[i] = v_i - sum_J c_{iJ} * v_J
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template <class VectorType> void ConstraintMatrix::condense(
VectorType &vec) const {
  CHECK_TEST(sorted);
  if (lines.empty()) return;
  for (CONSTRAINT_LINE_LIST_ENTRY *c=lines.first();c;c=lines.next(c)) {
    REAL &vr=vec[c->getValue()];
    CONSTRAINT_LINE_DATA_LIST *cldl=c->getData();
    for (CONSTRAINT_LINE_DATA_LIST_ENTRY *q=cldl->first();q;
    q=cldl->next(q)) {
      CONSTRAINT_LINE_DATA *qd=q->getData();
      vec[qd->first]+=(qd->second)*vr;
    }
    vr=0.;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template<class VectorType> void ConstraintMatrix::distribute(
VectorType &vec) const {
  CHECK_TEST(sorted);
  for (CONSTRAINT_LINE_LIST_ENTRY *c=lines.first();
  c;c=lines.next(c)) {
    REAL &v=vec[c->getValue()];
    v=0.;
    CONSTRAINT_LINE_DATA_LIST *cldl=c->getData();
    for (CONSTRAINT_LINE_DATA_LIST_ENTRY *i=cldl->first();i;
    i=cldl->next(i)) {
      CONSTRAINT_LINE_DATA *id=i->getData();
      v+=(id->second)*vec[id->first];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#define VECTOR_FUNCTIONS(VectorType) \
template void ConstraintMatrix::condense<VectorType>(VectorType&) \
  const;\
template void ConstraintMatrix::setResidual<VectorType>(const VectorType&,VectorType&) const; \
template void ConstraintMatrix::distribute<VectorType>(VectorType&) \
  const;

VECTOR_FUNCTIONS(Vector)

#include "NIDLList.C"
template class NIDLListNode<int>;
template class NIDLList<int>;
template class NIDLListNode<pair<int,REAL> >;
template class NIDLList<pair<int,REAL> >;
template class NIDLListC<pair<int,REAL> >;
template class NIDLListNode<NIDLListC<pair<int,REAL> > >;
template class NIDLList<NIDLListC<pair<int,REAL> > >;
template class 
  NIDLValueListNode<NIDLListC<pair<int,REAL> >,int>;
template class NIDLValueList<NIDLListC<pair<int,REAL> >,
  int>;

#include "NumPtr.C"
template class NumPtrBase<
  NIDLValueListNode<NIDLListC<pair<int, double> >, int>*,int>;
template class NumPtr<
  NIDLValueListNode<NIDLListC<pair<int, double> >, int>*>;
