#ifndef CONSTRAINT_MATRIX_H
#define CONSTRAINT_MATRIX_H "$Header:$"
//----------------------------  dof_constraints.h  ---------------------------
//    $Id: dof_constraints.h,v 1.54 2003/03/22 19:36:03 wolf Exp $
//    Version: $Name: Version-4-0-0 $
//
//    Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//    This file is subject to QPL and may not be  distributed
//    without copyright and license information. Please refer
//    to the file deal.II/doc/license.html for the  text  and
//    further information on this license.
//
//----------------------------  dof_constraints.h  ---------------------------
//
//modified from deal.II/deal.II/include/dofs/dof_constraints.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <DofHandler.H>
#include <NIDLList.H>
#include <SparseMatrix.H>
#include <SparsityPattern.H>
#include <Types.H>

//compare to DEAL ConstraintMatrix
class ConstraintMatrix {
  private:
    typedef pair<int,REAL> CONSTRAINT_LINE_DATA;
    typedef NIDLListNode<CONSTRAINT_LINE_DATA> 
      CONSTRAINT_LINE_DATA_LIST_ENTRY;
    typedef NIDLListC<CONSTRAINT_LINE_DATA> CONSTRAINT_LINE_DATA_LIST;
    typedef NIDLValueListNode<CONSTRAINT_LINE_DATA_LIST,int> 
      CONSTRAINT_LINE_LIST_ENTRY;
    typedef NIDLValueList<CONSTRAINT_LINE_DATA_LIST,int> 
      CONSTRAINT_LINE_LIST;

    CONSTRAINT_LINE_LIST lines;
    bool sorted;
  protected:
    void clear();
  public:
    ConstraintMatrix() : sorted(false) {}
    ~ConstraintMatrix() { clear(); }
    void addLine(int);
    void addEntry(int,int,REAL);
    void addEntries(int,const CONSTRAINT_LINE_DATA_LIST&);
    void close();
    void merge(ConstraintMatrix&);
    void shift(int);
    int nConstraints() const { return lines.length(); }
    bool isConstrained(int) const;
    bool isIdentityConstrained(int) const;
    int maxConstraintIndirections() const;
    void condense(SparsityPattern&) const;
    void condense(SparseMatrix&) const;
    void makeHangingNodeConstraints(const DofHandler&);
    template <class VectorType> void condense(VectorType&) const;
    template <class VectorType> void distribute(VectorType&) const;
    template <class VectorType> void setResidual(const VectorType&,
      VectorType&) const;
    void printOn(ostream&) const;
};

#endif
