// "$Header:$"
//----------------------------  dof_handler.cc  ---------------------------
//  $Id: dof_handler.cc,v 1.96 2003/01/14 19:51:08 wolf Exp $
//  Version: $Name: Version-4-0-0 $
//
//  Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
//
//----------------------------  dof_handler.cc  ---------------------------
//----------------------------  dof_tools.cc  ---------------------------
//  $Id: dof_tools.cc,v 1.103 2003/05/12 16:05:20 wolf Exp $
//  Version: $Name: Version-4-0-0 $
//
//  Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
//
//----------------------------  dof_tools.cc  ---------------------------
//
//modified from deal.II/deal.II/source/dofs/dof_handler.cc and dof_tools.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <DofHandler.H>

#include <algorithm>
#include <Boundary.H>
#include <FEValues.H>

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
ostream& operator<<(ostream &os,NORM_TYPE nt) {
  switch (nt) {
    case MEAN_NORM:
      os << "mean norm";
      break;
    case L1_NORM:
      os << "L1 norm";
      break;
    case L2_NORM:
      os << "L2 norm";
      break;
    case LINFTY_NORM:
      os << "Linfinity norm";
      break;
    case H1_SEMINORM:
      os << "H1 seminorm";
      break;
    case H1_NORM:
      os << "H1 norm";
      break;
    case LP_NORM:
      os << "Lp norm";
      break;
    case W1P_SEMINORM:
      os << "W1p seminorm";
      break;
    case W1P_NORM:
      os << "W1p norm";
      break;
  }
  return os;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::clearSpace() {  
  for (int i=0;i<levels.getNumber();i++) {
    if (levels[i]) delete levels[i]; levels[i]=0;
  }
  levels.cleanup();
  vertex_dofs.cleanup();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
//assume that all finite elements have same number of dofs per whatever
void DofHandler::reserveSpace() {
  int number_levels=tess->numberLevels();
  CHECK_TEST(number_levels>0);
  clearSpace();
  vertex_dofs.allocate(tess->getVertices().getNumber()
                      *selected_fe->dofsPerVertex());
  vertex_dofs.initialize(invalid_dof_index);
  levels.allocate(number_levels);
  int dpl=selected_fe->dofsPerLine();
  CHECK_NONNEGATIVE(dpl);
  int i=0;
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel(),i++) {
    int d1=0;
    for (bool doit=level->first();doit;doit=level->next()) {
      level->getShape()->setDofIndex(d1);
      d1+=dpl;
    }
    levels[i]=OPERATOR_NEW DofLevel1(d1);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void DofHandler::reserveSpace() {
  int number_levels=tess->numberLevels();
  CHECK_TEST(number_levels>0);
  clearSpace();
  vertex_dofs.allocate(tess->getVertices().getNumber()
                      *selected_fe->dofsPerVertex());
  vertex_dofs.initialize(invalid_dof_index);
  levels.allocate(number_levels);
  int dpl=selected_fe->dofsPerLine();
  int i=0;
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel(),i++) {
    int d1=0;
    Level1 *level1=static_cast<Level1*>(level);
    for (bool doit=level1->first();doit;doit=level1->next()) {
      level1->getShape()->setDofIndex(d1);
      d1+=dpl;
    }
    int d2=0;
    for (bool doit=level->first();doit;doit=level->next()) {
      Shape2 *e=level->getShape();
      e->setDofIndex(d2);
      d2+=selected_fe->dofsPerQuad(e);
    }
    levels[i]=OPERATOR_NEW DofLevel2(d1,d2);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void DofHandler::reserveSpace() {
  int number_levels=tess->numberLevels();
  CHECK_TEST(number_levels>0);
  clearSpace();
  vertex_dofs.allocate(tess->getVertices().getNumber()
                      *selected_fe->dofsPerVertex());
  vertex_dofs.initialize(invalid_dof_index);
  levels.allocate(number_levels);
  int dpl=selected_fe->dofsPerLine();
  int i=0;
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel(),i++) {
    int d1=0;
    Level1 *level1=static_cast<Level1*>(level);
    for (bool doit=level1->first();doit;doit=level1->next()) {
      level1->getShape()->setDofIndex(d1);
      d1+=dpl;
    }
    int d2=0;
    Level2 *level2=static_cast<Level2*>(level);
    for (bool doit=level2->first();doit;doit=level2->next()) {
      Shape2 *e=level2->getShape();
      e->setDofIndex(d2);
      d2+=selected_fe->dofsPerQuad(e);
    }
    int d3=0;
    for (bool doit=level->first();doit;doit=level->next()) {
      Shape3 *e=level->getShape();
      e->setDofIndex(d3);
      d3+=selected_fe->dofsPerHex(e);
    }
    levels[i]=OPERATOR_NEW DofLevel3(d1,d2,d3);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::vertexDofIndices(const Shape *element,
NumPtr<int> &vertex_dof_indices) const {
  int dpv=selected_fe->dofsPerVertex();
  int nv=element->numberVertices();
  NumPtr<int> vdi(vertex_dof_indices);
  for (int v=0;v<nv;v++) {
    NumPtr<int> vd(vertex_dofs,element->vertexIndex(v)*dpv,dpv);
    vdi.copyFrom(vd);
    vdi+=dpv;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::getDofIndices(const Shape1 *element,
const Level1 *level,NumPtr<int> &dof_indices) const {
  CHECK_SAME(element->numberDimensions(),1);

  int dofs_per_vertex=selected_fe->dofsPerVertex();
  int dofs_per_line=selected_fe->dofsPerLine();
  int nv=element->numberVertices();

  dof_indices.allocate(nv*dofs_per_vertex+dofs_per_line);
#ifdef INDEF
  dof_indices.initialize(invalid_dof_index);
#endif
  vertexDofIndices(element,dof_indices);
  NumPtr<int> remaining_indices(dof_indices,nv*dofs_per_vertex);
  getLevel(level)->getDofs(element,remaining_indices);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
void DofHandler::getDofIndices(const Shape2 *element,
const Level2 *level,NumPtr<int> &dof_indices) const {
  CHECK_SAME(element->numberDimensions(),2);

  int dofs_per_vertex=selected_fe->dofsPerVertex();
  int dofs_per_line=selected_fe->dofsPerLine();
  int nv=element->numberVertices();
  int dofs_per_quad=selected_fe->dofsPerQuad(element);
  int nl=element->numberCodimension1Shapes();

  dof_indices.allocate(nv*dofs_per_vertex+nl*dofs_per_line+dofs_per_quad);
#ifdef INDEF
  dof_indices.initialize(invalid_dof_index);
#endif
  vertexDofIndices(element,dof_indices);
  int offset=nv*dofs_per_vertex;
  const Level1 *level1=static_cast<const Level1*>(level);
  for (int l=0;l<nl;l++) {
    int face_index=element->codimension1ShapeIndex(l);
    const Shape1 *face_element=level->Level1::getShape(face_index);
    NumPtr<int> face_dofs(dof_indices,offset,dofs_per_line);
    getLevel(level1)->getDofs(face_element,face_dofs);
    offset+=dofs_per_line;
  }
  NumPtr<int> remaining_dofs(dof_indices,offset);
  getLevel(level)->getDofs(element,remaining_dofs);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void DofHandler::getDofIndices(const Shape3 *element,
const Level3 *level,NumPtr<int> &dof_indices) const {
  CHECK_SAME(element->numberDimensions(),3);
  int dofs_per_vertex=selected_fe->dofsPerVertex();
  int dofs_per_line=selected_fe->dofsPerLine();
  int dofs_per_hex=selected_fe->dofsPerHex(element);
  int nv=element->numberVertices();
  int nq=element->numberCodimension1Shapes();
  int nl=element->numberCodimension2Shapes();
  const Level2 *level2=static_cast<const Level2*>(level);
  const Level1 *level1=static_cast<const Level1*>(level);
  int sz=nv*dofs_per_vertex+nl*dofs_per_line+dofs_per_hex;
  for (int f=0;f<nq;f++) {
    int face_index=element->codimension1ShapeIndex(f);
    const Shape2 *face_element=level2->getShape(face_index);
    sz+=selected_fe->dofsPerQuad(face_element);
  }
  dof_indices.allocate(sz);
#ifdef INDEF
  dof_indices.initialize(invalid_dof_index);
#endif

  vertexDofIndices(element,dof_indices);
  int offset=nv*dofs_per_vertex;
  for (int l=0;l<nl;l++) {
    int line_index=element->codimension2ShapeIndex(l);
    const Shape1 *line_element=level1->getShape(line_index);
    NumPtr<int> line_dofs(dof_indices,offset,dofs_per_line);
    getLevel(level1)->getDofs(line_element,line_dofs);
    offset+=dofs_per_line;
  }
  for (int f=0;f<nq;f++) {
    int face_index=element->codimension1ShapeIndex(f);
    const Shape2 *face_element=level2->getShape(face_index);
    int dofs_per_quad=selected_fe->dofsPerQuad(face_element);
    NumPtr<int> face_dofs(dof_indices,offset,dofs_per_quad);
    getLevel(level2)->getDofs(face_element,face_dofs);
    offset+=dofs_per_quad;
  }
  NumPtr<int> remaining_dofs(dof_indices,offset);
  getLevel(level)->getDofs(element,remaining_dofs);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
int DofHandler::numberDofsOnChildren(const CODIMENSION1_SHAPE_CLASS*) 
const {
  return selected_fe->dofsPerVertex()
        +childrenPerCodimension1Shape()*selected_fe->dofsPerLine();
}
void DofHandler::dofsOnChildren(
const CODIMENSION1_SHAPE_CLASS *face_element,
const CODIMENSION1_LEVEL_CLASS *face_level,
NumPtr<int> &dofs_on_children) const {
  CHECK_TEST(face_level->hasChildren(face_element));
  int dpv=selected_fe->dofsPerVertex();
  int dpl=selected_fe->dofsPerLine();
  dofs_on_children.allocate(numberDofsOnChildren(face_element));
  NumPtr<int> doc(dofs_on_children);
  const CODIMENSION1_LEVEL_CLASS *face_child_level=
    face_level->finerLevel();
  int face_index=face_element->elementDataIndex();
  if (face_element->numberInteriorChildVertices()>0) {
    int c(INT_MAX),v(INT_MAX);
    int vertex=face_level->interiorChildVertex(face_index,c,v);
    int face_child_index=face_level->childIndex(face_index,c);
    const Shape1 *face_child=
      face_child_level->getShape(face_child_index);
    NumPtr<int> face_child_vertex_dof_indices(
      face_child->numberVertices()*dpv);
    vertexDofIndices(face_child,face_child_vertex_dof_indices);
    NumPtr<int> face_interior_child_vertex_dof_indices(
      face_child_vertex_dof_indices,v*dpv,dpv);
    doc.copyFrom(face_interior_child_vertex_dof_indices);
    doc+=face_interior_child_vertex_dof_indices.getNumber();
  }
  for (int child=0;child<childrenPerCodimension1Shape();child++) {
    int face_child_index=face_level->childIndex(face_index,child);
    const Shape1 *face_child=
      face_child_level->getShape(face_child_index);
    NumPtr<int> face_child_dof_indices(dpl);
    getLevel(face_child_level)->
      getDofs(face_child,face_child_dof_indices);
    doc.copyFrom(face_child_dof_indices);
    doc+=face_child_dof_indices.getNumber();
  }
  CHECK_SAME(0,doc.getNumber());
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
int DofHandler::numberDofsOnChildren(
const CODIMENSION1_SHAPE_CLASS *face_element) const {
  return (face_element->numberInteriorChildVertices()
         +face_element->numberCodimension1Shapes())
         *selected_fe->dofsPerVertex()
        +(face_element->numberCodimension1Shapes()
         *childrenPerCodimension2Shape()
         +face_element->numberInteriorChildCodimension1Shapes())
         *selected_fe->dofsPerLine()
        +childrenPerCodimension1Shape()
        *selected_fe->dofsPerQuad(face_element);
}
void DofHandler::dofsOnChildren(
const CODIMENSION1_SHAPE_CLASS *face_element,
const CODIMENSION1_LEVEL_CLASS *face_level,
NumPtr<int> &dofs_on_children) const {
  CHECK_TEST(face_level->hasChildren(face_element));
  int dpv=selected_fe->dofsPerVertex();
  int dpl=selected_fe->dofsPerLine();
  dofs_on_children.allocate(numberDofsOnChildren(face_element));
  NumPtr<int> doc(dofs_on_children);
  const CODIMENSION1_LEVEL_CLASS *face_child_level=
    face_level->finerLevel();
  const Level1 *line_level=static_cast<const Level1*>(face_level);
  const Level1 *face_line_child_level=line_level->finerLevel();
//child vertices interior to coarse face
  int face_index=face_element->elementDataIndex();
  if (face_element->numberInteriorChildVertices()>0) {
    int c=0,v=0;
    face_level->interiorChildVertex(face_index,c,v);
    int face_child_index=face_level->childIndex(face_index,c);
    const Shape2 *face_child_element=
      face_child_level->getShape(face_child_index);
    NumPtr<int> face_child_vertex_dof_indices(
      face_child_element->numberVertices()*dpv);
    vertexDofIndices(face_child_element,
      face_child_vertex_dof_indices);
    NumPtr<int> face_interior_child_vertex_dof_indices(
      face_child_vertex_dof_indices,v*dpv,dpv);
    doc.copyFrom(face_interior_child_vertex_dof_indices);
    doc+=face_interior_child_vertex_dof_indices.getNumber();
  }
//child vertices interior to coarse face lines
  for (int line=0;
  line<face_element->numberCodimension1Shapes();line++) {
    int face_line_index=
      face_element->codimension1ShapeIndex(line);
    const Shape1 *face_line_element=
      line_level->getShape(face_line_index);
    for (int i=0;
    i<face_line_element->numberInteriorChildVertices();i++) {
      int c=0,v=0;
      line_level->interiorChildVertex(face_line_index,c,v);
      int face_line_child_index=
        line_level->childIndex(face_line_index,c);
      const Shape1 *face_line_child_element=
        face_line_child_level->getShape(face_line_child_index);
      NumPtr<int> face_line_child_vertex_dof_indices(
        face_line_child_element->numberVertices()*dpv);
      vertexDofIndices(face_line_child_element,
        face_line_child_vertex_dof_indices);
      NumPtr<int> face_line_interior_child_vertex_dof_indices(
        face_line_child_vertex_dof_indices,v*dpv,dpv);
      doc.copyFrom(face_line_interior_child_vertex_dof_indices);
      doc+=face_line_interior_child_vertex_dof_indices.getNumber();
    }
  }
//child lines interior to coarse face
  for (int i=0;
  i<face_element->numberInteriorChildCodimension1Shapes();i++) {
    int c0,c1;
    face_element->
      interiorChildCodimension1ShapeNeighbors(i,c0,c1);
    int face_child_index0=face_level->childIndex(face_index,c0);
    int face_child_index1=face_level->childIndex(face_index,c1);
    const Shape2 *face_child_element0=
      face_child_level->getShape(face_child_index0);
    const Shape2 *face_child_element1=
      face_child_level->getShape(face_child_index1);
    int face_child_line_index=face_child_element0->
      sharedFaceElementIndex(face_child_element1);
    const Shape1 *face_child_line_element=
      face_line_child_level->getShape(face_child_line_index);
    NumPtr<int> face_line_child_dof_indices(dpl);
    getLevel(face_line_child_level)->
      getDofs(face_child_line_element,face_line_child_dof_indices);
    doc.copyFrom(face_line_child_dof_indices);
    doc+=face_line_child_dof_indices.getNumber();
  }
//child lines in coarse face lines
  for (int line=0;
  line<face_element->numberCodimension1Shapes();line++) {
    int face_line_index=
      face_element->codimension1ShapeIndex(line);
    for (int child=0;
    child<childrenPerCodimension2Shape();child++) {
      int face_line_child_index=
        line_level->childIndex(face_line_index,child);
      const Shape1 *face_line_child_element=
        face_line_child_level->getShape(face_line_child_index);
      NumPtr<int> face_line_child_dof_indices(dpl);
      getLevel(face_line_child_level)->
        getDofs(face_line_child_element,face_line_child_dof_indices);
      doc.copyFrom(face_line_child_dof_indices);
      doc+=face_line_child_dof_indices.getNumber();
    }
  }
//face child elements
  for (int child=0;child<childrenPerCodimension1Shape();child++) {
    int face_child_index=face_level->childIndex(face_index,child);
    const Shape2 *face_child_element=
      face_child_level->getShape(face_child_index);
    NumPtr<int> face_child_dof_indices(
      selected_fe->dofsPerQuad(face_child_element));
    getLevel(face_child_level)->
      getDofs(face_child_element,face_child_dof_indices);
    doc.copyFrom(face_child_dof_indices);
    doc+=face_child_dof_indices.getNumber();
  }
  CHECK_SAME(0,doc.getNumber());
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
int DofHandler::distributeDofsOnCell(const Shape1 *element,Level1 *level,
int next_free_dof) {
  int dpv=selected_fe->dofsPerVertex();
  if (dpv>0) {
    for (int v=0;v<element->numberVertices();v++) {
      int dof_number=element->vertexIndex(v)*dpv;
      if (vertex_dofs[dof_number]==invalid_dof_index) {
        for (int d=0;d<dpv;d++,next_free_dof++) {
          vertex_dofs[dof_number+d]=next_free_dof;
        }
      }
    }
  }
  DofLevel1 *dof_level=levels[level->numberCoarserLevels()];
  int dpl=selected_fe->dofsPerLine();
  dof_level->setDofs(element,dpl,next_free_dof);
  level->setUserFlag(element);
  return next_free_dof;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
int DofHandler::distributeDofsOnCell(const Shape2 *element,
Level2 *level,int next_free_dof) {
  int dpv=selected_fe->dofsPerVertex();
  if (dpv>0) {
    for (int v=0;v<element->numberVertices();v++) {
      int dof_number=element->vertexIndex(v)*dpv;
      if (vertex_dofs[dof_number]==invalid_dof_index) {
        for (int d=0;d<dpv;d++,next_free_dof++) {
          vertex_dofs[dof_number+d]=next_free_dof;
        }
      }
    }
  }
  DofLevel2 *dof_level2=levels[level->numberCoarserLevels()];
  int dpl=selected_fe->dofsPerLine();
  if (dpl>0) {
    const Level1 *level1=static_cast<const Level1*>(level);
    DofLevel1 *dof_level1=static_cast<DofLevel1*>(dof_level2);
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int face_index=element->codimension1ShapeIndex(f);
      const Shape1 *face_element=level1->getShape(face_index);
      dof_level1->setDofs(face_element,dpl,next_free_dof);
    }
  }
  int dpq=selected_fe->dofsPerQuad(element);
  if (dpq>0) {
    dof_level2->setDofs(element,dpq,next_free_dof);
  }
  level->setUserFlag(element);
  return next_free_dof;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
int DofHandler::distributeDofsOnCell(const Shape3 *element,
Level3 *level,int next_free_dof) {
  int dpv=selected_fe->dofsPerVertex();
  if (dpv>0) {
    for (int v=0;v<element->numberVertices();v++) {
      int dof_number=element->vertexIndex(v)*dpv;
      if (vertex_dofs[dof_number]==invalid_dof_index) {
        for (int d=0;d<dpv;d++,next_free_dof++) {
          vertex_dofs[dof_number+d]=next_free_dof;
        }
      }
    }
  }
  DofLevel3 *dof_level3=levels[level->numberCoarserLevels()];
  int dpl=selected_fe->dofsPerLine();
  const Level2 *level2=static_cast<const Level2*>(level);
  if (dpl>0) {
    const Level1 *level1=static_cast<const Level1*>(level);
    DofLevel1 *dof_level1=static_cast<DofLevel1*>(dof_level3);
    for (int l=0;l<element->numberCodimension2Shapes();l++) {
      int line_index=element->codimension2ShapeIndex(l);
      const Shape1 *line_element=level1->getShape(line_index);
      dof_level1->setDofs(line_element,dpl,next_free_dof);
    }
  }
  DofLevel2 *dof_level2=static_cast<DofLevel2*>(dof_level3);
  for (int f=0;f<element->numberCodimension1Shapes();f++) {
    int face_index=element->codimension1ShapeIndex(f);
    const Shape2 *face_element=level2->getShape(face_index);
    int dpq=selected_fe->dofsPerQuad(face_element);
    if (dpq>0) {
      dof_level2->setDofs(face_element,dpq,next_free_dof);
    }
  }
  int dph=selected_fe->dofsPerHex(element);
  if (dph>0) {
    dof_level3->setDofs(element,dph,next_free_dof);
  }
  level->setUserFlag(element);
  return next_free_dof;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::distributeDofs() {
  CHECK_TEST(tess->numberLevels()>0);
  reserveSpace();
  NumPtr<bool> user_flags;
  tess->saveUserFlags(user_flags);
  const_cast<Tessellation*>(tess)->clearUserFlags();
  int next_free_dof=0;   
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    const SHAPE_CLASS *element=tess->presentLevel()->getShape();
    next_free_dof=distributeDofsOnCell(element,tess->presentLevel(),
      next_free_dof);
  }
  used_dofs=next_free_dof;
  const_cast<Tessellation*>(tess)->loadUserFlags(user_flags);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::renumberDofs(const NumPtr<int> &new_numbers) {
  CHECK_SAME(new_numbers.getNumber(),numberDofs());
#ifdef DEBUG
  {
    NumPtrWithValue<int> tmp(new_numbers.getNumber());
    for (int i=0;i<new_numbers.getNumber();i++) {
      tmp.setValue(i,new_numbers[i],new_numbers[i]);
    }
    bool descending=false;
    tmp.heapSort(descending);
    for (int i=0;i<tmp.getNumber();i++) {
      CHECK_SAME(tmp.getValue(i),i);
    }
  }
#endif
  for (int i=0;i<vertex_dofs.getNumber();i++) {
    int &vdi=vertex_dofs[i];
    if (vdi!=invalid_dof_index) vdi=new_numbers[vdi];
    else {
      CHECK_TEST(!tess->vertexUsed(i/selected_fe->dofsPerVertex()));
    }
  }
  for (int l=0;l<levels.getNumber();l++) { 
    levels[l]->setDofs(new_numbers);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::getDofValues(const SHAPE_CLASS *element,
const LEVEL_CLASS *level,const Vector &values,Vector &local_values) 
const {
  NumPtr<int> dof_indices;
  getDofIndices(element,level,dof_indices);
  local_values.reinit(dof_indices.getNumber(),true);
  for (int i=0;i<dof_indices.getNumber();i++) {
    local_values[i]=values[dof_indices[i]];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::getInterpolatedDofValues(const SHAPE_CLASS *element,
const LEVEL_CLASS *level,const Vector &values,
Vector &interpolated_values) const {
  CHECK_SAME(values.size(),numberDofs());
  int dofs_per_cell=selected_fe->dofsPerCell(element);
  if (!level->hasChildren(element)) {
    getDofValues(element,level,values,interpolated_values);
  } else {
    Vector tmp1(dofs_per_cell),tmp2(dofs_per_cell);
    int index=element->elementDataIndex();
    interpolated_values.reinit(dofs_per_cell);
    for (int c=0;c<childrenPerShape();c++) {
      int child_index=level->childIndex(index,c);
      const SHAPE_CLASS *child_element=
        level->finerLevel()->getShape(child_index);
      getInterpolatedDofValues(child_element,level->finerLevel(),values,
        tmp1);
      const Matrix &prolongation_matrix=
        selected_fe->prolongationMatrix(element,c,level);
      prolongation_matrix.transposeVmult(tmp2,tmp1);
      for (int i=0;i<dofs_per_cell;i++) {
        interpolated_values[i]+=tmp2[i];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::makeSparsityPattern(SparsityPatternBase &sp) const {
  CHECK_SAME(sp.nRows(),used_dofs);
  CHECK_SAME(sp.nCols(),used_dofs);
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int dofs_per_cell=selected_fe->dofsPerCell(element);
    NumPtr<int> dofs_on_this_cell;
    getDofIndices(element,present_level,dofs_on_this_cell);
    for (int i=0;i<dofs_per_cell;i++) {
      for (int j=0;j<dofs_per_cell;j++) {
        sp.add(dofs_on_this_cell[i],dofs_on_this_cell[j]);
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
int DofHandler::maxCouplingsBetweenDofs() const {
  return min(3*selected_fe->dofsPerVertex()
            +2*selected_fe->dofsPerLine(),used_dofs);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
int DofHandler::maxCouplingsBetweenDofs() const {
  int nuv=tess->getVertices().getNumber();
  NumPtr<set<pair<const SHAPE_CLASS*,LEVEL_CLASS*> > > 
    vertex_active_element_sets(nuv);
  NumPtr<set<pair<const SHAPE_CLASS*,LEVEL_CLASS*> > > 
    vertex_element_sets(nuv);
//for each used vertex, find set of active elements having it as vertex
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    for (int v=0;v<element->numberVertices();v++) {
      vertex_active_element_sets[element->vertexIndex(v)].
        insert(make_pair(element,present_level));
    }
  }
//for each vertex, find set of elements having it as a vertex
  for (bool doit=tess->firstShape();doit;doit=tess->nextShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    for (int v=0;v<element->numberVertices();v++) {
      vertex_element_sets[element->vertexIndex(v)].
        insert(make_pair(element,present_level));
    }
  }
  int max_couplings=0;
  LEVEL_CLASS *tess_coarsest_level=tess->coarsestLevel();
  LEVEL_CLASS *tess_finest_level=tess_coarsest_level->finestLevel();
  for (int v=0;v<nuv;v++) {
    LEVEL_CLASS *finest_level=tess_coarsest_level;
    LEVEL_CLASS *coarsest_level=tess_finest_level;
    const set<pair<const SHAPE_CLASS*,LEVEL_CLASS*> > 
      &vertex_active_element_set=vertex_active_element_sets[v];
//  find finest and coarsest levels of elements with this vertex
    for (set<pair<const SHAPE_CLASS*,LEVEL_CLASS*> >::const_iterator 
      ves_iterator=vertex_active_element_set.begin();
    ves_iterator!=vertex_active_element_set.end();ves_iterator++) {
      LEVEL_CLASS *level=ves_iterator->second;
      if (level->isFinerThan(finest_level)) finest_level=level; 
      if (level->isCoarserThan(coarsest_level)) coarsest_level=level; 
    }
    CHECK_TEST(finest_level->numberCoarserLevels()
              -coarsest_level->numberCoarserLevels()<=1);
    int couplings=0;
    if (finest_level==coarsest_level) {
//    all elements containing this vertex are in the same level
//    set of all indices of vertices in elements containing this vertex
      set<int> vertex_indices;
      set<int> codimension1_shape_indices;
#if (SPACEDIM==3)
      set<int> codimension2_shape_indices;
#endif
      for (set<pair<const SHAPE_CLASS*,LEVEL_CLASS*> >::const_iterator 
      ves_iterator=vertex_active_element_set.begin();
      ves_iterator!=vertex_active_element_set.end();ves_iterator++) {
        const SHAPE_CLASS *element=ves_iterator->first;
#if (SPACEDIM==2)
        couplings+=selected_fe->dofsPerQuad(element);
#else
        couplings+=selected_fe->dofsPerHex(element);
#endif
        for (int ev=0;ev<element->numberVertices();ev++) {
          vertex_indices.insert(element->vertexIndex(ev));
        }
        for (int ef=0;ef<element->numberCodimension1Shapes();ef++) {
          codimension1_shape_indices.
            insert(element->codimension1ShapeIndex(ef));
        }
#if (SPACEDIM==3)
        for (int el=0;el<element->numberCodimension2Shapes();el++) {
          codimension2_shape_indices.
            insert(element->codimension2ShapeIndex(el));
        }
#endif
      }
      couplings+=vertex_indices.size()*selected_fe->dofsPerVertex();
#if (SPACEDIM==2)
      couplings+=codimension1_shape_indices.size()
                    *selected_fe->dofsPerLine();
#else
      couplings+=codimension2_shape_indices.size()
                    *selected_fe->dofsPerLine();
      for (set<int>::const_iterator face_iterator=
      codimension1_shape_indices.begin();
      face_iterator!=codimension1_shape_indices.end();face_iterator++) {
        CODIMENSION1_SHAPE_CLASS *face=finest_level->
          CODIMENSION1_LEVEL_CLASS::getShape(*face_iterator);
        couplings+=selected_fe->dofsPerQuad(face);
      }
#endif
    } else {
//    elements containing this vertex are not in the same level
//    indices of vertices in coarse active elements with this vertex
      set<int> coarse_vertex_indices;
//    indices of faces in coarse active elements with this vertex
      set<int> coarse_codimension1_shape_indices;
#if (SPACEDIM==3)
//    indices of lines in coarse active elements with this vertex
      set<int> coarse_codimension2_shape_indices;
#endif
//    coarse elements containing this vertex and having no children
      for (set<pair<const SHAPE_CLASS*,LEVEL_CLASS*> >::const_iterator 
      ves_iterator=vertex_active_element_set.begin();
      ves_iterator!=vertex_active_element_set.end();ves_iterator++) {
        const LEVEL_CLASS *level=ves_iterator->second;
        if (level==coarsest_level) {
          const SHAPE_CLASS *element=ves_iterator->first;
#if (SPACEDIM==2)
          couplings+=selected_fe->dofsPerQuad(element);
#else
          couplings+=selected_fe->dofsPerHex(element);
#endif
          for (int ev=0;ev<element->numberVertices();ev++) {
            coarse_vertex_indices.insert(element->vertexIndex(ev));
          }
          for (int ef=0;ef<element->numberCodimension1Shapes();ef++) {
            coarse_codimension1_shape_indices.
              insert(element->codimension1ShapeIndex(ef));
          }
#if (SPACEDIM==3)
          for (int el=0;el<element->numberCodimension2Shapes();el++) {
            coarse_codimension2_shape_indices.
              insert(element->codimension2ShapeIndex(el));
          }
#endif
        }
      }

//    all coarse elements having children and containing this vertex
      set<pair<const SHAPE_CLASS*,LEVEL_CLASS*> > ves_difference;
      const set<pair<const SHAPE_CLASS*,LEVEL_CLASS*> > 
        &vertex_element_set=vertex_element_sets[v];
      for (set<pair<const SHAPE_CLASS*,LEVEL_CLASS*> >::const_iterator
      ves_iterator=vertex_element_set.begin();
      ves_iterator!=vertex_element_set.end();ves_iterator++) {
        const pair<const SHAPE_CLASS*,LEVEL_CLASS*> &p=*ves_iterator;
        if (p.second!=coarsest_level) continue;
        if (p.second->hasChildren(p.first)) {
          ves_difference.insert(*ves_iterator);
        }
      }
//    all indices of vertices in fine elements containing this vertex
      set<int> fine_vertex_indices;
      set<int> fine_codimension1_shape_indices;
#if (SPACEDIM==3)
      set<int> fine_codimension2_shape_indices;
#endif
      for (set<pair<const SHAPE_CLASS*,LEVEL_CLASS*> >::const_iterator 
      vesdi=ves_difference.begin();vesdi!=ves_difference.end();vesdi++){
        const LEVEL_CLASS *level=vesdi->second;
        if (level!=coarsest_level) continue;
        const SHAPE_CLASS *element=vesdi->first;
        CHECK_TEST(level->hasChildren(element));
        for (int c=0;c<childrenPerShape();c++) {
          int child_index=
            level->childIndex(element->elementDataIndex(),c);
          const SHAPE_CLASS *child_element=
            finest_level->getShape(child_index);
          for (int ev=0;ev<child_element->numberVertices();
          ev++) {
            fine_vertex_indices.
              insert(child_element->vertexIndex(ev));
          }
          for (int ef=0;ef<child_element->numberCodimension1Shapes();
          ef++) {
            fine_codimension1_shape_indices.
              insert(child_element->codimension1ShapeIndex(ef));
          }
#if (SPACEDIM==3)
          for (int el=0;el<child_element->numberCodimension2Shapes();
          el++) {
            fine_codimension2_shape_indices.
              insert(child_element->codimension2ShapeIndex(el));
          }
#endif
        }
      }
      couplings+=(fine_vertex_indices.size()
        +coarse_vertex_indices.size())*selected_fe->dofsPerVertex();
#if (SPACEDIM==2)
      couplings+=(fine_codimension1_shape_indices.size()
                 +coarse_codimension1_shape_indices.size())
                *selected_fe->dofsPerLine();
#else
      couplings+=(fine_codimension2_shape_indices.size()
                 +coarse_codimension2_shape_indices.size())
                 *selected_fe->dofsPerLine();
      for (set<int>::const_iterator face_iterator=
      coarse_codimension1_shape_indices.begin();
      face_iterator!=coarse_codimension1_shape_indices.end();
      face_iterator++) {
        CODIMENSION1_SHAPE_CLASS *face=coarsest_level->
          CODIMENSION1_LEVEL_CLASS::getShape(*face_iterator);
        couplings+=selected_fe->dofsPerQuad(face);
      }
      for (set<int>::const_iterator face_iterator=
      fine_codimension1_shape_indices.begin();
      face_iterator!=fine_codimension1_shape_indices.end();
      face_iterator++) {
        CODIMENSION1_SHAPE_CLASS *face=finest_level->
          CODIMENSION1_LEVEL_CLASS::getShape(*face_iterator);
        couplings+=selected_fe->dofsPerQuad(face);
      }
#endif
    }
    max_couplings=max(max_couplings,couplings);
  }
  return min(max_couplings,numberDofs());
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::assembleLinearSystem(const Mapping &mapping,
const map<int,Quadrature<SPACEDIM> > &qmap,
map<int,Quadrature<SPACEDIM-1> > &fqmap,SparseMatrix &system_matrix,
Vector &system_rhs) const {
  UPDATE_FLAGS update_flags= UPDATE_VALUES | UPDATE_GRADIENTS | 
    UPDATE_JxW_VALUES | UPDATE_Q_POINTS;
  FEValues fe_values(mapping,*selected_fe,qmap,update_flags);
#if (SPACEDIM>1)
  UPDATE_FLAGS face_update_flags=UPDATE_VALUES | UPDATE_Q_POINTS |
    UPDATE_NORMAL_VECTORS | UPDATE_JxW_VALUES;
  FEFaceValues fe_face_values(mapping,*selected_fe,fqmap,
    face_update_flags);
#endif
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int dofs_per_cell=selected_fe->dofsPerCell(element);

//  local stiffness matrix and right-hand side
    Matrix cell_matrix(dofs_per_cell,dofs_per_cell);
    Vector cell_rhs(dofs_per_cell);
    cell_matrix=0.;
    cell_rhs=0.;

//  shape functions and quadrature points are mapped 
//    from canonical elements to present element
    fe_values.mapShapes(*tess,element,present_level);
    const NumPtr<Point<SPACEDIM> > &q_points=
      fe_values.getQuadraturePoints(element);
    const NumPtr<REAL> &jxw_values=fe_values.getJxWValues(element);
    const Array<2,REAL> &sv=*fe_values.getShapeValues(element);
    const Array<2,Tensor<1,SPACEDIM> > &sg=
      *fe_values.getShapeGradients(element);

//  computation of local stiffness matrix and right-hand side
    MaterialModel *mm=present_level->getMaterialModel();
    CHECK_POINTER(mm);
    mm->elementStiffnessMatrixAndRHS(q_points,jxw_values,sv,sg,
      cell_matrix,cell_rhs);

//  neumann boundary conditions
#if (SPACEDIM==1)
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      if (present_level->neighborLevel(f)==0) {
        const BoundaryCondition *bc=
          present_level->Level0::getBoundaryCondition(f);
        CHECK_POINTER(bc);
        if (bc->isNatural()) {
//        the following 9 lines are awkward:
//        the use of FEValues of some sort is necessary in order to
//          get the shapeValue's in the correct order
//        DEAL has trouble with FEFaceValues in 1D because there is
//          no Shape associated with faces of Interval's
//        FEValues wants to compute Jacobians and quadrature points
//          when it computes shapeValue's
          NumPtr<Point<1> > p(1); 
          p[0]=Interval::canonicalVertex(f);
          map<int,Quadrature<1> > fqmap_1d;
          fqmap_1d[element->numberVertices()]=Quadrature<1>(p);
          FEValues face_fe_values(mapping,*selected_fe,fqmap_1d,
            UPDATE_VALUES | UPDATE_Q_POINTS | UPDATE_JxW_VALUES); 
          face_fe_values.mapShapes(*tess,element,present_level);
          const Array<2,REAL> &fsv=
            *face_fe_values.getShapeValues(element);
          const NumPtr<Point<1> > &face_q_points=
            face_fe_values.getQuadraturePoints(element);
          NumPtr<REAL> face_jxw_values(1); face_jxw_values[0]=1.;
          const TimeDependentBoundaryCondition *tdbc=
            dynamic_cast<const TimeDependentBoundaryCondition*>(bc);
          if (tdbc) {
//          the linear system is an elliptic projection for initial data
            tdbc->faceNeumannValues(face_q_points,0.,face_jxw_values,0,
              fsv,cell_rhs);
          } else {
            bc->faceNeumannValues(face_q_points,face_jxw_values,0,fsv,
              cell_rhs);
          }
        }
      }
    }
#else
    int offset=0;
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int face_index=element->codimension1ShapeIndex(f);
      const BoundaryCondition *bc=present_level->
        CODIMENSION1_LEVEL_CLASS::getBoundaryCondition(face_index);
      if (bc!=0) {
        if (bc->isNatural()) {
          const CODIMENSION1_SHAPE_CLASS *face_element=present_level->
            CODIMENSION1_LEVEL_CLASS::getShape(face_index);
          fe_face_values.mapShapes(*tess,element,present_level,f);
          const Array<2,REAL> &fsv=
            *fe_face_values.getShapeValues(element);
          NumPtr<Point<SPACEDIM> > face_q_points=
            fe_face_values.getQuadraturePoints(element,f);
          NumPtr<REAL> face_jxw_values=
            fe_face_values.getJxWValues(element,f);
          const TimeDependentBoundaryCondition *tdbc=
            dynamic_cast<const TimeDependentBoundaryCondition*>(bc);
          if (tdbc) {
//          the linear system is an elliptic projection for initial data
            tdbc->faceNeumannValues(face_q_points,0.,face_jxw_values,
              offset,fsv,cell_rhs);
          } else {
            bc->faceNeumannValues(face_q_points,face_jxw_values,
              offset,fsv,cell_rhs);
          }
        }
      }
      offset+=
        fe_face_values.getFaceQuadrature(element,f).numberPoints();
    }
#endif

//  add element-wise contributions to global arrays
    NumPtr<int> local_dof_indices;
    getDofIndices(element,present_level,local_dof_indices);
    for (int i=0;i<dofs_per_cell;i++) {
      int ldfi=local_dof_indices[i];
      for (int j=0;j<dofs_per_cell;j++) {
        system_matrix.add(ldfi,local_dof_indices[j],cell_matrix(i,j));
      }
      system_rhs[ldfi]+=cell_rhs[i];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::assembleL2ProjectionLinearSystem(const Mapping &mapping,
const map<int,Quadrature<SPACEDIM> > &qmap,SparseMatrix &system_matrix,
Vector &system_rhs) const {
  UPDATE_FLAGS update_flags=
    UPDATE_VALUES | UPDATE_JxW_VALUES | UPDATE_Q_POINTS;
  FEValues fe_values(mapping,*selected_fe,qmap,update_flags);
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int dofs_per_cell=selected_fe->dofsPerCell(element);

//  local stiffness matrix and right-hand side
    Matrix cell_matrix(dofs_per_cell,dofs_per_cell);
    Vector cell_rhs(dofs_per_cell);
    cell_matrix=0.;
    cell_rhs=0.;

//  shape functions and quadrature points are mapped 
//    from canonical elements to present element
    fe_values.mapShapes(*tess,element,present_level);
    const NumPtr<Point<SPACEDIM> > &q_points=
      fe_values.getQuadraturePoints(element);
    const NumPtr<REAL> &jxw_values=fe_values.getJxWValues(element);
    const Array<2,REAL> &sv=*fe_values.getShapeValues(element);

    TransientMaterialModel *mm=dynamic_cast<TransientMaterialModel*>(
      present_level->getMaterialModel());
    CHECK_POINTER(mm);
    mm->elementL2ProjectionMatrixAndRHS(q_points,jxw_values,sv,
      cell_matrix,cell_rhs);
//  add element-wise contributions to global arrays
    NumPtr<int> local_dof_indices;
    getDofIndices(element,present_level,local_dof_indices);
    for (int i=0;i<dofs_per_cell;i++) {
      int ldfi=local_dof_indices[i];
      for (int j=0;j<dofs_per_cell;j++) {
        system_matrix.add(ldfi,local_dof_indices[j],cell_matrix(i,j));
      }
      system_rhs[ldfi]+=cell_rhs[i];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::assembleEllipticProjectionLinearSystem(
const Mapping &mapping,const map<int,Quadrature<SPACEDIM> > &qmap,
SparseMatrix &system_matrix,Vector &system_rhs) const {
  UPDATE_FLAGS update_flags=
    UPDATE_GRADIENTS | UPDATE_JxW_VALUES | UPDATE_Q_POINTS;
  FEValues fe_values(mapping,*selected_fe,qmap,update_flags);
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int dofs_per_cell=selected_fe->dofsPerCell(element);

//  local stiffness matrix and right-hand side
    Matrix cell_matrix(dofs_per_cell,dofs_per_cell);
    Vector cell_rhs(dofs_per_cell);
    cell_matrix=0.;
    cell_rhs=0.;

//  shape functions and quadrature points are mapped 
//    from canonical elements to present element
    fe_values.mapShapes(*tess,element,present_level);
    const NumPtr<Point<SPACEDIM> > &q_points=
      fe_values.getQuadraturePoints(element);
    const NumPtr<REAL> &jxw_values=fe_values.getJxWValues(element);
    const Array<2,Tensor<1,SPACEDIM> > &sg=
      *fe_values.getShapeGradients(element);

//  computation of local stiffness matrix and right-hand side
    TransientMaterialModel *mm=dynamic_cast<TransientMaterialModel*>(
      present_level->getMaterialModel());
    CHECK_POINTER(mm);
    mm->elementEllipticProjectionMatrixAndRHS(q_points,jxw_values,sg,
      cell_matrix,cell_rhs);
//  add element-wise contributions to global arrays
    NumPtr<int> local_dof_indices;
    getDofIndices(element,present_level,local_dof_indices);
    for (int i=0;i<dofs_per_cell;i++) {
      int ldfi=local_dof_indices[i];
      for (int j=0;j<dofs_per_cell;j++) {
        system_matrix.add(ldfi,local_dof_indices[j],cell_matrix(i,j));
      }
      system_rhs[ldfi]+=cell_rhs[i];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::assembleTransientLinearSystem(REAL time,
const Mapping &mapping,const map<int,Quadrature<SPACEDIM> > &qmap,
map<int,Quadrature<SPACEDIM-1> > &fqmap,SparseMatrix &system_mass_matrix,
SparseMatrix &system_stiffness_matrix,Vector &system_rhs) const {
  UPDATE_FLAGS update_flags= UPDATE_VALUES | UPDATE_GRADIENTS | 
    UPDATE_JxW_VALUES | UPDATE_Q_POINTS;
  FEValues fe_values(mapping,*selected_fe,qmap,update_flags);
#if (SPACEDIM>1)
  UPDATE_FLAGS face_update_flags=UPDATE_VALUES | UPDATE_Q_POINTS |
    UPDATE_NORMAL_VECTORS | UPDATE_JxW_VALUES;
  FEFaceValues fe_face_values(mapping,*selected_fe,fqmap,
    face_update_flags);
#endif
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int dofs_per_cell=selected_fe->dofsPerCell(element);

//  local stiffness matrix and right-hand side
    Matrix cell_mass_matrix(dofs_per_cell,dofs_per_cell);
    Matrix cell_stiffness_matrix(dofs_per_cell,dofs_per_cell);
    Vector cell_rhs(dofs_per_cell);
    cell_mass_matrix=0.;
    cell_stiffness_matrix=0.;
    cell_rhs=0.;

//  shape functions and quadrature points are mapped 
//    from canonical elements to present element
    fe_values.mapShapes(*tess,element,present_level);
    const NumPtr<Point<SPACEDIM> > &q_points=
      fe_values.getQuadraturePoints(element);
    const NumPtr<REAL> &jxw_values=fe_values.getJxWValues(element);
    const Array<2,REAL> &sv=*fe_values.getShapeValues(element);
    const Array<2,Tensor<1,SPACEDIM> > &sg=
      *fe_values.getShapeGradients(element);

//  computation of local stiffness matrix and right-hand side
    TransientMaterialModel *mm=dynamic_cast<TransientMaterialModel*>(
      present_level->getMaterialModel());
    CHECK_POINTER(mm);
    mm->elementStiffnessMatrixAndRHS(q_points,time,jxw_values,sv,sg,
      cell_mass_matrix,cell_stiffness_matrix,cell_rhs);

//  neumann boundary conditions
#if (SPACEDIM==1)
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      if (present_level->neighborLevel(f)==0) {
        const BoundaryCondition *bc=
          present_level->Level0::getBoundaryCondition(f);
        CHECK_POINTER(bc);
        if (bc->isNatural()) {
//        the following 9 lines are awkward:
//        the use of FEValues of some sort is necessary in order to
//          get the shapeValue's in the correct order
//        DEAL has trouble with FEFaceValues in 1D because there is
//          no Shape associated with faces of Interval's
//        FEValues wants to compute Jacobians and quadrature points
//          when it computes shapeValue's
          NumPtr<Point<1> > p(1); 
          p[0]=Interval::canonicalVertex(f);
          map<int,Quadrature<1> > fqmap_1d;
          fqmap_1d[element->numberVertices()]=Quadrature<1>(p);
          FEValues face_fe_values(mapping,*selected_fe,fqmap_1d,
            UPDATE_VALUES | UPDATE_Q_POINTS | UPDATE_JxW_VALUES); 
          face_fe_values.mapShapes(*tess,element,present_level);
          const Array<2,REAL> &fsv=
            *face_fe_values.getShapeValues(element);
          const NumPtr<Point<1> > &face_q_points=
            face_fe_values.getQuadraturePoints(element);
          NumPtr<REAL> face_jxw_values(1); face_jxw_values[0]=1.;
          const TimeDependentBoundaryCondition *tdbc=
            dynamic_cast<const TimeDependentBoundaryCondition*>(bc);
          if (tdbc) {
            tdbc->faceNeumannValues(face_q_points,time,face_jxw_values,0,
              fsv,cell_rhs);
          } else {
            bc->faceNeumannValues(face_q_points,face_jxw_values,0,fsv,
              cell_rhs);
          }
        }
      }
    }
#else
    int offset=0;
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int face_index=element->codimension1ShapeIndex(f);
      const BoundaryCondition *bc=present_level->
        CODIMENSION1_LEVEL_CLASS::getBoundaryCondition(face_index);
      if (bc!=0) {
        if (bc->isNatural()) {
          const CODIMENSION1_SHAPE_CLASS *face_element=present_level->
            CODIMENSION1_LEVEL_CLASS::getShape(face_index);
          fe_face_values.mapShapes(*tess,element,present_level,f);
          const Array<2,REAL> &fsv=
            *fe_face_values.getShapeValues(element);
          NumPtr<Point<SPACEDIM> > face_q_points=
            fe_face_values.getQuadraturePoints(element,f);
          NumPtr<REAL> face_jxw_values=
            fe_face_values.getJxWValues(element,f);
          const TimeDependentBoundaryCondition *tdbc=
            dynamic_cast<const TimeDependentBoundaryCondition*>(bc);
          if (tdbc) {
            tdbc->faceNeumannValues(face_q_points,time,face_jxw_values,
              offset,fsv,cell_rhs);
          } else {
            bc->faceNeumannValues(face_q_points,face_jxw_values,
              offset,fsv,cell_rhs);
          }
        }
      }
      offset+=
        fe_face_values.getFaceQuadrature(element,f).numberPoints();
    }
#endif
//  add element-wise contributions to global arrays
    NumPtr<int> local_dof_indices;
    getDofIndices(element,present_level,local_dof_indices);
    for (int i=0;i<dofs_per_cell;i++) {
      int ldfi=local_dof_indices[i];
      for (int j=0;j<dofs_per_cell;j++) {
        system_mass_matrix.add(ldfi,local_dof_indices[j],
          cell_mass_matrix(i,j));
        system_stiffness_matrix.add(ldfi,local_dof_indices[j],
          cell_stiffness_matrix(i,j));
      }
      system_rhs[ldfi]+=cell_rhs[i];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::assembleTransientLinearSystem(REAL time,
const Mapping &mapping,const map<int,Quadrature<SPACEDIM> > &qmap,
map<int,Quadrature<SPACEDIM-1> > &fqmap,Vector &system_rhs) const {
  UPDATE_FLAGS update_flags=
    UPDATE_VALUES | UPDATE_JxW_VALUES | UPDATE_Q_POINTS;
  FEValues fe_values(mapping,*selected_fe,qmap,update_flags);
#if (SPACEDIM>1)
  UPDATE_FLAGS face_update_flags=UPDATE_VALUES | UPDATE_Q_POINTS |
    UPDATE_NORMAL_VECTORS | UPDATE_JxW_VALUES;
  FEFaceValues fe_face_values(mapping,*selected_fe,fqmap,
    face_update_flags);
#endif
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int dofs_per_cell=selected_fe->dofsPerCell(element);

//  local stiffness matrix and right-hand side
    Vector cell_rhs(dofs_per_cell);
    cell_rhs=0.;

//  shape functions and quadrature points are mapped 
//    from canonical elements to present element
    fe_values.mapShapes(*tess,element,present_level);
    const NumPtr<Point<SPACEDIM> > &q_points=
      fe_values.getQuadraturePoints(element);
    const NumPtr<REAL> &jxw_values=fe_values.getJxWValues(element);
    const Array<2,REAL> &sv=*fe_values.getShapeValues(element);

//  computation of local stiffness matrix and right-hand side
    TimeIndependentTransientMaterialModel *mm=
      dynamic_cast<TimeIndependentTransientMaterialModel*>(
        present_level->getMaterialModel());
    CHECK_POINTER(mm);
    mm->elementRHS(q_points,time,jxw_values,sv,cell_rhs);
//  neumann boundary conditions
#if (SPACEDIM==1)
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      if (present_level->neighborLevel(f)==0) {
        const BoundaryCondition *bc=
          present_level->Level0::getBoundaryCondition(f);
        CHECK_POINTER(bc);
        if (bc->isNatural()) {
//        the following 9 lines are awkward:
//        the use of FEValues of some sort is necessary in order to
//          get the shapeValue's in the correct order
//        DEAL has trouble with FEFaceValues in 1D because there is
//          no Shape associated with faces of Interval's
//        FEValues wants to compute Jacobians and quadrature points
//          when it computes shapeValue's
          NumPtr<Point<1> > p(1); 
          p[0]=Interval::canonicalVertex(f);
          map<int,Quadrature<1> > fqmap_1d;
          fqmap_1d[element->numberVertices()]=Quadrature<1>(p);
          FEValues face_fe_values(mapping,*selected_fe,fqmap_1d,
            UPDATE_VALUES | UPDATE_Q_POINTS | UPDATE_JxW_VALUES); 
          face_fe_values.mapShapes(*tess,element,present_level);
          const Array<2,REAL> &fsv=
            *face_fe_values.getShapeValues(element);
          const NumPtr<Point<1> > &face_q_points=
            face_fe_values.getQuadraturePoints(element);
          NumPtr<REAL> face_jxw_values(1); face_jxw_values[0]=1.;
          const TimeDependentBoundaryCondition *tdbc=
            dynamic_cast<const TimeDependentBoundaryCondition*>(bc);
          if (tdbc) {
            tdbc->faceNeumannValues(face_q_points,time,face_jxw_values,0,
              fsv,cell_rhs);
          } else {
            bc->faceNeumannValues(face_q_points,face_jxw_values,0,fsv,
              cell_rhs);
          }
        }
      }
    }
#else
    int offset=0;
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int face_index=element->codimension1ShapeIndex(f);
      const BoundaryCondition *bc=present_level->
        CODIMENSION1_LEVEL_CLASS::getBoundaryCondition(face_index);
      if (bc!=0) {
        if (bc->isNatural()) {
          const CODIMENSION1_SHAPE_CLASS *face_element=present_level->
            CODIMENSION1_LEVEL_CLASS::getShape(face_index);
          fe_face_values.mapShapes(*tess,element,present_level,f);
          const Array<2,REAL> &fsv=
            *fe_face_values.getShapeValues(element);
          NumPtr<Point<SPACEDIM> > face_q_points=
            fe_face_values.getQuadraturePoints(element,f);
          NumPtr<REAL> face_jxw_values=
            fe_face_values.getJxWValues(element,f);
          const TimeDependentBoundaryCondition *tdbc=
            dynamic_cast<const TimeDependentBoundaryCondition*>(bc);
          if (tdbc) {
            tdbc->faceNeumannValues(face_q_points,time,face_jxw_values,
              offset,fsv,cell_rhs);
          } else {
            bc->faceNeumannValues(face_q_points,face_jxw_values,
              offset,fsv,cell_rhs);
          }
        }
      }
      offset+=
        fe_face_values.getFaceQuadrature(element,f).numberPoints();
    }
#endif
//  add element-wise contributions to global arrays
    NumPtr<int> local_dof_indices;
    getDofIndices(element,present_level,local_dof_indices);
    for (int i=0;i<dofs_per_cell;i++) {
      system_rhs[local_dof_indices[i]]+=cell_rhs[i];
    }
  }
}
//see Deal.II code in VectorTools::interpolate_boundary_values
void DofHandler::interpolateBoundaryValues(const Mapping &mapping,
map<int,REAL> &boundary_values,REAL time) const {
  int n_components=selected_fe->nComponents();
  bool fe_is_system=(n_components!=1);
  int dpv=selected_fe->dofsPerVertex();
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *level=tess->presentLevel();
    const SHAPE_CLASS *element=level->getShape();
    NumPtr<int> vertex_dof_indices(
      element->numberVertices()*selected_fe->dofsPerVertex());
    vertexDofIndices(element,vertex_dof_indices);
    for (int face_no=0;
    face_no<element->numberCodimension1Shapes();face_no++) {
      if (level->neighborLevel(face_no)==0) {
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          static_cast<CODIMENSION1_LEVEL_CLASS*>(level);
#if (SPACEDIM==1)
        const BoundaryCondition *bc=
          codimension1_level->getBoundaryCondition(face_no);
        CHECK_POINTER(bc);
        if (bc->isEssential()) {
          Vector function_values(n_components);
          const Point<SPACEDIM> &boundary_point=tess->
            getVertices()[element->codimension1ShapeIndex(face_no)];
          FunctionSystem *fs=bc->getFunctionSystem();
          TimeDependentFunctionSystem *tdfs=
            dynamic_cast<TimeDependentFunctionSystem*>(fs);
          if (tdfs) {
            tdfs->value(boundary_point,time,function_values);
          } else {
            CHECK_POINTER(fs);
            fs->value(boundary_point,function_values);
          }
          NumPtr<int> vdi(vertex_dof_indices,face_no*dpv,dpv);
          for (int i=0;i<dpv;i++) {
            if (bc->isEssential(i)) {
              boundary_values[vdi[i]]=function_values[i];
            }
          }
        }
#else
        int face_index=element->codimension1ShapeIndex(face_no);
        const BoundaryCondition *bc=
          codimension1_level->getBoundaryCondition(face_index);
        CHECK_POINTER(bc);
        if (bc->isEssential()) {
          const CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape(face_index);
          NumPtr<Point<SPACEDIM> > element_lattice_points;
          equallySpacedLatticePoints(element->numberVertices(),
            selected_fe->getDegree(),element_lattice_points);
          NumPtr<int> element_face_lattice_point_indices;
          equallySpacedLatticePointIndicesOfFace(element->numberVertices(),
            face_no,selected_fe->getDegree(),
            element_face_lattice_point_indices);
          NumPtr<Point<SPACEDIM> > canonical_locations(
            element_face_lattice_point_indices.getNumber());
          for (int i=0;i<element_face_lattice_point_indices.getNumber();
          i++) {
            canonical_locations[i]=
             element_lattice_points[element_face_lattice_point_indices[i]];
          }
          NumPtr<Point<SPACEDIM> > dof_locations(
            canonical_locations.getNumber());
          mapping.elementPoints(element,*tess,level,canonical_locations,
            dof_locations);
          NumPtr<int> face_dofs;
          getDofIndices(face_element,codimension1_level,face_dofs);
          NumPtr<int> element_dofs;
          getDofIndices(element,level,element_dofs);
          const Boundary *boundary=bc->getBoundary();
          selected_fe->setBoundaryValues(element,face_element,face_no,
            element_dofs,face_dofs,canonical_locations,dof_locations,
            bc,boundary_values);
        }
#endif
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//set output values to given_function at mapped_locations
void DofHandler::setValues(const Mapping &mapping,
const FunctionSystem *given_function,Vector &values) const {
  map<int,FiniteElementDataBase*> fedb_map;
  map<int,Quadrature<SPACEDIM>*> quad_map;
  selected_fe->equallySpacedLatticePointDataBase(fedb_map,quad_map);
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *level=tess->presentLevel();
    const SHAPE_CLASS *element=level->getShape();
    map<int,FiniteElementDataBase*>::const_iterator fedb_iterator=
      fedb_map.find(element->numberVertices());
    CHECK_TEST(fedb_iterator!=fedb_map.end());
    FiniteElementDataBase *fedb=fedb_iterator->second;
    const NumPtr<Point<SPACEDIM> > &lattice_points=
      fedb->getQuadrature()->points();
    NumPtr<Point<SPACEDIM> > mapped_locations(
      lattice_points.getNumber());
    mapping.elementPoints(element,*tess,level,lattice_points,
      mapped_locations);
    NumPtr<int> element_dofs;
    getDofIndices(element,level,element_dofs);
    selected_fe->setValues(element,element_dofs,fedb,mapped_locations,
      given_function,values);
  }
  for (map<int,FiniteElementDataBase*>::iterator fedb_iterator=
  fedb_map.begin();fedb_iterator!=fedb_map.end();fedb_iterator++) {
    delete fedb_iterator->second; fedb_iterator->second=0;
  }
  for (map<int,Quadrature<SPACEDIM>*>::iterator quad_iterator=
  quad_map.begin();quad_iterator!=quad_map.end();quad_iterator++) {
    delete quad_iterator->second; quad_iterator->second=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::interpolateBoundaryValues(const Mapping &mapping,
map<int,REAL> &boundary_values,map<int,Quadrature<SPACEDIM-1> > &dqmap,
REAL time) const {
  int n_components=selected_fe->nComponents();
  bool fe_is_system=(n_components!=1);
  int dpv=selected_fe->dofsPerVertex();
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape()) {
    LEVEL_CLASS *level=tess->presentLevel();
    const SHAPE_CLASS *element=level->getShape();
    NumPtr<int> vertex_dof_indices(
      element->numberVertices()*selected_fe->dofsPerVertex());
    vertexDofIndices(element,vertex_dof_indices);
    for (int face_no=0;
    face_no<element->numberCodimension1Shapes();face_no++) {
      if (level->neighborLevel(face_no)==0) {
        int face_index=element->codimension1ShapeIndex(face_no);
        CODIMENSION1_LEVEL_CLASS *codimension1_level=
          static_cast<CODIMENSION1_LEVEL_CLASS*>(level);
#if (SPACEDIM==1)
        const BoundaryCondition *bc=
          codimension1_level->getBoundaryCondition(face_no);
        if (bc!=0) {
          Vector function_values(n_components);
          const Point<SPACEDIM> &boundary_point=
            tess->getVertices()[element->codimension1ShapeIndex(face_no)];
          FunctionSystem *fs=bc->getFunctionSystem();
          TimeDependentFunctionSystem *tdfs=
            dynamic_cast<TimeDependentFunctionSystem*>(fs);
          if (tdfs) {
            tdfs->value(boundary_point,time,function_values);
          } else { 
            CHECK_POINTER(fs);
            fs->value(boundary_point,function_values);
          }
          NumPtr<int> vdi(vertex_dof_indices,face_no*dpv,dpv);
          for (int i=0;i<dpv;i++) {
            if (bc->isEssential(i)) {
              boundary_values[vdi[i]]=function_values[i];
            }
          }
        }
#else
        const BoundaryCondition *bc=
          codimension1_level->getBoundaryCondition(face_index);
        if (bc!=0) {
          const CODIMENSION1_SHAPE_CLASS *face_element=
            codimension1_level->getShape(face_index);
//        dqmap = user specified quadrature rule on Dirichlet boundary
          map<int,Quadrature<SPACEDIM-1> >::const_iterator
            dqm_iterator=dqmap.find(face_element->numberVertices());
          CHECK_TEST(dqm_iterator!=dqmap.end());
          const NumPtr<Point<SPACEDIM-1> > &dqp=
            dqm_iterator->second.points();
          int dpf=selected_fe->dofsPerFace(element,face_no);
          CHECK_SAME(dpf,dqp.getNumber());
//        project Dirichlet quadrature rule to canonical boundary
          NumPtr<Point<SPACEDIM> > canonical_locations(dpf);
          element->projectToFace(face_no,dqp,canonical_locations);
//        map canonical boundary points to element boundary
          NumPtr<Point<SPACEDIM> > dof_locations(dpf);
          mapping.elementPoints(element,*tess,level,canonical_locations,
            dof_locations);
          NumPtr<int> face_dofs;
          getDofIndices(face_element,codimension1_level,face_dofs);
          NumPtr<int> element_dofs;
          getDofIndices(element,level,element_dofs);
          selected_fe->setBoundaryValues(element,face_element,face_no,
            element_dofs,face_dofs,canonical_locations,dof_locations,
            bc,boundary_values,time);
        }
#endif
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//eliminates boundary dof from matrix and right-hand-side
void DofHandler::applyBoundaryValues(
const map<int,REAL> &boundary_values,SparseMatrix &matrix,
Vector &right_hand_side,bool preserve_symmetry) const {
  int n_rows=matrix.size(0);
  CHECK_SAME(matrix.size(1),n_rows);
  CHECK_SAME(n_rows,right_hand_side.size());
  if (boundary_values.size()==0) return;
  int n_dofs=n_rows;
  const SparsityPattern &sparsity=matrix.getSparsityPattern();
  const NumPtr<int> &sparsity_colnums=
    sparsity.getColumnNumbers();
  CHECK_TEST(sparsity.storesDiagonalFirst());

  REAL first_nonzero_diagonal_entry=1.;
  for (int i=0;i<n_dofs;i++) {
    REAL diag=matrix.diagElement(i);
    if (abs(diag)>0.) {
      first_nonzero_diagonal_entry=diag;
      break;
    }
  }
  for (map<int,REAL>::const_iterator dof=
  boundary_values.begin();dof!=boundary_values.end();dof++) {
    int dof_number=dof->first;
    CHECK_TEST(dof_number<n_dofs);
    int last=sparsity.getRowstartIndex(dof_number+1);
//  assumes diagonal stored first:
    for (int j=sparsity.getRowstartIndex(dof_number)+1;j<last;j++) {
      matrix.globalEntry(j)=0.;
    }
    REAL new_rhs;
    REAL diag=matrix.diagElement(dof_number);
    REAL boundary_value=dof->second;
    if (abs(diag)>0.0) {
      new_rhs=right_hand_side[dof_number]=boundary_value*diag;
    } else {
      matrix.set(dof_number,dof_number,first_nonzero_diagonal_entry);
      new_rhs=right_hand_side[dof_number]
             =boundary_value*first_nonzero_diagonal_entry;
    }
    if (preserve_symmetry) {
      for (int j=sparsity.getRowstartIndex(dof_number)+1;j<last;j++) {
        int row=sparsity_colnums[j];
        int start_next_row=sparsity.getRowstartIndex(row+1);
        int p=start_next_row;
//      assumes diagonal stored first:
        for (int c=sparsity.getRowstartIndex(row)+1;c<start_next_row;c++) {
          if (sparsity_colnums[c]>=dof_number) {
            p=c; break;
          }
        }
        CHECK_TEST((sparsity_colnums[p]==dof_number) &&
         (p!=sparsity.getRowstartIndex(row+1)));
        int global_entry=p-sparsity.getRowstartIndex(0);
        right_hand_side[row]-=
          matrix.globalEntry(global_entry)*boundary_value;
        matrix.globalEntry(global_entry)=0.;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//sets solution at boundary dofs
void DofHandler::applyBoundaryValues(
const map<int,REAL> &boundary_values,Vector &solution) const {
  if (boundary_values.size()==0) return;
  for (map<int,REAL>::const_iterator dof=
  boundary_values.begin();dof!=boundary_values.end();dof++) {
    solution[dof->first]=dof->second;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL DofHandler::integrateDifference(const Mapping &mapping,
const Vector &fe_function,const FunctionSystem &exact_solution,
const map<int,Quadrature<SPACEDIM> > &qmap,NORM_TYPE norm,
const FunctionSystem *weight,REAL exponent_1) {
  int n_components=selected_fe->nComponents();
  bool fe_is_system=(n_components!=1);
  if (weight) {
    int wnc=weight->nComponents();
    CHECK_TEST(wnc==1 || wnc==n_components);
  }

  REAL exponent=exponent_1;
  switch (norm) {
    case L2_NORM:
    case H1_SEMINORM:
    case H1_NORM:
      exponent=2.;
      break;
    case L1_NORM:
      exponent=1.;
      break;
    default:
      break;
  }
  
  UPDATE_FLAGS update_flags=UPDATE_Q_POINTS | UPDATE_JxW_VALUES;
  switch (norm) {
    case H1_SEMINORM:
    case W1P_SEMINORM:
      update_flags|=UPDATE_GRADIENTS;
      break;
    case H1_NORM:
    case W1P_NORM:
      update_flags|=UPDATE_GRADIENTS;
//    no break!
    default:
      update_flags|=UPDATE_VALUES;
      break;
  }  

  FEValues fe_values(mapping,*selected_fe,qmap,update_flags);
  int index=0;
  REAL diff_sum=0.;
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape(),index++) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    fe_values.mapShapes(*tess,element,present_level);
    const NumPtr<Point<SPACEDIM> > &q_points=
      fe_values.getQuadraturePoints(element);
    int n_q_points=q_points.getNumber();

    NumPtr<REAL> tmp_values(n_q_points);
    NumPtr<Tensor<1,SPACEDIM> > tmp_gradients(n_q_points);
    NumPtr<REAL> weight_values(n_q_points);
    NumPtr<Vector> weight_vectors(n_q_points); 
    NumPtr<Vector> psi_values(n_q_points);
    NumPtr<NumPtr<Tensor<1,SPACEDIM> > > psi_grads(n_q_points);
    for (int q=0;q<n_q_points;q++) {
      weight_vectors[q].reinit(n_components,true);
      psi_values[q].reinit(n_components,true);
      psi_grads[q].allocate(n_components);
    }
    if (weight) {
      if (n_components>1) {
        weight->value(weight_vectors,q_points);
      } else {
        weight->value(weight_values,q_points,0);
        for (int q=0;q<n_q_points;q++) {
          weight_vectors[q]=weight_values[q];
        }
      }
    } else {
      for (int q=0;q<n_q_points;q++) {
        weight_vectors[q]=1.;
      }
    }
     
    if (update_flags & UPDATE_VALUES) {
      if (fe_is_system) {
        exact_solution.value(psi_values,q_points);
      } else {
        exact_solution.value(tmp_values,q_points,0);
        for (int q=0;q<n_q_points;q++) {
          psi_values[q][0]=tmp_values[q];
        }
      }
      NumPtr<Vector> function_values;
      getFunctionValues(fe_values,element,present_level,fe_function,
        function_values);
      for (int q=0; q<n_q_points; ++q) {
        psi_values[q]-=function_values[q];
      }
    }
    if (update_flags & UPDATE_GRADIENTS) {
      if (fe_is_system) {
        exact_solution.gradient(psi_grads,q_points);
      } else {
        exact_solution.gradient(tmp_gradients,q_points,0);
        for (int q=0;q<n_q_points;q++) {
          psi_grads[q][0]=tmp_gradients[q];
        }
      }      
      NumPtr<NumPtr<Tensor<1,SPACEDIM> > > function_grads;
      getFunctionDerivatives(fe_values,element,present_level,fe_function,
        function_grads);
      for (int k=0;k<n_components;k++) {
        for (int q=0;q<n_q_points;++q) {
          psi_grads[q][k]-=function_grads[q][k];
        }
      }
    }
    REAL diff=0;
    const NumPtr<REAL> &jxw_values=fe_values.getJxWValues(element);
    switch (norm) {
      case MEAN_NORM:
        for (int q=0;q<n_q_points;q++) {
          REAL psi_scalar=0.;
          for (int k=0;k<n_components;k++) {
            psi_scalar+=psi_values[q][k]*weight_vectors[q][k];
          }
          diff+=psi_scalar*jxw_values[q];
        }
        break;
      case LP_NORM:
      case L1_NORM:
      case W1P_NORM:
        for (int q=0;q<n_q_points;q++) {
          REAL psi_scalar=0.;
          for (int k=0;k<n_components;k++) {
            psi_scalar+=pow(abs(psi_values[q][k]),exponent)
                          *weight_vectors[q][k];
          }
          diff+=psi_scalar*jxw_values[q];
        }
        break;
      case L2_NORM:
      case H1_NORM: {
        for (int q=0;q<n_q_points;q++) {
          REAL psi_scalar=0.;
          for (int k=0;k<n_components;k++) {
            psi_scalar+=psi_values[q][k]*psi_values[q][k]
                          *weight_vectors[q][k];
          }
          diff+=psi_scalar*jxw_values[q];
        }
        break;
      }
      case LINFTY_NORM: {
        for (int q=0;q<n_q_points;q++) {
          REAL psi_scalar=0.;
          for (int k=0;k<n_components;k++) {
            REAL newval=abs(psi_values[q][k])*weight_vectors[q][k];
            if (psi_scalar<newval) psi_scalar=newval;
          }
          diff=max(diff,psi_scalar);
        }
        break;
      }
      case H1_SEMINORM:
      case W1P_SEMINORM:
        break;
      default:
        OBSOLETE("not implemented");
        break;
    }
    switch (norm) {
      case MEAN_NORM:
      case LP_NORM:
      case L1_NORM:
      case L2_NORM:
        diff_sum+=diff;
        break;
      case LINFTY_NORM: {
        diff_sum=max(diff_sum,diff);
        break;
      }
      case W1P_SEMINORM:
      case W1P_NORM:
        for (int q=0;q<n_q_points;q++) {
          REAL psi_scalar=0.;
          for (int k=0;k<n_components;k++) {
            psi_scalar+=pow(psi_grads[q][k]*psi_grads[q][k],
              exponent*0.5)*weight_vectors[q][k];
          }
          diff+=psi_scalar*jxw_values[q];
        }
        diff_sum+=diff;
        break;
      case H1_SEMINORM:
      case H1_NORM:
        for (int q=0;q<n_q_points;q++) {
          REAL psi_scalar=0.;
          for (int k=0;k<n_components;k++) {
            psi_scalar+=(psi_grads[q][k]*psi_grads[q][k])
                          *weight_vectors[q][k];
          }
          diff+=psi_scalar*jxw_values[q];
        }
        diff_sum+=diff;
        break;
      default:
        break;
     }
  }
  switch (norm) {
    case LP_NORM:
    case L1_NORM:
    case W1P_SEMINORM:
    case W1P_NORM:
      diff_sum=pow(diff_sum,1./exponent);
      break;
    case L2_NORM:
    case H1_SEMINORM:
    case H1_NORM:
      diff_sum=sqrt(diff_sum);
      break;
    default:
      break;
  }
  return diff_sum;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// MEAN_NORM, L1_NORM, L2_NORM, LINFTY_NORM, LP_NORM measure value
// H1_SEMINORM, W1P_SEMINORM measure normal derivative
// H1_NORM, W1P_NORM measure value and normal derivative
REAL DofHandler::integrateDifferenceOnBoundary(const Mapping &mapping,
const Vector &fe_function,const FunctionSystem &exact_solution,
const map<int,Quadrature<SPACEDIM> > &qmap,
const map<int,Quadrature<SPACEDIM-1> > &bc_qmap,NORM_TYPE norm,
const FunctionSystem *weight,REAL exponent_1) {
  int n_components=selected_fe->nComponents();
  bool fe_is_system=(n_components!=1);
  if (weight) {
    int wnc=weight->nComponents();
    CHECK_TEST(wnc==1 || wnc==n_components);
  }

  REAL exponent=exponent_1;
  switch (norm) {
    case L2_NORM:
    case H1_SEMINORM:
    case H1_NORM:
      exponent=2.;
      break;
    case L1_NORM:
      exponent=1.;
      break;
    default:
      break;
  }
  
  UPDATE_FLAGS update_flags=UPDATE_Q_POINTS;
  switch (norm) {
    case H1_SEMINORM:
    case W1P_SEMINORM:
      update_flags|=UPDATE_GRADIENTS;
      break;
    case H1_NORM:
    case W1P_NORM:
      update_flags|=UPDATE_GRADIENTS;
//    no break!
    default:
      update_flags|=UPDATE_VALUES;
      break;
  }  

  FEValues fe_values(mapping,*selected_fe,qmap,
    UPDATE_TRANSFORMATION_VALUES | UPDATE_TRANSFORMATION_GRADIENTS);
  int index=0;
  REAL diff_sum=0.;
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape(),index++) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int nv=element->numberVertices();
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      if (!present_level->neighborLevel(f)) {
#if (SPACEDIM==1)
        NumPtr<Point<1> > p(1); p[0]=Interval::canonicalVertex(f);
        NumPtr<Point<1> > normals(1); normals[0][0]=(f==0 ? -1. : 1.);
        NumPtr<REAL> w(1); w[0]=1.;
        NumPtr<REAL> surface_measure(1); surface_measure[0]=1.;
        Quadrature<1> fq(p,w);
        map<int,Quadrature<1> > fqmap;
        fqmap[element->numberVertices()]=fq;
        FEValues fev(mapping,*selected_fe,fqmap,update_flags);
        fev.mapShapes(*tess,element,present_level);
        int n_q_points=1;
        NumPtr<Point<SPACEDIM> > q_points(1);
          q_points[0]=tess->getVertices()[element->vertexIndex(f)];
#else
        int face_element_index=element->codimension1ShapeIndex(f);
        CODIMENSION1_LEVEL_CLASS *present_codimension1_level=
          static_cast<CODIMENSION1_LEVEL_CLASS*>(present_level);
        CODIMENSION1_SHAPE_CLASS *face_element=
          present_codimension1_level->getShape(face_element_index);
        int fnv=face_element->numberVertices();
        const BoundaryCondition *bc=present_codimension1_level->
          getBoundaryCondition(face_element_index);
        const Boundary *boundary=bc->getBoundary();
        NumPtr<Point<SPACEDIM> > q_points;
        NumPtr<Point<SPACEDIM> > normals;
        NumPtr<REAL> surface_measure;
        map<int,Quadrature<SPACEDIM-1> >::const_iterator
          bcq_iterator=bc_qmap.find(fnv);
        CHECK_TEST(bcq_iterator!=bc_qmap.end());
        tess->getMappedMeasureOn(*face_element,present_level,
          bcq_iterator->second.points(),q_points,normals,
          surface_measure);

        fe_values.mapShapes(*tess,element,present_level);
        int n_q_points=q_points.getNumber();
        NumPtr<Point<SPACEDIM> > c_points(n_q_points);
        const MappingDataBase *md=fe_values.mappingData(element);
        for (int q=0;q<n_q_points;q++) {
          REAL min_dist=HUGE_VAL;
          for (int v=0;v<face_element->numberVertices();v++) {
            const Point<SPACEDIM> &p=
              tess->getVertices()[face_element->vertexIndex(v)];
            if (p.distance(q_points[q])<min_dist) c_points[q]=p;
          }
          mapping.canonicalPoint(element,*tess,md,q_points[q],c_points[q]);
        }

        CHECK_TEST(bcq_iterator!=bc_qmap.end());
        Quadrature<SPACEDIM> fq(c_points,bcq_iterator->second.weights());
        map<int,Quadrature<SPACEDIM> > qm;
        for (map<int,Quadrature<SPACEDIM> >::const_iterator
        q_iterator=qmap.begin();q_iterator!=qmap.end();q_iterator++) {
          Quadrature<SPACEDIM> &qmi=
            const_cast<Quadrature<SPACEDIM>&>(q_iterator->second);
          qm[q_iterator->first]=(q_iterator->first==nv ? fq : qmi);
        }
        FEValues fev(mapping,*selected_fe,qm,update_flags);
        fev.mapShapes(*tess,element,present_level);
#endif
        NumPtr<REAL> tmp_values(n_q_points);
        NumPtr<Tensor<1,SPACEDIM> > tmp_gradients(n_q_points);
        NumPtr<REAL> weight_values(n_q_points);
        NumPtr<Vector> weight_vectors(n_q_points); 
        NumPtr<Vector> psi_values(n_q_points);
        NumPtr<NumPtr<Tensor<1,SPACEDIM> > > psi_grads(n_q_points);
        for (int q=0;q<n_q_points;q++) {
          weight_vectors[q].reinit(n_components,true);
          psi_values[q].reinit(n_components,true);
          psi_grads[q].allocate(n_components);
        }
        if (weight) {
          if (n_components>1) {
            weight->value(weight_vectors,q_points);
          } else {
            weight->value(weight_values,q_points,0);
            for (int q=0;q<n_q_points;q++) {
              weight_vectors[q]=weight_values[q];
            }
          }
        } else {
          for (int q=0;q<n_q_points;q++) {
            weight_vectors[q]=1.;
          }
        }
        if (update_flags & UPDATE_VALUES) {
          if (fe_is_system) {
            exact_solution.value(psi_values,q_points);
          } else {
            exact_solution.value(tmp_values,q_points,0);
            for (int q=0;q<n_q_points;q++) {
              psi_values[q][0]=tmp_values[q];
            }
          }
          NumPtr<Vector> function_values;
          getFunctionValues(fev,element,present_level,
            fe_function,function_values);
          for (int q=0; q<n_q_points; ++q) {
            psi_values[q]-=function_values[q];
          }
        }
        if (update_flags & UPDATE_GRADIENTS) {
          if (fe_is_system) {
            exact_solution.gradient(psi_grads,q_points);
          } else {
            exact_solution.gradient(tmp_gradients,q_points,0);
            for (int q=0;q<n_q_points;q++) {
              psi_grads[q][0]=tmp_gradients[q];
            }
          }      
          NumPtr<NumPtr<Tensor<1,SPACEDIM> > > function_grads;
          getFunctionDerivatives(fev,element,present_level,
            fe_function,function_grads);
          for (int k=0;k<n_components;k++) {
            for (int q=0;q<n_q_points;++q) {
              psi_grads[q][k]-=function_grads[q][k];
            }
          }
        }
        REAL diff=0;
        const NumPtr<REAL> &fq_weights=fq.weights();
        switch (norm) {
          case MEAN_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                psi_scalar+=psi_values[q][k]*weight_vectors[q][k];
              }
              diff+=psi_scalar*surface_measure[q]*fq_weights[q];
            }
            break;
          case LP_NORM:
          case L1_NORM:
          case W1P_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                psi_scalar+=pow(abs(psi_values[q][k]),exponent)
                              *weight_vectors[q][k];
              }
              diff+=psi_scalar*surface_measure[q]*fq_weights[q];
            }
            break;
          case L2_NORM:
          case H1_NORM: {
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                psi_scalar+=psi_values[q][k]*psi_values[q][k]
                              *weight_vectors[q][k];
              }
              diff+=psi_scalar*surface_measure[q]*fq_weights[q];
            }
            break;
          }
          case LINFTY_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                REAL newval=abs(psi_values[q][k])*weight_vectors[q][k];
                if (psi_scalar<newval) psi_scalar=newval;
              }
              diff=max(diff,psi_scalar);
            }
            break;
          case H1_SEMINORM:
          case W1P_SEMINORM:
            break;
          default:
            OBSOLETE("not implemented");
            break;
        }
        switch (norm) {
          case MEAN_NORM:
          case LP_NORM:
          case L1_NORM:
          case L2_NORM: {
            diff_sum+=diff;
            break;
          }
          case LINFTY_NORM:
            diff_sum=max(diff_sum,diff);
          case W1P_SEMINORM:
          case W1P_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                REAL dot=abs(psi_grads[q][k].getValues()*normals[q]);
                psi_scalar+=pow(dot,exponent)*weight_vectors[q][k];
              }
              diff+=psi_scalar*surface_measure[q]*fq_weights[q];
            }
            diff_sum+=diff;
            break;
          case H1_SEMINORM:
          case H1_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                REAL dot=psi_grads[q][k].getValues()*normals[q];
                psi_scalar+=dot*dot*weight_vectors[q][k];
              }
              diff+=psi_scalar*surface_measure[q]*fq_weights[q];
            }
            diff_sum+=diff;
            break;
          default:
            break;
        }
      }
    }
  }
  switch (norm) {
    case LP_NORM:
    case L1_NORM:
    case W1P_SEMINORM:
    case W1P_NORM:
      diff_sum=pow(diff_sum,1./exponent);
      break;
    case L2_NORM:
    case H1_SEMINORM:
    case H1_NORM:
      diff_sum=sqrt(diff_sum);
      break;
    default:
      break;
  }
  return diff_sum;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// MEAN_NORM, L1_NORM, L2_NORM, LINFTY_NORM, LP_NORM measure value
// H1_SEMINORM, W1P_SEMINORM measure normal derivative
// H1_NORM, W1P_NORM measure value and normal derivative
REAL DofHandler::integrateDifferenceOnMappedBoundary(
const Mapping &mapping,const Vector &fe_function,
const FunctionSystem &exact_solution,
const map<int,Quadrature<SPACEDIM> > &qmap,
map<int,Quadrature<SPACEDIM-1> > &bc_qmap,NORM_TYPE norm,
const FunctionSystem *weight,REAL exponent_1) {
  int n_components=selected_fe->nComponents();
  bool fe_is_system=(n_components!=1);
  if (weight) {
    int wnc=weight->nComponents();
    CHECK_TEST(wnc==1 || wnc==n_components);
  }

  REAL exponent=exponent_1;
  switch (norm) {
    case L2_NORM:
    case H1_SEMINORM:
    case H1_NORM:
      exponent=2.;
      break;
    case L1_NORM:
      exponent=1.;
      break;
    default:
      break;
  }
  
  UPDATE_FLAGS update_flags=UPDATE_NORMAL_VECTORS | UPDATE_JxW_VALUES | 
    UPDATE_Q_POINTS;
  switch (norm) {
    case H1_SEMINORM:
    case W1P_SEMINORM:
      update_flags |= UPDATE_GRADIENTS | UPDATE_MAPPING_JACOBIAN_INVERSE;
      break;
    case H1_NORM:
    case W1P_NORM:
      update_flags |= UPDATE_GRADIENTS | UPDATE_MAPPING_JACOBIAN_INVERSE;
//    no break!
    default:
      update_flags|=UPDATE_VALUES;
      break;
  }  

#if (SPACEDIM>=2)
  FEFaceValues bc_fe_values(mapping,*selected_fe,bc_qmap,update_flags);
#endif
  int index=0;
  REAL diff_sum=0.;
  for (bool doit=tess->firstActiveShape();doit;
  doit=tess->nextActiveShape(),index++) {
    LEVEL_CLASS *present_level=tess->presentLevel();
    const SHAPE_CLASS *element=present_level->getShape();
    int nv=element->numberVertices();
    int offset=0;
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      if (!present_level->neighborLevel(f)) {
#if (SPACEDIM==1)
        NumPtr<Point<1> > p(1); p[0]=Interval::canonicalVertex(f);
        NumPtr<Point<1> > normals(1); normals[0][0]=(f==0 ? -1. : 1.);
        NumPtr<REAL> w(1); w[0]=1.;
        NumPtr<REAL> surface_measure(1); surface_measure[0]=1.;
        Quadrature<1> fq(p,w);
        map<int,Quadrature<1> > fqmap;
        fqmap[element->numberVertices()]=fq;
        FEValues face_fe_values(mapping,*selected_fe,fqmap,update_flags);
        face_fe_values.mapShapes(*tess,element,present_level);
        int n_q_points=1;
        NumPtr<Point<SPACEDIM> > q_points(1);
          q_points[0]=tess->getVertices()[element->vertexIndex(f)];
#else
        bc_fe_values.mapShapes(*tess,element,present_level,f);
        const NumPtr<Point<SPACEDIM> > &q_points=
          bc_fe_values.getQuadraturePoints(element,f);
        int n_q_points=q_points.getNumber();
        const NumPtr<REAL> &surface_measure=
          bc_fe_values.getJxWValues(element,f);
        const NumPtr<Point<SPACEDIM> > &normals=
          bc_fe_values.getNormalVectors(element,f);
#endif
        NumPtr<REAL> tmp_values(n_q_points);
        NumPtr<Tensor<1,SPACEDIM> > tmp_gradients(n_q_points);
        NumPtr<REAL> weight_values(n_q_points);
        NumPtr<Vector> weight_vectors(n_q_points); 
        NumPtr<Vector> psi_values(n_q_points);
        NumPtr<NumPtr<Tensor<1,SPACEDIM> > > psi_grads(n_q_points);
        for (int q=0;q<n_q_points;q++) {
          weight_vectors[q].reinit(n_components,true);
          psi_values[q].reinit(n_components,true);
          psi_grads[q].allocate(n_components);
        }
        if (weight) {
          if (n_components>1) {
            weight->value(weight_vectors,q_points);
          } else {
            weight->value(weight_values,q_points,0);
            for (int q=0;q<n_q_points;q++) {
              weight_vectors[q]=weight_values[q];
            }
          }
        } else {
          for (int q=0;q<n_q_points;q++) {
            weight_vectors[q]=1.;
          }
        }
        if (update_flags & UPDATE_VALUES) {
          if (fe_is_system) {
            exact_solution.value(psi_values,q_points);
          } else {
            exact_solution.value(tmp_values,q_points,0);
            for (int q=0;q<n_q_points;q++) {
              psi_values[q][0]=tmp_values[q];
            }
          }
          NumPtr<Vector> function_values;
#if (SPACEDIM==1)
          getFunctionValues(face_fe_values,element,present_level,
            fe_function,function_values);
#else
          getFunctionValues(bc_fe_values,element,f,present_level,
            fe_function,function_values);
#endif
          for (int q=0; q<n_q_points; ++q) {
            psi_values[q]-=function_values[q];
          }
        }
        if (update_flags & UPDATE_GRADIENTS) {
          if (fe_is_system) {
            exact_solution.gradient(psi_grads,q_points);
          } else {
            exact_solution.gradient(tmp_gradients,q_points,0);
            for (int q=0;q<n_q_points;q++) {
              psi_grads[q][0]=tmp_gradients[q];
            }
          }      
          NumPtr<NumPtr<Tensor<1,SPACEDIM> > > function_grads;
#if (SPACEDIM==1)
          getFunctionDerivatives(face_fe_values,element,present_level,
            fe_function,function_grads);
#else
          getFunctionDerivatives(bc_fe_values,element,f,present_level,
            fe_function,function_grads);
#endif
          for (int k=0;k<n_components;k++) {
            for (int q=0;q<n_q_points;++q) {
              psi_grads[q][k]-=function_grads[q][k];
            }
          }
        }
        REAL diff=0;
        switch (norm) {
          case MEAN_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                psi_scalar+=psi_values[q][k]*weight_vectors[q][k];
              }
              diff+=psi_scalar*surface_measure[q];
            }
            break;
          case LP_NORM:
          case L1_NORM:
          case W1P_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                psi_scalar+=pow(abs(psi_values[q][k]),exponent)
                              *weight_vectors[q][k];
              }
              diff+=psi_scalar*surface_measure[q];
            }
            break;
          case L2_NORM:
          case H1_NORM: {
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                psi_scalar+=psi_values[q][k]*psi_values[q][k]
                              *weight_vectors[q][k];
              }
              diff+=psi_scalar*surface_measure[q];
            }
            break;
          }
          case LINFTY_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                REAL newval=abs(psi_values[q][k])*weight_vectors[q][k];
                if (psi_scalar<newval) psi_scalar=newval;
              }
              diff=max(diff,psi_scalar);
            }
            break;
          case H1_SEMINORM:
          case W1P_SEMINORM:
            break;
          default:
            OBSOLETE("not implemented");
            break;
        }
        switch (norm) {
          case MEAN_NORM:
          case LP_NORM:
          case L1_NORM:
          case L2_NORM: {
            diff_sum+=diff;
            break;
          }
          case LINFTY_NORM:
            diff_sum=max(diff_sum,diff);
          case W1P_SEMINORM:
          case W1P_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                REAL dot=abs(psi_grads[q][k].getValues()*normals[q]);
                psi_scalar+=pow(dot,exponent)*weight_vectors[q][k];
              }
              diff+=psi_scalar*surface_measure[q];
            }
            diff_sum+=diff;
            break;
          case H1_SEMINORM:
          case H1_NORM:
            for (int q=0;q<n_q_points;q++) {
              REAL psi_scalar=0.;
              for (int k=0;k<n_components;k++) {
                REAL dot=psi_grads[q][k].getValues()*normals[q];
                psi_scalar+=dot*dot*weight_vectors[q][k];
              }
              diff+=psi_scalar*surface_measure[q];
            }
            diff_sum+=diff;
            break;
          default:
            break;
        }
      }
#if (SPACEDIM>=2)
      offset+=bc_fe_values.getFaceQuadrature(element,f).numberPoints();
#endif
    }
  }
  switch (norm) {
    case LP_NORM:
    case L1_NORM:
    case W1P_SEMINORM:
    case W1P_NORM:
      diff_sum=pow(diff_sum,1./exponent);
      break;
    case L2_NORM:
    case H1_SEMINORM:
    case H1_NORM:
      diff_sum=sqrt(diff_sum);
      break;
    default:
      break;
  }
  return diff_sum;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::printOn(ostream &os) const {
  os << "DofHandler:" << endl;
  os << "\ttess = " << tess
     << "\n\tselected_fe = " << selected_fe
     << "\n\tused_dofs = " << used_dofs
     << "\n\tlevels.getNumber = " << levels.getNumber()
     << "\n\tvertex_dofs.getNumber = " << vertex_dofs.getNumber()
     << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void DofHandler::printDofs() const {
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel()) {
    for (bool doit=tess->firstShape(level);
    doit && tess->presentLevel()==level;doit=tess->nextShapeInLevel()) {
      const SHAPE_CLASS *element=level->getShape();
      NumPtr<int> dof_indices;
      getDofIndices(element,level,dof_indices);
      cout << "\tdof_indices[" << element->elementDataIndex() << "] = ";
      for (int d=0;d<dof_indices.getNumber();d++) {
        cout << dof_indices[d] << " ";
      }
      cout << endl;
    }
  }
#if (SPACEDIM>1)
  for (CODIMENSION1_LEVEL_CLASS *level=
  static_cast<CODIMENSION1_LEVEL_CLASS*>(tess->coarsestLevel());
  level;level=level->finerLevel()) {
    for (bool doit=tess->firstCodimension1Shape(level);
    doit && tess->presentCodimension1Level()==level;
    doit=tess->nextCodimension1ShapeInLevel()) {
      const CODIMENSION1_SHAPE_CLASS *element=level->getShape();
      NumPtr<int> dof_indices;
      getDofIndices(element,level,dof_indices);
      cout << "\tdof_indices[" << element->elementDataIndex() << "] = ";
      for (int d=0;d<dof_indices.getNumber();d++) {
        cout << dof_indices[d] << " ";
      }
      cout << endl;
    }
  }
#endif
#if (SPACEDIM==3)
  for (Level1 *level=static_cast<Level1*>(tess->coarsestLevel());
  level;level=level->finerLevel()) {
    for (bool doit=tess->firstCodimension2Shape(level);
    doit && tess->presentCodimension2Level()==level;
    doit=tess->nextCodimension2ShapeInLevel()) {
      const Shape1 *element=level->getShape();
      NumPtr<int> dof_indices;
      getDofIndices(element,level,dof_indices);
      cout << "\tdof_indices[" << element->elementDataIndex() << "] = ";
      for (int d=0;d<dof_indices.getNumber();d++) {
        cout << dof_indices[d] << " ";
      }
      cout << endl;
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template void DofHandler::getFunctionValues<REAL>(const FEValues&,
  const SHAPE_CLASS*,const LEVEL_CLASS*,const Vector&,NumPtr<REAL>&) 
  const;
template void DofHandler::getFunctionValues<Vector>(const FEValues&,
  const SHAPE_CLASS*,const LEVEL_CLASS*,const Vector&,NumPtr<Vector>&) 
  const;
#if (SPACEDIM>1)
template void DofHandler::getFunctionValues<REAL>(const FEFaceValues&,
  const SHAPE_CLASS*,int,const LEVEL_CLASS*,const Vector&,NumPtr<REAL>&) 
  const;
template void DofHandler::getFunctionValues<Vector>(const FEFaceValues&,
  const SHAPE_CLASS*,int,const LEVEL_CLASS*,const Vector&,NumPtr<Vector>&) 
  const;
#endif

typedef Tensor<1,SPACEDIM> SCALAR_GRAD_TYPE;
typedef NumPtr<Tensor<1,SPACEDIM> > VECTOR_GRAD_TYPE;
template void DofHandler::getFunctionDerivatives<SCALAR_GRAD_TYPE>(
  const FEValues&,const SHAPE_CLASS*,const LEVEL_CLASS*,const Vector&,
  NumPtr<SCALAR_GRAD_TYPE>&) const;
template void DofHandler::getFunctionDerivatives<VECTOR_GRAD_TYPE>(
  const FEValues&,const SHAPE_CLASS*,const LEVEL_CLASS*,const Vector&,
  NumPtr<VECTOR_GRAD_TYPE>&) const;
#if (SPACEDIM>1)
template void DofHandler::getFunctionDerivatives<SCALAR_GRAD_TYPE>(
  const FEFaceValues&,const SHAPE_CLASS*,int,const LEVEL_CLASS*,
  const Vector&,NumPtr<SCALAR_GRAD_TYPE>&) const;
template void DofHandler::getFunctionDerivatives<VECTOR_GRAD_TYPE>(
  const FEFaceValues&,const SHAPE_CLASS*,int,const LEVEL_CLASS*,
  const Vector&,NumPtr<VECTOR_GRAD_TYPE>&) const;
#endif

typedef Tensor<2,SPACEDIM> SCALAR_GRAD_GRAD_TYPE;
typedef NumPtr<Tensor<2,SPACEDIM> > VECTOR_GRAD_GRAD_TYPE;
template void DofHandler::getFunction2ndDerivatives<SCALAR_GRAD_GRAD_TYPE>(
  const FEValues&,const SHAPE_CLASS*,const LEVEL_CLASS*,const Vector&,
  NumPtr<SCALAR_GRAD_GRAD_TYPE>&) const;
template void DofHandler::getFunction2ndDerivatives<VECTOR_GRAD_GRAD_TYPE>(
  const FEValues&,const SHAPE_CLASS*,const LEVEL_CLASS*,const Vector&,
  NumPtr<VECTOR_GRAD_GRAD_TYPE>&) const;
#if (SPACEDIM>1)
template void DofHandler::getFunction2ndDerivatives<SCALAR_GRAD_GRAD_TYPE>(
  const FEFaceValues&,const SHAPE_CLASS*,int,const LEVEL_CLASS*,
  const Vector&,NumPtr<SCALAR_GRAD_GRAD_TYPE>&) const;
template void DofHandler::getFunction2ndDerivatives<VECTOR_GRAD_GRAD_TYPE>(
  const FEFaceValues&,const SHAPE_CLASS*,int,const LEVEL_CLASS*,
  const Vector&,NumPtr<VECTOR_GRAD_GRAD_TYPE>&) const;
#endif

#include "NumPtr.C"
INSTANTIATE_NUMPTR(set<int>);
template class NumPtrBase<set<pair<const SHAPE_CLASS*,LEVEL_CLASS*> >,
  int>;
template class NumPtr<set<pair<const SHAPE_CLASS*,LEVEL_CLASS*> > >;
template class NumPtrBase<DOF_LEVEL_CLASS*,int>;
template class NumPtr<DOF_LEVEL_CLASS*>;
template class NumPtrWithValue<int>;
template class NumPtrBase<NumPtr<Tensor<1,SPACEDIM> >,int>;
template class NumPtr<NumPtr<Tensor<1,SPACEDIM> > >;

template class _Rb_tree<
  int, 
  pair<int const, NumPtr<REAL> >, 
  _Select1st<pair<int const, NumPtr<REAL> > >, 
  less<int>, 
  allocator<pair<int const, NumPtr<REAL> > > 
>;

template class _Rb_tree<
  int, 
  int, 
  _Identity<int>, 
  less<int>, 
  allocator<int> 
>;

template class _Rb_tree<
  pair<SHAPE_CLASS const*, LEVEL_CLASS*>, 
  pair<SHAPE_CLASS const*, LEVEL_CLASS*>, 
  _Identity<pair<SHAPE_CLASS const*, LEVEL_CLASS*> >, 
  less<pair<SHAPE_CLASS const*, LEVEL_CLASS*> >, 
  allocator<pair<SHAPE_CLASS const*, LEVEL_CLASS*> > 
>;

template class _Rb_tree<
  int, 
  pair<int const, NumPtr<Point<SPACEDIM-1> > >,
  _Select1st<pair<int const, NumPtr<Point<SPACEDIM-1> > > >,
  less<int>, 
  allocator<pair<int const, NumPtr<Point<SPACEDIM-1> > > > 
>;

template class _Rb_tree<
  int,
  pair<int const, double>,
  _Select1st<pair<int const, double> >,
  less<int>,
  allocator<pair<int const, double> >
>;
