#ifndef DOF_LEVEL_H
#define DOF_LEVEL_H "$Header:$"
//----------------------------  dof_levels.h  ---------------------------
//   $Id: dof_levels.h,v 1.19 2003/01/08 17:57:50 wolf Exp $
//   Version: $Name: Version-4-0-0 $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//----------------------------  dof_levels.h  ---------------------------
//
//modified from deal.II/deal.II/include/dofs/dof_levels.h
//  by John Trangenstein, August 2009

#include "NumPtr.H"
#include "Shape.H"

#define DOF_LEVEL_CLASS D_CHOOSE(DofLevel1,DofLevel2,DofLevel3)
#define CODIMENSION1_DOF_LEVEL_CLASS D_CHOOSE(void,DofLevel1,DofLevel2)

static const int invalid_dof_index=INT_MAX;
class DofLevel1 {
  protected:
    NumPtr<int> dofs1;
  public:
    DofLevel1(int n1) : dofs1(n1) {
      dofs1.initialize(invalid_dof_index);
    }
    virtual ~DofLevel1() {}
    int numberDofs() const { return dofs1.getNumber(); }
    void getDofs(const Shape1 *e,NumPtr<int> &indices) const { 
      indices.copyFrom(dofs1+e->dofIndex());
    }
    void setDofs(const NumPtr<int>&);
    void setDofs(const Shape1*,int,int&);
    virtual void printOn(ostream&) const;
};

#if (SPACEDIM>1)
class DofLevel2 : public DofLevel1 {
  protected:
    NumPtr<int> dofs2;
  public:
    DofLevel2(int n1,int n2) : DofLevel1(n1),dofs2(n2) {
      dofs2.initialize(invalid_dof_index);
    }
    virtual ~DofLevel2() {}
    int numberDofs() const { return dofs2.getNumber(); }
    void getDofs(const Shape2 *e,NumPtr<int> &indices) const { 
      indices.copyFrom(dofs2+e->dofIndex());
    }
    void setDofs(const NumPtr<int>&);
    void setDofs(const Shape2*,int,int&);
    virtual void printOn(ostream&) const;
};
#endif

#if (SPACEDIM==3)
class DofLevel3 : public DofLevel2 {
  protected:
    NumPtr<int> dofs3;
  public:
    DofLevel3(int n1,int n2,int n3) : DofLevel2(n1,n2),dofs3(n3) {
      dofs3.initialize(invalid_dof_index);
    }
    virtual ~DofLevel3() {}
    int numberDofs() const { return dofs3.getNumber(); }
    void getDofs(const Shape3 *e,NumPtr<int> &indices) const { 
      indices.copyFrom(dofs3+e->dofIndex());
    }
    void setDofs(const NumPtr<int>&);
    void setDofs(const Shape3*,int,int&);
    virtual void printOn(ostream&) const;
};
#endif

#endif
