// "$Header:$"
//----------------------------  grid_generator.cc  ------------------------
//  $Id: grid_generator.cc,v 1.48 2003/04/07 15:41:14 wolf Exp $
//  Version: $Name: Version-4-0-0 $
//
//  Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
//
//----------------------------  grid_generator.cc  ------------------------
//modified from deal.II/deal.II/source/grid/grid_generator.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <GenerateTess.H>
#include <cmath>
#include <PermutationGenerator.H>
#include <Tracer.H>
#include <Tessellation.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BoundaryData* generateParallelpiped(Tessellation &tess,
const Point<SPACEDIM> &low,const Point<SPACEDIM> edge[SPACEDIM],
const int number_elements[SPACEDIM],bool set_face_ids,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  Point<SPACEDIM> dx[SPACEDIM];
  for (int i=0;i<SPACEDIM;++i) {
    CHECK_POSITIVE(number_elements[i]);
    dx[i]=edge[i]/static_cast<REAL>(number_elements[i]);
  }
#if (SPACEDIM==1)
  NumPtr<Point<1> > vertices(number_elements[0]+1);
  Point<1> x=low;
  for (int v=0;v<=number_elements[0];v++,x+=dx[0]) {
    vertices[v]=x;
  }
  NumPtr<CellData> cells(number_elements[0]);
  for (int c=0;c<number_elements[0];c++) {
    CellData &cell=cells[c];
    cell.allocate(2);
    cell.vertex(0)=c;
    cell.vertex(1)=c+1;
    cell.setMaterialModel(0);
  }
#endif
#if (SPACEDIM==2)
  NumPtr<Point<2> > 
    vertices((number_elements[0]+1)*(number_elements[1]+1));
  Point<2> xj=low;
  int v=0;
  for (int j=0;j<=number_elements[1];j++,xj+=dx[1]) {
    Point<2> x=xj;
    for (int i=0;i<=number_elements[0];i++,v++,x+=dx[0]) {
      vertices[v]=x;
    }
  }
  NumPtr<CellData> 
    cells(number_elements[0]*number_elements[1]);
  int c=0;
  for (int j=0;j<number_elements[1];j++) {
    for (int i=0;i<number_elements[0];i++,c++) {
      CellData &cell=cells[c];
      cell.allocate(4);
      cell.vertex(0)=c+j;
      cell.vertex(1)=cell.vertex(0)+1;
      cell.vertex(2)=cell.vertex(1)+number_elements[0];
      cell.vertex(3)=cell.vertex(2)+1;
      cell.setMaterialModel(0);
    }
  }
#endif
#if (SPACEDIM==3)
  NumPtr<Point<3> > vertices((number_elements[0]+1)
                        *(number_elements[1]+1)*(number_elements[2]+1));
  Point<3> xk=low;
  int v=0;
  for (int k=0;k<=number_elements[2];k++,xk+=dx[2]) {
    Point<3> xj=xk;
    for (int j=0;j<=number_elements[1];j++,xj+=dx[1]) {
      Point<3> x=xj;
      for (int i=0;i<=number_elements[0];i++,v++,x+=dx[0]) {
        vertices[v]=x;
      }
    }
  }
  NumPtr<CellData> cells(number_elements[0]
                                *number_elements[1]*number_elements[2]);
  int n0=number_elements[0]+1;
  int n01=n0*(number_elements[1]+1);
  int c=0;
  int nk=0;
  for (int k=0;k<number_elements[2];k++,nk+=n01) {
    int njk=nk;
    for (int j=0;j<number_elements[1];j++,njk+=n0) {
      for (int i=0;i<number_elements[0];i++,c++) {
        CellData &cell=cells[c];
        cell.allocate(8);
        cell.vertex(0)=i+njk;
        cell.vertex(1)=cell.vertex(0)+1;
        cell.vertex(2)=cell.vertex(1)+number_elements[0];
        cell.vertex(3)=cell.vertex(2)+1;

        cell.vertex(4)=cell.vertex(0)+n01;
        cell.vertex(5)=cell.vertex(4)+1;
        cell.vertex(6)=cell.vertex(5)+number_elements[0];
        cell.vertex(7)=cell.vertex(6)+1;
        cell.setMaterialModel(0);
      }
    }
  }
#endif
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(2*SPACEDIM);
    boundary_condition.allocate(2*SPACEDIM);
#if (SPACEDIM==1)
    Point<1> center(low+edge[0]*0.5);
#elif (SPACEDIM==2)
    Point<2> center(low+(edge[0]+edge[1])*0.5);
#else
    Point<3> center(low+(edge[0]+edge[1]+edge[2])*0.5);
#endif
#if (SPACEDIM==1)
    boundary[0]=OPERATOR_NEW PlaneBoundary(center,low);
    boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0]);
    boundary_condition[0]=OPERATOR_NEW BoundaryCondition(boundary[0]);
    boundary_condition[1]=OPERATOR_NEW BoundaryCondition(boundary[1]);
    for (int f=0;f<2;f++) {
      tess.coarsestLevel()->
        Level0::setBoundaryCondition(f,boundary_condition[f]);
    }
#else
# if (SPACEDIM==2)
    boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1]);
    boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
      low+edge[0]+edge[1]);
    boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0]);
    boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
      low+edge[0]+edge[1]);
# else
    boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1],
      low+edge[2]);
    boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
      low+edge[0]+edge[1],low+edge[0]+edge[2]);
    boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[2],
      low+edge[0]);
    boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
      low+edge[1]+edge[2],low+edge[1]+edge[0]);
    boundary[4]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0],
      low+edge[1]);
    boundary[5]=OPERATOR_NEW PlaneBoundary(center,low+edge[2],
      low+edge[2]+edge[0],low+edge[2]+edge[1]);
# endif
    for (int i=0;i<2*SPACEDIM;i++) {
      boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
    }
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        CODIMENSION1_SHAPE_CLASS *face_element=
          codimension1_level->getShape();
        Point<SPACEDIM> center=face_element->center(vertices);
        bool found=false;
        for (int i=0;i<boundary.getNumber();i++) {
          if (abs(boundary[i]->constraint(center))<1.e-5) {
            codimension1_level->
              setBoundaryCondition(boundary_condition[i]);
            found=true;
            break;
          }
        }
        CHECK_TEST(found);
      }
    }
    tess.synchronize(boundary_data);
# ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
#   if (SPACEDIM==3)
    tess.checkLineBoundaryCondition(boundary_condition);
#   endif
# endif
#endif
  }
  return boundary_data;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
BoundaryData* generateParallelpipedTriangles(Tessellation &tess,
const Point<2> &low,const Point<2> edge[2],
const int number_elements[2],bool set_face_ids,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  Point<2> dx[2];
  for (int i=0;i<2;++i) {
    CHECK_POSITIVE(number_elements[i]);
    dx[i]=edge[i]/static_cast<REAL>(number_elements[i]);
  }
  NumPtr<Point<2> > vertices(number_elements[0]*number_elements[1]
                        +(number_elements[0]+1)*(number_elements[1]+1));
  Point<2> xj=low;
  int v=0;
  for (int j=0;j<=number_elements[1];j++,xj+=dx[1]) {
    Point<2> x=xj;
    for (int i=0;i<=number_elements[0];i++,v++,x+=dx[0]) {
      vertices[v]=x;
    }
    if (j<number_elements[1]) {
      x=xj+(dx[0]+dx[1])*0.5;
      for (int i=0;i<number_elements[0];i++,v++,x+=dx[0]) {
        vertices[v]=x;
      }
    }
  }
  NumPtr<CellData > cells(4*number_elements[0]*number_elements[1]);
  int c=0;
  for (int j=0;j<number_elements[1];j++) {
    for (int i=0;i<number_elements[0];i++,c+=4) {
      CellData &cell0=cells[c];
      cell0.allocate(3);
      cell0.vertex(0)=i+j*(2*number_elements[0]+1);
      cell0.vertex(1)=cell0.vertex(0)+1;
      cell0.vertex(2)=cell0.vertex(0)+number_elements[0]+1;
      cell0.setMaterialModel(0);
      CellData &cell1=cells[c+1];
      cell1.allocate(3);
      cell1.vertex(0)=cell0.vertex(1);
      cell1.vertex(1)=cell0.vertex(1)+2*number_elements[0]+1;
      cell1.vertex(2)=cell0.vertex(2);
      cell1.setMaterialModel(0);
      CellData &cell2=cells[c+2];
      cell2.allocate(3);
      cell2.vertex(0)=cell1.vertex(1);
      cell2.vertex(1)=cell0.vertex(0)+2*number_elements[0]+1;
      cell2.vertex(2)=cell0.vertex(2);
      cell2.setMaterialModel(0);
      CellData &cell3=cells[c+3];
      cell3.allocate(3);
      cell3.vertex(0)=cell2.vertex(1);
      cell3.vertex(1)=cell0.vertex(0);
      cell3.vertex(2)=cell0.vertex(2);
      cell3.setMaterialModel(0);
    }
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(4);
    boundary_condition.allocate(4);
    Point<2> center(low+(edge[0]+edge[1])*0.5);
    boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1]);
    boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
      low+edge[0]+edge[1]);
    boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0]);
    boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
      low+edge[0]+edge[1]);
    for (int i=0;i<4;i++) {
      boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
    }

    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        CODIMENSION1_SHAPE_CLASS *face_element=
          codimension1_level->getShape();
        Point<SPACEDIM> center=face_element->center(vertices);
        for (int i=0;i<boundary.getNumber();i++) {
          if (abs(boundary[i]->constraint(center))<1.e-5) {
            codimension1_level->
              setBoundaryCondition(boundary_condition[i]);
            break;
          }
        }
      }
    }
# ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
# endif
    tess.synchronize(boundary_data);
  }
  return boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
BoundaryData* generateParallelpipedPrisms(Tessellation &tess,
const Point<3> &low,const Point<3> edge[3],const int number_elements[3],
int dir,bool set_face_ids,NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  int n0=number_elements[0]+1;
  int n01=(number_elements[0]+1)*(number_elements[1]+1);
  int iv=0;
  int ivf=n01*(number_elements[2]+1);
  Point<3> dx[3];
  for (int i=0;i<3;i++) {
    CHECK_POSITIVE(number_elements[i]);
    dx[i]=edge[i]/static_cast<REAL>(number_elements[i]);
  }
  NumPtr<Point<3> > vertices(ivf
    +(number_elements[dir]+1)*number_elements[(dir+1)%3]
                             *number_elements[(dir+2)%3]);
  Point<3> xk=low;
  int v=0;
  for (int k=0;k<=number_elements[2];k++,xk+=dx[2]) {
    Point<3> xj=xk;
    for (int j=0;j<=number_elements[1];j++,xj+=dx[1]) {
      Point<3> x=xj;
      for (int i=0;i<=number_elements[0];i++,x+=dx[0]) {
        vertices[v++]=x;
      }
    }
  }
  NumPtr<CellData> 
    cells(4*number_elements[0]*number_elements[1]*number_elements[2]);
  switch (dir) {
    case 0: {
      for (int k=0;k<number_elements[2];k++) {
        for (int j=0;j<number_elements[1];j++) {
          for (int i=0;i<=number_elements[0];i++) {
            int ijk=i+j*n0+k*n01;
            vertices[v++]=(vertices[ijk]+vertices[ijk+n0]
                          +vertices[ijk+n01]+vertices[ijk+n01+n0])*0.25;
          }
        }
      }
      int c=0;
      for (int k=0;k<number_elements[2];k++) {
        for (int j=0;j<number_elements[1];j++) {
          for (int i=0;i<number_elements[0];i++) {
            { // 0
              CellData &cell=cells[c++];
              cell.allocate(6);
              cell.vertex(0)=ivf
                +i+(number_elements[0]+1)*(j+k*number_elements[1]);
              cell.vertex(1)=iv+i+j*n0+k*n01;
              cell.vertex(2)=cell.vertex(1)+n0;
              cell.vertex(3)=cell.vertex(0)+1;
              cell.vertex(4)=cell.vertex(1)+1;
              cell.vertex(5)=cell.vertex(2)+1;
              cell.setMaterialModel(0);
            }
            { // 1
              CellData &cell=cells[c++];
              cell.allocate(6);
              cell.vertex(0)=ivf
                +i+(number_elements[0]+1)*(j+k*number_elements[1]);
              cell.vertex(1)=iv+i+(j+1)*n0+k*n01;
              cell.vertex(2)=cell.vertex(1)+n01;
              cell.vertex(3)=cell.vertex(0)+1;
              cell.vertex(4)=cell.vertex(1)+1;
              cell.vertex(5)=cell.vertex(2)+1;
              cell.setMaterialModel(0);
            }
            { // 2
              CellData &cell=cells[c++];
              cell.allocate(6);
              cell.vertex(0)=ivf
                +i+(number_elements[0]+1)*(j+k*number_elements[1]);
              cell.vertex(1)=iv+i+(j+1)*n0+(k+1)*n01;
              cell.vertex(2)=cell.vertex(1)-n0;
              cell.vertex(3)=cell.vertex(0)+1;
              cell.vertex(4)=cell.vertex(1)+1;
              cell.vertex(5)=cell.vertex(2)+1;
              cell.setMaterialModel(0);
            }
            { // 3
              CellData &cell=cells[c++];
              cell.allocate(6);
              cell.vertex(0)=ivf
                +i+(number_elements[0]+1)*(j+k*number_elements[1]);
              cell.vertex(1)=iv+i+j*n0+(k+1)*n01;
              cell.vertex(2)=cell.vertex(1)-n01;
              cell.vertex(3)=cell.vertex(0)+1;
              cell.vertex(4)=cell.vertex(1)+1;
              cell.vertex(5)=cell.vertex(2)+1;
              cell.setMaterialModel(0);
            }
          }
        }
      }
      break;
    }
    case 1: {
      for (int k=0;k<number_elements[2];k++) {
        for (int j=0;j<=number_elements[1];j++) {
          for (int i=0;i<number_elements[0];i++) {
            int ijk=i+j*n0+k*n01;
            vertices[v++]=(vertices[ijk]+vertices[ijk+1]
                          +vertices[ijk+n01]+vertices[ijk+n01+1])*0.25;
          }
        }
      }
      int c=0;
      for (int k=0;k<number_elements[2];k++) {
        for (int j=0;j<number_elements[1];j++) {
          for (int i=0;i<number_elements[0];i++) {
            { // 0
              CellData &cell=cells[c++];
              cell.allocate(6);
              cell.vertex(0)=ivf
                +i+number_elements[0]*(j+k*(number_elements[1]+1));
              cell.vertex(1)=iv+i+j*n0+k*n01;
              cell.vertex(2)=cell.vertex(1)+n01;
              cell.vertex(3)=cell.vertex(0)+number_elements[0];
              cell.vertex(4)=cell.vertex(1)+n0;
              cell.vertex(5)=cell.vertex(2)+n0;
              cell.setMaterialModel(0);
            }
            { // 1
              CellData &cell=cells[c++];
              cell.allocate(6);
              cell.vertex(0)=ivf
                +i+number_elements[0]*(j+k*(number_elements[1]+1));
              cell.vertex(1)=iv+i+j*n0+(k+1)*n01;
              cell.vertex(2)=cell.vertex(1)+1;
              cell.vertex(3)=cell.vertex(0)+number_elements[0];
              cell.vertex(4)=cell.vertex(1)+n0;
              cell.vertex(5)=cell.vertex(2)+n0;
              cell.setMaterialModel(0);
            }
            { // 2
              CellData &cell=cells[c++];
              cell.allocate(6);
              cell.vertex(0)=ivf
                +i+number_elements[0]*(j+k*(number_elements[1]+1));
              cell.vertex(1)=iv+i+1+j*n0+(k+1)*n01;
              cell.vertex(2)=cell.vertex(1)-n01;
              cell.vertex(3)=cell.vertex(0)+number_elements[0];
              cell.vertex(4)=cell.vertex(1)+n0;
              cell.vertex(5)=cell.vertex(2)+n0;
              cell.setMaterialModel(0);
            }
            { // 3
              CellData &cell=cells[c++];
              cell.allocate(6);
              cell.vertex(0)=ivf
                +i+number_elements[0]*(j+k*(number_elements[1]+1));
              cell.vertex(1)=iv+i+1+j*n0+k*n01;
              cell.vertex(2)=cell.vertex(1)-1;
              cell.vertex(3)=cell.vertex(0)+number_elements[0];
              cell.vertex(4)=cell.vertex(1)+n0;
              cell.vertex(5)=cell.vertex(2)+n0;
              cell.setMaterialModel(0);
            }
          }
        }
      }
      break;
    }
    case 2: {
      for (int k=0;k<=number_elements[2];k++) {
        for (int j=0;j<number_elements[1];j++) {
          for (int i=0;i<number_elements[0];i++) {
            int ijk=i+j*n0+k*n01;
            vertices[v++]=(vertices[ijk]+vertices[ijk+1]
                          +vertices[ijk+n0]+vertices[ijk+n0+1])*0.25;
          }
        }
      }
      int c=0;
      for (int k=0;k<number_elements[2];k++) {
        for (int j=0;j<number_elements[1];j++) {
          for (int i=0;i<number_elements[0];i++) {
            { // 0
              CellData &cell=cells[c++];
              cell.allocate(6);
              cell.vertex(0)=ivf
                +i+number_elements[0]*(j+k*number_elements[1]);
              cell.vertex(1)=iv+i+j*n0+k*n01;
              cell.vertex(2)=cell.vertex(1)+1;
              cell.vertex(3)=cell.vertex(0)
                +number_elements[0]*number_elements[1];;
              cell.vertex(4)=cell.vertex(1)+n01;
              cell.vertex(5)=cell.vertex(2)+n01;
              cell.setMaterialModel(0);
            }
            { // 1
              CellData &cell=cells[c++];
              cell.allocate(6);
              cell.vertex(0)=ivf
                +i+number_elements[0]*(j+k*number_elements[1]);
              cell.vertex(1)=iv+i+1+j*n0+k*n01;
              cell.vertex(2)=cell.vertex(1)+n0;
              cell.vertex(3)=cell.vertex(0)
                +number_elements[0]*number_elements[1];;
              cell.vertex(4)=cell.vertex(1)+n01;
              cell.vertex(5)=cell.vertex(2)+n01;
              cell.setMaterialModel(0);
            }
            { // 2
              CellData &cell=cells[c++];
              cell.allocate(6);
              cell.vertex(0)=ivf
                +i+number_elements[0]*(j+k*number_elements[1]);
              cell.vertex(1)=iv+i+1+(j+1)*n0+k*n01;
              cell.vertex(2)=cell.vertex(1)-1;
              cell.vertex(3)=cell.vertex(0)
                +number_elements[0]*number_elements[1];;
              cell.vertex(4)=cell.vertex(1)+n01;
              cell.vertex(5)=cell.vertex(2)+n01;
              cell.setMaterialModel(0);
            }
            { // 3
              CellData &cell=cells[c++];
              cell.allocate(6);
              cell.vertex(0)=ivf
                +i+number_elements[0]*(j+k*number_elements[1]);
              cell.vertex(1)=iv+i+(j+1)*n0+k*n01;
              cell.vertex(2)=cell.vertex(1)-n0;
              cell.vertex(3)=cell.vertex(0)
                +number_elements[0]*number_elements[1];;
              cell.vertex(4)=cell.vertex(1)+n01;
              cell.vertex(5)=cell.vertex(2)+n01;
              cell.setMaterialModel(0);
            }
          }
        }
      }
      break;
    }
    default:
      OBSOLETE("bad dir");
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(6);
    boundary_condition.allocate(6);
    Point<3> center(low+(edge[0]+edge[1]+edge[2])*0.5);
    boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1],
      low+edge[2]);
    boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
      low+edge[0]+edge[1],low+edge[0]+edge[2]);
    boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[2],
      low+edge[0]);
    boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
      low+edge[1]+edge[2],low+edge[1]+edge[0]);
    boundary[4]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0],
      low+edge[1]);
    boundary[5]=OPERATOR_NEW PlaneBoundary(center,low+edge[2],
      low+edge[2]+edge[0],low+edge[2]+edge[1]);
    for (int i=0;i<2*SPACEDIM;i++) {
      boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
    }

    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        CODIMENSION1_SHAPE_CLASS *face_element=
          codimension1_level->getShape();
        Point<SPACEDIM> center=face_element->center(vertices);
        bool found=false;
        for (int i=0;i<boundary.getNumber();i++) {
          if (abs(boundary[i]->constraint(center))<1.e-5) {
            codimension1_level->
              setBoundaryCondition(boundary_condition[i]);
            found=true;
            break;
          }
        }
#ifdef DEBUG
        if (!found) {
          cout << "\tface_element:" << endl;
          face_element->printOn(cout);
          cout << "\tcenter = " << center << endl;
          for (int i=0;i<boundary.getNumber();i++) {
            cout << "\tboundary[" << i << "]:" << endl;
            boundary[i]->printOn(cout);
            cout << "\tconstraint = " << boundary[i]->constraint(center)
                 << endl;
          }
        }
#endif
        CHECK_TEST(found);
      }
    }
    tess.synchronize(boundary_data);
# ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
    tess.checkLineBoundaryCondition(boundary_condition);
# endif
  }
  return boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
BoundaryData* generateParallelpipedTetrahedra(Tessellation &tess,
const Point<3> &low,const Point<3> edge[3],
const int number_elements[3],bool set_face_ids,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  int n0=number_elements[0]+1;
  int n1=number_elements[1]+1;
  int n2=number_elements[2]+1;
  int n01=n0*(number_elements[1]+1);
  int iv=0;
  int ivf0=iv+n0*n1*n2;
  int ivf1=ivf0+n0*number_elements[1]*number_elements[2];
  int ivf2=ivf1+n1*number_elements[2]*number_elements[0];
  int ivc=ivf2+n2*number_elements[0]*number_elements[1];
  Point<3> dx[3];
  for (int i=0;i<3;++i) {
    CHECK_POSITIVE(number_elements[i]);
    dx[i]=edge[i]/static_cast<REAL>(number_elements[i]);
  }
  NumPtr<Point<3> > vertices(n0*n1*n2
    +n0*number_elements[1]*number_elements[2]
    +n1*number_elements[2]*number_elements[0]
    +n2*number_elements[0]*number_elements[1]
    +number_elements[0]*number_elements[1]*number_elements[2]);
  int v=0;
//vertices of parallelpipeds
  Point<3> xk=low;
  for (int k=0;k<=number_elements[2];k++,xk+=dx[2]) {
    Point<3> xj=xk;
    for (int j=0;j<=number_elements[1];j++,xj+=dx[1]) {
      Point<3> x=xj;
      for (int i=0;i<=number_elements[0];i++,x+=dx[0]) {
        vertices[v++]=x;
      }
    }
  }
//centers of parallelpiped faces in direction 0
  for (int k=0;k<number_elements[2];k++) {
    for (int j=0;j<number_elements[1];j++) {
      for (int i=0;i<=number_elements[0];i++) {
        int vf=i+j*n0+k*n01;
        vertices[v++]=(vertices[vf]+vertices[vf+n0]
                      +vertices[vf+n01]+vertices[vf+n01+n0])*0.25;
      }
    }
  }
//centers of parallelpiped faces in direction 1
  for (int k=0;k<number_elements[2];k++) {
    for (int j=0;j<=number_elements[1];j++) {
      for (int i=0;i<number_elements[0];i++) {
        int vf=i+j*n0+k*n01;
        vertices[v++]=(vertices[vf]+vertices[vf+1]
                      +vertices[vf+n01]+vertices[vf+n01+1])*0.25;
      }
    }
  }
//centers of parallelpiped faces in direction 2
  for (int k=0;k<=number_elements[2];k++) {
    for (int j=0;j<number_elements[1];j++) {
      for (int i=0;i<number_elements[0];i++) {
        int vf=i+j*n0+k*n01;
        vertices[v++]=(vertices[vf]+vertices[vf+1]
                      +vertices[vf+n0]+vertices[vf+n0+1])*0.25;
      }
    }
  }
//centers of parallelpipeds
  for (int k=0;k<number_elements[2];k++) {
    for (int j=0;j<number_elements[1];j++) {
      for (int i=0;i<number_elements[0];i++) {
        int vf=i+j*n0+k*n01;
        vertices[v++]=(vertices[vf]+vertices[vf+1]
                      +vertices[vf+n0]+vertices[vf+n0+1]
                      +vertices[vf+n01]+vertices[vf+n01+1]
                      +vertices[vf+n01+n0]+vertices[vf+n01+n0+1])*0.125;
      }
    }
  }
  NumPtr<CellData> 
    cells(24*number_elements[0]*number_elements[1]*number_elements[2]);
  int c=0;
  for (int k=0;k<number_elements[2];k++) {
    for (int j=0;j<number_elements[1];j++) {
      for (int i=0;i<number_elements[0];i++) {
        { // 0
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf0+i+n0*(j+k*number_elements[1]);
          cell.vertex(1)=iv+i+j*n0+k*n01;
          cell.vertex(2)=cell.vertex(1)+n0;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 1
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf0+i+n0*(j+k*number_elements[1]);
          cell.vertex(1)=iv+i+(j+1)*n0+k*n01;
          cell.vertex(2)=cell.vertex(1)+n01;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 2
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf0+i+n0*(j+k*number_elements[1]);
          cell.vertex(1)=iv+i+(j+1)*n0+(k+1)*n01;
          cell.vertex(2)=cell.vertex(1)-n0;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 3
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf0+i+n0*(j+k*number_elements[1]);
          cell.vertex(1)=iv+i+j*n0+(k+1)*n01;
          cell.vertex(2)=cell.vertex(1)-n01;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 4
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf0+i+1+n0*(j+k*number_elements[1]);
          cell.vertex(1)=iv+i+1+(j+1)*n0+k*n01;
          cell.vertex(2)=cell.vertex(1)-n0;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 5
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf0+i+1+n0*(j+k*number_elements[1]);
          cell.vertex(1)=iv+i+1+(j+1)*n0+(k+1)*n01;
          cell.vertex(2)=cell.vertex(1)-n01;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 6
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf0+i+1+n0*(j+k*number_elements[1]);
          cell.vertex(1)=iv+i+1+j*n0+(k+1)*n01;
          cell.vertex(2)=cell.vertex(1)+n0;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 7
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf0+i+1+n0*(j+k*number_elements[1]);
          cell.vertex(1)=iv+i+1+j*n0+k*n01;
          cell.vertex(2)=cell.vertex(1)+n01;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 8
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf1
            +i+number_elements[0]*(j+k*(number_elements[1]+1));
          cell.vertex(1)=iv+i+j*n0+k*n01;
          cell.vertex(2)=cell.vertex(1)+n01;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 9
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf1
            +i+number_elements[0]*(j+k*(number_elements[1]+1));
          cell.vertex(1)=iv+i+j*n0+(k+1)*n01;
          cell.vertex(2)=cell.vertex(1)+1;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 10
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf1
            +i+number_elements[0]*(j+k*(number_elements[1]+1));
          cell.vertex(1)=iv+i+1+j*n0+(k+1)*n01;
          cell.vertex(2)=cell.vertex(1)-n01;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 11
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf1
            +i+number_elements[0]*(j+k*(number_elements[1]+1));
          cell.vertex(1)=iv+i+1+j*n0+k*n01;
          cell.vertex(2)=cell.vertex(1)-1;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 12
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf1
            +i+number_elements[0]*(j+1+k*(number_elements[1]+1));
          cell.vertex(1)=iv+i+(j+1)*n0+(k+1)*n01;
          cell.vertex(2)=cell.vertex(1)-n01;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 13
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf1
            +i+number_elements[0]*(j+1+k*(number_elements[1]+1));
          cell.vertex(1)=iv+i+1+(j+1)*n0+(k+1)*n01;
          cell.vertex(2)=cell.vertex(1)-1;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 14
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf1
            +i+number_elements[0]*(j+1+k*(number_elements[1]+1));
          cell.vertex(1)=iv+i+1+(j+1)*n0+k*n01;
          cell.vertex(2)=cell.vertex(1)+n01;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 15
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf1
            +i+number_elements[0]*(j+1+k*(number_elements[1]+1));
          cell.vertex(1)=iv+i+(j+1)*n0+k*n01;
          cell.vertex(2)=cell.vertex(1)+1;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 16
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf2
            +i+number_elements[0]*(j+k*number_elements[1]);
          cell.vertex(1)=iv+i+j*n0+k*n01;
          cell.vertex(2)=cell.vertex(1)+1;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 17
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf2
            +i+number_elements[0]*(j+k*number_elements[1]);
          cell.vertex(1)=iv+i+1+j*n0+k*n01;
          cell.vertex(2)=cell.vertex(1)+n0;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 18
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf2
            +i+number_elements[0]*(j+k*number_elements[1]);
          cell.vertex(1)=iv+i+1+(j+1)*n0+k*n01;
          cell.vertex(2)=cell.vertex(1)-1;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 19
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf2
            +i+number_elements[0]*(j+k*number_elements[1]);
          cell.vertex(1)=iv+i+(j+1)*n0+k*n01;
          cell.vertex(2)=cell.vertex(1)-n0;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 20
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf2
            +i+number_elements[0]*(j+(k+1)*number_elements[1]);
          cell.vertex(1)=iv+i+1+j*n0+(k+1)*n01;
          cell.vertex(2)=cell.vertex(1)-1;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 21
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf2
            +i+number_elements[0]*(j+(k+1)*number_elements[1]);
          cell.vertex(1)=iv+i+1+(j+1)*n0+(k+1)*n01;
          cell.vertex(2)=cell.vertex(1)-n0;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 22
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf2
            +i+number_elements[0]*(j+(k+1)*number_elements[1]);
          cell.vertex(1)=iv+i+(j+1)*n0+(k+1)*n01;
          cell.vertex(2)=cell.vertex(1)+1;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
        { // 23
          CellData &cell=cells[c++];
          cell.allocate(4);
          cell.vertex(0)=ivf2
            +i+number_elements[0]*(j+(k+1)*number_elements[1]);
          cell.vertex(1)=iv+i+j*n0+(k+1)*n01;
          cell.vertex(2)=cell.vertex(1)+n0;
          cell.vertex(3)=ivc+i+number_elements[0]*(j+k*number_elements[1]);
          cell.setMaterialModel(0);
          CHECK_POSITIVE(cell.numberVertices());
        }
      }
    }
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(6);
    boundary_condition.allocate(6);
    Point<3> center(low+(edge[0]+edge[1]+edge[2])*0.5);
    boundary[0]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[1],
      low+edge[2]);
    boundary[1]=OPERATOR_NEW PlaneBoundary(center,low+edge[0],
      low+edge[0]+edge[1],low+edge[0]+edge[2]);
    boundary[2]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[2],
      low+edge[0]);
    boundary[3]=OPERATOR_NEW PlaneBoundary(center,low+edge[1],
      low+edge[1]+edge[2],low+edge[1]+edge[0]);
    boundary[4]=OPERATOR_NEW PlaneBoundary(center,low,low+edge[0],
      low+edge[1]);
    boundary[5]=OPERATOR_NEW PlaneBoundary(center,low+edge[2],
      low+edge[2]+edge[0],low+edge[2]+edge[1]);
    for (int i=0;i<2*SPACEDIM;i++) {
      boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
    }

    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        CODIMENSION1_SHAPE_CLASS *face_element=
          codimension1_level->getShape();
        Point<SPACEDIM> center=face_element->center(vertices);
        for (int i=0;i<boundary.getNumber();i++) {
          if (abs(boundary[i]->constraint(center))<1.e-5) {
            codimension1_level->
              setBoundaryCondition(boundary_condition[i]);
            break;
          }
        }
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
    tess.checkLineBoundaryCondition(boundary_condition);
#endif
  }
  return boundary_data;
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
BoundaryData* generateSlitQuadrilateral(Tessellation &tess,
bool set_face_ids,NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  NumPtr<int> p(10);
  RandomPermutation(p); // hopefully get random face numbers on boundary
  NumPtr<Point<2> > vertices(10);
  vertices[p[0]]=Point<2>(0.,0.);
  vertices[p[1]]=Point<2>(.5,0.);
  vertices[p[2]]=Point<2>(1.,0.);
  vertices[p[3]]=Point<2>(0.,.5);
  vertices[p[4]]=Point<2>(.5,.5);
  vertices[p[5]]=Point<2>(1.,.5);
  vertices[p[6]]=Point<2>(1.,.5);
  vertices[p[7]]=Point<2>(0.,1.);
  vertices[p[8]]=Point<2>(.5,1.);
  vertices[p[9]]=Point<2>(1.,1.);
  const int cell_vertices[4][4] = { 
    { p[0],p[1],p[3],p[4] }, { p[1],p[2],p[4],p[5] }, 
    { p[3],p[4],p[7],p[8] }, { p[4],p[6],p[8],p[9] }
  };
  NumPtr<CellData> cells(4);
  for (int c=0;c<4;c++) {
    cells[c].allocate(4);
    for (int v=0;v<4;v++) {
      cells[c].vertex(v)=cell_vertices[c][v];
    }
    cells[c].setMaterialModel(0);
  }

  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(6);
    boundary_condition.allocate(7);
    Point<2> cen(0.25,0.25);
    boundary[0]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[0]],
      vertices[p[7]]); // left
    boundary[1]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[2]],
      vertices[p[9]]); //right
    boundary[2]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[0]],
      vertices[p[2]]); // bottom
    boundary[3]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[4]],
      vertices[p[5]]); // slit: outer normal points up
    cen[1]=0.75;
    boundary[4]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[4]],
      vertices[p[6]]); // slit: outer normal points down
    boundary[5]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[7]],
      vertices[p[9]]); // top

    boundary_condition[0]=OPERATOR_NEW BoundaryCondition(boundary[0]);
    boundary_condition[1]=OPERATOR_NEW BoundaryCondition(boundary[1]);
    boundary_condition[2]=OPERATOR_NEW BoundaryCondition(boundary[1]);
    boundary_condition[3]=OPERATOR_NEW BoundaryCondition(boundary[2]);
    boundary_condition[4]=OPERATOR_NEW BoundaryCondition(boundary[3]);
    boundary_condition[5]=OPERATOR_NEW BoundaryCondition(boundary[4]);
    boundary_condition[6]=OPERATOR_NEW BoundaryCondition(boundary[5]);

    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        CODIMENSION1_SHAPE_CLASS *face_element=
          codimension1_level->getShape();
        Point<SPACEDIM> center=face_element->center(vertices);
        if (abs(boundary[0]->constraint(center))<1.e-5) {
          codimension1_level->
            setBoundaryCondition(boundary_condition[0]);
        } else if (abs(boundary[1]->constraint(center))<1.e-5) {
          if (center[1]<0.5) {
            codimension1_level->
              setBoundaryCondition(boundary_condition[1]);
          } else {
            codimension1_level->
              setBoundaryCondition(boundary_condition[2]);
          }
        } else if (abs(boundary[2]->constraint(center))<1.e-5) {
          codimension1_level->
            setBoundaryCondition(boundary_condition[3]);
        } else if (abs(boundary[5]->constraint(center))<1.e-5) {
          codimension1_level->
            setBoundaryCondition(boundary_condition[6]);
        } else {
          if (face_element->vertexIndex(0)==p[5] ||
          face_element->vertexIndex(1)==p[5]) {
            codimension1_level->
              setBoundaryCondition(boundary_condition[4]);
          } else {
            codimension1_level->
              setBoundaryCondition(boundary_condition[5]);
          }
        }
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
#endif
  }
  return boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
BoundaryData* generateL(Tessellation &tess,bool set_face_ids,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  NumPtr<int> p(8);
  RandomPermutation(p); // hopefully get random face numbers on boundary
  NumPtr<Point<2> > vertices(8);
  vertices[p[0]]=Point<2>(0.,0.);
  vertices[p[1]]=Point<2>(.5,0.);
  vertices[p[2]]=Point<2>(1.,0.);
  vertices[p[3]]=Point<2>(0.,.5);
  vertices[p[4]]=Point<2>(.5,.5);
  vertices[p[5]]=Point<2>(1.,.5);
  vertices[p[6]]=Point<2>(0.,1.);
  vertices[p[7]]=Point<2>(.5,1.);
  const int cell_vertices[3][4] = {
    {p[0], p[1], p[3], p[4]}, 
    {p[1], p[2], p[4], p[5]}, 
    {p[3], p[4], p[6], p[7]}
  };
  NumPtr<CellData> cells(3);
  for (int c=0;c<3;c++) {
    cells[c].allocate(4);
    for (int v=0;v<4;v++) {
      cells[c].vertex(v)=cell_vertices[c][v];
    }
    cells[c].setMaterialModel(0);
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(6);
    boundary_condition.allocate(6);
    Point<2> cen(0.25,0.25);
    boundary[0]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[0]],
      vertices[p[6]]);
    boundary[1]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[2]],
      vertices[p[5]]);
    boundary[2]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[0]],
      vertices[p[2]]);
    boundary[3]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[6]],
      vertices[p[7]]);
    boundary[4]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[4]],
      vertices[p[7]]);
    boundary[5]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[4]],
      vertices[p[5]]);

    for (int i=0;i<6;i++) {
      boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
    }

    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        CODIMENSION1_SHAPE_CLASS *face_element=
          codimension1_level->getShape();
        Point<SPACEDIM> center=face_element->center(vertices);
        for (int i=0;i<boundary.getNumber();i++) {
          if (abs(boundary[i]->constraint(center))<1.e-5) {
            codimension1_level->
              setBoundaryCondition(boundary_condition[i]);
            break;
          }
        }
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
#endif
  }
  return boundary_data;
}
#endif
#if (SPACEDIM==3)
BoundaryData* generateL(Tessellation &tess,bool set_face_ids,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  NumPtr<int> p(26);
  RandomPermutation(p); // hopefully get random face numbers on boundary
  NumPtr<Point<3> > vertices(26);
  vertices[p[ 0]]=Point<3>(0.,0.,0.);
  vertices[p[ 1]]=Point<3>(.5,0.,0.);
  vertices[p[ 2]]=Point<3>(1.,0.,0.);
  vertices[p[ 3]]=Point<3>(0.,0.,.5);
  vertices[p[ 4]]=Point<3>(.5,0.,.5);
  vertices[p[ 5]]=Point<3>(1.,0.,.5);
  vertices[p[ 6]]=Point<3>(0.,0.,1.);
  vertices[p[ 7]]=Point<3>(.5,0.,1.);
  vertices[p[ 8]]=Point<3>(1.,0.,1.);
  vertices[p[ 9]]=Point<3>(0.,.5,0.);
  vertices[p[10]]=Point<3>(.5,.5,0.);
  vertices[p[11]]=Point<3>(1.,.5,0.);
  vertices[p[12]]=Point<3>(0.,.5,.5);
  vertices[p[13]]=Point<3>(.5,.5,.5);
  vertices[p[14]]=Point<3>(1.,.5,.5);
  vertices[p[15]]=Point<3>(0.,.5,1.);
  vertices[p[16]]=Point<3>(.5,.5,1.);
  vertices[p[17]]=Point<3>(1.,.5,1.);
  vertices[p[18]]=Point<3>(0.,1.,0.);
  vertices[p[19]]=Point<3>(.5,1.,0.);
  vertices[p[20]]=Point<3>(1.,1.,0.);
  vertices[p[21]]=Point<3>(0.,1.,.5);
  vertices[p[22]]=Point<3>(.5,1.,.5);
  vertices[p[23]]=Point<3>(1.,1.,.5);
  vertices[p[24]]=Point<3>(0.,1.,1.);
  vertices[p[25]]=Point<3>(.5,1.,1.);
  const int cell_vertices[7][8] = {
    {p[ 0],p[ 1],p[ 9],p[10],p[ 3],p[ 4],p[12],p[13]},
    {p[ 1],p[ 2],p[10],p[11],p[ 4],p[ 5],p[13],p[14]},
    {p[ 3],p[ 4],p[12],p[13],p[ 6],p[ 7],p[15],p[16]},
    {p[ 4],p[ 5],p[13],p[14],p[ 7],p[ 8],p[16],p[17]},
    {p[ 9],p[10],p[18],p[19],p[12],p[13],p[21],p[22]},
    {p[10],p[11],p[19],p[20],p[13],p[14],p[22],p[23]},
    {p[12],p[13],p[21],p[22],p[15],p[16],p[24],p[25]}};
  NumPtr<CellData> cells(7);
  for (int c=0;c<7;c++) {
    cells[c].allocate(8);
    for (int v=0;v<8;v++) {
      cells[c].vertex(v)=cell_vertices[c][v];
    }
    cells[c].setMaterialModel(0);
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(9);
    boundary_condition.allocate(9);
    Point<3> cen(0.25,0.25,0.25);
    boundary[0]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[0]],
      vertices[p[18]],vertices[p[6]]); // back
    boundary[1]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[2]],
      vertices[p[20]],vertices[p[8]]); // front
    boundary[2]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[0]],
      vertices[p[2]],vertices[p[6]]); // left
    boundary[3]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[18]],
      vertices[p[20]],vertices[p[24]]); // right
    boundary[4]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[0]],
      vertices[p[2]],vertices[p[18]]); // bottom
    boundary[5]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[6]],
      vertices[p[8]],vertices[p[24]]); // top
    boundary[6]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[13]],
      vertices[p[22]],vertices[p[16]]); // L back
    boundary[7]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[13]],
      vertices[p[14]],vertices[p[16]]); // L left
    boundary[8]=OPERATOR_NEW PlaneBoundary(cen,vertices[p[13]],
      vertices[p[14]],vertices[p[22]]); // L bottom
    for (int i=0;i<9;i++) {
      boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
    }
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        CODIMENSION1_SHAPE_CLASS *face_element=
          codimension1_level->getShape();
        Point<SPACEDIM> center=face_element->center(vertices);
        for (int i=0;i<boundary.getNumber();i++) {
          if (abs(boundary[i]->constraint(center))<1.e-5) {
            codimension1_level->
              setBoundaryCondition(boundary_condition[i]);
            break;
          }
        }
      }
    }
    tess.synchronize(boundary_data);
# ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
    tess.checkLineBoundaryCondition(boundary_condition);
# endif
  }
  return boundary_data;
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
//5 quadrilaterals in ball
BoundaryData* generateBall(Tessellation &tess,
const Point<SPACEDIM> &center,REAL radius,bool set_face_ids,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  REAL r=radius/sqrt(2.0);
  REAL a=r/sqrt(5.0);
  NumPtr<Point<2> > vertices(8);
  NumPtr<int> p(8);
  RandomPermutation(p); // hopefully get random face numbers on boundary
  vertices[p[0]]=center+Point<2>(-r,-r);
  vertices[p[1]]=center+Point<2>( r,-r);
  vertices[p[2]]=center+Point<2>(-a,-a);
  vertices[p[3]]=center+Point<2>( a,-a);
  vertices[p[4]]=center+Point<2>(-a, a);
  vertices[p[5]]=center+Point<2>( a, a);
  vertices[p[6]]=center+Point<2>(-r, r);
  vertices[p[7]]=center+Point<2>( r, r);
  int cell_vertices[5][4] = {
    {p[0], p[1], p[2], p[3]},{p[0], p[2], p[6], p[4]},
    {p[2], p[3], p[4], p[5]},{p[1], p[7], p[3], p[5]},
    {p[6], p[4], p[7], p[5]}};
  NumPtr<CellData> cells(5);
  for (int c=0;c<5;c++) {
    cells[c].allocate(4);
    for (int v=0;v<4;v++) {
      cells[c].vertex(v)=cell_vertices[c][v];
    }
    cells[c].setMaterialModel(0);
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(1);
    boundary[0]=OPERATOR_NEW BallBoundary(center,radius);
    boundary_condition.allocate(1);
    boundary_condition[0]=OPERATOR_NEW BoundaryCondition(boundary[0]);
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        codimension1_level->
          setBoundaryCondition(boundary_condition[0]);
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
#endif
  }
  return boundary_data;
}
#endif
#if (SPACEDIM==3)
//7 hexahedra in ball
BoundaryData* generateBall(Tessellation &tess,
const Point<SPACEDIM> &center,REAL radius,bool set_face_ids,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  REAL r=radius/sqrt(3.0);
  REAL q=pow(316.+30.*sqrt(114.),1./3.);
  REAL a=r*(q-14./q+1.)/15.;
  NumPtr<Point<3> > vertices(16);
  NumPtr<int> p(16);
  RandomPermutation(p); // hopefully get random face numbers on boundary
  vertices[p[ 0]]=center+Point<3>(-r,-r,-r);
  vertices[p[ 1]]=center+Point<3>( r,-r,-r);
  vertices[p[ 2]]=center+Point<3>(-r, r,-r);
  vertices[p[ 3]]=center+Point<3>( r, r,-r);
  vertices[p[ 4]]=center+Point<3>(-r,-r, r);
  vertices[p[ 5]]=center+Point<3>( r,-r, r);
  vertices[p[ 6]]=center+Point<3>(-r, r, r);
  vertices[p[ 7]]=center+Point<3>( r, r, r);
  vertices[p[ 8]]=center+Point<3>(-a,-a,-a);
  vertices[p[ 9]]=center+Point<3>( a,-a,-a);
  vertices[p[10]]=center+Point<3>(-a, a,-a);
  vertices[p[11]]=center+Point<3>( a, a,-a);
  vertices[p[12]]=center+Point<3>(-a,-a, a);
  vertices[p[13]]=center+Point<3>( a,-a, a);
  vertices[p[14]]=center+Point<3>(-a, a, a);
  vertices[p[15]]=center+Point<3>( a, a, a);
  int cell_vertices[7][8] = {
    {p[ 0],p[ 1],p[ 2],p[ 3],p[ 8],p[ 9],p[10],p[11]},
    {p[ 8],p[ 9],p[10],p[11],p[12],p[13],p[14],p[15]},
    {p[12],p[13],p[14],p[15],p[ 4],p[ 5],p[ 6],p[ 7]},
    {p[10],p[11],p[ 2],p[ 3],p[14],p[15],p[ 6],p[ 7]},
    {p[ 0],p[ 8],p[ 2],p[10],p[ 4],p[12],p[ 6],p[14]},
    {p[ 0],p[ 1],p[ 8],p[ 9],p[ 4],p[ 5],p[12],p[13]},
    {p[ 9],p[ 1],p[11],p[ 3],p[13],p[ 5],p[15],p[ 7]}};
  NumPtr<CellData> cells(7);
  for (int c=0;c<7;c++) {
    cells[c].allocate(8);
    for (int v=0;v<8;v++) {
      cells[c].vertex(v)=cell_vertices[c][v];
    }
    cells[c].setMaterialModel(0);
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(1);
    boundary[0]=OPERATOR_NEW BallBoundary(center,radius);
    boundary_condition.allocate(1);
    boundary_condition[0]=OPERATOR_NEW BoundaryCondition(boundary[0]);
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        codimension1_level->
          setBoundaryCondition(boundary_condition[0]);
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
    tess.checkLineBoundaryCondition(boundary_condition);
#endif
  }
  return boundary_data;
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
BoundaryData* generateBallTriangles(Tessellation &tess,
const Point<2> &center,REAL radius,bool set_face_ids,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  NumPtr<Point<2> > vertices(5);
  NumPtr<int> p(5);
  RandomPermutation(p); // hopefully get random face numbers on boundary
  vertices[p[0]]=center;
  vertices[p[1]]=center+Point<2>( radius,0.);
  vertices[p[2]]=center+Point<2>( 0.,radius);
  vertices[p[3]]=center+Point<2>(-radius,0.);
  vertices[p[4]]=center+Point<2>(0.,-radius);
  int cell_vertices[4][3] = { 
    {p[0], p[1], p[2]},{p[0], p[2], p[3]},
    {p[0], p[3], p[4]},{p[0], p[4], p[1]}
  };
  NumPtr<CellData> cells(4);
  for (int c=0;c<4;c++) {
    cells[c].allocate(3);
    for (int v=0;v<3;v++) {
      cells[c].vertex(v)=cell_vertices[c][v];
    }
    cells[c].setMaterialModel(0);
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(1);
    boundary[0]=OPERATOR_NEW BallBoundary(center,radius);
    boundary_condition.allocate(1);
    boundary_condition[0]=OPERATOR_NEW BoundaryCondition(boundary[0]);
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        codimension1_level->
          setBoundaryCondition(boundary_condition[0]);
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
#endif
  }
  return boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
BoundaryData* generateBallTetrahedra(Tessellation &tess,
const Point<3> &center,REAL radius,bool set_face_ids,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  NumPtr<Point<3> > vertices(7);
  NumPtr<int> p(7);
  RandomPermutation(p); // hopefully get random face numbers on boundary
  vertices[p[0]]=center;
  vertices[p[1]]=center+Point<3>(0.,0.,-radius);
  vertices[p[2]]=center+Point<3>(radius,0.,0.);
  vertices[p[3]]=center+Point<3>(0.,radius,0.);
  vertices[p[4]]=center+Point<3>(-radius,0.,0.);
  vertices[p[5]]=center+Point<3>(0.,-radius,0.);
  vertices[p[6]]=center+Point<3>(0.,0.,radius);
  int cell_vertices[8][4] = { 
    {p[0],p[1],p[2],p[5]},{p[0],p[1],p[3],p[2]},
    {p[0],p[1],p[4],p[3]},{p[0],p[1],p[5],p[4]},
    {p[0],p[2],p[6],p[5]},{p[0],p[2],p[3],p[6]},
    {p[0],p[3],p[4],p[6]},{p[0],p[4],p[5],p[6]}
  };
  NumPtr<CellData> cells(8);
  for (int c=0;c<8;c++) {
    cells[c].allocate(4);
    for (int v=0;v<4;v++) {
      cells[c].vertex(v)=cell_vertices[c][v];
    }
    cells[c].setMaterialModel(0);
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(1);
    boundary[0]=OPERATOR_NEW BallBoundary(center,radius);
    boundary_condition.allocate(1);
    boundary_condition[0]=OPERATOR_NEW BoundaryCondition(boundary[0]);
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        codimension1_level->
          setBoundaryCondition(boundary_condition[0]);
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
    tess.checkLineBoundaryCondition(boundary_condition);
#endif
  }
  return boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//prisms have quadrateral faces on boundary of domain
BoundaryData* generateBallPrismsWithTetrahedra(Tessellation &tess,
const Point<3> &center,REAL radius,bool set_face_ids,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  REAL r=radius/sqrt(3.);
  NumPtr<Point<3> > vertices(12);
  NumPtr<int> p(12);
  RandomPermutation(p); // hopefully get random face numbers on boundary
  vertices[p[0]]=center+Point<3>(0.,-radius,0.);
  vertices[p[1]]=center+Point<3>(0., radius,0.);
  vertices[p[2]]=center+Point<3>(-r,-r,-r);
  vertices[p[3]]=center+Point<3>( r,-r,-r);
  vertices[p[4]]=center+Point<3>(-r, r,-r);
  vertices[p[5]]=center+Point<3>( r, r,-r);
  vertices[p[6]]=center+Point<3>(-r,-r, r);
  vertices[p[7]]=center+Point<3>( r,-r, r);
  vertices[p[8]]=center+Point<3>(-r, r, r);
  vertices[p[9]]=center+Point<3>( r, r, r);
  vertices[p[10]]=center-Point<3>(0.,0.5*r,0.);
  vertices[p[11]]=center+Point<3>(0.,0.5*r,0.);
  int prism_vertices[4][6] = { 
    {p[10],p[3],p[2],p[11],p[5],p[4]},{p[10],p[6],p[7],p[11],p[8],p[9]},
    {p[10],p[2],p[6],p[11],p[4],p[8]},{p[10],p[7],p[3],p[11],p[9],p[5]}
  };
  int tetrahedron_vertices[8][4] = { 
    {p[0],p[2],p[10],p[3]},{p[1],p[4],p[5],p[11]},
    {p[0],p[6],p[7],p[10]},{p[1],p[8],p[11],p[9]},
    {p[0],p[2],p[6],p[10]},{p[1],p[4],p[11],p[8]},
    {p[0],p[3],p[10],p[7]},{p[1],p[5],p[9],p[11]}
  };
  NumPtr<CellData> cells(12);
  for (int c=0;c<4;c++) {
    cells[c].allocate(6);
    for (int v=0;v<6;v++) {
      cells[c].vertex(v)=prism_vertices[c][v];
    }
    cells[c].setMaterialModel(0);
  }
  for (int c=4;c<12;c++) {
    cells[c].allocate(4);
    for (int v=0;v<4;v++) {
      cells[c].vertex(v)=tetrahedron_vertices[c-4][v];
    }
    cells[c].setMaterialModel(0);
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(1);
    boundary[0]=OPERATOR_NEW BallBoundary(center,radius);
    boundary_condition.allocate(1);
    boundary_condition[0]=OPERATOR_NEW BoundaryCondition(boundary[0]);
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        codimension1_level->
          setBoundaryCondition(boundary_condition[0]);
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
    tess.checkLineBoundaryCondition(boundary_condition);
#endif
  }
  return boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//prisms have triangle faces on boundary of domain
BoundaryData* generateBallPrismsWithHexahedra(Tessellation &tess,
const Point<3> &center,REAL radius,bool set_face_ids,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  REAL r=radius/sqrt(3.);
  NumPtr<int> p(20);
  RandomPermutation(p); // hopefully get random face numbers on boundary
  NumPtr<Point<3> > vertices(20);
  vertices[p[ 0]]=center+Point<3>(0.,-radius,0.);
  vertices[p[ 1]]=center+Point<3>(0., radius,0.);
  vertices[p[ 2]]=center+Point<3>(-r,-r,-r);
  vertices[p[ 3]]=center+Point<3>( r,-r,-r);
  vertices[p[ 4]]=center+Point<3>(-r, r,-r);
  vertices[p[ 5]]=center+Point<3>( r, r,-r);
  vertices[p[ 6]]=center+Point<3>(-r,-r, r);
  vertices[p[ 7]]=center+Point<3>( r,-r, r);
  vertices[p[ 8]]=center+Point<3>(-r, r, r);
  vertices[p[ 9]]=center+Point<3>( r, r, r);
  vertices[p[10]]=center+Point<3>(-0.5*r,-0.5*r,-0.5*r);
  vertices[p[11]]=center+Point<3>( 0.5*r,-0.5*r,-0.5*r);
  vertices[p[12]]=center+Point<3>(-0.5*r, 0.5*r,-0.5*r);
  vertices[p[13]]=center+Point<3>( 0.5*r, 0.5*r,-0.5*r);
  vertices[p[14]]=center+Point<3>(-0.5*r,-0.5*r, 0.5*r);
  vertices[p[15]]=center+Point<3>( 0.5*r,-0.5*r, 0.5*r);
  vertices[p[16]]=center+Point<3>(-0.5*r, 0.5*r, 0.5*r);
  vertices[p[17]]=center+Point<3>( 0.5*r, 0.5*r, 0.5*r);
  vertices[p[18]]=center+Point<3>(    0.,-0.5*r,    0.);
  vertices[p[19]]=center+Point<3>(    0., 0.5*r,    0.);
  int prism_vertices[12][6] = { 
    {p[ 0],p[ 3],p[ 2],p[18],p[11],p[10]},
    {p[ 1],p[ 4],p[ 5],p[19],p[12],p[13]},
    {p[ 0],p[ 6],p[ 7],p[18],p[14],p[15]},
    {p[ 1],p[ 9],p[ 8],p[19],p[17],p[16]},
    {p[ 0],p[ 2],p[ 6],p[18],p[10],p[14]},
    {p[ 1],p[ 8],p[ 4],p[19],p[16],p[12]},
    {p[ 0],p[ 7],p[ 3],p[18],p[15],p[11]},
    {p[ 1],p[ 5],p[ 9],p[19],p[13],p[17]},
    {p[18],p[11],p[10],p[19],p[13],p[12]},
    {p[18],p[14],p[15],p[19],p[16],p[17]},
    {p[18],p[10],p[14],p[19],p[12],p[16]},
    {p[18],p[15],p[11],p[19],p[17],p[13]}
  };
  int hexahedron_vertices[4][8] = { 
    {p[2],p[3],p[ 4],p[ 5],p[10],p[11],p[12],p[13]},
    {p[6],p[7],p[14],p[15],p[ 8],p[ 9],p[16],p[17]},
    {p[2],p[4],p[ 6],p[ 8],p[10],p[12],p[14],p[16]},
    {p[3],p[7],p[ 5],p[ 9],p[11],p[15],p[13],p[17]}
  };
  NumPtr<CellData> cells(16);
  for (int c=0;c<12;c++) {
    cells[c].allocate(6);
    for (int v=0;v<6;v++) {
      cells[c].vertex(v)=prism_vertices[c][v];
    }
    cells[c].setMaterialModel(0);
  }
  for (int c=12;c<16;c++) {
    cells[c].allocate(8);
    for (int v=0;v<8;v++) {
      cells[c].vertex(v)=hexahedron_vertices[c-12][v];
    }
    cells[c].setMaterialModel(0);
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(1);
    boundary[0]=OPERATOR_NEW BallBoundary(center,radius);
    boundary_condition.allocate(1);
    boundary_condition[0]=OPERATOR_NEW BoundaryCondition(boundary[0]);
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        codimension1_level->
          setBoundaryCondition(boundary_condition[0]);
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
    tess.checkLineBoundaryCondition(boundary_condition);
#endif
  }
  return boundary_data;
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//prisms have quadrateral faces on cylinder boundary of domain
BoundaryData* generateCylinderTetrahedra(Tessellation &tess,
const Point<3> &axis,const Point<3> &center,REAL radius,
bool set_face_ids,NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  NumPtr<Point<3> > vertices(10);
  NumPtr<int> p(10);
  RandomPermutation(p); // hopefully get random face numbers on boundary
  for (int i=0;i<p.getNumber();i++) p[i]=i;

  Tensor<1,3> e0(Point<3>(1.,0.,0.));
  Tensor<1,3> e1(Point<3>(0.,1.,0.));
  Tensor<1,3> e2(Point<3>(0.,0.,1.));
  Tensor<1,3> cp0,cp1,cp2;
  Tensor<1,3> ta(axis);
  crossProduct(cp0,ta,e0);
  crossProduct(cp1,ta,e1);
  crossProduct(cp2,ta,e2);
  REAL cpn0=cp0*cp0;
  REAL cpn1=cp1*cp1;
  REAL cpn2=cp2*cp2;
  Point<3> cpa,cpb;
  if (cpn0<min(cpn1,cpn2)) {
    cpa=cp1.getValues()*(radius/sqrt(cpn1));
    cpb=cp2.getValues()*(radius/sqrt(cpn2));
  } else if (cpn1<min(cpn2,cpn0)) {
    cpa=cp2.getValues()*(radius/sqrt(cpn2));
    cpb=cp1.getValues()*(radius/sqrt(cpn0));
  } else {
    cpa=cp0.getValues()*(radius/sqrt(cpn0));
    cpb=cp1.getValues()*(radius/sqrt(cpn1));
  }
  vertices[p[0]]=center+axis+cpa;
  vertices[p[1]]=center+axis+cpb;
  vertices[p[2]]=center+axis-cpa;
  vertices[p[3]]=center+axis-cpb;
  vertices[p[4]]=center-axis+cpa;
  vertices[p[5]]=center-axis+cpb;
  vertices[p[6]]=center-axis-cpa;
  vertices[p[7]]=center-axis-cpb;
  vertices[p[8]]=center+axis;
  vertices[p[9]]=center-axis;
  int tetrahedra_vertices[12][4] = { 
    {p[4],p[5],p[9],p[8]},
    {p[0],p[1],p[5],p[8]},
    {p[0],p[4],p[8],p[5]},
    {p[5],p[6],p[9],p[8]},
    {p[1],p[2],p[6],p[8]},
    {p[1],p[5],p[8],p[6]},
    {p[6],p[7],p[9],p[8]},
    {p[2],p[3],p[7],p[8]},
    {p[2],p[6],p[8],p[7]},
    {p[4],p[7],p[8],p[9]},
    {p[0],p[3],p[8],p[4]},
    {p[3],p[4],p[7],p[8]} 
  };
  NumPtr<CellData> cells(12);
  for (int c=0;c<12;c++) {
    cells[c].allocate(4);
    for (int v=0;v<4;v++) {
      cells[c].vertex(v)=tetrahedra_vertices[c][v];
    }
    cells[c].setMaterialModel(0);
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    REAL an=axis*axis;
    boundary.allocate(3);
    boundary_condition.allocate(3);

    boundary[0]=OPERATOR_NEW CylinderBoundary(axis,center,radius);
    boundary[1]=OPERATOR_NEW PlaneBoundary(center,vertices[9],
      vertices[4],vertices[5]);
    boundary[2]=OPERATOR_NEW PlaneBoundary(center,vertices[8],
      vertices[0],vertices[1]);
    for (int i=0;i<3;i++) {
      boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
    }
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        CODIMENSION1_SHAPE_CLASS *face_element=
          codimension1_level->getShape();
        Point<SPACEDIM> face_center=face_element->center(vertices);
        if (abs(boundary[1]->constraint(face_center))<1.e-5) {
          codimension1_level->
            setBoundaryCondition(boundary_condition[1]);
        } else if (abs(boundary[2]->constraint(face_center))<1.e-5) {
          codimension1_level->
            setBoundaryCondition(boundary_condition[2]);
        } else {
          codimension1_level->
            setBoundaryCondition(boundary_condition[0]);
        }
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
    tess.checkLineBoundaryCondition(boundary_condition);
#endif
  }
  return boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//prisms have quadrateral faces on cylinder boundary of domain
BoundaryData* generateCylinderPrisms(Tessellation &tess,
const Point<3> &axis,const Point<3> &center,REAL radius,
bool set_face_ids,NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  NumPtr<Point<3> > vertices(10);
  NumPtr<int> p(10);
  RandomPermutation(p); // hopefully get random face numbers on boundary
  Tensor<1,3> e0(Point<3>(1.,0.,0.));
  Tensor<1,3> e1(Point<3>(0.,1.,0.));
  Tensor<1,3> e2(Point<3>(0.,0.,1.));
  Tensor<1,3> cp0,cp1,cp2;
  Tensor<1,3> ta(axis);
  crossProduct(cp0,ta,e0);
  crossProduct(cp1,ta,e1);
  crossProduct(cp2,ta,e2);
  REAL cpn0=cp0*cp0;
  REAL cpn1=cp1*cp1;
  REAL cpn2=cp2*cp2;
  Point<3> cpa,cpb;
  if (cpn0<min(cpn1,cpn2)) {
    cpa=cp1.getValues()*(radius/sqrt(cpn1));
    cpb=cp2.getValues()*(radius/sqrt(cpn2));
  } else if (cpn1<min(cpn2,cpn0)) {
    cpa=cp2.getValues()*(radius/sqrt(cpn2));
    cpb=cp1.getValues()*(radius/sqrt(cpn0));
  } else {
    cpa=cp0.getValues()*(radius/sqrt(cpn0));
    cpb=cp1.getValues()*(radius/sqrt(cpn1));
  }
  vertices[p[0]]=center+axis+cpa;
  vertices[p[1]]=center+axis+cpb;
  vertices[p[2]]=center+axis-cpa;
  vertices[p[3]]=center+axis-cpb;
  vertices[p[4]]=center-axis+cpa;
  vertices[p[5]]=center-axis+cpb;
  vertices[p[6]]=center-axis-cpa;
  vertices[p[7]]=center-axis-cpb;
  vertices[p[8]]=center+axis;
  vertices[p[9]]=center-axis;
  int prism_vertices[4][6] = { 
    {p[9],p[4],p[5],p[8],p[0],p[1]},
    {p[9],p[5],p[6],p[8],p[1],p[2]},
    {p[9],p[6],p[7],p[8],p[2],p[3]},
    {p[9],p[7],p[4],p[8],p[3],p[0]}
  };
  NumPtr<CellData> cells(4);
  for (int c=0;c<4;c++) {
    cells[c].allocate(6);
    for (int v=0;v<6;v++) {
      cells[c].vertex(v)=prism_vertices[c][v];
    }
    cells[c].setMaterialModel(0);
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    REAL an=axis*axis;
    boundary.allocate(3);
    boundary_condition.allocate(3);
    boundary[0]=OPERATOR_NEW CylinderBoundary(axis,center,radius);
    boundary[1]=OPERATOR_NEW PlaneBoundary(center,vertices[p[9]],
      vertices[p[4]],vertices[p[5]]);
    boundary[2]=OPERATOR_NEW PlaneBoundary(center,vertices[p[8]],
      vertices[p[0]],vertices[p[1]]);
    for (int i=0;i<3;i++) {
      boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
    }
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        CODIMENSION1_SHAPE_CLASS *face_element=
          codimension1_level->getShape();
        Point<SPACEDIM> face_center=face_element->center(vertices);
        if (abs(boundary[1]->constraint(face_center))<1.e-5) {
          codimension1_level->
            setBoundaryCondition(boundary_condition[1]);
        } else if (abs(boundary[2]->constraint(face_center))<1.e-5) {
          codimension1_level->
            setBoundaryCondition(boundary_condition[2]);
        } else {
          codimension1_level->
            setBoundaryCondition(boundary_condition[0]);
        }
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
    tess.checkLineBoundaryCondition(boundary_condition);
#endif
  }
  return boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//prisms have quadrateral faces on cylinder boundary of domain
BoundaryData* generateCylinderHexahedra(Tessellation &tess,
const Point<3> &axis,const Point<3> &center,REAL radius,
bool set_face_ids,NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  NumPtr<Point<3> > vertices(16);
  NumPtr<int> p(16);
  RandomPermutation(p); // hopefully get random face numbers on boundary
  Tensor<1,3> e0(Point<3>(1.,0.,0.));
  Tensor<1,3> e1(Point<3>(0.,1.,0.));
  Tensor<1,3> e2(Point<3>(0.,0.,1.));
  Tensor<1,3> cp0,cp1,cp2;
  Tensor<1,3> ta(axis);
  crossProduct(cp0,ta,e0);
  crossProduct(cp1,ta,e1);
  crossProduct(cp2,ta,e2);
  REAL cpn0=cp0*cp0;
  REAL cpn1=cp1*cp1;
  REAL cpn2=cp2*cp2;
  Point<3> cpa,cpb;
  if (cpn0<min(cpn1,cpn2)) {
    cpa=cp1.getValues()*(radius/sqrt(cpn1));
    cpb=cp2.getValues()*(radius/sqrt(cpn2));
  } else if (cpn1<min(cpn2,cpn0)) {
    cpa=cp2.getValues()*(radius/sqrt(cpn2));
    cpb=cp1.getValues()*(radius/sqrt(cpn0));
  } else {
    cpa=cp0.getValues()*(radius/sqrt(cpn0));
    cpb=cp1.getValues()*(radius/sqrt(cpn1));
  }
  REAL sqrt5=sqrt(5.);
  vertices[p[0]]=center+axis+cpa;
  vertices[p[1]]=center+axis+cpb;
  vertices[p[2]]=center+axis-cpa;
  vertices[p[3]]=center+axis-cpb;
  vertices[p[4]]=center+axis+cpa/sqrt5;
  vertices[p[5]]=center+axis+cpb/sqrt5;
  vertices[p[6]]=center+axis-cpa/sqrt5;
  vertices[p[7]]=center+axis-cpb/sqrt5;
  vertices[p[8]]=center-axis+cpa;
  vertices[p[9]]=center-axis+cpb;
  vertices[p[10]]=center-axis-cpa;
  vertices[p[11]]=center-axis-cpb;
  vertices[p[12]]=center-axis+cpa/sqrt5;
  vertices[p[13]]=center-axis+cpb/sqrt5;
  vertices[p[14]]=center-axis-cpa/sqrt5;
  vertices[p[15]]=center-axis-cpb/sqrt5;
  int hexahedra_vertices[5][8] = { 
    {p[4],p[5],p[12],p[13],p[7],p[ 6],p[15],p[14]},
    {p[0],p[1],p[ 8],p[ 9],p[4],p[ 5],p[12],p[13]},
    {p[1],p[2],p[ 9],p[10],p[5],p[ 6],p[13],p[14]},
    {p[2],p[3],p[10],p[11],p[6],p[ 7],p[14],p[15]},
    {p[0],p[3],p[ 4],p[ 7],p[8],p[11],p[12],p[15]}
  };
  NumPtr<CellData> cells(5);
  for (int c=0;c<5;c++) {
    cells[c].allocate(8);
    for (int v=0;v<8;v++) {
      cells[c].vertex(v)=hexahedra_vertices[c][v];
    }
    cells[c].setMaterialModel(0);
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    REAL an=axis*axis;
    boundary.allocate(3);
    boundary_condition.allocate(3);
    boundary[0]=OPERATOR_NEW CylinderBoundary(axis,center,radius);
    boundary[1]=OPERATOR_NEW PlaneBoundary(center,center-axis,
      vertices[p[8]],vertices[p[9]]);
    boundary[2]=OPERATOR_NEW PlaneBoundary(center,center+axis,
      vertices[p[0]],vertices[p[1]]);
    for (int i=0;i<3;i++) {
      boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
    }
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        CODIMENSION1_SHAPE_CLASS *face_element=
          codimension1_level->getShape();
        Point<SPACEDIM> face_center=face_element->center(vertices);
        if (abs(boundary[1]->constraint(face_center))<1.e-5) {
          codimension1_level->
            setBoundaryCondition(boundary_condition[1]);
        } else if (abs(boundary[2]->constraint(face_center))<1.e-5) {
          codimension1_level->
            setBoundaryCondition(boundary_condition[2]);
        } else {
          codimension1_level->
            setBoundaryCondition(boundary_condition[0]);
        }
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
    tess.checkLineBoundaryCondition(boundary_condition);
#endif
  }
  return boundary_data;
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
BoundaryData* generateBallSectorTriangles(Tessellation &tess,
const Point<2> &center,REAL radius,REAL angle,bool set_face_ids,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  CHECK_TEST(radius>0.);
  CHECK_TEST(0.<angle && angle<=2.*M_PI);
  NumPtr<Point<2> > vertices;
  NumPtr<CellData> cells;
  if (angle<=0.5*M_PI) {
    vertices.allocate(3);
    cells.allocate(1);
    vertices[0]=center;
    vertices[1]=center+Point<2>( radius,0.);
    vertices[2]=center+Point<2>( radius*cos(angle),radius*sin(angle) );
    int cell_vertices[1][3] = { {0,1,2} };
    cells[0].allocate(3);
    for (int v=0;v<3;v++) {
      cells[0].vertex(v)=cell_vertices[0][v];
    }
    cells[0].setMaterialModel(0);
  } else if (angle<=M_PI) {
    vertices.allocate(4);
    cells.allocate(2);
    vertices[0]=center;
    vertices[1]=center+Point<2>( radius,0.);
    vertices[2]=
      center+Point<2>( radius*cos(angle*0.5),radius*sin(angle*0.5) );
    vertices[3]=center+Point<2>( radius*cos(angle),radius*sin(angle) );
    int cell_vertices[2][3] = { {0,1,2},{0,2,3} };
    for (int c=0;c<2;c++) {
      cells[c].allocate(3);
      for (int v=0;v<3;v++) {
        cells[c].vertex(v)=cell_vertices[c][v];
      }
      cells[c].setMaterialModel(0);
    }
  } else if (angle<=1.5*M_PI) {
    vertices.allocate(5);
    cells.allocate(3);
    vertices[0]=center;
    vertices[1]=center+Point<2>( radius,0.);
    vertices[2]=center+Point<2>(radius*cos(angle/3.),radius*sin(angle/3.));
    vertices[3]=
      center+Point<2>(radius*cos(2.*angle/3.),radius*sin(2.*angle/3.));
    vertices[4]=center+Point<2>( radius*cos(angle),radius*sin(angle) );
    int cell_vertices[3][3] = { {0,1,2},{0,2,3}, {0,3,4} };
    for (int c=0;c<3;c++) {
      cells[c].allocate(3);
      for (int v=0;v<3;v++) {
        cells[c].vertex(v)=cell_vertices[c][v];
      }
      cells[c].setMaterialModel(0);
    }
  } else {
    vertices.allocate(6);
    cells.allocate(4);
    vertices[0]=center;
    vertices[1]=center+Point<2>( radius,0.);
    vertices[2]=
      center+Point<2>(radius*cos(angle*0.25),radius*sin(angle*0.25));
    vertices[3]=
      center+Point<2>(radius*cos(angle*0.5),radius*sin(angle*0.5));
    vertices[4]=
      center+Point<2>(radius*cos(angle*0.75),radius*sin(angle*0.75));
    vertices[5]=center+Point<2>( radius*cos(angle),radius*sin(angle) );
    int cell_vertices[4][3] = { {0,1,2},{0,2,3},{0,3,4},{0,4,5} };
    for (int c=0;c<4;c++) {
      cells[c].allocate(3);
      for (int v=0;v<3;v++) {
        cells[c].vertex(v)=cell_vertices[c][v];
      }
      cells[c].setMaterialModel(0);
    }
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(3);
    boundary_condition.allocate(3);
    boundary[0]=OPERATOR_NEW BallBoundary(center,radius);
    Point<2> point_inside_domain(cos(angle*0.5),sin(angle*0.5));
    point_inside_domain*=radius;
    boundary[1]=OPERATOR_NEW PlaneBoundary(point_inside_domain,
      center,vertices[1]);
    boundary[2]=OPERATOR_NEW PlaneBoundary(point_inside_domain,
      center,vertices[vertices.getNumber()-1]);
    for (int i=0;i<3;i++) {
      boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
    }
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        CODIMENSION1_SHAPE_CLASS *face_element=
          codimension1_level->getShape();
        Point<SPACEDIM> face_center=face_element->center(vertices);
        if (abs(boundary[1]->constraint(face_center))<1.e-5) {
          codimension1_level->
            setBoundaryCondition(boundary_condition[1]);
        } else if (abs(boundary[2]->constraint(face_center))<1.e-5) {
          codimension1_level->
            setBoundaryCondition(boundary_condition[2]);
        } else {
          codimension1_level->
            setBoundaryCondition(boundary_condition[0]);
        }
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
#endif
  }
  return boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
BoundaryData* generateBallSectorTetrahedra(Tessellation &tess,
Point<3> axis,const Point<3> &center,REAL radius,REAL angle,
bool set_face_ids,NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  CHECK_TEST(radius>0.);
  CHECK_TEST(0.<angle && angle<=M_PI);
  REAL an=sqrt(axis.square());
  CHECK_TEST(an>0.);
  axis*=radius/an;
  Tensor<1,3> e0(Point<3>(1.,0.,0.));
  Tensor<1,3> e1(Point<3>(0.,1.,0.));
  Tensor<1,3> e2(Point<3>(0.,0.,1.));
  Tensor<1,3> cp0,cp1,cp2;
  Tensor<1,3> ta(axis);
  crossProduct(cp0,ta,e0);
  crossProduct(cp1,ta,e1);
  crossProduct(cp2,ta,e2);
  REAL cpn0=cp0*cp0;
  REAL cpn1=cp1*cp1;
  REAL cpn2=cp2*cp2;
  Point<3> cpa,cpb;
  if (cpn0<min(cpn1,cpn2)) {
    cpa=cp1.getValues()*(radius/sqrt(cpn1));
    cpb=cp2.getValues()*(radius/sqrt(cpn2));
  } else if (cpn1<min(cpn2,cpn0)) {
    cpa=cp2.getValues()*(radius/sqrt(cpn2));
    cpb=cp1.getValues()*(radius/sqrt(cpn0));
  } else {
    cpa=cp0.getValues()*(radius/sqrt(cpn0));
    cpb=cp1.getValues()*(radius/sqrt(cpn1));
  }

  NumPtr<Point<3> > vertices;
  NumPtr<CellData> cells;
  if (angle<=0.5*M_PI) {
    vertices.allocate(6);
    cells.allocate(4);
    vertices[0]=center;
    vertices[1]=center+axis;
    axis*=cos(angle);
    cpa*=sin(angle);
    cpb*=sin(angle);
    vertices[2]=center+axis+cpa;
    vertices[3]=center+axis+cpb;
    vertices[4]=center+axis-cpa;
    vertices[5]=center+axis-cpb;
    int cell_vertices[4][4] = { 
      {0,1,2,3}, {0,1,3,4}, {0,1,4,5}, {0,1,5,2} 
    };
    for (int c=0;c<4;c++) {
      cells[c].allocate(4);
      for (int v=0;v<4;v++) {
        cells[c].vertex(v)=cell_vertices[c][v];
      }
      cells[c].setMaterialModel(0);
    }
  } else {
    vertices.allocate(10);
    cells.allocate(12);
    vertices[0]=center;
    vertices[1]=center+axis;
    vertices[2]=center+cpa;
    vertices[3]=center+cpb;
    vertices[4]=center-cpa;
    vertices[5]=center-cpb;
    axis*=cos(angle);
    Point<3> cpc=(cpb+cpa)*sin(angle)/sqrt(2.);
    Point<3> cpd=(cpb-cpa)*sin(angle)/sqrt(2.);
    vertices[6]=center+axis+cpc;
    vertices[7]=center+axis+cpd;
    vertices[8]=center+axis-cpc;
    vertices[9]=center+axis-cpd;
    int cell_vertices[12][4] = { 
      {0,1,2,3}, {0,1,3,4}, {0,1,4,5}, {0,1,5,2},
      {0,2,6,3}, {0,3,7,4}, {0,4,8,5}, {0,2,5,9},
      {0,2,9,6}, {0,3,6,7}, {0,4,7,8}, {0,5,8,9}
    };
    for (int c=0;c<12;c++) {
      cells[c].allocate(4);
      for (int v=0;v<4;v++) {
        cells[c].vertex(v)=cell_vertices[c][v];
      }
      cells[c].setMaterialModel(0);
    }
  }
  BoundaryData *boundary_data=tess.create(vertices,cells);
  if (set_face_ids) {
    boundary.allocate(2);
    boundary_condition.allocate(2);
    boundary[0]=OPERATOR_NEW BallBoundary(center,radius);
    Point<3> axis(0.,0.,1.);
    boundary[1]=OPERATOR_NEW ConeBoundary(axis,center,angle);
    for (int i=0;i<2;i++) {
      boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
    }
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
//      replace bogus_boundary_condition in Tessellation::create
        CODIMENSION1_SHAPE_CLASS *face_element=
          codimension1_level->getShape();
        Point<3> v0=vertices[face_element->vertexIndex(0)];
        Point<3> v1=vertices[face_element->vertexIndex(1)];
        Point<3> v2=vertices[face_element->vertexIndex(2)];
        REAL r2=min(v0.square(),min(v1.square(),v2.square()));
        if (r2>radius*radius*0.9) {
          codimension1_level->
            setBoundaryCondition(boundary_condition[0]);
        } else {
          codimension1_level->
            setBoundaryCondition(boundary_condition[1]);
        }
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
    tess.checkLineBoundaryCondition(boundary_condition);
#endif
  }
  return boundary_data;
}
#endif

#include "NumPtr.C"
INSTANTIATE_NUMPTR( Boundary* );
INSTANTIATE_NUMPTR(BoundaryCondition*)
