#ifndef GENERATE_TESS_H
#define GENERATE_TESS_H "$Header:$"
//---------------------------  grid_generator.h  ---------------------------
//   $Id: grid_generator.h,v 1.36 2003/04/07 15:41:14 wolf Exp $
//   Version: $Name: Version-4-0-0 $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//----------------------------  grid_generator.h  --------------------------
//
//modified from deal.II/deal.II/include/grid/grid_generator.h
//  by John Trangentein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <Tessellation.H>

//PlaneBoundary:
//generates Intervals/Quadrilaterals/Hexahedra
// SubCellData::boundary_faces arg:
//   .--3--.
//   |     |
//   0     1 Quadrilateral (2D)
//   |     |
//   .--2--.

//           .-----------------.
//          /|                /|
//         / |               / |
//        /  |     /        /  |
//       /   |   -5-       /   |
//      /    |   /    |   /    |
//     /     |       -0- /     |
//    /      |        | /      |
//   .-----------------.       |
//   |       |         |       |
//   |   |/  |         |   |/  | Hexahedron (3D)
//   |   2   |         |   3   |
//   |  /|   .---------|--/|---.
//   |      / |        |      /
//   |     / -1-       |     /
//   |    /   |    /   |    /
//   |   /       -4-   |   /
//   |  /        /     |  /
//   | /               | /
//   |/                |/
//   .-----------------.
BoundaryData* generateParallelpiped(Tessellation&,
  const Point<SPACEDIM>&,const Point<SPACEDIM> edge[SPACEDIM],
  const int[SPACEDIM],bool,NumPtr<Boundary*>&,
  NumPtr<BoundaryCondition*>&);
#if (SPACEDIM==2)
// BoundaryData::boundary_faces arg:
//   .--3--.
//   |\   /|
//   0  X  1 
//   |/   \|
//   .--2--.
BoundaryData* generateParallelpipedTriangles(Tessellation&,
  const Point<SPACEDIM>&,const Point<SPACEDIM> edge[SPACEDIM],
  const int[SPACEDIM],bool,NumPtr<Boundary*>&,NumPtr<BoundaryCondition*>&);
#endif
#if (SPACEDIM==3)
BoundaryData* generateParallelpipedPrisms(Tessellation&,
  const Point<SPACEDIM>&,const Point<SPACEDIM> edge[SPACEDIM],
  const int[SPACEDIM],int,bool,NumPtr<Boundary*>&,
  NumPtr<BoundaryCondition*>&);
BoundaryData* generateParallelpipedTetrahedra(Tessellation&,
  const Point<SPACEDIM>&,const Point<SPACEDIM> edge[SPACEDIM],
  const int[SPACEDIM],bool,NumPtr<Boundary*>&,NumPtr<BoundaryCondition*>&);
#endif
#if (SPACEDIM==2)
// BoundaryData::boundary_faces arg:
//   .----6----.
//   |         2
//   |      5  |
//   0    .----.
//   |      4  1
//   |         |
//   .----3----.
BoundaryData* generateSlitQuadrilateral(Tessellation&,bool,
  NumPtr<Boundary*>&,NumPtr<BoundaryCondition*>&);
#endif
#if (SPACEDIM>=2)
// BoundaryData::boundary_faces arg:
//   .--3-.
//   |    4
//   |    |
//   0    .-5--.
//   |         1
//   |         |
//   .----2----.

//           .-----------------.
//          /|   /     |      /|
//         / | -5-    -0-    / |
//        /  | /       |    /  |
//       /   |    .--------/   |
//      /    |   /|        |   |
//     /     |  / |   |    |   |
//    /      | /  |  -6-   |   |
//   .--------- |/|   |    |   |
//   |       || 7 |        |   |
//   |   |/  ||/| .--------. |/|
//   |   2   ||  /   /    /  3 |
//   |  /|   .|-/---8----/--/|-.
//   |      / |/   /    /     /
//   |     /  .--------.     /
//   |    /        /   |    /
//   |   /       -4-   |   /
//   |  /        /     |  /
//   | /               | /
//   |/                |/
//   .-----------------.
BoundaryData* generateL(Tessellation&,bool,NumPtr<Boundary*>&,
  NumPtr<BoundaryCondition*>&);
#endif

//BallBoundary:
// BoundaryData::boundary_faces arg = 0
#if (SPACEDIM>=2)
//generates Quadrilaterals/Hexahedra
BoundaryData* generateBall(Tessellation&,const Point<SPACEDIM>&,REAL,
  bool,NumPtr<Boundary*>&,NumPtr<BoundaryCondition*>&);
#endif
#if (SPACEDIM==2)
BoundaryData* generateBallTriangles(Tessellation&,const Point<SPACEDIM>&,
  REAL,bool,NumPtr<Boundary*>&,NumPtr<BoundaryCondition*>&);
#endif
#if (SPACEDIM==3)
BoundaryData* generateBallTetrahedra(Tessellation&,
  const Point<SPACEDIM>&,REAL,bool,NumPtr<Boundary*>&,
  NumPtr<BoundaryCondition*>&);
BoundaryData* generateBallPrismsWithTetrahedra(Tessellation&,
  const Point<SPACEDIM>&,REAL,bool,NumPtr<Boundary*>&,
  NumPtr<BoundaryCondition*>&);
BoundaryData* generateBallPrismsWithHexahedra(Tessellation&,
  const Point<SPACEDIM>&,REAL,bool,NumPtr<Boundary*>&,
  NumPtr<BoundaryCondition*>&);
#endif

#if (SPACEDIM==3)
//CylinderBoundary & 2 PlaneBoundary's:
// CylincerBoundary: BoundaryData::boundary_faces arg = 0
// PlaneBoundary: BoundaryData::boundary_faces arg = 1,2
BoundaryData* generateCylinderTetrahedra(Tessellation&,
  const Point<SPACEDIM>&,const Point<SPACEDIM>&,REAL,bool,
  NumPtr<Boundary*>&,NumPtr<BoundaryCondition*>&);
BoundaryData* generateCylinderPrisms(Tessellation&,
  const Point<SPACEDIM>&,const Point<SPACEDIM>&,REAL,bool,
  NumPtr<Boundary*>&,NumPtr<BoundaryCondition*>&);
BoundaryData* generateCylinderHexahedra(Tessellation&,
  const Point<SPACEDIM>&,const Point<SPACEDIM>&,REAL,bool,
  NumPtr<Boundary*>&,NumPtr<BoundaryCondition*>&);
#endif

//BallBoundary & ConeBoundary:
// BallBoundary: BoundaryData::boundary_faces arg = 0
// ConeBoundary: BoundaryData::boundary_faces arg = 1
#if (SPACEDIM==2)
BoundaryData* generateBallSectorTriangles(Tessellation&,
  const Point<SPACEDIM>&,REAL,REAL,bool,NumPtr<Boundary*>&,
  NumPtr<BoundaryCondition*>&);
#endif
#if (SPACEDIM==3)
BoundaryData* generateBallSectorTetrahedra(Tessellation&,
  Point<SPACEDIM>,const Point<SPACEDIM>&,REAL,REAL,bool,
  NumPtr<Boundary*>&,NumPtr<BoundaryCondition*>&);
#endif

#endif
