//#include <float.h>
//#include <fstream>
#include <iostream>
#include <stdlib.h>
#include "Boundary.H"
#include "DDecl.H"
#include "GenerateTess.H"
#include "GUIEnumInputParameter.H"
#include "GUIVirtualInput.H"
#include "InputParameter.H"
#include "Tessellation.H"
#include "Tracer.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKGUI.H"
#include "GTKGUIVirtualInput.H"
#include "GTKWindow.H"
#else
#include "GUI.H"
#include "GUIVirtualInput.H"
#endif

GUI_INPUT_PARAMETER_LIST_TYPE *param_list=0;
BOOLEAN skip_gui=FALSE;
double winsize=0.5;
char* display_name=0;

#if (SPACEDIM==1)
#elif (SPACEDIM==2)
enum DOMAIN_TYPE{PARALLELPIPED_DOMAIN,BALL_DOMAIN,SLIT_DOMAIN,L_DOMAIN,
  BALL_SECTOR_DOMAIN};
static const char *domain_name[5]=
  {"parallelpiped domain","ball domain","slit domain","L domain",
   "ball sector domain"};
enum SHAPE{TRIANGLE_SHAPE,QUADRILATERAL_SHAPE};
static const char *shape_name[2]={"triangle","quadrilateral"};
SHAPE shape=QUADRILATERAL_SHAPE;
#else
enum DOMAIN_TYPE{PARALLELPIPED_DOMAIN,BALL_DOMAIN,L_DOMAIN,
  CYLINDER_DOMAIN,BALL_SECTOR_DOMAIN};
static const char *domain_name[5]=
  {"parallelpiped domain","ball domain","L domain","cylinder domain",
   "ball sector domain"};
enum SHAPE{TETRAHEDRON_SHAPE,PRISM_SHAPE,HEXAHEDRON_SHAPE};
static const char *shape_name[3]={"tetrahedron","prism","hexahedron"};
SHAPE shape=HEXAHEDRON_SHAPE;
SHAPE second_shape=HEXAHEDRON_SHAPE;
int isecond_shape=second_shape;
#endif
#if (SPACEDIM>1)
DOMAIN_TYPE domain=PARALLELPIPED_DOMAIN;
int idomain=domain;
int ishape=shape;
#endif

int number_refinements=1;
Point<SPACEDIM> low(D_DECL(0.,0.,0.));
Point<SPACEDIM> edge[SPACEDIM];
int number_elements[SPACEDIM]={D_DECL(1,1,1)};
#if (SPACEDIM>1)
Point<SPACEDIM> center(D_DECL(0.,0.,0.));
double radius=1.;
double angle=M_PI*0.25;
#endif
#if (SPACEDIM==3)
Point<SPACEDIM> axis(D_DECL(0.,0.,1.));
int dir=0;
#endif

//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void checkMainInput() {
//TRACER_CALL(tr,"checkMainInput");
#if (SPACEDIM>1)
  domain=static_cast<DOMAIN_TYPE>(idomain);
  shape=static_cast<SHAPE>(ishape);
#endif
#if (SPACEDIM==3)
  second_shape=(isecond_shape==TETRAHEDRON_SHAPE ? TETRAHEDRON_SHAPE :
    HEXAHEDRON_SHAPE);
#endif
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void runMain(BOOLEAN /*called_before*/) {
//TRACER_CALL(tr,"runMain");
  Tessellation tess;
  bool set_face_ids=true;
  BoundaryData *sub_cell_data=0;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
#if (SPACEDIM==1)
  sub_cell_data=generateParallelpiped(tess,low,edge,number_elements,
    set_face_ids,boundary,boundary_condition);
#elif (SPACEDIM==2)
  switch (domain) {
    case BALL_DOMAIN: {
      switch (shape) {
        case TRIANGLE_SHAPE:
          sub_cell_data=generateBallTriangles(tess,center,radius,
            set_face_ids,boundary,boundary_condition);
          break;
        default:
          sub_cell_data=generateBall(tess,center,radius,set_face_ids,
            boundary,boundary_condition);
      }
//    if (number_refinements>0) {
//      BallBoundary *bb=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,bb);
//    }
      break;
    }
    case SLIT_DOMAIN: {
      sub_cell_data=generateSlitQuadrilateral(tess,set_face_ids,
        boundary,boundary_condition);
//    if (number_refinements>0) {
//      Point<2> centroid(0.25,0.25);
//      for (int i=0;i<5;i++) {
//        PlaneBoundary *sb=OPERATOR_NEW PlaneBoundary(centroid);
//        tess.setBoundary(i,sb);
//      }
//      centroid[1]=0.75;
//      for (int i=5;i<7;i++) {
//        PlaneBoundary *sb=OPERATOR_NEW PlaneBoundary(centroid);
//        tess.setBoundary(i,sb);
//      }
//    }
      break;
    }
    case L_DOMAIN: {
      sub_cell_data=generateL(tess,set_face_ids,boundary,
        boundary_condition);
//    if (number_refinements>0) {
//      Point<2> centroid(0.25,0.25);
//      for (int i=0;i<6;i++) {
//        PlaneBoundary *sb=OPERATOR_NEW PlaneBoundary(centroid);
//        tess.setBoundary(i,sb);
//      }
//    }
      break;
    }
    case BALL_SECTOR_DOMAIN: {
//    TRACER_CALL(tr,"runMain BALL_SECTOR_DOMAIN");
      sub_cell_data=
        generateBallSectorTriangles(tess,center,radius,angle,
          set_face_ids,boundary,boundary_condition);
//    if (number_refinements>0) {
//      BallBoundary *bb=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,bb);
//      REAL ang2=angle*0.5;
//      Point<2> axis(cos(ang2),sin(ang2));
//      ConeBoundary *cb=OPERATOR_NEW ConeBoundary(axis,center,ang2);
//      tess.setBoundary(1,cb);
//    }
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      break;
    }
    default: { // PARALLELPIPED_DOMAIN
      switch (shape) {
        case TRIANGLE_SHAPE:
          sub_cell_data=generateParallelpipedTriangles(tess,low,edge,
            number_elements,set_face_ids,boundary,boundary_condition);
          break;
        default:
          sub_cell_data=generateParallelpiped(tess,low,edge,
            number_elements,set_face_ids,boundary,boundary_condition);
      }
//    if (number_refinements>0) {
//      Point<SPACEDIM> centroid(low+(edge[0]+edge[1])*0.5);
//      for (int i=0;i<4;i++) {
//        PlaneBoundary *sb=OPERATOR_NEW PlaneBoundary(centroid);
//        tess.setBoundary(i,sb);
//      }
//    }
    }
  }
#else
  switch (domain) {
    case BALL_DOMAIN: {
      switch (shape) {
        case TETRAHEDRON_SHAPE:
          sub_cell_data=
            generateBallTetrahedra(tess,center,radius,set_face_ids,
              boundary,boundary_condition);
          break;
        case PRISM_SHAPE: {
          switch (second_shape) {
            case TETRAHEDRON_SHAPE:
              sub_cell_data=generateBallPrismsWithTetrahedra(tess,
                center,radius,set_face_ids,boundary,boundary_condition);
              break;
            default:
              sub_cell_data=generateBallPrismsWithHexahedra(tess,center,
                radius,set_face_ids,boundary,boundary_condition);
          }
          break;
        }
        default:
          sub_cell_data=generateBall(tess,center,radius,set_face_ids,
            boundary,boundary_condition);
      }
//    if (number_refinements>0) {
//      BallBoundary *bb=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,bb);
//    }
      break;
    }
    case L_DOMAIN: {
      sub_cell_data=generateL(tess,set_face_ids,boundary,
        boundary_condition);
//    if (number_refinements>0) {
//      Point<3> centroid(0.25,0.25,0.25);
//      for (int i=0;i<9;i++) {
//        PlaneBoundary *sb=OPERATOR_NEW PlaneBoundary(centroid);
//        tess.setBoundary(i,sb);
//      }
//    }
      break;
    }
    case CYLINDER_DOMAIN: {
      switch (shape) {
        case TETRAHEDRON_SHAPE:
          sub_cell_data=generateCylinderTetrahedra(tess,axis,center,
            radius,set_face_ids,boundary,boundary_condition);
          break;
        case PRISM_SHAPE: {
          sub_cell_data=
            generateCylinderPrisms(tess,axis,center,radius,
              set_face_ids,boundary,boundary_condition);
          break;
        }
        default:
          sub_cell_data=generateCylinderHexahedra(tess,axis,center,
            radius,set_face_ids,boundary,boundary_condition);
      }
//    if (number_refinements>0) {
//      CylinderBoundary *cb=
//        OPERATOR_NEW CylinderBoundary(axis,center,radius);
//      tess.setBoundary(0,cb);
//      for (int i=1;i<3;i++) {
//        PlaneBoundary *pb=OPERATOR_NEW PlaneBoundary(center);
//        tess.setBoundary(i,pb);
//      }
//    }
      break;
    }
    case BALL_SECTOR_DOMAIN: {
//    TRACER_CALL(tr,"runMain BALL_SECTOR_DOMAIN");
      sub_cell_data=generateBallSectorTetrahedra(tess,axis,center,
        radius,angle,set_face_ids,boundary,boundary_condition);
//    if (number_refinements>0) {
//      BallBoundary *bb=OPERATOR_NEW BallBoundary(center,radius);
//      tess.setBoundary(0,bb);
//      ConeBoundary *cb=OPERATOR_NEW ConeBoundary(axis,center,angle);
//      tess.setBoundary(1,cb);
//    }
      break;
    }
    default: { // PARALLELPIPED_DOMAIN
      switch (shape) {
        case TETRAHEDRON_SHAPE:
          sub_cell_data=generateParallelpipedTetrahedra(tess,low,edge,
            number_elements,set_face_ids,boundary,boundary_condition);
          break;
        case PRISM_SHAPE:
          sub_cell_data=generateParallelpipedPrisms(tess,low,edge,
            number_elements,dir,set_face_ids,boundary,boundary_condition);
          break;
        default:
          sub_cell_data=generateParallelpiped(tess,low,edge,
            number_elements,set_face_ids,boundary,boundary_condition);
      }
//    if (number_refinements>0) {
//      Point<SPACEDIM> centroid(low+(edge[0]+edge[1])*0.5);
//      for (int i=0;i<6;i++) {
//        PlaneBoundary *sb=OPERATOR_NEW PlaneBoundary(centroid);
//        tess.setBoundary(i,sb);
//      }
//    }
    }
  }
#endif
  tess.plotActiveShapes();
#if (SPACEDIM==3)
  if (domain==BALL_DOMAIN) tess.plotActiveShapesIndividually();
#endif
  for (int r=0;r<number_refinements;r++) {
    tess.refineGlobal(1);
    tess.executeCoarseningAndRefinement();
    tess.plotActiveShapes();
  }
  delete sub_cell_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void cleanup() {
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void shutdown() {
  while (param_list->notEmpty()) delete param_list->delAfter(0);
  delete param_list; param_list=0;
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int main(int argc,char* argv[]) {
#ifdef DEBUG
//setTraps();
#endif
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  {
    for (int id=0;id<SPACEDIM;id++) {
      for (int d=0;d<SPACEDIM;d++) edge[id][d]=0.;
      edge[id][id]=1.;
    }
//  define input parameters and bounds
    param_list=OPERATOR_NEW GUI_INPUT_PARAMETER_LIST_TYPE("Main List");
    { const char *group="Grid Generation Parameters";
#if (SPACEDIM==1)
//    param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
//      idomain,"domain",domain_name[0]));
//    param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
//      ishape,"shape",shape_name[0]));
#elif (SPACEDIM==2)
      param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
        idomain,"domain",domain_name[0],domain_name[1],domain_name[2],
        domain_name[3],domain_name[4]));
      param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
        ishape,"shape",shape_name[0],shape_name[1]));
#else
      param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
        idomain,"domain",domain_name[0],domain_name[1],domain_name[2],
        domain_name[3],domain_name[4]));
      param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
        ishape,"shape",shape_name[0],shape_name[1],shape_name[2]));
      param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
        isecond_shape,"second shape",shape_name[0],shape_name[2]));
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(
        dir,"dir",0,2,group));
#endif

      param_list->append(OPERATOR_NEW GUIInputParameter<int>(
        number_refinements,"number_refinements",0,INT_MAX,group));

      param_list->append(OPERATOR_NEW GUIInputParameter<int>(
        number_elements[0],"number elements[0]",1,INT_MAX,group));
#if (SPACEDIM>1)
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(
        number_elements[1],"number elements[1]",1,INT_MAX,group));
#endif
#if (SPACEDIM==3)
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(
        number_elements[2],"number elements[2]",1,INT_MAX,group));
#endif

#if (SPACEDIM>1)
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        radius,"radius",0.,DBL_MAX,group));
# if (SPACEDIM==2)
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        angle,"angle",0.,M_PI*2.,group));
# endif
# if (SPACEDIM==3)
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        angle,"angle",0.,M_PI,group));
# endif
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        center[0],"center[0]",-DBL_MAX,DBL_MAX,group));
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        center[1],"center[1]",-DBL_MAX,DBL_MAX,group));
# if (SPACEDIM==3)
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        center[2],"center[2]",-DBL_MAX,DBL_MAX,group));
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        axis[0],"axis[0]",-DBL_MAX,DBL_MAX,group));
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        axis[1],"axis[1]",-DBL_MAX,DBL_MAX,group));
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        axis[2],"axis[2]",-DBL_MAX,DBL_MAX,group));
# endif
#endif

      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        low[0],"low[0]",-DBL_MAX,DBL_MAX,group));
#if (SPACEDIM>1)
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        low[1],"low[1]",-DBL_MAX,DBL_MAX,group));
#endif
#if (SPACEDIM==3)
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        low[2],"low[2]",-DBL_MAX,DBL_MAX,group));
#endif

      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        edge[0][0],"edge0[0]",0.,DBL_MAX,group));
#if (SPACEDIM>1)
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        edge[0][1],"edge0[1]",0.,DBL_MAX,group));
# if (SPACEDIM==3)
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        edge[0][2],"edge0[2]",0.,DBL_MAX,group));
# endif
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        edge[1][0],"edge1[0]",0.,DBL_MAX,group));
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        edge[1][1],"edge1[1]",0.,DBL_MAX,group));
# if (SPACEDIM==3)
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        edge[1][2],"edge1[2]",0.,DBL_MAX,group));
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        edge[2][0],"edge2[0]",0.,DBL_MAX,group));
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        edge[2][1],"edge2[1]",0.,DBL_MAX,group));
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        edge[2][2],"edge2[2]",0.,DBL_MAX,group));
# endif
#endif
    }
    { const char *group="Graphics Parameters";
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(winsize,
        "winsize",0.,1.,group) );
    }

//  read input parameters 
    if (argc>1) {
      ifstream in_file;
      in_file.open(argv[1],ios::in);
      if (in_file) {
        istream &infile(in_file);
        infile.clear(ios::goodbit);
        infile.seekg(0,ios::beg);
        char name[LENGTH_NAME], comment[LENGTH_COMMENT];
        BOOLEAN found_main=FALSE;
        while ( in_file >> setw(LENGTH_NAME) >> name ) {
          if ( strcmp(name,"Main") == 0 ) {
            in_file.getline( comment, LENGTH_COMMENT);
            found_main=TRUE;
            while ( in_file >> setw(LENGTH_NAME) >> name ) {
              if ( strcmp(name,"end") == 0 ) break;
              else if (strcmp(name,"skip_gui") == 0) {
                int iskip_gui; // type bool not read correctly
                in_file >> iskip_gui;
                skip_gui= iskip_gui!=0;
              } else param_list->formattedRead(in_file,name);
              in_file.getline( comment, LENGTH_COMMENT);
            }
          } else in_file.getline(comment,LENGTH_COMMENT);
        }
        if ( !found_main ) skip_gui=FALSE;
      }
      if (skip_gui) checkMainInput();
    } else skip_gui=FALSE;

    if (skip_gui) {
      runMain(FALSE);
      cleanup();
      shutdown();
    } else {
#ifdef USE_GTK
      GTKGUI gui(argv[0],display_name,param_list,&runMain,
      &checkMainInput,&cleanup,&shutdown,TRUE);
#else
      GUI gui(argv[0],display_name,param_list,&runMain,&checkMainInput,
      &cleanup,&shutdown,TRUE);
#endif
      gui.createFileMenu();
      gui.createViewMenu();
      gui.createHelpMenu();
      gui.createMainWindow(argc,argv);
      gui.eventLoop();
    }
  } // scope of MemoryDebugger
  return EXIT_SUCCESS;
}

template class InputParameter<int>;
template class InputParameter<double>;
template class InputParameter<bool>;
