#ifndef HCURL_DOF_LEVEL_H
#define HCURL_DOF_LEVEL_H "$Header:$"

#include "DofLevel.H"
#include "NumPtr.H"
#include "Shape.H"

//static const int invalid_dof_index=INT_MAX;
class HcurlDofLevel {
  protected:
    NumPtr<int> dofs;
    NumPtr<int> codimension1_dofs;
#if (SPACEDIM==3)
    NumPtr<int> codimension2_dofs;
#endif
    NumPtr<int> multiplier_dofs;
//  NumPtr<int> multiplier_codimension1_dofs;
//  map<int,int> face_element_data_index_to_multiplier_index_map;
    map<int,int> element_data_index_to_multiplier_index_map;
  public:
//the arguments to the constructor are set in HcurlDofHandler::reserveSpace
#if (SPACEDIM<=2)
    HcurlDofLevel(int cd1n,int n,int mn,const map<int,int> &edi2mim) :
    codimension1_dofs(cd1n),dofs(n), multiplier_dofs(mn),
    element_data_index_to_multiplier_index_map(edi2mim) {
      codimension1_dofs.initialize(invalid_dof_index);
      dofs.initialize(invalid_dof_index);
      multiplier_dofs.initialize(invalid_dof_index);
    }
#else
    HcurlDofLevel(int cd2n,int cd1n,int n,int mn,
    const map<int,int> &edi2mim) : codimension2_dofs(cd2n),
    codimension1_dofs(cd1n),dofs(n),multiplier_dofs(mn),
    element_data_index_to_multiplier_index_map(edi2mim) {
      codimension2_dofs.initialize(invalid_dof_index);
      codimension1_dofs.initialize(invalid_dof_index);
      dofs.initialize(invalid_dof_index);
      multiplier_dofs.initialize(invalid_dof_index);
    }
#endif
    virtual ~HcurlDofLevel() {}
    int numberDofs() const { return dofs.getNumber(); }
    int numberCodimension1Dofs() const { 
      return codimension1_dofs.getNumber(); 
    }
    int numberMultiplierDofs() const { return multiplier_dofs.getNumber(); }
//  int numberMultiplierCodimension1Dofs() const { 
//    return multiplier_codimension1_dofs.getNumber();
//  }
#if (SPACEDIM==3)
    int numberCodimension2Dofs() const { 
      return codimension2_dofs.getNumber(); 
    }
#endif
    void getDofs(int dof_index,NumPtr<int> &indices) const { 
      indices.copyFrom(dofs+dof_index);
    }
    void getCodimension1Dofs(int dof_index,NumPtr<int> &indices) const { 
      indices.copyFrom(codimension1_dofs+dof_index);
    }
#if (SPACEDIM==3)
    void getCodimension2Dofs(int dof_index,NumPtr<int> &indices) const { 
      indices.copyFrom(codimension2_dofs+dof_index);
    }
#endif
    void getMultiplierDofs(int element_data_index,NumPtr<int> &indices)
    const {
      map<int,int>::const_iterator edi2mim_iterator=
        element_data_index_to_multiplier_index_map.
        find(element_data_index);
      CHECK_TEST(edi2mim_iterator!=
                 element_data_index_to_multiplier_index_map.end());
      indices.copyFrom(multiplier_dofs+edi2mim_iterator->second);
    }
/*
    void getMultiplierCodimension1Dofs(int face_element_data_index,
    NumPtr<int> &indices) const { 
      map<int,int>::const_iterator fedi2mim_iterator=
        face_element_data_index_to_multiplier_index_map.
        find(face_element_data_index);
      CHECK_TEST(fedi2mim_iterator!=
                 face_element_data_index_to_multiplier_index_map.end());
      indices.copyFrom(multiplier_dofs+fedi2mim_iterator->second);
    }
*/
    void setDofs(const NumPtr<int>&);
    void setDofs(int,int,int&);
    void setCodimension1Dofs(const NumPtr<int>&);
    void setCodimension1Dofs(int,int,int&);
#if (SPACEDIM==3)
    void setCodimension2Dofs(const NumPtr<int>&);
    void setCodimension2Dofs(int,int,int&);
#endif
    void setMultiplierDofs(const NumPtr<int>&);
    void setMultiplierDofs(int,int,int&);
//  void setMultiplierCodimension1Dofs(const NumPtr<int>&);
//  void setMultiplierCodimension1Dofs(int,int,int&);
    virtual void printOn(ostream&) const;
};

#endif
