//"$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include "HdivDofLevel.H"
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofLevel::setDofs(const NumPtr<int> &new_numbers) {
  int n=dofs.getNumber();
  for (int i=0;i<n;i++) {
    int &di=dofs[i];
    if (di!=invalid_dof_index) di=new_numbers[di];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofLevel::setDofs(int dof_index,int dps,
int &next_free_dof) {
  NumPtr<int> edn(dofs,dof_index,dps);
  if (edn[0]==invalid_dof_index) {
    for (int d=0;d<dps;d++,next_free_dof++) edn[d]=next_free_dof;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofLevel::setCodimension1Dofs(const NumPtr<int> &new_numbers) {
  int n=codimension1_dofs.getNumber();
  for (int i=0;i<n;i++) {
    int &di=codimension1_dofs[i];
    if (di!=invalid_dof_index) di=new_numbers[di];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofLevel::setCodimension1Dofs(int dof_index,int dpcs,
int &next_free_dof) {
  NumPtr<int> edn(codimension1_dofs,dof_index,dpcs);
  if (edn[0]==invalid_dof_index) {
    for (int d=0;d<dpcs;d++,next_free_dof++) edn[d]=next_free_dof;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofLevel::setMultiplierDofs(const NumPtr<int> &new_numbers) {
  int n=multiplier_dofs.getNumber();
  for (int i=0;i<n;i++) {
    int &di=multiplier_dofs[i];
    if (di!=invalid_dof_index) di=new_numbers[di];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofLevel::setMultiplierDofs(int element_data_index,int mdps,
int &next_free_dof) {
  map<int,int>::const_iterator edi2mim_iterator=
    element_data_index_to_multiplier_index_map.find(element_data_index);
  CHECK_TEST(edi2mim_iterator!=
             element_data_index_to_multiplier_index_map.end());
  NumPtr<int> edn(multiplier_dofs,edi2mim_iterator->second,mdps);
  if (edn[0]==invalid_dof_index) {
    for (int d=0;d<mdps;d++,next_free_dof++) edn[d]=next_free_dof;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void HdivDofLevel::printOn(ostream &os) const {
  os << "HdivDofLevel: " << endl;
  for (int i=0;i<dofs.getNumber();i++) {
    os << "\tdofs[" << i << "] = " << dofs[i] << endl;
  }
  for (int i=0;i<codimension1_dofs.getNumber();i++) {
    os << "\tcodimension1_dofs[" << i << "] = " << codimension1_dofs[i]
       << endl;
  }
  for (int i=0;i<multiplier_dofs.getNumber();i++) {
    os << "\tmultiplier_dofs[" << i << "] = " << multiplier_dofs[i]
       << endl;
  }
  for (map<int,int>::const_iterator edi2mim_iterator=
  element_data_index_to_multiplier_index_map.begin();
  edi2mim_iterator!=element_data_index_to_multiplier_index_map.end();
  edi2mim_iterator++) {
    cout << "\telement_data_index_to_multiplier_index_map["
         << edi2mim_iterator->first << "] = "
         << edi2mim_iterator->second << endl;
  }
}
