#ifndef HDIV_DOF_LEVEL_H
#define HDIV_DOF_LEVEL_H "$Header:$"

#include "DofLevel.H"
#include "NumPtr.H"
#include "Shape.H"

//static const int invalid_dof_index=INT_MAX;
class HdivDofLevel {
  protected:
    NumPtr<int> dofs;
    NumPtr<int> codimension1_dofs;
    NumPtr<int> multiplier_dofs;
    map<int,int> element_data_index_to_multiplier_index_map;
  public:
//the arguments to the constructor are set in HdivDofHandler::reserveSpace
    HdivDofLevel(int cd1n,int n,int mn,const map<int,int> &edi2mim) : 
    codimension1_dofs(cd1n),dofs(n),multiplier_dofs(mn),
    element_data_index_to_multiplier_index_map(edi2mim) {
      codimension1_dofs.initialize(invalid_dof_index);
      dofs.initialize(invalid_dof_index);
      multiplier_dofs.initialize(invalid_dof_index);
    }
    virtual ~HdivDofLevel() {}
    int numberDofs() const { return dofs.getNumber(); }
    int numberCodimension1Dofs() const { 
      return codimension1_dofs.getNumber(); 
    }
    int numberMultiplierDofs() const { 
      return multiplier_dofs.getNumber(); 
    }
    void getDofs(int dof_index,NumPtr<int> &indices) const { 
      indices.copyFrom(dofs+dof_index);
    }
    void getCodimension1Dofs(int dof_index,NumPtr<int> &indices) const { 
      indices.copyFrom(codimension1_dofs+dof_index);
    }
    void getMultiplierDofs(int element_data_index,NumPtr<int> &indices)
    const { 
      map<int,int>::const_iterator edi2mim_iterator=
        element_data_index_to_multiplier_index_map.
        find(element_data_index);
      CHECK_TEST(edi2mim_iterator!=
                 element_data_index_to_multiplier_index_map.end());
      indices.copyFrom(multiplier_dofs+edi2mim_iterator->second);
    }
    void setDofs(const NumPtr<int>&);
    void setDofs(int,int,int&);
    void setCodimension1Dofs(const NumPtr<int>&);
    void setCodimension1Dofs(int,int,int&);
    void setMultiplierDofs(const NumPtr<int>&);
    void setMultiplierDofs(int,int,int&);
    virtual void printOn(ostream&) const;
};

#endif
