// "$Header:$"
//----------------------------  tria_accessor.cc  -------------------------
//  $Id: tria_accessor.cc,v 1.57 2003/04/11 21:22:21 wolf Exp $
//  Version: $Name: Version-4-0-0 $
//
//  Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
//
//----------------------------  tria_accessor.cc  -------------------------
//
//modified from deal.II/deal.II/source/grid/tria_accessor.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <Level.H>
#include <AreaGraphTool.H>
#include <Tracer.H>
#include <XYGraphTool.H>

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void ShapeData::printOn(ostream &os) const {
  os << "ShapeData: child = " << child
     << " parent = " << parent
     << " used = " << used
     << " user_flag = " << user_flag
     << " material_model = " << material_model
     << " user_pointer = " << user_pointer << endl;
  element->printOn(os);
}

#if (SPACEDIM==1)
const BoundaryCondition* Level0::boundary_face_data[2]={0,0};
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Level0::Level0(Level0 *cl) : coarser_level(cl),finer_level(0) {
  if (coarser_level) coarser_level->finer_level=this;
}
Level0::~Level0() {
  if (finer_level) delete finer_level;
  finer_level=0;
  if (coarser_level) coarser_level->finer_level=0;
}
void Level0::copyLevel(const Level0 &old_level) {
  CHECK_TEST(refine_flags.getNumber()==0);
  CHECK_TEST(coarsen_flags.getNumber()==0);
  CHECK_TEST(neighbors.getNumber()==0);
  CHECK_TEST(subdomain_ids.getNumber()==0);
  refine_flags.allocate(old_level.refine_flags.getNumber());
  refine_flags.copyFrom(old_level.refine_flags);
  coarsen_flags.allocate(old_level.coarsen_flags.getNumber());
  coarsen_flags.copyFrom(old_level.coarsen_flags);
  neighbors.allocate(old_level.neighbors.getNumber());
  neighbors.copyFrom(old_level.neighbors);
  subdomain_ids.allocate(old_level.subdomain_ids.getNumber());
  subdomain_ids.copyFrom(old_level.subdomain_ids);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Level1::Level1(Level1 *cl) : Level0(cl),present_index1(-1) {
}
Level1::~Level1() {
}
void Level1::deleteElementDataShapes() {
  for (int i=0;i<element_data1.getNumber();i++) {
    delete element_data1[i].getShape();
    element_data1[i].setShape(0);
  }
}
void Level1::copyLevel(const Level1 &old_level) {
  CHECK_TEST(element_data1.getNumber()==0);
  element_data1.allocate(old_level.element_data1.getNumber());
  element_data1.copyFrom(old_level.element_data1);
  model_data.allocate(old_level.model_data.getNumber());
  model_data.copyFrom(old_level.model_data);
#if (SPACEDIM>=2)
  boundary_face_data.allocate(old_level.boundary_face_data.getNumber());
  boundary_face_data.copyFrom(old_level.boundary_face_data);
#endif
#if (SPACEDIM==3)
  boundary_line_data.allocate(old_level.boundary_line_data.getNumber());
  boundary_line_data.copyFrom(old_level.boundary_line_data);
#endif
  Level0::copyLevel(old_level);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
Level2::Level2(Level2 *cl) : Level1(cl),present_index2(-1) {
}
Level2::~Level2() {
}
void Level2::deleteElementDataShapes() {
  for (int i=0;i<element_data2.getNumber();i++) {
    delete element_data2[i].getShape();
    element_data2[i].setShape(0);
  }
  Level1::deleteElementDataShapes();
}
void Level2::copyLevel(const Level2 &old_level) {
  CHECK_TEST(element_data2.getNumber()==0);
  element_data2.allocate(old_level.element_data2.getNumber());
  element_data2.copyFrom(old_level.element_data2);
  Level1::copyLevel(old_level);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
Level3::Level3(Level3 *cl) : Level2(cl),present_index3(-1) {
}
Level3::~Level3() {
}
void Level3::deleteElementDataShapes() {
  for (int i=0;i<element_data3.getNumber();i++) {
    delete element_data3[i].getShape();
    element_data3[i].setShape(0);
  }
  Level2::deleteElementDataShapes();
}
void Level3::copyLevel(const Level3 &old_level) {
  CHECK_TEST(element_data3.getNumber()==0);
  element_data3.allocate(old_level.element_data3.getNumber());
  element_data3.copyFrom(old_level.element_data3);
  Level2::copyLevel(old_level);
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Level0::reserveSpace(int new_elements) {
  if (new_elements>refine_flags.getNumber()) {
    NumPtr<bool> refine_flags_extension(new_elements);
    refine_flags_extension.initialize(false);
    NumPtr<bool> coarsen_flags_extension(new_elements);
    coarsen_flags_extension.initialize(false);
    NumPtr<pair<Level0*,int> >
      neighbors_extension(new_elements*2*SPACEDIM);
    pair<Level0*,int> p; p.first=0; p.second=-1;
    neighbors_extension.initialize(p);
    NumPtr<int> subdomain_ids_extension(new_elements);

    refine_flags_extension.copyFrom(refine_flags);
    coarsen_flags_extension.copyFrom(coarsen_flags);
    neighbors_extension.copyFrom(neighbors);
    subdomain_ids_extension.copyFrom(subdomain_ids);

    refine_flags.copy(refine_flags_extension);
    coarsen_flags.copy(coarsen_flags_extension);
    neighbors.copy(neighbors_extension);
    subdomain_ids.copy(subdomain_ids_extension);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Level1::reserveSpace(int new_elements) {
  int new_size=new_elements+numberUsedShapes();
  if (new_size>element_data1.getNumber()) {
    NumPtr<ShapeData> element_data_extension(new_size);
#if (SPACEDIM==1)
    NumPtr<MaterialModel*> model_data_extension(new_size);
#elif (SPACEDIM==2)
    NumPtr<const BoundaryCondition*> 
      boundary_face_data_extension(new_size);
#else
    NumPtr<pair<const BoundaryCondition*,const BoundaryCondition*> > 
      boundary_line_data_extension(new_size);
#endif
    for (int i=0;i<element_data1.getNumber();i++) {
      if (element_data1[i].isUsed()) {
        element_data_extension[i]=element_data1[i];
#if (SPACEDIM==1)
        model_data_extension[i]=model_data[i];
#elif (SPACEDIM==2)
        boundary_face_data_extension[i]=boundary_face_data[i];
#else
        boundary_line_data_extension[i]=boundary_line_data[i];
#endif
      }
    }
    element_data1.copy(element_data_extension);
#if (SPACEDIM==1)
    model_data.copy(model_data_extension);
#elif (SPACEDIM==2)
    boundary_face_data.copy(boundary_face_data_extension);
#else
    boundary_line_data.copy(boundary_line_data_extension);
#endif
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void Level2::reserveSpace(int new_elements) {
  int new_size=new_elements+numberUsedShapes();
  if (new_size>element_data2.getNumber()) {
    NumPtr<ShapeData> element_data_extension(new_size);
#if (SPACEDIM==2)
    NumPtr<MaterialModel*> model_data_extension(new_size);
#else
    NumPtr<const BoundaryCondition*> 
      boundary_face_data_extension(new_size);
#endif
    for (int i=0;i<element_data2.getNumber();i++) {
      if (element_data2[i].isUsed()) {
        element_data_extension[i]=element_data2[i];
#if (SPACEDIM==2)
        model_data_extension[i]=model_data[i];
#else
        boundary_face_data_extension[i]=boundary_face_data[i];
#endif
      }
    }
    element_data2.copy(element_data_extension);
#if (SPACEDIM==2)
    model_data.copy(model_data_extension);
#else
    boundary_face_data.copy(boundary_face_data_extension);
#endif
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Level3::reserveSpace(int new_elements) {
  int new_size=new_elements+numberUsedShapes();
  if (new_size>element_data3.getNumber()) {
    NumPtr<ShapeData> element_data_extension(new_size);
    NumPtr<MaterialModel*> model_data_extension(new_size);
    for (int i=0;i<element_data3.getNumber();i++) {
      if (element_data3[i].isUsed()) {
        element_data_extension[i]=element_data3[i];
        model_data_extension[i]=model_data[i];
      }
    }
    element_data3.copy(element_data_extension);
    model_data.copy(model_data_extension);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int Level1::interiorChildVertex(int index,int &c,int &v) const {
  const Shape1 *element=getShape(index);
  CHECK_SAME(element->numberVertices(),2);
  CHECK_SAME(element->numberInteriorChildVertices(),1);
  int ev[2]={element->vertexIndex(0),element->vertexIndex(1)};

  c=0;
  int child_index=childIndex(index,c);
  const Shape1 *child_element=finerLevel()->getShape(child_index);
  int cv=child_element->vertexIndex(0);
  v=(cv==ev[0] || cv==ev[1] ? 1 : 0);
  return child_element->vertexIndex(v);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
int Level2::interiorChildVertex(int index,int &c,int &v) const {
  const Shape2 *element=getShape(index);
  const Quadrilateral *quadrilateral=
    dynamic_cast<const Quadrilateral*>(element);
  CHECK_POINTER(quadrilateral);
  CHECK_SAME(element->numberInteriorChildVertices(),1);
  set<int> element_vertex_set;
  for (int i=0;i<element->numberVertices();i++) {
    element_vertex_set.insert(element->vertexIndex(i));
  }
  c=0;
  int child_index=childIndex(index,c);
  const Shape2 *child_element=finerLevel()->getShape(child_index);
  for (int i=0;i<child_element->numberVertices();i++) {
    int cv=child_element->vertexIndex(i);
    if (element_vertex_set.find(cv)!=element_vertex_set.end()) {
      v=quadrilateral->oppositeVertex(i);
      return child_element->vertexIndex(v);;
    }
  }
  OBSOLETE("wierd: no child vertices are vertices of element");
  v=INT_MAX;
  return -1;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
int Level3::interiorChildVertex(int index,int &c,int &v) const {
  const Shape3 *element=getShape(index);
  const Hexahedron *hexahedron=
    dynamic_cast<const Hexahedron*>(element);
  CHECK_POINTER(hexahedron);
  CHECK_SAME(element->numberInteriorChildVertices(),1);
  set<int> element_vertex_set;
  for (int i=0;i<element->numberVertices();i++) {
    element_vertex_set.insert(element->vertexIndex(i));
  }
  c=0;
  int child_index=childIndex(index,c);
  const Shape3 *child_element=finerLevel()->getShape(child_index);
  for (int i=0;i<child_element->numberVertices();i++) {
    if (element_vertex_set.find(child_element->vertexIndex(i))!=
    element_vertex_set.end()) {
      v=hexahedron->oppositeVertex(i);
      return child_element->vertexIndex(v);;
    }
  }
  OBSOLETE("wierd: no child vertices are vertices of element");
  v=INT_MAX;
  return -1;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#define LEVEL_SET_INTERIOR_CHILD_NEIGHBORS(n) \
void Level##n::setInteriorChildNeighbors(int index, \
const NumPtr<int> &child_indices) { \
  const Shape##n *element=getShape(); \
  Level##n *finer_level=finerLevel(); \
  for (int i=0;i<element->numberInteriorChildCodimension1Shapes();i++) { \
    int c0,c1; \
    element->interiorChildCodimension1ShapeNeighbors(i,c0,c1); \
    const Shape##n *child_element0= \
      finer_level->getShape(child_indices[c0]); \
    const Shape##n *child_element1= \
      finer_level->getShape(child_indices[c1]); \
    int shared_face_index= \
      child_element0->sharedFaceElementIndex(child_element1); \
    for (int f=0;f<child_element0->numberCodimension1Shapes();f++) { \
      if (child_element0->codimension1ShapeIndex(f)==shared_face_index) { \
        finer_level-> \
          setNeighbor(child_indices[c0],f,finer_level,child_indices[c1]); \
        break; \
      } \
    } \
    for (int f=0;f<child_element1->numberCodimension1Shapes();f++) { \
      if (child_element1->codimension1ShapeIndex(f)==shared_face_index) { \
        finer_level-> \
          setNeighbor(child_indices[c1],f,finer_level,child_indices[c0]); \
        break; \
      } \
    } \
  } \
}
LEVEL_SET_INTERIOR_CHILD_NEIGHBORS(1)
#if (SPACEDIM>=2)
LEVEL_SET_INTERIOR_CHILD_NEIGHBORS(2)
#endif
#if (SPACEDIM==3)
LEVEL_SET_INTERIOR_CHILD_NEIGHBORS(3)
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#define LEVEL_NEIGHBOR_OF_NEIGHBOR_FUNCTION(n) \
void Level##n::neighborOfNeighbor(const Shape##n *e,int f, \
const Level##n *&neighbor_level,const Shape##n *&neighbor,int &nf) \
const { \
  CHECK_TEST(f<e->numberCodimension1Shapes()); \
  int neighbor_index=neighborIndex(e,f); \
  CHECK_TEST(neighbor_index>=0); \
  neighbor_level=neighborLevel(e,f); \
  neighbor=neighbor_level->getShape(neighbor_index); \
  for (int fn=0;fn<neighbor->numberCodimension1Shapes();fn++){\
    int neighbor_neighbor_index= \
      neighbor_level->neighborIndex(neighbor_index,fn); \
    if (neighbor_neighbor_index==e->elementDataIndex()) { \
      if (this==neighbor_level->neighborLevel(neighbor_index,fn)) { \
        nf=fn; \
        return; \
      } \
    } \
  } \
  OBSOLETE("failed to find Shape as neighbor of neighbor"); \
  nf=INT_MAX; \
}
#if (SPACEDIM>=2)
LEVEL_NEIGHBOR_OF_NEIGHBOR_FUNCTION(2)
#endif
#if (SPACEDIM==3)
LEVEL_NEIGHBOR_OF_NEIGHBOR_FUNCTION(3)
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#define LEVEL_INTERIOR_CHILD_VERTEX_ON_FACE_OF(n,m) \
void Level##n::interiorChildVertexOnFaceOf(const Shape##n *element, \
int f,int &c,int &v) const { \
  int face_element_index=element->codimension1ShapeIndex(f); \
  const Shape##m *face_element= \
    Level##m::getShape(face_element_index); \
  CHECK_TEST(face_element->numberInteriorChildVertices()>0); \
  int face_element_child_index0=Level##m::childIndex( \
    face_element_index,0); \
  int face_element_child_indexn=Level##m::childIndex( \
    face_element_index,childrenPerCodimension1Shape()-1); \
  const Shape##m *face_element_child0=finerLevel()-> \
    Level##m::getShape(face_element_child_index0); \
  const Shape##m *face_element_childn=finerLevel()-> \
    Level##m::getShape(face_element_child_indexn); \
  int shared_vertex= \
    face_element_child0->sharedVertexIndex(face_element_childn); \
  c=element->childOnFace(f,0); \
  int child_index=childIndex(element->elementDataIndex(),c); \
  const Shape *child_element=finerLevel()->getShape(child_index); \
  for (int cv=0;cv<child_element->numberVertices();cv++) { \
    if (child_element->vertexIndex(cv)==shared_vertex) { \
      v=cv; \
      return; \
    } \
  } \
  OBSOLETE("should have found shared_vertex on child_element"); \
}
#if (SPACEDIM>=2)
LEVEL_INTERIOR_CHILD_VERTEX_ON_FACE_OF(2,1)
#endif
#if (SPACEDIM==3)
LEVEL_INTERIOR_CHILD_VERTEX_ON_FACE_OF(3,2)
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#define LEVEL_CHILD_FACE_ON_FACE_OF(n,m) \
void Level##n::childFaceOnFaceOf(const Shape##n *element,int f,int s, \
int &c,int &cf) const { \
  int face_element_index=element->codimension1ShapeIndex(f); \
  CHECK_TEST(face_element_index>=0); \
  c=element->childOnFace(f,s); \
  int child_index=childIndex(element->elementDataIndex(),c); \
  CHECK_TEST(child_index>=0); \
  const Shape##n *child_element=finerLevel()->getShape(child_index); \
  set<int> face_element_child_indices; \
  for (int fc=0;fc<Shape##n::childrenPerCodimension1Shape();fc++) { \
    face_element_child_indices. \
      insert(Level##m::childIndex(face_element_index,fc)); \
  } \
  set<int> child_element_face_indices; \
  for (int fc=0;fc<child_element->numberCodimension1Shapes();fc++) { \
    child_element_face_indices. \
      insert(child_element->codimension1ShapeIndex(fc)); \
  } \
  vector<int> shared_face_index(1); \
  vector<int>::iterator sfi_iterator; \
  sfi_iterator=set_intersection(face_element_child_indices.begin(), \
    face_element_child_indices.end(),child_element_face_indices.begin(), \
    child_element_face_indices.end(),shared_face_index.begin()); \
  for (int fc=0;fc<child_element->numberCodimension1Shapes();fc++) { \
    if (child_element->codimension1ShapeIndex(fc)==shared_face_index[0]){ \
      cf=fc; \
      return; \
    } \
  } \
  OBSOLETE("should have found shared_face_index on child_element"); \
} \
void Level##n::childFaceOn(int face_child_index, \
const Shape##n *element,int f,int &c,int &cf) const { \
  CHECK_TEST(hasChildren(element)); \
  c=-1,cf=-1; \
  for (int s=0;s<Shape##n::childrenPerCodimension1Shape();s++) { \
    int cc=element->childOnFace(f,s); \
    int child_index=childIndex(element,cc); \
    CHECK_TEST(child_index>=0); \
    const Shape##n *child_element=finerLevel()->getShape(child_index); \
    for (int ff=0;ff<child_element->numberCodimension1Shapes();ff++) { \
      if (child_element->codimension1ShapeIndex(ff)==face_child_index) { \
        c=cc; \
        cf=ff; \
        return; \
      } \
    } \
  } \
  OBSOLETE("should have found face_child_index on child_element"); \
}
void Level1::childFaceOnFaceOf(const Shape1 *element,int f,int s,int &c,
int &cf) const {
  int face_vertex_index=element->codimension1ShapeIndex(f);
  CHECK_TEST(face_vertex_index>=0);
  c=element->childOnFace(f,s);
  int child_index=childIndex(element->elementDataIndex(),c);
  CHECK_TEST(child_index>=0);
  const Shape1 *child_element=finerLevel()->getShape(child_index);
  for (int fc=0;fc<child_element->numberCodimension1Shapes();fc++) {
    if (child_element->codimension1ShapeIndex(fc)==face_vertex_index) {
      cf=fc;
      return;
    }
  }
  OBSOLETE("should have found shared_face_index on child_element");
}
#if (SPACEDIM>=2)
LEVEL_CHILD_FACE_ON_FACE_OF(2,1)
#endif
#if (SPACEDIM==3)
LEVEL_CHILD_FACE_ON_FACE_OF(3,2)
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
bool Level2::lineTangentPointsBackward(const Shape2 *shape,int l) const
{
  if (hasParent(shape)) {
    const Level2 *coarser_level=coarserLevel();
    CHECK_POINTER(coarser_level);
    int parent_index=parentIndex(shape);
    CHECK_TEST(parent_index>=0);
    const Shape2 *parent_shape=coarser_level->getShape(parent_index);
    int line_index=shape->codimension1ShapeIndex(l);
    if (Level1::hasParent(line_index)) {
      int parent_line_index=Level1::parentIndex(line_index);
      const Shape1 *parent_shape_line=
        coarser_level->Level1::getShape(parent_line_index);
      int parent_line_number=
        parent_shape->faceWithIndexEqualTo(parent_line_index);
      if (shape->vertexIndex(shape->elementVertexFromLine(l,0))==
      parent_shape->vertexIndex(parent_shape->
      elementVertexFromLine(parent_line_number,0)) || 
      shape->vertexIndex(shape->elementVertexFromLine(l,1))==
      parent_shape->vertexIndex(parent_shape->
      elementVertexFromLine(parent_line_number,1))) {
        return parent_shape->
          lineTangentPointsBackward(parent_line_number);
      } else {
        return !parent_shape->
          lineTangentPointsBackward(parent_line_number);
      }
    }
  } 
  return shape->lineTangentPointsBackward(l);
}
#endif
#if (SPACEDIM==3)
bool Level3::lineTangentPointsBackward(const Shape3 *shape,int l) const
{
  if (hasParent(shape)) {
    const Level3 *coarser_level=coarserLevel();
    CHECK_POINTER(coarser_level);
    int parent_index=parentIndex(shape);
    CHECK_TEST(parent_index>=0);
    const Shape3 *parent_shape=coarser_level->getShape(parent_index);
    int line_index=shape->codimension2ShapeIndex(l);
    if (Level1::hasParent(line_index)) {
      int parent_line_index=Level1::parentIndex(line_index);
      const Shape1 *parent_shape_line=
        coarser_level->Level1::getShape(parent_line_index);
      int parent_line_number=
        parent_shape->lineWithIndexEqualTo(parent_line_index);
      if (shape->vertexIndex(shape->elementVertexFromLine(l,0))==
      parent_shape->vertexIndex(parent_shape->
      elementVertexFromLine(parent_line_number,0)) || 
      shape->vertexIndex(shape->elementVertexFromLine(l,1))==
      parent_shape->vertexIndex(parent_shape->
      elementVertexFromLine(parent_line_number,1))) {
        return parent_shape->
          lineTangentPointsBackward(parent_line_number);
      } else {
        return !parent_shape->
          lineTangentPointsBackward(parent_line_number);
      }
    }
  } 
  return shape->lineTangentPointsBackward(l);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
void Level1::checkElementData() const {
  int ed1n=element_data1.getNumber();
  if (finer_level!=0) {
    const Level1 *fl=finerLevel();
    int n=fl->element_data1.getNumber();
    for (int i=0;i<ed1n;i++) {
      element_data1[i].check(n);
    }
  }
  for (int i=0;i<ed1n;i++) {
    int ci=element_data1[i].childIndex();
    if (ci!=-1) {
      cout << "\telement_data1[" << i << "].child = " << ci << endl;
    }
  }
  if (present_index1>=0 && present_index1<ed1n && ed1n>0) {
    for (int c=0;c<2;c++) {
      int ci=childIndex(c);
    }
  } else {
    cout << "\tLevel1::checkElementData : present_index1 = " 
         << present_index1 << endl;
  }
}
#if (SPACEDIM>=2)
void Level2::checkElementData() const {
  int ed2n=element_data2.getNumber();
  if (finer_level!=0) {
    const Level2 *fl=finerLevel();
    int n=fl->element_data2.getNumber();
    for (int i=0;i<ed2n;i++) {
      element_data2[i].check(n);
    }
  }
  for (int i=0;i<ed2n;i++) {
    int ci=element_data2[i].childIndex();
    if (ci!=-1) {
      cout << "\telement_data2[" << i << "].child = " << ci << endl;
    }
  }
  if (present_index2>=0 && present_index2<ed2n && ed2n>0) {
    int ci=element_data2[present_index2].childIndex();
    for (int c=0;c<4;c++) {
      int ci=childIndex(c);
    }
  } else {
    cout << "\tLevel2::checkElementData : present_index2 = " 
         << present_index2 << endl;
  }
  Level1::checkElementData();
}
#endif
#if (SPACEDIM==3)
void Level3::checkElementData() const {
  int ed3n=element_data3.getNumber();
  if (finer_level!=0) {
    const Level3 *fl=finerLevel();
    int n=fl->element_data3.getNumber();
    for (int i=0;i<ed3n;i++) {
      element_data3[i].check(n);
    }
  }
  for (int i=0;i<ed3n;i++) {
    int c=element_data3[i].childIndex();
  }
  if (present_index3>=0 && present_index3<ed3n && ed3n>0) {
    for (int c=0;c<8;c++) {
      int ci=childIndex(c);
    }
  } else {
    cout << "\tLevel3::checkElementData : present_index3 = " 
         << present_index3 << endl;
  }
  Level2::checkElementData();
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//can't seem to do this with templates
#ifdef DEBUG
#define LEVEL_CHECK_NEIGHBORS(n) \
void Level##n::checkNeighbors(int index) const { \
  if (!used(index)) return; \
  const Shape##n *element=getShape(index); \
  for (int f=0;f<element->numberCodimension1Shapes();f++) { \
    int face_element_index=element->codimension1ShapeIndex(f); \
    int neighbor_index=neighborIndex(index,f); \
    if (neighbor_index>=0) { \
      const Level##n *neighbor_level=neighborLevel(index,f); \
      if (neighbor_level->isFinerThan(this)) { \
        cout << "\tindex = " << index << endl; \
        cout << "\tlevel = " << numberCoarserLevels() << endl; \
        cout << "\tf = " << f << endl; \
        cout << "\tneighbor_index = " << neighbor_index << endl; \
        cout << "\tneighbor_level = " \
             << neighbor_level->numberCoarserLevels() << endl; \
        cout << "\tneighbor level is finer than this level" << endl; \
        CHECK_TEST(!neighbor_level->isFinerThan(this)); \
      } \
      const Shape##n *neighbor_element= \
        neighbor_level->getShape(neighbor_index); \
      if (neighbor_level==this) { \
        bool found=false; \
        for (int nf=0; \
        nf<neighbor_element->numberCodimension1Shapes();nf++) { \
          if (neighbor_level->neighborIndex(neighbor_index,nf)==index) \
          { \
            if (this==neighbor_level->neighborLevel(neighbor_index,nf))\
            { \
              found=true; \
              int neighbor_face_element_index= \
                neighbor_element->codimension1ShapeIndex(nf); \
              if (face_element_index!=neighbor_face_element_index) { \
                cout << "\tindex = " << index << endl; \
                cout << "\tlevel = " << numberCoarserLevels() << endl; \
                cout << "\tf = " << f << endl; \
                cout << "\tface_element_index = " << face_element_index \
                     << endl; \
                cout << "\tneighbor_index = " << neighbor_index << endl; \
                cout << "\tneighbor_level = " \
                     << neighbor_level->numberCoarserLevels() << endl; \
                cout << "\tnf = " << nf << endl; \
                cout << "\tneighbor_face_element_index = " << " " \
                     << neighbor_face_element_index << endl; \
                CHECK_SAME(face_element_index,neighbor_face_element_index); \
              } \
              break; \
            } \
          } \
        } \
        if (!found) { \
          cout << "\tindex = " << index << endl; \
          cout << "\tlevel = " << numberCoarserLevels() << endl; \
          cout << "\tf = " << f << endl; \
          cout << "\tneighbor_index = " << neighbor_index << endl; \
          cout << "\tneighbor_level = " \
               << neighbor_level->numberCoarserLevels() << endl; \
          for (int nf=0; \
          nf<neighbor_element->numberCodimension1Shapes();nf++) { \
            int nni=neighbor_level->neighborIndex(neighbor_index,nf); \
            const Level##n *nnl= \
              neighbor_level->neighborLevel(neighbor_index,nf); \
            cout << "\tnf,nni,nnl = " << nf << " " << nni << " " \
                 << nnl->numberCoarserLevels() << endl; \
          } \
          cout << "\tdid not find neighbor of neighbor same as this" \
               << endl; \
          CHECK_TEST(found); \
        } \
      } else { \
        const Level##n *coarser_level=coarserLevel(); \
        bool found=false; \
        for (int nf=0; \
        nf<neighbor_element->numberCodimension1Shapes();nf++) { \
          int neighbor_neighbor_index= \
            neighbor_level->neighborIndex(neighbor_index,nf); \
          const Level##n *neighbor_neighbor_level= \
            neighbor_level->neighborLevel(neighbor_index,nf); \
          if (neighbor_neighbor_level==coarser_level) { \
            const Shape##n *neighbor_neighbor_element= \
              neighbor_neighbor_level-> \
                getShape(neighbor_neighbor_index); \
            for (int nnc=0;nnc<childrenPerShape();nnc++) { \
              int neighbor_neighbor_child_index= \
                neighbor_neighbor_level-> \
                  childIndex(neighbor_neighbor_index,nnc); \
              if (neighbor_neighbor_child_index==index) { \
                found=true; \
                break; \
              } \
            } \
          } \
          if (found) break; \
        } \
        if (!found) { \
          cout << "\tindex = " << index << endl; \
          cout << "\tlevel = " << numberCoarserLevels() << endl; \
          cout << "\tf = " << f << endl; \
          cout << "\tneighbor_index = " << neighbor_index << endl; \
          cout << "\tneighbor_level = " \
               << neighbor_level->numberCoarserLevels() << endl; \
          for (int nf=0; \
          nf<neighbor_element->numberCodimension1Shapes();nf++) { \
            int nni=neighbor_level->neighborIndex(neighbor_index,nf); \
            const Level##n *nnl= \
              neighbor_level->neighborLevel(neighbor_index,nf); \
            if (nnl==coarser_level) { \
              cout << "\tnf,nni,nnl = " << nf << " " \
                   << nni << " " << nnl->numberCoarserLevels() \
                   << " is coarser" << endl; \
              const Shape##n *neighbor_neighbor_element= \
                nnl->getShape(nni); \
              for (int nnc=0;nnc<childrenPerShape();nnc++) { \
                int nnci=nnl->childIndex(nni,nnc); \
                cout << "\t\tnnc,nnci = " << " " << nnc << " " << nnci \
                     << endl; \
              } \
            } \
          } \
          cout << "\tno face of neighbor is coarser and has child " \
               << "same as this" << endl; \
          CHECK_TEST(found); \
        } \
      } \
    } \
  } \
} \
void Level##n::checkNeighbors() { \
  for (bool doit=first();doit;doit=next()) { \
    checkNeighbors(presentIndex()); \
  } \
}
LEVEL_CHECK_NEIGHBORS(1)
#if (SPACEDIM>=2)
LEVEL_CHECK_NEIGHBORS(2)
#endif
#if (SPACEDIM==3)
LEVEL_CHECK_NEIGHBORS(3)
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM>1)
#define LEVEL_CHECK_NEIGHBOR_LEVEL_AND_INDEX(n,m) \
void Level##n::checkNeighborLevelAndIndex(int index,int f) const { \
  int face_index=getShape(index)->codimension1ShapeIndex(f); \
  const Level##n *neighbor_level=neighborLevel(index,f); \
  if (!neighbor_level) return; \
  int neighbor_index=neighborIndex(index,f); \
  CHECK_TEST(neighbor_index>=0); \
  if (neighbor_level->isFinerThan(this)) { \
    cout << "\tindex = " << index << endl; \
    cout << "\tlevel = " << numberCoarserLevels() << endl; \
    cout << "\tf = " << f << endl; \
    cout << "\tneighbor_index = " << neighbor_index << endl; \
    cout << "\tneighbor_level = " \
         << neighbor_level->numberCoarserLevels() << endl; \
    CHECK_TEST(!neighbor_level->isFinerThan(this)); \
  } else if (neighbor_level==this) { \
    const Shape##n *neighbor=neighbor_level->getShape(neighbor_index); \
    bool found=false; \
    for (int nf=0;nf<neighbor->numberCodimension1Shapes();nf++) { \
      if (neighbor->codimension1ShapeIndex(nf)==face_index) { \
        found=true; \
        break; \
      } \
    } \
    if (!found) { \
      cout << "\tindex = " << index << endl; \
      cout << "\tlevel = " << numberCoarserLevels() << endl; \
      cout << "\tf = " << f << endl; \
      cout << "\tneighbor_index = " << neighbor_index << endl; \
      cout << "\tneighbor_level = " \
           << neighbor_level->numberCoarserLevels() << endl; \
      cout << "\tneighbor face indices ="; \
      for (int nf=0;nf<neighbor->numberCodimension1Shapes();nf++) { \
        cout << " " << neighbor->codimension1ShapeIndex(nf); \
      } \
      cout << endl; \
      CHECK_TEST(found); \
    } \
  } else { \
    const Shape##n *neighbor=neighbor_level->getShape(neighbor_index); \
    bool found=false; \
    for (int nf=0;nf<neighbor->numberCodimension1Shapes();nf++) { \
      int neighbor_face_index=neighbor->codimension1ShapeIndex(nf); \
      if (neighbor_level->Level##m::hasChildren(neighbor_face_index)) { \
        const Shape##m *neighbor_face_element= \
          neighbor_level->Level##m::getShape(neighbor_face_index); \
        for (int c=0;c<neighbor_face_element->childrenPerShape();c++) { \
          if (face_index== \
          neighbor_level->Level##m::childIndex(neighbor_face_index,c)) { \
            found=true; \
            break; \
          } \
        } \
      } \
      if (found) break; \
    } \
    if (!found) { \
      cout << "\tindex = " << index << endl; \
      cout << "\tlevel = " << numberCoarserLevels() << endl; \
      cout << "\tf = " << f << endl; \
      cout << "\tneighbor_index = " << neighbor_index << endl; \
      cout << "\tneighbor_level = " \
           << neighbor_level->numberCoarserLevels() << endl; \
      for (int nf=0;nf<neighbor->numberCodimension1Shapes();nf++) { \
        int neighbor_face_index=neighbor->codimension1ShapeIndex(nf); \
        if (neighbor_level->Level##m::hasChildren(neighbor_face_index)) { \
          const Shape##m *neighbor_face_element= \
            neighbor_level->Level##m::getShape(neighbor_face_index); \
          cout << "\tneighbor face child indices ="; \
          for (int c=0;c<neighbor_face_element->childrenPerShape();c++) { \
            cout << " " << neighbor_level-> \
              Level##m::childIndex(neighbor_face_index,c); \
          } \
          cout << endl; \
        } \
      } \
      CHECK_TEST(found); \
    } \
  } \
} \
void Level##n::checkNeighborLevelAndIndex(int index) const { \
  const Shape##n *element=getShape(index); \
  for (int f=0;f<element->numberCodimension1Shapes();f++) { \
    checkNeighborLevelAndIndex(index,f); \
  } \
}
LEVEL_CHECK_NEIGHBOR_LEVEL_AND_INDEX(2,1)
#if (SPACEDIM==3)
LEVEL_CHECK_NEIGHBOR_LEVEL_AND_INDEX(3,2)
#endif
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
# if (SPACEDIM>1)
# define LEVEL_CHECK_FACE_VERTICES(n,m) \
void Level##n::checkFaceVertices(int index) const { \
  const Shape##n *element=getShape(index); \
  for (int f=0;f<element->numberCodimension1Shapes();f++) { \
    int face_index=element->codimension1ShapeIndex(f); \
    const Shape##m *face_element=Level##m::getShape(face_index); \
    for (int fv=0;fv<face_element->numberVertices();fv++) { \
      int fvi=face_element->vertexIndex(fv); \
      bool found=false; \
      for (int v=0;v<element->numberVertices();v++) { \
        if (element->vertexIndex(v)==fvi) { \
          found=true; \
          break; \
        } \
      } \
      if (!found) { \
        cout << "\telement vertices ="; \
        for (int v=0;v<element->numberVertices();v++) { \
          cout << " " << element->vertexIndex(v); \
        } \
        cout << endl; \
        cout << "\tf = " << f << endl; \
        cout << "\tface vertices ="; \
        for (int fv=0;fv<face_element->numberVertices();fv++) { \
          cout << " " << face_element->vertexIndex(fv); \
        } \
        cout << endl; \
        CHECK_TEST(found); \
      } \
    } \
  } \
} \
void Level##n::checkFaceVertices() { \
  for (bool doit=first();doit;doit=next()) { \
    checkFaceVertices(presentIndex()); \
  } \
}
LEVEL_CHECK_FACE_VERTICES(2,1)
#if (SPACEDIM==3)
LEVEL_CHECK_FACE_VERTICES(3,2)
#endif
# endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
# define LEVEL_PRINT_CHILDREN(n) \
void Level##n::printChildren() const { \
  for (int i=0;i<element_data##n.getNumber();i++) { \
    cout << "\telement_data[" << i << "].childIndex = " \
         << element_data##n[i].childIndex() << endl; \
  } \
}
LEVEL_PRINT_CHILDREN(1)
# if (SPACEDIM>=2)
LEVEL_PRINT_CHILDREN(2)
# endif
# if (SPACEDIM==3)
LEVEL_PRINT_CHILDREN(3)
# endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
# define LEVEL_PRINT_ELEMENT_DATA(n) \
void Level##n::printElementData() const { \
  cout << "\tpresent_index = " << present_index##n << endl; \
  cout << "\telement_data.getNumber = " << element_data##n.getNumber() \
       << endl; \
  element_data##n[present_index##n].printOn(cout); \
}
LEVEL_PRINT_ELEMENT_DATA(1)
# if (SPACEDIM>=2)
LEVEL_PRINT_ELEMENT_DATA(2)
# endif
# if (SPACEDIM==3)
LEVEL_PRINT_ELEMENT_DATA(3)
# endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#define LEVEL_PRINT_REFINE_FLAGS(n) \
void Level##n::printRefineFlags() const { \
  cout << "\tlevel number = " << numberCoarserLevels() << endl; \
  for (int i=0;i<refine_flags.getNumber();i++) { \
    cout << "\trefine_flags[" << i << "] = " \
         << PRINT_BOOLEAN(refine_flags[i]) << endl; \
  } \
}
LEVEL_PRINT_REFINE_FLAGS(1)
#if (SPACEDIM>=2)
LEVEL_PRINT_REFINE_FLAGS(2)
#endif
#if (SPACEDIM==3)
LEVEL_PRINT_REFINE_FLAGS(3)
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#define LEVEL_PRINT_COARSEN_FLAGS(n) \
void Level##n::printCoarsenFlags() const { \
  cout << "\tlevel number = " << numberCoarserLevels() << endl; \
  for (int i=0;i<coarsen_flags.getNumber();i++) { \
    cout << "\tcoarsen_flags[" << i << "] = " \
         << PRINT_BOOLEAN(coarsen_flags[i]) << endl; \
  } \
}
LEVEL_PRINT_COARSEN_FLAGS(1)
#if (SPACEDIM>=2)
LEVEL_PRINT_COARSEN_FLAGS(2)
#endif
#if (SPACEDIM==3)
LEVEL_PRINT_COARSEN_FLAGS(3)
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void LEVEL_CLASS::printMaterialModels() const {
  cout << "\tlevel number = " << numberCoarserLevels() << endl;
  for (int i=0;i<model_data.getNumber();i++) {
    cout << "\tmaterial_model[" << i << "] = " << model_data[i] << endl;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void CODIMENSION1_LEVEL_CLASS::printFaceBoundaryConditions() const {
  cout << "\tlevel number = " << numberCoarserLevels() << endl;
#if (SPACEDIM==1)
  int nbc=2;
#else
  int nbc=boundary_face_data.getNumber();
#endif
  for (int i=0;i<nbc;i++) {
    cout << "\tboundary_condition[" << i << "] = "
         << boundary_face_data[i];
    if (boundary_face_data[i]) {
      if (boundary_face_data[i]->isEssential()) cout << " essential";
      if (boundary_face_data[i]->isNatural()) cout << " natural";
    }
    cout << endl;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Level1::printLineBoundaryConditions() const {
  cout << "\tlevel number = " << numberCoarserLevels() << endl;
  for (int i=0;i<boundary_line_data.getNumber();i++) {
    cout << "\tboundary_condition[" << i << "] = "
         << boundary_line_data[i].first << " "
         << boundary_line_data[i].second << endl;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#define LEVEL_PRINT_USED_FLAGS(n) \
void Level##n::printUsedFlags() const { \
  cout << "\tlevel number = " << numberCoarserLevels() << endl; \
  for (int i=0;i<element_data##n.getNumber();i++) { \
    cout << "\tused_flag[" << i << "] = " \
         << PRINT_BOOLEAN(element_data##n[i].isUsed()) << endl; \
  } \
}
LEVEL_PRINT_USED_FLAGS(1)
#if (SPACEDIM>=2)
LEVEL_PRINT_USED_FLAGS(2)
#endif
#if (SPACEDIM==3)
LEVEL_PRINT_USED_FLAGS(3)
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#define LEVEL_PRINT_USER_FLAGS(n) \
void Level##n::printUserFlags() const { \
  cout << "\tlevel number = " << numberCoarserLevels() << endl; \
  for (int i=0;i<element_data##n.getNumber();i++) { \
    cout << "\tuser_flag[" << i << "] = " \
         << PRINT_BOOLEAN(element_data##n[i].userFlagSet()) << endl; \
  } \
}
LEVEL_PRINT_USER_FLAGS(1)
#if (SPACEDIM>=2)
LEVEL_PRINT_USER_FLAGS(2)
#endif
#if (SPACEDIM==3)
LEVEL_PRINT_USER_FLAGS(3)
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#define LEVEL_PRINT_NEIGHBORS(n) \
void Level##n::printNeighbors(int index) const { \
  const Shape##n *element=getShape(index); \
  if (element!=0) { \
    cout << "\tneighbors of element number " << index \
         << " on level number " << numberCoarserLevels() << " :" \
         << endl; \
    for (int f=0;f<element->numberCodimension1Shapes();f++) { \
      const Level##n *nl=neighborLevel(index,f); \
      cout << "\tneighbor[" << f << "] = " << neighborIndex(index,f) \
           << " on level number " \
           << (nl!=0 ? nl->numberCoarserLevels() : -1)  << endl; \
    } \
  } \
}
LEVEL_PRINT_NEIGHBORS(1)
#if (SPACEDIM>=2)
LEVEL_PRINT_NEIGHBORS(2)
#endif
#if (SPACEDIM==3)
LEVEL_PRINT_NEIGHBORS(3)
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
void Level1::plotShapes(const NumPtr<Point<SPACEDIM> > &vertices) {
  Level1 *fl=dynamic_cast<Level1*>(finer_level);
  Palette pal;
  const char *title="Level1::plotShapes";
  REAL winsize=0.5;
#if (SPACEDIM==1)
  REAL xlo=HUGE_VAL,xhi=-HUGE_VAL,ylo=-1.,yhi=1.;
  for (bool doit=first();doit;doit=next()) {
    const Shape1 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[vertexIndex(v)];
      xlo=min(xlo,p[0]);
      xhi=max(xhi,p[0]);
    }
  }
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool gt(title,"x","y",xlo,xhi,ylo,yhi,&cmap,0,winsize);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();
  gt.setfgColor("blue");
  for (bool doit=first();doit;doit=next()) {
    const Shape1 *element=getShape();
    Point<SPACEDIM> p=
      vertices[vertexIndex(element->numberVertices()-1)];
    gt.movePen(p[0],0.5);
    gt.drawLine(p[0],0.55);
    gt.movePen(p[0],0.5);
    for (int v=0;v<element->numberVertices();v++) {
      p=vertices[vertexIndex(v)];
      gt.drawLine(p[0],0.5);
      gt.drawLine(p[0],0.45);
    }
  }
  gt.flush();
  XYGraphTool::WINDOW_TYPE::QuitButton qb;
#endif
#if (SPACEDIM==2)
  REAL rlo[2]={HUGE_VAL,HUGE_VAL},rhi[2]={-HUGE_VAL,-HUGE_VAL};
  for (bool doit=first();doit;doit=next()) {
    const Shape1 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      rlo[0]=min(rlo[0],p[0]);
      rlo[1]=min(rlo[1],p[1]);
      rhi[0]=max(rhi[0],p[0]);
      rhi[1]=max(rhi[1],p[1]);
    }
  }
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  AreaGraphTool gt(title,"x","y",rlo,rhi,&cmap,0,winsize);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();
  gt.setfgColor("blue");
  for (bool doit=first();doit;doit=next()) {
    const Shape1 *element=getShape();
    element->draw(vertices,gt);
  }
  gt.flush();
  AreaGraphTool::WINDOW_TYPE::QuitButton qb;
#endif
#if (SPACEDIM==3)
  Vector3 low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Vector3 high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  for (bool doit=first();doit;doit=next()) {
    const Shape1 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      for (int i=0;i<3;i++) {
        low[i]=min(low[i],p[i]);
        high[i]=max(high[i],p[i]);
      }
    }
  }
  Vector3 dif=(high-low)*0.05;
  low-=dif;
  high+=dif;
  BOOLEAN use_lighting=FALSE;
  VolGraphTool gt(title,low,high,pal,use_lighting,winsize);
  const VolGraphTool::COLOR_ARRAY_TYPE *color_array=gt.getColorArray();
  gt.newPage();
  gt.expose();
  for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
    {
      VolGraphTool::DrawDelimiter dd(do_best,&gt);
      VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
        color_array->getColor(0.5);
      for (bool doit=first();doit;doit=next()) {
        getShape()->draw(xc,vertices,gt);
      }
      if (do_best==TRUE) break;
    }
  }
  gt.eventLoop();
#endif
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
void Level1::plotShapesIndividually(
const NumPtr<Point<SPACEDIM> > &vertices) {
  Level1 *fl=dynamic_cast<Level1*>(finer_level);
  Palette pal;
  const char *title="Level1::plotShapes";
  REAL winsize=0.5;
#if (SPACEDIM==1)
  REAL xlo=HUGE_VAL,xhi=-HUGE_VAL,ylo=-1.,yhi=1.;
  for (bool doit=first();doit;doit=next()) {
    const Shape1 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[vertexIndex(v)];
      xlo=min(xlo,p[0]);
      xhi=max(xhi,p[0]);
    }
  }
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool gt(title,"x","y",xlo,xhi,ylo,yhi,&cmap,0,winsize);
  for (bool doit=first();doit;doit=next()) {
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    gt.setfgColor("blue");
    const Shape1 *element=getShape();
    Point<SPACEDIM> p=
      vertices[vertexIndex(element->numberVertices()-1)];
    gt.movePen(p[0],0.5);
    for (int v=0;v<element->numberVertices();v++) {
      p=vertices[vertexIndex(v)];
      gt.drawLine(p[0],0.5);
    }
    if (hasChildren()) {
      for (int c=0;c<childrenPerShape();c++) {
        gt.setfgColor(c,childrenPerShape());
        int child_index=childIndex(c);
        const Shape1 *child_element=fl->getShape(child_index);
        p=vertices[fl->
            vertexIndex(child_index,child_element->numberVertices()-1)];
        gt.movePen(p[0],-0.5);
        for (int v=0;v<child_element->numberVertices();v++) {
          p=vertices[fl->vertexIndex(child_index,v)];
          gt.drawLine(p[0],-0.5);
        }
      }
    }
    gt.flush();
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#endif
#if (SPACEDIM==2)
  REAL rlo[2]={HUGE_VAL,HUGE_VAL},rhi[2]={-HUGE_VAL,-HUGE_VAL};
  for (bool doit=first();doit;doit=next()) {
    const Shape1 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      rlo[0]=min(rlo[0],p[0]);
      rlo[1]=min(rlo[1],p[1]);
      rhi[0]=max(rhi[0],p[0]);
      rhi[1]=max(rhi[1],p[1]);
    }
  }
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  AreaGraphTool gt(title,"x","y",rlo,rhi,&cmap,0,winsize);
  for (bool doit=first();doit;doit=next()) {
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    gt.setfgColor("blue");
    const Shape1 *element=getShape();
    element->draw(vertices,gt);
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#endif
#if (SPACEDIM==3)
  Vector3 low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Vector3 high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  for (bool doit=first();doit;doit=next()) {
    const Shape1 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      for (int i=0;i<3;i++) {
        low[i]=min(low[i],p[i]);
        high[i]=max(high[i],p[i]);
      }
    }
  }
  Vector3 dif=(high-low)*0.05;
  low-=dif;
  high+=dif;
  BOOLEAN use_lighting=FALSE;
  VolGraphTool gt(title,low,high,pal,use_lighting,winsize);
  const VolGraphTool::COLOR_ARRAY_TYPE *color_array=gt.getColorArray();
  for (bool doit=first();doit;doit=next()) {
    gt.newPage();
    gt.expose();
    for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
      {
        VolGraphTool::DrawDelimiter dd(do_best,&gt);
        VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
          color_array->getColor(0.5);
        getShape()->draw(xc,vertices,gt);
      }
      if (do_best==TRUE) break;
    }
    gt.eventLoop();
  }
#endif
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM>=2)
void Level2::plotShapes(const NumPtr<Point<SPACEDIM> > &vertices) {
  Level2 *fl=dynamic_cast<Level2*>(finer_level);
  const char *title="Level2::plotShapes";
  Palette pal;
  REAL winsize=0.5;
#if (SPACEDIM==2)
  REAL rlo[2]={HUGE_VAL,HUGE_VAL},rhi[2]={-HUGE_VAL,-HUGE_VAL};
  for (bool doit=first();doit;doit=next()) {
    const Shape2 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      rlo[0]=min(rlo[0],p[0]);
      rlo[1]=min(rlo[1],p[1]);
      rhi[0]=max(rhi[0],p[0]);
      rhi[1]=max(rhi[1],p[1]);
    }
  }
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  AreaGraphTool gt(title,"x","y",rlo,rhi,&cmap,0,winsize);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();
  gt.setfgColor("blue");
  for (bool doit=first();doit;doit=next()) {
    const Shape2 *element=getShape();
    element->draw(vertices,gt);
  }
  gt.flush();
  AreaGraphTool::WINDOW_TYPE::QuitButton qb;
#endif
#if (SPACEDIM==3)
  Vector3 low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Vector3 high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  for (bool doit=first();doit;doit=next()) {
    const Shape2 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      for (int i=0;i<3;i++) {
        low[i]=min(low[i],p[i]);
        high[i]=max(high[i],p[i]);
      }
    }
  }
  Vector3 dif=(high-low)*0.05;
  low-=dif;
  high+=dif;
  BOOLEAN use_lighting=FALSE;
  VolGraphTool gt(title,low,high,pal,use_lighting,winsize);
  const VolGraphTool::COLOR_ARRAY_TYPE *color_array=gt.getColorArray();
  gt.newPage();
  gt.expose();
  for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
    {
      VolGraphTool::DrawDelimiter dd(do_best,&gt);
      VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
        color_array->getColor(0.5);
      for (bool doit=first();doit;doit=next()) {
        getShape()->draw(xc,vertices,gt);
      }
    }
    if (do_best==TRUE) break;
  }
  gt.eventLoop();
#endif
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM>=2)
void Level2::plotShapesIndividually(
const NumPtr<Point<SPACEDIM> > &vertices) {
  Level2 *fl=dynamic_cast<Level2*>(finer_level);
  const char *title="Level2::plotShapes";
  Palette pal;
  REAL winsize=0.5;
#if (SPACEDIM==2)
  REAL rlo[2]={HUGE_VAL,HUGE_VAL},rhi[2]={-HUGE_VAL,-HUGE_VAL};
  for (bool doit=first();doit;doit=next()) {
    const Shape2 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      rlo[0]=min(rlo[0],p[0]);
      rlo[1]=min(rlo[1],p[1]);
      rhi[0]=max(rhi[0],p[0]);
      rhi[1]=max(rhi[1],p[1]);
    }
  }
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  AreaGraphTool gt(title,"x","y",rlo,rhi,&cmap,0,winsize);
  for (bool doit=first();doit;doit=next()) {
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    gt.setfgColor("blue");
    const Shape2 *element=getShape();
    element->draw(vertices,gt);
    if (hasChildren()) {
      for (int c=0;c<childrenPerShape();c++) {
        gt.setfgColor(c,childrenPerShape());
        int child_index=childIndex(c);
        const Shape2 *child_element=fl->getShape(child_index);
        child_element->draw(vertices,gt);
      }
    }
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#endif
#if (SPACEDIM==3)
  Vector3 low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Vector3 high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  for (bool doit=first();doit;doit=next()) {
    const Shape2 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      for (int i=0;i<3;i++) {
        low[i]=min(low[i],p[i]);
        high[i]=max(high[i],p[i]);
      }
    }
  }
  Vector3 dif=(high-low)*0.05;
  low-=dif;
  high+=dif;
  BOOLEAN use_lighting=FALSE;
  VolGraphTool gt(title,low,high,pal,use_lighting,winsize);
  const VolGraphTool::COLOR_ARRAY_TYPE *color_array=gt.getColorArray();
  for (bool doit=first();doit;doit=next()) {
    gt.newPage();
    gt.expose();
    for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
      {
        VolGraphTool::DrawDelimiter dd(do_best,&gt);
        VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
          color_array->getColor(0.5);
        getShape()->draw(xc,vertices,gt);
        if (hasChildren()) {
          for (int c=0;c<childrenPerCodimension2Shape();c++){
            int child_index=childIndex(c);
            const Shape2 *child_element=fl->getShape(child_index);
            xc=color_array->getColor(static_cast<REAL>(c+1)
                  /static_cast<REAL>(childrenPerCodimension2Shape()));
            child_element->draw(xc,vertices,gt);
          }
        }
      }
      if (do_best==TRUE) break;
    }
    gt.eventLoop();
  }
#endif
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==3)
void Level3::plotShapes(const NumPtr<Point<SPACEDIM> > &vertices) {
  Vector3 low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Vector3 high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  for (bool doit=first();doit;doit=next()) {
    const Shape3 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      for (int i=0;i<3;i++) {
        low[i]=min(low[i],p[i]);
        high[i]=max(high[i],p[i]);
      }
    }
  }
  Vector3 dif=(high-low)*0.05;
  low-=dif;
  high+=dif;
  Palette pal;
  BOOLEAN use_lighting=FALSE;
  const char *title="Level3::plotShapes";
  REAL winsize=0.5;
  VolGraphTool gt(title,low,high,pal,use_lighting,winsize);
  const VolGraphTool::COLOR_ARRAY_TYPE *color_array=gt.getColorArray();
  gt.newPage();
  gt.expose();
  for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
    {
      VolGraphTool::DrawDelimiter dd(do_best,&gt);
      VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
        color_array->getColor(0.5);
      for (bool doit=first();doit;doit=next()) {
        getShape()->draw(xc,vertices,gt);
        if (hasChildren()) {
          for (int c=0;c<childrenPerShape();c++){
            int child_index=childIndex(c);
            const Shape3 *child_element=
              finerLevel()->getShape(child_index);
            xc=color_array->getColor(static_cast<REAL>(c+1)
                  /static_cast<REAL>(childrenPerShape()));
            child_element->draw(xc,vertices,gt);
          }
        }
      }
      if (do_best==TRUE) break;
    }
  }
  gt.eventLoop();
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==3)
void Level3::plotShapesIndividually(
const NumPtr<Point<SPACEDIM> > &vertices) {
  Vector3 low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Vector3 high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  for (bool doit=first();doit;doit=next()) {
    const Shape3 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      for (int i=0;i<3;i++) {
        low[i]=min(low[i],p[i]);
        high[i]=max(high[i],p[i]);
      }
    }
  }
  Vector3 dif=(high-low)*0.05;
  low-=dif;
  high+=dif;
  Palette pal;
  BOOLEAN use_lighting=FALSE;
  const char *title="Level3::plotShapes";
  REAL winsize=0.5;
  VolGraphTool gt(title,low,high,pal,use_lighting,winsize);
  const VolGraphTool::COLOR_ARRAY_TYPE *color_array=gt.getColorArray();
  for (bool doit=first();doit;doit=next()) {
    gt.newPage();
    gt.expose();
    for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
      {
        VolGraphTool::DrawDelimiter dd(do_best,&gt);
        VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
          color_array->getColor(0.5);
        getShape()->draw(xc,vertices,gt);
        if (hasChildren()) {
          for (int c=0;c<childrenPerShape();c++){
            int child_index=childIndex(c);
            const Shape3 *child_element=
              finerLevel()->getShape(child_index);
            xc=color_array->getColor(static_cast<REAL>(c+1)
                  /static_cast<REAL>(childrenPerShape()));
            child_element->draw(xc,vertices,gt);
          }
        }
      }
      if (do_best==TRUE) break;
    }
    gt.eventLoop();
  }
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
void Level1::plotUsedShapes(const NumPtr<Point<SPACEDIM> > &vertices){
#if (SPACEDIM==1)
  REAL xlo=HUGE_VAL,xhi=-HUGE_VAL,ylo=-1.,yhi=1.;
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    const Shape1 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[vertexIndex(v)];
      xlo=min(xlo,p[0]);
      xhi=max(xhi,p[0]);
    }
  }
  if (xlo>xhi) {
    cout << "\tLevel1::plotUsedShapes: no used elements in this level"
         << endl;
    return;
  }
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  const char *title="Level1::plotUsedShapes";
  REAL winsize=0.5;
  XYGraphTool gt(title,"x","y",xlo,xhi,ylo,yhi,&cmap,0,winsize);
  Level1 *fl=dynamic_cast<Level1*>(finer_level);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();
  gt.setfgColor("blue");
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    const Shape1 *element=getShape();
    Point<SPACEDIM> p=
      vertices[vertexIndex(element->numberVertices()-1)];
    gt.movePen(p[0],0.5);
    gt.drawLine(p[0],0.55);
    gt.movePen(p[0],0.5);
    for (int v=0;v<element->numberVertices();v++) {
      p=vertices[vertexIndex(v)];
      gt.drawLine(p[0],0.5);
      gt.drawLine(p[0],0.45);
    }
  }
  gt.flush();
  XYGraphTool::WINDOW_TYPE::QuitButton qb;
#endif
#if (SPACEDIM==2)
  OBSOLETE("not programmed");
#endif
#if (SPACEDIM==3)
  OBSOLETE("not programmed");
#endif
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
void Level1::plotUsedShapesIndividually(
const NumPtr<Point<SPACEDIM> > &vertices) {
#if (SPACEDIM==1)
  REAL xlo=HUGE_VAL,xhi=-HUGE_VAL,ylo=-1.,yhi=1.;
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    const Shape1 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[vertexIndex(v)];
      xlo=min(xlo,p[0]);
      xhi=max(xhi,p[0]);
    }
  }
  if (xlo>xhi) {
    cout << "\tLevel1::plotUsedShapes: no used elements in this level"
         << endl;
    return;
  }
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  const char *title="Level1::plotUsedShapes";
  REAL winsize=0.5;
  XYGraphTool gt(title,"x","y",xlo,xhi,ylo,yhi,&cmap,0,winsize);
  Level1 *fl=dynamic_cast<Level1*>(finer_level);
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    gt.setfgColor("blue");
    const Shape1 *element=getShape();
    Point<SPACEDIM> p=
      vertices[vertexIndex(element->numberVertices()-1)];
    gt.movePen(p[0],0.5);
    for (int v=0;v<element->numberVertices();v++) {
      p=vertices[vertexIndex(v)];
      gt.drawLine(p[0],0.5);
    }
    if (hasChildren()) {
      for (int c=0;c<childrenPerShape();c++) {
        gt.setfgColor(c,childrenPerShape());
        int child_index=childIndex(c);
        const Shape *child_element=fl->getShape(child_index);
        p=vertices[fl->
            vertexIndex(child_index,child_element->numberVertices()-1)];
        gt.movePen(p[0],-0.5);
        for (int v=0;v<child_element->numberVertices();v++) {
          p=vertices[fl->vertexIndex(child_index,v)];
          gt.drawLine(p[0],-0.5);
        }
      }
    }
    gt.flush();
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#endif
#if (SPACEDIM==2)
  OBSOLETE("not programmed");
#endif
#if (SPACEDIM==3)
  OBSOLETE("not programmed");
#endif
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM>=2)
void Level2::plotUsedShapes(const NumPtr<Point<SPACEDIM> > &vertices){
#if (SPACEDIM==2)
  REAL rlo[2]={HUGE_VAL,HUGE_VAL},rhi[2]={-HUGE_VAL,-HUGE_VAL};
  for (bool doit=first();doit;doit=next()) {
    const Shape2 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      rlo[0]=min(rlo[0],p[0]);
      rlo[1]=min(rlo[1],p[1]);
      rhi[0]=max(rhi[0],p[0]);
      rhi[1]=max(rhi[1],p[1]);
    }
  }
  if (rlo[0]>rhi[0] || rlo[1]>rhi[1]) {
    cout << "\tLevel2::plotUsedShapes: no used elements in this level"
         << endl;
    return;
  }
  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  const char *title="Level2::plotUsedShapes";
  REAL winsize=0.5;
  AreaGraphTool gt(title,"x","y",rlo,rhi,&cmap,0,winsize);
  Level2 *fl=dynamic_cast<Level2*>(finer_level);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();
  gt.setfgColor("blue");
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    const Shape2 *element=getShape();
    element->draw(vertices,gt);
  }
  gt.flush();
  AreaGraphTool::WINDOW_TYPE::QuitButton qb;
#endif
#if (SPACEDIM==3)
  OBSOLETE("not programmed");
#endif
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM>=2)
void Level2::plotUsedShapesIndividually(
const NumPtr<Point<SPACEDIM> > &vertices){
#if (SPACEDIM==2)
  REAL rlo[2]={HUGE_VAL,HUGE_VAL},rhi[2]={-HUGE_VAL,-HUGE_VAL};
  for (bool doit=first();doit;doit=next()) {
    const Shape2 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      rlo[0]=min(rlo[0],p[0]);
      rlo[1]=min(rlo[1],p[1]);
      rhi[0]=max(rhi[0],p[0]);
      rhi[1]=max(rhi[1],p[1]);
    }
  }
  if (rlo[0]>rhi[0] || rlo[1]>rhi[1]) {
    cout << "\tLevel2::plotUsedShapes: no used elements in this level"
         << endl;
    return;
  }
  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  const char *title="Level2::plotUsedShapes";
  REAL winsize=0.5;
  AreaGraphTool gt(title,"x","y",rlo,rhi,&cmap,0,winsize);
  Level2 *fl=dynamic_cast<Level2*>(finer_level);
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    gt.setfgColor("blue");
    const Shape2 *element=getShape();
    element->draw(vertices,gt);
    if (hasChildren()) {
      for (int c=0;c<childrenPerShape();c++) {
        gt.setfgColor(c,childrenPerShape());
        int child_index=childIndex(c);
        const Shape2 *child_element=fl->getShape(child_index);
        child_element->draw(vertices,gt);
      }
    }
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#endif
#if (SPACEDIM==3)
  OBSOLETE("not programmed");
#endif
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==3)
void Level3::plotUsedShapes(const NumPtr<Point<SPACEDIM> > &vertices){
  OBSOLETE("not programmed");
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==3)
void Level3::plotUsedShapesIndividually(
const NumPtr<Point<SPACEDIM> > &vertices){
  OBSOLETE("not programmed");
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
void Level1::plotActiveShapes(
const NumPtr<Point<SPACEDIM> > &vertices) {
#if (SPACEDIM==1)
  REAL xlo=HUGE_VAL,xhi=-HUGE_VAL,ylo=-1.,yhi=1.;
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    const Shape1 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      xlo=min(xlo,p[0]);
      xhi=max(xhi,p[0]);
    }
  }
  if (xlo>xhi) {
    return;
  }
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  const char *title="Level1::plotActiveShapes";
  REAL winsize=0.5;
  XYGraphTool gt(title,"x","y",xlo,xhi,ylo,yhi,&cmap,0,winsize);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();
  gt.setfgColor("blue");
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    const Shape1 *element=getShape();
    Point<SPACEDIM> p=
      vertices[vertexIndex(element->numberVertices()-1)];
    gt.movePen(p[0],0.45);
    gt.drawLine(p[0],0.55);
    gt.movePen(p[0],0.5);
    for (int v=0;v<element->numberVertices();v++) {
      p=vertices[element->vertexIndex(v)];
      gt.drawLine(p[0],0.5);
      gt.movePen(p[0],0.45);
      gt.drawLine(p[0],0.55);
      gt.movePen(p[0],0.5);
    }
  }
  gt.flush();
  XYGraphTool::WINDOW_TYPE::QuitButton qb;
#endif
#if (SPACEDIM==2)
  REAL rlo[2]={HUGE_VAL,HUGE_VAL},rhi[2]={-HUGE_VAL,-HUGE_VAL};
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    const Shape1 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      rlo[0]=min(rlo[0],p[0]);
      rlo[1]=min(rlo[1],p[1]);
      rhi[0]=max(rhi[0],p[0]);
      rhi[1]=max(rhi[1],p[1]);
    }
  }
  if (rlo[0]>rhi[0] || rlo[1]>rhi[1]) {
    return;
  }
  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  const char *title="Level1::plotActiveShapes";
  REAL winsize=0.5;
  AreaGraphTool gt(title,"x","y",rlo,rhi,&cmap,0,winsize);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();
  gt.setfgColor("blue");
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    const Shape1 *element=getShape();
    Point<SPACEDIM> p=
      vertices[element->vertexIndex(element->numberVertices()-1)];
    gt.movePen(p[0],p[1]);
    for (int v=0;v<element->numberVertices();v++) {
      p=vertices[element->vertexIndex(v)];
      gt.drawLine(p[0],p[1]);
    }
  }
  gt.flush();
  AreaGraphTool::WINDOW_TYPE::QuitButton qb;
#endif
#if (SPACEDIM==3)
  OBSOLETE("not programmed");
#endif
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
void Level1::plotActiveShapesIndividually(
const NumPtr<Point<SPACEDIM> > &vertices) {
#if (SPACEDIM==1)
  REAL xlo=HUGE_VAL,xhi=-HUGE_VAL,ylo=-1.,yhi=1.;
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    const Shape1 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      xlo=min(xlo,p[0]);
      xhi=max(xhi,p[0]);
    }
  }
  if (xlo>xhi) {
    cout << "\tLevel1::plotActiveShapes:"
         << " no active elements in this level" << endl;
    return;
  }
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  const char *title="Level1::plotActiveShapes";
  REAL winsize=0.5;
  XYGraphTool gt(title,"x","y",xlo,xhi,ylo,yhi,&cmap,0,winsize);
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    gt.setfgColor("blue");
    const Shape1 *element=getShape();
    Point<SPACEDIM> p=
      vertices[vertexIndex(element->numberVertices()-1)];
    gt.movePen(p[0],0.5);
    for (int v=0;v<element->numberVertices();v++) {
      p=vertices[element->vertexIndex(v)];
      gt.drawLine(p[0],0.5);
    }
    gt.flush();
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#endif
#if (SPACEDIM==2)
  REAL rlo[2]={HUGE_VAL,HUGE_VAL},rhi[2]={-HUGE_VAL,-HUGE_VAL};
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    const Shape1 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      rlo[0]=min(rlo[0],p[0]);
      rlo[1]=min(rlo[1],p[1]);
      rhi[0]=max(rhi[0],p[0]);
      rhi[1]=max(rhi[1],p[1]);
    }
  }
  if (rlo[0]>rhi[0] || rlo[1]>rhi[1]) {
    cout << "\tLevel1::plotActiveShapes:"
         << " no active elements in this level" << endl;
    return;
  }
  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  const char *title="Level1::plotActiveShapes";
  REAL winsize=0.5;
  AreaGraphTool gt(title,"x","y",rlo,rhi,&cmap,0,winsize);
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    gt.setfgColor("blue");
    const Shape1 *element=getShape();
    Point<SPACEDIM> p=
      vertices[element->vertexIndex(element->numberVertices()-1)];
    gt.movePen(p[0],p[1]);
    for (int v=0;v<element->numberVertices();v++) {
      p=vertices[element->vertexIndex(v)];
      gt.drawLine(p[0],p[1]);
    }
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#endif
#if (SPACEDIM==3)
  OBSOLETE("not programmed");
#endif
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM>=2)
void Level2::plotActiveShapes(
const NumPtr<Point<SPACEDIM> > &vertices) {
  Palette pal;
  const char *title="Level2::plotActiveShapes";
  REAL winsize=0.5;
#if (SPACEDIM==2)
  REAL rlo[2]={HUGE_VAL,HUGE_VAL},rhi[2]={-HUGE_VAL,-HUGE_VAL};
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    const Shape2 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      rlo[0]=min(rlo[0],p[0]);
      rlo[1]=min(rlo[1],p[1]);
      rhi[0]=max(rhi[0],p[0]);
      rhi[1]=max(rhi[1],p[1]);
    }
  }
  if (rlo[0]>rhi[0] || rlo[1]>rhi[1]) {
    return;
  }
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  AreaGraphTool gt(title,"x","y",rlo,rhi,&cmap,0,winsize);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();
  gt.setfgColor("blue");
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    const Shape2 *element=getShape();
    element->draw(vertices,gt);
  }
  gt.flush();
  AreaGraphTool::WINDOW_TYPE::QuitButton qb;
#endif
#if (SPACEDIM==3)
  Vector3 low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Vector3 high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  for (bool doit=first();doit;doit=next()) {
    const Shape2 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      for (int i=0;i<3;i++) {
        low[i]=min(low[i],p[i]);
        high[i]=max(high[i],p[i]);
      }
    }
  }
  Vector3 dif=(high-low)*0.05;
  low-=dif;
  high+=dif;
  BOOLEAN use_lighting=TRUE;
  VolGraphTool gt(title,low,high,pal,use_lighting,winsize);
  const VolGraphTool::COLOR_ARRAY_TYPE *color_array=gt.getColorArray();
  gt.newPage();
  gt.expose();
  for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
    {
      VolGraphTool::DrawDelimiter dd(do_best,&gt);
      VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
        color_array->getColor(0.5);
      for (bool doit=first();doit;doit=next()) {
        if (!used()) continue;
        if (hasChildren()) continue;
        getShape()->draw(xc,vertices,gt);
      }
    }
    if (do_best==TRUE) break;
  }
  gt.eventLoop();
#endif
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM>=2)
void Level2::plotActiveShapesIndividually(
const NumPtr<Point<SPACEDIM> > &vertices) {
  Palette pal;
  const char *title="Level2::plotActiveShapes";
  REAL winsize=0.5;
#if (SPACEDIM==2)
  REAL rlo[2]={HUGE_VAL,HUGE_VAL},rhi[2]={-HUGE_VAL,-HUGE_VAL};
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    const Shape2 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      rlo[0]=min(rlo[0],p[0]);
      rlo[1]=min(rlo[1],p[1]);
      rhi[0]=max(rhi[0],p[0]);
      rhi[1]=max(rhi[1],p[1]);
    }
  }
  if (rlo[0]>rhi[0] || rlo[1]>rhi[1]) {
    return;
  }
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  AreaGraphTool gt(title,"x","y",rlo,rhi,&cmap,0,winsize);
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    gt.setfgColor("blue");
    const Shape2 *element=getShape();
    element->draw(vertices,gt);
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
#endif
#if (SPACEDIM==3)
  Vector3 low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Vector3 high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  for (bool doit=first();doit;doit=next()) {
    const Shape2 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      for (int i=0;i<3;i++) {
        low[i]=min(low[i],p[i]);
        high[i]=max(high[i],p[i]);
      }
    }
  }
  Vector3 dif=(high-low)*0.05;
  low-=dif;
  high+=dif;
  BOOLEAN use_lighting=FALSE;
  VolGraphTool gt(title,low,high,pal,use_lighting,winsize);
  const VolGraphTool::COLOR_ARRAY_TYPE *color_array=gt.getColorArray();
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    gt.newPage();
    gt.expose();
    for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
      {
        VolGraphTool::DrawDelimiter dd(do_best,&gt);
        VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
          color_array->getColor(0.5);
        getShape()->draw(xc,vertices,gt);
        if (hasChildren()) {
          for (int c=0;c<childrenPerCodimension1Shape();c++){
            int child_index=childIndex(c);
            const Shape2 *child_element=
              finerLevel()->getShape(child_index);
            xc=color_array->getColor(static_cast<REAL>(c+1)
                  /static_cast<REAL>(childrenPerCodimension2Shape()));
            child_element->draw(xc,vertices,gt);
          }
        }
      }
      if (do_best==TRUE) break;
    }
    gt.eventLoop();
  }
#endif
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==3)
void Level3::plotActiveShapes(
const NumPtr<Point<SPACEDIM> > &vertices) {
  Palette pal;
  const char *title="Level3::plotActiveShapes";
  REAL winsize=0.5;
  Vector3 low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Vector3 high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    const Shape3 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      for (int i=0;i<3;i++) {
        low[i]=min(low[i],p[i]);
        high[i]=max(high[i],p[i]);
      }
    }
  }
  if (low[0]>high[0] || low[1]>high[1] || low[2]>high[2]) {
    return;
  }
  Vector3 dif=(high-low)*0.05;
  low-=dif;
  high+=dif;
  BOOLEAN use_lighting=TRUE;
  VolGraphTool gt(title,low,high,pal,use_lighting,winsize);
  const VolGraphTool::COLOR_ARRAY_TYPE *color_array=gt.getColorArray();
  gt.newPage();
  gt.expose();
  for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
    {
      VolGraphTool::DrawDelimiter dd(do_best,&gt);
      VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
        color_array->getColor(0.5);
      for (bool doit=first();doit;doit=next()) {
        if (!used()) continue;
        if (hasChildren()) continue;
        getShape()->draw(xc,vertices,gt);
      }
    }
    if (do_best==TRUE) break;
  }
  gt.eventLoop();
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==3)
void Level3::plotActiveShapes(
const NumPtr<Point<SPACEDIM> > &vertices,VolGraphTool &gt) {
  Palette pal;
  Vector3 low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Vector3 high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  for (bool doit=first();doit;doit=next()) {
    const Shape3 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      for (int i=0;i<3;i++) {
        low[i]=min(low[i],p[i]);
        high[i]=max(high[i],p[i]);
      }
    }
  }
  Vector3 dif=(high-low)*0.05;
  low-=dif;
  high+=dif;
  gt.rescale(low,high);
  const VolGraphTool::COLOR_ARRAY_TYPE *color_array=gt.getColorArray();
  int count=0;
  for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
    {
      VolGraphTool::DrawDelimiter dd(do_best,&gt);
      VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
        color_array->getColor(0.5);
      for (bool doit=first();doit;doit=next()) {
        if (!used()) continue;
        if (hasChildren()) continue;
        getShape()->draw(xc,vertices,gt);
        count++;
      }
    }
    if (do_best==TRUE) break;
  }
  if (count==0) cout << "\tno active shapes on this level" << endl;
  gt.eventLoop();
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==3)
void Level3::plotActiveShapesIndividually(
const NumPtr<Point<SPACEDIM> > &vertices) {
  Palette pal;
  const char *title="Level3::plotActiveShapes";
  REAL winsize=0.5;
  Vector3 low(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Vector3 high(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  for (bool doit=first();doit;doit=next()) {
    const Shape3 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      for (int i=0;i<3;i++) {
        low[i]=min(low[i],p[i]);
        high[i]=max(high[i],p[i]);
      }
    }
  }
  Vector3 dif=(high-low)*0.05;
  low-=dif;
  high+=dif;
  BOOLEAN use_lighting=TRUE;
  VolGraphTool gt(title,low,high,pal,use_lighting,winsize);
  const VolGraphTool::COLOR_ARRAY_TYPE *color_array=gt.getColorArray();
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    gt.newPage();
    gt.expose();
    for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
      {
        VolGraphTool::DrawDelimiter dd(do_best,&gt);
        VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
          color_array->getColor(0.5);
        getShape()->draw(xc,vertices,gt);
        if (hasChildren()) {
          for (int c=0;c<childrenPerShape();c++){
            int child_index=childIndex(c);
            const Shape3 *child_element=
              finerLevel()->getShape(child_index);
            xc=color_array->getColor(static_cast<REAL>(c+1)
                  /static_cast<REAL>(childrenPerShape()));
            child_element->draw(xc,vertices,gt);
          }
        }
      }
      if (do_best==TRUE) break;
    }
    gt.eventLoop();
  }
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==1)
void Level1::plotNeighbors(const NumPtr<Point<SPACEDIM> > &vertices) {
  REAL xlo=HUGE_VAL,xhi=-HUGE_VAL,ylo=-1.,yhi=1.;
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    const Shape1 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[vertexIndex(v)];
      xlo=min(xlo,p[0]);
      xhi=max(xhi,p[0]);
    }
  }
  if (xlo>xhi) {
    cout << "\tLevel1::plotNeighbors: no active elements in this level"
         << endl;
    return;
  }
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  const char *title="Level1::plotNeighbors";
  REAL winsize=0.5;
  XYGraphTool gt(title,"x","y",xlo,xhi,ylo,yhi,&cmap,0,winsize);
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    const Shape1 *element=getShape();
    Point<SPACEDIM> p=
      vertices[vertexIndex(element->numberVertices()-1)];
    gt.movePen(p[0],0.0);
    for (int v=0;v<element->numberVertices();v++) {
      p=vertices[vertexIndex(v)];
      gt.drawLine(p[0],0.0);
    }
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int neighbor_index=neighborIndex(f);
      if (neighbor_index<0) continue;
      const Level1 *neighbor_level=neighborLevel(f);
      REAL y=0.;
      if (neighbor_level->isCoarserThan(this)) {
        gt.setfgColor("red");
        y=0.5;
      } else if (neighbor_level->isFinerThan(this)) {
        gt.setfgColor("blue");
        y=-0.5;
      } else {
        gt.setfgColor("green");
      }
      const Shape1 *neighbor_element=
        neighbor_level->getShape(neighbor_index);
      p=vertices[neighbor_element->
                 vertexIndex(neighbor_element->numberVertices()-1)];
      gt.movePen(p[0],y);
      for (int v=0;v<element->numberVertices();v++) {
        p=vertices[neighbor_element->vertexIndex(v)];
        gt.drawLine(p[0],y);
      }
    }
    gt.flush();
    XYGraphTool::WINDOW_TYPE::QuitButton qb;
  }
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==2)
void Level2::plotNeighbors(const NumPtr<Point<SPACEDIM> > &vertices) {
  REAL rlo[2]={HUGE_VAL,HUGE_VAL},rhi[2]={-HUGE_VAL,-HUGE_VAL};
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    const Shape2 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      rlo[0]=min(rlo[0],p[0]);
      rlo[1]=min(rlo[1],p[1]);
      rhi[0]=max(rhi[0],p[0]);
      rhi[1]=max(rhi[1],p[1]);
    }
  }
  if (rlo[0]>rhi[0] || rlo[1]>rhi[1]) {
    cout << "\tLevel2::plotNeighbors:"
         << " no active elements in this level" << endl;
    return;
  }
  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  const char *title="Level2::plotNeighbors";
  REAL winsize=0.5;
  AreaGraphTool gt(title,"x","y",rlo,rhi,&cmap,0,winsize);
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    const Shape2 *element=getShape();
    element->draw(vertices,gt);
    for (int f=0;f<element->numberCodimension1Shapes();f++) {
      int neighbor_index=neighborIndex(f);
      if (neighbor_index<0) continue;
      Level2 *neighbor_level=neighborLevel(f);
      if (neighbor_level->isCoarserThan(this)) {
        gt.setfgColor("red");
      } else if (neighbor_level->isFinerThan(this)) {
        gt.setfgColor("blue");
      } else {
        gt.setfgColor("green");
      }
      const Shape2 *neighbor_element=
        neighbor_level->getShape(neighbor_index);
      neighbor_element->draw(vertices,gt);
    }
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==1)
void Level1::plotNeighbors(int index,
const NumPtr<Point<SPACEDIM> > &vertices) {
  REAL xlo=HUGE_VAL,xhi=-HUGE_VAL,ylo=-1.,yhi=1.;
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    const Shape1 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[vertexIndex(v)];
      xlo=min(xlo,p[0]);
      xhi=max(xhi,p[0]);
    }
  }
  if (xlo>xhi) {
    cout << "\tLevel1::plotNeighbors: no active elements in this level"
         << endl;
    return;
  }
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  const char *title="Level1::plotNeighbors";
  REAL winsize=0.5;
  XYGraphTool gt(title,"x","y",xlo,xhi,ylo,yhi,&cmap,0,winsize);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();
  const Shape1 *element=getShape(index);
  for (int f=0;f<element->numberCodimension1Shapes();f++) {
    int neighbor_index=neighborIndex(index,f);
    if (neighbor_index<0) continue;
    const Level1 *neighbor_level=neighborLevel(index,f);
    Point<SPACEDIM> p=vertices[element->
      vertexIndex(element->numberVertices()-1)];
    gt.movePen(p[0],0.0);
    for (int v=0;v<element->numberVertices();v++) {
      p=vertices[element->vertexIndex(v)];
      gt.drawLine(p[0],0.0);
    }
    REAL y=0.;
    if (neighbor_level->isCoarserThan(this)) {
      gt.setfgColor("red");
      y=0.5;
    } else if (neighbor_level->isFinerThan(this)) {
      gt.setfgColor("blue");
      y=-0.5;
    } else {
      gt.setfgColor("green");
    }
    const Shape1 *neighbor_element=
      neighbor_level->getShape(neighbor_index);
    p=vertices[neighbor_element->
               vertexIndex(neighbor_element->numberVertices()-1)];
    gt.movePen(p[0],y);
    for (int v=0;v<element->numberVertices();v++) {
      p=vertices[neighbor_element->vertexIndex(v)];
      gt.drawLine(p[0],y);
    }
  }
  gt.flush();
  XYGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#ifdef DEBUG
#if (SPACEDIM==2)
void Level2::plotNeighbors(int index,
const NumPtr<Point<SPACEDIM> > &vertices) {
  REAL rlo[2]={HUGE_VAL,HUGE_VAL},rhi[2]={-HUGE_VAL,-HUGE_VAL};
  for (bool doit=first();doit;doit=next()) {
    if (!used()) continue;
    if (hasChildren()) continue;
    const Shape2 *element=getShape();
    for (int v=0;v<element->numberVertices();v++) {
      const Point<SPACEDIM> &p=vertices[element->vertexIndex(v)];
      rlo[0]=min(rlo[0],p[0]);
      rlo[1]=min(rlo[1],p[1]);
      rhi[0]=max(rhi[0],p[0]);
      rhi[1]=max(rhi[1],p[1]);
    }
  }
  if (rlo[0]>rhi[0] || rlo[1]>rhi[1]) {
    cout << "\tLevel2:::plotNeighbors"
         << " no active elements in this level" << endl;
    return;
  }
  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  const char *title="Level2::plotNeighbors";
  REAL winsize=0.5;
  AreaGraphTool gt(title,"x","y",rlo,rhi,&cmap,0,winsize);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();
  const Shape2 *element=getShape(index);
  for (int f=0;f<element->numberCodimension1Shapes();f++) {
    int neighbor_index=neighborIndex(index,f);
    if (neighbor_index<0) continue;
    Level2 *neighbor_level=neighborLevel(index,f);
    Point<SPACEDIM> p=
      vertices[element->vertexIndex(element->numberVertices()-1)];
    gt.movePen(p[0],p[1]);
    for (int v=0;v<element->numberVertices();v++) {
      p=vertices[element->vertexIndex(v)];
      gt.drawLine(p[0],p[1]);
    }
    if (neighbor_level->isCoarserThan(this)) {
      gt.setfgColor("red");
    } else if (neighbor_level->isFinerThan(this)) {
      gt.setfgColor("blue");
    } else {
      gt.setfgColor("green");
    }
    const Shape2 *neighbor_element=
      neighbor_level->getShape(neighbor_index);
    neighbor_element->draw(vertices,gt);
  }
  gt.flush();
  AreaGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Level0::printOn(ostream &os) const {
  os << "Level0: numberCoarserLevels = " << numberCoarserLevels()
     << "\n\tcoarser_level = " << coarser_level 
     << ", finer_level = " << finer_level << endl;
//for (int i=0;i<refine_flags.getNumber();i++) {
//  os << "\trefine_flags[" << i << "] = " << refine_flags[i] << endl;
//}
//for (int i=0;i<coarsen_flags.getNumber();i++) {
//  os << "\tcoarsen_flags[" << i << "] = " << coarsen_flags[i] << endl;
//}
//for (int i=0;i<neighbors.getNumber();i++) {
//  os << "\tneighbors[" << i << "] = " << neighbors[i].first << " "
//     << neighbors[i].second << endl;
//}
//for (int i=0;i<subdomain_ids.getNumber();i++) {
//  os << "\tsubdomain_ids[" << i << "] = " << subdomain_ids[i] << endl;
//}
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Level1::printOn(ostream &os) const {
  os << "Level1: numberCoarserLevels = " << numberCoarserLevels()
     << "\n\tpresent_index1 = " << present_index1 
     << "\n\telement_data1.getNumber = " << element_data1.getNumber()
     << endl;
//for (int i=0;i<element_data1.getNumber();i++) {
//  os << "\telement_data1[" << i << "] = " << &element_data1[i] 
//     << endl;
//  element_data1[i].printOn(os);
//  os << endl;
//}
//for (int i=0;i<element_data1.getNumber();i++) {
//  os << "\tShape[" << i << "] = " << element_data1[i].getShape() 
//     << endl;
//  if (element_data1[i].getShape()) {
//    element_data1[i].getShape()->printOn(os);
//  }
//}
//for (int i=0;i<element_data1.getNumber();i++) {
//  os << "\tchildIndex[" << i << "] = " 
//     << element_data1[i].childIndex() << endl;
//}
//for (int i=0;i<element_data1.getNumber();i++) {
//  os << "\tused[" << i << "] = " 
//     << PRINT_BOOLEAN(element_data1[i].isUsed()) << endl;
//}
//for (int i=0;i<element_data1.getNumber();i++) {
//  os << "\tuser_flag[" << i << "] = " 
//     << PRINT_BOOLEAN(element_data1[i].userFlagSet()) << endl;
//}
//for (int i=0;i<element_data1.getNumber();i++) {
//  os << "\tmaterial_id[" << i << "] = " 
//     << element_data1[i].getMaterialId() << endl;
//}
//for (int i=0;i<element_data1.getNumber();i++) {
//  os << "\tuserPointer[" << i << "] = " 
//     << PRINT_BOOLEAN(element_data1[i].userPointer()) << endl;
//}
  Level0::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>=2)
void Level2::printOn(ostream &os) const {
  os << "Level2: numberCoarserLevels = " << numberCoarserLevels()
     << "\n\tpresent_index2 = " << present_index2 
     << "\n\telement_data2.getNumber = " << element_data2.getNumber()
     << endl;
//for (int i=0;i<element_data2.getNumber();i++) {
//  os << "\telement_data2[" << i << "] = " << &element_data2[i] 
//     << endl;
//  element_data2[i].printOn(os);
//  os << endl;
//}
//for (int i=0;i<element_data2.getNumber();i++) {
//  os << "\tShape[" << i << "] = " << element_data2[i].getShape() 
//     << endl;
//  if (element_data2[i].getShape()) {
//    element_data2[i].getShape()->printOn(os);
//  }
//}
//for (int i=0;i<element_data2.getNumber();i++) {
//  os << "\tchildIndex[" << i << "] = " 
//     << element_data2[i].childIndex() << endl;
//}
//for (int i=0;i<element_data2.getNumber();i++) {
//  os << "\tused[" << i << "] = " 
//     << PRINT_BOOLEAN(element_data2[i].isUsed()) << endl;
//}
//for (int i=0;i<element_data2.getNumber();i++) {
//  os << "\tuser_flag[" << i << "] = " 
//     << PRINT_BOOLEAN(element_data2[i].userFlagSet()) << endl;
//}
//for (int i=0;i<element_data2.getNumber();i++) {
//  os << "\tmaterial_id[" << i << "] = " 
//     << element_data2[i].getMaterialId() << endl;
//}
//for (int i=0;i<element_data2.getNumber();i++) {
//  os << "\tuserPointer[" << i << "] = " 
//     << PRINT_BOOLEAN(element_data2[i].userPointer()) << endl;
//}
  Level1::printOn(os);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void Level3::printOn(ostream &os) const {
  os << "Level3: numberCoarserLevels = " << numberCoarserLevels()
     << "\n\tpresent_index3 = " << present_index3 << endl;
  for (int i=0;i<element_data3.getNumber();i++) {
    os << "\telement_data3[" << i << "] = " << &element_data3[i] 
       << endl;
//  element_data3[i].printOn(os);
//  os << endl;
  }
//for (int i=0;i<element_data3.getNumber();i++) {
//  os << "\tShape[" << i << "] = " << element_data3[i].getShape() 
//     << endl;
//  if (element_data3[i].getShape()) {
//    element_data3[i].getShape()->printOn(os);
//  }
//}
//for (int i=0;i<element_data3.getNumber();i++) {
//  os << "\tchildIndex[" << i << "] = " 
//     << element_data3[i].childIndex() << endl;
//}
//for (int i=0;i<element_data3.getNumber();i++) {
//  os << "\tused[" << i << "] = " 
//     << PRINT_BOOLEAN(element_data3[i].isUsed()) << endl;
//}
//for (int i=0;i<element_data3.getNumber();i++) {
//  os << "\tuser_flag[" << i << "] = " 
//     << PRINT_BOOLEAN(element_data3[i].userFlagSet()) << endl;
//}
//for (int i=0;i<element_data3.getNumber();i++) {
//  os << "\tmaterial_id[" << i << "] = " 
//     << element_data3[i].getMaterialId() << endl;
//}
//for (int i=0;i<element_data3.getNumber();i++) {
//  os << "\tuserPointer[" << i << "] = " 
//     << PRINT_BOOLEAN(element_data3[i].userPointer()) << endl;
//}
  Level2::printOn(os);
}
#endif

#include <NumPtr.C>
template class NumPtrBase<ShapeData,int>;
template class NumPtr<ShapeData>;
INSTANTIATE_NUMPTR( const BoundaryCondition* );
INSTANTIATE_NUMPTR( MaterialModel* );
template class NumPtrBase<pair<const BoundaryCondition*,const BoundaryCondition*>,int>;
template class NumPtr<pair<const BoundaryCondition*,const BoundaryCondition*> >;

template int* fill_n<int*,int,int>(int*,int,int const&);
template 
  __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > > 
  set_intersection<
    int const*, 
    int const*, 
    __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > > 
  >(int const*, 
    int const*, 
    int const*, 
    int const*, 
    __gnu_cxx::__normal_iterator<int*, vector<int, allocator<int> > >);
