// "$Header:$"
//------------------------------------------------------------------------
//  $Id: mapping.cc,v 1.7 2003/01/08 17:58:01 wolf Exp $
//  Version: $Name: Version-4-0-0 $
//
//  Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
//
//------------------------------------------------------------------------
//----------------------------  mapping_q.cc  ---------------------------
//  $Id: mapping_q.cc,v 1.46 2003/04/07 15:42:17 wolf Exp $
//  Version: $Name: Version-4-0-0 $
//
//  Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
//
//----------------------------  mapping_q.cc  ---------------------------
//
//modified from deal.II/deal.II/source/fe/mapping.cc and mapping_q.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <Mapping.H>
#include <Boundary.H>
#include <Matrix.H>

bool MappingDataBase::use_full_degree_mapping_on_all_cells=false;
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
ostream& operator<<(ostream &os,UPDATE_FLAGS f) {
  if (f==UPDATE_DEFAULT) os << "update_default";
  else {
    if (f & UPDATE_VALUES) os << "update_values ";
    if (f & UPDATE_GRADIENTS) os << "update_gradients ";
    if (f & UPDATE_SECOND_DERIVATIVES) {
      os << "update_second_derivatives ";
    }
    if (f & UPDATE_DIVERGENCE) os << "update_divergence ";
    if (f & UPDATE_CURL) os << "update_curl ";
    if (f & UPDATE_BOUNDARY_FORMS) os << "update_boundary_forms ";
    if (f & UPDATE_Q_POINTS) os << "update_q_points ";
    if (f & UPDATE_DETERMINANT) os << "update_determinant ";
    if (f & UPDATE_JxW_VALUES) os << "update_jxw_values ";
    if (f & UPDATE_NORMAL_VECTORS) os << "update_normal_vectors ";
    if (f & UPDATE_MAPPING_JACOBIAN_INVERSE) {
      os << "update_mapping_jacobian_inverse ";
    }
    if (f & UPDATE_MAPPING_JACOBIAN) {
      os << "update_mapping_jacobian ";
    }
    if (f & UPDATE_TRANSFORMATION_VALUES) {
      os << "update_transformation_values ";
    }
    if (f & UPDATE_TRANSFORMATION_GRADIENTS) {
      os << "update_transformation_gradients ";
    }
    if (f & UPDATE_TRANSFORMATION_SECOND_DERIVATIVE) {
      os << "update_transformation_second_derivative ";
    }
  }
  return os;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
MappingDataBase::MappingDataBase(int p) :
update_once(UPDATE_DEFAULT),update_each(UPDATE_DEFAULT),
shape_values(0,0),shape_derivatives(0,0),
n_shape_functions(0),n_higher_order_shape_functions(0),
map_higher_order(false),shape_values_ok(false),
higher_order_shape_values_ok(false),shape_derivatives_ok(false),
higher_order_shape_derivatives_ok(false),
shape_second_derivatives_ok(false),
higher_order_shape_second_derivatives_ok(false),
mapping_support_points_ok(false),
higher_order_mapping_support_points_ok(false),mapping_jacobian_ok(false),
higher_order_mapping_jacobian_ok(false),
mapping_jacobian_inverse_ok(false),
higher_order_mapping_jacobian_inverse_ok(false),
mapping_second_derivative_ok(false),
higher_order_mapping_second_derivative_ok(false),
mapping_determinant_ok(false),
higher_order_mapping_determinant_ok(false),
shape_function(0),higher_order_shape_function(0),
#if (SPACEDIM==1)
degree(1)
#else
degree(p)
#endif
{
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//Flags for quantities that DO NOT depend on the Shape
void MappingDataBase::setUpdateOnce(UPDATE_FLAGS in) {
  update_once=in & (UPDATE_TRANSFORMATION_VALUES |
                    UPDATE_TRANSFORMATION_GRADIENTS |
                    UPDATE_TRANSFORMATION_SECOND_DERIVATIVE);
  if (in & UPDATE_Q_POINTS) update_once |= UPDATE_TRANSFORMATION_VALUES;
  if (in & (UPDATE_MAPPING_JACOBIAN_INVERSE
  | UPDATE_MAPPING_JACOBIAN | UPDATE_JxW_VALUES
  | UPDATE_BOUNDARY_FORMS | UPDATE_NORMAL_VECTORS)) {
    update_once |= UPDATE_TRANSFORMATION_GRADIENTS | UPDATE_DETERMINANT;
  }
  if (in & UPDATE_MAPPING_SECOND_DERIVATIVE) {
    update_once |= UPDATE_TRANSFORMATION_SECOND_DERIVATIVE;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//Flags for quantities that depend on the Shape
void MappingDataBase::setUpdateEach(UPDATE_FLAGS in) {
  update_each=in & (UPDATE_Q_POINTS | UPDATE_MAPPING_JACOBIAN_INVERSE
    | UPDATE_MAPPING_JACOBIAN | UPDATE_MAPPING_SECOND_DERIVATIVE
    | UPDATE_JxW_VALUES | UPDATE_BOUNDARY_FORMS 
    | UPDATE_NORMAL_VECTORS | UPDATE_DETERMINANT );
  if (update_each & UPDATE_JxW_VALUES) {
    update_each |= UPDATE_DETERMINANT;
  }
  if (update_each & (UPDATE_NORMAL_VECTORS | UPDATE_JxW_VALUES)) {
    update_each |= UPDATE_BOUNDARY_FORMS;
  }
  if (update_each & 
  (UPDATE_BOUNDARY_FORMS | UPDATE_MAPPING_JACOBIAN_INVERSE) ) {
    update_each |= UPDATE_MAPPING_JACOBIAN;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingDataBase::allocateSpace(UPDATE_FLAGS uf,
const Quadrature<SPACEDIM> *q) {
  shape_values_ok=false;
  higher_order_shape_values_ok=false;
  shape_derivatives_ok=false;
  higher_order_shape_derivatives_ok=false;
  mapping_support_points_ok=false;
  higher_order_mapping_support_points_ok=false;

  int npts=q->numberPoints();
  n_shape_functions=shape_function->getNumber();
  n_higher_order_shape_functions=(higher_order_shape_function==0
    ? 0 : higher_order_shape_function->getNumber());
  mapping_support_points.allocate(n_shape_functions);
  higher_order_mapping_support_points.
    allocate(n_higher_order_shape_functions);

  setUpdateOnce(uf);
  setUpdateEach(uf);

  if (update_once & UPDATE_TRANSFORMATION_VALUES) {
    if (n_shape_functions>0) {
      shape_values.reinit(n_shape_functions,npts);
    }
    if (n_higher_order_shape_functions>0) {
      higher_order_shape_values.
        reinit(n_higher_order_shape_functions,npts);
    }
  }
  if (update_once & UPDATE_TRANSFORMATION_GRADIENTS) {
    if (n_shape_functions>0) {
      shape_derivatives.reinit(n_shape_functions,npts);
    }
    if (n_higher_order_shape_functions>0) {
      higher_order_shape_derivatives.
        reinit(n_higher_order_shape_functions,npts);
    }
  }
  if (update_once & UPDATE_TRANSFORMATION_SECOND_DERIVATIVE) {
    if (n_shape_functions>0) {
      shape_second_derivatives.reinit(n_shape_functions,npts);
    }
    if (n_higher_order_shape_functions>0) {
      higher_order_shape_second_derivatives.
        reinit(n_higher_order_shape_functions,npts);
    }
  }
#if (SPACEDIM>1)
  const int nfaces=facesPerShape(n_shape_functions);
  reference_normal.allocate(nfaces);
  for (int f=0;f<nfaces;f++) {
//  reference_normal = unit outer normal * face measure 
//                                       / canonical face measure
    reference_normal[f]=faceNormal(n_shape_functions,f);
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//q must not be destructed before all calls to computeShapesFor
void MappingDataBase::computeShapes(const Quadrature<SPACEDIM> *q) {
  quadrature=q;
  shape_values_ok=false;
  higher_order_shape_values_ok=false;
  shape_derivatives_ok=false;
  higher_order_shape_derivatives_ok=false;
  mapping_jacobian_ok=false;
  higher_order_mapping_jacobian_ok=false;
  mapping_jacobian_inverse_ok=false;
  higher_order_mapping_jacobian_inverse_ok=false;

  const NumPtr<Point<SPACEDIM> > &canonical_points=quadrature->points();
  int n_points=canonical_points.getNumber();
  if (shape_function && 
  !(shape_values.empty() && shape_derivatives.empty() &&
  shape_second_derivatives.empty())) {
    if (!shape_function->requiresShapeToComputeValues()) {
#ifdef DEBUG
      if (!shape_values.empty()) {
        CHECK_SAME(shape_values.size(0),n_shape_functions); 
        CHECK_SAME(shape_values.size(1),n_points); 
      }
      if (!shape_derivatives.empty()) {
        CHECK_SAME(shape_derivatives.size(0),n_shape_functions);
        CHECK_SAME(shape_derivatives.size(1),n_points);
      }
#endif
      for (int p=0;p<n_points;p++) {
        if (!shape_values.empty()) {
          NumPtr<REAL> values;
          shape_function->computeValues(canonical_points[p],values);
          for (int s=0;s<values.getNumber();s++) {
            shape_values(s,p)=values[s];
          }
        }
        if (!shape_derivatives.empty()) {
          NumPtr<Tensor<1,SPACEDIM> > grads;
          shape_function->computeGrads(canonical_points[p],grads);
          for (int s=0;s<grads.getNumber();s++) {
            shape_derivatives(s,p)=grads[s];
          }
        }
        if (!shape_second_derivatives.empty()) {
          NumPtr<Tensor<2,SPACEDIM> > grad_grads;
          shape_function->computeGradGrads(canonical_points[p],grad_grads);
          for (int s=0;s<grad_grads.getNumber();s++) {
            shape_second_derivatives(s,p)=grad_grads[s];
          }
        }
      }
      shape_values_ok=!shape_values.empty();
      shape_derivatives_ok=!shape_derivatives.empty();
      shape_second_derivatives_ok=!shape_second_derivatives.empty();
    }
  }
  if (higher_order_shape_function && !(higher_order_shape_values.empty()
  && higher_order_shape_derivatives.empty() &&
  higher_order_shape_second_derivatives.empty())) {
    if (!higher_order_shape_function->requiresShapeToComputeValues()) {
#ifdef DEBUG
      if (!higher_order_shape_values.empty()) {
        CHECK_SAME(higher_order_shape_values.size(0),
                   n_higher_order_shape_functions); 
        CHECK_SAME(higher_order_shape_values.size(1),n_points); 
      }
      if (!higher_order_shape_derivatives.empty()) {
        CHECK_SAME(higher_order_shape_derivatives.size(0),
                   n_higher_order_shape_functions);
        CHECK_SAME(higher_order_shape_derivatives.size(1),n_points);
      }
#endif
      for (int p=0;p<n_points;p++) {
        if (!higher_order_shape_values.empty()) {
          NumPtr<REAL> values;
          higher_order_shape_function->
            computeValues(canonical_points[p],values);
          for (int s=0;s<higher_order_shape_values.size(0);s++) {
            higher_order_shape_values(s,p)=values[s];
          }
        }
        if (!higher_order_shape_derivatives.empty()) {
          NumPtr<Tensor<1,SPACEDIM> > grads;
          higher_order_shape_function->
            computeGrads(canonical_points[p],grads);
          for (int s=0;s<higher_order_shape_derivatives.size(0);s++) {
            higher_order_shape_derivatives(s,p)=grads[s];
          }
        }
        if (!higher_order_shape_second_derivatives.empty()) {
          NumPtr<Tensor<2,SPACEDIM> > grad_grads;
          higher_order_shape_function->
            computeGradGrads(canonical_points[p],grad_grads);
          for (int s=0;s<higher_order_shape_second_derivatives.size(0);
          s++) {
            higher_order_shape_second_derivatives(s,p)=grad_grads[s];
          }
        }
      }
      higher_order_shape_values_ok=!higher_order_shape_values.empty();
      higher_order_shape_derivatives_ok=
        !higher_order_shape_derivatives.empty();
      higher_order_shape_second_derivatives_ok=
        !higher_order_shape_second_derivatives.empty();
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingDataBase::computeShapesFor(const Shape *e) {
  const NumPtr<Point<SPACEDIM> > &quad_points=quadrature->points();
  int n_points=quad_points.getNumber();
  if (shape_function && 
  !(shape_values.empty() && shape_derivatives.empty() &&
  shape_second_derivatives.empty())) {
    if (shape_function->requiresShapeToComputeValues()) {
#ifdef DEBUG
      if (!shape_values.empty()) {
        CHECK_SAME(shape_values.size(0),n_shape_functions); 
        CHECK_SAME(shape_values.size(1),n_points); 
      }
      if (!shape_derivatives.empty()) {
        CHECK_SAME(shape_derivatives.size(0),n_shape_functions);
        CHECK_SAME(shape_derivatives.size(1),n_points);
      }
      if (!shape_second_derivatives.empty()) {
        CHECK_SAME(shape_second_derivatives.size(0),n_shape_functions);
        CHECK_SAME(shape_second_derivatives.size(1),n_points);
      }
#endif
      for (int p=0;p<n_points;p++) {
        if (!shape_values.empty()) {
          NumPtr<REAL> values;
          shape_function->computeValuesFor(e,quad_points[p],values);
          for (int s=0;s<values.getNumber();s++) {
            shape_values(s,p)=values[s];
          }
        }
        if (!shape_derivatives.empty()) {
          NumPtr<Tensor<1,SPACEDIM> > grads;
          shape_function->computeGradsFor(e,quad_points[p],grads);
          for (int s=0;s<grads.getNumber();s++) {
            shape_derivatives(s,p)=grads[s];
          }
        }
        if (!shape_second_derivatives.empty()) {
          NumPtr<Tensor<2,SPACEDIM> > grad_grads;
          shape_function->computeGradGradsFor(e,quad_points[p],grad_grads);
          for (int s=0;s<grad_grads.getNumber();s++) {
            shape_second_derivatives(s,p)=grad_grads[s];
          }
        }
      }
#ifdef DEBUG
      for (int p=0;p<shape_values.size(1);p++) {
        for (int s=0;s<shape_values.size(0);s++) {
          cout << "\tshape_values(" << s << "," << p << ") = "
               << shape_values(s,p) << endl;
        }
      }
      for (int p=0;p<shape_derivatives.size(1);p++) {
        for (int s=0;s<shape_derivatives.size(0);s++) {
          cout << "\tshape_derivatives(" << s << "," << p << ") = "
               << shape_derivatives(s,p) << endl;
        }
      }
      for (int p=0;p<shape_second_derivatives.size(1);p++) {
        for (int s=0;s<shape_second_derivatives.size(0);s++) {
          cout << "\tshape_second_derivatives(" << s << "," << p << ") = "
               << shape_second_derivatives(s,p) << endl;
        }
      }
#endif
      shape_values_ok=!shape_values.empty();
      shape_derivatives_ok=!shape_derivatives.empty();
      shape_second_derivatives_ok=!shape_second_derivatives.empty();
    }
  }
  if (higher_order_shape_function && !(higher_order_shape_values.empty()
  && higher_order_shape_derivatives.empty())) {
    if (higher_order_shape_function->requiresShapeToComputeValues()) {
#ifdef DEBUG
      if (!higher_order_shape_values.empty()) {
        CHECK_SAME(higher_order_shape_values.size(0),
                   n_higher_order_shape_functions); 
        CHECK_SAME(higher_order_shape_values.size(1),n_points); 
      }
      if (!higher_order_shape_derivatives.empty()) {
        CHECK_SAME(higher_order_shape_derivatives.size(0),
                   n_higher_order_shape_functions);
        CHECK_SAME(higher_order_shape_derivatives.size(1),n_points);
      }
      if (!higher_order_shape_second_derivatives.empty()) {
        CHECK_SAME(higher_order_shape_second_derivatives.size(0),
                   n_higher_order_shape_functions);
        CHECK_SAME(higher_order_shape_second_derivatives.size(1),n_points);
      }
#endif
      for (int p=0;p<n_points;p++) {
        if (!higher_order_shape_values.empty()) {
          NumPtr<REAL> values;
          higher_order_shape_function->
            computeValuesFor(e,quad_points[p],values);
          for (int s=0;s<higher_order_shape_values.size(0);s++) {
            higher_order_shape_values(s,p)=values[s];
          }
        }
        if (!higher_order_shape_derivatives.empty()) {
          NumPtr<Tensor<1,SPACEDIM> > grads;
          higher_order_shape_function->
            computeGradsFor(e,quad_points[p],grads);
          for (int s=0;s<higher_order_shape_derivatives.size(0);s++) {
            higher_order_shape_derivatives(s,p)=grads[s];
          }
        }
        if (!higher_order_shape_second_derivatives.empty()) {
          NumPtr<Tensor<2,SPACEDIM> > grad_grads;
          higher_order_shape_function->
            computeGradGradsFor(e,quad_points[p],grad_grads);
          for (int s=0;s<higher_order_shape_second_derivatives.size(0);
          s++) {
            higher_order_shape_second_derivatives(s,p)=grad_grads[s];
          }
        }
      }
#ifdef DEBUG
      for (int p=0;p<higher_order_shape_derivatives.size(1);p++) {
        for (int s=0;s<higher_order_shape_derivatives.size(0);s++) {
          cout << "\thigher_order_shape_derivatives(" << s << "," << p 
               << ") = " << higher_order_shape_derivatives(s,p) << endl;
        }
      }
      for (int p=0;p<higher_order_shape_second_derivatives.size(1);p++) {
        for (int s=0;s<higher_order_shape_second_derivatives.size(0);s++) {
          cout << "\thigher_order_shape_second_derivatives(" << s
               << "," << p << ") = "
               << higher_order_shape_second_derivatives(s,p) << endl;
        }
      }
#endif
      higher_order_shape_values_ok=!higher_order_shape_values.empty();
      higher_order_shape_derivatives_ok=
        !higher_order_shape_derivatives.empty();
      higher_order_shape_second_derivatives_ok=
        !higher_order_shape_second_derivatives.empty();
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingDataBase::computeJacobian(const SHAPE_CLASS *element,
const Tessellation &tess,const LEVEL_CLASS *level,int npts,
int offset,NumPtr<Point<SPACEDIM> > &quadrature_points,
bool &quadrature_points_ok) {
  computeMappingSupportPoints(element,tess,level);
  if (update_each & UPDATE_Q_POINTS) {
    CHECK_SAME(quadrature_points.getNumber(),npts);
    Point<SPACEDIM> p(D_DECL(0.,0.,0.));
    quadrature_points.initialize(p);
    if (map_higher_order) {
      CHECK_TEST(higher_order_shape_values_ok);
      CHECK_TEST(higher_order_mapping_support_points_ok);
      for (int p=0;p<npts;p++) {
        for (int k=0;k<n_higher_order_shape_functions;k++) {
          quadrature_points[p]+= 
            higher_order_mapping_support_points[k]
            *higher_order_shape_values(k,p+offset);
        }
      }
    } else {
      CHECK_TEST(shape_values_ok);
      CHECK_TEST(mapping_support_points_ok);
      for (int p=0;p<npts;p++) {
        for (int k=0;k<n_shape_functions;k++) {
          quadrature_points[p]+= 
            mapping_support_points[k]*shape_values(k,p+offset);
        }
      }
    }
    quadrature_points_ok=true;
  }
  if (update_each & (UPDATE_MAPPING_JACOBIAN | 
  UPDATE_MAPPING_JACOBIAN_INVERSE | UPDATE_BOUNDARY_FORMS)) {
    if (map_higher_order) {
      CHECK_TEST(higher_order_shape_derivatives_ok);
      CHECK_TEST(higher_order_mapping_support_points_ok);
      if (higher_order_mapping_jacobian.getNumber()!=npts) {
        higher_order_mapping_jacobian.cleanup();
        higher_order_mapping_jacobian.allocate(npts);
      }
      Tensor<2,SPACEDIM> tensor;
      higher_order_mapping_jacobian.initialize(tensor);
      for (int p=0;p<npts;p++) {
        Tensor<2,SPACEDIM> &c=higher_order_mapping_jacobian[p];
        for (int k=0;k<n_higher_order_shape_functions;k++) {
          const Point<SPACEDIM> &msp=
            higher_order_mapping_support_points[k];
          const Tensor<1,SPACEDIM> &sd=
            higher_order_shape_derivatives(k,p+offset);
          for (int i=0; i<SPACEDIM; ++i) {
            Tensor<1,SPACEDIM> &ci=c[i];
            const REAL &mspi=msp[i];
            for (int j=0;j<SPACEDIM;j++) {
              ci[j]+=mspi*sd[j];
            }
          }
        }
      }
      higher_order_mapping_jacobian_ok=true;
    } else {
      CHECK_TEST(shape_derivatives_ok);
      CHECK_TEST(mapping_support_points_ok);
      if (mapping_jacobian.getNumber()!=npts) {
        mapping_jacobian.cleanup();
        mapping_jacobian.allocate(npts);
      }
      Tensor<2,SPACEDIM> tensor;
      mapping_jacobian.initialize(tensor);
      for (int p=0;p<npts;p++) {
        Tensor<2,SPACEDIM> &c=mapping_jacobian[p];
        for (int k=0;k<n_shape_functions;k++) {
          const Point<SPACEDIM> &msp=mapping_support_points[k];
          const Tensor<1,SPACEDIM> &sd=shape_derivatives(k,p+offset);
          for (int i=0; i<SPACEDIM; ++i) {
            Tensor<1,SPACEDIM> &ci=c[i];
            const REAL &mspi=msp[i];
            for (int j=0;j<SPACEDIM;j++) {
              ci[j]+=mspi*sd[j];
            }
          }
        }
      }
      mapping_jacobian_ok=true;
    }
  }
  if (update_each & UPDATE_DETERMINANT) {
    if (map_higher_order) {
      CHECK_TEST(higher_order_mapping_jacobian_ok);
      if (higher_order_mapping_determinant.getNumber()!=npts) {
        higher_order_mapping_determinant.cleanup();
        higher_order_mapping_determinant.allocate(npts);
      }
      for (int p=0;p<npts;p++) {
        higher_order_mapping_determinant[p]=
          determinant(higher_order_mapping_jacobian[p]);
      }
      higher_order_mapping_determinant_ok=true;
    } else {
      CHECK_TEST(mapping_jacobian_ok);
      if (mapping_determinant.getNumber()!=npts) {
        mapping_determinant.cleanup();
        mapping_determinant.allocate(npts);
      }
      for (int p=0;p<npts;p++) {
        mapping_determinant[p]=determinant(mapping_jacobian[p]);
      }
      mapping_determinant_ok=true;
    }
  }
  if (update_each & UPDATE_MAPPING_JACOBIAN_INVERSE) {
    if (map_higher_order) {
      if (higher_order_mapping_jacobian_inverse.getNumber()!=npts) {
        higher_order_mapping_jacobian_inverse.cleanup();
        higher_order_mapping_jacobian_inverse.allocate(npts);
      }
      for (int p=0;p<npts;p++) {
        higher_order_mapping_jacobian_inverse[p]
        =invert(higher_order_mapping_jacobian[p]);
      }
      higher_order_mapping_jacobian_inverse_ok=true;
    } else {
      if (mapping_jacobian_inverse.getNumber()!=npts) {
        mapping_jacobian_inverse.cleanup();
        mapping_jacobian_inverse.allocate(npts);
      }
      for (int p=0;p<npts;p++) {
        mapping_jacobian_inverse[p]=invert(mapping_jacobian[p]);
      }
      mapping_jacobian_inverse_ok=true;
    }
  }
  if (update_each & UPDATE_MAPPING_SECOND_DERIVATIVE) {
    if (map_higher_order) {
      CHECK_TEST(higher_order_shape_second_derivatives_ok);
      CHECK_TEST(higher_order_mapping_support_points_ok);
      if (higher_order_mapping_second_derivative.getNumber()!=npts) {
        higher_order_mapping_second_derivative.cleanup();
        higher_order_mapping_second_derivative.allocate(npts);
      }
      Tensor<3,SPACEDIM> tensor;
      higher_order_mapping_second_derivative.initialize(tensor);
      for (int p=0;p<npts;p++) {
        Tensor<3,SPACEDIM> &gg=higher_order_mapping_second_derivative[p];
        for (int k=0;k<n_higher_order_shape_functions;k++) {
          const Point<SPACEDIM> &msp=
            higher_order_mapping_support_points[k];
          const Tensor<2,SPACEDIM> &sd=
            higher_order_shape_second_derivatives(k,p+offset);
          for (int i=0; i<SPACEDIM; ++i) {
            Tensor<2,SPACEDIM> &ggi=gg[i];
            const REAL &mspi=msp[i];
            for (int j=0;j<SPACEDIM;j++) {
              for (int k=0;k<SPACEDIM;k++) {
                ggi[j][k]+=mspi*sd[j][k];
              }
            }
          }
        }
      }
      higher_order_mapping_second_derivative_ok=true;
    } else {
      CHECK_TEST(shape_second_derivatives_ok);
      CHECK_TEST(mapping_support_points_ok);
      if (mapping_second_derivative.getNumber()!=npts) {
        mapping_second_derivative.cleanup();
        mapping_second_derivative.allocate(npts);
      }
      Tensor<3,SPACEDIM> tensor;
      mapping_second_derivative.initialize(tensor);
      for (int p=0;p<npts;p++) {
        Tensor<3,SPACEDIM> &gg=mapping_second_derivative[p];
        for (int k=0;k<n_shape_functions;k++) {
          const Point<SPACEDIM> &msp=mapping_support_points[k];
          const Tensor<2,SPACEDIM> &sd=
            shape_second_derivatives(k,p+offset);
          for (int i=0; i<SPACEDIM; ++i) {
            Tensor<2,SPACEDIM> &ggi=gg[i];
            const REAL &mspi=msp[i];
            for (int j=0;j<SPACEDIM;j++) {
              for (int k=0;k<SPACEDIM;k++) {
              ggi[j][k]+=mspi*sd[j][k];
              }
            }
          }
        }
      }
      mapping_second_derivative_ok=true;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingDataBase::computeFEValues(const SHAPE_CLASS *element,
const Tessellation &tess,const LEVEL_CLASS *level,
NumPtr<Point<SPACEDIM> > &quadrature_points,NumPtr<REAL> &JxW_values,
bool &quadrature_points_ok,bool &JxW_values_ok) {
  int npts=quadrature->numberPoints();
  if (requiresShapeToInitialize()) computeShapesFor(element);      
  computeJacobian(element,tess,level,npts,0,quadrature_points,
    quadrature_points_ok);
  if (update_each & UPDATE_JxW_VALUES) {
    const NumPtr<REAL> &weights=quadrature->weights();
    CHECK_SAME(JxW_values.getNumber(),npts);
    if (map_higher_order) {
      CHECK_TEST(higher_order_mapping_determinant_ok);
      for (int p=0;p<npts;p++) {
        JxW_values[p]=higher_order_mapping_determinant[p]*weights[p];
      }
    } else {
      CHECK_TEST(mapping_determinant_ok);
      for (int p=0;p<npts;p++) {
        JxW_values[p]=mapping_determinant[p]*weights[p];
      }
    }
    JxW_values_ok=true;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingDataBase::preMultiplyByJacobian(
Tensor<1,SPACEDIM> &dest,int q,const Tensor<1,SPACEDIM> &src) const {
  ASSERT((map_higher_order && higher_order_mapping_jacobian_ok) ||
         (!map_higher_order && mapping_jacobian_ok));
  product(dest,
    (map_higher_order ? higher_order_mapping_jacobian[q] 
                      : mapping_jacobian[q]),src);
}

void MappingDataBase::preMultiplyByJacobian(
Tensor<2,SPACEDIM> &dest,int q,const Tensor<2,SPACEDIM> &src) const {
  ASSERT((map_higher_order && higher_order_mapping_jacobian_ok) ||
         (!map_higher_order && mapping_jacobian_ok));
  product(dest,
    map_higher_order ? higher_order_mapping_jacobian[q] 
                     : mapping_jacobian[q], src);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingDataBase::preMultiplyByJacobianTranspose(
Tensor<1,SPACEDIM> &dest,int q,const Tensor<1,SPACEDIM> &src) const {
  ASSERT((map_higher_order && higher_order_mapping_jacobian_ok) ||
         (!map_higher_order && mapping_jacobian_ok));
  transposeProduct(dest,
    (map_higher_order ? higher_order_mapping_jacobian[q] 
                      : mapping_jacobian[q]),src);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingDataBase::postMultiplyByJacobianInverse(
Tensor<1,SPACEDIM> &dest,int q,const Tensor<1,SPACEDIM> &src) const {
  CHECK_TEST((map_higher_order && higher_order_mapping_jacobian_inverse_ok)
    || (!map_higher_order && mapping_jacobian_inverse_ok));
  product(dest,src,
    map_higher_order ? higher_order_mapping_jacobian_inverse[q] 
                     : mapping_jacobian_inverse[q]);
}
void MappingDataBase::postMultiplyByJacobianInverse(
Tensor<2,SPACEDIM> &dest,int q,const Tensor<2,SPACEDIM> &src) const {
  CHECK_TEST((map_higher_order && higher_order_mapping_jacobian_inverse_ok)
    || (!map_higher_order && mapping_jacobian_inverse_ok));
  product(dest,src,
    map_higher_order ? higher_order_mapping_jacobian_inverse[q] 
                     : mapping_jacobian_inverse[q]);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingDataBase::hdivTransform(Tensor<1,SPACEDIM> &dest,int q,
const Tensor<1,SPACEDIM> &src) const {
  preMultiplyByJacobian(dest,q,src);
  if (map_higher_order) {
    CHECK_TEST(higher_order_mapping_determinant_ok);
    dest/=higher_order_mapping_determinant[q];
  } else {
    CHECK_TEST(mapping_determinant_ok);
    dest/=mapping_determinant[q];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingDataBase::hdivTransformDerivative(
Tensor<2,SPACEDIM> &dest,int q,const Tensor<1,SPACEDIM> &src,
const Tensor<2,SPACEDIM> &src_deriv) const {
  if (map_higher_order) {
    CHECK_TEST(higher_order_mapping_determinant_ok);
    CHECK_TEST(higher_order_mapping_jacobian_ok);
    CHECK_TEST(higher_order_mapping_jacobian_inverse_ok);
    CHECK_TEST(higher_order_mapping_second_derivative_ok);
    Tensor<2,SPACEDIM> temp1=src_deriv;
    Tensor<2,SPACEDIM> determinant_derivative=
      determinantDerivative(higher_order_mapping_jacobian[q]);
    for (int k=0;k<SPACEDIM;k++) {
      REAL d_det_d_xi_k=0.;
      for (int i=0;i<SPACEDIM;i++) {
        for (int j=0;j<SPACEDIM;j++) {
          d_det_d_xi_k+=determinant_derivative[i][j]
            *higher_order_mapping_second_derivative[q][i][j][k];
        }
      }
      d_det_d_xi_k/=higher_order_mapping_determinant[q];;
      for (int i=0;i<SPACEDIM;i++) {
        temp1[i][k]-=src[i]*d_det_d_xi_k;
      }
    }
    Tensor<2,SPACEDIM> temp2;
    preMultiplyByJacobian(temp2,q,temp1);
    for (int k=0;k<SPACEDIM;k++) {
      REAL src_k=src[k];
      for (int i=0;i<SPACEDIM;i++) {
        for (int j=0;j<SPACEDIM;j++) {
          temp2[i][j]+=
            higher_order_mapping_second_derivative[q][i][j][k]*src_k;
        }
      }
    }
    temp2/=higher_order_mapping_determinant[q];;
    postMultiplyByJacobianInverse(dest,q,temp2);
  } else {
    CHECK_TEST(mapping_determinant_ok);
    CHECK_TEST(mapping_jacobian_ok);
    CHECK_TEST(mapping_jacobian_inverse_ok);
    CHECK_TEST(mapping_second_derivative_ok);
    Tensor<2,SPACEDIM> determinant_derivative=
      determinantDerivative(mapping_jacobian[q]);
    Tensor<2,SPACEDIM> temp1=src_deriv;
    for (int k=0;k<SPACEDIM;k++) {
      REAL d_det_d_xi_k=0.;
      for (int i=0;i<SPACEDIM;i++) {
        for (int j=0;j<SPACEDIM;j++) {
          d_det_d_xi_k+=determinant_derivative[i][j]
            *mapping_second_derivative[q][i][j][k];
        }
      }
      d_det_d_xi_k/=mapping_determinant[q];
      for (int i=0;i<SPACEDIM;i++) {
        temp1[i][k]-=src[i]*d_det_d_xi_k;
      }
    }
    Tensor<2,SPACEDIM> temp2;
    preMultiplyByJacobian(temp2,q,temp1);
    for (int k=0;k<SPACEDIM;k++) {
      REAL src_k=src[k];
      for (int i=0;i<SPACEDIM;i++) {
        for (int j=0;j<SPACEDIM;j++) {
          temp2[i][j]+=mapping_second_derivative[q][i][j][k]*src_k;
        }
      }
    }
    temp2/=mapping_determinant[q];
    postMultiplyByJacobianInverse(dest,q,temp2);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingDataBase::hdivTransformDivergence(
REAL &dest_divergence,int q,REAL src_divergence) const {
  REAL det=HUGE_VAL;
  if (map_higher_order) {
    CHECK_TEST(higher_order_mapping_determinant_ok);
    det=higher_order_mapping_determinant[q];
  } else {
    CHECK_TEST(mapping_determinant_ok);
    det=mapping_determinant[q];
  }
  dest_divergence=src_divergence/det;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingDataBase::hcurlTransformDerivative(
Tensor<2,SPACEDIM> &dest,int q,const Tensor<1,SPACEDIM> &src,
const Tensor<2,SPACEDIM> &src_deriv) const {
  if (map_higher_order) {
    CHECK_TEST(higher_order_mapping_jacobian_ok);
    CHECK_TEST(higher_order_mapping_jacobian_inverse_ok);
    CHECK_TEST(higher_order_mapping_second_derivative_ok);

    const Tensor<2,SPACEDIM> &homjiq=
      higher_order_mapping_jacobian_inverse[q];
    Tensor<1,SPACEDIM> mapping_deriv_inv_transpose_times_src;
    for (int i=0;i<SPACEDIM;i++) {
      REAL &sum=mapping_deriv_inv_transpose_times_src[i];
      for (int j=0;j<SPACEDIM;j++) {
        sum+=homjiq[j][i]*src[j];
      }
    }
    Tensor<2,SPACEDIM> work;
    for (int k=0;k<SPACEDIM;k++) {
      Tensor<1,SPACEDIM> temp;
      for (int i=0;i<SPACEDIM;i++) temp[i]=src_deriv[i][k];
      for (int i=0;i<SPACEDIM;i++) {
        for (int j=0;j<SPACEDIM;j++) {
          temp[i]-=higher_order_mapping_second_derivative[q][j][i][k]
                  *mapping_deriv_inv_transpose_times_src[j];
        }
      }
      for (int i=0;i<SPACEDIM;i++) {
        for (int j=0;j<SPACEDIM;j++) work[i][j]+=temp[i]*homjiq[k][j];
      }
    }
    for (int i=0;i<SPACEDIM;i++) {
      for (int j=0;j<SPACEDIM;j++) {
        REAL &sum=dest[i][j];
        sum=0.;
        for (int k=0;k<SPACEDIM;k++) sum+=homjiq[k][i]*work[k][j];
      }
    }
  } else {
    CHECK_TEST(mapping_jacobian_ok);
    CHECK_TEST(mapping_jacobian_inverse_ok);
    CHECK_TEST(mapping_second_derivative_ok);

    const Tensor<2,SPACEDIM> &mjiq=mapping_jacobian_inverse[q];
    Tensor<1,SPACEDIM> mapping_deriv_inv_transpose_times_src;
    for (int i=0;i<SPACEDIM;i++) {
      REAL &sum=mapping_deriv_inv_transpose_times_src[i];
      for (int j=0;j<SPACEDIM;j++) {
        sum+=mjiq[j][i]*src[j];
      }
    }
    Tensor<2,SPACEDIM> work;
    for (int k=0;k<SPACEDIM;k++) {
      Tensor<1,SPACEDIM> temp;
      for (int i=0;i<SPACEDIM;i++) temp[i]=src_deriv[i][k];
      for (int i=0;i<SPACEDIM;i++) {
        for (int j=0;j<SPACEDIM;j++) {
          temp[i]-=mapping_second_derivative[q][j][i][k]
                  *mapping_deriv_inv_transpose_times_src[j];
        }
      }
      for (int i=0;i<SPACEDIM;i++) {
        for (int j=0;j<SPACEDIM;j++) work[i][j]+=temp[i]*mjiq[k][j];
      }
    }
    for (int i=0;i<SPACEDIM;i++) {
      for (int j=0;j<SPACEDIM;j++) {
        REAL &sum=dest[i][j];
        sum=0.;
        for (int k=0;k<SPACEDIM;k++) sum+=mjiq[k][i]*work[k][j];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingDataBase::hcurlTransform(Tensor<1,SPACEDIM> &dest,int q,
const Tensor<1,SPACEDIM> &src) const {
  dest.initialize(0.);
  if (map_higher_order) {
    CHECK_TEST(higher_order_mapping_jacobian_inverse_ok);
    for (int i=0;i<SPACEDIM;i++) {
      for (int j=0;j<SPACEDIM;j++) {
        dest[i]+=higher_order_mapping_jacobian_inverse[q][j][i]*src[j];
      }
    }
  } else {
    CHECK_TEST(mapping_jacobian_inverse_ok);
    for (int i=0;i<SPACEDIM;i++) {
      for (int j=0;j<SPACEDIM;j++) {
        dest[i]+=mapping_jacobian_inverse[q][j][i]*src[j];
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# if (SPACEDIM==1)
void MappingDataBase::hcurlTransformCurl(REAL &dest_curl,int q,
REAL src_curl,REAL src_fcn) const {
  if (map_higher_order) {
    dest_curl=(src_curl
      -higher_order_mapping_second_derivative[q][0][0][0]*src_fcn)
      /pow(higher_order_mapping_determinant[q],2);
  } else {
    dest_curl=(src_curl-mapping_second_derivative[q][0][0][0]*src_fcn)
             /pow(mapping_determinant[q],2);
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# if (SPACEDIM==2)
void MappingDataBase::hcurlTransformCurl(REAL &dest_curl,int q,
REAL src_curl) const {
  hdivTransformDivergence(dest_curl,q,src_curl);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
# if (SPACEDIM==3)
void MappingDataBase::hcurlTransformCurl(Tensor<1,3> &dest_curl,
int q,const Tensor<1,3> &src_curl) const {
  hdivTransform(dest_curl,q,src_curl);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//the canonical point was determined by the one-point quadrature
//  used in allocateSpace
Point<SPACEDIM> MappingDataBase::elementPoint() const {
  Point<SPACEDIM> element_point(D_DECL(0.,0.,0.));
  if (map_higher_order) {
    CHECK_TEST(higher_order_shape_values_ok);
    CHECK_TEST(higher_order_mapping_support_points_ok);
    int l=higher_order_mapping_support_points.getNumber();
    CHECK_SAME(l,higher_order_shape_values.size(0));
    CHECK_SAME(1,higher_order_shape_values.size(1));
    for (int i=0;i<l;i++) {
      element_point+=higher_order_mapping_support_points[i]
                    *higher_order_shape_values(i,0);
    }
  } else {
    CHECK_TEST(shape_values_ok);
    CHECK_TEST(mapping_support_points_ok);
    int l=mapping_support_points.getNumber();
    CHECK_SAME(l,shape_values.size(0));
    CHECK_SAME(1,shape_values.size(1));
    for (int i=0;i<l;i++) {
      element_point+=mapping_support_points[i]*shape_values(i,0);
    }
  }
  return element_point;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//canonical points were determined by the quadrature 
//  in allocateSpace
void MappingDataBase::elementPoints(
NumPtr<Point<SPACEDIM> > &element_points) const {
  Point<SPACEDIM> zero_point(D_DECL(0.,0.,0.));
  if (map_higher_order) {
    CHECK_TEST(higher_order_shape_values_ok);
    CHECK_TEST(higher_order_mapping_support_points_ok);
    CHECK_SAME(element_points.getNumber(),
               higher_order_shape_values.size(1));
    int l=higher_order_mapping_support_points.getNumber();
    CHECK_SAME(l,higher_order_shape_values.size(0));
    for (int q=0;q<element_points.getNumber();q++) {
      element_points[q]=zero_point;
      for (int i=0;i<l;i++) {
        element_points[q]+=higher_order_mapping_support_points[i]
          *higher_order_shape_values(i,q);
      }
    }
  } else {
    CHECK_TEST(shape_values_ok);
    CHECK_TEST(mapping_support_points_ok);
    CHECK_SAME(element_points.getNumber(),shape_values.size(1));
    int l=mapping_support_points.getNumber();
    CHECK_SAME(l,shape_values.size(0));
    for (int q=0;q<element_points.getNumber();q++) {
      element_points[q]=zero_point;
      for (int i=0;i<l;i++) {
        element_points[q]+=mapping_support_points[i]*shape_values(i,q);
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//requires a sufficiently good initial guess for canonical_point
//this changes the quadrature
void MappingDataBase::canonicalPoint(const SHAPE_CLASS *element,
REAL diameter,const Point<SPACEDIM> &element_point,
Point<SPACEDIM> &canonical_point) {
//make sure that this MappingDataBase is special for canonicalPoint
// Newton iteration to solve f(x)=p(x)-p=0
// The start value is set to be the center of the unit cell.
// The shape values and derivatives of the mapping at this point are
// previously computed.
  Quadrature<SPACEDIM> q(canonical_point);
  computeShapes(&q);
  if (requiresShapeToInitialize()) computeShapesFor(element);      
  Point<SPACEDIM> f=elementPoint()-element_point;
  REAL eps=DBL_EPSILON*diameter*2.;
  int loop=0;
  while (f.square()>eps*eps && loop++<30) {      
    Tensor<2,SPACEDIM> df;
    if (map_higher_order) {
      CHECK_TEST(higher_order_shape_derivatives_ok);
      for (int k=0;k<n_higher_order_shape_functions;k++) {
        const Tensor<1,SPACEDIM> &grad_transform=
          higher_order_shape_derivatives(k,0);
        const Point<SPACEDIM> &point=
          higher_order_mapping_support_points[k];
        for (int i=0;i<SPACEDIM;++i) {
          for (int j=0;j<SPACEDIM;j++) {
            df[i][j]+=point[i]*grad_transform[j];
          }
        }
      }
    } else {
      CHECK_TEST(shape_derivatives_ok);
      for (int k=0;k<n_shape_functions;k++) {
        const Tensor<1,SPACEDIM> &grad_transform=
          shape_derivatives(k,0);
        const Point<SPACEDIM> &point=mapping_support_points[k];
        for (int i=0;i<SPACEDIM;++i) {
          for (int j=0;j<SPACEDIM;j++) {
            df[i][j]+=point[i]*grad_transform[j];
          }
        }
      }
    }
    Tensor<2,SPACEDIM> df_1=invert(df);
    Tensor<1,SPACEDIM> d;
    product(d,df_1,Tensor<1,SPACEDIM>(f)); 
    canonical_point-=d.getValues();
    q.setPoint(0,canonical_point);
    computeShapes(&q);
    if (requiresShapeToInitialize()) computeShapesFor(element);      
    f=elementPoint()-element_point;
  }
  quadrature=0; // q goes out of scope
  shape_values_ok=false;
  higher_order_shape_values_ok=false;
  shape_derivatives_ok=false;
  higher_order_shape_derivatives_ok=false;
  mapping_jacobian_ok=false;
  higher_order_mapping_jacobian_ok=false;
  mapping_jacobian_inverse_ok=false;
  higher_order_mapping_jacobian_inverse_ok=false;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
//offset works on the quadrature point index
void MappingDataBase::computeFEFaceValues(const SHAPE_CLASS *element,
const Tessellation &tess,const LEVEL_CLASS *level,int face_no,
bool is_subface,int npts,int offset,
const NumPtr<REAL> &weights,NumPtr<Point<SPACEDIM> > &quadrature_points,
NumPtr<REAL> &JxW_values,NumPtr<Tensor<1,SPACEDIM> > &boundary_forms,
NumPtr<Point<SPACEDIM> > &normal_vectors,bool &quadrature_points_ok,
bool &jxw_values_ok,bool &boundary_forms_ok,bool &normal_vectors_ok) {
  if (requiresShapeToInitialize()) computeShapesFor(element);      
  computeJacobian(element,tess,level,npts,offset,quadrature_points,
    quadrature_points_ok);
  if (update_each & UPDATE_BOUNDARY_FORMS) {
    CHECK_SAME(boundary_forms.getNumber(),npts);
    if (update_each & UPDATE_NORMAL_VECTORS) {
      CHECK_SAME(normal_vectors.getNumber(),npts);
    }
    if (update_each & UPDATE_JxW_VALUES) {
      CHECK_SAME(JxW_values.getNumber(),npts);
    }
// aux & unit_tangentials allocated in MappingDataBase::computeFaceTangents
//  reference_normal_f = unit outer normal_f * face measure 
//                                           / canonical face measure
    for (int q=0;q<npts;q++) {
      adjointProduct(boundary_forms[q],
        (map_higher_order ? higher_order_mapping_jacobian[q]
                        : mapping_jacobian[q]),reference_normal[face_no]);
    }
    boundary_forms_ok=true;
//  element tangent = aux(xi_q) = ( d x / d xi )(xi_q) * t( xi_q )
//  aux_ok=true;
    if (update_each & (UPDATE_NORMAL_VECTORS | UPDATE_JxW_VALUES)) {
      for (int i=0;i<boundary_forms.getNumber();++i) {
        REAL f;
        product(f,boundary_forms[i], boundary_forms[i]);
        f=sqrt(f);
        if (update_each & UPDATE_JxW_VALUES) {
          JxW_values[i]=f*weights[i];
          if (is_subface) {
            JxW_values[i]/=childrenPerCodimension1Shape();
          }
          jxw_values_ok=true;
        }
        if (update_each & UPDATE_NORMAL_VECTORS) {
          normal_vectors[i]=boundary_forms[i].getValues();
          normal_vectors[i]/=f;
          normal_vectors_ok=true;
        }
      }
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//offset works on the quadrature point index
void MappingDataBase::computeFEEdgeValues(const SHAPE_CLASS *element,
const Tessellation &tess,const LEVEL_CLASS *level,int npts,int offset,
NumPtr<Point<SPACEDIM> > &quadrature_points,bool &quadrature_points_ok) {
  if (requiresShapeToInitialize()) computeShapesFor(element);      
  computeJacobian(element,tess,level,npts,offset,quadrature_points,
    quadrature_points_ok);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingDataBase::printOn(ostream &os) const {
  os << "MappingDataBase:"
     << "\n\tupdate_once = " << update_once
     << "\n\tupdate_each = " << update_each
     << endl;
  os << "\tmap_higher_order = " << PRINT_BOOLEAN(map_higher_order)
     << "\n\tshape_values_ok = " << PRINT_BOOLEAN(shape_values_ok)
     << "\n\thigher_order_shape_values_ok = " 
     << PRINT_BOOLEAN(higher_order_shape_values_ok)
     << "\n\tshape_derivatives_ok = " 
     << PRINT_BOOLEAN(shape_derivatives_ok)
     << "\n\thigher_order_shape_derivatives_ok = " 
     << PRINT_BOOLEAN(higher_order_shape_derivatives_ok)
     << "\n\tmapping_support_points_ok = " 
     << PRINT_BOOLEAN(mapping_support_points_ok)
     << "\n\thigher_order_mapping_support_points_ok = " 
     << PRINT_BOOLEAN(higher_order_mapping_support_points_ok)
     << "\n\tmapping_jacobian_ok = " << PRINT_BOOLEAN(mapping_jacobian_ok)
     << "\n\thigher_order_mapping_jacobian_ok = " 
     << PRINT_BOOLEAN(higher_order_mapping_jacobian_ok)
     << "\n\tmapping_jacobian_inverse_ok = " 
     << PRINT_BOOLEAN(mapping_jacobian_inverse_ok)
     << "\n\thigher_order_mapping_jacobian_inverse_ok = " 
     << PRINT_BOOLEAN(higher_order_mapping_jacobian_inverse_ok)
     << endl;
  os << "\n\tn_shape_functions = " << n_shape_functions
     << "\n\tn_higher_order_shape_functions = " 
     << n_higher_order_shape_functions << endl;
  os << "\tquadrature:" << endl;
  quadrature->printOn(cout);
/*
  if (shape_values_ok) {
    for (int i=0;i<shape_values.size(0);i++) {
      os << "\tshape_values[" << i << "] =";
      for (int j=0;j<shape_values.size(1);j++) {
         os << " " << shape_values(i,j);
      }
      os << endl;
    }
  }
*/
/*
  if (higher_order_shape_values_ok) {
    for (int i=0;i<higher_order_shape_values.size(0);i++) {
      os << "\thigher_order_shape_values[" << i << "] =";
      for (int j=0;j<higher_order_shape_values.size(1);j++) {
         os << " " << higher_order_shape_values(i,j);
      }
      os << endl;
    }
  }
*/
/*
  if (shape_derivatives_ok) {
    for (int i=0;i<shape_derivatives.size(0);i++) {
      os << "\tshape_derivatives[" << i << "] =";
      for (int j=0;j<shape_derivatives.size(1);j++) {
         os << " " << shape_derivatives(i,j);
      }
      os << endl;
    }
  }
  if (higher_order_shape_derivatives_ok) {
    for (int i=0;i<higher_order_shape_derivatives.size(0);i++) {
      os << "\thigher_order_shape_derivatives[" << i << "] =";
      for (int j=0;j<higher_order_shape_derivatives.size(1);j++) {
         os << " " << higher_order_shape_derivatives(i,j);
      }
      os << endl;
    }
  }
*/
  if (mapping_support_points_ok) {
    for (int i=0;i<mapping_support_points.getNumber();i++) {
      os << "\tmapping_support_points[" << i << "] = " 
         << mapping_support_points[i] << endl;
    }
  }
  if (higher_order_mapping_support_points_ok) {
    for (int i=0;i<higher_order_mapping_support_points.getNumber();i++) {
      os << "\thigher_order_mapping_support_points[" << i << "] = " 
         << higher_order_mapping_support_points[i] << endl;
    }
  }
  if (mapping_jacobian_ok) {
    for (int i=0;i<mapping_jacobian.getNumber();i++) {
      os << "\tmapping_jacobian[" << i << "] = " << mapping_jacobian[i] 
         << endl;
    }
  }
  if (higher_order_mapping_jacobian_ok) {
    for (int i=0;i<higher_order_mapping_jacobian.getNumber();i++) {
      os << "\thigher_order_mapping_jacobian[" << i << "] = " 
         << higher_order_mapping_jacobian[i] << endl;
    }
  }
//
  if (mapping_jacobian_inverse_ok) {
    for (int i=0;i<mapping_jacobian_inverse.getNumber();i++) {
      os << "\tmapping_jacobian_inverse[" << i << "] = " 
         << mapping_jacobian_inverse[i] << endl;
    }
  }
  if (higher_order_mapping_jacobian_inverse_ok) {
    for (int i=0;i<higher_order_mapping_jacobian_inverse.getNumber();i++) {
      os << "\thigher_order_mapping_jacobian_inverse[" << i << "] = " 
         << higher_order_mapping_jacobian_inverse[i] << endl;
    }
  }
//
/*
#if (SPACEDIM>1)
  for (int i=0;i<reference_normal.getNumber();i++) {
    os << "\treference_normal[" << i << "] = " << reference_normal[i]
       << endl;
  }
#endif
*/
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Point<SPACEDIM> Mapping::elementPoint(const SHAPE_CLASS *e,
const Tessellation &tess,const LEVEL_CLASS *level,
const Point<SPACEDIM> &canonical_point) const {
  const Quadrature<SPACEDIM> point_quadrature(canonical_point);
  MappingDataBase *mdata=getMappingData(e->numberVertices(),
    UPDATE_TRANSFORMATION_VALUES,&point_quadrature);
  CHECK_POINTER(mdata);
  mdata->computeMappingSupportPoints(e,tess,level);
  if (mdata->requiresShapeToInitialize()) mdata->computeShapesFor(e);
  const Point<SPACEDIM> element_point=mdata->elementPoint();
  delete mdata;
  return element_point;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Mapping::elementPoints(const SHAPE_CLASS *e,
const Tessellation &tess,const LEVEL_CLASS *level,
const NumPtr<Point<SPACEDIM> > &canonical_points,
NumPtr<Point<SPACEDIM> > &element_points) const {
  const Quadrature<SPACEDIM> quadrature(canonical_points);
  MappingDataBase *mdata=getMappingData(e->numberVertices(),
    UPDATE_TRANSFORMATION_VALUES,&quadrature);
  CHECK_POINTER(mdata);
  mdata->computeMappingSupportPoints(e,tess,level);
  if (mdata->requiresShapeToInitialize()) mdata->computeShapesFor(e);
  mdata->elementPoints(element_points);
  delete mdata;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//this can converge to a point outside the canonical Shape
//  if the initial guess is not good enough
void Mapping::canonicalPoint(const SHAPE_CLASS *e,
const Tessellation &tess,const MappingDataBase *previous_md,
const Point<SPACEDIM> &element_point,Point<SPACEDIM> &canonical_point)
const {
  const Quadrature<SPACEDIM> point_quadrature(canonical_point);
  MappingDataBase *mdata=getMappingData(e->numberVertices(),
    UPDATE_TRANSFORMATION_VALUES | UPDATE_TRANSFORMATION_GRADIENTS,
    &point_quadrature);
  CHECK_POINTER(mdata);
  mdata->copyMappingSupportPointsFrom(previous_md);
  if (mdata->requiresShapeToInitialize()) mdata->computeShapesFor(e);
  mdata->canonicalPoint(e,e->diameter(tess.getVertices()),element_point,
    canonical_point);
  delete mdata;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void Mapping::printOn(ostream &os) const {
  os << "Mapping: " << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
MappingL::InternalData::InternalData(int nv,int p,
UPDATE_FLAGS update_flags,const Quadrature<SPACEDIM> *q) : 
MappingDataBase(p) {
#if (SPACEDIM==1)
  shape_function=OPERATOR_NEW C0LagrangeIntervalPolynomials(1);
  if (degree>1) {
    higher_order_shape_function=
      OPERATOR_NEW C0LagrangeIntervalPolynomials(degree);
  }
#elif (SPACEDIM==2)
  if (nv==Quadrilateral::verticesPerShape()) {
    shape_function=OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(1);
    if (degree>1) {
      higher_order_shape_function=
        OPERATOR_NEW C0LagrangeQuadrilateralPolynomials(degree);
    }
  } else {
    CHECK_SAME(nv,Triangle::verticesPerShape());
    shape_function=OPERATOR_NEW C0LagrangeTrianglePolynomials(1);
    if (degree>1) {
      higher_order_shape_function=
        OPERATOR_NEW C0LagrangeTrianglePolynomials(degree);
    }
  }
#else
  if (nv==Hexahedron::verticesPerShape()) {
    shape_function=OPERATOR_NEW C0LagrangeHexahedronPolynomials(1);
    if (degree>1) {
      higher_order_shape_function=
        OPERATOR_NEW C0LagrangeHexahedronPolynomials(degree);
    }
  } else if (nv==Tetrahedron::verticesPerShape()) {
    shape_function=OPERATOR_NEW C0LagrangeTetrahedronPolynomials(1);
    if (degree>1) {
      higher_order_shape_function=
        OPERATOR_NEW C0LagrangeTetrahedronPolynomials(degree);
    }
  } else {
    CHECK_SAME(nv,Prism::verticesPerShape());
    shape_function=OPERATOR_NEW C0LagrangePrismPolynomials(1);
    if (degree>1) {
      higher_order_shape_function=
        OPERATOR_NEW C0LagrangePrismPolynomials(degree);
    }
  }
#endif
  allocateSpace(update_flags,q);
  computeShapes(q);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingL::InternalData::computeMappingSupportPoints(
const SHAPE_CLASS *element,const Tessellation &tess,
const LEVEL_CLASS *level) {
  map_higher_order=false;
  if (higher_order_mapping_support_points.getNumber()>0) {
    if (use_full_degree_mapping_on_all_cells) map_higher_order=true;
    else {
      for (int f=0;f<element->numberCodimension1Shapes();f++) {
        int fei=element->codimension1ShapeIndex(f);
        if (level->CODIMENSION1_LEVEL_CLASS::getBoundaryCondition(fei)!=
        0) {
          map_higher_order=true;
          break;
        }
      }
    }
  }
  int nv=element->numberVertices();
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  if (!map_higher_order) {
    if (mapping_support_points.getNumber()>0) {
      CHECK_SAME(nv,mapping_support_points.getNumber());
      for (int v=0;v<nv;v++) {
        mapping_support_points[v]=vertices[element->vertexIndex(v)];
      }
      mapping_support_points_ok=true;
    }
  } else {
#if (SPACEDIM==1)
    OBSOLETE("not programmed");
#else
    tess.mappingSupportPointsOnBoundaryOf(element,degree,level,
      higher_order_mapping_support_points);
#endif
    higher_order_mapping_support_points_ok=true;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MappingL::InternalData::printOn(ostream &os) const {
  os << "MappingL::InternalData:" << endl;
  MappingDataBase::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

template class ArrayBase<2,REAL>;
template class Array<2,REAL>;
template class ArrayBase<2,Tensor<1,SPACEDIM> >;
template class Array<2,Tensor<1,SPACEDIM> >;
template class ArrayBase<2,Tensor<2,SPACEDIM> >;
template class Array<2,Tensor<2,SPACEDIM> >;
