#include <iostream>
#include <stdlib.h>
#include "Boundary.H"
#include "DDecl.H"
#include "GenerateTess.H"
#include "GUIEnumInputParameter.H"
#include "GUIVirtualInput.H"
#include "InputParameter.H"
#include "Mapping.H"
#include "Quadrature.H"
#include "SetTraps.H"
#include "Tessellation.H"
#include "Tracer.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKGUI.H"
#include "GTKGUIVirtualInput.H"
#include "GTKWindow.H"
#else
#include "GUI.H"
#include "GUIVirtualInput.H"
#endif

GUI_INPUT_PARAMETER_LIST_TYPE *param_list=0;
BOOLEAN skip_gui=FALSE;
double winsize=0.5;
char* display_name=0;

#if (SPACEDIM==1)
barf
#elif (SPACEDIM==2)
enum SHAPE{TRIANGLE_SHAPE,QUADRILATERAL_SHAPE};
static const char *shape_name[2]={"triangle","quadrilateral"};
SHAPE shape=QUADRILATERAL_SHAPE;
#else
enum SHAPE{TETRAHEDRON_SHAPE,PRISM_SHAPE,HEXAHEDRON_SHAPE};
static const char *shape_name[3]={"tetrahedron","prism","hexahedron"};
SHAPE shape=HEXAHEDRON_SHAPE;
SHAPE second_shape=HEXAHEDRON_SHAPE;
int isecond_shape=second_shape;
#endif
int ishape=shape;

/*
enum MAPPING_TYPE{LAGRANGE_MAPPING,HIERARCHICAL_MAPPING};
static const char *mapping_name[2]=
  {"lagrange mapping","hierarchical mapping"};
MAPPING_TYPE mapping_type=LAGRANGE_MAPPING;
int imapping_type=mapping_type;
*/
int degree=1;

int number_refinements=1;
Point<SPACEDIM> center(D_DECL(0.,0.,0.));
double radius=1.;

//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void checkMainInput() {
//TRACER_CALL(tr,"checkMainInput");
#if (SPACEDIM>1)
  shape=static_cast<SHAPE>(ishape);
//mapping_type=static_cast<MAPPING_TYPE>(imapping_type);
#endif
#if (SPACEDIM==3)
  second_shape=(isecond_shape==TETRAHEDRON_SHAPE ? TETRAHEDRON_SHAPE :
    HEXAHEDRON_SHAPE);
#endif
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void runMain(BOOLEAN /*called_before*/) {
//TRACER_CALL(tr,"runMain");
  Tessellation tess;
  bool set_face_ids=true;
  BoundaryData *sub_cell_data=0;
  UPDATE_FLAGS update_flags(UPDATE_VALUES | UPDATE_BOUNDARY_FORMS
    | UPDATE_Q_POINTS | UPDATE_JxW_VALUES
    | UPDATE_NORMAL_VECTORS | UPDATE_TRANSFORMATION_VALUES);
  Mapping *mapping=0;
//switch (mapping_type) {
//  case HIERARCHICAL_MAPPING:
//    mapping=OPERATOR_NEW MappingH(degree);
//    break;
//  default:
      mapping=OPERATOR_NEW MappingL(degree);
//}
  map<int,MappingDataBase*> mapping_data_map;
  NumPtr<Boundary*> boundary;
  NumPtr<BoundaryCondition*> boundary_condition;
#if (SPACEDIM==2)
  NewtonCotesQuadrature<1> ncq(mapping->getDegree()+1);
  {
    int nfp=3*ncq.numberPoints();
    NumPtr<Point<SPACEDIM> > q_points(nfp);
    NumPtr<REAL> weights(nfp);
    NumPtr<Point<SPACEDIM> > q_points_face=q_points;
    NumPtr<REAL> weights_face=weights;
    for (int f=0;f<3;f++) {
      Triangle::staticProjectToFace(f,ncq.points(),q_points_face);
      weights_face.copyFrom(ncq.weights());
      q_points_face+=ncq.numberPoints();
      weights_face+=ncq.numberPoints();
    }
    Quadrature<SPACEDIM> qf(q_points,weights);
    mapping_data_map[3]=mapping->getMappingData(3,update_flags,&qf);
  }
  {
    int nfp=4*ncq.numberPoints();
    NumPtr<Point<SPACEDIM> > q_points(nfp);
    NumPtr<REAL> weights(nfp);
    NumPtr<Point<SPACEDIM> > q_points_face=q_points;
    NumPtr<REAL> weights_face=weights;
    for (int f=0;f<4;f++) {
      Quadrilateral::staticProjectToFace(f,ncq.points(),q_points_face);
      weights_face.copyFrom(ncq.weights());
      q_points_face+=ncq.numberPoints();
      weights_face+=ncq.numberPoints();
    }
    Quadrature<SPACEDIM> qf(q_points,weights);
    mapping_data_map[4]=mapping->getMappingData(4,update_flags,&qf);
  }
  switch (shape) {
    case TRIANGLE_SHAPE:
      sub_cell_data=generateBallTriangles(tess,center,radius,set_face_ids,
        boundary,boundary_condition);
      break;
    default:
      sub_cell_data=generateBall(tess,center,radius,set_face_ids,
        boundary,boundary_condition);
  }
//if (number_refinements>0) {
//  BallBoundary *bb=OPERATOR_NEW BallBoundary(center,radius);
//  tess.setBoundary(0,bb);
//}
#else
  map<int,Quadrature<SPACEDIM>*> quadrature_map;
  TriangleFullNewtonCotesQuadrature tncq(mapping->getDegree()+1);
  NewtonCotesQuadrature<2> ncq(mapping->getDegree()+1);
  {
    int nfp=4*tncq.numberPoints();
    NumPtr<Point<SPACEDIM> > q_points(nfp);
    NumPtr<REAL> weights(nfp);
    NumPtr<Point<SPACEDIM> > q_points_face=q_points;
    NumPtr<REAL> weights_face=weights;
    for (int f=0;f<4;f++) {
      Tetrahedron::staticProjectToFace(f,tncq.points(),q_points_face);
      weights_face.copyFrom(tncq.weights());
      q_points_face+=tncq.numberPoints();
      weights_face+=tncq.numberPoints();
    }   
    Quadrature<SPACEDIM> *qf=
      OPERATOR_NEW Quadrature<SPACEDIM>(q_points,weights);
    quadrature_map[4]=qf;
    mapping_data_map[4]=mapping->getMappingData(4,update_flags,qf);
  }
  {
    int nfp=2*tncq.numberPoints()+3*ncq.numberPoints();
    NumPtr<Point<SPACEDIM> > q_points(nfp);
    NumPtr<REAL> weights(nfp);
    NumPtr<Point<SPACEDIM> > q_points_face=q_points;
    NumPtr<REAL> weights_face=weights;
    for (int f=0;f<2;f++) {
      Prism::staticProjectToFace(f,tncq.points(),q_points_face);
      weights_face.copyFrom(tncq.weights());
      q_points_face+=tncq.numberPoints();
      weights_face+=tncq.numberPoints();
    }   
    for (int f=2;f<5;f++) {
      Prism::staticProjectToFace(f,ncq.points(),q_points_face);
      weights_face.copyFrom(ncq.weights());
      q_points_face+=ncq.numberPoints();
      weights_face+=ncq.numberPoints();
    }   
    Quadrature<SPACEDIM> *qf=
      OPERATOR_NEW Quadrature<SPACEDIM>(q_points,weights);
    quadrature_map[6]=qf;
    mapping_data_map[6]=mapping->getMappingData(6,update_flags,qf);
  }
  {
    int nfp=6*ncq.numberPoints();
    NumPtr<Point<SPACEDIM> > q_points(nfp);
    NumPtr<REAL> weights(nfp);
    NumPtr<Point<SPACEDIM> > q_points_face=q_points;
    NumPtr<REAL> weights_face=weights;
    for (int f=0;f<6;f++) {
      Hexahedron::staticProjectToFace(f,ncq.points(),q_points_face);
      weights_face.copyFrom(ncq.weights());
      q_points_face+=ncq.numberPoints();
      weights_face+=ncq.numberPoints();
    }   
    Quadrature<SPACEDIM> *qf=
      OPERATOR_NEW Quadrature<SPACEDIM>(q_points,weights);
    quadrature_map[8]=qf;
    mapping_data_map[8]=mapping->getMappingData(8,update_flags,qf);
  }
  switch (shape) {
    case TETRAHEDRON_SHAPE:
      sub_cell_data=generateBallTetrahedra(tess,center,radius,
        set_face_ids,boundary,boundary_condition);
      break;
    case PRISM_SHAPE: {
      switch (second_shape) {
        case TETRAHEDRON_SHAPE:
          sub_cell_data=
            generateBallPrismsWithTetrahedra(tess,center,radius,
              set_face_ids,boundary,boundary_condition);
          break;
        default:
          sub_cell_data=
            generateBallPrismsWithHexahedra(tess,center,radius,
              set_face_ids,boundary,boundary_condition);
      }
      break;
    }
    default:
      sub_cell_data=generateBall(tess,center,radius,set_face_ids,
        boundary,boundary_condition);
  }
//if (number_refinements>0) {
//  BallBoundary *bb=OPERATOR_NEW BallBoundary(center,radius);
//  tess.setBoundary(0,bb);
//}
#endif
  Palette pal;
#if (SPACEDIM==2)
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]; rlo[0]=center[0]-radius; rlo[1]=center[1]-radius;
  REAL rhi[2]; rhi[0]=center[0]+radius; rhi[1]=center[1]+radius;
  AreaGraphTool gt("MappingGUI","x","y",rlo,rhi,&cmap,0,0.5);
#else
  Vector3 low(-radius*1.25,-radius*1.25,-radius*1.25);
  Vector3 high(radius*1.25,radius*1.25,radius*1.25);
  VolGraphTool gt("debugTess",low,high,pal,TRUE,0.5);
#endif
  for (int r=0;r<=number_refinements;r++) {
    gt.newPage();
    const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
#if (SPACEDIM==2)
//  draw ball boundary
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    gt.movePen(center[0]+radius,center[1]);
    REAL angle=0.;
    REAL dangle=2.*M_PI/256.;
    for (int i=0;i<=256;i++,angle+=dangle) {
      gt.drawLine(center[0]+radius*cos(angle),
                  center[1]+radius*sin(angle));
    }
    for (bool doit=tess.firstActiveShape();doit;doit=tess.nextActiveShape())
    {
      const SHAPE_CLASS *e=tess.presentShape();
      const LEVEL_CLASS *level=tess.presentLevel();
      map<int,MappingDataBase*>::iterator mdm_iterator=
        mapping_data_map.find(e->numberVertices());
      CHECK_TEST(mdm_iterator!=mapping_data_map.end());
      MappingDataBase *md=mdm_iterator->second;
//    draw normals to elements on domain boundary
      gt.setfgColor("red");
      for (int f=0;f<e->numberCodimension1Shapes();f++) {
        if (level->neighborLevel(e,f)==0) {
          NumPtr<Point<SPACEDIM> > quadrature_points(ncq.numberPoints());
          NumPtr<REAL> JxW_values(ncq.numberPoints());
          NumPtr<Tensor<1,SPACEDIM> > boundary_forms(ncq.numberPoints());
          NumPtr<Point<SPACEDIM> > normal_vectors(ncq.numberPoints());
          bool is_subface=false;
          bool quadrature_points_ok=false;
          bool jxw_values_ok=false;
          bool boundary_forms_ok=false;
          bool normal_vectors_ok=false;
          md->computeFEFaceValues(e,tess,level,f,is_subface,
            ncq.numberPoints(),f*ncq.numberPoints(),ncq.weights(),
            quadrature_points,JxW_values,boundary_forms,normal_vectors,
            quadrature_points_ok,jxw_values_ok,boundary_forms_ok,
            normal_vectors_ok);
          CHECK_TEST(quadrature_points_ok);
          gt.setfgColor("red");
          for (int q=0;q<quadrature_points.getNumber();q++) {
            Point<SPACEDIM> hq=//quadrature_points[q]+
                               normal_vectors[q]*.1;
            COMPLEX base(quadrature_points[q][0],quadrature_points[q][1]);
            COMPLEX head(hq[0],hq[1]);
            COMPLEX arrow(-sqrt(.05),-sqrt(.05));
            gt.drawVector(base,head,arrow);
          }
        }
      }
//    draw elements
      int npts=32;
      gt.setfgColor("green");
      for (int f=0;f<e->numberCodimension1Shapes();f++) {
        if (level->neighborLevel(e,f)==0) {
          Point<SPACEDIM> cv,dcv;
          if (e->numberVertices()==4) {
            cv=
              Quadrilateral::canonicalVertex(e->elementVertexFromFace(f,0));
            dcv=(Quadrilateral::canonicalVertex(
              e->elementVertexFromFace(f,1))-cv)/static_cast<REAL>(npts);
          } else {
            cv=Triangle::canonicalVertex(e->elementVertexFromFace(f,0));
            dcv=(Triangle::canonicalVertex(e->elementVertexFromFace(f,1))
                -cv)/static_cast<REAL>(npts);
          }
          Point<SPACEDIM> ep=
            mapping->elementPoint(e,tess,tess.presentLevel(),cv);
          gt.movePen(ep[0],ep[1]);
          cv+=dcv;
          for (int i=0;i<npts;i++,cv+=dcv) {
            ep=mapping->elementPoint(e,tess,tess.presentLevel(),cv);
            gt.drawLine(ep[0],ep[1]);
          }
        } else {
          int fi=e->codimension1ShapeIndex(f);
          const CODIMENSION1_SHAPE_CLASS *face=
            level->CODIMENSION1_LEVEL_CLASS::getShape(fi);
          const Point<SPACEDIM> &v0=vertices[face->vertexIndex(0)];
          gt.movePen(v0[0],v0[1]);
          const Point<SPACEDIM> &v1=vertices[face->vertexIndex(1)];
          gt.drawLine(v1[0],v1[1]);
        }
      }
    }
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
#else
//  int npts[2]={2,16};
    int npts[2]={2,degree+1};
    gt.expose();
    for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
      VolGraphTool::DrawDelimiter dd(do_best,&gt);
      for (bool doit=tess.firstActiveShape();doit;
      doit=tess.nextActiveShape()) {
        const SHAPE_CLASS *e=tess.presentShape();
        const LEVEL_CLASS *level=tess.presentLevel();
        for (int f=0;f<e->numberCodimension1Shapes();f++) {
          if (level->neighborLevel(e,f)==0) {
            int face_index=e->codimension1ShapeIndex(f);
            NumPtr<Point<SPACEDIM> > cv(e->numberVerticesOnFace(f));
            switch (e->numberVertices()) {
              case 4: {
                for (int v=0;v<cv.getNumber();v++) {
                  cv[v]=Tetrahedron::canonicalVertex(
                    e->elementVertexFromFace(f,v));
                }
                break;
              }
              case 6: {
                for (int v=0;v<cv.getNumber();v++) {
                  cv[v]=Prism::canonicalVertex(
                    e->elementVertexFromFace(f,v));
                }
                break;
              }
              case 8: {
                for (int v=0;v<cv.getNumber();v++) {
                  cv[v]=Hexahedron::canonicalVertex(
                    e->elementVertexFromFace(f,v));
                }
                break;
              }
              default:
                OBSOLETE("bad numberVertices");
            }
            Point<SPACEDIM> dc1=
              (cv[1]-cv[0])/static_cast<REAL>(npts[do_best]);
            Point<SPACEDIM> dc2=
              (cv[2]-cv[0])/static_cast<REAL>(npts[do_best]);
            Point<SPACEDIM> cj=cv[0];
            if (cv.getNumber()==4) {
              for (int j=0;j<npts[do_best];j++,cj+=dc2) {
                Point<SPACEDIM> ci=cj;
                for (int i=0;i<npts[do_best];i++,ci+=dc1) {
                  Point<SPACEDIM> ep0=
                    mapping->elementPoint(e,tess,tess.presentLevel(),ci);
                  Point<SPACEDIM> ep1=
                   mapping->elementPoint(e,tess,tess.presentLevel(),ci+dc1);
                  Point<SPACEDIM> ep2=
                   mapping->elementPoint(e,tess,tess.presentLevel(),ci+dc2);
                  Point<SPACEDIM> ep3=mapping->elementPoint(e,tess,
                    tess.presentLevel(),ci+dc1+dc2);
                  Vector3 v0(ep0[0],ep0[1],ep0[2]);
                  Vector3 v1(ep1[0],ep1[1],ep1[2]);
                  Vector3 v2(ep2[0],ep2[1],ep2[2]);
                  Vector3 v3(ep3[0],ep3[1],ep3[2]);
                  Vector3 vave=(v0+v1+v2+v3)*0.25;
                  if (dot(vave,cross(v0-vave,v1-vave))>0.) {
                    gt.colorTriangle(0.5,v0,v1,vave);
                  } else gt.colorTriangle(0.5,vave,v1,v0);
                  if (dot(vave,cross(v1-vave,v3-vave))>0.) {
                    gt.colorTriangle(0.5,v1,v3,vave);
                  } else gt.colorTriangle(0.5,vave,v3,v1);
                  if (dot(vave,cross(v3-vave,v2-vave))>0.) {
                    gt.colorTriangle(0.5,v3,v2,vave);
                  } else gt.colorTriangle(0.5,vave,v2,v3);
                  if (dot(vave,cross(v2-vave,v0-vave))>0.) {
                    gt.colorTriangle(0.5,v2,v0,vave);
                  } else gt.colorTriangle(0.5,vave,v0,v2);
                }
              }
            } else {
              for (int j=0;j<npts[do_best];j++,cj+=dc2) {
                Point<SPACEDIM> ci=cj;
                for (int i=0;i+j<npts[do_best];i++,ci+=dc1) {
                  Point<SPACEDIM> ep0=
                    mapping->elementPoint(e,tess,tess.presentLevel(),ci);
                  Point<SPACEDIM> ep1=
                   mapping->elementPoint(e,tess,tess.presentLevel(),ci+dc1);
                  Point<SPACEDIM> ep2=
                   mapping->elementPoint(e,tess,tess.presentLevel(),ci+dc2);
                  Vector3 v0(ep0[0],ep0[1],ep0[2]);
                  Vector3 v1(ep1[0],ep1[1],ep1[2]);
                  Vector3 v2(ep2[0],ep2[1],ep2[2]);
                  REAL d=dot(v0+(v1+v2),cross(v2-v0,v1-v0));
                  if (d<0.) gt.colorTriangle(0.5,v0,v1,v2);
                  else gt.colorTriangle(0.5,v2,v1,v0);
                  if (i+j+1<npts[do_best]) {
                    Point<SPACEDIM> ep3=mapping->elementPoint(e,tess,
                      tess.presentLevel(),ci+dc1+dc2);
                    Vector3 v3(ep3[0],ep3[1],ep3[2]);
                    if (d<0.) gt.colorTriangle(0.5,v3,v2,v1);
                    else gt.colorTriangle(0.5,v1,v2,v3);
                  }
                }
              }
            }
          }
        }
      }
      if (do_best==TRUE) break;
    }
    gt.eventLoop();
#endif
    if (r<number_refinements) {
      tess.refineGlobal(1);
      tess.executeCoarseningAndRefinement();
    }
  }
#if (SPACEDIM==3)
  for (map<int,Quadrature<SPACEDIM>*>::iterator
  q_iterator=quadrature_map.begin();
  q_iterator!=quadrature_map.end();q_iterator++) {
    delete q_iterator->second; q_iterator->second=0;
  }
#endif
  for (map<int,MappingDataBase*>::iterator
  mdm_iterator=mapping_data_map.begin();
  mdm_iterator!=mapping_data_map.end();mdm_iterator++) {
    delete mdm_iterator->second; mdm_iterator->second=0;
  }
  delete sub_cell_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
  delete mapping;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void cleanup() {
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void shutdown() {
  while (param_list->notEmpty()) delete param_list->delAfter(0);
  delete param_list; param_list=0;
}
//:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int main(int argc,char* argv[]) {
#ifdef DEBUG
//setTraps();
#endif
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  {
    param_list=OPERATOR_NEW GUI_INPUT_PARAMETER_LIST_TYPE("Main List");
    { const char *group="Mapping Parameters";
//    param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
//      imapping_type,"mapping",mapping_name[0],mapping_name[1]));
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(
        degree,"degree",1,INT_MAX,group));
    }
    { const char *group="Grid Generation Parameters";
#if (SPACEDIM==2)
      param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
        ishape,"shape",shape_name[0],shape_name[1]));
#else
      param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
        ishape,"shape",shape_name[0],shape_name[1],shape_name[2]));
      param_list->append(OPERATOR_NEW GUIEnumInputParameter(group,
        isecond_shape,"second shape",shape_name[0],shape_name[2]));
#endif
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(
        number_refinements,"number refinements",0,INT_MAX,group));
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        radius,"radius",0.,DBL_MAX,group));
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        center[0],"center[0]",-DBL_MAX,DBL_MAX,group));
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        center[1],"center[1]",-DBL_MAX,DBL_MAX,group));
#if (SPACEDIM==3)
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(
        center[2],"center[2]",-DBL_MAX,DBL_MAX,group));
#endif
    }
    { const char *group="Graphics Parameters";
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(winsize,
        "winsize",0.,1.,group) );
    }

//  read input parameters 
    if (argc>1) {
      ifstream in_file;
      in_file.open(argv[1],ios::in);
      if (in_file) {
        istream &infile(in_file);
        infile.clear(ios::goodbit);
        infile.seekg(0,ios::beg);
        char name[LENGTH_NAME], comment[LENGTH_COMMENT];
        BOOLEAN found_main=FALSE;
        while ( in_file >> setw(LENGTH_NAME) >> name ) {
          if ( strcmp(name,"Main") == 0 ) {
            in_file.getline( comment, LENGTH_COMMENT);
            found_main=TRUE;
            while ( in_file >> setw(LENGTH_NAME) >> name ) {
              if ( strcmp(name,"end") == 0 ) break;
              else if (strcmp(name,"skip_gui") == 0) {
                int iskip_gui; // type bool not read correctly
                in_file >> iskip_gui;
                skip_gui= iskip_gui!=0;
              } else param_list->formattedRead(in_file,name);
              in_file.getline( comment, LENGTH_COMMENT);
            }
          } else in_file.getline(comment,LENGTH_COMMENT);
        }
        if ( !found_main ) skip_gui=FALSE;
      }
      if (skip_gui) checkMainInput();
    } else skip_gui=FALSE;

    if (skip_gui) {
      runMain(FALSE);
      cleanup();
      shutdown();
    } else {
#ifdef USE_GTK
      GTKGUI gui(argv[0],display_name,param_list,&runMain,
      &checkMainInput,&cleanup,&shutdown,TRUE);
#else
      GUI gui(argv[0],display_name,param_list,&runMain,&checkMainInput,
      &cleanup,&shutdown,TRUE);
#endif
      gui.createFileMenu();
      gui.createViewMenu();
      gui.createHelpMenu();
      gui.createMainWindow(argc,argv);
      gui.eventLoop();
    }
  } // scope of MemoryDebugger
  return EXIT_SUCCESS;
}

template class InputParameter<int>;
template class InputParameter<double>;
template class InputParameter<bool>;
