#ifndef NONLINEAR_DOF_HANDLER_H
#define NONLINEAR_DOF_HANDLER_H "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "ConstraintMatrix.H"
#include "DofHandler.H"

class NonlinearDofHandler : public DofHandler {
  private:
    const Mapping &mapping;
    ConstraintMatrix &hanging_node_constraints;

    map<int,Quadrature<SPACEDIM> > qmap;
    map<int,Quadrature<SPACEDIM-1> > nqmap; // natural bc's
    FEValues *fe_values;
#if (SPACEDIM>=2)
    FEFaceValues *fe_face_values;
#endif
    map<int,REAL> boundary_values;
    Tensor<1,SPACEDIM> (*boundaryForce)(const Point<SPACEDIM>&);
    Vector (*rightHandSideValue)(const Point<SPACEDIM>&);
  public:
    NonlinearDofHandler(Tessellation&,FiniteElement&,const Mapping&,
      ConstraintMatrix&);
    virtual ~NonlinearDofHandler();

    bool nonlinearSystemEvaluation(const Vector&,Vector&) const;
    virtual void printOn(ostream&) const;
};

#endif
