#ifndef PATCH_H
#define PATCH_H "$Header:$"
//----------------------------  data_out_base.h  ---------------------------
//   $Id: data_out_base.h,v 1.56 2003/01/08 17:57:47 wolf Exp $
//   Version: $Name: Version-4-0-0 $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//----------------------------  data_out_base.h  ---------------------------
//
//modified from deal.II/base/include/base/data_out_base.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "Array.H"
#include "DofHandler.H"
#include "Mapping.H"
#include "NumPtr.H"
#include "Point.H"
#include "Quadrature.H"
#include "Shape.H"
#include "XYGraphTool.H"

//compare to DEAL DataOutBase::Patch
class Patch {
  protected:
    const SHAPE_CLASS *shape;
    const LEVEL_CLASS *level;
    NumPtr<Point<SPACEDIM> > ep;
    NumPtr<int> neighbors;
    int patch_index;
    Array<2,REAL> data;

    static int n_subdivisions;
    static map<int,Quadrature<SPACEDIM> > q_map;
    const Quadrature<SPACEDIM>& getQuadrature() const;
#if (SPACEDIM==3)
    void plotClipPlane(VolGraphTool&,REAL,REAL,
      const REAL value[2][2][2],const Vector3 vertex[2][2][2],
      const DofHandler*,const Mapping*,const Vector*) const;
    void plotTetrahedralClipPlane(VolGraphTool&,REAL,REAL,REAL,int,
      HAND,const DofHandler*,const Mapping*,const Vector*,
      const Point<SPACEDIM> x[4]) const;
    void plotPrismClipPlane(VolGraphTool&,REAL,REAL,REAL,int,HAND,
      const DofHandler*,const Mapping*,const Vector*,
      const Point<SPACEDIM> x[6]) const;
#endif
  public:
    static const int no_neighbor;
    static void setQuadratures(int);
    static const map<int,Quadrature<SPACEDIM> >& 
      getQuadratureMap() { return q_map; }

    Patch(const SHAPE_CLASS*,const LEVEL_CLASS*,const Tessellation&,
      const Mapping&,int,int);
    virtual ~Patch();
    int numberQuadraturePoints() const {
      return getQuadrature().numberPoints();
    }
    REAL getData(int dataset,int q_point) const {
      return data(dataset,q_point);
    }
    void setData(int dataset,REAL val) {
      for (int q=0;q<data.size(1);q++) {
        data(dataset,q)=val;
      }
    }
    void setData(int dataset,const NumPtr<REAL> &values) {
      for (int q=0;q<values.getNumber();q++) {
        data(dataset,q)=values[q];
      }
    }
    void setData(int dataset,int n_components,
    const NumPtr<Vector> &values) {
      int dc=dataset*n_components;
      for (int c=0;c<n_components;c++) {
        for (int q=0;q<values.getNumber();q++) {
          data(dc+c,q)=values[q][c];
        }
      }
    }
    int dataSize(int i) const { return data.size(i); }
    void setNeighbor(int f,int n) {
      neighbors[f]=n;
    }
    int nSubdivisions() const { return n_subdivisions; }
    void getMeshBounds(Point<SPACEDIM>&,Point<SPACEDIM>&) const;
    void getDataBounds(int,REAL&,REAL&) const;
#if (SPACEDIM==1)
    void plot(XYGraphTool&,int) const;
#endif
#if (SPACEDIM==2)
    void plot(AreaGraphTool&,int,REAL,REAL) const;
    void plotSurface(VolGraphTool&,int,REAL,REAL,BOOLEAN) const;
    void draw(AreaGraphTool&) const;
    void contour(AreaGraphTool&,REAL,int) const;
#endif
#if (SPACEDIM==3)
    void plotClipPlane(VolGraphTool&,REAL,REAL,REAL,int,HAND,
      const DofHandler*,const Mapping*,const Vector*) const;
    void plotSurface(VolGraphTool&,int,REAL,REAL,REAL,BOOLEAN) const;
    void draw(VolGraphTool&) const;
#endif
    virtual void printOn(ostream&) const;
};

#if (SPACEDIM==3)
void drawSurfaceWithinCube(VolGraphTool&,REAL,REAL value[2][2][2],
  Vector3 vertex[2][2][2]);
#endif

#endif
