// "$Header: /home/faculty/johnt/cvs/deal/grid/testGridGenerator2.C,v 1.1.1.1 2004/08/23 18:42:51 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <fstream>
#include <iomanip>
#include <iostream>
#include <stdlib.h>
#include <unistd.h>

#include "AreaGraphTool.H"
#include "Boundary.H"
#include "MemoryDebugger.H"
#include "Palette.H"
#include "Quadrature.H"
#include "SetTraps.H"
#include "Tessellation.H"
#include "Tracer.H"
#include "VolGraphTool.H"
#include "XYGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKGUI.H"
#include "GTKGUIVirtualInput.H"
#include "GTKWindow.H"
#else
#include "GUI.H"
#include "GUIVirtualInput.H"
#endif

REAL winsize=0.5;

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
/*
#if (SPACEDIM==2)
void plot(Tessellation &tess,const NumPtr<Point<2> > &vertices,
const char *title) {
  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]; rlo[0]=0.; rlo[1]=0.;
  REAL rhi[2]; rhi[0]=1.; rhi[1]=1.;
  AreaGraphTool gt(title,"x","y",rlo,rhi,&cmap,0,winsize);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();

  for (bool test=tess.firstActiveShape();test;
  test=tess.nextActiveShape()) {
    const SHAPE_CLASS *element=tess.presentShape();
#ifdef DEBUG
//  cout << "\telement = " << element->elementDataIndex()
//       << " has vertices numbered";
//  for (int i=0;i<element->numberVertices();i++) {
//    cout << " " << element->codimension1ShapeIndex(i);
//  }
//  cout << endl;
#endif
    int nv=element->numberVertices();
    Point<2> vertex=vertices[tess.presentShapeVertexIndex(nv-1)];
    gt.movePen(vertex[0],vertex[1]);
    for (int i=0;i<element->numberVertices();i++) {
      vertex=vertices[tess.presentShapeVertexIndex(i)];
      gt.drawLine(vertex[0],vertex[1]);
    }
  }
  gt.flush();
  AreaGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
*/
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void getNewPointOnPlot(AreaGraphTool &gt,Tessellation &tess) {
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool doit=tess.firstCodimension1Shape();doit;
  doit=tess.nextCodimension1Shape()) {
    const BoundaryCondition *bc=
      tess.presentCodimension1Level()->getBoundaryCondition();
    if (bc!=0) {
      const CODIMENSION1_SHAPE_CLASS *e=tess.presentCodimension1Shape();
      Point<SPACEDIM> p=bc->getBoundary()->getNewPointOn(*e,vertices);
      char s[6];
      snprintf(s,6,"%d",e->elementDataIndex());
      gt.putString(p[0],p[1],s);
    }
  }
}
void getIntermediatePointsOnPlot(AreaGraphTool &gt,Tessellation &tess,
int n) {
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool doit=tess.firstCodimension1Shape();doit;
  doit=tess.nextCodimension1Shape()) {
    const BoundaryCondition *bc=
      tess.presentCodimension1Level()->getBoundaryCondition();
    if (bc!=0) {
      const CODIMENSION1_SHAPE_CLASS *e=tess.presentCodimension1Shape();
      NumPtr<Point<SPACEDIM> > points(n);
      bc->getBoundary()->getIntermediatePointsOn(*e,vertices,points);
      for (int i=0;i<points.getNumber();i++) {
        char s[8];
        snprintf(s,8,"%d.%d",e->elementDataIndex(),i);
        gt.putString(points[i][0],points[i][1],s);
      }
    }
  }
}
/*
void getMappedPointsOnPlot(AreaGraphTool &gt,Tessellation &tess,int n)
{
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  GaussianQuadrature<1> gq(GaussianQuadrature<1>::argGivenNPoints(n));
  for (bool doit=tess.firstCodimension1Shape();doit;
  doit=tess.nextCodimension1Shape()) {
    const BoundaryCondition *bc=
      tess.presentCodimension1Level()->getBoundaryCondition();
    if (bc!=0) {
      const CODIMENSION1_SHAPE_CLASS *e=tess.presentCodimension1Shape();
      NumPtr<Point<SPACEDIM> > points;
      bc->getBoundary()->
        getMappedPointsOn(*e,vertices,gq.points(),points);
      for (int i=0;i<points.getNumber();i++) {
        char s[8];
        snprintf(s,8,"%d.%d",e->elementDataIndex(),i);
        gt.putString(points[i][0],points[i][1],s);
      }
    }
  }
}
*/
void getMappedMeasureOnPlot(AreaGraphTool &gt,Tessellation &tess,int n)
{
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  GaussianQuadrature<1> gq(GaussianQuadrature<1>::argGivenNPoints(n));
  REAL angle=2.6;
  COMPLEX arrow=COMPLEX(cos(angle),sin(angle))*ARROW_SIZE;
  for (bool doit=tess.firstCodimension1Shape();doit;
  doit=tess.nextCodimension1Shape()) {
    const BoundaryCondition *bc=
      tess.presentCodimension1Level()->getBoundaryCondition();
    if (bc!=0) {
      const CODIMENSION1_SHAPE_CLASS *e=tess.presentCodimension1Shape();
      NumPtr<Point<SPACEDIM> > points;
      NumPtr<Point<SPACEDIM> > normals;
      NumPtr<REAL> surface_measure;
      bc->getBoundary()->getMappedMeasureOn(*e,vertices,gq.points(),
        points,normals,surface_measure);
      REAL len=0.;
      for (int i=0;i<points.getNumber();i++) {
        char s[8];
        snprintf(s,8,"%d.%d",e->elementDataIndex(),i);
        gt.putString(points[i][0],points[i][1],s);
        COMPLEX base(points[i][0],points[i][1]);
        COMPLEX head(normals[i][0]*0.05,normals[i][1]*0.05);
        gt.drawVector(base,head,arrow);
        len+=surface_measure[i]*gq.weight(i);
      }
      cout << "\tface with vertices " << e->vertexIndex(0) << " "
           << e->vertexIndex(1) << " has length " << len << endl;
    }
  }
}
/*
void getNormalsAtVerticesOfPlot(AreaGraphTool &gt,Tessellation &tess) {
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  REAL angle=2.6;
  COMPLEX arrow=COMPLEX(cos(angle),sin(angle))*ARROW_SIZE;
  for (bool doit=tess.firstCodimension1Shape();doit;
  doit=tess.nextCodimension1Shape()) {
    const BoundaryCondition *bc=
      tess.presentCodimension1Level()->getBoundaryCondition();
    if (bc!=0) {
      const CODIMENSION1_SHAPE_CLASS *e=tess.presentCodimension1Shape();
      NumPtr<Tensor<1,SPACEDIM> > normals;
      bc->getBoundary()->getNormalsAtVerticesOf(*e,vertices,normals);
      for (int v=0;v<e->numberVertices();v++) {
        const Point<2> &vertex=vertices[e->vertexIndex(v)];
        COMPLEX base(vertex[0],vertex[1]);
        COMPLEX head(normals[v][0]*0.05,normals[v][1]*0.05);
        gt.drawVector(base,head,arrow);
      }
    }
  }
}
*/
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void plotTriangleElementsAndVertices(AreaGraphTool &gt,
Tessellation &tess) {
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool test=tess.firstActiveShape();test;
  test=tess.nextActiveShape()) {
    const SHAPE_CLASS *element=tess.presentShape();
    int nv=element->numberVertices();
    Point<2> vertex=vertices[tess.presentShapeVertexIndex(nv-1)];
    gt.movePen(vertex[0],vertex[1]);
    for (int i=0;i<element->numberVertices();i++) {
      vertex=vertices[tess.presentShapeVertexIndex(i)];
      gt.drawLine(vertex[0],vertex[1]);
    }
  }
  gt.setfgColor("blue");
  for (int v=0;v<vertices.getNumber();v++) {
    char s[4];
    snprintf(s,4,"%d",v);
    gt.putString(vertices[v][0],vertices[v][1],s);
  }
}
void oneTriangle(int bf) {
//TRACER_CALL(t,"oneTriangle with BallBoundary");
  cout << "\toneTriangle, bf = " << bf << endl;
  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]; rlo[0]=-1.; rlo[1]=-1.;
  REAL rhi[2]; rhi[0]=1.; rhi[1]=1.;
  AreaGraphTool gt("oneTriangle with BallBoundary","x","y",rlo,rhi,
    &cmap,0,winsize);

  Point<2> p0(1.,0.);
  Point<2> p1(-0.5, sqrt(.75));
  Point<2> p2(-0.5,-sqrt(.75));
  NumPtr<Point<2> > vertices(3);
  vertices[0]=p0;
  vertices[1]=p1;
  vertices[2]=p2;

  NumPtr<CellData> cells(1);
  cells[0].allocate(3);
  for (int j=0;j<3;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  NumPtr<const BoundaryCondition*> boundary_condition(3);
  NumPtr<Boundary*> pb(3);
  Point<2> centroid(0.,0.);
  pb[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],vertices[1]);
  pb[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],vertices[2]);
  pb[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[2],vertices[0]);
  Boundary *bb=OPERATOR_NEW BallBoundary(centroid,1.);
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<2> center=face_element->center(vertices);
      if (center[0]<-0.4) {
        boundary_condition[1]=
          OPERATOR_NEW BoundaryCondition( (bf==1 ? bb : pb[1]) );
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else if (center[1]<0.) {
        boundary_condition[2]=
          OPERATOR_NEW BoundaryCondition( (bf==2 ? bb : pb[2]) );
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      } else {
        boundary_condition[0]=
          OPERATOR_NEW BoundaryCondition( (bf==0 ? bb : pb[0]) );
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      }
    }
  }

  {
    cout << "\t\tgetNewPointOn" << endl;
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    plotTriangleElementsAndVertices(gt,tess);
    gt.setfgColor("red");
    getNewPointOnPlot(gt,tess);
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
//

  cout << "\t\tgetIntermediatePointsOn" << endl;
  for (int n=1;n<4;n++) {
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    plotTriangleElementsAndVertices(gt,tess);
    gt.setfgColor("red");
    getIntermediatePointsOnPlot(gt,tess,n);
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
//

/*
  cout << "\t\tgetMappedPointsOn" << endl;
  for (int n=1;n<4;n++) {
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    plotTriangleElementsAndVertices(gt,tess);
    gt.setfgColor("red");
    getMappedPointsOnPlot(gt,tess,n);
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
*/

  cout << "\t\tgetMappedMeasureOn" << endl;
  for (int n=1;n<4;n++) {
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    plotTriangleElementsAndVertices(gt,tess);
    gt.setfgColor("red");
    getMappedMeasureOnPlot(gt,tess,n);
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
//

/*
  {
    cout << "\t\tgetNormalsAtVerticesOf" << endl;
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    plotTriangleElementsAndVertices(gt,tess);
    gt.setfgColor("red");
    getNormalsAtVerticesOfPlot(gt,tess);
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
*/
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  for (int f=0;f<boundary_condition.getNumber();f++) {
    delete boundary_condition[f]; boundary_condition[f]=0;
  }
  for (int f=0;f<pb.getNumber();f++) {
    delete pb[f]; pb[f]=0;
  }
  delete bb; bb=0;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void plotCircleAndInteriorLines(AreaGraphTool &gt,Tessellation &tess,
Point<2> &center,REAL radius) {
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  gt.movePen(center[0]+radius,center[1]);
  REAL angle=0.;
  REAL dangle=2.*M_PI/256.;
  for (int i=0;i<=256;i++,angle+=dangle) {
    gt.drawLine(center[0]+radius*cos(angle),
                center[1]+radius*sin(angle));
  }

  for (bool doit=tess.firstCodimension1Shape();doit;
  doit=tess.nextCodimension1Shape()) {
    const CODIMENSION1_SHAPE_CLASS *e=tess.presentCodimension1Shape();
    if (tess.presentCodimension1Level()->getBoundaryCondition()==0) {
      Point<2> vertex=vertices[e->vertexIndex(0)];
      gt.movePen(vertex[0],vertex[1]);
      vertex=vertices[e->vertexIndex(1)];
      gt.drawLine(vertex[0],vertex[1]);
    }
  }
  gt.setfgColor("blue");
  for (int v=0;v<vertices.getNumber();v++) {
    char s[4];
    snprintf(s,4,"%d",v);
    gt.putString(vertices[v][0],vertices[v][1],s);
  }
}
void twoTriangles() {
//TRACER_CALL(t,"twoTriangles with BallBoundary");
  cout << "\ttwoTriangles" << endl;
  REAL radius=sqrt(0.5);
#ifdef DEBUG
  cout << "\tarc length of quarter circle is " << 0.5*M_PI*radius 
       << endl;
#endif
  NumPtr<Point<2> > vertices(4);
  vertices[0]=Point<2>(0.,0.);
  vertices[1]=Point<2>(1.,0.);
  vertices[2]=Point<2>(1.,1.);
  vertices[3]=Point<2>(0.,1.);

  NumPtr<CellData> cells(2);
  for (int i=0;i<2;i++) {
    cells[i].allocate(3);
    cells[i].setMaterialModel(0);
  }
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=1;
  cells[0].vertex(2)=2;
  cells[1].vertex(0)=0;
  cells[1].vertex(1)=2;
  cells[1].vertex(2)=3;

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  Point<2> center(0.5,0.5);
  BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
  BoundaryCondition *bc=OPERATOR_NEW BoundaryCondition(b);
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      codimension1_level->setBoundaryCondition(bc);
    }
  }
  tess.synchronize(boundary_data);

  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]; rlo[0]=-0.2; rlo[1]=-0.2;
  REAL rhi[2]; rhi[0]=1.2; rhi[1]=1.2;
  AreaGraphTool gt("twoTriangles with BallBoundary","x","y",rlo,rhi,
    &cmap,0,winsize);
//
  {
    cout << "\t\tgetNewPointOn" << endl;
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    plotCircleAndInteriorLines(gt,tess,center,radius);
    gt.setfgColor("red");
    getNewPointOnPlot(gt,tess);
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
//
//
  cout << "\t\tgetIntermediatePointsOn" << endl;
  for (int n=1;n<4;n++) {
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    plotCircleAndInteriorLines(gt,tess,center,radius);
    gt.setfgColor("red");
    getIntermediatePointsOnPlot(gt,tess,n);
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
/*
  cout << "\t\tgetMappedPointsOn" << endl;
  for (int n=1;n<4;n++) {
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    plotCircleAndInteriorLines(gt,tess,center,radius);
    gt.setfgColor("red");
    getMappedPointsOnPlot(gt,tess,n);
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
*/

  cout << "\t\tgetMappedMeasureOn" << endl;
  for (int n=1;n<4;n++) {
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    plotCircleAndInteriorLines(gt,tess,center,radius);
    gt.setfgColor("red");
    getMappedMeasureOnPlot(gt,tess,n);
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
//
/*
  cout << "\t\tgetNormalsAtVerticesOf" << endl;
  {
    gt.newPage();
    gt.setbgColor("white");
    gt.setfgColor("black");
    gt.drawAxes();
    plotCircleAndInteriorLines(gt,tess,center,radius);
    gt.setfgColor("red");
    getNormalsAtVerticesOfPlot(gt,tess);
    gt.flush();
    AreaGraphTool::WINDOW_TYPE::QuitButton qb;
  }
*/
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  delete bc;
  delete b;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
/*
#if (SPACEDIM==3)
void plot(Tessellation &tess,const NumPtr<Point<3> > &vertices,
const char *title) {
  Vector3 low(0.,0.,0.);
  Vector3 high(1.,1.,1.);
  Palette pal;
  BOOLEAN use_lighting=FALSE;
  VolGraphTool gt(title,low,high,pal,use_lighting,winsize);

  gt.newPage();
  gt.expose();

  for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
    {
      VolGraphTool::DrawDelimiter dd(do_best,&gt);
      for (bool test=tess.firstActiveShape();test;
      test=tess.nextActiveShape()) {
        const SHAPE_CLASS *element=tess.presentShape();
        tess.setCodimension1Level();
#ifdef DEBUG
//      cout << "\telement = " << element->elementDataIndex()
//           << " has vertices numbered";
//      for (int i=0;i<element->numberVertices();i++) {
//        cout << " " << element->codimension1ShapeIndex(i);
//      }
//      cout << endl;
#endif
        for (int f=0;f<element->numberCodimension1Shapes();
        f++) {
          int face_index=element->codimension1ShapeIndex(f);
          const CODIMENSION1_SHAPE_CLASS *face=
            tess.presentCodimension1Shape(face_index);
          int nv=face->numberVertices();
          int vertex_index=face->vertexIndex(nv-1);
          Point<3> start=vertices[vertex_index];
          Vector3 s(start[0],start[1],start[2]);
          for (int v=0;v<face->numberVertices();v++) {
            vertex_index=face->vertexIndex(v);
            Point<3> finish=vertices[vertex_index];
            Vector3 f(finish[0],finish[1],finish[2]);
            gt.drawLine(s,f);
            s=f;
          }
        }
      }
    }
    if (do_best==TRUE) break;
  }
  gt.eventLoop();
}
#endif
*/
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void plotShapes(VolGraphTool &gt,Tessellation &tess) {
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
    gt.getWindow()->getColor("black");
  for (bool test=tess.firstActiveShape();test;
  test=tess.nextActiveShape()) {
    const SHAPE_CLASS *element=tess.presentShape();
    element->draw(xc,vertices,gt);
  }
  gt.setfgColor("blue");
  for (int v=0;v<vertices.getNumber();v++) {
    char s[4];
    snprintf(s,4,"V%d",v);
    Vector3 p(vertices[v][0],vertices[v][1],vertices[v][2]);
    gt.putString(&p,s);
  }
}
void getNewPointOnLinesPlot(VolGraphTool &gt,Tessellation &tess) {
//TRACER_CALL(t,"getNewPointOnLinesPlot");
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool doit=tess.firstCodimension2Shape();doit;
  doit=tess.nextCodimension2Shape()) {
    pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
      tess.presentCodimension2Level()->getBoundaryConditions();
    if (bcs.first!=0) {
#ifdef DEBUG
//    cout << "\tline index = " 
//         << tess.presentCodimension2Level()->presentIndex() << endl;
#endif
      const Shape1 *e=tess.presentCodimension2Shape();
      Point<SPACEDIM> p;
      if (bcs.second==0) {
//      TRACER_CALL(t,"getNewPointOnLinesPlot getNewPointOn");
        p=bcs.first->getBoundary()->getNewPointOn(*e,vertices);
      } else {
//      TRACER_CALL(t,"getNewPointOnLinesPlot getNewPointOnIntersectionWith");
        p=bcs.first->getBoundary()->getNewPointOnIntersectionWith(
          bcs.second->getBoundary(),*e,vertices);
      }
      char s[6];
      snprintf(s,6,"L%d",e->elementDataIndex());
      Vector3 p3(p[0],p[1],p[2]);
      gt.putString(&p3,s);
    }
  }
}
void getIntermediatePointsOnLinesPlot(VolGraphTool &gt,Tessellation &tess,
int n) {
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool doit=tess.firstCodimension2Shape();doit;
  doit=tess.nextCodimension2Shape()) {
    pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
      tess.presentCodimension2Level()->getBoundaryConditions();
    if (bcs.first!=0) {
      const Shape1 *e=tess.presentCodimension2Shape();
      NumPtr<Point<SPACEDIM> > points(n);
      if (bcs.second==0) {
        bcs.first->getBoundary()->
          getIntermediatePointsOn(*e,vertices,points);
      } else {
        bcs.first->getBoundary()->
          getIntermediatePointsOnIntersectionWith(
            bcs.second->getBoundary(),*e,vertices,points);
      }
      for (int i=0;i<points.getNumber();i++) {
        char s[8];
        snprintf(s,8,"L%d.%d",e->elementDataIndex(),i);
        Vector3 p(points[i][0],points[i][1],points[i][2]);
        gt.putString(&p,s);
      }
    }
  }
}
/*
void getMappedPointsOnLinesPlot(VolGraphTool &gt,Tessellation &tess,
int n) {
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  GaussianQuadrature<1> gq(GaussianQuadrature<1>::argGivenNPoints(n));
  for (bool doit=tess.firstCodimension2Shape();doit;
  doit=tess.nextCodimension2Shape()) {
    pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
      tess.presentCodimension2Level()->getBoundaryConditions();
    if (bcs.first!=0) {
      const Shape1 *e=tess.presentCodimension2Shape();
      NumPtr<Point<SPACEDIM> > points;
      if (bcs.second==0) {
        bcs.first->getBoundary()->
          getMappedPointsOn(*e,vertices,gq.points(),points);
      } else {
        bcs.first->getBoundary()->
          getMappedPointsOnIntersectionWith(
            bcs.second->getBoundary(),*e,vertices,gq.points(),points);
      }
      for (int i=0;i<points.getNumber();i++) {
        char s[8];
        snprintf(s,8,"L%d.%d",e->elementDataIndex(),i);
        Vector3 p(points[i][0],points[i][1],points[i][2]);
        gt.putString(&p,s);
      }
    }
  }
}
*/
void getNewPointOnFacesPlot(VolGraphTool &gt,Tessellation &tess) {
//TRACER_CALL(t,"getNewPointOnFacesPlot");
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool doit=tess.firstCodimension1Shape();doit;
  doit=tess.nextCodimension1Shape()) {
    const BoundaryCondition *bc=
      tess.presentCodimension1Level()->getBoundaryCondition();
    if (bc!=0) {
      const CODIMENSION1_SHAPE_CLASS *e=tess.presentCodimension1Shape();
#ifdef DEBUG
//    cout << "\tface index = " << e->elementDataIndex() << endl; 
//    e->printOn(cout);
#endif
      Point<SPACEDIM> p=bc->getBoundary()->getNewPointOn(*e,vertices);
      char s[6];
      snprintf(s,6,"F%d",e->elementDataIndex());
      Vector3 p3(p[0],p[1],p[2]);
      gt.putString(&p3,s);
    }
  }
}
void getMappingSupportPointsOnFacePlot(VolGraphTool &gt,Tessellation &tess,
int n) {
//TRACER_CALL(t,"getMappingSupportPointsOnFacePlot");
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
    const SHAPE_CLASS *e=tess.presentShape();
    LEVEL_CLASS *level=tess.presentLevel();
    NumPtr<Point<SPACEDIM> > 
      points(e->numberEquallySpacedLatticePoints(n));
    tess.mappingSupportPointsOnBoundaryOf(e,n,level,points);
    for (int f=0;f<e->numberCodimension1Shapes();f++) {
      int face_index=e->codimension1ShapeIndex(f);
//    const BoundaryCondition *bc=level->CODIMENSION1_LEVEL_CLASS::
//      getBoundaryCondition(face_index);
//    if (bc!=0) {
        NumPtr<int> face_lattice_indices;
        e->equallySpacedLatticePointIndicesOfFace(f,n,
          face_lattice_indices);
        for (int i=0;i<face_lattice_indices.getNumber();i++) {
          char s[8];
          snprintf(s,8,"F%d.%d",face_index,i);
          const Point<SPACEDIM> &point=points[face_lattice_indices[i]];
          Vector3 p(point[0],point[1],point[2]);
          gt.putString(&p,s);
        }
//    }
    }
  }
}
/*
void getMappedPointsOnFacesPlot(VolGraphTool &gt,Tessellation &tess,
const map<int,Quadrature<2>*> &qmap) {
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool doit=tess.firstCodimension1Shape();doit;
  doit=tess.nextCodimension1Shape()) {
    const BoundaryCondition *bc=
      tess.presentCodimension1Level()->getBoundaryCondition();
    if (bc!=0) {
      const CODIMENSION1_SHAPE_CLASS *e=
        tess.presentCodimension1Shape();
      map<int,Quadrature<2>*>::const_iterator qmap_iterator=
        qmap.find(e->numberVertices());
      CHECK_TEST(qmap_iterator!=qmap.end());
      NumPtr<Point<SPACEDIM> > points;
      bc->getBoundary()->getMappedPointsOn(*e,vertices,
        qmap_iterator->second->points(),points);
      for (int i=0;i<points.getNumber();i++) {
        char s[8];
        snprintf(s,8,"F%d.%d",e->elementDataIndex(),i);
        Vector3 p(points[i][0],points[i][1],points[i][2]);
        gt.putString(&p,s);
      }
    }
  }
}
*/
void getMappedMeasureOnPlot(VolGraphTool &gt,Tessellation &tess,
const map<int,Quadrature<2>*> &qmap) {
//TRACER_CALL(t,"getMappedMeasureOnPlot");
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  for (bool doit=tess.firstCodimension1Shape();doit;
  doit=tess.nextCodimension1Shape()) {
    const BoundaryCondition *bc=
      tess.presentCodimension1Level()->getBoundaryCondition();
    if (bc!=0) {
      const CODIMENSION1_SHAPE_CLASS *e=
        tess.presentCodimension1Shape();
#ifdef DEBUG
//    e->printOn(cout);
#endif
      NumPtr<Point<SPACEDIM> > points;
      NumPtr<Point<SPACEDIM> > normals;
      NumPtr<REAL> surface_measure;
      map<int,Quadrature<2>*>::const_iterator qmap_iterator=
        qmap.find(e->numberVertices());
      CHECK_TEST(qmap_iterator!=qmap.end());
      Quadrature<2> *q=qmap_iterator->second;
      bc->getBoundary()->getMappedMeasureOn(*e,vertices,q->points(),
        points,normals,surface_measure);
      REAL area=0.;
      for (int i=0;i<points.getNumber();i++) {
        char s[8];
        snprintf(s,8,"F%d.%d",e->elementDataIndex(),i);
        Vector3 p(points[i][0],points[i][1],points[i][2]);
        gt.putString(&p,s);
        area+=surface_measure[i]*q->weight(i); 
      }
      cout << "\tface with vertices ";
      for (int v=0;v<e->numberVertices();v++) {
        cout << e->vertexIndex(v) << " ";
      }
      cout << " has area " << area << endl;
    }
  }
}
void getMappedMeasureOnPlotNormals(VolGraphTool &gt,Tessellation &tess,
const map<int,Quadrature<2>*> &qmap) {
//TRACER_CALL(t,"getMappedMeasureOnPlotNormals");
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  gt.setfgColor("red");
  for (bool doit=tess.firstCodimension1Shape();doit;
  doit=tess.nextCodimension1Shape()) {
    const BoundaryCondition *bc=
      tess.presentCodimension1Level()->getBoundaryCondition();
    if (bc!=0) {
      const CODIMENSION1_SHAPE_CLASS *e=
        tess.presentCodimension1Shape();
      NumPtr<Point<SPACEDIM> > points;
      NumPtr<Point<SPACEDIM> > normals;
      NumPtr<REAL> surface_measure;
      map<int,Quadrature<2>*>::const_iterator qmap_iterator=
        qmap.find(e->numberVertices());
      CHECK_TEST(qmap_iterator!=qmap.end());
      Quadrature<2> *q=qmap_iterator->second;
      bc->getBoundary()->getMappedMeasureOn(*e,vertices,q->points(),
        points,normals,surface_measure);
      for (int i=0;i<points.getNumber();i++) {
        Point<3> vertex=points[i];
        Vector3 base(vertex[0],vertex[1],vertex[2]);
        Vector3 head(normals[i][0],normals[i][1],normals[i][2]);
        gt.drawVector(base,base+head*.1);
      }
    }
  }
}
/*
void getNormalsAtVerticesOfPlot(VolGraphTool &gt,Tessellation &tess) {
  const NumPtr<Point<SPACEDIM> > &vertices=tess.getVertices();
  gt.setfgColor("red");
  for (bool doit=tess.firstCodimension1Shape();doit;
  doit=tess.nextCodimension1Shape()) {
    const BoundaryCondition *bc=
      tess.presentCodimension1Level()->getBoundaryCondition();
    if (bc!=0) {
      const CODIMENSION1_SHAPE_CLASS *e=tess.presentCodimension1Shape();
      NumPtr<Tensor<1,SPACEDIM> > normals;
      bc->getBoundary()->getNormalsAtVerticesOf(*e,vertices,normals);
      for (int i=0;i<normals.getNumber();i++) {
        Point<3> vertex=vertices[e->vertexIndex(i)];
        Vector3 base(vertex[0],vertex[1],vertex[2]);
        Vector3 head(normals[i][0],normals[i][1],normals[i][2]);
        gt.drawVector(base,base+head*.1);
      }
    }
  }
}
*/
void plotEverything(VolGraphTool &gt,Tessellation &tess) {
//
  cout << "\tgetNewPointOn lines" << endl;
  for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
    VolGraphTool::DrawDelimiter dd(do_best,&gt);
    plotShapes(gt,tess);
    getNewPointOnLinesPlot(gt,tess);
    if (do_best==TRUE) break;
  }
  gt.eventLoop();
//

//
  for (int n=1;n<4;n++) {
    cout << "\tgetIntermediatePointsOn lines, n = " << n << endl;
    for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
      VolGraphTool::DrawDelimiter dd(do_best,&gt);
      plotShapes(gt,tess);
      getIntermediatePointsOnLinesPlot(gt,tess,n);
      if (do_best==TRUE) break;
    }
    gt.eventLoop();
  }
//

/*obsolete:
  for (int n=1;n<4;n++) {
    cout << "\tgetMappedPointsOn lines, n = " << n << endl;
    for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
      VolGraphTool::DrawDelimiter dd(do_best,&gt);
      plotShapes(gt,tess);
      getMappedPointsOnLinesPlot(gt,tess,n);
      if (do_best==TRUE) break;
    }
    gt.eventLoop();
  }
*/

//
  cout << "\tgetNewPointOn face" << endl;
  for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
    VolGraphTool::DrawDelimiter dd(do_best,&gt);
    plotShapes(gt,tess);
    getNewPointOnFacesPlot(gt,tess);
    if (do_best==TRUE) break;
  }
  gt.eventLoop();
//

//
  for (int n=1;n<5;n++) 
//int n=3;
  {
    cout << "\tgetMappingSupportPointsOn face, n = " << n << endl;
    for (BOOLEAN do_best=FALSE;;do_best=!do_best) 
//  bool do_best=FALSE;
    {
      VolGraphTool::DrawDelimiter dd(do_best,&gt);
      plotShapes(gt,tess);
      getMappingSupportPointsOnFacePlot(gt,tess,n);
      if (do_best==TRUE) break;
    }
    gt.eventLoop();
  }
//

//
  for (int n=1;n<4;n++)
//int n=2;
  {
    TriangleGaussianQuadrature triangle_quadrature((n==1 ? 2 : n+3));
    GaussianQuadrature<2> quadrilateral_quadrature(n);
    map<int,Quadrature<2>*> qmap;
    qmap[3]=&triangle_quadrature;
    qmap[4]=&quadrilateral_quadrature;

//  obsolete:
//  cout << "\tgetMappedPointsOn face, n = " << n << endl;
//  for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
//    VolGraphTool::DrawDelimiter dd(do_best,&gt);
//    plotShapes(gt,tess);
//    getMappedPointsOnFacesPlot(gt,tess,qmap);
//    if (do_best==TRUE) break;
//  }
//  gt.eventLoop();

//
    cout << "\tgetMappedMeasureOn face, n = " << n << endl;
    for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
      VolGraphTool::DrawDelimiter dd(do_best,&gt);
      plotShapes(gt,tess);
      getMappedMeasureOnPlot(gt,tess,qmap);
      if (do_best==TRUE) break;
    }
    gt.eventLoop();

    cout << "\tgetMappedMeasureOn face normals, n = " << n << endl;
    for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
      VolGraphTool::DrawDelimiter dd(do_best,&gt);
      plotShapes(gt,tess);
      getMappedMeasureOnPlotNormals(gt,tess,qmap);
      if (do_best==TRUE) break;
    }
    gt.eventLoop();
  }
//

/*obsolete:
  cout << "\tgetNormalsAtVerticesOf" << endl;
  for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
    VolGraphTool::DrawDelimiter dd(do_best,&gt);
    plotShapes(gt,tess);
    getNormalsAtVerticesOfPlot(gt,tess);
    if (do_best==TRUE) break;
  }
  gt.eventLoop();
*/
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void oneHexahedronWithPlanes(VolGraphTool &gt) {
  cout << "\toneHexahedronWithPlanes" << endl;
  NumPtr<Point<3> > vertices(8);
  vertices[0]=Point<3>(0.,0.,0.);
  vertices[1]=Point<3>(1.,0.,0.);
  vertices[2]=Point<3>(0.,1.,0.);
  vertices[3]=Point<3>(1.,1.,0.);
  vertices[4]=Point<3>(0.,0.,1.);
  vertices[5]=Point<3>(1.,0.,1.);
  vertices[6]=Point<3>(0.,1.,1.);
  vertices[7]=Point<3>(1.,1.,1.);

  NumPtr<CellData> cells(1);
  cells[0].allocate(8);
  for (int j=0;j<8;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  NumPtr<Boundary*> boundary(6);
  NumPtr<BoundaryCondition*> boundary_condition(6);
  Point<3> centroid(0.5,0.5,0.5);
  boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[2],vertices[4]); // back
  boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
    vertices[3],vertices[5]); // front
  boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[4]); // left
  boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[2],
    vertices[3],vertices[6]); // right
  boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[2]); // bottom
  boundary[5]=OPERATOR_NEW PlaneBoundary(centroid,vertices[4],
    vertices[5],vertices[6]); // top
  for (int f=0;f<6;f++) {
    boundary_condition[f]=OPERATOR_NEW BoundaryCondition(boundary[f]);
#ifdef DEBUG
//  cout << "\tboundary[" << f << "] = " << boundary[f] << endl;
#endif
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
      codimension1_level->getShape();
      Point<3> center=face_element->center(vertices);
      if (center[0]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      } else if (center[0]>0.9) {
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else if (center[1]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      } else if (center[1]>0.9) {
        codimension1_level->setBoundaryCondition(boundary_condition[3]);
      } else if (center[2]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[4]);
      } else {
        codimension1_level->setBoundaryCondition(boundary_condition[5]);
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif
  Vector3 low(-0.05,-0.05,-0.05);
  Vector3 high(1.05,1.05,1.05);
  gt.rescale(low,high);
  plotEverything(gt,tess);
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void oneHexahedronInBall(VolGraphTool &gt) {
  cout << "\toneHexahedronInBall" << endl;
  REAL radius=sqrt(.75);
  cout << "\tsurface area of face = " << 4.*M_PI*radius*radius/6.
       << endl;
  NumPtr<Point<3> > vertices(8);
  vertices[0]=Point<3>(0.,0.,0.);
  vertices[1]=Point<3>(1.,0.,0.);
  vertices[2]=Point<3>(0.,1.,0.);
  vertices[3]=Point<3>(1.,1.,0.);
  vertices[4]=Point<3>(0.,0.,1.);
  vertices[5]=Point<3>(1.,0.,1.);
  vertices[6]=Point<3>(0.,1.,1.);
  vertices[7]=Point<3>(1.,1.,1.);

  NumPtr<CellData> cells(1);
  cells[0].allocate(8);
  for (int j=0;j<8;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  Point<3> centroid(0.5,0.5,0.5);
  Boundary *boundary=OPERATOR_NEW BallBoundary(centroid,radius);
  BoundaryCondition* boundary_condition=
    OPERATOR_NEW BoundaryCondition(boundary);;
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      codimension1_level->setBoundaryCondition(boundary_condition);
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif
  Vector3 low(-0.45,-0.45,-0.45);
  Vector3 high(1.45,1.45,1.45);
  gt.rescale(low,high);
  plotEverything(gt,tess);
  delete boundary_data;
  delete boundary_condition; boundary_condition=0;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//should fail Lenoir test for at most one face on boundary
void oneTetrahedronInBallFace(VolGraphTool &gt,int bf) {
  cout << "\toneTetrahedronInBallFace bf = " << bf << endl;
  REAL radius=1.;
  cout << "\tsurface area of face = " << 4.*M_PI*radius*radius/4.
       << endl;
  NumPtr<Point<3> > vertices(4);
  vertices[0]=Point<3>(0.,0.,-radius);
  vertices[1]=Point<3>(radius*sqrt(8./9.),0.,1./3.);
  vertices[2]=Point<3>(-radius*sqrt(2./9.), radius*sqrt(2./3.),1./3.);
  vertices[3]=Point<3>(-radius*sqrt(2./9.),-radius*sqrt(2./3.),1./3.);

  NumPtr<CellData> cells(1);
  cells[0].allocate(4);
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=1;
  cells[0].vertex(2)=2;
  cells[0].vertex(3)=3;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  Point<3> centroid(0.,0.,0.);
  Boundary *boundary=OPERATOR_NEW BallBoundary(centroid,radius);
  NumPtr<Boundary*> pb(4);
  pb[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],vertices[2],
    vertices[3]);
  pb[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],vertices[1],
    vertices[3]);
  pb[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],vertices[1],
    vertices[2]);
  pb[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],vertices[2],
    vertices[3]);
  NumPtr<BoundaryCondition*> boundary_condition(4);
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      Point<3> face_center=
        tess.presentCodimension1Shape()->center(vertices);
      for (int f=0;f<pb.getNumber();f++) {
        if (abs(pb[f]->constraint(face_center))<1.e-5) {
          boundary_condition[f]=
            OPERATOR_NEW BoundaryCondition(bf==f ? boundary : pb[f]);
          codimension1_level->setBoundaryCondition(boundary_condition[f]);
          break;
        }
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//for (int v=0;v<vertices.getNumber();v++) {
//  cout << "\tconstraint(" << v << ") = " 
//       << boundary->constraint(vertices[v]) << endl;
//}
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif
  Vector3 low(-radius*1.25,-radius*1.25,-radius*1.25);
  Vector3 high(radius*1.25,radius*1.25,radius*1.25);
  gt.rescale(low,high);
  plotEverything(gt,tess);
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<pb.getNumber();i++) {
    delete pb[i]; pb[i]=0;
  }
  delete boundary; boundary=0;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
//should fail Lenoir test for at most one face on boundary
void oneTetrahedronInBallLine(VolGraphTool &gt,int bl) {
  cout << "\toneTetrahedronInBallLine bl = " << bl << endl;
  REAL radius=1.;
  cout << "\tsurface area of face = " << 4.*M_PI*radius*radius/4.
       << endl;
  NumPtr<Point<3> > vertices(4);
  vertices[0]=Point<3>(0.,0.,-radius);
  vertices[1]=Point<3>(radius*sqrt(8./9.),0.,1./3.);
  vertices[2]=Point<3>(-radius*sqrt(2./9.), radius*sqrt(2./3.),1./3.);
  vertices[3]=Point<3>(-radius*sqrt(2./9.),-radius*sqrt(2./3.),1./3.);

  NumPtr<CellData> cells(1);
  cells[0].allocate(4);
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=1;
  cells[0].vertex(2)=2;
  cells[0].vertex(3)=3;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  Point<3> centroid(0.,0.,0.);
  Boundary *boundary=OPERATOR_NEW BallBoundary(centroid,radius);
  BoundaryCondition* boundary_condition=
    OPERATOR_NEW BoundaryCondition(boundary);
  for (bool doit=tess.firstCodimension2ActiveShape();doit;
  doit=tess.nextCodimension2ActiveShape()) {
    Level1 *codimension2_level=tess.presentCodimension2Level();
    codimension2_level->setBoundaryCondition(0);
    codimension2_level->setBoundaryCondition(
      codimension2_level->presentIndex()==bl ?  boundary_condition : 0);
  }
  for (bool doit=tess.firstCodimension1Shape();doit;
  doit=tess.nextCodimension1Shape()) {
    tess.presentCodimension1Level()->setBoundaryCondition(0);
  }
#ifdef DEBUG
//for (int v=0;v<vertices.getNumber();v++) {
//  cout << "\tconstraint(" << v << ") = " 
//       << boundary->constraint(vertices[v]) << endl;
//}
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif
  Vector3 low(-radius*1.25,-radius*1.25,-radius*1.25);
  Vector3 high(radius*1.25,radius*1.25,radius*1.25);
  gt.rescale(low,high);
  plotEverything(gt,tess);
  delete boundary_data;
  delete boundary_condition; boundary_condition=0;
  delete boundary; boundary=0;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void fourTetrahedraInBall(VolGraphTool &gt) {
  cout << "\tfourTetrahedraInBall" << endl;
  REAL radius=1.;
  cout << "\tsurface area of face = " << 4.*M_PI*radius*radius/4.
       << endl;
  NumPtr<Point<3> > vertices(5);
  vertices[0]=Point<3>(0.,0.,0.);
  vertices[1]=Point<3>(0.,0.,-radius);
  vertices[2]=Point<3>(radius*sqrt(8./9.),0.,1./3.);
  vertices[3]=Point<3>(-radius*sqrt(2./9.), radius*sqrt(2./3.),1./3.);
  vertices[4]=Point<3>(-radius*sqrt(2./9.),-radius*sqrt(2./3.),1./3.);

  NumPtr<CellData> cells(4);
  cells[0].allocate(4);
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=1;
  cells[0].vertex(2)=3;
  cells[0].vertex(3)=2;
  cells[0].setMaterialModel(0);

  cells[1].allocate(4);
  cells[1].vertex(0)=0;
  cells[1].vertex(1)=1;
  cells[1].vertex(2)=4;
  cells[1].vertex(3)=3;
  cells[1].setMaterialModel(0);

  cells[2].allocate(4);
  cells[2].vertex(0)=0;
  cells[2].vertex(1)=1;
  cells[2].vertex(2)=2;
  cells[2].vertex(3)=4;
  cells[2].setMaterialModel(0);

  cells[3].allocate(4);
  cells[3].vertex(0)=0;
  cells[3].vertex(1)=2;
  cells[3].vertex(2)=3;
  cells[3].vertex(3)=4;
  cells[3].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  Point<3> centroid(0.,0.,0.);
  Boundary *boundary=OPERATOR_NEW BallBoundary(centroid,radius);
  BoundaryCondition* boundary_condition=
    OPERATOR_NEW BoundaryCondition(boundary);;
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      codimension1_level->setBoundaryCondition(boundary_condition);
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif
  Vector3 low(-radius*1.05,-radius*1.05,-radius*1.05);
  Vector3 high(radius*1.05,radius*1.05,radius*1.05);
  gt.rescale(low,high);
  plotEverything(gt,tess);
  delete boundary_data;
  delete boundary_condition; boundary_condition=0;
  delete boundary; boundary=0;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void onePrismInBallFace(VolGraphTool &gt,int bf) {
  TRACER_CALL(t,"onePrismInBallFace");
  cout << "\tonePrismInBallFace bf = " << bf << endl;
  REAL radius=sqrt(7);
  NumPtr<Point<3> > vertices(6);
  vertices[0]=Point<3>( 2.,       0.,-sqrt(3.));
  vertices[1]=Point<3>(-1., sqrt(3.),-sqrt(3.));
  vertices[2]=Point<3>(-1.,-sqrt(3.),-sqrt(3.));
  vertices[3]=Point<3>( 2.,       0., sqrt(3.));
  vertices[4]=Point<3>(-1., sqrt(3.), sqrt(3.));
  vertices[5]=Point<3>(-1.,-sqrt(3.), sqrt(3.));

  NumPtr<CellData> cells(1);
  cells[0].allocate(6);
  for (int j=0;j<6;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  Point<3> center(0.,0.,0.);
  Boundary *boundary=OPERATOR_NEW BallBoundary(center,radius);;
  NumPtr<Boundary*> pb(5);
  pb[0]=OPERATOR_NEW PlaneBoundary(center,vertices[0],vertices[1],
    vertices[2]);
  pb[1]=OPERATOR_NEW PlaneBoundary(center,vertices[3],vertices[4],
    vertices[5]);
  pb[2]=OPERATOR_NEW PlaneBoundary(center,vertices[0],vertices[2],
    vertices[3]);
  pb[3]=OPERATOR_NEW PlaneBoundary(center,vertices[0],vertices[1],
    vertices[3]);
  pb[4]=OPERATOR_NEW PlaneBoundary(center,vertices[1],vertices[2],
    vertices[4]);
  NumPtr<BoundaryCondition*> boundary_condition(5);
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      Point<3> face_center=
        tess.presentCodimension1Shape()->center(vertices);
      for (int f=0;f<pb.getNumber();f++) {
        if (abs(pb[f]->constraint(face_center))<1.e-5) {
          boundary_condition[f]=
            OPERATOR_NEW BoundaryCondition(bf==f ? boundary : pb[f]);
          codimension1_level->setBoundaryCondition(boundary_condition[f]);
          break;
        }
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif

  Vector3 low(-radius*1.25,-radius*1.25,-radius*1.25);
  Vector3 high(radius*1.25,radius*1.25,radius*1.25);
  gt.rescale(low,high);
  plotEverything(gt,tess);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<pb.getNumber();i++) {
    delete pb[i]; pb[i]=0;
  }
  delete boundary; boundary=0;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void onePrismInBallLine(VolGraphTool &gt,int bl) {
//TRACER_CALL(t,"onePrismInBallLine");
  cout << "\tonePrismInBallLine bl = " << bl << endl;
  REAL radius=sqrt(7);
  NumPtr<Point<3> > vertices(6);
  vertices[0]=Point<3>( 2.,       0.,-sqrt(3.));
  vertices[1]=Point<3>(-1., sqrt(3.),-sqrt(3.));
  vertices[2]=Point<3>(-1.,-sqrt(3.),-sqrt(3.));
  vertices[3]=Point<3>( 2.,       0., sqrt(3.));
  vertices[4]=Point<3>(-1., sqrt(3.), sqrt(3.));
  vertices[5]=Point<3>(-1.,-sqrt(3.), sqrt(3.));

  NumPtr<CellData> cells(1);
  cells[0].allocate(6);
  for (int j=0;j<6;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  Point<3> center(0.,0.,0.);
  Boundary *boundary=OPERATOR_NEW BallBoundary(center,radius);;
  BoundaryCondition* boundary_condition=
    OPERATOR_NEW BoundaryCondition(boundary);
  for (bool doit=tess.firstCodimension2ActiveShape();doit;
  doit=tess.nextCodimension2ActiveShape()) {
    Level1 *codimension2_level=tess.presentCodimension2Level();
    codimension2_level->setBoundaryCondition(0);
    codimension2_level->setBoundaryCondition(
      codimension2_level->presentIndex()==bl ?  boundary_condition : 0);
  }
  for (bool doit=tess.firstCodimension1Shape();doit;
  doit=tess.nextCodimension1Shape()) {
    tess.presentCodimension1Level()->setBoundaryCondition(0);
  }
#ifdef DEBUG
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif

  Vector3 low(-radius*1.25,-radius*1.25,-radius*1.25);
  Vector3 high(radius*1.25,radius*1.25,radius*1.25);
  gt.rescale(low,high);
  plotEverything(gt,tess);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  delete boundary_condition; boundary_condition=0;
  delete boundary; boundary=0;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void threePrismsAndTwoTetrahedraInBall(VolGraphTool &gt) {
  TRACER_CALL(t,"threePrismsAndTwoTetrahedraInBall");
  cout << "\tthreePrismsAndTwoTetrahedraInBall" << endl;
  REAL radius=1.;
  NumPtr<Point<3> > vertices(8);
  vertices[0]=Point<3>( 0.,       0.,-radius/6.);
  vertices[1]=Point<3>( 0.,       0., radius/6.);
  vertices[2]=Point<3>(radius*sqrt(8./9.),0.,-radius/3.);
  vertices[3]=Point<3>(-radius*sqrt(2./9.), radius*sqrt(2./3.),-radius/3.);
  vertices[4]=Point<3>(-radius*sqrt(2./9.),-radius*sqrt(2./3.),-radius/3.);
  vertices[5]=Point<3>(radius*sqrt(8./9.),0., radius/3.);
  vertices[6]=Point<3>(-radius*sqrt(2./9.), radius*sqrt(2./3.), radius/3.);
  vertices[7]=Point<3>(-radius*sqrt(2./9.),-radius*sqrt(2./3.), radius/3.);

  NumPtr<CellData> cells(5);
  cells[0].allocate(4);
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=2;
  cells[0].vertex(2)=4;
  cells[0].vertex(3)=3;
  cells[0].setMaterialModel(0);

  cells[1].allocate(4);
  cells[1].vertex(0)=1;
  cells[1].vertex(1)=5;
  cells[1].vertex(2)=6;
  cells[1].vertex(3)=7;
  cells[1].setMaterialModel(0);

  cells[2].allocate(6);
  cells[2].vertex(0)=0;
  cells[2].vertex(1)=2;
  cells[2].vertex(2)=3;
  cells[2].vertex(3)=1;
  cells[2].vertex(4)=5;
  cells[2].vertex(5)=6;
  cells[2].setMaterialModel(0);

  cells[3].allocate(6);
  cells[3].vertex(0)=0;
  cells[3].vertex(1)=3;
  cells[3].vertex(2)=4;
  cells[3].vertex(3)=1;
  cells[3].vertex(4)=6;
  cells[3].vertex(5)=7;
  cells[3].setMaterialModel(0);

  cells[4].allocate(6);
  cells[4].vertex(0)=0;
  cells[4].vertex(1)=4;
  cells[4].vertex(2)=2;
  cells[4].vertex(3)=1;
  cells[4].vertex(4)=7;
  cells[4].vertex(5)=5;
  cells[4].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  Point<3> center(0.,0.,0.);
  Boundary *boundary=OPERATOR_NEW BallBoundary(center,radius);;
  BoundaryCondition *boundary_condition=
    OPERATOR_NEW BoundaryCondition(boundary);
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      codimension1_level->setBoundaryCondition(boundary_condition);
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif

  Vector3 low(-radius*1.25,-radius*1.25,-radius*1.25);
  Vector3 high(radius*1.25,radius*1.25,radius*1.25);
  gt.rescale(low,high);
  plotEverything(gt,tess);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  delete boundary_condition; boundary_condition=0;
  delete boundary; boundary=0;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void oneHexahedronInCylinder(VolGraphTool &gt) {
  cout << "\toneHexahedronInCylinder" << endl;
  NumPtr<Point<3> > vertices(8);
  vertices[0]=Point<3>(0.,0.,0.);
  vertices[1]=Point<3>(1.,0.,0.);
  vertices[2]=Point<3>(0.,1.,0.);
  vertices[3]=Point<3>(1.,1.,0.);
  vertices[4]=Point<3>(0.,0.,1.);
  vertices[5]=Point<3>(1.,0.,1.);
  vertices[6]=Point<3>(0.,1.,1.);
  vertices[7]=Point<3>(1.,1.,1.);

  NumPtr<CellData> cells(1);
  cells[0].allocate(8);
  for (int j=0;j<8;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  NumPtr<Boundary*> boundary(3);
  NumPtr<BoundaryCondition*> boundary_condition(3);
  Point<3> axis(0.,0.,1.);
  Point<3> centroid(0.5,0.5,0.5);
  boundary[0]=OPERATOR_NEW CylinderBoundary(axis,centroid,sqrt(0.5));
  boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[2]); // bottom
  boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[4],
    vertices[5],vertices[6]); // top
  for (int f=0;f<3;f++) {
    boundary_condition[f]=OPERATOR_NEW BoundaryCondition(boundary[f]);
#ifdef DEBUG
//  cout << "\tboundary[" << f << "] = " << boundary[f] << endl;
#endif
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
      codimension1_level->getShape();
      Point<3> center=face_element->center(vertices);
      if (center[2]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else if (center[2]>0.9) {
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      } else {
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif
  Vector3 low(-0.25,-0.25,-0.25);
  Vector3 high(1.25,1.25,1.25);
  gt.rescale(low,high);
  plotEverything(gt,tess);
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void onePrismWithPlanes(VolGraphTool &gt) {
  TRACER_CALL(t,"onePrismWithPlanes");
  cout << "\tonePrismWithPlanes" << endl;
  NumPtr<Point<3> > vertices(6);
  vertices[0]=Point<3>(0.,0.,0.);
  vertices[1]=Point<3>(0.,0.,1.);
  vertices[2]=Point<3>(1.,0.,0.);
  vertices[3]=Point<3>(0.,1.,0.);
  vertices[4]=Point<3>(0.,1.,1.);
  vertices[5]=Point<3>(1.,1.,0.);

  NumPtr<CellData> cells(1);
  cells[0].allocate(6);
  for (int j=0;j<6;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  NumPtr<Boundary*> boundary(5);
  NumPtr<BoundaryCondition*> boundary_condition(5);
  Point<3> centroid((vertices[0]+vertices[1]+vertices[2]
                    +vertices[3]+vertices[4]+vertices[5])/6.);
  boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[2]);
  boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[3],
    vertices[4],vertices[5]);
  boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[2],vertices[3]);
  boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[3]);
  boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
    vertices[2],vertices[4]);
  for (int f=0;f<5;f++) {
    boundary_condition[f]=OPERATOR_NEW BoundaryCondition(boundary[f]);
#ifdef DEBUG
//  cout << "\tboundary[" << f << "] = " << boundary[f] << endl;
#endif
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
      codimension1_level->getShape();
      Point<3> center=face_element->center(vertices);
      if (center[1]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      } else if (center[1]>0.9) {
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else if (center[2]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      } else if (center[0]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[3]);
      } else {
        codimension1_level->setBoundaryCondition(boundary_condition[4]);
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif

  Vector3 low(-0.05,-0.05,-0.05);
  Vector3 high(1.05,1.05,1.05);
  gt.rescale(low,high);
  plotEverything(gt,tess);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void oneTetrahedronWithPlanes(VolGraphTool &gt) {
  TRACER_CALL(t,"oneTetrahedronWithPlanes");
  cout << "\toneTetrahedronWithPlanes" << endl;
  NumPtr<Point<3> > vertices(4);
  vertices[0]=Point<3>(0.,0.,0.);
  vertices[1]=Point<3>(1.,0.,0.);
  vertices[2]=Point<3>(0.,1.,0.);
  vertices[3]=Point<3>(0.,0.,1.);

  NumPtr<CellData> cells(1);
  cells[0].allocate(4);
  for (int j=0;j<4;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  NumPtr<Boundary*> boundary(4);
  NumPtr<BoundaryCondition*> boundary_condition(4);
  Point<3> centroid((vertices[0]+vertices[1]+vertices[2]+vertices[3])/4.);
  boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[2],vertices[3]);
  boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[3]);
  boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[2]);
  boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
    vertices[2],vertices[3]);
  for (int f=0;f<4;f++) {
    boundary_condition[f]=OPERATOR_NEW BoundaryCondition(boundary[f]);
#ifdef DEBUG
//  cout << "\tboundary[" << f << "] = " << boundary[f] << endl;
#endif
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
      codimension1_level->getShape();
      Point<3> center=face_element->center(vertices);
      if (center[0]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      } else if (center[1]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else if (center[2]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      } else {
        codimension1_level->setBoundaryCondition(boundary_condition[3]);
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif

  Vector3 low(-0.05,-0.05,-0.05);
  Vector3 high(1.05,1.05,1.05);
  gt.rescale(low,high);
  plotEverything(gt,tess);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<boundary.getNumber();i++) {
    delete boundary[i]; boundary[i]=0;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void onePrismInCylinder(VolGraphTool &gt,int bf) {
  TRACER_CALL(t,"onePrismInCylinder");
  cout << "\tonePrismInCylinder, bf = " << bf << endl;
  REAL radius=2.;
  NumPtr<Point<3> > vertices(6);
  vertices[0]=Point<3>( 2.,       0.,-2.);
  vertices[1]=Point<3>(-1., sqrt(3.),-2.);
  vertices[2]=Point<3>(-1.,-sqrt(3.),-2.);
  vertices[3]=Point<3>( 2.,       0., 2.);
  vertices[4]=Point<3>(-1., sqrt(3.), 2.);
  vertices[5]=Point<3>(-1.,-sqrt(3.), 2.);

  NumPtr<CellData> cells(1);
  cells[0].allocate(6);
  for (int j=0;j<6;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  Point<3> axis(0.,0.,1.);
  Point<3> center(0.,0.,0.);
  NumPtr<Boundary*> pb(5);
  NumPtr<BoundaryCondition*> boundary_condition(5);
  Boundary *cb=OPERATOR_NEW CylinderBoundary(axis,center,radius);;
  pb[0]=OPERATOR_NEW PlaneBoundary(center,vertices[0],vertices[1],
    vertices[2]);
  pb[1]=OPERATOR_NEW PlaneBoundary(center,vertices[3],vertices[4],
    vertices[5]);
  pb[2]=OPERATOR_NEW PlaneBoundary(center,vertices[0],vertices[2],
    vertices[3]);
  pb[3]=OPERATOR_NEW PlaneBoundary(center,vertices[0],vertices[1],
    vertices[3]);
  pb[4]=OPERATOR_NEW PlaneBoundary(center,vertices[1],vertices[2],
    vertices[4]);
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<3> face_center=face_element->center(vertices);
      if (abs(pb[0]->constraint(face_center))<1.e-5) {
        boundary_condition[0]=OPERATOR_NEW BoundaryCondition(pb[0]);
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      } else if (abs(pb[1]->constraint(face_center))<1.e-5) {
        boundary_condition[1]=OPERATOR_NEW BoundaryCondition(pb[1]);
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else if (abs(pb[2]->constraint(face_center))<1.e-5) {
        boundary_condition[2]=
          OPERATOR_NEW BoundaryCondition(bf==2 ? cb : pb[2]);
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      } else if (abs(pb[3]->constraint(face_center))<1.e-5) {
        boundary_condition[3]=
          OPERATOR_NEW BoundaryCondition(bf==3 ? cb : pb[3]);
        codimension1_level->setBoundaryCondition(boundary_condition[3]);
      } else {
        boundary_condition[4]=
          OPERATOR_NEW BoundaryCondition(bf==4 ? cb : pb[4]);
        codimension1_level->setBoundaryCondition(boundary_condition[4]);
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif

  Vector3 low(-radius*1.25,-radius*1.25,-radius*1.25);
  Vector3 high(radius*1.25,radius*1.25,radius*1.25);
  gt.rescale(low,high);
  plotEverything(gt,tess);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
  for (int i=0;i<pb.getNumber();i++) {
    delete pb[i]; pb[i]=0;
  }
  delete cb; cb=0;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void threePrismsInCylinder(VolGraphTool &gt) {
  TRACER_CALL(t,"threePrismsInCylinder");
  cout << "\tthreePrismsInCylinder" << endl;
  REAL radius=2.;
  NumPtr<Point<3> > vertices(8);
  vertices[0]=Point<3>(0.,0.,-radius);
  vertices[1]=Point<3>(0.,0.,radius);
  vertices[2]=Point<3>(radius,       0.,-radius);
  vertices[3]=Point<3>(-radius*0.5, radius*sqrt(0.75),-radius);
  vertices[4]=Point<3>(-radius*0.5,-radius*sqrt(0.75),-radius);
  vertices[5]=Point<3>(radius,       0.,radius);
  vertices[6]=Point<3>(-radius*0.5, radius*sqrt(0.75),radius);
  vertices[7]=Point<3>(-radius*0.5,-radius*sqrt(0.75),radius);

  NumPtr<CellData> cells(3);
  cells[0].allocate(6);
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=2;
  cells[0].vertex(2)=3;
  cells[0].vertex(3)=1;
  cells[0].vertex(4)=5;
  cells[0].vertex(5)=6;
  cells[0].setMaterialModel(0);

  cells[1].allocate(6);
  cells[1].vertex(0)=0;
  cells[1].vertex(1)=3;
  cells[1].vertex(2)=4;
  cells[1].vertex(3)=1;
  cells[1].vertex(4)=6;
  cells[1].vertex(5)=7;
  cells[1].setMaterialModel(0);

  cells[2].allocate(6);
  cells[2].vertex(0)=0;
  cells[2].vertex(1)=4;
  cells[2].vertex(2)=2;
  cells[2].vertex(3)=1;
  cells[2].vertex(4)=7;
  cells[2].vertex(5)=5;
  cells[2].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  Point<3> axis(0.,0.,1.);
  Point<3> center(0.,0.,0.);
  NumPtr<Boundary*> boundary(3);
  NumPtr<BoundaryCondition*> boundary_condition(3);
  boundary[0]=OPERATOR_NEW CylinderBoundary(axis,center,radius);;
  boundary[1]=OPERATOR_NEW PlaneBoundary(center,vertices[0],vertices[2],
    vertices[3]);
  boundary[2]=OPERATOR_NEW PlaneBoundary(center,vertices[1],vertices[5],
    vertices[6]);
  for (int i=0;i<3;i++) {
    boundary_condition[i]=
      OPERATOR_NEW BoundaryCondition(boundary[i]);
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<3> face_center=face_element->center(vertices);
      if (face_center[2]<-radius*0.9) {
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else if (face_center[2]>radius*0.9) {
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      } else {
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif

  Vector3 low(-radius*1.25,-radius*1.25,-radius*1.25);
  Vector3 high(radius*1.25,radius*1.25,radius*1.25);
  gt.rescale(low,high);
  plotEverything(gt,tess);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  for (int i=0;i<3;i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void oneHexahedronInCone(VolGraphTool &gt) {
  TRACER_CALL(t,"oneHexahedronInCone");
  REAL angle=M_PI*0.25; // must have angle<=pi/4 for plotting
  REAL height=1.;
  NumPtr<Point<3> > vertices(8);
  REAL ht=height*tan(angle);
  REAL bottom=0.1;
  vertices[0]=Point<3>( 0.,-ht*bottom, height*bottom );
  vertices[1]=Point<3>( ht*bottom, 0., height*bottom );
  vertices[2]=Point<3>( 0., ht*bottom, height*bottom );
  vertices[3]=Point<3>(-ht*bottom, 0., height*bottom );
  vertices[4]=Point<3>( 0.,-ht, height );
  vertices[5]=Point<3>( ht, 0., height );
  vertices[6]=Point<3>( 0., ht, height );
  vertices[7]=Point<3>(-ht, 0., height );

  NumPtr<CellData> cells(1);
  cells[0].allocate(8);
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=1;
  cells[0].vertex(2)=3;
  cells[0].vertex(3)=2;
  cells[0].vertex(4)=4;
  cells[0].vertex(5)=5;
  cells[0].vertex(6)=7;
  cells[0].vertex(7)=6;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus_boundary_condition in Tessellation::create
  Point<3> axis(0.,0.,1.);
  Point<3> center(0.,0.,0.);
  ConeBoundary *cb=OPERATOR_NEW ConeBoundary(axis,center,angle);
  BoundaryCondition *cbbc=OPERATOR_NEW BoundaryCondition(cb);

  Point<3> pid(0.,0.,height*(1.+bottom)*0.5);
  NumPtr<Boundary*> pb(2);
  pb[0]=OPERATOR_NEW PlaneBoundary(pid,vertices[0],
    vertices[1],vertices[3]);
  pb[1]=OPERATOR_NEW PlaneBoundary(pid,vertices[4],
    vertices[5],vertices[7]);
  NumPtr<BoundaryCondition*> pbbc(2);
  pbbc[0]=OPERATOR_NEW BoundaryCondition(pb[0]);
  pbbc[1]=OPERATOR_NEW BoundaryCondition(pb[1]);

  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<3> face_center=face_element->center(vertices);
      if (abs(pb[0]->constraint(face_center))<1.e-5) {
        codimension1_level->setBoundaryCondition(pbbc[0]);
      } else if (abs(pb[1]->constraint(face_center))<1.e-5) {
        codimension1_level->setBoundaryCondition(pbbc[1]);
      } else {
        codimension1_level->setBoundaryCondition(cbbc);
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//cout << "\tcbbc = " << cbbc << endl;
//cout << "\tpbbc[0] = " << pbbc[0] << endl;
//cout << "\tpbbc[1] = " << pbbc[1] << endl;
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif

  Vector3 low (-ht*1.05,-ht*1.05,-0.05);
  Vector3 high(ht*1.05,ht*1.05,height*1.05);
  gt.rescale(low,high);
  plotEverything(gt,tess);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  delete cbbc;
  delete cb;
  delete pbbc[0];
  delete pbbc[1];
  delete pb[0];
  delete pb[1];
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void threeTetrahedraInCone(VolGraphTool &gt) {
  TRACER_CALL(t,"threeTetrahedraInCone");
  REAL angle=M_PI/6.; // must have angle< pi/4 for plotting
  REAL height=1.;
  REAL ht=height*tan(angle);
  NumPtr<Point<3> > vertices(5);
  vertices[0]=Point<3>(0.,0.,0.);
  vertices[1]=Point<3>(0.,0.,height);
  vertices[2]=Point<3>( ht,0.,height);
  vertices[3]=Point<3>(-ht*0.5, ht*sqrt(0.75),height);
  vertices[4]=Point<3>(-ht*0.5,-ht*sqrt(0.75),height);

  NumPtr<CellData> cells(3);
  cells[0].allocate(4);
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=1;
  cells[0].vertex(2)=2;
  cells[0].vertex(3)=3;
  cells[0].setMaterialModel(0);

  cells[1].allocate(4);
  cells[1].vertex(0)=0;
  cells[1].vertex(1)=1;
  cells[1].vertex(2)=3;
  cells[1].vertex(3)=4;
  cells[1].setMaterialModel(0);

  cells[2].allocate(4);
  cells[2].vertex(0)=0;
  cells[2].vertex(1)=1;
  cells[2].vertex(2)=4;
  cells[2].vertex(3)=2;
  cells[2].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus_boundary_condition in Tessellation::create
  Point<3> axis(0.,0.,1.);
  ConeBoundary *cb=OPERATOR_NEW ConeBoundary(axis,vertices[0],angle);
  BoundaryCondition *cbbc=OPERATOR_NEW BoundaryCondition(cb);

  Point<3> pid(0.,0.,height*0.5);
  PlaneBoundary *pb=OPERATOR_NEW PlaneBoundary(pid,vertices[2],
    vertices[3],vertices[4]);
  BoundaryCondition *pbbc=OPERATOR_NEW BoundaryCondition(pb);

  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<3> face_center=face_element->center(vertices);
      if (abs(pb->constraint(face_center))<1.e-5) {
        codimension1_level->setBoundaryCondition(pbbc);
      } else {
        codimension1_level->setBoundaryCondition(cbbc);
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//cout << "\tcbbc = " << cbbc << endl;
//cout << "\tbbbc = " << bbbc << endl;
//cout << "\tpbbc = " << pbbc << endl;
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif

  Vector3 low (-ht*1.05,-ht*1.05,-0.05);
  Vector3 high(ht*1.05,ht*1.05,height*1.05);
  gt.rescale(low,high);
  plotEverything(gt,tess);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  delete cbbc;
  delete pbbc;
  delete cb;
  delete pb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void TetrahedraInCylinder(VolGraphTool &gt) {
  TRACER_CALL(t,"TetrahedraInCylinder");
  cout << "\tTetrahedraInCylinder" << endl;
  REAL radius=2.;
  NumPtr<Point<3> > vertices(7);
  vertices[0]=Point<3>(0.,0.,0.);
  vertices[1]=Point<3>(radius*0.5, radius*sqrt(0.75),-radius);
  vertices[2]=Point<3>(-radius,       0.,-radius);
  vertices[3]=Point<3>(radius*0.5,-radius*sqrt(0.75),-radius);
  vertices[4]=Point<3>(radius,       0.,radius);
  vertices[5]=Point<3>(-radius*0.5, radius*sqrt(0.75),radius);
  vertices[6]=Point<3>(-radius*0.5,-radius*sqrt(0.75),radius);

  NumPtr<CellData> cells(8);
  cells[0].allocate(4);
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=1;
  cells[0].vertex(2)=3;
  cells[0].vertex(3)=2;
  cells[0].setMaterialModel(0);

  cells[1].allocate(4);
  cells[1].vertex(0)=0;
  cells[1].vertex(1)=1;
  cells[1].vertex(2)=2;
  cells[1].vertex(3)=5;
  cells[1].setMaterialModel(0);

  cells[2].allocate(4);
  cells[2].vertex(0)=0;
  cells[2].vertex(1)=2;
  cells[2].vertex(2)=3;
  cells[2].vertex(3)=6;
  cells[2].setMaterialModel(0);

  cells[3].allocate(4);
  cells[3].vertex(0)=0;
  cells[3].vertex(1)=1;
  cells[3].vertex(2)=4;
  cells[3].vertex(3)=3;
  cells[3].setMaterialModel(0);

  cells[4].allocate(4);
  cells[4].vertex(0)=0;
  cells[4].vertex(1)=1;
  cells[4].vertex(2)=5;
  cells[4].vertex(3)=4;
  cells[4].setMaterialModel(0);

  cells[5].allocate(4);
  cells[5].vertex(0)=0;
  cells[5].vertex(1)=2;
  cells[5].vertex(2)=6;
  cells[5].vertex(3)=5;
  cells[5].setMaterialModel(0);

  cells[6].allocate(4);
  cells[6].vertex(0)=0;
  cells[6].vertex(1)=3;
  cells[6].vertex(2)=4;
  cells[6].vertex(3)=6;
  cells[6].setMaterialModel(0);

  cells[7].allocate(4);
  cells[7].vertex(0)=0;
  cells[7].vertex(1)=4;
  cells[7].vertex(2)=5;
  cells[7].vertex(3)=6;
  cells[7].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  Point<3> axis(0.,0.,1.);
  Point<3> center(0.,0.,0.);
  NumPtr<Boundary*> boundary(3);
  NumPtr<BoundaryCondition*> boundary_condition(3);
  boundary[0]=OPERATOR_NEW CylinderBoundary(axis,center,radius);;
  boundary[1]=OPERATOR_NEW PlaneBoundary(center,vertices[1],vertices[2],
    vertices[3]);
  boundary[2]=OPERATOR_NEW PlaneBoundary(center,vertices[4],vertices[5],
    vertices[6]);
  for (int i=0;i<3;i++) {
    boundary_condition[i]=
      OPERATOR_NEW BoundaryCondition(boundary[i]);
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<3> face_center=face_element->center(vertices);
      if (abs(boundary[1]->constraint(face_center))<1.e-5) {
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else if (abs(boundary[2]->constraint(face_center))<1.e-5) {
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      } else {
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif

  Vector3 low(-radius*1.25,-radius*1.25,-radius*1.25);
  Vector3 high(radius*1.25,radius*1.25,radius*1.25);
  gt.rescale(low,high);
  plotEverything(gt,tess);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  for (int i=0;i<3;i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//GLUquadric *quadric=gluNewQuadric();
//gluQuadricTexture(quadric,GL_FALSE);
//gluQuadricDrawStyle(quadric,GLU_FILL);
//gluQuadricNormals(quadric,GLU_SMOOTH);
//gluQuadricOrientation(quadric,GLU_INSIDE);
//gluCylinder(quadric,radius,radius,1.,20,20);
//gluDeleteQuadric(quadric);
#if (SPACEDIM==3)
void fourPrisms(VolGraphTool &gt) {
//TRACER_CALL(t,"fourPrisms on CylinderBoundary");
  REAL radius=sqrt(.5);

  cout << "\tfour Prisms and Eight Tetrahedra in Cylinder" << endl;
  cout << "\ttop, bottom area = " << M_PI*radius*radius << endl;
  cout << "\tside quadrilateral area = " << 0.5*M_PI*radius << endl;
  cout << "\tside triangle area = " << 0.25*M_PI*radius << endl;

  NumPtr<Point<3> > vertices(12);
  vertices[ 0]=Point<3>(0.,0.,0.);
  vertices[ 1]=Point<3>(radius,0.,0.);
  vertices[ 2]=Point<3>(0.,radius,0.);
  vertices[ 3]=Point<3>(-radius,0.,0.);
  vertices[ 4]=Point<3>(0.,-radius,0.);
  
  vertices[ 5]=Point<3>(0.,0.,1./3.);
  vertices[ 6]=Point<3>(0.,0.,2./3.);

  vertices[ 7]=Point<3>(0.,0.,1.);
  vertices[ 8]=Point<3>(radius,0.,1.);
  vertices[ 9]=Point<3>(0.,radius,1.);
  vertices[10]=Point<3>(-radius,0.,1.);
  vertices[11]=Point<3>(0.,-radius,1.);
  Point<3> axis(0.,0.,1.);
  Point<3> center(0.,0.,0.5);

  NumPtr<CellData> cells(12);
  cells[0].allocate(6);
  cells[0].vertex(0)=1;
  cells[0].vertex(1)=2;
  cells[0].vertex(2)=5;
  cells[0].vertex(3)=8;
  cells[0].vertex(4)=9;
  cells[0].vertex(5)=6;
  cells[0].setMaterialModel(0);

  cells[1].allocate(6);
  cells[1].vertex(0)=1;
  cells[1].vertex(1)=5;
  cells[1].vertex(2)=4;
  cells[1].vertex(3)=8;
  cells[1].vertex(4)=6;
  cells[1].vertex(5)=11;
  cells[1].setMaterialModel(0);

  cells[2].allocate(6);
  cells[2].vertex(0)=2;
  cells[2].vertex(1)=3;
  cells[2].vertex(2)=5;
  cells[2].vertex(3)=9;
  cells[2].vertex(4)=10;
  cells[2].vertex(5)=6;
  cells[2].setMaterialModel(0);

  cells[3].allocate(6);
  cells[3].vertex(0)=3;
  cells[3].vertex(1)=4;
  cells[3].vertex(2)=5;
  cells[3].vertex(3)=10;
  cells[3].vertex(4)=11;
  cells[3].vertex(5)=6;
  cells[3].setMaterialModel(0);

  cells[4].allocate(4);
  cells[4].vertex(0)=0;
  cells[4].vertex(1)=1;
  cells[4].vertex(2)=2;
  cells[4].vertex(3)=5;
  cells[4].setMaterialModel(0);

  cells[5].allocate(4);
  cells[5].vertex(0)=0;
  cells[5].vertex(1)=2;
  cells[5].vertex(2)=3;
  cells[5].vertex(3)=5;
  cells[5].setMaterialModel(0);

  cells[6].allocate(4);
  cells[6].vertex(0)=0;
  cells[6].vertex(1)=3;
  cells[6].vertex(2)=4;
  cells[6].vertex(3)=5;
  cells[6].setMaterialModel(0);

  cells[7].allocate(4);
  cells[7].vertex(0)=0;
  cells[7].vertex(1)=1;
  cells[7].vertex(2)=5;
  cells[7].vertex(3)=4;
  cells[7].setMaterialModel(0);

  cells[8].allocate(4);
  cells[8].vertex(0)=6;
  cells[8].vertex(1)=7;
  cells[8].vertex(2)=8;
  cells[8].vertex(3)=9;
  cells[8].setMaterialModel(0);

  cells[9].allocate(4);
  cells[9].vertex(0)=6;
  cells[9].vertex(1)=7;
  cells[9].vertex(2)=9;
  cells[9].vertex(3)=10;
  cells[9].setMaterialModel(0);

  cells[10].allocate(4);
  cells[10].vertex(0)=6;
  cells[10].vertex(1)=7;
  cells[10].vertex(2)=10;
  cells[10].vertex(3)=11;
  cells[10].setMaterialModel(0);

  cells[11].allocate(4);
  cells[11].vertex(0)=6;
  cells[11].vertex(1)=7;
  cells[11].vertex(2)=11;
  cells[11].vertex(3)=8;
  cells[11].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus_boundary_condition in Tessellation::create
  REAL an=axis*axis;
  NumPtr<Boundary*> boundary(3);
  NumPtr<BoundaryCondition*> boundary_condition(3);

  boundary[0]=OPERATOR_NEW CylinderBoundary(axis,center,radius);
  boundary[1]=OPERATOR_NEW PlaneBoundary(center,
    vertices[0],vertices[1],vertices[2]);
  boundary[2]=OPERATOR_NEW PlaneBoundary(center,
    vertices[7],vertices[8],vertices[9]);
  for (int i=0;i<3;i++) {
    boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<SPACEDIM> fc=face_element->center(vertices);
      if (fc[2]>0.9) {
        codimension1_level->
          setBoundaryCondition(boundary_condition[1]);
      } else if (fc[2]<0.1) {
        codimension1_level->
          setBoundaryCondition(boundary_condition[2]);
      } else {
        codimension1_level->
          setBoundaryCondition(boundary_condition[0]);
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//for (int i=0;i<boundary.getNumber();i++) {
//  boundary[i]->printOn(cout);
//}
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif

  Vector3 low(-radius-0.05,-radius-0.05,-0.05);
  Vector3 high(radius+.05,radius+.05,1.05);
  gt.rescale(low,high);
  plotEverything(gt,tess);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void fourteenTetrahedra(VolGraphTool &gt) {
//TRACER_CALL(t,"fourteen Tetrahedra in CylinderBoundary");
  REAL radius=sqrt(.5);

  cout << "\tEight Tetrahedra in Cylinder" << endl;
  cout << "\ttop, bottom area = " << M_PI*radius*radius << endl;
  cout << "\tside quadrilateral area = " << 0.5*M_PI*radius << endl;
  cout << "\tside triangle area = " << 0.25*M_PI*radius << endl;

  NumPtr<Point<3> > vertices(10);
  vertices[0]=Point<3>(radius,0.,0.);
  vertices[1]=Point<3>(-0.5*radius,radius*sqrt(0.75),0.);
  vertices[2]=Point<3>(-0.5*radius,-radius*sqrt(0.75),0.);
  
  vertices[3]=Point<3>(0.5*radius,-radius*sqrt(0.75),0.5);
  vertices[4]=Point<3>(0.5*radius, radius*sqrt(0.75),0.5);
  vertices[5]=Point<3>(-radius,0.,0.5);
  
  vertices[6]=Point<3>(radius,0.,1.);
  vertices[7]=Point<3>(-0.5*radius,radius*sqrt(0.75),1.);
  vertices[8]=Point<3>(-0.5*radius,-radius*sqrt(0.75),1.);

  vertices[9]=Point<3>(0.,0.,0.5);
  Point<3> axis(0.,0.,1.);
  Point<3> center(0.,0.,0.5);

  NumPtr<CellData> cells(14);
  cells[0].allocate(4);
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=1;
  cells[0].vertex(2)=2;
  cells[0].vertex(3)=9;
  cells[0].setMaterialModel(0);

  cells[1].allocate(4);
  cells[1].vertex(0)=0;
  cells[1].vertex(1)=1;
  cells[1].vertex(2)=9;
  cells[1].vertex(3)=4;
  cells[1].setMaterialModel(0);

  cells[2].allocate(4);
  cells[2].vertex(0)=1;
  cells[2].vertex(1)=2;
  cells[2].vertex(2)=9;
  cells[2].vertex(3)=5;
  cells[2].setMaterialModel(0);

  cells[3].allocate(4);
  cells[3].vertex(0)=0;
  cells[3].vertex(1)=3;
  cells[3].vertex(2)=4;
  cells[3].vertex(3)=9;
  cells[3].setMaterialModel(0);

  cells[4].allocate(4);
  cells[4].vertex(0)=1;
  cells[4].vertex(1)=4;
  cells[4].vertex(2)=5;
  cells[4].vertex(3)=9;
  cells[4].setMaterialModel(0);

  cells[5].allocate(4);
  cells[5].vertex(0)=2;
  cells[5].vertex(1)=3;
  cells[5].vertex(2)=9;
  cells[5].vertex(3)=5;
  cells[5].setMaterialModel(0);

  cells[6].allocate(4);
  cells[6].vertex(0)=0;
  cells[6].vertex(1)=2;
  cells[6].vertex(2)=3;
  cells[6].vertex(3)=9;
  cells[6].setMaterialModel(0);

  cells[7].allocate(4);
  cells[7].vertex(0)=3;
  cells[7].vertex(1)=4;
  cells[7].vertex(2)=9;
  cells[7].vertex(3)=6;
  cells[7].setMaterialModel(0);

  cells[8].allocate(4);
  cells[8].vertex(0)=4;
  cells[8].vertex(1)=5;
  cells[8].vertex(2)=9;
  cells[8].vertex(3)=7;
  cells[8].setMaterialModel(0);

  cells[9].allocate(4);
  cells[9].vertex(0)=3;
  cells[9].vertex(1)=5;
  cells[9].vertex(2)=8;
  cells[9].vertex(3)=9;
  cells[9].setMaterialModel(0);

  cells[10].allocate(4);
  cells[10].vertex(0)=6;
  cells[10].vertex(1)=7;
  cells[10].vertex(2)=9;
  cells[10].vertex(3)=8;
  cells[10].setMaterialModel(0);

  cells[11].allocate(4);
  cells[11].vertex(0)=3;
  cells[11].vertex(1)=6;
  cells[11].vertex(2)=9;
  cells[11].vertex(3)=8;
  cells[11].setMaterialModel(0);

  cells[12].allocate(4);
  cells[12].vertex(0)=4;
  cells[12].vertex(1)=6;
  cells[12].vertex(2)=7;
  cells[12].vertex(3)=9;
  cells[12].setMaterialModel(0);

  cells[13].allocate(4);
  cells[13].vertex(0)=5;
  cells[13].vertex(1)=7;
  cells[13].vertex(2)=8;
  cells[13].vertex(3)=9;
  cells[13].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus_boundary_condition in Tessellation::create
  REAL an=axis*axis;
  NumPtr<Boundary*> boundary(3);
  NumPtr<BoundaryCondition*> boundary_condition(3);

  boundary[0]=OPERATOR_NEW CylinderBoundary(axis,center,radius);
  boundary[1]=OPERATOR_NEW PlaneBoundary(center,
    (vertices[6]+vertices[7]+vertices[8])/3.,vertices[6],vertices[7]);
  boundary[2]=OPERATOR_NEW PlaneBoundary(center,
    (vertices[0]+vertices[1]+vertices[2])/3.,vertices[0],vertices[1]);
  for (int i=0;i<3;i++) {
    boundary_condition[i]=OPERATOR_NEW BoundaryCondition(boundary[i]);
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<SPACEDIM> fc=face_element->center(vertices);
      if (fc[2]>0.9) {
        codimension1_level->
          setBoundaryCondition(boundary_condition[1]);
      } else if (fc[2]<0.1) {
        codimension1_level->
          setBoundaryCondition(boundary_condition[2]);
      } else {
        codimension1_level->
          setBoundaryCondition(boundary_condition[0]);
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//for (int i=0;i<boundary.getNumber();i++) {
//  boundary[i]->printOn(cout);
//}
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif

  Vector3 low(-radius-0.05,-radius-0.05,-0.05);
  Vector3 high(radius+.05,radius+.05,1.05);
  gt.rescale(low,high);
  plotEverything(gt,tess);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  for (int i=0;i<boundary_condition.getNumber();i++) {
    delete boundary_condition[i]; boundary_condition[i]=0;
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void twoPrismsInBall(VolGraphTool &gt) {
  TRACER_CALL(t,"twoPrismsInBall");
  NumPtr<Point<3> > vertices(8);
  Point<3> center(0.,0.,0.);
  REAL radius=sqrt(0.75);
  vertices[0]=Point<3>( 0.5, 0.5,-0.5);
  vertices[1]=Point<3>(-0.5, 0.5,-0.5);
  vertices[2]=Point<3>(-0.5,-0.5,-0.5);
  vertices[3]=Point<3>( 0.5,-0.5,-0.5);
  vertices[4]=Point<3>( 0.5, 0.5, 0.5);
  vertices[5]=Point<3>(-0.5, 0.5, 0.5);
  vertices[6]=Point<3>(-0.5,-0.5, 0.5);
  vertices[7]=Point<3>( 0.5,-0.5, 0.5);

  NumPtr<CellData> cells(2);
  cells[0].allocate(6);
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=3;
  cells[0].vertex(2)=7;
  cells[0].vertex(3)=1;
  cells[0].vertex(4)=2;
  cells[0].vertex(5)=6;
  cells[0].setMaterialModel(0);

  cells[1].allocate(6);
  cells[1].vertex(0)=1;
  cells[1].vertex(1)=5;
  cells[1].vertex(2)=6;
  cells[1].vertex(3)=0;
  cells[1].vertex(4)=4;
  cells[1].vertex(5)=7;
  cells[1].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus_boundary_condition in Tessellation::create
  BallBoundary *b=OPERATOR_NEW BallBoundary(center,radius);
  BoundaryCondition *bc=OPERATOR_NEW BoundaryCondition(b);
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      codimension1_level->setBoundaryCondition(bc);
    }
  }
  tess.synchronize(boundary_data);

  Vector3 low(-radius-0.05,-radius-0.05,-radius-0.05);
  Vector3 high(radius+.05,radius+.05,radius+.05);
  gt.rescale(low,high);
  plotEverything(gt,tess);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  delete bc;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void iceCreamCone(VolGraphTool &gt) { // 2 tetrahedra; cone + ball
  TRACER_CALL(t,"iceCreamCone");
  REAL radius=sqrt(2.);
  NumPtr<Point<3> > vertices(5);
  vertices[0]=Point<3>( radius, 0., 0. );
  vertices[1]=Point<3>(-1.,-1.,0.);
  vertices[2]=Point<3>(-1., 1.,0.);
  vertices[3]=Point<3>( 0., 0.,-radius);
  vertices[4]=Point<3>( 0., 0., radius);

  NumPtr<CellData> cells(2);
  cells[0].allocate(4);
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=1;
  cells[0].vertex(2)=4;
  cells[0].vertex(3)=2;
  cells[0].setMaterialModel(0);

  cells[1].allocate(4);
  cells[1].vertex(0)=0;
  cells[1].vertex(1)=1;
  cells[1].vertex(2)=2;
  cells[1].vertex(3)=3;
  cells[1].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus_boundary_condition in Tessellation::create
  Point<3> axis(0.,0.,1.);
  REAL angle=acos(sqrt(0.5));
  ConeBoundary *cb=OPERATOR_NEW ConeBoundary(axis,vertices[3],angle);
  BoundaryCondition *cbbc=OPERATOR_NEW BoundaryCondition(cb);

  Point<3> center(0.,0.,0.);
  BallBoundary *bb=OPERATOR_NEW BallBoundary(center,radius);
  BoundaryCondition *bbbc=OPERATOR_NEW BoundaryCondition(bb);

  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<3> face_center=face_element->center(vertices);
      if (face_center[2]<-0.) {
        codimension1_level->setBoundaryCondition(cbbc);
      } else {
        codimension1_level->setBoundaryCondition(bbbc);
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//cout << "\tcbbc = " << cbbc << endl;
//cout << "\tbbbc = " << bbbc << endl;
//cout << "\tpbbc = " << pbbc << endl;
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif

  Vector3 low (-radius*1.1,-radius*1.1,-radius*1.1);
  Vector3 high( radius*1.1, radius*1.1, radius*1.1);
  gt.rescale(low,high);
  plotEverything(gt,tess);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  delete cbbc;
  delete bbbc;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void silo(VolGraphTool &gt) { // 2 prisms + 2 tetrahedra; cylinder + ball
  TRACER_CALL(t,"silo");
  REAL radius=sqrt(0.5);
  NumPtr<Point<3> > vertices(9);
  vertices[0]=Point<3>( 0.5, 0.5,-0.5);
  vertices[1]=Point<3>(-0.5, 0.5,-0.5);
  vertices[2]=Point<3>(-0.5,-0.5,-0.5);
  vertices[3]=Point<3>( 0.5,-0.5,-0.5);
  vertices[4]=Point<3>( 0.5, 0.5, 0.5);
  vertices[5]=Point<3>(-0.5, 0.5, 0.5);
  vertices[6]=Point<3>(-0.5,-0.5, 0.5);
  vertices[7]=Point<3>( 0.5,-0.5, 0.5);
  vertices[8]=Point<3>( 0.5+radius, 0.0, 0.0);

  NumPtr<CellData> cells(4);
  cells[0].allocate(6);
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=3;
  cells[0].vertex(2)=7;
  cells[0].vertex(3)=1;
  cells[0].vertex(4)=2;
  cells[0].vertex(5)=6;
  cells[0].setMaterialModel(0);

  cells[1].allocate(6);
  cells[1].vertex(0)=1;
  cells[1].vertex(1)=5;
  cells[1].vertex(2)=6;
  cells[1].vertex(3)=0;
  cells[1].vertex(4)=4;
  cells[1].vertex(5)=7;
  cells[1].setMaterialModel(0);

  cells[2].allocate(4);
  cells[2].vertex(0)=0;
  cells[2].vertex(1)=3;
  cells[2].vertex(2)=8;
  cells[2].vertex(3)=7;
  cells[2].setMaterialModel(0);

  cells[3].allocate(4);
  cells[3].vertex(0)=0;
  cells[3].vertex(1)=4;
  cells[3].vertex(2)=7;
  cells[3].vertex(3)=8;
  cells[3].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus_boundary_condition in Tessellation::create
  Point<3> axis(1.,0.,0.);
  Point<3> center(0.,0.,0.);
  CylinderBoundary *cb=OPERATOR_NEW CylinderBoundary(axis,center,radius);
  BoundaryCondition *cbbc=OPERATOR_NEW BoundaryCondition(cb);
  Point<3> ball_center(0.5,0.,0.);
  BallBoundary *bb=OPERATOR_NEW BallBoundary(ball_center,radius);
  BoundaryCondition *bbbc=OPERATOR_NEW BoundaryCondition(bb);
  PlaneBoundary *pb=OPERATOR_NEW PlaneBoundary(center,vertices[1],
    vertices[5],vertices[2]);
  BoundaryCondition *pbbc=OPERATOR_NEW BoundaryCondition(pb);
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<3> face_center=face_element->center(vertices);
      if (face_center[0]<-0.45) {
        codimension1_level->setBoundaryCondition(pbbc);
      } else if (face_center[0]>0.5) {
        codimension1_level->setBoundaryCondition(bbbc);
      } else {
        codimension1_level->setBoundaryCondition(cbbc);
      }
    }
  }
  tess.synchronize(boundary_data);
#ifdef DEBUG
//cout << "\tcbbc = " << cbbc << endl;
//cout << "\tbbbc = " << bbbc << endl;
//cout << "\tpbbc = " << pbbc << endl;
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif

  Vector3 low(-0.75,-0.75,-0.75);
  Vector3 high(0.55+radius,0.75,0.75);
  gt.rescale(low,high);
  plotEverything(gt,tess);
  gt.eventLoop();
#ifdef DEBUG
//tess.printOn(cout);
#endif
  delete boundary_data;
  delete cbbc;
  delete bbbc;
  delete pbbc;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

int main(int argc,char* argv[]) {
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
#ifdef USE_GTK
    GTKWindow::gtkInit(argc,argv);
#endif
#ifdef DEBUG
    void setTraps();
#endif

#if (SPACEDIM==2)
    for (int bf=-1;bf<3;bf++) {
      oneTriangle(bf);
    }
    twoTriangles(); // BallBoundary
#endif
#if (SPACEDIM==3)
    Vector3 low(-0.05,-0.05,-0.05);
    Vector3 high(1.05,1.05,1.05);
    Palette pal;
    VolGraphTool gt("testBoundary",low,high,pal,TRUE,0.5);
    gt.newPage();
    gt.expose();

    oneHexahedronWithPlanes(gt);
    onePrismWithPlanes(gt);
    oneTetrahedronWithPlanes(gt);

    oneHexahedronInBall(gt);
    for (int bf=-1;bf<4;bf++) {
      oneTetrahedronInBallFace(gt,bf);
    }
    for (int bl=0;bl<6;bl++) {
      oneTetrahedronInBallLine(gt,bl);
    }
//
    for (int bf=-1;bf<5;bf++) 
//  int bf=2;
    {
      onePrismInBallFace(gt,bf);
    }
//
//
    for (int bl=0;bl<9;bl++) 
//  int bl=2;
    {
      onePrismInBallLine(gt,bl);
    }
//
    fourTetrahedraInBall(gt);
    threePrismsAndTwoTetrahedraInBall(gt);

    oneHexahedronInCylinder(gt);
//
    for (int bf=0;bf<5;bf++) 
//  int bf=2;
    {
      onePrismInCylinder(gt,bf);
    }
//
    threePrismsInCylinder(gt);
//  TetrahedraInCylinder(gt); // fails Lenoir test

    oneHexahedronInCone(gt);
    threeTetrahedraInCone(gt);

//  fourteenTetrahedra(gt); // CylinderBoundary + PlaneBoundary's
//  fourPrisms(gt); // CylinderBoundary + PlaneBoundary's
//  twoPrismsInBall(gt); // Ball
//  iceCreamCone(gt); // Cone + Ball
//  silo(gt); // Cylinder + Ball + Plane
#endif
  }
  return EXIT_SUCCESS;
}

template class NumPtrBase<Tensor<1,SPACEDIM>,int>;
template class NumPtr<Tensor<1,SPACEDIM> >;
template class _Rb_tree<
  int, 
  pair<int const, Quadrature<2>*>, 
  _Select1st<pair<int const, Quadrature<2>*> >, 
  less<int>, 
  allocator<pair<int const, Quadrature<2>*> >
>;
