// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "Boundary.H"
#include "FEValues.H"
#include "HcurlFEValues.H"
#include "HdivFEValues.H"
#include "MemoryDebugger.H"
#include "SetTraps.H"
#include "Tracer.H"
//#include "XYGraphTool.H"
#ifdef USE_GTK
//#include <gtk/gtkmain.h>
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void richardsonShapeDerivs(int nmax,const Tessellation &tess,
const LEVEL_CLASS *level,const Mapping &mapping,
const MappingDataBase *md,FiniteElement &finite_element,
const SHAPE_CLASS *e,const Point<SPACEDIM> &cp,const Point<SPACEDIM> &x,
const Vector &dof_values,REAL f,Tensor<1,SPACEDIM> &fp) {
//TRACER_CALL(t,"richardsonShapeDerivs");
#ifdef DEBUG
//e->printOn(cout);
//cout << "\tcp = " << cp << endl;
//cout << "\tx = " << x << endl;
//for (int s=0;s<dof_values.size();s++) {
//  cout << "\tdof_values[" << s << "] = " << dof_values[s] << endl;
//}
//cout << "\tf = " << f << endl;
#endif
  UPDATE_FLAGS update_flags=UPDATE_GRADIENTS | UPDATE_Q_POINTS;
  nmax=max(1,nmax);
  for (int i=0;i<SPACEDIM;i++) {
    NumPtr<REAL> table(nmax);
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      Point<SPACEDIM> xnew; xnew=x; xnew[i]+=h;
      NumPtr<Point<SPACEDIM> > cpnew(1);
      cpnew[0]=cp;
      mapping.canonicalPoint(e,tess,md,xnew,cpnew[0]);
#ifdef DEBUG
//    cout << "\n\txnew = " << xnew << endl;
//    cout << "\tcpnew = " << cpnew[0] << endl;
#endif
      Quadrature<SPACEDIM> q(cpnew);
      map<int,Quadrature<SPACEDIM> > qmap;
      qmap[e->numberVertices()]=q;
      FEValues fev(mapping,finite_element,qmap,UPDATE_VALUES);
      fev.mapShapes(tess,e,level);
      int dpc=fev.dofsPerCell(e);
      NumPtr<REAL> values;
      fev.getFunctionValues(dof_values,e,values);
#ifdef DEBUG
//    cout << "\tvalues = " << values[0] << endl;
#endif
      table[nmax-n]=(values[0]-f)/h;
      REAL p=2.;
      for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
        table[m]=table[m-1]+(table[m-1]-table[m])/(p-1.);
      }
      fp[i]=table[nmax-1];
#ifdef DEBUG
//    cout << "\tfp = " << fp[i] << endl;
#endif
    }
  }
#ifdef DEBUG
//cout << "\tfp = " << fp << endl;
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void richardsonHdivShapeDerivs(int nmax,const Tessellation &tess,
const LEVEL_CLASS *level,const Mapping &mapping,
const MappingDataBase *md,HdivFiniteElement &finite_element,
const SHAPE_CLASS *e,const Point<SPACEDIM> &cp,const Point<SPACEDIM> &x,
const Vector &dof_values,const Tensor<1,SPACEDIM> &f,
Tensor<2,SPACEDIM> &fp) {
//TRACER_CALL(t,"richardsonShapeDerivs");
#ifdef DEBUG
//e->printOn(cout);
//cout << "\tcp = " << cp << endl;
//cout << "\tx = " << x << endl;
//for (int s=0;s<dof_values.size();s++) {
//  cout << "\tdof_values[" << s << "] = " << dof_values[s] << endl;
//}
//cout << "\tf = " << f << endl;
#endif
  UPDATE_FLAGS update_flags=UPDATE_DETERMINANT | UPDATE_VALUES |
    UPDATE_Q_POINTS | UPDATE_MAPPING_JACOBIAN_INVERSE;
  nmax=max(1,nmax);
  for (int j=0;j<SPACEDIM;j++) {
    NumPtr<Tensor<1,SPACEDIM> > table(nmax);
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      Point<SPACEDIM> xnew; xnew=x; xnew[j]+=h;
      NumPtr<Point<SPACEDIM> > cpnew(1);
      cpnew[0]=cp;
      mapping.canonicalPoint(e,tess,md,xnew,cpnew[0]);
#ifdef DEBUG
//    cout << "\n\txnew = " << xnew << endl;
//    cout << "\tcpnew = " << cpnew[0] << endl;
#endif
      Quadrature<SPACEDIM> q(cpnew);
      map<int,Quadrature<SPACEDIM> > qmap;
      qmap[e->numberVertices()]=q;
      HdivFEValues fev(mapping,finite_element,qmap,update_flags);
      fev.mapShapes(tess,e,level);
      int dpc=fev.dofsPerCell(e);
      NumPtr<Tensor<1,SPACEDIM> > values;
      fev.getFunctionValues(dof_values,e,values);
#ifdef DEBUG
//      cout << "\tvalues = " << values[0] << endl;
#endif
      for (int i=0;i<SPACEDIM;i++) {
        table[nmax-n][i]=(values[0][i]-f[i])/h;
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          table[m][i]=table[m-1][i]+(table[m-1][i]-table[m][i])/(p-1.);
        }
        fp[i][j]=table[nmax-1][i];
#ifdef DEBUG
//      cout << "\tfp = " << fp[i] << endl;
#endif
      }
    }
  }
#ifdef DEBUG
//cout << "\tfp = " << fp << endl;
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//#if (SPACEDIM>=2)
void richardsonHcurlShapeDerivs(int nmax,const Tessellation &tess,
const LEVEL_CLASS *level,const Mapping &mapping,
const MappingDataBase *md,HcurlFiniteElement &finite_element,
const SHAPE_CLASS *e,const Point<SPACEDIM> &cp,const Point<SPACEDIM> &x,
const Vector &dof_values,const Tensor<1,SPACEDIM> &f,
Tensor<2,SPACEDIM> &fp) {
//TRACER_CALL(t,"richardsonShapeDerivs");
# ifdef DEBUG
//e->printOn(cout);
//cout << "\tcp = " << cp << endl;
//cout << "\tx = " << x << endl;
//for (int s=0;s<dof_values.size();s++) {
//  cout << "\tdof_values[" << s << "] = " << dof_values[s] << endl;
//}
//cout << "\tf = " << f << endl;
# endif
  UPDATE_FLAGS update_flags=UPDATE_DETERMINANT | UPDATE_VALUES |
    UPDATE_Q_POINTS | UPDATE_MAPPING_JACOBIAN_INVERSE;
  nmax=max(1,nmax);
  for (int j=0;j<SPACEDIM;j++) {
    NumPtr<Tensor<1,SPACEDIM> > table(nmax);
    REAL h=0.1;
    for (int n=1;n<=nmax;n++,h*=0.5) {
      Point<SPACEDIM> xnew; xnew=x; xnew[j]+=h;
      NumPtr<Point<SPACEDIM> > cpnew(1);
      cpnew[0]=cp;
      mapping.canonicalPoint(e,tess,md,xnew,cpnew[0]);
# ifdef DEBUG
//    cout << "\n\txnew = " << xnew << endl;
//    cout << "\tcpnew = " << cpnew[0] << endl;
# endif
      Quadrature<SPACEDIM> q(cpnew);
      map<int,Quadrature<SPACEDIM> > qmap;
      qmap[e->numberVertices()]=q;
      HcurlFEValues fev(mapping,finite_element,qmap,update_flags);
      fev.mapShapes(tess,e,level);
      int dpc=fev.dofsPerCell(e);
      NumPtr<Tensor<1,SPACEDIM> > values;
      fev.getFunctionValues(dof_values,e,values);
# ifdef DEBUG
//      cout << "\tvalues = " << values[0] << endl;
# endif
      for (int i=0;i<SPACEDIM;i++) {
        table[nmax-n][i]=(values[0][i]-f[i])/h;
        REAL p=2.;
        for (int m=nmax+1-n;m<nmax;m++,p*=2.) {
          table[m][i]=table[m-1][i]+(table[m-1][i]-table[m][i])/(p-1.);
        }
        fp[i][j]=table[nmax-1][i];
# ifdef DEBUG
//      cout << "\tfp = " << fp[i] << endl;
# endif
      }
    }
  }
# ifdef DEBUG
//cout << "\tfp = " << fp << endl;
# endif
}
//#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void debugDerivatives(Tessellation &tess,Mapping &mapping,
FiniteElement &finite_element,
map<int,Quadrature<SPACEDIM> > &quadrature_map) {
//TRACER_CALL(t,"debugDerivatives");
  UPDATE_FLAGS update_flags=
    UPDATE_VALUES | UPDATE_GRADIENTS | UPDATE_Q_POINTS;
  FEValues fev(mapping,finite_element,quadrature_map,update_flags);
#ifdef DEBUG
//cout << "\tupdate_flags = " << update_flags << endl;
//cout << "\t&fev = " << &fev << endl;
#endif
  for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
//  TRACER_CALL(t,"debugDerivatives loop");
    const SHAPE_CLASS *e=tess.presentShape();
    const LEVEL_CLASS *level=tess.presentLevel();
    fev.mapShapes(tess,e,level);
    const MappingDataBase *md=fev.mappingData(e);
    Point<SPACEDIM> element_point=md->elementPoint();
    const Quadrature<SPACEDIM> q=fev.getQuadrature(e);
    Point<SPACEDIM> q_point=q.point(0);
#ifdef DEBUG
//  md->printOn(cout);
//  cout << "\tq_point = " << q_point << endl;
//  cout << "\telement_point = " << element_point << endl;
#endif

    int dpc=fev.dofsPerCell(e);
    Vector dof_values(dpc);
    for (int s=0;s<dpc;s++) {
      dof_values=0.;
      dof_values[s]=1.;
      NumPtr<REAL> values;
      fev.getFunctionValues(dof_values,e,values);
      NumPtr<Tensor<1,SPACEDIM> > derivatives;
      fev.getFunctionDerivatives(dof_values,e,derivatives);
#ifdef DEBUG
//    cout << "\tvalues[" << s << "] = " << values[0] << endl;
//    cout << "\tderivatives[" << s << "] = " << derivatives[0] << endl;
#endif
      Tensor<1,SPACEDIM> diff;
//    int nmax=finite_element.getDegree()+SPACEDIM;
      int nmax=9;
      richardsonShapeDerivs(nmax,tess,level,mapping,md,finite_element,
        e,q_point,element_point,dof_values,values[0],diff);
      for (int d=0;d<SPACEDIM;d++) {
        REAL exact=derivatives[0][d];
        REAL approx=diff[d];
        REAL error=abs(exact-approx);
        if (abs(exact)>0.) error/=abs(exact);
        if (error>2097152.*DBL_EPSILON) {
          if (abs(exact)>2097152.*DBL_EPSILON) {
            cout << "\tfunctionDerivative(" << s << "," << d << ") = "
                 << exact << " " << approx << " error = "
                 << error << " " << error/DBL_EPSILON << endl;
            CHECK_TEST(error<=2097152.*DBL_EPSILON);
          }
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void debugDerivativesForMapping(Tessellation &tess,Mapping &mapping,
map<int,Quadrature<SPACEDIM> > &quadrature_map) {
  TRACER_CALL(t,"debugDerivativesForMapping");
  for (int degree_f=1;degree_f<=D_CHOOSE(6,6,4);degree_f++) {
    cout << "\tdegree_f = " << degree_f << endl;
    {
//    TRACER_CALL(t,"FE_L")
      FE_L finite_element(degree_f);
      debugDerivatives(tess,mapping,finite_element,quadrature_map);
    }
    {
//    TRACER_CALL(t,"FE_H")
      FE_H finite_element(degree_f);
      debugDerivatives(tess,mapping,finite_element,quadrature_map);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void debugHdivDerivatives(Tessellation &tess,Mapping &mapping,
HdivFiniteElement &finite_element,
map<int,Quadrature<SPACEDIM> > &quadrature_map) {
//TRACER_CALL(t,"debugDerivatives");
  UPDATE_FLAGS update_flags= UPDATE_DETERMINANT |
    UPDATE_MAPPING_SECOND_DERIVATIVE | UPDATE_VALUES | UPDATE_GRADIENTS
    | UPDATE_DIVERGENCE | UPDATE_Q_POINTS;
  HdivFEValues fev(mapping,finite_element,quadrature_map,update_flags);
#ifdef DEBUG
//cout << "\tupdate_flags = " << update_flags << endl;
//cout << "\t&fev = " << &fev << endl;
#endif
  for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
//  TRACER_CALL(t,"debugDerivatives loop");
    const SHAPE_CLASS *e=tess.presentShape();
    const LEVEL_CLASS *level=tess.presentLevel();
    fev.mapShapes(tess,e,level);
    const MappingDataBase *md=fev.mappingData(e);
    Point<SPACEDIM> element_point=md->elementPoint();
    const Quadrature<SPACEDIM> q=fev.getQuadrature(e);
    Point<SPACEDIM> q_point=q.point(0);
#ifdef DEBUG
//  md->printOn(cout);
//  cout << "\tq_point = " << q_point << endl;
//  cout << "\telement_point = " << element_point << endl;
#endif

    int dpc=fev.dofsPerCell(e);
    Vector dof_values(dpc);
    for (int s=0;s<dpc;s++) {
      dof_values=0.;
      dof_values[s]=1.;
      NumPtr<Tensor<1,SPACEDIM> > values;
      fev.getFunctionValues(dof_values,e,values);
      NumPtr<Tensor<2,SPACEDIM> > derivatives;
      fev.getFunctionDerivatives(dof_values,e,derivatives);
#ifdef DEBUG
//    cout << "\tvalues[" << s << "] = " << values[0] << endl;
//    cout << "\tderivatives[" << s << "] = " << derivatives[0] << endl;
#endif
      Tensor<2,SPACEDIM> diff;
//    int nmax=finite_element.getDegree()+SPACEDIM;
      int nmax=9;
      richardsonHdivShapeDerivs(nmax,tess,level,mapping,md,finite_element,
        e,q_point,element_point,dof_values,values[0],diff);
      for (int i=0;i<SPACEDIM;i++) {
        for (int j=0;j<SPACEDIM;j++) {
          REAL exact=derivatives[0][i][j];
          REAL approx=diff[i][j];
          REAL error=abs(exact-approx);
          if (abs(exact)>0.) error/=abs(exact);
          if (error>134217728.*DBL_EPSILON) {
            if (abs(exact)>134217728.*DBL_EPSILON) {
              cout << "\tfunctionDerivative(" << s << "," << i << ","
                   << j << ") = " << exact << " " << approx
                   << " error = " << error << " " << error/DBL_EPSILON
                   << endl;
              CHECK_TEST(error<=134217728.*DBL_EPSILON);
            }
          }
        }
      }
      NumPtr<REAL> divergence;
      fev.getFunctionDivergence(dof_values,e,divergence);
      REAL exact=divergence[0];
      REAL approx=0.;
      for (int d=0;d<SPACEDIM;d++) approx+=diff[d][d];
      REAL error=abs(exact-approx);
      if (abs(exact)>0.) error/=abs(exact);
      if (error>134217728.*DBL_EPSILON) {
        if (abs(exact)>134217728.*DBL_EPSILON) {
          cout << "\tfunctionDivergence(" << s << ") = " << exact << " "
               << approx << " error = " << error << " "
               << error/DBL_EPSILON << endl;
          CHECK_TEST(error<=134217728.*DBL_EPSILON);
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//#if (SPACEDIM>=2)
void debugHcurlDerivatives(Tessellation &tess,Mapping &mapping,
HcurlFiniteElement &finite_element,
map<int,Quadrature<SPACEDIM> > &quadrature_map) {
//TRACER_CALL(t,"debugDerivatives");
  UPDATE_FLAGS update_flags= UPDATE_DETERMINANT |
    UPDATE_MAPPING_SECOND_DERIVATIVE | UPDATE_VALUES | UPDATE_GRADIENTS 
    | UPDATE_CURL | UPDATE_Q_POINTS;
  HcurlFEValues fev(mapping,finite_element,quadrature_map,update_flags);
# ifdef DEBUG
//finite_element.printOn(cout);
//cout << "\tupdate_flags = " << update_flags << endl;
//cout << "\t&fev = " << &fev << endl;
# endif
  for (bool doit=tess.firstShape();doit;doit=tess.nextShape()) {
//  TRACER_CALL(t,"debugDerivatives loop");
    const SHAPE_CLASS *e=tess.presentShape();
# ifdef DEBUG
//  cout << "\te->numberVertices = " << e->numberVertices() << endl;
//  fev.printFEData(cout);
#  endif
    const LEVEL_CLASS *level=tess.presentLevel();
    fev.mapShapes(tess,e,level);
    const MappingDataBase *md=fev.mappingData(e);
    Point<SPACEDIM> element_point=md->elementPoint();
    const Quadrature<SPACEDIM> q=fev.getQuadrature(e);
    Point<SPACEDIM> q_point=q.point(0);
# ifdef DEBUG
//  e->printOn(cout);
//  md->printOn(cout);
//  cout << "\tq_point = " << q_point << endl;
//  cout << "\telement_point = " << element_point << endl;
# endif

    int dpc=fev.dofsPerCell(e);
    Vector dof_values(dpc);
    for (int s=0;s<dpc;s++) {
      dof_values=0.;
      dof_values[s]=1.;
      NumPtr<Tensor<1,SPACEDIM> > values;
      fev.getFunctionValues(dof_values,e,values);
      NumPtr<Tensor<2,SPACEDIM> > derivatives;
      fev.getFunctionDerivatives(dof_values,e,derivatives);
# ifdef DEBUG
//    cout << "\tvalues[" << s << "] = " << values[0] << endl;
//    cout << "\tderivatives[" << s << "] = " << derivatives[0] << endl;
# endif
      Tensor<2,SPACEDIM> diff;
//    int nmax=finite_element.getDegree()+SPACEDIM;
      int nmax=9;
      richardsonHcurlShapeDerivs(nmax,tess,level,mapping,md,finite_element,
        e,q_point,element_point,dof_values,values[0],diff);
      for (int i=0;i<SPACEDIM;i++) {
        for (int j=0;j<SPACEDIM;j++) {
          REAL exact=derivatives[0][i][j];
          REAL approx=diff[i][j];
          REAL error=abs(exact-approx);
          if (abs(exact)>0.) error/=abs(exact);
          if (error>536870912.*DBL_EPSILON) {
            if (abs(exact)>536870912.*DBL_EPSILON) {
              cout << "\tfunctionDerivative(" << s << "," << i << ","
                   << j << ") = " << exact << " " << approx
                   << " error = " << error << " " << error/DBL_EPSILON
                   << endl;
              CHECK_TEST(error<=536870912.*DBL_EPSILON);
            }
          }
        }
      }
# if (SPACEDIM==1)
      NumPtr<REAL> curl;
      fev.getFunctionCurl(dof_values,e,curl);
      REAL exact=curl[0];
      REAL approx=diff[0][0];
      REAL error=abs(exact-approx);
      if (abs(exact)>0.) error/=abs(exact);
      if (error>1024.*DBL_EPSILON) {
        if (abs(exact)>1024.*DBL_EPSILON) {
          cout << "\tfunctionCurl(" << s << ") = " << exact << " "
               << approx << " error = " << error << " "
               << error/DBL_EPSILON << endl;
          CHECK_TEST(error<=1024.*DBL_EPSILON);
        }
      }
# elif (SPACEDIM==2)
      NumPtr<REAL> curl;
      fev.getFunctionCurl(dof_values,e,curl);
      REAL exact=curl[0];
      REAL approx=diff[1][0]-diff[0][1];
      REAL error=abs(exact-approx);
      if (abs(exact)>0.) error/=abs(exact);
      if (error>134217728.*DBL_EPSILON) {
        if (abs(exact)>134217728.*DBL_EPSILON) {
          cout << "\tfunctionCurl(" << s << ") = " << exact << " "
               << approx << " error = " << error << " "
               << error/DBL_EPSILON << endl;
          CHECK_TEST(error<=134217728.*DBL_EPSILON);
        }
      }
# else
      NumPtr<Tensor<1,3> > curl;
      fev.getFunctionCurl(dof_values,e,curl);
      Tensor<1,3> exact=curl[0];
      Tensor<1,3> approx;
        approx[0]=diff[2][1]-diff[1][2];
        approx[1]=diff[0][2]-diff[2][0];
        approx[2]=diff[1][0]-diff[0][1];
      REAL error=(exact-approx)*(exact-approx);
      if (exact*exact>0.) error/=(exact*exact);
      if (error>134217728.*DBL_EPSILON) {
        if (sqrt(exact*exact)>134217728.*DBL_EPSILON) {
          cout << "\tfunctionCurl(" << s << ") = " << exact << " "
               << approx << " error = " << error << " "
               << error/DBL_EPSILON << " "
               << sqrt(exact*exact)/DBL_EPSILON << endl;
          CHECK_TEST(error<=134217728.*DBL_EPSILON);
        }
      }
# endif
    }
  }
}
//#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void debugHdivDerivativesForMapping(Tessellation &tess,Mapping &mapping,
map<int,Quadrature<SPACEDIM> > &quadrature_map) {
  TRACER_CALL(t,"debugHdivDerivativesForMapping");
  for (int degree_f=1;degree_f<=D_CHOOSE(6,6,4);degree_f++) {
    cout << "\tdegree_f = " << degree_f << endl;
    {
//    TRACER_CALL(t,"RTFiniteElement")
      RTFiniteElement finite_element(degree_f);
      debugHdivDerivatives(tess,mapping,finite_element,quadrature_map);
    }
    {
//    TRACER_CALL(t,"BDFMFiniteElement")
      BDFMFiniteElement finite_element(degree_f);
      debugHdivDerivatives(tess,mapping,finite_element,quadrature_map);
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//#if (SPACEDIM>=2)
void debugHcurlDerivativesForMapping(Tessellation &tess,Mapping &mapping,
map<int,Quadrature<SPACEDIM> > &quadrature_map) {
  TRACER_CALL(t,"debugHcurlDerivativesForMapping");
  for (int degree_f=0;degree_f<=6;degree_f++)
  {
//  int degree_f=2;
    cout << "\tdegree_f = " << degree_f << endl;
    {
//    TRACER_CALL(t,"Nedelec1FiniteElement")
      Nedelec1FiniteElement finite_element(degree_f);
      debugHcurlDerivatives(tess,mapping,finite_element,quadrature_map);
    }
    if (degree_f>=1) {
//    TRACER_CALL(t,"Nedelec2FiniteElement")
      Nedelec2FiniteElement finite_element(degree_f);
      debugHcurlDerivatives(tess,mapping,finite_element,quadrature_map);
    }
  }
}
//#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int main(int argc,char* argv[]) {
#ifdef DEBUG
//setTraps();
#endif
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
#ifdef USE_GTK
//  GTKWindow::gtkInit(argc,argv);
#endif
    cout.precision(16);
    REAL r=drand48(); // flush the tiny first random number
    NumPtr<Point<SPACEDIM> > points(1);
    Point<SPACEDIM> &p=points[0];
    for (int d=0;d<SPACEDIM;d++) p[d]=drand48();
//  for (int d=0;d<SPACEDIM;d++) p[d]=0.;
#if (SPACEDIM>=2)
    p[1]*=1.-p[0];
#endif
#if (SPACEDIM==3)
    p[2]*=1.-p[0]-p[1];
#endif
    Quadrature<SPACEDIM> q(points);
    map<int,Quadrature<SPACEDIM> > quadrature_map;
#if (SPACEDIM==1)
    Point<1> p0(0.);
    Point<1> p1(1.);
    NumPtr<Point<1> > vertices(2);
    vertices[0]=p0;
    vertices[1]=p1;

    NumPtr<CellData> cells(1);

    cells[0].allocate(2);
    for (int i=0;i<2;i++) cells[0].vertex(i)=i;
    cells[0].setMaterialModel(0);

    Tessellation tess;
    BoundaryData *boundary_data=tess.create(vertices,cells);
    quadrature_map[2]=q;
#endif
#if (SPACEDIM==2)
    NumPtr<Point<2> > vertices(5);
    vertices[0]=Point<2>(0.,0.);
    vertices[1]=Point<2>(.75,.25);
    vertices[2]=Point<2>(0.,.5);
    vertices[3]=Point<2>(1.,.5);
    vertices[4]=Point<2>(.5,1.);

    NumPtr<CellData> cells(2);
    for (int i=0;i<2;i++) {
      cells[i].setMaterialModel(0);
    }
    cells[0].allocate(4);
    cells[0].vertex(0)=0;
    cells[0].vertex(1)=1;
    cells[0].vertex(2)=2;
    cells[0].vertex(3)=3;

    cells[1].allocate(3);
    cells[1].vertex(0)=2;
    cells[1].vertex(1)=3;
    cells[1].vertex(2)=4;

    Tessellation tess;
    BoundaryData *boundary_data=tess.create(vertices,cells);
    Point<2> centroid((vertices[0]+vertices[1]
                      +vertices[2]+vertices[3])*0.25);
    NumPtr<BoundaryCondition*> boundary_condition(5);
    NumPtr<Boundary*> boundary(5);
    boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
      vertices[2]);
    boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
      vertices[3]);
    boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
      vertices[1]);
    boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[2],
      vertices[4]);
    boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,vertices[3],
      vertices[4]);
    for (int f=0;f<5;f++) {
      boundary_condition[f]=
        OPERATOR_NEW BoundaryCondition(boundary[f]);
    }
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
        CODIMENSION1_SHAPE_CLASS *face_element=
          codimension1_level->getShape();
        Point<2> center=face_element->center(vertices);
        for (int i=0;i<boundary.getNumber();i++) {
          if (abs(boundary[i]->constraint(center))<1.e-5) {
            codimension1_level->
              setBoundaryCondition(boundary_condition[i]);
            break;
          }
        }
      }
    }
    tess.synchronize(boundary_data);
    quadrature_map[3]=q;
    quadrature_map[4]=q;
#endif
#if (SPACEDIM==3)
    NumPtr<Point<3> > vertices(12);
    vertices[ 0]=Point<3>(0.,0.,0.);
    vertices[ 1]=Point<3>(1.,0.,0.);
    vertices[ 2]=Point<3>(0.,1.,0.);
    vertices[ 3]=Point<3>(1.,1.,0.);
    vertices[ 4]=Point<3>(0.,0.,.5);
    vertices[ 5]=Point<3>(1.,0.,.5);
    vertices[ 6]=Point<3>(0.,1.,.5);
    vertices[ 7]=Point<3>(1.,1.,.5);
    vertices[ 8]=Point<3>(0.,.5,1.);
    vertices[ 9]=Point<3>(.3,.5,1.);
    vertices[10]=Point<3>(.6,.5,1.);
    vertices[11]=Point<3>(1.,.5,1.);

    NumPtr<CellData> cells(4);
    cells[0].allocate(8);
    for (int j=0;j<8;j++) cells[0].vertex(j)=j;
    cells[0].setMaterialModel(0);

    cells[1].allocate(6);
    cells[1].vertex(0)=4;
    cells[1].vertex(1)=6;
    cells[1].vertex(2)=9;
    cells[1].vertex(3)=5;
    cells[1].vertex(4)=7;
    cells[1].vertex(5)=10;
    cells[1].setMaterialModel(0);

    cells[2].allocate(4);
    cells[2].vertex(0)=4;
    cells[2].vertex(1)=6;
    cells[2].vertex(2)=8;
    cells[2].vertex(3)=9;
    cells[2].setMaterialModel(0);

    cells[3].allocate(4);
    cells[3].vertex(0)=5;
    cells[3].vertex(1)=7;
    cells[3].vertex(2)=10;
    cells[3].vertex(3)=11;
    cells[3].setMaterialModel(0);

    Tessellation tess;
    BoundaryData *boundary_data=tess.create(vertices,cells);
    NumPtr<Boundary*> boundary(7);
    NumPtr<BoundaryCondition*> boundary_condition(7);
    Point<3> centroid((vertices[0]+vertices[1]
                      +vertices[2]+vertices[3])*0.25);
    boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
      vertices[2],vertices[4]);
    boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
      vertices[3],vertices[5]);
    boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
      vertices[1],vertices[4]);
    boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[2],
      vertices[3],vertices[6]);
    boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
      vertices[1],vertices[2]);
    boundary[5]=OPERATOR_NEW PlaneBoundary(centroid,vertices[4],
      vertices[5],vertices[8]);
    boundary[6]=OPERATOR_NEW PlaneBoundary(centroid,vertices[6],
      vertices[7],vertices[8]);
    for (int f=0;f<7;f++) {
      boundary_condition[f]=OPERATOR_NEW BoundaryCondition(boundary[f]);
#ifdef DEBUG
//    cout << "\tboundary_condition[" << f << "] = "
//         << boundary_condition[f] << " boundary = " << boundary[f] 
//         << endl;
#endif
    }
    for (bool doit=tess.firstCodimension1ActiveShape();doit;
    doit=tess.nextCodimension1ActiveShape()) {
      CODIMENSION1_LEVEL_CLASS *codimension1_level=
        tess.presentCodimension1Level();
      if (codimension1_level->getBoundaryCondition()!=0) {
        CODIMENSION1_SHAPE_CLASS *face_element=
          codimension1_level->getShape();
        Point<3> center=face_element->center(vertices);
        for (int i=0;i<boundary.getNumber();i++) {
          if (abs(boundary[i]->constraint(center))<1.e-5) {
            codimension1_level->setBoundaryCondition(boundary_condition[i]);
            break;
          }
        }
      }
    }
    tess.synchronize(boundary_data);
#ifdef DEBUG
    tess.checkFaceBoundaryCondition(boundary_condition);
    tess.checkLineBoundaryCondition(boundary_condition);
#endif
    quadrature_map[4]=q;
    quadrature_map[6]=q;
    quadrature_map[8]=q;
#endif
#ifdef DEBUG
//tess.printOn(cout);
//  for (map<int,Quadrature<SPACEDIM> >::const_iterator q_iterator=
//  quadrature_map.begin();q_iterator!=quadrature_map.end();q_iterator++)
//  {
//    cout << "\tnv = " << q_iterator->first << endl;
//    q_iterator->second.printOn(cout);
//  }
#endif
//  PlaneBoundary *sb=OPERATOR_NEW PlaneBoundary(p);
//  tess.setBoundary(0,sb);
//  for (int degree_m=1;degree_m<=D_CHOOSE(6,4,3);degree_m++)
    {
      int degree_m=1;
      cout << "\tdegree_m = " << degree_m << endl;
      {
        TRACER_CALL(t,"MappingL")
        MappingL mapping(degree_m);
//      debugDerivativesForMapping(tess,mapping,quadrature_map);
//      debugHdivDerivativesForMapping(tess,mapping,quadrature_map);
        debugHcurlDerivativesForMapping(tess,mapping,quadrature_map);
      }
/* obsolete:
      {
//      TRACER_CALL(t,"MappingH")
        MappingH mapping(degree_m);
        debugDerivativesForMapping(tess,mapping,quadrature_map);
      }
*/
    }
    delete boundary_data;
#if (SPACEDIM>1)
    for (int i=0;i<boundary_condition.getNumber();i++) {
      delete boundary_condition[i]; boundary_condition[i]=0;
    }
    for (int i=0;i<boundary.getNumber();i++) {
      delete boundary[i]; boundary[i]=0;
    }
#endif
  }
  return EXIT_SUCCESS;
}
