// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "GenerateTess.H"
#include "MemoryDebugger.H"
#include "Tessellation.H"
#include "Tracer.H"
#ifdef USE_GTK
#include "GTKWindow.H"
#include <gtk/gtkmain.h>
#endif


int main(int argc,char* argv[]) {
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
#ifdef USE_GTK
    GTKWindow::gtkInit(argc,argv);
#endif

#if (SPACEDIM==1)
    Point<SPACEDIM> low(0.);
    Point<SPACEDIM> edge[1];
      edge[0]=Point<SPACEDIM>(1.);
    int number_elements[SPACEDIM];
      number_elements[0]=2;
    bool set_face_ids=true;

    Tessellation tess;
    NumPtr<Boundary*> boundary;
    NumPtr<BoundaryCondition*> boundary_condition;
    BoundaryData *sub_cell_data=generateParallelpiped(tess,low,edge,
      number_elements,set_face_ids,boundary,boundary_condition);
#ifdef DEBUG
//  tess.printVertices();
//  tess.printElements();
//  for (int f=0;f<2;f++) {
//    cout << "\tmaterial id[" << f << "] = " << static_cast<int>(
//      tess.coarsestLevel()->Level0::getMaterialId(f)) << endl;
//  }
#endif
//  tess.plotShapes();
//  tess.plotUsedShapes();
    tess.plotActiveShapes();

    tess.refineGlobal(1);
//  tess.plotShapes();
//  tess.plotUsedShapes();
    tess.plotActiveShapes();
    delete sub_cell_data;
    for (int i=0;i<boundary_condition.getNumber();i++) {
      delete boundary_condition[i]; boundary_condition[i]=0;
    }
    for (int i=0;i<boundary.getNumber();i++) {
      delete boundary[i]; boundary[i]=0;
    }
#endif

  }
  return EXIT_SUCCESS;
}
