// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include "Boundary.H"
#include "GenerateTess.H"
#include "MemoryDebugger.H"
#include "Tessellation.H"
#include "Tracer.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BoundaryData* oneTriangle(Tessellation &tess,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
//TRACER_CALL(t,"oneTriangle");
  NumPtr<Point<2> > vertices(3);
  vertices[0]=Point<2>(0.,0.);
  vertices[1]=Point<2>(1.,0.);
  vertices[2]=Point<2>(0.5,1.);

  NumPtr<CellData> cells(1);
  cells[0].allocate(3);
  for (int j=0;j<3;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  BoundaryData *subcell_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  boundary.allocate(3);
  boundary_condition.allocate(3);
  Point<2> centroid((vertices[0]+vertices[1]+vertices[2])/3.);
  boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],vertices[2]);
  boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],vertices[1]);
  boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],vertices[2]);
  for (int f=0;f<3;f++) {
    boundary_condition[f]=OPERATOR_NEW BoundaryCondition(boundary[f]);
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<2> center=face_element->center(vertices);
      for (int i=0;i<boundary.getNumber();i++) {
        if (abs(boundary[i]->constraint(center))<1.e-5) {
          codimension1_level->setBoundaryCondition(boundary_condition[i]);
          break;
        }
      }
/*
      if (center[0]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      } else if (center[1]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else {
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      }
*/
    }
  }
  return subcell_data;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BoundaryData* oneQuadrilateral(Tessellation &tess,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
//TRACER_CALL(t,"oneQuadrilateral");
  NumPtr<Point<2> > vertices(4);
  vertices[0]=Point<2>(0.,0.);
  vertices[1]=Point<2>(1.,0.);
  vertices[2]=Point<2>(0.,1.);
  vertices[3]=Point<2>(1.,1.);

  NumPtr<CellData> cells(1);
  cells[0].allocate(4);
  for (int j=0;j<4;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  BoundaryData *subcell_data=tess.create(vertices,cells);
//must replace bogus boundary condition from Tessellation::create
  boundary.allocate(4);
  boundary_condition.allocate(4);
  Point<2> centroid((vertices[0]+vertices[1]
                    +vertices[2]+vertices[3])*0.25);
  boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],vertices[2]);
  boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],vertices[3]);
  boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],vertices[1]);
  boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[2],vertices[3]);
  for (int f=0;f<4;f++) {
    boundary_condition[f]=OPERATOR_NEW BoundaryCondition(boundary[f]);
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<2> center=face_element->center(vertices);
      for (int i=0;i<boundary.getNumber();i++) {
        if (abs(boundary[i]->constraint(center))<1.e-5) {
          codimension1_level->setBoundaryCondition(boundary_condition[i]);
          break;
        }
      }
/*
      if (center[0]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      } else if (center[0]>0.9) {
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else if (center[1]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      } else {
        codimension1_level->setBoundaryCondition(boundary_condition[3]);
      }
*/
    }
  }
#ifdef DEBUG
//tess.printCodimension1Elements();
#endif
  return subcell_data;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BoundaryData* oneTriangleOneQuadrilateral(Tessellation &tess,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
//TRACER_CALL(t,"oneTriangleOneQuadrilateral");
  NumPtr<Point<2> > vertices(5);
  vertices[0]=Point<2>(0.,0.);
  vertices[1]=Point<2>(1.,0.);
  vertices[2]=Point<2>(0.,.5);
  vertices[3]=Point<2>(1.,.5);
  vertices[4]=Point<2>(.5,1.);

  NumPtr<CellData> cells(2);
  for (int i=0;i<2;i++) {
    cells[i].setMaterialModel(0);
  }
  cells[0].allocate(4);
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=1;
  cells[0].vertex(2)=2;
  cells[0].vertex(3)=3;

  cells[1].allocate(3);
  cells[1].vertex(0)=2;
  cells[1].vertex(1)=3;
  cells[1].vertex(2)=4;

  BoundaryData *subcell_data=tess.create(vertices,cells);
//must replace bogus boundary condition from Tessellation::create
  boundary.allocate(5);
  boundary_condition.allocate(5);
  Point<2> centroid((vertices[0]+vertices[1]
                    +vertices[2]+vertices[3])*0.25);
  boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],vertices[2]);
  boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],vertices[3]);
  boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],vertices[1]);
  boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[2],vertices[4]);
  boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,vertices[3],vertices[4]);
  for (int f=0;f<5;f++) {
    boundary_condition[f]=OPERATOR_NEW BoundaryCondition(boundary[f]);
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<2> center=face_element->center(vertices);
      for (int i=0;i<boundary.getNumber();i++) {
        if (abs(boundary[i]->constraint(center))<1.e-5) {
          codimension1_level->setBoundaryCondition(boundary_condition[i]);
          break;
        }
      }
/*
      if (center[0]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      } else if (center[0]>0.9) {
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else if (center[1]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      } else if (center[0]<0.4) {
        codimension1_level->setBoundaryCondition(boundary_condition[3]);
      } else {
        codimension1_level->setBoundaryCondition(boundary_condition[4]);
      }
*/
    }
  }
#ifdef DEBUG
//tess.printCodimension1Elements();
#endif
  return subcell_data;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BoundaryData* parallelpiped(Tessellation &tess,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
//TRACER_CALL(t,"parallelpiped");
  Point<2> low(0.,0.);
  Point<2> edge[2];
    edge[0]=Point<SPACEDIM>(1.,0.);
    edge[1]=Point<SPACEDIM>(0.,1.);
  int number_elements[2];
    number_elements[0]=2;
    number_elements[1]=2;
  bool set_face_ids=true;

  BoundaryData *subcell_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,boundary,boundary_condition);
#ifdef DEBUG
//tess.printVertices();
//tess.printElements();
//tess.printCodimension1Elements();
#endif
  return subcell_data;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BoundaryData* parallelpipedTriangles(Tessellation &tess,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
//TRACER_CALL(t,"parallelpipedTriangles");
  Point<2> low(0.,0.);
  Point<2> edge[2];
    edge[0]=Point<SPACEDIM>(1.,0.);
    edge[1]=Point<SPACEDIM>(0.,1.);
  int number_elements[2];
    number_elements[0]=2;
    number_elements[1]=2;
  bool set_face_ids=true;

  BoundaryData *subcell_data=generateParallelpipedTriangles(tess,low,
    edge,number_elements,set_face_ids,boundary,boundary_condition);
#ifdef DEBUG
//tess.printVertices();
//tess.printElements();
//tess.printCodimension1Elements();
#endif
  return subcell_data;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::


int main(int argc,char* argv[]) {
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
#ifdef USE_GTK
    GTKWindow::gtkInit(argc,argv);
#endif
    {
      cout << "\toneTriangle" << endl;
      Tessellation tess;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=
        oneTriangle(tess,boundary,boundary_condition);
      tess.plotActiveShapes();

      tess.refineGlobal(1);
      tess.plotActiveShapes();
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\toneQuadrilateral" << endl;
      Tessellation tess;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=
        oneQuadrilateral(tess,boundary,boundary_condition);
      tess.plotActiveShapes();

      tess.refineGlobal(1);
      tess.plotActiveShapes();
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\toneTriangleOneQuadrilateral" << endl;
      Tessellation tess;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=
        oneTriangleOneQuadrilateral(tess,boundary,boundary_condition);
      tess.plotActiveShapes();

      tess.refineGlobal(1);
      tess.plotActiveShapes();
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tparallelpiped" << endl;
      Tessellation tess;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=
        parallelpiped(tess,boundary,boundary_condition);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes();

      tess.refineGlobal(1);
      tess.plotActiveShapes();
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tparallelpipedTriangles" << endl;
      Tessellation tess;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=
        parallelpipedTriangles(tess,boundary,boundary_condition);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes();
//    tess.plotUsedShapes();
//    tess.plotActiveShapes();

      tess.refineGlobal(1);
      tess.plotActiveShapes();
//    tess.plotUsedShapes();
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    bool set_face_ids=true;
    {
      cout << "\tgenerateBall" << endl;
      Tessellation tess;
      Point<2> center(0.,0.);
      REAL radius=1.;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateBall(tess,center,radius,
        set_face_ids,boundary,boundary_condition);
//    BallBoundary *bb=OPERATOR_NEW BallBoundary(center,radius);
//    tess.setBoundary(0,bb);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes();

      tess.refineGlobal(1);
      tess.plotActiveShapes();
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateBallTriangles" << endl;
      Tessellation tess;
      Point<2> center(0.,0.);
      REAL radius=1.;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateBallTriangles(tess,center,
        radius,set_face_ids,boundary,boundary_condition);
//    BallBoundary *bb=OPERATOR_NEW BallBoundary(center,radius);
//    tess.setBoundary(0,bb);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes();

      tess.refineGlobal(1);
      tess.plotActiveShapes();
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateSlitQuadrilateral" << endl;
      Tessellation tess;
      Point<2> center(0.,0.);
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=
        generateSlitQuadrilateral(tess,set_face_ids,boundary,
        boundary_condition);
//    Point<2> centroid(0.25,0.25);
//    for (int i=0;i<5;i++) {
//      PlaneBoundary *sb=OPERATOR_NEW PlaneBoundary(centroid);
//      tess.setBoundary(i,sb);
//    }
//    centroid[1]=0.75;
//    for (int i=5;i<7;i++) {
//      PlaneBoundary *sb=OPERATOR_NEW PlaneBoundary(centroid);
//      tess.setBoundary(i,sb);
//    }
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes();

      tess.refineGlobal(1);
      tess.plotActiveShapes();
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateL" << endl;
      Tessellation tess;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=
        generateL(tess,set_face_ids,boundary,boundary_condition);
//    Point<2> centroid(0.25,0.25);
//    for (int i=0;i<6;i++) {
//      PlaneBoundary *sb=OPERATOR_NEW PlaneBoundary(centroid);
//      tess.setBoundary(i,sb);
//    }
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes();

      tess.refineGlobal(1);
      tess.plotActiveShapes();
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateBallSectorTriangles" << endl;
      Tessellation tess;
      Point<2> center(0.,0.);
      REAL angle=M_PI*0.25;
      REAL radius=1.;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateBallSectorTriangles(tess,
        center,radius,angle,set_face_ids,boundary,boundary_condition);
//    BallBoundary *bb=OPERATOR_NEW BallBoundary(center,radius);
//    tess.setBoundary(0,bb);
//    angle*=0.5;
//    Point<2> axis(cos(angle),sin(angle));
//    ConeBoundary *cb=OPERATOR_NEW ConeBoundary(axis,center,angle);
//    tess.setBoundary(1,cb);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes();

      tess.refineGlobal(1);
      tess.plotActiveShapes();
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateBallSectorTriangles" << endl;
      Tessellation tess;
      Point<2> center(0.,0.);
      REAL angle=M_PI*0.75;
      REAL radius=1.;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateBallSectorTriangles(tess,
        center,radius,angle,set_face_ids,boundary,boundary_condition);
//    BallBoundary *bb=OPERATOR_NEW BallBoundary(center,radius);
//    tess.setBoundary(0,bb);
//    angle*=0.5;
//    Point<2> axis(cos(angle),sin(angle));
//    ConeBoundary *cb=OPERATOR_NEW ConeBoundary(axis,center,angle);
//    tess.setBoundary(1,cb);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes();

      tess.refineGlobal(1);
      tess.plotActiveShapes();
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateBallSectorTriangles" << endl;
      Tessellation tess;
      Point<2> center(0.,0.);
      REAL angle=M_PI*1.25;
      REAL radius=1.;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateBallSectorTriangles(tess,
        center,radius,angle,set_face_ids,boundary,boundary_condition);
//    BallBoundary *bb=OPERATOR_NEW BallBoundary(center,radius);
//    tess.setBoundary(0,bb);
//    angle*=0.5;
//    Point<2> axis(cos(angle),sin(angle));
//    ConeBoundary *cb=OPERATOR_NEW ConeBoundary(axis,center,angle);
//    tess.setBoundary(1,cb);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes();

      tess.refineGlobal(1);
      tess.plotActiveShapes();
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateBallSectorTriangles" << endl;
      Tessellation tess;
      Point<2> center(0.,0.);
      REAL angle=M_PI*1.75;
      REAL radius=1.;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateBallSectorTriangles(tess,
        center,radius,angle,set_face_ids,boundary,boundary_condition);
//    BallBoundary *bb=OPERATOR_NEW BallBoundary(center,radius);
//    tess.setBoundary(0,bb);
//    angle*=0.5;
//    Point<2> axis(cos(angle),sin(angle));
//    ConeBoundary *cb=OPERATOR_NEW ConeBoundary(axis,center,angle);
//    tess.setBoundary(1,cb);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes();

      tess.refineGlobal(1);
      tess.plotActiveShapes();
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
  }
  return EXIT_SUCCESS;
}
