// "$Header:$"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include "Boundary.H"
//#include "Debug.H"
#include "GenerateTess.H"
#include "MemoryDebugger.H"
#include "Tessellation.H"
#include "Tracer.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BoundaryData* oneTetrahedron(Tessellation &tess,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
//TRACER_CALL(t,"oneTetrahedron");
  NumPtr<Point<3> > vertices(4);
  vertices[0]=Point<3>(0.,0.,0.);
  vertices[1]=Point<3>(1.,0.,0.);
  vertices[2]=Point<3>(0.,1.,0.);
  vertices[3]=Point<3>(0.,0.,1.);
  
  NumPtr<CellData> cells(1);
  cells[0].allocate(4);
  for (int j=0;j<4;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);
  
  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  boundary.allocate(4);
  boundary_condition.allocate(4);
  Point<3> centroid((vertices[0]+vertices[1]
                    +vertices[2]+vertices[3])*0.25);
  boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[2],vertices[3]);
  boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[3]);
  boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[2]);
  boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
    vertices[2],vertices[3]);
  for (int f=0;f<4;f++) {
    boundary_condition[f]=OPERATOR_NEW BoundaryCondition(boundary[f]);
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
      codimension1_level->getShape();
      Point<3> center=face_element->center(vertices);
      for (int i=0;i<boundary.getNumber();i++) {
        if (abs(boundary[i]->constraint(center))<1.e-5) {
          codimension1_level->setBoundaryCondition(boundary_condition[i]);
          break;
        }
      }
/*
      if (center[0]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      } else if (center[1]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else if (center[2]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      } else {
        codimension1_level->setBoundaryCondition(boundary_condition[3]);
      }
*/
    }
  }
  tess.synchronize(boundary_data);
  return boundary_data;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BoundaryData* onePrism(Tessellation &tess,NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
//TRACER_CALL(t,"onePrism");
  NumPtr<Point<3> > vertices(6);
  vertices[0]=Point<3>(0.,0.,0.);;
  vertices[1]=Point<3>(1.,0.,0.);;
  vertices[2]=Point<3>(0.,1.,0.);;
  vertices[3]=Point<3>(0.,0.,1.);;
  vertices[4]=Point<3>(1.,0.,1.);;
  vertices[5]=Point<3>(0.,1.,1.);;
  
  NumPtr<CellData> cells(1);
  cells[0].allocate(6);
  for (int j=0;j<6;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  boundary.allocate(5);
  boundary_condition.allocate(5);
  Point<3> centroid((vertices[0]+vertices[1]+vertices[2]
                    +vertices[3]+vertices[4]+vertices[5])/6.);
  boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[2]);
  boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[3],
    vertices[4],vertices[5]);
  boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[2],vertices[3]);
  boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[3]);
  boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
    vertices[2],vertices[4]);
  for (int f=0;f<5;f++) {
    boundary_condition[f]=OPERATOR_NEW BoundaryCondition(boundary[f]);
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
      codimension1_level->getShape();
      Point<3> center=face_element->center(vertices);
      for (int i=0;i<boundary.getNumber();i++) {
        if (abs(boundary[i]->constraint(center))<1.e-5) {
          codimension1_level->setBoundaryCondition(boundary_condition[i]);
          break;
        }
      }
/*
      if (center[2]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      } else if (center[0]>0.9) {
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else if (center[0]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      } else if (center[1]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[3]);
      } else {
        codimension1_level->setBoundaryCondition(boundary_condition[4]);
      }
*/
    }
  }
  tess.synchronize(boundary_data);
  return boundary_data;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BoundaryData* oneHexahedron(Tessellation &tess,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
//TRACER_CALL(t,"oneHexahedron");
  NumPtr<Point<3> > vertices(8);
  vertices[0]=Point<3>(0.,0.,0.);
  vertices[1]=Point<3>(1.,0.,0.);
  vertices[2]=Point<3>(0.,1.,0.);
  vertices[3]=Point<3>(1.,1.,0.);
  vertices[4]=Point<3>(0.,0.,1.);
  vertices[5]=Point<3>(1.,0.,1.);
  vertices[6]=Point<3>(0.,1.,1.);
  vertices[7]=Point<3>(1.,1.,1.);
  
  NumPtr<CellData> cells(1);
  cells[0].allocate(8);
  for (int j=0;j<8;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  boundary.allocate(6);
  boundary_condition.allocate(6);
  Point<3> centroid((vertices[0]+vertices[1]+vertices[2]+vertices[3]
              +vertices[4]+vertices[5]+vertices[6]+vertices[7])*0.125);
  boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[2],vertices[4]);
  boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
    vertices[3],vertices[5]);
  boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[4]);
  boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[2],
    vertices[3],vertices[6]);
  boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[2]);
  boundary[5]=OPERATOR_NEW PlaneBoundary(centroid,vertices[4],
    vertices[5],vertices[6]);
  for (int f=0;f<6;f++) {
    boundary_condition[f]=OPERATOR_NEW BoundaryCondition(boundary[f]);
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
      codimension1_level->getShape();
      Point<3> center=face_element->center(vertices);
      for (int i=0;i<boundary.getNumber();i++) {
        if (abs(boundary[i]->constraint(center))<1.e-5) {
          codimension1_level->setBoundaryCondition(boundary_condition[i]);
          break;
        }
      }
/*
      if (center[0]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      } else if (center[0]>0.9) {
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else if (center[1]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      } else if (center[1]>0.9) {
        codimension1_level->setBoundaryCondition(boundary_condition[3]);
      } else if (center[2]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[4]);
      } else {
        codimension1_level->setBoundaryCondition(boundary_condition[5]);
      }
*/
    }
  }
  tess.synchronize(boundary_data);
  return boundary_data;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BoundaryData* twoTetrahedraOnePrismOneHexahedron(Tessellation &tess,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
//TRACER_CALL(t,"twoTetrahedraOnePrismOneHexahedron");
  NumPtr<Point<3> > vertices(12);
  vertices[ 0]=Point<3>(0.,0.,0.);
  vertices[ 1]=Point<3>(1.,0.,0.);
  vertices[ 2]=Point<3>(0.,1.,0.);
  vertices[ 3]=Point<3>(1.,1.,0.);
  vertices[ 4]=Point<3>(0.,0.,.5);
  vertices[ 5]=Point<3>(1.,0.,.5);
  vertices[ 6]=Point<3>(0.,1.,.5);
  vertices[ 7]=Point<3>(1.,1.,.5);

  vertices[ 8]=Point<3>(0.,.5,1.);
  vertices[ 9]=Point<3>(.3,.5,1.);
  vertices[10]=Point<3>(.6,.5,1.);
  vertices[11]=Point<3>(1.,.5,1.);
  
  NumPtr<CellData> cells(4);
  cells[0].allocate(8);
  for (int j=0;j<8;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);
  
  cells[1].allocate(6);
  cells[1].vertex(0)=4;
  cells[1].vertex(1)=6;
  cells[1].vertex(2)=9;
  cells[1].vertex(3)=5;
  cells[1].vertex(4)=7;
  cells[1].vertex(5)=10;
  cells[1].setMaterialModel(0);
   
  cells[2].allocate(4);
  cells[2].vertex(0)=4;
  cells[2].vertex(1)=6;
  cells[2].vertex(2)=8;
  cells[2].vertex(3)=9;
  cells[2].setMaterialModel(0);
  
  cells[3].allocate(4);
  cells[3].vertex(0)=5;
  cells[3].vertex(1)=7;
  cells[3].vertex(2)=10;
  cells[3].vertex(3)=11;
  cells[3].setMaterialModel(0);

  BoundaryData *boundary_data=tess.create(vertices,cells);

//must replace bogus boundary condition from Tessellation::create
  boundary.allocate(7);
  boundary_condition.allocate(7);
  Point<3> centroid((vertices[0]+vertices[1]
                    +vertices[2]+vertices[3])*0.25);
  boundary[0]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[2],vertices[4]);
  boundary[1]=OPERATOR_NEW PlaneBoundary(centroid,vertices[1],
    vertices[3],vertices[5]);
  boundary[2]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[4]);
  boundary[3]=OPERATOR_NEW PlaneBoundary(centroid,vertices[2],
    vertices[3],vertices[6]);
  boundary[4]=OPERATOR_NEW PlaneBoundary(centroid,vertices[0],
    vertices[1],vertices[2]);
  boundary[5]=OPERATOR_NEW PlaneBoundary(centroid,vertices[4],
    vertices[5],vertices[8]);
  boundary[6]=OPERATOR_NEW PlaneBoundary(centroid,vertices[6],
    vertices[7],vertices[8]);
  for (int f=0;f<7;f++) {
    boundary_condition[f]=OPERATOR_NEW BoundaryCondition(boundary[f]);
#ifdef DEBUG
//  cout << "\tboundary_condition[" << f << "] = "
//       << boundary_condition[f] << " boundary = " << boundary[f] << endl;
#endif
  }
  for (bool doit=tess.firstCodimension1ActiveShape();doit;
  doit=tess.nextCodimension1ActiveShape()) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      tess.presentCodimension1Level();
    if (codimension1_level->getBoundaryCondition()!=0) {
      CODIMENSION1_SHAPE_CLASS *face_element=
        codimension1_level->getShape();
      Point<3> center=face_element->center(vertices);
      for (int i=0;i<boundary.getNumber();i++) {
        if (abs(boundary[i]->constraint(center))<1.e-5) {
          codimension1_level->setBoundaryCondition(boundary_condition[i]);
          break;
        }
      }
/*
      if (center[0]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[0]);
      } else if (center[0]>0.9) {
        codimension1_level->setBoundaryCondition(boundary_condition[1]);
      } else if (center[1]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[2]);
      } else if (center[1]>0.9) {
        codimension1_level->setBoundaryCondition(boundary_condition[3]);
      } else if (center[2]<0.1) {
        codimension1_level->setBoundaryCondition(boundary_condition[4]);
      } else if (center[0]<0.4) {
        codimension1_level->setBoundaryCondition(boundary_condition[5]);
      } else {
        codimension1_level->setBoundaryCondition(boundary_condition[6]);
      }
*/
    }
  }
#ifdef DEBUG
  tess.checkFaceBoundaryCondition(boundary_condition);
#endif
  tess.synchronize(boundary_data);
#ifdef DEBUG
//tess.printVertices();
//tess.printCodimension2Elements();
//tess.printCodimension1Elements();
//tess.printElements();
#endif
  return boundary_data;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BoundaryData* parallelpiped(Tessellation &tess,
NumPtr<Boundary*> &boundary,
NumPtr<BoundaryCondition*> &boundary_condition) {
  TRACER_CALL(t,"parallelpiped");
  Point<3> low(0.,0.,0.);
  Point<3> edge[3];
    edge[0]=Point<SPACEDIM>(1.,0.,0.);
    edge[1]=Point<SPACEDIM>(0.,1.,0.);
    edge[2]=Point<SPACEDIM>(0.,0.,1.);
  int number_elements[3];
    number_elements[0]=2;
    number_elements[1]=2;
    number_elements[2]=2;
  bool set_face_ids=true;
  
  BoundaryData *boundary_data=generateParallelpiped(tess,low,edge,
    number_elements,set_face_ids,boundary,boundary_condition);
  return boundary_data;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

int main(int argc,char* argv[]) {
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
#ifdef USE_GTK
    GTKWindow::gtkInit(argc,argv);
#endif
    Palette pal;
    const char *title="testGridGenerator3";
    REAL winsize=0.5;
    Vector3 low(-1.,-1.,-1.);
    Vector3 high(1.,1.,1.);
    VolGraphTool gt("debugTess",low,high,pal,TRUE,0.5);
    gt.newPage();
    gt.expose();
//
    {
      cout << "\toneTetrahedron" << endl;
      Tessellation tess;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=
        oneTetrahedron(tess,boundary,boundary_condition);
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tonePrism" << endl;
      Tessellation tess;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=
        onePrism(tess,boundary,boundary_condition);
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\toneHexahedron" << endl;
      Tessellation tess;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=
        oneHexahedron(tess,boundary,boundary_condition);
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
//
    {
      cout << "\ttwoTetrahedraOnePrismOneHexahedron" << endl;
      Tessellation tess;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=twoTetrahedraOnePrismOneHexahedron(
        tess,boundary,boundary_condition);
#ifdef DEBUG
//    pair<const BoundaryCondition*,const BoundaryCondition*> bcs=
//      tess.coarsestLevel()->Level1::getBoundaryConditions(8);
//    WatchPointer::set((const void**) &(bcs.second));
//    WatchPointer::print();
#endif

//    tess.plotShapes();
//    tess.plotUsedShapes();
      tess.plotActiveShapes(gt);
#ifdef DEBUG
//    WatchPointer::check();
#endif

      tess.refineGlobal(1);
#ifdef DEBUG
//    WatchPointer::check();
#endif
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
//
    bool set_face_ids=true;
//
    {
      cout << "\tgenerateParallelpiped" << endl;
      Tessellation tess;
      Point<3> low(-1.,-1.,-1.);
      Point<3> edge[3]={
        Point<3>(2.,0.,0.),Point<3>(0.,2.,0.),Point<3>(0.,0.,2.)
      };
      int number_elements[3]={2,2,2};
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateParallelpiped(tess,low,edge,
        number_elements,set_face_ids,boundary,boundary_condition);
#ifdef DEBUG
//    for (int i=0;i<boundary_condition.getNumber();i++) {
//      cout << "\tboundary_condition[" << i << "] = "
//           << boundary_condition[i] << endl;
//    }
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    for (int dir=0;dir<3;dir++) 
//  int dir=0;
    {
      cout << "\tgenerateParallelpipedPrisms, dir = " << dir << endl;
      Tessellation tess;
      Point<3> low(-1.,-1.,-1.);
      Point<3> edge[3]={
        Point<3>(2.,0.,0.),Point<3>(0.,2.,0.),Point<3>(0.,0.,2.)
      };
      int number_elements[3]={2,2,2};
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateParallelpipedPrisms(tess,
        low,edge,number_elements,dir,set_face_ids,boundary,
        boundary_condition);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateParallelpipedTetrahedra" << endl;
      Tessellation tess;
      Point<3> low(-1.,-1.,-1.);
      Point<3> edge[3]={
        Point<3>(2.,0.,0.),Point<3>(0.,2.,0.),Point<3>(0.,0.,2.)
      };
      int number_elements[3]={2,2,2};
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateParallelpipedTetrahedra(tess,
        low,edge,number_elements,set_face_ids,boundary,
        boundary_condition);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
//
//
    {
      cout << "\tgenerateBall" << endl;
      Tessellation tess;
      Point<3> center(0.,0.,0.);
      REAL radius=1.;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=
        generateBall(tess,center,radius,set_face_ids,boundary,
        boundary_condition);
#ifdef DEBUG
//    for (int i=0;i<boundary_condition.getNumber();i++) {
//      cout << "\tboundary_condition[" << i << "] = "
//           << boundary_condition[i] << endl;
//    }
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateBallTetrahedra" << endl;
      Tessellation tess;
      Point<3> center(0.,0.,0.);
      REAL radius=1.;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateBallTetrahedra(tess,center,radius,
        set_face_ids,boundary,boundary_condition);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateBallPrismsWithTetrahedra" << endl;
      Tessellation tess;
      Point<3> center(0.,0.,0.);
      REAL radius=1.;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateBallPrismsWithTetrahedra(tess,
        center,radius,set_face_ids,boundary,boundary_condition);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateBallPrismsWithHexahedra" << endl;
      Tessellation tess;
      Point<3> center(0.,0.,0.);
      REAL radius=1.;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateBallPrismsWithHexahedra(tess,
        center,radius,set_face_ids,boundary,boundary_condition);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
//
    {
      cout << "\tgenerateL" << endl;
      Tessellation tess;
      Point<3> center(0.,0.,0.);
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=
        generateL(tess,set_face_ids,boundary,boundary_condition);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateBallSectorTetrahedra" << endl;
      Tessellation tess;
      Point<3> center(0.,0.,0.);
      Point<3> axis(0.,0.,1.);
      REAL radius=1.;
      REAL angle=M_PI*0.25;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateBallSectorTetrahedra(tess,axis,
        center,radius,angle,set_face_ids,boundary,boundary_condition);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateBallSectorTetrahedra" << endl;
      Tessellation tess;
      Point<3> center(0.,0.,0.);
      Point<3> axis(0.,0.,1.);
      REAL radius=1.;
      REAL angle=M_PI*0.75;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateBallSectorTetrahedra(tess,axis,
        center,radius,angle,set_face_ids,boundary,boundary_condition);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateCylinderTetrahedra" << endl;
      Tessellation tess;
      Point<3> center(0.,0.,0.);
      Point<3> axis(0.,0.,1.);
      REAL radius=1.;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateCylinderTetrahedra(tess,
        axis,center,radius,set_face_ids,boundary,boundary_condition);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
//
    {
      cout << "\tgenerateCylinderPrisms" << endl;
      Tessellation tess;
      Point<3> center(0.,0.,0.);
      Point<3> axis(0.,0.,1.);
      REAL radius=1.;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateCylinderPrisms(tess,axis,
        center,radius,set_face_ids,boundary,boundary_condition);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
    {
      cout << "\tgenerateCylinderHexahedra" << endl;
      Tessellation tess;
      Point<3> center(0.,0.,0.);
      Point<3> axis(0.,0.,1.);
      REAL radius=1.;
      NumPtr<Boundary*> boundary;
      NumPtr<BoundaryCondition*> boundary_condition;
      BoundaryData *sub_cell_data=generateCylinderHexahedra(tess,
        axis,center,radius,set_face_ids,boundary,boundary_condition);
#ifdef DEBUG
//    tess.printVertices();
//    tess.printElements();
//    tess.printCodimension1Elements();
#endif
      tess.plotActiveShapes(gt);

      tess.refineGlobal(1);
      tess.coarsestLevel()->finestLevel()->
        plotActiveShapes(tess.getVertices(),gt);
      delete sub_cell_data;
      for (int i=0;i<boundary_condition.getNumber();i++) {
        delete boundary_condition[i]; boundary_condition[i]=0;
      }
      for (int i=0;i<boundary.getNumber();i++) {
        delete boundary[i]; boundary[i]=0;
      }
    }
  }
  return EXIT_SUCCESS;
}
