// "$Header: /home/faculty/johnt/cvs/deal/grid/testGridGenerator2.C,v 1.1.1.1 2004/08/23 18:42:51 johnt Exp $"
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#include <fstream>
#include <iomanip>
#include <iostream>
#include <stdlib.h>
#include <unistd.h>

#include "AreaGraphTool.H"
//nclude "BoundaryLib.H"
//nclude "GridGenerator.H"
#include "MemoryDebugger.H"
#include "Palette.H"
#include "Tessellation.H"
#include "Tracer.H"
#include "VolGraphTool.H"
#include "XYGraphTool.H"
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKGUI.H"
#include "GTKGUIVirtualInput.H"
#include "GTKWindow.H"
#else
#include "GUI.H"
#include "GUIVirtualInput.H"
#endif

REAL winsize=0.5;
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
void plot(Tessellation &tess,const NumPtr<Point<1> > vertices,
const char *title) {
  Palette pal;
  XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  XYGraphTool gt(title,"x","y",0.,1.,-1.,1.,&cmap,0,winsize);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();

  for (bool test=tess.firstActiveShape();test;
  test=tess.nextActiveShape()) {
    const Shape *element=tess.presentShape();
    Point<1> vertex=vertices[element->vertexIndex(0)];
    gt.movePen(vertex[0],0.);
    gt.drawDiamondGivenCenter(vertex[0],0.,0.01);
    vertex=vertices[element->vertexIndex(1)];
    gt.drawLine(vertex[0],0.);
    gt.drawDiamondGivenCenter(vertex[0],0.,0.01);
#ifdef DEBUG
    cout << "\telement = " << element->elementDataIndex()
         << " has vertices numbered";
    for (int i=0;i<element->numberVertices();i++) {
      cout << " " << element->vertexIndex(i);
    }
    cout << endl;
#endif
  }
  gt.flush();
  XYGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
void oneInterval() {
  TRACER_CALL(t,"oneInterval");
  Point<1> p0(0.);
  Point<1> p1(1.);
  NumPtr<Point<1> > vertices(2);
  vertices[0]=p0;
  vertices[1]=p1;

  NumPtr<CellData> cells(1);
  cells[0].allocate(2);
  for (int i=0;i<2;i++) cells[0].vertex(i)=i;
  cells[0].setMaterialModel(0);


  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  plot(tess,vertices,"one interval");
  delete boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
void twoIntervals() {
  TRACER_CALL(t,"twoIntervals");
  Point<1> p0(0.);
  Point<1> p1(0.5);
  Point<1> p2(1.);
  NumPtr<Point<1> > vertices(3);
  vertices[0]=p0;
  vertices[1]=p1;
  vertices[2]=p2;

  NumPtr<CellData> cells(2);
  for (int i=0;i<cells.getNumber();i++) {
    cells[i].allocate(2);
    for (int j=0;j<2;j++) cells[i].vertex(j)=i+j;
    cells[i].setMaterialModel(0);
  }

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  plot(tess,vertices,"two intervals");
  delete boundary_data;
}
#endif

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void plot(Tessellation &tess,const NumPtr<Point<2> > &vertices,
const char *title) {
  Palette pal;
  AreaGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
  REAL rlo[2]; rlo[0]=0.; rlo[1]=0.;
  REAL rhi[2]; rhi[0]=1.; rhi[1]=1.;
  AreaGraphTool gt(title,"x","y",rlo,rhi,&cmap,0,winsize);
  gt.newPage();
  gt.setbgColor("white");
  gt.setfgColor("black");
  gt.drawAxes();

  for (bool test=tess.firstActiveShape();test;
  test=tess.nextActiveShape()) {
    const Shape *element=tess.presentShape();
#ifdef DEBUG
    cout << "\telement = " << element->elementDataIndex()
         << " has vertices numbered";
    for (int i=0;i<element->numberVertices();i++) {
      cout << " " << element->vertexIndex(i);
    }
    cout << endl;
#endif
    element->draw(vertices,gt);
  }
  gt.flush();
  AreaGraphTool::WINDOW_TYPE::QuitButton qb;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void oneTriangle() {
  TRACER_CALL(t,"oneTriangle");
  Point<2> p0(0.,0.);
  Point<2> p1(1.,0.);
  Point<2> p2(0.5,1.);
  NumPtr<Point<2> > vertices(3);
  vertices[0]=p0;
  vertices[1]=p1;
  vertices[2]=p2;

  NumPtr<CellData> cells(1);
  cells[0].allocate(3);
  for (int j=0;j<3;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  plot(tess,vertices,"one triangle");
  delete boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void twoTriangles() {
  TRACER_CALL(t,"twoTriangles");
  NumPtr<Point<2> > vertices(4);
  vertices[0]=Point<2>(0.,0.);
  vertices[1]=Point<2>(1.,0.);
  vertices[2]=Point<2>(1.,1.);
  vertices[3]=Point<2>(0.,1.);

  NumPtr<CellData> cells(2);
  for (int i=0;i<2;i++) {
    cells[i].allocate(3);
    cells[i].setMaterialModel(0);
  }
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=1;
  cells[0].vertex(2)=2;
  cells[1].vertex(0)=0;
  cells[1].vertex(1)=2;
  cells[1].vertex(2)=3;

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  plot(tess,vertices,"two triangles");
  delete boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void fourTriangles() {
  TRACER_CALL(t,"fourTriangles");
  NumPtr<Point<2> > vertices(5);
  vertices[0]=Point<2>(0.,0.);
  vertices[1]=Point<2>(1.,0.);
  vertices[2]=Point<2>(1.,1.);
  vertices[3]=Point<2>(0.,1.);
  vertices[4]=Point<2>(.5,.5);

  NumPtr<CellData> cells(4);
  for (int i=0;i<4;i++) {
    cells[i].allocate(3);
    cells[i].setMaterialModel(0);
    cells[i].vertex(0)=i;
    cells[i].vertex(1)=(i+1)%4;
    cells[i].vertex(2)=4;
  }

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  plot(tess,vertices,"four triangles");
  delete boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void oneQuadrilateral() {
  TRACER_CALL(t,"oneQuadrilateral");
  NumPtr<Point<2> > vertices(4);
  vertices[0]=Point<2>(0.,0.);
  vertices[1]=Point<2>(1.,0.);
  vertices[2]=Point<2>(0.,1.);
  vertices[3]=Point<2>(1.,1.);

  NumPtr<CellData> cells(1);
  cells[0].allocate(4);
  for (int j=0;j<4;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  plot(tess,vertices,"one quadrilateral");
  delete boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void twoQuadrilaterals() {
  TRACER_CALL(t,"twoQuadrilaterals");
  NumPtr<Point<2> > vertices(6);
  vertices[0]=Point<2>(0.,0.);
  vertices[1]=Point<2>(.5,0.);
  vertices[2]=Point<2>(1.,0.);
  vertices[3]=Point<2>(0.,1.);
  vertices[4]=Point<2>(.5,1.);
  vertices[5]=Point<2>(1.,1.);

  NumPtr<CellData> cells(2);
  for (int i=0;i<2;i++) {
    cells[i].allocate(4);
    cells[i].setMaterialModel(0);
    cells[i].vertex(0)=i;
    cells[i].vertex(1)=i+1;
    cells[i].vertex(2)=i+3;
    cells[i].vertex(3)=i+4;
  }

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  plot(tess,vertices,"two quadrilaterals");
  delete boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void slitQuadrilateral() {
  TRACER_CALL(t,"slitQuadrilateral");
  NumPtr<Point<2> > vertices(10);
  vertices[0]=Point<2>(0.,0.);
  vertices[1]=Point<2>(.5,0.);
  vertices[2]=Point<2>(1.,0.);
  vertices[3]=Point<2>(0.,.5);
  vertices[4]=Point<2>(.5,.5);
  vertices[5]=Point<2>(1.,.5);
  vertices[6]=Point<2>(1.,.5);
  vertices[7]=Point<2>(0.,1.);
  vertices[8]=Point<2>(.5,1.);
  vertices[9]=Point<2>(1.,1.);

  NumPtr<CellData> cells(4);
  for (int i=0;i<4;i++) {
    cells[i].allocate(4);
    cells[i].setMaterialModel(0);
  }
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=1;
  cells[0].vertex(2)=3;
  cells[0].vertex(3)=4;
  cells[1].vertex(0)=1;
  cells[1].vertex(1)=2;
  cells[1].vertex(2)=4;
  cells[1].vertex(3)=5;
  cells[2].vertex(0)=3;
  cells[2].vertex(1)=4;
  cells[2].vertex(2)=7;
  cells[2].vertex(3)=8;
  cells[3].vertex(0)=4;
  cells[3].vertex(1)=6;
  cells[3].vertex(2)=8;
  cells[3].vertex(3)=9;

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  plot(tess,vertices,"slit quadrilateral");
  delete boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void oneTriangleOneQuadrilateral() {
  TRACER_CALL(t,"oneTriangleOneQuadrilateral");
  NumPtr<Point<2> > vertices(5);
  vertices[0]=Point<2>(0.,0.);
  vertices[1]=Point<2>(1.,0.);
  vertices[2]=Point<2>(0.,.5);
  vertices[3]=Point<2>(1.,.5);
  vertices[4]=Point<2>(.5,1.);

  NumPtr<CellData> cells(2);
  for (int i=0;i<2;i++) {
    cells[i].setMaterialModel(0);
  }
  cells[0].allocate(4);
  cells[0].vertex(0)=0;
  cells[0].vertex(1)=1;
  cells[0].vertex(2)=2;
  cells[0].vertex(3)=3;

  cells[1].allocate(3);
  cells[1].vertex(0)=2;
  cells[1].vertex(1)=3;
  cells[1].vertex(2)=4;

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  plot(tess,vertices,"one triangle one quadrilateral");
  delete boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void fourTrianglesOneQuadrilateral() {
  TRACER_CALL(t,"fourTrianglesOneQuadrilateral");
  NumPtr<Point<2> > vertices(8);
  vertices[0]=Point<2>(.5,0.);
  vertices[1]=Point<2>(.25,.25);
  vertices[2]=Point<2>(.75,.25);
  vertices[3]=Point<2>(0.,.5);
  vertices[4]=Point<2>(1.,.5);
  vertices[5]=Point<2>(.25,.75);
  vertices[6]=Point<2>(.75,.75);
  vertices[7]=Point<2>(.5,1.);

  NumPtr<CellData> cells(5);
  for (int i=0;i<5;i++) {
    cells[i].setMaterialModel(0);
  }
  cells[0].allocate(4);
  cells[0].vertex(0)=1;
  cells[0].vertex(1)=2;
  cells[0].vertex(2)=5;
  cells[0].vertex(3)=6;

  for (int i=1;i<5;i++) cells[i].allocate(3);
  cells[1].vertex(0)=0;
  cells[1].vertex(1)=2;
  cells[1].vertex(2)=1;
  cells[2].vertex(0)=1;
  cells[2].vertex(1)=5;
  cells[2].vertex(2)=3;
  cells[3].vertex(0)=2;
  cells[3].vertex(1)=4;
  cells[3].vertex(2)=6;
  cells[4].vertex(0)=5;
  cells[4].vertex(1)=6;
  cells[4].vertex(2)=7;

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  plot(tess,vertices,"four triangles one quadrilateral");
  delete boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void plot(Tessellation &tess,const NumPtr<Point<3> > &vertices,
VolGraphTool &gt) {
//TRACER_CALL(t,"plot");
#ifdef DEBUG
  for (bool test=tess.firstActiveShape();test;
  test=tess.nextActiveShape()) {
    const Shape *element=tess.presentShape();
    cout << "\telement = " << element->elementDataIndex()
         << " has vertices numbered";
    for (int i=0;i<element->numberVertices();i++) {
      cout << " " << element->vertexIndex(i);
    }
    cout << endl;
  }
#endif
  gt.newPage();
  gt.expose();

  VolGraphTool::WINDOW_TYPE::COLOR_TYPE xc=
    gt.getWindow()->getColor("blue");
  for (BOOLEAN do_best=FALSE;;do_best=!do_best) {
    {
      VolGraphTool::DrawDelimiter dd(do_best,&gt);
      for (bool test=tess.firstActiveShape();test;
      test=tess.nextActiveShape()) {
        const Shape *element=tess.presentShape();
        element->draw(xc,vertices,gt);
      }
    }
    if (do_best==TRUE) break;
  }
  gt.eventLoop();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void oneTetrahedron(VolGraphTool &gt) {
  TRACER_CALL(t,"oneTetrahedron");
  NumPtr<Point<3> > vertices(4);
  vertices[0]=Point<3>(0.,0.,0.);
  vertices[1]=Point<3>(1.,0.,0.);
  vertices[2]=Point<3>(0.,1.,0.);
  vertices[3]=Point<3>(0.,0.,1.);

  NumPtr<CellData> cells(1);
  cells[0].allocate(4);
  for (int j=0;j<4;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  plot(tess,vertices,gt);
  delete boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void onePrism(VolGraphTool &gt) {
  TRACER_CALL(t,"onePrism");
  NumPtr<Point<3> > vertices(6);
  vertices[0]=Point<3>(0.,0.,0.);;
  vertices[1]=Point<3>(1.,0.,0.);;
  vertices[2]=Point<3>(0.,1.,0.);;
  vertices[3]=Point<3>(0.,0.,1.);;
  vertices[4]=Point<3>(1.,0.,1.);;
  vertices[5]=Point<3>(0.,1.,1.);;

  NumPtr<CellData> cells(1);
  cells[0].allocate(6);
  for (int j=0;j<6;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  plot(tess,vertices,gt);
  delete boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void oneHexahedron(VolGraphTool &gt) {
  TRACER_CALL(t,"oneHexahedron");
  NumPtr<Point<3> > vertices(8);
  vertices[0]=Point<3>(0.,0.,0.);
  vertices[1]=Point<3>(1.,0.,0.);
  vertices[2]=Point<3>(0.,1.,0.);
  vertices[3]=Point<3>(1.,1.,0.);
  vertices[4]=Point<3>(0.,0.,1.);
  vertices[5]=Point<3>(1.,0.,1.);
  vertices[6]=Point<3>(0.,1.,1.);
  vertices[7]=Point<3>(1.,1.,1.);

  NumPtr<CellData> cells(1);
  cells[0].allocate(8);
  for (int j=0;j<8;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  plot(tess,vertices,gt);
  delete boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void twoTetrahedraOnePrismOneHexahedron(VolGraphTool &gt) {
  TRACER_CALL(t,"twoTetrahedraOnePrismOneHexahedron");
  NumPtr<Point<3> > vertices(12);
  vertices[ 0]=Point<3>(0.,0.,0.);
  vertices[ 1]=Point<3>(1.,0.,0.);
  vertices[ 2]=Point<3>(0.,1.,0.);
  vertices[ 3]=Point<3>(1.,1.,0.);
  vertices[ 4]=Point<3>(0.,0.,.5);
  vertices[ 5]=Point<3>(1.,0.,.5);
  vertices[ 6]=Point<3>(0.,1.,.5);
  vertices[ 7]=Point<3>(1.,1.,.5);
  vertices[ 8]=Point<3>(0.,.5,1.);
  vertices[ 9]=Point<3>(.3,.5,1.);
  vertices[10]=Point<3>(.6,.5,1.);
  vertices[11]=Point<3>(1.,.5,1.);

  NumPtr<CellData> cells(4);
  cells[0].allocate(8);
  for (int j=0;j<8;j++) cells[0].vertex(j)=j;
  cells[0].setMaterialModel(0);

  cells[1].allocate(6);
  cells[1].vertex(0)=4;
  cells[1].vertex(1)=6;
  cells[1].vertex(2)=9;
  cells[1].vertex(3)=5;
  cells[1].vertex(4)=7;
  cells[1].vertex(5)=10;
  cells[1].setMaterialModel(0);

  cells[2].allocate(4);
  cells[2].vertex(0)=4;
  cells[2].vertex(1)=6;
  cells[2].vertex(2)=8;
  cells[2].vertex(3)=9;
  cells[2].setMaterialModel(0);

  cells[3].allocate(4);
  cells[3].vertex(0)=5;
  cells[3].vertex(1)=7;
  cells[3].vertex(2)=10;
  cells[3].vertex(3)=11;
  cells[3].setMaterialModel(0);

  Tessellation tess;
  BoundaryData *boundary_data=tess.create(vertices,cells);
#ifdef DEBUG
//tess.printOn(cout);
#endif
  plot(tess,vertices,gt);
  delete boundary_data;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

int main(int argc,char* argv[]) {
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
#ifdef USE_GTK
    GTKWindow::gtkInit(argc,argv);
#endif

//  window size, as fraction of min(screen width, screen height)

#if (SPACEDIM==1)
    oneInterval();
    twoIntervals();
#endif
#if (SPACEDIM==2)
    oneTriangle();
    twoTriangles();
    fourTriangles();
    oneQuadrilateral();
    twoQuadrilaterals();
    slitQuadrilateral();
    oneTriangleOneQuadrilateral();
    fourTrianglesOneQuadrilateral();
#endif
#if (SPACEDIM==3)
    Vector3 low(-0.05,-0.05,-0.05);
    Vector3 high(1.05,1.05,1.05);
    Palette pal;
    BOOLEAN use_lighting=TRUE;
    VolGraphTool gt("3D Tessellation",low,high,pal,use_lighting,winsize);
    oneTetrahedron(gt);
    onePrism(gt);
    oneHexahedron(gt);
    twoTetrahedraOnePrismOneHexahedron(gt);
#endif
  }
  return EXIT_SUCCESS;
}
