//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
// "$Header: /home/faculty/johnt/cvs/deal_new/graphics/GLWindow.C,v 1.1 2009/08/20 17:31:46 johnt Exp $
#include "GLWindow.H"
#ifndef USE_GTK
#if (SPACEDIM>1)
#include <math.h>

#define NTSC_WIDTH 600
#define NTSC_HEIGHT 420

#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <string.h>
#define LABEL_LENGTH 10
#include <float.h>
#include <math.h>
#include <stdlib.h>

#include <GL/gl.h>
#include <GL/glu.h>
#if defined (__digital__) && defined (__unix__)
#else
#include <GL/glut.h>
#endif
#include <GL/glx.h>
#include <GLwMDrawA.h>

#include <X11/keysym.h>
#include <Xm/Xm.h>
#include <Xm/MainW.h>
#include <Xm/Frame.h>
#include <Xm/PushB.h>
#include <Xm/ToggleB.h>
#include <Xm/CascadeB.h>
#include <Xm/Separator.h>
#include <Xm/MessageB.h>
#include <Xm/RowColumn.h>
#include <Xm/Form.h> 
#include <Xm/DrawingA.h>

#include "Debug.H"
#include "MyInline.H"
#include "XWindow.H"

#define LABEL_LENGTH 10
#define BUFSIZE 512
#define DEFAULT_FONT "-adobe-times-medium-r-normal-*-10-100-*-*-p-*-iso8859-*"

static const char*   fallbackResources[] = {
#ifdef IRIX_5_2_or_higher
  "*sgiMode: true",       /* try to enable IRIX 5.2+ look & feel */ 
  "*useSchemes: all",     /* and SGI schemes */
#endif
//"*background: LightGrey",
  "*title: OpenGL Box",
  "!*glxarea*width: 300", "*glxarea*height: 300", 
  "*selectlabel.labelString:  (0, 0, 0)", 
  "*selectbutton.labelString:  Select",
  "*lxscale.titleString:  X lx",  
  "*lyscale.titleString:  Y ly",  
  "*lzscale.titleString:  Z lz",  
  "*rxscale.titleString:  X rx",  
  "*ryscale.titleString:  Y ry",  
  "*rzscale.titleString:  Z rz",  
  "*drawing area.titleString: drawing area",
  reinterpret_cast<char*>(0)
};
void beep() { system("/bin/echo \"\007\""); }

unsigned int StackCount::count=0;
GLenum GLWindow::FeedbackRecord::feedback_type=GL_3D;
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::updateMat() {
  if (!mat_valid) {
    glGetDoublev(GL_MODELVIEW_MATRIX, ModelMat);
    glGetDoublev(GL_PROJECTION_MATRIX, ProjMat);
    glGetIntegerv(GL_VIEWPORT, viewport);
    mat_valid=TRUE;
  }
  CHECK_POSITIVE(detMatrix(ModelMat))
  CHECK_POSITIVE(-detMatrix(ProjMat))
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Vector3 GLWindow::clipCoord(const Vector3 &object_coord) {
  updateMat();
  GLdouble winx, winy, winz;
  GLint fake_viewport[4];
    fake_viewport[0]=0; fake_viewport[1]=0;
    fake_viewport[2]=2; fake_viewport[3]=2;
  gluProject(object_coord[0],object_coord[1],object_coord[2],
              ModelMat, ProjMat, fake_viewport, &winx, &winy, &winz );
  return Vector3(winx-ONE,winy-ONE,TWO*winz-ONE);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Vector3 GLWindow::faceLocation(const Pair<GLdouble> &screen_coord) {
  double max_depth=DBL_MAX;
  BOOLEAN is_interior=FALSE;
  Vector3 best=bound;

  Pair<GLdouble> plo[3],phi[3];
  Vector3 vlo[3],vhi[3];
  Pair<GLdouble> plow=screenCoords(-bound);
  Pair<GLdouble> phigh=screenCoords(bound);
  for (int i=0;i<3;i++) {
    vlo[i]=-bound; vlo[i][i]= bound[i]; plo[i]=screenCoords(vlo[i]);
    vhi[i]= bound; vhi[i][i]=-bound[i]; phi[i]=screenCoords(vhi[i]);
  }
  for (int i=0;i<3;i++) {
    int i1=(i+1)%3,i2=(i+2)%3;
    Pair<GLdouble> tlo=
      closestInteriorPosition(plow,plo[i1],plo[i2],screen_coord);
    Vector3 object_coord=
      -bound+(vlo[i1]+bound)*tlo[0]+(vlo[i2]+bound)*tlo[1];
    Vector3 f=clipCoord(object_coord);
    if (min(tlo[0],tlo[1])>ZERO && max(tlo[0],tlo[1])<ONE) {
      if (!is_interior || f[2]>max_depth) {
        max_depth=f[2];
        best=object_coord;
      }
      is_interior=TRUE;
    } else if (!is_interior && f[2]>max_depth) {
      max_depth=f[2];
      best=object_coord;
    }

    Pair<GLdouble> thi=
      closestInteriorPosition(phigh,phi[i1],phi[i2],screen_coord);
    object_coord=bound+(vhi[i1]-bound)*thi[0]+(vhi[i2]-bound)*thi[1];
    f=clipCoord(object_coord);
    if (min(thi[0],thi[1])>ZERO && max(thi[0],thi[1])<ONE) {
      if (!is_interior || f[2]>max_depth) {
        max_depth=f[2];
        best=object_coord;
      }
      is_interior=TRUE;
    } else if (!is_interior && f[2]>max_depth) {
      max_depth=f[2];
      best=object_coord;
    }
  }
  return best;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::crossHairs(const Vector3 &object_coord) {
  if  (!mouse_list) mouse_list=OPERATOR_NEW MyDisplayList();
  { ListDelimiter ld(mouse_list->getBase(),GL_COMPILE);
    REAL fudge=radius/30.;
    char label[LABEL_LENGTH];
    Vector3 user_coord=userCoords(object_coord);
    for (int i=0;i<3;i++) {
      if (i==0) setfgColor("red");
      if (i==1) setfgColor("green");
      if (i==2) setfgColor("blue");
      Vector3 low=object_coord; low[i]=-bound[i];
      Vector3 high=object_coord; high[i]=bound[i];
      drawLine(&low,&high);
      for (int j=0;j<LABEL_LENGTH;j++) label[j]=0;
      snprintf(label,LABEL_LENGTH,"%g",user_coord[i]);
      label[LABEL_LENGTH-1]=0;
      Vector3 pos=low; 
      pos[i]=-bound[i]-fudge;
      glRasterPos3f(pos[0],pos[1],pos[2]);
      printString(label);
      pos[i]=bound[i]+fudge;
      glRasterPos3f(pos[0],pos[1],pos[2]);
      printString(label);
    }
  }
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
GLWindow::GLWindow(const char *na,const Vector3 &b,BOOLEAN ul,REAL rw,
REAL rh,BOOLEAN rb,BOOLEAN persp,BOOLEAN td,BOOLEAN sbb) : 
VirtualGLWindow(na,b,rb),glarea(0),
#if (SPACEDIM==3)
slide_bar(0),toggle_button_array(0),
surface_normal_direction(INCREASING_FUNCTION),
#endif
dpy(0),screen(0),vi(0),cx(0),double_buffer(GL_TRUE),W(0),H(0),
rwidth(ZERO),rheight(ZERO),radius(0.),active_clip_direction(0),
active_clip_hand(LEFT_HAND),closest_edge(-bound,Vector3(ONE,ZERO,ZERO)),
use_lighting(ul),perspective(persp),three_dimensional(td),
show_bounding_box(sbb),
button_pressed(FALSE),break_flag(FALSE),break_mouse_flag(FALSE),
finding_mouse(FALSE),cursor(bound),new_model(TRUE),mat_valid(FALSE),
moved(FALSE),cursor_selected(FALSE),font(0),fontbase(0),FontHeight(0),
FontDescent(0),bounding_box_list(0),mouse_list(0),user_list(0),
best_list(0),best_list_exposed(TRUE),display(0),image(0) {
  Arg args[20];
  Cardinal n;

  n=0;
  XtSetArg(args[n],XmNallowShellResize,True); n++;

  int i=0;
  char **t=reinterpret_cast<char**>(XtMalloc(sizeof(char*)));
  topshell=XtOpenApplication(&appcontext,const_cast<char*>("VolGraphTool"),
    0, 0, &i, t, const_cast<char**>(fallbackResources),
    applicationShellWidgetClass,args,n);
  XtFree(reinterpret_cast<char *>(t));

  if (name!=0) {
    XtVaSetValues(topshell, XtNtitle, name, 
      reinterpret_cast<XtPointer>(0));
  }

  graph_tool = XmCreateMainWindow(topshell,const_cast<char*>("VolGraphTool"),
    0, 0);

  createMenubar();

  Widget parent = XtVaCreateWidget ( "Form",
    xmFormWidgetClass, graph_tool,
    XmNresizePolicy, XmRESIZE_GROW,
    reinterpret_cast<XtPointer>(0));

  dpy = XtDisplay(parent);
  screen=XDefaultScreen(dpy);
  XtVaGetValues(parent,XmNcolormap,&cmap,NULL);
  int dw=XDisplayWidth(dpy,screen);
  int dh=XDisplayHeight(dpy,screen);
  int dm=min(dw,dh);
  if (rw>ONE) rw=ONE;
  if (rh>ONE) rh=ONE;
  if (rw<=ZERO || rh<=ZERO) {
    W=NTSC_WIDTH;
    H=NTSC_HEIGHT;
  } else {
    W=static_cast<int>(rw*dm);
    H=static_cast<int>(rh*dm);
  }
  rwidth=static_cast<REAL>(W);
  rheight=static_cast<REAL>(H);

  int errorBase,eventBase;

  radius=norm(bound);
  
  XtAppContext app=XtDisplayToApplicationContext(dpy);
  if (!glXQueryExtension(dpy, &errorBase, &eventBase)){
    XtAppError(app, "No glx extension on this display");
  }
    
  int dblBuf[]={GLX_DOUBLEBUFFER,GLX_RGBA,GLX_DEPTH_SIZE,16,
    GLX_RED_SIZE,1,GLX_GREEN_SIZE,1,GLX_BLUE_SIZE,1,None};
  vi=glXChooseVisual(dpy,DefaultScreen(dpy),dblBuf);
  if (!vi) {
    vi=glXChooseVisual(dpy,DefaultScreen(dpy),&dblBuf[1]);
    if (!vi) XtAppError(app,"no RGB visual with depth buffer found");
    double_buffer = GL_FALSE;
  }

  cx = glXCreateContext(dpy, vi, 
         None,     //no display list sharing
         GL_TRUE); //favor direct rendering
  if (cx == 0) XtAppError(app, "could not create rendering context");

  glarea = XtVaCreateManagedWidget(name,
    glwMDrawingAreaWidgetClass, parent, 
    GLwNvisualInfo, vi, 
    XmNwidth, W, XmNheight, H,
    XmNtraversalOn,True,// ask for keyboard input
    0);

// glwMDrawingAreaWidgetClass also allocate a colormap unless one is
// proviede by the application. The same colormap uses for multiple
// drawing areas with the same visual. For multiple gl drawing areas
// sharing a single colormap, set GLwNinstallColormap TRUE only for
// exactly one area.
        
    XtAddCallback(glarea, GLwNexposeCallback, 
                  &GLWindow::exposeCallback, 
                  reinterpret_cast<XtPointer>(this) );
    XtAddCallback(glarea, GLwNresizeCallback, 
                  &GLWindow::resizeCallback, 
                  reinterpret_cast<XtPointer>(this) );
    XtAddCallback(glarea, GLwNinputCallback, 
                  &GLWindow::inputCallback, 
                  reinterpret_cast<XtPointer>(this) );
    XtAddCallback(glarea, GLwNginitCallback, 
                  &GLWindow::ginitCallback, 
                  reinterpret_cast<XtPointer>(this) );

// XtWindow(glarea) exists only after the widget has been realized.
// No GL operations should be performed until the window has been
// realized. The ginit callback is called when the window has been
// created.

  XtVaSetValues(glarea,
    XmNtopAttachment, XmATTACH_FORM,
    XmNbottomAttachment, XmATTACH_NONE,
    XmNleftAttachment, XmATTACH_FORM,
    XmNrightAttachment, XmATTACH_FORM,
    reinterpret_cast<XtPointer>(0) );

  XtManageChild(parent);
  XtManageChild(graph_tool);
  XtRealizeWidget(topshell);

#if (SPACEDIM==2)
//y direction up, z direction normal to screen
  Pair<REAL> zero_pair;
  curquat=trackBall(zero_pair,zero_pair);
#else
//z direction up, y normal to screen
  curquat=Quaternion<REAL>(ONE,ZERO,ZERO,ONE);
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
GLWindow::~GLWindow() {
  clearClipPlanes();
#if (SPACEDIM==3)
  if (slide_bar) delete slide_bar; slide_bar=0;
  if (toggle_button_array) delete toggle_button_array; 
    toggle_button_array=0;
#endif
  if (bounding_box_list) {
    delete bounding_box_list; bounding_box_list=0;
  }
  clearDisplayLists();

  XDestroyWindow( dpy, XtWindow(glarea) );
  XDestroyWindow( XtDisplay(topshell), XtWindow(topshell) );
  XtDestroyApplicationContext(appcontext);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Widget GLWindow::createMenubar() {
  Arg args[20];
  Widget menubar = XmCreateMenuBar(graph_tool, 
    const_cast<char*>("graph_tool_menu"), 0, 0);
  XtManageChild(menubar);

  Widget menupane = XmCreatePulldownMenu(menubar, 
    const_cast<char*>("Action"), 0, 0);
  XtSetArg(args[0], XmNsubMenuId, menupane);
  Widget cascade = XmCreateCascadeButton(menubar, 
    const_cast<char*>("Action"), args, 1);
  XtManageChild(cascade);

  Widget break_btn = XmCreatePushButton(menupane, 
    const_cast<char*>("Continue"), 0, 0);
  XtAddCallback(break_btn, XmNactivateCallback,
                      &GLWindow::breakoutCallback,
                      reinterpret_cast<XtPointer>(this) );
  XtManageChild(break_btn);

  Widget btn = XmCreatePushButton(menupane, 
    const_cast<char*>("Quit"), 0, 0);
  XtAddCallback(btn, XmNactivateCallback,
                      &GLWindow::quitCallback, 
                      reinterpret_cast<XtPointer>(this) );
  XtManageChild(btn);

  return menubar;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::quit(Widget w) {
  int result=FALSE;
  char str[]="Do you want to quit?";
  Widget dialog = XmCreateQuestionDialog (w, 
    const_cast<char*>("question"), 0, 0 );

  XtVaSetValues ( dialog,
    XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
    XtVaTypedArg, XmNmessageString,
    XmRString, str, strlen ( str ) + 1,
    0 );
  XtAddCallback(dialog,XmNokCallback,YesCallback,
    reinterpret_cast<XtPointer>(&result) );
  XtManageChild ( dialog );
  while ( XtIsManaged ( dialog ) ) {
    XEvent event;
    XtAppNextEvent(XtWidgetToApplicationContext(dialog),&event);
    XtDispatchEvent ( &event );
  }
  XtDestroyWidget ( dialog );
  if (result) break_flag=TRUE;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::eventLoop(BOOLEAN loop_only_if_escaped) {
  XEvent event;
  if (loop_only_if_escaped) {
    BOOLEAN found_escape=FALSE;
    while (XtAppPending(appcontext)) {
      XtAppNextEvent(appcontext,&event);
      if (event.type==KeyPress) {
        char buf;
        KeySym keysym;
        if (XLookupString(reinterpret_cast<XKeyEvent*>(&event),&buf,1,
        &keysym,0)==1) {
          found_escape= (keysym==XK_Escape);
        }
      } else XtDispatchEvent(&event);
    }
    if (!found_escape) return;
    beep();
  }

  break_flag=FALSE;
  while (!break_flag) {
    XtAppNextEvent(appcontext,&event);
    XtDispatchEvent(&event);
  }
  break_flag=FALSE;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int GLWindow::numberSurfaces() const {
#if (SPACEDIM==3)
  if (toggle_button_array!=0) {
    CHECK_SAME(number_surfaces,toggle_button_array->numberSurfaces());
  }
  return number_surfaces;
#else
  return 1;
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Vector3 GLWindow::getMouse() {
  XEvent event;
  break_mouse_flag=FALSE;
  finding_mouse=TRUE;
  while (!break_mouse_flag) {
    XtAppNextEvent(appcontext,&event);
    XtDispatchEvent(&event);
  }
  break_mouse_flag=FALSE;
  finding_mouse=FALSE;
  return cursor;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//called when the widget is first realized.
// Perform any graphics initialization

// Bind the GLX context to the window (make this window the current one)
// Need to be called only once if you only have one GL window, otherwise
// it has to be called at start of each callback.
//glXMakeCurrent( dpy, XtWindow(glarea), cx );

void GLWindow::ginit() {
  GLwDrawingAreaMakeCurrent(glarea,cx);

  makeRasterFont();

  if (use_lighting) {
    glShadeModel(GL_SMOOTH);
    GLfloat light0_diffuse[] = {1.0f, 1.0f, 1.0f, 1.0};
    GLfloat position[]={0.0f, 0.0f, 1.0, 0.0f};
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
    glLightfv(GL_LIGHT0, GL_POSITION, position);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
    {
      const GLfloat mat_diffuse1[]={1.0f, 0.5f, 0.5f, 1.0f};
      glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse1);
      const GLfloat mat_diffuse2[]={0.5f, 1.0f, 0.5f, 1.0f};
      glMaterialfv(GL_BACK, GL_DIFFUSE, mat_diffuse2);
    }
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);
  } else {
    glShadeModel(GL_FLAT);
  }

  if (three_dimensional) {
    resize();
    glEnable(GL_DEPTH_TEST);
  } else {
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-radius,radius,-radius,radius,-radius,radius);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glDisable(GL_DEPTH_TEST);
  }

  VirtualGLWindow::ginit();

  setbgColor(0.7,0.7,0.7);
  makeBoundingBox();

  resize();
  updateMat();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::rescale(const Vector3 &l,const Vector3 &h) {
  VirtualGLWindow::rescale(l,h);
  makeBoundingBox();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::makeBoundingBox() {
  if  (!bounding_box_list) {
    bounding_box_list=OPERATOR_NEW MyDisplayList();
  }
  { ListDelimiter ld(bounding_box_list->getBase(),GL_COMPILE);
    if (use_lighting) {
      glDisable(GL_LIGHTING);
      glDisable(GL_LIGHT0);
    }
    setfgColor("black");
    glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
    {
      Vector3 low=-bound;
      GLenum mode=GL_RENDER;
      GLenum type=GL_QUADS;
      drawBox(&low,&bound,&mode,&type);
    }
    INTEGER three=3;
    setLineWidth(&three);
    for (int i=0;i<3;i++) {
      int i1=(i+1)%3,i2=(i+2)%3;
      if (i==0) setfgColor("red");
      if (i==1) setfgColor("green");
      if (i==2) setfgColor("blue");
      Vector3 low(-bound),high(-bound);
      low[i]=clip_plane[i][LEFT_HAND]->getLocation();
      high[i]=clip_plane[i][RIGHT_HAND]->getLocation();

      drawLine(&low,&high);
      low[i1]=high[i1]=bound[i1];
      drawLine(&low,&high);
      low[i2]=high[i2]=bound[i2];
      drawLine(&low,&high);
      low[i1]=high[i1]=-bound[i1];
      drawLine(&low,&high);
    }
    INTEGER one=1;
    setLineWidth(&one);

    float fudge=radius/30.,len=radius/10.,len1=radius/10.+radius/50.;
    Vector3 fudged_origin=-bound;
    for (int i=0;i<3;i++) fudged_origin[i]-=fudge;

    Vector3 vector_end=fudged_origin; vector_end[0]+=len;
    setfgColor("red"); // setfgColor(1.,0.,0.);
    drawLine(&fudged_origin,&vector_end);
    glRasterPos3f(-bound[0]-fudge+len1,-bound[1]-fudge,-bound[2]-fudge);
    printString("x");

    vector_end=fudged_origin; vector_end[1]+=len;
    setfgColor("green"); // setfgColor(0.,1.,0.);
    drawLine(&fudged_origin,&vector_end);
    glRasterPos3f(-bound[0]-fudge,-bound[1]-fudge+len1,-bound[2]-fudge);
    printString("y");

    vector_end=fudged_origin; vector_end[2]+=len;
    setfgColor("blue"); // setfgColor(0.,0.,1.);
    drawLine(&fudged_origin,&vector_end);
    glRasterPos3f(-bound[0]-fudge,-bound[1]-fudge,-bound[2]-fudge+len1);
    printString("z");
    if (use_lighting) {
      glEnable(GL_LIGHTING);
      glEnable(GL_LIGHT0);
    }
  }
}
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
GLWindow::SlideBar* GLWindow::slideBar() {
  CHECK_SAME(number_surfaces,1)
  if (!slide_bar) slide_bar=OPERATOR_NEW SlideBar(glarea,H); 
  return slide_bar;
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::deleteSlideBar() {
  if (slide_bar) delete slide_bar; slide_bar=0;
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
GLWindow::ToggleButtonArray* GLWindow::toggleButtonArray(int ns) {
  CHECK_TEST(ns>1)
  if (toggle_button_array==0) {
    CHECK_SAME(number_surfaces,1)
    number_surfaces=ns;
    toggle_button_array=
      OPERATOR_NEW ToggleButtonArray(glarea,number_surfaces); 
  }
  return toggle_button_array;
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::deleteToggleButtonArray() {
  if (toggle_button_array) delete toggle_button_array; 
  toggle_button_array=0;
  number_surfaces=1;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::makeRasterFont() {
  strcpy(FontName, DEFAULT_FONT);// can set it to a different font
  
  font = XLoadQueryFont(dpy, 
    "-adobe-helvetica-medium-r-normal--17-120-100-100-p-88-iso8859-1");
  if (!font) {
    cerr<<"Unable to load font: "<<FontName<<endl;
    exit(0);
  }
  fontbase = glGenLists( (GLuint) font->max_char_or_byte2+1 );
  if (fontbase == 0) {
    cerr<<"out of display lists"<<endl;
    exit(0);
  }
  glXUseXFont( font->fid, font->min_char_or_byte2, 
          font->max_char_or_byte2 - font->min_char_or_byte2 + 1, 
          fontbase + font->min_char_or_byte2 );
  FontHeight = font->ascent + font->descent;
  FontDescent = font->descent;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::printString(const char *s) {
  glListBase(fontbase);
  glCallLists(strlen(s),GL_UNSIGNED_BYTE,
    reinterpret_cast<const GLubyte *>(s));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::resize() {
  XtVaGetValues(glarea, XmNwidth, &W, XmNheight, &H, 0);
  rwidth=static_cast<REAL>(W);
  rheight=static_cast<REAL>(H);

  makeCurrent();

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  GLfloat min_W_H=min(W,H);
  GLfloat max_W_H=min(W,H);
  GLfloat aspect_ratio=max_W_H/min_W_H;
  if (W <= H) {
    glOrtho(-radius,radius,-radius*aspect_ratio,radius*aspect_ratio,
      -radius,radius);
  } else {
    glOrtho(-radius*aspect_ratio,radius*aspect_ratio,-radius,radius,
      -radius,radius);
  }
  glViewport(-5, -5, W,H);
  updateMat();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::expose() {
  makeCurrent();

  glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
  disableClipPlanes();

  if (new_model) recalcModelView();
  if (button_pressed || show_bounding_box) bounding_box_list->call();
  if (mouse_list) mouse_list->call();

  enableClipPlanes();
  if ((best_list_exposed || rotate_best) && best_list!=0) { 
    best_list->call();
  } else if (user_list) {
    user_list->call();
  }
  plotClipPlanes();

  if (double_buffer) {
    GLwDrawingAreaSwapBuffers(glarea);
  }

  if (!glXIsDirect(dpy, cx)) glFinish(); 
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::input(XEvent *ev) {
  makeCurrent();

  switch (ev->type) {
    case ButtonPress: {
      Pair<GLdouble> new_screen_coord=
        xyToScreen(ev->xbutton.x,ev->xbutton.y);
      buttonPress(ev->xbutton,new_screen_coord);
      break;
    }
    case ButtonRelease: {
      Pair<GLdouble> new_screen_coord=
        xyToScreen(ev->xbutton.x,ev->xbutton.y);
      buttonRelease(ev->xbutton,new_screen_coord);
      break;
    }
    case MotionNotify: {
      Pair<GLdouble> new_screen_coord=
        xyToScreen(ev->xmotion.x,ev->xmotion.y);
      moved=TRUE;
      if (ev->xmotion.state & Button1Mask) { //Left button
        rotateImage(new_screen_coord);
      } else if (ev->xmotion.state & Button2Mask) { // Middle button
        moveClipPlane(new_screen_coord);
      } else if (ev->xmotion.state & Button3Mask) { // Right button
        moveCursor(new_screen_coord);
      }
      break;
    }
    case KeyPress:
      char buf;
      KeySym keysym;
      if (XLookupString(reinterpret_cast<XKeyEvent*>(ev),&buf,1,&keysym,
      0)==1){
        switch (keysym) {
          case XK_space:
            break;
          case XK_q:
          case XK_Escape:
              exit(0);
        }
      }
      break;
    default:
      break;
  }
  expose();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::buttonPress(const XButtonEvent &xbutton,
const Pair<GLdouble> &new_screen_coord) {
  button_pressed=TRUE;
  moved=FALSE; cursor_selected=FALSE;
  switch (xbutton.button) {
    case Button1: // Left 
      XmProcessTraversal(glarea,XmTRAVERSE_CURRENT); //for keyboard
      break;
    case Button2: // Middle
      selectClipPlane(new_screen_coord);
      break;
    case Button3: // Right
      markFaceLocation(new_screen_coord);
      break;
    default:
      break;
  }
  old_screen_coord=new_screen_coord;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//left button motion
void GLWindow::rotateImage(const Pair<GLdouble> &new_screen_coord) {
  Quaternion<REAL> rotate_quat=
    trackBall(screenToUnitCircle(old_screen_coord),
              screenToUnitCircle(new_screen_coord));
  curquat=curquat*rotate_quat;
  old_screen_coord=new_screen_coord;
  new_model = TRUE;
  best_list_exposed=rotate_best;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//right button motion
void GLWindow::moveCursor(const Pair<GLdouble> &new_screen_coord) {
  if ((new_screen_coord-old_screen_coord).norm()<4.) return;

  GLdouble max_inner_prod=ZERO;
  unsigned int axis=0;
  for (unsigned int i=0;i<3;i++) {
    GLdouble d=innerProduct(wuvec[i],new_screen_coord-old_screen_coord);
    GLdouble abs_d=d>=ZERO?d:-d;
    if (abs_d>max_inner_prod) {
      max_inner_prod=abs_d; axis=i; 
    }
  }

  Vector3 axis_vector; axis_vector[axis]=ONE;
  Line new_cursor_line=objectCoords(new_screen_coord);
  Line allowed_cursor_line(cursor,axis_vector);
  allowed_cursor_line.nearestPositionTo(new_cursor_line);
  cursor=allowed_cursor_line.getPosition();
  
  if (cursor[axis]<-bound[axis]) cursor[axis]=-bound[axis];
  else if (cursor[axis]>bound[axis]) cursor[axis]=bound[axis];

  crossHairs(cursor);

  old_screen_coord=screenCoords(cursor);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//middle button motion
void GLWindow::moveClipPlane(const Pair<GLdouble> &new_screen_coord) {
  Line new_cursor_line=objectCoords(new_screen_coord);
  closest_edge.nearestPositionTo(new_cursor_line);
  cursor=closest_edge.getPosition();
  AxisClipPlane *cp=clip_plane[active_clip_direction][active_clip_hand];

  if (cp->getHand()==LEFT_HAND) {
    cursor[active_clip_direction]=max(-bound[active_clip_direction],
      min(clip_plane[active_clip_direction][RIGHT_HAND]->getLocation(),
          cursor[active_clip_direction]));
  } else {
    cursor[active_clip_direction]=min(bound[active_clip_direction],
      max(clip_plane[active_clip_direction][LEFT_HAND]->getLocation(),
          cursor[active_clip_direction]));
  }

  old_screen_coord=screenCoords(cursor);
  cp->setLocation(cursor[active_clip_direction]);

  makeBoundingBox();
  crossHairs(cursor);
#if (SPACEDIM==3)
  if (plot_obj!=0) {
    if (!plot_obj->drawSurface()) {
      VirtualGLWindow::DrawDelimiter dd(this,cp);
      plot_obj->plot(FALSE,cp);
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::buttonRelease(const XButtonEvent &xbutton,
const Pair<GLdouble> &new_screen_coord) {
  button_pressed=FALSE;
  moved=FALSE; cursor_selected=FALSE;
  switch (xbutton.button) {
    case Button1: // Left 
      break;
    case Button2: // Middle
#if (SPACEDIM==3)
      if (plot_obj!=0) {
        if (!plot_obj->drawSurface()) {
          VirtualGLWindow::DrawDelimiter 
           dd(this,clip_plane[active_clip_direction][active_clip_hand]);
          plot_obj->plot(TRUE,
            clip_plane[active_clip_direction][active_clip_hand]);
        }
      }
#endif
      break;
    case Button3: // Right
      if (finding_mouse) break_mouse_flag=TRUE;
      break;
    default:
      break;
  }
  if (mouse_list) delete mouse_list; mouse_list=0;
  old_screen_coord=new_screen_coord;

  best_list_exposed=TRUE;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::drawLine(const Vector3 *start,const Vector3 *end) const {
  { PrimitiveDelimiter pd(GL_LINES);
    glVertex3f((*start)[0],(*start)[1],(*start)[2]);
    glVertex3f((*end)[0],(*end)[1],(*end)[2]);
  }
}
#if (SPACEDIM==2)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::colorTriangle(const XColor *xc,const Vector3 *v0,
const Vector3 *v1,const Vector3 *v2) const {
  setfgColor(*xc);
  setPolygonMode(GL_FRONT,GL_FILL,GL_CCW);
  Vector3 normal=cross((*v1)-(*v0),(*v2)-(*v0));
  REAL nrm=norm(normal);
  if (nrm>ZERO) normal/=nrm;
  { TriangleDelimiter td;
    if (normal[2]>ZERO) {
      glNormal3d(normal[0],normal[1],normal[2]);
      glVertex3d((*v0)[0],(*v0)[1],(*v0)[2]);
      glVertex3d((*v1)[0],(*v1)[1],(*v1)[2]);
      glVertex3d((*v2)[0],(*v2)[1],(*v2)[2]);
    } else {
      normal=-normal;
      glNormal3d(normal[0],normal[1],normal[2]);
      glVertex3d((*v0)[0],(*v0)[1],(*v0)[2]);
      glVertex3d((*v2)[0],(*v2)[1],(*v2)[2]);
      glVertex3d((*v1)[0],(*v1)[1],(*v1)[2]);
    }
  }
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::colorTriangle(const XColor *xc,const Vector3 *v0,
const Vector3 *v1,const Vector3 *v2) const {
  CHECK_TEST(isDrawing())
  setfgColor(*xc);
  setPolygonMode(GL_FRONT,GL_FILL,GL_CCW);
  Vector3 normal=cross((*v1)-(*v0),(*v2)-(*v0));
  REAL nrm=norm(normal);
  if (nrm>ZERO) normal/=nrm;
  CHECK_TEST((*v0)>=-bound && (*v0)<=bound)
  CHECK_TEST((*v1)>=-bound && (*v1)<=bound)
  CHECK_TEST((*v2)>=-bound && (*v2)<=bound)

  GLfloat emit_max=GLfloat(maxIntensity)*5.f;
  const GLfloat mat_emission1[]=
    {GLfloat(xc->red)/emit_max,GLfloat(xc->green)/emit_max,
     GLfloat(xc->blue)/emit_max,1.f};
  glMaterialfv(GL_FRONT, GL_EMISSION, mat_emission1);
  const GLfloat mat_emission2[]=
   {1.f-mat_emission1[0],1.f-mat_emission1[1],1.f-mat_emission1[2],1.f};
  glMaterialfv(GL_BACK, GL_EMISSION, mat_emission2);

  { TriangleDelimiter td;
    if (surface_normal_direction==INCREASING_FUNCTION) {
      glNormal3d(normal[0],normal[1],normal[2]);
      glVertex3d((*v0)[0],(*v0)[1],(*v0)[2]);
      glVertex3d((*v1)[0],(*v1)[1],(*v1)[2]);
      glVertex3d((*v2)[0],(*v2)[1],(*v2)[2]);
    } else {
      glNormal3d(-normal[0],-normal[1],-normal[2]);
      glVertex3d((*v2)[0],(*v2)[1],(*v2)[2]);
      glVertex3d((*v1)[0],(*v1)[1],(*v1)[2]);
      glVertex3d((*v0)[0],(*v0)[1],(*v0)[2]);
    }
  }
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::drawVector(const Vector3 *base,const Vector3 *head) const
{
  drawLine(base,head);

  REAL c=cos(2.6),s=sin(2.6);
  Vector3 arrow=*head,dif=(*head)-(*base);
  Vector3 adif(abs(dif[0]),abs(dif[1]),abs(dif[2]));
  int d=0; adif.max(d);
  int d1=(d+1)%3,d2=(d+2)%3;
  arrow[d1]+=(c*dif[d1]-s*dif[d2])*ARROW_SIZE;
  arrow[d2]+=(s*dif[d1]+c*dif[d2])*ARROW_SIZE;
  drawLine(head,&arrow);
}
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::drawVector(const Vector3 *base,const Vector3 *head,
const AxisClipPlane *cp) const {
  unsigned int clip_dir=cp->getDirection();
  GLdouble clip_location=cp->getLocation();
  Vector3 b=*base; b[clip_dir]=clip_location;
  Vector3 h=*head; h[clip_dir]=clip_location;
  drawLine(&b,&h);

  int d1=(clip_dir+1)%3,d2=(clip_dir+2)%3;
  REAL c=cos(2.6),s=sin(2.6);
  Vector3 a=h,dif=h-b;
  a[d1]+=(c*dif[d1]-s*dif[d2])*ARROW_SIZE;
  a[d2]+=(s*dif[d1]+c*dif[d2])*ARROW_SIZE;
  drawLine(&h,&a);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::plotObject(BOOLEAN do_best) {
  if (rotate_best) do_best=TRUE;
  if (plot_obj!=0) {
    if (plot_obj->drawSurface()) {
      freeDisplayList(do_best);
      newPage();
      { VirtualGLWindow::DrawDelimiter dd(do_best,this);
        plot_obj->plot(do_best);
      }
      best_list_exposed=do_best;
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::colorRect(const XColor *xc,const Vector3 *low,
const Vector3 *high,const int *dir) const {
  CHECK_TEST(inside_draw_delimiter)
  setfgColor(*xc);
  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  { QuadDelimiter qd;
    Vector3 v=*low; v[*dir]=HALF*((*low)[*dir]+(*high)[*dir]);
    int dir1=(*dir+1)%3,dir2=(*dir+2)%3;
    glVertex3d(v[0],v[1],v[2]);
    v[dir2]=(*high)[dir2];
    glVertex3d(v[0],v[1],v[2]);
    v[dir1]=(*high)[dir1];
    glVertex3d(v[0],v[1],v[2]);
    v[dir2]=(*low)[dir2];
    glVertex3d(v[0],v[1],v[2]);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::makeCurrent() {
  GLwDrawingAreaMakeCurrent(glarea,cx);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::recalcModelView() {
  GLfloat m[4][4];

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  curquat.rotationMatrix(m);
  glMultMatrixf(&m[0][0]);
  glTranslatef(ZERO,ZERO,ZERO);
  new_model = FALSE; 
  mat_valid = FALSE;
  updateMat();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::printVisual() {
  const char *visual_class[]={
    "StaticGray",   // 0
    "GrayScale",    // 1
    "StaticColor",  // 2
    "PseudoColor",  // 3
    "TrueColor",    // 4
    "DirectColor",  // 5
  }; // XVisualInfo struct is in X11/Xutil.h
  
  printf("Visual Info:\n  visualid: %lu, screen: %d, depth: %u, class: %s, colormap_size: %d\n", 
    vi->visualid, vi->screen, vi->depth, 
    visual_class[vi->c_class], vi->colormap_size);
  printf("  red_mask: %lX, green_mask: %lX, blue_mask: %lX, bits_per_rgb: %d\n", 
    vi->red_mask, vi->green_mask, vi->blue_mask, 
    vi->bits_per_rgb);
          
  int bufferSize,level,redSize,greenSize,blueSize,alphaSize,depthSize ;
  Boolean doublebuffer,rgba ;
  
  XtVaGetValues(glarea, 
    GLwNbufferSize, &bufferSize,
    GLwNlevel, &level,
    GLwNdoublebuffer,&doublebuffer,
    GLwNrgba,&rgba,
    GLwNredSize,&redSize,
    GLwNgreenSize,&greenSize,
    GLwNblueSize,&blueSize,
    GLwNalphaSize,&alphaSize,
    GLwNdepthSize,&depthSize,
    0);
  printf("---------------\n bufferSize: %d,level: %d,redSize: %d,greenSize: %d,blueSize: %d,alphaSize: %d,depthSize: %d",
    bufferSize,level,redSize,greenSize,blueSize,alphaSize,depthSize);
  printf(" doublebuffer: %d,rgba: %d\n",doublebuffer,rgba);

  Dimension depth;
  Colormap cmap;

  XtVaGetValues(glarea,
    XmNdepth,&depth,
    XmNcolormap,&cmap,
    0);

  printf("------------\n depth: %d, Colormap: %d\n",depth,cmap);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// project an object coord into X win coord
Pair<GLdouble> GLWindow::screenCoords(const Vector3 &object_coord) {
  GLdouble winx, winy, winz;
  updateMat();
  gluProject(object_coord[0],object_coord[1],object_coord[2],
              ModelMat, ProjMat, viewport, &winx, &winy, &winz );
  return Pair<GLdouble>(winx,winy);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// convert a X win coord into a line in obj coord
Line GLWindow::objectCoords(Pair<GLdouble> screen_coord) {
  updateMat();

  Vector3 pos0,pos1;
  GLint success=gluUnProject(screen_coord[0],screen_coord[1],ZERO,
    ModelMat,ProjMat,viewport,&pos0[0],&pos0[1],&pos0[2]);
  ASSERT(success);

  success=gluUnProject(screen_coord[0],screen_coord[1],ONE,ModelMat,
    ProjMat,viewport,&pos1[0],&pos1[1],&pos1[2]);
  ASSERT(success);

  return Line(pos0,pos1-pos0);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::markFaceLocation(const Pair<double> &new_cursor_coord) {
  cursor=faceLocation(new_cursor_coord);
  crossHairs(cursor);

  Pair<GLdouble> screen_coord_low=screenCoords(-bound);
  for (int i=0;i<3;i++) { // around the low vertex -bound
    Vector3 neighbor(-bound); neighbor[i]=bound[i];

    Pair<GLdouble> screen_coord=screenCoords(neighbor);
    wuvec[i]=screen_coord-screen_coord_low;
    GLdouble d=wuvec[i].norm();
    if (d>ZERO) wuvec[i] /= d;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::selectClipPlane(const Pair<GLdouble> &new_screen_coord) {
  Line new_cursor_line=objectCoords(new_screen_coord);
  GLdouble min_dist=DBL_MAX;
  for (unsigned int i=0;i<3;i++) {
    Vector3 axis; axis[i]=ONE;
    Vector3 vertex(-bound);
    int i1=(i+1)%3,i2=(i+2)%3,c1=0;
    for (vertex[i1]=-bound[i1];c1<2;vertex[i1]=-vertex[i1],c1++) {
      int c2=0;
      for (vertex[i2]=-bound[i2];c2<2;vertex[i2]=-vertex[i2],c2++) {
        Line current_edge(vertex,axis);
        REAL dist=current_edge.nearestPositionTo(new_cursor_line);
        if (dist<min_dist) {
          closest_edge=current_edge;
          active_clip_direction=i;
          min_dist=dist;
        }
      }
    }
  }
  cursor=closest_edge.getPosition();
  if (cursor[active_clip_direction]<-bound[active_clip_direction]) {
    cursor[active_clip_direction]=-bound[active_clip_direction];
  } else if(cursor[active_clip_direction]>bound[active_clip_direction]){
    cursor[active_clip_direction]=bound[active_clip_direction];
  }

  GLdouble middle_pos=
    HALF*(clip_plane[active_clip_direction][LEFT_HAND]->getLocation()
         +clip_plane[active_clip_direction][RIGHT_HAND]->getLocation());
  if (cursor[active_clip_direction]<=middle_pos) {
    active_clip_hand=LEFT_HAND;
  } else active_clip_hand=RIGHT_HAND;
  old_screen_coord=screenCoords(cursor);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::initialize(const VirtualGLWindow::PlotObj *po) {
  plot_obj=po;
#if (SPACEDIM==3)
  if (plot_obj) {
    if (!plot_obj->drawSurface()) {
      disableClipPlanes();
      BOOLEAN do_best=TRUE;
      for (unsigned int i=0;i<3;i++) {
        { VirtualGLWindow::DrawDelimiter dd(this,
            clip_plane[i][LEFT_HAND]);
          plot_obj->plot(do_best,clip_plane[i][LEFT_HAND]);
        }
        { VirtualGLWindow::DrawDelimiter dd(this,
            clip_plane[i][RIGHT_HAND]);
          plot_obj->plot(do_best,clip_plane[i][RIGHT_HAND]);
        }
      }
    }
  }
#endif
  enableClipPlanes();
//the following is necessary for rotation of the initial image
  if (!rotate_best) plotObject(FALSE);
//the following provides the initial image
  plotObject(TRUE);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::plotClipPlanes() const {
#if (SPACEDIM==3)
  if (plot_obj) {
    if (!plot_obj->drawSurface()) {
      enableClipPlanes();
      for (unsigned int i=0;i<3;i++) {
        clip_plane[i][LEFT_HAND]->callPlotList();
        clip_plane[i][RIGHT_HAND]->callPlotList();
      }
    }
  }
#endif
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::terminatePlotObj() {
  if (plot_obj) delete plot_obj;
  plot_obj=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//this must write what openRaster will read
void GLWindow::createRaster(FILE *os) {
  CHECK_POINTER(os);
  CHECK_TEST(!image);
//these correspond to GlobalMain::runMovie
  Vector3 window_coord(HALF,HALF,HALF);
  Vector3 user_high=userCoords(window_coord);
  Vector3 user_low=userCoords(-window_coord);
  fwrite(reinterpret_cast<char*>(&user_low[0]), sizeof(REAL),1,os);
  fwrite(reinterpret_cast<char*>(&user_low[1]), sizeof(REAL),1,os);
  fwrite(reinterpret_cast<char*>(&user_high[0]), sizeof(REAL),1,os);
  fwrite(reinterpret_cast<char*>(&user_high[1]), sizeof(REAL),1,os);

  const char *str="";
  int len=1;
  fwrite(reinterpret_cast<char*>(&len),sizeof(int),1,os);
  fwrite(str,1,1,os);

//this corresponds to code in VGT12::openRaster
//if (double_buffer) GLwDrawingAreaSwapBuffers(glarea);
  display=glXGetCurrentDisplay();
  image=XGetImage(display,XtWindow(glarea),0,0,W,H,AllPlanes,ZPixmap); 
  ASSERT(image!=0);

  int screen=DefaultScreen(display);
  REAL dw=static_cast<REAL>(DisplayWidth(display,screen));
  REAL dh=static_cast<REAL>(DisplayHeight(display,screen));
  REAL dm=min(dw,dh);
  REAL rw=rwidth/dm;
  REAL rh=rheight/dm;
  fwrite( reinterpret_cast<char*>(&rw), sizeof(REAL), 1, os);
  fwrite( reinterpret_cast<char*>(&rh), sizeof(REAL), 1, os);
  int length=strlen(name)+1;
  fwrite( reinterpret_cast<char*>(&length), sizeof(int), 1, os);
  fwrite( name, sizeof(char), length, os);

  fwrite(reinterpret_cast<char*>(&(image->xoffset)),sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->format)),sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->byte_order)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bitmap_unit)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bitmap_bit_order)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bitmap_pad)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->depth)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bytes_per_line)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bits_per_pixel)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->red_mask)),
    sizeof(unsigned long),1,os);
  fwrite(reinterpret_cast<char*>(&(image->green_mask)),
    sizeof(unsigned long),1,os);
  fwrite(reinterpret_cast<char*>(&(image->blue_mask)),
    sizeof(unsigned long),1,os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::openRaster(FILE *is) {
  CHECK_POINTER(is);
  CHECK_TEST(!image);
  int length;
  fread( reinterpret_cast<char*>(&length), sizeof(int), 1, is);
  char *wn=OPERATOR_NEW_BRACKET(char,length);
  fread( wn, 1, length, is);
  ASSERT(strcmp(name,wn)==0); 
   delete wn;
  wn=0;

  int image_width=W,image_height=H;
  int image_xoffset,image_format;
  int image_byte_order,image_bitmap_unit,image_bitmap_bit_order;
  int image_bitmap_pad,image_depth,image_bytes_per_line;
  int image_bits_per_pixel;
  fread(reinterpret_cast<char*>(&image_xoffset),sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_format),sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_byte_order),sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bitmap_unit),sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bitmap_bit_order),
    sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bitmap_pad),
    sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_depth),
    sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bytes_per_line),
    sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bits_per_pixel),
    sizeof(int),1,is);

#ifdef DEBUG
  CHECK_TEST(ImageByteOrder(display)==image_byte_order);
  CHECK_TEST(BitmapUnit(display)==image_bitmap_unit);
  CHECK_TEST(BitmapBitOrder(display)==image_bitmap_bit_order);
  CHECK_TEST(BitmapBitOrder(display)==image_bitmap_bit_order);

  int number_depths;
  int *depth_list=XListDepths(display,screen,&number_depths);
  if (depth_list) {
    BOOLEAN found=FALSE;
    for (int i=0;i<number_depths;i++) {
      if (image_depth==depth_list[i]) found=TRUE;
    }
    CHECK_TEST(found);
  }
  XFree(depth_list);

  CHECK_TEST(image_width<=W);
  CHECK_TEST(image_height<=H);
#endif

  image=XCreateImage(display,0,image_depth,image_format,
    image_xoffset,0,image_width,image_height,image_bitmap_pad,
    image_bytes_per_line);
  ASSERT(image!=0);
  image->data=0;

  fread(reinterpret_cast<char*>(&image->red_mask),
    sizeof(unsigned long),1,is);
  fread(reinterpret_cast<char*>(&image->green_mask),
    sizeof(unsigned long),1,is);
  fread(reinterpret_cast<char*>(&image->blue_mask),
    sizeof(unsigned long),1,is);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::writeRaster(FILE *os) {
  CHECK_POINTER(image);
  CHECK_POINTER(os);
  XDestroyImage(image);
  image=XGetImage(display,XtWindow(glarea),0,0,W,H,AllPlanes,ZPixmap); 
  ASSERT(image!=0);

  int buffer_size=(image->bytes_per_line)*(image->height);
  if (image->format!=ZPixmap) buffer_size *= image->depth;
  fwrite(reinterpret_cast<char*>(&buffer_size), sizeof(int), 1, os);
  fwrite( image->data, buffer_size, 1, os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::readRaster(FILE *is) {
#ifdef DEBUG
  CHECK_TEST(image!=0);
  CHECK_TEST(is!=0);
# ifdef ALPHA
//alpha crashes without the following
  cout << "\tis position = " << ftell(is) << endl;
# endif
#else
  if (!image) return;
  if (!is) return;
#endif
  ASSERT(display!=0);
  
  XGCValues gcvalues;
  gcvalues.foreground=BlackPixel(display,screen);
  gcvalues.background=WhitePixel(display,screen);
  gcvalues.line_width=1;               // also
  gcvalues.line_style=LineSolid;       //   see
  gcvalues.cap_style=CapRound;         //     XSetLineAttributes
  gcvalues.join_style=JoinRound;       //       .
  unsigned long mask= GCForeground | GCBackground | GCLineWidth
    | GCLineStyle | GCCapStyle | GCJoinStyle | GCFont;
  GC gc = XCreateGC(display,XtWindow(glarea),mask,&gcvalues);

  size_t buffer_size=(image->bytes_per_line)*(image->height);
  if (image->format!=ZPixmap) buffer_size *= image->depth;
  size_t bs;
  char *buffer=OPERATOR_NEW_BRACKET(char,buffer_size);
  image->data=buffer;
  while (TRUE) {
    if (fread(reinterpret_cast<char*>(&bs),sizeof(int),1,is)<1) break;
#ifdef DEBUG
# ifdef ALPHA
//alpha crashes without the following
    cout << "\tis position = " << ftell(is) << endl;
# endif
    CHECK_TEST(buffer_size==bs);
#else
    if (buffer_size!=bs) break;
#endif

    if (fread( buffer, 1, bs, is)<bs) break;

//  if (double_buffer) GLwDrawingAreaSwapBuffers(glarea);
    XPutImage(display,XtWindow(glarea),gc,image,0,0,0,0,W,H); 
    flush(TRUE);
  }
  delete buffer;
  XFreeGC(display,gc);
  image->data=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::closeRaster() {
  CHECK_POINTER(image);
  XDestroyImage(image);
  image=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::printOn(ostream &os) const {
  os << "GLWindow: "
//   << "\n\trwidth,rheight = " << rwidth << " " << rheight
//   << "\n\tcursor = " << cursor
//   << "\n\tbound = " << bound
//   << "\n\tradius = " << radius
//   << "\n\tactive_clip_direction = " << active_clip_direction
//   << "\n\tactive_clip_hand = " << active_clip_hand
     << endl;
//os << "\n\tperspective = " << PRINT_BOOLEAN(perspective)
//   << "\n\tthree_dimensional = " << PRINT_BOOLEAN(three_dimensional)
//   << "\n\tbutton_pressed = " << PRINT_BOOLEAN(button_pressed)
//   << "\n\tnew_model = " << PRINT_BOOLEAN(new_model)
//   << "\n\tmat_valid = " << PRINT_BOOLEAN(mat_valid)
//   << "\n\tmoved = " << PRINT_BOOLEAN(moved)
//   << "\n\tcursor_selected = " << PRINT_BOOLEAN(cursor_selected)
//   << "\n\tbest_list_exposed = " << PRINT_BOOLEAN(best_list_exposed)
//   << endl;
//os << "\n\tplot_obj = " << (void*) plot_obj
//   << "\n\tbounding_box_list = " << bounding_box_list
//   << "\n\tmouse_list = " << mouse_list
//   << "\n\tuser_list = " << user_list
//   << "\n\tbest_list = " << best_list
//   << endl;
  VirtualGLWindow::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::drawBox(const Vector3 *low,const Vector3 *high,
const GLenum *mode,const GLenum *type) const {
  CHECK_TEST(*low<*high)
  switch (*type) {
    case GL_POINTS: // normal along radius from center
      { PrimitiveDelimiter pd(*type);
        glNormal3f(-ONE,-ONE,-ONE);
        glVertex3d((*low) [0],(*low) [1],(*low) [2]);
        glNormal3f( ONE,-ONE,-ONE);
        glVertex3d((*high)[0],(*low) [1],(*low) [2]);

        glNormal3f(-ONE, ONE,-ONE);
        glVertex3d((*low) [0],(*high)[1],(*low) [2]);
        glNormal3f( ONE, ONE,-ONE);
        glVertex3d((*high)[0],(*high)[1],(*low) [2]);

        glNormal3f(-ONE,-ONE, ONE);
        glVertex3d((*low) [0],(*low) [1],(*high)[2]);
        glNormal3f( ONE,-ONE, ONE);
        glVertex3d((*high)[0],(*low) [1],(*high)[2]);

        glNormal3f(-ONE, ONE, ONE);
        glVertex3d((*low) [0],(*high)[1],(*high)[2]);
        glNormal3f( ONE, ONE, ONE);
        glVertex3d((*high)[0],(*high)[1],(*high)[2]);
      }
      break;
    case GL_LINES:
      { PrimitiveDelimiter pd(*type);
        { glNormal3f(ZERO,-ONE,-ONE);
          glVertex3d((*low) [0],(*low) [1],(*low) [2]);
          glVertex3d((*high)[0],(*low) [1],(*low) [2]);

          glNormal3f(ZERO, ONE,-ONE);
          glVertex3d((*low) [0],(*high)[1],(*low) [2]);
          glVertex3d((*high)[0],(*high)[1],(*low) [2]);

          glNormal3f(ZERO,-ONE, ONE);
          glVertex3d((*low) [0],(*low) [1],(*high)[2]);
          glVertex3d((*high)[0],(*low) [1],(*high)[2]);

          glNormal3f(ZERO, ONE, ONE);
          glVertex3d((*low) [0],(*high)[1],(*high)[2]);
          glVertex3d((*high)[0],(*high)[1],(*high)[2]);
        }
        { glNormal3f(-ONE,ZERO,-ONE);
          glVertex3d((*low) [0],(*low) [1],(*low) [2]);
          glVertex3d((*low) [0],(*high)[1],(*low) [2]);

          glNormal3f( ONE,ZERO,-ONE);
          glVertex3d((*high)[0],(*low) [1],(*low) [2]);
          glVertex3d((*high)[0],(*high)[1],(*low) [2]);

          glNormal3f(-ONE,ZERO, ONE);
          glVertex3d((*low) [0],(*low) [1],(*high)[2]);
          glVertex3d((*low) [0],(*high)[1],(*high)[2]);

          glNormal3f( ONE,ZERO, ONE);
          glVertex3d((*high)[0],(*low) [1],(*high)[2]);
          glVertex3d((*high)[0],(*high)[1],(*high)[2]);
        }
        { glNormal3f(-ONE,-ONE,ZERO);
          glVertex3d((*low) [0],(*low) [1],(*low) [2]);
          glVertex3d((*low) [0],(*low) [1],(*high)[2]);

          glNormal3f( ONE,-ONE,ZERO);
          glVertex3d((*high)[0],(*low) [1],(*low) [2]);
          glVertex3d((*high)[0],(*low) [1],(*high)[2]);

          glNormal3f(-ONE, ONE,ZERO);
          glVertex3d((*low) [0],(*high)[1],(*low) [2]);
          glVertex3d((*low) [0],(*high)[1],(*high)[2]);

          glNormal3f( ONE, ONE,ZERO);
          glVertex3d((*high)[0],(*high)[1],(*low) [2]);
          glVertex3d((*high)[0],(*high)[1],(*high)[2]);
        }
      }
      break;
    case GL_POLYGON:
    case GL_TRIANGLE_STRIP:
    case GL_TRIANGLE_FAN:
    case GL_QUADS:
      { QuadDelimiter qd;
        { glNormal3f(-ONE,ZERO,ZERO);
          if (*mode==GL_SELECT) glLoadName(0); // left
          glVertex3d((*low) [0],(*low) [1],(*low) [2]);
          glVertex3d((*low) [0],(*high)[1],(*low) [2]);
          glVertex3d((*low) [0],(*high)[1],(*high)[2]);
          glVertex3d((*low) [0],(*low) [1],(*high)[2]);
          if (*mode==GL_FEEDBACK) glPassThrough(1.0); // left->right
          glNormal3f( ONE,ZERO,ZERO);
          if (*mode==GL_SELECT) glLoadName(1); // right
          glVertex3d((*high)[0],(*low) [1],(*low) [2]);
          glVertex3d((*high)[0],(*high)[1],(*low) [2]);
          glVertex3d((*high)[0],(*high)[1],(*high)[2]);
          glVertex3d((*high)[0],(*low) [1],(*high)[2]);
        }
        if (*mode==GL_FEEDBACK) glPassThrough(2.0); // right->bottom
        { glNormal3f(ZERO,-ONE,ZERO);
          if (*mode==GL_SELECT) glLoadName(2); // bottom
          glVertex3d((*low)[0] ,(*low) [1],(*low) [2]);
          glVertex3d((*high)[0],(*low) [1],(*low) [2]);
          glVertex3d((*high)[0],(*low) [1],(*high)[2]);
          glVertex3d((*low)[0] ,(*low) [1],(*high)[2]);
          if (*mode==GL_FEEDBACK) glPassThrough(3.0); // bottom->top
          glNormal3f(ZERO, ONE,ZERO);
          if (*mode==GL_SELECT) glLoadName(3); // top
          glVertex3d((*low) [0],(*high)[1],(*low) [2]);
          glVertex3d((*high)[0],(*high)[1],(*low) [2]);
          glVertex3d((*high)[0],(*high)[1],(*high)[2]);
          glVertex3d((*low) [0],(*high)[1],(*high)[2]);
        }
        if (*mode==GL_FEEDBACK) glPassThrough(4.0); // top->back
        { glNormal3f(ZERO,ZERO,-ONE);
          if (*mode==GL_SELECT) glLoadName(4); // back
          glVertex3d((*low) [0],(*low) [1],(*low) [2]);
          glVertex3d((*high)[0],(*low) [1],(*low) [2]);
          glVertex3d((*high)[0],(*high)[1],(*low) [2]);
          glVertex3d((*low) [0],(*high)[1],(*low) [2]);
          if (*mode==GL_FEEDBACK) glPassThrough(5.0); // back->front
          glNormal3f(ZERO,ZERO, ONE);
          if (*mode==GL_SELECT) glLoadName(5); // front
          glVertex3d((*low) [0],(*low) [1],(*high)[2]);
          glVertex3d((*high)[0],(*low) [1],(*high)[2]);
          glVertex3d((*high)[0],(*high)[1],(*high)[2]);
          glVertex3d((*low) [0],(*high)[1],(*high)[2]);
        }
      }
      break;
    case GL_TRIANGLES:
      { PrimitiveDelimiter pd(*type);
        { glNormal3f(-ONE,ZERO,ZERO);
          glVertex3d((*low) [0],(*low) [1],(*low) [2]);
          glVertex3d((*low) [0],(*high)[1],(*low) [2]);
          glVertex3d((*low) [0],(*low) [1],(*high)[2]);

          glVertex3d((*low) [0],(*high)[1],(*high)[2]);
          glVertex3d((*low) [0],(*low) [1],(*high)[2]);
          glVertex3d((*low) [0],(*high)[1],(*low) [2]);

          glNormal3f( ONE,ZERO,ZERO);
          glVertex3d((*high)[0],(*low) [1],(*low) [2]);
          glVertex3d((*high)[0],(*high)[1],(*low) [2]);
          glVertex3d((*high)[0],(*low) [1],(*high)[2]);

          glVertex3d((*high)[0],(*high)[1],(*high)[2]);
          glVertex3d((*high)[0],(*low) [1],(*high)[2]);
          glVertex3d((*high)[0],(*high)[1],(*low) [2]);
        }
        { glNormal3f(ZERO,-ONE,ZERO);
          glVertex3d((*low)[0] ,(*low) [1],(*low) [2]);
          glVertex3d((*high)[0],(*low) [1],(*low) [2]);
          glVertex3d((*low)[0] ,(*low) [1],(*high)[2]);

          glVertex3d((*high)[0],(*low) [1],(*high)[2]);
          glVertex3d((*low)[0] ,(*low) [1],(*high)[2]);
          glVertex3d((*high)[0],(*low) [1],(*low) [2]);

          glNormal3f(ZERO, ONE,ZERO);
          glVertex3d((*low) [0],(*high)[1],(*low) [2]);
          glVertex3d((*high)[0],(*high)[1],(*low) [2]);
          glVertex3d((*low) [0],(*high)[1],(*high)[2]);

          glVertex3d((*high)[0],(*high)[1],(*high)[2]);
          glVertex3d((*low) [0],(*high)[1],(*high)[2]);
          glVertex3d((*high)[0],(*high)[1],(*low) [2]);
        }
        { glNormal3f(ZERO,ZERO,-ONE);
          glVertex3d((*low) [0],(*low) [1],(*low) [2]);
          glVertex3d((*high)[0],(*low) [1],(*low) [2]);
          glVertex3d((*low) [0],(*high)[1],(*low) [2]);

          glVertex3d((*high)[0],(*high)[1],(*low) [2]);
          glVertex3d((*low) [0],(*high)[1],(*low) [2]);
          glVertex3d((*high)[0],(*low) [1],(*low) [2]);

          glNormal3f(ZERO,ZERO, ONE);
          glVertex3d((*low) [0],(*low) [1],(*high)[2]);
          glVertex3d((*high)[0],(*low) [1],(*high)[2]);
          glVertex3d((*low) [0],(*high)[1],(*high)[2]);

          glVertex3d((*high)[0],(*high)[1],(*high)[2]);
          glVertex3d((*low) [0],(*high)[1],(*high)[2]);
          glVertex3d((*high)[0],(*low) [1],(*high)[2]);
        }
      }
      break;
    case GL_LINE_STRIP:
    case GL_LINE_LOOP: // normal is bogus
      { PrimitiveDelimiter pd(*type);
        glNormal3f(ZERO,ZERO,-ONE);
        glVertex3d((*low) [0],(*low) [1],(*low) [2]);
        glVertex3d((*low) [0],(*high)[1],(*low) [2]);
        glVertex3d((*low) [0],(*high)[1],(*high)[2]);
        glVertex3d((*low) [0],(*low) [1],(*high)[2]);
        glVertex3d((*low) [0],(*low) [1],(*low) [2]);
        glVertex3d((*high)[0],(*low) [1],(*low) [2]);
        glVertex3d((*high)[0],(*low) [1],(*high)[2]);
        glVertex3d((*low) [0],(*low) [1],(*high)[2]);
      }
      { PrimitiveDelimiter pd(*type);
        glNormal3f(ZERO,ZERO, ONE);
        glVertex3d((*high)[0],(*high)[1],(*high)[2]);
        glVertex3d((*high)[0],(*low) [1],(*high)[2]);
        glVertex3d((*high)[0],(*low) [1],(*low) [2]);
        glVertex3d((*high)[0],(*high)[1],(*low) [2]);
        glVertex3d((*high)[0],(*high)[1],(*high)[2]);
        glVertex3d((*low) [0],(*high)[1],(*high)[2]);
        glVertex3d((*low) [0],(*high)[1],(*low) [2]);
        glVertex3d((*high)[0],(*high)[1],(*low) [2]);
      }
      break;
    case GL_QUAD_STRIP: // normal is bogus
      { PrimitiveDelimiter pd(*type);
        glNormal3f(ZERO,ZERO,-ONE);
        glVertex3d((*low) [0],(*low) [1],(*low) [2]);
        glVertex3d((*low) [0],(*low) [1],(*high)[2]);
        glVertex3d((*low) [0],(*high)[1],(*low) [2]);
        glVertex3d((*low) [0],(*high)[1],(*high)[2]);
        glVertex3d((*high)[0],(*high)[1],(*low) [2]);
        glVertex3d((*high)[0],(*high)[1],(*high)[2]);
        glVertex3d((*high)[0],(*low) [1],(*low) [2]);
        glVertex3d((*high)[0],(*low) [1],(*high)[2]);
      }
      break;
    default:
      break;
  }
}



#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
GLWindow::SlideBar::SlideBar(Widget glarea,Dimension H) : 
var_min(HUGE_VAL),var_max(-HUGE_VAL),frac(HALF),
sig_digit(UNDEFINED),slide_bar(0),decimal_points(0) {
  Widget toplevel=XtParent(glarea);
  XmString title=
    XmStringCreateLocalized(const_cast<char*>("Iso-surface Value"));
  slide_bar = XtVaCreateManagedWidget ( const_cast<char*>("slide_bar"),
    xmScaleWidgetClass,  toplevel,
    XmNtopWidget,        glarea,
    XmNtopAttachment,    XmATTACH_WIDGET,
    XmNbottomAttachment, XmATTACH_FORM,
    XmNleftAttachment,   XmATTACH_FORM,
    XmNrightAttachment,  XmATTACH_FORM,
    XmNshowValue,        TRUE,
    XmNtitleString,      title,
    XmNorientation,      XmHORIZONTAL,
    XmNrightOffset,      20,
    XmNleftOffset,       20,
    XmNtraversalOn,      True,
    XmNheight,           H/10,
    0 );
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::SlideBar::rescale(REAL low,REAL high) {
  CHECK_TEST(low<high)
  var_min=low;
  var_max=high;
  decimal_points=short(floor(log10(var_max-var_min)));
  sig_digit=pow(10.,decimal_points);
  int slide_min=0,slide_max=1;
  if (decimal_points<0) {
    slide_max=static_cast<int>(ceil(var_max/sig_digit));
    slide_min=static_cast<int>(floor(var_min/sig_digit));
    decimal_points=-decimal_points+1;
  } else {
    slide_max=static_cast<int>(ceil(var_max));
    slide_min=static_cast<int>(floor(var_min));
    decimal_points=1;
  }
  XtVaSetValues(slide_bar,
    XmNminimum, slide_min*10,
    XmNmaximum, slide_max*10,
    XmNvalue,   slide_min*10,
    XmNdecimalPoints,decimal_points,
    0);
  REAL v=var_min+frac*(var_max-var_min);
  int slide_val=0;
  if (decimal_points>0) slide_val=static_cast<int>(ceil(v/sig_digit));
  else slide_val=static_cast<int>(ceil(v));
  XtVaSetValues(slide_bar,
    XmNvalue, slide_val,
    0);
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::SlideBar::setValue(REAL v) {
  v=max(var_min,min(var_max,v));
  frac=(v-var_min)/(var_max-var_min);
  int slide_val=0;
  if (decimal_points>0) slide_val=static_cast<int>(ceil(v/sig_digit));
  else slide_val=static_cast<int>(ceil(v));
  XtVaSetValues(slide_bar,
    XmNvalue, slide_val*10,
    0);
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
REAL GLWindow::SlideBar::getValue(int scale_struct_val) const {
  if (decimal_points>0) {
    return static_cast<REAL>(scale_struct_val)*sig_digit*0.1;
  }
  return static_cast<REAL>(scale_struct_val);
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::SlideBar::setCallback(XtCallbackProc callback,
XtPointer client_data) const {
  XtAddCallback(slide_bar,
    XmNvalueChangedCallback, callback,
    client_data);
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::SlideBar::removeCallback(XtCallbackProc callback,
XtPointer client_data) const {
  XtCallbackStatus status=
    XtHasCallbacks(slide_bar,XmNvalueChangedCallback);
  if (status==XtCallbackHasSome) {
    XtRemoveCallback(slide_bar,
      XmNvalueChangedCallback, callback,
      client_data);
  }
}
#endif
#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::SlideBar::printOn(ostream &os) const {
  os << "GLWindow::SlideBar: slide_bar = " << slide_bar << endl;
  os << "\tvar_min = " << var_min << endl;
  os << "\tvar_max = " << var_max << endl;
  os << "\tfrac = " << frac << endl;
  os << "\tsig_digit = " << sig_digit << endl;
  os << "\tdecimal_points = " << decimal_points << endl;
}
#endif

#if (SPACEDIM==3)
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
GLWindow::ToggleButtonArray::ToggleButtonArray(Widget glarea,int n) : 
toggle_button_array(n),client_data_save(0) {
  Widget toplevel=XtParent(glarea);
  XmString title=
    XmStringCreateLocalized(const_cast<char*>("Iso-Surface Numbers"));
  XmString st=XmStringCreateSimple(const_cast<char*>("0"));
  toggle_button_array[0]=
    XtVaCreateManagedWidget(const_cast<char*>("toggle_button"),
    xmToggleButtonWidgetClass, toplevel,
    XmNtitleString,            title,
    XmNlabelString,            st,
    XmNtopWidget,              glarea,
    XmNtopAttachment,          XmATTACH_WIDGET,
    XmNbottomAttachment,       XmATTACH_FORM,
    XmNleftAttachment,         XmATTACH_FORM,
    XmNrightAttachment,        XmATTACH_NONE,
    XmNindicatorOn,            True,
    XmNvisibleWhenOff,         True,
    XmNset,                    True, 0);
  XmStringFree(st);
  for (int i=1;i<n-1;i++) {
    char label[LABEL_LENGTH];
    snprintf(label,LABEL_LENGTH,"%d",i);
    XmString st=XmStringCreateSimple(const_cast<char*>(label));
    toggle_button_array[i]=
      XtVaCreateManagedWidget(const_cast<char*>("toggle_button"),
      xmToggleButtonWidgetClass, toplevel,
      XmNtitleString,            title,
      XmNlabelString,            st,
      XmNtopWidget,              glarea,
      XmNtopAttachment,          XmATTACH_WIDGET,
      XmNbottomAttachment,       XmATTACH_FORM,
      XmNleftWidget,             toggle_button_array[i-1],
      XmNleftAttachment,         XmATTACH_WIDGET,
      XmNrightAttachment,        XmATTACH_NONE,
      XmNindicatorOn,            True,
      XmNvisibleWhenOff,         True,
      XmNset,                    True, 0);
    XmStringFree(st);
  }
  char label[LABEL_LENGTH];
  snprintf(label,LABEL_LENGTH,"%d%",n-1);
  st=XmStringCreateSimple(const_cast<char*>(label));
  toggle_button_array[n-1]=XtVaCreateManagedWidget("toggle_button",
    xmToggleButtonWidgetClass, toplevel,
    XmNtitleString,            title,
    XmNlabelString,            st,
    XmNtopWidget,              glarea,
    XmNtopAttachment,          XmATTACH_WIDGET,
    XmNbottomAttachment,       XmATTACH_FORM,
    XmNleftWidget,             toggle_button_array[n-2],
    XmNleftAttachment,         XmATTACH_WIDGET,
    XmNrightAttachment,        XmATTACH_NONE,
    XmNindicatorOn,            True,
    XmNvisibleWhenOff,         True,
    XmNset,                    True, 0);
  XmStringFree(st);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void GLWindow::ToggleButtonArray::printOn(ostream &os) const {
  os << "GLWindow::ToggleButtonArray: "
     << "\n\tnumberSurfaces = " << numberSurfaces()
     << "\n\tclient_data_save = " << client_data_save
     << endl;
  for (int i=0;i<numberSurfaces();i++) {
    os << "\t drawSurface(" << i << ") = "
       << PRINT_BOOLEAN(drawSurface(i)) << endl;
  }
}
#endif

#if (SPACEDIM==3)
#include "NumPtr.C"
INSTANTIATE_NUMPTR(Widget)
#endif
#endif
#endif
