//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#ifndef GLWINDOW_H
#define GLWINDOW_H "$Header: /home/faculty/johnt/cvs/deal_new/graphics/GLWindow.H,v 1.1 2009/08/20 17:31:46 johnt Exp $"
#ifndef USE_GTK
#if (SPACEDIM>1)
#include <iostream>
#include <X11/Intrinsic.h>
#include <Xm/DrawingA.h>
#include <Xm/DrawnB.h>
#include <Xm/Form.h> 
#include <Xm/Label.h>
#include <Xm/Scale.h>
#include <Xm/ScrollBar.h>
#include <Xm/RowColumn.h>
#include <Xm/ToggleB.h>
#include <Xm/XmP.h>
#include <GL/gl.h>
#if defined (__digital__) && defined (__unix__)
#else
#include <GL/glut.h>
#endif
#include <GL/glx.h>
#include "Const.H"
#include "Line.H"
#include "MemoryDebugger.H"
#include "NumPtr.H"
#include "Pair.H"
#include "Palette.H"
#include "Quaternion.H"
#include "ToggleCallback.H"
#include "Tracer.H"
#include "Vector3.H"
#include "VGT.H"
#include "VirtualGLWindow.H"

// middle button: clip
// left button: rotate
// right button: cursor position
// extra widgets: level surface selection, colormap editor

/*
//use of GLTracer can make code perform incorrectly:
//  e.g. in GLWindow::makeCurrent
class GLTracer : public Tracer {
  private:
    void clearErrors() const;
  public:
    GLTracer(const char *str) : Tracer(str) { 
//    clearErrors(); 
    }
    GLTracer(const char *str,const char *f,int l) : Tracer(str,f,l) { 
//    clearErrors(); 
    }
    ~GLTracer() { 
//    clearErrors(); 
    }
};
#define GLTRACER_CALL(n,s) GLTracer n(s,__FILE__,__LINE__);
*/

class GLWindow : public VirtualGLWindow {
  private:
  public:
    class AttributeDelimiter;
    class CullDelimiter;
    class DisablerDelimiter;
    class FeedbackRecord;
    class FeedbackDelimiter;
    class HitRecord;
    class LineDelimiter;
    class ListDelimiter;
    class MatrixDelimiter;
    class PrimitiveDelimiter;
    class QuadDelimiter;
    class SelectDelimiter;
#if (SPACEDIM==3)
    class SlideBar;
    class ToggleButtonArray;
    enum SURFACE_NORMAL_DIRECTION{INCREASING_FUNCTION,
      DECREASING_FUNCTION,INCREASING_OR_DECREASING_FUNCTION};
#endif
    class TriangleDelimiter;
  private:
    Widget topshell,graph_tool;
    Widget glarea;
    XtAppContext appcontext;

    Display *display;
    XImage *image;
#if (SPACEDIM==3)
    SlideBar *slide_bar;
    ToggleButtonArray *toggle_button_array;
    SURFACE_NORMAL_DIRECTION surface_normal_direction;
#endif
    
    Display *dpy;
    int screen;
    Colormap cmap;
    XVisualInfo *vi;
    GLXContext cx;
    GLboolean  double_buffer;

    Dimension W,H;
    REAL rwidth,rheight;

    GLfloat radius; // radius of the boundary box
    
    unsigned int active_clip_direction;
    HAND active_clip_hand;
    Line closest_edge;

    Pair<GLdouble> wuvec[3];

    BOOLEAN use_lighting,perspective,three_dimensional;
    BOOLEAN button_pressed;
    BOOLEAN break_flag,break_mouse_flag;
    BOOLEAN finding_mouse;
    BOOLEAN best_list_exposed;
    BOOLEAN show_bounding_box;

    Vector3 cursor;

    Quaternion<REAL> curquat;
    Pair<GLdouble> old_screen_coord;
    BOOLEAN new_model,mat_valid,moved,cursor_selected;
    GLdouble ModelMat[16], ProjMat[16];
    GLint viewport[4];
    
    char FontName[100];
    XFontStruct *font;
    GLuint fontbase;
    int FontHeight;
    int FontDescent;

    MyDisplayList *bounding_box_list,*mouse_list,*user_list; 
    MyDisplayList *best_list;
    Pair<GLdouble> xyToScreen(int x,int y) const {
      return Pair<GLdouble>(x,H-y);
    }
    Pair<GLdouble> screenCoords(const Vector3&);

    GLdouble screenToX(Pair<GLdouble> screen) const {
      return screen[0];
    }
//  using H-screen[1] was inconsistent with GL's gluProject
    GLdouble screenToY(Pair<GLdouble> screen) const {
      return rheight-screen[1];
    }
    Line objectCoords(Pair<GLdouble>);

    Vector3 clipCoord(const Vector3&);

//  see Open GL Ref Manual, p. 427
    Pair<GLdouble> screenToUnitCircle(Pair<GLdouble> screen) const {
      return Pair<GLdouble>(TWO*screen[0]/rwidth-ONE,
                            TWO*screen[1]/rheight-ONE);
    }
    void clearErrors() const;
    void markFaceLocation(const Pair<GLdouble>&);
    int textWidth(char *str) const {
      int dir,ascent,descent;
      XCharStruct overall;
      XTextExtents(font,str,strlen(str),&dir,&ascent,&descent,&overall);
      return overall.width;
    }
    void ginit();
    Widget createMenubar();
    void input(XEvent*);
    void resize();
    void makeBoundingBox();
    void recalcModelView();

    void makeCurrent();
    
    void selectClipPlane(const Pair<GLdouble>&);
    void moveClipPlane(const Pair<GLdouble>&);
    
    void makeRasterFont();

    Vector3 faceLocation(const Pair<GLdouble>&);
    void crossHairs(const Vector3&);
    void rotateImage(const Pair<GLdouble>&);
    void moveCursor(const Pair<GLdouble>&);
    void updateMat();
      
    void buttonPress(const XButtonEvent&,const Pair<GLdouble>&);
    void buttonRelease(const XButtonEvent&,const Pair<GLdouble>&);

    static void breakoutCallback(Widget w,XtPointer client_data,
    XtPointer) {
      GLWindow *obj=reinterpret_cast<GLWindow*>(client_data);
      if (obj) obj->breakout(w);
    }
    void breakout(Widget) { break_flag=TRUE; }
    static void quitCallback(Widget w,XtPointer client_data,
    XtPointer) {
      GLWindow *obj=reinterpret_cast<GLWindow*>(client_data);
      if (obj) obj->quit(w);
    }
    void quit(Widget);
    static void YesCallback(Widget,XtPointer client_data,XtPointer) {
      int *result=reinterpret_cast<int*>(client_data);
      *result=TRUE;
    }

  public:
    GLWindow(const char*,const Vector3&,BOOLEAN,REAL=HALF,REAL=HALF,
      BOOLEAN=FALSE,BOOLEAN=FALSE,BOOLEAN=TRUE,BOOLEAN=FALSE);
    ~GLWindow();
    void eventLoop(BOOLEAN);
    XColor getColor(const char *str) const {
      XColor xc_exact,xc_screen;
      Status st=XLookupColor(dpy,cmap,str,&xc_exact,&xc_screen);
      return xc_screen;
    }
    void rescale(const Vector3&,const Vector3&);
    int getWidth() const { return static_cast<int>(W); }
    int getHeight() const { return static_cast<int>(H); }

    void setfgColor(const XColor &xc) const {
      glColor3us(xc.red,xc.green,xc.blue);
    }
    void setfgColor(const char *str) const {
      GLWindow *copy=const_cast<GLWindow*>(this);
      copy->setfgColor(getColor(str));
    }
    void setfgColor(GLclampf r,GLclampf g,GLclampf b) const {
      glColor3f(r,g,b);
    }

    void setbgColor(GLclampf r,GLclampf g,GLclampf b,GLclampf a=0.)
    const {
      glClearColor(r,g,b,a);
    }
    void setbgColor(const XColor &xc) {
      setbgColor(xc.red,xc.green,xc.blue);
    }
    void setbgColor(const char *str) const {
      GLWindow *copy=const_cast<GLWindow*>(this);
      copy->setbgColor(getColor(str));
    }

    void setLineWidth(const INTEGER *lw) {
//    see glGetFloatv(GL_LINE_WIDTH_RANGE)
      glLineWidth(static_cast<GLfloat>(*lw));
    }
    void setLineWidth(const INTEGER *lw) const {
      glLineWidth(static_cast<GLfloat>(*lw));
    }
    void setPointSize(int ps) const {
//    see glGetFloatv(GL_POINT_SIZE_RANGE)
      glPointSize(static_cast<GLfloat>(ps));
    }
    void setPolygonMode(GLenum face=GL_FRONT_AND_BACK,
    GLenum draw_mode=GL_FILL,GLenum face_mode=GL_CCW) const {
      CHECK_TEST(face==GL_FRONT_AND_BACK || face==GL_FRONT ||face==GL_BACK);
      CHECK_TEST(draw_mode==GL_POINT || draw_mode==GL_LINE || draw_mode==GL_FILL);
      CHECK_TEST(face_mode==GL_CCW || face_mode==GL_CW);
      glPolygonMode(face,draw_mode);
      glFrontFace(face_mode);
    }
    void newPage() { 
      glClear(GL_COLOR_BUFFER_BIT); 
    }
    void flush(BOOLEAN) { 
      glFlush(); 
    }
    void putString(const Vector3 *window_coords,const char *str) {
      glRasterPos3f((*window_coords)[0],(*window_coords)[1],
                    (*window_coords)[2]);
      printString(str);
    }
    void drawBox(const Vector3*,const Vector3*,const GLenum*,
      const GLenum*) const;
    void expose();
    
//  Can use the same visual and same context in other GL widgets
//    to share colormap and display lists
    const XVisualInfo *getVisual() { return vi; }
    GLXContext getContext() { return cx; }

    void drawLine(const Vector3*,const Vector3*) const;
    void colorRect(const XColor*,const Vector3*,const Vector3*,
      const int*) const;
    void colorRect(const char *str,const Vector3 &l,const Vector3 &h,
    int d) const {
      XColor xc=getColor(str);
      colorRect(&xc,&l,&h,&d);
    }
    void colorTriangle(const XColor*,const Vector3*,const Vector3*,
      const Vector3*) const;
    void drawVector(const Vector3*,const Vector3*) const;
#if (SPACEDIM==3)
    void drawVector(const Vector3*,const Vector3*,const AxisClipPlane*) 
      const;
#endif

    void printVisual();
    void printString(const char*);

    GLuint getDisplayList(BOOLEAN do_best) { 
      if (do_best || rotate_best) {
        if (!best_list) best_list=OPERATOR_NEW MyDisplayList();
        return best_list->getBase(); 
      }
      if (!user_list) user_list=OPERATOR_NEW MyDisplayList();
      return user_list->getBase(); 
    }
    void freeDisplayList(BOOLEAN do_best) { 
      if (do_best || rotate_best) {
        if (best_list) delete best_list; best_list=0;
      } else {
        if (user_list) delete user_list; user_list=0;
      }
    }
    void clearDisplayLists() {
      freeDisplayList(FALSE);
      freeDisplayList(TRUE);
    }
    
    const Vector3& getCursor() const { return cursor; } 

    void initialize(const VirtualGLWindow::PlotObj*);
    void terminatePlotObj();
    void plotClipPlanes() const;

    void plotObject(BOOLEAN);

    Display* getDisplay() const {return dpy; };
    int numberSurfaces() const;
#if (SPACEDIM==3)
    GLWindow::SlideBar* slideBar();
    BOOLEAN hasSlideBar() const { return slide_bar!=0; }
    void deleteSlideBar();
    GLWindow::ToggleButtonArray* toggleButtonArray(int=2);
    BOOLEAN hasToggleButtonArray() const { 
      return toggle_button_array!=0; 
    }
    void deleteToggleButtonArray();
    void setSurfaceNormalDirection(SURFACE_NORMAL_DIRECTION snd) {
      surface_normal_direction=snd;
    }
#endif

    void startDrawing(BOOLEAN do_best) {
      CHECK_TEST(!isDrawing())
      VirtualGLWindow::startDrawing(do_best);
      glNewList(getDisplayList(do_best || rotate_best),GL_COMPILE);
      if (use_lighting) {
        glEnable(GL_LIGHTING);
        glEnable(GL_LIGHT0);
      }
    }
    void finishDrawing() {
      CHECK_TEST(isDrawing())
      if (use_lighting) {
        glDisable(GL_LIGHT0);
        glDisable(GL_LIGHTING);
      }
      VirtualGLWindow::finishDrawing();
      glEndList();
    }

    static void resizeCallback(Widget,XtPointer client_data,XtPointer) {
      GLWindow *win=reinterpret_cast<GLWindow*>(client_data);
      if (win) win->resize();
    }
    static void inputCallback(Widget,XtPointer client_data,
    XtPointer call_data) {
      GLWindow *win=reinterpret_cast<GLWindow*>(client_data);
      if (win) {
        XmDrawingAreaCallbackStruct *cd=
          reinterpret_cast<XmDrawingAreaCallbackStruct*>(call_data);
        win->input(cd->event);
      }
    }
    static void exposeCallback(Widget,XtPointer client_data,XtPointer) {
      GLWindow *win=reinterpret_cast<GLWindow*>(client_data);
      if (win) win->expose();
    }
    static void ginitCallback(Widget,XtPointer client_data, XtPointer) {
      GLWindow *win=reinterpret_cast<GLWindow*>(client_data);
      if (win) win->ginit();
    }
    Vector3 getMouse();
    BOOLEAN buttonIsPressed(Vector3 &where) const {
      where=cursor; return button_pressed;
    }
    void createRaster(FILE*);
    void writeRaster(FILE*);
    void closeRaster();
    void openRaster(FILE*);
    void readRaster(FILE*);

    void printOn(ostream &os=cerr) const;
};

// p. 360
class GLWindow::HitRecord {
  private:
    GLuint number_names;
    const GLuint *names;
  public:
    HitRecord(const GLuint *pos) : number_names(pos[0]),names(pos+3) {;}
    virtual ~HitRecord() {;}
    GLuint bottomName() const { return names[0]; }
    GLuint topName() const { return names[number_names-1]; }
    HitRecord next() const { return HitRecord(names+number_names); }
    virtual void printOn(ostream &os) const {
      os << "\tGLWindow::HitRecord:" << endl;
    }
};

// p. 378
class GLWindow::FeedbackRecord {
  private:
    GLfloat token;
    const GLfloat *vertices;
  public:
    static GLenum feedback_type;

    FeedbackRecord(const GLfloat *pos):token(pos[0]),vertices(pos+1){;}
    virtual ~FeedbackRecord() {;}
    const GLfloat* position() const { return &token; }
    GLfloat getToken() const { return token; }
    FeedbackRecord next() const { 
      int nvertices=0;
      int extraneous=0;
      if (token==GL_POINT_TOKEN || token==GL_BITMAP_TOKEN ||
      token==GL_DRAW_PIXEL_TOKEN || token==GL_COPY_PIXEL_TOKEN) {
        nvertices=1;
      } else if (token==GL_LINE_TOKEN || token==GL_LINE_RESET_TOKEN) {
        nvertices=2;
      } else if (token==GL_POLYGON_TOKEN) {
        OBSOLETE("what were we trying to do?")
      } else if (token==GL_PASS_THROUGH_TOKEN) extraneous=1;
      switch (feedback_type) {
        case GL_2D:
          return FeedbackRecord(vertices+2*nvertices+extraneous);
        case GL_3D:
          return FeedbackRecord(vertices+3*nvertices+extraneous);
        case GL_3D_COLOR:
        case GL_3D_COLOR_TEXTURE:
        case GL_4D_COLOR_TEXTURE:
        default:
          OBSOLETE("have to determine whether in RGBA or color-index");
          return FeedbackRecord(&token);
      }
    }
    virtual void printOn(ostream &os) const {
      os << "\tGLWindow::BeedbackRecord:" << endl;
    }
};

class StackCount {
  private:
    static unsigned int count;
  public:
    StackCount() { count++; }
    virtual ~StackCount() { 
      CHECK_POSITIVE(count)
      count--;
    }
    virtual void printOn(ostream &os) const {
      os << "\tStackCount:" << endl;
    }
};

class GLWindow::AttributeDelimiter : public StackCount {
  public:
    AttributeDelimiter(GLbitfield mask) { glPushAttrib(mask); }
    virtual ~AttributeDelimiter() { glPopAttrib(); }
};

class GLWindow::MatrixDelimiter : public StackCount {
  public:
    MatrixDelimiter() { glPushMatrix(); }
    virtual ~MatrixDelimiter() { glPopMatrix(); }
};

// GL_COMPILE              = 4864 = 0x1300
// GL_COMPILE_AND_EXECUTE  = 4865 = 0x1301
// GL_RENDER   = 7168 = 0x1C00
// GL_FEEDBACK = 7169 = 0x1C01
// GL_SELECT   = 7170 = 0x1C02
class GLWindow::PrimitiveDelimiter {
  private:
    GLint list_mode,render_mode;
  public:
    PrimitiveDelimiter(GLenum mode) {
      glGetIntegerv(GL_RENDER_MODE,&render_mode);
      glGetIntegerv(GL_LIST_MODE,&list_mode);
      if (render_mode==GL_RENDER && list_mode==GL_COMPILE) 
        glBegin(mode); 
    }
    ~PrimitiveDelimiter() { 
      if (render_mode==GL_RENDER && list_mode==GL_COMPILE) 
        glEnd(); 
    }
};

//eg, for cap=GL_DITHER
class GLWindow::DisablerDelimiter {
  private:
    GLenum capability;
  public:
    DisablerDelimiter(GLenum c) : capability(c) {glDisable(capability);}
    ~DisablerDelimiter() { glEnable(capability); }
};

class GLWindow::LineDelimiter : public GLWindow::PrimitiveDelimiter {
  public:
    LineDelimiter() : PrimitiveDelimiter(GL_LINES) {;}
};

// GL_QUADS = 7
class GLWindow::QuadDelimiter : public GLWindow::PrimitiveDelimiter {
  public:
    QuadDelimiter() : PrimitiveDelimiter(GL_QUADS) {;}
};

class GLWindow::TriangleDelimiter : public GLWindow::PrimitiveDelimiter
{
  public:
    TriangleDelimiter() : PrimitiveDelimiter(GL_TRIANGLES) {;}
};

class GLWindow::FeedbackDelimiter {
  public:
    FeedbackDelimiter() { glRenderMode(GL_FEEDBACK); }
    ~FeedbackDelimiter() { }
};

class GLWindow::SelectDelimiter {
  public:
    SelectDelimiter() {
      glRenderMode(GL_SELECT);
      glInitNames();
      glPushName(UINT_MAX);
    }
    ~SelectDelimiter() { }
};

class GLWindow::ListDelimiter {
  private:
    GLuint active_list;
  public:
    ListDelimiter(GLuint list,GLenum mode) : active_list(list) {
      CHECK_TEST(mode==GL_COMPILE || mode==GL_COMPILE_AND_EXECUTE);
      CHECK_TEST(glIsList(list));
      glNewList(list,mode); 
    }
    ~ListDelimiter() { 
      glEndList(); 
    }
};

#if (SPACEDIM==3)
class GLWindow::SlideBar {
  private:
    REAL frac,var_min,var_max,sig_digit;
    Widget slide_bar;
    short decimal_points;
  public:
    SlideBar(Widget,Dimension);
    void rescale(REAL,REAL);
    void setValue(REAL);
    REAL getValue(int) const;
    void setCallback(XtCallbackProc,XtPointer) const;
    void removeCallback(XtCallbackProc,XtPointer) const;
    void printOn(ostream&) const;
};

class GLWindow::ToggleButtonArray {
  private:
    NumPtr<Widget> toggle_button_array;
    void **client_data_save;
  public:
    ToggleButtonArray(Widget,int);
    int numberSurfaces () const { 
      return toggle_button_array.getNumber(); 
    }
    BOOLEAN drawSurface(int n) const {
      return XmToggleButtonGetState(toggle_button_array[n]);
    }
    template<class T> void setCallback(XtCallbackProc callback,
    T *data) {
      int nt=toggle_button_array.getNumber();
      ToggleButtonArrayCallbackStructure<T> **client_data=
        OPERATOR_NEW_BRACKET(ToggleButtonArrayCallbackStructure<T>*,nt);
      client_data_save=reinterpret_cast<void**>(client_data);
      for (int i=0;i<nt;i++) {
        client_data[i]=
          OPERATOR_NEW ToggleButtonArrayCallbackStructure<T>(data,i);
        XtAddCallback(toggle_button_array[i],
          XmNvalueChangedCallback,callback,client_data[i]);
      }
    }
    template<class T> void removeCallback(T *data,
    XtCallbackProc callback) {
      if (client_data_save!=0) {
        ToggleButtonArrayCallbackStructure<T> **client_data=
          reinterpret_cast< ToggleButtonArrayCallbackStructure<T>** >
          (client_data_save);
        for (int i=0;i<toggle_button_array.getNumber();i++) {
          XtCallbackStatus status=XtHasCallbacks(toggle_button_array[i],
              XmNvalueChangedCallback);
          if (status==XtCallbackHasSome) {
            XtRemoveCallback(toggle_button_array[i],
              XmNvalueChangedCallback,callback,client_data[i]);
          }
          delete client_data[i];
          client_data[i]=0;
        }
        delete [] client_data;
        client_data_save=0;
      }
    }
    void printOn(ostream&) const;
};
#endif
#endif
#endif
#endif
