//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#ifndef VIRTUAL_GRAPH_TOOL_H
#define VIRTUAL_GRAPH_TOOL_H "$Header: /home/faculty/johnt/cvs/deal_new/graphics/VGT.H,v 1.1 2009/08/20 17:31:47 johnt Exp $"
#include <stdio.h>
#include <iostream>
#include <string.h>
#include "Const.H"
#include "Errors.H"
#include "ISLList.H"
#include "MemoryDebugger.H"
#include "VirtualWindow.H"

class VirtualColormap;
class VirtualGraphTool : public ISLListNode {
  private:
    char* name;
  protected:
    VirtualGraphTool(const VirtualGraphTool&) { 
      OBSOLETE("copy constructor disallowed"); 
    }
  public:
    VirtualGraphTool(const char *str) : name(0) {
      name = OPERATOR_NEW_BRACKET(char,strlen(str) + 1); 
      strcpy(name,str);
    }
    virtual ~VirtualGraphTool() { 
      if (name) delete [] name; name=0; 
    }
    virtual const char* getName() const { return name; }
    virtual VirtualWindow* getWindow() { return 0; }
    virtual const VirtualWindow* getWindow() const { return 0; }
    virtual void newPage() { 
      VirtualWindow *w=getWindow();
      CHECK_POINTER(w)
      w->newPage(); 
    }
    virtual void setLineWidth(int lw) { 
      VirtualWindow *w=getWindow();
      CHECK_POINTER(w)
      w->setLineWidth(&lw); 
    }
    virtual void setfgColor(const char *color) {
      VirtualWindow *w=getWindow();
      CHECK_POINTER(w)
      w->setfgColor(color);
    }
    virtual void setbgColor(const char *color) {
      VirtualWindow *w=getWindow();
      CHECK_POINTER(w)
      w->setbgColor(color);
    }
    virtual void flush(BOOLEAN use_buffer=TRUE) { 
      VirtualWindow *w=getWindow();
      CHECK_POINTER(w)
      w->flush(use_buffer); 
    }
    virtual void createRasterFile(BOOLEAN) {;}
    virtual void writeRasterFile() {;}
    virtual void closeRasterFile(BOOLEAN) {;}
    virtual void openRasterFile(FILE*,BOOLEAN) {;}
    virtual void readRasterFile() {;}
    virtual void printOn(ostream& os = cerr) const;
};

class VirtualGraphTool12 : public VirtualGraphTool {
  private:
  protected:
    FILE *raster_file;
    REAL xlo,ylo,xhi,yhi; //user coordinates for window corners
    REAL xlen,ylen;       //width and height in user coordinates
    BOOLEAN black_and_white;
    char *xlabel,*ylabel;

    virtual void initGT(BOOLEAN);
    VirtualGraphTool12(const VirtualGraphTool12 &g) : 
      VirtualGraphTool(g) { OBSOLETE("copy constructor disallowed"); }
  public:
    VirtualGraphTool12(const char *str,const char *xl,const char *yl);
    void delayedConstructor(REAL,REAL,REAL,REAL,BOOLEAN =TRUE,
      BOOLEAN =FALSE);
    void delayedConstructor(REAL =HALF,REAL =HALF,BOOLEAN =FALSE);
    virtual ~VirtualGraphTool12();
    virtual VirtualWindow12* getWindow() { return 0; }
    virtual const VirtualWindow12* getWindow() const { return 0; }
    void setFont(char *fn) { getWindow()->setFont(fn); }
    void putString(REAL x,REAL y,const char *str,REAL angle=0.) {
      REAL xx=(x-xlo)/xlen;
      REAL yy=(y-ylo)/ylen;
      getWindow()->putString(&xx,&yy,str,&angle);
    }
    const char* getXLabel() const { return xlabel; }
    const char* getYLabel() const { return ylabel; }
    REAL getLowX() const { return xlo; }
    REAL getLowY() const { return ylo; }
    REAL getHighX() const { return xhi; }
    REAL getHighY() const { return yhi; }
    REAL getPixelWidth() const { 
      return (xhi-xlo)/(1.1*static_cast<REAL>(getWindow()->getWidth()));
    }
    REAL getPixelHeight() const { 
     return (xhi-xlo)/(1.1*static_cast<REAL>(getWindow()->getHeight()));
    }
    BOOLEAN monochrome() const { 
      return getWindow()->monochrome(); 
    }
    void setfgColor(const char *color) { 
      VirtualGraphTool::setfgColor(color); 
    }
    void setfgColor(INTEGER ic,INTEGER im) {
      getWindow()->setfgColor(&ic,&im);
    }

    virtual void rescale(REAL,REAL,REAL,REAL);
    virtual BOOLEAN buttonIsPressed(REAL&,REAL&);
    virtual int getMouse(REAL&,REAL&);
    void movePen(REAL x,REAL y) {
      REAL xfrac=(x-xlo)/xlen; REAL yfrac=(y-ylo)/ylen;
      getWindow()->movePen(&xfrac,&yfrac);
    }
    void drawLine(REAL x,REAL y,BOOLEAN use_buffer=TRUE) {
      REAL xfrac=(x-xlo)/xlen; REAL yfrac=(y-ylo)/ylen;
      getWindow()->drawLine(&xfrac,&yfrac,&use_buffer);
    }
    void drawPlusGivenCorners(REAL xxlo,REAL yylo,REAL xxhi,REAL yyhi) {
      REAL xmid=HALF*(xxlo+xxhi),ymid=HALF*(yylo+yyhi);
      movePen(xxlo,ymid); drawLine(xxhi,ymid);
      movePen(xmid,yylo); drawLine(xmid,yyhi);
    }
    void drawPlusGivenCenter(REAL x,REAL y,REAL dx,REAL dy) {
      movePen(x-dx,y); drawLine(x+dx,y);
      movePen(x,y-dy); drawLine(x,y+dy);
    }
    void drawPlus(REAL x,REAL y,REAL dx=HALF) {
      drawPlusGivenCenter(x,y,dx,dx*(yhi-ylo)/(xhi-xlo));
    }
    void drawCrossGivenCorners(REAL xxlo,REAL yylo,REAL xxhi,REAL yyhi){
      movePen(xxlo,yylo); drawLine(xxhi,yyhi);
      movePen(xxhi,yylo); drawLine(xxlo,yyhi);
    }
    void drawCrossGivenCenter(REAL x,REAL y,REAL dx,REAL dy) {
      drawCrossGivenCorners(x-dx,y-dy,x+dx,y+dy);
    }
    void drawCross(REAL x,REAL y,REAL dx=HALF) {
      drawCrossGivenCenter(x,y,dx,dx*(yhi-ylo)/(xhi-xlo));
    }
    void drawDiamondGivenCorners(REAL xxlo,REAL yylo,REAL xxhi,
    REAL yyhi) {
      REAL xmid=HALF*(xxlo+xxhi),ymid=HALF*(yylo+yyhi);
      movePen(xxhi,ymid);
      drawLine(xmid,yyhi); drawLine(xxlo,ymid);
      drawLine(xmid,yylo); drawLine(xxhi,ymid);
    }
    void drawDiamondGivenCenter(REAL x,REAL y,REAL dx=HALF) {
      REAL xxhi=x+dx;
      REAL dy=dx*(yhi-ylo)/(xhi-xlo);
      movePen(xxhi,y);
      drawLine(x,y+dy); drawLine(x-dx,y);
      drawLine(x,y-dy); drawLine(xxhi,y);
    }
    void drawBoxGivenCorners(REAL xxlo,REAL yylo,REAL xxhi,REAL yyhi) {
      movePen(xxlo,yylo);
      drawLine(xxhi,yylo); drawLine(xxhi,yyhi);
      drawLine(xxlo,yyhi); drawLine(xxlo,yylo);
    }
    void drawBoxGivenCenter(REAL x,REAL y,REAL dx=HALF) {
      REAL dy=dx*(yhi-ylo)/(xhi-xlo);
      drawBoxGivenCorners(x-dx,y-dy,x+dx,y+dy);
    }
    void colorPolygon(int,REAL*,REAL*);
    void colorPolygon(REAL,int,REAL*,REAL*);
    void newPage() { getWindow()->newPage(); }
    void createRasterFile(BOOLEAN);
    void writeRasterFile();
    void closeRasterFile(BOOLEAN);
    void openRasterFile(FILE*,BOOLEAN);
    void readRasterFile();
#if (SPACEDIM>1)
    virtual void drawVector(COMPLEX base,COMPLEX head,COMPLEX arrow) {
      COMPLEX position=base+head; movePen(real(base),imag(base));
      drawLine(real(position),imag(position));
      position=position+head*arrow;
      drawLine(real(position),imag(position));
    }
    virtual void colorQuad(REAL,REAL*,REAL*);
#endif
    void drawAxes();
    virtual void printOn(ostream& os = cerr) const;
};

double ticSpacing(double,double);
#endif
