//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#ifndef VIRTUAL_WINDOW_H
#define VIRTUAL_WINDOW_H "$Header: /home/faculty/johnt/cvs/deal_new/graphics/VirtualWindow.H,v 1.1 2009/08/20 17:31:47 johnt Exp $"
#include <stdio.h>
#include <iostream>
#ifdef USE_GTK
#include <gdk/gdkcolor.h>
#endif
#include <X11/Xlib.h>
#include "Const.H"
#include "Errors.H"
#include "Types.H"

class VirtualWindow {
  public:
#ifdef USE_GTK
    typedef GdkColor COLOR_TYPE;
#else 
    typedef XColor COLOR_TYPE;
#endif
  protected:
    VirtualWindow(const VirtualWindow&) { 
      OBSOLETE("copy constructor disallowed");
    }
  public:
    VirtualWindow() {}
    virtual ~VirtualWindow() {}
    virtual VirtualWindow* getWindow() = 0;
    virtual const VirtualWindow* getWindow() const = 0;
    virtual const char* getName() const = 0;
    virtual int getWidth() const = 0;
    virtual int getHeight() const = 0;
    virtual void newPage() = 0 ;
    virtual void setLineWidth(const INTEGER*) const=0 ;
    virtual void setfgColor(const char*) const=0;
    virtual void setbgColor(const char*) const=0;
    virtual void flush(BOOLEAN=TRUE) = 0 ;
    virtual void printOn(ostream& os = cerr) const {
      os << "VirtualWindow:" << getName() << endl;
    }
};

class VirtualWindow12 : public VirtualWindow {
  protected:
    VirtualWindow12(const VirtualWindow12&) { 
      OBSOLETE("copy constructor disallowed");
    }
  public:
    VirtualWindow12() {}
    virtual ~VirtualWindow12() {}
    virtual const char* getDisplayName() const = 0 ;
    virtual void setFont(const char*) const = 0 ;
    virtual BOOLEAN monochrome() const = 0 ;
    virtual void setfgColor(const INTEGER*,const INTEGER*) const=0 ;
    virtual void putString(const REAL*,const REAL*,const char*,
      const REAL*) const=0;
    virtual void writeXBorder(const char *str) const=0;
    virtual void writeYBorder(const char *str) const=0;
    virtual void movePen(const REAL*,const REAL*) const = 0 ;
    virtual void drawLine(const REAL*,const REAL*,const BOOLEAN* =0)
      const=0;
    virtual BOOLEAN buttonIsPressed(REAL&,REAL&) = 0 ;
    virtual int getMouse(REAL&,REAL&) = 0 ;
    virtual void colorPolygon(const REAL*,const INTEGER *,const REAL*,
      const REAL*) const=0;
    virtual void colorPolygon(const INTEGER*,const REAL*,const REAL*)
      const=0;
    virtual void colorRect(const REAL*,const REAL*,const REAL*,
      const REAL*,const REAL*) const=0;
    virtual int pickColor(REAL) const=0 ;
    virtual void setfgColor(const char*) const = 0;
    virtual void setfgColor(const REAL*) const = 0;
    virtual void setbgColor(const char*) const = 0;
    virtual void setbgColor(const REAL*) const = 0;
    virtual void writePaletteName(FILE*) = 0 ;
    virtual void createRaster(FILE*) = 0 ;
    virtual void writeRaster(FILE*) = 0 ;
    virtual void closeRaster() = 0 ;
    virtual void openRaster(FILE*) = 0 ;
    virtual void readRaster(FILE*) = 0 ;
    virtual void printOn(ostream& os = cerr) const {
      os << "VirtualWindow12: display_name=" << getDisplayName()
         << endl;
      VirtualWindow::printOn(os);
    }
};
#endif
