//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
// "$Header: /home/faculty/johnt/cvs/deal_new/graphics/XColorArray.C,v 1.1 2009/08/20 17:31:47 johnt Exp $"
#include "XColorArray.H"
#include "Palette.H"

XColorArray::XColorArray(const Palette &pal) : 
NumPtr<XColor>(pal.getMapIndex(pal.getNumNamedColors()-1)+1) {
  int num_named_colors=pal.getNumNamedColors();
  XColor exact,actual;
  Display *display=XOpenDisplay(0);
  const Colormap &cmap=DefaultColormap(display,DefaultScreen(display));

  XLookupColor(display,cmap,pal.getColorName(0),&exact,&actual);
  (*this)[0]=actual;

  XLookupColor(display,cmap,pal.getColorName(1),&exact,&actual);
  (*this)[1]=actual;

  for (int i=2;i<num_named_colors;i++) {
    XLookupColor(display,cmap,pal.getColorName(i),&exact,&actual);
    (*this)[pal.getMapIndex(i)]=actual;
    XColor low=(*this)[pal.getMapIndex(i-1)];
    XColor high=(*this)[pal.getMapIndex(i)];
    int n=pal.getMapIndex(i)-pal.getMapIndex(i-1);
    for (int j=pal.getMapIndex(i-1)+1;j<pal.getMapIndex(i);j++) {
      float frac=float(j-pal.getMapIndex(i-1))/float(n);
      actual.red=low.red
        +(unsigned short)(float(high.red-low.red)*frac);
      actual.green=low.green
        +(unsigned short)(float(high.green-low.green)*frac);
      actual.blue=low.blue
        +(unsigned short)(float(high.blue-low.blue)*frac);
      (*this)[j]=actual;
    }
  }
}

#include "NumPtr.C"
INSTANTIATE_NUMPTR(XColor)
