//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
// "$Header: /home/faculty/johnt/cvs/deal_new/graphics/XColormap.C,v 1.1 2009/08/20 17:31:47 johnt Exp $"
#include "XColormap.H"
#include <string.h>
#include <stdlib.h>
#include "MemoryDebugger.H"
#include "Tracer.H"
#define MAX_X_COLORS 248
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
XColormap::XColormap(Palette *pal,const char *dn,Display* dpy) : 
VirtualColormap(pal),cmap_colors(0),cmap(None),display(dpy),visual(0),
depth(0),screen(0) {
  int num_named_colors=pal->getNumNamedColors();

//need background color and two Palette colors
  CHECK_TEST(num_named_colors>=3);
//background color should be map_index[0]
  CHECK_SAME(pal->getMapIndex(0),0);
//first Palette color should be map_index[1]
  CHECK_SAME(pal->getMapIndex(1),1);
#ifdef DEBUG
  for (int i=2;i<num_named_colors;i++) {
    CHECK_TEST(pal->getMapIndex(i-1)<pal->getMapIndex(i));
  }
#endif

  if(!display) display=XOpenDisplay(dn);
#ifdef DEBUG
  if (!display) {
    cout << "attempted to open display on " << XDisplayName(dn) << endl;
  }
#endif
  ASSERT(display!=0);

  screen=DefaultScreen(display);
  depth=DefaultDepth(display,screen);
  cmap=DefaultColormap(display,screen);

  if (depth==1) cout << "Monochrome display" << endl; 
  ASSERT(depth>1);

  int k=DirectColor;
  XVisualInfo visual_info;
  while (!XMatchVisualInfo(display,screen,depth,k--,&visual_info) && 
    k>=0);
  if (k<StaticColor) {
    cout << "No color visual available at default depth = " << depth 
         << endl;
  }
  ASSERT(k>=StaticColor);
  visual=visual_info.visual;
    
  XColor *xc=OPERATOR_NEW_BRACKET(XColor,getNumberColors());
  cmap_colors=OPERATOR_NEW Pixel[getNumberColors()];

  if (allocColorCells(xc)) {
//  do not know why, but this has to be called to get right colors
    XQueryColors(display,cmap,xc,getNumberColors());
    for (int i=0;i<getNumberColors();i++) {
      cmap_colors[i]=xc[i].pixel;
    }
  } else {
    XColor exact;
    exact.flags=DoRed | DoGreen | DoBlue;
    for (int i=0;i<getNumberColors();i++) {
      exact.red=xc[i].red /* & visual_info.red_mask*/;
      exact.green=xc[i].green /* & visual_info.green_mask*/;
      exact.blue=xc[i].blue /* & visual_info.blue_mask*/;
      if (!XAllocColor(display,cmap,&exact)) {
        cout << "\ti = " << i << endl;
        OBSOLETE("could not allocate color number");
      }
      cmap_colors[i]=exact.pixel;
    }
  }
  delete [] xc;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BOOLEAN XColormap::allocColorCells(XColor *xc) {
  Palette *pal=getPalette();
  int num_named_colors=pal->getNumNamedColors();
  for (int i=0;i<num_named_colors;i++) {
    XColor exact;
    XColor *xci=xc+pal->getMapIndex(i);
    if (!XLookupColor(display,cmap,pal->getColorName(i),&exact,xci)) {
      cout << "color name = " << pal->getColorName(i) << endl;
      OBSOLETE("XColormap: cannot lookup color")
    }
  }
  for (int i=2;i<num_named_colors;i++) {
    int map_lo=pal->getMapIndex(i-1);
    int map_hi=pal->getMapIndex(i);
    float denom=float(map_hi - map_lo);
    float lo_red=float(xc[map_lo].red);
    float hi_red=float(xc[map_hi].red);
    float lo_green=float(xc[map_lo].green);
    float hi_green=float(xc[map_hi].green);
    float lo_blue=float(xc[map_lo].blue);
    float hi_blue=float(xc[map_hi].blue);
    for (int j=map_lo+1;j<map_hi;j++) {
      float lo_ratio=float(map_hi - j)/denom;
      float hi_ratio=float(j - map_lo)/denom;
      xc[j].red=int(lo_red*lo_ratio+hi_red*hi_ratio);
      xc[j].green=int(lo_green*lo_ratio+hi_green*hi_ratio);
      xc[j].blue=int(lo_blue*lo_ratio+hi_blue*hi_ratio);
    }
  }

  Bool contig=False;
  unsigned long *plane_masks=0;
  unsigned int nplanes=0;
  if (XAllocColorCells(display,cmap,contig,plane_masks,nplanes,
  cmap_colors,getNumberColors())) {
    for (int i=0;i<getNumberColors();i++){
      xc[i].pixel=cmap_colors[i];
      xc[i].flags=DoRed|DoGreen|DoBlue;
    }
    XStoreColors(display,cmap,xc,getNumberColors());
    return TRUE;
  }
  return FALSE;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XColormap::installNewVisual(Window w,Visual *v) {
  XWindowAttributes a;
  Status status=XGetWindowAttributes(display,w,&a);
  depth=a.depth;

  visual=v;
  XFreeColormap(display,cmap);
  cmap=XCreateColormap(display,w,visual,AllocNone);
  XColor *xc=OPERATOR_NEW_BRACKET(XColor,getNumberColors());
  ASSERT(allocColorCells(xc));
//  do not know why, but this has to be called to get right colors
  XQueryColors(display,cmap,xc,getNumberColors());
  for (int i=0;i<getNumberColors();i++) {
    cmap_colors[i]=xc[i].pixel;
  }
  delete [] xc;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XColormap::installNewMap(Visual *v,int &num_colors,XColor *xc) {
// if we wanted to make sure that things fit into the old color map,
// we could follow the approach in imageToXImage in send.c of xloadimage
  visual=v;
  XFreeColormap(display,cmap);
  cmap=XCreateColormap(display,RootWindow(display, screen),visual,
                       AllocNone);
  int i;
  for (i=0;i<getNumberColors();i++) {
    if (!XAllocColor(display,cmap,xc+i)) {
      cout << "XColormap::XColormap: allocation in new color map"
           << "\nfailed at " << i+1 << " colors" << endl;
          break;
    } else cmap_colors[i]=xc[i].pixel;
  }
  num_colors=i;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Pixel XColormap::allocColor(const char *colorname) {
  int color_index=getPalette()->findEntry(colorname);
  if (color_index<0) {
    getPalette()->insertEntry(getNumberColors(),colorname);
    XColor xc,exact; 
    Pixel px;
    ASSERT(XAllocColorCells(display,cmap,FALSE,0,0,&px,1));
    ASSERT(XLookupColor(display,cmap,colorname,&exact,&xc));
    xc.flags = DoRed | DoGreen | DoBlue;
    xc.pixel=px;
    XStoreColor(display,cmap,&xc);
    
    cmap_colors[getNumberColors()]=xc.pixel;
    VirtualColormap::allocColor();

    return xc.pixel;
  } else return cmap_colors[color_index];
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
XColormap::~XColormap() {
//XFreeColors(display,cmap,cmap_colors,getNumberColors(),0);
//XUninstallColormap(display,cmap);
  XFreeColormap(display,cmap);
  XFlush(display);
  delete [] cmap_colors;
  cmap_colors=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XColormap::printOn(ostream &os) const {
  os << "XColormap: " << endl;
  os << "\n\tnum_cmap_colors = " << getNumberColors()
     << "\n\tdisplay = " << display 
     << "\n\tcmap = " << cmap 
     << "\n\tpalette = " << getPalette() << endl;

  XColor *xc=OPERATOR_NEW_BRACKET(XColor,getNumberColors());
  for (int i=0;i<getNumberColors();i++) {
    xc[i].pixel=cmap_colors[i];
  }
  XQueryColors(display, cmap, xc, getNumberColors());
  for (int i=0;i<getNumberColors();i++) {
    cout << "\tpixel,rgb = " << xc[i].pixel << " " << xc[i].red << " " 
         << xc[i].green << " " << xc[i].blue << endl;
  }
  delete [] xc;
//printDisplayInfo(os,0,display);
//printScreenInfo(os,display,screen,0);
//printWindowInfo(os,display,window);
//printWMProperties(os,display,window);
//printHosts(display,os);
  VirtualColormap::printOn(os);
}
