//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
// "$Header: /home/faculty/johnt/cvs/deal_new/graphics/XTWindow.C,v 1.1 2009/08/20 17:31:47 johnt Exp $"
#include "XTWindow.H"
#include <string.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xlibint.h>
#include <Xm/DrawingA.h>
#include <Xm/Print.h>
#include <Xm/Xm.h>
#include <Xm/MessageB.h>
#include <Xm/XmStrDefs.h>
#include <Xm/XpmP.h>
#include "Const.H"
#include "MemoryDebugger.H"
#include "Palette.H"
#include "Tracer.H"
#include "XWindow.H"

#define NTSC_WIDTH 600
#define NTSC_HEIGHT 420

static Atom wm_delete_window;

static char default_font[] = "8x13";
//static char default_font[] = "-adobe-helvetica-medium-o-normal--24-240-75-75-p-130-iso8859-1";
#define NUMBER_ERRORS 18
// these appear in /usr/local/X11/include/X.h
static const char *ErrorCodes[NUMBER_ERRORS] = {
  "Success",           // 0
  "BadRequest",        // 1
  "BadValue",          // 2
  "BadWindow",         // 3
  "BadPixmap",         // 4
  "BadAtom",           // 5
  "BadCursor",         // 6
  "BadFont",           // 7
  "BadMatch",          // 8
  "BadDrawable",       // 9
  "BadAccess",         // 10
  "BadAlloc",          // 11
  "BadColor",          // 12
  "BadGC",             // 13
  "BadIDChoice",       // 14
  "BadName",           // 15
  "BadLength",         // 16
  "BadImplementation"  // 17
};

//viewres and listres will list resources for widgets
static const char *fallback_resources[] = {
//"*background: LightGrey",
  "*title: XTWindow",
  "!*glxarea*width: 300",
  "*glxarea*height: 300", 
  "*selectlabel.labelString:  (0, 0, 0)", 
  "*selectbutton.labelString:  Select",
  "*lxscale.titleString:  X lx",  
  "*lyscale.titleString:  Y ly",  
  "*lzscale.titleString:  Z lz",  
  "*rxscale.titleString:  X rx",  
  "*ryscale.titleString:  Y ry",  
  "*rzscale.titleString:  Z rz",  
  "*drawing area.titleString: drawing area",
  (const char*)0
};

XtAppContext XTWindow::general_appcontext;
Widget XTWindow::grand_parent=0;
unsigned int XTWindow::ref_ctr=0;

// *********************************************************************

/* xvertext 5.0, Copyright (c) 1993 Alan Richardson (mppa3@uk.ac.sussex.syma)
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both 
 * the copyright notice and this permission notice appear in supporting
 * documentation.  All work developed as a consequence of the use of
 * this program should duly acknowledge such use. No representations are
 * made about the suitability of this software for any purpose.  It is
 * provided "as is" without express or implied warranty.
 */

// xvertext had several memory leaks, 
//   failed to clear Pixmaps before writing text, and
//   generated segmentation faults by failing to delete data before
//     destroying XImage's.

//#include <X11/Xlib.h>
//#include <X11/Xutil.h>
//#include <X11/Xatom.h>
//#include <stdio.h>
//#include <math.h>
//#include "rotated.h"

//#define XV_VERSION      5.0
//#define XV_COPYRIGHT \
//      "xvertext routines Copyright (c) 1993 Alan Richardson"

// text alignment
#define NONE             0
#define TLEFT            1
#define TCENTRE          2
#define TRIGHT           3
#define MLEFT            4
#define MCENTRE          5
#define MRIGHT           6
#define BLEFT            7
#define BCENTRE          8
#define BRIGHT           9

// Make sure cache size is set
#ifndef CACHE_SIZE_LIMIT
# define CACHE_SIZE_LIMIT 0
#endif //CACHE_SIZE_LIMIT
    
// Make sure a cache method is specified
#ifndef CACHE_XIMAGES
# ifndef CACHE_BITMAPS
#   define CACHE_BITMAPS
# endif //CACHE_BITMAPS
#endif //CACHE_XIMAGES

// Debugging macros

//#ifdef DEBUG
//static int debug=1;
//#else
//static int debug=0;
//#endif //DEBUG

//#define DEBUG_PRINT1(a) if (debug) printf (a)
//#define DEBUG_PRINT2(a, b) if (debug) printf (a, b)
//#define DEBUG_PRINT3(a, b, c) if (debug) printf (a, b, c)
//#define DEBUG_PRINT4(a, b, c, d) if (debug) printf (a, b, c, d)
//#define DEBUG_PRINT5(a, b, c, d, e) if (debug) printf (a, b, c, d, e)

// A structure holding everything needed for a rotated string
typedef struct rotated_text_item_template {
  Pixmap bitmap;
  XImage *ximage;
  
  char *text;
  char *font_name;
  Font fid;
  float angle;
  int align;
  float magnify;
  
  int cols_in;
  int rows_in;
  int cols_out;
  int rows_out;
  
  int nl;
  int max_width;
  float *corners_x;
  float *corners_y;
  
  long int size;
  int cached;

  struct rotated_text_item_template *next;
} RotatedTextItem;
RotatedTextItem *first_text_item=0;


// A structure holding current magnification and bounding box padding
static struct style_template {
  float magnify;
  int bbx_pad;
} style={
  1.,
  0
};

//static char *my_strtok();
//int XRotDrawString(Display*,XFontStruct*,float,Drawable,GC,int,int,
//  char*);

//**********************************************************************
// Routine to mimic `strdup()' (some machines don't have it)
//**********************************************************************
//static char* my_strdup(char *str) {
//  if (!str) return 0;
//  char *s=OPERATOR_NEW_BRACKET(char,strlen(str)+1);
//  if (s) strcpy(s, str);
//  return s;
//}


//**********************************************************************
// Routine to replace `strtok' : 
//   this one returns a zero length string 
//   if it encounters two consecutive delimiters 
//**********************************************************************
static char* my_strtok(char *str1,const char *str2) {
  static int start, len;
  static char *stext;
  
  if (!str2) return 0;
  /* initialise if str1 not NULL */
  if (str1) {
    start=0;
    stext=str1;
    len=strlen(str1);
  }
  /* run out of tokens ? */
  if (start>=len) return 0;
  /* loop through characters */
  int i;
  for (i=start; i<len; i++) {
    /* loop through delimiters */
    int stop=0;
    for (int j=0;j<strlen(str2);j++) if (stext[i]==str2[j]) stop=1;
    if (stop) break;
  }
  stext[i]='\0';
  char *ret=stext+start;
  start=i+1;
  return ret;
}

//**********************************************************************
// Return version/copyright information
//**********************************************************************
//float XRotVersion(char *str,int n) {
//  if (str) strncpy(str,XV_COPYRIGHT,n);
//  return XV_VERSION;
//}

//**********************************************************************
// Set the font magnification factor for all subsequent operations
//**********************************************************************
void XRotSetMagnification(float m) {
  if (m>0.) style.magnify=m;
}

//**********************************************************************
// Set the padding used when calculating bounding boxes
//**********************************************************************
void XRotSetBoundingBoxPad(int p) {
  if (p>=0) style.bbx_pad=p;
}

//**********************************************************************
//  Create an XImage structure and allocate memory for it
//**********************************************************************
static XImage* MakeXImage(Display *dpy,int w,int h) {
// reserve memory for image
  char *data=OPERATOR_NEW_BRACKET(char,((w-1)/8+1)*h);
  if (!data) return 0;
  for (int i=0;i<((w-1)/8+1)*h;i++) data[i]=0;
  
// create the XImage
  XImage *I=XCreateImage(dpy,DefaultVisual(dpy,DefaultScreen(dpy)),1,
    XYBitmap,0,data,w,h,8,0);
  if (!I) return 0;
  
  I->byte_order=I->bitmap_bit_order=MSBFirst;
  return I;
}

//**********************************************************************
//  Draw a horizontal string in a quick fashion
//**********************************************************************
static int XRotDrawHorizontalString(Display *dpy,XFontStruct *font,
Drawable drawable,GC gc,int x,int y,const char *text,int align,int bg) {
  int nl=1;
  const char *str2_a="\0", *str2_b="\n\0";

//DEBUG_PRINT1("**\nHorizontal text.\n");

//this gc has similar properties to the user's gc (including stipple)
  GC my_gc=XCreateGC(dpy, drawable, 0L, 0);
  XCopyGC(dpy, gc, GCForeground|GCBackground|GCFunction|GCStipple|
    GCFillStyle|GCTileStipXOrigin|GCTileStipYOrigin|GCPlaneMask, my_gc);
  XSetFont(dpy, my_gc, font->fid);

// count number of sections in string
  if (align!=NONE) {
    for (int i=0; i<strlen(text)-1; i++) if (text[i]=='\n') nl++;
  }
    
//ignore newline characters if not doing alignment
  const char *str2=(align==NONE ? str2_a : str2_b);
    
// overall font height
  int height=font->ascent+font->descent;
    
// y position 
  int yp;
  if(align==TLEFT || align==TCENTRE || align==TRIGHT) {
    yp=y+font->ascent;
  } else if(align==MLEFT || align==MCENTRE || align==MRIGHT) {
    yp=y-nl*height/2+font->ascent;
  } else if(align==BLEFT || align==BCENTRE || align==BRIGHT) {
    yp=y-nl*height+font->ascent;
  } else yp=y;
    
  char *str1=strdup(text);
  if (!str1) return 1;
    
  char *str3=my_strtok(str1, str2);
// loop through each section in the string
  do {
    int dir, asc, desc;
    XCharStruct overall;
    XTextExtents(font, str3, strlen(str3), &dir, &asc, &desc, &overall);

//  where to draw section in x ?
    int xp;
    if (align==TLEFT || align==MLEFT || align==BLEFT || align==NONE) {
      xp=x;
    } else if (align==TCENTRE || align==MCENTRE || align==BCENTRE) {
      xp=x-overall.rbearing/2;
    } else xp=x-overall.rbearing;

//  draw string onto bitmap
    if (!bg) XDrawString(dpy,drawable,my_gc,xp,yp,str3,strlen(str3));
    else XDrawImageString(dpy,drawable,my_gc,xp,yp,str3,strlen(str3));
      
//  move to next line
    yp+=height;

    str3=my_strtok((char *)NULL, str2);
  } while (str3);
    
  free(str1);
  XFreeGC(dpy, my_gc);

  return 0;
}

//**********************************************************************
// Magnify an XImage using bilinear interpolation
//**********************************************************************
static XImage *XRotMagnifyImage(Display *dpy,XImage *ximage) {
//size of input image
  int cols_in=ximage->width;
  int rows_in=ximage->height;
//size of final image
  int cols_out=
    static_cast<int>(static_cast<float>(cols_in)*style.magnify);
  int rows_out=
    static_cast<int>(static_cast<float>(rows_in)*style.magnify);
//this will hold final image
  XImage *I_out=MakeXImage(dpy, cols_out, rows_out);
  if (!I_out) return 0;

// width in bytes of input, output images
  int byte_width_in=(cols_in-1)/8+1;
  int byte_width_out=(cols_out-1)/8+1;
//for speed
  float mag_inv=1./style.magnify;
  float y=0.;

//loop over magnified image
  for (int j2=0; j2<rows_out; j2++) {
    float x=0;
    int j=static_cast<int>(y);

    for (int i2=0; i2<cols_out; i2++) {
      int i=static_cast<int>(x);
//    bilinear interpolation - where are we on bitmap ?
//    right edge
      float u,t;
      float z1, z2, z3, z4;
      if (i==cols_in-1 && j!=rows_in-1) {
        t=0;
        u=y-(float)j;
        z1=(ximage->data[j*byte_width_in+i/8] & 128>>(i%8))>0;
        z2=z1;
        z3=(ximage->data[(j+1)*byte_width_in+i/8] & 128>>(i%8))>0;
        z4=z3;
//    top edge */
      } else if (i!=cols_in-1 && j==rows_in-1) {
        t=x-(float)i;
        u=0;
        z1=(ximage->data[j*byte_width_in+i/8] & 128>>(i%8))>0;
        z2=(ximage->data[j*byte_width_in+(i+1)/8] & 128>>((i+1)%8))>0;
        z3=z2;
        z4=z1;
// top right corner
      } else if (i==cols_in-1 && j==rows_in-1) {
        u=0;
        t=0;
        z1=(ximage->data[j*byte_width_in+i/8] & 128>>(i%8))>0;
        z2=z1;
        z3=z1;
        z4=z1;
// somewhere `safe'
      } else {
        t=x-(float)i;
        u=y-(float)j;
        z1=(ximage->data[j*byte_width_in+i/8] & 128>>(i%8))>0;
        z2=(ximage->data[j*byte_width_in+(i+1)/8] & 128>>((i+1)%8))>0;
        z3=(ximage->data[(j+1)*byte_width_in+(i+1)/8]&128>>((i+1)%8))>0;
        z4=(ximage->data[(j+1)*byte_width_in+i/8] & 128>>(i%8))>0;
      }

// if interpolated value is greater than 0.5, set bit
      if (((1-t)*(1-u)*z1 + t*(1-u)*z2 + t*u*z3 + (1-t)*u*z4)>0.5) {
        I_out->data[j2*byte_width_out+i2/8]|=128>>i2%8;
      }
      x+=mag_inv;
    }
    y+=mag_inv;
  }
//destroy original
  XDestroyImage(ximage);
//return big image
  return I_out;
}

//**********************************************************************
//  Create a rotated text item
//**********************************************************************

static RotatedTextItem *XRotCreateTextItem(Display *dpy,
XFontStruct *font,float angle,const char *text,int align) {
  const char *str2_a="\0", *str2_b="\n\0";
    
// allocate memory
  RotatedTextItem *item=OPERATOR_NEW RotatedTextItem;
  if (!item) return 0;

// count number of sections in string
  item->nl=1;
  if (align!=NONE) {
    for (int i=0; i<strlen(text)-1; i++) if(text[i]=='\n') item->nl++;
  }
    
// ignore newline characters if not doing alignment
  const char *str2=(align==NONE ? str2_a : str2_b);
    
// find width of longest section
  char *str1=strdup(text);
  if (!str1) return 0;
    
  char *str3=my_strtok(str1, str2);

  int dir, asc, desc;
  XCharStruct overall;
  XTextExtents(font, str3, strlen(str3), &dir, &asc, &desc, &overall);
    
  item->max_width=overall.rbearing;
    
// loop through each section
  do {
    str3=my_strtok((char *)NULL, str2);

    if (str3) {
      XTextExtents(font,str3,strlen(str3),&dir,&asc,&desc,&overall);

      if(overall.rbearing>item->max_width) {
          item->max_width=overall.rbearing;
      }
    }
  } while(str3);
    
  free(str1);
    
// overall font height
  int height=font->ascent+font->descent;
    
// dimensions horizontal text will have
  item->cols_in=item->max_width;
  item->rows_in=item->nl*height;
    
// bitmap for drawing on
  Pixmap canvas=XCreatePixmap(dpy, DefaultRootWindow(dpy),
                       item->cols_in, item->rows_in, 1);
    
// create a GC for the bitmap
  GC font_gc=XCreateGC(dpy, canvas, 0L, 0);
  XSetBackground(dpy, font_gc, 0);
  XSetFont(dpy, font_gc, font->fid);
    
// make sure the bitmap is blank
  XSetForeground(dpy, font_gc, 0);
  XFillRectangle(dpy,canvas,font_gc,0,0,item->cols_in+1,
    item->rows_in+1);
  XSetForeground(dpy, font_gc, 1);
    
// pre-calculate sin and cos
  float sin_angle=sin(angle);
  float cos_angle=cos(angle);
    
// text background will be drawn using XFillPolygon
  item->corners_x=OPERATOR_NEW_BRACKET(float,4*item->nl);
  if (!item->corners_x) return 0;
    
  item->corners_y=OPERATOR_NEW_BRACKET(float,4*item->nl);
  if (!item->corners_y) return 0;
    
// draw text horizontally
    
// start at top of bitmap
  int yp=font->ascent;
    
  str1=strdup(text);
  if (!str1) return 0;
  str3=my_strtok(str1, str2);
    
// loop through each section in the string
  int ic=0;
  do {
    XTextExtents(font,str3,strlen(str3),&dir,&asc,&desc,&overall);

// where to draw section in x ?
    int xp;
    if (align==TLEFT || align==MLEFT || align==BLEFT || align==NONE) {
        xp=0;
    } else if(align==TCENTRE || align==MCENTRE || align==BCENTRE) {
      xp=(item->max_width-overall.rbearing)/2;
    } else xp=item->max_width-overall.rbearing;

// draw string onto bitmap
    XDrawString(dpy, canvas, font_gc, xp, yp, str3, strlen(str3));
    
// keep a note of corner positions of this string
    item->corners_x[ic]=(static_cast<float>(xp)
                    -static_cast<float>(item->cols_in)/2)*style.magnify;
    item->corners_y[ic]=(static_cast<float>(yp-font->ascent)
                    -static_cast<float>(item->rows_in)/2)*style.magnify;
    item->corners_x[ic+1]=item->corners_x[ic];
    item->corners_y[ic+1]=item->corners_y[ic]
                         +static_cast<float>(height)*style.magnify;
    item->corners_x[item->nl*4-1-ic]=item->corners_x[ic]
                    +static_cast<float>(overall.rbearing)*style.magnify;
    item->corners_y[item->nl*4-1-ic]=item->corners_y[ic];
    item->corners_x[item->nl*4-2-ic]=item->corners_x[item->nl*4-1-ic];
    item->corners_y[item->nl*4-2-ic]=item->corners_y[ic+1];
    ic+=2;
    
// move to next line
    yp+=height;
    
    str3=my_strtok(0, str2);
  } while(str3);
  free(str1);
    
// create image to hold horizontal text
  XImage *I_in=MakeXImage(dpy, item->cols_in, item->rows_in);
  if (!I_in) return 0;
    
// extract horizontal text
  XGetSubImage(dpy, canvas, 0, 0, item->cols_in, item->rows_in,
               1, XYPixmap, I_in, 0, 0);
  I_in->format=XYBitmap;
    
// magnify horizontal text
  int old_cols_in=0, old_rows_in=0;
  if (style.magnify!=1.) {
    I_in=XRotMagnifyImage(dpy, I_in);

    old_cols_in=item->cols_in;
    old_rows_in=item->rows_in;
    item->cols_in=
      static_cast<int>(static_cast<float>(item->cols_in)*style.magnify);
    item->rows_in=
      static_cast<int>(static_cast<float>(item->rows_in)*style.magnify);
  }

// how big will rotated text be
  item->cols_out=
    static_cast<int>(fabs(static_cast<float>(item->rows_in)*sin_angle) 
                    +fabs(static_cast<float>(item->cols_in)*cos_angle) 
                    +0.99999) +2;

  item->rows_out=
    static_cast<int>(fabs(static_cast<float>(item->rows_in)*cos_angle)
                    +fabs(static_cast<float>(item->cols_in)*sin_angle)
                    +0.99999) +2;

  if (item->cols_out%2==0) item->cols_out++;
  if (item->rows_out%2==0) item->rows_out++;
    
// create image to hold rotated text
  item->ximage=MakeXImage(dpy, item->cols_out, item->rows_out);
  if (!item->ximage) return 0;
    
  int byte_w_in=(item->cols_in-1)/8+1;
  int byte_w_out=(item->cols_out-1)/8+1;
    
// we try to make this bit as fast as possible 
//  - which is why it looks a bit over-the-top
    
// vertical distance from centre
  float dj=0.5-static_cast<float>(item->rows_out)/2;

// where abouts does text actually lie in rotated image?
  float xl,xr,xinc;
  if (angle==0 || angle==M_PI/2 || angle==M_PI || angle==3*M_PI/2) {
    xl=0;
    xr=static_cast<float>(item->cols_out);
    xinc=0;
  } else if(angle<M_PI) {
    xl=static_cast<float>(item->cols_out)/2+
        (dj-static_cast<float>(item->rows_in)/(2*cos_angle))/
            tan(angle)-2;
    xr=static_cast<float>(item->cols_out)/2+
        (dj+static_cast<float>(item->rows_in)/(2*cos_angle))/
            tan(angle)+2;
    xinc=1./tan(angle);
  }
  else {
    xl=static_cast<float>(item->cols_out)/2
     +(dj+static_cast<float>(item->rows_in)/(2*cos_angle))/tan(angle)-2;
    xr=static_cast<float>(item->cols_out)/2
     +(dj-static_cast<float>(item->rows_in)/(2*cos_angle))/tan(angle)+2;
    xinc=1./tan(angle);
  }

//loop through all relevent bits in rotated image
  for (int j=0; j<item->rows_out; j++) {

//  no point re-calculating these every pass
    float di=static_cast<float>( xl<0 ? 0 : static_cast<int>(xl) )
            +0.5-static_cast<float>(item->cols_out)/2;
    int byte_out=(item->rows_out-j-1)*byte_w_out;

//  loop through meaningful columns
    for (int i=((xl<0)?0:(int)xl); 
    i<((xr>=item->cols_out)?item->cols_out:(int)xr); i++) {
    
//    rotate coordinates
      int it=static_cast<int>(static_cast<float>(item->cols_in)/2 
                             + ( di*cos_angle + dj*sin_angle));
      int jt=static_cast<int>(static_cast<float>(item->rows_in)/2 
                             - (-di*sin_angle + dj*cos_angle));
    
//    set pixel if required
      if (it>=0 && it<item->cols_in && jt>=0 && jt<item->rows_in) {
        if((I_in->data[jt*byte_w_in+it/8] & 128>>(it%8))>0) {
            item->ximage->data[byte_out+i/8]|=128>>i%8;
        }
      }
      di+=1;
    }
    dj+=1;
    xl+=xinc;
    xr+=xinc;
  }
  delete [] I_in->data; I_in->data=0;
  XDestroyImage(I_in); I_in=0;
    
  if (style.magnify!=1.) {
    item->cols_in=old_cols_in;
    item->rows_in=old_rows_in;
  }

#ifdef CACHE_BITMAPS
//create a bitmap to hold rotated text
  item->bitmap=XCreatePixmap(dpy, DefaultRootWindow(dpy),
                             item->cols_out, item->rows_out, 1);
    
//make the text bitmap from XImage
  XPutImage(dpy, item->bitmap, font_gc, item->ximage, 0, 0, 0, 0,
            item->cols_out, item->rows_out);

  delete [] item->ximage->data; item->ximage->data=0;
  XDestroyImage(item->ximage);
#endif /*CACHE_BITMAPS*/

  XFreeGC(dpy, font_gc);
  XFreePixmap(dpy, canvas);

  return item;
}

//**********************************************************************
//  Free the resources used by a text item
//**********************************************************************
static void XRotFreeTextItem(Display *dpy, RotatedTextItem *item) {
  free(item->text);
  if (item->font_name) free(item->font_name);
  delete [] item->corners_x; item->corners_x=0;
  delete [] item->corners_y; item->corners_y=0;
#ifdef CACHE_BITMAPS
  XFreePixmap(dpy, item->bitmap);
#else
  delete [] item->ximage->data; item->ximage->data=0;
  XDestroyImage(item->ximage);
#endif /* CACHE_BITMAPS */
  delete item;
}

//**********************************************************************
//  Adds a text item to the end of the cache, removing as many items
//      from the front as required to keep cache size below limit
//**********************************************************************
static void XRotAddToLinkedList(Display *dpy, RotatedTextItem *item) {
  static long int current_size=0;
  static RotatedTextItem *last=0;
  RotatedTextItem *i1=first_text_item, *i2=0;

#ifdef CACHE_BITMAPS
//I don't know how much memory a pixmap takes in the server 
//  - probably this + a bit more we can't account for
  item->size=((item->cols_out-1)/8+1)*item->rows_out;
#else
// this is pretty much the size of a RotatedTextItem
  item->size=((item->cols_out-1)/8+1)*item->rows_out 
    +sizeof(XImage) + strlen(item->text) 
    + item->nl*8*sizeof(float) + sizeof(RotatedTextItem);

  if(item->font_name) item->size+=strlen(item->font_name);
  else item->size+=sizeof(Font);
#endif /*CACHE_BITMAPS */

#ifdef DEBUG
// count number of items in cache, for debugging
  {
    int i=0;
    while(i1) {
      i++;
      i1=i1->next;
    }
//  DEBUG_PRINT2("Cache has %d items.\n", i);
    i1=first_text_item;
  }
#endif

//DEBUG_PRINT4("current cache size=%ld, new item=%ld, limit=%ld\n",
//             current_size, item->size, CACHE_SIZE_LIMIT*1024);

// if this item is bigger than whole cache, forget it
  if (item->size>CACHE_SIZE_LIMIT*1024) {
//  DEBUG_PRINT1("Too big to cache\n\n");
    item->cached=0;
    return;
  }

// remove elements from cache as needed
  while (i1 && current_size+item->size>CACHE_SIZE_LIMIT*1024) {
//  DEBUG_PRINT2("Removed %d bytes\n", i1->size);

//  if (i1->font_name!=NULL) {
//    DEBUG_PRINT5("  (`%s'\n   %s\n   angle=%f align=%d)\n",
//                 i1->text, i1->font_name, i1->angle, i1->align);
//  }

#ifdef CACHE_FID
//  if (!i1->font_name) {
//    DEBUG_PRINT5("  (`%s'\n  FID=%ld\n   angle=%f align=%d)\n",
//                 i1->text, i1->fid, i1->angle, i1->align);
//  }
#endif /*CACHE_FID*/
    current_size-=i1->size;
    i2=i1->next;

//  free resources used by the unlucky item
    XRotFreeTextItem(dpy, i1);

//  remove it from linked list */
    first_text_item=i2;
    i1=i2;
  }

// add new item to end of linked list
  if (!first_text_item) {
    item->next=0;
    first_text_item=item;
    last=item;
  } else {
    item->next=0;
    last->next=item;
    last=item;
  }

// new cache size
  current_size+=item->size;
  item->cached=1;
//DEBUG_PRINT1("Added item to cache.\n");
}

//**********************************************************************
//   Query cache for a match with this font/text/angle/alignment
//       request, otherwise arrange for its creation
//**********************************************************************

static RotatedTextItem* XRotRetrieveFromCache(Display *dpy,
XFontStruct *font,float angle,const char *text,int align) {
  Font fid;
  char *font_name=NULL;
  unsigned long name_value;
  RotatedTextItem *item=NULL;
  RotatedTextItem *i1=first_text_item;
    
//get font name, if it exists
  if (XGetFontProperty(font, XA_FONT, &name_value)) {
//  DEBUG_PRINT1("got font name OK\n");
    font_name=XGetAtomName(dpy, name_value);
    fid=0;
  }
#ifdef CACHE_FID
// otherwise rely (unreliably?) on font ID
  else {
//  DEBUG_PRINT1("can't get fontname, caching FID\n");
    font_name=0;
    fid=font->fid;
  }
#else
//not allowed to cache font ID's
  else {
//  DEBUG_PRINT1("can't get fontname, can't cache\n");
    font_name=0;
    fid=0;
  }
#endif /*CACHE_FID*/
    
//look for a match in cache

//matching formula:
//  identical text;
//  identical fontname (if defined, font ID's if not);
//  angles close enough (<0.00001 here, could be smaller);
//  HORIZONTAL alignment matches, OR it's a one line string;
//  magnifications the same

  while (i1 && !item) {
//  match everything EXCEPT fontname/ID
    if (strcmp(text, i1->text)==0 && fabs(angle-i1->angle)<0.00001 &&
    style.magnify==i1->magnify && 
    (i1->nl==1 || 
    ((align==0)?9:(align-1))%3==((i1->align==0)?9:(i1->align-1))%3)) {
//    now match fontname/ID
      if (font_name!=0 && i1->font_name!=0) {
        if (strcmp(font_name, i1->font_name)==0) {
          item=i1;
//        DEBUG_PRINT1("Matched against font names\n");
        } else i1=i1->next;
      }
#ifdef CACHE_FID
      else if (font_name==NULL && i1->font_name==NULL) {
        if (fid==i1->fid) {
          item=i1;
//        DEBUG_PRINT1("Matched against FID's\n");
        } else i1=i1->next;
      }
#endif /*CACHE_FID*/
      else i1=i1->next;
    } else i1=i1->next;
  }
    
//if (item) DEBUG_PRINT1("**\nFound target in cache.\n");
//if (!item) DEBUG_PRINT1("**\nNo match in cache.\n");

//no match
  if (!item) {
//  create new item
    item=XRotCreateTextItem(dpy, font, angle, text, align);
    if (!item) return 0;

//  record what it shows
    item->text=strdup(text);

//  fontname or ID
    if (font_name) {
      item->font_name=strdup(font_name);
      item->fid=0;
    } else {
      item->font_name=NULL;
      item->fid=fid;
    }
    item->angle=angle;
    item->align=align;
    item->magnify=style.magnify;

//  cache it 
    XRotAddToLinkedList(dpy, item);
  }

  if (font_name) XFree(font_name);

//if XImage is cached, need to recreate the bitmap

#ifdef CACHE_XIMAGES
  {
//  create bitmap to hold rotated text
    item->bitmap=XCreatePixmap(dpy, DefaultRootWindow(dpy),
                               item->cols_out, item->rows_out, 1);

//  depth one gc */
    GC depth_one_gc=XCreateGC(dpy, item->bitmap, NULL, 0);
    XSetBackground(dpy, depth_one_gc, 0);
    XSetForeground(dpy, depth_one_gc, 1);

//  make the text bitmap from XImage */
    XPutImage(dpy, item->bitmap, depth_one_gc, item->ximage, 0, 0, 0, 0,
              item->cols_out, item->rows_out);

    XFreeGC(dpy, depth_one_gc);
  }
#endif /*CACHE_XIMAGES*/
    
  return item;
}

//**********************************************************************
//  Aligns and paints a rotated string
//**********************************************************************
static int XRotPaintAlignedString(Display *dpy,XFontStruct *font,
float angle,Drawable drawable,GC gc,int x,int y,const char *text,
int align,int bg) {
  float hot_x, hot_y;
  float hot_xp, hot_yp;
  
// return early for NULL/empty strings
  if (!text) return 0;
  if (strlen(text)==0) return 0;

// manipulate angle to 0<=angle<360 degrees
  while (angle<0) angle+=360.;
  while(angle>=360) angle-=360.;
  angle*=M_PI/180.;
  
// horizontal text made easy
  if (angle==0. && style.magnify==1.) { 
    return 
      XRotDrawHorizontalString(dpy,font,drawable,gc,x,y,text,align,bg);
  }
  
// get a rotated bitmap
  RotatedTextItem *item=
    XRotRetrieveFromCache(dpy,font,angle,text,align);
  if (!item) return 0;
  
// this gc has similar properties to the user's gc
  GC my_gc=XCreateGC(dpy,drawable,0L,0);
  XCopyGC(dpy,gc,GCForeground|GCBackground|GCFunction|GCPlaneMask,
          my_gc);

//alignment : which point (hot_x, hot_y) relative to bitmap centre
//coincides with user's specified point?
  
// y position
  if (align==TLEFT || align==TCENTRE || align==TRIGHT) {
    hot_y=static_cast<float>(item->rows_in)/2*style.magnify;
  } else if (align==MLEFT || align==MCENTRE || align==MRIGHT) hot_y=0;
  else if (align==BLEFT || align==BCENTRE || align==BRIGHT) {
    hot_y=-static_cast<float>(item->rows_in)/2*style.magnify;
  } else {
    hot_y=-(static_cast<float>(item->rows_in)/2
           -static_cast<float>(font->descent))*style.magnify;
  }
  
// x position
  if (align==TLEFT || align==MLEFT || align==BLEFT || align==NONE) {
    hot_x=-static_cast<float>(item->max_width)/2*style.magnify;
  } else if(align==TCENTRE || align==MCENTRE || align==BCENTRE) hot_x=0;
  else hot_x=static_cast<float>(item->max_width)/2*style.magnify;
  
// pre-calculate sin and cos
  float sin_angle=sin(angle);
  float cos_angle=cos(angle);
  
// rotate hot_x and hot_y around bitmap centre
  hot_xp= hot_x*cos_angle - hot_y*sin_angle;
  hot_yp= hot_x*sin_angle + hot_y*cos_angle;
  
// text background will be drawn using XFillPolygon
  if (bg) {
//  reserve space for XPoints
    XPoint *xpoints=OPERATOR_NEW_BRACKET(XPoint,4*item->nl);
    if (!xpoints) return 1;
    
//  rotate corner positions
    for (int i=0; i<4*item->nl; i++) {
      xpoints[i].x=
        static_cast<short int>(static_cast<float>(x) 
                              +( (item->corners_x[i]-hot_x)*cos_angle 
                                +(item->corners_y[i]+hot_y)*sin_angle));
      xpoints[i].y=
        static_cast<short int>(static_cast<float>(y) 
                              +(-(item->corners_x[i]-hot_x)*sin_angle 
                                +(item->corners_y[i]+hot_y)*cos_angle));
    }
    
//  we want to swap foreground and background colors here;
//    XGetGCValues() is only available in R4+
    Pixmap empty_stipple=XCreatePixmap(dpy, drawable, 1, 1, 1);
    GC depth_one_gc=XCreateGC(dpy, empty_stipple, 0L, 0);
    XSetForeground(dpy, depth_one_gc, 0);
    XFillRectangle(dpy, empty_stipple, depth_one_gc, 0, 0, 2, 2);
    XSetStipple(dpy, my_gc, empty_stipple);
    XSetFillStyle(dpy, my_gc, FillOpaqueStippled);
    XFillPolygon(dpy, drawable, my_gc, xpoints, 4*item->nl, Nonconvex,
                 CoordModeOrigin);
//  free our resources
    delete [] xpoints; xpoints=0;
    XFreeGC(dpy, depth_one_gc);
    XFreePixmap(dpy, empty_stipple);
  }
  
//where should top left corner of bitmap go ?
  int xp=static_cast<int>(static_cast<float>(x)
                       -(static_cast<float>(item->cols_out)/2 +hot_xp));
  int yp=static_cast<int>(static_cast<float>(y)
                       -(static_cast<float>(item->rows_out)/2 -hot_yp));
  
// by default we draw the rotated bitmap, solid
  Pixmap bitmap_to_paint=item->bitmap;

// handle user stippling
#ifndef X11R3
  {
    XGCValues values;
//  try and get some GC properties
    if (XGetGCValues(dpy, gc, GCStipple|GCFillStyle|GCForeground|
    GCBackground|GCTileStipXOrigin|GCTileStipYOrigin, &values)) {
//    only do this if stippling requested
      if ((values.fill_style==FillStippled ||
      values.fill_style==FillOpaqueStippled) && !bg) {
//      opaque stipple: draw rotated text in background colour
        if (values.fill_style==FillOpaqueStippled) {
          XSetForeground(dpy, my_gc, values.background);
          XSetFillStyle(dpy, my_gc, FillStippled);
          XSetStipple(dpy, my_gc, item->bitmap);
          XSetTSOrigin(dpy, my_gc, xp, yp);
          XFillRectangle(dpy, drawable, my_gc, xp, yp,
                         item->cols_out, item->rows_out);
          XSetForeground(dpy, my_gc, values.foreground);
        }
//      this will merge the rotated text and the user's stipple
        Pixmap new_bitmap=
          XCreatePixmap(dpy,drawable,item->cols_out,item->rows_out,1);

//      create a GC
        GC depth_one_gc=XCreateGC(dpy, new_bitmap, 0L, 0);
        XSetForeground(dpy, depth_one_gc, 1);
        XSetBackground(dpy, depth_one_gc, 0);

//      set the relative stipple origin 
        XSetTSOrigin(dpy, depth_one_gc, values.ts_x_origin-xp,
          values.ts_y_origin-yp);

//      fill the whole bitmap with the user's stipple */
        XSetStipple(dpy, depth_one_gc, values.stipple);
        XSetFillStyle(dpy, depth_one_gc, FillOpaqueStippled);
        XFillRectangle(dpy, new_bitmap, depth_one_gc,
                       0, 0, item->cols_out, item->rows_out);

// set stipple origin back to normal
        XSetTSOrigin(dpy, depth_one_gc, 0, 0);

// this will contain an inverse copy of the rotated text
        Pixmap inverse=XCreatePixmap(dpy, drawable, item->cols_out, 
          item->rows_out, 1);

// invert text
        XSetFillStyle(dpy, depth_one_gc, FillSolid);
        XSetFunction(dpy, depth_one_gc, GXcopyInverted);
        XCopyArea(dpy, item->bitmap, inverse, depth_one_gc,
                  0, 0, item->cols_out, item->rows_out, 0, 0);

// now delete user's stipple everywhere EXCEPT on text
        XSetForeground(dpy, depth_one_gc, 0);
        XSetBackground(dpy, depth_one_gc, 1);
        XSetStipple(dpy, depth_one_gc, inverse);
        XSetFillStyle(dpy, depth_one_gc, FillStippled);
        XSetFunction(dpy, depth_one_gc, GXcopy);
        XFillRectangle(dpy, new_bitmap, depth_one_gc,
                       0, 0, item->cols_out, item->rows_out);

// free resources 
        XFreePixmap(dpy, inverse);
        XFreeGC(dpy, depth_one_gc);

// this is the new bitmap
        bitmap_to_paint=new_bitmap;
      }
    }
  }
#endif /*X11R3*/

// paint text using stipple technique
  XSetFillStyle(dpy, my_gc, FillStippled);
  XSetStipple(dpy, my_gc, bitmap_to_paint);
  XSetTSOrigin(dpy, my_gc, xp, yp);
  XFillRectangle(dpy,drawable,my_gc,xp,yp,item->cols_out,
    item->rows_out);
  
// free our resources
  XFreeGC(dpy, my_gc);

// stippled bitmap no longer needed
  if (bitmap_to_paint!=item->bitmap) XFreePixmap(dpy, bitmap_to_paint);

#ifdef CACHE_XIMAGES
  XFreePixmap(dpy, item->bitmap);
#endif /*CACHE_XIMAGES*/

// if item isn't cached, destroy it completely
  if (!item->cached) XRotFreeTextItem(dpy,item);
  return 0;
}

//**********************************************************************
//  A front end to XRotPaintAlignedString:
//      -no alignment, no background
//**********************************************************************
int XRotDrawString(Display *dpy,XFontStruct *font,float angle,
Drawable drawable,GC gc,int x,int y,const char *str) {
  return XRotPaintAlignedString(dpy,font,angle,drawable,gc,x,y,str,NONE,
    0);
}

//**********************************************************************
//  A front end to XRotPaintAlignedString:
//      -no alignment, paints background
//**********************************************************************
int XRotDrawImageString(Display *dpy,XFontStruct *font,float angle,
Drawable drawable,GC gc,int x,int y,char *str) {
  return XRotPaintAlignedString(dpy,font,angle,drawable,gc,x,y,str,NONE,
    1);
}

//**********************************************************************
//  A front end to XRotPaintAlignedString:
//      -does alignment, no background
//**********************************************************************
int XRotDrawAlignedString(Display *dpy,XFontStruct *font,float angle,
Drawable drawable,GC gc,int x,int y,char *text,int align) {
  return XRotPaintAlignedString(dpy,font,angle,drawable,gc,x,y,text,
    align,0);
}

//**********************************************************************
//  A front end to XRotPaintAlignedString:
//      -does alignment, paints background
//**********************************************************************
int XRotDrawAlignedImageString(Display *dpy,XFontStruct *font,
float angle,Drawable drawable,GC gc,int x,int y,char * text,int align) {
  return XRotPaintAlignedString(dpy,font,angle,drawable,gc,x,y,text,
    align,1);
}

//**********************************************************************
// Calculate the bounding box some text will have when painted
//**********************************************************************
XPoint *XRotTextExtents(Display *dpy,XFontStruct *font,float angle,
int x,int y,char *text,int align) {
  const char *str2_a="\0", *str2_b="\n\0";
    
// manipulate angle to 0<=angle<360 degrees
  while(angle<0) angle+=360;
  while(angle>360) angle-=360;
  angle*=M_PI/180;
    
// count number of sections in string
  int nl=1;
  if (align!=NONE) {
    for (int i=0; i<strlen(text)-1; i++) if(text[i]=='\n') nl++;
  }
    
// ignore newline characters if not doing alignment
  const char *str2=(align==NONE ? str2_a : str2_b);
    
// find width of longest section
  char *str1=strdup(text);
  if (!str1) return 0;
    
  char *str3=my_strtok(str1, str2);

  int dir, asc, desc;
  XCharStruct overall;
  XTextExtents(font, str3, strlen(str3), &dir, &asc, &desc, &overall);

  int max_width=overall.rbearing;
    
// loop through each section
  do {
    str3=my_strtok(0, str2);
    if (str3) {
      XTextExtents(font,str3,strlen(str3),&dir,&asc,&desc,&overall);
      if (overall.rbearing>max_width) max_width=overall.rbearing;
    }
  } while(str3);
  free(str1);
    
// overall font height
  int height=font->ascent+font->descent;
    
// dimensions horizontal text will have
  int cols_in=max_width;
  int rows_in=nl*height;
    
// pre-calculate sin and cos
  float sin_angle=sin(angle);
  float cos_angle=cos(angle);
    
  float hot_x, hot_y;
// y position
  if (align==TLEFT || align==TCENTRE || align==TRIGHT) {
    hot_y=static_cast<float>(rows_in)/2*style.magnify;
  } else if(align==MLEFT || align==MCENTRE || align==MRIGHT) {
    hot_y=0;
  } else if(align==BLEFT || align==BCENTRE || align==BRIGHT) {
    hot_y=-static_cast<float>(rows_in)/2*style.magnify;
  } else {
    hot_y=-(static_cast<float>(rows_in)/2
           -static_cast<float>(font->descent))*style.magnify;
  }
    
// x position
  if (align==TLEFT || align==MLEFT || align==BLEFT || align==NONE) {
    hot_x=-static_cast<float>(max_width)/2*style.magnify;
  } else if(align==TCENTRE || align==MCENTRE || align==BCENTRE) hot_x=0;
  else hot_x=static_cast<float>(max_width)/2*style.magnify;
    
// reserve space for XPoints
  XPoint *xp_in=OPERATOR_NEW_BRACKET(XPoint,5);
  if (!xp_in) return 0;

  XPoint *xp_out=OPERATOR_NEW_BRACKET(XPoint,5);
  if (!xp_out) return 0;

// bounding box when horizontal, relative to bitmap centre
  xp_in[0].x=static_cast<short int>(
    -static_cast<float>(cols_in)*style.magnify/2-style.bbx_pad);
  xp_in[0].y=static_cast<short int>(
     static_cast<float>(rows_in)*style.magnify/2+style.bbx_pad);
  xp_in[1].x=static_cast<short int>(
     static_cast<float>(cols_in)*style.magnify/2+style.bbx_pad);
  xp_in[1].y=static_cast<short int>(
     static_cast<float>(rows_in)*style.magnify/2+style.bbx_pad);
  xp_in[2].x=static_cast<short int>(
     static_cast<float>(cols_in)*style.magnify/2+style.bbx_pad);
  xp_in[2].y=static_cast<short int>(
     -static_cast<float>(rows_in)*style.magnify/2-style.bbx_pad);
  xp_in[3].x=static_cast<short int>(
     -static_cast<float>(cols_in)*style.magnify/2-style.bbx_pad);
  xp_in[3].y=static_cast<short int>(
     -static_cast<float>(rows_in)*style.magnify/2-style.bbx_pad);
  xp_in[4].x=xp_in[0].x;
  xp_in[4].y=xp_in[0].y;

// rotate and translate bounding box
  for(int i=0; i<5; i++) {
    xp_out[i].x=static_cast<short int>(static_cast<float>(x) 
      + ( (static_cast<float>(xp_in[i].x)-hot_x)*cos_angle 
         +(static_cast<float>(xp_in[i].y)+hot_y)*sin_angle));
    xp_out[i].y=static_cast<short int>(static_cast<float>(y) 
      + (-(static_cast<float>(xp_in[i].x)-hot_x)*sin_angle 
         +(static_cast<float>(xp_in[i].y)+hot_y)*cos_angle));
  }
  delete [] xp_in; xp_in=0;
  return xp_out;
}



//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
XTWindow::XTWindow(const char *str,XColormap *cm,REAL rw,REAL rh,
BOOLEAN mono,BOOLEAN db) : 
name(0),double_buffer(db),display(0),screen(0),window(0),
parent(0),drawa(0),gc(0),xborder(0),yborder(0),width(0),height(0),
rwidth(ZERO),rheight(ZERO),x0(0),y0(0),grey_scale(mono),depth(0),
xcolormap(cm),cmap(None),buffer_map(0),buffer_gc(0),image(0) 
/* ,allow_expose(FALSE) */ {
//TRACER_CALL(t,"XTWindow::XTWindow");
  if (str) {
    name=OPERATOR_NEW_BRACKET(char,strlen(str)+1);
    strcpy(name,str);
  } else {
    name=OPERATOR_NEW_BRACKET(char,1);
    name[0]='\0';
  }

  Arg args[20];
  Cardinal num_args=0;
  XtSetArg(args[num_args],XtNtitle,name); num_args++;
  XtSetArg(args[num_args],XtNmappedWhenManaged,False); num_args++;
  XtSetArg(args[num_args],XtNallowShellResize,False); num_args++;
  if (cm->getColormap()!=
  DefaultColormap(cm->getDisplay(),DefaultScreen(cm->getDisplay()))) {
    XtSetArg(args[num_args],XmNvisual,cm->getVisual()); num_args++;
  }
  XtSetArg(args[num_args],XtNdepth,cm->getDepth()); num_args++;
  XtSetArg(args[num_args],XmNcolormap,cm->getColormap()); num_args++;
  if (grey_scale) {
//  XtSetArg(args[num_args],XtNdepth,1); num_args++;
    ASSERT(cm->getDepth()==1);
  }
  if (grand_parent==0) {
//  TRACER_CALL(t,"XTWindow::XTWindow XtOpenApplication");
    int argc=0;
    String *argv=reinterpret_cast<String*>(XtMalloc(sizeof(String)));
    grand_parent=XtOpenApplication(0,"XtWindowApplicationShell",
      reinterpret_cast<XrmOptionDescRec*>(0),0,&argc,argv,
      const_cast<char**>(fallback_resources),applicationShellWidgetClass,
      args,num_args);
    general_appcontext=XtWidgetToApplicationContext(grand_parent);
    XtFree(reinterpret_cast<String>(argv));
  } 
#ifdef DEBUG
//cout << "\tgrand_parent = " << grand_parent << endl;
//cout << "\tgeneral_appcontext = " << general_appcontext 
//     << " " << &general_appcontext << endl;
#endif
  parent=XtCreatePopupShell("parent popup",topLevelShellWidgetClass,
    grand_parent,args,num_args);
  ref_ctr++;

  {
    display=XtDisplay(parent);
    CHECK_POINTER(display);
//  screen=DefaultScreen(display);
    screen=cm->getScreen();
    int dw=DisplayWidth(display,screen);
    int dh=DisplayHeight(display,screen);
    int dm=min(dw,dh);
    if (rw>ONE) rw=ONE;
    if (rh>ONE) rh=ONE;
    if (rw<=ZERO || rh<=ZERO) {
      width=NTSC_WIDTH;
      height=NTSC_HEIGHT;
    } else {
      width=int(rw*dm);
      height=int(rh*dm);
    }
    xborder=width/20;
    yborder=height/20;
    rwidth=REAL(width);
    rheight=REAL(height);
    XtVaSetValues(parent,
      XtNwidth,width,
      XtNheight,height,
      reinterpret_cast<XtPointer>(0));

    depth=0;
    XtVaGetValues(parent,
      XmNdepth,&depth,
      reinterpret_cast<XtPointer>(0));
    grey_scale=(xcolormap==0 || depth==1);

    if (xcolormap) {
      XtVaGetValues(parent,
        XmNcolormap,&cmap,
        reinterpret_cast<XtPointer>(0));
      Colormap xcm=xcolormap->getColormap();
      if (cmap!=xcm) {
        cmap=xcm;
        XtVaSetValues(parent,
          XmNcolormap,xcm,
          reinterpret_cast<XtPointer>(0));
      }
    }

//  options are XmRESIZE_ANY,XmRESIZE_GROW,XmRESIZE_NONE
    Cardinal num_args=0;
    XtSetArg(args[num_args],XmNresizePolicy,XmRESIZE_ANY); num_args++;
    XtSetArg(args[num_args],XtNsaveUnder,True); num_args++;
    drawa=XmCreateDrawingArea(parent,
      const_cast<char*>("XtWindowDrawingArea"),args,num_args);
    XtManageChild(drawa);
    XtRealizeWidget(parent);
    window=XtWindow(drawa);
#ifdef DEBUG
//  cout << "\tdisplay = " << display << endl;
//  cout << "\twindow = " << window << endl;
#endif
    XtVaSetValues(drawa,
      XmNresizePolicy,XmRESIZE_NONE,
      reinterpret_cast<XtPointer>(0));
  }
  
  XSelectInput(display, window, ExposureMask | KeyPressMask | 
               ButtonPressMask | Button1MotionMask | 
               Button2MotionMask | Button3MotionMask | 
               ButtonReleaseMask | StructureNotifyMask | 
               SubstructureRedirectMask | VisibilityChangeMask );

  gc=makeGC(window);
  ASSERT(gc);
  if (double_buffer) {
    buffer_map=makePixmap();
    buffer_gc=makeGC(buffer_map);
    ASSERT(buffer_gc);
  }

  XtAddCallback(drawa, XmNexposeCallback, &exposeCallback, this);
  XtAddCallback(drawa, XmNinputCallback, &inputCallback, this);
  XtAddCallback(drawa, XmNresizeCallback, &resizeCallback, this);

  x0=0;
  y0=0;

  XtMapWidget(parent);
  XFlush(display);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
XTWindow::~XTWindow() {
//TRACER_CALL(t,"XTWindow::~XTWindow");
  if (display) {
    XDestroyWindow(display,XtWindow(drawa));
    ref_ctr--;
    XDestroyWindow(XtDisplay(parent),XtWindow(parent));
    parent=0;
    XFlush(display);
    if (double_buffer) {
      XFreeGC(display,buffer_gc);
      XFreePixmap(display,buffer_map);
    }
    XFreeGC( display, gc );
    display=0;
    if (ref_ctr==0) {
      XDestroyWindow(XtDisplay(grand_parent),XtWindow(grand_parent));
      grand_parent=0;
//    XtDestroyApplicationContext(general_appcontext);
      general_appcontext=0;
    }
  }
  delete [] name; name=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::expose() {
//TRACER_CALL(t,"XTWindow::expose");
//if (double_buffer && allow_expose) refresh(0,0,width,height);
  if (double_buffer) refresh(0,0,width,height);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Pixmap XTWindow::makePixmap() const {
  Pixmap pm=XCreatePixmap(display,window,width,height,depth);
  ASSERT(pm);
  return pm;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
GC XTWindow::makeGC(Drawable drawable) const {
//TRACER_CALL(t,"XTWindow::makeGC");
  ASSERT(display);
  ASSERT(drawable);

  XGCValues gcvalues;
//gcvalues.foreground=BlackPixel(display,screen); // also XSetForeground
//gcvalues.background=WhitePixel(display,screen); // also XSetBackground
  gcvalues.foreground=xcolormap->getColor(1);
  gcvalues.background=xcolormap->getColor(0);
  gcvalues.line_width=1;               // also
  gcvalues.line_style=LineSolid;       //   see
  gcvalues.cap_style=CapRound;         //     XSetLineAttributes
  gcvalues.join_style=JoinRound;       //       .
  XFontStruct *the_font = XLoadQueryFont(display,default_font);
  if (!the_font) {
    the_font = XLoadQueryFont(display,"fixed");
    ASSERT(the_font);
  }
  gcvalues.font = the_font->fid;                  // also see XSetFont
  unsigned long mask= GCForeground | GCBackground | GCLineWidth 
                    | GCLineStyle | GCCapStyle | GCJoinStyle | GCFont;
  GC the_gc = XCreateGC(display,drawable,mask,&gcvalues);
  XFreeFont(display,the_font);
  return the_gc;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::setfgColour(Pixel pixel,GC the_gc) const {
  XSetForeground(display,the_gc,pixel);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::setfgColour(Pixel pixel) {
  if (double_buffer) setfgColour(pixel,buffer_gc);
  else setfgColour(pixel,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::setfgColor(const char *colorname,GC the_gc) {
  if (monochrome()) return;
  setfgColour(allocColor(colorname),the_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::setbgColor(const char *colorname,Drawable drawable,
GC the_gc) {
  if (monochrome()) return;
  setbgColour(allocColor(colorname),drawable,the_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::setfgColor(REAL frac,GC the_gc) const {
  if (monochrome()) return;
  frac=max(ZERO,min(ONE,frac));
  int color=pickColor(frac);
  setfgColour(xcolormap->getColor(color),the_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::setfgColor(const REAL *frac) const {
  if (double_buffer) setfgColor(*frac,buffer_gc);
  else setfgColor(*frac,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::setbgColor(REAL frac,Drawable drawable,GC the_gc) const {
  if (monochrome()) return;
  frac=max(ZERO,min(ONE,frac));
  int color=pickColor(frac);
  color=max(0,min(color,xcolormap->getNumberColors()));
  setbgColour(xcolormap->getColor(color),drawable,the_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::setbgColor(const REAL *frac) const {
  if (double_buffer) setbgColor(*frac,buffer_map,buffer_gc);
  else setbgColor(*frac,window,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM == 1)
void XTWindow::setfgColor(INTEGER ic,INTEGER,GC the_gc) const {
  ic=max(1,min(xcolormap->getNumberColors(),ic));
  Pixel pix=xcolormap->getColor(ic);
  setfgColour(pix,the_gc);
}
#else
void XTWindow::setfgColor(INTEGER ic,INTEGER im,GC the_gc) const {
  setfgColor(REAL(ic+1)/REAL(im+1),the_gc);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM == 1)
void XTWindow::setfgColor(const INTEGER *ic,const INTEGER *im) const {
  if (double_buffer) setfgColor(*ic,*im,buffer_gc);
  else setfgColor(*ic,*im,gc);
}
#else
void XTWindow::setfgColor(const INTEGER *ic,const INTEGER *im) const {
  REAL val=REAL(*ic+1)/REAL(*im+1);
  setfgColor(&val);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void  XTWindow::setFont(const char *font) const {
  XGCValues gcvalues;
  XFontStruct *the_font;
  if ((the_font = XLoadQueryFont(display,font)) == 0) {
     cerr << "Cannot open font: " << font << "\n";
     exit(1);
  };
  gcvalues.font = the_font->fid;
  if (double_buffer) XChangeGC( display, buffer_gc, GCFont, &gcvalues );
  else XChangeGC( display, gc, GCFont, &gcvalues );
  XFreeFont(display,the_font);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Pixel XTWindow::allocColor(const char *colorname) {
  CHECK_TEST(!monochrome());
  CHECK_POINTER(colorname);
  XColor color, exact;
  ASSERT(XAllocNamedColor(display,cmap,colorname,&color,&exact));
  return color.pixel;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::putString(const REAL *px,const REAL *py,const char *str,
const REAL *angle) const {
  XFontStruct *the_font;
  if ((the_font = XLoadQueryFont(display,default_font)) == 0) {
     cerr << "Cannot open font " << default_font << "\n";
     exit(1);
  };
  int x=ixLoc(*px);
  int y=iyLoc(*py);
  float ang=static_cast<float>(*angle);
  if (doubleBuffer()) {
    XRotDrawString(display,the_font,ang,buffer_map,buffer_gc,x,y,str);
  } else {
    XRotDrawString(display,the_font,ang,window,gc,x,y,str);
  }
  XFreeFont(display,the_font);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::writeXBorder(const char *str) const {
  XFontStruct *the_font;
  if ((the_font = XLoadQueryFont(display,default_font)) == 0) {
     cerr << "Cannot open font " << default_font << "\n";
     exit(1);
  };
  int x=width/2;
  int y=myY(yborder/2);
  if (doubleBuffer()) {
    XRotDrawString(display,the_font,0.,buffer_map,buffer_gc,x,y,str);
  } else {
    XRotDrawString(display,the_font,0.,window,gc,x,y,str);
  }
  XFreeFont(display,the_font);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::writeYBorder(const char *str) const {
  XFontStruct *the_font;
  if ((the_font = XLoadQueryFont(display,default_font)) == 0) {
     cerr << "Cannot open font " << default_font << "\n";
     exit(1);
  };
  if (doubleBuffer()) {
    XRotDrawString(display,the_font,90.,buffer_map,buffer_gc,xborder/2,
      myY(height/2),str);
  } else {
    XRotDrawString(display,the_font,90.,window,gc,xborder/2,
      myY(height/2),str);
  }
  XFreeFont(display,the_font);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::setbgColour(Pixel pixel) {
  if (double_buffer) {
    XGCValues gcvalues;
    unsigned long mask=GCForeground;
    ASSERT(XGetGCValues(display,buffer_gc,mask,&gcvalues));
    Pixel foreground=gcvalues.foreground;
    XSetForeground(display,buffer_gc,pixel);
    XFillRectangle(display,buffer_map,buffer_gc,0,0,width,height);
    XSetForeground(display,buffer_gc,foreground);
  } else {
    XSetWindowBackground(display,window,pixel);
    XClearWindow(display, window);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::setbgColour(Pixel pixel,Drawable drawable,GC the_gc) 
const {
  XGCValues gcvalues;
  unsigned long mask=GCForeground;
  ASSERT(XGetGCValues(getDisplay(),the_gc,mask,&gcvalues));
  Pixel foreground=gcvalues.foreground;
  XSetForeground(getDisplay(),the_gc,pixel);
  XFillRectangle(getDisplay(),drawable,the_gc,0,0,width,height);
  XSetForeground(getDisplay(),the_gc,foreground);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::setfgColor(const char *colorname) const {
  if (monochrome()) return;
  XTWindow *copy=const_cast<XTWindow*>(this);
  copy->setfgColour(copy->allocColor(colorname));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::setbgColor(const char *colorname) const {
  if (monochrome()){
    return;
  }
  XTWindow *copy=const_cast<XTWindow*>(this);
  copy->setbgColour(copy->allocColor(colorname));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int XTWindow::pickColor(REAL frac) const {
  if (xcolormap) {
#if (SPACEDIM==1)
    return static_cast<int>(
      frac*(static_cast<REAL>(xcolormap->getNumberColors())-HALF) );
#else
    return static_cast<int>(
      2.+frac*static_cast<REAL>(xcolormap->getNumberColors()-3) );
#endif
  } else return 1;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::newPage(Drawable drawable,GC the_gc) {
//allow_expose=FALSE;
  if (!monochrome()) {
    setbgColour(xcolormap->getColor(0),drawable,the_gc);
  }
  XGCValues gcvalues;
  unsigned long mask=GCForeground | GCBackground;
  ASSERT(XGetGCValues(display,the_gc,mask,&gcvalues));
  Pixel foreground=gcvalues.foreground;
  Pixel background=gcvalues.background;
  XSetForeground(display,the_gc,background);
  XFillRectangle(display,drawable,the_gc,0,0,width,height);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::newPage() {
//allow_expose=FALSE;
  if (!monochrome()) setbgColour(xcolormap->getColor(0));
  if (double_buffer) {
    XGCValues gcvalues;
    unsigned long mask=GCForeground | GCBackground;
    ASSERT(XGetGCValues(display,buffer_gc,mask,&gcvalues));
    Pixel foreground=gcvalues.foreground;
    Pixel background=gcvalues.background;
    XSetForeground(display,buffer_gc,background);
    XFillRectangle(display,buffer_map,buffer_gc,0,0,width,height);
    XSetForeground(display,buffer_gc,foreground);
  } else XClearWindow( display, window );
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::drawLine(const REAL *px,const REAL *py,
const BOOLEAN *puse_buffer) const {
  XTWindow *copy=const_cast<XTWindow*>(this);
  BOOLEAN use_buffer=(puse_buffer ? *puse_buffer : TRUE);
  int x1=ixLoc(*px); int y1=iyLoc(*py);
  if (doubleBuffer() && use_buffer) {
    XDrawLine(display,buffer_map,buffer_gc,x0,y0,x1,y1);
  } else {
    XDrawLine(display, window, gc, x0, y0, x1, y1 );
  }
  copy->x0 = x1; copy->y0 = y1;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::refresh(int x,int y,int w,int h) const {
  if (double_buffer) {
    XCopyArea(display,buffer_map,window,gc,x,y,w,h,x,y);
  }
} // refresh & expose can't be called from the main process
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::refreshFrom(const Buffer &xb) const {
  XCopyArea(display,xb.getPixmap(),window,gc,0,0,width,height,0,0);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::flush(BOOLEAN use_buffer) { 
  if (double_buffer && use_buffer) refresh(0,0,width,height);
  XFlush( display ); 
//allow_expose=TRUE;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//this should be called before createRaster
void XTWindow::writePaletteName(FILE *os) {
  const char* pal_name=xcolormap->getPalette()->getName();
  int length=strlen(pal_name)+1;
  fwrite(reinterpret_cast<char*>(&length), sizeof(int), 1, os);
  fwrite( pal_name, sizeof(char), length, os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::createRaster(FILE *os) {
  CHECK_POINTER(os);
  CHECK_TEST(!image);

  if (double_buffer) {
    image=
      XGetImage(display,buffer_map,0,0,width,height,AllPlanes,ZPixmap); 
  } else {
    image=XGetImage(display,window,0,0,width,height,AllPlanes,ZPixmap); 
  }
  ASSERT(image);

  REAL dw=REAL(DisplayWidth(display,screen));
  REAL dh=REAL(DisplayHeight(display,screen));
  REAL dm=min(dw,dh);
  REAL rw=rwidth/dm;
  REAL rh=rheight/dm;
  fwrite(reinterpret_cast<char*>(&rw),sizeof(REAL),1,os);
  fwrite(reinterpret_cast<char*>(&rh),sizeof(REAL),1,os);
  int length=strlen(name)+1;
  fwrite(reinterpret_cast<char*>(&length),sizeof(int),1,os);
  fwrite( name, sizeof(char), length, os);

  fwrite(reinterpret_cast<char*>(&(image->xoffset)),sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->format)),sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->byte_order)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bitmap_unit)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bitmap_bit_order)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bitmap_pad)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->depth)),sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bytes_per_line)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bits_per_pixel)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->red_mask)),
    sizeof(unsigned long),1,os);
  fwrite(reinterpret_cast<char*>(&(image->green_mask)),
    sizeof(unsigned long),1,os);
  fwrite(reinterpret_cast<char*>(&(image->blue_mask)),
    sizeof(unsigned long),1,os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::openRaster(FILE *is) {
  CHECK_POINTER(is);
  CHECK_TEST(!image);
  int length;
  fread(reinterpret_cast<char*>(&length),sizeof(int),1,is);
  char *wn=OPERATOR_NEW_BRACKET(char,length);
  fread( wn, 1, length, is);
  ASSERT(strcmp(name,wn)==0); 
  delete wn;
  wn=0;

  int image_width=width,image_height=height;
  int image_xoffset,image_format;
  int image_byte_order,image_bitmap_unit,image_bitmap_bit_order;
  int image_bitmap_pad,image_depth,image_bytes_per_line;
  int image_bits_per_pixel;
  fread(reinterpret_cast<char*>(&image_xoffset),sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_format),sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_byte_order),sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bitmap_unit),sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bitmap_bit_order),
    sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bitmap_pad),
    sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_depth),
    sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bytes_per_line),
    sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bits_per_pixel),
    sizeof(int),1,is);

#ifdef DEBUG
  CHECK_TEST(ImageByteOrder(display)==image_byte_order);
  CHECK_TEST(BitmapUnit(display)==image_bitmap_unit);
  CHECK_TEST(BitmapBitOrder(display)==image_bitmap_bit_order);
  CHECK_TEST(BitmapBitOrder(display)==image_bitmap_bit_order);

  int number_depths;
  int *depth_list=XListDepths(display,screen,&number_depths);
  if (depth_list) {
    BOOLEAN found=FALSE;
    for (int i=0;i<number_depths;i++) {
      if (image_depth==depth_list[i]) found=TRUE;
    }
    CHECK_TEST(found);
  }
  XFree(depth_list);
  CHECK_TEST(image_width<=width);
  CHECK_TEST(image_height<=height);
#endif

  image=XCreateImage(display,0,image_depth,image_format,
    image_xoffset,0,image_width,image_height,image_bitmap_pad,
    image_bytes_per_line);
  ASSERT(image);
  image->data=0;

  fread(reinterpret_cast<char*>(&image->red_mask),
    sizeof(unsigned long),1,is);
  fread(reinterpret_cast<char*>(&image->green_mask),
    sizeof(unsigned long),1,is);
  fread(reinterpret_cast<char*>(&image->blue_mask),
    sizeof(unsigned long),1,is);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::writeRaster(FILE *os) {
  CHECK_POINTER(image);
  CHECK_POINTER(os);

  XDestroyImage(image);
  if (double_buffer) {
    image=
      XGetImage(display,buffer_map,0,0,width,height,AllPlanes,ZPixmap); 
  } else {
    image=XGetImage(display,window,0,0,width,height,AllPlanes,ZPixmap); 
  }
  ASSERT(image);

  int buffer_size=(image->bytes_per_line)*(image->height);
  if (image->format!=ZPixmap) buffer_size *= image->depth;

  fwrite(reinterpret_cast<char*>(&buffer_size), sizeof(int), 1, os);
  fwrite( image->data, buffer_size, 1, os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::readRaster(FILE *is) {
#ifdef DEBUG
  CHECK_TEST(image);
  CHECK_TEST(is);
#else
  if (!image) return;
  if (!is) return;
#endif

  size_t buffer_size=(image->bytes_per_line)*(image->height);
  if (image->format!=ZPixmap) buffer_size *= image->depth;
  size_t bs;
  char *buffer=OPERATOR_NEW_BRACKET(char,buffer_size);
  image->data=buffer;
  while (TRUE) {
    if (fread(reinterpret_cast<char*>(&bs),sizeof(int),1,is)<1) break;
#ifdef DEBUG
    CHECK_TEST(buffer_size==bs);
#else
    if (buffer_size!=bs) break;
#endif

    if (fread( buffer, 1, bs, is)<bs) break;

    if (double_buffer) {
     XPutImage(display,buffer_map,buffer_gc,image,0,0,0,0,width,height);
    } else XPutImage(display,window,gc,image,0,0,0,0,width,height); 
    flush();
  }
  delete buffer;
  image->data=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::closeRaster() {
  CHECK_POINTER(image);
  XDestroyImage(image);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::writeXPM(const char *filename) {
  CHECK_POINTER(filename);
  CHECK_TEST(!image);
  if (double_buffer) {
    image=
      XGetImage(display,buffer_map,0,0,width,height,AllPlanes,ZPixmap);
  } else {
    image=XGetImage(display,window,0,0,width,height,AllPlanes,ZPixmap);
  }
  CHECK_TEST(image!=0);

  char string[LENGTH_NAME];
  snprintf(string,LENGTH_NAME,"%s.xpm",filename);
  int ret=XpmWriteFileFromImage(display,string,image,0,0);
  CHECK_TEST(ret==XpmSuccess);
  XDestroyImage(image);
  image=0;
}
/*
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::nextEvent(long event_mask) {
  Status status=TRUE;
  XEvent event;
  while (status==FALSE) {
    XNextEvent(display, &event);
    status=parseEvent(event,event_mask);
  }
}
*/
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::eventLoop() {
//TRACER_CALL(t,"XTWindow::eventLoop");
  XtAppMainLoop(general_appcontext);
/*
  XEvent event;
  long mask= ExposureMask | KeyPressMask | 
             ButtonPressMask | Button1MotionMask | 
             Button2MotionMask | Button3MotionMask | 
             ButtonReleaseMask | StructureNotifyMask | 
             SubstructureRedirectMask | VisibilityChangeMask;
  while (TRUE) {
    XtAppNextEvent(appcontext,&event);
#ifdef DEBUG
//  parseEvent(event,mask);
#endif
    XtDispatchEvent(&event);
  }
*/
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Status XTWindow::parseEvent(const XEvent &event,long event_mask) {
//TRACER_CALL(t,"XTWindow::parseEvent");
  Status status=FALSE;
  switch (event.type) {
    case KeyPress:
#ifdef DEBUG
//    cout << "\tKeyPress event" << endl;
#endif
      if (event_mask & KeyPressMask) status=TRUE;
      break;
    case KeyRelease:
#ifdef DEBUG
//    cout << "\tKeyRelease event" << endl;
#endif
      if (event_mask & KeyReleaseMask) status=TRUE;
      break;
    case ButtonPress:
#ifdef DEBUG
//    cout << "\tButtonPress event" << endl;
#endif
      if (event_mask & ButtonPressMask) status=TRUE;
      break;
    case ButtonRelease:
#ifdef DEBUG
//    cout << "\tButtonRelease event" << endl;
#endif
      if (event_mask & ButtonReleaseMask) status=TRUE;
      break;
    case MotionNotify:
#ifdef DEBUG
//    cout << "\tMotionNotify event" << endl;
#endif
      if (event_mask & (PointerMotionMask | ButtonMotionMask | 
      Button1MotionMask | Button2MotionMask | Button3MotionMask | 
      Button4MotionMask | Button5MotionMask)) status=TRUE;
      break;
    case EnterNotify:
#ifdef DEBUG
//    cout << "\tEnterNotify event" << endl;
#endif
      if (event_mask & EnterWindowMask) status=TRUE;
      break;
    case LeaveNotify:
#ifdef DEBUG
//    cout << "\tLeaveNotify event" << endl;
#endif
      if (event_mask & LeaveWindowMask) status=TRUE;
      break;
    case FocusIn:
    case FocusOut:
#ifdef DEBUG
//    cout << "\tFocusIn or FocusOut event" << endl;
#endif
      if (event_mask & FocusChangeMask) status=TRUE;
      break;
    case KeymapNotify:
#ifdef DEBUG
//    cout << "\tKeymapNotify event" << endl;
#endif
      if (event_mask & KeymapStateMask) status=TRUE;
      break;
    case Expose:
#ifdef DEBUG
//    cout << "\tExpose event" << endl;
#endif
      if (event_mask & ExposureMask) status=TRUE;
#ifdef DEBUG
//    {
//    const XExposeEvent *xee=
//      reinterpret_cast<const XExposeEvent*>(&event);
//    cout << "\nExposeEvent : " << endl;
//    cout << "\tdisplay = " << xee->display << endl;
//    cout << "\twindow = " << xee->window << endl;
//    cout << "\tx,y = " << xee->x << " " << xee->y << endl;
//    cout << "\twidth,height = " << xee->width << " " << xee->height
//         << endl;
//    }
#endif
      break;
    case GraphicsExpose:
    case NoExpose:
#ifdef DEBUG
//    cout << "\tGraphicsExpose or NoExpose event" << endl;
#endif
      break;
    case CirculateRequest:
    case ConfigureRequest:
    case MapRequest:
#ifdef DEBUG
//    cout << "\tCirculateRequest or ConfigureRequest or MapRequest event" << endl;
#endif
      if (event_mask & SubstructureRedirectMask) status=TRUE;
      break;
    case ResizeRequest:
#ifdef DEBUG
//    cout << "\tResizeRequest event" << endl;
#endif
      if (event_mask & ResizeRedirectMask) status=TRUE;
      break;
    case CirculateNotify:
    case ConfigureNotify:
    case CreateNotify:
    case DestroyNotify:
    case GravityNotify:
    case MapNotify:
    case ReparentNotify:
    case UnmapNotify:
#ifdef DEBUG
//    cout << "\tCirculateNotify or ConfigureNotify or CreateNotify or DestroyNotify or GravityNotify or MapNotify or ReparentNotify or UnmapNotify event" << endl;
#endif
      if (event_mask & (StructureNotifyMask | SubstructureNotifyMask))
        status=TRUE;
      break;
    case VisibilityNotify:
#ifdef DEBUG
//    cout << "\tVisibilityNotify event" << endl;
#endif
      if (event_mask & VisibilityChangeMask) status=TRUE;
      break;
    case ColormapNotify:
#ifdef DEBUG
//    cout << "\tColormapNotify event" << endl;
#endif
      if (event_mask & ColormapChangeMask) status=TRUE;
      break;
    case ClientMessage:
    case MappingNotify:
    case SelectionClear:
    case SelectionNotify:
    case SelectionRequest:
#ifdef DEBUG
//    cout << "\tClientMessage or MappingNotify or SelectionClear or SelectionNotify or SelectionRequest event" << endl;
#endif
      break;
    case PropertyNotify:
#ifdef DEBUG
//    cout << "\tPropertyNotify event" << endl;
#endif
      if (event_mask & PropertyChangeMask) status=TRUE;
      break;
  }
  return status;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BOOLEAN XTWindow::buttonIsPressed(int &x,int &y) {
  if (which_button_pressed==0) return FALSE;
  XEvent event;
  while (TRUE) {
    XNextEvent(display, &event);
    if (event.xany.window == window) {
      switch (event.type) {
        case ButtonRelease:
          if (which_button_pressed==event.xbutton.button) return FALSE;
          break;
        case MotionNotify:
          x = event.xbutton.x;
          y = event.xbutton.y;
          return TRUE;
          break;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BOOLEAN XTWindow::buttonIsPressed(REAL &rx,REAL &ry) {
  int x=0,y=0;
  BOOLEAN pressed=buttonIsPressed(x,y);
  if (pressed) {
    rx=static_cast<REAL>(x-xborder)/static_cast<REAL>(width-xborder);
    ry=static_cast<REAL>(myY(y+yborder))
      /static_cast<REAL>(height-yborder);
  }
  return pressed;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int XTWindow::getMouse(REAL &rx, REAL &ry) {
  int ib,x,y;
  ib=getMouse(x,y);
  rx=static_cast<REAL>(x-xborder)/static_cast<REAL>(width-xborder);
  ry=static_cast<REAL>(myY(y+yborder))
    /static_cast<REAL>(height-yborder);
  return ib;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int XTWindow::getMouse(int &x,int &y) {
  XEvent event;
  while(XCheckTypedWindowEvent(display,window,ButtonPress,&event));

  which_button_pressed=0;
  while (which_button_pressed == 0) {
    XNextEvent(display, &event);
    if (event.xany.window == window && event.type == ButtonPress) {
      which_button_pressed = event.xbutton.button;
      x = event.xbutton.x;
      y = event.xbutton.y;
    }
  }
  return which_button_pressed;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::colorRect(int color,int w,int h,int x,int y,
Drawable drawable,GC the_gc) const {
  if (monochrome()) return;
  color=max(0,min(color,xcolormap->getNumberColors()));
  XSetForeground(display,the_gc,xcolormap->getColor(color));
  XFillRectangle(display,drawable,the_gc,x,y,w,h);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::colorRect(int color,int w,int h,int x,int y) const {
  if (double_buffer) colorRect(color,w,h,x,y,buffer_map,buffer_gc);
  else colorRect(color,w,h,x,y,window,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::colorRect(REAL ratio,REAL rw,REAL rh,REAL rx,REAL ry,
Drawable drawable,GC the_gc) const {
  if (monochrome()) return;
  ratio=max(ZERO,min(ONE,ratio));
  int color=1+int(ratio*(xcolormap->getNumberColors()-2));
  int lox=ixLoc(rx);
  int hix=ixLoc(rx+rw);
  int loy=iyLoc(ry);
  int hiy=iyLoc(ry+rh);
  colorRect(color,hix-lox,hiy-loy,lox,loy,drawable,the_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::colorRect(const REAL *ratio,const REAL *rw,
const REAL *rh,const REAL *rx,const REAL *ry) const {
  if (double_buffer) {
    colorRect(*ratio,*rw,*rh,*rx,*ry,buffer_map,buffer_gc);
  } else colorRect(*ratio,*rw,*rh,*rx,*ry,window,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::colorRect(const char *color_name,REAL rw,REAL rh,REAL rx,
REAL ry,Drawable drawable,GC the_gc) {
  Pixel pix=allocColor(color_name);
  if (monochrome()) return;
  XSetForeground(display,the_gc,pix);
  int lox=ixLoc(rx);
  int hix=ixLoc(rx+rw);
  int loy=iyLoc(ry);
  int hiy=iyLoc(ry+rh);
  XFillRectangle(display,drawable,the_gc,lox,loy,hix-lox,hiy-loy);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::colorRect(const char *color_name,REAL rw,REAL rh,REAL rx,
REAL ry) {
  if (double_buffer) {
    colorRect(color_name,rw,rh,rx,ry,buffer_map,buffer_gc);
  } else colorRect(color_name,rw,rh,rx,ry,window,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::colorPolygon(int npts,XPoint p[],Drawable drawable,
GC the_gc) const {
  XFillPolygon(display,drawable,the_gc,p,npts,Nonconvex,
               CoordModeOrigin);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::colorPolygon(int color,int npts,XPoint p[],
Drawable drawable,GC the_gc) const {
  if (monochrome()) return;
  int num_cmap_colors=xcolormap->getNumberColors();
  CHECK_BOUNDS(color,0,num_cmap_colors) ;
  XSetForeground(display,the_gc,xcolormap->getColor(color));
  XFillPolygon(display,drawable,the_gc,p,npts,Nonconvex,
               CoordModeOrigin);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::colorPolygon(int color,int npts,XPoint p[]) const {
  if (double_buffer) colorPolygon(color,npts,p,buffer_map,buffer_gc);
  else colorPolygon(color,npts,p,window,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::colorPolygon(INTEGER npts,const REAL *xfrac,
const REAL *yfrac,Drawable drawable,GC the_gc) const {
  XPoint *p=OPERATOR_NEW_BRACKET(XPoint,npts);
  for (int i=0;i<npts;i++) {
    p[i].x=short(ixLoc(xfrac[i]));
    p[i].y=short(iyLoc(yfrac[i]));
  }
  colorPolygon(npts,p,drawable,the_gc);
  delete p;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::colorPolygon(REAL ratio,INTEGER npts,const REAL *xfrac,
const REAL *yfrac,Drawable drawable,GC the_gc) const {
//TRACER_CALL(t,"XTWindow::colorPolygon");
  if (monochrome()) return;
  int num_cmap_colors=xcolormap->getNumberColors();
  int color=0;
  if (ratio>=ZERO && ratio<=ONE) {
    color=2+static_cast<int>(ratio*(num_cmap_colors-3));
  } else if (ratio<ZERO) color=2;
  else color=num_cmap_colors-1;
#ifdef DEBUG
//cout << "\tratio = " << ratio << endl;
//cout << "\tcolor = " << color << endl;
#endif
  XSetForeground(display,the_gc,xcolormap->getColor(color));
  colorPolygon(npts,xfrac,yfrac,drawable,the_gc);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::colorPolygon(const REAL *ratio,const INTEGER *npts,
const REAL *xfrac,const REAL *yfrac) const {
//TRACER_CALL(t,"XTWindow::colorPolygon");
  if (double_buffer) {
    colorPolygon(*ratio,*npts,xfrac,yfrac,buffer_map,buffer_gc);
  } else colorPolygon(*ratio,*npts,xfrac,yfrac,window,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::colorPolygon(const INTEGER *npts,const REAL *xfrac,
const REAL *yfrac) const {
  if (double_buffer) {
    colorPolygon(*npts,xfrac,yfrac,buffer_map,buffer_gc);
  } else colorPolygon(*npts,xfrac,yfrac,window,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::printOn(ostream& os) const {
  os << "XTWindow: name = " << name << endl; 
//os << "\tdisplay_name = " << display_name << endl;
//printDisplayInfo(os,display_name,display);

  int num_screens=ScreenCount(display);
  os << "number of screens available = " << num_screens << "\n"
     << "default screen = " << DefaultScreen(display) << "\n"
     << "screen used by this window = " << screen << endl;
  for (int i=0;i<num_screens;i++) printScreenInfo(os,display,screen,i);

//printWindowInfo(os,display,window);
//printWMProperties(os,display,window);
//printHosts(display,os);
//printGC(display,gc,os);
//if (double_buffer) printGC(display,buffer_gc,os);
//char *dn;
//if (XFetchName(display,window,&dn)) {
//  os << "display_name = " << *dn <<endl;
//} else os << "name is unknown" << endl;
//XFree(dn);
  os << "x0 = " << x0 << "\n"
     << "y0 = " << y0 << "\n"
     << "depth = " << depth << endl;
//os << "cmap = " << cmap << "\n"
//   << "xcolormap = " << (XColormap*) xcolormap << endl;
  os << "double_buffer = " << PRINT_BOOLEAN(double_buffer) << endl;
  VirtualWindow12::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
XTWindow::Buffer::Buffer(XTWindow *xw) : VirtualWindow12(),parent(xw),
buffer_map(0),buffer_gc(0),x0(0),y0(0) {
  buffer_map=parent->makePixmap();
  buffer_gc=parent->makeGC(buffer_map);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
XTWindow::Buffer::~Buffer() {
  if (getDisplay()) {
    XFreeGC(getDisplay(),buffer_gc);
    XFreePixmap(getDisplay(),buffer_map);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::Buffer::setLineWidth(const INTEGER *lw) const { 
  XSetLineAttributes(getDisplay(),buffer_gc,(*lw+1)/2,LineSolid,
    CapRound,JoinRound);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::Buffer::putString(const REAL *px,const REAL *py,
const char *str,const REAL *angle) const {
  XFontStruct *the_font;
  if ((the_font = XLoadQueryFont(getDisplay(),default_font)) == 0) {
     cerr << "Cannot open font " << default_font << "\n";
     exit(1);
  };
  XRotDrawString(getDisplay(),the_font,static_cast<float>(*angle),
    buffer_map,buffer_gc,parent->ixLoc(*px),parent->iyLoc(*py),str);
  XFreeFont(getDisplay(),the_font);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::Buffer::writeXBorder(const char *str) const {
  XDrawString(getDisplay(),buffer_map,buffer_gc,parent->width/2,
    parent->myY(parent->yborder/2),str,strlen(str));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::Buffer::writeYBorder(const char *str) const {
  XDrawString(getDisplay(),buffer_map,buffer_gc,parent->xborder/2,
    parent->myY(parent->height/2),str,strlen(str));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::Buffer::colorRect(const REAL *ratio,const REAL *rw,
const REAL *rh,const REAL *rx,const REAL *ry) const {
  parent->colorRect(*ratio,*rw,*rh,*rx,*ry,buffer_map,buffer_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::Buffer::colorRect(const char *color_name,REAL rw,
REAL rh,REAL rx,REAL ry) {
  parent->colorRect(color_name,rw,rh,rx,ry,buffer_map,buffer_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::Buffer::colorPolygon(int color,int npts,XPoint p[]) const
{
  parent->colorPolygon(color,npts,p,buffer_map,buffer_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::Buffer::colorPolygon(const REAL *ratio,
const INTEGER *npts,const REAL *xfrac,const REAL *yfrac) const {
  parent->colorPolygon(*ratio,*npts,xfrac,yfrac,buffer_map,buffer_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::Buffer::colorPolygon(const INTEGER *npts,
const REAL *xfrac,const REAL *yfrac) const {
  parent->colorPolygon(*npts,xfrac,yfrac,buffer_map,buffer_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XTWindow::Buffer::printOn(ostream& os) const {
  os << "XTWindow::Buffer : parent = " << parent << endl;
  printGC(getDisplay(),buffer_gc,os);
  VirtualWindow12::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
XTWindow::QuitButton::QuitButton() : dialog(0) {
//TRACER_CALL(t,"QuitButton::QuitButton");
  int yes_called=FALSE;

  char str[]="Are you finished viewing the results?";
  Widget dialog=XmCreateQuestionDialog(XTWindow::grand_parent,
    const_cast<char*>("question"),0,0);

  XtVaSetValues(dialog,
    XmNdialogStyle,XmDIALOG_FULL_APPLICATION_MODAL,
    XtVaTypedArg,XmNmessageString,
    XmRString,str,strlen(str)+1,
    0 );
  XtAddCallback(dialog,XmNokCallback,YesCallback,
    reinterpret_cast<XtPointer>(&yes_called) );
  XtManageChild(dialog);
  XEvent event;
  long mask= ExposureMask | KeyPressMask | 
             ButtonPressMask | Button1MotionMask | 
             Button2MotionMask | Button3MotionMask | 
             ButtonReleaseMask | StructureNotifyMask | 
             SubstructureRedirectMask | VisibilityChangeMask;
  while (XtIsManaged(dialog) && !yes_called) {
    XtAppNextEvent(XTWindow::general_appcontext,&event);
#ifdef DEBUG
//  XTWindow::parseEvent(event,mask);
#endif
    XtDispatchEvent(&event);
  }
  XtDestroyWidget(dialog); 
  dialog=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
XTWindow::QuitButton::~QuitButton() { }
