//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#ifndef XT_WINDOW_H
#define XT_WINDOW_H "$Header: /home/faculty/johnt/cvs/deal_new/graphics/XTWindow.H,v 1.1 2009/08/20 17:31:47 johnt Exp $"
#include <algorithm>
#include <iostream>
#include <X11/Intrinsic.h>
#include "Const.H"
#include "Types.H"
#include "VirtualWindow.H"
#include "XColormap.H"

class XTWindow : public VirtualWindow12 {
  public:
    typedef XColormap COLOR_MAP_TYPE;
    class Buffer;
    friend class Buffer;
    class QuitButton;
    friend class QuitButton;
  private:
    char *name;
    BOOLEAN double_buffer;
    int mouseX,mouseY,mouseB,mouseCurrent;

    static XtAppContext general_appcontext;
    static Widget grand_parent;
    static unsigned int ref_ctr;

    Display *display;
    int screen;
    Window window;
    Widget parent,drawa;
    GC gc;

    int xborder,yborder;
    int width,height;
    REAL rwidth, rheight;
    int x0, y0;
    BOOLEAN grey_scale;
    int which_button_pressed;
    BOOLEAN allow_expose;

    Cardinal depth;
    XColormap*const xcolormap;
    Colormap cmap;

    Pixmap buffer_map;
    GC buffer_gc;

    XImage *image;

    int myY(int y) const { return height - 1 - y; }
    int ixLoc(REAL x) const { 
      return xborder+static_cast<int>(min(ONE,max(ZERO,x))
                    *static_cast<REAL>(width-xborder));
    }
    int iyLoc(REAL y) const { 
      return myY(yborder+static_cast<int>(min(ONE,max(ZERO,y))
                        *static_cast<REAL>(height-yborder))); 
    }

    Pixmap makePixmap() const;
    GC makeGC(Drawable) const;
    void setfgColour(Pixel);
    void setfgColour(Pixel,GC) const;
    void setbgColour(Pixel,Drawable,GC) const;
    void setfgColor(const char*,GC);
    void setbgColor(const char*,Drawable,GC);
    void setfgColor(REAL,GC) const;
    void setbgColor(REAL,Drawable,GC) const;
    void setbgColour(Pixel);
    void setfgColor(INTEGER,INTEGER,GC) const;
    void newPage(Drawable,GC);
    void colorRect(int,int,int,int,int,Drawable,GC) const;
    void colorRect(int,int,int,int,int) const;
    void colorRect(REAL,REAL,REAL,REAL,REAL,Drawable,GC) const ;
    void colorRect(const char*, REAL,REAL,REAL,REAL,Drawable,GC);
    void colorPolygon(int,int,XPoint[],Drawable,GC) const;
    void colorPolygon(int,XPoint[],Drawable,GC) const;
    void colorPolygon(int,int,XPoint[]) const;
    void colorPolygon(REAL,int,const REAL*,const REAL*,Drawable,GC)
      const;
    void colorPolygon(int,const REAL*,const REAL*,Drawable,GC) const;
    static Status parseEvent(const XEvent&,long);

    BOOLEAN buttonIsPressed(int&,int&);
    int getMouse(int&,int&);
  public:
    XTWindow(const char*,XColormap* =0,REAL=HALF,REAL=HALF,
      BOOLEAN =FALSE,BOOLEAN=TRUE);
    ~XTWindow();
    virtual XTWindow* getWindow() { return this; }
    virtual const XTWindow* getWindow() const { return this; }
    const char* getName() const { return name; }
    const char* getDisplayName() const { 
      char *dn;
      Status st=XFetchName(display,window,&dn);
      CHECK_TEST(st!=0);
      return dn; 
    }
    BOOLEAN doubleBuffer() const { return double_buffer; }
    int getWidth() const { return width; }
    int getHeight() const { return height; }

    BOOLEAN buttonIsPressed(REAL&,REAL&);
    void newPage();
    void movePen(const REAL *px,const REAL *py) const {
      XTWindow *copy=const_cast<XTWindow*>(this);
      copy->x0 = ixLoc(*px);
      copy->y0 = iyLoc(*py);
    }
    void drawLine(const REAL*,const REAL*,const BOOLEAN* =0) const;
    BOOLEAN monochrome() const { return grey_scale; }
    Display* getDisplay() const { return display; }
    int getMouse(REAL&,REAL&);
    void colorRect(const REAL*,const REAL*,const REAL*,const REAL*,
      const REAL*) const;
    void colorRect(const char *name, REAL rw,REAL rh,REAL rx,REAL ry);
    void colorPolygon(const REAL*,const INTEGER*,const REAL*,
      const REAL*) const;
    void colorPolygon(const INTEGER*,const REAL*,const REAL*) const;
    void setLineWidth(const INTEGER *plw) const {
      if (doubleBuffer()) {
        XSetLineAttributes(display,buffer_gc,(*plw+1)/2,LineSolid,
          CapRound, JoinRound );
      } else {
        XSetLineAttributes(display,gc,(*plw+1)/2,LineSolid,CapRound,
          JoinRound );
      }
    }
    void expose();
    void input(XtPointer call_data) {;}
    void resize() {;}
    static void setGrandParent(Widget gp) {
      if (grand_parent==0) { 
        grand_parent=gp; 
        general_appcontext=XtWidgetToApplicationContext(grand_parent);
        ref_ctr=1; 
      }
    }
    static void exposeCallback(Widget,XtPointer client_data,XtPointer) {
      XTWindow *win=reinterpret_cast<XTWindow*>(client_data);
      if (win) win->expose();
    }
    static void inputCallback(Widget,XtPointer client_data,
    XtPointer call_data) {
      XTWindow *win=reinterpret_cast<XTWindow*>(client_data);
      if (win) win->input(call_data);
    }
    static void resizeCallback(Widget,XtPointer client_data,XtPointer) {
      XTWindow *win=reinterpret_cast<XTWindow*>(client_data);
      if (win) win->resize();
    }
    void setFont(const char*) const;
    Pixel allocColor(const char*);
    void putString(const REAL *px,const REAL *py,const char *str,
      const REAL *angle) const;
    void writeXBorder(const char *str) const;
    void writeYBorder(const char *str) const;
    void setfgColor(const char*) const;
    void setbgColor(const char*) const;
    int pickColor(REAL) const ;
    void setfgColor(const REAL*) const;
    void setbgColor(const REAL*) const;
    void setfgColor(const INTEGER*,const INTEGER*) const;
    void refresh(int,int,int,int) const ;
    void refreshFrom(const Buffer&) const;
    void eventLoop();
    void flush(BOOLEAN=TRUE);
    void flush(const BOOLEAN *pval) { return flush(*pval); }
    void sync() { XSync( display , FALSE ); }
    XColormap* getXColormap(void) {return xcolormap;}

    void writePaletteName(FILE*);
    void createRaster(FILE*);
    void writeRaster(FILE*);
    void closeRaster();
    void openRaster(FILE*);
    void readRaster(FILE*);
    void writeXPM(const char*);
    void printOn(ostream& os = cerr) const;
};

class XTWindow::Buffer : public VirtualWindow12 {
  friend class XTWindow;
  private:
    XTWindow*const parent;
    Pixmap buffer_map;
    GC buffer_gc;
    int x0, y0;

    Pixmap getPixmap() const { return buffer_map; }
    void setFont(const char*) const {;}
    void flush(BOOLEAN) {;}
    BOOLEAN buttonIsPressed(REAL &x,REAL &y) { x=y=ZERO; return FALSE; }
    int getMouse(REAL &x,REAL &y) { x=y=ZERO; return -1; }
    void writePaletteName(FILE*) {;}
    void createRaster(FILE*) {;}
    void writeRaster(FILE*) {;}
    void closeRaster() {;}
    void openRaster(FILE*) {;}
    void readRaster(FILE*) {;}
  protected:
    Buffer(const Buffer&) : parent(0) { 
      OBSOLETE("copy constructor disallowed"); 
    }
  public:
    Buffer(XTWindow*);
    virtual ~Buffer();
    XTWindow::Buffer* getWindow() { return this; }
    const XTWindow::Buffer* getWindow() const { return this; }
    const char* getName() const { return parent->getName(); }
    const char* getDisplayName() const { 
      return parent->getDisplayName(); 
    }
    int getWidth() const { return parent->getWidth(); }
    int getHeight() const { return parent->getHeight(); }
  
    Display* getDisplay() const { return parent->getDisplay(); }
    BOOLEAN monochrome() const { return parent->monochrome(); }
    void setLineWidth(const INTEGER*) const;
    void setfgColour(Pixel pixel) { 
      parent->setfgColour(pixel,buffer_gc); 
    }
    void setbgColour(Pixel pixel) {
      parent->setbgColour(pixel,buffer_map,buffer_gc);
    }
    void setfgColor(const char *colorname) const {
      parent->setfgColor(colorname,buffer_gc);
    }
    void setbgColor(const char *colorname) const {
      parent->setbgColor(colorname,buffer_map,buffer_gc);
    }
    void setfgColor(const REAL *pfrac) const {
      parent->setfgColor(*pfrac,buffer_gc); 
    }
    void setbgColor(const REAL *pfrac) const {
      parent->setbgColor(*pfrac,buffer_map,buffer_gc); 
    }
    void setfgColor(const INTEGER *ic,const INTEGER* im) const {
      parent->setfgColor(*ic,*im,buffer_gc);
    }
    void putString(const REAL*,const REAL*,const char*,const REAL*) 
      const;
    void writeXBorder(const char *str) const;
    void writeYBorder(const char *str) const;
    void movePen(const REAL *px,const REAL *py) const {
      XTWindow::Buffer *copy=const_cast<XTWindow::Buffer*>(this);
      copy->x0=parent->ixLoc(*px); copy->y0=parent->iyLoc(*py);
    }
    void drawLine(const REAL *px,const REAL *py,const BOOLEAN* =0) const
    {
      XTWindow::Buffer *copy=const_cast<XTWindow::Buffer*>(this);
      int x1=parent->ixLoc(*px); int y1=parent->iyLoc(*py);
      XDrawLine(getDisplay(),buffer_map,buffer_gc,x0,y0,x1,y1);
      copy->x0 = x1; copy->y0 = y1;
    }

    void newPage() { parent->newPage(buffer_map,buffer_gc); }
    void colorPolygon(int,int,XPoint[]) const;
    void colorPolygon(const REAL*,const INTEGER*,const REAL*,
      const REAL*) const;
    void colorPolygon(const INTEGER*,const REAL*,const REAL*) const;
    void colorRect(const REAL*,const REAL*,const REAL*,const REAL*,
      const REAL*) const;
    void colorRect(const char*,REAL,REAL,REAL,REAL);
    int pickColor(REAL frac) const { return parent->pickColor(frac); }
    void printOn(ostream& os = cerr) const;
};

class XTWindow::QuitButton {
  private:
    Widget dialog;
  public:
    QuitButton();
    ~QuitButton();
    static void YesCallback(Widget,XtPointer client_data,XtPointer) {
      int *yes_called=reinterpret_cast<int*>(client_data);
      *yes_called=TRUE;
    }
};

#endif
