//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
// "$Header: /home/faculty/johnt/cvs/deal_new/graphics/XWindow.C,v 1.1 2009/08/20 17:31:47 johnt Exp $"
#include "XWindow.H"
#include <string.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <Xm/VendorS.h>
#include <X11/xpm.h>
#include "Const.H"
#include "MemoryDebugger.H"
#include "Palette.H"

#define NTSC_WIDTH 600
#define NTSC_HEIGHT 420


static char default_font[] = "8x13";
#define NUMBER_ERRORS 18
// these appear in /usr/local/X11/include/X.h
static const char *ErrorCodes[NUMBER_ERRORS] = {
  "Success",           // 0
  "BadRequest",        // 1
  "BadValue",          // 2
  "BadWindow",         // 3
  "BadPixmap",         // 4
  "BadAtom",           // 5
  "BadCursor",         // 6
  "BadFont",           // 7
  "BadMatch",          // 8
  "BadDrawable",       // 9
  "BadAccess",         // 10
  "BadAlloc",          // 11
  "BadColor",          // 12
  "BadGC",             // 13
  "BadIDChoice",       // 14
  "BadName",           // 15
  "BadLength",         // 16
  "BadImplementation"  // 17
};
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void printHosts(Display *display,ostream& os) {
  int access_enabled;
  int num_hosts;
  XHostAddress *hosts;
  hosts=XListHosts(display,&num_hosts,&access_enabled);
  if (hosts) {
    if (access_enabled) os << "\naccess is enabled to:" << endl;
    else os << "access disabled to:" << endl;
    for (int i=0;i<num_hosts;i++) {
//    if (hosts[i].family==FamilyInternet) {
//      struct hostent *thehostent=
//        gethostbyaddr(hosts[i].address,hosts[i].length,AF_INET);
//      os << "\t" << thehostent->h_name << " "
//         << inet_ntoa(*reinterpret_cast<struct in_addr*>(
//            hosts[i].address)) << endl;
//    } else 
      {
        switch (hosts[i].family) {
          case FamilyDECnet:
            os << "Family is DECnet" << endl;
            break;
          case FamilyChaos:
            os << "Family is Chaos" << endl;
            break;
          default:
            os << "Family is unknown" << endl;
            break;
        }
      }
    }
    XFree(hosts);
  } else os << "cannot get list of hosts" << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int MyXErrorHandler(Display *display,XErrorEvent *error_event) {
  char error_string[LENGTH_COMMENT+1];
  XGetErrorText(display,error_event->error_code,error_string,
                LENGTH_COMMENT);
  cerr << "X Error on display " << DisplayString(display)
       << ".\nResource " << error_event->resourceid << endl;
  if ( (error_event->error_code>0) &&
  (error_event->error_code<NUMBER_ERRORS) ) {
    cerr << error_string << " (" 
         << ErrorCodes[error_event->error_code] << ")"<<endl;
  } else cerr << error_string << endl;
  cerr << "Op code " 
       << static_cast<unsigned int>(error_event->request_code);
  cerr << "." << static_cast<unsigned int>(error_event->minor_code);
  cerr << ", Error code " 
       << static_cast<unsigned int>(error_event->error_code);
  cerr <<", type "<<error_event->type<<", serial "
       << error_event->serial << endl;
  if (error_event) abort();
  return TRUE;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void printGC(Display *display,GC gc,ostream& os) {
  os << "\ngc = " << gc << endl;
  XGCValues gcvalues;
//the GCClipMask and GCDashList bits cannot be requested
  unsigned long valuemask=GCFunction | GCPlaneMask | GCForeground
    | GCBackground | GCLineWidth | GCLineStyle | GCCapStyle
    | GCJoinStyle | GCFillStyle | GCFillRule | GCArcMode | GCTile
    |  GCStipple | GCTileStipXOrigin | GCTileStipYOrigin | GCFont
    | GCSubwindowMode | GCGraphicsExposures | GCClipXOrigin
    | GCClipYOrigin | GCDashOffset;
  int status=XGetGCValues(display,gc,valuemask,&gcvalues);
  if (status) {
    os << "graphics context values:\tfunction = ";
    switch (gcvalues.function) {
      case GXclear:
        os << "GXclear";
        break;
      case GXand:
        os << "GXand";
        break;
      case GXandReverse:
        os << "GXandReverse";
        break;
      case GXcopy:
        os << "GXcopy";
        break;
      case GXandInverted:
        os << "GXandInverted";
        break;
      case GXnoop:
        os << "GXnoop";
        break;
      case GXor:
        os << "GXor";
        break;
      case GXnor:
        os << "GXnor";
        break;
      case GXequiv:
        os << "GXequiv";
        break;
      case GXinvert:
        os << "GXinvert";
        break;
      case GXorReverse:
        os << "GXorReverse";
        break;
      case GXcopyInverted:
        os << "GXcopyInverted";
        break;
      case GXorInverted:
        os << "GXorInverted";
        break;
      case GXnand:
        os << "GXnand";
        break;
      case GXset:
        os << "GXset";
        break;
      default:
        os << "unknown";
        break;
    }
    os << "\n\tplane_mask = " << gcvalues.plane_mask
       << "\n\tforeground = " << gcvalues.foreground
       << "\n\tbackground = " << gcvalues.background
       << "\n\tline_width = " << gcvalues.line_width
       << "\n\tline_style = ";
    switch (gcvalues.line_style) {
      case LineSolid:
        os << "LineSolid";
        break;
      case LineDoubleDash:
        os << "LineDoubleDash";
        break;
      case LineOnOffDash:
        os << "LineOnOffDash";
        break;
      default:
        os << "unknown";
        break;
    }
    os << "\n\tcap_style = ";
    switch (gcvalues.cap_style) {
      case CapNotLast:
        os << "CapNotLast";
        break;
      case CapButt:
        os << "CapButt";
        break;
      case CapRound:
        os << "CapRound";
        break;
      case CapProjecting:
        os << "CapProjecting";
        break;
      default:
        os << "unknown";
        break;
    }
    os << "\n\tjoin_style = ";
    switch (gcvalues.join_style) {
      case JoinMiter:
        os << "JoinMiter";
        break;
      case JoinRound:
        os << "JoinRound";
        break;
      case JoinBevel:
        os << "JoinBevel";
        break;
      default:
        os << "unknown";
        break;
    }
    os << "\n\tfill_style = ";
    switch (gcvalues.fill_style) {
      case FillSolid:
        os << "FillSolid";
        break;
      case FillTiled:
        os << "FillTiled";
        break;
      case FillOpaqueStippled:
        os << "FillOpaqueStippled";
        break;
      case FillStippled:
        os << "FillStippled";
        break;
      default:
        os << "unknown";
        break;
    }
    os << "\n\tfill_rule = ";
    switch (gcvalues.fill_rule) {
      case EvenOddRule:
        os << "EvenOddRule";
        break;
      case WindingRule:
        os << "WindingRule";
        break;
      default:
        os << "unknown";
        break;
    }
    os << "\n\tarc_mode = ";
    switch (gcvalues.arc_mode) {
      case ArcPieSlice:
        os << "ArcPieSlice";
        break;
      case ArcChord:
        os << "ArcChord";
        break;
      default:
        os << "unknown";
        break;
    }
    os << "\n\ttile = " << gcvalues.tile
       << "\n\tstipple = " << gcvalues.stipple
       << "\n\tts_x_origin = " << gcvalues.ts_x_origin
       << "\n\tfont = " << gcvalues.font
       << "\n\tsubwindow_mode = ";
    switch (gcvalues.subwindow_mode) {
      case ClipByChildren:
        os << "ClipByChildren";
        break;
      case IncludeInferiors:
        os << "IncludeInferiors";
        break;
      default:
        os << "unknown";
        break;
    }
    os << "\n\tgraphics_exposures = " << gcvalues.graphics_exposures
       << "\n\tclip_x_origin = " << gcvalues.clip_x_origin
       << "\n\tclip_y_origin = " << gcvalues.clip_y_origin
       << "\n\tts_x_origin = " << gcvalues.ts_x_origin
       << "\n\tdash_offset = " << gcvalues.dash_offset << endl;
  } else os << "cannot get gcvalues" << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void printPixmapFormatValues(ostream& os,Display *display) {
  XPixmapFormatValues *p;
  int count;
  p=XListPixmapFormats(display,&count);
  if (p) {
    os << "XPixmapFormatValues = " << p
       << "\n\tdepth = " << p->depth
       << "\n\tbits_per_pixel = " << p->bits_per_pixel
       << "\n\tscanline_pad = " << p->scanline_pad << endl;
    XFree(p);
  } else os << "cannot get XPixmapFormatValues" << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void printDisplayInfo(ostream& os,const char *display_name,
Display *display) {
  os << "display name = " << display_name << "\n"
     << "display = " << display << endl;
  os << "X server from " << ServerVendor(display)
     << " at release " << VendorRelease(display)
     << ", protocol " << ProtocolVersion(display)
     << "." << ProtocolRevision(display) << endl;

  int number_extensions;
  char **extension_list;
  extension_list=XListExtensions(display,&number_extensions);
  if (extension_list) {
    cout << "server extensions include" << endl;
    for (int i=0;i<number_extensions;i++) 
      cout << "\t" << extension_list[i] << endl;
  }
  XFreeExtensionList(extension_list);

  os << "connection number = " << ConnectionNumber(display)
     << "\ndefault screen of display = " 
     << DefaultScreenOfDisplay(display)
     << "\ndefault screen = " << DefaultScreen(display)
     << "\ndisplay string = " << DisplayString(display)
     << "\nlast known request processed = "
     << LastKnownRequestProcessed(display)
     << "\nnext request = " << NextRequest(display)
     << "\nqueue length = " << QLength(display) << endl;

  printPixmapFormatValues(os,display);

  os << "image byte order = ";
  switch (ImageByteOrder(display)) {
    case LSBFirst:
      os << " LSBFirst" << endl;
      break;
    case MSBFirst:
      os << " MSBFirst" << endl;
      break;
    default:
      os << "unknown" << endl;
      break;
  }
  os << "bitmap unit = " << BitmapUnit(display)
     << "\nbitmap bit order = ";
  switch (BitmapBitOrder(display)) {
    case LSBFirst:
      os << " LSBFirst" << endl;
      break;
    case MSBFirst:
      os << " MSBFirst" << endl;
      break;
    default:
      os << "unknown" << endl;
      break;
  }
  os << "bitmap pad = " << BitmapPad(display) << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void printVisualInfo(ostream& os,Display *display,int scrn) {
  XVisualInfo visual_template;
  visual_template.screen=scrn;
  int number_visuals;
  XVisualInfo *visual_list=XGetVisualInfo(display,VisualScreenMask,
                             &visual_template,&number_visuals);
  if (visual_list) {
    os << "\tscrn " << scrn << " has " << number_visuals << " Visuals" 
       << "\n";
    for (int i=0;i<number_visuals;i++) {
       switch (visual_list[i].c_class) {
        case PseudoColor:
          os << "PseudoColor (read/write) ";
          break;
        case DirectColor:
          os << "DirectColor (read/write) ";
          break;
        case TrueColor:
          os << "TrueColor (read only)";
          break;
        case StaticColor:
          os << "StaticColor (read only)";
          break;
        case GrayScale:
          os << "GrayScale (read/write) ";
          break;
        case StaticGray:
          os << "StaticGray (read only)";
          break;
        default:
          os << "and something else ";
          break;
      }
      os << "visual with id " << visual_list[i].visualid 
         << "\n\t\tplanes depth = " << visual_list[i].depth 
         << "\n\t\tcolormap entries = " << visual_list[i].colormap_size
         << "\n\t\tbits per pixel = " << visual_list[i].bits_per_rgb 
         << endl;
    }
  } else os << "could not find a visual list for this screen" << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void printScreenInfo(ostream& os,Display *display,int screen,int scrn) {
  Screen *scrnptr=ScreenOfDisplay(display,scrn);
  os << "scrn " << scrn << " is " << DisplayWidth(display,scrn)
     << " by " << DisplayHeight(display,scrn) << " pixels\n"
     << "\tbit planes = " << DefaultDepth(display,scrn) << "\n"
     << "\t" << DisplayCells(display,scrn) << " color cells and "
     << MaxCmapsOfScreen(scrnptr) << " colormaps\n"
     << "\tblack is defined as " << BlackPixel(display,scrn) 
     << "; white as " << WhitePixel(display,scrn) 
     << "\n\tdisplay planes = " << DisplayPlanes(display,screen) <<endl;
  os << "\tdefault cells = " << CellsOfScreen(scrnptr)
     << "\n\tdefault depth = " << DefaultDepthOfScreen(scrnptr)
     << "\n\tevent mask = " << EventMaskOfScreen(scrnptr)
     << "\n\tmax color maps = " << MaxCmapsOfScreen(scrnptr)
     << "\n\tmin color maps = " << MinCmapsOfScreen(scrnptr) << endl;

  int number_depths;
  int *depth_list=XListDepths(display,scrn,&number_depths);
  if (depth_list) {
    os << "\tsupports ";
    for (int i=0;i<number_depths;i++) os << depth_list[i] << " ";
    os << "depths" << endl;
    XFree(depth_list);
  }

  if (DoesSaveUnders(scrnptr)) {
    os << "\twill provide save-unders" << endl;
  } else {
    os << "does not provide save-unders" << endl;
  }

  switch (DoesBackingStore(scrnptr)) {
    case WhenMapped:
      os << "\tbacking store is provided for mapped windows" << endl;
      break;
    case Always:
      os << "\tbacking store is always provided" << endl;
      break;
    case NotUseful:
    default:
      os << "\tbacking store is not provided" << endl;
      break;
  }

  printVisualInfo(os,display,scrn);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void printXWindowAttributes(ostream& os,Display *display,Window window){
  XWindowAttributes a;
  Status status=XGetWindowAttributes(display,window,&a);
  if (status) {
    os << "\nwindow attributes:\n\tx, y = " << a.x << ", " << a.y
       << "\n\twidth, height = " << a.width << " " << a.height
       << "\n\tborder_width = " << a.border_width
       << "\n\tdepth = " << a.depth
       << "\n\tvisual = " << a.visual
       << "\n\tclass = ";
    switch (a.c_class) {
      case InputOutput:
        os << "InputOutput";
        break;
      case InputOnly:
        os << "InputOnly";
        break;
      default:
        os << "unknown";
        break;
    }
    os << "\n\tbit_gravity = ";
    switch (a.bit_gravity) {
      case ForgetGravity:
        os << "ForgetGravity";
        break;
      case NorthWestGravity:
        os << "NorthWestGravity";
        break;
      case NorthGravity:
        os << "NorthGravity";
        break;
      case NorthEastGravity:
        os << "NorthEastGravity";
        break;
      case WestGravity:
        os << "WestGravity";
        break;
      case CenterGravity:
        os << "CenterGravity";
        break;
      case EastGravity:
        os << "EastGravity";
        break;
      case SouthWestGravity:
        os << "SouthWestGravity";
        break;
      case SouthGravity:
        os << "SouthGravity";
        break;
      case SouthEastGravity:
        os << "SouthEastGravity";
        break;
      case StaticGravity:
        os << "StaticGravity";
        break;
      default:
        os << "unknown";
        break;
    }
    os << "\n\twin_gravity = ";
    switch (a.win_gravity) {
      case UnmapGravity:
        os << "UnmapGravity";
        break;
      case NorthWestGravity:
        os << "NorthWestGravity";
        break;
      case NorthGravity:
        os << "NorthGravity";
        break;
      case NorthEastGravity:
        os << "NorthEastGravity";
        break;
      case WestGravity:
        os << "WestGravity";
        break;
      case CenterGravity:
        os << "CenterGravity";
        break;
      case EastGravity:
        os << "EastGravity";
        break;
      case SouthWestGravity:
        os << "SouthWestGravity";
        break;
      case SouthGravity:
        os << "SouthGravity";
        break;
      case SouthEastGravity:
        os << "SouthEastGravity";
        break;
      case StaticGravity:
        os << "StaticGravity";
        break;
      default:
        os << "unknown";
        break;
    }
    os << "\n\tbacking_store = ";
    switch (a.backing_store) {
      case NotUseful:
        os << "NotUseful";
        break;
      case WhenMapped:
        os << "WhenMapped";
        break;
      case Always:
        os << "Always";
        break;
      default:
        os << "unknown";
        break;
    }
    os << "\n\tbacking_planes = " << a.backing_planes
       << "\n\tbacking_pixel = " << a.backing_pixel
       << "\n\tsave_under = " << a.save_under
       << "\n\tmap_installed = " << a.map_installed
       << "\n\tmap_state = ";
    switch (a.map_state) {
      case IsUnmapped:
        os << "IsUnmapped";
        break;
      case IsUnviewable:
        os << "IsUnviewable";
        break;
      case IsViewable:
        os << "IsViewable";
        break;
      default:
        os << "unknown";
        break;
    }
    os << "\n\tall_event_masks = " << a.all_event_masks
       << "\n\tyour_event_mask = " << a.your_event_mask
       << "\n\tdo_not_propagate_mask = " << a.do_not_propagate_mask
       << "\n\toverride_redirect = " << a.override_redirect
       << "\n\tscreen = " << a.screen << endl;
  } else os << "cannot get XSetWindowAttributes" << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void printWidgetInfo(ostream& os,Widget widget) {
  os << "widget = " << widget << "\n";
  os << "\tdisplay = " << XtDisplay(widget) << endl;
  os << "\tscreen = " << XtScreen(widget) << endl;
  os << "\tparent = " << XtParent(widget) << endl;
  os << "\twindow = " << XtWindow(widget) << endl;
  os << "\tname = " << XtName(widget) << endl;
  if (XtIsWidget(widget)) cout << "\tis widget" << endl;
  if (XtIsComposite(widget)) cout << "\tis composite" << endl;
  if (XtIsConstraint(widget)) cout << "\tis constraint" << endl;
  if (XtIsShell(widget)) cout << "\tis shell" << endl;
  if (XtIsOverrideShell(widget)) cout << "\tis override shell" << endl;
  if (XtIsWMShell(widget)) cout << "\tis window manager shell" << endl;
  if (XtIsVendorShell(widget)) cout << "\tis vendor shell" << endl;
  if (XtIsTransientShell(widget)) cout << "\tis transient shell" <<endl;
  if (XtIsTopLevelShell(widget)) cout << "\tis top level shell" << endl;
  if (XtIsApplicationShell(widget)) {
    cout << "\tis application shell" <<endl;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void printWindowInfo(ostream& os,Display *display,Window window) {
  os << "window = " << window << "\n";

  char *wn;
  XFetchName(display,window,&wn);
  os << "window_name = " << wn;
  XFree(wn);

  int x,y;
  unsigned int wi,h;
  unsigned int b,d;
  Window rootwindow;
  int status=
    XGetGeometry(display,window,&rootwindow,&x,&y,&wi,&h,&b,&d);
  if (status) {
    os << " at local " << x << " " << y << "; " << wi << " by " << h
       << endl;
  } else os << " will not return geometry information" << endl;

  printXWindowAttributes(os,display,window);

  XClassHint *classhint;
  classhint=XAllocClassHint();
  if (classhint) {
    status=XGetClassHint(display,window,classhint);
    if (status) {
      os << "CLASS " << classhint->res_name << "/" 
         << classhint->res_class << endl;
    } else os << "will not return classhint information" << endl;
    XFree(classhint);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void printWMProperties(ostream& os,Display *display,Window window) {
  Atom *properties;
  int num_properties;
  properties=XListProperties(display,window,&num_properties);
  if (properties) {
    os << "\nwindow manager properties are ";
    for (int i=0;i<num_properties;i++) {
      char *name=XGetAtomName(display,properties[i]);
      if (name) os << "\n\t" << name;
      else os << "\n\tunknown";
      XFree(name);
    }
    os << endl;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//XEvent types are defined in X.h
void printXEventType(XEvent *event) {
  switch (event->type) {
    case KeyPress:
      cout << "KeyPress";
      break;
    case KeyRelease:
      cout << "KeyRelease";
      break;
    case ButtonPress:
      cout << "ButtonPress";
      break;
    case ButtonRelease:
      cout << "ButtonRelease";
      break;
    case MotionNotify:
      cout << "MotionNotify";
      break;
    case EnterNotify:
      cout << "EnterNotify";
      break;
    case LeaveNotify:
      cout << "LeaveNotify";
      break;
    case FocusIn:
      cout << "FocusIn";
      break;
    case FocusOut:
      cout << "FocusOut";
      break;
    case KeymapNotify:
      cout << "KeymapNotify";
      break;
    case Expose:
      cout << "Expose";
      break;
    case GraphicsExpose:
      cout << "GraphicsExpose";
      break;
    case NoExpose:
      cout << "NoExpose";
      break;
    case VisibilityNotify:
      cout << "VisibilityNotify";
      break;
    case CreateNotify:
      cout << "CreateNotify";
      break;
    case DestroyNotify:
      cout << "DestroyNotify";
      break;
    case UnmapNotify:
      cout << "UnmapNotify";
      break;
    case MapNotify:
      cout << "MapNotify";
      break;
    case MapRequest:
      cout << "MapRequest";
      break;
    case ReparentNotify:
      cout << "ReparentNotify";
      break;
    case ConfigureNotify:
      cout << "ConfigureNotify";
      break;
    case ConfigureRequest:
      cout << "ConfigureRequest";
      break;
    case GravityNotify:
      cout << "GravityNotify";
      break;
    case ResizeRequest:
      cout << "ResizeRequest";
      break;
    case CirculateNotify:
      cout << "CirculateNotify";
      break;
    case CirculateRequest:
      cout << "CirculateRequest";
      break;
    case PropertyNotify:
      cout << "PropertyNotify";
      break;
    case SelectionClear:
      cout << "SelectionClear";
      break;
    case SelectionRequest:
      cout << "SelectionRequest";
      break;
    case SelectionNotify:
      cout << "SelectionNotify";
      break;
    case ColormapNotify:
      cout << "ColormapNotify";
      break;
    case ClientMessage:
      cout << "ClientMessage";
      break;
    case MappingNotify:
      cout << "MappingNotify";
      break;
    default:
      cout << "unknown";
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
XWindow::XWindow(const char *str,XColormap *cm,REAL rw,REAL rh,
BOOLEAN mono,BOOLEAN db,Visual *visual) : name(0),double_buffer(db),
display(0),screen(0),window(0),gc(0),width(0),height(0),rwidth(ZERO),
rheight(ZERO),x0(0),y0(0),grey_scale(mono),nplanes(0),xcolormap(cm),
cmap(None),buffer_map(0),buffer_gc(0),image(0) {
  if (str) {
    name=OPERATOR_NEW_BRACKET(char,strlen(str)+1);
    strcpy(name,str);
  } else {
    name=OPERATOR_NEW_BRACKET(char,1);
    name[0]='\0';
  }

  ASSERT(cm!=0);
  if (cm) {
    display=cm->getDisplay();
    screen=cm->getScreen();
  } else {
    char *dn=0;
    if (!(display=XOpenDisplay(dn))) {
      cout << "attempted to open display on " << XDisplayName(dn) 
           << endl;
      cout << "cannot connect to X server " << endl;
      CHECK_TEST(display!=0);
    }
    screen=DefaultScreen(display);
  }
  CHECK_TEST(display!=0);

  int x,y;
  REAL rx=HALF,ry=HALF;
  {
    nplanes=DisplayPlanes(display,screen);
    if (!xcolormap) grey_scale=TRUE;
    if (nplanes==1) grey_scale=TRUE;

    Window parent=RootWindow(display,screen);

    int dw=DisplayWidth(display,screen);
    int dh=DisplayHeight(display,screen);
    int dm=min(dw,dh);
    REAL rdm=static_cast<REAL>(dm);

    if (rw>ONE) rw=ONE;
    if (rh>ONE) rh=ONE;
    if (rw<=ZERO || rh<=ZERO) {
      width=NTSC_WIDTH;
      height=NTSC_HEIGHT;
    } else {
      width=static_cast<int>(rw*rdm-HALF)+1;
      height=static_cast<int>(rh*rdm-HALF)+1;
    }
    rwidth=REAL(width);
    rheight=REAL(height);

    if (rx<ZERO || rx>=ONE) rx=ZERO;
    if (ry<ZERO || ry>=ONE) ry=ZERO;
    x=dw/2;
    y=dh/2;

    XSetWindowAttributes attributes;
    Screen *scrnptr=ScreenOfDisplay(display,screen);
    attributes.background_pixel=WhitePixel(display,screen);
    attributes.border_pixel=BlackPixel(display,screen);
    attributes.backing_store=Always;
    attributes.save_under=XDoesSaveUnders(scrnptr);
    unsigned long attributes_mask=
      CWBackPixel | CWBorderPixel | CWBackingStore | CWSaveUnder;

    unsigned int border_width=4;
    window=XCreateWindow(display,parent,x,y,width,height,border_width,
      nplanes,InputOutput,visual,attributes_mask,&attributes);
    CHECK_TEST(window);

    XWindowAttributes a;
    Status st=XGetWindowAttributes(display,window,&a);
    CHECK_TEST(st!=0);
    nplanes=a.depth;
    if (xcolormap) {
      cmap=xcolormap->getColormap();
      if (cmap!=a.colormap) {
        XSetWindowColormap(display,window,cmap);
      }
    } else {
      cmap=a.colormap;
    }

    if (double_buffer) buffer_map=makePixmap();
  }
  
  Atom wm_delete_window=XInternAtom(display, "WM_DELETE_WINDOW", False);
  Pixmap icon=None;
  XWMHints *wmhints;
  wmhints=XAllocWMHints();
  CHECK_TEST(wmhints!=0);

  wmhints->input=True;
  wmhints->initial_state=NormalState;
  wmhints->flags=StateHint | InputHint;

  if (icon) {
    wmhints->icon_pixmap=icon;
    wmhints->icon_mask=icon;
    wmhints->flags=wmhints->flags | IconPixmapHint | IconMaskHint ;
  }

  XSetWMHints(display,window,wmhints);
  XFree(wmhints);
  XSetWMProtocols(display, window, &wm_delete_window, 1);

  XSizeHints *sizehints;
  sizehints=XAllocSizeHints();
  CHECK_TEST(sizehints!=0);

  sizehints->min_width=1;  // needed in R4 or the base size
  sizehints->min_height=1;//   might become the min size
  sizehints->flags=PMinSize;
  sizehints->base_width=width;
  sizehints->base_height=height;
  sizehints->flags|=PBaseSize;
  XSetWMNormalHints(display,window,sizehints);
  XFree(sizehints);

  if (str) {
    XStoreName(display,window,str);
    XSetIconName(display,window,str);
  } else {
    XStoreName(display,window,"xterm");
    XSetIconName(display,window,"xterm");
  }
  
  XSelectInput(display, window, ExposureMask | KeyPressMask | 
               ButtonPressMask | Button1MotionMask | Button2MotionMask |
               Button3MotionMask | ButtonReleaseMask | 
               StructureNotifyMask | SubstructureRedirectMask | 
               VisibilityChangeMask );

  XMapRaised(display,window);
  XMapSubwindows(display,window);
  XFlush(display);

  gc=makeGC(window);
  CHECK_TEST(gc!=0);
  if (double_buffer) {
    buffer_gc=makeGC(buffer_map);
    CHECK_TEST(buffer_gc!=0);
  }

  x0=0;
  y0=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
XWindow::~XWindow() {
  if (display) {
    XDestroySubwindows(display,window);
    XDestroyWindow(display,window);
    XFlush(display);
    if (double_buffer) {
      XFreeGC(display,buffer_gc);
      XFreePixmap(display,buffer_map);
    }
    XFreeGC( display, gc );
    display=0;
  }
  delete [] name; name=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::expose() {
  if (double_buffer) refresh(0,0,width,height);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Pixmap XWindow::makePixmap() const {
  Pixmap pm=XCreatePixmap(display,window,width,height,nplanes);
  CHECK_TEST(pm);
  return pm;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
GC XWindow::makeGC(Drawable drawable) const {
  CHECK_TEST(display!=0);

  XGCValues gcvalues;
  gcvalues.foreground=BlackPixel(display,screen);
  gcvalues.background=WhitePixel(display,screen);
  gcvalues.line_width=1;               // also
  gcvalues.line_style=LineSolid;       //   see
  gcvalues.cap_style=CapRound;         //     XSetLineAttributes
  gcvalues.join_style=JoinRound;       //       .
  XFontStruct *the_font = XLoadQueryFont(display,default_font);
  if (!the_font) {
    the_font = XLoadQueryFont(display,"fixed");
    CHECK_TEST(the_font!=0);
  }
  gcvalues.font = the_font->fid;                  // also see XSetFont
  unsigned long mask= GCForeground | GCBackground | GCLineWidth 
                    | GCLineStyle | GCCapStyle | GCJoinStyle | GCFont;
  GC the_gc = XCreateGC(display,drawable,mask,&gcvalues);
  XFreeFont(display,the_font);
  return the_gc;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::setfgColour(Pixel pixel,GC the_gc) const {
  XSetForeground(display,the_gc,pixel);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::setfgColour(Pixel pixel) {
  if (double_buffer) setfgColour(pixel,buffer_gc);
  else setfgColour(pixel,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::setfgColor(const char *colorname,GC the_gc) {
  if (monochrome()) return;
  setfgColour(allocColor(colorname),the_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::setbgColor(const char *colorname,Drawable drawable,
GC the_gc) {
  if (monochrome()) return;
  setbgColour(allocColor(colorname),drawable,the_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::setfgColor(REAL frac,GC the_gc) const {
  if (monochrome()) return;
  frac=max(ZERO,min(ONE,frac));
  int color=pickColor(frac);
  setfgColour(xcolormap->getColor(color),the_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::setfgColor(const REAL *frac) {
  if (double_buffer) setfgColor(*frac,buffer_gc);
  else setfgColor(*frac,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::setbgColor(REAL frac,Drawable drawable,GC the_gc) const {
  if (monochrome()) return;
  frac=max(ZERO,min(ONE,frac));
  int color=pickColor(frac);
  color=max(0,min(color,xcolormap->getNumberColors()));
  setbgColour(xcolormap->getColor(color),drawable,the_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::setbgColor(const REAL *frac) {
  if (double_buffer) setbgColor(*frac,buffer_map,buffer_gc);
  else setbgColor(*frac,window,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM == 1)
void XWindow::setfgColor(INTEGER ic,INTEGER,GC the_gc) const {
  ic=max(1,min(xcolormap->getNumberColors(),ic));
  Pixel pix=xcolormap->getColor(ic);
  setfgColour(pix,the_gc);
}
#else
void XWindow::setfgColor(INTEGER ic,INTEGER im,GC the_gc) const {
  setfgColor(REAL(ic+1)/REAL(im+1),the_gc);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM == 1)
void XWindow::setfgColor(const INTEGER *ic,const INTEGER *im) {
  if (double_buffer) setfgColor(*ic,*im,buffer_gc);
  else setfgColor(*ic,*im,gc);
}
#else
void XWindow::setfgColor(const INTEGER *ic,const INTEGER *im) {
  REAL val=REAL(*ic+1)/REAL(*im+1);
  setfgColor(&val);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void  XWindow::setFont(const char *font) {
  XGCValues gcvalues;
  XFontStruct *the_font;
  if ((the_font = XLoadQueryFont(display,font)) == 0) {
     cerr << "Cannot open font: " << font << "\n";
     exit(1);
  };
  gcvalues.font = the_font->fid;
  if (double_buffer) XChangeGC( display, buffer_gc, GCFont, &gcvalues );
  else XChangeGC( display, gc, GCFont, &gcvalues );
  XFreeFont(display,the_font);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Pixel XWindow::allocColor(const char *colorname) {
  CHECK_TEST(!monochrome());
  CHECK_POINTER(colorname);
  XColor color, exact;
  Status status=XAllocNamedColor(display,cmap,colorname,&color,&exact);
  CHECK_TEST(status!=0)
  return color.pixel;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::setbgColour(Pixel pixel) {
  if (double_buffer) {
    XGCValues gcvalues;
    unsigned long mask=GCForeground;
    Status status=XGetGCValues(display,buffer_gc,mask,&gcvalues);
    CHECK_TEST(status!=0)
    Pixel foreground=gcvalues.foreground;
    XSetForeground(display,buffer_gc,pixel);
    XFillRectangle(display,buffer_map,buffer_gc,0,0,width,height);
    XSetForeground(display,buffer_gc,foreground);
  } else {
    XSetWindowBackground(display,window,pixel);
    XClearWindow(display, window);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::setbgColour(Pixel pixel,Drawable drawable,GC the_gc) const
{
  XGCValues gcvalues;
  unsigned long mask=GCForeground;
  Status status=XGetGCValues(getDisplay(),the_gc,mask,&gcvalues);
  CHECK_TEST(status!=0)
  Pixel foreground=gcvalues.foreground;
  XSetForeground(getDisplay(),the_gc,pixel);
  XFillRectangle(getDisplay(),drawable,the_gc,0,0,width,height);
  XSetForeground(getDisplay(),the_gc,foreground);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::setfgColor(const char *colorname) const {
  if (monochrome()) return;
  XWindow *copy=const_cast<XWindow*>(this);
  copy->setfgColour(copy->allocColor(colorname));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::setbgColor(const char *colorname) const {
  if (monochrome()) return;
  XWindow *copy=const_cast<XWindow*>(this);
  copy->setbgColour(copy->allocColor(colorname));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int XWindow::pickColor(REAL frac) const {
  if (xcolormap)
    return int(frac*(REAL(xcolormap->getNumberColors())-HALF));
  else return 1;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::newPage(Drawable drawable,GC the_gc) const {
  if (!monochrome()) {
    setbgColour(xcolormap->getColor(0),drawable,the_gc);
  }
  XGCValues gcvalues;
  unsigned long mask=GCForeground | GCBackground;
  Status status=XGetGCValues(display,the_gc,mask,&gcvalues);
  CHECK_TEST(status!=0)
  Pixel background=gcvalues.background;
  XSetForeground(display,the_gc,background);
  XFillRectangle(display,drawable,the_gc,0,0,width,height);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::newPage() {
  if (!monochrome()) setbgColour(xcolormap->getColor(0));
  if (double_buffer) {
    XGCValues gcvalues;
    unsigned long mask=GCForeground | GCBackground;
    Status status=XGetGCValues(display,buffer_gc,mask,&gcvalues);
    CHECK_TEST(status!=0)
    Pixel foreground=gcvalues.foreground;
    Pixel background=gcvalues.background;
    XSetForeground(display,buffer_gc,background);
    XFillRectangle(display,buffer_map,buffer_gc,0,0,width,height);
    XSetForeground(display,buffer_gc,foreground);
  } else XClearWindow( display, window );
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::drawLine(const REAL *px,const REAL *py,
const BOOLEAN *puse_buffer) {
  BOOLEAN use_buffer=(puse_buffer ? *puse_buffer : TRUE);
  int x1=ixLoc(*px); int y1=iyLoc(*py);
  if (doubleBuffer() && use_buffer) {
    XDrawLine(display,buffer_map,buffer_gc,x0,y0,x1,y1);
  } else {
    XDrawLine(display, window, gc, x0, y0, x1, y1 );
  }
  x0 = x1; y0 = y1;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::refresh(int x,int y,int w,int h) const {
  if (double_buffer) {
    XCopyArea(display,buffer_map,window,gc,x,y,w,h,x,y);
  }
} 
// refresh & expose can't be called from the main process
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::refreshFrom(const Buffer &xb) const {
  XCopyArea(display,xb.getPixmap(),window,gc,0,0,width,height,0,0);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::flush(BOOLEAN use_buffer) { 
  if (double_buffer && use_buffer) refresh(0,0,width,height);
  XFlush( display ); 
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//this should be called before createRaster
void XWindow::writePaletteName(FILE *os) {
  const char* pal_name=xcolormap->getPalette()->getName();
  int length=strlen(pal_name)+1;
  fwrite( reinterpret_cast<char*>(&length), sizeof(int), 1, os);
  fwrite( pal_name, sizeof(char), length, os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//this must write what openRaster will read
void XWindow::createRaster(FILE *os) {
  CHECK_POINTER(os);
  CHECK_TEST(!image);
  if (double_buffer) {
    image=
      XGetImage(display,buffer_map,0,0,width,height,AllPlanes,ZPixmap); 
  } else {
    image=XGetImage(display,window,0,0,width,height,AllPlanes,ZPixmap); 
  }
  CHECK_TEST(image!=0);

  REAL dw=REAL(DisplayWidth(display,screen));
  REAL dh=REAL(DisplayHeight(display,screen));
  REAL dm=min(dw,dh);
  REAL rw=rwidth/dm;
  REAL rh=rheight/dm;
  fwrite( reinterpret_cast<char*>(&rw), sizeof(REAL), 1, os);
  fwrite( reinterpret_cast<char*>(&rh), sizeof(REAL), 1, os);
  int length=strlen(name)+1;
  fwrite( reinterpret_cast<char*>(&length), sizeof(int), 1, os);
  fwrite( name, sizeof(char), length, os);

  fwrite(reinterpret_cast<char*>(&(image->xoffset)),sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->format)),sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->byte_order)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bitmap_unit)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bitmap_bit_order)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bitmap_pad)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->depth)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bytes_per_line)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->bits_per_pixel)),
    sizeof(int),1,os);
  fwrite(reinterpret_cast<char*>(&(image->red_mask)),
    sizeof(unsigned long),1,os);
  fwrite(reinterpret_cast<char*>(&(image->green_mask)),
    sizeof(unsigned long),1,os);
  fwrite(reinterpret_cast<char*>(&(image->blue_mask)),
    sizeof(unsigned long),1,os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::openRaster(FILE *is) {
  CHECK_POINTER(is);
  CHECK_TEST(!image);
  int length;
  fread( reinterpret_cast<char*>(&length), sizeof(int), 1, is);
  char *wn=OPERATOR_NEW_BRACKET(char,length);
  fread( wn, 1, length, is);
  CHECK_TEST(strcmp(name,wn)==0); 
  delete wn;
  wn=0;

  int image_width=width,image_height=height;
  int image_xoffset,image_format;
  int image_byte_order,image_bitmap_unit,image_bitmap_bit_order;
  int image_bitmap_pad,image_depth,image_bytes_per_line;
  int image_bits_per_pixel;
  fread(reinterpret_cast<char*>(&image_xoffset),sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_format),sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_byte_order),sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bitmap_unit),sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bitmap_bit_order),
    sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bitmap_pad),
    sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_depth),
    sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bytes_per_line),
    sizeof(int),1,is);
  fread(reinterpret_cast<char*>(&image_bits_per_pixel),
    sizeof(int),1,is);

#ifdef DEBUG
  CHECK_TEST(ImageByteOrder(display)==image_byte_order);
  CHECK_TEST(BitmapUnit(display)==image_bitmap_unit);
  CHECK_TEST(BitmapBitOrder(display)==image_bitmap_bit_order);
  CHECK_TEST(BitmapBitOrder(display)==image_bitmap_bit_order);

  int number_depths;
  int *depth_list=XListDepths(display,screen,&number_depths);
  if (depth_list) {
    BOOLEAN found=FALSE;
    for (int i=0;i<number_depths;i++) {
      if (image_depth==depth_list[i]) found=TRUE;
    }
    CHECK_TEST(found);
  }
  XFree(depth_list);

  CHECK_TEST(image_width<=width);
  CHECK_TEST(image_height<=height);
#endif

  image=XCreateImage(display,0,image_depth,image_format,
    image_xoffset,0,image_width,image_height,image_bitmap_pad,
    image_bytes_per_line);
  CHECK_TEST(image!=0);
  image->data=0;

  fread(reinterpret_cast<char*>(&image->red_mask),
    sizeof(unsigned long),1,is);
  fread(reinterpret_cast<char*>(&image->green_mask),
    sizeof(unsigned long),1,is);
  fread(reinterpret_cast<char*>(&image->blue_mask),
    sizeof(unsigned long),1,is);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::writeRaster(FILE *os) {
  CHECK_POINTER(image);
  CHECK_POINTER(os);
  XDestroyImage(image);
  if (double_buffer) {
    image=
      XGetImage(display,buffer_map,0,0,width,height,AllPlanes,ZPixmap); 
  } else {
    image=XGetImage(display,window,0,0,width,height,AllPlanes,ZPixmap); 
  }
  CHECK_TEST(image!=0);

  int buffer_size=(image->bytes_per_line)*(image->height);
  if (image->format!=ZPixmap) buffer_size *= image->depth;
  fwrite(reinterpret_cast<char*>(&buffer_size), sizeof(int), 1, os);
  fwrite( image->data, buffer_size, 1, os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::readRaster(FILE *is) {
#ifdef DEBUG
  CHECK_TEST(image!=0);
  CHECK_TEST(is!=0);
# ifdef ALPHA
//alpha crashes without the following
  cout << "\tis position = " << ftell(is) << endl;
# endif
#else
  if (!image) return;
  if (!is) return;
#endif

  size_t buffer_size=(image->bytes_per_line)*(image->height);
  if (image->format!=ZPixmap) buffer_size *= image->depth;
  size_t bs;
  char *buffer=OPERATOR_NEW_BRACKET(char,buffer_size);
  image->data=buffer;
  while (TRUE) {
    if (fread(reinterpret_cast<char*>(&bs),sizeof(int),1,is)<1) break;
#ifdef DEBUG
# ifdef ALPHA
//alpha crashes without the following
    cout << "\tis position = " << ftell(is) << endl;
# endif
    CHECK_TEST(buffer_size==bs);
#else
    if (buffer_size!=bs) break;
#endif

    if (fread( buffer, 1, bs, is)<bs) break;

    if (double_buffer) {
     XPutImage(display,buffer_map,buffer_gc,image,0,0,0,0,width,height);
    } else XPutImage(display,window,gc,image,0,0,0,0,width,height); 
    flush();
  }
  delete buffer;
  image->data=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::closeRaster() {
  CHECK_POINTER(image);
  XDestroyImage(image);
  image=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::writeXPM(const char *filename) {
  CHECK_POINTER(filename);
  CHECK_TEST(!image);
  if (double_buffer) {
    image=
      XGetImage(display,buffer_map,0,0,width,height,AllPlanes,ZPixmap); 
  } else {
    image=XGetImage(display,window,0,0,width,height,AllPlanes,ZPixmap); 
  }
  CHECK_TEST(image!=0);

  char string[LENGTH_NAME];
  snprintf(string,LENGTH_NAME,"%s.xpm",filename);
  int ret=XpmWriteFileFromImage(display,string,image,0,0);
  CHECK_TEST(ret==XpmSuccess);
  XDestroyImage(image);
  image=0;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::nextEvent(long event_mask) {
  Status status=TRUE;
  while (status==FALSE) {
    XNextEvent(display, &event);
    status=parseEvent(event_mask);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
Status XWindow::parseEvent(long event_mask) const {
  Status status=FALSE;
  switch (event.type) {
    case KeyPress:
      if (event_mask & KeyPressMask) status=TRUE;
      break;
    case KeyRelease:
      if (event_mask & KeyReleaseMask) status=TRUE;
      break;
    case ButtonPress:
      if (event_mask & ButtonPressMask) status=TRUE;
      break;
    case ButtonRelease:
      if (event_mask & ButtonReleaseMask) status=TRUE;
      break;
    case MotionNotify:
      if (event_mask & (PointerMotionMask | ButtonMotionMask | 
      Button1MotionMask | Button2MotionMask | Button3MotionMask | 
      Button4MotionMask | Button5MotionMask)) status=TRUE;
      break;
    case EnterNotify:
      if (event_mask & EnterWindowMask) status=TRUE;
      break;
    case LeaveNotify:
      if (event_mask & LeaveWindowMask) status=TRUE;
      break;
    case FocusIn:
    case FocusOut:
      if (event_mask & FocusChangeMask) status=TRUE;
      break;
    case KeymapNotify:
      if (event_mask & KeymapStateMask) status=TRUE;
      break;
    case Expose:
      if (event_mask & ExposureMask) status=TRUE;
      break;
    case GraphicsExpose:
    case NoExpose:
      break;
    case CirculateRequest:
    case ConfigureRequest:
    case MapRequest:
      if (event_mask & SubstructureRedirectMask) status=TRUE;
      break;
    case ResizeRequest:
      if (event_mask & ResizeRedirectMask) status=TRUE;
      break;
    case CirculateNotify:
    case ConfigureNotify:
    case CreateNotify:
    case DestroyNotify:
    case GravityNotify:
    case MapNotify:
    case ReparentNotify:
    case UnmapNotify:
      if (event_mask & (StructureNotifyMask | SubstructureNotifyMask))
        status=TRUE;
      break;
    case VisibilityNotify:
      if (event_mask & VisibilityChangeMask) status=TRUE;
      break;
    case ColormapNotify:
      if (event_mask & ColormapChangeMask) status=TRUE;
      break;
    case ClientMessage:
    case MappingNotify:
    case SelectionClear:
    case SelectionNotify:
    case SelectionRequest:
      break;
    case PropertyNotify:
      if (event_mask & PropertyChangeMask) status=TRUE;
      break;
  }
  return status;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BOOLEAN XWindow::buttonIsPressed(int &x,int &y) {
  if (which_button_pressed==0) return FALSE;
  while (TRUE) {
    XNextEvent(display, &event);
    if (event.xany.window == window) {
      switch (event.type) {
        case ButtonRelease:
          if (which_button_pressed==event.xbutton.button) return FALSE;
          break;
        case MotionNotify:
          x = event.xbutton.x;
          y = event.xbutton.y;
          return TRUE;
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
BOOLEAN XWindow::buttonIsPressed(REAL &rx,REAL &ry) {
  int x=0,y=0;
  BOOLEAN pressed=buttonIsPressed(x,y);
  if (pressed) {
    rx=REAL(x)/rwidth;
    ry=REAL(myY(y))/rheight;
  }
  return pressed;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int XWindow::getMouse(REAL &rx, REAL &ry) {
  int ib,x,y;
  ib=getMouse(x,y);
  rx=REAL(x)/rwidth;
  ry=REAL(myY(y))/rheight;
  return ib;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
int XWindow::getMouse(int &x,int &y) {
  while(XCheckTypedWindowEvent(display,window,ButtonPress,&event));

  which_button_pressed=0;
  while (which_button_pressed == 0) {
    XNextEvent(display, &event);
    if (event.xany.window == window && event.type == ButtonPress) {
      which_button_pressed = event.xbutton.button;
      x = event.xbutton.x;
      y = event.xbutton.y;
    }
  }
  return which_button_pressed;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::colorRect(int color,int w,int h,int x,int y,
Drawable drawable,GC the_gc) const {
  if (monochrome()) return;
  color=max(0,min(color,xcolormap->getNumberColors()));
  XSetForeground(display,the_gc,xcolormap->getColor(color));
  XFillRectangle(display,drawable,the_gc,x,y,w,h);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::colorRect(int color,int w,int h,int x,int y) const {
  if (double_buffer) colorRect(color,w,h,x,y,buffer_map,buffer_gc);
  else colorRect(color,w,h,x,y,window,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::colorRect(REAL ratio,REAL rw,REAL rh,REAL rx,REAL ry,
Drawable drawable,GC the_gc) const {
  if (monochrome()) return;
  ratio=max(ZERO,min(ONE,ratio));
  int color;
  color=1+int(ratio*(xcolormap->getNumberColors()-2));
  rw=min(ONE,max(ZERO,rw));
  rh=min(ONE,max(ZERO,rh));
  rx=min(ONE,max(ZERO,rx));
  ry=min(ONE,max(ZERO,ry));
  colorRect(color,int(rw*rwidth),int(rh*rheight),int(rx*rwidth),
            int(ry*rheight),drawable,the_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::colorRect(const REAL *ratio,const REAL *rw,const REAL *rh,
const REAL *rx,const REAL *ry) {
  if (double_buffer) {
    colorRect(*ratio,*rw,*rh,*rx,*ry,buffer_map,buffer_gc);
  } else colorRect(*ratio,*rw,*rh,*rx,*ry,window,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::colorRect(const char *color_name,REAL rw,REAL rh,REAL rx,
REAL ry,Drawable drawable,GC the_gc) {
  Pixel pix=allocColor(color_name);
  if (monochrome()) return;
  XSetForeground(display,the_gc,pix);
  XFillRectangle(display,drawable,the_gc,int(rx*rwidth),int(ry*rheight),
    int(rw*rwidth),int(rh*rheight));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::colorRect(const char *color_name,REAL rw,REAL rh,REAL rx,
REAL ry) {
  if (double_buffer) {
    colorRect(color_name,rw,rh,rx,ry,buffer_map,buffer_gc);
  } else colorRect(color_name,rw,rh,rx,ry,window,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::colorPolygon(int npts,XPoint p[],Drawable drawable,
GC the_gc) const {
  XFillPolygon(display,drawable,the_gc,p,npts,Nonconvex,
               CoordModeOrigin);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::colorPolygon(int color,int npts,XPoint p[],
Drawable drawable,GC the_gc) const {
  if (monochrome()) return;
  int num_cmap_colors=xcolormap->getNumberColors();
  CHECK_BOUNDS(color,0,num_cmap_colors) ;
  XSetForeground(display,the_gc,xcolormap->getColor(color));
  XFillPolygon(display,drawable,the_gc,p,npts,Nonconvex,
               CoordModeOrigin);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::colorPolygon(int color,int npts,XPoint p[]) const {
  if (double_buffer) colorPolygon(color,npts,p,buffer_map,buffer_gc);
  else colorPolygon(color,npts,p,window,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::colorPolygon(INTEGER npts,const REAL *xfrac,
const REAL *yfrac,Drawable drawable,GC the_gc) const {
  XPoint *p=OPERATOR_NEW_BRACKET(XPoint,npts);
  for (int i=0;i<npts;i++) {
    REAL rx=min(ONE,max(ZERO,xfrac[i]));
    REAL ry=min(ONE,max(ZERO,yfrac[i]));
    p[i].x=short(ixLoc(rx));
    p[i].y=short(iyLoc(ry));
  }
  colorPolygon(npts,p,drawable,gc);
  delete p;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::colorPolygon(REAL ratio,INTEGER npts,const REAL *xfrac,
const REAL *yfrac,Drawable drawable,GC the_gc) const {
  if (monochrome()) return;
  int num_cmap_colors=xcolormap->getNumberColors();
  int color=0;
  if (ratio>=ZERO && ratio<=ONE) color=1+int(ratio*(num_cmap_colors-2));
  XSetForeground(display,the_gc,xcolormap->getColor(color));
  colorPolygon(npts,xfrac,yfrac,drawable,the_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::colorPolygon(const REAL *ratio,const INTEGER *npts,
const REAL *xfrac,const REAL *yfrac) const {
  XWindow *copy=const_cast<XWindow*>(this);
  if (double_buffer) {
    copy->colorPolygon(*ratio,*npts,xfrac,yfrac,buffer_map,buffer_gc);
  } else copy->colorPolygon(*ratio,*npts,xfrac,yfrac,window,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::colorPolygon(const INTEGER *npts,const REAL *xfrac,
const REAL *yfrac) const {
  XWindow *copy=const_cast<XWindow*>(this);
  if (double_buffer) {
    copy->colorPolygon(*npts,xfrac,yfrac,buffer_map,buffer_gc);
  } else copy->colorPolygon(*npts,xfrac,yfrac,window,gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::printOn(ostream& os) const {
  os << "XWindow: name = " << name << endl; 

//int num_screens=ScreenCount(display);
//os << "number of screens available = " << num_screens << "\n"
//   << "default screen = " << DefaultScreen(display) << "\n"
//   << "screen used by this window = " << screen << endl;
//for (int i=0;i<num_screens;i++) printScreenInfo(os,display,screen,i);

//printWindowInfo(os,display,window);
//printWMProperties(os,display,window);
//printHosts(display,os);
//printGC(display,gc,os);
//if (double_buffer) printGC(display,buffer_gc,os);
//char *dn;
//if (XFetchName(display,window,&dn)) {
//  os << "display_name = " << *dn <<endl;
//  printDisplayInfo(os,dn,display);
//} else os << "name is unknown" << endl;
//XFree(dn);
  os << "x0 = " << x0 << "\n"
     << "y0 = " << y0 << "\n"
     << "nplanes = " << nplanes << endl;
//os << "cmap = " << cmap << "\n"
//   << "xcolormap = " << (XColormap*) xcolormap << endl;
  os << "double_buffer = " << PRINT_BOOLEAN(double_buffer) << endl;
  VirtualWindow12::printOn(os);
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
XWindow::Buffer::Buffer(XWindow *xw) : VirtualWindow12(),parent(xw),
buffer_map(0),buffer_gc(0),x0(0),y0(0) {
  buffer_map=parent->makePixmap();
  buffer_gc=parent->makeGC(buffer_map);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
XWindow::Buffer::~Buffer() {
  if (getDisplay()) {
    XFreeGC(getDisplay(),buffer_gc);
    XFreePixmap(getDisplay(),buffer_map);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::Buffer::setLineWidth(const INTEGER *lw) { 
  XSetLineAttributes(getDisplay(),buffer_gc,(*lw+1)/2,LineSolid,
    CapRound,JoinRound);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::Buffer::putString(const REAL *px,const REAL *py,
const char *str,const REAL * /*angle*/) {
  XDrawString(getDisplay(),buffer_map,buffer_gc,parent->ixLoc(*px),
    parent->iyLoc(*py),str,strlen(str));
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::Buffer::colorRect(const REAL *ratio,const REAL *rw,
const REAL *rh,const REAL *rx,const REAL *ry) {
  parent->colorRect(*ratio,*rw,*rh,*rx,*ry,buffer_map,buffer_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::Buffer::colorRect(const char *color_name,REAL rw,
REAL rh,REAL rx,REAL ry) {
  parent->colorRect(color_name,rw,rh,rx,ry,buffer_map,buffer_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::Buffer::colorPolygon(int color,int npts,XPoint p[]) const
{
  parent->colorPolygon(color,npts,p,buffer_map,buffer_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::Buffer::colorPolygon(const REAL *ratio,
const INTEGER *npts,const REAL *xfrac,const REAL *yfrac) const {
  parent->colorPolygon(*ratio,*npts,xfrac,yfrac,buffer_map,buffer_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::Buffer::colorPolygon(const INTEGER *npts,
const REAL *xfrac,const REAL *yfrac) const {
  parent->colorPolygon(*npts,xfrac,yfrac,buffer_map,buffer_gc);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void XWindow::Buffer::printOn(ostream& os) const {
  os << "XWindow::Buffer : parent = " << parent << endl;
  printGC(getDisplay(),buffer_gc,os);
  VirtualWindow12::printOn(os);
}
