//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#ifndef X_WINDOW_H
#define X_WINDOW_H "$Header: /home/faculty/johnt/cvs/deal_new/graphics/XWindow.H,v 1.1 2009/08/20 17:31:47 johnt Exp $"
#include <algorithm>
#include <iostream>
#include <X11/Intrinsic.h>
#include "Const.H"
#include "Types.H"
#include "VirtualWindow.H"
#include "XColormap.H"

const long maxIntensity = 65535;

class XWindow : public VirtualWindow12 {
  public:
    typedef XColormap COLOR_MAP_TYPE;
    class Buffer;
    friend class Buffer;
  private:
    char *name;
    BOOLEAN double_buffer;
    int mouseX,mouseY,mouseB,mouseCurrent;

    XtAppContext appcontext;
    Display *display;
    int screen;
    Window window;
    Widget parent,drawa;
    GC gc;

    int width,height;
    REAL rwidth, rheight;
    int x0, y0;
    BOOLEAN grey_scale;
    int which_button_pressed;

    int nplanes;
    XColormap*const xcolormap;
    Colormap cmap;

    Pixmap buffer_map;
    GC buffer_gc;
    XEvent event;

    XImage *image;

    int myY(int y) const { return height - 1 - y; }
    int ixLoc(REAL x) const { 
      return int(min(ONE,max(ZERO,x))*rwidth); 
    }
    int iyLoc(REAL y) const { 
      return myY(int(min(ONE,max(ZERO,y))*rheight)); 
    }

    Pixmap makePixmap() const;
    GC makeGC(Drawable) const;
    void setfgColour(Pixel);
    void setfgColour(Pixel,GC) const;
    void setbgColour(Pixel,Drawable,GC) const;
    void setfgColor(const char*,GC);
    void setbgColor(const char*,Drawable,GC);
    void setfgColor(REAL,GC) const;
    void setbgColor(REAL,Drawable,GC) const;
    void setbgColour(Pixel);
    void setfgColor(INTEGER,INTEGER,GC) const;
    void newPage(Drawable,GC) const;
    void colorRect(int,int,int,int,int,Drawable,GC) const;
    void colorRect(int,int,int,int,int) const;
    void colorRect(REAL,REAL,REAL,REAL,REAL,Drawable,GC) const ;
    void colorRect(const char*, REAL,REAL,REAL,REAL,Drawable,GC);
    void colorPolygon(int,int,XPoint[],Drawable,GC) const;
    void colorPolygon(int,XPoint[],Drawable,GC) const;
    void colorPolygon(int,int,XPoint[]) const;
    void colorPolygon(REAL,int,const REAL*,const REAL*,Drawable,GC)
      const;
    void colorPolygon(int,const REAL*,const REAL*,Drawable,GC) const;
    Status parseEvent(long) const ;

    BOOLEAN buttonIsPressed(int&,int&);
    int getMouse(int&,int&);
  public:
    XWindow(const char*,XColormap*,REAL=HALF,REAL=HALF,BOOLEAN =FALSE,
      BOOLEAN=TRUE,Visual* =CopyFromParent);
    ~XWindow();
    virtual XWindow* getWindow() { return this; }
    virtual const XWindow* getWindow() const { return this; }
    const char* getName() const { return name; }
    const char* getDisplayName() const { 
      char *dn;
      ASSERT(XFetchName(display,window,&dn));
      return dn; 
    }
    BOOLEAN doubleBuffer() const { return double_buffer; }
    int getWidth() const { return width; }
    int getHeight() const { return height; }

    BOOLEAN buttonIsPressed(REAL&,REAL&);
    void newPage();
    void movePen(const REAL *px,const REAL *py) {
      x0 = ixLoc(*px);
      y0 = iyLoc(*py);
    }
    void drawLine(const REAL*,const REAL*,const BOOLEAN* =0);
    BOOLEAN monochrome() const { return grey_scale; }
    Display* getDisplay() const { return display; }
    int getMouse(REAL&,REAL&);
    void colorRect(const REAL*,const REAL*,const REAL*,const REAL*,
      const REAL*);
    void colorRect(const char *name, REAL rw,REAL rh,REAL rx,REAL ry);
    void setLineWidth(const INTEGER *plw) {
      if (doubleBuffer()) {
        XSetLineAttributes(display,buffer_gc,(*plw+1)/2,LineSolid,
          CapRound, JoinRound );
      } else {
        XSetLineAttributes(display,gc,(*plw+1)/2,LineSolid,CapRound,
          JoinRound );
      }
    }
    void expose();
    void input(XtPointer call_data) {;}
    void resize() {;}
    static void exposeCallback(Widget,XtPointer client_data,XtPointer) {
      XWindow *win=reinterpret_cast<XWindow*>(client_data);
      if (win) win->expose();
    }
    static void inputCallback(Widget,XtPointer client_data,
    XtPointer call_data) {
      XWindow *win=reinterpret_cast<XWindow*>(client_data);
      if (win) win->input(call_data);
    }
    static void resizeCallback(Widget,XtPointer client_data,XtPointer) {
      XWindow *win=reinterpret_cast<XWindow*>(client_data);
      if (win) win->resize();
    }
    void setFont(const char*);
    Pixel allocColor(const char*);
    void putString(const REAL *px,const REAL *py,const char *str,
    const REAL * /*angle*/) {
      REAL x=min(ONE,max(ZERO,*px)); 
      REAL y=min(ONE,max(ZERO,*py));
      if (doubleBuffer()) {
        XDrawString(display,buffer_map,buffer_gc,ixLoc(x),iyLoc(y),str,
          strlen(str));
      } else {
        XDrawString(display,window,gc,ixLoc(x),iyLoc(y),str,
          strlen(str));
      }
    }
    void setfgColor(const char*) const;
    void setbgColor(const char*) const;
    int pickColor(REAL) const ;
    void setfgColor(const REAL*);
    void setbgColor(const REAL*);
    void setfgColor(const INTEGER*,const INTEGER*);
    void refresh(int,int,int,int) const ;
    void refreshFrom(const Buffer&) const;
    void nextEvent(long);
    void flush(BOOLEAN=TRUE);
    void flush(const BOOLEAN *pval) { flush(*pval); }
    void sync() { XSync( display , FALSE ); }
    XColormap* getXColormap(void) {return xcolormap;}
    void copyevent(XEvent* ev){event=*ev;}

    void writePaletteName(FILE*);
    void createRaster(FILE*);
    void writeRaster(FILE*);
    void closeRaster();
    void openRaster(FILE*);
    void readRaster(FILE*);
    void writeXPM(const char*);
    void printOn(ostream& os = cerr) const;
//  here is stuff added to deal with const in the distributed situation
    void setLineWidth(const INTEGER *plw) const {
      const_cast<XWindow*>(this)->setLineWidth(plw);
    }
    void setFont(const char *str) const {
      const_cast<XWindow*>(this)->setFont(str);
    }
    void setfgColor(const INTEGER *pi1,const INTEGER *pi2) const {
      const_cast<XWindow*>(this)->setfgColor(pi1,pi2);
    }
    void setfgColor(const REAL *pr1) const {
      const_cast<XWindow*>(this)->setfgColor(pr1);
    }
    void setbgColor(const REAL *pr1) const {
      const_cast<XWindow*>(this)->setbgColor(pr1);
    }
    void colorPolygon(const INTEGER*,const REAL*,const REAL*) const;
    void colorPolygon(const REAL*,const INTEGER*,const REAL*,
      const REAL*)  const;
    void movePen(const REAL *px,const REAL *py) const {
      const_cast<XWindow*>(this)->movePen(px,py);
    }
    void putString(const REAL *px,const REAL *py,const char *str,
    const REAL *angle) const{
      const_cast<XWindow*>(this)->putString(px,py,str,angle);
    }
    void drawLine(const REAL *pr1,const REAL *pr2,const BOOLEAN *pb1=0)
    const {
      const_cast<XWindow*>(this)->drawLine(pr1,pr2,pb1);
    }
    void colorRect(const REAL *pr1,const REAL *pr2,const REAL *pr3,
      const REAL *pr4,const REAL *pr5) const {
      const_cast<XWindow*>(this)->colorRect(pr1,pr2,pr3,pr4,pr5);
    }
};

class XWindow::Buffer : public VirtualWindow12 {
  friend class XWindow;
  private:
    XWindow*const parent;
    Pixmap buffer_map;
    GC buffer_gc;
    int x0, y0;

    Pixmap getPixmap() const { return buffer_map; }
    void setFont(const char*) const {;}
    void flush(BOOLEAN) {;}
    BOOLEAN buttonIsPressed(REAL &x,REAL &y) { x=y=ZERO; return FALSE; }
    int getMouse(REAL &x,REAL &y) { x=y=ZERO; return -1; }
    void writePaletteName(FILE*) {;}
    void createRaster(FILE*) {;}
    void writeRaster(FILE*) {;}
    void closeRaster() {;}
    void openRaster(FILE*) {;}
    void readRaster(FILE*) {;}
  protected:
    Buffer(const Buffer&) : parent(0) { 
      OBSOLETE("copy constructor disallowed"); 
    }
  public:
    Buffer(XWindow*);
    virtual ~Buffer();
  XWindow::Buffer* getWindow() { return this; }
  const XWindow::Buffer* getWindow() const { return this; }
    const char* getName() const { return parent->getName(); }
    const char* getDisplayName() const { 
      return parent->getDisplayName(); 
    }
  
    Display* getDisplay() const { return parent->getDisplay(); }
    BOOLEAN monochrome() const { return parent->monochrome(); }
    void setLineWidth(const INTEGER*);
    void setfgColour(Pixel pixel) { 
      parent->setfgColour(pixel,buffer_gc); 
    }
    void setbgColour(Pixel pixel) {
      parent->setbgColour(pixel,buffer_map,buffer_gc);
    }
    void setfgColor(const char *colorname) const {
      parent->setfgColor(colorname,buffer_gc);
    }
    void setbgColor(const char *colorname) const {
      parent->setbgColor(colorname,buffer_map,buffer_gc);
    }
    void setfgColor(const REAL *pfrac) { 
      parent->setfgColor(*pfrac,buffer_gc); 
    }
    void setbgColor(const REAL *pfrac) { 
      parent->setbgColor(*pfrac,buffer_map,buffer_gc); 
    }
    void setfgColor(const INTEGER *ic,const INTEGER* im) {
      parent->setfgColor(*ic,*im,buffer_gc);
    }
    void putString(const REAL*,const REAL*,const char*,
      const REAL * /*angle*/);
    void movePen(const REAL *px,const REAL *py) {
      x0=parent->ixLoc(*px); y0=parent->iyLoc(*py);
    }
    void drawLine(const REAL *px,const REAL *py,const BOOLEAN* =0) {
      int x1=parent->ixLoc(*px); int y1=parent->iyLoc(*py);
      XDrawLine(getDisplay(),buffer_map,buffer_gc,x0,y0,x1,y1);
      x0 = x1; y0 = y1;
    }

    void newPage() { parent->newPage(buffer_map,buffer_gc); }
    void colorPolygon(int,int,XPoint[]) const;
    void colorRect(const REAL*,const REAL*,const REAL*,const REAL*,
      const REAL*);
    void colorRect(const char*,REAL,REAL,REAL,REAL);
    int pickColor(REAL frac) const { return parent->pickColor(frac); }
    void printOn(ostream& os = cerr) const;
//  here is stuff added to deal with const in the distributed situation
    void setLineWidth(const INTEGER *plen) const {
      const_cast<XWindow::Buffer*>(this)->setLineWidth(plen);
    }
    void setfgColor(const INTEGER *p,const INTEGER *q) const {
      const_cast<XWindow::Buffer*>(this)->setfgColor(p,q);
    }
    void setfgColor(const REAL *pr) const {
      const_cast<XWindow::Buffer*>(this)->setfgColor(pr);
    }
    void setbgColor(const REAL *pr) const {
      const_cast<XWindow::Buffer*>(this)->setbgColor(pr);
    }
    void putString(const REAL *x,const REAL *y,const char *str,
    const REAL *angle) const {
      const_cast<XWindow::Buffer*>(this)->putString(x,y,str,angle);
    }
    void movePen(const REAL *x,const REAL *y) const {
      const_cast<XWindow::Buffer*>(this)->movePen(x,y);
    }
    void drawLine(const REAL *x,const REAL *y,const BOOLEAN *pb) const {
      const_cast<XWindow::Buffer*>(this)->drawLine(x,y,pb);
    }
    void colorPolygon(const REAL*,const INTEGER*,const REAL*,
      const REAL*)  const;
    void colorPolygon(const INTEGER*,const REAL*,const REAL*)  const;
    void colorRect(const REAL *t,const REAL *x,
      const REAL *y,const REAL *z,const REAL *w) const {
      const_cast<XWindow::Buffer*>(this)->colorRect(t,x,y,z,w);
    }

};

extern void printHosts(Display*,ostream&);
extern void printDisplayInfo(ostream&,const char*,Display*);
extern void printScreenInfo(ostream&,Display*,int,int);
extern void printVisualInfo(ostream&,Display*,int);
extern void printWidgetInfo(ostream&,Widget);
extern void printWindowInfo(ostream&,Display*,Window);
extern void printXWindowAttributes(ostream&,Display*,Window);
extern void printWMProperties(ostream&,Display*,Window);
extern void printPixmapFormatValues(ostream&,Display*);
extern void printXEventType(XEvent*);
extern void printGC(Display*,GC,ostream&);

#endif
