//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#ifndef DEBUG_H
#define DEBUG_H "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/Debug.H,v 1.1 2009/08/20 17:33:32 johnt Exp $"
#include <fstream>
#include <iostream>
#include <stdlib.h>
#include "Arch.H"
#include "Errors.H"

#ifdef DEBUG
#define DEBUG_ONLY(debug) debug
#define DEBUG_COMMA(debug) ,debug
#define DEBUG_CHOICE(debug,optimize) debug
#else
#define DEBUG_ONLY(debug)
#define DEBUG_COMMA(debug)
#define DEBUG_CHOICE(debug,optimize) optimize
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
inline void openHandler(const char *file_name,fstream &file) {
  cout << "cannot open " << file_name << "\n"
       << "rdstate = " << file.rdstate() << endl;
  cerr << "cannot open " << file_name << "\n"
       << "rdstate = " << file.rdstate() << endl;
  abort();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
inline void openHandler(const char *file_name,ifstream &file) {
  cout << "cannot open " << file_name << "\n"
       << "rdstate = " << file.rdstate() << endl;
  cerr << "cannot open " << file_name << "\n"
       << "rdstate = " << file.rdstate() << endl;
  abort();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
inline void openHandler(const char *file_name,ofstream &file) {
  cout << "cannot open " << file_name << "\n"
       << "rdstate = " << file.rdstate() << endl;
  cerr << "cannot open " << file_name << "\n"
       << "rdstate = " << file.rdstate() << endl;
  abort();
}

extern "C" {
  void F77_NAME(print_loc)(const void *p);
  void F77_NAME(print_watch_loc)();
  void F77_NAME(check_watch_loc)();
}

class WatchPoint {
  public:
    static void *watch_point;
    static void *watch_point_address;

    WatchPoint() {;}
//  template<class T> void set(T *p);
    static void set(float **p);
    static void set(double **p);
    static void set(int **p);
    static void set(unsigned int **p);
    static void set(long **p);
    static void set(unsigned long **p);
    static void set(short **p);
    static void set(unsigned short **p);
    static void set(void **p);
    static void unset();
    static void print();
    static void check();
};

class WatchPointer {
  public:
    static const void *watch_pointer;
    static const void **watch_pointer_address;

    WatchPointer();
    static void set(const void **t) { 
      watch_pointer=*t; watch_pointer_address=t; 
    }
    static void unset() { watch_pointer=0; watch_pointer_address=0; }
    static void print() {
      if (watch_pointer_address!=0) {
        cout << "\twatch_pointer = " << watch_pointer << endl;
      }
    }
    static void check() {
      if (watch_pointer_address!=0) {
        ASSERT(*watch_pointer_address==watch_pointer);
      }
    }
};

extern void wait();
//extern void beep();
#endif
