//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
// "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/IDLList.C,v 1.1 2009/08/20 17:33:32 johnt Exp $"
#include "IDLList.H"
#include <limits.h>
//#include <values.h>
#include <stdlib.h>
//do not use Tracer in this class: we use this for MemoryDebugger
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template<class T> void IDLList<T>::reverse() {
  INTEGER l=length();
  IDLListNode *tbk=0;
  IDLListNode *tfd=0;
  IDLListNode *t=h;
  for (INTEGER i=0;i<l;i++,t=tbk) t->reverseConnections(tbk,tfd);
  h=h->next();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template<class T> int IDLList<T>::length() const {
  int l = 0;
  IDLListNode *t = h;
  if (t != 0) do { ++l; t = t->next(); } while (t != h);
  return l;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template<class T> BOOLEAN IDLList<T>::owns(const IDLListNode *p) const {
  if (h != 0 && p != 0) {
    IDLListNode *t = h;
    do {
      if (t == p) return TRUE;
      t = t->next();
    } while (t != h);
  }
  return FALSE;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template<class T> BOOLEAN IDLList<T>::progress(const IDLListNode *p,
const IDLListNode *q) const {
  CHECK_TEST(!empty());
  CHECK_TEST(owns(p));

  do {
    if (p==q) return TRUE;
    p=p->next();
  } while (p!=h);
  return FALSE;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template<class T> BOOLEAN IDLList<T>::regress(const IDLListNode *p,
const IDLListNode *q) const {
  CHECK_TEST(!empty());
  CHECK_TEST(owns(p));

  do {
    if (p==q) return TRUE;
    p=p->prev();
  } while (p!=h->prev());
  return FALSE;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template<class T> BOOLEAN IDLList<T>::ok() const {
  BOOLEAN v = TRUE;
  if (h) {
    IDLListNode *t = h;
    long count = LONG_MAX;      // Lots of chances to find h!
    do {
      count--;
      v &= t->prev()->next() == t;
      v &= t->next()->prev() == t;
      t = t->next();
    } while (v && count > 0 && t != h);
    v &= count > 0;
  }
  return v;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template<class T> void IDLList<T>::printOn(ostream& os) const {
  if (ok()) {
    int l=0;
    os << "IDLList: " << this << endl;
    for (T *p = first();p;p=next(p), l++) {
      os << l << " : "; p->printOn(os);
    }
  } else {
    os << "IDLLIst: list is not ok" << endl;
  }
  os << "\n";
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template<class T> ostream& operator<<(ostream& s,const IDLList<T> &dl) {
  T *p=dl.first();
  if (p) {
    for (int l=0; p; p=dl.next(p), l++) {
      s << l << " : "; p->printOn(s); s << endl;
    }
  } else s << "list is empty" << endl;
  return s;
}
