//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#ifndef IDLLIST_H
#define IDLLIST_H "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/IDLList.H,v 1.1 2009/08/20 17:33:32 johnt Exp $"
#include "Types.H"
#include "IDLListNode.H"
#include "Errors.H"
//do not use Tracer in this class: we use this for MemoryDebugger

//the instantiation will fail unless T is derived from IDLListNode
template<class T> class IDLList {
  private:
    IDLListNode *h;
  public:
    IDLList() : h(0) {;}
    ~IDLList() {;}
    BOOLEAN empty() const { return h==0; }
    BOOLEAN notEmpty() const { return h!=0; }

    T* first() const { return dynamic_cast<T*>(h); }
    T* next(const IDLListNode *p) const {
      return dynamic_cast<T*>((p==0||h==0||p==h->prev())?0 :p->next());
    }
    T* prev(const IDLListNode *p) const {
      return dynamic_cast<T*>((p == 0 || p == h) ? 0 : p->prev());
    }
    T* last() const { 
      return dynamic_cast<T*>(h == 0 ? 0 : h->prev());
    }

    void prepend(T *item) {
      if (!h) h=item->selfConnect(); 
      else h=item->placeBefore(h);
    }
    void append(T *item) {
      if (!h) h=item->selfConnect();
      else item->placeAfter(h->prev());
    }
    void insAfter(T *p,T *item) {
      if (!h) h=item->selfConnect();
      else if (!p) h=item->placeBefore(h);
      else item->placeAfter(p);
    }
    void insBefore(T *p,T *item) {
      if (!h) h=item->selfConnect();
      else if (!p) item->placeAfter(h->prev());
      else { item->placeBefore(p); if (p==h) h=item; }
    }
    void replace(T *oldp,T *newt) {
      CHECK_TEST(owns(oldp));
      if (oldp->next() == oldp) newt->selfConnect();
      else newt->replace(oldp);
      if (h == oldp) h = newt;
      oldp = newt;
    }
    void join(IDLList &b) {
      IDLListNode *t=b.h;
      if (!h) h = t;
      else if (t) h->switchBackConnectionWith(t);
      b.h = 0;
    }
    void joinAtFront(IDLList &b) {
      IDLListNode *t=b.h;
      if (!h) h=t;
      else if (t) { h->switchBackConnectionWith(t); h=t; }
      b.h=0;
    }
    T* delAfter(IDLListNode *p) {
      CHECK_POINTER(h);
      CHECK_TEST(p != h->prev());
      CHECK_TEST(p == 0 || p->next() != 0);
      if (!p) p = h->prev();
      T *t = static_cast<T*>(p->next());
      if (t==p) h=0;
      else { if (t == h) h = t->next(); t->disconnect(); }
      return t;
    }
    T* delBefore(IDLListNode *p) {
      CHECK_POINTER(h);
      CHECK_TEST(p != h);
      CHECK_TEST(p == 0 || p->prev() != 0);
      if (!p) p=h;
      T *t = static_cast<T*>(p->prev());
      if (t==p) h=0;
      else { if (t == h) h = t->next(); t->disconnect(); }
      return t;
    }
    T* delAndNext(T *&p) {
      CHECK_POINTER(h);
      if (!p) p = static_cast<T*>(h);
      T *next_node = static_cast<T*>(p->next());
      T* deleted_node = p;
      if (p == h->prev()) next_node = 0;
      if (next_node == p) h = 0;
      else if (p == h) h = next_node;
      p->disconnect();
      p=next_node;
      return deleted_node;
    }
    T* delAndPrev(T *&p) {
      CHECK_POINTER(h);
      if (!p) p = static_cast<T*>(h->prev());
      T *prev_node = static_cast<T*>(p->prev());
      T* deleted_node = p;
      if (prev_node == p) h = 0;
      else if (p == h) { h = p->next(); prev_node=0; }
      p->disconnect();
      p=prev_node;
      return deleted_node;
    }
    void remove(T *p) {
      CHECK_TEST(owns(p));
      if (p==h) h = (h->next()==h ? static_cast<T*>(0) : h->next());
      p->disconnect();
    }
//  if p==h we are swapping first with last:
    void swapWithPrev(IDLListNode *p) {
      CHECK_TEST(!empty());
      CHECK_POINTER(p);
      CHECK_TEST(owns(p));
      IDLListNode *pbk=p->prev();
      if (pbk==h) h=p;
      else if (p==h) h=pbk;
      p->switchWithPrevious();
    }
//  if p==h->prev() we are swapping first with last:
    void swapWithNext(IDLListNode *p) {
      CHECK_TEST(!empty());
      CHECK_POINTER(p);
      CHECK_TEST(owns(p));
//    CHECK_TEST(p!=h->prev());
      IDLListNode *pfd=p->next();
      if (p==h) h=pfd;
      else if (pfd==h) h=p;
      p->switchWithNext();
    }
    void swap(IDLListNode *p,IDLListNode *q) {
      if (p==q) return;
      CHECK_TEST(!empty());
      CHECK_TEST(p!=0 && q!=0);
      CHECK_TEST(owns(p) && owns(q));
      IDLListNode *pbk=p->prev();
      if (pbk==q) { swapWithPrev(p); q=p->next(); return; }
      IDLListNode *pfd=p->next();
      if (pfd==q) { swapWithNext(p); q=p->prev(); return; }
      if (p==h) h=q;
      else if (q==h) h=p;
      p->switchWith(q);
    }
    void reverse();
    int length() const;
    BOOLEAN owns(const IDLListNode*) const;
    BOOLEAN progress(const IDLListNode*,const IDLListNode*) const ;
    BOOLEAN regress(const IDLListNode*,const IDLListNode*) const ;
    BOOLEAN ok() const;

    void printOn(ostream& os = cerr) const;
};
template<class T> ostream& operator<<(ostream&,const IDLList<T> &);
#endif
