//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#ifndef IDLLIST_NODE_H
#define IDLLIST_NODE_H "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/IDLListNode.H,v 1.1 2009/08/20 17:33:32 johnt Exp $"
#include <iostream>

class IDLListNode {
  private:
    IDLListNode* bk;
    IDLListNode* fd;
  public:
    IDLListNode() : bk(0), fd(0) {;}
    virtual ~IDLListNode() {;}
    IDLListNode* prev() const { return bk; }
    IDLListNode* next() const { return fd; }
    IDLListNode* selfConnect() { return fd=bk=this; }
    IDLListNode* placeBefore(IDLListNode *n) {
      fd=n; bk=n->bk; n->bk->fd=this; return n->bk=this;
    }
    IDLListNode* placeAfter(IDLListNode *n) {
      fd=n->fd; bk=n; n->fd->bk=this; return n->fd=this;
    }
    IDLListNode* replace(IDLListNode *n) {
      n->bk->fd=this; n->fd->bk=this; bk=n->bk; fd=n->fd; return n;
    }
    void switchBackConnectionWith(IDLListNode *n) {
      IDLListNode *t=n->bk; bk->fd=n; n->bk=bk; bk=t; t->fd=this;
    }
    IDLListNode* disconnect() {
      bk->fd=fd; fd->bk=bk; return this;
    }
    void reverseConnections(IDLListNode *&p,IDLListNode *&n) {
      p=fd; n=bk; fd=n; bk=p;
    }
    void switchWithPrevious() {
      if (bk!=fd) {
        IDLListNode *bkbk=bk->bk; fd->bk=bk; bkbk->fd=this; 
        bk->bk=this; bk->fd=fd; fd=bk; bk=bkbk;
      }
    }
    void switchWithNext() {
      if (bk!=fd) {
        IDLListNode *fdfd=fd->fd; bk->fd=fd; fdfd->bk=this;
        fd->fd=this; fd->bk=bk; bk=fd; fd=fdfd;
      }
    }
    void switchWith(IDLListNode *n) {
      IDLListNode *b=bk; IDLListNode *f=fd; fd->bk=bk->fd=n; bk=n->bk;
      fd=n->fd; n->fd->bk=n->bk->fd=this; n->bk=b; n->fd=f;
    }
    virtual void printOn(ostream &os = cerr) const {
      os << "IDLListNode: this = " << this
         << " bk = " << bk << " fd = " << fd;
    }
};   
inline ostream& operator<<(ostream &os,const IDLListNode &node) {
  node.printOn(os); return os;
}
#endif
