//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#ifndef MYINLINE_H
#define MYINLINE_H "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/MyInline.H,v 1.1 2009/08/20 17:33:32 johnt Exp $"
#include <sys/param.h>

#ifdef __SVR4
#include <sunmath.h>
#endif

#include <math.h>

#ifndef __linux__
#include <generic.h>
#endif

#include "Arch.H"
extern "C" {
  double F77_NAME(my_powd)(double&,int&);
  int F77_NAME(my_powi)(int&,int&);
  float F77_NAME(my_pows)(float&,int&);
# ifdef __GNUC__
  float F77_NAME(my_powf)(float&,float&);
# endif
}

//use min,max in <algorithm>

#ifdef __DECCXX
// from /usr/include/g++-2/cmath.h
inline float  abs(float  x) { return (x>=0. ? x : -x); }
inline double abs(double x) { return (x>=0. ? x : -x); }
inline long double abs(long double x) { return (x>=0. ? x : -x); }
#endif

inline int coarsen(int v,int r) { return (v<0 ? (v+1)/r-1 : v/r); }

//can't use power in <numeric>: won't handle negative exponent
//although /usr/include/libftn.h has double**int and int**long,
//  libftn.h leaves cilist64 undefined
//inline double power(double x,double y) { return pow(x,y); }
//inline double power(double x,int y) { return F77_NAME(my_powd)(x,y); }
#ifdef __GNUC__
//inline float power(float x,float y) { return F77_NAME(my_powf)(x,y); }
#else
//inline float power(float x,float y) { return powf(x,y); }
#endif
//inline float power(float x,int y) { return F77_NAME(my_pows)(x,y); }
//inline int power(int x,int y) { return F77_NAME(my_powi)(x,y); }

inline bool between(double test,double bound1,double bound2) {
  return (test-bound1)*(test-bound2)>=0.;
}
inline double RadToDeg(double radians) {
  return radians*180./M_PI;
}
inline double DegToRa(double degrees) {
  return degrees*M_PI/180.;
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#include <iostream>
#include <fstream>
inline void iosError(ios &s,const char *message ) {
//if (s.good()) return 0;
  cout << message << endl;
  if (s.eof()) cout << "eofbit set" << endl;
  if (s.fail()) cout << "failbit set" << endl;
  if (s.bad()) cout << "badbit set" << endl;
  abort();
//return s.rdstate();
}
inline void iosDebug(ios &s) {
  cout << "\trdstate = " << s.rdstate() << endl;
  cout << "\tbad = " << s.bad() << endl;
  cout << "\tfail = " << s.fail() << endl;
  cout << "\teof = " << s.eof() << endl;
  cout << "\tgood = " << s.good() << endl;
}    
inline void ofstreamDebug(ofstream &s) {
#ifndef __DECCXX
  cout << "\tis_open = " << s.is_open() << endl;
#endif
}
#endif
