//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
// "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/NIDLList.C,v 1.1 2009/08/20 17:33:32 johnt Exp $"
#include "NIDLList.H"
#include "Errors.H"
#include <limits.h>
#include <stdlib.h>
#include "MemoryDebugger.H"

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template<class T> ostream& operator<<(ostream& strm,
const NIDLList<T> &list) {
  NIDLListNode<T>* p = list.first();
  if (p) {
    for( int l=0; p; p=list.next(p), l++ ){
      strm << l << " : " << *p << "\n";
    }
  } else strm << "list is empty" << endl;
  return strm;
}

template<class T> void NIDLList<T>::printOn(ostream& os) const {
  if (ok()) {
    int l=0;
    os << "NIDLList<T>: " << this << endl;
    for (NIDLListNode<T> *p = first();p;p=next(p), l++) {
      os << l << " : " << p << "\n";
    }
  } else {
    os << "NIDLList: list is not ok" << endl;
  }
  os << "\n";
}

template<class T> void NIDLList<T>::reverse() {
  INTEGER l=length();
  NIDLListNode<T> *tbk=0;
  NIDLListNode<T> *tfd=0;
  NIDLListNode<T> *t=h;
  for (INTEGER i=0;i<l;i++,t=tbk) t->reverseConnections(tbk,tfd);
  h=h->next();
}

template<class T> void NIDLList<T>::swap(NIDLListNode<T> *p,
NIDLListNode<T> *q) {
  if (p==q) return;
  CHECK_TEST(!empty());
  CHECK_TEST(p!=0 && q!=0);
  CHECK_TEST(owns(p) && owns(q));

  NIDLListNode<T> *pbk=p->prev();
  if (pbk==q) {
    swapWithPrev(p);
    q=p->next();
    return;
  }
  NIDLListNode<T> *pfd=p->next();
  if (pfd==q) {
    swapWithNext(p);
    q=p->prev();
    return;
  }
  if (p==h) h=q;
  else if (q==h) h=p;
  p->switchWith(q);
}

template<class T> int NIDLList<T>::length() const {
  int l = 0;
  NIDLListNode<T>* t = h;
  if (t != 0) do { ++l; t = t->next(); } while (t != h);
  return l;
}

template<class T> BOOLEAN NIDLList<T>::owns(const NIDLListNode<T> *p) 
const {
  if (h != 0 && p != 0) {
    NIDLListNode<T>* t = h;
    do {
      if (t == p) return TRUE;
      t = t->next();
    } while (t != h);
  }
  return FALSE;
}

template<class T> BOOLEAN NIDLList<T>::progress(
const NIDLListNode<T> *p,const NIDLListNode<T> *q) const {
  CHECK_TEST(!empty());
  CHECK_TEST(owns(p));

  do {
    if (p==q) return TRUE;
    p=p->next();
  } while (p!=h);
  return FALSE;
}

template<class T> BOOLEAN NIDLList<T>::regress(const NIDLListNode<T> *p,
const NIDLListNode<T> *q) const {
  CHECK_TEST(!empty());
  CHECK_TEST(owns(p));

  do {
    if (p==q) return TRUE;
    p=p->prev();
  } while (p!=h->prev());
  return FALSE;
}

template<class T> BOOLEAN NIDLList<T>::ok() const {
  BOOLEAN v = TRUE;
  if (h) {
    NIDLListNode<T> *t = h;
    long count = LONG_MAX;
    do {
      count--;
      v &= t->prev()->next() == t;
      v &= t->next()->prev() == t;
      t = t->next();
    } while (v && count > 0 && t != h);
    v &= count > 0;
  }
//make sure we found the end of the list
  ASSERT(v);
  return v;
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template<class T> void NIDLListC<T>::sort(BOOLEAN largest_first) {
  if (this->empty()) return;
  for (NIDLListNode<T> *p=this->first(),*pfd=next(p);pfd;
  p=pfd,pfd=next(pfd)){
    NIDLListNode<T> *choice=p;
    if (largest_first) {
//    find largest after p
      for (NIDLListNode<T> *q=pfd;q;q=next(q)) {
        if (!(*(choice->getData()) >= *(q->getData()))) choice=q;
      }
    } else  {
      for (NIDLListNode<T> *q=pfd;q;q=next(q)) {
        if (!(*(choice->getData()) <= *(q->getData()))) choice=q;
      }
    }
    if ( p != choice) swap(p,choice);
  }
}

template<class T> void NIDLListC<T>::printOn(ostream& os) const {
  os << "NIDLListC<T>: " << this << endl;
  NIDLList<T>::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
template<class T,class V> void NIDLValueList<T,V>::selectionSort(
BOOLEAN largest_first) {
  if (empty()) return;
  for (NIDLValueListNode<T,V> *p=first(),*pfd=next(p);pfd;
  p=pfd,pfd=next(pfd)){
    NIDLValueListNode<T,V> *choice=p;
    if (largest_first) {
//    find largest after p
      for (NIDLValueListNode<T,V> *q=pfd;q;q=next(q)) {
        if (!(choice->getValue() >= q->getValue())) choice=q;
      }
    } else  {
      for (NIDLValueListNode<T,V> *q=pfd;q;q=next(q)) {
        if (!(choice->getValue() <= q->getValue())) choice=q;
      }
    }
    if ( p != choice) {
      swap(p,choice);
    }
  }
}

template<class T,class V> void NIDLValueList<T,V>::insertionSort(
BOOLEAN largest_first) {
  if (empty()) return;
  NIDLValueListNode<T,V> *pfd=next(first());
  while (pfd) {
    NIDLValueListNode<T,V> *p=pfd->prev();
    NIDLValueListNode<T,V> *pfdfd=next(pfd);
    if (largest_first)  {
      NIDLValueListNode<T,V> *q=p;
      NIDLValueListNode<T,V> *qfd=pfd;
      while (!(q->getValue()>=qfd->getValue())) {
        NIDLValueListNode<T,V> *qbk=prev(q);
        swapWithPrev(qfd);
        if (!qbk) break;
        q=qbk;
        qfd=q->next();
      }
    } else {
      NIDLValueListNode<T,V> *q=p;
      NIDLValueListNode<T,V> *qfd=pfd;
      while (!(q->getValue()>=qfd->getValue())) {
        NIDLValueListNode<T,V> *qbk=prev(q);
        swapWithPrev(qfd);
        if (!qbk) break;
        q=qbk;
        qfd=q->next();
      }
    }
    pfd=pfdfd;
  }
}

//this is probably programmed wrong:
//  it always does length*(length-1)/2 compares
template<class T,class V> void NIDLValueList<T,V>::bubbleSort(
BOOLEAN largest_first) {
  if (empty()) return;
  NIDLValueListNode<T,V> *p=last();
  while (p) {
    NIDLValueListNode<T,V> *start=first();
    NIDLValueListNode<T,V> *q=start,*qfd=next(q);
    while (q!=p && qfd!=0) {
      if (largest_first ? !(q->getValue()>=qfd->getValue()) : 
      !(q->getValue()<=qfd->getValue()) ) {
        swapWithNext(q);
        if (qfd==p) break;
      } else q=next(q);
      qfd=next(q);
    }
    if (p!=prev(q)) p=prev(p);
  }
}

template<class T,class V> void NIDLValueList<T,V>::quickSort(
NIDLValueListNode<T,V> *left,NIDLValueListNode<T,V> *right,
BOOLEAN largest_first) {
  if (left==right) return;
  if (empty()) return;
  CHECK_TEST(left!=0&&right!=0);
  CHECK_TEST(progress(left,right));

  NIDLValueListNode<T,V> *low=left;
  NIDLValueListNode<T,V> *high=right;
  if (largest_first) {
    while (TRUE) {
//    find first node after left that is < left
      do {
        low=low->next();
        if (low==right->next()) break;
      } while (low->getValue() >= left->getValue());
//    find last node after left that is >= left
      while (!(high->getValue() >= left->getValue())) {
        high=high->prev();
      }
      if (low==high->next()) break;
      swap(low,high);
      if (right==low) right=high;
      else if (right==high) right=low;
      NIDLValueListNode<T,V> *t=low;
      low=high;
      high=t;
      if (low==high->next()) break;
    }
  } else {
    while (TRUE) {
//    find first node after left that is > left
      do {
        low=low->next();
        if (low==right->next()) break;
      } while (low->getValue() <= left->getValue());
//    find last node after left that is <= left
//    compares++;
      while (!(high->getValue() <= left->getValue())) {
        high=high->prev();
      }
      if (low==high->next()) break;
      swap(low,high);
      if (right==low) right=high;
      else if (right==high) right=low;
      NIDLValueListNode<T,V> *t=low;
      low=high;
      high=t;
      if (low==high->next()) break;
    }
  }
  if (left!=high) {
    swap(left,high);
    if (right==high) right=left;
    quickSort(high,left->prev(),largest_first);
  }
  if (low != right->next()) quickSort(low,right,largest_first);
}

template<class T,class V> void NIDLValueList<T,V>::balance(INTEGER bins,
NIDLValueList<T,V> *&bin_list) {
  while (NIDLList<T>::first()) {
    V min_value=bin_list[0].value;
    INTEGER min_bin=0;
//  find bin list with minimum value
    for (INTEGER i=1;i<bins;i++) {
      if (bin_list[i].value < min_value) {
        min_bin=i;
        min_value=bin_list[i].value;
      }
    }
//  h cannot be on this list and a bin list simultaneously
    V val=first()->getValue();
    bin_list[min_bin].append(remove(first()),val);
  }
}

template<class T,class V> void NIDLValueList<T,V>::printOn(ostream& os)
const {
  if (ok()) {
    int l=0;
    os << "NIDLValueList<T,V>: " << this << endl;
    os << "list value: " << value << endl;
    for (NIDLValueListNode<T,V> *p = first();p;p=next(p), l++) {
      os << l << " : " << *p << "\n";
    }
  } else {
    os << "list is not ok" << endl;
  }
  os << "\n";
}

template<class T,class V> ostream& operator<<(
ostream& strm ,const NIDLValueList<T,V> &list) {
  strm << "value = " << list.getValue() << endl;
  NIDLValueListNode<T,V>* p = list.first();
  if (p) {
    for( int l=0; p; p=list.next(p), l++ ){
      strm << l << " : " << *p << "\n";
    }
  } else strm << "list is empty" << endl;
  return strm;
}
