//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
#ifndef TIMEDOBJECT_H
#define TIMEDOBJECT_H "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/TimedObject.H,v 1.1 2009/08/20 17:33:33 johnt Exp $"
#ifdef TIMING
#include <stdlib.h>
#include <sys/times.h>
#include <sys/param.h>
#include "Const.H"
#include "Errors.H"
#include "ISLList.H"

class TimedObjectList;
class TimedObject : public ISLListNode {
  private:
  protected:
    char *name;
    INTEGER number_calls;
    BOOLEAN is_on;
    REAL total_run_time;
    clock_t start;
    struct tms usage;

    TimedObject( const TimedObject& ) { 
      OBSOLETE("copy constructor disallowed"); 
    }
  public:
    TimedObject(const char *);
    TimedObject(const char*,TimedObjectList&);
    virtual ~TimedObject() { if (name) delete name; name=0; }

    void turnOn();
    REAL turnOff();
    BOOLEAN isOn() const { return is_on; }
    INTEGER numberCalls() const { return number_calls; }
    REAL totalRunTime() const { return total_run_time; }
    virtual void reset() {
      ASSERT(!is_on);
      number_calls=0; is_on=FALSE; total_run_time=ZERO; start=0;
    }
    virtual void printOn(ostream&) const;
};

class TimedObjectList : public ISLList<TimedObject> {
  private:
  protected:
    TimedObjectList(const TimedObjectList&) { 
      OBSOLETE("copy constructor disallowed"); 
    }
  public:
    TimedObjectList() {;}
    virtual ~TimedObjectList();
    virtual void reset() const;
    virtual void printOn(ostream&) const;
};

class Timer {
  private:
  protected:
    TimedObject *timed_object;
  public:
    Timer(TimedObject *t) : timed_object(t) { timed_object->turnOn(); }
    virtual ~Timer() { timed_object->turnOff(); }
    virtual void printOn(ostream &os) const {
      timed_object->printOn(os);
    }
};

class IndexedTimedObjectList;
class IndexedTimedObject : public ISLListNode {
  private:
    int max_indices;
  protected:
    TimedObject **object_array;
    long *total_length;

    IndexedTimedObject( const IndexedTimedObject& ) { 
      OBSOLETE("copy constructor disallowed"); 
    }
  public: 
    IndexedTimedObject(const char*,int,IndexedTimedObjectList&);
    IndexedTimedObject(const char*,int);
    virtual ~IndexedTimedObject();
    void turnOn(INTEGER,INTEGER =0);
    void turnOff(INTEGER =0);
    BOOLEAN isOn(INTEGER index =0) const { 
      return (index>=0 && index<max_indices 
             ? object_array[index]->isOn() : FALSE); 
    }
    void reset();
    virtual void printOn(ostream&) const;
};

class IndexedTimedObjectList : public ISLList<IndexedTimedObject> {
  protected:
    IndexedTimedObjectList(const IndexedTimedObjectList&) {
      OBSOLETE("copy constructor disallowed");
    }
  public: 
    IndexedTimedObjectList() {;}
    virtual ~IndexedTimedObjectList();
    void reset() const;
    virtual void printOn(ostream&) const;
};

class IndexedTimer {
  protected:
    IndexedTimedObject *timed_object;
    INTEGER index;
  public: 
    IndexedTimer(IndexedTimedObject *t,INTEGER length,INTEGER l=0) :
      timed_object(t),index(l) { timed_object->turnOn(length,l);}
    virtual ~IndexedTimer() { timed_object->turnOff(index); }
    virtual void printOn(ostream &os) const {timed_object->printOn(os);}
};

#define TIMER(t,s,l) if (!t) { t=OPERATOR_NEW TimedObject(s,l); } Timer z(t);
#define INDEXED_TIMER(t,s,l,n,m,k)                                             \
  if (!t) { t=OPERATOR_NEW IndexedTimedObject(s,n,l); } IndexedTimer z(t,m,k);
#else
#define TIMER(t,s,l)
#define INDEXED_TIMER(t,s,l,n,m,k)
#endif
#endif
