//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
// "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/testIDLList.C,v 1.1 2009/08/20 17:33:33 johnt Exp $"
#include <stdlib.h>
#include "Debug.H"
#include "IDLListNode.H"
#include "IDLList.H"
#include "MemoryDebugger.H"
#include "MemoryDebugger.H"
#include "IDLList.C"

class IntNode : public IDLListNode {
  private:
    int data;
  public:
    IntNode(int i) : data(i) {;}
    int value() const { return data; }
    virtual void printOn(ostream &os=cerr) const {
      os << "IntNode: this = " << this
         << " prev = " << prev() << " next = " << next() 
         << " value = " << data;
    }
};

template class IDLList<IntNode>;
template ostream& operator<<(ostream&,const IDLList<IntNode> &);

int main( int /*argc*/, char** /*argv*/) {
  IDLList<IntNode> idllist;
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  CHECK_TEST(idllist.ok())
  CHECK_TEST(idllist.empty())
  CHECK_TEST(!idllist.notEmpty())
  cout << "\tconstructor ok at line " << __LINE__ << endl;

  for (int i=0;i<10;i++) {
    IntNode *ii=OPERATOR_NEW IntNode(i);
    idllist.append(ii);
  }
  CHECK_TEST(idllist.ok())
  CHECK_TEST(!idllist.empty())
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  CHECK_SAME(idllist.length(),10)
  cout << "\tappend ok at line " << __LINE__ << endl;

  IntNode *t=(idllist.next(idllist.first()));
  CHECK_TEST(idllist.owns(t))

  IntNode *td=idllist.delAfter(t);
  CHECK_SAME(td->value(),2)
  CHECK_SAME(t->value(),1)
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  delete td;
  td=0;

  CHECK_TEST(idllist.ok())
  while (idllist.notEmpty()) {
    td=idllist.delAfter(0);
//  cout << "deleted entry = " << *td << " at line " << __LINE__ <<endl;
    delete td;
    td=0;
  }
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  cout << "\tdelAfter ok at line " << __LINE__ << endl;

  for (int i=0;i<10;i++) {
    IntNode *ii=OPERATOR_NEW IntNode(i);
    idllist.append(ii);
  }
//cout << "idllist at line " << __LINE__ << endl;
//cout << "at line " << __LINE__ << endl;
//cout << idllist;
  t=(idllist.prev(idllist.last()));
  td=idllist.delBefore(t);
  CHECK_SAME(td->value(),7)
  CHECK_SAME(t->value(),8)
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  delete td;
  td=0;

  CHECK_TEST(idllist.ok())
  while (idllist.notEmpty()) {
    td=idllist.delBefore(0);
//  cout << "deleted entry = " << *td << " at line " << __LINE__ <<endl;
    delete td;
    td=0;
  }
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  cout << "\tdelBefore ok at line " << __LINE__ << endl;

  for (int i=0;i<10;i++) {
    IntNode *ii=OPERATOR_NEW IntNode(i);
    idllist.append(ii);
  }
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  t=idllist.next(idllist.first());
  td=idllist.delAndNext(t);
  CHECK_SAME(td->value(),1)
  CHECK_SAME(t->value(),2)
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  delete td;
  td=0;

  CHECK_TEST(idllist.ok())
  t=idllist.first();
  while (idllist.notEmpty()) {
    td=idllist.delAndNext(t);
//  cout << "deleted entry = " << *td << " at line " << __LINE__ <<endl;
    delete td;
    td=0;
  }
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  cout << "\tdelAndNext ok at line " << __LINE__ << endl;
  

  for (int i=0;i<10;i++) {
    IntNode *ii=OPERATOR_NEW IntNode(i);
    idllist.append(ii);
  }
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  t=idllist.next(idllist.first());
  td=idllist.delAndPrev(t);
  CHECK_SAME(td->value(),1)
  CHECK_SAME(t->value(),0)
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  delete td;
  td=0;

  CHECK_TEST(idllist.ok())
  t=idllist.last();
  while (idllist.notEmpty()) {
    td=idllist.delAndPrev(t);
//  cout << "deleted entry = " << *td << " at line " << __LINE__<< endl;
    delete td;
    td=0;
  }
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  cout << "\tdelAndPrev ok at line " << __LINE__ << endl;

  for (int i=0;i<10;i++) {
    IntNode *ii=OPERATOR_NEW IntNode(i);
    idllist.prepend(ii);
  }
  CHECK_SAME(idllist.length(),10)
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
//cout << "\tprepend ok at line " << __LINE__ << endl;

  IntNode *im1=OPERATOR_NEW IntNode(-1);
  IntNode *im2=OPERATOR_NEW IntNode(-2);
  idllist.insAfter(0,im1);
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  CHECK_SAME(idllist.length(),11)
  CHECK_SAME(idllist.first()->value(),-1)

  idllist.insAfter(idllist.first(),im2);
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  CHECK_SAME(idllist.length(),12)
  CHECK_SAME(idllist.next(idllist.first())->value(),-2)
  cout << "\tinsAfter ok at line " << __LINE__ << endl;

  IntNode *jm1=OPERATOR_NEW IntNode(-1);
  IntNode *jm2=OPERATOR_NEW IntNode(-2);
  idllist.insBefore(0,jm1);
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  CHECK_SAME(idllist.length(),13)
  CHECK_SAME(idllist.last()->value(),-1)

  idllist.insBefore(idllist.last(),jm2);
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  CHECK_SAME(idllist.length(),14)
  CHECK_SAME(idllist.prev(idllist.last())->value(),-2)
  cout << "\tinsBefore ok at line " << __LINE__ << endl;

  t=idllist.first();
  IDLList<IntNode> idllist2;
  idllist.join(idllist2);
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  CHECK_TEST(idllist.owns(t))
  CHECK_TEST(!idllist2.owns(t))

  idllist2.join(idllist);
//cout << "idllist2 at line " << __LINE__ << endl;
//cout << idllist2;
  CHECK_TEST(!idllist.owns(t))
  CHECK_TEST(idllist2.owns(t))

//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;

//cout << "idllist2 at line " << __LINE__ << endl;
//cout << idllist2;

  for (int i=0;i<10;i++) {
    IntNode *ii=OPERATOR_NEW IntNode(i);
    idllist.prepend(ii);
  }
  idllist.join(idllist2);
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  CHECK_SAME(idllist.length(),24)
  CHECK_TEST(idllist2.empty())
  cout << "\tjoin ok at line " << __LINE__ << endl;

  for (int i=0;i<10;i++) {
    IntNode *ii=OPERATOR_NEW IntNode(i);
    idllist2.prepend(ii);
  }
//cout << "idllist2 at line " << __LINE__ << endl;
//cout << idllist2;
  idllist.joinAtFront(idllist2);
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
//cout << "idllist2 at line " << __LINE__ << endl;
//cout << idllist2;
  CHECK_SAME(idllist.length(),34)
  CHECK_TEST(idllist2.empty())
  cout << "\tjoinAtFront ok at line " << __LINE__ << endl;

  IntNode *km1=OPERATOR_NEW IntNode(-10);
//cout << "km1 = " << km1 << "  at line " << __LINE__ << endl;
  IntNode *rl=idllist.last();
  idllist.replace(rl,km1);
//cout << "rl = " << rl << "  at line " << __LINE__ << endl;
  delete rl;
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  CHECK_SAME(idllist.length(),34)
  CHECK_SAME(idllist.last()->value(),-10)
  CHECK_SAME(km1->value(),-10)
  CHECK_SAME(rl->value(),-1)

  IntNode *km2=OPERATOR_NEW IntNode(-2);
  IntNode *rf=idllist.first();
  idllist.replace(rf,km2);
//cout << "idllist at line " << __LINE__ << endl;
//cout << "rf = " << rf << "  at line " << __LINE__ << endl;
  delete rf;
//cout << idllist;
  CHECK_SAME(idllist.length(),34)
  CHECK_SAME(km2->value(),-2)
  CHECK_SAME(rf->value(),9)
  cout << "\treplace ok at line " << __LINE__ << endl;

  rl=idllist.last();
  idllist.remove(rl);
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
//cout << "rl = " << rl << "  at line " << __LINE__ << endl;
  delete rl;
  CHECK_SAME(idllist.length(),33)
  CHECK_SAME(idllist.last()->value(),-2)

  rf=idllist.first();
  idllist.remove(rf);
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
//cout << "rf = " << rf << "  at line " << __LINE__ << endl;
  delete rf;
  CHECK_SAME(idllist.length(),32)
  CHECK_SAME(idllist.first()->value(),8)
  cout << "\tremove ok" << endl;

  idllist.reverse();
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  CHECK_SAME(idllist.length(),32)
  CHECK_SAME(idllist.first()->value(),-2)
  CHECK_SAME(idllist.last()->value(),8)
  cout << "\treverse ok" << endl;

  while (idllist.notEmpty()) {
    td=idllist.delAfter(0);
//  cout << "deleted data = " << *td << " at line " << __LINE__ << endl;
    delete td;
    td=0;
  }
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  CHECK_TEST(idllist.empty())

  while (idllist2.notEmpty()) {
    td=idllist2.delAfter(0);
//  cout << "deleted data = " << *td << " at line " << __LINE__ << endl;
    delete td;
    td=0;
  }
//cout << "idllist2 at line " << __LINE__ << endl;
//cout << idllist2;
  CHECK_TEST(idllist2.empty())

  for (int i=0;i<10;i++) {
    IntNode *ii=OPERATOR_NEW IntNode(i);
    idllist.prepend(ii);
  }
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  idllist.swapWithNext(idllist.first());
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  CHECK_SAME(idllist.length(),10)
  CHECK_SAME(idllist.first()->value(),8)
  cout << "\tswapWithNext ok at line " << __LINE__ << endl;

  idllist.swapWithPrev(idllist.last());
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  CHECK_SAME(idllist.length(),10)
  CHECK_SAME(idllist.last()->value(),1)
  cout << "\tswapWithPrev ok at line " << __LINE__ << endl;

  idllist.swap(idllist.first(),idllist.last());
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;
  CHECK_SAME(idllist.length(),10)
  CHECK_SAME(idllist.first()->value(),1)
  CHECK_SAME(idllist.last()->value(),8)
  cout << "\tswap ok at line " << __LINE__ << endl;

  CHECK_TEST(idllist.progress(idllist.first(),idllist.last()))
  CHECK_TEST(!idllist.progress(idllist.last(),idllist.first()))
  CHECK_TEST(!idllist.regress(idllist.first(),idllist.last()))
  CHECK_TEST(idllist.regress(idllist.last(),idllist.first()))
  cout << "\tprogress, regress ok at line " << __LINE__ << endl;

  while (idllist.notEmpty()) {
    td=idllist.delAfter(0);
    delete td;
    td=0;
  }
//cout << "idllist at line " << __LINE__ << endl;
//cout << idllist;

//cout << "idllist2 at line " << __LINE__ << endl;
//cout << idllist2;

  return 1;
}
