//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
// "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/testISLList.C,v 1.1 2009/08/20 17:33:33 johnt Exp $"
#include <stdlib.h>
#include "Debug.H"
#include "ISLListNode.H"
#include "ISLList.H"
#include "MemoryDebugger.H"
#include "ISLList.C"

class IntNode : public ISLListNode {
  private:
    int data;
  public:
    IntNode(int i) : data(i) {;}
    int value() const { return data; }
    virtual void printOn(ostream &os=cerr) const {
      os << "IntNode: this = " << this
         << " after = " << after() << " value = " << data;
    }
};

template class ISLList<IntNode>;
template ostream& operator<<(ostream&,const ISLList<IntNode> &);

int main( int /*argc*/, char** /*argv*/) {
  ISLList<IntNode> isllist;
//cout << isllist;
  CHECK_TEST(isllist.ok())
  CHECK_TEST(isllist.empty())
  CHECK_TEST(!isllist.notEmpty())
  cout << "\tconstructor ok" << endl;

  for (int i=0;i<10;i++) {
    IntNode *ii=OPERATOR_NEW IntNode(i);
    isllist.append(*ii);
  }
//cout << isllist;
  CHECK_TEST(isllist.ok())
  CHECK_TEST(!isllist.empty())
  CHECK_SAME(isllist.length(),10)
  cout << "\tappend ok" << endl;

  IntNode *t=isllist.next(isllist.first());
  CHECK_TEST(isllist.owns(t))

  IntNode *td;
  td=isllist.delAfter(t);
//cout << isllist;
  CHECK_TEST(isllist.ok())
  CHECK_SAME(isllist.length(),9)
  CHECK_SAME(td->value(),2)
  CHECK_SAME(t->value(),1)
  delete td;
  td=0;

  while (isllist.notEmpty()) {
    td=isllist.delAfter(0);
//  cout << "deleted entry = " << *td << endl;
    delete td;
    td=0;
  }
//cout << isllist;
  CHECK_TEST(isllist.empty())
  CHECK_TEST(isllist.ok())
  cout << "\tdelAfter ok" << endl;

  for (int i=0;i<10;i++) {
    IntNode *ii=OPERATOR_NEW IntNode(i);
    isllist.prepend(*ii);
  }
//cout << isllist;
  CHECK_TEST(isllist.ok())
  CHECK_SAME(isllist.length(),10)
  cout << "\tprepend ok" << endl;

  IntNode *im1=OPERATOR_NEW IntNode(-1);
  IntNode *im2=OPERATOR_NEW IntNode(-2);
  isllist.insAfter(0,*im1);
//cout << isllist;
  CHECK_SAME(isllist.length(),11)
  CHECK_SAME(isllist.first()->value(),-1)

  isllist.insAfter(isllist.first(),*im2);
//cout << isllist;
  CHECK_SAME(isllist.length(),12)
  CHECK_SAME(isllist.next(isllist.first())->value(),-2)
  cout << "\tinsAfter ok" << endl;

  t=isllist.first();
  ISLList<IntNode> isllist2;
  isllist.join(isllist2);
//cout << isllist;
  CHECK_TEST(isllist.ok())
  CHECK_SAME(isllist.length(),12)
  CHECK_TEST(isllist2.ok())
  CHECK_TEST(isllist2.empty())
  CHECK_TEST(isllist.owns(t))
  CHECK_TEST(!isllist2.owns(t))

  isllist2.join(isllist);
//cout << "\tisllist2:" << endl;
//cout << isllist2;
//cout << "isllist:" << endl;
//cout << isllist;
  CHECK_TEST(isllist2.ok())
  CHECK_SAME(isllist2.length(),12)
  CHECK_TEST(!isllist2.empty())
  CHECK_TEST(isllist.ok())
  CHECK_TEST(isllist.empty())
  CHECK_TEST(!isllist.owns(t))
  CHECK_TEST(isllist2.owns(t))

  for (int i=0;i<10;i++) {
    IntNode *ii=OPERATOR_NEW IntNode(i);
    isllist.prepend(*ii);
  }
  isllist.join(isllist2);
//cout << "isllist:" << endl;
//cout << isllist;
//cout << "\tisllist2:" << endl;
//cout << isllist2;
  CHECK_TEST(isllist.ok())
  CHECK_SAME(isllist.length(),22)
  CHECK_TEST(isllist2.ok())
  CHECK_TEST(isllist2.empty())
  cout << "\tjoin ok" << endl;

//cout << "\tclearing isllist" << endl;
  while (isllist.notEmpty()) {
    td=isllist.delAfter(0);
//  cout << "deleted data = " << *td << endl;
    delete td;
    td=0;
  }
//cout << isllist;

//cout << "\tclearing isllist2" << endl;
  while (isllist2.notEmpty()) {
    td=isllist2.delAfter(0);
//  cout << "deleted data = " << *td << endl;
    delete td;
    td=0;
  }
//cout << isllist2;

  return 1;
}
