//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
// "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/testNIDLList.C,v 1.1 2009/08/20 17:33:33 johnt Exp $"
#include <stdlib.h>
#include "Debug.H"
#include "Types.H"
#include "Tracer.H"

#include "NIDLList.C"

//#define SORTING
#ifndef SORTING
# define node_name NIDLListNode<int>
# define list_name NIDLList<int>
#else
# define node_name NIDLValueListNode<int,int>
# define list_name NIDLValueList<int,int>
  template class NIDLListNode<int>;
  template ostream& operator<<(ostream&,const NIDLListNode<int>&);
  template class NIDLList<int>;
#endif

template class node_name;
template class list_name;
template ostream& operator<<(ostream&,const node_name&);
template ostream& operator<<(ostream&,const list_name&);

int main(int /*argc*/,char** /*argv*/) {
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"constructors, append, delAfter");
      list_name nidllist;
      cout << nidllist;
      CHECK_TEST(nidllist.ok())
      CHECK_TEST(nidllist.empty())
      CHECK_TEST(!nidllist.notEmpty())

      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
#ifndef SORTING
        nidllist.append(ii);
#else
        nidllist.append(ii,i);
#endif
      }
      CHECK_TEST(nidllist.ok())
      CHECK_TEST(!nidllist.empty())
      CHECK_SAME(nidllist.length(),10)
//    cout << nidllist;

      node_name *t=(nidllist.next(nidllist.first()));
      CHECK_TEST(nidllist.owns(t))

      int *td=nidllist.delAfter(t);
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),9)
      CHECK_SAME(*td,2)
      delete td;
      td=0;

      while (nidllist.notEmpty()) {
        td=nidllist.delAfter(0);
        delete td;
        td=0;
      }
      CHECK_TEST(nidllist.ok())
      CHECK_TEST(nidllist.empty())
    }
    cout << "\tNIDLList constructors, append, delAfter ok" << endl;
  }

  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
      NIDLList<int>::Iterator q;
      CHECK_TEST(!q)

      list_name nidllist;
      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
#ifndef SORTING
        nidllist.append(ii);
#else
        nidllist.append(ii,i);
#endif
      }
      int i=0;
      for (NIDLList<int>::Iterator p(nidllist);p;p++,i++) {
        CHECK_SAME(*p,i);
        CHECK_TEST(p)
        CHECK_TEST(p!=q)
        CHECK_TEST(!(p==q))
      }

      NIDLList<int>::Iterator p(nidllist); p.fastForward();
      for (i=9;p;p--,i--) CHECK_SAME(*p,i);

      for (p.rewind(),i=0;p;p++,i++) CHECK_SAME(*p,i);

//    cf BoxList::foundIntersection :
      for (p.rewind(),i=0;p;p++,i++) {
        NIDLList<int>::Iterator q(p); q++;
        for (int j=i+1;q;q++,j++) CHECK_SAME(*q,j);
      }

      for (NIDLList<int>::Iterator pp(nidllist);pp;pp++) {
        delete nidllist.replace(pp,OPERATOR_NEW int(100));
      }
      for (p.rewind();p;p++,i++) CHECK_SAME(*p,100);

      while (nidllist.notEmpty()) delete nidllist.delAfter(0);
    }
    cout << "\tNIDLList Iterator ok" << endl;
  }

  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"delBefore");
      list_name nidllist;
      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
#ifndef SORTING
        nidllist.append(ii);
#else
        nidllist.append(ii,i);
#endif
      }
      node_name *t=nidllist.prev(nidllist.last());
      int *td=nidllist.delBefore(t);
      CHECK_TEST(nidllist.ok())
      CHECK_TEST(!nidllist.empty())
      CHECK_SAME(nidllist.length(),9)
      CHECK_SAME(*td,7)
//    cout << "\tdelBefore argument = " << *t << endl;
      delete td;
      td=0;

      while (nidllist.notEmpty()) {
        td=nidllist.delBefore(0);
        delete td;
        td=0;
      }
      CHECK_TEST(nidllist.ok())
      CHECK_TEST(nidllist.empty())
    }
    cout << "\tNIDLList delBefore ok" << endl;
  }

  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"delAndNext");
      list_name nidllist;

      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
#ifndef SORTING
        nidllist.append(ii);
#else
        nidllist.append(ii,i);
#endif
      }

      node_name *t=nidllist.next(nidllist.first());
      int *td=nidllist.delAndNext(t);
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),9)
      CHECK_SAME(*td,1)
      CHECK_SAME(*t->getData(),2)
      delete td;
      td=0;

      t=nidllist.first();
      while (nidllist.notEmpty()) {
        td=nidllist.delAndNext(t);
        delete td;
        td=0;
      }
      CHECK_TEST(nidllist.ok())
      CHECK_TEST(nidllist.empty())
    }
    cout << "\tNIDLList delAndNext ok" << endl;
  }

  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"delAndPrev");
      list_name nidllist;

      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
#ifndef SORTING
        nidllist.append(ii);
#else
        nidllist.append(ii,i);
#endif
      }
      node_name *t=nidllist.next(nidllist.first());
      int *td=nidllist.delAndPrev(t);
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),9)
      CHECK_SAME(*td,1)
      CHECK_SAME(*t->getData(),0)
      delete td;
      td=0;

      t=nidllist.last();
      while (nidllist.notEmpty()) {
        td=nidllist.delAndPrev(t);
        delete td;
        td=0;
      }
      CHECK_TEST(nidllist.ok())
      CHECK_TEST(nidllist.empty())
    }
    cout << "\tNIDLList delAndPrev ok" << endl;
  }

  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"prepend, insAfter, insBefore, join, joinAtFront");
      list_name nidllist;

      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
#ifndef SORTING
        nidllist.prepend(ii);
#else
        nidllist.prepend(ii,i);
#endif
      }
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),10)
      CHECK_TEST(!nidllist.empty())

      int *im1=OPERATOR_NEW int(-1);
      int *im2=OPERATOR_NEW int(-2);
#ifndef SORTING
      nidllist.insAfter(0,im1);
#else
      nidllist.insAfter(0,im1,-1);
#endif
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),11)
      CHECK_SAME(*nidllist.first()->getData(),-1)
      CHECK_SAME(*nidllist.next(nidllist.first())->getData(),9)

#ifndef SORTING
      nidllist.insAfter(nidllist.first(),im2);
#else
      nidllist.insAfter(nidllist.first(),im2,-2);
#endif
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),12)
      CHECK_SAME(*nidllist.next(nidllist.first())->getData(),-2)
      CHECK_SAME(*nidllist.next(nidllist.next(nidllist.first()))->getData(),9)

      int *jm1=OPERATOR_NEW int(-1);
      int *jm2=OPERATOR_NEW int(-2);
#ifndef SORTING
      nidllist.insBefore(0,jm1);
#else
      nidllist.insBefore(0,jm1,-1);
#endif
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),13)
      CHECK_SAME(*nidllist.last()->getData(),-1)
      CHECK_SAME(*nidllist.prev(nidllist.last())->getData(),0)

#ifndef SORTING
      nidllist.insBefore(nidllist.last(),jm2);
#else
      nidllist.insBefore(nidllist.last(),jm2,-2);
#endif
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),14)
      CHECK_SAME(*nidllist.last()->getData(),-1)
      CHECK_SAME(*nidllist.prev(nidllist.last())->getData(),-2)

      node_name *t=nidllist.first();
      list_name nidllist2;
      nidllist.join(nidllist2);
      CHECK_TEST(nidllist2.ok())
      CHECK_TEST(nidllist2.empty())
      CHECK_SAME(nidllist2.length(),0)
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),14)
      CHECK_TEST(nidllist.owns(t))
      CHECK_TEST(!nidllist2.owns(t))

      nidllist2.join(nidllist);
      CHECK_TEST(nidllist.ok())
      CHECK_TEST(nidllist.empty())
      CHECK_TEST(nidllist2.ok())
      CHECK_SAME(nidllist2.length(),14)
      CHECK_TEST(nidllist2.owns(t))
      CHECK_TEST(!nidllist.owns(t))

      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
#ifndef SORTING
        nidllist.prepend(ii);
#else
        nidllist.prepend(ii,i);
#endif
      }
      nidllist.join(nidllist2);
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),24)
      CHECK_TEST(nidllist2.ok())
      CHECK_TEST(nidllist2.empty())

      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
#ifndef SORTING
        nidllist2.prepend(ii);
#else
        nidllist2.prepend(ii,i);
#endif
      }
      nidllist.joinAtFront(nidllist2);
      CHECK_TEST(nidllist2.ok())
      CHECK_TEST(nidllist2.empty())
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),34)

      t=nidllist.last();
      while (nidllist.notEmpty()) {
        int *td=nidllist.delAndPrev(t);
        delete td;
        td=0;
      }
    }
    cout << "\tNIDLList prepend, insAfter, insBefore, join, "
         << "joinAtFront ok" << endl;
  }

#ifndef SORTING
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
      list_name nidllist;

      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
        nidllist.append(ii);
      }

      NIDLList<int>::Iterator p(nidllist);
      nidllist.insAfter(p,OPERATOR_NEW int(-1));
      p++;
      CHECK_SAME( *p, -1)

      nidllist.insBefore(p,OPERATOR_NEW int(-2));
      p--;
      CHECK_SAME( *p, -2)

      int *d=nidllist.delAndNext(p);
      CHECK_SAME( *d, -2)
      delete d;

      d=nidllist.delAndPrev(p);
      CHECK_SAME( *d, -1)
      CHECK_TEST( p==NIDLList<int>::Iterator(nidllist) )
      delete d;

      d=nidllist.remove(p);
      CHECK_SAME( *d, 0)
      delete d;

      while (nidllist.notEmpty()) delete nidllist.delAfter(0);
    }
    cout << "\tNIDLList Iterator insAfter, insBefore ok" << endl;
  }
#endif

  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"replace, remove");
      list_name nidllist;

      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
#ifndef SORTING
        nidllist.prepend(ii);
#else
        nidllist.prepend(ii,i);
#endif
      }

      int *km1=OPERATOR_NEW int(-1);
      node_name *rl=nidllist.last();
#ifndef SORTING
      int *j=nidllist.replace(rl,km1);
#else
      int *j=nidllist.replace(rl,km1,-1);
#endif
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),10)
      CHECK_SAME(*nidllist.last()->getData(),-1)
      CHECK_SAME(*j,0)
      delete j;
      j=0;

      int *km2=OPERATOR_NEW int(-2);
      node_name *rf=nidllist.first();
#ifndef SORTING
      j=nidllist.replace(rf,km2);
#else
      j=nidllist.replace(rf,km2,-2);
#endif
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),10)
      CHECK_SAME(*nidllist.first()->getData(),-2)
      CHECK_SAME(*j,9)
      delete j;
      j=0;

      j=nidllist.remove(nidllist.last());
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),9)
      CHECK_SAME(*nidllist.last()->getData(),1)
      CHECK_SAME(*j,-1)
      delete j;
      j=0;

      j=nidllist.remove(nidllist.first());
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),8)
      CHECK_SAME(*nidllist.first()->getData(),8)
      CHECK_SAME(*j,-2)
      delete j;
      j=0;

      node_name *t=nidllist.last();
      while (nidllist.notEmpty()) {
        int *td=nidllist.delAndPrev(t);
        delete td;
        td=0;
      }
    }
    cout << "\tNIDLList replace, remove ok" << endl;
  }

  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"reverse");
      list_name nidllist;

      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
#ifndef SORTING
        nidllist.prepend(ii);
#else
        nidllist.prepend(ii,i);
#endif
      }

      nidllist.reverse();
      CHECK_TEST(nidllist.ok())
      CHECK_SAME(nidllist.length(),10)
      int i=0;
      for (node_name *t=nidllist.first();t;t=nidllist.next(t),i++) {
        CHECK_SAME(*t->getData(),i)
      }

      while (nidllist.notEmpty()) {
        int *td=nidllist.delAfter(0);
        delete td;
        td=0;
      }
    }
    cout << "\tNIDLList reverse ok" << endl;
  }

  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"swap, swapWithNext, swapWithPrev, progress, regress");
      list_name nidllist;

      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
#ifndef SORTING
        nidllist.prepend(ii);
#else
        nidllist.prepend(ii,i);
#endif
      }
      nidllist.swapWithNext(nidllist.first());
      CHECK_SAME(*nidllist.first()->getData(),8)
      CHECK_SAME(*nidllist.next(nidllist.first())->getData(),9)

      nidllist.swapWithNext(nidllist.last()); // error
      CHECK_SAME(*nidllist.first()->getData(),0)
      CHECK_SAME(*nidllist.last()->getData(),8)

      nidllist.swapWithPrev(nidllist.last());
      CHECK_SAME(*nidllist.last()->getData(),1)
      CHECK_SAME(*nidllist.prev(nidllist.last())->getData(),8)

      nidllist.swapWithPrev(nidllist.first());
      CHECK_SAME(*nidllist.first()->getData(),1)
      CHECK_SAME(*nidllist.last()->getData(),0)

      nidllist.swap(nidllist.first(),nidllist.last());
      CHECK_SAME(*nidllist.first()->getData(),0)
      CHECK_SAME(*nidllist.last()->getData(),1)

      CHECK_TEST(nidllist.progress(nidllist.first(),nidllist.last()))
      CHECK_TEST(!nidllist.progress(nidllist.last(),nidllist.first()))
      CHECK_TEST(!nidllist.regress(nidllist.first(),nidllist.last()))
      CHECK_TEST(nidllist.regress(nidllist.last(),nidllist.first()))

      while (nidllist.notEmpty()) {
        int *td=nidllist.delAfter(0);
        delete td;
        td=0;
      }
    }
    cout << "\tNIDLList swap, swapWithNext, swapWithPrev, progress, "
         << "regress ok" << endl;
  }

#ifdef SORTING
  int nv=0;
  cout << "\tenter number of values for sorts" << endl;
  cin >> nv;
  cout << "\tnumber values = " << nv << endl;
  unsigned int seed=0;
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"selectionSort");
      list_name nidllist;

      srand(seed);
      for (int i=0;i<nv;i++) {
        int *ii=OPERATOR_NEW int(i);
        int r=int( double(nv)*rand() / (RAND_MAX+1.));
        nidllist.append(ii,r);
      }
      nidllist.selectionSort();
      CHECK_TEST(nidllist.ok())

      int v=nidllist.first()->getValue();
      while (nidllist.notEmpty()) {
        int *td=nidllist.delAfter(0); 
        if (nidllist.notEmpty()) {
          CHECK_TEST(nidllist.first()->getValue()<=v)
          v=nidllist.first()->getValue();
        }
        delete td;
        td=0;
      }
    }
    cout << "\tNIDLList selectionSort ok" << endl;
  }

  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"insertionSort");
      list_name nidllist;

      srand(seed);
      for (int i=0;i<nv;i++) {
        int *ii=OPERATOR_NEW int(i);
        int r=int( double(nv)*rand() / (RAND_MAX+1.));
        nidllist.append(ii,r);
      }
      nidllist.insertionSort();
      CHECK_TEST(nidllist.ok())

      int v=nidllist.first()->getValue();
      while (nidllist.notEmpty()) {
        int *td=nidllist.delAfter(0); 
        if (nidllist.notEmpty()) {
          CHECK_TEST(nidllist.first()->getValue()<=v)
          v=nidllist.first()->getValue();
        }
        delete td; 
        td=0;
      }
    }
    cout << "\tNIDLList insertionSort ok" << endl;
  }

  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"bubbleSort");
      list_name nidllist;

      srand(seed);
      for (int i=0;i<nv;i++) {
        int *ii=OPERATOR_NEW int(i);
        int r=int( double(nv)*rand() / (RAND_MAX+1.));
        nidllist.append(ii,r);
      }
      nidllist.bubbleSort();
      CHECK_TEST(nidllist.ok())

      int v=nidllist.first()->getValue();
      while (nidllist.notEmpty()) {
        int *td=nidllist.delAfter(0); 
        if (nidllist.notEmpty()) {
          CHECK_TEST(nidllist.first()->getValue()<=v)
          v=nidllist.first()->getValue();
        }
        delete td; 
        td=0;
      }
    }
    cout << "\tNIDLList bubbleSort ok" << endl;
  }

  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"quickSort");
      list_name nidllist;

      srand(seed);
      for (int i=0;i<nv;i++) {
        int *ii=OPERATOR_NEW int(i);
        int r=int( double(nv)*rand() / (RAND_MAX+1.));
        nidllist.append(ii,r);
      }
      nidllist.quickSort(nidllist.first(),nidllist.last());
      CHECK_TEST(nidllist.ok())

      int v=nidllist.first()->getValue();
      while (nidllist.notEmpty()) {
        int *td=nidllist.delAfter(0); 
        if (nidllist.notEmpty()) {
          CHECK_TEST(nidllist.first()->getValue()<=v)
          v=nidllist.first()->getValue();
        }
        delete td; 
        td=0;
      }
    }
    cout << "\tNIDLList quickSort ok" << endl;
  }

  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"balance");
      list_name nidllist;
      srand(seed);
      for (int i=0;i<nv;i++) {
        int *ii=OPERATOR_NEW int(i);
        int r=int( double(nv)*rand() / (RAND_MAX+1.));
        nidllist.append(ii,r);
      }
      nidllist.sort();

      int nb=0;
      cout << "\tenter number of bins for balance" << endl;
      cin >> nb;
      cout << "\tnumber bins = " << nb << endl;

      int v=nidllist.getValue();
      cout << "\n\tnidllist.value = " << v << endl;
      list_name *nidllist_array=OPERATOR_NEW_BRACKET(list_name,nb);
      nidllist.balance(nb,nidllist_array);
      for (int i=0;i<nb;i++) {
        cout << "\tlist number[" << i << "].value = " 
             << nidllist_array[i].getValue() << endl;
        v-=nidllist_array[i].getValue();
        while (nidllist_array[i].notEmpty()) {
          int *td=nidllist_array[i].delAfter(0); delete td; 
          td=0;
        }
      }
      CHECK_SAME(v,0)
      delete [] nidllist_array;
    }
  }
#endif

  return 1;
}
