//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
// "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/testNISLList.C,v 1.1 2009/08/20 17:33:33 johnt Exp $"
#include <stdlib.h>
#include "Debug.H"
#include "NISLList.C"

template class NISLListNode<int>;
template ostream& operator<<(ostream&,const NISLListNode<int> &);
template class NISLList<int>;
template ostream& operator<<(ostream&,const NISLList<int> &);

int main( int /*argc*/, char** /*argv*/) {
  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"constructors, append, delAfter");
      NISLList<int> nisllist;
//    cout << nisllist;
      CHECK_TEST(nisllist.ok())
      CHECK_TEST(nisllist.empty())
      CHECK_TEST(!nisllist.notEmpty())

      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
        nisllist.append(ii);
      }
//    cout << nisllist;
      CHECK_TEST(nisllist.ok())
      CHECK_TEST(!nisllist.empty())
      CHECK_TEST(nisllist.notEmpty())
      CHECK_SAME(nisllist.length(),10)

      NISLListNode<int> *t=(nisllist.next(nisllist.first()));
      CHECK_TEST(nisllist.owns(t))

      int *td;
      td=nisllist.delAfter(t);
      CHECK_TEST(nisllist.ok())
      CHECK_SAME(nisllist.length(),9)
      CHECK_SAME(*td,2)
      delete td;
      td=0;

      while (nisllist.notEmpty()) {
        td=nisllist.delAfter(0);
        delete td;
        td=0;
      }
      CHECK_TEST(nisllist.ok())
      CHECK_TEST(nisllist.empty())
    }
    cout << "\tconstructors, append, delAfter ok" << endl;
  }

  {
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(tr,"prepend, insAfter, join");
      NISLList<int> nisllist;

      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
        nisllist.prepend(ii);
      }
      CHECK_TEST(nisllist.ok())
      CHECK_TEST(!nisllist.empty())
      CHECK_SAME(nisllist.length(),10)

      int *im1=OPERATOR_NEW int(-1);
      int *im2=OPERATOR_NEW int(-2);
      nisllist.insAfter(0,im1);
      CHECK_TEST(nisllist.ok())
      CHECK_SAME(nisllist.length(),11)
      CHECK_SAME(*nisllist.first()->getData(),-1)

      nisllist.insAfter(nisllist.first(),im2);
      CHECK_TEST(nisllist.ok())
      CHECK_SAME(nisllist.length(),12)
      CHECK_SAME(*nisllist.next(nisllist.first())->getData(),-2)

      NISLListNode<int> *t=nisllist.first();
      NISLList<int> nisllist2;
      nisllist.join(nisllist2);
      CHECK_TEST(nisllist2.ok())
      CHECK_TEST(nisllist2.empty())
      CHECK_SAME(nisllist2.length(),0)
      CHECK_TEST(!nisllist2.owns(t))
      CHECK_TEST(nisllist.ok())
      CHECK_SAME(nisllist.length(),12)
      CHECK_TEST(nisllist.owns(t))

      nisllist2.join(nisllist);
      CHECK_TEST(nisllist2.ok())
      CHECK_TEST(!nisllist2.empty())
      CHECK_SAME(nisllist2.length(),12)
      CHECK_TEST(nisllist2.owns(t))
      CHECK_TEST(nisllist.ok())
      CHECK_SAME(nisllist.length(),0)
      CHECK_TEST(!nisllist.owns(t))

      for (int i=0;i<10;i++) {
        int *ii=OPERATOR_NEW int(i);
        nisllist.prepend(ii);
      }
      nisllist.join(nisllist2);
      CHECK_TEST(nisllist2.ok())
      CHECK_TEST(nisllist2.empty())
      CHECK_SAME(nisllist2.length(),0)
      CHECK_TEST(!nisllist2.owns(t))
      CHECK_TEST(nisllist.ok())
      CHECK_SAME(nisllist.length(),22)
      CHECK_TEST(nisllist.owns(t))

      while (nisllist.notEmpty()) {
        int *td=nisllist.delAfter(0);
        delete td;
        td=0;
      }

      while (nisllist2.notEmpty()) {
        int *td=nisllist2.delAfter(0);
        delete td;
        td=0;
      }
    }
    cout << "\tprepend, insAfter, join ok" << endl;
  }

  return 1;
}
