//**********************************************************************
// Copyright 2006 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************
// "$Header: /home/faculty/johnt/cvs/deal_new/memdebug/testNumPtr.C,v 1.1 2009/08/20 17:33:33 johnt Exp $"
#include <stdlib.h>
#include "Debug.H"
#include "NumPtr.H"
#include "MemoryDebugger.H"
#include "NumPtr.C"

template class NumPtrBase<void,UINT>;
typedef double pointer_type;
INSTANTIATE_NUMPTR(pointer_type)
INSTANTIATE_NUMPTR(char)

INSTANTIATE_NUMPTR(int)
template class NumPtrWithValue<pointer_type>;

INSTANTIATE_NUMPTR(int*)
INSTANTIATE_NUMPTR(pointer_type*)
template class NumPtrWithCompare<pointer_type*,int*>;

int main( int /*argc*/, char** /*argv*/) {
  { 
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    {
//    TRACER_CALL(t,"testNumPtr constructors and inequalities");
      NumPtrBase<void,UINT> ptr0;
//    cout << "\tNumPtrBase() :"; ptr0.printOn(cout);

      char *data=OPERATOR_NEW_BRACKET(char,5);
      NumPtrBase<void,UINT> p1(5,data);
//    cout << "\tNumPtrBase(5,data) :"; p1.printOn(cout);
      CHECK_SAME(!p1,FALSE);
      CHECK_SAME(p1.getNumber(),5)

      CHECK_SAME(p1<data,FALSE)
      CHECK_SAME(p1<p1,FALSE)
      CHECK_SAME(p1>data,FALSE)
      CHECK_SAME(p1>p1,FALSE)
      CHECK_SAME(p1<=data,TRUE)
      CHECK_SAME(p1<=p1,TRUE)
      CHECK_SAME(p1>=data,TRUE)
      CHECK_SAME(p1>=p1,TRUE)
      CHECK_SAME(p1==p1,TRUE)
      CHECK_SAME(p1!=p1,FALSE)

      NumPtrBase<void,UINT> p4(p1);
//    cout << "\nNumPtrBase(p) :"; p4.printOn(cout);
      CHECK_TEST(p4==p1)
      CHECK_TEST(p4!=ptr0)
      CHECK_SAME(p4.getNumber(),5)
      CHECK_SAME(p4.getData(),p1.getData())

      delete [] data;
    }
    cout << "\tNumPtrBase constructors and inequalities ok" << endl;
  }

  { 
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(t,"testNumPtr constructors and inequalities");
      NumPtr<pointer_type> ptr0;
//    cout << "\tNumPtr() :"; ptr0.printOn(cout);
      ptr0.allocate(20);
//    cout << "\tafter allocate :"; ptr0.printOn(cout);
      CHECK_SAME(!ptr0,FALSE);
      CHECK_SAME(ptr0.getNumber(),20)

      NumPtr<pointer_type> p(10);
//    cout << "\tNumPtr(10) :"; p.printOn(cout);
      CHECK_SAME(!p,FALSE);
//    cout << "\tgetData = " << reinterpret_cast<void*>(p.getData()) 
//         << endl;
      CHECK_SAME(p.getNumber(),10)
      CHECK_TEST(p.ownsData())

      NumPtr<pointer_type> p4(p);
//    cout << "\tNumPtr(p) :"; p4.printOn(cout);
      CHECK_TEST(p4==p)
      CHECK_TEST(!(p4==ptr0))
      CHECK_TEST(!p4.ownsData())
      CHECK_SAME(p4.getNumber(),10)
      CHECK_SAME(p4.getData(),p.getData())

      NumPtr<pointer_type> p2(p,3);
//    cout << "\tNumPtr(p,3) :"; p2.printOn(cout);
      CHECK_SAME(p2.getNumber(),7)
      CHECK_SAME((p2+2).getNumber(),5)

      NumPtr<pointer_type> p3(p,3,5);
      CHECK_SAME(p3.getNumber(),5)
      CHECK_SAME((++p3).getNumber(),4)

      pointer_type *data=OPERATOR_NEW_BRACKET(pointer_type,5);
      NumPtr<pointer_type> p1(5,data);
//    cout << "\tNumPtr(5,data) :"; p1.printOn(cout);
      CHECK_SAME(p1.contains(p),FALSE)
      CHECK_SAME(p1.contains(p1),TRUE)
      CHECK_SAME(p1.contains(data),TRUE)
      CHECK_SAME(p1<data,FALSE)
      CHECK_SAME(p1<p1,FALSE)
      CHECK_SAME(p1>data,FALSE)
      CHECK_SAME(p1>p1,FALSE)
      CHECK_SAME(p1<=data,TRUE)
      CHECK_SAME(p1<=p1,TRUE)
      CHECK_SAME(p1>=data,TRUE)
      CHECK_SAME(p1>=p1,TRUE)
      CHECK_SAME(p1==p1,TRUE)
      CHECK_SAME(p1!=p1,FALSE)
      CHECK_SAME(p==p1,FALSE)
      CHECK_SAME(p!=p1,TRUE)

      delete [] data;
    }
    cout << "\tNumPtr constructors and inequalities ok" << endl;
  }

  { 
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(t,"testNumPtr initialize and copyFrom");
      NumPtr<pointer_type> p(10);
      pointer_type *data=OPERATOR_NEW_BRACKET(pointer_type,5);
      NumPtr<pointer_type> p1(5,data);
      p1.initialize(-1);

      for (int i=0;i<p.getNumber();i++) p[i]=i;
//    for (int i=0;i<p.getNumber();i++) {
//      cout << "\tp.operator[] = " << p[i] << endl; 
//    }
//    for (int i=0;i<p1.getNumber();i++) {
//      cout << "\tp1.operator[] = " << p1[i] << endl; 
//    }

      p.copyFrom(p1);
//    cout << "\tafter copyFrom :" << endl;
      for (int i=0;i<p.getNumber();i++) {
//      cout << "\tp.operator[] = " << p[i] << endl; 
        if (i<5) {
          CHECK_SAME(p[i],-1);
        } else CHECK_SAME(p[i],i);
      }

      delete [] data;
    }
    cout << "\tNumPtr initialize and copyFrom ok" << endl;
  }

  { 
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(t,"testNumPtr copy and undefine");
      pointer_type *data=OPERATOR_NEW_BRACKET(pointer_type,5);
      NumPtr<pointer_type> p(5,data);
      NumPtr<pointer_type> p1(10);
      NumPtr<pointer_type> ptr0;

      ptr0.copy(p);
      CHECK_TEST(!p.ownsData())
      CHECK_SAME(p.getNumber(),5)
      CHECK_SAME(p.getData(),ptr0.getData())
      CHECK_SAME(p.getNumber(),ptr0.getNumber())
      CHECK_TEST(!ptr0.ownsData())

      ptr0.copy(p1);
      CHECK_TEST(!p1.ownsData())
      CHECK_SAME(p1.getNumber(),10)
      CHECK_SAME(p1.getData(),ptr0.getData())
      CHECK_SAME(p1.getNumber(),ptr0.getNumber())
      CHECK_TEST(ptr0.ownsData())

      p.copy(ptr0);
      CHECK_TEST(!ptr0.ownsData())
      CHECK_SAME(ptr0.getNumber(),10)
      CHECK_SAME(ptr0.getData(),p.getData())
      CHECK_SAME(ptr0.getNumber(),p.getNumber())
      CHECK_TEST(p.ownsData())

      ptr0.undefine();
      CHECK_SAME(ptr0.getNumber(),0)
      CHECK_SAME(ptr0.getData(),0)

      delete [] data;
    }
    cout << "\tNumPtr copy and undefine ok" << endl;
  }

  { 
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(t,"testNumPtr write and read");
      NumPtr<pointer_type> p(10);
      for (int i=0;i<p.getNumber();i++) {
        p[i]=static_cast<pointer_type>(i);
      }
//    p.printOn(cout);
      NumPtr<pointer_type> ptr0(10);
//    ptr0.printOn(cout);

      ofstream ofile;
      ofile.open("numptrtest");
      CHECK_TEST(!ofile.fail());
      p.write(ofile);
      ofile.close();

      ifstream ifile;
      ifile.open("numptrtest");
      CHECK_TEST(!ifile.fail());
      ptr0.read(ifile);
      ifile.close();

//    cout << "\tafter read :"; ptr0.printOn(cout);
      for (int i=0;i<ptr0.getNumber();i++) {
//      cout << "\tptr0[" << i << "] = " << ptr0[i] << endl; 
        CHECK_SAME(ptr0[i],static_cast<pointer_type>(i))
      }

//    cout << "\tpointers go out of scope now" << endl;
    }
    cout << "\tNumPtr write and read ok" << endl;
  }

  { 
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(t,"testNumPtr constructors and inequalities");
      NumPtrWithValue<pointer_type> ptr0;
      CHECK_SAME(ptr0.getTotalValue(),0)
//    cout << "\tNumPtr() :"; ptr0.printOn(cout);
      ptr0.allocate(20);
//    cout << "\tafter allocate :"; ptr0.printOn(cout);
      CHECK_SAME(!ptr0,FALSE);
      CHECK_SAME(ptr0.getNumber(),20)
      CHECK_SAME(ptr0.getTotalValue(),0)

      NumPtrWithValue<pointer_type> p(10);
      for (int i=0;i<p.getNumber();i++) {
        p.setValue(i,static_cast<pointer_type>(-i),i);
      }
//    cout << "\tNumPtr(10) :"; p.printOn(cout);
      CHECK_SAME(!p,FALSE);
//    cout << "\tgetData = " << reinterpret_cast<void*>(p.getData()) 
//         << endl;
      CHECK_SAME(p.getNumber(),10)
      CHECK_TEST(p.ownsData())
      CHECK_SAME(p.getTotalValue(),45)

      NumPtrWithValue<pointer_type> p4(p);
//    cout << "\tNumPtr(p) :"; p4.printOn(cout);
      CHECK_TEST(p4==p)
      CHECK_TEST(!(p4==ptr0))
      CHECK_TEST(!p4.ownsData())
      CHECK_SAME(p4.getNumber(),10)
      CHECK_SAME(p4.getData(),p.getData())
      CHECK_SAME(p4.getTotalValue(),45)

      NumPtrWithValue<pointer_type> p2(p,3);
//    cout << "\tNumPtr(p,3) :"; p2.printOn(cout);
      CHECK_SAME(p2.getNumber(),7)
      CHECK_SAME(p2.getTotalValue(),42)
      CHECK_SAME((p2+2).getNumber(),5)
      CHECK_SAME((p2+2).getTotalValue(),35)

      NumPtrWithValue<pointer_type> p3(p,3,5);
      CHECK_SAME(p3.getNumber(),5)
      CHECK_SAME(p3.getTotalValue(),25)
      CHECK_SAME((++p3).getNumber(),4)
      CHECK_SAME(p3.getTotalValue(),22)

      pointer_type *data=OPERATOR_NEW_BRACKET(pointer_type,5);
      int *value_data=OPERATOR_NEW_BRACKET(int,5);
      for (int i=0;i<5;i++) value_data[i]=i;
      NumPtrWithValue<pointer_type> p1(5,data,value_data);
//    cout << "\tNumPtr(5,data) :"; p1.printOn(cout);
      CHECK_SAME(p1.getTotalValue(),10)
      CHECK_SAME(p1.contains(p),FALSE)
      CHECK_SAME(p1.contains(p1),TRUE)
      CHECK_SAME(p1.contains(data),TRUE)
      CHECK_SAME(p1<data,FALSE)
      CHECK_SAME(p1<p1,FALSE)
      CHECK_SAME(p1>data,FALSE)
      CHECK_SAME(p1>p1,FALSE)
      CHECK_SAME(p1<=data,TRUE)
      CHECK_SAME(p1<=p1,TRUE)
      CHECK_SAME(p1>=data,TRUE)
      CHECK_SAME(p1>=p1,TRUE)
      CHECK_SAME(p1==p1,TRUE)
      CHECK_SAME(p1!=p1,FALSE)
      CHECK_SAME(p==p1,FALSE)
      CHECK_SAME(p!=p1,TRUE)

      delete [] data;
      delete [] value_data;
      
    }
    cout << "\tNumPtrWithValue constructors and inequalities ok" <<endl;
  }

  { 
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(t,"testNumPtr initialize and copyFrom");
      NumPtrWithValue<pointer_type> p(10);
      pointer_type *data=OPERATOR_NEW_BRACKET(pointer_type,5);
      int *value_data=OPERATOR_NEW_BRACKET(int,5);
      NumPtrWithValue<pointer_type> p1(5,data,value_data);
      p1.initialize(-1,1);
      CHECK_SAME(p1.getTotalValue(),5)

      for (int i=0;i<p.getNumber();i++) {
        p.setValue(i,static_cast<pointer_type>(-i),i);
      }
//    for (int i=0;i<p.getNumber();i++) {
//      cout << "\tp.operator[] = " << p[i] << endl; 
//    }
//    for (int i=0;i<p1.getNumber();i++) {
//      cout << "\tp1.operator[] = " << p1[i] << endl; 
//    }

      p.copyFrom(p1);
//    cout << "\tafter copyFrom :" << endl;
      for (int i=0;i<p.getNumber();i++) {
//      cout << "\tp.operator[] = " << p[i] << endl; 
        pointer_type t;
        int v=INT_MAX;
        p.getValue(i,t,v);
        if (i<5) {
          CHECK_SAME(t,static_cast<pointer_type>(-1))
          CHECK_SAME(v,1)
        }
        else {
          CHECK_SAME(t,static_cast<pointer_type>(-i))
          CHECK_SAME(v,i)
        }
      }

      delete [] data;
      delete [] value_data;
    }
    cout << "\tNumPtrWithValue initialize and copyFrom ok" <<endl;
  }

  { 
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(t,"testNumPtr copy and undefine");
      pointer_type *data=OPERATOR_NEW_BRACKET(pointer_type,5);
      int *value_data=OPERATOR_NEW_BRACKET(int,5);
      NumPtrWithValue<pointer_type> p(5,data,value_data);
      NumPtrWithValue<pointer_type> p1(10);
      NumPtrWithValue<pointer_type> ptr0;

      ptr0.copy(p);
      CHECK_TEST(!p.ownsData())
      CHECK_SAME(p.getNumber(),5)
      CHECK_SAME(p.getData(),ptr0.getData())
      CHECK_SAME(p.getNumber(),ptr0.getNumber())
      CHECK_TEST(!ptr0.ownsData())

      ptr0.copy(p1);
      CHECK_TEST(!p1.ownsData())
      CHECK_SAME(p1.getNumber(),10)
      CHECK_SAME(p1.getData(),ptr0.getData())
      CHECK_SAME(p1.getNumber(),ptr0.getNumber())
      CHECK_TEST(ptr0.ownsData())

      p.copy(ptr0);
      CHECK_TEST(!ptr0.ownsData())
      CHECK_SAME(ptr0.getNumber(),10)
      CHECK_SAME(ptr0.getData(),p.getData())
      CHECK_SAME(ptr0.getNumber(),p.getNumber())
      CHECK_TEST(p.ownsData())

      ptr0.undefine();
      CHECK_SAME(ptr0.getNumber(),0)
      CHECK_SAME(ptr0.getData(),0)

      delete [] data;
      delete [] value_data;
    }
    cout << "\tNumPtrWithValue copy and undefine ok" <<endl;
  }

  { 
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(t,"testNumPtr write and read");
      NumPtrWithValue<pointer_type> p(10);
      for (int i=0;i<p.getNumber();i++) {
        p.setValue(i,static_cast<pointer_type>(-i),i);
      }
//    p.printOn(cout);
      NumPtrWithValue<pointer_type> ptr0(10);
//    ptr0.printOn(cout);

      ofstream ofile;
      ofile.open("numptrtest");
      CHECK_TEST(!ofile.fail());
      p.write(ofile);
      ofile.close();

      ifstream ifile;
      ifile.open("numptrtest");
      CHECK_TEST(!ifile.fail());
      ptr0.read(ifile);
      ifile.close();

//    cout << "\tafter read :"; ptr0.printOn(cout);
      for (int i=0;i<ptr0.getNumber();i++) {
        pointer_type t;
        int v;
        ptr0.getValue(i,t,v);
        CHECK_SAME(t,static_cast<pointer_type>(-i))
        CHECK_SAME(v,i);
      }
//    cout << "\tpointers go out of scope now" << endl;
    }
    cout << "\tNumPtrWithValue write and read ok" <<endl;
  }

//
  { 
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(t,"testNumPtr sort");
      int nv=101;
      NumPtrWithValue<pointer_type> p(nv),q(nv);
      unsigned int seed=0;
      srand(seed);
      for (int i=0;i<p.getNumber();i++) {
        pointer_type t=static_cast<pointer_type>(nv)*rand() 
                        / (RAND_MAX+1.);
//      pointer_type t=static_cast<pointer_type>(i);
        int v=static_cast<int>(t); 
        p.setValue(i,t,v);
        q.setValue(i,t,v);
      }
      p.heapSort(TRUE);
      q.heapSort(FALSE);
    }
    cout << "\tNumPtrWithValue sort ok" <<endl;
  }
//

//
  { 
#ifdef MEM_DEBUG
    MemoryDebugger md(1);
#endif
    { 
//    TRACER_CALL(t,"testNumPtr sort");
      int nv=101;
      NumPtrWithCompare<pointer_type*,int*> p(nv),q(nv);
      unsigned int seed=0;
      srand(seed);
      for (int i=0;i<p.getNumber();i++) {
        pointer_type t=static_cast<pointer_type>(nv)*rand() 
                        / (RAND_MAX+1.);
//      pointer_type t=static_cast<pointer_type>(i);
        int v=static_cast<int>(t); 
        p.setValue(i,OPERATOR_NEW pointer_type(t),OPERATOR_NEW int(v));
        q.setValue(i,OPERATOR_NEW pointer_type(t),OPERATOR_NEW int(v));
      }
      p.heapSort(TRUE);
      q.heapSort(FALSE);
      for (int i=0;i<p.getNumber();i++) {
        REAL *t=0;
        int *v=0;
        p.getValue(i,t,v);
        delete t;
        delete v;
        q.getValue(i,t,v);
        delete t;
        delete v;
      }
    }
    cout << "\tNumPtrWithCompare sort ok" <<endl;
  }
//

  return 1;
}
