//  "$Header: /home/faculty/johnt/cvs/deal_new/mg/MGDofHandler.C,v 1.1 2009/08/20 17:34:10 johnt Exp $"
//----------------------------  mg_dof_handler.cc  ------------------------
//  $Id: MGDofHandler.C,v 1.1 2009/08/20 17:34:10 johnt Exp $
//  Version: $Name:  $
//
//  Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//  This file is subject to QPL and may not be  distributed
//  without copyright and license information. Please refer
//  to the file deal.II/doc/license.html for the  text  and
//  further information on this license.
//
//----------------------------  mg_dof_handler.cc  ------------------------
//
//modified from deal.II/deal.II/source/multigrid/mg_dof_handler.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <MGDofHandler.H>
#include <MGTransfer.H>
#include <MemoryDebugger.H>
#include <NIDLList.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGDofHandler::MGVertexDofs::init(int cl,int fl,int dpv) {
  CHECK_SAME(indices.getNumber(),0);
  coarsest_level=cl;
  finest_level=fl;
  int n_levels=finest_level-coarsest_level+1;
  indices.allocate(n_levels*dpv);
  indices.initialize(invalid_dof_index);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGDofHandler::MGVertexDofs::printOn(ostream &os) const {
  os << "MGDofHandler::MGVertexDofs:" << endl;
  os << "\tcoarsest_level = " << coarsest_level << endl;
  os << "\tfinest_level = " << finest_level << endl;
  os << "\tindices.getNumber = " << indices.getNumber() << endl;
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGDofHandler::mgVertexDofIndices(const Shape *element,int ln,
NumPtr<int> &vertex_dof_indices) const {
  int dpv=selected_fe->dofsPerVertex();
  int nv=element->numberVertices();
  int i=0;
  NumPtr<int> vdi(vertex_dof_indices);
  for (int v=0;v<nv;v++) {
    mg_vertex_dofs[element->vertexIndex(v)].getIndices(ln,dpv,vdi);
    vdi+=dpv;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGDofHandler::getMGDofIndices(const Shape1 *element,
const Level1 *level,NumPtr<int> &dof_indices) const {
  CHECK_SAME(element->numberDimensions(),1);

  int dpv=selected_fe->dofsPerVertex();
  int dpl=selected_fe->dofsPerLine();
  int nv=element->numberVertices();
  int ln=level->numberCoarserLevels();

  dof_indices.allocate(nv*dpv+dpl);
  mgVertexDofIndices(element,ln,dof_indices);
  NumPtr<int> remaining_indices(dof_indices,nv*dpv);
  mg_levels[ln]->DofLevel1::getDofs(element,remaining_indices);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM>1)
void MGDofHandler::getMGDofIndices(const Shape2 *element,
const Level2 *level,NumPtr<int> &dof_indices) const {
  CHECK_SAME(element->numberDimensions(),2);
  int dpv=selected_fe->dofsPerVertex();
  int dpl=selected_fe->dofsPerLine();
  int nv=element->numberVertices();
  int dpq=selected_fe->dofsPerQuad(element);
  int nl=element->numberCodimension1Shapes();
  int ln=level->numberCoarserLevels();

  dof_indices.allocate(nv*dpv+nl*dpl+dpq);
  mgVertexDofIndices(element,ln,dof_indices);
  int offset=nv*dpv;
  DofLevel2 *mgl=mg_levels[ln];
  for (int l=0;l<nl;l++) {
    int face_index=element->codimension1ShapeIndex(l);
    const Shape1 *face_element=level->Level1::getShape(face_index);
    NumPtr<int> face_dofs(dof_indices,offset,dpl);
    mgl->DofLevel1::getDofs(face_element,face_dofs);
    offset+=dpl;
  }
  NumPtr<int> remaining_dofs(dof_indices,offset);
  mgl->getDofs(element,remaining_dofs);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void MGDofHandler::getMGDofIndices(const Shape3 *element,
const Level3 *level,NumPtr<int> &dof_indices) const {
  CHECK_SAME(element->numberDimensions(),3);
  int dpv=selected_fe->dofsPerVertex();
  int dpl=selected_fe->dofsPerLine();
  int dph=selected_fe->dofsPerHex(element);
  int nv=element->numberVertices();
  int nq=element->numberCodimension1Shapes();
  int nl=element->numberCodimension2Shapes();
  int ln=level->numberCoarserLevels();

  int sz=nv*dpv+nl*dpl+dph;
  for (int f=0;f<nq;f++) {
    int face_index=element->codimension1ShapeIndex(f);
    const Shape2 *face_element=level->Level2::getShape(face_index);
    sz+=selected_fe->dofsPerQuad(face_element);
  }
  dof_indices.allocate(sz);
  mgVertexDofIndices(element,ln,dof_indices);
  int offset=nv*dpv;
  DofLevel3 *mgl=mg_levels[ln];
  for (int l=0;l<nl;l++) {
    int line_index=element->codimension2ShapeIndex(l);
    const Shape1 *line_element=level->Level1::getShape(line_index);
    NumPtr<int> line_dofs(dof_indices,offset,dpl);
    mgl->DofLevel1::getDofs(line_element,line_dofs);
    offset+=dpl;
  }
  for (int f=0;f<nq;f++) {
    int face_index=element->codimension1ShapeIndex(f);
    const Shape2 *face_element=level->Level2::getShape(face_index);
    int dpq=selected_fe->dofsPerQuad(face_element);
    NumPtr<int> face_dofs(dof_indices,offset,dpq);
    mgl->DofLevel2::getDofs(face_element,face_dofs);
    offset+=dpq;
  }
  NumPtr<int> remaining_dofs(dof_indices,offset);
  mgl->getDofs(element,remaining_dofs);
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGDofHandler::setMGVertexDofs() {
  int number_levels=tess->numberLevels();
  int number_vertices=tess->getVertices().getNumber();
  mg_vertex_dofs.allocate(number_vertices);
  NumPtr<int> min_level(number_vertices);
  min_level.initialize(number_levels);
  NumPtr<int> max_level(number_vertices);
  max_level.initialize(0);
  for (bool doit=tess->firstShape();doit;doit=tess->nextShape()) {
    const LEVEL_CLASS *level=tess->presentLevel();
    int ncl=level->numberCoarserLevels();
    const SHAPE_CLASS *element=level->getShape();
    for (int v=0;v<element->numberVertices();v++) {
      int vertex_index=element->vertexIndex(v);
      min_level[vertex_index]=min(min_level[vertex_index],ncl);
      max_level[vertex_index]=max(max_level[vertex_index],ncl);
    }
  }
  for (int v=0;v<number_vertices;v++) {
    if (tess->vertexUsed(v)) {
      CHECK_TEST(min_level[v]<number_levels);
      CHECK_TEST(max_level[v]>=min_level[v]);
      mg_vertex_dofs[v].init(min_level[v],max_level[v],
        selected_fe->dofsPerVertex());
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
void MGDofHandler::reserveSpace() {
  CHECK_POINTER(selected_fe);
  int number_levels=tess->numberLevels();
  CHECK_TEST(number_levels>0);
  clearSpace();
  mg_levels.allocate(number_levels);
  int dpl=selected_fe->dofsPerLine();
  int i=0;
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel(),i++) {
    int d1=0;
    for (bool doit=level->first();doit;doit=level->next()) {
      level->getShape()->setDofIndex(d1);
      d1+=dpl;
    }
    mg_levels[i]=OPERATOR_NEW DofLevel1(d1);
  }
  setMGVertexDofs();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
void MGDofHandler::reserveSpace() {
  CHECK_POINTER(selected_fe);
  int number_levels=tess->numberLevels();
  CHECK_TEST(number_levels>0);
  clearSpace();
  mg_levels.allocate(number_levels);
  int dpl=selected_fe->dofsPerLine();
  int i=0;
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel(),i++) {
    int d1=0;
    Level1 *level1=static_cast<Level1*>(level);
    for (bool doit=level1->first();doit;doit=level1->next()) {
      level1->getShape()->setDofIndex(d1);
      d1+=dpl;
    }
    int d2=0;
    for (bool doit=level->first();doit;doit=level->next()) {
      Shape2 *e=level->getShape();
      e->setDofIndex(d2);
      d2+=selected_fe->dofsPerQuad(e);
    }
    mg_levels[i]=OPERATOR_NEW DofLevel2(d1,d2);
  }
  setMGVertexDofs();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
void MGDofHandler::reserveSpace() {
  CHECK_POINTER(selected_fe);
  int number_levels=tess->numberLevels();
  CHECK_TEST(number_levels>0);
  clearSpace();
  mg_levels.allocate(number_levels);
  int dpl=selected_fe->dofsPerLine();
  int i=0;
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel(),i++) {
    int d1=0;
    Level1 *level1=static_cast<Level1*>(level);
    for (bool doit=level1->first();doit;doit=level1->next()) {
      level1->getShape()->setDofIndex(d1);
      d1+=dpl;
    }
    int d2=0;
    Level2 *level2=static_cast<Level2*>(level);
    for (bool doit=level2->first();doit;doit=level2->next()) {
      Shape2 *e=level2->getShape();
      e->setDofIndex(d2);
      d2+=selected_fe->dofsPerQuad(e);
    }
    int d3=0;
    for (bool doit=level->first();doit;doit=level->next()) {
      Shape3 *e=level->getShape();
      e->setDofIndex(d3);
      d3+=selected_fe->dofsPerHex(e);
    }
    mg_levels[i]=OPERATOR_NEW DofLevel3(d1,d2,d3);
  }
  setMGVertexDofs();
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==1)
int MGDofHandler::distributeDofsOnCell(const Shape1 *element,
Level1 *level,int next_free_dof) { 
  int dpv=selected_fe->dofsPerVertex();
  int ln=level->numberCoarserLevels();
  if (dpv>0) {
    for (int v=0;v<element->numberVertices();v++) {
      Level1 *neighbor_level=level->neighborLevel(element,v);
      int vi=element->vertexIndex(v);
      if (neighbor_level==level) {
        int neighbor_index=level->neighborIndex(element,v);
        if (level->userFlagSet(neighbor_index)) {
          const Shape1 *neighbor_element=
            level->getShape(neighbor_index);
          int nvi=neighbor_element->vertexIndex(1-v);
          NumPtr<int> neighbor_vertex_dofs(dpv);
          mg_vertex_dofs[nvi].getIndices(ln,dpv,neighbor_vertex_dofs);
          for (int d=0;d<dpv;d++) {
            mg_vertex_dofs[vi].setIndex(ln,d,dpv,neighbor_vertex_dofs[d]);
          }
          continue;
        }
      }
      for (int d=0;d<dpv;d++,next_free_dof++) {
        mg_vertex_dofs[vi].setIndex(ln,d,dpv,next_free_dof);
      }
    }
  }
  int dpl=selected_fe->dofsPerLine();
  mg_levels[ln]->setDofs(element,dpl,next_free_dof);
  level->setUserFlag(element);
  return next_free_dof;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==2)
int MGDofHandler::distributeDofsOnCell(const Shape2 *element,
Level2 *level,int next_free_dof) { 
  int dpv=selected_fe->dofsPerVertex();
  int ln=level->numberCoarserLevels();
  if (dpv>0) {
    for (int v=0;v<element->numberVertices();v++) {
      int vi=element->vertexIndex(v);
      NumPtr<int> indices(dpv);
      mg_vertex_dofs[vi].getIndices(ln,dpv,indices);
      if (indices[0]==invalid_dof_index) {
        for (int d=0;d<dpv;d++,next_free_dof++) {
          mg_vertex_dofs[vi].setIndex(ln,d,dpv,next_free_dof);
        }
      }
    }
  }
  DofLevel2 *dof_level2=mg_levels[ln];
  int dpl=selected_fe->dofsPerLine();
  if (dpl>0) {
    Level1 *level1=static_cast<Level1*>(level);
    DofLevel1 *dof_level1=static_cast<DofLevel1*>(dof_level2);
    for (int f=0;f<element->numberCodimension1Shapes();f++) { 
      int face_index=element->codimension1ShapeIndex(f);
      const Shape1 *face_element=level1->getShape(face_index);
      dof_level1->setDofs(face_element,dpl,next_free_dof);
    }
  }
  int dpq=selected_fe->dofsPerQuad(element);
  if (dpq>0) {
    dof_level2->setDofs(element,dpq,next_free_dof);
  }
  level->setUserFlag(element);
  return next_free_dof;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#if (SPACEDIM==3)
int MGDofHandler::distributeDofsOnCell(const Shape3 *element,
Level3 *level,int next_free_dof) { 
  int dpv=selected_fe->dofsPerVertex();
  int ln=level->numberCoarserLevels();
  if (dpv>0) {
    for (int v=0;v<element->numberVertices();v++) {
      int vi=element->vertexIndex(v);
      NumPtr<int> indices(dpv);
      mg_vertex_dofs[vi].getIndices(ln,dpv,indices);
      if (indices[0]==invalid_dof_index) {
        for (int d=0;d<dpv;d++,next_free_dof++) {
          mg_vertex_dofs[vi].setIndex(ln,d,dpv,next_free_dof);
        }
      }
    }
  }
  DofLevel3 *dof_level3=mg_levels[ln];
  int dpl=selected_fe->dofsPerLine();
  if (dpl>0) {
    const Level1 *level1=static_cast<const Level1*>(level);
    DofLevel1 *dof_level1=static_cast<DofLevel1*>(dof_level3);
    for (int l=0;l<element->numberCodimension2Shapes();l++) { 
      int line_index=element->codimension2ShapeIndex(l);
      const Shape1 *line_element=level1->getShape(line_index);
      dof_level1->setDofs(line_element,dpl,next_free_dof);
    }
  }
  Level2 *level2=static_cast<Level2*>(level);
  DofLevel2 *dof_level2=static_cast<DofLevel2*>(dof_level3);
  for (int f=0;f<element->numberCodimension1Shapes();f++) {
    int face_index=element->codimension1ShapeIndex(f);
    const Shape2 *face_element=level2->getShape(face_index);
    int dpq=selected_fe->dofsPerQuad(face_element);
    if (dpq>0) {
      dof_level2->setDofs(face_element,dpq,next_free_dof);
    }
  }
  int dph=selected_fe->dofsPerHex(element);
  if (dph>0) {
    dof_level3->setDofs(element,dph,next_free_dof);
  }
  level->setUserFlag(element);
  return next_free_dof;
}
#endif
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGDofHandler::distributeDofs() {
  DofHandler::distributeDofs();
  reserveSpace();
  mg_used_dofs.cleanup();
  mg_used_dofs.allocate(tess->numberLevels());
  mg_used_dofs.initialize(0);

  NumPtr<bool> user_flags;
  tess->saveUserFlags(user_flags);
  const_cast<Tessellation*>(tess)->clearUserFlags();
  int i=0;
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel(),i++) {
    int next_free_dof=0;
    for (bool doit=level->first();doit;doit=level->next()) {
      const SHAPE_CLASS *element=level->getShape();
      next_free_dof=distributeDofsOnCell(element,level,next_free_dof);
    }
    mg_used_dofs[i]=next_free_dof;
  }
  const_cast<Tessellation*>(tess)->loadUserFlags(user_flags);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGDofHandler::clearSpace() {
  for (int i=0;i<mg_levels.getNumber();i++) {
    delete mg_levels[i];
  }
  mg_levels.cleanup();
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//this is very similar to DofHandler::makeSparsityPattern
//the differences are 
//  this routine works with all elements on a single level
//  dofs_on_this_cell come from getMGDofIndices, not getDofIndices
void MGDofHandler::makeSparsityPattern(SparsityPatternBase &sparsity,
int ln) const {
  int n_dofs=numberDofs(ln);
  CHECK_SAME(sparsity.nRows(),n_dofs);
  CHECK_SAME(sparsity.nCols(),n_dofs);
  Tessellation &tess=getTessellation();
  FiniteElement &fe=getFe();
  int level_number=0;
  for (LEVEL_CLASS *level=tess.coarsestLevel();level;
  level=level->finerLevel(),level_number++) {
    if (level_number<ln) continue;
    for (bool doit=level->first();doit;doit=level->next()) {
      const SHAPE_CLASS *element=level->getShape();
      int dpc=fe.dofsPerCell(element);
      NumPtr<int> dofs_on_this_cell;
      getMGDofIndices(element,level,dofs_on_this_cell);
      for (int i=0; i<dpc;i++) {
        for (int j=0; j<dpc;j++) {
          sparsity.add(dofs_on_this_cell[i],dofs_on_this_cell[j]);
        }
      }
    }
    break;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//this is called from Multigrid::Multigrid only
//  src is a global stiffness matrix, so it uses DofHandler::getDofIndices
//  dst is a multigrid matrix, so it uses MGDofHandler::getMGDofIndices 
void MGDofHandler::copyMatrix(int ln,const SolverMatrix &src,
SolverMatrix &dst) const {
  FiniteElement &fe=getFe();
  Tessellation &tess=getTessellation();
  int level_number=0;
  for (LEVEL_CLASS *level=tess.coarsestLevel();level;
  level=level->finerLevel(),level_number++) {
    if (level_number<ln) continue;
    for (bool doit=level->first();doit;doit=level->next()) {
      if (!level->hasChildren()) {
        const SHAPE_CLASS *element=level->getShape();
        int dpc=fe.dofsPerCell(element);
        NumPtr<int> dofs_on_this_cell;
        DofHandler::getDofIndices(element,level,dofs_on_this_cell);
        NumPtr<int> mg_dofs_on_this_cell;
        getMGDofIndices(element,level,mg_dofs_on_this_cell);
        for (int i=0; i<dpc;i++) {
          int mg_di=mg_dofs_on_this_cell[i];
          int di=dofs_on_this_cell[i];
          for (int j=0; j<dpc;j++) {
            int mg_dj=mg_dofs_on_this_cell[j];
            int dj=dofs_on_this_cell[j];
            dst(mg_di,mg_dj)=src(di,dj); 
          }
        }
      }
    }
    break;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGDofHandler::findDirichletBoundaryDofs(const Mapping &mapping,
NumPtr<set<int> > &dirichlet_boundary_dofs) const {
  dirichlet_boundary_dofs.cleanup();
  dirichlet_boundary_dofs.allocate(tess->numberLevels());
  int ln=0;
  int dpv=selected_fe->dofsPerVertex();
  int dpl=selected_fe->dofsPerLine();
  for (LEVEL_CLASS *level=tess->coarsestLevel();level;
  level=level->finerLevel(),ln++) {
    CODIMENSION1_LEVEL_CLASS *codimension1_level=
      static_cast<CODIMENSION1_LEVEL_CLASS*>(level);
    DOF_LEVEL_CLASS *mg_level=mg_levels[ln];
    for (bool doit=level->first();doit;doit=level->next()) {
      const SHAPE_CLASS *element=level->getShape();
#if (SPACEDIM==1)
      for (int f=0;f<element->numberCodimension1Shapes();f++)
      {
        if (!level->neighborLevel(f)) {
          const BoundaryCondition *bc=
            codimension1_level->getBoundaryCondition(f);
          if (bc->isEssential()) {
            int vi=element->vertexIndex(f);
            NumPtr<int> indices(dpv);
            mg_vertex_dofs[vi].getIndices(ln,dpv,indices);
            for (int i=0;i<dpv;i++) {
              dirichlet_boundary_dofs[ln].insert(indices[i]);
            }
          }
        }
      }
#else
      for (int f=0;f<element->numberCodimension1Shapes();f++) {
        int face_index=element->codimension1ShapeIndex(f);
        const BoundaryCondition *bc=
          codimension1_level->getBoundaryCondition(face_index);
        if (bc!=0) {
          if (bc->isEssential()) {
            const CODIMENSION1_SHAPE_CLASS *face_element=
              codimension1_level->getShape(face_index);
            for (int v=0;v<face_element->numberVertices();v++) {
              int vi=face_element->vertexIndex(v);
              NumPtr<int> indices(dpv);
              mg_vertex_dofs[vi].getIndices(ln,dpv,indices);
              for (int i=0;i<dpv;i++) {
                dirichlet_boundary_dofs[ln].insert(indices[i]);
              }
            }
#if (SPACEDIM==2)
            NumPtr<int> face_indices(dpl);
            mg_level->DofLevel1::getDofs(face_element,face_indices);
            for (int d=0;d<dpl;d++) {
              dirichlet_boundary_dofs[ln].insert(face_indices[d]);
            }
#else
            for (int l=0;
            l<face_element->numberCodimension1Shapes();l++) {
              int line_index=face_element->codimension1ShapeIndex(l);
              const Shape1 *line_element=
                level->Level1::getShape(line_index);
              NumPtr<int> line_indices(dpl);
              mg_level->DofLevel1::getDofs(line_element,line_indices);
              for (int d=0;d<dpl;d++) {
                dirichlet_boundary_dofs[ln].insert(line_indices[d]);
              }
            }
            int dpq=selected_fe->dofsPerQuad(face_element);
            NumPtr<int> face_indices(dpq);
            mg_level->DofLevel2::getDofs(face_element,face_indices);
            for (int d=0;d<dpq;d++) {
              dirichlet_boundary_dofs[ln].insert(face_indices[d]);
            }
#endif
          }
        }
      }
#endif
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGDofHandler::coarsenMatrix(int coarse_ln,
const MGTransfer &transfer,const SolverMatrix &fine_matrix,
SolverMatrix &coarse_matrix) const {
  int number_levels=getTessellation().numberLevels();
  FiniteElement &fe=getFe();
  Tessellation &tess=getTessellation();
  int fine_ln=coarse_ln+1;
  Vector xc(transfer.prolongationSize(coarse_ln,1));
  Vector xf(transfer.prolongationSize(coarse_ln,0));
  Vector yf(transfer.prolongationSize(coarse_ln,0));
  NumPtr<NIDLList<int> > 
    vertex_to_element(tess.getVertices().getNumber());
  NIDLList<NIDLList<int> > shape_colors;
  int level_number=0;
  for (LEVEL_CLASS *level=tess.coarsestLevel();level;
  level=level->finerLevel(),level_number++) {
    if (level_number<coarse_ln) continue;
    for (bool doit=level->first();doit;doit=level->next()) {
      if (level->hasChildren()) {
        const SHAPE_CLASS *element=level->getShape();
        int edi=element->elementDataIndex();
        for (int v=0;v<element->numberVertices();v++) {
          vertex_to_element[element->vertexIndex(v)].
            append(OPERATOR_NEW int(edi));
        }
      }
    }
  NumPtr<NIDLList<int> > 
    vertex_to_element_original(vertex_to_element.getNumber());
  for (int i=0;i<vertex_to_element.getNumber();i++) {
    NIDLList<int> &vte=vertex_to_element[i];
    NIDLList<int> &vteo=vertex_to_element_original[i];
    for (NIDLListNode<int> *p=vte.first();p;p=vte.next(p)) {
      vteo.append(OPERATOR_NEW int(*(p->getData())));
    }
  }
    bool done=false;
    while (!done) {
      NIDLList<int> *current_shape_color_list=
        OPERATOR_NEW NIDLList<int>;
      NumPtr<NIDLList<int> > 
        vertex_to_element_copy(tess.getVertices().getNumber());
      for (int i=0;i<vertex_to_element.getNumber();i++) {
        NIDLList<int> &vte=vertex_to_element[i];
        for (NIDLListNode<int> *p=vte.first();p;p=vte.next(p)) {
          vertex_to_element_copy[i].
            append(OPERATOR_NEW int(*(p->getData())));
        }
      }
      for (int i=0;i<vertex_to_element_copy.getNumber();i++) {
        NIDLList<int> &vtec=vertex_to_element_copy[i];
        if (vtec.empty()) continue;
        int new_shape_to_color= *vtec.first()->getData();
        current_shape_color_list->
          append(OPERATOR_NEW int(new_shape_to_color));
        set<int> touched_vertices;
        set<int> neighbor_elements;
        neighbor_elements.insert(new_shape_to_color);
        const SHAPE_CLASS *element=level->getShape(new_shape_to_color);
        for (int v=0;v<element->numberVertices();v++) {
          int vi=element->vertexIndex(v);
          NIDLList<int> &vteo=vertex_to_element_original[vi];
          for (NIDLListNode<int> *p=vteo.first();p;p=vteo.next(p)) {
            int new_shape_to_color_neighbor= *(p->getData());
            const SHAPE_CLASS *neighbor_element=
              level->getShape(new_shape_to_color_neighbor);
            for (int nv=0;
            nv<neighbor_element->numberVertices();nv++) {
              int nvi=neighbor_element->vertexIndex(nv);
              touched_vertices.insert(nvi);
              NIDLList<int> &vtecn=vertex_to_element_copy[nvi];
              for (NIDLListNode<int> *pn=vtecn.first();pn;
              pn=vtecn.next(pn)) {
                neighbor_elements.insert(*(pn->getData()));
              }
            }
          }
          NIDLList<int> &vtev=vertex_to_element[vi];
          for (NIDLListNode<int> *p=vtev.first();p;p=vtev.next(p)) {
            if (new_shape_to_color== *(p->getData())) {
              delete vtev.delAndNext(p);
              break;
            }
          }
        }
        for (set<int>::const_iterator ne_iterator=
        neighbor_elements.begin();ne_iterator!=neighbor_elements.end();
        ne_iterator++) {
          int neighbor_index=*ne_iterator;
          const SHAPE_CLASS *neighbor_element=
            level->getShape(neighbor_index);
          for (int v=0;v<neighbor_element->numberVertices();v++) {
            int vi=neighbor_element->vertexIndex(v);
            NIDLList<int> &vtec=vertex_to_element_copy[vi];
            for (NIDLListNode<int> *p=vtec.first();p;p=vtec.next(p)) {
              if (*p->getData()==neighbor_index) {
                delete vtec.delAndNext(p);
                break;
              }
            }
          }
        }
      }
      done=current_shape_color_list->empty();
      if (!done) shape_colors.append(current_shape_color_list);
      else delete current_shape_color_list;
    }

    for (int i=0;i<vertex_to_element.getNumber();i++) {
      CHECK_TEST(vertex_to_element[i].empty());
    }
    vertex_to_element.cleanup();
    for (int i=0;i<vertex_to_element_original.getNumber();i++){
      NIDLList<int> vteo=vertex_to_element_original[i];
      while (!vteo.empty()) delete vteo.delAfter(0);
    }
//  the following will recompute columns of coarse matrix corresponding
//    to dofs shared by elements (vertices, face dofs)
//  since a shared dof could be combined with a previously unshared dof
//    on a different element with the same color, 
//    the recomputation is hard to prevent
    for(NIDLListNode<NIDLList<int> > *sc=shape_colors.first();sc;
    sc=shape_colors.next(sc)) {
      NIDLList<int> *scl=sc->getData();
      NumPtr<NumPtr<int> > dofs_on_cells(scl->length());
      int max_dpc=0;
      int n=0;
      for (NIDLListNode<int> *e=scl->first();e;e=scl->next(e),n++) {
        int edi= *(e->getData());
        const SHAPE_CLASS *element=level->getShape(edi);
        int dpc=fe.dofsPerCell(element);
        max_dpc=max(max_dpc,dpc);
//      dofs_on_cells[n].allocate(dpc); 
        getMGDofIndices(element,level,dofs_on_cells[n]);
      }
      for (int j=0;j<max_dpc;j++) {
        xc=0.;
        for (int n=0;n<dofs_on_cells.getNumber();n++) {
          const NumPtr<int> &docn=dofs_on_cells[n];
          if (j<docn.getNumber()) xc[docn[j]]=1.;
        }
        xf=0.;
        transfer.prolongationAndAdd(fine_ln,xf,xc);
        fine_matrix.vmult(yf,xf);
        transfer.restriction(fine_ln,xc,yf);
        for (int n=0;n<dofs_on_cells.getNumber();n++) {
          const NumPtr<int> &docn=dofs_on_cells[n];
          if (j<docn.getNumber()) {
            int dj=docn[j];
            for (int i=0;i<docn.getNumber();i++) {
              int di=docn[i];
              coarse_matrix(di,dj)=xc[di]; 
            }
          }
        }
      }
    }
    break;
  }
  while (!shape_colors.empty()) {
    NIDLList<int> *sc=shape_colors.first()->getData();
    while (!sc->empty()) delete sc->delAfter(0);
    delete shape_colors.delAfter(0);
  }
//the prolongation and restriction contribute nothing at Dirichlet bc's:
//  the following prevents floating point exceptions in preconditioners
  int m=coarse_matrix.size(0);
  for (int i=0;i<m;i++) {
    REAL &diag=coarse_matrix(i,i);
    if (abs(diag)<=0.) diag=1.;
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGDofHandler::printOn(ostream &os) const {
  os << "MGDofHandler: " << endl;
  os << "\tmg_levels.getNumber = " << mg_levels.getNumber() << endl;
  os << "\tmg_vertex_dofs.getNumber = " << mg_vertex_dofs.getNumber() 
     << endl;
  os << "\tmg_used_dofs.getNumber = " << mg_used_dofs.getNumber() 
     << endl;
  DofHandler::printOn(os);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGDofHandler::printDofs() const {
  for (bool doit=tess->firstShape();doit;doit=tess->nextShape()) {
    const LEVEL_CLASS *level=tess->presentLevel();
    int ncl=level->numberCoarserLevels();
    const SHAPE_CLASS *element=level->getShape();
    NumPtr<int> dof_indices;
    getMGDofIndices(element,level,dof_indices);
    cout << "\n\telement = " << element->elementDataIndex() 
         << ", vertices = ";
    for (int v=0;v<element->numberVertices();v++) {
      cout << element->vertexIndex(v) << " ";
    }
    cout << endl;
    for (int d=0;d<dof_indices.getNumber();d++) {
      cout << "\tmg_dof_indices[" << d << "] = " << dof_indices[d]
           << endl;
    }
  } 
#if (SPACEDIM>1)
  for (bool doit=tess->firstCodimension1Shape();doit;
  doit=tess->nextCodimension1Shape()) {
    const CODIMENSION1_LEVEL_CLASS *level=
      tess->presentCodimension1Level();
    int ncl=level->numberCoarserLevels();
    const CODIMENSION1_SHAPE_CLASS *face_element=level->getShape();
    NumPtr<int> dof_indices;
    getMGDofIndices(face_element,level,dof_indices);
    cout << "\n\tface_element = " << face_element->elementDataIndex() 
         << ", vertices = ";
    for (int v=0;v<face_element->numberVertices();v++) {
      cout << face_element->vertexIndex(v) << " ";
    }
    cout << endl;
    for (int d=0;d<dof_indices.getNumber();d++) {
      cout << "\tmg_dof_indices[" << d << "] = " << dof_indices[d]
           << endl;
    }
  } 
#endif
#if (SPACEDIM==3)
  for (bool doit=tess->firstCodimension2Shape();doit;
  doit=tess->nextCodimension2Shape()) {
    const Level1 *level=tess->presentCodimension2Level();
    int ncl=level->numberCoarserLevels();
    const Shape1 *edge_element=level->getShape();
    NumPtr<int> dof_indices;
    getMGDofIndices(edge_element,level,dof_indices);
    cout << "\n\tedge_element = " << edge_element->elementDataIndex() 
         << ", vertices = ";
    for (int v=0;v<edge_element->numberVertices();v++) {
      cout << edge_element->vertexIndex(v) << " ";
    }
    cout << endl;
    for (int d=0;d<dof_indices.getNumber();d++) {
      cout << "\tmg_dof_indices[" << d << "] = " << dof_indices[d]
           << endl;
    }
  } 
#endif
  DofHandler::printDofs();
}

#include "NIDLList.C"
template class NIDLListNode<int>;
template class NIDLList<int>;
template class NIDLListNode<NIDLList<int> >;
template class NIDLList<NIDLList<int> >;

#include "NumPtr.C"
INSTANTIATE_NUMPTR(MGDofHandler::MGVertexDofs);
template class NumPtrBase<NIDLList<int>,int>;
template class NumPtr<NIDLList<int> >;
