#ifndef MGDOF_HANDLER_H
#define MGDOF_HANDLER_H "$Header: /home/faculty/johnt/cvs/deal_new/mg/MGDofHandler.H,v 1.1 2009/08/20 17:34:10 johnt Exp $"
//----------------------------  mg_dof_handler.h  --------------------------
//   $Id: MGDofHandler.H,v 1.1 2009/08/20 17:34:10 johnt Exp $
//   Version: $Name:  $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//----------------------------  mg_dof_handler.h  --------------------------
//
//modified from deal.II/deal.II/include/multigrid/mg_dof_handler.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <DofHandler.H>
class MGTransfer;

//this should replace the DofHandler in user programs
//the only virtual DofHandler functions are 
//  ~DofHandler, distributeDofs and printOn
class MGDofHandler : public DofHandler {
  private:
    class MGVertexDofs {
      private:
        int coarsest_level;
        int finest_level;
        NumPtr<int> indices;
      public:
        MGVertexDofs() : coarsest_level(INT_MAX),finest_level(0) {}
        virtual ~MGVertexDofs() {}
        void init(int,int,int);
        void setIndex(int level,int dof_number,int dofs_per_vertex,
        int index) {
          CHECK_BOUNDS(level,coarsest_level,finest_level+1);
          CHECK_TEST(dof_number<dofs_per_vertex);
          indices[(level-coarsest_level)*dofs_per_vertex+dof_number]=
            index;
        }
        int getIndices(int level,int dpv,NumPtr<int> &dof_numbers) const {
          CHECK_BOUNDS(level,coarsest_level,finest_level+1);
          NumPtr<int> vertex_indices(indices,(level-coarsest_level)*dpv,
            dpv);
          dof_numbers.copyFrom(vertex_indices);
        }
        int getCoarsestLevel() const { return coarsest_level; }
        int getFinestLevel() const { return finest_level; }
        virtual void printOn(ostream&) const;
    };

    NumPtr<DOF_LEVEL_CLASS*> mg_levels;
    NumPtr<MGVertexDofs> mg_vertex_dofs;
    NumPtr<int> mg_used_dofs;

    void setMGVertexDofs();
    void mgVertexDofIndices(const Shape*,int,NumPtr<int>&) const;
    int distributeDofsOnCell(const SHAPE_CLASS*,LEVEL_CLASS*,int);
    void reserveSpace();
    void clearSpace();
  public:
    MGDofHandler(Tessellation &tess,FiniteElement &fe) : 
      DofHandler(tess,fe) {}
    virtual ~MGDofHandler() { 
      clearSpace();
      DofHandler::clearSpace();
    }
    virtual void distributeDofs();
    void getMGDofIndices(const Shape1*,const Level1*,NumPtr<int>&) const;
#if (SPACEDIM>1)
    void getMGDofIndices(const Shape2*,const Level2*,NumPtr<int>&) const;
#endif
#if (SPACEDIM==3)
    void getMGDofIndices(const Shape3*,const Level3*,NumPtr<int>&) const;
#endif
    void makeSparsityPattern(SparsityPatternBase&,int) const;
    void copyMatrix(int,const SolverMatrix&,SolverMatrix&) const;
    void findDirichletBoundaryDofs(const Mapping&,NumPtr<set<int> >&) 
      const;
    void coarsenMatrix(int,const MGTransfer&,const SolverMatrix&,
      SolverMatrix&) const;
    DOF_LEVEL_CLASS* getLevel(int l) const { return mg_levels[l]; }
    int numberDofs(int level) const {
      return mg_used_dofs[level];
    }
    int numberDofs() const { return DofHandler::numberDofs(); }

    virtual void printOn(ostream&) const;
    virtual void printDofs() const;
};

#endif
