#ifndef MGLevelObject_H
#define MGLevelObject_H "$Header: /home/faculty/johnt/cvs/deal_new/mg/MGLevelObject.H,v 1.1 2009/08/20 17:34:10 johnt Exp $"
//--------------------------------------------------------------------
//   $Id: MGLevelObject.H,v 1.1 2009/08/20 17:34:10 johnt Exp $
//   Version: $Name:  $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//--------------------------------------------------------------------
//
//modified from deal.II/deal.II/include/multigrid/mg_level_object.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <MemoryDebugger.H>
#include <NumPtr.H>

template<class Object> class MGLevelObject {
  private:
    int minlevel;
    NumPtr<Object*> objects;
    const Object *cloner;
  public:
    MGLevelObject(const Object &c,int mn=0,int mx=0) : cloner(&c) {
      resize(mn,mx); 
      objects.initialize(0);
    }
    ~MGLevelObject() {
      for (int i=0;i<objects.getNumber();i++) {
        if (objects[i]) delete objects[i]; objects[i]=0;
      }
    }
    Object* operator[](int i) { return objects[i-minlevel]; }
    const Object* operator[](int i) const { return objects[i-minlevel]; }
    void deleteObjects() {
      for (int i=0;i<objects.getNumber();i++) {
        delete objects[i]; objects[i]=0;
      }
    }
    void resize(int new_minlevel,int new_maxlevel) {
      if (objects.getNumber()>0) {
        for (int i=0;i<objects.getNumber();i++) {
          delete objects[i];
        }
        objects.cleanup();
      }
      CHECK_TEST(new_minlevel<=new_maxlevel);
      minlevel=new_minlevel;
      objects.allocate(new_maxlevel-new_minlevel+1);
      objects.initialize(0);
    }
    int getMinlevel() const { return minlevel; }
    int getMaxlevel() const { 
      return minlevel+objects.getNumber()-1; 
    }
    void clone(int level) {
      int ln=level-minlevel;
      CHECK_TEST(ln>=0);
      CHECK_TEST(!objects[ln]);
      objects[ln]=cloner->cloneEmpty();
    }
    void printOn(ostream &os) const {
      os << "MGLevelObject = " << this << endl;
      os << "\tminlevel = " << minlevel << endl;
      os << "cloner = " << cloner << endl;
      for (int i=0;i<objects.getNumber();i++) {
        os << "\tobjects[" << i << "] = " << objects[i] << endl;
        objects[i]->printOn(os);
      }
    }
};

#endif
