// "$Header: /home/faculty/johnt/cvs/deal_new/mg/MGTransfer.C,v 1.1 2009/08/20 17:34:10 johnt Exp $"
//-----------------------------------------------------------------------
//   $Id: MGTransfer.C,v 1.1 2009/08/20 17:34:10 johnt Exp $
//   Version: $Name:  $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//-----------------------------------------------------------------------
//
//modified from deal.II/deal.II/source/multigrid/mg_transfer_prebuilt.cc
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <MGTransfer.H>
#include <FiniteElement.H>
#include <MemoryDebugger.H>
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
MGTransfer::~MGTransfer() {
  for (int i=0;i<prolongation_matrices.getNumber();i++) {
    delete prolongation_matrices[i]; prolongation_matrices[i]=0; 
  }
  for (int i=0;i<prolongation_sparsity_patterns.getNumber();i++) {
    delete prolongation_sparsity_patterns[i];
    prolongation_sparsity_patterns[i]=0; 
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGTransfer::buildMatrices(
const NumPtr<set<int> > &dirichlet_boundary_dofs) {
  Tessellation &tess=mg_dof_handler.getTessellation();
  int n_levels=tess.numberLevels();
  sizes.cleanup();
  sizes.allocate(n_levels);
  for (int l=0;l<n_levels;l++) sizes[l]=mg_dof_handler.numberDofs(l);
  prolongation_sparsity_patterns.cleanup();
  prolongation_sparsity_patterns.allocate(n_levels-1);
  prolongation_matrices.cleanup();
  prolongation_matrices.allocate(n_levels-1);
  for (int i=0;i+1<n_levels;i++) {
    prolongation_matrices[i]=OPERATOR_NEW SparseMatrix;
  }
  slave_mg_dof_indices_on_level.cleanup();
  slave_mg_dof_indices_on_level.allocate(n_levels-1);
  FiniteElement &fe=mg_dof_handler.getFe();
  int max_dofs_per_cell=fe.maxDofsPerCell();
  int l=0;
  for (LEVEL_CLASS *level=tess.coarsestLevel();l+1<n_levels;
  level=level->finerLevel(),l++) {
    SparsityPattern *prolongation_sparsity=
      prolongation_sparsity_patterns[l]=OPERATOR_NEW 
      SparsityPattern(sizes[l+1],sizes[l],max_dofs_per_cell+1);
    LEVEL_CLASS *finer_level=level->finerLevel();
    set<int> all_dof_indices_on_level;
    set<int> master_dof_indices_on_level;
    for (bool doit=level->first();doit;doit=level->next()) {
      const SHAPE_CLASS *element=level->getShape();
      if (level->hasChildren()) {
        NumPtr<int> renumber;
//      renumber[fe_s] = dof_s
        fe.shapeFunctionFamily(element)->
          shapeFunctionsOrder(element,renumber);
        NumPtr<int> renumber_inverse(renumber.getNumber());
        for (int i=0;i<renumber.getNumber();i++) {
//        renumber_inverse[dof_s] = fe_s
          renumber_inverse[abs(renumber[i])]=i;
        }
        NumPtr<int> dof_indices_parent;
        mg_dof_handler.getMGDofIndices(element,level,dof_indices_parent);
        for (int d=0;d<dof_indices_parent.getNumber();d++) {
          all_dof_indices_on_level.insert(dof_indices_parent[d]);
        }
#if (SPACEDIM==1)
        NumPtr<int> line_interior_child_vertices(0);
#endif
#if (SPACEDIM==2)
        NumPtr<int> 
          line_interior_child_vertices(element->numberCodimension1Shapes());
        for (int li=0;li<line_interior_child_vertices.getNumber();
        li++) {
          int line_index=element->codimension1ShapeIndex(li);
          int c=0;
          int v=0;
          line_interior_child_vertices[li]=
            level->Level1::interiorChildVertex(line_index,c,v);
        }
#endif
#if (SPACEDIM==3)
        NumPtr<int> 
          line_interior_child_vertices(element->numberCodimension2Shapes());
        for (int li=0;li<line_interior_child_vertices.getNumber();
        li++) {
          int line_index=element->codimension2ShapeIndex(li);
          int c=0;
          int v=0;
          line_interior_child_vertices[li]=
            level->Level1::interiorChildVertex(line_index,c,v);
        }
#endif
        int dpc=fe.dofsPerCell(element);
        for (int c=0;c<childrenPerShape();c++) {
//        FE_L: prolongation_matrix(q,S)=
//              shape_function_S(parent_point(lattice_point_q))
//          shape_functions ordered same as lattice_points
//          permutation from lattice order to parentCoord order given by
//            Shape::reorderChildEquallySpacedLatticePoints
//          this permutation is used in processConstraints (FiniteElement.C)
//          permutation from lattice order to dof order in
//            shapeFunctionsOrder, which is just call to
//            Shape::reorderEquallySpacedLatticePoints
//            renumber[lattice_order]=dof_order
//        FE_H: prolongation_matrix_{s,S}=
//              [ shape_function_s(lattice_point_q) ]^{-1}
//              [ shape_function_S(parent_point(lattice_point_q)) ]
//          shape_functions in dof order, with possible sign flip
//          permutation from generic order to dof order for specific Shape 
//            in shapeFunctionsOrder; this just flips some signs
//        FiniteElement::prolongationMatrix returns matrix M so that
//          for all parent shapes e
//          for all child_point's in child shape c
//          [ M_{s,S} ]^\top [ shape_function_s(child_point) ] = 
//            [ shape_function_S(parent_point(child_point) ]
//          this is tested in testFE, procedure prolongationCase
          const Matrix &prolongation=
            fe.prolongationMatrix(element,c,level);
          CHECK_TEST(prolongation.size(1)!=0);
          int child_index=level->childIndex(c);
          const SHAPE_CLASS *child_element=
            finer_level->getShape(child_index);
          const Shape *reoriented_child_element=
            fe.shapeFunctionFamily(child_element)->
              reorientChildForParent(element,child_element,c,
                                     line_interior_child_vertices);
          const SHAPE_CLASS *rce=
            dynamic_cast<const SHAPE_CLASS*>(reoriented_child_element);
          NumPtr<int> dof_indices_child;
          mg_dof_handler.getMGDofIndices(rce,finer_level,
            dof_indices_child);
          NumPtr<int> child_renumber;
          fe.shapeFunctionFamily(child_element)->
            shapeFunctionsOrder(child_element,child_renumber);
          NumPtr<int> child_renumber_inverse(child_renumber.getNumber());
          for (int i=0;i<renumber.getNumber();i++) {
            child_renumber_inverse[abs(child_renumber[i])]=i;
          }
          if (reoriented_child_element!=child_element) {
            delete reoriented_child_element;
          }
          reoriented_child_element=0;
          for (int dof_row=0;dof_row<dpc;dof_row++) {
            int row=dof_indices_child[dof_row];
            int prow=child_renumber_inverse[dof_row];
            for (int dof_col=0;dof_col<dpc;dof_col++) {
              int pcol=renumber_inverse[dof_col];
              if (abs(prolongation(prow,pcol))>0.) {
                prolongation_sparsity->add(row,dof_indices_parent[dof_col]);
              }
            }
          }
        }
      } else {
//      dofs on all cells without children
        NumPtr<int> dof_indices;
        mg_dof_handler.getMGDofIndices(element,level,dof_indices);
        for (int d=0;d<dof_indices.getNumber();d++) {
          int dof=dof_indices[d];
          all_dof_indices_on_level.insert(dof);
          master_dof_indices_on_level.insert(dof);
        }
      }
    }
    prolongation_sparsity->compress();

    set<int> &slave_mg_dof_indices=slave_mg_dof_indices_on_level[l]; 
    set_difference(all_dof_indices_on_level.begin(),
      all_dof_indices_on_level.end(),
      master_dof_indices_on_level.begin(),
      master_dof_indices_on_level.end(),
      inserter(slave_mg_dof_indices,slave_mg_dof_indices.begin()));
    SparseMatrix *pml=prolongation_matrices[l];
    pml->reinit(prolongation_sparsity);
    for (bool doit=level->first();doit;doit=level->next()) {
      if (level->hasChildren()) {
        const SHAPE_CLASS *element=level->getShape();
        NumPtr<int> renumber;
        fe.shapeFunctionFamily(element)->
          shapeFunctionsOrder(element,renumber);
        NumPtr<int> renumber_inverse(renumber.getNumber());
        for (int i=0;i<renumber.getNumber();i++) {
          renumber_inverse[abs(renumber[i])]=i;
        }
        int dpc=fe.dofsPerCell(element);
        NumPtr<int> dof_indices_parent;
        mg_dof_handler.getMGDofIndices(element,level,dof_indices_parent);
#if (SPACEDIM==1)
        NumPtr<int> line_interior_child_vertices(0);
#endif
#if (SPACEDIM==2)
        NumPtr<int> 
          line_interior_child_vertices(element->numberCodimension1Shapes());
        for (int li=0;li<line_interior_child_vertices.getNumber();
        li++) {
          int line_index=element->codimension1ShapeIndex(li);
          int c=0;
          int v=0;
          line_interior_child_vertices[li]=
            level->Level1::interiorChildVertex(line_index,c,v);
        }
#endif
#if (SPACEDIM==3)
        NumPtr<int> 
          line_interior_child_vertices(element->numberCodimension2Shapes());
        for (int li=0;li<line_interior_child_vertices.getNumber();
        li++) {
          int line_index=element->codimension2ShapeIndex(li);
          int c=0;
          int v=0;
          line_interior_child_vertices[li]=
            level->Level1::interiorChildVertex(line_index,c,v);
        }
#endif
        for (int c=0;c<childrenPerShape();c++) {
          const Matrix &prolongation=
            fe.prolongationMatrix(element,c,level);
          int child_index=level->childIndex(c);
          const SHAPE_CLASS *child_element=
            finer_level->getShape(child_index);
          const Shape *reoriented_child_element=
            fe.shapeFunctionFamily(child_element)->
              reorientChildForParent(element,child_element,c,
                                     line_interior_child_vertices);
          const SHAPE_CLASS *rce=
            dynamic_cast<const SHAPE_CLASS*>(reoriented_child_element);
          int child_dpc=fe.dofsPerCell(rce);
          NumPtr<int> dof_indices_child;
          mg_dof_handler.getMGDofIndices(rce,finer_level,
            dof_indices_child);
          if (reoriented_child_element!=child_element) {
            delete reoriented_child_element; 
          }
          reoriented_child_element=0;

          NumPtr<int> child_renumber;
          fe.shapeFunctionFamily(child_element)->
            shapeFunctionsOrder(child_element,child_renumber);
          NumPtr<int> child_renumber_inverse(child_renumber.getNumber());
          for (int i=0;i<renumber.getNumber();i++) {
            child_renumber_inverse[abs(child_renumber[i])]=i;
          }
          for (int dof_row=0;dof_row<dpc;dof_row++) {
            int row=dof_indices_child[dof_row];
            int prow=child_renumber_inverse[dof_row];
            for (int dof_col=0;dof_col<dpc;dof_col++) {
              int pcol=renumber_inverse[dof_col];
              if (abs(prolongation(prow,pcol))>0.) {
                pml->set(row,dof_indices_parent[dof_col],
                         prolongation(prow,pcol));
              }
            }
          }
        }
      }
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//src is a global system vector, so it uses DofHandler::getDofIndices
//dst is a multigrid vector, so it uses MGDofHandler::getMGDofIndices
void MGTransfer::copyToMG(MGLevelObject<Vector> &dst,const Vector &src) 
const {
  FiniteElement &fe=mg_dof_handler.getFe();
  int minlevel=dst.getMinlevel();
  int maxlevel=dst.getMaxlevel();
  CHECK_SAME(sizes.getNumber(),
             mg_dof_handler.getTessellation().numberLevels());
  for (int l=minlevel;l<=maxlevel;l++) {
    dst[l]->reinit(sizes[l]);
  }
  const Tessellation &tess=mg_dof_handler.getTessellation();
  for (LEVEL_CLASS *level=tess.coarsestLevel()->finestLevel();level;
  level=level->coarserLevel()) {
    int ln=level->numberCoarserLevels();
    if (ln>maxlevel) continue;
    if (ln<minlevel) break;
    Vector *dstln=dst[ln];
    CHECK_POINTER(dstln);
    for (bool doit=level->first();doit;doit=level->next()) {
      if (!level->hasChildren()) {
        const SHAPE_CLASS *element=level->getShape();
        NumPtr<int> global_dof_indices;
        mg_dof_handler.
          DofHandler::getDofIndices(element,level,global_dof_indices);
        NumPtr<int> level_dof_indices;
        mg_dof_handler.getMGDofIndices(element,level,level_dof_indices);
        for (int i=0;i<level_dof_indices.getNumber();i++) {
          (*dstln)[level_dof_indices[i]]=src[global_dof_indices[i]];
        }
      }
    }
    if (ln<maxlevel) restriction(ln+1,*dstln,*dst[ln+1]);
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGTransfer::copyFromMG(Vector &dst,
const MGLevelObject<Vector> &src) const {
  FiniteElement &fe=mg_dof_handler.getFe();
  Tessellation &tess=mg_dof_handler.getTessellation();
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape()) {
    LEVEL_CLASS *level=tess.presentLevel();
    int ln=level->numberCoarserLevels();
    const SHAPE_CLASS *element=level->getShape();
    int dpc=fe.dofsPerCell(element);
    NumPtr<int> global_dof_indices;
    mg_dof_handler.
      DofHandler::getDofIndices(element,level,global_dof_indices);
    NumPtr<int> level_dof_indices;
    mg_dof_handler.getMGDofIndices(element,level,level_dof_indices);
    for (int i=0;i<dpc;i++) {
      dst[global_dof_indices[i]]=(*src[ln])[level_dof_indices[i]];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGTransfer::copyFromMGAdd(Vector &dst,
const MGLevelObject<Vector> &src) const {
  FiniteElement &fe=mg_dof_handler.getFe();
  Tessellation &tess=mg_dof_handler.getTessellation();
  for (bool doit=tess.firstActiveShape();doit;
  doit=tess.nextActiveShape()) {
    LEVEL_CLASS *level=tess.presentLevel();
    int ln=level->numberCoarserLevels();
    const SHAPE_CLASS *element=level->getShape();
    int dpc=fe.dofsPerCell(element);
    NumPtr<int> global_dof_indices;
    mg_dof_handler.
      DofHandler::getDofIndices(element,level,global_dof_indices);
    NumPtr<int> level_dof_indices;
    mg_dof_handler.getMGDofIndices(element,level,level_dof_indices);
    for (int i=0;i<dpc;i++) {
      dst[global_dof_indices[i]]+=(*src[ln])[level_dof_indices[i]];
    }
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGTransfer::prolongationAndAdd(int fine_level,Vector &dst,
const Vector &src) const {
  CHECK_TEST((fine_level>=1) && 
             (fine_level<=prolongation_matrices.getNumber()));
  Vector slave_src(src.size(),0.);
  const set<int> &slave_mg_dof_indices=
    slave_mg_dof_indices_on_level[fine_level-1];
  for (set<int>::const_iterator smdi_iterator=
  slave_mg_dof_indices.begin();smdi_iterator!=slave_mg_dof_indices.end();
  smdi_iterator++) {
    int i= *smdi_iterator;
    slave_src[i]=src[i];
  }
  prolongation_matrices[fine_level-1]->vmultAdd(dst,slave_src);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
//assumes that the coarse grid nodes are master
//  OK for standard finite elements, but not for Raviart-Thomas
void MGTransfer::restriction(int fine_level,Vector &dst,
const Vector &src) const {
  CHECK_TEST((fine_level>=1) && 
             (fine_level<=prolongation_matrices.getNumber()));
  Vector rv(dst.size());
//the following zeros rv, then does the matrix multiply
  prolongation_matrices[fine_level-1]->transposeVmult(rv,src);
//copy restriction only at slave dofs 
//  (not in coarse element with no children)
  const set<int> &slave_mg_dof_indices=
    slave_mg_dof_indices_on_level[fine_level-1];
  for (set<int>::const_iterator smdi_iterator=
  slave_mg_dof_indices.begin();smdi_iterator!=slave_mg_dof_indices.end();
  smdi_iterator++) {
    int i= *smdi_iterator;
    dst[i]=rv[i];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGTransfer::copySlaveDofs(int level,Vector &dst,const Vector &src) 
const {
  const set<int> &slave_mg_dof_indices=
    slave_mg_dof_indices_on_level[level];
  for (set<int>::const_iterator smdi_iterator=
  slave_mg_dof_indices.begin();smdi_iterator!=
  slave_mg_dof_indices.end();smdi_iterator++) {
    int d= *smdi_iterator;
    dst[d]=src[d];
  }
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void MGTransfer::printOn(ostream &os) const {
  os << "MGTransfer:" << endl;
  os << "\n\tsizes.getNumber = " << sizes.getNumber()
     << "\n\tprolongation_matrices.getNumber = "
     << prolongation_matrices.getNumber() << endl;
}

#include "NumPtr.C"
INSTANTIATE_NUMPTR(SparsityPattern*);
INSTANTIATE_NUMPTR(SparseMatrix*);

template insert_iterator<set<int,less<int>,allocator<int> > > 
  set_difference<
    _Rb_tree_const_iterator<int>, 
    _Rb_tree_const_iterator<int>, 
    insert_iterator<set<int, less<int>, allocator<int> > > >(
      _Rb_tree_const_iterator<int>, 
      _Rb_tree_const_iterator<int>, 
      _Rb_tree_const_iterator<int>, 
      _Rb_tree_const_iterator<int>, 
      insert_iterator<set<int, less<int>, allocator<int> > >
    );
