#ifndef MULTIGRID_H
#define MULTIGRID_H "$Header: /home/faculty/johnt/cvs/deal_new/mg/Multigrid.H,v 1.1 2009/08/20 17:34:11 johnt Exp $"
//----------------------------  multigrid.h  ---------------------------
//   $Id: Multigrid.H,v 1.1 2009/08/20 17:34:11 johnt Exp $
//   Version: $Name:  $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//----------------------------  multigrid.h  ---------------------------
//
//modified from deal.II/deal.II/include/multigrid/multigrid.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include <MGDofHandler.H>
#include <MGLevelObject.H>
#include <MGTransfer.H>
#include <Precondition.H>
#include <Types.H>

class Multigrid {
  private:
    const MGDofHandler *mg_dof_handler;
    int minlevel;
    int maxlevel;
    Vector *vector_cloner;
    MGLevelObject<Vector> initial_residual;
    MGLevelObject<Vector> initial_residual_save;
    MGLevelObject<Vector> residual;
    MGLevelObject<Vector> solution;
    MGLevelObject<Vector> solution_save;
    MGLevelObject<Vector> t;    
    NumPtr<SparsityPattern*> sparsity_patterns;
    MGLevelObject<SolverMatrix> matrices;
    MGLevelObject<SolverPreconditioner> smoothers;
    Solver *coarse_solver;
    SolverPreconditioner *coarse_solver_preconditioner;
    MGTransfer *transfer;

    void levelMGStep(int,int,int);
  public:
    Multigrid(const MGDofHandler&,const SolverMatrix&,Vector&,
      MGTransfer&,const SolverPreconditioner&,Solver&,
      SolverPreconditioner&,const NumPtr<set<int> >&,int mnl=0,
      int mxl=INT_MAX);
    virtual ~Multigrid();
    void vcycle() { levelMGStep(maxlevel,1,1); }
    void wcycle(int si,int cgp) { levelMGStep(maxlevel,si,cgp); }
    void solve(SolverControl&,const SolverMatrix&,const Vector&,
      Vector&,int si=1,int cgp=1);
    void copyCycleCopy(const Vector &src,Vector &dst) {
      transfer->copyToMG(initial_residual,src);
      vcycle();
      transfer->copyFromMG(dst,solution);
    }
#ifdef DEBUG
    void checkProlongationAndRestrictionRandom() const;
    void checkMatrixSymmetry() const;
    void checkMatrixCoarsening(const NumPtr<set<int> >&) const;
    void checkCoarseGridProjection();
    void checkSmootherAdjoints();
    void checkSmootherConvergence();
    void checkVCycleSymmetry();
    void checkVCycleMagnitude();
    void checkMatricesPositiveDefinite(const NumPtr<set<int> >&,
      NumPtr<NumPtr<REAL> >&,NumPtr<REAL>&) const;
    void checkSpectralRadiusOfSmootherIterativeImprovement(
      const NumPtr<NumPtr<REAL> >&) const;
    void checkSmootherErrorReduction(const NumPtr<NumPtr<REAL> >&,
      const NumPtr<REAL>&) const;
    REAL estimateEllipticProjection(const NumPtr<NumPtr<REAL> >&,
      const NumPtr<REAL>&,NumPtr<REAL>&) const;
#endif
    void printOn(ostream&) const;
};

class PreconditionMG : public SolverPreconditioner {
  private:
    Multigrid *multigrid;
  public:
    PreconditionMG(Multigrid &mg) : multigrid(&mg) {}
    virtual ~PreconditionMG() {}
    void vmult(Vector &dst,const Vector &src) const {
      multigrid->copyCycleCopy(src,dst);
    }
    virtual void printOn(ostream&) const;
};

#endif
