#include <float.h>
#include <fstream>
#include <iostream>
#include <math.h>
#include "Arch.H"
#include "Debug.H"
#include "SetTraps.H"
#include "GUIEnumInputParameter.H"
#include "GUIVirtualInput.H"
#include "Palette.H"
#include "TimedObject.H"
#include "Tracer.H"
#include "VGT.H"
#include "VolGraphTool.H"
#include <unistd.h>
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#include "GTKGUI.H"
#include "GTKGUIVirtualInput.H"
#include "GTKWindow.H"
#else
#include "GUI.H"
#include "GUIVirtualInput.H"
#endif

extern "C" {
  void F77NAME(initialize)(const double &dt,
    const int &n0,const int &n1,const int &order,
    double *fmat0,double *fmat1,double *mat0,double *mat1,double *u,
    double *x0,double *x1);
  void F77_NAME(stabilization_scheme)(const double &dt,
    const int &n0,const int &n1,
    double *fmat0,double *fmat1,double *mat0,double *mat1,double *uold,
    double *x0,double *x1,
    double *unew);
  void F77NAME(print_loc)(void *p) {
    cout << p << endl;
    cout << "\tdouble = " << *(double*) p << endl;
    cout << "\tfloat = " << *(float*) p << endl;
    cout << "\tint = " << *(int*) p << endl;
  }
}
struct constant_common {
  double pi;
};
extern constant_common F77NAME(constant);

double heatSolution(const double &x,const double &y,const double &t) {
  return sin(2.*M_PI*x)*sin(4.*y)*exp(-20.*M_PI*M_PI*t);
}

BOOLEAN skip_gui=FALSE;
int n0=10;
int n1=10;
int nsteps=10;
int order=2;

double dt=0.001;
double tmax=0.01;
double winsize=0.5;
char* display_name=0;
Palette *pal=0;
VolGraphTool *g=0;
GUI_INPUT_PARAMETER_LIST_TYPE *param_list=0;

void checkMainInput() {
}

void plot(VolGraphTool *g,int n0,int n1,const Vector3 &low,
const Vector3 &high,const double *x0,const double *x1,const double *u) {
//TRACER_CALL(z,"plot");
#ifdef DEBUG
//cout << "\tn0 = " << n0 << endl;
//cout << "\tn1 = " << n1 << endl;
//cout << "\tlow = " << low << endl;
//cout << "\thigh = " << high << endl;
//for (int ie0=0;ie0<=n0;ie0++) {
//  cout << "\tx0[" << ie0 << "] = " << x0[ie0] << endl;
//}
//for (int ie1=0;ie1<=n1;ie1++) {
//  cout << "\tx1[" << ie1 << "] = " << x1[ie1] << endl;
//}
//for (int ie1=0;ie1<=n1;ie1++) {
//  for (int ie0=0;ie0<=n0;ie0++) {
//    cout << "\tu[" << ie0 << "," << ie1 << "] = " << u[ie0+ie1*(n0+1)]
//         << endl;
//  }
//}
#endif
  double udif=1./(high[2]-low[2]);
  for (int ie1=0;ie1<n1;ie1++) {
    for (int ie0=0;ie0<n0;ie0++) {
      Vector3 u0(x0[ie0  ],x1[ie1  ],u[ie0  + ie1   *(n0+1)]);
      Vector3 u1(x0[ie0+1],x1[ie1  ],u[ie0+1+ ie1   *(n0+1)]);
      Vector3 u2(x0[ie0+1],x1[ie1+1],u[ie0+1+(ie1+1)*(n0+1)]);
      Vector3 u3(x0[ie0  ],x1[ie1+1],u[ie0  +(ie1+1)*(n0+1)]);
      Vector3 uave=(u0+u1+u2+u3)*0.25;

      double frac=((u0[2]+u1[2]+uave[2])/3.-low[2])*udif;
      frac=max(0.,min(1.,frac));
      g->colorTriangle(frac,u0,u1,uave);

      frac=((u1[2]+u2[2]+uave[2])/3.-low[2])*udif;
      frac=max(0.,min(1.,frac));
      g->colorTriangle(frac,u1,u2,uave);

      frac=((u2[2]+u3[2]+uave[2])/3.-low[2])*udif;
      frac=max(0.,min(1.,frac));
      g->colorTriangle(frac,u2,u3,uave);

      frac=((u3[2]+u0[2]+uave[2])/3.-low[2])*udif;
      frac=max(0.,min(1.,frac));
      g->colorTriangle(frac,u3,u0,uave);
    }
  }
}

void runMain(BOOLEAN /*called_before*/) {
//TRACER_CALL(z,"runMain");
  double *u0=OPERATOR_NEW_BRACKET(double,(n0+1)*(n1+1));
  double *u1=OPERATOR_NEW_BRACKET(double,(n0+1)*(n1+1));
  double *x0=OPERATOR_NEW_BRACKET(double,n0+1);
  double *x1=OPERATOR_NEW_BRACKET(double,n1+1);
  double *mat0=OPERATOR_NEW_BRACKET(double,3*(n0-1));
  double *mat1=OPERATOR_NEW_BRACKET(double,3*(n1-1));
  double *fmat0=OPERATOR_NEW_BRACKET(double,3*(n0-1));
  double *fmat1=OPERATOR_NEW_BRACKET(double,3*(n1-1));

  if (tmax>static_cast<double>(nsteps)*dt) {
    tmax=static_cast<double>(nsteps)*dt;
  }
//if (tmax<static_cast<double>(nsteps)*dt) {
//  nsteps=static_cast<int>(tmax/dt);
//}
  dt=min(dt,tmax/static_cast<double>(nsteps));
#ifdef DEBUG
//cout << "\ttmax = " << tmax << endl;
//cout << "\tnsteps = " << nsteps << endl;
//cout << "\tdt = " << dt << endl;
#endif

  {
//  TRACER_CALL(tr,"initialization");
#ifdef INDEF
//  initialize arrays to help IEEE exception handling catch unassigned
    for (int i=0;i<(n0+1)*(n1+1);i++) {
      u0[i]=HUGE_VAL;
      u1[i]=HUGE_VAL;
    }
    for (int i=0;i<=n0;i++) {
      x0[i]=HUGE_VAL;
    }
    for (int i=0;i<=n1;i++) {
      x1[i]=HUGE_VAL;
    }
    for (int i=0;i<3*(n0-1);i++) {
      mat0[i]=HUGE_VAL;
      fmat0[i]=HUGE_VAL;
    }
    for (int i=0;i<3*(n1-1);i++) {
      mat1[i]=HUGE_VAL;
      fmat1[i]=HUGE_VAL;
    }
#endif
    F77NAME(initialize)(dt,n0,n1,order,fmat0,fmat1,mat0,mat1,u0,x0,x1);
  }

//find min,max data values
  Vector3 low(x0[0],x1[0],DBL_MAX);
  Vector3 high(x0[n0],x1[n1],-DBL_MAX);
  for (int i=0;i<(n0+1)*(n1+1);i++) {
    double ui=u0[i];
    if (ui<low[2]) low[2]=ui;
    if (ui>high[2]) high[2]=ui;
  }

  pal=OPERATOR_NEW Palette();
  BOOLEAN use_lighting=FALSE;
  g= OPERATOR_NEW VolGraphTool("Operator Splitting",
    low,high,*pal,use_lighting,winsize);
  {
//  TRACER_CALL(tr0,"initialize graphics");
    { VolGraphTool::DrawDelimiter dd(true,g);
      g->drawBox(low,high);
    }
    g->newPage();
    g->expose();

    for (bool do_best=false;;do_best=!do_best) {
      VolGraphTool::DrawDelimiter dd(do_best,g);
      plot(g,n0,n1,low,high,x0,x1,u0);
      if (do_best) break;
    }
    g->eventLoop(false);
  }

//TimedObject integrate_timing("integrate");
  double t=0.;
  double *uold=u0;
  double *unew=u1;
  for (int step=0;step<nsteps && t<tmax;step++) {
//  cout << "t = " << t << endl;
    F77_NAME(stabilization_scheme)(dt,n0,n1,
      fmat0,fmat1,mat0,mat1,uold,x0,x1,
      unew);
    t+=dt;

//  find min,max data values
//  low[2]=DBL_MAX;
//  high[2]=-DBL_MAX;
//  for (int i=0;i<(n0+1)*(n1+1);i++) {
//    double ui=u0[i];
//    if (ui<low[2]) low[2]=ui;
//    if (ui>high[2]) high[2]=ui;
//  }

    g->newPage();
    for (bool do_best=false;;do_best=!do_best) {
      VolGraphTool::DrawDelimiter dd(do_best,g);
      plot(g,n0,n1,low,high,x0,x1,u0);
      if (do_best) break;
    }
    g->expose();

    double *temp=uold;
    uold=unew;
    unew=temp;
  } // end of time step loop
  g->eventLoop(false);

  delete [] u0; u0=0;
  delete [] u1; u1=0;
  delete [] x0; x0=0;
  delete [] x1; x1=0;
  delete [] mat0; mat0=0;
  delete [] mat1; mat1=0;
  delete [] fmat0; fmat0=0;
  delete [] fmat1; fmat1=0;
}

void cleanup() {
  delete g; g=0;
  delete pal; pal=0;
}

void shutdown() {
  while (param_list->notEmpty()) delete param_list->delAfter(0);
  delete param_list; param_list=0;
}

int main(int argc,char* argv[]) {
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  {
//  set machine-dependent constants for Fortran
    F77NAME(constant).pi=M_PI;

//  define input parameters and bounds
    param_list=OPERATOR_NEW GUI_INPUT_PARAMETER_LIST_TYPE("Param List");
    { const char *group="Problem Parameters";
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(tmax,
        "tmax",0.,DBL_MAX,group) );
    }
    { const char *group="Method Parameters";
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(dt,
        "dt",0.,DBL_MAX,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(order,
        "order",1,3,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(n0,
        "n0",1,INT_MAX,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(n1,
        "n1",1,INT_MAX,group) );
      param_list->append(OPERATOR_NEW GUIInputParameter<int>(nsteps,
        "nsteps",1,INT_MAX,group) );
    }
    { const char *group="Graphics Parameters";
      param_list->append(OPERATOR_NEW GUIInputParameter<double>(winsize,
        "winsize",0.,1.,group) );
    }

//  read input file, if it exists
    if (argc>1) {
      ifstream in_file;
      in_file.open(argv[1],ios::in);
      if (in_file) {
        istream &infile(in_file);
        infile.clear(ios::goodbit);
        infile.seekg(0,ios::beg);
        char name[LENGTH_NAME], comment[LENGTH_COMMENT];
        BOOLEAN found_main=FALSE;
        while ( in_file >> setw(LENGTH_NAME) >> name ) {
          if ( strcmp(name,"Main") == 0 ) {
            in_file.getline( comment, LENGTH_COMMENT);
            found_main=TRUE;
            while ( in_file >> setw(LENGTH_NAME) >> name ) {
              if ( strcmp(name,"end") == 0 ) break;
              else if (strcmp(name,"skip_gui") == 0) {
                int iskip_gui; // type bool not read correctly
                in_file >> iskip_gui;
                skip_gui= iskip_gui!=0;
              }
              else param_list->formattedRead(in_file,name);
              in_file.getline( comment, LENGTH_COMMENT);
            }
          } else in_file.getline(comment,LENGTH_COMMENT);
        }
        if ( !found_main ) skip_gui=FALSE;
      }
      if (skip_gui) checkMainInput();
    } else skip_gui=FALSE;

    if (skip_gui) {
      runMain(FALSE);
      cleanup();
      shutdown();
    } else {
#ifdef USE_GTK
      GTKGUI gui(argv[0],display_name,param_list,&runMain,
        &checkMainInput,&cleanup,&shutdown,TRUE);
#else
      GUI gui(argv[0],display_name,param_list,&runMain,&checkMainInput,
        &cleanup,&shutdown,TRUE);
#endif
      gui.createFileMenu();
      gui.createViewMenu();
      gui.createHelpMenu();
      gui.createMainWindow(argc,argv);
      gui.eventLoop();
    }

  } // scope of MemoryDebugger
  return EXIT_SUCCESS;
}

template class InputParameter<int>;
template class InputParameter<double>;
template class InputParameter<bool>;
