      program main
c+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      integer ncells
      parameter (ncells=100)
c     computational arrays
      double precision
     &  uold(0:ncells),
     &  unew(0:ncells),
     &  x(0:ncells),
     &  matrix(0:ncells,0:1)

c     input:
      integer nsteps
      double precision tmax,decay,diffusion

c     local variables
      integer ic,fi,la,ifirst,ilast
      integer istep
      double precision dt,t,x3,xi
c     double precision pi2
c+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c     input:
      nsteps=10000
      tmax=0.05d0
      decay=0.45d0
      diffusion=1.d0

c     array dimension bounds
      fi=0
      la=ncells
c     interior mesh bounds
      ifirst=0
      ilast=ncells

c     initialize 
c     pi2=1.57079632679489661923d0
      dx=1.d0/dble(ilast-ifirst)
      dt=decay*dx**2/diffusion
      xi=0.d0
      do ic=ifirst,ilast
        x(ic)=xi

c       constant; if we use this, we have to change the Dirichlet bc below
c       uold(ic)=1.d0

c       linear; if we use this, we have to change the Neumann bc below
c       uold(ic)=xi

c       first separable solution:
c       uold(ic)=sin( pi2*xi )

c       square wave:
        x3=3.d0*xi
        if (x3.ge.1.d0 .and. x3.le.2.d0) then
          uold(ic)=1.d0
        else
          uold(ic)=0.d0
        endif
        xi=xi+dx
      enddo

      istep=0
      t=0.d0
      do while (istep.lt.nsteps .and. t.lt.tmax)

c       explicit centered differences:
c       homogeneous Dirichlet boundary condition at left
        ic=ifirst
        unew(ic)=0.d0

c       interior
        do ic=ifirst+1,ilast-1
          unew(ic)=uold(ic)
     &            +decay*((uold(ic+1)-uold(ic)) - (uold(ic)-uold(ic-1)))
        enddo

c       homogeneous Neumann condition at right
        ic=ilast
        unew(ic)=uold(ic) -2.d0*decay*(uold(ic)-uold(ic-1))

        t=t+dt
        istep=istep+1

c       since fortran doesn't have pointers, have to copy arrays
        do ic=fi,la
          uold(ic)=unew(ic)
        enddo
      enddo

c     write results, plot later
c     print *, 'xaxis label "x"'
c     print *, 'yaxis label "u"'
      do ic=ifirst,ilast
        print *,x(ic),uold(ic)
      enddo
     
      stop
      end
