c     given x and order,
c     compute Legendre polynomial f(x), and derivatives f' and f''
      subroutine legendre_polys(order,x,f,dfdx,d2fdx2)
      integer order
      double precision x
      double precision f,dfdx,d2fdx2

      double precision dfm1dx,dfp1dx,d2fm1dx2,d2fp1dx2,factor,factorm1,
     &  fm1,fp1
      integer i
c     print *, "entering legendre_polys"
c     print *, "order = ",order
c     print *, "x = ",x

      if (order.le.0) then
        f=1.d0
        dfdx=0.d0
        d2fdx2=0.d0
        return
      endif
      fm1=1.d0
      f=x
      dfm1dx=0.d0
      dfdx=1.d0
      d2fm1dx2=0.d0
      d2fdx2=0.d0
c     print *, " "
c     print *, "i = 1"
c     print *, "f,dfdx,d2fdx2 = ",f,dfdx,d2fdx2
c     print *, "fm1,dfm1dx,d2fm1dx2 = ",fm1,dfm1dx,d2fm1dx2
      do i=2,order
        factor=dble(2*i-1)/dble(i)
        factorm1=dble(i-1)/dble(i)
        fp1=factor*x*f-factorm1*fm1
        dfp1dx=factor*(x*dfdx+f)-factorm1*dfm1dx
        d2fp1dx2=factor*(x*d2fdx2+2.d0*dfdx)-factorm1*d2fm1dx2
        fm1=f
        f=fp1
        dfm1dx=dfdx
        dfdx=dfp1dx
        d2fm1dx2=d2fdx2
        d2fdx2=d2fp1dx2
c       print *, " "
c       print *, "i = ",i
c       print *, "f,dfdx,d2fdx2 = ",f,dfdx,d2fdx2
c       print *, "fm1,dfm1dx,d2fm1dx2 = ",fm1,dfm1dx,d2fm1dx2
      enddo
c     print *, "leaving legendre_polys"
c     print *, "f,dfdx,d2fdx2 = ",f,dfdx,d2fdx2
      return
      end

c***********************************************************************

c     given maximum number of iterations maxit and Legendre poly order,
c     find x between xlo and xhi
c     so that Legendre polynomial f of given order satisfies f'(x)=0
      subroutine legendre_derivative_zero(maxit,order,xlo,xhi,x)
      include 'const.i'
      integer maxit,order
      double precision xhi,xlo
      double precision x
      double precision a,b,dfadx,dfbdx,dfdx,d2fadx2,d2fbdx2,d2fdx2,f,fa,
     &  fb,xnew
      integer it
c     print *, "entering legendre_derivative_zero"
c     print *, "maxit = ",maxit
c     print *, "order = ",order
c     print *, "xlo,xhi = ",xlo,xhi

      if (order.le.1) then
        x=undefind
        return
      endif
      a=xlo
      b=xhi
      call legendre_polys(order,a,fa,dfadx,d2fadx2)
      call legendre_polys(order,b,fb,dfbdx,d2fbdx2)
c     print *, "a,dfadx,d2fadx2 = ",a,dfadx,d2fadx2
c     print *, "b,dfbdx,d2fbdx2 = ",b,dfbdx,d2fbdx2
      if (dfadx*dfbdx.gt.0.d0) then
        print *, "legendre_derivative_zero not necessarily bracketed"
        call flush(6)
        call abort()
      endif
      x=0.5d0*(a+b)
      do it=1,maxit
        call legendre_polys(order,x,f,dfdx,d2fdx2)
c       print *, " "
c       print *, "a,x,b = ",a,x,b
c       print *, "dfadx,dfdx,dfbdx = ",dfadx,dfdx,dfbdx
c       print *, "x,dfdx,d2fdx2 = ",x,dfdx,d2fdx2
        if (abs(dfdx).le.roundoff*abs(f)) goto 10
        if (b-a.le.roundoff) goto 10
        if (dfdx*d2fdx2.ge.0.d0) then
          if (abs(dfdx).lt.(x-a)*abs(d2fdx2)) then
c           print *, "newton going left"
            xnew=x-dfdx/d2fdx2
          else
c           print *, "Newton going left too big: bisection"
c           if (abs(d2fdx2).ne.0.d0) then
c             print *, "x-a,dfdx/d2fdx2 = ",x-a,dfdx/d2fdx2
c           endif
            xnew=0.5d0*(a+x)
          endif
          b=x
          fb=f
          dfbdx=dfdx
          d2fbdx2=d2fdx2
        else
          xnew=x-dfdx/d2fdx2
          if (xnew.ge.b) then
c           print *, "Newton going right too big: bisection"
c           if (abs(d2fdx2).ne.0.d0) then
c             print *, "b-x,-dfdx/d2fdx2 = ",b-x,-(dfdx/d2fdx2)
c           endif
            xnew=0.5d0*(x+b)
          endif
          a=x
          fa=f
          dfadx=dfdx
          d2fadx2=d2fdx2
        endif
        x=xnew
      enddo
   10 continue
c     print *, "leaving legendre_derivative_zero"
c     print *, "x = ",x
c     call flush(6)
      return
      end

c***********************************************************************

c     given max iterations maxit and number of Lobatto nodes n,
c     find array of Lobatto nodes for all rules of order at most n
c     the nodes for the rule with i nodes with 2 <= i <= n are in
c       nodes(k+j) for 0 <=j <= i
c     where k=((i-2)*(i+1))/2
      subroutine lobatto_nodes(maxit,n,nodes)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      integer maxit,n
      double precision nodes((n*(n+3))/2)
      integer i,j,k
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c     print *, "entering lobatto_nodes"
c     print *, "maxit = ",maxit
c     print *, "n = ",n
      k=0
      do i=0,n-1
        k=k+1
        nodes(k)=-1.d0
        k=k+i+1
        nodes(k)=1.d0
      enddo
      k=1
      do i=1,n-1
        k=k+2 ! 1st node(=-1) of lobatto rule using i zeros of legendre'
c       print *, " "
c       print *, "i = ",i
c       print *, "node indices ",k," to ",k+i+1
        do j=1,i
          k=k+1
          call legendre_derivative_zero(maxit,i+1,
     &      nodes(k-i-2),nodes(k-i-1), nodes(k))
c         if (j.eq.1 .or. j.eq.i) print *, "k = ",k
        enddo
      enddo
c     print *, "leaving lobatto_nodes"
c     k=1
c     do i=0,n-1
c       print *, "nodes = ",(nodes(k+j),j=0,i+1)
c       k=k+i+2
c     enddo
c     call flush(6)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      return
      end

c***********************************************************************

c     given Lobatto rule with order+2 nodes and array of nodes,
c     find array of weights
c     so that Lobatto rule exact for all polynomials of degree<=order+1
c     matrix and ipiv are storage for Lapack routine dgesv
      subroutine lobatto_weights(order,nodes, weights, matrix,ipiv)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      integer order
      double precision nodes(0:order)
      double precision weights(0:order)
      double precision matrix(0:order,0:order)
      integer ipiv(0:order)
      double precision dfdx,d2fdx2,f
      integer i,j,info
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c     print *, "entering lobatto_weights"
c     print *, "order = ",order
c     print *, "nodes = ",(nodes(i),i=0,order)
      do i=0,order
        do j=0,order
          call legendre_polys(i+1,nodes(j),f,dfdx,d2fdx2)
          matrix(i,j)=dfdx
          if (j.eq.0) weights(i)=-f
          if (j.eq.order) weights(i)=weights(i)+f
        enddo
      enddo
      call dgesv(order+1,1,matrix,order+1,ipiv,weights,order+1,info)
c     print *, "leaving lobatto_weights"
c     print *, "weights = ",(weights(i),i=0,order)
c     call flush(6)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      return
      end

c***********************************************************************

c     spectral deferred correction needs to compute integrals of
c       Lagrange interpolation polynomials
c     given Lobatto rule with order+2 nodes, arrays of nodes and weights
c     compute
c       integrals(i,j) = int_(-1)^t_i Lagrange polynomial_j(t) d t
      subroutine lagrange_polynomial_integrals(order,nodes,weights,
     &  integrals)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      integer order
      double precision nodes(0:order),weights(0:order)
      double precision integrals(1:order,0:order)
      integer i,j,k,n
      double precision denom,dt,num,sum,t,tbar
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c     print *, "entering lagrange_polynomial_integrals"
c     print *, "order = ",order
c     print *, "nodes = ",(nodes(i),i=0,order)
      do k=1,order ! integrate lagrange poly from t_0 to t_k
        dt=0.5d0*(nodes(k)-nodes(0))
        tbar=0.5d0*(nodes(k)+nodes(0))
        do j=0,order ! lagrange poly omits node j
          sum=0.d0
          denom=1.d0
          do i=0,order
            if (i.ne.j) denom=denom*(nodes(j)-nodes(i))
          enddo
          do n=0,order ! quadrature point in lobatto rule
            t=nodes(n)*dt+tbar ! mapped quadrature point in (t_0,t_k)
            num=1.d0
            do i=0,order
              if (i.ne.j) num=num*(t-nodes(i))
            enddo
            sum=sum+weights(n)*num/denom
          enddo
          integrals(k,j)=sum*dt
        enddo
      enddo
c     print *, "leaving lagrange_polynomial_integrals"
c     do k=1,order
c       print *, " integrals from",nodes(0)," to ",nodes(k)," = ",
c    &    (integrals(k,j),j=0,order)
c     enddo
c     call flush(6)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      return
      end

c***********************************************************************

      subroutine deferred_correction(order,nodes,integrals,dt,t,y,f,
     &  farray,fnewarray,tarray,yarray,ynewarray)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      integer order
      double precision nodes(0:order),integrals(1:order,0:order),
     &  dt,t,y,f
      external f
      double precision farray(0:order),fnewarray(0:order),
     &  tarray(0:order),yarray(0:order),ynewarray(0:order)
      integer i
      double precision d,eold,enew
      integer j,k
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c     print *, "entering deferred_correction"
c     print *, "order = ",order
c     print *, "t,dt = ",t,dt
c     print *, "y = ",y
c     print *, "nodes = ",(nodes(i),i=0,order)
c     do k=1,order
c       print *, " integrals from",nodes(0)," to ",nodes(k)," = ",
c    &    (integrals(k,j),j=0,order)
c     enddo
c     call flush(6)
      yarray(0)=y
      tarray(0)=t
      farray(0)=f(tarray(0),yarray(0))
      fnewarray(0)=farray(0)
      ynewarray(0)=yarray(0)
      do i=0,order-1
        tarray(i+1)=t+0.5d0*dt*(nodes(i+1)+1.d0)
        yarray(i+1)=yarray(i)+(tarray(i+1)-tarray(i))*farray(i)
        farray(i+1)=f(tarray(i+1),yarray(i+1))
c       print *, "forward euler y(",tarray(i+1),") = ",yarray(i+1)
      enddo
      do j=1,order-1
        eold=0.d0
        d=0.d0
c       print *, "correction ",j
        do i=0,order-1
          enew=0.d0
          do k=0,order
            enew=enew+farray(k)*integrals(i+1,k)
          enddo
          enew=enew*0.5d0*dt+(yarray(0)-yarray(i+1))
          d=d+(tarray(i+1)-tarray(i))*(fnewarray(i)-farray(i))
     &     +(enew-eold)
          eold=enew
          ynewarray(i+1)=yarray(i+1)+d
          fnewarray(i+1)=f(tarray(i+1),ynewarray(i+1))
c         print *, "at t = ",tarray(i+1)," d,e,y,f = ",d,enew,
c    &      ynewarray(i+1),fnewarray(i+1)
        enddo
        do i=1,order
          yarray(i)=ynewarray(i)
          farray(i)=fnewarray(i)
        enddo
c       print *, "at t = ",tarray(order)," y = ",yarray(order)
      enddo
      y=yarray(order)
c     print *, "leaving deferred_correction"
c     print *, "y = ",y
c     call flush(6)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      return
      end
