#include <float.h>
#include <fstream>
#include <iostream>
#include <math.h>
#include "Arch.H"
#include "Debug.H"
#include "SetTraps.H"
#include "InputParameter.H"
#include "Palette.H"
#include "Tracer.H"
#include "VGT.H"
#include "XColormap.H"
#include "XYGraphTool.H"
#include <unistd.h>
#ifdef USE_GTK
#include <gtk/gtkmain.h>
#endif

#define NPLOT 100
#define NDECAY 6
//double decay[]={.7};
//double decay_base=1.;
double decay[]={0.1,0.5,1.,2.,10.,100.};

extern "C" {
  void F77NAME(fourier)(const int&,const double&,const double&,
    const double&,const double&,const double&,
    double*,double*,double*);
}

struct machine_common {
  double roundoff,small,huge,undefind,lnrndoff,lnsmall;
  int ihuge,inhuge;
};
extern machine_common F77NAME(machine);

int main(int argc,char* argv[]) {
//cout << "\tin main" << endl;
#ifdef DEBUG
//setTraps();
#endif
#ifdef MEM_DEBUG
  MemoryDebugger md(1);
#endif
#ifdef USE_GTK
  GTKWindow::gtkInit(argc,argv);
#endif
  {

//  set machine-dependent constants for Fortran
    F77NAME(machine).roundoff=DBL_EPSILON;
    F77NAME(machine).small=DBL_MIN;
    F77NAME(machine).huge=DBL_MAX;
    F77NAME(machine).undefind=HUGE_VAL;

    double theta[NPLOT+1];
    double dissipation[(NPLOT+1)*NDECAY];
    double dispersion[(NPLOT+1)*NDECAY];

    double xlo=0.;
//  double xlo=-M_PI;
    double xhi=M_PI;
    double dtheta=(xhi-xlo)/double(NPLOT);
    theta[0]=xlo;
    for (int i=1;i<=NPLOT;i++) {
      theta[i]=theta[i-1]+dtheta;
    }

    double dissipation_lo=DBL_MAX;
    double dissipation_hi=-DBL_MAX;
    double dispersion_lo=DBL_MAX;
    double dispersion_hi=-DBL_MAX;
//  int j=0;
    for (int j=0;j<NDECAY;j++) {
//    TRACER_CALLV(t,"fourier call, j = ",j)
      int n=NPLOT;
      double the_decay=decay[j];
      F77NAME(fourier)(n,the_decay,dispersion_lo,dispersion_hi,
        dissipation_lo,dissipation_hi,
        theta,dissipation+j*(NPLOT+1),dispersion+j*(NPLOT+1));
    }

    {
//    TRACER_CALL(t,"graphics")
//    setup interactive graphics
      Palette pal;
      XYGraphTool::WINDOW_TYPE::COLOR_MAP_TYPE cmap(&pal);
      double winsize=0.5;
      XYGraphTool gt_dissipation("dissipation","theta",
        "dissipation error",xlo,xhi,dissipation_lo,dissipation_hi,&cmap,
        0,winsize);
      XYGraphTool gt_dispersion("dispersion","theta","phase error",
        xlo,xhi,dispersion_lo,dispersion_hi,&cmap,0,winsize);

//    initialize graphics display
      gt_dissipation.setbgColor("white");
      gt_dissipation.setfgColor("black");
      gt_dissipation.drawAxes();
      gt_dispersion.setbgColor("white");
      gt_dispersion.setfgColor("black");
      gt_dispersion.drawAxes();

//    draw numerical solution
      for (int j=0;j<NDECAY;j++) {
        const char *color_name=pal.getColorName(j+1);
        gt_dissipation.setfgColor(color_name);
        gt_dispersion.setfgColor(color_name);
        cout << color_name << " decay = " << decay[j] <<endl;
        double *dissipation_j=dissipation+j*(NPLOT+1);
        double *dispersion_j=dispersion+j*(NPLOT+1);

        gt_dissipation.movePen(theta[0],*dissipation_j);
        gt_dispersion.movePen(theta[0],*dispersion_j);
        for (int i=1;i<=NPLOT;i++) {
          gt_dissipation.drawLine(theta[i],dissipation_j[i]);
          gt_dispersion.drawLine(theta[i],dispersion_j[i]);
        }
      }
      gt_dissipation.setfgColor("black");
      cout << "black = stability limit" <<endl;
      gt_dissipation.movePen(theta[0],exp(-theta[0]*theta[0])-1.);
      for (int i=1;i<=NPLOT;i++) {
        gt_dissipation.drawLine(theta[i],exp(-theta[i]*theta[i])-1.);
      }

//    force X to perform the requests
      gt_dissipation.flush();
      gt_dispersion.flush();

      XYGraphTool::WINDOW_TYPE::QuitButton qb;
    } // pal,cmap,gt go out of scope here 
  } // scope of MemoryDebugger
  return EXIT_SUCCESS;
}

#ifdef __GNUC__
/*
extern "C" {
  void F77NAME(MAIN)(void) {;}
  void F77_NAME(MAIN)(void) {;}
}
*/

template class InputParameter<int>;
template class InputParameter<double>;
template class InputParameter<bool>;
#endif
