      subroutine initialize(fi,la,ifirst,ilast,
     &  u,x)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      double precision roundoff,small,huge,undefind
      common/machine/roundoff,small,huge,undefind
      double precision pi,time
      common/constant/pi,time
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      integer fi,la,ifirst,ilast
      double precision u(fi:la),x(fi:la)

      integer ic
      double precision dx
      double precision x3,xi
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c     print *, "in initialize, fi,la,ifirst,ilast = ",fi,la,ifirst,ilast
      dx=1.d0/dble(ilast-ifirst)
      xi=0.d0

      do ic=ifirst,ilast
        x(ic)=xi

c       constant; if we use this, we have to change the Dirichlet bc below
c       uold(ic)=1.d0

c       line; if we use this, we have to change the Neumann bc below
c       uold(ic)=xi

c       first separable solution:
c       u(ic)=sin( 0.5d0*pi*x(ic) )

c       square wave:
        x3=3.d0*xi
        if (x3.ge.1.d0 .and. x3.le.2.d0) then
          u(ic)=1.d0
        else
          u(ic)=0.d0
        endif

        xi=xi+dx
      enddo
c     print *, "dx = ",dx
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      return
      end 

c***********************************************************************

      subroutine explicit_centered(fi,la,ifirst,ilast, diffusion,dt,dx,
     &  uold,
     &  unew)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      double precision pi,time
      common/constant/pi,time
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      integer fi,la,ifirst,ilast
      double precision diffusion,dt,dx
      double precision uold(fi:la)
      double precision unew(fi:la)

      integer ic
      double precision decay

c     double precision soln,t,x
c     soln(x,t)=exp(-0.25d0*diffusion*t*pi**2)*sin(0.5d0*pi*x)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c     print *, "in explicit_centered, diffusion,dt,dx = ",
c    &  diffusion,dt,dx
c     call print_loc(dx)
c     do ic=ifirst,ilast
c       print *, "uold[",ic,"] = ",uold(ic)
c     enddo
      decay=diffusion*dt/dx**2
c     print *, "decay = ",decay

      if (decay.le.0.d0 .or. decay.ge.0.5d0) then
        print *, "in explicit_centered, decay<=0 or decay>=1/2",
     &    " is unstable"
        print *, "decay = ",decay
        call abort()
      endif

c     homogeneous Dirichlet condition at left
      ic=ifirst
      unew(ic)=0.d0

c     interior
      do ic=ifirst+1,ilast-1
c       print *, "ic,uold = ",uold(ic-1),uold(ic),uold(ic+1)
c       call flush()
        unew(ic)=uold(ic) 
     &          +decay*((uold(ic+1)-uold(ic)) - (uold(ic)-uold(ic-1)))
c       print*, "error(",ic,") = ",
c    &    soln(dble(ic)*dx,time+dt)-soln(dble(ic)*dx,time)
c    &    - decay*((soln(dble(ic+1)*dx,time)-soln(dble(ic)*dx,time))
c    &            -(soln(dble(ic)*dx,time)-soln(dble(ic-1)*dx,time)))
      enddo

c     homogeneous Neumann condition at right
      ic=ilast
      unew(ic)=uold(ic)-2.d0*decay*(uold(ic)-uold(ic-1))
c       print*, "error(",ic,") = ",
c    &    soln(dble(ic)*dx,time+dt)-soln(dble(ic)*dx,time)
c    &    + 2.d0*decay*(soln(dble(ic)*dx,time)-soln(dble(ic-1)*dx,time))
c       print*, "unew-uold,delta uold = ",
c    &    soln(dble(ic)*dx,time+dt)-soln(dble(ic)*dx,time),
c    &    2.d0*decay*(soln(dble(ic)*dx,time)-soln(dble(ic-1)*dx,time))
c       print *, "x = ",dble(ic)*dx

c     call print_loc(unew)
c     do ic=ifirst,ilast
c       print *, "unew[",ic,"] = ",unew(ic)
c     enddo
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      return
      end

c***********************************************************************

      subroutine implicit_centered(fi,la,ifirst,ilast, diffusion,dt,dx,
     &  uold,
     &  matrix,unew)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      integer fi,la,ifirst,ilast
      double precision diffusion,dt,dx
      double precision
     &  uold(fi:la)
      double precision
     &  matrix(fi:la,0:1),
     &  unew(fi:la)

      integer ic
      double precision decaydx
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      decaydx=diffusion*dt/dx

c     homogeneous Dirichlet condition at left
      ic=ifirst
        unew(ic)=0.d0

c     interior; right-hand side stored in unew
      do ic=ifirst+1,ilast-1
        matrix(ic,0)=dx+2.d0*decaydx
        matrix(ic,1)=-decaydx
        unew(ic)=dx*uold(ic)
      enddo

c     homogeneous Neumann condition at right
      ic=ilast
        matrix(ic,0)=0.5d0*dx+decaydx
        unew(ic)=0.5d0*dx*uold(ic)

c     factor and forward-solve
      do ic=ifirst+1,ilast-1
        matrix(ic,0)=1.d0/sqrt(matrix(ic,0))
        matrix(ic,1)=matrix(ic,1)*matrix(ic,0)
        matrix(ic+1,0)=matrix(ic+1,0)-matrix(ic,1)**2

        unew(ic)=unew(ic)*matrix(ic,0)
        unew(ic+1)=unew(ic+1)-matrix(ic,1)*unew(ic)
      enddo
      ic=ilast
        unew(ic)=unew(ic)/matrix(ic,0)

c     back-solve
      do ic=ilast-1,ifirst+1,-1
        unew(ic)=(unew(ic)-matrix(ic,1)*unew(ic+1))*matrix(ic,0)
      enddo
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      return
      end

c***********************************************************************

      subroutine forward_euler_fem(fi,la,ifirst,ilast, diffusion,dt,dx,
     &  uold,
     &  matrix,unew)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      integer fi,la,ifirst,ilast
      double precision diffusion,dt,dx
      double precision
     &  uold(fi:la)
      double precision
     &  matrix(fi:la,0:1),
     &  unew(fi:la)

      integer ic
      double precision decay,diag,offdiag
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c     do ic=ifirst,ilast
c       print *, "uold[",ic,"] = ",uold(ic)
c     enddo
      decay=diffusion*dt/dx**2

c     homogeneous Dirichlet condition at left
      ic=ifirst
        unew(ic)=0.d0

c     interior
      diag=2.d0/3.d0
      offdiag=1.d0/6.d0
      do ic=ifirst+1,ilast-1
        matrix(ic,0)=diag
        matrix(ic,1)=offdiag
        unew(ic)=diag*uold(ic)+offdiag*(uold(ic-1)+uold(ic+1))
     &    +decay*((uold(ic+1)-uold(ic))-(uold(ic)-uold(ic-1)))
      enddo

c     homogeneous Neumann condition at right
      ic=ilast
        matrix(ic,0)=1.d0/3.d0
        unew(ic)=offdiag*uold(ic-1)+uold(ic)/3.d0
     &    -decay*(uold(ic)-uold(ic-1))
c     do ic=ifirst,ilast
c       print *, "rhs[",ic,"] = ",unew(ic)
c     enddo
c     do ic=ifirst,ilast
c       print *, "matrix[",ic,"] = ",matrix(ic,0),matrix(ic,1)
c     enddo

c     factor and forward-solve
      do ic=ifirst+1,ilast-1
        matrix(ic,0)=1.d0/sqrt(matrix(ic,0))
        matrix(ic,1)=matrix(ic,1)*matrix(ic,0)
        matrix(ic+1,0)=matrix(ic+1,0)-matrix(ic,1)**2

        unew(ic)=unew(ic)*matrix(ic,0)
        unew(ic+1)=unew(ic+1)-matrix(ic,1)*unew(ic)
      enddo
      ic=ilast
        unew(ic)=unew(ic)/matrix(ic,0)

c     back-solve
      do ic=ilast-1,ifirst+1,-1
        unew(ic)=(unew(ic)-matrix(ic,1)*unew(ic+1))*matrix(ic,0)
      enddo
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c     do ic=ifirst,ilast
c       print *, "unew[",ic,"] = ",unew(ic)
c     enddo
      return
      end

c***********************************************************************

      subroutine implicit_centered_periodic(fi,la,ifirst,ilast, 
     &  diffusion,dt,dx,
     &  uold,
     &  matrix,unew)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      integer fi,la,ifirst,ilast
      double precision diffusion,dt,dx
      double precision
     &  uold(fi:la)
      double precision
     &  matrix(fi:la,0:2),
     &  unew(fi:la)

      integer ic
      double precision decay
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
c     do ic=ifirst,ilast
c       print *, "uold[",ic,"] = ",uold(ic)
c     enddo

      decay=diffusion*dt/dx**2

c     interior
c     matrix(*,0) <==> diagonal
c     matrix(*,1) <==> super-diagonal
c     matrix(*,2) <==> last column
      do ic=ifirst,ilast-1
        matrix(ic,0)=1.d0+2.d0*decay
        matrix(ic,1)=-decay
        matrix(ic,2)=0.d0
        unew(ic)=uold(ic)
      enddo
      matrix(ifirst,2)=-decay

c     factor and forward-solve
      do ic=ifirst,ilast-4
        matrix(ic,0)=1.d0/sqrt(matrix(ic,0))
        matrix(ic,1)=matrix(ic,1)*matrix(ic,0)
        matrix(ic,2)=matrix(ic,2)*matrix(ic,0)
        matrix(ic+1,0)=matrix(ic+1,0)-matrix(ic,1)**2
        matrix(ic+1,2)=matrix(ic+1,2)-matrix(ic,1)*matrix(ic,2)
        matrix(ilast-1,0)=matrix(ilast-1,0)-matrix(ic,2)**2

        unew(ic)=unew(ic)*matrix(ic,0)
        unew(ic+1)=unew(ic+1)-matrix(ic,1)*unew(ic)
        unew(ilast-1)=unew(ilast-1)-matrix(ic,2)*unew(ic)
      enddo
      ic=ilast-3
        matrix(ic,0)=1.d0/sqrt(matrix(ic,0))
        matrix(ic,1)=matrix(ic,1)*matrix(ic,0)
        matrix(ic,2)=matrix(ic,2)*matrix(ic,0)
        matrix(ic+1,0)=matrix(ic+1,0)-matrix(ic,1)**2
        matrix(ic+1,1)=matrix(ic+1,1)-matrix(ic,1)*matrix(ic,2)
        matrix(ic+1,2)=matrix(ic+1,1)
        matrix(ilast-1,0)=matrix(ilast-1,0)-matrix(ic,2)**2
        unew(ic)=unew(ic)*matrix(ic,0)
        unew(ic+1)=unew(ic+1)-matrix(ic,1)*unew(ic)
        unew(ilast-1)=unew(ilast-1)-matrix(ic,2)*unew(ic)

      ic=ilast-2
        matrix(ic,0)=1.d0/sqrt(matrix(ic,0))
        matrix(ic,1)=matrix(ic,1)*matrix(ic,0)
        matrix(ic,2)=matrix(ic,1)
        matrix(ic+1,0)=matrix(ic+1,0)-matrix(ic,1)**2
        unew(ic)=unew(ic)*matrix(ic,0)
        unew(ic+1)=unew(ic+1)-matrix(ic,1)*unew(ic)

      ic=ilast-1
        unew(ic)=unew(ic)/matrix(ic,0)
      do ic=ilast-2,ifirst,-1
        unew(ic)=unew(ic)-matrix(ic,2)*unew(ilast-1)
      enddo
      ic=ilast-2
        unew(ic)=unew(ic)*matrix(ic,0)

      do ic=ilast-3,ifirst,-1
        unew(ic)=(unew(ic)-matrix(ic,1)*unew(ic+1))*matrix(ic,0)
      enddo
      unew(ilast)=unew(ifirst)

c     do ic=ifirst,ilast
c       print *, "unew[",ic,"] = ",unew(ic)
c     enddo
c     ic=ifirst
c       print *, "error[",ic,"] = ",(unew(ic)-uold(ic))
c    &    -decay*( (unew(ic+1)-unew(ic)) - (unew(ic)-unew(ilast-1)) )
c     do ic=ifirst+1,ilast-2
c       print *, "error[",ic,"] = ",(unew(ic)-uold(ic))
c    &    -decay*( (unew(ic+1)-unew(ic)) - (unew(ic)-unew(ic-1)) )
c     enddo
c     ic=ilast-1
c       print *, "error[",ic,"] = ",(unew(ic)-uold(ic))
c    &    -decay*( (unew(ifirst)-unew(ic)) - (unew(ic)-unew(ic-1)) )
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      return
      end

c***********************************************************************

      subroutine crank_nicolson(fi,la,ifirst,ilast, diffusion,dt,dx,
     &  uold,
     &  matrix,unew)
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      integer fi,la,ifirst,ilast
      double precision diffusion,dt,dx
      double precision uold(fi:la)
      double precision
     &  matrix(fi:la,0:1),
     &  unew(fi:la)

      integer ic
      double precision decaydx
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      decaydx=diffusion*dt/dx

c     homogeneous Dirichlet condition at left
      ic=ifirst
        unew(ic)=0.d0

c     interior
      do ic=ifirst+1,ilast-1
        matrix(ic,0)=dx+decaydx
        matrix(ic,1)=-(0.5d0*decaydx)
        unew(ic) = dx*uold(ic) 
     &    +0.5d0*decaydx*((uold(ic+1)-uold(ic)) - (uold(ic)-uold(ic-1)))
      enddo

c     homogeneous Neumann condition at right
      ic=ilast
        matrix(ic,0)=0.5d0*(dx+decaydx)
        unew(ic)=0.5d0*(dx*uold(ic)-decaydx*(uold(ic)-uold(ic-1)))

c     factor and forward-solve
      do ic=ifirst+1,ilast-1
        matrix(ic,0)=1.d0/sqrt(matrix(ic,0))
        matrix(ic,1)=matrix(ic,1)*matrix(ic,0)
        matrix(ic+1,0)=matrix(ic+1,0)-matrix(ic,1)**2

        unew(ic)=unew(ic)*matrix(ic,0)
        unew(ic+1)=unew(ic+1)-matrix(ic,1)*unew(ic)
      enddo
      ic=ilast
        unew(ic)=unew(ic)/matrix(ic,0)

c     back-solve
      do ic=ilast-1,ifirst+1,-1
        unew(ic)=(unew(ic)-matrix(ic,1)*unew(ic+1))*matrix(ic,0)
      enddo
c     ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      return
      end
