// "Header:$"
//----------------------------  eigen.h  ---------------------------
//   $Id: eigen.h,v 1.22 2003/04/30 23:01:57 wolf Exp $
//   Version: $Name: Version-4-0-0 $
//
//   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 by the deal.II authors
//
//   This file is subject to QPL and may not be  distributed
//   without copyright and license information. Please refer
//   to the file deal.II/doc/license.html for the  text  and
//   further information on this license.
//
//----------------------------  eigen.h  ---------------------------
//
//modified fromdeal.II/lac/include/lac/eigen.h
//  by John Trangenstein, August 2009
//**********************************************************************
// Copyright 2009 John A. Trangenstein
//
// This software is made available for research and instructional use 
// only. 
// You may copy and use this software without charge for these 
// non-commercial purposes, provided that the copyright notice and 
// associated text is reproduced on all copies.  
// For all other uses (including distribution of modified versions), 
// please contact the author at
//   John A. Trangenstein
//   Department of Mathematics
//   Duke University
//   Durham, NC 27708-0320
//   USA
// or
//   johnt@math.duke.edu
// 
// This software is made available "as is" without any assurance that it
// is completely correct, or that it will work for your purposes.  
// Use the software at your own risk.
//**********************************************************************

#include "Eigen.H"
#include <Precondition.H>
#include <ShiftedMatrix.H>

void EigenPower::solve(REAL &value,const SolverMatrix &A,Vector &x) {
  SOLVER_STATE conv=SOLVER_ITERATE;
  Vector *y=x.clone();
  Vector *r=x.clone();
  REAL length=x.l2Norm();
  REAL old_length=0.;
  x.scale(1./length);
  A.vmult(*y,x);
  for (int iter=0;conv==SOLVER_ITERATE;iter++) {
    y->add(shift,x);
    old_length=length;
    length=y->l2Norm();
    REAL entry=0.;
    REAL thresh=length/static_cast<REAL>(x.size());
    int i=0;
    do {
      CHECK_TEST(i<x.size());
      entry=(*y)[i++];
    } while (abs(entry)<thresh);
    --i;
    value=(entry*x[i]<0.) ? -length : length;
    value-=shift;
    x.equ(1/length,*y);
    A.vmult(*y,x);
    conv=control().check(iter,abs(1./length-1./old_length));
  }
  delete y; y=0;
  delete r; r=0;
  CHECK_TEST(conv==SOLVER_SUCCESS);
}
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
void EigenInverse::solve(REAL &value,const SolverMatrix &A,Vector &x) {
  SOLVER_STATE conv=SOLVER_ITERATE;
  ShiftedMatrix A_s(A,-value);
  ReductionControl inner_control(5000,1.e-5,false,false);
  PreconditionIdentity prec;
  GMRESSolver solver(inner_control);
  int goal=start_adaption;
  Vector *y=x.clone();
  Vector *r=x.clone();
  REAL length=x.l2Norm();
  REAL old_length=0.;
  REAL old_value=value;
  x.scale(1./length);
  for (int iter=0;conv==SOLVER_ITERATE;iter++) {
    solver.solve(A_s,*y,x,prec);
    old_length=length;
    length=y->l2Norm();
    REAL entry=0.;
    int i=0;
    REAL thresh=length/static_cast<REAL>(x.size());
    do {
      CHECK_TEST(i<x.size());
      entry=(*y)[i++];
    } while (abs(entry)<thresh);
    --i;
    value=(entry*x[i]<0.) ? -length : length;
    value=1./value;      
    value-=A_s.getShift();
    if (iter==goal) {
      REAL new_shift=-relaxation*value+(1.-relaxation)*A_s.getShift();
      A_s.setShift(new_shift);
      ++goal;
    }
    x.equ(1./length,*y);
    if (use_residual) {
      y->equ(value,x);
      A.vmult(*r,x);
      r->scale(-1.);
      r->add(value,x);
      REAL res=r->l2Norm();
      conv=control().check(iter,res);
    } else {
      conv=control().check(iter,abs(1./value-1./old_value));
    }
    old_value=value;
  }
  delete y;
  delete r;
  CHECK_TEST(conv==SOLVER_SUCCESS);
}
